/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: eip_linux.c,v 1.96.2.3 2005/09/07 14:34:58 mike Exp $"

#include <qsnet/kernel.h>
#include <qsnet/debug.h>

#include <qsnet/module.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/version.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>

#undef ASSERT
#include <net/sock.h>
#include <net/ip.h>



#include <elan/epcomms.h>
#include <elan/epsvc.h>

#include "eip_linux.h"
#include "eip_stats.h"

#ifdef UNUSED
static void eip_skb_display(struct sk_buff *);
#endif
static void eip_iph_display(struct iphdr *);
#ifdef UNUSED
static void eip_eiph_display(EIP_HEADER *);
static void eip_packet_display(unsigned char *);
#endif
static void eip_tmd_display(EIP_TMD *);
static void eip_tmd_head_display(EIP_TMD_HEAD *);
static void eip_rmd_display(EIP_RMD *);
static void eip_rmd_head_display(EIP_RMD_HEAD *);

static void eip_rmd_reclaim(EIP_RMD *);

static inline EP_NMH *eip_dma_reserve(int, int);
static inline void __eip_tmd_load(EIP_TMD *, EP_RAILMASK *);
static inline void __eip_tmd_unload(EIP_TMD *);
static inline unsigned long eip_buff_alloc(int, int);
static inline void eip_buff_free(unsigned long, int);
static struct iphdr *eip_ipfrag_get(char *);
static inline void eip_rmd_free(EIP_RMD *);
static inline void eip_skb_load(EIP_RMD *);
static inline void eip_skb_unload(EIP_RMD *);
static inline void eip_rmd_requeue(EIP_RMD *);
static EIP_RMD *eip_rmd_alloc(int, int);
static int eip_rmd_alloc_replace(EIP_RMD *, int, int);
static int eip_rmd_alloc_queue(int, int, int, int);
static int eip_rmds_alloc(void);
static void eip_rxhandler(EP_RXD *);
static void eip_rx_tasklet(unsigned long);
static inline void eip_tmd_init(EIP_TMD *, unsigned long, EIP_TMD_HEAD *, unsigned long, int);
static inline EIP_TMD *eip_tmd_get(int);
static inline void eip_tmd_put(EIP_TMD *);
static inline void eip_tmd_load(EIP_TMD *);
static inline void eip_tmd_unload(EIP_TMD *);
static inline EIP_TMD *eip_tmd_alloc_queue(EIP_TMD *, EIP_TMD_HEAD *, int);
static inline EIP_TMD *eip_tmd_alloc_queue_copybreak(EIP_TMD_HEAD *, int);
static inline EIP_TMD *eip_tmd_alloc_queue_aggreg(EIP_TMD_HEAD *, int);
static int eip_tmds_alloc(void);
int eip_hard_start_xmit(struct sk_buff *, struct net_device *);
static inline int eip_do_xmit(EIP_TMD *, EP_NMD *i, EP_PAYLOAD *);
static void eip_txhandler(EP_TXD *, void *, EP_STATUS);
static void eip_tx_tasklet(unsigned long);
void eip_stop_queue(void);
void eip_start_queue(void);
static int eip_open(struct net_device *);
static int eip_close(struct net_device *);
static struct net_device_stats *eip_get_stats(struct net_device *);
static int eip_change_mtu(struct net_device *, int);

static int eip_rx_dropping = 0;
static int eip_rx_tasklet_locked = 1;

/* Global */
struct timer_list eip_rx_tasklet_timer;
	
EIP_RX *eip_rx = NULL;
EIP_TX *eip_tx = NULL;
int  eip_checksum_state=CHECKSUM_NONE;

int tmd_max = EIP_TMD_MAX_NR;
int rmd_max = EIP_RMD_MAX_NR;
int rx_envelope_nr = EIP_RX_ENVELOPE_NR;
int rx_granularity = EIP_RX_GRANULARITY;
int tx_copybreak_max = EIP_TX_COPYBREAK_MAX;
EP_RAILMASK tx_railmask = EP_RAILMASK_ALL;
int eipdebug = 0;

#ifdef UNUSED
static void eip_skb_display(struct sk_buff *skb)
{
	if (skb) {
		__EIP_DBG_PRINTF("SKB [%p] : len %d truesize %d  proto %x pkt type %x cloned %d users %d summed %d\n", 
			skb, skb->len, skb->truesize, skb->protocol, skb->pkt_type, skb->cloned, atomic_read(&skb->users), skb->ip_summed);
		__EIP_DBG_PRINTF("SKB [%p] : skb_shinfo dataref %d nr_frags %d frag_list[%p] (device %p)\n", skb,
			 atomic_read(&skb_shinfo(skb)->dataref), skb_shinfo(skb)->nr_frags, skb_shinfo(skb)->frag_list, skb->dev);
		__EIP_DBG_PRINTF("SKB [%p] : head[%p] data[%p] tail [%p] end [%p] data_len [%d]\n", skb, skb->head, skb->data, 
				skb->tail, skb->end, skb->data_len);
		__EIP_DBG_PRINTF("SKB [%p] : Transport Layer h.(th, uh, icmph, raw)[%p]\n", skb, skb->h.th);
		__EIP_DBG_PRINTF("SKB [%p] : Network Layer      nh.(iph, arph, raw)[%p]\n", skb, skb->nh.iph);
		__EIP_DBG_PRINTF("SKB [%p] : Link Layer         mac.(ethernet, raw)[%p]\n", skb, skb->mac.ethernet);
		return;
	}
	EIP_ERR_PRINTF("SKB IS NULL - NO SKB TO DISPLAY\n");
}
#endif
static void eip_iph_display(struct iphdr *iph)
{
	if (iph) {
		__EIP_DBG_PRINTF("IPH [%p] : version %d header len %d TOS 0x%x Total len %d\n", 
			iph, iph->version, iph->ihl, htons(iph->tos), htons(iph->tot_len));
		__EIP_DBG_PRINTF("IPH [%p] : id %d frag flags 0x%x offset %d\n",
				iph, htons(iph->id), (iph->frag_off & htons(IP_CE | IP_DF | IP_MF)) >> 4, 
				(htons(iph->frag_off) << 3) & IP_OFFSET);
		__EIP_DBG_PRINTF("IPH [%p] : TTL %d proto %d header checksum 0x%x\n", iph, iph->ttl, iph->protocol, iph->check);
		__EIP_DBG_PRINTF("IPH [%p] : IP src %u.%u.%u.%u dest %u.%u.%u.%u\n", iph, 
				 ((unsigned char *)&(iph->saddr))[0],((unsigned char *)&(iph->saddr))[1], ((unsigned char *)&(iph->saddr))[2],((unsigned char *)&(iph->saddr))[3],
				 ((unsigned char *)&(iph->daddr))[0],((unsigned char *)&(iph->daddr))[1], ((unsigned char *)&(iph->daddr))[2],((unsigned char *)&(iph->daddr))[3]);
		return;
	}
	EIP_ERR_PRINTF("IPH IS NULL - NO IPH TO DISPLAY\n");
}
#ifdef UNUSED
static void eip_eiph_display(EIP_HEADER * eiph)
{
	if (eiph) {
		__EIP_DBG_PRINTF("EIPH [%p] : dhost %04x.%04x.%04x sap %x\n", eiph, eiph->h_dhost.ip_bcast, eiph->h_dhost.ip_inst, 
				eiph->h_dhost.ip_addr, eiph->h_sap);
		__EIP_DBG_PRINTF("EIPH [%p] : shost %04x.%04x.%04x \n", eiph, eiph->h_shost.ip_bcast, eiph->h_shost.ip_inst,
				 eiph->h_shost.ip_addr);
		return;
	}
	EIP_ERR_PRINTF("EIPH IS NULL - NO EIPH TO DISPLAY\n");
}
static void eip_packet_display(unsigned char *data)
{
	eip_eiph_display((EIP_HEADER *) data);
	eip_iph_display((struct iphdr *) (data + EIP_HEADER_PAD + ETH_HLEN));
}
#endif
static void eip_tmd_display(EIP_TMD * tmd)
{
	if (tmd) {
		__EIP_DBG_PRINTF("\t\tTMD [%p] : next[%p] skb[%p] DVMA[%d]\n", tmd, tmd->chain.next, tmd->skb, tmd->dvma_idx);
		if (tmd->dma_base)
			__EIP_DBG_PRINTF("TMD [%p] : head[%p] *data 0x%lx\n", tmd, tmd->head, *((unsigned long *) tmd->dma_base));
		else
			__EIP_DBG_PRINTF("TMD [%p] : head[%p] NO DATA !!!\n", tmd, tmd->head);
		__EIP_DBG_PRINTF("TMD [%p] : DMA(%lx,%d,%d) ebase[%x]\n",tmd,  tmd->dma_base, tmd->dma_len, tmd->nmd.nmd_len,
				 tmd->nmd.nmd_addr);
		return;
	}
	EIP_ERR_PRINTF("TMD IS NULL - NO TMD TO DISPLAY\n");
	
}
static void eip_ipf_display(EIP_IPFRAG * ipf)
{
	if (ipf) {
		__EIP_DBG_PRINTF("IPF[%p] : datagram len %d dma correction %d uts %lx frag_nr %d\n", ipf, ipf->datagram_len,
				ipf->dma_correction, ipf->timestamp.tv_usec, ipf->frag_nr);
		eip_tmd_display((EIP_TMD *) ipf);
		return;
	}
	EIP_ERR_PRINTF("IPF IS NULL - NO IPF TO DISPLAY\n");
}

static void eip_tmd_head_display(EIP_TMD_HEAD * head)
{
	if (head) {
		__EIP_DBG_PRINTF("TMD HEAD [%p] : handle[%p] tmds[%p] %3.3d/%3.3d/%3.3d\n", head, head->handle, head->tmd, 
			EIP_STAT_QUEUED_GET(&head->stats), EIP_STAT_ALLOC_GET(&head->stats),
			eip_tx->tmd_max_nr);
		return;
	}
	EIP_ERR_PRINTF("TMD HEAD IS NULL - NO TMD HEAD TO DISPLAY\n");
}
static void eip_rmd_display(EIP_RMD * rmd)
{
	if (rmd) {
		__EIP_DBG_PRINTF("RMD [%p] : next[%p] rxd[%p] DVMA[%d]\n", rmd, rmd->chain.next, rmd->rxd, rmd->dvma_idx);
		__EIP_DBG_PRINTF("RMD [%p] : head[%p]\n", rmd, rmd->head); 
		__EIP_DBG_PRINTF("RMD [%p] : ebase[%x]\n", rmd,  rmd->nmd.nmd_addr); 
		return;
	}
	EIP_ERR_PRINTF("RMD IS NULL - NO RMD TO DISPLAY\n");
}
static void eip_rmd_head_display(EIP_RMD_HEAD * head)
{
	if (head) {
		__EIP_DBG_PRINTF("RMD HEAD [%p] : rcvr[%p] handle[%p] busy list[%p]\n", head, head->rcvr, head->handle, head->busy_list);
		__EIP_DBG_PRINTF("RMD HEAD [%p] : %3.3d/%3.3d/%3.3d\n", head, 
				EIP_STAT_QUEUED_GET(&head->stats), EIP_STAT_ALLOC_GET(&head->stats), eip_rx->rmd_max_nr);
		return;
	}
	EIP_ERR_PRINTF("RMD HEAD IS NULL - NO RMD HEAD TO DISPLAY\n");
}

/* END  - DISPLAY FUNCTIONS */
static inline EP_NMH *eip_dma_reserve(int pages_nr, int perm)
{
	EP_NMH *handle = ep_dvma_reserve(eip_tx->ep_system, pages_nr, perm);
	
	if (handle)
		EIP_DBG_PRINTF(EIP_DBG_EP_DVMA, "HANDLE [%p] %d pages of elan address space reserved\n", 
				handle, pages_nr);
	else
		EIP_ERR_PRINTF("cannot reserve %d page(s) of elan address space\n", pages_nr);

	return handle;
}

static inline void __eip_tmd_load(EIP_TMD * tmd, EP_RAILMASK *rmask)
{
	EIP_ASSERT(tmd->nmd.nmd_len > 0);
	
	ep_dvma_load(eip_tx->ep_system, NULL, (caddr_t) tmd->dma_base, tmd->nmd.nmd_len, tmd->head->handle,
			tmd->dvma_idx, rmask, &tmd->nmd);
}

static inline void __eip_tmd_unload(EIP_TMD * tmd)
{
	EIP_ASSERT(tmd->nmd.nmd_addr && tmd->head->handle);
	
	ep_dvma_unload(eip_tx->ep_system, tmd->head->handle, &tmd->nmd);
	tmd->nmd.nmd_addr = 0;
}
static inline unsigned long eip_buff_alloc(int buff_len, int gfp)
{
	unsigned long buff_base = (buff_len < PAGE_SIZE) ? 
				(unsigned long) kmalloc(buff_len, gfp) :
				__get_dma_pages(gfp, get_order(buff_len));
	
	if (likely(buff_base))
		return buff_base;

	EIP_ERR_PRINTF("cannot allocate %db of memory\n", buff_len);
	return 0;
}
static inline void eip_buff_free(unsigned long buff_base, int buff_len)
{
	(buff_len < PAGE_SIZE) ?  kfree((void *) buff_base) :
		free_pages(buff_base, get_order(buff_len));
}
static struct iphdr *eip_ipfrag_get(char *data)
{
	struct ethhdr *eh = (struct ethhdr *) (data);
	struct iphdr *iph;

	if (eh->h_proto == htons(ETH_P_IP)) {
		iph = (struct iphdr *) ((char *) eh + ETH_HLEN);

		/* EIP_DBG(eip_iph_display(iph)); */

		if ((iph->frag_off & htons(IP_MF | IP_OFFSET)))
			return iph;
	}
	return NULL;
}

static inline void eip_rmd_free(EIP_RMD * rmd)
{
	EIP_ASSERT2(rmd->nmd.nmd_addr == 0, eip_rmd_display, rmd);
	
	if ( rmd->skb != NULL) 
		kfree_skb (rmd->skb);
	
	kfree(rmd);

	EIP_DBG_PRINTF(EIP_DBG_MEMFREE, "RMD [%p] : FREED\n", rmd);
}
static inline void eip_skb_load(EIP_RMD * rmd)
{
	EP_RAILMASK rmask = rmd->rxd ? ep_rxd_railmask (rmd->rxd) : 0;

	EIP_ASSERT(skb_tailroom(rmd->skb) > 0);

	ep_dvma_load(eip_tx->ep_system, NULL, (caddr_t) rmd->skb->data, skb_tailroom(rmd->skb), rmd->head->handle,
		     rmd->dvma_idx, &rmask, &rmd->nmd);
	
	EIP_DBG_PRINTF(EIP_DBG_RMD_EP_DVMA, "RMD [%p] : LOADED\n", rmd);
}
static inline void eip_skb_unload(EIP_RMD * rmd)
{
	EIP_ASSERT(rmd->nmd.nmd_addr && rmd->head->handle);
	
	ep_dvma_unload(eip_tx->ep_system, rmd->head->handle, &rmd->nmd);
	rmd->nmd.nmd_addr = 0;
	
	EIP_DBG_PRINTF(EIP_DBG_RMD_EP_DVMA, "RMD [%p] : UNLOADED\n", rmd);
}
static inline void eip_rmd_requeue(EIP_RMD * rmd)
{
	EIP_ASSERT(rmd->rxd);

	rmd->chain.next    = NULL;

	ep_requeue_receive(rmd->rxd, eip_rxhandler, rmd, &rmd->nmd, EP_NO_ALLOC|EP_NO_SLEEP );

	atomic_inc(&rmd->head->stats);
	
	EIP_DBG_PRINTF(EIP_DBG_RMD_QUEUE, "RMD [%p] : REQUEUED\n", rmd);
}
static EIP_RMD * eip_rmd_alloc(int svc, int gfp)
{
	int buff_len = EIP_SVC_SMALLEST_LEN << svc;
	EIP_RMD *rmd;
	struct sk_buff *skb;

	if (!(skb = alloc_skb((buff_len -  EIP_EXTRA), gfp)))
		return NULL;
	
	skb_reserve(skb, 2);

	if (!(rmd = (EIP_RMD *) kmalloc(buff_len, gfp))) {
		kfree_skb(skb);
		return NULL;
	}

	rmd->skb = skb;

	rmd->chain.next = NULL;
	rmd->rxd = NULL;
	rmd->head = &eip_rx->head[svc];

	return rmd;
}

static int eip_rmd_alloc_replace(EIP_RMD *rmd, int svc, int gfp) 
{
	struct sk_buff *skb,*old;
	int buff_len = EIP_SVC_SMALLEST_LEN << svc;

	if (!(skb = alloc_skb(buff_len, gfp)))
		return 1;
	
	skb_reserve(skb, 2);

	eip_skb_unload(rmd);

	old      = rmd->skb;
	rmd->skb = skb;

	eip_skb_load(rmd);

	eip_rmd_requeue(rmd);

	kfree_skb(old);

	return 0;
}

static int eip_rmd_alloc_queue(int svc, int dvma_idx, int gfp, int attr)
{
	EIP_RMD * rmd = eip_rmd_alloc(svc, gfp);

	if (!rmd)
		return 1;

	EIP_STAT_ALLOC_ADD(&rmd->head->stats, 1);

	rmd->dvma_idx = dvma_idx;
 	eip_skb_load(rmd);

	EIP_DBG2(EIP_DBG_RMD, eip_rmd_display, rmd, "RMD [%p] : ALLOCATED for SVC 0x%x\n", rmd, svc);

	if (ep_queue_receive(rmd->head->rcvr, eip_rxhandler, (void *) rmd, &rmd->nmd, attr) == ESUCCESS) {
		atomic_inc(&rmd->head->stats);
		EIP_DBG_PRINTF(EIP_DBG_RMD_QUEUE, "RMD [%p] : QUEUED on SVC 0x%x\n", rmd, svc);
		return 0;
	}
	
	EIP_ERR_PRINTF("RMD [%p] : couldn't be QUEUED on SVC 0x%x\n", rmd, svc);

	EIP_STAT_ALLOC_SUB(&rmd->head->stats, 1);

	eip_skb_unload(rmd);
	eip_rmd_free(rmd);

	return 1;
}

static int eip_rmds_alloc(void)
{
	int idx, svc;

	eip_rx->irq_list = NULL;
	eip_rx->irq_list_nr = 0;

    	for (svc = 0; svc < EIP_SVC_NR; svc++) {
		eip_rx->head[svc].rcvr = ep_alloc_rcvr(eip_tx->ep_system, EIP_SVC_EP(svc), rx_envelope_nr);
		if (!eip_rx->head[svc].rcvr) {
			EIP_ERR_PRINTF("Cannot install receiver for SVC 0x%x - maybe cable is disconnected\n", svc);
			return -EAGAIN;
		}

		eip_rx->head[svc].handle =
		    eip_dma_reserve(EIP_DVMA_PAGES((EIP_SVC_SMALLEST_LEN << svc)) * eip_rx->rmd_max_nr,
				    EP_PERM_WRITE);
		if (!eip_rx->head[svc].handle)
			return -ENOMEM;
		
		EIP_DBG(EIP_DBG_RMD_HEAD, eip_rmd_head_display, &eip_rx->head[svc]);

		for (idx = 0; idx < EIP_RMD_NR; idx++) {
			if (eip_rmd_alloc_queue(svc, idx * EIP_DVMA_PAGES((EIP_SVC_SMALLEST_LEN << svc)), 
						GFP_KERNEL, EP_NO_SLEEP))
				return -ENOMEM;
		}
	}
	return 0;
}
static void eip_rmds_free(void)
{
	unsigned long flags;
	EIP_RMD *rmd;
	int svc; 
	
	spin_lock_irqsave(&eip_rx->lock, flags);
	rmd = eip_rx->irq_list;
	eip_rx->irq_list = NULL;
	eip_rx->irq_list_nr = 0;
	spin_unlock_irqrestore(&eip_rx->lock, flags);

	eip_rmd_reclaim(rmd);
	
	for (svc = 0; svc < EIP_SVC_NR ; svc++) {
		
		while ((rmd = eip_rx->head[svc].busy_list)) {
			eip_rx->head[svc].busy_list = NULL;
			eip_rmd_reclaim(rmd);
			if (eip_rx->head[svc].busy_list) {
				EIP_DBG_PRINTF(EIP_DBG_RMD_QUEUE, "Still RMD [%p] on BUSY list SVC 0x%d - Scheduling\n", rmd, svc);	
				schedule();
			}
		}

		EIP_ASSERT(EIP_STAT_QUEUED_GET(&eip_rx->head[svc].stats) == EIP_STAT_ALLOC_GET(&eip_rx->head[svc].stats));
		
		EIP_DBG_PRINTF(EIP_DBG_GEN, "HEAD[%p] : FREEING RCVR [%p]\n", &eip_rx->head[svc],
				eip_rx->head[svc].rcvr);
		
		ep_free_rcvr(eip_rx->head[svc].rcvr);

		EIP_DBG_PRINTF(EIP_DBG_EP_DVMA, "HEAD[%p] : RELEASING DVMA [%p]\n", &eip_rx->head[svc], 
				eip_rx->head[svc].handle);

		ep_dvma_release(eip_tx->ep_system, eip_rx->head[svc].handle);
	}

}
static int eip_rx_queues_low (void) {
	int svc;
	for (svc = 0; svc < EIP_SVC_NR; svc++) 
		if (EIP_STAT_QUEUED_GET(&eip_rx->head[svc].stats)  < EIP_RMD_ALLOC_THRESH) 
			return (1);
	return (0);
}
static void eip_rxhandler(EP_RXD * rxd)
{
	EIP_RMD *rmd  		= (EIP_RMD *) ep_rxd_arg(rxd);
	EP_STATUS ret 		= ep_rxd_status(rxd);
	EP_PAYLOAD * payload 	= ep_rxd_payload(rxd);
	unsigned long data	= (unsigned long) rmd->skb->data; 
	int frag_nr 		= 0;
	int len;

	struct sk_buff *skb;
	static char count = 0;

	atomic_dec(&rmd->head->stats);
	rmd->rxd = rxd;

	if (likely(ret == EP_SUCCESS)) {

		rmd->head->dma++;

		if ( eip_rx_dropping) {
		    eip_rmd_requeue(rmd);
		    return;
		}

		len = (payload) ? payload->Data[frag_nr++] : ep_rxd_len(rxd);

		EIP_DBG(EIP_DBG_RMD, eip_rmd_display, rmd);

again:
		if ( (skb = skb_clone(rmd->skb, GFP_ATOMIC)) ) {
			unsigned int off = (data - (unsigned long) rmd->skb->data);

			/* have to set the length before calling
			 * skb pull as it will not allow you to
			 * pull past the end */

			skb_put (skb, off + len);
			skb_pull (skb, off);

			skb->protocol = eth_type_trans(skb, eip_rx->net_device);
			skb->ip_summed = eip_checksum_state;
			skb->dev = eip_rx->net_device;

			/* Fabien/David/Mike this is a hack/fix to allow aggrigation of packets to work.
			 * The problem is ip_frag looks at the truesize to see if it is caching too much space.
			 * As we are reusing a large skb (cloned) for a number of small fragments, they appear to take up alot of space.
			 * so ip_frag dropped them after 4 frags (not good). So we lie and set the truesize to just bigger than the data. 
			 */
			if (payload) 
				skb->truesize = SKB_DATA_ALIGN(skb->len + EIP_HEADER_PAD) +sizeof(struct sk_buff);

		}
		if ( (skb) && 
		     (netif_rx(skb) != NET_RX_DROP)){

			eip_rx->bytes += len;
			
			if (payload && payload->Data[frag_nr] ) {
				data += EIP_IP_ALIGN(len);
				len   = payload->Data[frag_nr++];
				goto again;
			}
			eip_rx->packets += ++frag_nr;
		} else if ( (eip_rx->dropped++ % 20) == 0)
				__EIP_DBG_PRINTK("Packet dropped by the TCP/IP stack - increase /proc/sys/net/core/netdev_max_backlog\n");
	} else if (ret == EP_SHUTDOWN ) {
		EIP_DBG2(EIP_DBG_RMD, eip_rmd_display, rmd, "ABORTING\n");
                ep_complete_receive(rxd);
                eip_skb_unload(rmd);
		EIP_STAT_ALLOC_SUB(&rmd->head->stats, 1);
                eip_rmd_free(rmd);
        	return;
	} else {
		EP_ENVELOPE *env = ep_rxd_envelope(rxd);
		EP_NMD *nmd ;
		
		EIP_ERR_PRINTF("RMD[%p] : RECEIVE ret = %d\n", rmd, ret);

		for (len = 0 ; len < env->nFrags ; len++) {
			nmd = &env->Frags[len];
			EIP_ERR_PRINTF("RMD[%p] : ep_frag #%d nmd_addr [%x] nmd_len %d\n", rmd, len, 
					(unsigned int) nmd->nmd_addr, nmd->nmd_len);
		}
		eip_rx->errors++;
		EIP_ASSERT2(atomic_read(&skb_shinfo(rmd->skb)->dataref) == 1, eip_rmd_display, rmd);
	}

	/* data is used to store the irq flags */
	spin_lock_irqsave(&eip_rx->lock, data);
	rmd->chain.next = eip_rx->irq_list;
	eip_rx->irq_list = rmd;
	eip_rx->irq_list_nr++;
	spin_unlock_irqrestore(&eip_rx->lock, data);

	if (((count++ % eip_rx->sysctl_granularity) == 0) /* and either we have passed up a number of them */
	    || eip_rx_queues_low())                       /* or we are low                                 */
		tasklet_schedule(&eip_rx->tasklet);
	else
	{
		if ( !timer_pending (&eip_rx_tasklet_timer)  )    /* the timer not already set  */
			mod_timer (&eip_rx_tasklet_timer, lbolt);
	}
}

/* dest ; if the buffer still reference on it mocve the rmd to the dest list */
static void eip_rmd_reclaim(EIP_RMD *rmd) 
{
	EIP_RMD *rmd_next = rmd;
	int dataref;

	while (rmd_next) {
		rmd = rmd_next;
		rmd_next = rmd_next->chain.next;

		dataref = atomic_read(&skb_shinfo(rmd->skb)->dataref);
		EIP_ASSERT(dataref > 0);
		
		if (dataref == 1) {
			eip_rmd_requeue(rmd);
		} else {
			rmd->chain.next = rmd->head->busy_list;
			rmd->head->busy_list = rmd;
		}
	}
}
static void eip_rx_tasklet(unsigned long arg)
{
	EIP_RMD *rmd, *rmd_next;
	unsigned long flags;
	short svc, queued;
	int   needs_reschedule;

	if (eip_rx_tasklet_locked) /* we dont want the tasklet to do anything when we are finishing */
	    return;

	for (svc = 0; svc < EIP_SVC_NR; svc++) {
		rmd = eip_rx->head[svc].busy_list;
		eip_rx->head[svc].busy_list = NULL;
		eip_rmd_reclaim(rmd);
	}

	spin_lock_irqsave(&eip_rx->lock, flags);
	rmd = eip_rx->irq_list;
	eip_rx->irq_list = NULL;
	eip_rx->irq_list_nr = 0;
	spin_unlock_irqrestore(&eip_rx->lock, flags);
	
	eip_rmd_reclaim(rmd);

	needs_reschedule = 0;

	for (svc = 0; svc < EIP_SVC_NR; svc++) {
		/* the plan is : allocate some more if possible or steall some dvma space from those on the EIP_BUSY_LIST */
		queued = EIP_STAT_QUEUED_GET(&eip_rx->head[svc].stats);

		EIP_ASSERT(queued >= 0 && queued <= EIP_RMD_MAX_NR);	
		
		if (queued < EIP_RMD_ALLOC_THRESH) {
			short allocated = EIP_STAT_ALLOC_GET(&eip_rx->head[svc].stats);
			short how_many; 

			EIP_ASSERT(allocated >= 0 && allocated <= EIP_RMD_MAX_NR);
			
			if (likely(allocated < eip_rx->rmd_max_nr)) {

				how_many = (((allocated / EIP_RMD_ALLOC_STEP) + 1) * EIP_RMD_ALLOC_STEP);
				if (how_many > eip_rx->rmd_max_nr)
					how_many = eip_rx->rmd_max_nr;

				for (; allocated < how_many &&  
							(eip_rmd_alloc_queue(svc, allocated * EIP_DVMA_PAGES((EIP_SVC_SMALLEST_LEN << svc)), 
									      GFP_ATOMIC, EP_NO_ALLOC|EP_NO_SLEEP) == 0) ; allocated++);
				if ( allocated != how_many ) {
					eip_rx->reschedule++;
					needs_reschedule = 1;
				}
			} else {
				/* steal how_many rmds and put them on the aside list */
                		how_many = EIP_RMD_ALLOC_THRESH - queued;

				EIP_ASSERT(how_many >= 0 && how_many <= EIP_RMD_ALLOC_THRESH);
				
				rmd_next = eip_rx->head[svc].busy_list;
				eip_rx->head[svc].busy_list = NULL;

				while (how_many-- && rmd_next) {
					rmd = rmd_next;
					rmd_next = rmd_next->chain.next;

					if (eip_rmd_alloc_replace(rmd, svc, GFP_ATOMIC)) {
						rmd_next = rmd;
						break;
					}
				}
				eip_rx->head[svc].busy_list = rmd_next;
				if ( how_many )
					needs_reschedule = 1;
            		}
		}
	}
	
	if (needs_reschedule) 
	{
		if ( !timer_pending (&eip_rx_tasklet_timer)) 
			mod_timer (&eip_rx_tasklet_timer, lbolt);
	}
}
static void eip_rx_tasklet_resched(unsigned long arg)
{
	tasklet_schedule(&eip_rx->tasklet);	
}

static inline void eip_tmd_init(EIP_TMD * tmd, unsigned long buff_base, EIP_TMD_HEAD * head, unsigned long buff_len,
				int dvma_idx)
{
	tmd->dvma_idx = dvma_idx;
	tmd->dma_base = buff_base;
	tmd->dma_len = -1;
	tmd->skb = NULL;
	tmd->head = head;
	tmd->chain.next = NULL;

	if (tmd->head != &eip_tx->head[EIP_TMD_STD]) {
		tmd->nmd.nmd_len = buff_len;
		eip_tmd_load(tmd);
	} else  {
		tmd->nmd.nmd_len  = -1;
		tmd->nmd.nmd_addr = 0;
	}
}

static inline EIP_TMD *eip_tmd_get(int id)
{
	unsigned long flags;
	EIP_TMD *tmd = NULL;
	spin_lock_irqsave(&eip_tx->lock, flags);
	while ((tmd = eip_tx->head[id].tmd) == NULL) {
		spin_unlock_irqrestore(&eip_tx->lock, flags);
		if (ep_enable_txcallbacks(eip_tx->xmtr) == 0) {

			spin_lock_irqsave (&eip_tx->lock, flags);
			if (eip_tx->head[id].tmd == NULL) {
				__EIP_DBG_PRINTF("Cannot get a TMD on head %d ... stopping queue\n", id);
				
				eip_stop_queue ();
				
				spin_unlock_irqrestore (&eip_tx->lock, flags);

				return NULL;
			}
			spin_unlock_irqrestore (&eip_tx->lock, flags);
		}

		ep_disable_txcallbacks(eip_tx->xmtr);
		spin_lock_irqsave(&eip_tx->lock, flags);
	}
	eip_tx->head[id].tmd = tmd->chain.next;
	spin_unlock_irqrestore(&eip_tx->lock, flags);
	atomic_dec(&tmd->head->stats);
	return tmd;
}

static inline void eip_tmd_put(EIP_TMD * tmd)
{
	unsigned long flags;

	tmd->skb = NULL;

	spin_lock_irqsave(&eip_tx->lock, flags);
	tmd->chain.next = tmd->head->tmd;
	tmd->head->tmd = tmd;
	spin_unlock_irqrestore(&eip_tx->lock, flags);
	atomic_inc(&tmd->head->stats);

	eip_start_queue();

	EIP_DBG_PRINTF(EIP_DBG_TMD_QUEUE, "TMD [%p] : REQUEUED\n", tmd);
}
static inline void eip_tmd_load(EIP_TMD * tmd)
{
	EP_RAILMASK rmask = tx_railmask;

	__eip_tmd_load(tmd, &rmask);
	
	EIP_DBG_PRINTF(EIP_DBG_EP_DVMA, "TMD [%p] : LOADED\n", tmd);
}
static inline void eip_tmd_unload(EIP_TMD * tmd)
{
	__eip_tmd_unload(tmd);
	
	EIP_DBG_PRINTF(EIP_DBG_EP_DVMA, "TMD [%p] : UNLOADED\n", tmd);
}
static inline void eip_tmd_free(EIP_TMD * tmd)
{
	eip_buff_free(tmd->dma_base, tmd->nmd.nmd_len);
	
	EIP_DBG_PRINTF(EIP_DBG_MEMFREE, "TMD [%p] : FREED\n", tmd);
	
	EIP_STAT_ALLOC_SUB(&tmd->head->stats, 1);
}

/* tmd on a separate block */
static inline EIP_TMD *eip_tmd_alloc_queue(EIP_TMD * tmd, EIP_TMD_HEAD * head, int dvma_idx)
{
	eip_tmd_init(tmd, 0, head, -1, dvma_idx);

	eip_tmd_put(tmd);

	EIP_STAT_ALLOC_ADD(&tmd->head->stats, 1);
	EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
	return tmd;
}
/* tmd on the buffer */
static inline EIP_TMD *eip_tmd_alloc_queue_copybreak(EIP_TMD_HEAD * head, int dvma_idx)
{
	EIP_TMD *tmd;
	unsigned long buff_base;

	if (!(buff_base = eip_buff_alloc(tx_copybreak_max + sizeof(EIP_TMD), GFP_KERNEL)))
		return NULL;

	tmd = (EIP_TMD *) (buff_base + tx_copybreak_max);
	eip_tmd_init(tmd, buff_base, head, tx_copybreak_max, dvma_idx);

	eip_tmd_put(tmd);
	EIP_STAT_ALLOC_ADD(&tmd->head->stats, 1);
	EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
	return tmd;
}

/* ipf are on the buffer */
static inline EIP_TMD *eip_tmd_alloc_queue_aggreg(EIP_TMD_HEAD * head, int dvma_idx)
{
	EIP_TMD *tmd;
	unsigned long buff_base;

	if (!(buff_base = eip_buff_alloc(EIP_SVC_BIGGEST_LEN, GFP_KERNEL)))
		return NULL;

	tmd = (EIP_TMD *) (buff_base + EIP_SVC_BIGGEST_LEN - sizeof(EIP_IPFRAG));
	eip_tmd_init(tmd, buff_base, head, EIP_SVC_BIGGEST_LEN - sizeof(EIP_IPFRAG), dvma_idx);

	eip_tmd_put(tmd);
	EIP_STAT_ALLOC_ADD(&tmd->head->stats, 1);
	EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
	return tmd;
}

static int eip_tmds_alloc()
{
	int i;
	int page_nr;
	EIP_TMD *tmd;

	page_nr = EIP_DVMA_PAGES(tx_copybreak_max);

	eip_tx->head[EIP_TMD_COPYBREAK].handle = eip_dma_reserve(page_nr * eip_tx->tmd_max_nr, EP_PERM_READ);
	
	EIP_DBG(EIP_DBG_TMD_HEAD, eip_tmd_head_display, &eip_tx->head[EIP_TMD_COPYBREAK]);

	for (i = 0; i < EIP_TMD_NR; i++) {
		if (!eip_tmd_alloc_queue_copybreak(&eip_tx->head[EIP_TMD_COPYBREAK], i * page_nr))
			return -ENOMEM;
	}

	eip_tx->head[EIP_TMD_STD].handle =
	    eip_dma_reserve(EIP_DVMA_PAGES(EIP_SVC_BIGGEST_LEN) * eip_tx->tmd_max_nr, EP_PERM_READ);
	
	EIP_DBG(EIP_DBG_TMD_HEAD, eip_tmd_head_display, &eip_tx->head[EIP_TMD_STD]);

	tmd = kmalloc(sizeof(EIP_TMD) * EIP_TMD_NR, GFP_KERNEL);
	if (!tmd) {
		EIP_ERR_PRINTF("Cannot ALLOCATE %d of tmds\n", (int) sizeof(EIP_TMD) * EIP_TMD_NR);
		return -ENOMEM;
	}
	
	page_nr = EIP_DVMA_PAGES(EIP_SVC_BIGGEST_LEN);
	
	for (i = 0; i < EIP_TMD_NR; i++, tmd++) {
		if (!eip_tmd_alloc_queue(tmd, &eip_tx->head[EIP_TMD_STD], i * page_nr))
			return -ENOMEM;
	}

	page_nr = EIP_DVMA_PAGES(EIP_SVC_BIGGEST_LEN);

	eip_tx->head[EIP_TMD_AGGREG].handle = eip_dma_reserve(page_nr * eip_tx->tmd_max_nr, EP_PERM_READ);
	EIP_DBG(EIP_DBG_TMD_HEAD, eip_tmd_head_display, &eip_tx->head[EIP_TMD_AGGREG]);

	for (i = 0; i < EIP_TMD_NR; i++) {
		if (!eip_tmd_alloc_queue_aggreg(&eip_tx->head[EIP_TMD_AGGREG], i * page_nr))
			return -ENOMEM;
	}
	return 0;
}

static void eip_tmds_free(void) 
{
	EIP_TMD *tmd;
	EIP_TMD *tmd_next;
	int i;
	
    	ep_poll_transmits(eip_tx->xmtr);

	for (i = 0 ; i < 3 ; i++) {
again:
		if (EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats) < EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats)) {
			EIP_DBG_PRINTF(EIP_DBG_TMD, "Polling XMTR [%p]\n", eip_tx->xmtr);	
    			ep_poll_transmits(eip_tx->xmtr);
			goto again;
		}
	}
	/* everything should be queued */
        if ((tmd = eip_tx->head[EIP_TMD_COPYBREAK].tmd)) {
            do {
                    	tmd_next = tmd->chain.next;
                        eip_tmd_unload(tmd);
			
            		EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
			
                        eip_tmd_free(tmd);
            } while (tmd_next && (tmd = tmd_next));
        }
	
	EIP_DBG_PRINTF(EIP_DBG_TMD_EP_DVMA, "HEAD[EIP_TMD_COPYBREAK] release DVMA [%p]\n",
			eip_tx->head[EIP_TMD_COPYBREAK].handle);	
	
        ep_dvma_release(eip_tx->ep_system, eip_tx->head[EIP_TMD_COPYBREAK].handle);
	
	/* these ones have been allocated as a block */
    	if ((tmd = eip_tx->head[EIP_TMD_STD].tmd)) {
        	do {
			if (tmd->dvma_idx == 0 ) {
				kfree(tmd);
                            	/* eip_tmd_free(tmd); */
				EIP_STAT_ALLOC_SUB(&tmd->head->stats, EIP_TMD_NR);
				tmd_next = NULL;
				EIP_DBG_PRINTF(EIP_DBG_TMD_EP_DVMA, "TMD HEAD[%p] : [EIP_TMD_STD] BLOCK FREED\n", tmd);	
			} else 
	                  	tmd_next = tmd->chain.next;
        	} while (tmd_next && (tmd = tmd_next));
    	}
	EIP_DBG_PRINTF(EIP_DBG_TMD_EP_DVMA, "HEAD[EIP_TMD_STD] release DVMA [%p]\n", 
			eip_tx->head[EIP_TMD_STD].handle);	
	
        ep_dvma_release(eip_tx->ep_system, eip_tx->head[EIP_TMD_STD].handle);
	
    	if ((tmd = eip_tx->head[EIP_TMD_AGGREG].tmd)) {
        	do {
			tmd_next = tmd->chain.next;

			EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
			
			eip_tmd_unload(tmd);
			eip_tmd_free(tmd);
        	} while (tmd_next && (tmd = tmd_next));
    	}
	EIP_DBG_PRINTF(EIP_DBG_TMD_EP_DVMA, "TMD HEAD[%p] : [EIP_TMD_AGGREG] release DVMA\n", 
			eip_tx->head[EIP_TMD_AGGREG].handle);	
	
        ep_dvma_release(eip_tx->ep_system, eip_tx->head[EIP_TMD_AGGREG].handle);

    	ep_free_xmtr(eip_tx->xmtr);
	EIP_DBG_PRINTF(EIP_DBG_TMD, "XMTR[%p] : FREED\n", eip_tx->xmtr);
}

static inline void eip_ipf_skb_add(EIP_IPFRAG * ipf, struct sk_buff *skb)
{
	int align = EIP_IP_ALIGN(skb->len);
	
	
	if (ipf->dma_len == -1) {	/* like a virgin; touched for the very first time */
		do_gettimeofday(&ipf->timestamp);
		/* FIXE ME put that in release tmd code */
		ipf->frag_nr 		= 0;
		ipf->dma_len 		= 0;
		ipf->datagram_len	= -1;
		ipf->dma_correction 	= 0;
	}
	
	memcpy((void *) (ipf->dma_base + ipf->dma_len), skb->data, skb->len);
	
	if (ipf->datagram_len == -1) {
		struct iphdr * iph = skb->nh.iph;
		int offset = ntohs(iph->frag_off);

		/* last one ? ;  offset & ~IP_OFFSET = IP fragment flags */
		if (((offset & ~IP_OFFSET) & IP_MF) == 0) {
			offset &= IP_OFFSET;
			offset <<= 3;    
			ipf->datagram_len = offset + htons(iph->tot_len) - sizeof(struct iphdr);
		}
	}

	skb->next 			= ipf->skb;
	ipf->skb 			= skb;
	ipf->payload.Data[ipf->frag_nr] = skb->len;
    	ipf->dma_len        	       += align;
    	ipf->dma_correction            += align - skb->len  + ETH_HLEN + sizeof(struct iphdr);
	/* FIXME ; Count got wrong if ip header has options */

    	ipf->frag_nr++;

	EIP_DBG2(EIP_DBG_TMD, eip_ipf_display, ipf, "ADDED skb[%p] len %db ALIGNED(%db)\n", skb, skb->len, EIP_IP_ALIGN(skb->len));
}

#define eip_ipf_hasroom(ipf, skb) ((ipf->dma_len + EIP_IP_ALIGN(skb->len) < eip_tx->sysctl_ipfrag_copybreak))
int eip_hard_start_xmit(struct sk_buff *skb, struct net_device *devnet) 
{

	EIP_TMD *tmd;
	EP_NMD nmd;
	struct iphdr *iph;
	int j;

	if (skb->destructor){
		atomic_inc(&eip_tx->destructor);
		tasklet_schedule(&eip_tx->tasklet);
	} 

	if (!(iph = eip_ipfrag_get(skb->data)) || (eip_tx->sysctl_aggregation == 0)) { /* not ip fragment */
no_aggreg:
		j = (skb->len < eip_tx->sysctl_copybreak) ? EIP_TMD_COPYBREAK : EIP_TMD_STD; /* j = head id */
		
		if (!(tmd = eip_tmd_get(j))) {
			if (skb->destructor)
				atomic_dec(&eip_tx->destructor);
			return 1;
		}
		
		tmd->dma_len 	= skb->len;
		tmd->skb	= skb;
		tmd->skb->next 	= NULL;
		tmd->chain.next = NULL;
		
		if (j == EIP_TMD_COPYBREAK) {
			memcpy((void *) tmd->dma_base, skb->data, skb->len);
			
			ep_nmd_subset(&nmd, &tmd->nmd, 0, skb->len);
#ifdef EIP_MORE_STATS
			eip_tx->sent_copybreak++;
#endif
			return eip_do_xmit(tmd, &nmd, NULL);
		}
		tmd->dma_base 		= (unsigned long) skb->data;
		tmd->nmd.nmd_len	= skb->len;
		eip_tmd_load(tmd);

#ifdef EIP_MORE_STATS
		eip_tx->sent_std++;
#endif
		return eip_do_xmit(tmd, &tmd->nmd, NULL);
	} else if ( skb->len > EIP_SVC_BIGGEST_LEN/2 ) { 
		/* don't aggregate when we have a full mtu of data */
		/* or more than 32k ; in this case it is cheaper   */
		/* to just map the buffer and send it              */
		goto no_aggreg;
	} else {
		EIP_IPFRAG *ipf = NULL;
		unsigned long flags;
		struct list_head *l;
		struct iphdr *iph2;
		int i;
		__u16 id = iph->id;
		__u32 saddr = iph->saddr;
		__u32 daddr = iph->daddr;
		__u8 protocol = iph->protocol;

       		EIP_DBG(EIP_DBG_IPH, eip_iph_display, iph);

		j = 0;

		/* here we can't have full mtu size aggregated packet */
		EIP_ASSERT_RET(skb->len < eip_tx->sysctl_ipfrag_copybreak, 0);

		spin_lock_irqsave(&eip_tx->ipfraglock, flags);
		list_for_each(l, &eip_tx->ipfrag) {
			ipf = list_entry(l, EIP_IPFRAG, list);
			iph2 = eip_ipfrag_get((char *) ipf->dma_base);
			
                        EIP_ASSERT(iph2);
			
			if ((iph2->id == id) && 
					(get_unaligned(&iph2->saddr) == saddr) && 
					(get_unaligned(&iph2->daddr) == daddr) && 
					(iph2->protocol == protocol)) {
                		/* || timeout */
				if (eip_ipf_hasroom(ipf, skb)) {
					
					eip_ipf_skb_add(ipf, skb);
					
					if ((ipf->datagram_len != -1) && 
							(ipf->dma_len == (ipf->datagram_len + ipf->dma_correction) || 
							 ipf->frag_nr == (128 / sizeof(uint32_t)))) {
send_aggreg:
						ipf->payload.Data[ipf->frag_nr] = 0;
						list_del(&ipf->list);
						eip_tx->ipfrag_count--;
						spin_unlock_irqrestore(&eip_tx->ipfraglock, flags);
					
						ep_nmd_subset(&nmd, &ipf->nmd, 0, ipf->dma_len);
						
#ifdef EIP_MORE_STATS
						eip_tx->sent_aggreg++;
#endif
     					 	if ((i = eip_do_xmit((EIP_TMD *) ipf, &nmd, &ipf->payload)) != EP_SUCCESS)
							return i;
						if (j)
							goto new;
						return 0;
					}
					
					spin_unlock_irqrestore(&eip_tx->ipfraglock, flags);
        				tasklet_schedule(&eip_tx->tasklet);
					return 0;
				} else {
					EIP_DBG_PRINTF(EIP_DBG_TMD, "IPF[%p] : FULL %db full - sending it\n", ipf, ipf->dma_len);
					j = 1;
					goto send_aggreg;
				}
			}
		}
		spin_unlock_irqrestore(&eip_tx->ipfraglock, flags);
new:
		if (!(ipf = (EIP_IPFRAG *) eip_tmd_get(EIP_TMD_AGGREG)))
			goto no_aggreg;

		eip_ipf_skb_add(ipf, skb);
		
		spin_lock_irqsave(&eip_tx->ipfraglock, flags);
		list_add_tail(&ipf->list, &eip_tx->ipfrag);
		eip_tx->ipfrag_count++;
		spin_unlock_irqrestore(&eip_tx->ipfraglock, flags);
		tasklet_schedule(&eip_tx->tasklet);
	}
	return 0;
}
static int eip_do_xmit(EIP_TMD * tmd, EP_NMD *nmd, EP_PAYLOAD *payload)
{
	EIP_HEADER *eiph = (EIP_HEADER *) tmd->dma_base;
	int         attr = EP_SET_DATA((EP_NO_SLEEP | EP_NO_INTERRUPT | EP_NO_FAILOVER), EP_TYPE_SVC_INDICATOR, EP_SVC_EIP);
	unsigned long flags;
	int svc, rnum;

	SIZE_TO_SVC(nmd->nmd_len, svc);

	EIP_DBG(EIP_DBG_TMD, eip_tmd_display, tmd);
	/* EIP_DBG(eip_eiph_display(eiph)); */
	
	if (unlikely (eiph->h_dhost.ip_bcast))
		rnum = ep_pickRail (EP_NMD_RAILMASK (nmd) & tx_railmask & ep_xmtr_availrails(eip_tx->xmtr));
	else
		rnum = ep_pickRail (EP_NMD_RAILMASK (nmd) & tx_railmask & ep_xmtr_noderails(eip_tx->xmtr, ntohs(eiph->h_dhost.ip_addr)));

	if (rnum >= 0)
		attr = EP_SET_PREFRAIL(attr, rnum);

	/* add to inuse list  */
	spin_lock_irqsave (&eip_tx->lock, flags);
	list_add_tail (&tmd->chain.link, &eip_tx->inuse);
	spin_unlock_irqrestore (&eip_tx->lock, flags);

	/* ENOMEM EINVAL ECONNREFUSED ESUCCESS */
	svc = (unlikely(eiph->h_dhost.ip_bcast)) ? 
		ep_multicast_message(eip_tx->xmtr, -1, -1, NULL, EIP_SVC_EP(svc), attr | EP_NOT_MYSELF, eip_txhandler, tmd, payload, nmd, 1) :

		ep_transmit_message(eip_tx->xmtr, ntohs(eiph->h_dhost.ip_addr), EIP_SVC_EP(svc),  attr, eip_txhandler, tmd, payload, nmd, 1);
		
	if (likely(svc == EP_SUCCESS))
		return 0;
	else if (svc == ENOMEM) {
		EIP_ERR_PRINTF("%s", "Memory allocation error ...\n");
		eip_tx->errors++;
	}
	else
	{
		/* EP_EINVAL occurs when the svc has a bad value or the iovec has too many frag; */
		/* we don't use the latter option here                                        */
		__EIP_DBG_PRINTF("TMD [%p] : DROPPED skb[%p] status = %d from ep_?_message\n", tmd, tmd->skb, svc);

		eip_tx->dropped++;
	}

	eip_txhandler(NULL, tmd, -99);

	/* Quadrics GNAT sw-elan/4397 - since we will "never" be able to send this packet to the */
	/* destination node, we drop it and feign success - this has the same behaviour as an    */
	/* ethernet where it sticks the packet on the wire, but no-one receives it.              */
	return 0;
}

static void eip_txhandler(EP_TXD * txd, void *arg, EP_STATUS status)
{
	EIP_TMD *tmd = (EIP_TMD *) arg;
	struct sk_buff *skb_next;
	unsigned long flags;
	int svc = 0;
	
	if (likely(status == EP_SUCCESS)) {
		SIZE_TO_SVC(tmd->dma_len, svc);
		eip_tx->dma[svc]++;
		eip_tx->bytes += tmd->dma_len;
		
		if (tmd->head == &eip_tx->head[EIP_TMD_AGGREG]) {
			EIP_IPFRAG *ipf = (EIP_IPFRAG *) tmd;
			eip_tx->packets += ipf->frag_nr;
		} else
			eip_tx->packets++;
	} else {
		if (tmd->head == &eip_tx->head[EIP_TMD_AGGREG]) {
			EIP_IPFRAG *ipf = (EIP_IPFRAG *) tmd;
			eip_tx->dropped += ipf->frag_nr;
			EIP_DBG_PRINTF(EIP_DBG_TMD, "txhandler aggreg packet dropped status = %d\n", status);
		} else  {
			eip_tx->dropped++;
			EIP_DBG_PRINTF(EIP_DBG_TMD, "txhandler packet dropped status = %d\n", status);
		}
	}

	if (tmd->head == &eip_tx->head[EIP_TMD_STD]) {
		eip_tmd_unload(tmd);
		tmd->dma_base = 0;
		tmd->nmd.nmd_len = -1;
	}
		
	tmd->dma_len = -1;
	
	svc = 0;
	while (tmd->skb) {
        	svc++;
		
		if (tmd->skb->destructor)
			atomic_dec(&eip_tx->destructor);

        	skb_next = tmd->skb->next;
        	dev_kfree_skb_any(tmd->skb);
        	tmd->skb = skb_next;
    	}
	EIP_DBG_PRINTF(EIP_DBG_TMD, "IPF/TMD [%p] : %d skb RELEASE/FREED\n", tmd, svc);

	/* remove from inuse list  */
	spin_lock_irqsave (&eip_tx->lock, flags);
	list_del (&tmd->chain.link);
	spin_unlock_irqrestore (&eip_tx->lock, flags);

	eip_tmd_put(tmd);
}

static void eip_tx_tasklet(unsigned long arg)
{
	struct timeval now;
	unsigned long flags;
	EIP_IPFRAG *ipf, *ipfq = NULL;
	EP_NMD nmd;
	struct list_head *list;
	struct list_head *tmp;
	char resched = 0;
	char poll = 1;
	
	do_gettimeofday(&now);
	
	spin_lock_irqsave(&eip_tx->ipfraglock, flags);
	if (eip_tx->ipfrag_count) {
		list_for_each_safe(list, tmp, &eip_tx->ipfrag) {
			ipf = list_entry(list, EIP_IPFRAG, list);
        		/* delta = (((now.tv_sec - ipf->timestamp.tv_sec) * 1000000UL) + now.tv_usec) - ipf->timestamp.tv_usec; */
			if (((((now.tv_sec - ipf->timestamp.tv_sec) * 1000000UL) + now.tv_usec) - 
					ipf->timestamp.tv_usec) >= (1000UL * eip_tx->sysctl_ipfrag_to)) {
				list_del(&ipf->list);
				eip_tx->ipfrag_count--;
				ipf->chain.next = (EIP_TMD *) ipfq;
				ipfq = ipf;
			}
		}
	}
	if (eip_tx->ipfrag_count)
		resched = 1;
	spin_unlock_irqrestore(&eip_tx->ipfraglock, flags);

	while (ipfq) {
		poll = 0;

		ep_nmd_subset(&nmd, &ipfq->nmd, 0, ipfq->dma_len);
		
		ipfq->payload.Data[ipfq->frag_nr] = 0;
		
#ifdef EIP_MORE_STATS
		eip_tx->sent_aggreg++;
#endif
		ipf = (EIP_IPFRAG *) ipfq->chain.next;
		eip_do_xmit((EIP_TMD *) ipfq, &nmd, &ipfq->payload);
		ipfq = ipf;
       }
       
       if (poll)
               ep_poll_transmits(eip_tx->xmtr);

       if (atomic_read(&eip_tx->destructor) || resched )
               tasklet_schedule(&eip_tx->tasklet);
}
void eip_start_queue()
{
	if (netif_queue_stopped(eip_tx->net_device)) {
		EIP_DBG_PRINTK(EIP_DBG_GEN, "Waking up %s queue\n", eip_tx->net_device->name);
		netif_wake_queue(eip_tx->net_device);
	}
}
void eip_stop_queue()
{
	EIP_DBG_PRINTK(EIP_DBG_GEN, "Stopping %s queue\n", eip_tx->net_device->name);
	netif_stop_queue(eip_tx->net_device);
}

static int eip_open(struct net_device *devnet)
{
	if (devnet->flags & IFF_PROMISC)
		EIP_DBG_PRINTK(EIP_DBG_GEN, "%s entering in promiscuous mode\n", devnet->name);

	netif_start_queue(devnet);
	EIP_DBG_PRINTK(EIP_DBG_GEN, "iface %s MAC %02x:%02x:%02x:%02x:%02x:%02x up\n",
			devnet->name, (devnet->dev_addr[0]) & 0xff,
			(devnet->dev_addr[1]) & 0xff, (devnet->dev_addr[2]) & 0xff, (devnet->dev_addr[3]) & 0xff,
		   	(devnet->dev_addr[4]) & 0xff, (devnet->dev_addr[5]) & 0xff);
	return 0;
}

static int eip_close(struct net_device *devnet)
{
	if (devnet->flags & IFF_PROMISC)
		EIP_DBG_PRINTK(EIP_DBG_GEN, "%s leaving promiscuous mode\n", devnet->name);

	netif_stop_queue(devnet);

	eip_rx_tasklet(0);

	EIP_DBG_PRINTK(EIP_DBG_GEN, "iface %s MAC %02x:%02x:%02x:%02x:%02x:%02x down\n", 
		devnet->name, (devnet->dev_addr[0]) & 0xff,
		(devnet->dev_addr[1]) & 0xff, (devnet->dev_addr[2]) & 0xff, (devnet->dev_addr[3]) & 0xff,
		(devnet->dev_addr[4]) & 0xff, (devnet->dev_addr[5]) & 0xff);
	return 0;
}

static struct net_device_stats *eip_get_stats(struct net_device *devnet)
{
	static struct net_device_stats stats;

	stats.rx_packets = eip_rx->packets;
	stats.rx_bytes = eip_rx->bytes;
	stats.rx_errors = eip_rx->errors;
	stats.rx_dropped = eip_rx->dropped;

	stats.tx_packets = eip_tx->packets;
	stats.tx_bytes = eip_tx->bytes;
	stats.tx_errors = eip_tx->errors;
	stats.tx_dropped = eip_tx->dropped;
	return &stats;
}

static int eip_change_mtu(struct net_device *devnet, int mtu)
{
	if (mtu <= EIP_MTU_MAX) {
		EIP_DBG_PRINTK(EIP_DBG_GEN, "MTU size changed from %d to %d\n", devnet->mtu, mtu);
		devnet->mtu = mtu;
	}
	return 0;
}

#ifdef MODULE
int eip_init(void)
{
	struct net_device *devnet;
	int errno = 0;

	eip_rx_dropping = 0; 
	eip_rx_tasklet_locked = 1;

	/* timer up but not started */
	init_timer (&eip_rx_tasklet_timer);
	eip_rx_tasklet_timer.function = eip_rx_tasklet_resched;
	eip_rx_tasklet_timer.data     = (unsigned long) 0;
	eip_rx_tasklet_timer.expires  = lbolt + hz;

	devnet = alloc_etherdev(sizeof(EIP_RX) + sizeof(EIP_TX));
	if (!devnet) {
		EIP_ERR_PRINTF("Unable to ALLOCATE etherdev structure\n");
		return -ENOMEM;
	}
	strcpy (devnet->name, "eip0");

	EIP_DBG_PRINTK(EIP_DBG_GEN, "Enabling aggregation code\n");
	devnet->change_mtu = eip_change_mtu;
	devnet->mtu = EIP_MTU_MAX;
	devnet->open = eip_open;
	devnet->stop = eip_close;
	devnet->hard_start_xmit = eip_hard_start_xmit;
	devnet->get_stats = eip_get_stats;

        /* devnet->features |= (NETIF_F_DYNALLOC); */
        /* devnet->features = (NETIF_F_SG | NETIF_F_FRAGLIST | NETIF_F_HIGHDMA); */
        /* devnet->features |= (NETIF_F_SG|NETIF_F_FRAGLIST|NETIF_F_HIGHDMA|NETIF_F_HW_CSUM); */

	eip_rx = (EIP_RX *) devnet->priv;
	eip_tx = (EIP_TX *) (eip_rx + 1);

	/* instance 0 */
	eip_tx->ep_system = ep_system();
	if (eip_tx->ep_system == NULL) {
		EIP_ERR_PRINTF("kernel comms for iface %s does not exist\n", devnet->name);
		errno = -ENXIO;
		goto out;
	}
	if (ep_waitfor_nodeid(eip_tx->ep_system) == ELAN_INVALID_NODE) {
		EIP_ERR_PRINTF("network position not found\n");
		errno = -EAGAIN;
		goto out;
	}
	eip_tx->xmtr = ep_alloc_xmtr(eip_tx->ep_system);
	if (!eip_tx->xmtr) {
		EIP_ERR_PRINTF("Cannot create allocated transmitter - maybe cable is disconnected\n");
		errno = -EAGAIN;
		goto out;
	}
	/* assign MAC address */
	*((int *) &devnet->dev_addr[4]) = htons(ep_nodeid(eip_tx->ep_system));
	eip_rx->net_device = devnet;
	eip_tx->net_device = devnet;

	atomic_set(&eip_tx->destructor, 0);

	if ((tmd_max >= EIP_TMD_MIN_NR) && (tmd_max <= EIP_TMD_MAX_NR)) {
		EIP_DBG_PRINTF(EIP_DBG_GEN, "Setting tmd_max_nr to %d\n", tmd_max);
		eip_tx->tmd_max_nr = tmd_max;
	} else {
		EIP_ERR_PRINTF("parameter error : %d <= tmd_max(%d) <= %d using default %d\n", 
				EIP_TMD_MIN_NR, tmd_max, EIP_TMD_MAX_NR, EIP_TMD_MAX_NR);
		eip_tx->tmd_max_nr = EIP_TMD_MAX_NR;
	}

	if ((rmd_max >= EIP_RMD_MIN_NR) && (rmd_max <= EIP_RMD_MAX_NR)) {
		EIP_DBG_PRINTF(EIP_DBG_GEN, "Setting rmd_max_nr to %d\n", rmd_max);
		eip_rx->rmd_max_nr = rmd_max;
	} else {
		EIP_ERR_PRINTF("parameter error : %d <= rmd_max(%d) <= %d using default %d\n", EIP_RMD_MIN_NR,
			   rmd_max, EIP_RMD_MAX_NR, EIP_RMD_MAX_NR);
		eip_rx->rmd_max_nr = EIP_RMD_MAX_NR;
	}

	if ((rx_envelope_nr > 0) && (rx_envelope_nr <= 1024)) {	/* > 1024 don't be silly */
		EIP_DBG_PRINTK(EIP_DBG_GEN, "Setting rx_envelope_nr to %d\n", rx_envelope_nr);
	} else {
		EIP_ERR_PRINTF("parameter error : 0 < rx_envelope_nr(%d) <= 1024 using default %d\n",
			   rx_envelope_nr, EIP_RX_ENVELOPE_NR);
		rx_envelope_nr = EIP_RX_ENVELOPE_NR;
	}

	if (tx_copybreak_max <= EIP_TX_COPYBREAK_MAX) {
		EIP_DBG_PRINTF(EIP_DBG_GEN, "Setting tx_copybreak_max to %d\n", tx_copybreak_max);
	} else {
		EIP_ERR_PRINTF("parameter error : tx_copybreak_max > %d using default %d\n",
			   EIP_TX_COPYBREAK_MAX, EIP_TX_COPYBREAK_MAX);
		tx_copybreak_max = EIP_TX_COPYBREAK_MAX;
	}
#ifdef EIP_MORE_STATS
	eip_tx->sent_copybreak = 0;
	eip_tx->sent_std = 0;
	eip_tx->sent_aggreg = 0;
#endif

	eip_tx->ipfrag_count = 0;
	eip_aggregation_set(1);
	eip_rx_granularity_set(rx_granularity);
	eip_tx_copybreak_set(EIP_TX_COPYBREAK);
	eip_ipfrag_to_set(EIP_IPFRAG_TO);
	eip_ipfrag_copybreak_set(EIP_IPFRAG_COPYBREAK);

	spin_lock_init(&eip_tx->lock);
	spin_lock_init(&eip_tx->ipfraglock);
	spin_lock_init(&eip_rx->lock);
	tasklet_init(&eip_rx->tasklet, eip_rx_tasklet, 0);
	tasklet_init(&eip_tx->tasklet, eip_tx_tasklet, 0);
	INIT_LIST_HEAD(&eip_tx->ipfrag);
	INIT_LIST_HEAD(&eip_tx->inuse);

	/* if we fail here cannot do much yet; waiting for rcvr remove code in ep. */
	errno = eip_tmds_alloc();
	if (errno)
		goto out;

	errno = eip_rmds_alloc();
	if (errno)
		goto out;

	errno = eip_stats_init();
	if (errno)
		goto out;

	if (ep_svc_indicator_set(eip_tx->ep_system, EP_SVC_EIP) != EP_SUCCESS) {
		EIP_ERR_PRINTF("Cannot set the service indicator\n");
		errno = -EINVAL;
		goto out;
	}

	eip_rx_tasklet_locked = 0;
	tasklet_schedule(&eip_rx->tasklet);

	SET_MODULE_OWNER(eip_tx->net_device);

	if (register_netdev(devnet)) {
		printk("eip: failed to register netdev\n");
		goto out;
	}

	EIP_DBG_PRINTK(EIP_DBG_GEN, "iface %s MAC %02x:%02x:%02x:%02x:%02x:%02x ready\n", 
		devnet->name, (devnet->dev_addr[0]) & 0xff,
		(devnet->dev_addr[1]) & 0xff, (devnet->dev_addr[2]) & 0xff, (devnet->dev_addr[3]) & 0xff,
		(devnet->dev_addr[4]) & 0xff, (devnet->dev_addr[5]) & 0xff);

	return 0;
      out:
	unregister_netdev(devnet);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 25)
	kfree(devnet);
#else
	free_netdev(devnet);
#endif

	return errno;
}
void eip_exit(void)
{
	int i;

	eip_rx_dropping = 1;                /* means that new messages wont be sent to tcp stack */
	eip_rx_tasklet_locked = 1;

	netif_stop_queue(eip_tx->net_device);

    	if (ep_svc_indicator_clear(eip_tx->ep_system, EP_SVC_EIP) != EP_SUCCESS) {
		EIP_ERR_PRINTF("Cannot unset the service indicator\n");
	}

	schedule_timeout(10);
	
	del_timer_sync (&eip_rx_tasklet_timer);

	tasklet_disable(&eip_rx->tasklet);
	tasklet_disable(&eip_tx->tasklet);

	tasklet_kill(&eip_tx->tasklet);
	tasklet_kill(&eip_rx->tasklet);

        eip_rmds_free();
        eip_tmds_free();

	/* that things freed */
	for (i = 0 ; i < EIP_SVC_NR ; i++) {
		if ( EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats) != 0 )
			EIP_ERR_PRINTF("%d RMDs not FREED on SVC[%d]\n", EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats), i);
	}
	for (i = 0 ; i < 3 ; i++) {
		if ( EIP_STAT_ALLOC_GET(&eip_tx->head[i].stats) != 0 )
			EIP_ERR_PRINTF("%d TMDs not freed on TX HEAD[%d]\n", EIP_STAT_ALLOC_GET(&eip_tx->head[i].stats), i);
		
	}
	unregister_netdev(eip_tx->net_device);
    	kfree(eip_tx->net_device);
	
    	eip_stats_cleanup();
}

module_init(eip_init);
module_exit(eip_exit);

module_param(eipdebug, uint, 0);
MODULE_PARM_DESC(eipdebug, "Set debug flags");

module_param(rx_envelope_nr, uint, 0);
MODULE_PARM_DESC(rx_enveloppe_nr, "Number of allocated enveloppe on the rx side");

module_param(tx_copybreak_max, uint, 0);
MODULE_PARM_DESC(tx_copybreak_max, "Maximum size of the tx copybreak limit (default 512)");

module_param(tmd_max, uint, 0);
module_param(rmd_max, uint, 0);
MODULE_PARM_DESC(tmd_max, "Maximun number of transmit buffers (default 64)");
MODULE_PARM_DESC(rmd_max, "Maximun number of receive buffers (default 64)");

module_param(tx_railmask, ushort, 0);
MODULE_PARM_DESC(tx_railmask, "Mask of which rails transmits can be queued on");

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("Elan IP driver");
MODULE_LICENSE("GPL");
#endif	/* MODULE */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
