/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: eip_linux.h,v 1.47 2004/10/01 10:49:29 mike Exp $"

#ifndef __EIP_LINUX_H
#define __EIP_LINUX_H

#define EIP_WATERMARK			(0xfab1e)

#define EIP_PAGES(s)			(((s - 1) >> PAGE_SHIFT) + 1)
#define EIP_DVMA_PAGES(s)		((s < PAGE_SIZE) ? EIP_PAGES(s) + 1 : EIP_PAGES(s))

#define EIP_SVC_SMALLEST_LEN		(1 << 9)	/* 512 */
#define EIP_SVC_BIGGEST_LEN		(1 << 16)	/* 64k */

#define EIP_SVC_SMALLEST		(0)
#define EIP_SVC_BIGGEST			(7)

#define EIP_SVC_NR			(8)
#define EIP_SVC_EP(s)			(s + EP_MSG_SVC_EIP512)

#define EIP_STAT_ALLOC_SHIFT		(8)
#define EIP_STAT_ALLOC_GET(atomicp)	((int) atomic_read(atomicp) >> EIP_STAT_ALLOC_SHIFT)
#define EIP_STAT_ALLOC_ADD(atomicp, v)	(atomic_add((v << EIP_STAT_ALLOC_SHIFT), atomicp))
#define EIP_STAT_ALLOC_SUB(atomicp, v)	(atomic_sub((v << EIP_STAT_ALLOC_SHIFT), atomicp))

#define EIP_STAT_QUEUED_MASK		(0xff)
#define EIP_STAT_QUEUED_GET(atomicp)	((int) atomic_read(atomicp) & EIP_STAT_QUEUED_MASK)

#define EIP_RMD_NR			(8)
#define EIP_RMD_MIN_NR			(8)
#define EIP_RMD_MAX_NR			(64)	/* should be < than (1 << EIP_STAT_ALLOC_SHIFT) */

#define EIP_RMD_ALLOC_STEP		(8)
#define EIP_RMD_ALLOC_THRESH		(16)

#define EIP_RMD_ALLOC			(1)
#define EIP_RMD_REPLACE			(0)

#define EIP_TMD_NR			(64)
#define EIP_TMD_MIN_NR			(16)
#define EIP_TMD_MAX_NR			(64)	/* should be < than (1 << EIP_STAT_ALLOC_SHIFT) */

#define EIP_TMD_TYPE_NR			(3)
#define EIP_TMD_COPYBREAK		(0x0)
#define EIP_TMD_STD			(0x1)
#define EIP_TMD_AGGREG			(0x2)

#define EIP_TX_COPYBREAK		(512)
#define EIP_TX_COPYBREAK_MAX		(1024)

#define EIP_IPFRAG_TO			(50)	/* time out before a frag is sent in msec */
#define EIP_IPFRAG_COPYBREAK		(EIP_SVC_BIGGEST_LEN - sizeof(EIP_IPFRAG) - EIP_HEADER_PAD)

#define EIP_RX_ENVELOPE_NR		((EIP_RMD_MAX_NR*EIP_SVC_NR)/2)
#define EIP_RX_GRANULARITY		(1)

#define EIP_IP_ALIGN(X)			(((X) + (15)) & ~(15))
#define EIP_EXTRA			roundup (sizeof(EIP_RMD), 256)
#define EIP_RCV_DMA_LEN(s)             	(s - EIP_EXTRA - EIP_HEADER_PAD)
#define EIP_MTU_MAX			(EIP_RCV_DMA_LEN(EIP_SVC_BIGGEST_LEN) - (ETH_HLEN))

#define SIZE_TO_SVC(s, svc) 									\
	do { 											\
					if (s <= EIP_RCV_DMA_LEN((1 << 9)))  {svc = 0;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 10))) {svc = 1;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 11))) {svc = 2;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 12))) {svc = 3;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 13))) {svc = 4;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 14))) {svc = 5;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 15))) {svc = 6;break;}	\
					if (s <= EIP_RCV_DMA_LEN((1 << 16))) {svc = 7;break;}	\
					svc = -666;						\
					EIP_ASSERT(1 == 0);					\
	} while (0)

extern int eipdebug;
#define EIP_ASSERT_ON 
/* #define NO_DEBUG */


/* ######################## */
#ifdef NO_DEBUG
#define __EIP_DBG_PRINTF(fmt, args...)
#define EIP_DBG_PRINTF(flag, fmt, args...)
#else

#define EIP_DBG_RMD		0x1
#define EIP_DBG_TMD		0x2
#define EIP_DBG_RMD_HEAD	0x4
#define EIP_DBG_TMD_HEAD	0x8
#define EIP_DBG_EIPH		0x10
#define EIP_DBG_IPH		0x20
#define EIP_DBG_RMD_EP_DVMA	0x40
#define EIP_DBG_TMD_EP_DVMA	0x80
#define EIP_DBG_EP_DVMA		(EIP_DBG_RMD_EP_DVMA|EIP_DBG_TMD_EP_DVMA)
#define EIP_DBG_MEMALLOC	0x100
#define EIP_DBG_MEMFREE		0x200
#define EIP_DBG_RMD_QUEUE	0x400
#define EIP_DBG_TMD_QUEUE	0x800
#define EIP_DBG_GEN		0x1000
#define EIP_DBG_DEBUG		0x2000
	
#define __EIP_DBG_PRINTF(fmt, args...)	(qsnet_debugf (QSNET_DEBUG_BUFFER, " CPU #%d %s: " fmt, smp_processor_id(), __func__, ## args))
#define EIP_DBG_PRINTF(flag, fmt, args...) (unlikely(eipdebug & flag) ? __EIP_DBG_PRINTF(fmt, ## args):(void)0)

#define __EIP_DBG_PRINTK(fmt, args...)	(qsnet_debugf (QSNET_DEBUG_BUF_CON, " CPU #%d %s: " fmt, smp_processor_id(), __func__, ## args))
#define EIP_DBG_PRINTK(flag, fmt, args...) (unlikely(eipdebug & flag) ? __EIP_DBG_PRINTF(fmt, ## args):(void)0)
	    
#define EIP_ERR_PRINTF(fmt, args...) 	__EIP_DBG_PRINTK("!!! ERROR !!! - " fmt, ## args)

	
#define EIP_DBG2(flag, fn, fn_arg, fmt, args...)								\
    if (unlikely(eipdebug & flag)) {                                    					\
	    qsnet_debugf (QSNET_DEBUG_BUFFER, "+CPU #%d %s: " fmt, smp_processor_id(), __func__, ##args);	\
            (void)(fn)(fn_arg);                                            					\
	    qsnet_debugf (QSNET_DEBUG_BUFFER, "-CPU #%d %s: " fmt, smp_processor_id(), __func__, ##args);	\
    }


#define EIP_DBG(flag, fn, args...)								\
    if (unlikely(eipdebug & flag)) {                                    			\
	    qsnet_debugf (QSNET_DEBUG_BUFFER, "+CPU #%d %s\n", smp_processor_id(), __func__);	\
            (void)(fn)(args);                                            			\
	    qsnet_debugf (QSNET_DEBUG_BUFFER, "-CPU #%d %s :\n", smp_processor_id(), __func__);	\
    }
#endif /* NO_DEBUG */


#ifdef EIP_ASSERT_ON

#define __EIP_ASSERT_PRINT(exp) 				\
		eipdebug = 0xffff;				\
		EIP_ERR_PRINTF("ASSERT : %s, %s::%d\n",		\
		       #exp, __BASE_FILE__, __LINE__); 		

#define EIP_ASSERT(exp)							\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
		        netif_stop_queue(eip_tx->net_device);	        \
		}

#define EIP_ASSERT2(exp, f, arg)					\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
		        f(arg);					        \
		} 							\
	} while (0)

#define EIP_ASSERT_BUG(exp) 						\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
			BUG();						\
		}							\
	} while (0)

#define EIP_ASSERT_GOTO(exp, label, f, arg)				\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
		        f(arg);						\
		        goto label;					\
		} 							\
	} while (0)

#define EIP_ASSERT_RET(exp, ret)					\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
			return ret;					\
		} 							\
	} while (0)

#define EIP_ASSERT_RETURN(exp, f, arg)					\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
		        f(arg);						\
			return;						\
		} 							\
	} while (0)

#define EIP_ASSERT_RETNULL(exp, f, arg)					\
	do {								\
		if (!(exp)) { 						\
			__EIP_ASSERT_PRINT(exp);                        \
		        f(arg);						\
			return NULL;					\
		} 							\
	} while (0)

#else

#define EIP_ASSERT(exp) 		do {} while(0)
#define EIP_ASSERT_OUT(exp) 		do {} while(0)
#define EIP_ASSERT_RETURN(exp) 		do {} while(0)
#define EIP_ASSERT_RETNULL(exp)		do {} while(0)
#define EIP_ASSERT_BUG(exp) 		do {} while(0)

#endif /* EIP_ASSERT */



typedef struct {
	u_short ip_bcast;
	u_short ip_inst;
	u_short ip_addr;
} EIP_ADDRESS;

typedef struct {
	EIP_ADDRESS h_dhost;
	EIP_ADDRESS h_shost;
	u_short h_sap;
} EIP_HEADER;
#define EIP_HEADER_PAD			(2)

typedef struct eip_proc_fs {
	const char *name;
	struct proc_dir_entry **parent;
	read_proc_t *read;
	write_proc_t *write;
	unsigned char allocated;
	struct proc_dir_entry *entry;
} EIP_PROC_FS;

#define EIP_PROC_ROOT_DIR		"eip"

#define EIP_PROC_DEBUG_DIR		"debug"
#define EIP_PROC_DEBUG_RX_FLUSH		"rx_flush"
#define EIP_PROC_DEBUG_TX_FLUSH		"tx_flush"

#define EIP_PROC_AGGREG_DIR		"aggregation"
#define EIP_PROC_AGGREG_ONOFF		"enable"
#define EIP_PROC_AGGREG_TO		"timeout"
#define EIP_PROC_AGGREG_COPYBREAK	"copybreak"

#define EIP_PROC_TX_COPYBREAK		"tx_copybreak"
#define EIP_PROC_STATS			"stats"
#define EIP_PROC_RX_GRAN		"rx_granularity"
#define EIP_PROC_TX_RAILMASK		"tx_railmask"
#define EIP_PROC_TMD_INUSE		"tmd_inuse"
#define EIP_PROC_EIPDEBUG		"eipdebug"
#define EIP_PROC_CHECKSUM               "checksum"

/* RX */
/* dma_len is used to keep the len of a received packet */
/* nmd.nmd_len is the max dma that can be received      */
/*                                                      */
struct eip_rmd {
	struct sk_buff *skb;

	EP_NMD nmd;
	u16 dvma_idx;

	EP_RXD *rxd;
	struct eip_rmd_head *head;
	union {
		struct list_head link;				/* when on "busy" list */
		struct eip_rmd  *next;				/* all other lists */
	} chain;
};
typedef struct eip_rmd EIP_RMD;
struct eip_rmd_head {
	EP_NMH *handle;

	EP_RCVR *rcvr;
	EIP_RMD *busy_list;

	/* stats */
	atomic_t stats;
	unsigned long dma;
};

typedef struct eip_rmd_head EIP_RMD_HEAD;
typedef struct eip_rx {
	struct eip_rmd_head head[EIP_SVC_NR];

	EIP_RMD *irq_list;
	short    irq_list_nr;	

	/* stats */
	unsigned long packets;
	unsigned long bytes;
	unsigned long errors;
	unsigned long dropped;
	unsigned long reschedule;

	spinlock_t lock;
	struct tasklet_struct tasklet;
	unsigned char rmd_max_nr;
	unsigned char sysctl_granularity;
	struct net_device *net_device;
} EIP_RX;

/* TX */
/* dma_len_max is the maximum len for a given DMA                      */
/* where mnd.nmd_len is the len of the packet to send ~> than skb->len */
typedef struct eip_ipfrag_handle {
	/* common with tmd */
	unsigned long dma_base;
	int dma_len;
	EP_NMD nmd;
	u16 dvma_idx;

	struct sk_buff *skb;
	struct eip_tmd_head *head;
	union {
		struct list_head link;				/* when on "busy" list */
		struct eip_tmd  *next;				/* all other lists */
	} chain;

	/* private */
	struct list_head list;
	struct timeval timestamp;
	unsigned int frag_nr;
	int datagram_len; /* Ip data */
	int dma_correction;
	EP_PAYLOAD payload;
} EIP_IPFRAG;

struct eip_tmd {
	unsigned long dma_base;
	int dma_len;
	EP_NMD nmd;
	u16 dvma_idx;

	struct sk_buff *skb;
	struct eip_tmd_head *head;
	union {
		struct list_head link;				/* when on "busy" list */
		struct eip_tmd  *next;				/* all other lists */
	} chain;
};

struct eip_tmd_head {
	EP_NMH *handle;

	struct eip_tmd *tmd;
	atomic_t stats;
};

typedef struct eip_tmd EIP_TMD;
typedef struct eip_tmd_head EIP_TMD_HEAD;

/* #define EIP_MORE_STATS */

typedef struct eip_tx {
	struct net_device *net_device;
	EP_XMTR *xmtr;
	EP_SYS *ep_system;

	struct eip_tmd_head head[EIP_TMD_TYPE_NR];
	struct list_head inuse;
	atomic_t destructor;

	/* stats */
	unsigned long packets;
	unsigned long bytes;
	unsigned long errors;
	unsigned long dropped;
	unsigned long dma[EIP_SVC_NR];
	
#ifdef EIP_MORE_STATS
	unsigned long sent_copybreak;
	unsigned long sent_std;
	unsigned long sent_aggreg;
#endif

	unsigned char tmd_max_nr;

	unsigned short sysctl_copybreak;
	unsigned short sysctl_ipfrag_to;
	unsigned short sysctl_ipfrag_copybreak;
	unsigned short sysctl_aggregation;

	unsigned short ipfrag_count;
	struct list_head ipfrag;
	spinlock_t ipfraglock;

	spinlock_t lock;
	struct tasklet_struct tasklet;
} EIP_TX;

/* =============================================== */
    /* unsigned long   multicast; */
#endif				/* __EIP_LINUX_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
