/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

/*
 * $Id: eip_stats.c,v 1.36.2.2 2005/09/07 14:34:58 mike Exp $
 * $Source: /cvs/master/quadrics/eipmod/eip_stats.c,v $
 */

#include <qsnet/kernel.h>
#include <qsnet/module.h>
#include <elan/epcomms.h>

#include <linux/netdevice.h>

#include <linux/kernel.h>
#include <linux/proc_fs.h>

#include <asm/atomic.h>

#include <qsnet/procfs_linux.h>

#include "eip_linux.h"
#include "eip_stats.h"

extern EIP_RX *eip_rx;
extern EIP_TX *eip_tx;
extern int tx_copybreak_max;
extern EP_RAILMASK tx_railmask;
extern int  eip_checksum_state;
extern void eip_stop_queue(void);
extern void eip_start_queue(void);

static int eip_stats_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
{
	int i, outlen = 0;

	*buf = '\0';
	strcat(buf, "\n");
	strcat(buf, "--------------------------------------------+------------+-----------------+\n");
	strcat(buf, "    SKB/DMA    |               | Rx         | Tx         |  TMD TYPE       |\n");
	strcat(buf, "--------------------------------------------+------------|-----------------+\n");

	i = 0;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld | #1[%3.3d/%3.3d/%3.3d] |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i],
		EIP_STAT_QUEUED_GET(&eip_tx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_tx->head[i].stats),
		 eip_tx->tmd_max_nr);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld | #2[%3.3d/%3.3d/%3.3d] |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i],
		EIP_STAT_QUEUED_GET(&eip_tx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_tx->head[i].stats),
		eip_tx->tmd_max_nr);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld | #3[%3.3d/%3.3d/%3.3d] |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i],
		EIP_STAT_QUEUED_GET(&eip_tx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_tx->head[i].stats),
		eip_tx->tmd_max_nr);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld +-----------------+\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i]);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i]);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i]);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i]);

	i++;
	sprintf(buf + strlen(buf), " [%5d/%5d] | [%3.3d/%3.3d/%3.3d] | %10ld | %10ld |\n",
		EIP_SVC_SMALLEST_LEN << i, (int) EIP_RCV_DMA_LEN((EIP_SVC_SMALLEST_LEN << i)),
		EIP_STAT_QUEUED_GET(&eip_rx->head[i].stats), EIP_STAT_ALLOC_GET(&eip_rx->head[i].stats),
		eip_rx->rmd_max_nr, eip_rx->head[i].dma, eip_tx->dma[i]);

	strcat(buf, "--------------------------------------------+------------+\n");
	sprintf(buf + strlen(buf), " RMD IRQ %4.4d                    %10lu | %10lu |\n",
		eip_rx->irq_list_nr, 
		eip_rx->packets, eip_tx->packets);
	strcat(buf, "--------------------------------------------+------------+\n");

#ifdef EIP_MORE_STATS
	strcat(buf, "\n");
	sprintf(buf + strlen(buf), " Copybreak %10ld Std %10ld Aggreg %10ld\n",
			eip_tx->sent_copybreak, eip_tx->sent_std, eip_tx->sent_aggreg);
#endif


	strcat(buf, "\n");
	sprintf(buf + strlen(buf), "Rx bytes: %lu (%lu Mb) errors: %lu dropped: %lu reschedule: %lu\n",
		eip_rx->bytes, eip_rx->bytes / (1024 * 1024), eip_rx->errors, eip_rx->dropped, eip_rx->reschedule);
	sprintf(buf + strlen(buf), "Tx bytes: %lu (%lu Mb) errors: %lu dropped: %lu\n",
		eip_tx->bytes, eip_tx->bytes / (1024 * 1024), eip_tx->errors, eip_tx->dropped);
	strcat(buf, "\n");

	outlen = strlen(buf);
	ASSERT(outlen < PAGE_SIZE);
	*eof = 1;
	return outlen;
}

void eip_stats_dump(void)
{
    int eof;

    char *buf = kmalloc(PAGE_SIZE, GFP_KERNEL);

    if (buf == NULL)
    {
	printk("no memory to produce eip_stats\n");
	return;
    }

    eip_stats_read(buf, NULL, 0, 0, &eof, NULL);

    printk(buf);

    kfree(buf);
}

static int eip_stats_write(struct file *file, const char *buf, unsigned long count, void *data)
{
	int i;
	unsigned long flags;

	spin_lock_irqsave(&eip_rx->lock, flags);
	eip_rx->packets = 0;
	eip_rx->bytes = 0;
	eip_rx->errors = 0;
	eip_rx->dropped = 0;
	eip_rx->reschedule = 0;
	for (i = 0; i < EIP_SVC_NR; eip_rx->head[i].dma = 0, i++);
	spin_unlock_irqrestore(&eip_rx->lock, flags);

	spin_lock_irqsave(&eip_tx->lock, flags);
	eip_tx->packets = 0;
	eip_tx->bytes = 0;
	eip_tx->errors = 0;
	eip_tx->dropped = 0;
#ifdef EIP_MORE_STATS
	eip_tx->sent_copybreak = 0;
	eip_tx->sent_std = 0;
	eip_tx->sent_aggreg = 0;
#endif
	for (i = 0; i < EIP_SVC_NR; eip_tx->dma[i] = 0, i++);
	spin_unlock_irqrestore(&eip_tx->lock, flags);

	return count;
}

#define		eip_stats_var_write(name)									\
static int eip_stats_##name##_write(struct file *file, const char *buf, unsigned long count, void *data)	\
{														\
	char * b = (char *) buf;										\
	*(b + count) = '\0';											\
	eip_##name##_set((int) simple_strtoul(b, NULL, 10));							\
	return count;												\
}

#define 	eip_stats_var_read(name, var)									\
static int eip_stats_##name##_read(char *buf, char **start, off_t off, int count, int *eof, void *data)		\
{														\
	sprintf(buf, "%d\n", var);										\
	*eof = 1;												\
	return strlen(buf);											\
}


#define		eip_stats_var_set(name, min, max, default, var)									\
void eip_##name##_set(int i)													\
{																\
	if ( (i >= min) && (i <= max)) {											\
		EIP_DBG_PRINTK(EIP_DBG_GEN, "Setting " #name " to %d\n", i);							\
        	var =(unsigned short) i;											\
	}															\
	else {															\
		EIP_ERR_PRINTF("parameter error : %d <= " #name "(%d) <= %d using default %d\n", min, i, (int) max, (int) default);	\
	}															\
}

eip_stats_var_set(tx_copybreak, 0, tx_copybreak_max, EIP_TX_COPYBREAK, eip_tx->sysctl_copybreak);
eip_stats_var_set(rx_granularity, 1, EIP_RMD_MIN_NR, EIP_RX_GRANULARITY, eip_rx->sysctl_granularity);
eip_stats_var_set(tx_railmask, 0, EP_RAILMASK_ALL, EP_RAILMASK_ALL, tx_railmask);
eip_stats_var_set(ipfrag_to, 0, (1 << 16), EIP_IPFRAG_TO, eip_tx->sysctl_ipfrag_to);
eip_stats_var_set(aggregation, 0, 1, 1, eip_tx->sysctl_aggregation);
eip_stats_var_set(ipfrag_copybreak, 0, EIP_IPFRAG_COPYBREAK, EIP_IPFRAG_COPYBREAK, eip_tx->sysctl_ipfrag_copybreak);
/* eip_stats_var_set(eipdebug, 0, , 0, eipdebug); */

eip_stats_var_read(aggregation, eip_tx->sysctl_aggregation);
eip_stats_var_read(ipfrag_count, eip_tx->ipfrag_count);
eip_stats_var_read(ipfrag_to, eip_tx->sysctl_ipfrag_to);
eip_stats_var_read(ipfrag_copybreak, eip_tx->sysctl_ipfrag_copybreak);
eip_stats_var_read(tx_copybreak, eip_tx->sysctl_copybreak);
eip_stats_var_read(rx_granularity, eip_rx->sysctl_granularity);
eip_stats_var_read(tx_railmask, tx_railmask);

eip_stats_var_write(aggregation);
eip_stats_var_write(ipfrag_to);
eip_stats_var_write(ipfrag_copybreak);
eip_stats_var_write(tx_copybreak);
eip_stats_var_write(rx_granularity);
eip_stats_var_write(tx_railmask);


static int eip_checksum_write(struct file *file, const char *buf, unsigned long count, void *data)
{
	char * b = (char *) buf;
	int    value;

	*(b + count) = '\0';

	value = (int) simple_strtoul(b, NULL, 10);
	if  ((value >= CHECKSUM_NONE) && (value <= CHECKSUM_UNNECESSARY)) 
		eip_checksum_state = value;
	else 
		EIP_ERR_PRINTF("%d <= checksum(%d) <= %d using old value %d\n", CHECKSUM_NONE, value, CHECKSUM_UNNECESSARY, eip_checksum_state);

	return count;
}

static int eip_checksum_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
{
	switch ( eip_checksum_state ) 
	{
	case 0  : sprintf(buf, "0 CHECKSUM_NONE\n");                      break;
	case 1  : sprintf(buf, "1 CHECKSUM_HW\n");                        break;
	case 2  : sprintf(buf, "2 CHECKSUM_UNNECESSARY\n");               break;
	default : sprintf(buf, "%d INVALID VALUE\n", eip_checksum_state); break;
	}
	*eof = 1;
	return strlen(buf);
}

static int eip_stats_eipdebug_read(char *buf, char **start, off_t off, int count, int *eof, void *data)
{
	*buf = '\0';
	sprintf(buf + strlen(buf), "0x%x\n", eipdebug);
	*eof = 1;
	return strlen(buf);
}
static int eip_stats_eipdebug_write(struct file *file, const char *buf, unsigned long count, void *data)
{
	char * p = (char *) buf;
    	*(p + count - 1) = '\0';
	eipdebug = simple_strtoul(p, NULL, 0);
	__EIP_DBG_PRINTK("Setting eipdebug to 0x%x\n", eipdebug);
	return count;
}

static int eip_stats_tmd_inuse_read(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	struct list_head *lp;
	unsigned long flags;
	unsigned int len = 0;

	spin_lock_irqsave(&eip_tx->lock, flags);
	list_for_each (lp, &eip_tx->inuse) {
		EIP_TMD *tmd = list_entry (lp, EIP_TMD, chain.link);
		EIP_HEADER *eiph = (EIP_HEADER *) tmd->dma_base;
		
                len += sprintf(page+len, "tmd=%p id=%d len=%d\n",
			       tmd, eiph ? ntohs(eiph->h_dhost.ip_addr) : -1,
			       tmd->dma_len);

                if (len + 40 >= count)
                        break;
        }
        spin_unlock_irqrestore(&eip_tx->lock, flags);

	return qsnet_proc_calc_metrics (page, start, off, count, eof, len);
}

static int eip_stats_debug_rx_flush(struct file *file, const char *buf, unsigned long count, void *data)
{
	EIP_DBG_PRINTF(EIP_DBG_GEN, "Flushing rx ...\n");
	tasklet_schedule(&eip_rx->tasklet);
	return count;
}
static int eip_stats_debug_tx_flush(struct file *file, const char *buf, unsigned long count, void *data)
{
	EIP_DBG_PRINTF(EIP_DBG_GEN, "Flushing tx ... %d tmds reclaimed\n", ep_enable_txcallbacks(eip_tx->xmtr));
	ep_disable_txcallbacks(eip_tx->xmtr);
	tasklet_schedule(&eip_tx->tasklet);
	return count;
}

#define EIP_PROC_PARENT_NR	(3)
/* NOTE : the parents should be declared b4 the children */
static EIP_PROC_FS eip_procs[] = {
	/* {name, parent, read fn, write fn, allocated, entry}, */
	{EIP_PROC_ROOT_DIR, &qsnet_procfs_root, NULL, NULL, 0, NULL},
	{EIP_PROC_DEBUG_DIR, &eip_procs[0].entry, NULL, NULL, 0, NULL},
	{EIP_PROC_AGGREG_DIR, &eip_procs[0].entry, NULL, NULL, 0, NULL},	/* end of parents */
	{EIP_PROC_STATS, &eip_procs[0].entry, eip_stats_read, eip_stats_write, 0, NULL},
	{EIP_PROC_TX_COPYBREAK, &eip_procs[0].entry, eip_stats_tx_copybreak_read, eip_stats_tx_copybreak_write, 0, NULL},
	{EIP_PROC_RX_GRAN, &eip_procs[0].entry, eip_stats_rx_granularity_read, eip_stats_rx_granularity_write, 0, NULL},
	{EIP_PROC_TX_RAILMASK, &eip_procs[0].entry, eip_stats_tx_railmask_read, eip_stats_tx_railmask_write, 0, NULL},
	{EIP_PROC_TMD_INUSE, &eip_procs[0].entry, eip_stats_tmd_inuse_read, NULL, 0, NULL},
	{EIP_PROC_EIPDEBUG, &eip_procs[0].entry, eip_stats_eipdebug_read, eip_stats_eipdebug_write, 0, NULL},
	{EIP_PROC_CHECKSUM, &eip_procs[0].entry, eip_checksum_read, eip_checksum_write, 0, NULL},
	{EIP_PROC_DEBUG_RX_FLUSH, &eip_procs[1].entry, NULL, eip_stats_debug_rx_flush, 0, NULL},
	{EIP_PROC_DEBUG_TX_FLUSH, &eip_procs[1].entry, NULL, eip_stats_debug_tx_flush, 0, NULL},
	{"ipfrag_count", &eip_procs[2].entry, eip_stats_ipfrag_count_read, NULL, 0, NULL},
	{EIP_PROC_AGGREG_TO, &eip_procs[2].entry, eip_stats_ipfrag_to_read, eip_stats_ipfrag_to_write, 0, NULL},
	{EIP_PROC_AGGREG_ONOFF, &eip_procs[2].entry, eip_stats_aggregation_read, eip_stats_aggregation_write, 0, NULL},
	{EIP_PROC_AGGREG_COPYBREAK, &eip_procs[2].entry, eip_stats_ipfrag_copybreak_read, eip_stats_ipfrag_copybreak_write, 0, NULL},
	{NULL, NULL, NULL, NULL, 1, NULL},
};

int eip_stats_init(void)
{
	int p;

	for (p = 0; !eip_procs[p].allocated; p++) {
		if (p < EIP_PROC_PARENT_NR)
			eip_procs[p].entry = proc_mkdir(eip_procs[p].name, *eip_procs[p].parent);
		else
			eip_procs[p].entry = create_proc_entry(eip_procs[p].name, 0, *eip_procs[p].parent);

		if (!eip_procs[p].entry) {
			EIP_ERR_PRINTF("%s\n", "Cannot allocate proc entry");
			eip_stats_cleanup();
			return -ENOMEM;
		}

		eip_procs[p].entry->owner = THIS_MODULE;
		eip_procs[p].entry->write_proc = eip_procs[p].write;
		eip_procs[p].entry->read_proc = eip_procs[p].read;
		eip_procs[p].allocated = 1;
	}
	eip_procs[p].allocated = 0;
	return 0;
}

void eip_stats_cleanup(void)
{
	int p;
	for (p = (sizeof (eip_procs)/sizeof (eip_procs[0]))-1; p >= 0; p--)
		if (eip_procs[p].allocated) {
			EIP_DBG_PRINTF(EIP_DBG_GEN, "Removing %s from proc\n", eip_procs[p].name);
			remove_proc_entry(eip_procs[p].name, *eip_procs[p].parent);
		}
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
