/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: bitmap.c,v 1.5 2004/01/20 17:32:17 david Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/shared/bitmap.c,v $*/

#if defined(__KERNEL__)
#include <qsnet/kernel.h>
#endif
#include <qsnet/config.h>
#include <elan/bitmap.h>

/*
 * Return the index of the first available bit in the 
 * bitmap , or -1 for failure
 */
int
bt_freebit (bitmap_t *bitmap, int nbits)
{
    int last = (--nbits) >> BT_ULSHIFT;
    int maxbit;
    int	i, j;

    /* look for a word with a bit off */
    for (i = 0; i <= last; i++)
	if (bitmap[i] != ~((bitmap_t) 0))
	    break;

    if (i <= last)
    {
	/* found an word with a bit off,  now see which bit it is */
	maxbit = (i == last) ? (nbits & BT_ULMASK) : (BT_NBIPUL-1);
	for (j = 0; j <= maxbit; j++)
	    if ((bitmap[i] & (1 << j)) == 0)
		return ((i << BT_ULSHIFT) | j);
    }
    return (-1);
    
}

/*
 * bt_lowbit:
 *     Return the index of the lowest set bit in the
 *     bitmap, or -1 for failure.
 */
int
bt_lowbit (bitmap_t *bitmap, int nbits)
{
    int last = (--nbits) >> BT_ULSHIFT;
    int maxbit;
    int i, j;
    
    /* look for a word with a bit on */
    for (i = 0; i <= last; i++)
	if (bitmap[i] != 0)
	    break;
    if (i <= last)
    {
	/* found a word bit a bit on, now see which bit it is */
	maxbit = (i == last) ? (nbits & BT_ULMASK) : (BT_NBIPUL-1);
	for (j = 0; j <= maxbit; j++)
	    if (bitmap[i] & (1 << j))
		return ((i << BT_ULSHIFT) | j);
    }

    return (-1);
}

/*
 * Return the index of the first available bit in the 
 * bitmap , or -1 for failure
 */
int
bt_nextbit (bitmap_t *bitmap, int nbits, int last, int isset)
{
    int first = ((last+1) + BT_NBIPUL-1) >> BT_ULSHIFT;
    int end   = (--nbits) >> BT_ULSHIFT;
    int maxbit;
    int	i, j;

    /* look for bits before the first whole word */
    if (((last+1) & BT_ULMASK) != 0)
    {
	maxbit = ((first-1) == last) ? (nbits & BT_ULMASK) : (BT_NBIPUL-1);
	for (j = ((last+1) & BT_ULMASK); j <= maxbit; j++)
	    if ((bitmap[first-1] & (1 << j)) == (isset << j))
		return (((first-1) << BT_ULSHIFT) | j);
    }

    /* look for a word with a bit off */
    for (i = first; i <= end; i++)
	if (bitmap[i] != (isset ? 0 : ~((bitmap_t) 0)))
	    break;

    if (i <= end)
    {
	/* found an word with a bit off,  now see which bit it is */
	maxbit = (i == end) ? (nbits & BT_ULMASK) : (BT_NBIPUL-1);
	for (j = 0; j <= maxbit; j++)
	    if ((bitmap[i] & (1 << j)) == (isset << j))
		return ((i << BT_ULSHIFT) | j);
    }
    return (-1);
}

void
bt_copy (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	b[i] = a[i];

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	if (BT_TEST(a, i))
	    BT_SET(b,i);
	else
	    BT_CLEAR(b,i);
}

void
bt_zero (bitmap_t *bitmap, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	bitmap[i] = 0;

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	BT_CLEAR(bitmap,i);
}

void
bt_fill (bitmap_t *bitmap, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	bitmap[i] = ~((bitmap_t) 0);

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	BT_SET(bitmap,i);
}

int
bt_cmp (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	if (a[i] != b[i])
	    return (1);

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	if (BT_TEST (a, i) != BT_TEST(b, i))
	    return (1);
    return (0);
}

void
bt_intersect (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;
    
    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	a[i] &= b[i];

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	if (BT_TEST (a, i) && BT_TEST (b, i))
	    BT_SET (a, i);
	else
	    BT_CLEAR (a, i);
}

void
bt_remove (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	a[i] &= ~b[i];

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	if (BT_TEST (b, i))
	    BT_CLEAR (a, i);
}

void
bt_add (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;

    for (i = 0; i < (nbits>>BT_ULSHIFT); i++)
	a[i] |= b[i];

    for (i <<= BT_ULSHIFT; i < nbits; i++)
	if (BT_TEST(b, i))
	    BT_SET (a, i);
}

/*
 * bt_spans : partition a spans partition b
 *    == all bits set in 'b' are set in 'a'
 */
int
bt_spans (bitmap_t *a, bitmap_t *b, int nbits)
{
    int i;
    
    for (i = 0; i < nbits; i++)
	if (BT_TEST (b, i) && !BT_TEST (a, i))
	    return (0);
    return (1);
}

/*
 * bt_subset: copy [base,base+nbits-1] from 'a' to 'b'
 */
void
bt_subset (bitmap_t *a, bitmap_t *b, int base, int nbits)
{
    int i;

    for (i = 0; i < nbits; i++)
    {
	if (BT_TEST (a, base+i))
	    BT_SET(b,i);
	else
	    BT_CLEAR (b,i);
    }
}

void 
bt_up (bitmap_t *a, bitmap_t *b, bitmap_t *c, int nbits)
{
    int i;
    
    for (i = 0; i < nbits; i++)
    {
	if (!BT_TEST (a, i) && BT_TEST (b, i))
	{
	    BT_SET (c, i);
        }
	else
	{
	    BT_CLEAR (c, i);
        }
    }
}

void 
bt_down (bitmap_t *a, bitmap_t *b, bitmap_t *c, int nbits)
{
    int i;
    
    for (i = 0; i < nbits; i++)
    {
	if (BT_TEST (a, i) && !BT_TEST (b, i))
	{
	    BT_SET (c, i);
        }
	else
	{
	    BT_CLEAR (c, i);
        }
    }
}

int
bt_nbits (bitmap_t *a, int nbits)
{
    int i, c;
    for (i = 0, c = 0; i < nbits; i++)
	if (BT_TEST (a, i))
	    c++;
    return (c);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
