/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: capability.c,v 1.19.2.2 2005/07/21 10:42:36 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/capability.c,v $ */


#include <qsnet/kernel.h>
#include <elan/elanmod.h>

static LIST_HEAD(elan_cap_list); 

typedef struct elan_vp_struct
{
	struct list_head list;
	ELAN_CAPABILITY  vp;
} ELAN_VP_NODE_STRUCT;

/* There is an array of these structs for each process/context in the CAP 
 * This is then replicated for each rail. The usercopy handle stuff is 
 * only maintained in rail 0 though
 */
typedef struct elan_attached_struct
{
	void               *cb_args;
	ELAN_DESTROY_CB  cb_func;
	struct task_struct *handle;		/* usercopy: attached task handle */
	struct task_struct *owner;		/* usercopy: attached task handle owner */
} ELAN_ATTACHED_STRUCT;

typedef struct elan_cap_node_struct
{
	struct list_head list;
	ELAN_CAP_STRUCT     node;
	ELAN_ATTACHED_STRUCT *attached[ELAN_MAX_RAILS];
     	struct list_head vp_list;
} ELAN_CAP_NODE_STRUCT;


ELAN_CAP_NODE_STRUCT *
find_cap_node(ELAN_CAPABILITY *cap)
{
	struct list_head        *tmp;
	ELAN_CAP_NODE_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_cap_list) {
		ptr = list_entry(tmp, ELAN_CAP_NODE_STRUCT , list);
		/* is it an exact match (key not checked) */
		if ( ELAN_CAP_TYPE_MATCH(&ptr->node.cap,cap) 
		     && ELAN_CAP_GEOM_MATCH(&ptr->node.cap,cap)) {
			return ptr;
		}
	}
	return ptr;
}

ELAN_VP_NODE_STRUCT *
find_vp_node( ELAN_CAP_NODE_STRUCT *cap_node,ELAN_CAPABILITY *map)
{
	struct list_head       * tmp;
	ELAN_VP_NODE_STRUCT * ptr = NULL;

	list_for_each(tmp, &cap_node->vp_list) {
		ptr = list_entry(tmp, ELAN_VP_NODE_STRUCT , list);
		/* is it an exact match (key not checked) */
		if ( ELAN_CAP_TYPE_MATCH(&ptr->vp,map) 
		     && ELAN_CAP_GEOM_MATCH(&ptr->vp,map)){
			return ptr;
		}
	}
	return ptr;
}

int 
elan_validate_cap(ELAN_CAPABILITY *cap)
{
	char                      space[127];

	ELAN_DEBUG1 (ELAN_DBG_VP,"elan_validate_cap %s\n",elan_capability_string(cap,space));

	/* check versions */
	if (cap->cap_version != ELAN_CAP_VERSION_NUMBER)
	{
		ELAN_DEBUG2 (ELAN_DBG_VP,"elan_validate_cap: (cap->Version != ELAN_CAP_VERSION) %d %d\n", cap->cap_version, ELAN_CAP_VERSION_NUMBER);
		return (EINVAL);
	}

	/* check its not HWTEST */
	if ( cap->cap_type & ELAN_CAP_TYPE_HWTEST )
	{
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_cap: failed type = ELAN_CAP_TYPE_HWTEST \n");	
		return (EINVAL);
	}
	
	/* check its type */
	switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_KERNEL :	
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_cap: failed type = ELAN_CAP_TYPE_KERNEL \n");	
		return (EINVAL);

		/* check it has a valid type */
	case ELAN_CAP_TYPE_BLOCK:
	case ELAN_CAP_TYPE_CYCLIC:
		break;

		/* all others are failed as well */
	default:
		ELAN_DEBUG1 (ELAN_DBG_VP,"elan_validate_cap: failed unknown type = %x \n", (cap->cap_type & ELAN_CAP_TYPE_MASK));	
		return (EINVAL);
	}
	
	if ((cap->cap_lowcontext == ELAN_CAP_UNINITIALISED) || (cap->cap_highcontext == ELAN_CAP_UNINITIALISED)
	    || (cap->cap_lownode == ELAN_CAP_UNINITIALISED) || (cap->cap_highnode    == ELAN_CAP_UNINITIALISED))
	{
		
		ELAN_DEBUG4 (ELAN_DBG_VP,"elan_validate_cap: ELAN_CAP_UNINITIALISED   LowNode %d   HighNode %d   LowContext %d   highContext %d\n",
			     cap->cap_lownode , cap->cap_highnode,
			     cap->cap_lowcontext , cap->cap_highcontext);
		return (EINVAL);
	}	

	if (cap->cap_lowcontext > cap->cap_highcontext)
	{
		ELAN_DEBUG2 (ELAN_DBG_VP,"elan_validate_cap: (cap->cap_lowcontext > cap->cap_highcontext) %d %d\n",cap->cap_lowcontext , cap->cap_highcontext);
		return (EINVAL);
	}
	
	if (cap->cap_lownode > cap->cap_highnode)
	{
		ELAN_DEBUG2 (ELAN_DBG_VP,"elan_validate_cap: (cap->cap_lownode > cap->cap_highnode) %d %d\n",cap->cap_lownode, cap->cap_highnode);
		return (EINVAL);
	}

	if (cap->cap_mycontext != ELAN_CAP_UNINITIALISED) 
	{
		ELAN_DEBUG1 (ELAN_DBG_VP,"elan_validate_cap: failed cap->cap_mycontext is set %d  \n", cap->cap_mycontext);
		return (EINVAL);
	}


     	if ((ELAN_CAP_NUM_NODES(cap) * ELAN_CAP_NUM_CONTEXTS(cap)) > ELAN_MAX_VPS)
	{
		ELAN_DEBUG6 (ELAN_DBG_VP,"elan_validate_cap: too many vps  LowNode %d   HighNode %d   LowContext %d   highContext %d,  %d >% d\n",
			     cap->cap_lownode , cap->cap_highnode,
			     cap->cap_lowcontext , cap->cap_highcontext,
			     (ELAN_CAP_NUM_NODES(cap) * ELAN_CAP_NUM_CONTEXTS(cap)),
			     ELAN_MAX_VPS);
		
		return (EINVAL);
	}

	return (ESUCCESS);
}

int
elan_validate_map(ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map)
{
	ELAN_CAP_NODE_STRUCT * ptr  = NULL;
	ELAN_VP_NODE_STRUCT  * vptr = NULL;
	char space[256];

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map \n");
	ELAN_DEBUG1 (ELAN_DBG_VP,"elan_validate_map cap = %s \n",elan_capability_string(cap,space));
	ELAN_DEBUG1 (ELAN_DBG_VP,"elan_validate_map map = %s \n",elan_capability_string(map,space));

	/* does cap exist    */
	ptr = find_cap_node(cap);
	if ( ptr == NULL ) 
	{
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map: cap not found \n");
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return EINVAL;
	}
	/* is it active */
	if ( ! ptr->node.active ) 
	{
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map: cap not active \n");
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return EINVAL;
	}

	/* are they the same */
	if ( ELAN_CAP_TYPE_MATCH(cap,map) 
	     && ELAN_CAP_GEOM_MATCH(cap,map)) 
	{
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map: cap == map  passed\n");
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return ESUCCESS;
	}

	/* is map in map list */
	vptr = find_vp_node(ptr, map);
	if ( vptr == NULL ) 
	{
		ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map:  map not found\n");
		ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
		return EINVAL;
	}
	
	ELAN_DEBUG0 (ELAN_DBG_VP,"elan_validate_map:  map passed\n");
	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return ESUCCESS;
}

int
elan_create_cap(ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap)
{
	char                      space[127];
	struct list_head        * tmp;
	ELAN_CAP_NODE_STRUCT * ptr = NULL;
	int                       i, rail;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	ELAN_DEBUG1 (ELAN_DBG_VP,"elan_create_cap %s\n",elan_capability_string(cap,space));	

	/* need to check that the cap does not over lap another one 
	   or is an exact match with only the userkey changing */
	list_for_each(tmp, &elan_cap_list) {
		ptr = list_entry(tmp, ELAN_CAP_NODE_STRUCT , list);

		/* is it an exact match (key not checked) */
		if ( ELAN_CAP_TYPE_MATCH(&ptr->node.cap,cap) 
		     && ELAN_CAP_GEOM_MATCH(&ptr->node.cap,cap)
		     && (ptr->node.owner == owner)) {
			if ( ptr->node.active ) {
				/* dont inc attached count as its like a create */
				ptr->node.cap.cap_userkey = cap->cap_userkey;
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return ESUCCESS;
			}
			else
			{
				ELAN_DEBUG (ELAN_DBG_VP,"elan_create_cap failed %s\n",
					    elan_capability_string(&ptr->node.cap,space));
				ELAN_DEBUG (ELAN_DBG_VP,"elan_create_cap failed ptr %p owner %p attached %d\n",
					    ptr, owner, ptr->node.attached);
					     
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return EINVAL;
			}
		}
		
		/* does it overlap, even with ones being destroyed */
		if (elan_cap_overlap(&ptr->node.cap,cap))
		{
			ELAN_DEBUG (ELAN_DBG_VP,"elan_create_cap failed overlap %s\n",
				    elan_capability_string(&ptr->node.cap,space));
			ELAN_DEBUG (ELAN_DBG_VP,"elan_create_cap failed overlap ptr %p owner %p attached %d active %d\n",
				    ptr, owner, ptr->node.attached, ptr->node.active);

			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
			return  EACCES;
		}
	}

	/* create it */
	KMEM_ALLOC(ptr, ELAN_CAP_NODE_STRUCT *, sizeof(ELAN_CAP_NODE_STRUCT), 1);
	if (ptr == NULL)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  ENOMEM;
	}

	/* create per rail space for the attached array */
	for(rail=0;rail<ELAN_MAX_RAILS;rail++)
	{
		ptr->attached[rail]=NULL;
		/* GNAT 7685: Always need to allocate an attached structure in rail 0 for the usercopy device */
		if ( ELAN_CAP_IS_RAIL_SET(cap,rail) || rail == 0 ) 
		{
			KMEM_ALLOC(ptr->attached[rail], ELAN_ATTACHED_STRUCT *, sizeof(ELAN_ATTACHED_STRUCT) *  ELAN_CAP_NUM_CONTEXTS(cap), 1);
			if (ptr->attached[rail] == NULL) 
			{
				for(;rail>=0;rail--)
					if ( ptr->attached[rail] )
						KMEM_FREE(ptr->attached[rail], sizeof(ELAN_ATTACHED_STRUCT) *  ELAN_CAP_NUM_CONTEXTS(cap));

				KMEM_FREE(ptr, sizeof(ELAN_CAP_NODE_STRUCT));
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return  ENOMEM;
			}
			/* blank the per context attached array */
			for(i=0;i<ELAN_CAP_NUM_CONTEXTS(cap);i++)
			{
				ptr->attached[rail][i].cb_func = NULL;
				/* user-to-user copy */
				ptr->attached[rail][i].handle  = NULL;
				ptr->attached[rail][i].owner   = NULL;
			}
		}
	}	
	
	ptr->node.owner     = owner;
	ptr->node.cap       = *cap;
	ptr->node.attached  = 1;    /* creator counts as attached */
	ptr->node.active    = 1;
	ptr->vp_list.next   = &(ptr->vp_list);
	ptr->vp_list.prev   = &(ptr->vp_list);

	list_add_tail(&ptr->list, &elan_cap_list);	

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

void
elan_destroy_cap_test(ELAN_CAP_NODE_STRUCT *cap_ptr)
{
	/* called by someone holding the mutex   */
	struct list_head       * vp_tmp;
	ELAN_VP_NODE_STRUCT * vp_ptr = NULL;
	int                      rail;

	ASSERT(cap_ptr->node.attached >= 0);

	/* check to see if it can be deleted now */
	if ( cap_ptr->node.attached == 0 ) {
		
		ELAN_DEBUG1(ELAN_DBG_CAP,"elan_destroy_cap_test: %p attached == 0\n", cap_ptr);	
		
		/* delete the vp list */
		list_for_each(vp_tmp, &(cap_ptr->vp_list)) {
			vp_ptr = list_entry(vp_tmp, ELAN_VP_NODE_STRUCT , list);
			list_del(&vp_ptr->list);
			KMEM_FREE( vp_ptr, sizeof(ELAN_VP_NODE_STRUCT));
		}
		
		list_del(&cap_ptr->list);

		/* delete space for the attached array */
		for(rail=0;rail<ELAN_MAX_RAILS;rail++)
			if (cap_ptr->attached[rail]) 
				KMEM_FREE(cap_ptr->attached[rail], sizeof(ELAN_ATTACHED_STRUCT) * ELAN_CAP_NUM_CONTEXTS(&(cap_ptr->node.cap)));
			
		KMEM_FREE(cap_ptr, sizeof(ELAN_CAP_NODE_STRUCT));		
	}
	else
		ELAN_DEBUG2(ELAN_DBG_CAP,"elan_destroy_cap_test: %p attached = %d\n",
			    cap_ptr, cap_ptr->node.attached);	

}

int
elan_destroy_cap(ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap)
{
	char                      space[127];
	struct list_head        * el;
	struct list_head        * nel;
	ELAN_CAP_NODE_STRUCT * ptr = NULL;
	int                       i, rail;
	int                       found = 0;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	ELAN_DEBUG1 (ELAN_DBG_CAP,"elan_destroy_cap %s\n",elan_capability_string(cap,space));	

	list_for_each_safe (el, nel, &elan_cap_list) {
		ptr = list_entry(el, ELAN_CAP_NODE_STRUCT , list);
		
		/* is it an exact match */
		if ( (ptr->node.owner == owner )
		     && (  (cap == NULL) 
			   || (ELAN_CAP_TYPE_MATCH(&ptr->node.cap,cap) && ELAN_CAP_GEOM_MATCH(&ptr->node.cap,cap)))) {

			if ( ptr->node.active ) {

				/* mark as in active and dec attached count */
				ptr->node.active = 0;
				ptr->node.attached--;
				ptr->node.owner  = 0; /* no one own's it now */

				ASSERT(ptr->node.attached >= 0);
				
				/* need to tell any one who was attached that this has been destroy'd */
				for(rail=0;rail<ELAN_MAX_RAILS;rail++)
					if (ELAN_CAP_IS_RAIL_SET( &(ptr->node.cap), rail)) {
						for(i=0;i< ELAN_CAP_NUM_CONTEXTS(&(ptr->node.cap));i++)
							if ( ptr->attached[rail][i].cb_func != NULL) 
								ptr->attached[rail][i].cb_func(ptr->attached[rail][i].cb_args, cap, NULL);
					}
				
				/* now try to destroy it */
				elan_destroy_cap_test(ptr);
				
				/* found it */
				found = 1;
			}
		}
	}
	
	if ( found )
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return ESUCCESS;
	}

	/* failed */
	ELAN_DEBUG1(ELAN_DBG_CAP,"elan_destroy_cap: %p didnt find it \n", cap);	

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return EINVAL;
}

int 
elan_get_caps(uint *number_of_results, uint array_size, ELAN_CAP_STRUCT *caps)
{
	uint                      results = 0;
	struct list_head        * tmp;
	ELAN_CAP_NODE_STRUCT * ptr = NULL;
	

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	ELAN_DEBUG0(ELAN_DBG_CAP,"elan_get_caps\n");	

	list_for_each(tmp, &elan_cap_list) {
		ptr = list_entry(tmp, ELAN_CAP_NODE_STRUCT , list);
		
		copyout(&ptr->node, &caps[results], sizeof (ELAN_CAP_STRUCT));
		
		results++;
		
		if ( results >= array_size )
		{
			copyout(&results, number_of_results, sizeof(uint));	
			ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
			return ESUCCESS;
		}
	}

	copyout(&results, number_of_results, sizeof(uint));	

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return ESUCCESS;
}

int
elan_create_vp(ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map)
{
	ELAN_CAP_NODE_STRUCT * cap_ptr = NULL;
	ELAN_VP_NODE_STRUCT  * vp_ptr  = NULL;
	
	ELANMOD_RWLOCK_WRITE(&elan_rwlock);


	ELAN_DEBUG0(ELAN_DBG_CAP,"elan_create_vp\n");

	/* the railmasks must match */
	if ( cap->cap_railmask != map->cap_railmask)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  EINVAL;
	}

	/* does the cap exist */
	cap_ptr = find_cap_node(cap);
	if ((cap_ptr == NULL) || ( cap_ptr->node.owner != owner ) || (! cap_ptr->node.active) )
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  EINVAL;
	}
	
	/* is there already a mapping */
	vp_ptr = find_vp_node(cap_ptr,map);
	if ( vp_ptr != NULL) 
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  EINVAL;
	}

	/* create space for mapping */
	KMEM_ALLOC(vp_ptr, ELAN_VP_NODE_STRUCT *, sizeof(ELAN_VP_NODE_STRUCT), 1);
	if (vp_ptr == NULL)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  ENOMEM;
	}
			
	/* copy map */
	vp_ptr->vp = *map;
	list_add_tail(&vp_ptr->list, &(cap_ptr->vp_list));	
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

int
elan_destroy_vp(ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map)
{
	ELAN_CAP_NODE_STRUCT * cap_ptr = NULL;
	ELAN_VP_NODE_STRUCT  * vp_ptr  = NULL;
	int                       i, rail;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	ELAN_DEBUG0(ELAN_DBG_CAP,"elan_destroy_vp\n");	

	cap_ptr = find_cap_node(cap);
	if ((cap_ptr!=NULL) && (cap_ptr->node.owner == owner) && ( cap_ptr->node.active))
	{		
		vp_ptr = find_vp_node( cap_ptr, map );
		if ( vp_ptr != NULL ) 
		{
			list_del(&vp_ptr->list);
			KMEM_FREE(vp_ptr, sizeof(ELAN_VP_NODE_STRUCT));
	      
			/* need to tell those who are attached that map is nolonger in use */
			for(rail=0;rail<ELAN_MAX_RAILS;rail++)
				if (ELAN_CAP_IS_RAIL_SET(cap, rail))
				{
					for(i=0;i< ELAN_CAP_NUM_CONTEXTS(&(cap_ptr->node.cap));i++)
						if ( cap_ptr->attached[rail][i].cb_func != NULL) 
							cap_ptr->attached[rail][i].cb_func( cap_ptr->attached[rail][i].cb_args, cap, map);
				}

			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
			return  ESUCCESS;
		}
	}	
	
	/* didnt find it */
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  EINVAL;
}

int 
elan_attach_cap(ELAN_CAPABILITY *cap, unsigned int rail, void *args, ELAN_DESTROY_CB func)
{
	char                  space[127];
	struct list_head     *el;

	ELAN_DEBUG1 (ELAN_DBG_CAP,"elan_attach_cap %s\n",elan_capability_string(cap,space));

	/* currently must provide a call back, as null mean something */
	if ( func == NULL)
		return (EINVAL);

	/* mycontext must be set and correct */
	if ( ! ELAN_CAP_VALID_MYCONTEXT(cap))
		return (EINVAL);

	/* rail must be one of the rails in railmask */
	if (((1 << rail) & cap->cap_railmask) == 0)
		return (EINVAL);
	
	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	list_for_each(el, &elan_cap_list) {
		ELAN_CAP_NODE_STRUCT *cap_ptr = list_entry(el, ELAN_CAP_NODE_STRUCT , list);
		
		/* is it an exact match */
		if (ELAN_CAP_MATCH(&cap_ptr->node.cap,cap) && cap_ptr->node.active) {
			unsigned int attached_index = cap->cap_mycontext - cap->cap_lowcontext;
			
			if ( cap_ptr->attached[rail][attached_index].cb_func != NULL ) /* only one per ctx per rail */
			{
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return   EBUSY;
			}

			/* keep track of who attached as we might need to tell them when */
			/* cap or maps get destroyed                                     */
			cap_ptr->attached[rail][ attached_index ].cb_func = func;
			cap_ptr->attached[rail][ attached_index ].cb_args = args;
			cap_ptr->node.attached++;

			ELAN_DEBUG0(ELAN_DBG_CAP,"elan_attach_cap: passed\n");
			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
			return ESUCCESS;
		}
	}
	
	ELAN_DEBUG0(ELAN_DBG_CAP,"elan_attach_cap: failed to find \n");

	/* didnt find one */
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return EINVAL;
}

int 
elan_detach_cap(ELAN_CAPABILITY *cap, unsigned int rail)
{
	struct list_head *el, *nel;
	char              space[256];

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	ELAN_DEBUG1(ELAN_DBG_CAP,"elan_detach_cap %s\n",elan_capability_string(cap,space));
	list_for_each_safe (el, nel, &elan_cap_list) {
		ELAN_CAP_NODE_STRUCT *ptr = list_entry (el, ELAN_CAP_NODE_STRUCT, list);

		/* is it an exact match (key not checked) */
		if (ELAN_CAP_TYPE_MATCH(&ptr->node.cap,cap) &&
		    ELAN_CAP_GEOM_MATCH(&ptr->node.cap,cap) &&
		    (ptr->node.cap.cap_railmask & cap->cap_railmask) == cap->cap_railmask) {
		
			unsigned int attached_index = cap->cap_mycontext - cap->cap_lowcontext;

			if ( ptr->attached[rail][ attached_index ].cb_func == NULL ) {
				ELAN_DEBUG0(ELAN_DBG_CAP,"elanmod_detach_cap already removed \n");
				
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return  ESUCCESS;
			}

			ptr->attached[rail][ attached_index ].cb_func = NULL;
			ptr->attached[rail][ attached_index ].cb_args = (void *)0;

			ptr->node.attached--;
			
			ASSERT(ptr->node.attached >= 0);

			ELAN_DEBUG1(ELAN_DBG_CAP,"elanmod_detach_cap new attach count %d \n", ptr->node.attached);

			elan_destroy_cap_test(ptr);

			ELAN_DEBUG0(ELAN_DBG_CAP,"elan_detach_cap: success\n");	

			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
			return  ESUCCESS;
		}
	}

	ELAN_DEBUG0(ELAN_DBG_CAP,"elan_detach_cap: failed to find\n");
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  EINVAL;
}

int
elan_cap_dump()
{
	struct list_head        * tmp;
	ELAN_CAP_NODE_STRUCT * ptr = NULL;
	
	ELANMOD_RWLOCK_READ(&elan_rwlock);
	
	list_for_each(tmp, &elan_cap_list) {
		ptr = list_entry(tmp, ELAN_CAP_NODE_STRUCT , list);

		ELAN_DEBUG2 (ELAN_DBG_ALL, "cap dump: owner %p type %x\n", ptr->node.owner, ptr->node.cap.cap_type);
			
		ELAN_DEBUG5 (ELAN_DBG_ALL, "cap dump: LowNode %d   HighNode %d   LowContext %d   mycontext %d   highContext %d\n",
			     ptr->node.cap.cap_lownode , ptr->node.cap.cap_highnode,
			     ptr->node.cap.cap_lowcontext , ptr->node.cap.cap_mycontext, ptr->node.cap.cap_highcontext);

	}

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

int
elan_usercopy_attach(ELAN_CAPABILITY *cap, ELAN_CAP_NODE_STRUCT **node_ptr, void *handle, void *owner)
{
	struct list_head     *el;

	/* mycontext must be set and correct */
	if ( ! ELAN_CAP_VALID_MYCONTEXT(cap))
		return -EINVAL;
	
	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	/* Search all cap node structs looking for an exact match (including key) */
	list_for_each(el, &elan_cap_list) {
		ELAN_CAP_NODE_STRUCT *cap_ptr = list_entry(el, ELAN_CAP_NODE_STRUCT , list);
		
		/* is it an exact match */
		if (ELAN_CAP_MATCH(&cap_ptr->node.cap,cap) && cap_ptr->node.active) {
			char space[127];
			/* Work out which local process index we are */
			unsigned int attached_index = cap->cap_mycontext - cap->cap_lowcontext;

			ELAN_DEBUG(ELAN_DBG_CAP, "usercopy_attach: %s\n",
				   elan_capability_string(cap,space));

			ELAN_DEBUG(ELAN_DBG_CAP, 
				   "usercopy_attach: cap_ptr %p handle %p owner %p idx %d\n", 
				   cap_ptr, handle, owner, attached_index);
			
			/* Check we're not being called multiple times for the same local process */
			if (cap_ptr->attached[0][attached_index].handle)
			{
				ELAN_DEBUG(ELAN_DBG_CAP, 
					   "usercopy_attach: cap_ptr %p idx %d already attached handle %p owner %p\n",
					   cap_ptr, attached_index, 
					   cap_ptr->attached[0][attached_index].handle,
					   cap_ptr->attached[0][attached_index].owner);
					    
				ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
				return -EAGAIN;
			}

			/* Reference count node struct */
			cap_ptr->node.attached++;
			
			/* Stash our task handle/owner off the cap node array */
			cap_ptr->attached[0][attached_index].handle = handle;
			cap_ptr->attached[0][attached_index].owner = owner;
			
			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);

			/* Return node pointer to caller */
			*node_ptr = cap_ptr;

			return ESUCCESS;
		}
	}

	/* failed to match a cap */
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return -EINVAL;
}

int
elan_usercopy_detach(ELAN_CAP_NODE_STRUCT *cap_ptr, void *owner)
{
	int i;

	/* NB: The usercopy code holds a read lock on this rwlock and
	 * hence we will block here if exit_fs() gets called during a
	 * copy to this process
	 */
	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	/* Find this process in the attached task handle/owner array */
	for(i=0; i< ELAN_CAP_NUM_CONTEXTS((&cap_ptr->node.cap)); i++)
	{
		if (cap_ptr->attached[0][i].owner == owner)
		{
			ELAN_DEBUG(ELAN_DBG_CAP,
				   "usercopy_detach: cap_ptr %p handle %p owner %p id %d\n",
				   cap_ptr, cap_ptr->attached[0][i].handle, owner, i);

			/* Clear our task handle/owner off the cap node array */
			cap_ptr->attached[0][i].handle = NULL;
			cap_ptr->attached[0][i].owner  = NULL;
			
			/* Reference count node struct */
			cap_ptr->node.attached--;

			ASSERT(cap_ptr->node.attached >= 0);
			
			/* May need to destroy cap if reference count has hit zero */
			elan_destroy_cap_test(cap_ptr);

			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);

			return ESUCCESS;
		}
	}

	ELAN_DEBUG(ELAN_DBG_CAP, "usercopy_detach: cap_ptr %p[%d] failed owner %p\n", 
		   cap_ptr, cap_ptr->node.attached, owner);

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);

	return -EINVAL;
}

/* Returns the associated handle for the supplied ctxId process in the cap node */
/* Should be called holding a read lock on the elan_rwlock */
int
elan_usercopy_handle(ELAN_CAP_NODE_STRUCT *cap_ptr, int ctxId, void **handlep)
{
	int res = ESUCCESS;
	void *handle;

	/* Sanity check argument */
	if (ctxId < 0 || ctxId >= ELAN_CAP_NUM_CONTEXTS(&(cap_ptr->node.cap)))
		return -EINVAL;
	
//	ELANMOD_RWLOCK_READ(&elan_rwlock);

	/* Get the task handle for the remote process */
	if ((handle = cap_ptr->attached[0][ctxId].handle) == NULL)
		res = -EAGAIN;
	
//	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);

	*handlep = handle;

	return res;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
