/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: capability_general.c,v 1.10 2004/02/25 13:47:59 daniel Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/shared/capability_general.c,v $ */

#if defined(__KERNEL__)

#include <qsnet/kernel.h>

#else

#include <stdlib.h>
#include <stdio.h>
#include <sys/param.h>

#endif

#include <elan/elanmod.h>


void
elan_nullcap (ELAN_CAPABILITY *cap)
{
	register int i;

	for (i = 0; i < sizeof (cap->cap_userkey)/sizeof(cap->cap_userkey.key_values[0]); i++)
		cap->cap_userkey.key_values[i] = ELAN_CAP_UNINITIALISED;
    
	cap->cap_lowcontext  = ELAN_CAP_UNINITIALISED;
	cap->cap_highcontext = ELAN_CAP_UNINITIALISED;
	cap->cap_mycontext   = ELAN_CAP_UNINITIALISED;
	cap->cap_lownode     = ELAN_CAP_UNINITIALISED;
	cap->cap_highnode    = ELAN_CAP_UNINITIALISED;
	cap->cap_railmask    = ELAN_CAP_UNINITIALISED;
	cap->cap_type        = ELAN_CAP_UNINITIALISED;
	cap->cap_spare       = 0;
	cap->cap_version     = ELAN_CAP_VERSION_NUMBER;
	
	for (i = 0; i < sizeof (cap->cap_bitmap)/sizeof (cap->cap_bitmap[0]); i++)
		cap->cap_bitmap[i] = 0;
}

char *
elan_capability_string (ELAN_CAPABILITY *cap, char *str)
{
	if (cap == NULL) 
		sprintf (str, "[-.-.-.-] cap = NULL\n");
	else
		sprintf (str, "[%x.%x.%x.%x] Version %x Type %x \n"
			 "Context %x.%x.%x Node %x.%x\n",
			 cap->cap_userkey.key_values[0], cap->cap_userkey.key_values[1],
			 cap->cap_userkey.key_values[2], cap->cap_userkey.key_values[3],
			 cap->cap_version, cap->cap_type, 
			 cap->cap_lowcontext, cap->cap_mycontext, cap->cap_highcontext,
			 cap->cap_lownode, cap->cap_highnode);
	
	return (str);
}

ELAN_LOCATION
elan_vp2location (u_int process, ELAN_CAPABILITY *cap)
{
	ELAN_LOCATION location;
	int i, vp, node, context, nnodes, nctxs;

	vp = 0;

	location.loc_node    = ELAN_INVALID_NODE;
	location.loc_context = -1;
       
	nnodes = cap->cap_highnode - cap->cap_lownode + 1;
	nctxs  = cap->cap_highcontext - cap->cap_lowcontext + 1;
       
	switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_BLOCK:
		for (node = 0, i = 0; node < nnodes; node++)
		{
			for (context = 0; context < nctxs; context++)
			{
				if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, context + (node * nctxs)))
				{
					if (vp == process)
					{
						/* Return relative indices within the capability box */
						location.loc_node    = node;
						location.loc_context = context;

						return (location);
					}
		       
					vp++;
				}
			}
		}
		break;
	
	case ELAN_CAP_TYPE_CYCLIC:
		for (context = 0, i = 0; context < nctxs; context++)
		{
			for (node = 0; node < nnodes; node++)
			{
				if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, node + (context * nnodes)))
				{
					if (vp == process)
					{
						location.loc_node    = node;
						location.loc_context = context;

						return (location);
					}
		    
					vp++;
				}
			}
		}
		break;
	}
    
	return( location );
}

int
elan_location2vp (ELAN_LOCATION location, ELAN_CAPABILITY *cap)
{
    int  vp, node, context, nnodes, nctxs;

    nnodes = cap->cap_highnode - cap->cap_lownode + 1;
    nctxs  = cap->cap_highcontext - cap->cap_lowcontext + 1;

    vp = 0;
    
    switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
	for (node = 0 ; node < nnodes ; node++)
	{
	    for (context = 0; context < nctxs; context++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, context + (node * nctxs)))
		{
		    if ((location.loc_node == node) && (location.loc_context == context))
		    {
			/* Found it ! */
			return( vp );
		    }
		    
		    vp++;
		}
	    }
	}
	break;
	
    case ELAN_CAP_TYPE_CYCLIC:
	for (context = 0; context < nctxs; context++)
	{
	    for (node = 0; node < nnodes; node++)
	    {
		if ((cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP) || BT_TEST (cap->cap_bitmap, node + (context * nnodes)))
		{
		    if ((location.loc_node == node) && (location.loc_context == context))
		    {
			/* Found it ! */
			return( vp );
		    }
		    
		    vp++;
		}
	    }
	}
	break;
    }
    
    /* Failed to find it */
    return( -1 );
}

/* Return the number of processes as described by a capability */
int
elan_nvps (ELAN_CAPABILITY *cap)
{
	int i, c, nbits = ELAN_CAP_BITMAPSIZE(cap);

	if (cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP)
		return (nbits);

	for (i = 0, c = 0; i < nbits; i++)
		if (BT_TEST (cap->cap_bitmap, i))
			c++;

	return (c);
}

/* Return the number of local processes on a given node as described by a capability */
int
elan_nlocal (int node, ELAN_CAPABILITY *cap)
{
	int vp;
	ELAN_LOCATION loc;
	int nLocal = 0;

	for (vp = 0; vp < elan_nvps(cap); vp++)
	{
		loc = elan_vp2location(vp, cap);
		if (loc.loc_node == node)
			nLocal++;
	}

	return (nLocal);
}

/* Return the maximum number of local processes on any node as described by a capability */
int
elan_maxlocal (ELAN_CAPABILITY *cap)
{
	return(cap->cap_highcontext - cap->cap_lowcontext + 1);
}

/* Return the vps of the local processes on a given node as described by a capability */
int
elan_localvps (int node, ELAN_CAPABILITY *cap, int *vps, int size)
{
	int context;
	ELAN_LOCATION loc;
	int nLocal = 0;
    
	loc.loc_node = node;

	for (context = 0; context < MIN(size, elan_maxlocal(cap)); context++)
	{
		loc.loc_context = context;
	
		/* Should return -1 if none found */
		if ( (vps[context] = elan_location2vp( loc, cap )) != -1)
			nLocal++;
	}

	return (nLocal);
}

/* Return the number of rails that this capability utilises */
int
elan_nrails (ELAN_CAPABILITY *cap)
{
	int nrails = 0;
	unsigned int railmask;

	/* Test for a multi-rail capability */
	if (cap->cap_type & ELAN_CAP_TYPE_MULTI_RAIL)
	{
		/* Grab rail bitmask from capability */
		railmask = cap->cap_railmask;
	
		while (railmask)
		{
			if (railmask & 1)
				nrails++;
	    
			railmask >>= 1;
		}
	}
	else 
		/* Default to just one rail */
		nrails = 1;
	
	return (nrails);
}

/* Fill out an array giving the physical rail numbers utilised by a capability */
int
elan_rails (ELAN_CAPABILITY *cap, int *rails)
{
	int nrails, rail;
	unsigned int railmask;

	/* Test for a multi-rail capability */
	if (cap->cap_type & ELAN_CAP_TYPE_MULTI_RAIL)
	{
		/* Grab rail bitmask from capability */
		railmask = cap->cap_railmask;
	
		nrails = rail = 0;
		while (railmask)
		{
			if (railmask & 1)
				rails[nrails++] = rail;
	    
			rail++;
			railmask >>= 1;
		}
	}
	else
	{
		/* Default to just one rail */
		rails[0] = 0;
		nrails = 1;
	}

	return( nrails );
}

int 
elan_cap_overlap(ELAN_CAPABILITY *cap1, ELAN_CAPABILITY *cap2)
{
	/* by context */
	if ( cap1->cap_highcontext < cap2->cap_lowcontext ) return (0);
	if ( cap1->cap_lowcontext  > cap2->cap_highcontext) return (0);
	
	/* by node */
	if ( cap1->cap_highnode < cap2->cap_lownode ) return (0);
	if ( cap1->cap_lownode  > cap2->cap_highnode) return (0);

	/* by rail */
	/* they overlap if they have a rail in common */
	return (cap1->cap_railmask & cap2->cap_railmask);
}

#if !defined(__KERNEL__)

/* Fill out an array that hints at the best use of the rails on a
 * per process basis. The library user can then decide whether or not
 * to take this into account (e.g. TPORTs)
 * All processes calling this fn will be returned the same information.
 */
int
elan_prefrails(ELAN_CAPABILITY *cap, int *pref, int nvp)
{
	int i;
	int nrails = elan_nrails(cap);
	int maxlocal = elan_maxlocal(cap);

	/* Test for a multi-rail capability */
	if (! (cap->cap_type & ELAN_CAP_TYPE_MULTI_RAIL))
	{
		/* Default to just one rail */
		for (i = 0; i < nvp; i++)
			pref[i] = 0;

		return( 0 );
	}

	/*
	 * We allocate rails on a per node basis sharing our the rails
	 * equally amongst the local processes. However, if there is only
	 * one process per node and multiple rails, then we use a different
	 * algorithm where rails are allocated across all the processes in 
	 * a round-robin fashion
	 */
    
	if (maxlocal == 1)
	{
		/* Allocate rails in a round-robin manner */
		for (i = 0; i < nvp; i++)
			*pref++ = i % nrails;
	}
	else
	{
		int node;
		int *vps;
		int nnodes = cap->cap_highnode - cap->cap_lownode + 1;

		vps = (int *) malloc(sizeof(int)*maxlocal);

		/* Grab the local process info for each node and allocate
		 * rails to those vps on an equal basis
		 */
		for (node = 0; node < nnodes; node++)
		{
			int nlocal;
			int pprail;

			/* Grab an array of local vps */
			nlocal = elan_localvps(node, cap, vps, maxlocal);
	    
			/* Calculate the number processes per rail */
			if ((pprail = nlocal/nrails) == 0)
				pprail = 1;

			/* Allocate processes to rails */
			for (i = 0; i < nlocal; i++)
			{
				pref[vps[i]] = (i / pprail) % nrails;
			}
		}
	
		free(vps);
	}

	return( 0 );
}

void 
elan_get_random_key(ELAN_USERKEY *key)
{
    int i;
    for (i = 0; i < sizeof(key->key_values) / sizeof(key->key_values[0]); i++)
	key->key_values[i] = lrand48();
}

int elan_lowcontext(ELAN_CAPABILITY *cap)
{
    return(cap->cap_lowcontext);
}

int elan_mycontext(ELAN_CAPABILITY *cap)
{
    return(cap->cap_mycontext);
}

int elan_highcontext(ELAN_CAPABILITY *cap)
{
    return(cap->cap_highcontext);
}

int elan_lownode(ELAN_CAPABILITY *cap)
{
    return(cap->cap_lownode);
}

int elan_highnode(ELAN_CAPABILITY *cap)
{
    return(cap->cap_highnode);
}

int elan_captype(ELAN_CAPABILITY *cap)
{
    return(cap->cap_type);
}

int elan_railmask(ELAN_CAPABILITY *cap)
{
    return(cap->cap_railmask);
}

#endif

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
