/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: device.c,v 1.6 2005/04/13 09:31:47 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/device.c,v $*/

#include <qsnet/kernel.h>
#include <elan/elanmod.h>

static LIST_HEAD(elan_dev_list);

ELAN_DEV_STRUCT *
elan_dev_find (ELAN_DEV_IDX devidx)
{
	struct list_head   *tmp;
	ELAN_DEV_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_dev_list) {
		ptr = list_entry(tmp, ELAN_DEV_STRUCT , node);
		if (ptr->devidx == devidx) 
			return ptr;
		if (ptr->devidx > devidx)
			return ERR_PTR(-ENXIO);
	}
	
	return ERR_PTR(-EINVAL);
}

ELAN_DEV_STRUCT *
elan_dev_find_byrail (unsigned short deviceid, unsigned rail)
{
	struct list_head   *tmp;
	ELAN_DEV_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_dev_list) {
		ptr = list_entry(tmp, ELAN_DEV_STRUCT , node);

		ELAN_DEBUG5 (ELAN_DBG_ALL,"elan_dev_find_byrail devidx %d - %04x %04x,  %d %d \n", ptr->devidx, 
			     ptr->devinfo->dev_device_id, deviceid, ptr->devinfo->dev_rail, rail);

		if (ptr->devinfo->dev_device_id == deviceid && ptr->devinfo->dev_rail == rail)
			return ptr;
	}
	
	return NULL;
}

ELAN_DEV_IDX
elan_dev_register (ELAN_DEVINFO *devinfo, ELAN_DEV_OPS *ops, void * user_data)
{
	ELAN_DEV_STRUCT *ptr;
	ELAN_DEV_IDX        devidx = 0;
	struct list_head   *tmp;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	/* is it already registered */
	if ((ptr = elan_dev_find_byrail(devinfo->dev_device_id, devinfo->dev_rail)) != NULL) 
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return EINVAL;
	}

	/* find a free device idx */
	list_for_each (tmp, &elan_dev_list) {
		if (list_entry (tmp, ELAN_DEV_STRUCT, node)->devidx != devidx)
			break;
		devidx++;
	}

	/* create it and add */
	KMEM_ALLOC(ptr, ELAN_DEV_STRUCT *, sizeof(ELAN_DEV_STRUCT), 1);
	if (ptr == NULL)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return ENOMEM;
	}

	ptr->devidx    = devidx;
	ptr->ops       = ops;
	ptr->devinfo   = devinfo;
	ptr->user_data = user_data;

	/* insert this entry *before* the last entry we've found */
	list_add_tail(&ptr->node, tmp);

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

int
elan_dev_deregister (ELAN_DEVINFO *devinfo)
{
	ELAN_DEV_STRUCT *target;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	if ((target = elan_dev_find_byrail (devinfo->dev_device_id, devinfo->dev_rail)) == NULL)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  EINVAL;
	}

	list_del(&target->node);

	/* delete target entry */
	KMEM_FREE(target, sizeof(ELAN_DEV_STRUCT));

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

int
elan_dev_dump ()
{
	struct list_head   *tmp;
	ELAN_DEV_STRUCT *ptr=NULL;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	list_for_each(tmp, &elan_dev_list) {
		ptr = list_entry(tmp, ELAN_DEV_STRUCT , node);

		ELAN_DEBUG3 (ELAN_DBG_ALL,"dev dump: index %u rail %u elan%c\n", 
			     ptr->devidx, ptr->devinfo->dev_rail, '3' + ptr->devinfo->dev_device_id);
		ELAN_DEBUG5 (ELAN_DBG_ALL,"dev dump: Vid %x   Did %x  Rid %x  DR %d  DVal %x\n",
			     ptr->devinfo->dev_vendor_id,
			     ptr->devinfo->dev_device_id,
			     ptr->devinfo->dev_revision_id,
			     ptr->devinfo->dev_driver_version,
			     ptr->devinfo->dev_num_down_links_value);

	}

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return ESUCCESS;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
