/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */
#ident "@(#)$Id: elanmod.c,v 1.12 2005/04/13 09:31:47 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/elanmod.c,v $*/

#include <qsnet/kernel.h>
#include <elan/elanmod.h>

ELANMOD_RWLOCK elan_rwlock;

int 
elan_init()
{
	ELANMOD_RWLOCK_INIT(&elan_rwlock);
	return (ESUCCESS);
}

int 
elan_fini()
{
	ELANMOD_RWLOCK_DESTROY(&elan_rwlock);
	return (ESUCCESS);
}

int 
elanmod_classify_cap (ELAN_POSITION *position, ELAN_CAPABILITY *cap, unsigned use)
{
	if (cap->cap_version != ELAN_CAP_VERSION_NUMBER)
	{
		ELAN_DEBUG2 (ELAN_DBG_VP, "elanmod_classify_cap: (cap->Version != ELAN_CAP_VERSION) %d %d\n", cap->cap_version, ELAN_CAP_VERSION_NUMBER);
		return (-EINVAL);
	}
	
	if (cap->cap_lowcontext == ELAN_CAP_UNINITIALISED || cap->cap_highcontext == ELAN_CAP_UNINITIALISED)
	{
		ELAN_DEBUG3 (ELAN_DBG_VP, "elanmod_classify_cap: LowContext %d    HighContext %d MyContext %d\n",
			     cap->cap_lowcontext , cap->cap_highcontext, cap->cap_mycontext);
		return (-EINVAL);
	}
	
	if (cap->cap_lowcontext > cap->cap_highcontext)
	{
		ELAN_DEBUG2 (ELAN_DBG_VP, "elanmod_classify_cap: (cap->cap_lowcontext > cap->cap_highcontext) %d %d\n",cap->cap_lowcontext , cap->cap_highcontext);
		return (-EINVAL);
	}
	
	
	switch (cap->cap_type & ELAN_CAP_TYPE_MASK)
	{
	case ELAN_CAP_TYPE_BLOCK:
	case ELAN_CAP_TYPE_CYCLIC:
		if (position->pos_mode == ELAN_POS_UNKNOWN)
		{
			ELAN_DEBUG0 (ELAN_DBG_VP, "elanmod_classify_cap: Position Unknown \n");
			return (-EAGAIN);
		}
		
		if ( ! ( ELAN_USER_CONTEXT(cap->cap_lowcontext) && ELAN_USER_CONTEXT(cap->cap_highcontext)))
		{
			ELAN_DEBUG4 (ELAN_DBG_VP, "elanmod_classify_cap:  USER_BASE_CONTEXT %d %d %d %d \n" ,  ELAN_USER_BASE_CONTEXT_NUM,cap->cap_lowcontext, cap->cap_highcontext ,ELAN_USER_TOP_CONTEXT_NUM);
			return (-EINVAL);
		}
		if (cap->cap_lownode == ELAN_CAP_UNINITIALISED)
			cap->cap_lownode = position->pos_nodeid;
		if (cap->cap_highnode == ELAN_CAP_UNINITIALISED)
			cap->cap_highnode = position->pos_nodeid;
		
		if (cap->cap_lownode < 0 || cap->cap_highnode >= position->pos_nodes || cap->cap_lownode > cap->cap_highnode)
		{
			ELAN_DEBUG3 ( ELAN_DBG_VP,"elanmod_classify_cap: low %d high %d pos %d \n" , cap->cap_lownode  ,cap->cap_highnode, position->pos_nodes);
			
			return (-EINVAL);
		}
		
		if ((cap->cap_highnode < position->pos_nodeid) || (cap->cap_lownode > position->pos_nodeid))
		{
			ELAN_DEBUG3 (ELAN_DBG_VP, "elanmod_classify_cap: node not i range low %d high %d this %d\n",
				     cap->cap_lownode, cap->cap_highnode, position->pos_nodeid);
			return (-EINVAL);
		}

		break;
	default:
		ELAN_DEBUG1 (ELAN_DBG_VP, "elanmod_classify_cap: cant decode type %x \n", cap->cap_type & ELAN_CAP_TYPE_MASK);
		return (-EINVAL);

	}

	switch (use)
	{
	case ELAN_USER_ATTACH:
	case ELAN_USER_DETACH:
		if (cap->cap_mycontext == ELAN_CAP_UNINITIALISED)
		{
			ELAN_DEBUG0 (ELAN_DBG_VP, "elanmod_classify_cap: cap->cap_mycontext == ELAN_CAP_UNINITIALISED");
			return (-EINVAL);
		}
	
		if ((cap->cap_mycontext != ELAN_CAP_UNINITIALISED) && 
		    (cap->cap_mycontext < cap->cap_lowcontext || cap->cap_mycontext > cap->cap_highcontext))
		{
			ELAN_DEBUG3 (ELAN_DBG_VP, "elanmod_classify_cap: cap->cap_mycontext out of range %d %d %d \n", cap->cap_lowcontext,cap->cap_mycontext,cap->cap_highcontext);
			return (-EINVAL);
		}   
		break;

	case ELAN_USER_P2P:
		break;

	case ELAN_USER_BROADCAST:
		if (! (cap->cap_type & ELAN_CAP_TYPE_BROADCASTABLE)) {
			ELAN_DEBUG0 (ELAN_DBG_VP, "elanmod_classify_cap: use ELAN_USER_BROADCAST but cap not ELAN_CAP_TYPE_BROADCASTABLE\n");
			return (-EINVAL);
		}
		break;

	default:
		ELAN_DEBUG1 (ELAN_DBG_VP, "elanmod_classify_cap: unknown use (%d)\n",use);
		return (-EINVAL);
	}



	/* is any ctxt an rms one ?? */
	if (ELAN_RMS_CONTEXT(cap->cap_lowcontext) || ELAN_RMS_CONTEXT(cap->cap_highcontext))
	{
		/* so both low and high must be */
		if (!(ELAN_RMS_CONTEXT(cap->cap_lowcontext) && ELAN_RMS_CONTEXT(cap->cap_highcontext))) 
		{
			ELAN_DEBUG2 (ELAN_DBG_VP, "elanmod_classify_cap: not rms ctxt %x %x\n",cap->cap_lowcontext,cap->cap_highcontext );
			return (-EINVAL);
		}
		ELAN_DEBUG0 (ELAN_DBG_VP, "elanmod_classify_cap: returning ELAN_CAP_RMS\n");
		return (ELAN_CAP_RMS);
	}

	ELAN_DEBUG0 (ELAN_DBG_VP, "elanmod_classify_cap: returning ELAN_CAP_OK\n");
	return (ELAN_CAP_OK);
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
