/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elanmod_linux.c,v 1.23.2.6 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/elanmod_linux.c,v $*/

#include <qsnet/kernel.h>

#include <elan/elanmod.h>
#include <elan/elanmod_linux.h>

#include <qsnet/module.h>
#include <linux/sysctl.h>
#include <linux/init.h>

#include <qsnet/procfs_linux.h>

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("Elan support module");

MODULE_LICENSE("GPL");

/* elanmod.c */
EXPORT_SYMBOL(elanmod_classify_cap);

/* bitmap.c */
#include <elan/bitmap.h>

EXPORT_SYMBOL(bt_freebit);
EXPORT_SYMBOL(bt_lowbit); 
EXPORT_SYMBOL(bt_nextbit);
EXPORT_SYMBOL(bt_copy);
EXPORT_SYMBOL(bt_zero); 
EXPORT_SYMBOL(bt_fill); 
EXPORT_SYMBOL(bt_cmp); 
EXPORT_SYMBOL(bt_intersect);
EXPORT_SYMBOL(bt_remove); 
EXPORT_SYMBOL(bt_add); 
EXPORT_SYMBOL(bt_spans);
EXPORT_SYMBOL(bt_subset);  
EXPORT_SYMBOL(bt_up);
EXPORT_SYMBOL(bt_down);
EXPORT_SYMBOL(bt_nbits);

/* capability.c */
EXPORT_SYMBOL(elan_nullcap);
EXPORT_SYMBOL(elan_detach_cap);
EXPORT_SYMBOL(elan_attach_cap);
EXPORT_SYMBOL(elan_validate_map);

/* stats.c */
EXPORT_SYMBOL(elan_stats_register);
EXPORT_SYMBOL(elan_stats_deregister);

/* device.c */
EXPORT_SYMBOL(elan_dev_deregister);
EXPORT_SYMBOL(elan_dev_register);

/* debug */
int  elan_debug_mode = QSNET_DEBUG_BUFFER; 
int  elan_debug_mask;

static struct proc_dir_entry *elan_procfs_root;

extern void elan_procfs_init(void);
extern void elan_procfs_fini(void);

static int elan_open    (struct inode *ino, struct file *fp);
static int elan_release (struct inode *ino, struct file *fp);
static int elan_ioctl   (struct inode *ino, struct file *fp, unsigned int cmd, unsigned long arg);

static int elan_user_open    (struct inode *ino, struct file *fp);
static int elan_user_release (struct inode *ino, struct file *fp);
static int elan_user_ioctl (struct inode *ino, struct file *fp, unsigned int cmd, unsigned long arg);

static struct file_operations elan_fops = 
{
	ioctl:	 elan_ioctl,
	open:	 elan_open,
	release: elan_release,
};

static struct file_operations elan_user_fops = 
{
	ioctl:	 elan_user_ioctl,
	open:	 elan_user_open,
	release: elan_user_release,
};

static int __init elan_start(void)
{
	int res;

	elan_procfs_init(); 

	if ((res = elan_init()) != ESUCCESS)
	{
		elan_procfs_fini();
		return (-res);
	}

	return (0);
}

static void __exit elan_exit(void)
{
	elan_fini();
	elan_procfs_fini();
}


/* Declare the module init and exit functions */
void
elan_procfs_init()
{
	struct proc_dir_entry  *p;
	
	elan_procfs_root = proc_mkdir("elan",   qsnet_procfs_root);

	if (elan_procfs_root == NULL)
		return;
	
	qsnet_proc_register_hex(elan_procfs_root, "debug_mask", &elan_debug_mask, 0);
	qsnet_proc_register_hex(elan_procfs_root, "debug_mode", &elan_debug_mode, 0);

	if ((p = create_proc_entry ("ioctl", 0, elan_procfs_root)) != NULL)
	{
		p->proc_fops = &elan_fops;
		p->data      = 0;
		p->owner     = THIS_MODULE;
	}   

	/* user entry point */
	if ((p = create_proc_entry ("user", 0, elan_procfs_root)) != NULL)
	{
		p->proc_fops = &elan_user_fops;
		p->data      = 0;
		p->owner     = THIS_MODULE;
	}   
}

void
elan_procfs_fini()
{
	if (elan_procfs_root == NULL)
		return;

	remove_proc_entry ("debug_mask", elan_procfs_root);
	remove_proc_entry ("debug_mode", elan_procfs_root);
	
	remove_proc_entry ("ioctl",   elan_procfs_root); 

	/* remove user entry point */
	remove_proc_entry ("user",   elan_procfs_root); 
	
	remove_proc_entry ("elan",   qsnet_procfs_root);
}

module_init(elan_start);
module_exit(elan_exit);

static int
elan_open (struct inode *inode, struct file *fp)
{
	MOD_INC_USE_COUNT;
	fp->private_data = NULL;
	return (0);
}

static int
elan_release (struct inode *inode, struct file *fp)
{
	/* mark all caps owned by fp to be destroyed */
	elan_destroy_cap(fp,NULL);

	MOD_DEC_USE_COUNT;
	return (0);
}

static int 
elan_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
	int rep = 0;

	switch (cmd) 
	{
	case ELANCTRL_STATS_GET_NEXT :
	{
		ELANCTRL_STATS_GET_NEXT_STRUCT args;

		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_GET_NEXT_STRUCT)))
			return (-EFAULT); 

		/* uses copyin/copyout */
		if (elan_stats_get_next_index(args.statidx, args.next_statidx) != 0 ) 
			return (-EINVAL);	

		break;
	}
	case ELANCTRL_STATS_FIND_INDEX :
	{
		ELANCTRL_STATS_FIND_INDEX_STRUCT args;
		char block_name[ELAN_STATS_NAME_MAX_LEN+1];
		int res;

		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_FIND_INDEX_STRUCT)))
			return (-EFAULT); 

		res = strncpy_from_user (block_name, args.block_name, sizeof (block_name));

		if (res == 0 || res  == sizeof (block_name))
			return -ERANGE;
		if (res < 0)
			return res;

		/* uses copyin/copyout */
		if (elan_stats_find_index(block_name, args.statidx, args.num_entries) != 0 ) 
			return (-EINVAL);	

		break;
	}
	case ELANCTRL_STATS_GET_BLOCK_INFO :
	{
		ELANCTRL_STATS_GET_BLOCK_INFO_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_GET_BLOCK_INFO_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_stats_get_block_info(args.statidx, args.block_name, args.num_entries) != 0 ) 
			return (-EINVAL);
		break;		
	}
	case ELANCTRL_STATS_GET_INDEX_NAME :
	{
		ELANCTRL_STATS_GET_INDEX_NAME_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_GET_INDEX_NAME_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_stats_get_index_name(args.statidx, args.index, args.name) != 0 )
			return (-EINVAL);
		break;
	}
	case ELANCTRL_STATS_CLEAR_BLOCK :
	{
		ELANCTRL_STATS_CLEAR_BLOCK_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_CLEAR_BLOCK_STRUCT)))
			return (-EFAULT);

		/* statidx is not a pointer */
		if (elan_stats_clear_block(args.statidx) != 0 )
			return (-EINVAL);
		break;
	}
	case ELANCTRL_STATS_GET_BLOCK :
	{
		ELANCTRL_STATS_GET_BLOCK_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_STATS_GET_BLOCK_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_stats_get_block(args.statidx, args.entries, args.values) != 0 )
			return (-EINVAL);
		break;
	}
	case ELANCTRL_GET_DEVINFO :
	{
		ELANCTRL_GET_DEVINFO_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_GET_DEVINFO_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_get_devinfo(args.devidx, args.devinfo) != 0 )
			return (-EINVAL);
		break;		
	}
	case ELANCTRL_GET_POSITION :
	{
		ELANCTRL_GET_POSITION_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_GET_POSITION_STRUCT)))
			return (-EFAULT); 

		/* uses copyin/copyout */
		if (elan_get_position(args.devidx, args.position) != 0 )
			return (-EINVAL);
		break;		
	}
	case ELANCTRL_SET_POSITION :
	{
		ELANCTRL_SET_POSITION_STRUCT args;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_SET_POSITION_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_set_position(args.devidx, args.nodeId, args.numNodes) != 0 )
			return (-EINVAL);	
		break;		
	}
	case ELANCTRL_CREATE_CAP  :
	{
		ELANCTRL_CREATE_CAP_STRUCT *args;

		/* get space for args */
		KMEM_ALLOC(args, ELANCTRL_CREATE_CAP_STRUCT *, sizeof(ELANCTRL_CREATE_CAP_STRUCT), 1);
		if (args == NULL)
			return(-ENOMEM);	

		/* copy them */
		if (copy_from_user (args, (void *) arg, sizeof (ELANCTRL_CREATE_CAP_STRUCT)))
			return (-EFAULT);
		else 
		{
			if (((rep = elan_validate_cap(&args->cap)) != 0) || ((rep = elan_create_cap(fp,&args->cap)) != 0)) 
				rep = (-rep);
		}

		/* free the space */
		KMEM_FREE(args, sizeof(ELANCTRL_CREATE_CAP_STRUCT));

		break;		
	}
	case ELANCTRL_DESTROY_CAP  :
	{
		ELANCTRL_DESTROY_CAP_STRUCT *args;

		/* get space for args */
		KMEM_ALLOC(args, ELANCTRL_DESTROY_CAP_STRUCT *, sizeof(ELANCTRL_DESTROY_CAP_STRUCT), 1);
		if (args == NULL)
			return(-ENOMEM);	

		/* copy them */
		if (copy_from_user (args, (void *) arg, sizeof (ELANCTRL_DESTROY_CAP_STRUCT)))
			rep = (-EFAULT);
		else 
		{
			if (elan_destroy_cap(fp, &args->cap) != 0 )
				rep = (-EINVAL);
		}

		/* free the space */
		KMEM_FREE(args, sizeof(ELANCTRL_DESTROY_CAP_STRUCT));

		break;		
	}
	case ELANCTRL_CREATE_VP  :
	{
		ELANCTRL_CREATE_VP_STRUCT *args;

		/* get space for args */
		KMEM_ALLOC(args, ELANCTRL_CREATE_VP_STRUCT *, sizeof(ELANCTRL_CREATE_VP_STRUCT), 1);
		if (args == NULL)
			return(-ENOMEM);	

		/* copy them */
		if (copy_from_user (args, (void *) arg, sizeof (ELANCTRL_CREATE_VP_STRUCT)))
			return (-EFAULT);
		else
		{
			if ((elan_validate_cap( &args->map) != 0) || (elan_create_vp(fp, &args->cap, &args->map) != 0 ))
				rep = (-EINVAL);	
		}

		KMEM_FREE(args, sizeof(ELANCTRL_CREATE_VP_STRUCT ));

		break;		
	}
	case ELANCTRL_DESTROY_VP  :
	{
		ELANCTRL_DESTROY_VP_STRUCT *args;

		/* get space for args */
		KMEM_ALLOC(args, ELANCTRL_DESTROY_VP_STRUCT *, sizeof(ELANCTRL_DESTROY_VP_STRUCT), 1);
		if (args == NULL)
			return(-ENOMEM);	
		
		/* copy them */
		if (copy_from_user (args, (void *) arg, sizeof (ELANCTRL_DESTROY_VP_STRUCT)))
		 	rep = (-EFAULT);
		else 
		{
			if (elan_destroy_vp(fp, &args->cap, &args->map) != 0 )
				rep = (-EINVAL);	
		}

		KMEM_FREE(args, sizeof(ELANCTRL_DESTROY_VP_STRUCT ));

		break;		
	}

	case ELANCTRL_GET_CAPS  :
	{
		ELANCTRL_GET_CAPS_STRUCT args;
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_GET_CAPS_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if (elan_get_caps(args.number_of_results, args.array_size, args.caps) != 0 )
			return (-EINVAL);
		break;		
	}
	case ELANCTRL_DEBUG_DUMP :
	{
		elan_cap_dump();
		elan_dev_dump();

		break;
	}
	case ELANCTRL_DEBUG_BUFFER :
	{
		ELANCTRL_DEBUG_BUFFER_STRUCT args;

		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_DEBUG_BUFFER_STRUCT)))
			return (-EFAULT);

		/* uses copyin/copyout */
		if ((args.size = qsnet_debug_buffer (args.buffer, args.size)) != -1 &&
		    copy_to_user ((void *) arg, &args, sizeof (ELANCTRL_DEBUG_BUFFER_STRUCT)))
			return (-EFAULT);
		break;
	}
	default:
		return (-EINVAL);
		break;
	}

	return (rep);
}


static int
elan_user_open (struct inode *inode, struct file *fp)
{
	MOD_INC_USE_COUNT;
	fp->private_data = NULL;
	return (0);
}

static int
elan_user_release (struct inode *inode, struct file *fp)
{
	struct elan_cap_node_struct *cap_ptr = (struct elan_cap_node_struct *)fp->private_data;

	if (cap_ptr) {
		/* Remove this process from usercopy system */
		/* GNAT 7498: New to pass in a common owner pointer */
		if (elan_usercopy_detach (cap_ptr, fp) == 0)
			fp->private_data = NULL;
	}
	
	MOD_DEC_USE_COUNT;
	return (0);
}

static int 
elan_user_ioctl (struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
	int rep = 0;
#if !defined(NO_MMPUT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
	struct elan_cap_node_struct *cap_ptr = (struct elan_cap_node_struct *)fp->private_data;
#endif

	switch (cmd) 
	{
#if !defined(NO_MMPUT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
	case ELANCTRL_USERCOPY_ATTACH:
	{
		ELANCTRL_USERCOPY_ATTACH_STRUCT args;
		
		/* Are we already attached ? */
		if (cap_ptr != NULL)
			return -EAGAIN;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_USERCOPY_ATTACH_STRUCT)))
			return -EFAULT; 
	
		/* Lookup the associated cap node and can check we are allowed to
		 * access it using the supplied capability. If allowed, then associate 
		 * our task with that cap node
		 * We also reference count it as we then hang it off the file pointer
		 */
		/* GNAT 7498: New to pass in a common owner pointer */
		if ((rep = elan_usercopy_attach(&args.cap, &cap_ptr, current, fp)) < 0)
			return -EAGAIN;
		
		/* Hang cap node off file pointer for future usercopy ioctls */
		fp->private_data = (void *) cap_ptr;

		break;	
	}
	case ELANCTRL_USERCOPY_DETACH:
	{
		/* Detach process */
		if (cap_ptr) {
			/* Remove this process from usercopy system */
			/* GNAT 7498: New to pass in a common owner pointer */
			if ((rep = elan_usercopy_detach (cap_ptr, fp)) == 0)
				fp->private_data = NULL;
		}
		else
			rep = -EINVAL;
		
		break;	
	}
	case ELANCTRL_USERCOPY:
	{
		ELANCTRL_USERCOPY_STRUCT args;
		
		/* Check that we have previously successfully attached */
		if (cap_ptr == NULL)
			return -EAGAIN;
		
		if (copy_from_user (&args, (void *) arg, sizeof (ELANCTRL_USERCOPY_STRUCT)))
			return (-EFAULT); 
		
		/* Perform user-to-user copy */
		rep = elan_usercopy(args.remote, args.local, args.len, args.write, args.ctxId, cap_ptr);

		break;
	}
#endif /* !defined(NO_MMPUT) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9) */
	default:
		return (-EINVAL);
		break;
	}

	return (rep);
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
