/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: stats.c,v 1.7 2005/04/13 09:31:47 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/stats.c,v $*/

#include <qsnet/kernel.h>
#include <elan/elanmod.h>

static LIST_HEAD(elan_stats_list);
static ELAN_STATS_IDX elan_next_statidx=0;

ELAN_STATS_STRUCT *
elan_stats_find(ELAN_STATS_IDX statidx)
{
	struct list_head     *tmp;
	ELAN_STATS_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_stats_list) {
		ptr = list_entry(tmp, ELAN_STATS_STRUCT , node);
		if ( ptr->statidx == statidx ) 
			return ptr;
	}

	ELAN_DEBUG1 (ELAN_DBG_CTRL, "elan_stats_find failed %d\n", statidx);	
	return NULL;
}

ELAN_STATS_STRUCT *
elan_stats_find_by_name(caddr_t block_name)
{
	struct list_head     *tmp;
	ELAN_STATS_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_stats_list)	{
		ptr = list_entry(tmp, ELAN_STATS_STRUCT , node);
		if (!strcmp(ptr->block_name, block_name)) 
		{
			ELAN_DEBUG3 (ELAN_DBG_CTRL, "elan_stats_find_by_name found %s (%d,%d)\n", block_name, ptr->statidx, ptr->num_entries);	
			return ptr;
		}
	}

	ELAN_DEBUG1 (ELAN_DBG_CTRL, "elan_stats_find_by_name failed %s\n", block_name);
	return NULL;
}

ELAN_STATS_STRUCT *
elan_stats_find_next(ELAN_STATS_IDX statidx)
{
	struct list_head     *tmp;
	ELAN_STATS_STRUCT *ptr=NULL;

	list_for_each(tmp, &elan_stats_list) {
		ptr = list_entry(tmp, ELAN_STATS_STRUCT , node);
	  
		if ( ptr->statidx > statidx ) 
			return ptr;	  
	}	

	return NULL;
}

int 
elan_stats_get_next_index (ELAN_STATS_IDX statidx, ELAN_STATS_IDX *next_block)
{
	ELAN_STATS_STRUCT *target;
	ELAN_STATS_IDX        next = 0;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	if ((target = elan_stats_find_next(statidx)) != NULL)
		next = target->statidx;

	copyout(&next, next_block, sizeof(ELAN_STATS_IDX) );

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return 0;
}

int 
elan_stats_find_index  (caddr_t  block_name, ELAN_STATS_IDX *statidx,  uint *num_entries)

{
	ELAN_STATS_STRUCT *target;
	ELAN_STATS_IDX        index   = 0;
	uint                  entries = 0;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	ELAN_DEBUG1(ELAN_DBG_CTRL, "elan_stats_find_index %s \n", block_name);

	if ((target = elan_stats_find_by_name(block_name)) != NULL)
	{
		index   = target->statidx;
		entries = target->num_entries;
	}

	ELAN_DEBUG3(ELAN_DBG_CTRL, "elan_stats_find_index found %d %d (target=%p)\n", index, entries, target);

	copyout(&index,   statidx,     sizeof(ELAN_STATS_IDX));
	copyout(&entries, num_entries, sizeof(uint));

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return  ESUCCESS;
}

int 
elan_stats_get_block_info (ELAN_STATS_IDX statidx, caddr_t  block_name, uint *num_entries)
{
	ELAN_STATS_STRUCT *target;
	int                   res=EINVAL;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	ELAN_DEBUG1(ELAN_DBG_CTRL, "elan_stats_get_block_info statidx %d\n",statidx);

	if ((target = elan_stats_find(statidx)) != NULL)
	{
		ELAN_DEBUG2(ELAN_DBG_CTRL, "elan_stats_get_block_info name %s entries %d\n",block_name, *num_entries);
		
		copyout( target->block_name,  block_name,  ELAN_STATS_NAME_MAX_LEN);
		copyout(&target->num_entries, num_entries, sizeof(uint));

		res = ESUCCESS;
	}

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return res;
}

int 
elan_stats_get_index_name (ELAN_STATS_IDX statidx, uint index, caddr_t name)
{
	ELAN_STATS_STRUCT *target;
	int                   res=EINVAL;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	ELAN_DEBUG2(ELAN_DBG_CTRL, "elan_stats_get_index_name statidx %d index %d\n",statidx, index);

	if ((target = elan_stats_find(statidx)) != NULL)
	{
		if ( target->ops->elan_stats_get_name== NULL) 
		{
			ELAN_DEBUG0(ELAN_DBG_CTRL, "elan_stats_get_index_name no callback\n");	
			ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
			return  res;
		}

		if ((res = target->ops->elan_stats_get_name(target->arg, index, name)) == 0)
			ELAN_DEBUG1(ELAN_DBG_CTRL, "elan_stats_get_index_name name %s\n",name);	

	}
	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return  res;
}

int 
elan_stats_get_block (ELAN_STATS_IDX statidx, uint entries, ulong *values)
{
	ELAN_STATS_STRUCT *target;
	int                   res=EINVAL;

	ELANMOD_RWLOCK_READ(&elan_rwlock);

	
	if ((target = elan_stats_find(statidx)) != NULL)
	{
		if ( target->ops->elan_stats_get_block == NULL) 
		{
			ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
			return  res;
		}

		res = target->ops->elan_stats_get_block(target->arg, entries, values);
	}

	ELANMOD_RWLOCK_READ_UNLOCK(&elan_rwlock);
	return  res;
}

int 
elan_stats_clear_block (ELAN_STATS_IDX statidx)
{
	ELAN_STATS_STRUCT *target;
	int                   res=EINVAL;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	if ((target = elan_stats_find(statidx)) != NULL)
	{
		if ( target->ops->elan_stats_clear_block == NULL) 
		{
			ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
			return  res;
		}
	
		res = target->ops->elan_stats_clear_block(target->arg);
	}
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  res;
}

void
elan_stats_next_statidx(void)
{
	/* XXXXX need to put not in use check here incase we loop MRH */
	/* tho its a bigish loop :)                                   */
	elan_next_statidx++;
	if (!elan_next_statidx)
		elan_next_statidx++;
}

int 
elan_stats_register (ELAN_STATS_IDX    *statidx, 
			char              *block_name, 
			uint               num_entries,
			ELAN_STATS_OPS *ops,
			void              *arg)
{
	ELAN_STATS_STRUCT *target;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);

	/* create it and add */
	KMEM_ALLOC(target, ELAN_STATS_STRUCT *, sizeof(ELAN_STATS_STRUCT), 1);
	if (target == NULL)
	{
		ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
		return  ENOMEM;
	}

	elan_stats_next_statidx();

	*statidx = elan_next_statidx;

	target->statidx     = elan_next_statidx;
	target->num_entries = num_entries;
	target->ops         = ops;
	target->arg         = arg;
	strcpy(target->block_name, block_name);
	
	list_add_tail(&target->node, &elan_stats_list);

	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);
	return  0;
}

int
elan_stats_deregister (ELAN_STATS_IDX statidx)
{
	ELAN_STATS_STRUCT *target;

	ELANMOD_RWLOCK_WRITE(&elan_rwlock);
	if ((target = elan_stats_find(statidx)) != NULL)
	{

		list_del(&target->node);
		
		/* delete target entry */
		KMEM_FREE(target, sizeof(ELAN_STATS_STRUCT));
	}
	ELANMOD_RWLOCK_WRITE_UNLOCK(&elan_rwlock);

	return  target == NULL ? EINVAL : 0;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
