/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: context.c,v 1.117.2.1 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/context.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>
#include <elan/elanmod.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elan3mmu.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/vmseg.h>
#include <elan3/elan3ops.h>
#include <elan3/elansyscall.h>
/*
 * Global variables configurable from /etc/system file
 *     (OR /etc/sysconfigtab on Digital UNIX)
 */
int ntrapped_threads   = 64;
int ntrapped_dmas      = 64;
int ntrapped_events    = E3_NonSysCntxQueueSize + 128;
int ntrapped_commands  = 64;
int noverflow_commands = 1024;
int nswapped_threads   = 64;
int nswapped_dmas      = 64;

#define NUM_HALTOPS	8

void *SwapListsLockInfo;
void *CmdLockInfo;

static void HaltSwapContext (ELAN3_DEV *dev, void *arg);

static char *OthersStateStrings[]  = {"others_running", "others_halting", "others_swapping", 
				      "others_halting_more", "others_swapping_more", "others_swapped"};

ELAN3_CTXT *
elan3_alloc (ELAN3_DEV *dev, int  kernel)
{
    ELAN3_CTXT    *ctxt;
    int           i;
    unsigned long flags;

    PRINTF1 (DBG_DEVICE, DBG_FN, "elan3_alloc: %s\n", kernel ? "kernel" : "user");

    KMEM_ZALLOC (ctxt, ELAN3_CTXT *, sizeof (ELAN3_CTXT), TRUE);
    
    if (ctxt == NULL)
	return (NULL);

    elan_nullcap (&ctxt->Capability);

    ctxt->Device      = dev;
    ctxt->OthersState = CTXT_OTHERS_SWAPPED;
    ctxt->RefCnt      = 1;
    ctxt->Position    = dev->Position;

    if (kernel)
	ctxt->Status = CTXT_DETACHED | CTXT_SWAPPED_OUT | CTXT_KERNEL;
    else
	ctxt->Status = CTXT_DETACHED | CTXT_SWAPPED_OUT | CTXT_NO_LWPS;

    ctxt->Elan3mmu = elan3mmu_alloc (ctxt);

    kcondvar_init (&ctxt->Wait);
    kcondvar_init (&ctxt->CommandPortWait);
    kcondvar_init (&ctxt->LwpWait);
    kcondvar_init (&ctxt->HaltWait);

    spin_lock_init (&ctxt->InputFaultLock);

    kmutex_init (&ctxt->SwapListsLock);
    kmutex_init (&ctxt->CmdPortLock);
    kmutex_init (&ctxt->NetworkErrorLock);
    kmutex_init (&ctxt->CmdLock);

    krwlock_init (&ctxt->VpLock);

    KMEM_GETPAGES (ctxt->FlagPage, ELAN3_FLAGSTATS *, 1, TRUE);
    if (!ctxt->FlagPage)
	goto error;
    bzero ((char *) ctxt->FlagPage, PAGESIZE);

    KMEM_ZALLOC (ctxt->CommandTraps, COMMAND_TRAP *,    sizeof (COMMAND_TRAP)    * ntrapped_commands, TRUE);
    if (!ctxt->CommandTraps)
	goto error;

    KMEM_ZALLOC (ctxt->ThreadTraps,  THREAD_TRAP *,     sizeof (THREAD_TRAP)     * ntrapped_threads,  TRUE);
    if (!ctxt->ThreadTraps)
	goto error;

    KMEM_ZALLOC (ctxt->DmaTraps,     DMA_TRAP *,        sizeof (DMA_TRAP)        * ntrapped_dmas,     TRUE);
    if (!ctxt->DmaTraps)
	goto error;

    KMEM_ZALLOC (ctxt->EventCookies, EVENT_COOKIE *,    sizeof (EVENT_COOKIE)    * ntrapped_events,   TRUE);
    if (!ctxt->EventCookies)
	goto error;

    KMEM_ZALLOC (ctxt->Commands,     CProcTrapBuf_BE *, sizeof (CProcTrapBuf_BE) * noverflow_commands,TRUE);
    if (!ctxt->Commands)
	goto error;

    KMEM_ZALLOC (ctxt->SwapThreads,  E3_Addr *,         sizeof (E3_Addr)         * nswapped_threads,  TRUE);
    if (!ctxt->SwapThreads)
	goto error;

    KMEM_ZALLOC (ctxt->SwapDmas,     E3_DMA_BE *,       sizeof (E3_DMA_BE)       * nswapped_dmas,     TRUE);
    if (!ctxt->SwapDmas)
	goto error;

    /*
     * "slop" is defined as follows :
     *     number of entries REQUIRED to be left spare to consume all other traps
     *     up until the time that the context can be swapped out.
     *  
     * CommandTrapQ : 1 command issued by main + 1 issued by the thread processor per elan
     * ThreadTrapQ  : 2 from command + 2 input
     * DmaTrapQ     : 2 from command + 2 input
     * EventTrapQ   : 2 from command + 1 thread + 1 dma + 2 input + E3_NonSysCntxQueueSize
     */
    spin_lock_irqsave (&dev->IntrLock, flags);
    ELAN3_QUEUE_INIT (ctxt->CommandTrapQ, ntrapped_commands,  2);
    ELAN3_QUEUE_INIT (ctxt->ThreadTrapQ,  ntrapped_threads,   4);
    ELAN3_QUEUE_INIT (ctxt->DmaTrapQ,     ntrapped_dmas,      4);
    ELAN3_QUEUE_INIT (ctxt->EventCookieQ, ntrapped_events,    MIN(E3_NonSysCntxQueueSize + 6, ntrapped_events - 6));
    ELAN3_QUEUE_INIT (ctxt->CommandQ,     noverflow_commands, 0);
    ELAN3_QUEUE_INIT (ctxt->SwapThreadQ,  nswapped_threads,   0);
    ELAN3_QUEUE_INIT (ctxt->SwapDmaQ,     nswapped_dmas,      0);
    spin_unlock_irqrestore (&dev->IntrLock, flags);

#if defined(DIGITAL_UNIX)
    /* Allocate the segelan for the command port */
    if (! kernel && elan3_segelan3_create (ctxt) == NULL)
    {
	elan3_detach(ctxt);
	elan3_free (ctxt);
	return ((ELAN3_CTXT *) NULL);
    }
#endif

    /*
     * Initialise the Input Fault list 
     */
    spin_lock (&ctxt->InputFaultLock);
    for (i = 0; i < NUM_INPUT_FAULT_SAVE; i++)
	ctxt->InputFaults[i].Next = (i == (NUM_INPUT_FAULT_SAVE-1)) ? NULL : &ctxt->InputFaults[i+1];
    ctxt->InputFaultList = &ctxt->InputFaults[0];
    spin_unlock (&ctxt->InputFaultLock);

    ReserveHaltOperations (dev, NUM_HALTOPS, TRUE);
    
    if ((ctxt->RouteTable = AllocateRouteTable (ctxt->Device, ELAN3_MAX_VPS)) == NULL)
    {
	PRINTF0 (DBG_DEVICE, DBG_FN, "elan3_alloc: cannot map route table\n");
	elan3_detach(ctxt);
	elan3_free (ctxt);
	return ((ELAN3_CTXT *) NULL);
    }  

    return (ctxt);


 error:

    elan3_detach(ctxt);
    elan3_free (ctxt);
    if (ctxt->FlagPage)
	KMEM_FREEPAGES ((void *) ctxt->FlagPage, 1);
    if (ctxt->CommandTraps)
	KMEM_FREE ((void *) ctxt->CommandTraps, sizeof (COMMAND_TRAP)    * ntrapped_commands);
    if (ctxt->ThreadTraps)
	KMEM_FREE ((void *) ctxt->ThreadTraps,  sizeof (THREAD_TRAP)     * ntrapped_threads);
    if (ctxt->DmaTraps)
	KMEM_FREE ((void *) ctxt->DmaTraps,     sizeof (DMA_TRAP)        * ntrapped_dmas);
    if (ctxt->EventCookies)
	KMEM_FREE ((void *) ctxt->EventCookies, sizeof (EVENT_COOKIE)    * ntrapped_events);
    if (ctxt->Commands)
	KMEM_FREE ((void *) ctxt->Commands,     sizeof (CProcTrapBuf_BE) * noverflow_commands);
    if (ctxt->SwapThreads)
	KMEM_FREE ((void *) ctxt->SwapThreads,  sizeof (E3_Addr)         * nswapped_threads);
    if (ctxt->SwapDmas)
	KMEM_FREE ((void *) ctxt->SwapDmas,     sizeof (E3_DMA_BE)       * nswapped_dmas);

    kcondvar_destroy (&ctxt->Wait);
    kcondvar_destroy (&ctxt->CommandPortWait);
    kcondvar_destroy (&ctxt->LwpWait);
    kcondvar_destroy (&ctxt->HaltWait);

    kmutex_destroy (&ctxt->SwapListsLock);
    kmutex_destroy (&ctxt->CmdLock);
    kmutex_destroy (&ctxt->NetworkErrorLock);
    spin_lock_destroy  (&ctxt->InputFaultLock);

    krwlock_destroy (&ctxt->VpLock);

    KMEM_FREE (ctxt, sizeof (ELAN3_CTXT));

    return (NULL);
}

void
elan3_free (ELAN3_CTXT *ctxt)
{
    ELAN3_DEV     *dev = ctxt->Device;
    NETERR_FIXUP *nef;
    
    PRINTF1 (ctxt, DBG_FN, "elan3_free: %p \n", ctxt);
   
    elan3_removevp (ctxt, ELAN3_INVALID_PROCESS);			/* Remove any virtual process mappings */

#if defined(DIGITAL_UNIX)
    WaitForContext (ctxt);					/* wait for all references to this context to go away */
#endif

    if (ctxt->RouteTable)
	FreeRouteTable (dev, ctxt->RouteTable);
    ctxt->RouteTable = NULL;

    elan3mmu_free (ctxt->Elan3mmu);				/* free of our Elan3mmu  */

    if (ctxt->Private)						/* Call back to "user" to free off  */
	ELAN3_OP_FREE_PRIVATE (ctxt);				/* private data */

#if defined(DIGITAL_UNIX)
    if (! CTXT_IS_KERNEL(ctxt))
	elan3_segelan3_destroy (ctxt);				/* Unmap the command port from the users address space. */
#endif
   
    ReleaseHaltOperations (dev, NUM_HALTOPS);

    if (ctxt->Input0Resolver)
	CancelNetworkErrorResolver (ctxt->Input0Resolver);

    if (ctxt->Input1Resolver)
	CancelNetworkErrorResolver (ctxt->Input1Resolver);

    while ((nef = ctxt->NetworkErrorFixups) != NULL)
    {
	ctxt->NetworkErrorFixups = nef->Next;

	CompleteNetworkErrorFixup (ctxt, nef, ESRCH);
    }

    KMEM_FREEPAGES ((void *) ctxt->FlagPage, 1);

    KMEM_FREE ((void *) ctxt->CommandTraps, sizeof (COMMAND_TRAP)    * ntrapped_commands);
    KMEM_FREE ((void *) ctxt->ThreadTraps,  sizeof (THREAD_TRAP)     * ntrapped_threads);
    KMEM_FREE ((void *) ctxt->DmaTraps,     sizeof (DMA_TRAP)        * ntrapped_dmas);
    KMEM_FREE ((void *) ctxt->EventCookies, sizeof (EVENT_COOKIE)    * ntrapped_events);
    KMEM_FREE ((void *) ctxt->Commands,     sizeof (CProcTrapBuf_BE) * noverflow_commands);
    KMEM_FREE ((void *) ctxt->SwapThreads,  sizeof (E3_Addr)         * nswapped_threads);
    KMEM_FREE ((void *) ctxt->SwapDmas,     sizeof (E3_DMA_BE)       * nswapped_dmas);

    kcondvar_destroy (&ctxt->Wait);
    kcondvar_destroy (&ctxt->CommandPortWait);
    kcondvar_destroy (&ctxt->LwpWait);
    kcondvar_destroy (&ctxt->HaltWait);

    kmutex_destroy (&ctxt->SwapListsLock);
    kmutex_destroy (&ctxt->CmdLock);
    kmutex_destroy (&ctxt->NetworkErrorLock);
    spin_lock_destroy  (&ctxt->InputFaultLock);

    krwlock_destroy (&ctxt->VpLock);

    KMEM_FREE (ctxt, sizeof (ELAN3_CTXT));
}

int 
elan3_doattach(ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap)
{
    unsigned long pgnum = ((cap->cap_mycontext & MAX_ROOT_CONTEXT_MASK) * sizeof (E3_CommandPort)) / PAGE_SIZE;
    unsigned long pgoff = ((cap->cap_mycontext & MAX_ROOT_CONTEXT_MASK) * sizeof (E3_CommandPort)) & (PAGE_SIZE-1);
    ELAN3_DEV     *dev   = ctxt->Device;
    int           res   = ESUCCESS;
    unsigned long flags;

    /* Map in the command port for this context */
    if (MapDeviceRegister (dev, ELAN3_BAR_COMMAND_PORT, &ctxt->CommandPage, pgnum * PAGE_SIZE, PAGE_SIZE, &ctxt->CommandPageHandle) != ESUCCESS)
    {
	PRINTF0 (ctxt, DBG_FN, "elan3_doattach: MapDeviceRegister failed");
	return (EINVAL);
    }

    ctxt->CommandPort = ctxt->CommandPage + pgoff;

    spin_lock_irqsave (&dev->IntrLock, flags);

    res = 0;
    if (ELAN3_DEV_CTX_TABLE(dev,cap->cap_mycontext) != NULL)
	res = EBUSY;
    else
    {
	if ((res = elan3mmu_attach (ctxt->Device, cap->cap_mycontext, ctxt->Elan3mmu, 
				    ctxt->RouteTable->Table, ctxt->RouteTable->Size-1)) == 0)
	{
	    ELAN3_DEV_CTX_TABLE(dev,cap->cap_mycontext) = ctxt;
	    ctxt->Capability                            = *cap;
	}
    }

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    if (res == ESUCCESS)
	elan3_swapin (ctxt, CTXT_DETACHED);
    else 
    {
	UnmapDeviceRegister (dev, &ctxt->CommandPageHandle);
	ctxt->CommandPage = (ioaddr_t) 0; 
	ctxt->CommandPort = (ioaddr_t) 0;
    }

    return (res);
}

void
elan3_destroy_callback( void * args, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map)
{
    if (map == NULL) 
    {
	/* the cap is being destroyed */
	PRINTF0 (NULL, DBG_VP, "elan3_destroy_callback: the cap is being destroyed \n");
    }
    else
    {
	/* the map is being destroyed */
	PRINTF0 (NULL, DBG_VP, "elan3_destroy_callback: the map is being destroyed \n");
    }
}

int
elan3_attach (ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap)
{
    ELAN3_DEV *dev = ctxt->Device;
    int type;
    int res;

    switch (type = elan3_validate_cap (dev, cap, ELAN_USER_ATTACH))
    {
    case ELAN_CAP_OK:
	/* nothing */
	break;

    case ELAN_CAP_RMS:
	if ((res = elan_attach_cap(cap, dev->Devinfo.dev_rail, ctxt, elan3_destroy_callback)) != 0)
	    return res;
	break;

    default:
	return (EINVAL);
    }

    if (((res = elan3_doattach(ctxt,cap)) != ESUCCESS) && (type == ELAN_CAP_RMS))
	elan_detach_cap(cap, dev->Devinfo.dev_rail);

    return res;
}

void
elan3_detach ( ELAN3_CTXT *ctxt )
{
    ELAN3_DEV   *dev                 = ctxt->Device;
    int need_to_call_elanmod_detach = 0;
    unsigned long flags;

    PRINTF1 (ctxt, DBG_FN, "elan3_detach: %p \n", ctxt );
    
    if (ctxt->Capability.cap_mycontext == ELAN_CAP_UNINITIALISED)
    {
	PRINTF0 (ctxt, DBG_FN, "elan3_detach: context not attached \n");
	return ;
    }

    /* must you be in the ctx_table ?? */
    
    switch (ctxt->Capability.cap_type & ELAN_CAP_TYPE_MASK)
    {
    case ELAN_CAP_TYPE_BLOCK:
    case ELAN_CAP_TYPE_CYCLIC:
    {
	if (ELAN3_SYSTEM_CONTEXT (ctxt->Capability.cap_mycontext))
	    return ;

	if (! (ctxt->Capability.cap_type & ELAN_CAP_TYPE_HWTEST))
	    need_to_call_elanmod_detach = 1;

	break;
    }	
    default:
	return ;
    }

    elan3_swapout (ctxt, CTXT_DETACHED);

    spin_lock_irqsave (&dev->IntrLock, flags);

    elan3mmu_detach (dev, ctxt->Capability.cap_mycontext);
    ELAN3_DEV_CTX_TABLE(dev,ctxt->Capability.cap_mycontext) = NULL;

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    if (ctxt->CommandPage)
    {
	UnmapDeviceRegister (dev, &ctxt->CommandPageHandle);
	ctxt->CommandPage = (ioaddr_t) 0;
    }
    
    if (need_to_call_elanmod_detach) 
	elan_detach_cap(&ctxt->Capability, dev->Devinfo.dev_rail);

    elan_nullcap (&ctxt->Capability);

}

void
elan3_dodetach ( ELAN3_CTXT *ctxt )
{
    ELAN3_DEV     *dev = ctxt->Device;
    unsigned long flags;

    PRINTF1 (ctxt, DBG_FN, "elan3_dodetach: %p \n", ctxt );
    
    if (ctxt->Capability.cap_mycontext == ELAN_CAP_UNINITIALISED)
    {
	PRINTF0 (ctxt, DBG_FN, "elan3_dodetach: context not attached \n");
	return ;
    }

    elan3_swapout (ctxt, CTXT_DETACHED);

    spin_lock_irqsave (&dev->IntrLock, flags);

    elan3mmu_detach (dev, ctxt->Capability.cap_mycontext);
    ELAN3_DEV_CTX_TABLE(dev,ctxt->Capability.cap_mycontext) = NULL;

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    if (ctxt->CommandPage)
    {
	UnmapDeviceRegister (dev, &ctxt->CommandPageHandle);
	ctxt->CommandPage = (ioaddr_t) 0;
    }
    
    elan_nullcap (&ctxt->Capability);
}

void
elan3_swapin (ELAN3_CTXT *ctxt, int reason)
{
    ELAN3_DEV *dev = ctxt->Device;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    ASSERT (ctxt->Status & CTXT_SWAPPED_REASONS);

    PRINTF3 (ctxt, DBG_SWAP, "elan3_swapin: status %x State %s reason %x\n", 
	     ctxt->Status, OthersStateStrings[ctxt->OthersState], reason);

    while (ctxt->Status & CTXT_SWAPPING_OUT)			/* In transition */
	kcondvar_wait (&ctxt->LwpWait, &dev->IntrLock, &flags);

    if (reason == CTXT_NO_LWPS && ctxt->LwpCount++ != 0)	/* Added another LWP */
    {
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	return;
    }

    if ((ctxt->Status & ~reason) & CTXT_SWAPPED_REASONS)
	ctxt->Status &= ~reason;
    else
    {
	ASSERT (ctxt->Status & CTXT_SWAPPED_OUT);
	ASSERT (ctxt->OthersState == CTXT_OTHERS_SWAPPED);
	
	/*
	 * Will not be swapped out anymore, so ask the "user" to perform 
	 * any swapping in he needs before letting the context run again.
	 */
	
	ctxt->Status &= ~(CTXT_SWAPPED_OUT | CTXT_QUEUES_EMPTY | reason);
	ctxt->OthersState = CTXT_OTHERS_RUNNING;

	if (ctxt->Input0Trap.State == CTXT_STATE_OK && ctxt->Input1Trap.State == CTXT_STATE_OK)
	    SetInputterStateForContext (ctxt, 0, NULL);
	
	kcondvar_wakeupall (&ctxt->Wait, &dev->IntrLock);
    }

    PRINTF2 (ctxt, DBG_SWAP, "elan3_swapin: all done - status %x state %s\n",
	     ctxt->Status, OthersStateStrings[ctxt->OthersState]);
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}


void
elan3_swapout (ELAN3_CTXT *ctxt, int reason)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int           cansleep;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    PRINTF3 (ctxt, DBG_SWAP, "elan3_swapout: status %x state %s reason %x\n", 
	     ctxt->Status, OthersStateStrings[ctxt->OthersState], reason);

    if (reason == CTXT_NO_LWPS)
    {
	if (--ctxt->LwpCount != 0)				/* Still other LWPs running */
	{
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    return;
	}

	kcondvar_wakeupall (&ctxt->LwpWait, &dev->IntrLock);		/* Wakeup anyone waiting on LwpCount */
    }
    
    ctxt->Status |= reason;
    
    while (ctxt->Status & CTXT_SWAPPING_OUT)			/* wait for someone else to finish swapping */
	kcondvar_wait (&ctxt->LwpWait, &dev->IntrLock, &flags);		/* out */

    if (ctxt->Status & CTXT_SWAPPED_OUT)
    {
	if (reason == CTXT_NO_LWPS)				/* Wakeup other thread waiting on LWP exit */
	    kcondvar_wakeupall (&ctxt->LwpWait, &dev->IntrLock);
	
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	return;
    }
    
    /*
     * mark the context as swapping out.
     */
    ctxt->Status |= CTXT_SWAPPING_OUT;
    
    if (reason != CTXT_FIXUP_NETERR)
    {
	/*
	 * Stop all of the lwps.
	 */
	while (ctxt->LwpCount)
	{
	    kcondvar_wakeupall (&ctxt->Wait, &dev->IntrLock);		/* Wake up any lwps */
	    kcondvar_wait (&ctxt->LwpWait, &dev->IntrLock, &flags);		/* then wait for them to enter elan3_swapout */
	}
    }
    
    StartSwapoutContext (ctxt, 0, NULL);
    for (;;)
    {
	PRINTF0 (ctxt, DBG_SWAP, "elan3_swapout: HandleExceptions\n");

	cansleep = (HandleExceptions(ctxt, &flags) == ESUCCESS);

	PRINTF2 (ctxt, DBG_SWAP, "elan3_swapout: OthersState=%d cansleep=%d\n", ctxt->OthersState, cansleep);

	if (ctxt->OthersState == CTXT_OTHERS_SWAPPED)
	    break;

	if (cansleep)
	    kcondvar_wait (&ctxt->Wait, &dev->IntrLock, &flags);
    }
    PRINTF0 (ctxt, DBG_SWAP, "elan3_swapout: swapped out\n");
    
    ASSERT (ELAN3_QUEUE_EMPTY (ctxt->DmaTrapQ));
    ASSERT (ELAN3_QUEUE_EMPTY (ctxt->ThreadTrapQ));

    ctxt->Status |=  CTXT_SWAPPED_OUT;
    ctxt->Status &= ~CTXT_SWAPPING_OUT;

    kcondvar_wakeupall (&ctxt->LwpWait, &dev->IntrLock);

    PRINTF2 (ctxt, DBG_SWAP, "elan3_swapout: all done - status %x state %s\n",
	     ctxt->Status, OthersStateStrings[ctxt->OthersState]);

    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

int
elan3_pagefault (ELAN3_CTXT *ctxt, E3_FaultSave_BE *FaultSave, int npages)
{
    E3_Addr     elanAddr = FaultSave->s.FaultAddress;
    int		writeable;
    int		res;

    PRINTF3 (ctxt, DBG_FAULT, "elan3_pagefault: elanAddr %08x FSR %08x : %s\n", elanAddr, FaultSave->s.FSR.Status,
	     FaultSave->s.FSR.s.ProtFault ? "protection fault" : "pte invalid");
    
    /* Look at the FSR to determine the fault type etc */
    
    if (FaultSave->s.FSR.Status == 0)				/* this is a target abort/parity error, so look */
    {								/* at the PCI config space registers to determine  */
	ElanBusError (ctxt->Device);
	return (EFAULT);					
    }
    
    if (FaultSave->s.FSR.s.AlignmentErr)			/* Alignment errors are always fatal. */
    {
	PRINTF0 (ctxt, DBG_FAULT, "elan3_pagefault: Alignment error\n");
	return (EFAULT);
    }

    if (FaultSave->s.FSR.s.WalkBadData)				/* Memory ECC error during a walk */
    {
	PRINTF0 (ctxt, DBG_FAULT, "elan3_pagefault: Memory ECC error during walk\n");
	return (EFAULT);
    }

    if (!FaultSave->s.FSR.s.ProtFault &&			/* DMA memory type changed */
	!FaultSave->s.FSR.s.Walking)
    {
	PRINTF0 (ctxt, DBG_FAULT, "elan3_pagefault: DMA memory type changed\n");
	return (EFAULT);
    }

    ASSERT (FaultSave->s.FSR.s.ProtFault ? 			/* protection errors, should always have a valid pte */
	    (!FaultSave->s.FSR.s.Walking || !(FaultSave->s.FSR.s.Level==3) ||  FaultSave->s.FSR.s.FaultPte == ELAN3_ET_PTE) : 
	    FaultSave->s.FSR.s.FaultPte == ELAN3_ET_INVALID);	/* otherwise it must be an invalid pte */

    /*
     * Determine whether to fault for a 'write' from the access permissions we need, and not
     * from the access type (WrAcc).
     */
    writeable = (FaultSave->s.FSR.s.AccTypePerm & (1 << FSR_WritePermBit));

    /* Check that we have the right permissions for this access type. */
    if ((res = elan3mmu_checkperm (ctxt->Elan3mmu, (elanAddr&PAGEMASK), npages*PAGESIZE, FaultSave->s.FSR.s.AccTypePerm)) != 0)
    {
	PRINTF1 (ctxt, DBG_FAULT, "elan3_pagefault: %s\n", (res == ENOMEM) ? "no protection mapping" : "protection error");
	
	return (res);
    }

    res = LoadElanTranslation (ctxt, (elanAddr&PAGEMASK), npages*PAGESIZE, FaultSave->s.FSR.s.ProtFault, writeable);

    if (res == ESUCCESS)
    {
	BumpStat (ctxt->Device, PageFaults);
	BumpUserStat (ctxt, PageFaults);
    }

    PRINTF1 (ctxt, DBG_FAULT, "elan3_pagefault: -> %d\n", res);

    return (res);
}

void
elan3_block_inputter (ELAN3_CTXT *ctxt, int block)
{
    ELAN3_DEV *dev = ctxt->Device;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    
    if (block)
	ctxt->Status |= CTXT_USER_FILTERING;
    else
	ctxt->Status &= ~CTXT_USER_FILTERING;

    if (ctxt->Capability.cap_mycontext != ELAN_CAP_UNINITIALISED)
	SetInputterStateForContext (ctxt, 0, NULL);
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

int
FixupNetworkErrors (ELAN3_CTXT *ctxt, unsigned long *flags)
{
    ELAN3_DEV 	 *dev = ctxt->Device;
    NETERR_FIXUP *nef;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    
    if (ctxt->NetworkErrorFixups == NULL)
	return (ESUCCESS);

    spin_unlock_irqrestore (&dev->IntrLock, *flags);
    
    kmutex_lock (&ctxt->NetworkErrorLock);			/* single thread while fixing up errors */
    elan3_swapout (ctxt, CTXT_FIXUP_NETERR);

    spin_lock_irqsave (&dev->IntrLock, *flags);
    while ((nef = ctxt->NetworkErrorFixups) != NULL)
    {
	ctxt->NetworkErrorFixups = nef->Next;
	spin_unlock_irqrestore (&dev->IntrLock, *flags);

	if (ELAN3_OP_FIXUP_NETWORK_ERROR (ctxt, nef) == OP_FAILED)
	    CompleteNetworkErrorFixup (ctxt, nef, EINVAL);

	spin_lock_irqsave (&dev->IntrLock, *flags);
    }
    spin_unlock_irqrestore (&dev->IntrLock, *flags);

    elan3_swapin (ctxt, CTXT_FIXUP_NETERR);

    kmutex_unlock (&ctxt->NetworkErrorLock);
    spin_lock_irqsave (&dev->IntrLock, *flags);
    return (EAGAIN);
}

int
CompleteNetworkErrorResolver (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, NETERR_RESOLVER *rvp)
{
    int state;

    switch (rvp->Status)
    {
    case ESUCCESS:
	/*
	 * the item still existed at the source - if it's a wait for EOP transaction
	 * then the source will retry - otherwise the remote event will have been
	 * cleared and we should execute it
	 */
	PRINTF1 (ctxt, DBG_NETERR, "CompleteNetworkErrorResolver: ESUCCESS zero WaitForEopTransaction %p\n", trap->WaitForEopTransaction);

	state = trap->WaitForEopTransaction ? CTXT_STATE_OK : CTXT_STATE_NEEDS_RESTART;

	break;

    case ESRCH:	
	/*
	 * the item was not found at the source - we should always execute the transaction
	 * since it will never be resent
	 */
	PRINTF1 (ctxt, DBG_NETERR, "CompleteNetworkErrorResolver: ESRCH execute WaitForEopTransaction %p\n", trap->WaitForEopTransaction);
	state = CTXT_STATE_NEEDS_RESTART;
	break;

    default:							/* other errors */
	PRINTF1 (ctxt, DBG_NETERR, "CompleteNetworkErrorResolver: %d\n", rvp->Status);
	if (ElanException (ctxt, EXCEPTION_NETWORK_ERROR, INPUT_PROC, trap, &rvp) == OP_HANDLED)
	    state = CTXT_STATE_NEEDS_RESTART;
	else
	    state = CTXT_STATE_OK;
	break;
    }

    FreeNetworkErrorResolver (rvp);

    return (state);
}

int
HandleExceptions (ELAN3_CTXT *ctxt, unsigned long *flags)
{
    ELAN3_DEV        *dev    = ctxt->Device;
    THREAD_TRAP      tproc;
    DMA_TRAP         dproc;
    NETERR_RESOLVER *rvp;
    int		     state;

    if (ctxt->Status & CTXT_COMMAND_OVERFLOW_ERROR)
    {
	ctxt->Status &= ~CTXT_COMMAND_OVERFLOW_ERROR;
	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ElanException (ctxt, EXCEPTION_COMMAND_OVERFLOW, COMMAND_PROC, NULL);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }
    
    if (! ELAN3_QUEUE_BACK_EMPTY (ctxt->CommandTrapQ))
    {
	/* XXXX: unmap translations to the command port */

	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ResolveCProcTrap (ctxt);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }
    
    if (ctxt->Input0Trap.State == CTXT_STATE_TRAPPED)
    {
	ctxt->Input0Trap.State = CTXT_STATE_RESOLVING;

	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ResolveIProcTrap (ctxt, &ctxt->Input0Trap, &ctxt->Input0Resolver);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }

    if (ctxt->Input1Trap.State == CTXT_STATE_TRAPPED)
    {
	ctxt->Input1Trap.State = CTXT_STATE_RESOLVING;

	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ResolveIProcTrap (ctxt, &ctxt->Input1Trap, &ctxt->Input1Resolver);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }

    if ((rvp = ctxt->Input0Resolver) != NULL && rvp->Completed)
    {
	ASSERT (ctxt->Input0Trap.State == CTXT_STATE_NETWORK_ERROR);

	ctxt->Input0Resolver = NULL;
	
	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	state = CompleteNetworkErrorResolver (ctxt, &ctxt->Input0Trap, rvp);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	ctxt->Input0Trap.State = state;
	return (EAGAIN);
    }

    if ((rvp = ctxt->Input1Resolver) != NULL && rvp->Completed)
    {
	ASSERT (ctxt->Input1Trap.State == CTXT_STATE_NETWORK_ERROR);

	ctxt->Input1Resolver = NULL;
	
	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	state = CompleteNetworkErrorResolver (ctxt,&ctxt->Input1Trap, rvp);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	ctxt->Input1Trap.State = state;
	return (EAGAIN);
    }

    if (NextTProcTrap (ctxt, &tproc))
    {
	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ResolveTProcTrap (ctxt, &tproc);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }
    ctxt->Status &= ~CTXT_THREAD_QUEUE_FULL;

    if (NextDProcTrap (ctxt, &dproc))
    {
	spin_unlock_irqrestore (&dev->IntrLock, *flags);
	ResolveDProcTrap (ctxt, &dproc);
	spin_lock_irqsave (&dev->IntrLock, *flags);
	return (EAGAIN);
    }
    ctxt->Status &= ~CTXT_DMA_QUEUE_FULL;

    /* Handle all event interrupts. */
    if (! ELAN3_QUEUE_EMPTY (ctxt->EventCookieQ))
    {
	while (! ELAN3_QUEUE_EMPTY (ctxt->EventCookieQ))
	{
	    E3_uint32 cookie = *ELAN3_QUEUE_FRONT (ctxt->EventCookieQ, ctxt->EventCookies);

	    ELAN3_QUEUE_REMOVE (ctxt->EventCookieQ);

	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    if (ELAN3_OP_EVENT (ctxt, cookie, OP_LWP) != OP_DEFER)
		spin_lock_irqsave (&dev->IntrLock, *flags);
	    else
	    {
		spin_lock_irqsave (&dev->IntrLock, *flags);	/* place the cookie back on the queue. */
								/* note we place it on the front to ensure  */
		ELAN3_QUEUE_ADD_FRONT (ctxt->EventCookieQ);	/* event ordering. */
		*ELAN3_QUEUE_FRONT (ctxt->EventCookieQ, ctxt->EventCookies) = cookie;
	    }
	}
	return (EAGAIN);
    }
    ctxt->Status &= ~CTXT_EVENT_QUEUE_FULL;

    if (! ELAN3_QUEUE_EMPTY (ctxt->SwapDmaQ))
    {
	while (! ELAN3_QUEUE_EMPTY (ctxt->SwapDmaQ))
	{
	    E3_DMA_BE DmaDesc = *ELAN3_QUEUE_FRONT (ctxt->SwapDmaQ, ctxt->SwapDmas);

	    ELAN3_QUEUE_REMOVE (ctxt->SwapDmaQ);

	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    RestartDmaDesc (ctxt, &DmaDesc);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	}
	return (EAGAIN);
    }
    
    if (! ELAN3_QUEUE_EMPTY (ctxt->SwapThreadQ))
    {
	while (! ELAN3_QUEUE_EMPTY (ctxt->SwapThreadQ))
	{
	    E3_Addr StackPointer = *ELAN3_QUEUE_FRONT (ctxt->SwapThreadQ, ctxt->SwapThreads);

	    ELAN3_QUEUE_REMOVE (ctxt->SwapThreadQ);

	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    ReissueStackPointer (ctxt, StackPointer);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	}
	return (EAGAIN);
    }
    
    switch (ctxt->OthersState)
    {
    case CTXT_OTHERS_SWAPPING:
	if (! (ctxt->Status & CTXT_OTHERS_REASONS))
	    ctxt->OthersState = CTXT_OTHERS_RUNNING;
	else
	    ctxt->OthersState = CTXT_OTHERS_SWAPPED;

	PRINTF1 (ctxt, DBG_LWP, "HandleExceptions: OthersState : swapping -> %s\n", OthersStateStrings[ctxt->OthersState]);
	    
	break;

    case CTXT_OTHERS_SWAPPING_MORE:
	ctxt->OthersState = CTXT_OTHERS_HALTING_MORE;
	QueueHaltOperation (dev, 0, NULL, INT_DProcHalted | INT_TProcHalted, HaltSwapContext, ctxt);

	PRINTF1 (ctxt, DBG_LWP, "HandleExceptions: OthersState : swapping_more -> %s\n", OthersStateStrings[ctxt->OthersState]);
	break;
    }
    return (ESUCCESS);
}

int
RestartContext (ELAN3_CTXT *ctxt, unsigned long *flags)
{
    ELAN3_DEV *dev = ctxt->Device;
    int       res;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));

    PRINTF1 (ctxt, DBG_LWP, "RestartContext: status %x\n", ctxt->Status);

    if (! (ctxt->Status & CTXT_OTHERS_REASONS))
    {
	if (! ELAN3_QUEUE_FRONT_EMPTY (ctxt->CommandTrapQ) || ! ELAN3_QUEUE_EMPTY(ctxt->CommandQ))
	{
	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    RestartCProcTrap (ctxt);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	    return (EAGAIN);
	}

	if (ctxt->Input0Trap.State == CTXT_STATE_NEEDS_RESTART)
	{
	    ctxt->Input0Trap.State = CTXT_STATE_EXECUTING;

	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    res = RestartIProcTrap (ctxt, &ctxt->Input0Trap);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	    
	    if (res == ESUCCESS)
		ctxt->Input0Trap.State = CTXT_STATE_OK;
	    else
		ctxt->Input0Trap.State = CTXT_STATE_NEEDS_RESTART;
	    return (EAGAIN);
	}

	if (ctxt->Input1Trap.State == CTXT_STATE_NEEDS_RESTART)
	{
	    ctxt->Input1Trap.State = CTXT_STATE_EXECUTING;

	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    res = RestartIProcTrap (ctxt, &ctxt->Input1Trap);
	    spin_lock_irqsave (&dev->IntrLock, *flags);

	    if (res == ESUCCESS)
		ctxt->Input1Trap.State = CTXT_STATE_OK;
	    else
		ctxt->Input1Trap.State = CTXT_STATE_NEEDS_RESTART;
	    return (EAGAIN);
	}

	if (SetEventsNeedRestart (ctxt))
	{
	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    RestartSetEvents (ctxt);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	    return (EAGAIN);
	}

	SetInputterStateForContext (ctxt, 0, NULL);

	if (TProcNeedsRestart (ctxt))
	{
	    spin_unlock_irqrestore (&dev->IntrLock, *flags);

	    LoadCommandPortTranslation (ctxt);
	    RestartTProcItems (ctxt);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	    return (EAGAIN);
	}

	if (DProcNeedsRestart (ctxt))
	{
	    spin_unlock_irqrestore (&dev->IntrLock, *flags);
	    RestartDProcItems (ctxt);
	    spin_lock_irqsave (&dev->IntrLock, *flags);
	    return (EAGAIN);
	}

	if (ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ))
	{
	    PRINTF1 (ctxt, DBG_LWP, "RestartContext: setting Command Flag at %p to 0\n", &ctxt->FlagPage->CommandFlag);

	    ctxt->FlagPage->CommandFlag = 0;

	    if (ctxt->Status & CTXT_WAITING_COMMAND)
	    {
		PRINTF0 (ctxt, DBG_LWP, "RestartContext: waking up threads waiting for commandport\n");
		
		ctxt->Status &= ~CTXT_WAITING_COMMAND;
		
		kcondvar_wakeupall (&ctxt->CommandPortWait, &dev->IntrLock);
	    }
	}
    }

    return (ESUCCESS);
}

static void
HaltSwapContext (ELAN3_DEV *dev, void *arg)
{
    ELAN3_CTXT        *ctxt    = (ELAN3_CTXT *) arg;
    int	              SysCntx = (ctxt->Capability.cap_mycontext & SYS_CONTEXT_BIT);
    E3_ThreadQueue_BE thread;
    E3_DMA_BE         dma;
    sdramaddr_t       FPtr, BPtr;
    sdramaddr_t	      Base, Top;
    u_int	     *runCount;
    unsigned long     flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    ASSERT (ctxt->OthersState == CTXT_OTHERS_HALTING || ctxt->OthersState == CTXT_OTHERS_HALTING_MORE);

    PRINTF2 (ctxt, DBG_SWAP, "HaltSwapContext: status %x state %s\n", ctxt->Status, OthersStateStrings[ctxt->OthersState]);

    if (! (ctxt->Status & CTXT_OTHERS_REASONS))
    {
	if (ctxt->OthersState == CTXT_OTHERS_HALTING_MORE)
	{
	    runCount = SysCntx ? &dev->HaltAllCount : &dev->HaltNonContext0Count;

	    if (--(*runCount) == 0)
		SetSchedStatusRegister (dev, 0, NULL);
	}
	ctxt->OthersState = CTXT_OTHERS_RUNNING;
	
	PRINTF0 (ctxt, DBG_SWAP, "HaltSwapContext: no more reason to swap -> others_running\n");

	kcondvar_wakeupall (&ctxt->Wait, &dev->IntrLock);
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	return;
    }

    /*
     * Capture all other processors since we're not being responsive to 
     * the command processor interrupt.
     */
    CAPTURE_CPUS();

    if (SysCntx)
    {
	FPtr = read_reg32 (dev, TProc_SysCntx_FPtr);
	BPtr = read_reg32 (dev, TProc_SysCntx_BPtr);
	Base = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxThreadQueue[0]);
	Top  = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxThreadQueue[E3_SysCntxQueueSize-1]);
    }
    else
    {
	FPtr  = read_reg32 (dev, TProc_NonSysCntx_FPtr);
	BPtr  = read_reg32 (dev, TProc_NonSysCntx_BPtr);
	Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxThreadQueue[0]);
	Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxThreadQueue[E3_NonSysCntxQueueSize-1]);
    }

    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, (void *) &thread, sizeof (E3_ThreadQueue_BE));
	
	if (thread.s.Context == ctxt->Capability.cap_mycontext)
	{
	    if (ELAN3_QUEUE_FULL (ctxt->SwapThreadQ))
		break;
	    
	    *ELAN3_QUEUE_BACK(ctxt->SwapThreadQ, ctxt->SwapThreads) = thread.s.Thread;
	    ELAN3_QUEUE_ADD (ctxt->SwapThreadQ);
	    
	    /*
	     * Remove this entry from the queue by replacing it with 
	     * the "magic" thread value.
	     *
	     * NOTE: we must preserve the SYS_CONTEXT_BIT since the Elan uses this
	     * to mark the approriate run queue as empty.
	     */
	    thread.s.Context = SysCntx ? SYS_CONTEXT_BIT : 0;
	    thread.s.Thread  = VanishingStackPointer;

	    elan3_sdram_copyq_to_sdram (dev, (void *) &thread, FPtr, sizeof (E3_ThreadQueue_BE));
	}
	
	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_ThreadQueue);
    }

    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc.s.FSR)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData1.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData2.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData3.s.FSR.Status)) == 0);

    if (SysCntx)
    {
	FPtr  = read_reg32 (dev, DProc_SysCntx_FPtr);
	BPtr  = read_reg32 (dev, DProc_SysCntx_BPtr);
	Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0]);
	Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[E3_SysCntxQueueSize-1]);
    }
    else
    {
	FPtr  = read_reg32 (dev, DProc_NonSysCntx_FPtr);
	BPtr  = read_reg32 (dev, DProc_NonSysCntx_BPtr);
	Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[0]);
	Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[E3_NonSysCntxQueueSize-1]);
    }

    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, &dma, sizeof (E3_DMA_BE));
	    
	if (dma.s.dma_u.s.Context == ctxt->Capability.cap_mycontext)
	{
	    if (ELAN3_QUEUE_FULL (ctxt->SwapDmaQ))
		break;
	    
	    *ELAN3_QUEUE_BACK (ctxt->SwapDmaQ, ctxt->SwapDmas) = dma;
	    ELAN3_QUEUE_ADD (ctxt->SwapDmaQ);

	    /*
	     * Remove the DMA from the queue by replacing it with one with
	     * zero size and no events.
	     *
	     * NOTE: we must preserve the SYS_CONTEXT_BIT since the Elan uses this
	     * to mark the approriate run queue as empty.
	     */
	    dma.s.dma_type            = ((SysCntx ? SYS_CONTEXT_BIT : 0) << 16);
	    dma.s.dma_size            = 0;
	    dma.s.dma_source          = (E3_Addr) 0;
	    dma.s.dma_dest            = (E3_Addr) 0;
	    dma.s.dma_destCookieVProc = (E3_Addr) 0;
	    dma.s.dma_srcEvent        = (E3_Addr) 0;
	    dma.s.dma_srcCookieVProc  = (E3_Addr) 0;

	    elan3_sdram_copyq_to_sdram (dev, &dma, FPtr, sizeof (E3_DMA_BE));
	}

	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_DMA);
    }

    /*
     * Release the other processors now before signalling the LWP.
     */
    RELEASE_CPUS();

    if (! ELAN3_QUEUE_FULL (ctxt->SwapDmaQ) && !ELAN3_QUEUE_FULL (ctxt->SwapThreadQ))
    {
	/*
	 * We've compleletly emptied the elan queues of items in this
	 * context, so we now mark it as fully swapped out.
	 */
	if (ctxt->OthersState == CTXT_OTHERS_HALTING_MORE)
	{
	    runCount = SysCntx ? &dev->HaltAllCount : &dev->HaltNonContext0Count;
	    
	    if (--(*runCount) == 0)
		SetSchedStatusRegister (dev, 0, NULL);
	    
	}
	PRINTF0 (ctxt, DBG_SWAP, "HaltSwapContext: queues emptied -> others_swapping\n");

	ctxt->OthersState = CTXT_OTHERS_SWAPPING;
	kcondvar_wakeupall (&ctxt->Wait, &dev->IntrLock);
    }
    else
    {
	if (ctxt->OthersState == CTXT_OTHERS_HALTING)
	{
	    runCount = SysCntx ? &dev->HaltAllCount : &dev->HaltNonContext0Count;
	    
	    if ((*runCount)++ == 0)
		SetSchedStatusRegister (dev, 0, NULL);
	}
	PRINTF0 (ctxt, DBG_SWAP, "HaltSwapContext: queues not emptied -> others_swapping_more\n");

	ctxt->OthersState = CTXT_OTHERS_SWAPPING_MORE;
	kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

void
UnloadCommandPageMapping (ELAN3_CTXT *ctxt)
{
    /*
     * Unload the Elan translations,  and flag the main processor to stall after 
     * issueing its next command.
     */
    if (ctxt->CommandPageMapping != NULL && (ctxt->Status & CTXT_COMMAND_MAPPED_ELAN))
    {
	ELAN3MMU_RGN *rgn = elan3mmu_rgnat_main (ctxt->Elan3mmu, ctxt->CommandPageMapping);
	
	if (rgn != NULL)
	{
	    E3_Addr eaddr = rgn->rgn_ebase + (ctxt->CommandPageMapping - rgn->rgn_mbase);
	    
	    PRINTF1 (ctxt, DBG_INTR, "UnloadCommandPageMapping: unmapping command port at addr %08x\n", eaddr);
	    
	    elan3mmu_unload (ctxt->Elan3mmu, eaddr, PAGESIZE, PTE_UNLOAD);
	}
	
	ctxt->Status &= ~CTXT_COMMAND_MAPPED_ELAN;
    }
}

void
StartSwapoutContext (ELAN3_CTXT *ctxt, E3_uint32 Pend, E3_uint32 *Maskp)
{
    ELAN3_DEV   *dev     = ctxt->Device;
    int		SysCntx = (ctxt->Capability.cap_mycontext & SYS_CONTEXT_BIT);
    u_int      *runCount;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));

    PRINTF2 (ctxt, DBG_SWAP, "StartSwapoutContext: Status %x OthersState %s\n",
	     ctxt->Status, OthersStateStrings [ctxt->OthersState]);
    /*
     * Disable the inputters,  we should already have a reason for it.
     */
    SetInputterStateForContext (ctxt, Pend, Maskp);

    UnloadCommandPageMapping (ctxt);

    /* 
     * Flag main processor to stall after issueing next command
     */
    PRINTF1 (ctxt, DBG_SWAP, "StartSwapoutContext: setting Command Flag at %p to 1\n", &ctxt->FlagPage->CommandFlag);

    ctxt->FlagPage->CommandFlag = 1;

    PRINTF1 (ctxt, DBG_SWAP, "StartSwapoutContext: OthersState=%d\n", ctxt->OthersState);

    /*
     * And queue a haltop to stop the queues and clear it out.
     */
    switch (ctxt->OthersState)
    {
    case CTXT_OTHERS_RUNNING:
	PRINTF0 (ctxt, DBG_SWAP, "StartSwapoutContext: -> others_halting\n");

	ctxt->OthersState = CTXT_OTHERS_HALTING;

	QueueHaltOperation (dev, Pend, Maskp, INT_DProcHalted | INT_TProcHalted, HaltSwapContext, ctxt);
	break;
	
    case CTXT_OTHERS_SWAPPING:
	PRINTF0 (ctxt, DBG_SWAP, "StartSwapoutContext: -> others_swapping_more\n");
	ctxt->OthersState = CTXT_OTHERS_SWAPPING_MORE;

	runCount = SysCntx ? &dev->HaltAllCount : &dev->HaltNonContext0Count;
	    
	if ((*runCount)++ == 0)
	    SetSchedStatusRegister (dev, Pend, Maskp);
	break;
    default:
	PRINTF1 (ctxt, DBG_SWAP, "StartSwapoutContext: OthersState=%d\n", ctxt->OthersState);
	break;
    }
}

#if defined(DIGITAL_UNIX)
/* temporary tweaks to priority bump */
int lwp_do_prio = 1;
int lwp_do_nxm = 1;
int lwp_prio = BASEPRI_USER-1;
#elif defined(LINUX)
/* This is the default nice level for the helper LWP */
int LwpNice = -1;
#endif

int
elan3_lwp (ELAN3_CTXT *ctxt)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int	          res;
    unsigned long flags;

    PRINTF1 (ctxt, DBG_LWP, "elan3_lwp: started, context 0x%x\n", ctxt->Capability.cap_mycontext);

#if defined(DIGITAL_UNIX)
    {
        thread_t mythread = current_thread();
        if (lwp_do_prio && (lwp_do_nxm || !IS_NXM_TASK(mythread->task)))
        {
            mythread->priority = mythread->sched_pri = lwp_prio;
            mythread->max_priority = BASEPRI_HIGHEST;
            (void) thread_priority(mythread, lwp_prio, 0, 1);
        }
    }
#elif defined(LINUX)
    {
	/* Do the priority trick for the helper LWP so that it
	 * runs in preferance to the user threads which may be
	 * burning CPU waiting for a trap to be fixed up
	 */
#ifdef NO_O1_SCHED
	if (LwpNice >= -20 && LwpNice < 20)
	    current->nice = LwpNice;
#else
	set_user_nice(current, LwpNice);
#endif
    }
#endif

    elan3_swapin (ctxt, CTXT_NO_LWPS);

    spin_lock_irqsave (&dev->IntrLock, flags);

    /* If we're swapped out, and not detached (or exiting) then wait until we're swapped back in */
    /* since otherwise we could "spin" forever continually calling elan3_lwp() */
    if ((ctxt->Status & CTXT_SWAPPED_REASONS) && ! (ctxt->Status & (CTXT_DETACHED|CTXT_EXITING)))
	kcondvar_waitsig (&ctxt->Wait, &dev->IntrLock, &flags);

    for (;;)
    {
#if defined(DIGITAL_UNIX)
        if (thread_should_halt(current_thread()) || 
            CURSIG_CHECK(task_to_proc(current_thread()->task), u.np_uthread))
	{
	    PRINTF1 (ctxt, DBG_LWP, "elan3_lwp: exiting on %s\n", 
		     thread_should_halt(current_thread()) ? "halt" : "signal");
            break;
	}
#endif

	if (ctxt->Status & CTXT_SWAPPED_REASONS)
	{
	    PRINTF0 (ctxt, DBG_LWP, "elan3_lwp: exiting on swapped reasons\n");
	    break;
	}

	if (! (ctxt->inhibit))
	{
	    if (FixupNetworkErrors (ctxt, &flags) == ESUCCESS &&
	 	HandleExceptions (ctxt, &flags) == ESUCCESS &&
	    	RestartContext (ctxt, &flags) == ESUCCESS)
		{
    	            if (kcondvar_waitsig (&ctxt->Wait, &dev->IntrLock, &flags) == 0)
    	            {
	                PRINTF0 (ctxt, DBG_LWP, "elan3_lwp: exiting by kcondvar_wait_sig()\n");
	                break;
	            }
	        }
	}
	else
	{
	    printk("elan3_lwp :: skipping as inhibited\n");
	    if (kcondvar_waitsig (&ctxt->Wait, &dev->IntrLock, &flags) == 0)
	    {
	        PRINTF0 (ctxt, DBG_LWP, "elan3_lwp: exiting by kcondvar_wait_sig()\n");
	        break;
	    }
	}

    }

    /* Return EINVAL to elan3_syscall_lwp() when we want it to exit */
    res = (ctxt->Status & (CTXT_DETACHED|CTXT_EXITING)) ? EINVAL : 0;

    spin_unlock_irqrestore (&dev->IntrLock, flags);
    
    elan3_swapout (ctxt, CTXT_NO_LWPS);

    spin_lock_irqsave (&dev->IntrLock, flags);
    FixupNetworkErrors (ctxt, &flags);
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

void
SetInputterStateForContext (ELAN3_CTXT *ctxt, E3_uint32 Pend, E3_uint32 *Maskp)
{
    ELAN3_DEV  *dev          = NULL;
    int        new_disabled = 0;
    int	       ctxnum;

    ASSERT (ctxt != NULL);
    dev  = ctxt->Device;
    ASSERT (SPINLOCK_HELD (&dev->IntrLock));

    new_disabled = (ctxt->Input0Trap.State != CTXT_STATE_OK ||
		    ctxt->Input1Trap.State != CTXT_STATE_OK ||
		    (ctxt->Status & CTXT_INPUTTER_REASONS) != 0);
    

    ctxnum   = ctxt->Capability.cap_mycontext;

#ifndef __lock_lint  
    PRINTF2 (ctxt , DBG_IPROC, "SetInputterState: ctxnum %x %s attached\n", ctxnum, ctxt->Disabled ? "disabled " : "");
#endif /* __lock_lint */
        
    if (ctxt->Disabled != new_disabled)
    {
	PRINTF2 (ctxt, DBG_IPROC, "SetInputterState: ctxnum %x change %s\n", ctxnum, new_disabled ? "enabled to disabled" : "disabled to enabled");
	
	ctxt->Disabled = new_disabled;

	/* synchronize the context filter for this context */
	elan3mmu_set_context_filter (dev, ctxnum, new_disabled, Pend, Maskp);
    }
}

int
CheckCommandQueueFlushed (ELAN3_CTXT *ctxt, E3_uint32 cflags, int how, unsigned long *flags)
{
    ELAN3_DEV *dev    = ctxt->Device;
    int       delay  = 1;
    int i, SeenComQueueEmpty;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    ASSERT (cflags != DmaComQueueNotEmpty || dev->HaltDmaDequeueCount != 0);

    /*
     * Flush the command processor queues and poll the queue to see it it empties.
     */
    if (dev->FlushCommandCount++ == 0)
	SetSchedStatusRegister (dev, 0, NULL);

    /* 
     * Ensure previous writes have been flushed through the write buffers
     */
    wmb(); mmiob();

    /*
     * If the command processor traps,  or it's taking too long to observe
     * the queue as emtpy,  then we need to force the interrupt handler to 
     * run for us.  So queue a halt operation for the dma processor.
     */
    SeenComQueueEmpty = !(read_reg32 (dev, ComQueueStatus) & cflags);
    for (i = 20; i > 0 || (how & ISSUE_COMMAND_CANT_WAIT); i--)
    {
	if (SeenComQueueEmpty || (read_reg32 (dev, Exts.InterruptReg) & (INT_CProc | INT_ComQueue)))
	    break;
	
	mb();
	DELAY (delay);

	if ((delay <<= 1) == 0) delay = 1;

	SeenComQueueEmpty = !(read_reg32 (dev, ComQueueStatus) & cflags);
    }

    if (--dev->FlushCommandCount == 0)
	SetSchedStatusRegister (dev, 0, NULL);

    /*
     * If we've seen the command queue that we're interested in with nothing in it
     * and the command processor has not trapped then the commands we've
     * issued have been successfully processed.
     */
    if (SeenComQueueEmpty && ! (read_reg32 (dev, Exts.InterruptReg) & (INT_CProc | INT_ComQueue)))
    {
	PRINTF0 (ctxt, DBG_CMD, "CheckCommandQueueFlushed: observed dma queue empty and command proc not trapped\n");

	if (cflags == DmaComQueueNotEmpty && --dev->HaltDmaDequeueCount == 0)
	    SetSchedStatusRegister (dev, 0, NULL);

	return (ISSUE_COMMAND_OK);
    }

    if ((how & ISSUE_COMMAND_CANT_WAIT) != 0)
	return (ISSUE_COMMAND_WAIT);
    
    /*
     * Halt the dma processor and wait for it to halt,  if the command we've issued has
     * trapped then the interrupt handler will have moved it to the context structure.
     */
    PRINTF0 (ctxt, DBG_CMD, "CheckCommandQueueFlushed: waiting for dproc to halt\n");
    QueueHaltOperation (dev, 0, NULL, INT_DProcHalted, WakeupLwp, ctxt);
    while (! ctxt->Halted)
    {
	PRINTF1 (ctxt, DBG_CMD, "CheckCommandQueueFlushed: waiting for Halted - %d\n", ctxt->Halted);

	kcondvar_wait (&ctxt->HaltWait, &dev->IntrLock, flags);

	PRINTF1 (ctxt, DBG_CMD, "CheckCommandQueueFlushed: woken for Halted - %d\n", ctxt->Halted);
    }
    ctxt->Halted = 0;
    
    PRINTF0 (ctxt, DBG_CMD, "CheckCommandQueueFlushed: dproc halted, checking for trap\n");
    
    if (cflags == DmaComQueueNotEmpty && --dev->HaltDmaDequeueCount == 0)
	SetSchedStatusRegister (dev, 0, NULL);

    return (ELAN3_QUEUE_BACK_EMPTY (ctxt->CommandTrapQ) ? ISSUE_COMMAND_OK : ISSUE_COMMAND_TRAPPED);
}

int
WaitForCommandPort (ELAN3_CTXT *ctxt)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int	          res;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    if (ctxt->Status & CTXT_DETACHED)
	res = EINVAL;
    else 
    {
	if (! ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ) || (ctxt->Status & CTXT_OTHERS_REASONS))
	{
	    ctxt->Status |= CTXT_WAITING_COMMAND;
	    if (CTXT_IS_KERNEL(ctxt))
		kcondvar_wait (&ctxt->CommandPortWait, &dev->IntrLock, &flags);
	    else 
		kcondvar_waitsig (&ctxt->CommandPortWait, &dev->IntrLock, &flags);
	}
	
	res = (!ELAN3_QUEUE_EMPTY(ctxt->CommandTrapQ) || (ctxt->Status & CTXT_OTHERS_REASONS)) ? EAGAIN : 0;
    }
	
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

static char *
CommandName (int offset)
{
    switch (offset)
    {
    case offsetof (E3_CommandPort, PutDma):	return ("PutDma");
    case offsetof (E3_CommandPort, GetDma):	return ("GetDma");
    case offsetof (E3_CommandPort, RunThread):	return ("RunThread");
    case offsetof (E3_CommandPort, WaitEvent0):	return ("WaitEvent0");
    case offsetof (E3_CommandPort, WaitEvent1):	return ("WaitEvent1");
    case offsetof (E3_CommandPort, SetEvent):	return ("SetEvent");
    default:					return ("Bad Command");
    }
}

int
IssueCommand (ELAN3_CTXT *ctxt, unsigned cmdoff, E3_Addr value, int cflags)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int	          res;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    if ((! (cflags & ISSUE_COMMAND_FOR_CPROC) && !ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ)) || (ctxt->Status & CTXT_OTHERS_REASONS))
    {
	/*
	 * Cannot issue commands for non-cproc traps if command port is trapped, 
	 * nor if the dma/thread trap queues are full, or we're swapping out
	 */
	PRINTF2 (ctxt, DBG_CMD, "IssueCommand: %s %08x -> ISSUE_COMMAND_RETRY\n",
		 CommandName (cmdoff), value);

	res = ISSUE_COMMAND_RETRY;
    }
    else
    {
	PRINTF2 (ctxt, DBG_CMD, "IssueCommand: %s %08x -> ISSUE_COMMAND_OK\n",
		 CommandName (cmdoff), value);

	mb();							/* ensure writes to main memory completed */
	writel (value, (void *)(ctxt->CommandPort + cmdoff));		/* issue command */
	mmiob();						/* and flush through IO writes */

	res = ISSUE_COMMAND_OK;
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);
    
    return (res);
}

int
IssueDmaCommand (ELAN3_CTXT *ctxt, E3_Addr value, void *item, int how)
{
    ELAN3_DEV     *dev    = ctxt->Device;
    int	          res;
    unsigned long flags;

    /*
     * Since we may be issuing a command that could trap, and we're interested in
     * the outcome, the command port trap resolving code must be locked out.
     */
    kmutex_lock (&ctxt->CmdLock);
    spin_lock_irqsave (&dev->IntrLock, flags);

    if ((! (how & ISSUE_COMMAND_FOR_CPROC) && !ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ)) || (ctxt->Status & CTXT_OTHERS_REASONS))
    {
	PRINTF2 (ctxt, DBG_CMD, "IssueDmaCommand: PutDma %08x [%p] -> ISSUE_COMMAND_RETRY\n", value, item);

	/*
	 * Cannot issue commands for non-cproc traps if command port is trapped, 
	 * nor if the dma/thread trap queues are full, or we're swapping out
	 */
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	kmutex_unlock (&ctxt->CmdLock);
	return (ISSUE_COMMAND_RETRY);
    }
    
    ASSERT (item == NULL || ctxt->CommandPortItem == NULL);

    /*
     * Stop the DMA processor from removing entries from the 
     * command port, and force the command processor to do this.
     * This means that if a trap occurs then it will be the command
     * processor that traps.
     */
    if (dev->HaltDmaDequeueCount++ == 0)
	SetSchedStatusRegister (dev, 0, NULL);

    PRINTF2 (ctxt, DBG_CMD, "IssueDmaCommand: PutDma %08x [%p]\n", value, item);

    /*
     * Always issue the DMA to the 'write' command,  since we've asserted HaltDmaDequeue
     * the command processor will read the descriptor and transfer it to the run queue. 
     * The command processor looks at the dma_direction field to determine whether it is
     * a read or a write and whether to alter the dma_souce of the descriptr on the run 
     * queue
     */
    mb();							/* ensure writes to main memory ccompleted */
    writel (value, (void *) (ctxt->CommandPort + offsetof (E3_CommandPort, PutDma)));
    mmiob();							/* and flush through IO writes */
    
    res = CheckCommandQueueFlushed (ctxt, DmaComQueueNotEmpty, how, &flags);

    if (res == ISSUE_COMMAND_TRAPPED)
    {
	PRINTF2 (ctxt, DBG_CMD, "IssueDmaCommand: PutDma %08x [%p] -> ISSUE_COMMAND_TRAPPED\n", value, item);
	/*
	 * Remember the item we're issueing so that if the command port traps the item will not
	 * get freed off until the descriptor has been read after the command trap has been fixed
	 * up.
	 */
	if (item != NULL)
	    ctxt->CommandPortItem = item;
    }

    spin_unlock_irqrestore (&dev->IntrLock, flags);
    kmutex_unlock (&ctxt->CmdLock);

    return (res);
}

int
WaitForDmaCommand (ELAN3_CTXT *ctxt, void *item, int how)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int           res;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    res = CheckCommandQueueFlushed (ctxt, DmaComQueueNotEmpty, how, &flags);

    if (res == ISSUE_COMMAND_TRAPPED && item != NULL)
	ctxt->CommandPortItem = item;

    spin_unlock_irqrestore (&dev->IntrLock, flags);
    
    return (res);
}

void
FixupEventTrap (ELAN3_CTXT *ctxt, int proc, void *trap, E3_uint32 TrapType, E3_FaultSave_BE *FaultSaveArea, int flags)
{
    ASSERT (! CTXT_IS_KERNEL (ctxt));

    /*
     * This code re-issues the part of the set event that trapped.
     */
    switch (TrapType)
    {
    case MI_ChainedEventError:
	ElanException (ctxt, EXCEPTION_CHAINED_EVENT, proc, trap, FaultSaveArea->s.EventAddress);
	break;
	

    case MI_SetEventReadWait:
	/*
	 * Fault occured on the read for the event location. Just re-issue
	 * setevent using EventAddress in E3_FaultSave
	 */
	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_SetEventReadWait: re-issuing setevent %08x\n", 
		 FaultSaveArea->s.EventAddress);
	
	ReissueEvent (ctxt, (E3_Addr) FaultSaveArea->s.EventAddress, flags);
	break;

    case MI_DoSetEvent:
    {
	/*
	 * Fault occured because the block write of a block copy event trapped.
	 * Must grab the event type, source and dest then simulate the block copy and then
	 * perform the set. Once the block copy is started the event location cannot be read
	 * again.
	 */
	E3_Event *EventPtr  = (E3_Event *) elan3mmu_mainaddr (ctxt->Elan3mmu, FaultSaveArea->s.EventAddress);
	E3_uint32 EventType = fuword (&EventPtr->ev_Type);
	
	/*
	 * Check that the event has the block copy bit
	 * set in it,  since we couldn't trap here if it
	 * didn't
	 */
	if ((EventType & EV_TYPE_BCOPY) != EV_TYPE_BCOPY)
	{
	    PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_DoSetEvent: Unexpected type=%x\n", EventType);
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	    break;
	}
	
	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_DoSetEvent: RunEventType %x\n", EventType);

	if (RunEventType (ctxt, FaultSaveArea, EventType))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);

	break;
    }
    
    case MI_ThreadUpdateNonSysCntxBack:
    case MI_ThreadUpdateSysCntxBack:
    {
	/*
	 * Fault occured because the block write of a block copy event trapped.
	 * Must grab the event type, source and dest then simulate the block copy and then
	 * run the thread. Once the block copy is started the event location cannot be read
	 * again.
	 */
	E3_Event *EventPtr = (E3_Event *) elan3mmu_mainaddr (ctxt->Elan3mmu, FaultSaveArea->s.EventAddress);
	E3_uint32 EventType = fuword (&EventPtr->ev_Type);

	/*
	 * Check for the correct EventPtr type
	 */
	if ((EventType & (EV_TYPE_MASK_THREAD|EV_TYPE_MASK_BCOPY)) != (EV_TYPE_BCOPY | EV_TYPE_THREAD))
	{
	    PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_ThreadUpdateCntx0Back: Unexpected type=%x for setevent trap. Should be thread\n", EventType);
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	    break;
	}
	
	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_ThreadUpdateCntx0Back: RunEventType %x\n", EventType);
	if (RunEventType (ctxt, FaultSaveArea, EventType))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	break;
    }
    
    case MI_EventIntUpdateBPtr:
    {
	/*
	 * Fault occured because the block write of a block copy event trapped.
	 * Must grab the event type, source and dest then simulate the block copy and then
	 * run the dma. Once the block copy is started the event location cannot be read
	 * again.
	 */
	E3_Event *EventPtr = (E3_Event *) elan3mmu_mainaddr (ctxt->Elan3mmu, FaultSaveArea->s.EventAddress);
	E3_uint32 EventType = fuword (&EventPtr->ev_Type);

	/*
	 * Check for the correct EventPtr type
	 */
	if ((EventType & (EV_TYPE_MASK_EVIRQ|EV_TYPE_MASK_BCOPY)) != (EV_TYPE_BCOPY | EV_TYPE_EVIRQ))
	{
	    PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_EventIntUpdateBPtr: Unexpected type=%x\n", EventType);
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	    break;
	}

	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_EventIntUpdateBPtr: RunEventType %x\n", EventType);
	if (RunEventType(ctxt, FaultSaveArea, EventType))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	break;
    }
    
    case MI_RunDmaDesc:
    {
	/*
	 * Fault occured because the block write of a block copy event trapped.
	 * Must grab the event type, source and dest then simulate the block copy and then
	 * run the dma. Once the block copy is started the event location cannot be read
	 * again.
	 */
	E3_Event *EventPtr = (E3_Event *) elan3mmu_mainaddr (ctxt->Elan3mmu, FaultSaveArea->s.EventAddress);
	E3_uint32 EventType = fuword (&EventPtr->ev_Type);

	/*
	 * Check for the correct EventPtr type
	 */
	if ((EventType & (EV_TYPE_MASK_DMA|EV_TYPE_MASK_BCOPY)) != (EV_TYPE_BCOPY | EV_TYPE_DMA))
	{
	    PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_RunDmaDesc: Unexpected type=%x\n", EventType);
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	    break;
	}

	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_RunDmaDesc: RunEventType %x\n", EventType);
	if (RunEventType(ctxt, FaultSaveArea, EventType))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	break;
    }
    
    case MI_WaitForCntxDmaDescRead:
    case MI_WaitForNonCntxDmaDescRead:
	/*
	 * Fault occured on the read of the dma descriptor. Run dma using the
	 * Fault Address in FaultSave.
	 */
	PRINTF1 (ctxt, DBG_EVENT, "FixupEventTrap: MI_WaitForCntxDmaDescRead: re-issue dma at %08x\n", FaultSaveArea->s.FaultAddress);
	
	RestartDmaPtr (ctxt, FaultSaveArea->s.FaultAddress);
	break;
    
    case MI_FinishedSetEvent:
	/*
	 * Fault occured because the block write of a block copy event trapped.
	 * Simulate the block copy.
	 */
	if (SimulateBlockCopy (ctxt, FaultSaveArea->s.EventAddress))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	break;
	
    case MI_BlockCopyEvent:
    case MI_BlockCopyWaitForReadData:
    {
	/*
	 * Fault occured on the read or write of the data for a block copy
	 * event. Simulate the block copy using EventAddress in E3_FaultSave. Must also sample
	 * the event type and then perform a run.
	 */
	E3_Event *EventPtr = (E3_Event *) elan3mmu_mainaddr (ctxt->Elan3mmu, FaultSaveArea->s.EventAddress);
	E3_uint32 EventType = fuword (&EventPtr->ev_Type);

	PRINTF0 (ctxt, DBG_EVENT, "FixupEventTrap: MI_BlockCopyWaitForReadData: BCopy read fault in BCopy event. Simulating BCopy.\n");
	
	if (RunEventType(ctxt, FaultSaveArea, EventType))
	    ElanException (ctxt, EXCEPTION_BAD_EVENT, proc, trap, FaultSaveArea, TrapType);
	break;
    }
    
    case MI_EventQueueOverflow:
    case MI_ThreadQueueOverflow:
    case MI_DmaQueueOverflow:
	/* XXXX: should handle queue overflow */
	PRINTF0 (ctxt, DBG_EVENT, "FixupEventTrap: Queue overflow\n");

	ElanException (ctxt, EXCEPTION_QUEUE_OVERFLOW, proc, trap, FaultSaveArea, TrapType);
	break;

    default:
	ElanException (ctxt, EXCEPTION_BUS_ERROR, proc, trap, FaultSaveArea, TrapType);
	break;
    }
}

int
SimulateBlockCopy (ELAN3_CTXT *ctxt, E3_Addr EventAddress)
{
    E3_Addr  SourcePtrElan;
    E3_Addr  DestPtrElan;
    unsigned DataType;
    int      i;

    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	ElanException (ctxt, EXCEPTION_FAULTED, EVENT_PROC, NULL, EventAddress);
	return (TRUE);
    }

    SourcePtrElan = ELAN3_OP_LOAD32 (ctxt, EventAddress + offsetof (E3_BlockCopyEvent, ev_Source));
    DestPtrElan   = ELAN3_OP_LOAD32 (ctxt, EventAddress + offsetof (E3_BlockCopyEvent, ev_Dest));
    DataType      = DestPtrElan & EV_BCOPY_DTYPE_MASK;
    DestPtrElan  &= ~EV_BCOPY_DTYPE_MASK;


    PRINTF3 (ctxt, DBG_EVENT, "SimulateBlockCopy: Event %08x SourcePtr %08x DestPtr %08x\n",
	     EventAddress, SourcePtrElan, DestPtrElan);

    if (SourcePtrElan & EV_WCOPY)
	ELAN3_OP_STORE32 (ctxt, DestPtrElan, SourcePtrElan);
    else
    {
	/*
	 * NOTE: since the block copy could be to sdram, we issue the writes backwards,
	 *       except we MUST ensure that the last item in the block is written last.
	 */
#if defined(__LITTLE_ENDIAN__)
	/*
	 * For little endian cpu's we don't need to worry about the data type.
	 */
	for (i = E3_BLK_SIZE-(2*sizeof (E3_uint64)); i >= 0; i -= sizeof (E3_uint64))
	    ELAN3_OP_STORE64 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD64 (ctxt, SourcePtrElan + i));

	i = E3_BLK_SIZE - sizeof (E3_uint64);
	ELAN3_OP_STORE64 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD64 (ctxt, SourcePtrElan + i));
#else
	switch (DataType)
	{
	case EV_TYPE_BCOPY_BYTE:
	    for (i = E3_BLK_SIZE-(2*sizeof (E3_uint8)); i >= 0; i -= sizeof (E3_uint8))
		ELAN3_OP_STORE8 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD8 (ctxt, SourcePtrElan + i));
	    
	    i = E3_BLK_SIZE - sizeof (E3_uint8);
	    ELAN3_OP_STORE8 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD8 (ctxt, SourcePtrElan + i));
	    break;

	case EV_TYPE_BCOPY_HWORD: 
	    for (i = E3_BLK_SIZE-(2*sizeof (E3_uint16)); i >= 0; i -= sizeof (E3_uint16))
		ELAN3_OP_STORE16 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD16 (ctxt, SourcePtrElan + i));
	    
	    i = E3_BLK_SIZE - sizeof (E3_uint16);
	    ELAN3_OP_STORE16 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD16 (ctxt, SourcePtrElan + i));
	    break;
	    
	case EV_TYPE_BCOPY_WORD:  
	    for (i = E3_BLK_SIZE-(2*sizeof (E3_uint32)); i >= 0; i -= sizeof (E3_uint32))
		ELAN3_OP_STORE32 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD32 (ctxt, SourcePtrElan + i));
	    
	    i = E3_BLK_SIZE - sizeof (E3_uint32);
	    ELAN3_OP_STORE32 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD32 (ctxt, SourcePtrElan + i));
	    break;
	    
	case EV_TYPE_BCOPY_DWORD: 
	    for (i = E3_BLK_SIZE-(2*sizeof (E3_uint64)); i >= 0; i -= sizeof (E3_uint64))
		ELAN3_OP_STORE64 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD64 (ctxt, SourcePtrElan + i));
	    
	    i = E3_BLK_SIZE - sizeof (E3_uint64);
	    ELAN3_OP_STORE64 (ctxt, DestPtrElan + i, ELAN3_OP_LOAD64 (ctxt, SourcePtrElan + i));
	    break;
	}
#endif
    }
    ELAN3_OP_END_FAULT_CHECK (ctxt);

    return (FALSE);
}

void
ReissueEvent (ELAN3_CTXT *ctxt, E3_Addr addr, int flags)
{
    PRINTF1 (ctxt, DBG_CMD, "ReissueEvent : Event=%08x\n", addr);

    if (IssueCommand (ctxt, offsetof (E3_CommandPort, SetEvent), addr, flags) == ISSUE_COMMAND_RETRY)
    {
	PRINTF1 (ctxt, DBG_CMD, "ReissueEvent: queue event %08x\n", addr);

	kmutex_lock (&ctxt->SwapListsLock);
	ctxt->ItemCount[LIST_SETEVENT]++;
	ELAN3_OP_PUT_WORD_ITEM (ctxt, LIST_SETEVENT, addr);
	kmutex_unlock (&ctxt->SwapListsLock);
    }
}

int
SetEventsNeedRestart (ELAN3_CTXT *ctxt)
{
    return (ctxt->ItemCount[LIST_SETEVENT] != 0);
}

void
RestartSetEvents (ELAN3_CTXT *ctxt)
{
    void     *item;
    E3_uint32 EventPointer;

    kmutex_lock (&ctxt->SwapListsLock);
    
    while (ctxt->ItemCount[LIST_SETEVENT])
    {
	if (! ELAN3_OP_GET_WORD_ITEM (ctxt, LIST_SETEVENT, &item, &EventPointer))
	    ctxt->ItemCount[LIST_SETEVENT] = 0;
	else
	{
	    if (IssueCommand (ctxt, offsetof (E3_CommandPort, SetEvent), EventPointer, FALSE) == ISSUE_COMMAND_RETRY)
	    {
		ELAN3_OP_PUTBACK_ITEM (ctxt, LIST_SETEVENT, item);
		kmutex_unlock (&ctxt->SwapListsLock);
		return;
	    }
	    
	    ctxt->ItemCount[LIST_SETEVENT]--;
	    ELAN3_OP_FREE_WORD_ITEM (ctxt, item);
	}
    }
    kmutex_unlock (&ctxt->SwapListsLock);
}

int
RunEventType(ELAN3_CTXT *ctxt, E3_FaultSave_BE *FaultSaveArea, E3_uint32 EventType)
{
    int failed = FALSE;

    if ((EventType & EV_TYPE_BCOPY) != 0)
	failed = SimulateBlockCopy(ctxt, FaultSaveArea->s.EventAddress);
    
    if ((EventType & EV_TYPE_MASK) == EV_TYPE_THREAD)
	ReissueStackPointer (ctxt, EventType & ~(EV_TYPE_MASK_THREAD|EV_TYPE_MASK_BCOPY));
    else if ((EventType & EV_TYPE_MASK) == EV_TYPE_DMA)
	RestartDmaPtr (ctxt, EventType & ~(EV_TYPE_MASK_DMA|EV_TYPE_MASK_BCOPY));
    else if ((EventType & EV_TYPE_EVIRQ) != 0)
	QueueEventInterrupt (ctxt, EventType & ~(EV_TYPE_MASK_EVIRQ|EV_TYPE_MASK_BCOPY));
    else /* Chained event */
    {
	if ((EventType & ~EV_TYPE_BCOPY) != 0) /* not null setevent */
	    ReissueEvent (ctxt, EventType & ~(EV_TYPE_MASK_CHAIN|EV_TYPE_MASK_BCOPY), FALSE);
    }

    return (failed);
}

void
WakeupLwp (ELAN3_DEV *dev, void *arg)
{
    ELAN3_CTXT    *ctxt = (ELAN3_CTXT *) arg;
    unsigned long flags;

    PRINTF1 (ctxt, DBG_INTR, "WakeupLwp: %d\n", SPINLOCK_HELD (&dev->IntrLock));

    spin_lock_irqsave (&dev->IntrLock, flags);
    ctxt->Halted = 1;
    kcondvar_wakeupone (&ctxt->HaltWait, &dev->IntrLock);

    PRINTF0 (ctxt, DBG_INTR, "WakeupLwp: woken up context\n");
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

void
QueueEventInterrupt (ELAN3_CTXT *ctxt, E3_uint32 cookie)
{
    ELAN3_DEV     *dev = ctxt->Device;
    unsigned long flags;

    PRINTF1 (ctxt, DBG_EVENT, "QueueEventInterrupt: cookie %08x\n", cookie);

    if (ELAN3_OP_EVENT (ctxt, cookie, OP_INTR) == OP_DEFER)
    {
	spin_lock_irqsave (&ctxt->Device->IntrLock, flags);

	if (ELAN3_QUEUE_REALLY_FULL (ctxt->EventCookieQ))
	{
	    ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
	    StartSwapoutContext (ctxt, 0, NULL);
	}
	else
	{
	    *(ELAN3_QUEUE_BACK (ctxt->EventCookieQ, ctxt->EventCookies)) = cookie;
	    
	    ELAN3_QUEUE_ADD (ctxt->EventCookieQ);
	    kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
	    if (ELAN3_QUEUE_FULL (ctxt->EventCookieQ))
	    {
		ctxt->Status |= CTXT_EVENT_QUEUE_FULL;
		StartSwapoutContext (ctxt, 0, NULL);
	    }
	}
	spin_unlock_irqrestore (&ctxt->Device->IntrLock, flags);
    }
}

int
ElanException (ELAN3_CTXT *ctxt, int type, int proc, void *trap, ...)
{
    int     res;
    va_list ap;

    va_start (ap, trap);

    PRINTF2 (ctxt, DBG_FN, "ElanException: proc %d type %d\n", proc, type);

    res = ELAN3_OP_EXCEPTION (ctxt, type, proc, trap, ap);

    va_end (ap);
    
    return (res);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
