/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: cproc.c,v 1.46 2004/02/10 15:05:10 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/cproc.c,v $ */


#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/vmseg.h>

void
HandleCProcTrap (ELAN3_DEV *dev, E3_uint32 Pend, E3_uint32 *Maskp)
{
    E3_FaultSave_BE     FaultSave;
    CProcTrapBuf_BE	TrapBuf;
    COMMAND_TRAP       *trap;
    ELAN3_CTXT	       *ctxt;
    sdramaddr_t         CurrTrap;
    sdramaddr_t         LastTrapAddr;
    int    		NTrapEntries;
    int			NewPend;
    unsigned long       flags;

    /* 
     * Temporarily mask out the command processor interrupt, since
     * we may cause it be re-asserted when we re-issue the commands
     * from the overflow queue area.
     */
    DISABLE_INT_MASK (dev, INT_CProc | INT_ComQueue);

    NewPend = read_reg32 (dev, Exts.InterruptReg);

    do {
	if (NewPend & INT_ComQueue)
	{
	    if ((read_reg32 (dev, ComQueueStatus) & ComQueueError) != 0)
	    {
		printk ("elan%d: InterruptReg=%x ComQueueStatus=%x\n", dev->Instance,
			read_reg32 (dev, Exts.InterruptReg), read_reg32 (dev, ComQueueStatus));
		panic ("elan: command queue has overflowed !!");
		/* NOTREACHED */
	    }

	    BumpStat (dev, ComQueueHalfFull);

	    /*
	     * Capture the other cpus and stop the threads processor then
	     * allow the command processor to eagerly flush the command queue.
	     */
	    dev->FlushCommandCount++; dev->HaltThreadCount++;
	    SetSchedStatusRegister (dev, Pend, Maskp);

	    CAPTURE_CPUS();

	    while ((read_reg32 (dev, ComQueueStatus) & ComQueueNotEmpty) != 0)
		mb();
	    
	    /*
	     * Let the threads processor run again, and release the cross call.
	     */
	    RELEASE_CPUS();

	    dev->FlushCommandCount--; dev->HaltThreadCount--;
	    SetSchedStatusRegister (dev, Pend, Maskp);

	    /*
	     * Re-sample the interrupt register to see if the command processor
	     * has trapped while flushing the queue.  Preserve the INT_ComQueue
	     * bit, so we can clear the ComQueueStatus register later.
	     */
	    NewPend = (read_reg32 (dev, Exts.InterruptReg) | INT_ComQueue);
	}
	
	CurrTrap = dev->CommandPortTraps[dev->CurrentCommandPortTrap];
	
	if (NewPend & INT_CProc)
	{
	    BumpStat (dev, CProcTraps);

	    /*
	     * Copy the MMU Fault Save area and zero it out for future traps.
	     */
	    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, CProc), &FaultSave, sizeof (E3_FaultSave));
	    elan3_sdram_zeroq_sdram      (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, CProc), sizeof (E3_FaultSave));

	    /*
	     * First entry in the cproc trap save area is the value of Areg and Breg for the
	     * uWord before the address fault.
	     */
	    TrapBuf.Align64 = elan3_sdram_readq (dev, CurrTrap); CurrTrap += sizeof (TrapBuf.Align64);

	    ctxt = ELAN3_DEV_CTX_TABLE(dev, (TrapBuf.r.Breg >> 16));
	    if (ctxt == NULL)
	    {
		PRINTF2 (DBG_DEVICE, DBG_INTR, "HandleCProcTrap: context invalid [%08x.%08x]\n", TrapBuf.r.Areg, TrapBuf.r.Breg);
		BumpStat (dev, InvalidContext);
	    }
	    else
	    {
		if (ELAN3_QUEUE_REALLY_FULL (ctxt->CommandTrapQ))
		{
		    if ((ctxt->Status & CTXT_COMMAND_OVERFLOW_ERROR) == 0)
		    {
			ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
			StartSwapoutContext (ctxt, Pend, Maskp);
		    }
		}
		else
		{
		    trap = ELAN3_QUEUE_BACK (ctxt->CommandTrapQ, ctxt->CommandTraps);
		    
		    trap->FaultSave     = FaultSave;
		    trap->Status.Status = read_reg32 (dev, Exts.CProcStatus.Status);
		    trap->TrapBuf       = TrapBuf;
		    
		    /*
		     * The command processor does not stop after it has trapped. It will continue
		     * to save commands for other contexts into the commands port save area.
		     * The valid context for the trap is held in FaultSave. As some of this
		     * trap code uses the context in the status register the local copy must be
		     * updated with the trap context.
		     */
		    trap->Status.s.Context = (TrapBuf.r.Breg >> 16);
		    
		    PRINTF4 (ctxt, DBG_INTR, "HandleCProcTrap: WakeupFnt=%x Cntx=%x SuspAddr=%x TrapType=%s\n",
			     trap->Status.s.WakeupFunction, trap->Status.s.Context,
			     trap->Status.s.SuspendAddr, MiToName(trap->Status.s.TrapType));
		    PRINTF2 (ctxt, DBG_INTR, "HandleCProcTrap: Areg=%08x Breg=%08x\n", 
			     trap->TrapBuf.r.Areg, trap->TrapBuf.r.Breg);
		    
		    if (ELAN3_OP_CPROC_TRAP (ctxt, trap) == OP_DEFER)
		    {
			ELAN3_QUEUE_ADD (ctxt->CommandTrapQ);
			
			PRINTF1 (ctxt, DBG_INTR, "HandleCProcTrap: setting Command Flag at %p to 1\n", &ctxt->FlagPage->CommandFlag);
			
			ctxt->FlagPage->CommandFlag = 1;
			
			kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
		    }
		}

		UnloadCommandPageMapping (ctxt);
	    }
	}
	
	/*
	 * Now change the CommandPortTrap queue.
	 * Must stop the command processor, wait for it to stop, find the final
	 * entry in the current cproc trap save area, reset the comm port
	 * trap save address to the other queue, clear the command port interrupt and
	 * set it running normally again, and then let it go again. This is not very
	 * time critical but it would be a good idea to prevent a higher priority
	 * interrupt from slowing down the process to prevent to fifos filling.
	 */
	spin_lock_irqsave (&dev->CProcLock, flags);

	SET_SCHED_STATUS (dev, CProcStop);

	while ((read_reg32 (dev, Exts.SchCntReg) & CProcStopped) == 0)
	{
	    PRINTF0 (DBG_DEVICE, DBG_INTR, "HandleCProcTrap: waiting for command processor to stop\n");
	    mb();
	}
	
	/*
	 * Remember how many entries are in the saved command queue,  and 
	 * re-initialise it, before restarting the command processor.
	 */
	NTrapEntries = (read_reg32 (dev, CProc_TrapSave_Addr) - dev->CommandPortTraps[dev->CurrentCommandPortTrap])/sizeof (E3_uint64);
	LastTrapAddr = dev->CommandPortTraps[dev->CurrentCommandPortTrap] + NTrapEntries*sizeof (TrapBuf);

	dev->CurrentCommandPortTrap ^= 1;
	write_reg32 (dev, CProc_TrapSave_Addr, dev->CommandPortTraps[dev->CurrentCommandPortTrap]);

	PRINTF1 (DBG_DEVICE, DBG_INTR, "HandleCProcTrap: command trap queue has %d entries\n", NTrapEntries);

	if (NTrapEntries > ELAN3_COMMAND_TRAP_SIZE/sizeof (E3_uint64))
	    panic ("HandleCProcTrap: command trap queue has overflowed\n");
	
	if (NewPend & INT_CProc)
	{
	    /*
	     * Clear the CProc interrupt and set it running normally again. Nothing should
	     * be running now that could issue commands apart from this trap handler.
	     */
	    PULSE_SCHED_STATUS (dev, RestartCProc);
	}
	
	if (NewPend & INT_ComQueue)
	{
	    /*
	     * Write any value here to clear out the half full and error bits of the command
	     * overflow queues. This will also remove the overflow interrupt.
	     */
	    write_reg32 (dev, ComQueueStatus, 0);
	}
	
	/*
	 * And let the command processor start again
	 */
	CLEAR_SCHED_STATUS (dev, CProcStop);
	
	/*
	 * Now re-issue all the commands that were issued after the command port trapped.
	 * Should halt the dma processor and force command sto be put onto the run queues
	 * to ensure that a remote re-issued command is handled correctly. NOTE it is
	 * not necessary to wait for the dma processor to stop and this will reduce the
	 * performance impact. As CProcHalt is asserted all commands will be flushed
	 * to the queues.
	 */
	dev->HaltDmaDequeueCount++; dev->FlushCommandCount++;
	SetSchedStatusRegister (dev, Pend, Maskp);
	
	/*
	 * XXXX: should we do a capture/release if the trap overflow
	 *       area has a "large" number of commands in it,  since
	 *       we will just stuff them all back in, together with 
	 *       all those issued by the other cpus/thread processors.
	 */
	while (CurrTrap != LastTrapAddr)
	{
	    /* Read the next saved (but not trapped) command */
	    TrapBuf.Align64 = elan3_sdram_readq (dev, CurrTrap); CurrTrap += sizeof (TrapBuf);
	    

	    ctxt = ELAN3_DEV_CTX_TABLE(dev, (TrapBuf.s.ContextType >> 16));
	    
	    if (ctxt == NULL)
	    {
		PRINTF1 (DBG_DEVICE, DBG_INTR, "HandleCProcTrap: context %x invalid\n", TrapBuf.s.ContextType >> 16);
		BumpStat (dev, InvalidContext);
	    }
	    else
	    {
		if (!ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ) || (ctxt->Status & CTXT_OTHERS_REASONS))
		{
		    PRINTF3 (ctxt, DBG_INTR, "HandleCProcTrap: save command %x context %x - %08x\n",
			     (TrapBuf.s.ContextType>>3) & 0x3ff, TrapBuf.s.ContextType >> 17, TrapBuf.s.Addr);
		    
		    if (ELAN3_QUEUE_REALLY_FULL (ctxt->CommandQ))
		    {
			ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
			StartSwapoutContext (ctxt, Pend, Maskp);
		    }
		    else
		    {
			*ELAN3_QUEUE_BACK(ctxt->CommandQ, ctxt->Commands) = TrapBuf;

			ELAN3_QUEUE_ADD (ctxt->CommandQ);
		    }
		    continue;
		}
		
		/* Reissue the command to the command port for this context */
		PRINTF2 (ctxt, DBG_INTR, "HandleCProcTrap: re-issue command %x - %08x\n",
			 (TrapBuf.s.ContextType>>5) & 0xff, TrapBuf.s.Addr);
		
		mb();
		if (ELAN3_OP_CPROC_REISSUE(ctxt, &TrapBuf) != OP_HANDLED)
		    ((E3_uint32 *) ctxt->CommandPort)[(TrapBuf.s.ContextType>>5) & 0xff] = TrapBuf.s.Addr;
		mmiob();
	    }
	}
	
	while ((read_reg32 (dev, ComQueueStatus) & ComQueueNotEmpty) != 0)
	{
	    PRINTF0 (DBG_DEVICE, DBG_INTR, "HandleCProcTrap: waiting for queues to empty after reissueing commands\n");
	    mb();
	}
	
	dev->HaltDmaDequeueCount--; dev->FlushCommandCount--;
	SetSchedStatusRegister (dev, Pend, Maskp);
	
	spin_unlock_irqrestore (&dev->CProcLock, flags);

	/*
	 * Re-read the interrupt register and see if we've got another command
	 * port interrupt
	 */
	NewPend = read_reg32 (dev, Exts.InterruptReg);
    } while ((NewPend & (INT_CProc | INT_ComQueue)) != 0);


    /*
     * Re-enable the command processor interrupt as we've finished 
     * polling it.
     */
    ENABLE_INT_MASK (dev, INT_CProc | INT_ComQueue);
}

void
ResolveCProcTrap (ELAN3_CTXT *ctxt)
{
    ELAN3_DEV     *dev = ctxt->Device;
    COMMAND_TRAP *trap;
    int	          res;
    unsigned long flags;

    kmutex_lock (&ctxt->CmdLock);
    spin_lock_irqsave (&dev->IntrLock, flags);

    while (! ELAN3_QUEUE_BACK_EMPTY (ctxt->CommandTrapQ))
    {
	trap = ELAN3_QUEUE_MIDDLE(ctxt->CommandTrapQ, ctxt->CommandTraps);
	spin_unlock_irqrestore (&dev->IntrLock, flags);

	switch (trap->Status.s.TrapType)
	{
	case MI_EventIntUpdateBPtr:
	case MI_ChainedEventError:
	case MI_EventQueueOverflow:
	case MI_ThreadQueueOverflow:
	case MI_DmaQueueOverflow:
	    PRINTF1 (ctxt, DBG_CPROC, "ResolveCProcTrap: %s\n", MiToName (trap->Status.s.TrapType));
	    break;
	    
	default:
	    /* All other traps are MMU related, we should have a fault address and FSR */
	    if ((res = elan3_pagefault (ctxt, &trap->FaultSave, 1)) != ESUCCESS)
	    {
		PRINTF1 (ctxt, DBG_CPROC, "ResolveCProcTrap: elan3_pagefault failed for address %08x\n", 
			 trap->FaultSave.s.FaultAddress);
		ElanException (ctxt, EXCEPTION_INVALID_ADDR, COMMAND_PROC, trap, &trap->FaultSave, res);
		
		/* Set the trap type to 0 so the command does not get re-issued */
		trap->Status.s.TrapType = 0;
	    }
	    break;
	}
	
	spin_lock_irqsave (&dev->IntrLock, flags);

	ELAN3_QUEUE_CONSUME (ctxt->CommandTrapQ);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);
    kmutex_unlock (&ctxt->CmdLock);
}

int
RestartCProcTrap (ELAN3_CTXT *ctxt)
{
    ELAN3_DEV     *dev      = ctxt->Device;
    COMMAND_TRAP  trap;
    void	 *item;
    int		  res;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    while (! ELAN3_QUEUE_FRONT_EMPTY (ctxt->CommandTrapQ))
    {
	trap = (*ELAN3_QUEUE_FRONT (ctxt->CommandTrapQ, ctxt->CommandTraps));
	ELAN3_QUEUE_REMOVE (ctxt->CommandTrapQ);
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	
	BumpUserStat (ctxt, CProcTraps);

	switch (trap.Status.s.TrapType)
	{
	case 0:
	    res = ISSUE_COMMAND_OK;
	    break;
	    
	case MI_WaitForWaitEventDesc:
	    /*
	     * Fault occured on the read of wait event descriptor for wait event type 0.
	     * Fault already fixed. Just re-issue the wait command. Wait event descriptor addr
	     * is in the Areg save value.
	     */
	    PRINTF1 (ctxt, DBG_CPROC, "RestartCProcTrap: WaitEvent type0 desc read fault %08x\n", 
		     trap.TrapBuf.r.Areg);
	    
	    res = IssueCommand (ctxt, offsetof (E3_CommandPort, WaitEvent0), trap.TrapBuf.r.Areg, ISSUE_COMMAND_FOR_CPROC);
	    break;

	case MI_WaitForEventReadTy0:
	    /*
	     * Fault occured on the read of event location for wait event type 0.
	     * Fault already fixed. Just re-issue the wait command. Wait event descriptor addr
	     * is in the Areg save value.
	     */
	    PRINTF1 (ctxt, DBG_CPROC, "RestartCProcTrap: WaitEvent type0 event loc fault %08x\n",
		     trap.TrapBuf.r.Areg);
	    
	    res = IssueCommand (ctxt, offsetof (E3_CommandPort, WaitEvent0), trap.TrapBuf.r.Areg, ISSUE_COMMAND_FOR_CPROC);
	    break;
	    
	case MI_WaitForEventReadTy1:
	    /*
	     * Fault occured on the read of the event location for wait event type 1.
	     * Areg has the original ptr and count.
	     * Fault already fixed. Just re-issue the wait command using Areg and context.
	     */
	    PRINTF1 (ctxt, DBG_CPROC, "RestartCProcTrap: WaitEvent type1 event location read fault %08x\n",
		     trap.TrapBuf.r.Areg);
	    res = IssueCommand (ctxt, offsetof (E3_CommandPort, WaitEvent1), trap.TrapBuf.r.Areg, ISSUE_COMMAND_FOR_CPROC);
	    break;
	    
	case MI_WaitForCntxDmaDescRead:
	case MI_WaitForNonCntxDmaDescRead:
	    /*
	     * Fault occured on the read of the dma descriptor. Run dma using the
	     * Fault Address in FaultSave.
	     */
	    PRINTF1 (ctxt, DBG_CPROC, "RestartCProcTrap: MI_WaitForCntxDmaDescRead: re-issue dma at %08x\n", 
		     trap.FaultSave.s.FaultAddress);
	    
	    res = IssueDmaCommand (ctxt, trap.FaultSave.s.FaultAddress, NULL, ISSUE_COMMAND_FOR_CPROC);
	    break;
	    
	default:
	    /*
	     * Assume the fault will be fixed by FixupEventTrap.
	     */
	    FixupEventTrap (ctxt, COMMAND_PROC, &trap, trap.Status.s.TrapType, &trap.FaultSave, ISSUE_COMMAND_FOR_CPROC);

	    res = ISSUE_COMMAND_OK;
	    break;
	}

	switch (res)
	{
	case ISSUE_COMMAND_OK:					/* command re-issued ok*/
	    break;

	case ISSUE_COMMAND_TRAPPED:				/* command trapped,  it will have been copied */
	    return (EAGAIN);					/* to the back of the trap queue */

	case ISSUE_COMMAND_RETRY:				/* didn't issue command, so place back at front for */
	    spin_lock_irqsave (&dev->IntrLock, flags);		/* later (after resolving other traps */

	    if (ELAN3_QUEUE_REALLY_FULL (ctxt->CommandTrapQ))
		ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
	    else
	    {
		ELAN3_QUEUE_ADD_FRONT(ctxt->CommandTrapQ);
		(*ELAN3_QUEUE_FRONT (ctxt->CommandTrapQ, ctxt->CommandTraps)) = trap;
	    }
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    return (EAGAIN);

	default:
	    return (EINVAL);
	}
	spin_lock_irqsave (&dev->IntrLock, flags);
    }	

    /*
     * GNAT 5409 - if CommandPortItem was not NULL, but other reasons were set,
     *             then we'd not free the CommandPortItem even though we'd re-
     *             issued all trapped and overflowed commands.  Hence only return
     *             without clearing CommandPortItem if we will be called again as
     *             either CommandTrapQ or CommandQ is not empty.
     */

    /* Now run the overflowed commands for this context */
    if (! ELAN3_QUEUE_EMPTY (ctxt->CommandQ))
    {
	if (! ELAN3_QUEUE_EMPTY (ctxt->CommandTrapQ) || (ctxt->Status & CTXT_OTHERS_REASONS))
	{
	    PRINTF0 (ctxt, DBG_CPROC, "RestartCProcTrap: cannot issue overflowed commands\n");
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    return (EAGAIN);
	}

	/*
	 * Just re-issue the commands,  if one traps then the remainder will 
	 * just get placed in the overflow queue again and the interrupt handler
	 * will copy them back in here.
	 *
	 * Stop the dma processor from taking commands,  since one of the commands
	 * could be a re-issued remote dma, which must be processed by the command
	 * processor.
	 */
	
	if (dev->HaltDmaDequeueCount++ == 0)
	    SetSchedStatusRegister (dev, 0, NULL);
	
	while (! ELAN3_QUEUE_EMPTY (ctxt->CommandQ))
	{
	    CProcTrapBuf_BE *TrapBuf = ELAN3_QUEUE_FRONT (ctxt->CommandQ, ctxt->Commands);
	    
	    PRINTF2 (ctxt, DBG_CPROC, "RestartCProcTrap: re-issue command %x - %08x\n",
		     (TrapBuf->s.ContextType>>5) & 0xff, TrapBuf->s.Addr);
	    mb();							/* ensure writes to main memory completed */
	    ((E3_uint32 *) ctxt->CommandPort)[(TrapBuf->s.ContextType>>5) & 0xff] = TrapBuf->s.Addr;
	    mmiob();		     				/* and flush through IO writes */
	    
	    ELAN3_QUEUE_REMOVE (ctxt->CommandQ);
	}
	
	/* observe the command processor having halted */
	res = CheckCommandQueueFlushed (ctxt, DmaComQueueNotEmpty, 0, &flags);
	
	if (res != ISSUE_COMMAND_OK)
	{
	    PRINTF0 (ctxt, DBG_CPROC, "RestartCProcTrap: trapped after issueing overflowed commands\n");
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    return (EAGAIN);
	}
    }

    /* remove the command port item, while holding the lock */
    item = ctxt->CommandPortItem;
    ctxt->CommandPortItem = NULL;

    spin_unlock_irqrestore (&dev->IntrLock, flags);
	
    if (item != NULL)						/* Free of any item that may have been stored */
    {								/* because of the commandport trap */
	PRINTF1 (ctxt, DBG_CPROC, "RestartCProcTrap: commandPortItem %p\n", item);

	kmutex_lock (&ctxt->SwapListsLock);
	ELAN3_OP_FREE_BLOCK_ITEM (ctxt, item);
	kmutex_unlock (&ctxt->SwapListsLock);
    }

    return (ESUCCESS);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
