/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: dproc.c,v 1.52 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/dproc.c,v $ */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/intrinsics.h>
#include <elan3/dma.h>
#include <elan3/vmseg.h>

#define DMA_RETRY_FAIL_COUNT	8

static void PrintUserDma (ELAN3_CTXT *ctxt, E3_Addr addr);

int
HandleDProcTrap (ELAN3_DEV *dev, E3_uint32 *RestartBits)
{
    DMA_TRAP	 *trap   = dev->DmaTrap;

    ASSERT(SPINLOCK_HELD (&dev->IntrLock));

    /* Scoop out the trap information, before restarting the Elan */
    trap->Status.Status = read_reg32 (dev, Exts.DProcStatus.Status);
    
    ASSERT(trap->Status.s.WakeupFunction == WakeupNever);

    /* copy the normal dma access fault type */
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc), &trap->FaultSave, sizeof (E3_FaultSave_BE));
    
    /* copy all 4 of the dma data fault type */
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0), &trap->Data0, 4*sizeof (E3_FaultSave_BE));
    
    /* Copy the DMA descriptor */
    copy_dma_regs (dev, &trap->Desc);
    
    /* Copy the packet info */
    trap->PacketInfo.Value = read_reg32 (dev, Exts.Dmas.DmaRds.DMA_PacketInfo.Value);

    /* update device statistics */
    BumpStat (dev, DProcTraps);
    switch (trap->Status.s.TrapType)
    {
    case MI_DmaPacketTimedOutOrPacketError:
	if (trap->PacketInfo.s.PacketTimeout)
	    BumpStat (dev, DmaOutputTimeouts);
	else if (trap->PacketInfo.s.PacketAckValue == C_ACK_ERROR)
	    BumpStat (dev, DmaPacketAckErrors);
	break;
	
    case MI_DmaFailCountError:
	BumpStat (dev, DmaRetries);
	break;
    }

    /* Must now zero all the FSRs so that a subsequent fault can be seen */
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc), sizeof (E3_FaultSave));
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0), 4*sizeof (E3_FaultSave));
	    
    *RestartBits |= RestartDProc;
    return (TRUE);
}

void
DeliverDProcTrap (ELAN3_DEV *dev, DMA_TRAP *dmaTrap, E3_uint32 Pend)
{
    ELAN3_CTXT	    *ctxt;
    E3_FaultSave_BE *FaultArea;
    DMA_TRAP	    *trap;
    register int     i;

    ASSERT(SPINLOCK_HELD (&dev->IntrLock));

    ctxt = ELAN3_DEV_CTX_TABLE(dev, dmaTrap->Status.s.Context);

    if (ctxt == NULL)
    {
	PRINTF1 (DBG_DEVICE, DBG_INTR, "DeliverDProcTrap: context %x invalid\n", dmaTrap->Status.s.Context);
	BumpStat (dev, InvalidContext);
    }
    else
    {
	if (ELAN3_OP_DPROC_TRAP (ctxt, dmaTrap) == OP_DEFER)
	{
	    if (ELAN3_QUEUE_REALLY_FULL (ctxt->DmaTrapQ))
	    {
		ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
		StartSwapoutContext (ctxt, Pend, NULL);
	    }
	    else
	    {
		trap = ELAN3_QUEUE_BACK (ctxt->DmaTrapQ, ctxt->DmaTraps);
		
		bcopy (dmaTrap, trap, sizeof (DMA_TRAP));
		
		PRINTF5 (ctxt, DBG_INTR, "DeliverDProcTrap: WakeupFnt=%x Cntx=%x SuspAddr=%x PacketInfo=%x TrapType=%s\n",
			 trap->Status.s.WakeupFunction, trap->Status.s.Context, 
			 trap->Status.s.SuspendAddr, trap->PacketInfo.Value, MiToName (trap->Status.s.TrapType));
		PRINTF3 (ctxt, DBG_INTR, "                    FaultAddr=%x EventAddr=%x FSR=%x\n",
			 trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress,
			 trap->FaultSave.s.FSR.Status);
		for (i = 0, FaultArea = &trap->Data0; i < 4; i++, FaultArea++)
		    PRINTF4 (ctxt, DBG_INTR, "                  %d FaultAddr=%x EventAddr=%x FSR=%x\n", i,
			     FaultArea->s.FaultAddress, FaultArea->s.EventAddress, FaultArea->s.FSR.Status);
		
		PRINTF4 (ctxt, DBG_INTR, "                 type %08x size %08x source %08x dest %08x\n",
			 trap->Desc.s.dma_type, trap->Desc.s.dma_size, trap->Desc.s.dma_source, trap->Desc.s.dma_dest);
		PRINTF2 (ctxt, DBG_INTR, "                 Dest event %08x cookie/proc %08x\n",
			 trap->Desc.s.dma_destEvent, trap->Desc.s.dma_destCookieVProc);
		PRINTF2 (ctxt, DBG_INTR, "                 Source event %08x cookie/proc %08x\n",
			 trap->Desc.s.dma_srcEvent, trap->Desc.s.dma_srcCookieVProc);
		ELAN3_QUEUE_ADD (ctxt->DmaTrapQ);
		kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
		
		if (ELAN3_QUEUE_FULL (ctxt->DmaTrapQ))
		{
		    PRINTF0 (ctxt, DBG_INTR, "DeliverDProcTrap: dma queue full, must swap out\n");
		    ctxt->Status |= CTXT_DMA_QUEUE_FULL;
		    
		    StartSwapoutContext (ctxt, Pend, NULL);
		}
	    }
	}
    }
}

int
NextDProcTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap)
{
    ELAN3_DEV *dev = ctxt->Device;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    
    if (ELAN3_QUEUE_EMPTY (ctxt->DmaTrapQ))
	return (0);

    *trap = *ELAN3_QUEUE_FRONT (ctxt->DmaTrapQ, ctxt->DmaTraps);
    ELAN3_QUEUE_REMOVE (ctxt->DmaTrapQ);
    
    return (1);
}

void
ResolveDProcTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap)
{
    E3_FaultSave_BE *FaultArea;
    int		     FaultHandled = 0;
    int		     res;
    register int     i;
    
    PRINTF4 (ctxt, DBG_DPROC, "ResolveDProcTrap: WakeupFnt=%x Cntx=%x SuspAddr=%x TrapType=%s\n",
	     trap->Status.s.WakeupFunction, trap->Status.s.Context, 
	     trap->Status.s.SuspendAddr, MiToName (trap->Status.s.TrapType));
    PRINTF3 (ctxt, DBG_DPROC, "                    FaultAddr=%x EventAddr=%x FSR=%x\n",
	     trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress,
	     trap->FaultSave.s.FSR.Status);
    for (i = 0, FaultArea = &trap->Data0; i < 4; i++, FaultArea++)
	PRINTF4 (ctxt, DBG_DPROC, "                  %d FaultAddr=%x EventAddr=%x FSR=%x\n", i,
		 FaultArea->s.FaultAddress, FaultArea->s.EventAddress, FaultArea->s.FSR.Status);

    PRINTF4 (ctxt, DBG_DPROC, "                  type %08x size %08x source %08x dest %08x\n",
	     trap->Desc.s.dma_type, trap->Desc.s.dma_size, trap->Desc.s.dma_source, trap->Desc.s.dma_dest);
    PRINTF2 (ctxt, DBG_DPROC, "                  Dest event %08x cookie/proc %08x\n",
	     trap->Desc.s.dma_destEvent, trap->Desc.s.dma_destCookieVProc);
    PRINTF2 (ctxt, DBG_DPROC, "                  Source event %08x cookie/proc %08x\n",
	     trap->Desc.s.dma_srcEvent, trap->Desc.s.dma_srcCookieVProc);
    
    BumpUserStat (ctxt, DProcTraps);

    switch (trap->Status.s.TrapType)
    {
    case MI_DmaPacketTimedOutOrPacketError:
	/*
	 * Faulted due to packet timeout or a PAckError.
	 * Reset fail count and reissue the same desc.
	 */
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: got a PAckError or the output timed out. Rescheduling dma.\n");
	if (ElanException (ctxt, EXCEPTION_PACKET_TIMEOUT, DMA_PROC, trap) == OP_IGNORE)
	{
	    BumpUserStat (ctxt, DmaRetries);

	    trap->Desc.s.dma_failCount = DMA_RETRY_FAIL_COUNT;

	    RestartDmaTrap (ctxt, trap);
	}
	return;

    case MI_DmaFailCountError:
	/*
	 * Faulted due to dma fail count.
	 * Reset fail count and reissue the same desc.
	 */
	PRINTF1 (ctxt, DBG_DPROC, "ResolveDProcTrap: Reset dma fail count to %d\n", DMA_RETRY_FAIL_COUNT);
	
	if (ElanException (ctxt, EXCEPTION_DMA_RETRY_FAIL, DMA_PROC, trap) == OP_IGNORE)
	{
	    BumpUserStat (ctxt, DmaRetries);

	    trap->Desc.s.dma_failCount = DMA_RETRY_FAIL_COUNT;

	    RestartDmaTrap (ctxt, trap);
	}
	return;

    case MI_TimesliceDmaQueueOverflow:
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: dma timeslice queue overflow\n");
	RestartDmaTrap (ctxt, trap);
	return;
	
    case MI_UnimplementedError:
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: unimplemented dma trap\n");
	if (ElanException (ctxt, EXCEPTION_UNIMPLEMENTED, DMA_PROC, trap) == OP_IGNORE)
	    RestartDmaTrap (ctxt, trap);
	return;

    case MI_EventQueueOverflow:
    case MI_ThreadQueueOverflow:
    case MI_DmaQueueOverflow:
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: trapped on a write set event.\n");
	FixupEventTrap (ctxt, DMA_PROC, trap, trap->Status.s.TrapType, &trap->FaultSave, 0);
	return;

    case MI_RemoteDmaCommand:
    case MI_RunDmaCommand:
    case MI_DequeueNonSysCntxDma:
    case MI_DequeueSysCntxDma:
	/*
	 * The DMA processor has trapped due to outstanding prefetches from the previous 
	 * dma.  The "current" dma has not been consumed, so we just ignore the trap
	 */
	return;

    case MI_WaitForRemoteDescRead2:
    case MI_ExecuteDmaDescriptorForRun:
	/*
	 * The DMA processor has trapped while fetching the dma descriptor, so
	 * zero it out to not confuse the user on an error
	 */
	bzero (&trap->Desc, sizeof (trap->Desc));
	break;
    }

    /*
     * All other uWords will have updated one of the fault areas,  so fix
     * any faults found in them.  If there were no faults found then it 
     * must have been a bus error
     */
    for (i = 0, FaultArea = &trap->Data0; i < 4; i++, FaultArea++)
    {
	if (FaultArea->s.FSR.Status != 0)
	{
	    FaultHandled++;

	    ASSERT ((FaultArea->s.FSR.Status & FSR_SizeMask) == FSR_Block64 ||
		    (FaultArea->s.FSR.Status & FSR_SizeMask) == FSR_Block32);
	    
	    ASSERT (FaultArea->s.FaultContext == trap->Status.s.Context);
	    
	    if (((trap->Desc.s.dma_source & PAGEOFFSET) >= (PAGESIZE-E3_BLK_SIZE)) &&
		((trap->Desc.s.dma_source & PAGEMASK) != ((trap->Desc.s.dma_source + trap->Desc.s.dma_size-1) & PAGEMASK)))
	    {
		/* XXXX: dma started within last 64 bytes of the page
		 *       terminate the process if it has pagefaulted */
		if (FaultArea->s.FaultAddress == (trap->Desc.s.dma_source & ~(E3_BLK_SIZE-1)))
		{
		    printk ("elan%d: invalid dma - context=%x source=%x\n", ctxt->Device->Instance, 
			    ctxt->Capability.cap_mycontext, trap->Desc.s.dma_source);
		    
		    if (ElanException (ctxt, EXCEPTION_BAD_DMA, DMA_PROC, trap, NULL, 0) != OP_IGNORE)
			return;
		}
	    }

	    if (trap->Desc.s.dma_size != 0 && (res = elan3_pagefault (ctxt, FaultArea, 1)) != ESUCCESS)
	    {
		/* XXXX: Rev B Elans can prefetch data passed the end of the dma descriptor */
		/*       if the fault relates to this, then just ignore it */
		if (FaultArea->s.FaultAddress < (trap->Desc.s.dma_source+trap->Desc.s.dma_size) ||
		    FaultArea->s.FaultAddress > (trap->Desc.s.dma_source+trap->Desc.s.dma_size+E3_BLK_SIZE*2))
		{
		    PRINTF1 (ctxt, DBG_DPROC, "ResolveDProcTrap: elan3_pagefault failed for address %x\n",
			     FaultArea->s.FaultAddress);
		    
		    if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, DMA_PROC, trap, FaultArea, res) != OP_IGNORE)
			return;
		}
	    }
	}
    }
    
    if (trap->FaultSave.s.FSR.Status != 0)
    {
	FaultHandled++;

	ASSERT (trap->FaultSave.s.FaultContext == trap->Status.s.Context);

	if ((trap->FaultSave.s.FSR.Status & FSR_SizeMask) == FSR_RouteFetch)
	{
	    res = ResolveVirtualProcess (ctxt, trap->FaultSave.s.FaultAddress & 0xffff); /* mask out cookie */

	    switch (res)
	    {
	    default:
		if (ElanException (ctxt, EXCEPTION_INVALID_PROCESS, DMA_PROC, trap, trap->FaultSave.s.FaultAddress, res) != OP_IGNORE)
		    return;
		
	    case EAGAIN:
		/* XXXX; wait on trail blazing code */

	    case 0:
		break;
	    }
	}
	else
	{
	    if ((res = elan3_pagefault (ctxt, &trap->FaultSave, 1)) != ESUCCESS)
	    {
		PRINTF1 (ctxt, DBG_DPROC, "ResolveDProcTrap: elan3_pagefault failed for address %x\n",
			 trap->FaultSave.s.FaultAddress);

		if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, DMA_PROC, trap, &trap->FaultSave, res) != OP_IGNORE)
		    return;
	    }
	}
    }

    if (! FaultHandled)
    {
	ElanBusError (ctxt->Device);

	if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, DMA_PROC, trap, &trap->FaultSave, EFAULT) != OP_IGNORE)
	    return;
    }

    switch (trap->Status.s.TrapType)
    {
    case MI_WaitForRemoteDescRead2:
	/*
	 * Faulted while trying to read the dma descriptor for a read dma.
	 * Fix fault and re-issue using FaultAddress.
	 */
	PRINTF1 (ctxt, DBG_DPROC, "ResolveDProcTrap: trapped reading a remote dma descriptor at %x.\n",
		 trap->FaultSave.s.FaultAddress);
	
	RestartDmaPtr (ctxt, trap->FaultSave.s.FaultAddress);
	break;
	
    case MI_ExecuteDmaDescriptorForRun:
	/*
	 * Faulted while trying to read the dma descriptor for a write dma.
	 * Fix fault and re-issue using FaultAddress.
	 */
	PRINTF1 (ctxt, DBG_DPROC, "ResolveDProcTrap: trapped reading a write dma descriptor at %x.\n", 
		 trap->FaultSave.s.FaultAddress);
	
	RestartDmaPtr (ctxt, trap->FaultSave.s.FaultAddress);
	break;
	
    case MI_WaitForRemoteRoutes1:
    case MI_WaitForRemoteRoutes2:
    case MI_SendRemoteDmaDesc:
    case MI_SendDmaIdentify:
    case MI_SendRemoteDmaRoutes2:
    case MI_WaitForDmaRoutes1:
    case MI_DmaLoop:
    case MI_ExitDmaLoop:
    case MI_GetDestEventValue:
    case MI_SendFinalUnlockTrans:
    case MI_SendNullSetEvent:
    case MI_SendFinalSetEvent:
    case MI_SendDmaEOP:
	/*
	 * Faulted either fetching routes or fetching dma data.
	 * Fix fault and re-issue using FaultAddress.
	 */

    case MI_SendEOPforRemoteDma:
    case MI_LookAtRemoteAck:
    case MI_FailedAckIfCCis0:
	/*
	 * Possible fault when reading the remote desc into the dma data buffers
	 */
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap:  trapped reading a dma data or fetching a route\n");
	RestartDmaTrap (ctxt, trap);
	break;
	
    case MI_DequeueSysCntxDma:
    case MI_DequeueNonSysCntxDma:
    case MI_RemoteDmaCommand:
    case MI_RunDmaCommand:
	/*
	 * It is possible that a dma can get back onto the queue while outstanding dma
	 * have not finished trapping. In this case the trap can be ignored as the dma
	 * state has been saved. It might trap again the next time it comes to the front
	 * of the queue and be fixed then.
	 */
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: trap after dma has finished. ignored\n");
	break;
	
    default:
	PRINTF0 (ctxt, DBG_DPROC, "ResolveDProcTrap: trapped on a write set event.\n");
	FixupEventTrap (ctxt, DMA_PROC, trap, trap->Status.s.TrapType, &trap->FaultSave, 0);
	break;
    }
}

int
DProcNeedsRestart (ELAN3_CTXT *ctxt)
{
    return (ctxt->ItemCount[LIST_DMA_PTR] != 0 ||
	    ctxt->ItemCount[LIST_DMA_DESC] != 0);
}

void
RestartDProcItems (ELAN3_CTXT *ctxt)
{
    void      *item;
    E3_Addr    value;
    int	       res;
    
    kmutex_lock (&ctxt->SwapListsLock);
    while (ctxt->ItemCount[LIST_DMA_PTR])
    {
	if (! ELAN3_OP_GET_WORD_ITEM (ctxt, LIST_DMA_PTR, &item, &value))
	    ctxt->ItemCount[LIST_DMA_PTR] = 0;
	else
	{
	    PRINTF1 (ctxt, DBG_DPROC, "RestartDProc: issue write dma at %x\n", value);
	    PrintUserDma (ctxt, value);

	    res = IssueDmaCommand (ctxt, value, NULL, 0);
	    
	    if (res == ISSUE_COMMAND_RETRY)
	    {
		ELAN3_OP_PUTBACK_ITEM (ctxt, LIST_DMA_PTR, item);
		kmutex_unlock (&ctxt->SwapListsLock);
		return;
	    }
	    
	    ctxt->ItemCount[LIST_DMA_PTR]--;
	    ELAN3_OP_FREE_WORD_ITEM (ctxt, item);
	}
    }
    
    while (ctxt->ItemCount[LIST_DMA_DESC])
    {
	if (! ELAN3_OP_GET_BLOCK_ITEM (ctxt, LIST_DMA_DESC, &item, &value))
	    ctxt->ItemCount[LIST_DMA_DESC] = 0;
	else
	{
	    PRINTF1 (ctxt, DBG_DPROC, "RestartDProc: issue dma desc at %x\n", value);
	    PrintUserDma (ctxt, value);

	    res = IssueDmaCommand (ctxt, value, item, 0);

	    switch (res)
	    {
	    case ISSUE_COMMAND_OK:
		ctxt->ItemCount[LIST_DMA_DESC]--;
		ELAN3_OP_FREE_BLOCK_ITEM (ctxt, item);
		break;
		
	    case ISSUE_COMMAND_RETRY:
		ELAN3_OP_PUTBACK_ITEM (ctxt, LIST_DMA_DESC, item);
		kmutex_unlock (&ctxt->SwapListsLock);
		return;
		
	    case ISSUE_COMMAND_TRAPPED:
		ctxt->ItemCount[LIST_DMA_DESC]--;
		/* The item will be freed off when the command port trap */
		/* fixed up and the command successfully re-issued */
		break;
	    }
	}
    }

    kmutex_unlock (&ctxt->SwapListsLock);
}

void
RestartDmaDesc(ELAN3_CTXT *ctxt, E3_DMA_BE *desc)
{
    kmutex_lock (&ctxt->SwapListsLock);
    if (desc->s.dma_direction != DMA_WRITE)
	desc->s.dma_direction = (desc->s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;

    ELAN3_OP_PUT_BLOCK_ITEM (ctxt, LIST_DMA_DESC, (E3_uint32 *) desc);
    ctxt->ItemCount[LIST_DMA_DESC]++;

    kmutex_unlock (&ctxt->SwapListsLock);
}

void
RestartDmaTrap(ELAN3_CTXT *ctxt, DMA_TRAP *trap)
{
    /* Negative length DMAs are illegal, since they hangup the dma processor,
     * if they got generated then they will have been spotted by PollForDmahungup,
     * and delivered to us with a Dequeue  suspend address,
     *
     * GNAT sw-elan3/3908: Moved this check into this new function to avoid
     * it sampling old or invalid register state
     */
    if (trap->Desc.s.dma_size > E3_MAX_DMA_SIZE)
	ElanException (ctxt, EXCEPTION_BAD_DMA, DMA_PROC, trap, NULL, 0);
    else
	RestartDmaDesc (ctxt, &trap->Desc);
}

void
RestartDmaPtr (ELAN3_CTXT *ctxt, E3_Addr ptr)
{
    kmutex_lock (&ctxt->SwapListsLock);
    ELAN3_OP_PUT_WORD_ITEM (ctxt, LIST_DMA_PTR, ptr);
    ctxt->ItemCount[LIST_DMA_PTR]++;
    kmutex_unlock (&ctxt->SwapListsLock);
}

static void
PrintUserDma (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    E3_DMA *dma;

    /* Dont call a function which takes locks unless we need to */
    if (!(elan3_debug & DBG_DPROC))
        return;

    dma = (E3_DMA *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    PRINTF4 (ctxt, DBG_DPROC, "DMA: type %08x size %08x source %08x dest %08x\n",
	     fuword ((int *) &dma->dma_type), fuword ((int *) &dma->dma_size), 
	     fuword ((int *) &dma->dma_source), fuword ((int *) &dma->dma_dest));
    PRINTF4 (ctxt, DBG_DPROC, "DMA: Dest %08x %08x  Local %08x %08x\n",
	     fuword ((int *) &dma->dma_destEvent), fuword ((int *) &dma->dma_destCookieProc), 
	     fuword ((int *) &dma->dma_srcEvent), fuword ((int *) &dma->dma_srcCookieProc));
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
