/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elan3mmu_generic.c,v 1.76 2004/12/14 10:19:38 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/vm/elan3mmu_generic.c,v $*/

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>

#ifdef CONFIG_MPSAS
#  define zero_all_ptbls
#endif

/*
 * Debugging
 */
int	elan3mmu_debug = 0;

#define	N_L3PTBL_MTX	(0x20)
#define	N_L2PTBL_MTX	(0x40)
#define	N_L1PTBL_MTX	(0x20)

#define	L3PTBL_MTX_HASH(p) \
	((((uintptr_t)(p) >> 12) ^ ((uintptr_t)(p) >> 2)) & (N_L3PTBL_MTX - 1))
static	spinlock_t l3ptbl_lock[N_L3PTBL_MTX];

#define	L2PTBL_MTX_HASH(p)   \
	((((uintptr_t)(p) >> 12) ^ ((uintptr_t)(p) >> 2)) & (N_L2PTBL_MTX - 1))
static	spinlock_t l2ptbl_lock[N_L2PTBL_MTX];

#define	L1PTBL_MTX_HASH(p)   \
	((((uintptr_t)(p) >> 12) ^ ((uintptr_t)(p) >> 2)) & (N_L1PTBL_MTX - 1))
static	spinlock_t l1ptbl_lock[N_L1PTBL_MTX];


#define	BASE2VA(p)	((E3_Addr)((p)->ptbl_base << 16))
#define	VA2BASE(v)	((u_short)(((uintptr_t)(v)) >> 16))

ELAN3MMU_GLOBAL_STATS	elan3mmu_global_stats;

static void          elan3mmu_flush_context_filter (ELAN3_DEV *dev, void *);
static void          elan3mmu_unload_loop (ELAN3MMU *elan3mmu, ELAN3_PTBL *ptbl, int first_valid, int nptes, int flags);

static ELAN3_PTBL    *elan3mmu_create_ptbls (ELAN3_DEV *dev, int level, int attr, int keep);
static ELAN3_PTBL    *elan3mmu_ta_to_ptbl (ELAN3MMU *elan3mmu, ELAN3_PTP *ptp);

static ELAN3_PTBL    *elan3mmu_alloc_pte    (ELAN3_DEV *dev, ELAN3MMU *elan3mmu, int *idx);
void                 elan3mmu_free_lXptbl  (ELAN3_DEV *dev, ELAN3_PTBL *ptbl);

void                 elan3mmu_free_pte  (ELAN3_DEV *dev,  ELAN3MMU *elan3mmu,  ELAN3_PTBL *ptbl_ptr, int idx);

static ELAN3_PTBL    *elan3mmu_alloc_l1ptbl (ELAN3_DEV *dev, int attr, ELAN3MMU *elan3mmu);
static ELAN3_PTBL    *elan3mmu_alloc_l2ptbl (ELAN3_DEV *dev, int attr, ELAN3_PTBL *parent, ELAN3MMU *elan3mmu,
					    E3_Addr base, spinlock_t **plock, unsigned long *flags);
static ELAN3_PTBL    *elan3mmu_alloc_l3ptbl (ELAN3_DEV *dev, int attr, ELAN3_PTBL *parent, ELAN3MMU *elan3mmu,
					    E3_Addr base, spinlock_t **plock, unsigned long *flags);

static int	     elan3mmu_steal_this_ptbl (ELAN3_DEV *dev, ELAN3_PTBL *l3ptbl);
static ELAN3_PTBL    *elan3mmu_steal_l3ptbl (ELAN3_DEV *dev, int attr);

static spinlock_t   *elan3mmu_ptbl_to_lock (int level, ELAN3_PTBL *ptbl);

/*
 * Encoding of MMU permissions against access type,
 * to allow quick permission checking against access 
 * type.
 */
u_char elan3mmu_permissionTable[] =
{
    0xcc,	/* 11001100 ELAN3_PERM_NULL        */
    0x01,	/* 00000001 ELAN3_PERM_LOCALREAD   */
    0x05,	/* 00000101 ELAN3_PERM_READ        */
    0x33,	/* 00110011 ELAN3_PERM_NOREMOTE    */
    0x37,	/* 00110111 ELAN3_PERM_REMOTEREAD  */
    0x3f,	/* 00111111 ELAN3_PERM_REMOTEWRITE */
    0xf7,	/* 11110111 ELAN3_PERM_REMOTEEVENT */
    0xff,	/* 11111111 ELAN3_PERM_REMOTEALL	  */
} ;

void
elan3mmu_init()
{
    register int i;

    HAT_PRINTF0 (1, "elan3mmu_init: initialising elan mmu\n");

    for (i = 0; i < N_L1PTBL_MTX; i++)
	spin_lock_init (&l1ptbl_lock[i]);

    for (i = 0; i < N_L2PTBL_MTX; i++)
	spin_lock_init (&l2ptbl_lock[i]);

    for (i = 0; i < N_L3PTBL_MTX; i++)
	spin_lock_init (&l3ptbl_lock[i]);

    elan3mmu_global_stats.version = ELAN3MMU_STATS_VERSION;

    elan3mmu_init_osdep();
}

void
elan3mmu_fini()
{
    register int i;

    HAT_PRINTF0 (1, "elan3mmu_fini: finalising elan mmu\n");

    for (i = 0; i < N_L1PTBL_MTX; i++)
	spin_lock_destroy (&l1ptbl_lock[i]);

    for (i = 0; i < N_L2PTBL_MTX; i++)
	spin_lock_destroy (&l2ptbl_lock[i]);

    for (i = 0; i < N_L3PTBL_MTX; i++)
	spin_lock_destroy (&l3ptbl_lock[i]);

    elan3mmu_fini_osdep();
}

ELAN3MMU *
elan3mmu_alloc (ELAN3_CTXT *ctxt)
{
    ELAN3MMU  *elan3mmu;
    ELAN3_PTBL *l1ptbl;

    ALLOC_ELAN3MMU (elan3mmu, TRUE);
    
    spin_lock_init (&elan3mmu->elan3mmu_lock);

    spin_lock (&elan3mmu->elan3mmu_lock);			/* lock_lint */

    elan3mmu->elan3mmu_ergns    = NULL;
    elan3mmu->elan3mmu_etail    = NULL;
    elan3mmu->elan3mmu_ergnlast = NULL;
    elan3mmu->elan3mmu_mrgns    = NULL;
    elan3mmu->elan3mmu_mtail    = NULL;
    elan3mmu->elan3mmu_mrgnlast = NULL;
    elan3mmu->elan3mmu_ctxt     = ctxt;

    spin_lock_init (&elan3mmu->elan3mmu_lXptbl_lock);
    elan3mmu->elan3mmu_lXptbl   = NULL;

    spin_unlock (&elan3mmu->elan3mmu_lock);			/* lock_lint */

    l1ptbl = elan3mmu_alloc_l1ptbl(ctxt->Device, 0, elan3mmu);

    elan3mmu->elan3mmu_ctp      = (sdramaddr_t) 0;
    elan3mmu->elan3mmu_dev      = ctxt->Device;
    elan3mmu->elan3mmu_l1ptbl   = l1ptbl;

    /* Ensure that there are at least some level 3 page tables,  since if a level 2 and */
    /* a level 3 table are allocated together, then the level 3 is allocated with the NO_ALLOC */
    /* flag,  thus there MUST be at least one that can be stolen or on the free list */
    if (elan3mmu->elan3mmu_dev->Level[PTBL_LEVEL_3].PtblFreeList == NULL)
	elan3mmu_create_ptbls (elan3mmu->elan3mmu_dev, PTBL_LEVEL_3, 0, 0);

    HAT_PRINTF1 (1, "elan3mmu_alloc: elan3mmu %p\n", elan3mmu);

    elan3mmu_alloc_osdep (elan3mmu);

    return (elan3mmu);
}

void 
elan3mmu_free (ELAN3MMU *elan3mmu)
{
    ELAN3MMU_RGN   *rgn;
    ELAN3_PTBL	   *l1ptbl;
    spinlock_t	   *l1lock;
    unsigned long   l1flags;
    unsigned long   flags;

    HAT_PRINTF1 (1, "elan3mmu_free : elan3mmu %p\n", elan3mmu);
    
    /*
     * Invalidate the level1 page table,  since it's already removed
     * from the context table, there is no need to flush the tlb.
     */
    l1ptbl = elan3mmu->elan3mmu_l1ptbl;
    elan3mmu->elan3mmu_l1ptbl = NULL;
    
    if (elan3mmu_lock_ptbl (l1ptbl, LK_PTBL_FAILOK, elan3mmu, (E3_Addr) 0, PTBL_LEVEL_1, &l1lock, &l1flags) == LK_PTBL_OK)
    {
	elan3mmu_l1inval (elan3mmu, l1ptbl, PTE_UNLOAD_NOFLUSH);
	elan3mmu_free_l1ptbl (elan3mmu->elan3mmu_dev, l1ptbl, l1lock, l1flags);
    }

    /*
     * Free of any permission regions.
     */
    spin_lock (&elan3mmu->elan3mmu_lock);					/* lock_lint */
    while ((rgn = elan3mmu->elan3mmu_mrgns) != NULL)
    {
	spin_lock_irqsave (&elan3mmu->elan3mmu_dev->IntrLock, flags);		/* lock_lint */
	elan3mmu_removergn_elan (elan3mmu, rgn->rgn_ebase);
	elan3mmu_removergn_main (elan3mmu, rgn->rgn_mbase);
	spin_unlock_irqrestore (&elan3mmu->elan3mmu_dev->IntrLock, flags);	/* lock_lint */
	
	FREE_ELAN3MMU_RGN (rgn);
    }
    elan3mmu->elan3mmu_mrgnlast = NULL;
    elan3mmu->elan3mmu_ergnlast = NULL;

    /* 
     * Free the lXptbl list
     */
    ASSERT (elan3mmu->elan3mmu_lXptbl == NULL); /* XXXX MRH need to add list removal */
 
    elan3mmu->elan3mmu_lXptbl = NULL;
    spin_lock_destroy (&elan3mmu->elan3mmu_lXptbl_lock);


    spin_unlock (&elan3mmu->elan3mmu_lock);					/* lock_lint */

    spin_lock_destroy (&elan3mmu->elan3mmu_lock);

    FREE_ELAN3MMU (elan3mmu);
}

/*================================================================================*/
/* Interface routines to device driver */
static void
elan3mmu_flush_context_filter (ELAN3_DEV *dev, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    ASSERT ((read_reg32 (dev, Exts.InterruptReg) & (INT_DiscardingSysCntx | INT_DiscardingNonSysCntx)) == 
	    (INT_DiscardingSysCntx | INT_DiscardingNonSysCntx));

    dev->FilterHaltQueued = 0;

    write_reg32 (dev, Input_Context_Fil_Flush, 0);

    HAT_PRINTF0 (1, "elan3mmu_flush_context_filter completed\n");

    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

void
elan3mmu_set_context_filter (ELAN3_DEV *dev, int ctx, int disabled, E3_uint32 Pend, E3_uint32 *Maskp)
{
    int         mctx = ctx & MAX_ROOT_CONTEXT_MASK;
    sdramaddr_t ctp  = dev->ContextTable + mctx * sizeof (E3_ContextControlBlock);

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));

    ASSERT ((mctx < 32 || mctx >= ELAN3_KCOMM_BASE_CONTEXT_NUM) ? (ctx & SYS_CONTEXT_BIT) : ! (ctx & SYS_CONTEXT_BIT));

    elan3_sdram_writel (dev, ctp + offsetof (E3_ContextControlBlock, filter), 
		  ((ctx & SYS_CONTEXT_BIT) ? E3_CCB_CNTX0 : 0) | (disabled ? E3_CCB_DISCARD_ALL : 0));

    HAT_PRINTF4 (1, "elan3mmu_set_context_filter: ctx %x [%lx] -> %s (%x)\n", ctx, ctp,
		 disabled ? "up" : "down", elan3_sdram_readl (dev, ctp + offsetof (E3_ContextControlBlock, filter)));

    /* queue a halt operation to flush the context filter while the inputter is halted */
    if (dev->FilterHaltQueued == 0)
    {
	dev->FilterHaltQueued = 1;
	QueueHaltOperation (dev, Pend, Maskp, INT_DiscardingSysCntx | INT_DiscardingNonSysCntx, 
			    elan3mmu_flush_context_filter, NULL);
    }
}

int
elan3mmu_attach (ELAN3_DEV *dev, int ctx, ELAN3MMU *elan3mmu, sdramaddr_t routeTable, E3_uint32 routeMask)
{
    sdramaddr_t ctp;
    ELAN3_PTP    trootptp;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));

    ctx &= MAX_ROOT_CONTEXT_MASK;						/* Mask out all high bits in context */
    
    if (ctx < 0 || ctx >= dev->ContextTableSize)
	return (EINVAL);

    ctp = dev->ContextTable + ctx * sizeof (E3_ContextControlBlock);
    
    trootptp = elan3_readptp (dev, ctp + offsetof (E3_ContextControlBlock, rootPTP));
    
    if (ELAN3_PTP_TYPE(trootptp) != ELAN3_ET_INVALID)
	return (EBUSY);

    elan3mmu->elan3mmu_ctp = ctp;
    
    trootptp = PTBL_TO_PTADDR (elan3mmu->elan3mmu_l1ptbl) | ELAN3_ET_PTP;
    
    HAT_PRINTF4 (1, "elan3mmu_attach: ctp at %08lx : trootptp=%08x VPT_ptr=%08lx VPT_mask=%08x\n",
		 ctp, trootptp, routeTable, routeMask);
    
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, rootPTP), trootptp);
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, VPT_ptr), routeTable);
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, VPT_mask), routeMask);
    
    return (ESUCCESS);
}

void
elan3mmu_detach (ELAN3_DEV *dev, int ctx)
{
    ELAN3_PTP    invalidptp = ELAN3_INVALID_PTP;
    sdramaddr_t ctp;
    
    ctx &= MAX_ROOT_CONTEXT_MASK;						/* Mask out all high bits in context */
    
    if (ctx < 0 || ctx >= dev->ContextTableSize)
	return;
    
    ctp = dev->ContextTable + ctx * sizeof (E3_ContextControlBlock);
    
    HAT_PRINTF1 (1, "elan3mmu_detach: clearing ptp at %lx\n", ctp);
    
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, rootPTP), invalidptp);
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, VPT_mask), 0);
    elan3_writeptp (dev, ctp + offsetof (E3_ContextControlBlock, VPT_ptr), 0);
    
    ElanFlushTlb (dev);
}

int
elan3mmu_reference (ELAN3MMU *elan3mmu, int ctx)
{
    ELAN3_DEV              *dev = elan3mmu->elan3mmu_dev;
    sdramaddr_t            ctp;
    E3_ContextControlBlock ccb;
    ELAN3_PTP               trootptp;

    ctx &= MAX_ROOT_CONTEXT_MASK;						/* Mask out all high bits in context */
    
    if (ctx < 0 || ctx >= dev->ContextTableSize)
	return (EINVAL);

    ctp = dev->ContextTable + ctx * sizeof (E3_ContextControlBlock);

    trootptp = elan3_readptp (dev, ctp + offsetof (E3_ContextControlBlock, rootPTP));
    
    if (ELAN3_PTP_TYPE(trootptp) != ELAN3_ET_INVALID)
	return (EBUSY);
    
    elan3_sdram_copyl_from_sdram (dev, elan3mmu->elan3mmu_ctp, &ccb, sizeof (E3_ContextControlBlock));
    elan3_sdram_copyl_to_sdram (dev, &ccb, ctp, sizeof (E3_ContextControlBlock));
    
    return (ESUCCESS);
    
}
/*================================================================================*/
/* Elan permission regions. */

/* elan address region management */
ELAN3MMU_RGN *
elan3mmu_findrgn_elan (ELAN3MMU *elan3mmu,
		       E3_Addr addr, int tail)
{
    ELAN3MMU_RGN *next = NULL;
    ELAN3MMU_RGN *rgn;
    ELAN3MMU_RGN *hirgn;
    ELAN3MMU_RGN *lorgn;
    E3_Addr       base;
    E3_Addr       lastaddr;
    int		  forward;

    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) || SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (elan3mmu->elan3mmu_ergns == NULL)
	return (NULL);

    rgn = elan3mmu->elan3mmu_ergnlast;
    if (rgn == NULL)
	rgn = elan3mmu->elan3mmu_ergns;

    forward = 0;
    if ((u_long) (base = rgn->rgn_ebase) < (u_long)addr)
    {
	if ((u_long)addr <= ((u_long) base + rgn->rgn_len - 1))
	    return (rgn);					/* ergnlast contained addr */

	hirgn = elan3mmu->elan3mmu_etail;

	if ((u_long) (lastaddr = (hirgn->rgn_ebase + hirgn->rgn_len - 1)) < (u_long) addr)
	    return (tail ? hirgn : NULL);			/* addr is out of range */
	
	if ((u_long) (addr - base) > (u_long) (lastaddr - addr))
	    rgn = hirgn;
	else
	{
	    rgn = rgn->rgn_enext;
	    forward++;
	}
    }
    else
    {
	lorgn = elan3mmu->elan3mmu_ergns;

	if ((u_long)lorgn->rgn_ebase > (u_long) addr)
	    return (lorgn);					/* lowest regions is higher than addr */
	if ((u_long)(addr - lorgn->rgn_ebase) < (u_long) (base - addr))
	{
	    rgn = lorgn;					/* search forward from head */
	    forward++;
	}
    }
    if (forward)
    {
	while ((u_long)(rgn->rgn_ebase + rgn->rgn_len - 1) < (u_long)addr)
	    rgn = rgn->rgn_enext;

	if ((u_long)rgn->rgn_ebase <= (u_long)addr)
	    elan3mmu->elan3mmu_ergnlast = rgn;
	return (rgn);
    }
    else
    {
	while ((u_long)rgn->rgn_ebase > (u_long)addr)
	{
	    next = rgn;
	    rgn = rgn->rgn_eprev;
	}

	if ((u_long) (rgn->rgn_ebase + rgn->rgn_len - 1) < (u_long)addr)
	    return (next);
	else
	{
	    elan3mmu->elan3mmu_ergnlast = rgn;
	    return (rgn);
	}
    }
}

int
elan3mmu_addrgn_elan (ELAN3MMU *elan3mmu, ELAN3MMU_RGN *nrgn)
{
    ELAN3MMU_RGN *rgn   = elan3mmu_findrgn_elan (elan3mmu, nrgn->rgn_ebase, 1);
    E3_Addr       nbase = nrgn->rgn_ebase;
    E3_Addr	  ntop  = nbase + nrgn->rgn_len - 1; /* avoid wrap */
    E3_Addr	  base;

    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) && SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (rgn == NULL)
    {
	elan3mmu->elan3mmu_ergns = elan3mmu->elan3mmu_etail = nrgn;
	nrgn->rgn_enext = nrgn->rgn_eprev = NULL;
    }
    else
    {
	base = rgn->rgn_ebase;

	if ((u_long)(base + rgn->rgn_len - 1) < (u_long)nbase)	/* top of region below requested address */
	{							/* so insert after region (and hence at end */
	    nrgn->rgn_eprev = rgn;				/* of list */
	    nrgn->rgn_enext = NULL;
	    rgn->rgn_enext = elan3mmu->elan3mmu_etail = nrgn;
	}
	else
	{
	    if ((u_long)nbase >= (u_long)base || (u_long)ntop >= (u_long)base)
		return (-1);					/* overlapping region */

	    nrgn->rgn_enext = rgn;				/* insert before region */
	    nrgn->rgn_eprev = rgn->rgn_eprev;
	    rgn->rgn_eprev  = nrgn;
	    if (elan3mmu->elan3mmu_ergns == rgn)
		elan3mmu->elan3mmu_ergns = nrgn;
	    else
		nrgn->rgn_eprev->rgn_enext = nrgn;
	}
    }
    elan3mmu->elan3mmu_ergnlast = nrgn;
    
    return (0);
}

ELAN3MMU_RGN *
elan3mmu_removergn_elan (ELAN3MMU *elan3mmu, E3_Addr addr)
{
    ELAN3MMU_RGN *rgn = elan3mmu_findrgn_elan (elan3mmu, addr, 0);
    
    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) && SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (rgn == NULL || rgn->rgn_ebase != addr)
	return (NULL);
    
    elan3mmu->elan3mmu_ergnlast = rgn->rgn_enext;
    if (rgn == elan3mmu->elan3mmu_etail)
	elan3mmu->elan3mmu_etail = rgn->rgn_eprev;
    else
	rgn->rgn_enext->rgn_eprev = rgn->rgn_eprev;
    
    if (rgn == elan3mmu->elan3mmu_ergns)
	elan3mmu->elan3mmu_ergns = rgn->rgn_enext;
    else
	rgn->rgn_eprev->rgn_enext = rgn->rgn_enext;

    return (rgn);
}

ELAN3MMU_RGN *
elan3mmu_rgnat_elan (ELAN3MMU *elan3mmu, E3_Addr addr)
{
    ELAN3MMU_RGN *rgn = elan3mmu_findrgn_elan (elan3mmu, addr, 0);
    E3_Addr       base;

    if (rgn != NULL && (u_long)(base = rgn->rgn_ebase) <= (u_long)addr && (u_long)addr <= (u_long)(base + rgn->rgn_len - 1))
	return (rgn);
    return (NULL);
}

/* main address region management */
ELAN3MMU_RGN *
elan3mmu_findrgn_main (ELAN3MMU *elan3mmu,
		       caddr_t addr, int tail)
{
    ELAN3MMU_RGN *next = NULL;
    ELAN3MMU_RGN *rgn;
    ELAN3MMU_RGN *hirgn;
    ELAN3MMU_RGN *lorgn;
    caddr_t       lastaddr;
    caddr_t       base;
    int		  forward;

    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) || SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (elan3mmu->elan3mmu_mrgns == NULL)
	return (NULL);

    rgn = elan3mmu->elan3mmu_mrgnlast;
    if (rgn == NULL)
	rgn = elan3mmu->elan3mmu_mrgns;

    forward = 0;
    if ((base = rgn->rgn_mbase) < addr)
    {
	if (addr <= (base + rgn->rgn_len - 1))
	    return (rgn);					/* ergnlast contained addr */

	hirgn = elan3mmu->elan3mmu_mtail;
	if ((lastaddr = hirgn->rgn_mbase + hirgn->rgn_len - 1) < addr)
	    return (tail ? hirgn : NULL);			/* addr is out of range */
	
	if ((addr - base) > (lastaddr - addr))
	    rgn = hirgn;
	else
	{
	    rgn = rgn->rgn_mnext;
	    forward++;
	}
    }
    else
    {
	lorgn = elan3mmu->elan3mmu_mrgns;
	if (lorgn->rgn_mbase > addr)
	    return (lorgn);					/* lowest regions is higher than addr */
	if ((addr - lorgn->rgn_mbase) < (base - addr))
	{
	    rgn = lorgn;					/* search forward from head */
	    forward++;
	}
    }
    if (forward)
    {
	while ((rgn->rgn_mbase + rgn->rgn_len - 1) < addr)
	    rgn = rgn->rgn_mnext;

	if (rgn->rgn_mbase <= addr)
	    elan3mmu->elan3mmu_mrgnlast = rgn;
	return (rgn);
    }
    else
    {
	while (rgn->rgn_mbase > addr)
	{
	    next = rgn;
	    rgn = rgn->rgn_mprev;
	}
	if ((rgn->rgn_mbase + rgn->rgn_len - 1) < addr)
	    return (next);
	else
	{
	    elan3mmu->elan3mmu_mrgnlast = rgn;
	    return (rgn);
	}
    }
}

int
elan3mmu_addrgn_main (ELAN3MMU *elan3mmu, ELAN3MMU_RGN *nrgn)
{
    ELAN3MMU_RGN *rgn   = elan3mmu_findrgn_main (elan3mmu, nrgn->rgn_mbase, 1);
    caddr_t       nbase = nrgn->rgn_mbase;
    caddr_t	  ntop  = nbase + nrgn->rgn_len - 1;
    caddr_t	  base;

    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) && SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (rgn == NULL)
    {
	elan3mmu->elan3mmu_mrgns = elan3mmu->elan3mmu_mtail = nrgn;
	nrgn->rgn_mnext = nrgn->rgn_mprev = NULL;
    }
    else
    {
	base = rgn->rgn_mbase;

	if ((base + rgn->rgn_len - 1) < nbase)			/* top of region below requested address */
	{							/* so insert after region (and hence at end */
	    nrgn->rgn_mprev = rgn;				/* of list */
	    nrgn->rgn_mnext = NULL;
	    rgn->rgn_mnext = elan3mmu->elan3mmu_mtail = nrgn;
	}
	else
	{
	    if (nbase >= base || ntop >= base)
		return (-1);					/* overlapping region */

	    nrgn->rgn_mnext = rgn;				/* insert before region */
	    nrgn->rgn_mprev = rgn->rgn_mprev;
	    rgn->rgn_mprev  = nrgn;
	    if (elan3mmu->elan3mmu_mrgns == rgn)
		elan3mmu->elan3mmu_mrgns = nrgn;
	    else
		nrgn->rgn_mprev->rgn_mnext = nrgn;
	}
    }
    elan3mmu->elan3mmu_mrgnlast = nrgn;
    
    return (0);
}

ELAN3MMU_RGN *
elan3mmu_removergn_main (ELAN3MMU *elan3mmu, caddr_t addr)
{
    ELAN3MMU_RGN *rgn = elan3mmu_findrgn_main (elan3mmu, addr, 0);
    
    ASSERT (SPINLOCK_HELD (&elan3mmu->elan3mmu_dev->IntrLock) && SPINLOCK_HELD (&elan3mmu->elan3mmu_lock));

    if (rgn == NULL || rgn->rgn_mbase != addr)
	return (NULL);
    
    elan3mmu->elan3mmu_mrgnlast = rgn->rgn_mnext;
    if (rgn == elan3mmu->elan3mmu_mtail)
	elan3mmu->elan3mmu_mtail = rgn->rgn_mprev;
    else
	rgn->rgn_mnext->rgn_mprev = rgn->rgn_mprev;
    
    if (rgn == elan3mmu->elan3mmu_mrgns)
	elan3mmu->elan3mmu_mrgns = rgn->rgn_mnext;
    else
	rgn->rgn_mprev->rgn_mnext = rgn->rgn_mnext;

    return (rgn);
}

ELAN3MMU_RGN *
elan3mmu_rgnat_main (ELAN3MMU *elan3mmu, caddr_t addr)
{
    ELAN3MMU_RGN *rgn = elan3mmu_findrgn_main (elan3mmu, addr, 0);
    caddr_t       base;

    if (rgn != NULL && (base = rgn->rgn_mbase) <= addr && addr <= (base + rgn->rgn_len - 1))
	return (rgn);
    return (NULL);
}

int
elan3mmu_setperm (ELAN3MMU *elan3mmu,
		  caddr_t   maddr,
		  E3_Addr   eaddr,
		  u_int	    len,
		  u_int	    perm)
{
    ELAN3_DEV     *dev = elan3mmu->elan3mmu_dev;
    ELAN3MMU_RGN *nrgn;
    unsigned long  flags;

    HAT_PRINTF4 (1, "elan3mmu_setperm: user %p elan %08x len %x perm %x\n", maddr, eaddr, len, perm);

    if ((((uintptr_t) maddr) & PAGEOFFSET) || (eaddr & PAGEOFFSET) || (len & PAGEOFFSET)) 
    {
        HAT_PRINTF0 (1, "elan3mmu_setperm:  alignment failure\n");
	return (EINVAL);
    }

    if (((uintptr_t) maddr + len - 1) < (uintptr_t) maddr || ((u_long)eaddr + len - 1) < (u_long)eaddr) 
    {
	HAT_PRINTF0 (1, "elan3mmu_setperm:  range failure\n");
	return (EINVAL);
    }

    ALLOC_ELAN3MMU_RGN(nrgn, TRUE);
    
    spin_lock (&elan3mmu->elan3mmu_lock);
    nrgn->rgn_mbase = maddr;
    nrgn->rgn_ebase = eaddr;
    nrgn->rgn_len   = len;
    nrgn->rgn_perm  = perm;

    spin_lock_irqsave (&dev->IntrLock, flags);
    if (elan3mmu_addrgn_elan (elan3mmu, nrgn) < 0)
    {
	HAT_PRINTF0 (1, "elan3mmu_setperm:  elan address exists\n");
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	spin_unlock (&elan3mmu->elan3mmu_lock);

	FREE_ELAN3MMU_RGN (nrgn);
	return (EINVAL);
    }
    
    if (elan3mmu_addrgn_main (elan3mmu, nrgn) < 0)
    {
	HAT_PRINTF0 (1, "elan3mmu_setperm:  main address exists\n");
	elan3mmu_removergn_elan (elan3mmu, eaddr);

	spin_unlock_irqrestore (&dev->IntrLock, flags);
	spin_unlock (&elan3mmu->elan3mmu_lock);

	FREE_ELAN3MMU_RGN (nrgn);
	return (EINVAL);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);
    spin_unlock (&elan3mmu->elan3mmu_lock);

    return (ESUCCESS);
}

void
elan3mmu_clrperm (ELAN3MMU *elan3mmu,
		  E3_Addr   addr,
		  u_int	    len)
{
    E3_Addr       raddr;
    E3_Addr       rtop;
    ELAN3MMU_RGN *nrgn;
    ELAN3MMU_RGN *rgn;
    ELAN3MMU_RGN *rgn_next;
    u_int	  ssize;
    unsigned long flags;
    int		  res;

    HAT_PRINTF2 (1, "elan3mmu_clrperm: elan %08x len %x\n", addr, len);

    raddr = (addr & PAGEMASK);
    rtop = ((addr + len - 1) & PAGEMASK) + PAGEOFFSET;

    ALLOC_ELAN3MMU_RGN (nrgn, TRUE);

    spin_lock (&elan3mmu->elan3mmu_lock);
    
    for (rgn = elan3mmu_findrgn_elan (elan3mmu, addr, 0); rgn != NULL; rgn = rgn_next)
    {
	if (rtop < rgn->rgn_ebase)				/* rtop was in a gap */
	    break;
	
	rgn_next = rgn->rgn_enext;				/* Save next region pointer */
	
	if (raddr <= rgn->rgn_ebase && rtop >= (rgn->rgn_ebase + rgn->rgn_len - 1)) 
	{
	    /* whole region is cleared */
	    elan3mmu_unload (elan3mmu, rgn->rgn_ebase, rgn->rgn_len, PTE_UNLOAD);
	    
	    spin_lock_irqsave (&elan3mmu->elan3mmu_dev->IntrLock, flags);
	    elan3mmu_removergn_elan (elan3mmu, rgn->rgn_ebase);
	    elan3mmu_removergn_main (elan3mmu, rgn->rgn_mbase);
	    spin_unlock_irqrestore (&elan3mmu->elan3mmu_dev->IntrLock, flags);

	    FREE_ELAN3MMU_RGN (rgn);
	}
	else if (raddr <= rgn->rgn_ebase)
	{
	    /* clearing at beginning, so shrink size and increment base ptrs */
	    ssize = rtop - rgn->rgn_ebase + 1;

	    elan3mmu_unload (elan3mmu, rgn->rgn_ebase, ssize, PTE_UNLOAD);
	    
	    spin_lock_irqsave (&elan3mmu->elan3mmu_dev->IntrLock, flags);
	    rgn->rgn_mbase += ssize;
	    rgn->rgn_ebase += ssize;
	    rgn->rgn_len   -= ssize;
	    spin_unlock_irqrestore (&elan3mmu->elan3mmu_dev->IntrLock, flags);
	    
	}
	else if (rtop >= (rgn->rgn_ebase + rgn->rgn_len - 1))
	{
	    /* clearing at end, so just shrink length of region */
	    ssize = ((rgn->rgn_ebase + rgn->rgn_len - 1) - raddr) + 1;

	    elan3mmu_unload (elan3mmu, raddr, ssize, PTE_UNLOAD);

	    spin_lock_irqsave (&elan3mmu->elan3mmu_dev->IntrLock, flags);
	    rgn->rgn_len -= ssize;
	    spin_unlock_irqrestore (&elan3mmu->elan3mmu_dev->IntrLock, flags);
	}
	else
	{
	    /* the section to go is in the middle,  so need to  */
	    /* split it into two regions */
	    elan3mmu_unload (elan3mmu, raddr, rtop - raddr + 1, PTE_UNLOAD);

	    spin_lock_irqsave (&elan3mmu->elan3mmu_dev->IntrLock, flags);

	    ASSERT (nrgn != NULL);

	    nrgn->rgn_mbase = rgn->rgn_mbase + (rtop - rgn->rgn_ebase + 1);;
	    nrgn->rgn_ebase = rtop + 1;
	    nrgn->rgn_len   = ((rgn->rgn_ebase + rgn->rgn_len - 1) - rtop);
	    nrgn->rgn_perm  = rgn->rgn_perm;

	    rgn->rgn_len = (raddr - rgn->rgn_ebase);		/* shrink original region */

	    res = elan3mmu_addrgn_elan (elan3mmu, nrgn);	/* insert new region */
	    ASSERT (res == 0);					/* which cannot fail */

	    res = elan3mmu_addrgn_main (elan3mmu, nrgn);	
	    ASSERT (res == 0);
	    spin_unlock_irqrestore (&elan3mmu->elan3mmu_dev->IntrLock, flags);

	    nrgn = NULL;
	}
    }
    spin_unlock (&elan3mmu->elan3mmu_lock);

    if (nrgn != NULL)
	FREE_ELAN3MMU_RGN (nrgn);
}

int
elan3mmu_checkperm (ELAN3MMU *elan3mmu,
		    E3_Addr   addr,
		    u_int     len,
		    u_int     access)
{
    E3_Addr	 raddr = (((E3_Addr) addr) & PAGEMASK);
    u_int        rtop = ((addr + len - 1) & PAGEMASK) + PAGEOFFSET;
    u_int	 rsize = rtop - raddr + 1;
    ELAN3MMU_RGN *rgn;

    HAT_PRINTF3 (1, "elan3mmu_checkperm: user %08x len %x access %x\n", addr, len, access);
    
    
    if ((raddr + rsize - 1) < raddr)
	return (ENOMEM);
    
    spin_lock (&elan3mmu->elan3mmu_lock);
    if ((rgn = elan3mmu_rgnat_elan (elan3mmu, raddr)) == (ELAN3MMU_RGN *) NULL)
    {
	spin_unlock (&elan3mmu->elan3mmu_lock);
	return (ENOMEM);
    }
    else
    {
	register int ssize;
	
	for (; rsize != 0; rsize -= ssize, raddr += ssize)
	{
	    if (raddr > (rgn->rgn_ebase + rgn->rgn_len - 1))
	    {
		rgn  = rgn->rgn_enext;
		
		if (rgn == NULL || raddr != rgn->rgn_ebase)
		{
		    spin_unlock (&elan3mmu->elan3mmu_lock);
		    return (ENOMEM);
		}
	    }
	    if ((raddr + rsize - 1) > (rgn->rgn_ebase + rgn->rgn_len - 1))
		ssize = ((rgn->rgn_ebase + rgn->rgn_len - 1) - raddr) + 1;
	    else
		ssize = rsize;
	    
	    HAT_PRINTF4 (1, "elan3mmu_checkperm : rgn %x -> %x perm %x access %x\n",
			 rgn->rgn_ebase, rgn->rgn_ebase + rgn->rgn_len, rgn->rgn_perm, access);

	    if (ELAN3_INCOMPAT_ACCESS (rgn->rgn_perm, access))
	    {
		spin_unlock (&elan3mmu->elan3mmu_lock);
		return (EACCES);
	    }
	}
    }
    
    spin_unlock (&elan3mmu->elan3mmu_lock);
    
    return (ESUCCESS);
}

caddr_t
elan3mmu_mainaddr (ELAN3MMU *elan3mmu, E3_Addr addr)
{
    ELAN3MMU_RGN *rgn;
    caddr_t	  raddr;
    
    spin_lock (&elan3mmu->elan3mmu_lock);
    if ((rgn = elan3mmu_rgnat_elan (elan3mmu, addr)) == (ELAN3MMU_RGN *) NULL)
	raddr = NULL;
    else
	raddr = rgn->rgn_mbase + (addr - rgn->rgn_ebase);
    spin_unlock (&elan3mmu->elan3mmu_lock);

    return (raddr);
}

E3_Addr
elan3mmu_elanaddr (ELAN3MMU *elan3mmu, caddr_t addr)
{
    ELAN3MMU_RGN *rgn;
    E3_Addr       raddr;

    spin_lock (&elan3mmu->elan3mmu_lock);
    if ((rgn = elan3mmu_rgnat_main (elan3mmu, addr)) == (ELAN3MMU_RGN *) NULL)
	raddr = (E3_Addr) 0;
    else
	raddr = rgn->rgn_ebase + (addr - rgn->rgn_mbase);
    spin_unlock (&elan3mmu->elan3mmu_lock);

    return (raddr);
}

void
elan3mmu_displayrgns(ELAN3MMU *elan3mmu)
{
    ELAN3MMU_RGN *rgn;

    spin_lock (&elan3mmu->elan3mmu_lock);
    HAT_PRINTF0 (1, "elan3mmu_displayrgns: main regions\n");
    for (rgn = elan3mmu->elan3mmu_mrgns; rgn; rgn = (rgn->rgn_mnext == elan3mmu->elan3mmu_mrgns) ? NULL : rgn->rgn_mnext)
	HAT_PRINTF5 (1, "    RGN %p ebase %08x mbase %p len %08x perm %08x\n", rgn, rgn->rgn_ebase, rgn->rgn_mbase, rgn->rgn_len, rgn->rgn_perm);
    HAT_PRINTF0 (1, "elan3mmu_displayrgns: elan regions\n");
    for (rgn = elan3mmu->elan3mmu_ergns; rgn; rgn = (rgn->rgn_enext == elan3mmu->elan3mmu_ergns) ? NULL : rgn->rgn_enext)
	HAT_PRINTF5 (1, "    RGN %p ebase %08x mbase %p len %08x perm %08x\n", rgn, rgn->rgn_ebase, rgn->rgn_mbase, rgn->rgn_len, rgn->rgn_perm);

    spin_unlock (&elan3mmu->elan3mmu_lock);
}

/*============================================================================*/
/* Private functions */
#define ELAN3_PTE_IS_VALID(ptbl, pte)	\
          ((ptbl->ptbl_flags & PTBL_KERNEL) ? \
          (pte&(~ELAN3_PTE_REF)) != elan3mmu_kernel_invalid_pte(ptbl->ptbl_elan3mmu) : \
          ELAN3_PTE_VALID(pte))

void
elan3mmu_expand (ELAN3MMU *elan3mmu, E3_Addr addr, int len, int level, int attr)
{
    ELAN3_PTBL          *ptbl;
    sdramaddr_t		pte;
    spinlock_t	       *lock;
    u_int		span;
    unsigned long       flags;

    HAT_PRINTF3 (1, "elan3mmu_expand: elan3mmu %p %08x to %08x\n", elan3mmu, 
		 addr, addr + len);

    for ( ; len != 0; addr += span, len -= span)
    {
	/* as we asked for level 3 we know its a pte */
	pte = elan3mmu_ptealloc (elan3mmu, addr, level, &ptbl, &lock, attr, &flags);

	switch (level)
	{
	case PTBL_LEVEL_3:
	    span = MIN(len, ELAN3_L3_PTSIZE - ((E3_Addr) addr & ELAN3_L3_PTOFFSET));
	    break;
	case PTBL_LEVEL_2:
	    span = MIN(len, ELAN3_L2_PTSIZE - ((E3_Addr) addr & ELAN3_L2_PTOFFSET));
	    break;
	default:
	    span = len;
	    break;
	}
	
	if (pte != (sdramaddr_t) 0)
	    elan3mmu_unlock_ptbl (ptbl, lock, flags);
    }
}

void
elan3mmu_reserve (ELAN3MMU *elan3mmu, E3_Addr addr, u_int npages, sdramaddr_t *ptes)
{
    ELAN3_PTBL          *ptbl;
    sdramaddr_t		pte;
    spinlock_t	       *lock;
    u_int		span;
    int			len;
    int			i;
    unsigned long       flags;

    HAT_PRINTF3 (1, "elan3mmu_reserve: elan3mmu %p %08x to %08x\n", elan3mmu, 
		 addr, addr + (npages << ELAN3_PAGE_SHIFT));

    for (len = (npages << ELAN3_PAGE_SHIFT); len != 0; addr += span, len -= span)
    {
	/* as we asked for level 3 we know its a pte */
	pte = elan3mmu_ptealloc (elan3mmu, addr, 3, &ptbl, &lock, 0, &flags);

	span = MIN(len, ELAN3_L3_PTSIZE - ((E3_Addr) addr & ELAN3_L3_PTOFFSET));
	
	if (ptes != NULL)
	{
	    for (i = 0; i < span; i += ELAN3_PAGE_SIZE, pte += ELAN3_PTE_SIZE)
		*ptes++ = pte;
	    ptbl->ptbl_valid += (span >> ELAN3_PAGE_SHIFT);

	    HAT_PRINTF4 (2, "elan3mmu_reserve: inc valid for level %d ptbl %p to %d   (%d)\n", 
		     PTBL_LEVEL(ptbl->ptbl_flags), ptbl, ptbl->ptbl_valid, (span >> ELAN3_PAGE_SHIFT));

	}

	elan3mmu_unlock_ptbl (ptbl, lock, flags);
    }
}

void
elan3mmu_release (ELAN3MMU *elan3mmu, E3_Addr addr, u_int npages, sdramaddr_t *ptes)
{
    ELAN3_DEV           *dev = elan3mmu->elan3mmu_dev;
    ELAN3_PTBL          *ptbl;
    sdramaddr_t		pte;
    ELAN3_PTE		tpte;
    spinlock_t	       *lock;
    u_int		span;
    int			len;
    int			i;
    int			level;
    unsigned long       flags;
    
    HAT_PRINTF3 (1, "elan3mmu_release: elan3mmu %p %08x to %08x\n", elan3mmu, 
		 addr, addr + (npages << ELAN3_PAGE_SHIFT));

    if (ptes == NULL)
	return;

    tpte = elan3mmu_kernel_invalid_pte (elan3mmu);

    for (len = (npages << ELAN3_PAGE_SHIFT); len != 0; addr += span, len -= span)
    {
	/* as we asked for level 3 we know its a pte */
	pte = elan3mmu_ptefind(elan3mmu, addr, &level, &ptbl, &lock, &flags);
	ASSERT (level == PTBL_LEVEL_3);

	span = MIN(len, ELAN3_L3_PTSIZE - ((E3_Addr) addr & ELAN3_L3_PTOFFSET));


	for (i = 0 ; i < span; i += ELAN3_PAGE_SIZE, pte += ELAN3_PTE_SIZE)
	    elan3_writepte (dev, pte, tpte);
	ptbl->ptbl_valid -= (span >> ELAN3_PAGE_SHIFT);

	HAT_PRINTF3 (2, "elan3mmu_release: inc valid for level %d ptbl %p to %d\n", 
		     PTBL_LEVEL(ptbl->ptbl_flags), ptbl, ptbl->ptbl_valid);

	elan3mmu_unlock_ptbl (ptbl, lock, flags);
    }
    ElanFlushTlb (elan3mmu->elan3mmu_dev);
}

void
elan3mmu_pteload (ELAN3MMU *elan3mmu, int level, E3_Addr addr, physaddr_t paddr, int perm, int attr)
    
{
    ELAN3_DEV     *dev;
    ELAN3_PTBL    *ptbl;
    spinlock_t   *lock;
    unsigned long flags;
    ELAN3_PTE      newpte;
    ELAN3_PTE      oldpte;
    sdramaddr_t   pte;

    ASSERT((level == PTBL_LEVEL_2) || (level == PTBL_LEVEL_3));

    /* Generate the new pte which we're going to load */
    dev = elan3mmu->elan3mmu_dev;

    newpte = elan3mmu_phys_to_pte (dev, paddr, perm);
    
    if (attr & PTE_LOAD_BIG_ENDIAN)
	newpte |= ELAN3_PTE_BIG_ENDIAN;

    HAT_PRINTF4 (1, "elan3mmu_pteload: elan3mmu %p level %d addr %x pte %llx\n", elan3mmu, level, addr, (long long) newpte);
    HAT_PRINTF5 (1, "elan3mmu_pteload:%s%s%s perm=%d phys=%llx\n",
		 (newpte & ELAN3_PTE_LOCAL)  ? " local" : "",
		 (newpte & ELAN3_PTE_64_BIT)     ? " 64 bit" : "",
		 (newpte & ELAN3_PTE_BIG_ENDIAN) ? " big-endian" : " little-endian",
		 (u_int) (newpte & ELAN3_PTE_PERM_MASK) >> ELAN3_PTE_PERM_SHIFT,
		 (unsigned long long) (newpte & ELAN3_PTE_PFN_MASK));
		  
    if (level == PTBL_LEVEL_3)
	pte = elan3mmu_ptealloc (elan3mmu, addr, level, &ptbl, &lock, attr, &flags);
    else
    {
	sdramaddr_t ptp = elan3mmu_ptealloc (elan3mmu, addr, level, &ptbl, &lock, attr, &flags);

	pte = elan3mmu_ptp2pte (elan3mmu, ptp, level);

	HAT_PRINTF3 (2, "elan3mmu_pteload: level %d ptp at %lx => pte at %lx\n", level, ptp, pte);
    }

    if (pte == (sdramaddr_t) 0)
    {
	ASSERT (level == PTBL_LEVEL_3 && (attr & (PTE_NO_SLEEP | PTE_NO_STEAL)) == (PTE_NO_SLEEP | PTE_NO_STEAL));
	return;
    }

    ASSERT (ptbl->ptbl_elan3mmu == elan3mmu);
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == level);
    ASSERT (PTBL_IS_LOCKED (ptbl->ptbl_flags));
    
    oldpte = elan3_readpte (dev, pte);

    HAT_PRINTF3 (2, "elan3mmu_pteload: modify pte at %lx from %llx to %llx\n", pte, (long long) oldpte, (long long) newpte);

    if (ELAN3_PTE_IS_VALID(ptbl, oldpte))
    {
	ELAN3MMU_STAT(ptereload);

	ASSERT ((newpte & ~((E3_uint64)ELAN3_PTE_PERM_MASK | ELAN3_RM_MASK)) == (oldpte & ~((E3_uint64)ELAN3_PTE_PERM_MASK | ELAN3_RM_MASK)));
	
	if ((newpte & ~ELAN3_RM_MASK) != (oldpte & ~ELAN3_RM_MASK))
	{
	    /* We're modifying a valid translation, it must be mapping the same page */
	    /* so we use elan3_modifypte to not affect the referenced and modified bits */
	    elan3_modifypte (dev, pte, newpte);


	    ElanFlushTlb (elan3mmu->elan3mmu_dev);
	}
    }
    else
    {
	ELAN3MMU_STAT(pteload);

	ptbl->ptbl_valid++;

	HAT_PRINTF3 (2, "elan3mmu_pteload: inc valid for level %d ptbl %p to %d\n", 
		     PTBL_LEVEL(ptbl->ptbl_flags), ptbl, ptbl->ptbl_valid);

	HAT_PRINTF2 (2, "elan3mmu_pteload: write pte %lx to %llx\n", pte, (long long) newpte);

	elan3_writepte (dev, pte, newpte);

	if (ptbl->ptbl_flags & PTBL_KERNEL)
	    ElanFlushTlb (elan3mmu->elan3mmu_dev);

    }

    elan3mmu_unlock_ptbl (ptbl, lock, flags);
}

void
elan3mmu_unload (ELAN3MMU *elan3mmu, E3_Addr addr, u_int len, int attr)
{
    ELAN3_PTBL          *ptbl;
    sdramaddr_t         ptp;
    spinlock_t	       *lock;
    int			level;
    u_int		span;
    unsigned long	flags;

    HAT_PRINTF3(1, "elan3mmu_unload (elan3mmu %p addr %x -> %x)\n", elan3mmu, addr, addr+len-1);

    for (; len != 0; addr += span, len -= span)
    {
	ptp  = elan3mmu_ptefind(elan3mmu, addr, &level, &ptbl, &lock, &flags);

	span = MIN(len, ELAN3_L3_PTSIZE - ((E3_Addr) addr & ELAN3_L3_PTOFFSET));

	if (ptp != (sdramaddr_t) 0)
	{
	    HAT_PRINTF2 (2, "elan3mmu_unload: unload [%x,%x]\n", addr, addr + span);
	    
	    if ( level ==  PTBL_LEVEL_3 ) 
		elan3mmu_unload_loop (elan3mmu, ptbl, ptp - PTBL_TO_PTADDR(ptbl), span >> ELAN3_PAGE_SHIFT, attr);
	    else
	    {
		ELAN3_PTP    invalidptp = ELAN3_INVALID_PTP;
		ELAN3_DEV   *dev = elan3mmu->elan3mmu_dev;
		ELAN3_PTBL  *lXptbl;
		ELAN3_PTP    tptp;
		int         idx;

		tptp = elan3_readptp (elan3mmu->elan3mmu_dev, ptp);

		ASSERT (ELAN3_PTP_TYPE(tptp) == ELAN3_ET_PTE);

		lXptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tptp);
		idx    = (PTP_TO_PT_PADDR(tptp) - PTBL_TO_PTADDR(lXptbl))/ELAN3_PTE_SIZE;

		if ( level == PTBL_LEVEL_1) 
		    span = MIN(len, ELAN3_L2_PTSIZE - ((E3_Addr) addr & ELAN3_L2_PTOFFSET));
		else
		    span = MIN(len, ELAN3_L3_PTSIZE - ((E3_Addr) addr & ELAN3_L3_PTOFFSET));

		/* invalidate the ptp. */
		elan3_writeptp (dev, ptp, invalidptp);
		if (! (attr & PTE_UNLOAD_NOFLUSH))
		    ElanFlushTlb (dev);	    
    
		elan3mmu_free_pte ( dev, elan3mmu, lXptbl, idx); 

		ptbl->ptbl_valid--;

		HAT_PRINTF3 (2, "elan3mmu_unload: dec valid for level %d ptbl %p to %d\n", 
			     PTBL_LEVEL(ptbl->ptbl_flags), ptbl, ptbl->ptbl_valid);	

	    }
	    elan3mmu_unlock_ptbl (ptbl, lock, flags);
	}
    }
}

static void
elan3mmu_unload_loop (ELAN3MMU *elan3mmu, ELAN3_PTBL *ptbl, int first_valid, int nptes, int flags)
{
    ELAN3_DEV   *dev = elan3mmu->elan3mmu_dev;
    sdramaddr_t pte;
    ELAN3_PTE    tpte;
    int         last_valid = first_valid + nptes;
    int	        i;
    
    HAT_PRINTF3 (1, "elan3mmu_unloadloop: ptbl %p entries [%d->%d]\n", ptbl, first_valid, last_valid);

    ASSERT (PTBL_IS_LOCKED (ptbl->ptbl_flags));
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == PTBL_LEVEL_3);
    
    pte = PTBL_TO_PTADDR(ptbl) + first_valid;
    
    for (i = first_valid; i < last_valid; i++, pte += ELAN3_PTE_SIZE)
    {
	if (ptbl->ptbl_valid == 0)
	    break;

	tpte = elan3_readpte (dev, pte);
	if (! ELAN3_PTE_IS_VALID(ptbl, tpte))
	    continue;
	
	elan3mmu_pteunload (ptbl, pte, flags, NO_MLIST_LOCK);
    }
}

void
elan3mmu_pteunload (ELAN3_PTBL *ptbl, sdramaddr_t pte, int flags, int got_mlist_lock)
{
    ELAN3_DEV   *dev = ptbl->ptbl_elan3mmu->elan3mmu_dev;
    ELAN3_PTE    tpte;

    ASSERT (PTBL_LEVEL (ptbl->ptbl_flags) == PTBL_LEVEL_3);
    ASSERT (PTBL_IS_LOCKED (ptbl->ptbl_flags));

    HAT_PRINTF2 (1, "elan3mmu_pteunload: ptbl %p pte %lx\n", ptbl, pte);

    ELAN3MMU_STAT (pteunload);

    elan3_invalidatepte (dev, pte);

    if (! (flags & PTE_UNLOAD_NOFLUSH))
	ElanFlushTlb (dev);
    
    tpte = ELAN3_INVALID_PTE;
    elan3_writepte (dev, pte, tpte);
    
    if (ptbl->ptbl_flags & PTBL_KERNEL)
    {
	tpte = elan3mmu_kernel_invalid_pte(ptbl->ptbl_elan3mmu);

	elan3_writepte (dev, pte, tpte);
    }

    ptbl->ptbl_valid--;

    HAT_PRINTF3 (2, "elan3mmu_pteunload: dec valid for level %d ptbl %p to %d\n", 
		 PTBL_LEVEL(ptbl->ptbl_flags), ptbl, ptbl->ptbl_valid);

}

void
elan3mmu_ptesync (ELAN3_PTBL *ptbl, sdramaddr_t pte, int flags, int got_mlist_lock)
{

}

/*
 * Create more page tables at a given level for this Elan.
 */
static ELAN3_PTBL *
elan3mmu_create_ptbls (ELAN3_DEV *dev, int level, int attr, int keep)
{
    sdramaddr_t	  pts;
    ELAN3_PTBL    *ptbl;
    ELAN3_PTBL    *first;
    ELAN3_PTBL    *last;
    ELAN3_PTBL_GR *ptg;
    register int  i;
    register int  inc;
    
    HAT_PRINTF1 (2, "elan3mmu_create_ptbls: create level %d ptbls\n", level);

    pts = elan3_sdram_alloc (dev, PTBL_GROUP_SIZE);
    if (pts == (sdramaddr_t) 0)
    {
	HAT_PRINTF0 (2, "elan3mmu_create_ptbls: cannot map elan pages\n");

	ELAN3MMU_STAT (create_ptbl_failed);
	return (NULL);
    }
    
    HAT_PRINTF1 (2, "elan3mmu_create_ptbls: pts at %lx\n", pts);
    
    ALLOC_PTBL_GR (ptg, !(attr & PTE_NO_SLEEP)); 	        /* Allocate the group of page tables */
    if (ptg == NULL)						/* for this page */
    {
	HAT_PRINTF0 (2, "elan3mmu_create_ptbls: cannot allocate page table group\n");

	elan3_sdram_free (dev, pts, PTBL_GROUP_SIZE);

	ELAN3MMU_STAT (create_ptbl_failed);
	return (NULL);
    }

    HAT_PRINTF1 (2, "elan3mmu_create_ptbls: ptg is %p\n", ptg);
    
    ElanSetPtblGr (dev, pts, ptg);
    
    HAT_PRINTF4 (2, "elan3mmu_create_ptbls: zeroing %d bytes at %lx, %d bytes at %p\n",
		 PTBL_GROUP_SIZE, pts, (int) sizeof (ELAN3_PTBL_GR), ptg);

#ifndef zero_all_ptbls
    elan3_sdram_zeroq_sdram (dev, pts, PTBL_GROUP_SIZE);		/* Ensure that all PTEs/PTPs are invalid */
#endif
    bzero ((caddr_t) ptg, sizeof (ELAN3_PTBL_GR));
    
    ptg->pg_addr  = pts;
    ptg->pg_level = level;

    ptbl = ptg->pg_ptbls;					/* Initialise the index in all page tables */
    for (i = 0; i < PTBLS_PER_GROUP_MAX; i++)
    {
	ptbl->ptbl_index = (u_char) i;
	ptbl->ptbl_next  = (ELAN3_PTBL *) 0xdeaddead;
	ptbl++;
    }
    
    switch (level)						/* Determine the number of ptbls we can  */
    {								/* allocate from this page, by jumping  */
    case PTBL_LEVEL_X: inc = PTBLS_PER_PTBL_LX; break;	        /* multiples of the smallest. */
    case PTBL_LEVEL_1: inc = PTBLS_PER_PTBL_L1; break;
    case PTBL_LEVEL_2: inc = PTBLS_PER_PTBL_L2; break;
    case PTBL_LEVEL_3: inc = PTBLS_PER_PTBL_L3; break;
    default:           inc = PTBLS_PER_PTBL_L3; break;
    }

    ptbl = ptg->pg_ptbls;					/* Chain them together */
    for (i = 0; i < PTBLS_PER_GROUP_MAX; i += inc, ptbl += inc)
	ptbl->ptbl_next = ptbl + inc;

    first = ptg->pg_ptbls;					/* Determine list of */
    last  = first + PTBLS_PER_GROUP_MAX - inc;			/* ptbls to add to free list */
    if (! keep)
	ptbl = NULL;
    else
    {
	ptbl  = first;
	first = first->ptbl_next;
    }
    
    spin_lock (&dev->Level[level].PtblLock);
    dev->Level[level].PtblTotal     += PTBLS_PER_GROUP_MAX/inc;		/* Increment the counts */
    dev->Level[level].PtblFreeCount += PTBLS_PER_GROUP_MAX/inc;

    ELAN3MMU_SET_STAT (num_ptbl_level[level], dev->Level[level].PtblTotal);

    if (keep)
	dev->Level[level].PtblFreeCount--;
    
    last->ptbl_next = dev->Level[level].PtblFreeList;			/* And add to free list */
    dev->Level[level].PtblFreeList = first;
    spin_unlock (&dev->Level[level].PtblLock);
    
    spin_lock (&dev->PtblGroupLock);
    ptg->pg_next = dev->Level[level].PtblGroupList;
    dev->Level[level].PtblGroupList = ptg;
    spin_unlock (&dev->PtblGroupLock);

    HAT_PRINTF1 (2, "elan3mmu_create_ptbls: returning ptbl %p\n", ptbl);
    
    return (ptbl);
}

static ELAN3_PTBL *
elan3mmu_ta_to_ptbl (ELAN3MMU *elan3mmu, ELAN3_PTP *ptp)
{
    E3_Addr	  ptpa  = PTP_TO_PT_PADDR(*ptp);
    ELAN3_PTBL_GR *pg    = ElanGetPtblGr (elan3mmu->elan3mmu_dev, (sdramaddr_t)ptpa & ~(PTBL_GROUP_SIZE-1));
    
    return (pg->pg_ptbls + ((ptpa - pg->pg_addr) >> ELAN3_PT_SHIFT));
}

static ELAN3_PTBL *
elan3mmu_alloc_lXptbl (ELAN3_DEV *dev, int attr,  ELAN3MMU *elan3mmu)
{
    ELAN3_PTBL *ptbl = NULL;

    spin_lock (&dev->Level[PTBL_LEVEL_X].PtblLock);
    if (dev->Level[PTBL_LEVEL_X].PtblFreeList)
    {
	ptbl = dev->Level[PTBL_LEVEL_X].PtblFreeList;

	HAT_PRINTF1 (2, "elan3mmu_alloc_lXptbl: found ptbl %p on free list\n", ptbl);

	dev->Level[PTBL_LEVEL_X].PtblFreeList = ptbl->ptbl_next;
	dev->Level[PTBL_LEVEL_X].PtblFreeCount--;
    }
    spin_unlock (&dev->Level[PTBL_LEVEL_X].PtblLock);
    
    if (ptbl == NULL) 
    {
	ptbl = elan3mmu_create_ptbls (dev, PTBL_LEVEL_X, attr, 1);

	HAT_PRINTF1 (2, "elan3mmu_alloc_lXptbl: created level X ptbl %p\n", ptbl);
    }

    if (ptbl == NULL)
    {
	if ((attr & PTE_NO_STEAL))
	{
	    HAT_PRINTF0 (2, "elan3mmu_alloc_lXptbl: not allowed to steal ptbl for use at level 2\n");
	    return NULL;
	}

	ELAN3MMU_STAT(lX_alloc_l3);

	ptbl = elan3mmu_steal_l3ptbl (dev, attr);
	
	HAT_PRINTF1 (2, "elan3mmu_alloc_lXptbl: stolen level3 ptbl %p used as level 2\n", ptbl);
    }

    ptbl->ptbl_elan3mmu = elan3mmu;
    ptbl->ptbl_base     = 0;
    ptbl->ptbl_parent   = 0;
    ptbl->ptbl_flags    = PTBL_LEVEL_X | PTBL_ALLOCED;
    
    HAT_PRINTF2 (2, "elan3mmu_alloc_lXptbl: ptbl %p dev %p\n", ptbl, dev);

#ifdef zero_all_ptbls
    elan3_sdram_zero_sdarm (dev, PTBL_TO_PTADDR(ptbl), ELAN3_LX_ENTRIES*ELAN3_PTE_SIZE);
#endif

    return (ptbl);
}

static ELAN3_PTBL *
elan3mmu_alloc_pte (ELAN3_DEV *dev, ELAN3MMU *elan3mmu, int *idx)
{
    ELAN3_PTBL   * ptbl_ptr;
    int           index;

    /* lock whilst looking for space */
    spin_lock (&elan3mmu->elan3mmu_lXptbl_lock);
    
    /* walk the lXptbl list */
    ptbl_ptr = elan3mmu->elan3mmu_lXptbl;
    while ( ptbl_ptr != NULL ) 
    {
	/* does this ptlb have any free ones */
	if (  (index = ptbl_ptr->ptbl_valid) < ELAN3_LX_ENTRIES) 
	{
	    /*  better to search  from valid count as its likly to be free */
	    index = ptbl_ptr->ptbl_valid; 
	    do {
		if ((ptbl_ptr->ptbl_base & (1 << index)) == 0)
		    goto found;

		/* move index on and wrap back to start if needed */
		if ((++index) == ELAN3_LX_ENTRIES) 
		    index = 0;
	    } while (index != ptbl_ptr->ptbl_valid);

	    panic ("elan3mmu_alloc_pte: has ptbl valid < 32 when but no free pte's");
 	}
	ptbl_ptr = ptbl_ptr->ptbl_parent;
    }
	
    /* unlock so we can create space */
    spin_unlock (&elan3mmu->elan3mmu_lXptbl_lock); 

    /* if create some more */
    ptbl_ptr = elan3mmu_alloc_lXptbl(dev, 0, elan3mmu);

    /* get the lock again */
    spin_lock (&elan3mmu->elan3mmu_lXptbl_lock);
	
    /* add to front of list as its obviously got free ones on it */
    ptbl_ptr->ptbl_parent     = elan3mmu->elan3mmu_lXptbl;
    elan3mmu->elan3mmu_lXptbl = ptbl_ptr;

    /* grap the first one */
    index = 0;
    
 found:
    ptbl_ptr->ptbl_base |= (1 << index);
    ptbl_ptr->ptbl_valid++;

    HAT_PRINTF3 (2, "elan3mmu_alloc_pte: inc valid for level %d ptbl %p to %d\n", 
		 PTBL_LEVEL(ptbl_ptr->ptbl_flags), ptbl_ptr, ptbl_ptr->ptbl_valid);

    /* release the loc and return it */
    spin_unlock (&elan3mmu->elan3mmu_lXptbl_lock); 

    *idx = index;
    return (ptbl_ptr);
}

static ELAN3_PTBL *
elan3mmu_alloc_l1ptbl (ELAN3_DEV *dev, int attr, ELAN3MMU *elan3mmu)
{
    ELAN3_PTBL *ptbl = NULL;
    ELAN3_PTBL *p;
    int i,j;
    
    spin_lock (&dev->Level[PTBL_LEVEL_1].PtblLock);
    if (dev->Level[PTBL_LEVEL_1].PtblFreeList)
    {
	ptbl = dev->Level[PTBL_LEVEL_1].PtblFreeList;
	dev->Level[PTBL_LEVEL_1].PtblFreeList = ptbl->ptbl_next;
	dev->Level[PTBL_LEVEL_1].PtblFreeCount--;
    }
    spin_unlock (&dev->Level[PTBL_LEVEL_1].PtblLock);
    
    if (ptbl == NULL)
	ptbl = elan3mmu_create_ptbls (dev, PTBL_LEVEL_1, attr, 1);
    
    if (ptbl == NULL)
	panic ("elan3mmu_alloc_l1ptbl: cannot alloc ptbl");
    
    for (p = ptbl, j = i = 0; i < PTBLS_PER_PTBL_L1; i++, p++)
    {
	p->ptbl_elan3mmu = elan3mmu;
	p->ptbl_base     = VA2BASE (j);
	p->ptbl_flags    = PTBL_LEVEL_1 | PTBL_GROUPED;
	p->ptbl_parent   = NULL;
	
	j += L1_VA_PER_PTBL;
    }
    
    /* Now mark the real page table as allocated */
    /* level 1 ptbls are returned unlocked */
    ptbl->ptbl_flags = PTBL_LEVEL_1 | PTBL_ALLOCED;
    
    HAT_PRINTF2 (2, "elan3mmu_alloc_l1ptbl: ptbl %p dev %p\n", ptbl, dev);

#ifdef zero_all_ptbls
    elan3_sdram_zeroq_sdram (dev, PTBL_TO_PTADDR(ptbl), ELAN3_L1_ENTRIES*ELAN3_PTP_SIZE);
#endif

    return (ptbl);
}

static ELAN3_PTBL *
elan3mmu_alloc_l2ptbl (ELAN3_DEV *dev, int attr, ELAN3_PTBL *parent, ELAN3MMU *elan3mmu, E3_Addr base, spinlock_t **plock, unsigned long *flags)
{
    ELAN3_PTBL *ptbl = NULL;
    ELAN3_PTBL *p;
    int        i;
    int        j;
    unsigned long ptbl_flags;

    spin_lock_irqsave (&dev->Level[PTBL_LEVEL_2].PtblLock, ptbl_flags);
    if (dev->Level[PTBL_LEVEL_2].PtblFreeList)
    {
	ptbl = dev->Level[PTBL_LEVEL_2].PtblFreeList;

	HAT_PRINTF1 (2, "elan3mmu_alloc_l2ptbl: found ptbl %p on free list\n", ptbl);

	dev->Level[PTBL_LEVEL_2].PtblFreeList = ptbl->ptbl_next;
	dev->Level[PTBL_LEVEL_2].PtblFreeCount--;
    }
    spin_unlock_irqrestore (&dev->Level[PTBL_LEVEL_2].PtblLock, ptbl_flags);
    
    if (ptbl == NULL) 
    {
	ptbl = elan3mmu_create_ptbls (dev, PTBL_LEVEL_2, attr, 1);

	HAT_PRINTF1 (2, "elan3mmu_alloc_l2ptbl: created level 2 ptbl %p\n", ptbl);
    }

    if (ptbl == NULL)
    {
	if ((attr & PTE_NO_STEAL))
	{
	    HAT_PRINTF0 (2, "elan3mmu_alloc_l2ptbl: not allowted to steal ptbl for use at level 2\n");
	    return (NULL);
	}

	ELAN3MMU_STAT(l2_alloc_l3);

	ptbl = elan3mmu_steal_l3ptbl (dev, attr);
	
	HAT_PRINTF1 (2, "elan3mmu_alloc_l2ptbl: stolen level3 ptbl %p used as level 2\n", ptbl);
    }
    
    *plock = elan3mmu_ptbl_to_lock (PTBL_LEVEL_2, ptbl);
    spin_lock_irqsave (*plock, *flags);
    
    for (p = ptbl, j = i = 0; i < PTBLS_PER_PTBL_L2; i++, p++)
    {
	p->ptbl_elan3mmu = elan3mmu;
	p->ptbl_base     = VA2BASE (base + j);
	p->ptbl_flags    = PTBL_LEVEL_2 | PTBL_GROUPED;
	p->ptbl_parent   = parent;
	
	j += L2_VA_PER_PTBL;
    }
    
    ptbl->ptbl_flags  = PTBL_LEVEL_2 | PTBL_ALLOCED | PTBL_LOCKED;
    
    HAT_PRINTF3 (2, "elan3mmu_alloc_l2ptbl: ptbl %p dev %p base %x\n", ptbl, dev, base);

#ifdef zero_all_ptbls
    elan3_sdram_zero_sdarm (dev, PTBL_TO_PTADDR(ptbl), ELAN3_L2_ENTRIES*ELAN3_PTP_SIZE);
#endif

    return (ptbl);
}

static ELAN3_PTBL *
elan3mmu_alloc_l3ptbl (ELAN3_DEV *dev, int attr, ELAN3_PTBL *parent, ELAN3MMU *elan3mmu, E3_Addr base, spinlock_t **plock, unsigned long *flags)
{
    ELAN3_PTBL *ptbl = NULL;
    ELAN3_PTBL *p;
    int	       i;
    int	       j;
    unsigned long ptbl_flags;

    spin_lock_irqsave (&dev->Level[PTBL_LEVEL_3].PtblLock, ptbl_flags);
    if (dev->Level[PTBL_LEVEL_3].PtblFreeList)
    {
	HAT_PRINTF1 (2, "elan3mmu_alloc_l3ptbl: found ptbl %p on free list\n", ptbl);

	ptbl = dev->Level[PTBL_LEVEL_3].PtblFreeList;
	dev->Level[PTBL_LEVEL_3].PtblFreeList = ptbl->ptbl_next;
	dev->Level[PTBL_LEVEL_3].PtblFreeCount--;
    }
    spin_unlock_irqrestore (&dev->Level[PTBL_LEVEL_3].PtblLock, ptbl_flags);
    
    if (ptbl == NULL)
    {
	ptbl = elan3mmu_create_ptbls (dev, PTBL_LEVEL_3, attr, 1);

	HAT_PRINTF1 (2, "elan3mmu_alloc_l3ptbl: created level 3 ptbl %p\n", ptbl);
    }

    if (ptbl == NULL)
    {
	if ((attr & PTE_NO_STEAL))
	{
	    HAT_PRINTF0 (2, "elan3mmu_alloc_l3ptbl: not allowed to steal ptbl for use at level 3\n");
	    return (NULL);
	}

	ptbl = elan3mmu_steal_l3ptbl (dev, attr);

	HAT_PRINTF1 (2, "elan3mmu_alloc_l3ptbl: stolen level3 ptbl %p\n", ptbl);
    }
    
    *plock = elan3mmu_ptbl_to_lock (PTBL_LEVEL_3, ptbl);
    spin_lock_irqsave (*plock,*flags);
    
    for (p = ptbl, j = i = 0; i < PTBLS_PER_PTBL_L3; i++, p++)
    {
	p->ptbl_elan3mmu = elan3mmu;
	p->ptbl_base     = VA2BASE (base + j);
	p->ptbl_flags    = PTBL_LEVEL_3 | PTBL_GROUPED;
	p->ptbl_parent   = parent;
	
	j += L3_VA_PER_PTBL;
    }
    
    ptbl->ptbl_flags = PTBL_LEVEL_3 | PTBL_ALLOCED | PTBL_LOCKED;
    
    HAT_PRINTF3 (2, "elan3mmu_alloc_l3ptbl: ptbl %p dev %p base %x\n", ptbl, dev, base);

#ifdef zero_all_ptbls
    elan3_sdram_zeroq_sdram (dev, PTBL_TO_PTADDR(ptbl), ELAN3_L3_ENTRIES*ELAN3_PTE_SIZE);
#endif

    return (ptbl);
}

void 
elan3mmu_free_pte  (ELAN3_DEV *dev,  ELAN3MMU *elan3mmu,  ELAN3_PTBL *ptbl_ptr, int idx)
{  
    sdramaddr_t pte  = PTBL_TO_PTADDR (ptbl_ptr) | (idx * sizeof (ELAN3_PTE));
    ELAN3_PTE    tpte = ELAN3_INVALID_PTE;
    ELAN3_PTBL *prev;

    /* ensure that the pte is invalid when free */
    elan3_writepte (dev, pte, tpte);

    /* lock whilst removing */
    spin_lock (&elan3mmu->elan3mmu_lXptbl_lock);

    HAT_PRINTF4 (2, "elan3mmu_free_pte idx %d   ptbl_ptr %p ptbl_base  %x  ptbl_ptr->ptbl_valid %d \n", 
		 idx, ptbl_ptr, ptbl_ptr->ptbl_base, ptbl_ptr->ptbl_valid);
    /* make sure it was set */
    ASSERT ( ptbl_ptr->ptbl_base & (1 << idx) ); 
    ASSERT ( ptbl_ptr->ptbl_valid > 0  );

    ptbl_ptr->ptbl_base &= ~(1 << idx);
    ptbl_ptr->ptbl_valid--;

    HAT_PRINTF3 (2, "elan3mmu_free_pte: dec valid for level %d ptbl %p to %d\n", 
		 PTBL_LEVEL(ptbl_ptr->ptbl_flags), ptbl_ptr, ptbl_ptr->ptbl_valid); 
 
    /* was that the last one on this page */
    if ( ! ptbl_ptr->ptbl_valid ) 
    {
	/* so no bits should be set then */
	ASSERT ( ptbl_ptr->ptbl_base == 0 );

	/* is this the first page ?? */
	if ( elan3mmu->elan3mmu_lXptbl == ptbl_ptr ) 
	{
	    /* make the list start at the second element */
	     elan3mmu->elan3mmu_lXptbl = ptbl_ptr->ptbl_parent;

	     /* put ptbl back on free list */
	     elan3mmu_free_lXptbl(dev, ptbl_ptr);

	     /* unlock and return */
	     spin_unlock (&elan3mmu->elan3mmu_lXptbl_lock);
	     return ;
	}

	/* scan thro list looking for this page */
	prev = elan3mmu->elan3mmu_lXptbl;
	while ( prev->ptbl_parent != NULL ) 
	{
	    if ( prev->ptbl_parent == ptbl_ptr ) /* its the next one */
	    {
		/* remove element from chain */
	        prev->ptbl_parent =  ptbl_ptr->ptbl_parent;

		/* put ptbl back on free list */
		elan3mmu_free_lXptbl(dev, ptbl_ptr);

		/* unlock and return */
		spin_unlock (&elan3mmu->elan3mmu_lXptbl_lock);
		return ;
	    }		
	    prev = prev->ptbl_parent;
	}
	
       	panic ("elan3mmu_free_pte: failed to find ptbl in chain");
	/* NOTREACHED */
    }
    
    spin_unlock (&elan3mmu->elan3mmu_lXptbl_lock);
}

void
elan3mmu_free_lXptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl)
{
    ELAN3_PTBL_GR *ptg;

    HAT_PRINTF2 (2, "elan3mmu_free_lXptbl: dev %p ptbl %p\n", dev, ptbl);

    ASSERT (ptbl->ptbl_flags & PTBL_ALLOCED);
    ASSERT ((ptbl->ptbl_flags & PTBL_KEEP) == 0);
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == PTBL_LEVEL_X);
    ASSERT (ptbl->ptbl_valid == 0);
   
    ptbl->ptbl_flags = 0;

    ptg = PTBL_TO_GR(ptbl);

    if (ptg->pg_level == PTBL_LEVEL_3)
    {
	ELAN3MMU_STAT(lX_freed_l3);

	HAT_PRINTF1 (2, "elan3mmu_free_lXptbl: freeing stolen level 3 ptbl %p\n", ptbl);

	/* this was really a level 3 ptbl which we had to steal */
	spin_lock (&dev->Level[PTBL_LEVEL_3].PtblLock);
	ptbl->ptbl_next = dev->Level[PTBL_LEVEL_3].PtblFreeList;
	dev->Level[PTBL_LEVEL_3].PtblFreeList = ptbl;
	dev->Level[PTBL_LEVEL_3].PtblFreeCount++;
	spin_unlock (&dev->Level[PTBL_LEVEL_3].PtblLock);
    }
    else
    {
	spin_lock (&dev->Level[PTBL_LEVEL_X].PtblLock);
	ptbl->ptbl_next = dev->Level[PTBL_LEVEL_X].PtblFreeList;
	dev->Level[PTBL_LEVEL_X].PtblFreeList = ptbl;
	dev->Level[PTBL_LEVEL_X].PtblFreeCount++;
	spin_unlock (&dev->Level[PTBL_LEVEL_X].PtblLock);
    }
}

void
elan3mmu_free_l1ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags)
{
    HAT_PRINTF3 (2, "elan3mmu_free_l1ptbl: dev %p ptbl %p ptbl->ptbl_valid %x \n", dev, ptbl, ptbl->ptbl_valid);

    ASSERT (ptbl->ptbl_flags & PTBL_ALLOCED);
    ASSERT ((ptbl->ptbl_flags & PTBL_KEEP) == 0);
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == PTBL_LEVEL_1);
    ASSERT (ptbl->ptbl_valid == 0);
    
    HAT_PRINTF2 (2, "elan3mmu_free_l1ptbl: dev %p ptbl %p\n", dev, ptbl);

    ptbl->ptbl_flags = 0;
    spin_unlock (lock);
    
    spin_lock (&dev->Level[PTBL_LEVEL_1].PtblLock);
    ptbl->ptbl_next = dev->Level[PTBL_LEVEL_1].PtblFreeList;
    dev->Level[PTBL_LEVEL_1].PtblFreeList = ptbl;
    dev->Level[PTBL_LEVEL_1].PtblFreeCount++;
    spin_unlock (&dev->Level[PTBL_LEVEL_1].PtblLock);

    local_irq_restore (flags);
}

void
elan3mmu_free_l2ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags)
{
    ELAN3_PTBL_GR *ptg;

    HAT_PRINTF2 (2, "elan3mmu_free_l2ptbl: dev %p ptbl %p\n", dev, ptbl);

    ASSERT (PTBL_IS_LOCKED(ptbl->ptbl_flags));
    ASSERT (ptbl->ptbl_flags & PTBL_ALLOCED);
    ASSERT ((ptbl->ptbl_flags & PTBL_KEEP) == 0);
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == PTBL_LEVEL_2);
    ASSERT (ptbl->ptbl_valid == 0);
   
    ptbl->ptbl_flags = 0;
    spin_unlock (lock);

    ptg = PTBL_TO_GR(ptbl);

    if (ptg->pg_level == PTBL_LEVEL_3)
    {
	ELAN3MMU_STAT(l2_freed_l3);

	HAT_PRINTF1 (2, "elan3mmu_free_l2ptbl: freeing stolen level 3 ptbl %p\n", ptbl);

	/* this was really a level 3 ptbl which we had to steal */
	spin_lock (&dev->Level[PTBL_LEVEL_3].PtblLock);
	ptbl->ptbl_next = dev->Level[PTBL_LEVEL_3].PtblFreeList;
	dev->Level[PTBL_LEVEL_3].PtblFreeList = ptbl;
	dev->Level[PTBL_LEVEL_3].PtblFreeCount++;
	spin_unlock (&dev->Level[PTBL_LEVEL_3].PtblLock);
    }
    else
    {
	spin_lock (&dev->Level[PTBL_LEVEL_2].PtblLock);
	ptbl->ptbl_next = dev->Level[PTBL_LEVEL_2].PtblFreeList;
	dev->Level[PTBL_LEVEL_2].PtblFreeList = ptbl;
	dev->Level[PTBL_LEVEL_2].PtblFreeCount++;
	spin_unlock (&dev->Level[PTBL_LEVEL_2].PtblLock);
    }  
    local_irq_restore (flags);
}

void
elan3mmu_free_l3ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags)
{
    ASSERT (PTBL_IS_LOCKED(ptbl->ptbl_flags));
    ASSERT (ptbl->ptbl_flags & PTBL_ALLOCED);
    ASSERT ((ptbl->ptbl_flags & PTBL_KEEP) == 0);
    ASSERT (PTBL_LEVEL(ptbl->ptbl_flags) == PTBL_LEVEL_3);
    ASSERT (ptbl->ptbl_valid == 0);
    
    HAT_PRINTF2 (2, "elan3mmu_free_l3ptbl: dev %p ptbl %p\n", dev, ptbl);

    if (ptbl->ptbl_flags & PTBL_KERNEL)				/* if the ptbl has been used by the kernel */
    {								/* then zero all the pte's, since they will */
	elan3_sdram_zeroq_sdram (dev, PTBL_TO_PTADDR(ptbl), ELAN3_L3_ENTRIES*ELAN3_PTE_SIZE);
    }

    ptbl->ptbl_flags = 0;
    spin_unlock (lock);
    
    spin_lock (&dev->Level[PTBL_LEVEL_3].PtblLock);
    ptbl->ptbl_next = dev->Level[PTBL_LEVEL_3].PtblFreeList;
    dev->Level[PTBL_LEVEL_3].PtblFreeList = ptbl;
    dev->Level[PTBL_LEVEL_3].PtblFreeCount++;
    spin_unlock (&dev->Level[PTBL_LEVEL_3].PtblLock);

    local_irq_restore (flags);
}

void
elan3mmu_kernel_l3ptbl (ELAN3_PTBL *ptbl)
{
    ELAN3_DEV   *dev  = ptbl->ptbl_elan3mmu->elan3mmu_dev;
    sdramaddr_t pte  = PTBL_TO_PTADDR(ptbl);
    ELAN3_PTE    tpte = elan3mmu_kernel_invalid_pte(ptbl->ptbl_elan3mmu);
    int	        i;

    ptbl->ptbl_flags |= PTBL_KERNEL;
    for (i = 0; i < ELAN3_L3_ENTRIES; i++, pte += ELAN3_PTE_SIZE)
    {
	elan3_writepte (dev, pte, tpte);
    }
}
	
#define PTBL_CAN_STEAL(flag)	(((flag) & (PTBL_KERNEL|PTBL_KEEP)) == 0 && (((flag) & PTBL_ALLOCED) && PTBL_LEVEL(flag) == PTBL_LEVEL_3))
#define PTBL_MAY_STEAL(flag)	(((flag) & (PTBL_KERNEL|PTBL_KEEP|PTBL_LOCKED)) == 0 && (((flag) & PTBL_ALLOCED) && PTBL_LEVEL(flag) == PTBL_LEVEL_3))

static int
elan3mmu_steal_this_ptbl (ELAN3_DEV *dev, ELAN3_PTBL *l3ptbl)
{
    ELAN3_PTBL  *l2ptbl     = l3ptbl->ptbl_parent;
    E3_Addr     l2addr     = BASE2VA(l2ptbl);
    E3_Addr     l3addr     = BASE2VA(l3ptbl);
    ELAN3_PTP    invalidptp = ELAN3_INVALID_PTP;
    sdramaddr_t l2ptp;
    spinlock_t *l2lock;
    unsigned long l2flags;

    HAT_PRINTF5 (1, "elan3mmu_steal_this_ptbl: l3ptbl %p (%x) l2ptbl %p (%x) l2addr %x\n",
		 l3ptbl, l3ptbl->ptbl_flags, l2ptbl, l2ptbl->ptbl_flags, l2addr);

    if (PTBL_CAN_STEAL (l3ptbl->ptbl_flags) &&
	elan3mmu_lock_ptbl (l2ptbl, LK_PTBL_NOWAIT, l3ptbl->ptbl_elan3mmu, l2addr, PTBL_LEVEL_2, &l2lock, &l2flags) == LK_PTBL_OK)
    {
	ELAN3MMU_STAT(stolen_ptbls);

	/* Locked both L3 and L2 page tables. */
	l2ptp = PTBL_TO_PTADDR (l2ptbl) + ELAN3_L2_INDEX(l3addr)*ELAN3_PTP_SIZE;
	
	/* detach the level 3 page table */
	elan3_writeptp (dev, l2ptp, invalidptp);
	ElanFlushTlb (dev);

	l2ptbl->ptbl_valid--;

	HAT_PRINTF3 (2, "elan3mmu_steal_this_ptbl: dec valid for level %d ptbl %p to %d\n", PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid); 

	elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);

	elan3mmu_unload_loop (l3ptbl->ptbl_elan3mmu, l3ptbl, 0, ELAN3_L3_ENTRIES, PTE_UNLOAD_NOFLUSH);

	ASSERT (l3ptbl->ptbl_valid == 0);

	l3ptbl->ptbl_flags = 0;
	return (1);
    }
    return (0);
}

static ELAN3_PTBL *
elan3mmu_steal_l3ptbl (ELAN3_DEV *dev, int attr)
{
    ELAN3_PTBL_GR	*ptg;
    ELAN3_PTBL		*ptbl;
    spinlock_t		*lock;
    unsigned long        group_flags;
    unsigned long        ptbl_flags;
    register int 	 i;

    HAT_PRINTF1 (2, "elan3mmu_steal_l3ptbl: attr %x\n", attr);

    spin_lock_irqsave (&dev->PtblGroupLock, group_flags);

    ptg = dev->Level3PtblGroupHand;

    if (ptg == NULL)
	ptg = dev->Level[PTBL_LEVEL_3].PtblGroupList;
    
    for (;;)
    {
	while (ptg)
	{
	    for (i = 0, ptbl = ptg->pg_ptbls; i < PTBLS_PER_GROUP_MAX; i++, ptbl++)
	    {
		if (PTBL_MAY_STEAL (ptbl->ptbl_flags) &&
		    elan3mmu_lock_this_ptbl (ptbl, LK_PTBL_NOWAIT, &lock, &ptbl_flags) == LK_PTBL_OK)
		{
		    if (elan3mmu_steal_this_ptbl (dev, ptbl ))
		    {
			HAT_PRINTF1 (2, "elan3mmu_steal_l3ptbl: stolen ptbl %p\n", ptbl);

			elan3mmu_unlock_ptbl (ptbl, lock,ptbl_flags);

			dev->Level3PtblGroupHand = ptg->pg_next;

			spin_unlock_irqrestore (&dev->PtblGroupLock, group_flags);

			return (ptbl);
		    }
		    elan3mmu_unlock_ptbl (ptbl, lock, ptbl_flags);
		}
	    }
	    ptg = ptg->pg_next;
	}
	
	if (dev->Level[PTBL_LEVEL_3].PtblFreeList)
	{
	    spin_lock (&dev->Level[PTBL_LEVEL_3].PtblLock);
	    ptbl = dev->Level[PTBL_LEVEL_3].PtblFreeList;
	    if (ptbl != NULL)
	    {
		dev->Level[PTBL_LEVEL_3].PtblFreeList = ptbl->ptbl_next;
		dev->Level[PTBL_LEVEL_3].PtblFreeCount--;
	    }
	    spin_unlock (&dev->Level[PTBL_LEVEL_3].PtblLock);

	    if (ptbl != NULL)
	    {
		HAT_PRINTF1 (2, "elan3mmu_steal_l3ptbl: found ptbl %p on free list\n", ptbl);
		break;
	    }
	}

	ptbl = elan3mmu_create_ptbls (dev, PTBL_LEVEL_3, attr, 1);

	if (ptbl != NULL)
	{
	    HAT_PRINTF1 (2, "elan3mmu_steal_l3ptbl: created new ptbl %p\n", ptbl);
	    break;
	}
	
	HAT_PRINTF0 (1, "elan3mmu_steal_l3ptbl: cannot find a ptbl, retrying\n");
	ptg = dev->Level[PTBL_LEVEL_3].PtblGroupList;
    }

    spin_unlock (&dev->PtblGroupLock);
    return (ptbl);
}

sdramaddr_t
elan3mmu_ptefind (ELAN3MMU *elan3mmu, E3_Addr addr, int *level, 
		  ELAN3_PTBL **pptbl, spinlock_t **plock, unsigned long *flags)
{
    ELAN3_DEV   *dev = elan3mmu->elan3mmu_dev;
    ELAN3_PTBL  *l1ptbl;
    sdramaddr_t l1ptp;
    ELAN3_PTP    tl1ptp;
    E3_Addr     l1base;
    ELAN3_PTBL  *l2ptbl;
    sdramaddr_t l2ptp;
    ELAN3_PTP    tl2ptp;
    E3_Addr     l2base;
    ELAN3_PTBL  *l3ptbl;
    sdramaddr_t l3pte;
    spinlock_t *l1lock;
    spinlock_t *l2lock;
    spinlock_t *l3lock;
    unsigned long l1flags;
    unsigned long l2flags;
    unsigned long l3flags;

    HAT_PRINTF2 (2, "elan3mmu_ptefind: elan3mmu %p addr %x\n", elan3mmu, addr);

    l1ptbl = elan3mmu->elan3mmu_l1ptbl;
    *level = 0;

    if (l1ptbl == NULL)
	return ((sdramaddr_t) NULL);

    l1ptp  = PTBL_TO_PTADDR(l1ptbl) + ELAN3_L1_INDEX(addr)*ELAN3_PTP_SIZE;
    l1base = ELAN3_L1_BASE(addr);
    
retryl1:
    tl1ptp = elan3_readptp (dev, l1ptp);
    
    HAT_PRINTF4 (2, "elan3mmu_ptefind: l1ptbl %p l1ptp %lx l1base %x : tl1ptp %x\n", l1ptbl, l1ptp, l1base, tl1ptp);

    switch (ELAN3_PTP_TYPE(tl1ptp))
    {
    case ELAN3_ET_PTE:
	elan3mmu_lock_ptbl (l1ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &l1flags);

	tl1ptp = elan3_readptp (dev, l1ptp);
	if (ELAN3_PTP_TYPE(tl1ptp) != ELAN3_ET_PTE)
	{
	    elan3mmu_unlock_ptbl (l1ptbl, l1lock, l1flags);
	    goto retryl1;
	}
	
	*level = 1;
	*pptbl = l1ptbl;
	*plock = l1lock;
	*flags = l1flags;
	
	/* return with l1lock */
	return (l1ptp);  

    case ELAN3_ET_INVALID:
	return ((sdramaddr_t) 0);
	
    case ELAN3_ET_PTP:
	break;

    default:
	panic ("elan3mmu_ptefind: found bad entry in level 1 page table");
	/* NOTREACHED */
    }
    
    HAT_PRINTF1 (2, "elan3mmu_ptefind: chain to level 2 ptbl from ptp %x\n", tl1ptp);

    l2ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
    l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr)*ELAN3_PTP_SIZE;
    l2base = ELAN3_L2_BASE(addr);
    
    tl2ptp = elan3_readptp (dev, l2ptp);
    
    HAT_PRINTF4 (2, "elan3mmu_ptefind: l2ptbl %p l2ptp %lx l2base %x : tl2ptp %x\n", l2ptbl, l2ptp, l2base, tl2ptp);

    switch (ELAN3_PTP_TYPE(tl2ptp))
    {
    case ELAN3_ET_PTE:
	switch (elan3mmu_lock_ptbl (l2ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags))
	{
	case LK_PTBL_OK:
	    tl2ptp = elan3_readptp (dev, l2ptp);
	    if (ELAN3_PTP_TYPE(tl2ptp) != ELAN3_ET_PTE)
	    {
		elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);
		goto retryl1;
	    }
	    
	    *level = 2;
	    *pptbl = l2ptbl;
	    *plock = l2lock;
	    *flags = l2flags;
	    
	    /* return with l2lock */
	    return (l2ptp); 
	    
	case LK_PTBL_MISMATCH:
	    HAT_PRINTF6 (2, "elan3mmu_ptefind: PTBL_MISMATCH : ptbl %p flags %x elan3mmu %p base %x (%p %x)\n",
			 l2ptbl, l2ptbl->ptbl_flags, l2ptbl->ptbl_elan3mmu, l2ptbl->ptbl_base, elan3mmu, addr);
	    
	    /*
	     * We've trogged down to this ptbl,  but someone has just
	     * stolen it,  so try all over again.
	     */
	    goto retryl1;
	    
	default:
	    panic ("elan3mmu_ptefind: elan3mmu_lock_ptbl returned bad value");
	    /* NOTREACHED */
	}
    case ELAN3_ET_INVALID:
	return ((sdramaddr_t) 0);
	
    case ELAN3_ET_PTP:
	break;
    default:
	panic ("elan3mmu_ptefind: found bad entry in level 2 page table");
	/* NOTREACHED */
    }
    
    HAT_PRINTF1 (2, "elan3mmu_ptefind: chain to level 3 page table from ptp %x\n", tl2ptp);

    l3ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
    l3pte  = PTBL_TO_PTADDR(l3ptbl) + ELAN3_L3_INDEX(addr)*ELAN3_PTE_SIZE;
    
    HAT_PRINTF2 (2, "elan3mmu_ptefind: l3ptbl %p l3pte %lx\n", l3ptbl, l3pte);
		 
    switch (elan3mmu_lock_ptbl (l3ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &l3flags))
    {
    case LK_PTBL_OK:
	*level = 3;
	*plock = l3lock;
	*pptbl = l3ptbl;
	*flags = l3flags;

	return (l3pte);
	
    case LK_PTBL_FAILED:
	panic ("elan3mmu_ptefind: l3 lock failed");
	/* NOTREACHED */

    case LK_PTBL_MISMATCH:
	HAT_PRINTF6 (2, "elan3mmu_ptefind: PTBL_MISMATCH : ptbl %p flags %x elan3mmu %p base %x (%p %x)\n",
		     l3ptbl, l3ptbl->ptbl_flags, l3ptbl->ptbl_elan3mmu, l3ptbl->ptbl_base, elan3mmu, addr);
		     
	/*
	 * We've trogged down to this ptbl,  but someone has just
	 * stolen it,  so try all over again.
	 */
	goto retryl1;
	
    default:
	panic ("elan3mmu_ptefind: elan3mmu_lock_ptbl returned bad value");
	/* NOTREACHED */
    }
    /* NOTREACHED */
    return ((sdramaddr_t) 0);
}

sdramaddr_t 
elan3mmu_ptp2pte (ELAN3MMU *elan3mmu, sdramaddr_t ptp, int level)
{
    ELAN3_PTP tptp = elan3_readptp (elan3mmu->elan3mmu_dev, ptp);

    ASSERT (level != 3 && ELAN3_PTP_TYPE(tptp) == ELAN3_ET_PTE);

    return PTP_TO_PT_PADDR(tptp);
}

sdramaddr_t
elan3mmu_ptealloc (ELAN3MMU *elan3mmu, E3_Addr addr, int level, 
		   ELAN3_PTBL **pptbl, spinlock_t **plock, int attr, unsigned long *flags)
{
    ELAN3_DEV   *dev     = elan3mmu->elan3mmu_dev;
    ELAN3_PTBL  *l1ptbl;
    ELAN3_PTBL  *lXptbl;
    int         idx;
    sdramaddr_t l1ptp;
    ELAN3_PTP    tl1ptp;
    E3_Addr     l1base;
    spinlock_t *l1lock;
    ELAN3_PTBL  *l2ptbl;
    sdramaddr_t l2ptp;
    ELAN3_PTP    tl2ptp;
    E3_Addr     l2base;
    spinlock_t *l2lock;
    ELAN3_PTBL  *l3ptbl;
    sdramaddr_t l3pte;
    E3_Addr     l3base;
    spinlock_t *l3lock;

    unsigned long l1flags;
    unsigned long l2flags;
    unsigned long l3flags;

    HAT_PRINTF2 (2, "elan3mmu_ptealloc: elan3mmu %p addr %x\n", elan3mmu, addr);

    l1ptbl = elan3mmu->elan3mmu_l1ptbl;
    if (l1ptbl == NULL)
	return ((sdramaddr_t) 0);

    l1ptp  = PTBL_TO_PTADDR(l1ptbl) + ELAN3_L1_INDEX(addr)*ELAN3_PTP_SIZE;
    l1base = ELAN3_L1_BASE(addr);
		
retryl1:
    tl1ptp = elan3_readptp (dev, l1ptp);

    HAT_PRINTF5 (2, "elan3mmu_ptealloc: l1ptbl %p 1ptp %lx l1base %x (%x) : tl1ptp %x\n",
		 l1ptbl, l1ptp, l1base, l1ptbl->ptbl_base, tl1ptp);

    switch (ELAN3_PTP_TYPE(tl1ptp))
    {
    case ELAN3_ET_PTE:
	if (level == PTBL_LEVEL_1)
	{
	    elan3mmu_lock_ptbl (l1ptbl, 0, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &l1flags);

	    tl1ptp = elan3_readptp (dev, l1ptp);
	    if (ELAN3_PTP_TYPE(tl1ptp) != ELAN3_ET_PTE)
	    {
		elan3mmu_unlock_ptbl (l1ptbl, l1lock, l1flags);
		goto retryl1;
	    }
	    
	    *pptbl = l1ptbl;
	    *plock = l1lock;
	    *flags = l1flags;

	    /* return holding l1lock */
	    return (l1ptp);
	}
	panic ("elan3mmu_ptealloc: found pte in level 1 page table");
	/* NOTREACHED */

    case ELAN3_ET_PTP:
	if (level == PTBL_LEVEL_1)
	    panic ("elan3mmu_ptealloc: found PTP when loading a level 1 PTE\n");
	break;

    case ELAN3_ET_INVALID:
	if (level == PTBL_LEVEL_1)
	{
	    if ((lXptbl = elan3mmu_alloc_pte (dev, elan3mmu,  &idx)) == NULL)
		return ((sdramaddr_t) 0);

	    elan3mmu_lock_ptbl (l1ptbl, 0, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &l1flags);

	    tl1ptp = elan3_readptp (dev, l1ptp);
	    if (ELAN3_PTP_TYPE(tl1ptp) != ELAN3_ET_INVALID)
	    {
		/* raced with someone else, whose got there first */
		elan3mmu_free_pte (dev, elan3mmu, lXptbl, idx);

		/* drop the l1lock and retry */
		elan3mmu_unlock_ptbl (l1ptbl, l1lock, l1flags);
		goto retryl1;
	    }
	    
	    tl1ptp = PTBL_TO_PTADDR(lXptbl) | (idx * ELAN3_PTE_SIZE) | ELAN3_ET_PTE;
	    
	    elan3_writeptp (dev, l1ptp, tl1ptp);

	    *pptbl = l1ptbl;
	    *plock = l1lock;
	    *flags = l1flags;

	    /* return holding l1lock */
	    return (l1ptp);
	}

	if (level == PTBL_LEVEL_2)
	{
	    if ((lXptbl = elan3mmu_alloc_pte (dev, elan3mmu, &idx)) == NULL)
		return ((sdramaddr_t) 0);

	    if ((l2ptbl = elan3mmu_alloc_l2ptbl (dev, attr, l1ptbl, elan3mmu, ELAN3_L2_BASE(addr), &l2lock, &l2flags)) == NULL)
	    {
		elan3mmu_free_pte (dev, elan3mmu, lXptbl, idx); 
		return ((sdramaddr_t) 0);
	    }

	    /* Connect l2ptbl to the new LX pte */
	    l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr) * ELAN3_PTP_SIZE;
	    tl2ptp = PTBL_TO_PTADDR(lXptbl) | (idx * ELAN3_PTE_SIZE) | ELAN3_ET_PTE;

	    elan3_writeptp (dev, l2ptp, tl2ptp);

	    /* Now need to lock the l1 ptbl */
	    elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);

	    elan3mmu_lock_ptbl (l1ptbl, 0, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &l1flags);
	    elan3mmu_lock_ptbl (l2ptbl, 0, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags);

	    tl1ptp = elan3_readptp (dev, l1ptp);
	    if (ELAN3_PTP_TYPE(tl1ptp) != ELAN3_ET_INVALID)
	    {
		HAT_PRINTF0 (2, "elan3mmu_ptealloc: beaten to it,  free l2 ptbl/lx pte\n");
		
		tl2ptp = ELAN3_INVALID_PTP;
		elan3_writeptp (dev, l2ptp, tl2ptp);
		
		HAT_PRINTF2 (2, "elan3mmu_ptealloc: write level 2 ptp %lx to %x\n", l2ptp, tl2ptp);
		HAT_PRINTF2 (2, "elan3mmu_ptealloc: freeing l2 ptbl %p (%x)\n", l2ptbl, l2ptbl->ptbl_flags);
		
		elan3mmu_free_l2ptbl (dev, l2ptbl, l2lock, l2flags);
		elan3mmu_free_pte (dev, elan3mmu, lXptbl, idx);

		elan3mmu_unlock_ptbl (l1ptbl, l1lock, l1flags);

		goto retryl1;
	    }
	    
	    /* Now have L1 locked,  so install the L2 ptbl */
	    l1ptp  = PTBL_TO_PTADDR(l1ptbl) + ELAN3_L1_INDEX(addr)*ELAN3_PTP_SIZE;
	    tl1ptp = PTBL_TO_PTADDR(l2ptbl) | ELAN3_ET_PTP;
	    l1ptbl->ptbl_valid++;

	    HAT_PRINTF3 (2, "elan3mmu_ptealloc: inc valid for level %d ptbl %p to %d\n", 
			 PTBL_LEVEL(l1ptbl->ptbl_flags), l1ptbl, l1ptbl->ptbl_valid);
	    
	    elan3_writeptp (dev, l1ptp, tl1ptp);
	    
	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: write l1ptp %lx to %x\n", l1ptp, tl1ptp);

	    /* unordered unlock - lock l1ptbl, lock l2ptbl, unlock l1ptbl */
	    elan3mmu_unlock_ptbl (l1ptbl, l1lock, l2flags); /* need to unlock with the l2flags to keep irq order correct */

	    *pptbl = l2ptbl;
	    *plock = l2lock;
	    *flags = l1flags; /* return the l1flags here as we have released the l2flags already to keep order */

	    /* return holding l2lock */
	    return (l2ptp);
	}

	HAT_PRINTF0 (2, "elan3mmu_ptealloc: allocating level 2 and level 3 page tables\n");

	/* Allocate a level 2 and level 3 page table and link them together */
	if ((l2ptbl = elan3mmu_alloc_l2ptbl (dev, attr, l1ptbl, elan3mmu, ELAN3_L2_BASE(addr), &l2lock, &l2flags)) == NULL)
	    return ((sdramaddr_t) 0);

	if ((l3ptbl = elan3mmu_alloc_l3ptbl (dev, attr | PTE_NO_SLEEP, l2ptbl, elan3mmu, ELAN3_L3_BASE(addr), &l3lock, &l3flags)) == NULL)
	{
	    elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);
	    return ((sdramaddr_t) 0);
	}

	ASSERT (PTBL_IS_LOCKED (l2ptbl->ptbl_flags));
	ASSERT (PTBL_LEVEL (l2ptbl->ptbl_flags) == PTBL_LEVEL_2);
	ASSERT (PTBL_IS_LOCKED (l3ptbl->ptbl_flags));
	ASSERT (PTBL_LEVEL (l3ptbl->ptbl_flags) == PTBL_LEVEL_3);

	HAT_PRINTF6 (2, "elan3mmu_ptealloc: l2ptbl %p (%x,%x) l3ptbl %p (%x,%x)\n",
		     l2ptbl, l2ptbl->ptbl_flags, l2ptbl->ptbl_base,
		     l3ptbl, l3ptbl->ptbl_flags, l3ptbl->ptbl_base);

	if (CTXT_IS_KERNEL (elan3mmu->elan3mmu_ctxt))
	{
	    l2ptbl->ptbl_flags |= PTBL_KERNEL;
	    elan3mmu_kernel_l3ptbl (l3ptbl);
	}
	
	/*
	 * Connect L3 ptbl to the new L2 ptbl.
	 */
	l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr) * ELAN3_PTP_SIZE;
	tl2ptp = PTBL_TO_PTADDR(l3ptbl) | ELAN3_ET_PTP;

	l2ptbl->ptbl_valid = 1;

	HAT_PRINTF3 (2, "elan3mmu_ptealloc: set valid for level %d ptbl %p to %d\n", 
		     PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid);

	HAT_PRINTF2 (2, "elan3mmu_ptealloc: write level 2 ptp %lx to %x\n", l2ptp, tl2ptp);

	elan3_writeptp (dev, l2ptp, tl2ptp);

	/* 
	 * Now need to lock the l1 ptbl - to maintain lock ordering
	 * we set the PTBL_KEEP bit to stop the l3 ptbl from being 
	 * stolen and drop the locks in the order we aquired them
	 */
	l3ptbl->ptbl_flags |= PTBL_KEEP;

	elan3mmu_unlock_ptbl (l3ptbl, l3lock, l3flags);
	elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);

	elan3mmu_lock_ptbl (l1ptbl, 0, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &l1flags);
	elan3mmu_lock_ptbl (l3ptbl, 0, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &l3flags);
	    
	l3ptbl->ptbl_flags &= ~PTBL_KEEP;
	   
	/* Now have l1 and l3 ptbls locked,  so install the new l2 ptbl into the l1. */
	tl1ptp = elan3_readptp (dev, l1ptp);

	HAT_PRINTF2 (2, "elan3mmu_ptealloc: l1ptp %lx is %x\n", l1ptp, tl1ptp);

	if (ELAN3_PTP_TYPE(tl1ptp) != ELAN3_ET_INVALID)
	{
	    HAT_PRINTF0 (2, "elan3mmu_ptealloc: beaten to it,  free l2/l3 ptbls\n");

	    /* free off the level 3 page table */
	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: freeing l3 ptbl %p (%x)\n", l3ptbl, l3ptbl->ptbl_flags);

	    l3ptbl->ptbl_flags &= ~PTBL_KEEP;
	    elan3mmu_free_l3ptbl (dev, l3ptbl, l3lock, l3flags);

	    /* and unlock the level 1 ptbl */
	    elan3mmu_unlock_ptbl (l1ptbl, l1lock, l1flags);
	    
	    /* lock the level 2 page table, and clear out the PTP, then free it */
	    (void) elan3mmu_lock_ptbl (l2ptbl, 0, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags);

	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: locked l2 ptbl %p (%x)\n", l2ptbl, l2ptbl->ptbl_flags);
	    
	    tl2ptp = ELAN3_INVALID_PTP;
	    elan3_writeptp (dev, l2ptp, tl2ptp);
	    l2ptbl->ptbl_valid = 0;

	    HAT_PRINTF3 (2, "elan3mmu_ptealloc: set to 0 valid for level %d ptbl %p to %d\n", PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid); 

	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: write level 2 ptp %lx to %x\n", l2ptp, tl2ptp);
	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: freeing l2 ptbl %p (%x)\n", l2ptbl, l2ptbl->ptbl_flags);

	    elan3mmu_free_l2ptbl (dev, l2ptbl, l2lock, l2flags);

	    goto retryl1;
	}
	
	HAT_PRINTF4 (2, "elan3mmu_ptealloc: l1ptbl is %p (%x), l3ptbl is %p (%x)\n", 
		     l1ptbl, l1ptbl->ptbl_flags, l3ptbl, l3ptbl->ptbl_flags);

	/* Now have L1 and L3 locked,  so install the L2 ptbl */
	l1ptp  = PTBL_TO_PTADDR(l1ptbl) + ELAN3_L1_INDEX(addr)*ELAN3_PTP_SIZE;
	tl1ptp = PTBL_TO_PTADDR(l2ptbl) | ELAN3_ET_PTP;
	l1ptbl->ptbl_valid++;

	HAT_PRINTF3 (2, "elan3mmu_ptealloc: inc valid for level %d ptbl %p to %d\n", 
		     PTBL_LEVEL(l1ptbl->ptbl_flags), l1ptbl, l1ptbl->ptbl_valid);

	elan3_writeptp (dev, l1ptp, tl1ptp);

	HAT_PRINTF2 (2, "elan3mmu_ptealloc: write l1ptp %lx to %x\n", l1ptp, tl1ptp);

	/* unordered unlock - lock l1ptbl, lock l3ptbl, unlock l1ptbl */
	elan3mmu_unlock_ptbl (l1ptbl, l1lock, l3flags); /* free using l3flags to keep irq ordering */

	l3pte = PTBL_TO_PTADDR (l3ptbl) + ELAN3_L3_INDEX(addr)*ELAN3_PTE_SIZE;

	/* Level 3 ptbl is already locked,  so just return the pte */
	*pptbl = l3ptbl;
	*plock = l3lock;
	*flags = l1flags; /* return l1flags to keep irq ordering */

	return (l3pte);

    default:
	panic ("elan3mmu_ptealloc: found bad entry in level 1 page table");
	/* NOTREACHED */
    }

    HAT_PRINTF1 (2, "elan3mmu_ptealloc: chain to level 2 ptbl from ptp %x\n", tl1ptp);

    l2ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
    l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr)*ELAN3_PTP_SIZE;
    l2base = ELAN3_L2_BASE(addr);

    tl2ptp = elan3_readptp (dev, l2ptp);

    HAT_PRINTF5 (2, "elan3mmu_ptealloc: l2ptbl %p l2ptp %lx l2base %x (%x) : tl2ptp %x\n",
		 l2ptbl, l2ptp, l2base, l2ptbl->ptbl_base, tl2ptp);

    switch (ELAN3_PTP_TYPE(tl2ptp))
    {
    case ELAN3_ET_PTE:
	if (level == PTBL_LEVEL_2) {
	    /* this is a pointer to a pte,  we should just return it */

	    switch (elan3mmu_lock_ptbl (l2ptbl, 0, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags))
	    {
	    case LK_PTBL_OK:
		break;
	
	    case LK_PTBL_FAILED:
		panic ("elan3mmu_ptealloc: l2 lock failed");
		/* NOTREACHED */
		
	    case LK_PTBL_MISMATCH:
		HAT_PRINTF6 (2, "elan3mmu_ptealloc: PTBL_MISMATCH : ptbl %p flags %x elan3mmu %p base %x (%p %x)\n",
			     l2ptbl, l2ptbl->ptbl_flags, l2ptbl->ptbl_elan3mmu, l2ptbl->ptbl_base, elan3mmu, addr);
		
		/*
		 * We've trogged down to this ptbl,  but someone has just
		 * stolen it,  so try all over again.
		 */
		goto retryl1;
		
	    default:
		panic ("elan3mmu_ptealloc: elan3mmu_lock_ptbl returned bad value");
		/* NOTREACHED */
	    }


	    tl2ptp = elan3_readptp (dev, l2ptp);
	    if (ELAN3_PTP_TYPE(tl2ptp) != ELAN3_ET_PTE)
	    {
		elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);
		goto retryl1;
	    }

	    *pptbl = l2ptbl;
	    *plock = l2lock;
	    *flags = l2flags;

	    /* return holdind l2lock */
	    return (l2ptp);
	}
	panic ("elan3mmu: found pte in level 2 page table");
	/* NOTREACHED */

    case ELAN3_ET_PTP:
	break;

    case ELAN3_ET_INVALID:
	if (level == PTBL_LEVEL_2) 
	{
	    if ((lXptbl = elan3mmu_alloc_pte (dev, elan3mmu, &idx)) == NULL)
		return ((sdramaddr_t) 0);

	    switch (elan3mmu_lock_ptbl (l2ptbl, 0, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags))
	    {
	    case LK_PTBL_OK:
		break;
	
	    case LK_PTBL_FAILED:
		panic ("elan3mmu_ptealloc: l2 lock failed");
		/* NOTREACHED */
		
	    case LK_PTBL_MISMATCH:
		HAT_PRINTF6 (2, "elan3mmu_ptealloc: PTBL_MISMATCH : ptbl %p flags %x elan3mmu %p base %x (%p %x)\n",
			     l2ptbl, l2ptbl->ptbl_flags, l2ptbl->ptbl_elan3mmu, l2ptbl->ptbl_base, elan3mmu, addr);
		
		/*
		 * We've trogged down to this ptbl,  but someone has just
		 * stolen it,  so try all over again.
		 */
		goto retryl1;
		
	    default:
		panic ("elan3mmu_ptealloc: elan3mmu_lock_ptbl returned bad value");
		/* NOTREACHED */
	    }

	    tl2ptp = elan3_readptp (dev, l2ptp);
	    if (ELAN3_PTP_TYPE(tl2ptp) != ELAN3_ET_INVALID)
	    {
		HAT_PRINTF0 (2, "elan3mmu_ptealloc: beaten to it,  free lx pte\n");

		elan3mmu_free_pte (dev, elan3mmu, lXptbl, idx);

		elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);
		goto retryl1;
	    }

	    /* Connect l2ptbl to the new LX pte */
	    tl2ptp = PTBL_TO_PTADDR(lXptbl) | (idx * ELAN3_PTE_SIZE) | ELAN3_ET_PTE;
	    	   
	    HAT_PRINTF3 (2, "elan3mmu_ptealloc: inc valid for level %d ptbl %p to %d\n", 
			 PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid);
	    
	    elan3_writeptp (dev, l2ptp, tl2ptp);
	    
	    HAT_PRINTF2 (2, "elan3mmu_ptealloc: write l2ptp %lx to %x\n", l2ptp, tl2ptp);

	    *pptbl = l2ptbl;
	    *plock = l2lock;
	    *flags = l2flags;

	    /* return holding l2lock */
	    return (l2ptp);
	}
	HAT_PRINTF0 (2, "elan3mmu_ptealloc: allocate level 3 page table\n");

	if ((l3ptbl = elan3mmu_alloc_l3ptbl (dev, attr, l2ptbl, elan3mmu, ELAN3_L3_BASE(addr), &l3lock, &l3flags)) == NULL)
	    return ((sdramaddr_t) 0);

	if (CTXT_IS_KERNEL (elan3mmu->elan3mmu_ctxt))
	    elan3mmu_kernel_l3ptbl (l3ptbl);

	/* 
	 * Now need to lock the l2 ptbl - to maintain lock ordering
	 * we set the PTBL_KEEP bit to stop the l3 ptbl from being 
	 * stolen and drop the locks in the order we aquired them
	 */
	l3ptbl->ptbl_flags |= PTBL_KEEP;

	elan3mmu_unlock_ptbl (l3ptbl, l3lock, l3flags);

	if (elan3mmu_lock_ptbl (l2ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &l2flags) == LK_PTBL_MISMATCH)
	{
	    HAT_PRINTF0 (2, "elan3mmu_ptealloc: l2ptbl freed, free l3 ptbl and try again\n");
	      
	    elan3mmu_lock_ptbl (l3ptbl, 0, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &l3flags);

	    /* free off the level 3 page table, and try again */
	    l3ptbl->ptbl_flags &= ~PTBL_KEEP;
	    elan3mmu_free_l3ptbl (dev, l3ptbl, l3lock, l3flags);
	    
	    goto retryl1;
	}

	elan3mmu_lock_ptbl (l3ptbl, 0, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &l3flags);

	l3ptbl->ptbl_flags &= ~PTBL_KEEP;

	/* Now have L2 and L3 ptbls locked, see if someone has beaten us to it. */
	tl2ptp = elan3_readptp (dev, l2ptp);

	HAT_PRINTF2 (2, "elan3mmu_ptealloc: l2ptp at %lx is %x\n", l2ptp, tl2ptp);

	if (ELAN3_PTP_TYPE(tl2ptp) != ELAN3_ET_INVALID)
	{
	    HAT_PRINTF0 (2, "elan3mmu_ptealloc: beaten to it, free l3 ptbl and try again\n");

	    /* free off the level 3 page table, and try again */
	    l3ptbl->ptbl_flags &= ~PTBL_KEEP;
	    elan3mmu_free_l3ptbl (dev, l3ptbl, l3lock, l3flags);
	    
	    /* Someone has allocated the ptbl before us */
	    elan3mmu_unlock_ptbl (l2ptbl, l2lock, l2flags);
	    
	    goto retryl1;
	}

	ASSERT (PTBL_IS_LOCKED (l2ptbl->ptbl_flags));

	/* Install the L3 ptbl into the L2 one */
	l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr)*ELAN3_PTP_SIZE;
	tl2ptp = PTBL_TO_PTADDR(l3ptbl) | ELAN3_ET_PTP;
	l2ptbl->ptbl_valid++;

	HAT_PRINTF3 (2, "elan3mmu_ptealloc: inc valid for level %d ptbl %p to %d\n",
		     PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid);

	elan3_writeptp (dev, l2ptp, tl2ptp);

	HAT_PRINTF2 (2, "elan3mmu_ptealloc: write level 2 ptp %lx to %x\n", l2ptp, tl2ptp);

	/* unordered unlock - lock l2ptbl, lock l3ptbl, unlock l2ptbl */
	elan3mmu_unlock_ptbl (l2ptbl, l2lock, l3flags); /* free with the l3flags to keep irq ordering */

	l3pte = PTBL_TO_PTADDR(l3ptbl) + ELAN3_L3_INDEX(addr)*ELAN3_PTE_SIZE;
	
	/* Level 3 ptbl is already locked, so just return the pte */
	*pptbl = l3ptbl;
	*plock = l3lock;
	*flags = l2flags; /* return l2flags to keep irq ordering */

	return (l3pte);

    default:
	panic ("elan3mmu_ptealloc: found bad entry in level 2 page table");
	/* NOTREACHED */
    }

    HAT_PRINTF1 (2, "elan3mmu_ptealloc: chain to level 3 page table from ptp %x\n", tl2ptp);

    l3ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
    l3pte  = PTBL_TO_PTADDR(l3ptbl) + ELAN3_L3_INDEX(addr)*ELAN3_PTE_SIZE;
    l3base = ELAN3_L3_BASE(addr);

    HAT_PRINTF4 (2, "elan3mmu_ptealloc: l3ptbl %p 3pte %lx l3base %x (%x)\n",
		 l3ptbl, l3pte, l3base, l3ptbl->ptbl_base);
		 
    if (elan3mmu_lock_ptbl (l3ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &l3flags) == LK_PTBL_OK)
    {
	*pptbl = l3ptbl;
	*plock = l3lock;
	*flags = l3flags;

	return (l3pte);
    }

    /* got all the way down here,  but its been nicked before we could lock it */
    /* so try all over again */
    goto retryl1;
}

void
elan3mmu_l1inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l1ptbl, int attr)
{
    ELAN3_DEV     *dev = elan3mmu->elan3mmu_dev;
    ELAN3_PTP      invalidptp = ELAN3_INVALID_PTP;
    ELAN3_PTP      tl1ptp;
    sdramaddr_t   l1ptp;
    E3_Addr       addr;
    spinlock_t   *l2lock;
    ELAN3_PTBL    *l2ptbl;
    ELAN3_PTBL    *lXptbl;
    int           idx;
    int	          i;
    int	          ret;
    unsigned long flags;

    l1ptp = PTBL_TO_PTADDR(l1ptbl);

    HAT_PRINTF2 (1, "elan3mmu_l1inval: l1ptbl %p l1ptp %lx\n", l1ptbl, l1ptp);

    for (i = 0, addr = 0; i < ELAN3_L1_ENTRIES; i++, l1ptp += ELAN3_PTP_SIZE)
    {
	tl1ptp = elan3_readptp (dev, l1ptp);
	switch (ELAN3_PTP_TYPE(tl1ptp))
	{
	case ELAN3_ET_PTE:
	    lXptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
	    idx    = (PTP_TO_PT_PADDR(tl1ptp) - PTBL_TO_PTADDR(lXptbl))/ELAN3_PTE_SIZE;  

	    HAT_PRINTF3 (2, "elan3mmu_l1inval: l1ptbl %p : lXptbl %p idx %d\n",
			 l1ptbl, lXptbl, idx);

	    /* invalidate the L1 pte. */
	    elan3_writeptp (dev, l1ptp, invalidptp);
	    if (! (attr & PTE_UNLOAD_NOFLUSH))
		ElanFlushTlb (dev);	    

	    l1ptbl->ptbl_valid--;
	    elan3mmu_free_pte ( dev, elan3mmu,  lXptbl, idx); 

	    HAT_PRINTF3 (2, "elan3mmu_l1inval: dec valid for level %d ptbl %p to %d\n",
		     PTBL_LEVEL(l1ptbl->ptbl_flags), l1ptbl, l1ptbl->ptbl_valid);
	    
	    break;

	case ELAN3_ET_PTP:
	    HAT_PRINTF5 (2, "elan3mmu_l1inval: l1ptbl %p : ptp %lx (%x) addr %x (%d)\n",
			 l1ptbl, l1ptp, tl1ptp, addr, i);

	    /* invalidate the L1 ptp. */
	    elan3_writeptp (dev, l1ptp, invalidptp);
	    if (! (attr & PTE_UNLOAD_NOFLUSH))
		ElanFlushTlb (dev);

	    /* invalidate the level 2 page table */
	    l2ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
	    ret    = elan3mmu_l2inval (elan3mmu, l2ptbl, attr | PTE_UNLOAD_NOFLUSH, addr, &l2lock, &flags);

	    ASSERT ((l2ptbl->ptbl_flags & PTBL_KEEP) == 0);
	    
	    if (ret == LK_PTBL_OK)
	    {
		if (((l2ptbl->ptbl_flags & PTBL_KEEP) == 0) && l2ptbl->ptbl_valid == 0)
		{
		    HAT_PRINTF1 (2, "elan3mmu_l1inval: free l2ptbl %p\n", l2ptbl);
		    
		    l1ptbl->ptbl_valid--;
		    elan3mmu_free_l2ptbl (elan3mmu->elan3mmu_dev, l2ptbl, l2lock, flags);

		    HAT_PRINTF3 (2, "elan3mmu_l1inval: dec valid for level %d ptbl %p to %d\n", 
				 PTBL_LEVEL(l1ptbl->ptbl_flags), l1ptbl, l1ptbl->ptbl_valid);
		}
		else
		{
		    /* need to keep this page table,  so even though its now empty, */
		    /* chain it back in */
		    HAT_PRINTF1 (2, "elan3mmu_l1inval: keep l2ptbl %p\n", l2ptbl);

		    elan3_writeptp (dev, l1ptp, tl1ptp);
		    elan3mmu_unlock_ptbl (l2ptbl, l2lock, flags);
		}    
	    }
	    else
	    {
		l1ptbl->ptbl_valid--;

		HAT_PRINTF3 (2, "elan3mmu_l1inval: dec valid for level %d ptbl %p to %d\n", 
			     PTBL_LEVEL(l1ptbl->ptbl_flags), l1ptbl, l1ptbl->ptbl_valid);
	    }
	    break;
	    
	case ELAN3_ET_INVALID:
	    break;

	default:
	    panic ("elan3mmu_l1inval: found invalid entry in level 1 page table");
	    /* NOTREACHED */
	}

	if (l1ptbl->ptbl_valid == 0)
	    break;

	addr += ELAN3_L1_SIZE;
    }
}

int
elan3mmu_l2inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l2ptbl, int attr, E3_Addr addr, spinlock_t **pl2lock, unsigned long *flags)
{
    ELAN3_DEV   *dev = elan3mmu->elan3mmu_dev;
    ELAN3_PTP    invalidptp = ELAN3_INVALID_PTP;
    ELAN3_PTP    tl2ptp;
    sdramaddr_t l2ptp;
    spinlock_t *l3lock;
    unsigned long l3flags;
    ELAN3_PTBL  *l3ptbl;
    ELAN3_PTBL  *lXptbl;
    int         idx;
    int	        i;
    int	        ret;

    HAT_PRINTF2 (1, "elan3mmu_l2inval: l2ptbl %p addr %x\n", l2ptbl, addr);

    ASSERT (PTBL_LEVEL (l2ptbl->ptbl_flags) == PTBL_LEVEL_2);
    ASSERT (PTBL_LEVEL (l2ptbl->ptbl_parent->ptbl_flags) == PTBL_LEVEL_1);

    ret = elan3mmu_lock_ptbl (l2ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_2, pl2lock, flags);

    ASSERT (ret == LK_PTBL_OK);
    ASSERT (l2ptbl->ptbl_elan3mmu == elan3mmu);
    ASSERT (l2ptbl->ptbl_parent->ptbl_elan3mmu == elan3mmu);

    l2ptp = PTBL_TO_PTADDR(l2ptbl);

    for (i = 0; i < ELAN3_L2_ENTRIES; i++, l2ptp += ELAN3_PTP_SIZE)
    {
	tl2ptp = elan3_readptp (dev, l2ptp);
	switch (ELAN3_PTP_TYPE(tl2ptp))
	{
	case ELAN3_ET_PTE:
	    lXptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
	    idx    = (PTP_TO_PT_PADDR(tl2ptp) - PTBL_TO_PTADDR(lXptbl))/ELAN3_PTE_SIZE;  

	    HAT_PRINTF3 (2, "elan3mmu_l2inval: l2ptbl %p : lXptbl %p idx %d\n",
			 l2ptbl, lXptbl, idx);

	    /* invalidate the L2 pte. */
	    elan3_writeptp (dev, l2ptp, invalidptp);
	    if (! (attr & PTE_UNLOAD_NOFLUSH))
		ElanFlushTlb (dev);

	    l2ptbl->ptbl_valid--;
	    elan3mmu_free_pte ( dev, elan3mmu, lXptbl, idx); 

	    HAT_PRINTF3 (2, "elan3mmu_l2inval: dec valid for level %d ptbl %p to %d\n", PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid); 

	    break;
	    
	case ELAN3_ET_PTP:
	    HAT_PRINTF5 (2, "elan3mmu_l2inval: l2ptbl %p : ptp %lx (%x) addr %x (%d)\n",
			 l2ptbl, l2ptp, tl2ptp, addr, i);

	    /* invalidate the L2 ptp. */
	    elan3_writeptp (dev, l2ptp, invalidptp);
	    if (! (attr & PTE_UNLOAD_NOFLUSH))
		ElanFlushTlb (dev);
	    
	    /* unload the level 3 page table */
	    l3ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
	    ret = elan3mmu_l3inval (elan3mmu, l3ptbl, attr | PTE_UNLOAD_NOFLUSH, addr, &l3lock, &l3flags);
	    
	    if (ret == LK_PTBL_OK)
	    {
		if ((l3ptbl->ptbl_flags & PTBL_KEEP) == 0 && l3ptbl->ptbl_valid == 0)
		{
		    /* decrement the valid count of the level 2 page table, and */
		    /* free off the level 3 page table */
		    HAT_PRINTF1 (2, "elan3mmu_l2inval: free l3ptbl %p\n", l3ptbl);

		    l2ptbl->ptbl_valid--;
		    elan3mmu_free_l3ptbl (elan3mmu->elan3mmu_dev, l3ptbl, l3lock, l3flags);

		    HAT_PRINTF3 (2, "elan3mmu_l2inval: dec valid for level %d ptbl %p to %d\n", 
				 PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid);
		}
		else
		{
		    /* need to keep this page table,  so even though its now empty, */
		    /* chain it back in */
		    HAT_PRINTF1 (2, "elan3mmu_l2inval: keep l3ptbl %p\n", l3ptbl);

		    elan3_writeptp (dev, l2ptp, tl2ptp);
		    elan3mmu_unlock_ptbl (l3ptbl, l3lock, l3flags);
		}
	    }
	    else
	    {
		l2ptbl->ptbl_valid--;

		HAT_PRINTF3 (2, "elan3mmu_l2inval: dec valid for level %d ptbl %p to %d\n", 
			     PTBL_LEVEL(l2ptbl->ptbl_flags), l2ptbl, l2ptbl->ptbl_valid);
	    }
	    break;
	    
	case ELAN3_ET_INVALID:
	    break;

	default:
	    panic ("elan3mmu_l2inval: found pte in level 2 page table");
	    /* NOTREACHED */
	}

	if (l2ptbl->ptbl_valid == 0)
	    break;

	addr += ELAN3_L2_SIZE;
    }

    ASSERT (PTBL_IS_LOCKED(l2ptbl->ptbl_flags));

    return (ret);
}

int 
elan3mmu_l3inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l3ptbl, int attr, E3_Addr addr, spinlock_t **pl3lock, unsigned long *flags)
{
    int ret;

    HAT_PRINTF3 (2, "elan3mmu_l3inval: l3ptbl %p parent %p addr %x\n", l3ptbl, l3ptbl->ptbl_parent, addr);

    ASSERT (PTBL_IS_LOCKED (l3ptbl->ptbl_parent->ptbl_flags));
    ASSERT (PTBL_LEVEL (l3ptbl->ptbl_parent->ptbl_flags) == PTBL_LEVEL_2);
    ASSERT (l3ptbl->ptbl_parent->ptbl_elan3mmu == elan3mmu);
    ASSERT (l3ptbl->ptbl_parent->ptbl_base == VA2BASE (ELAN3_L2_BASE(addr)));
    
    ret = elan3mmu_lock_ptbl (l3ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_3, pl3lock, flags);

    ASSERT (ret == LK_PTBL_OK);
    ASSERT (PTBL_LEVEL (l3ptbl->ptbl_flags) == PTBL_LEVEL_3);

    elan3mmu_unload_loop (elan3mmu, l3ptbl, 0, ELAN3_L3_ENTRIES, attr);

    ASSERT (PTBL_IS_LOCKED (l3ptbl->ptbl_flags));

    return (ret);
 }

int
elan3mmu_lock_this_ptbl (ELAN3_PTBL *ptbl, int flag, spinlock_t **plock, unsigned long *flags)
{
    int         level = PTBL_LEVEL (ptbl->ptbl_flags);
    spinlock_t *lock  = elan3mmu_ptbl_to_lock (level, ptbl);

    local_irq_save (*flags);

    if ((flag & LK_PTBL_NOWAIT) == 0)
	spin_lock (lock);
    else if (! spin_trylock (lock)) {
	local_irq_restore (*flags);
	return (LK_PTBL_FAILED);
    }
    
    if (level != PTBL_LEVEL (ptbl->ptbl_flags))
    {
	spin_unlock (lock);	
	local_irq_restore (*flags);
	return (LK_PTBL_MISMATCH);
    }

    ptbl->ptbl_flags |= PTBL_LOCKED;
    *plock = lock;
    return (LK_PTBL_OK);
}

int
elan3mmu_lock_ptbl (ELAN3_PTBL *ptbl, u_int flag, ELAN3MMU *elan3mmu, E3_Addr va, int level, spinlock_t **plock, unsigned long *flags)
{
    spinlock_t *lock = elan3mmu_ptbl_to_lock (level, ptbl);
    int         res  = LK_PTBL_MISMATCH;

    local_irq_save (*flags);
    
    if ((flag & LK_PTBL_NOWAIT) == 0)
	spin_lock (lock);
    else if (spin_trylock (lock) == 0) {
	local_irq_restore(*flags);
	return (LK_PTBL_FAILED);
    }
    
    if (PTBL_LEVEL (ptbl->ptbl_flags) != level)
    {
	res = LK_PTBL_MISMATCH;
	goto mismatch;
    }
    
    /* We have the right mutex,  so check that its the ptbl we want. */
    switch (level)
    {
    case PTBL_LEVEL_1: va = ELAN3_L1_BASE(va); break;
    case PTBL_LEVEL_2: va = ELAN3_L2_BASE(va); break;
    case PTBL_LEVEL_3: va = ELAN3_L3_BASE(va); break;
    }

    if (ptbl->ptbl_elan3mmu != elan3mmu || ptbl->ptbl_base != VA2BASE(va))
    {
	res = LK_PTBL_MISMATCH;
	goto mismatch;
    }

    ASSERT ((ptbl->ptbl_flags & PTBL_LOCKED) == 0);
    ptbl->ptbl_flags |= PTBL_LOCKED;

    *plock = lock;
    return (LK_PTBL_OK);

mismatch:
    if (! (flag & LK_PTBL_FAILOK))
	panic ("elan3mmu: failed to lock ptbl\n");
	
    spin_unlock (lock);
    local_irq_restore(*flags);
    return (res);
}

void
elan3mmu_unlock_ptbl (ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags)
{
    ptbl->ptbl_flags &= ~PTBL_LOCKED;
    spin_unlock_irqrestore (lock,flags);
}

static spinlock_t *
elan3mmu_ptbl_to_lock (int level, ELAN3_PTBL *ptbl)
{
    switch (level)
    {
    case PTBL_LEVEL_3: return (&l3ptbl_lock[L3PTBL_MTX_HASH(ptbl)]);
    case PTBL_LEVEL_2: return (&l2ptbl_lock[L2PTBL_MTX_HASH(ptbl)]);
    case PTBL_LEVEL_1: return (&l1ptbl_lock[L1PTBL_MTX_HASH(ptbl)]);
    case PTBL_LEVEL_X: 
	panic ("elan3mmu: ptbl_to_lock, bad level X");
    default:
	panic ("elan3mmu: ptbl_to_lock, bad level");
	/* NOTREACHED */
    }
    return (NULL);
}

void
elan3mmu_display (ELAN3MMU *elan3mmu, E3_Addr addr)
{
    ELAN3_DEV   *dev = elan3mmu->elan3mmu_dev;
    ELAN3_PTBL  *l1ptbl;
    sdramaddr_t l1ptp;
    spinlock_t *l1lock;
    ELAN3_PTE    tl1pte;
    ELAN3_PTP    tl1ptp;
    E3_Addr     l1base;
    ELAN3_PTBL  *l2ptbl;
    sdramaddr_t l2ptp;
    ELAN3_PTE    tl2pte;
    spinlock_t *l2lock;
    ELAN3_PTP    tl2ptp;
    E3_Addr     l2base;
    ELAN3_PTBL  *l3ptbl;
    sdramaddr_t l3pte;
    ELAN3_PTE    tl3pte;
    spinlock_t *l3lock;
    ELAN3_PTBL  *lXptbl;
    int         idx;
    unsigned long flags;

    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: elan3mmu %p addr %x\n", elan3mmu, addr);

    l1ptbl = elan3mmu->elan3mmu_l1ptbl;

    if (l1ptbl == NULL)
	return;

    l1ptp  = PTBL_TO_PTADDR(l1ptbl) + ELAN3_L1_INDEX(addr)*ELAN3_PTP_SIZE;
    l1base = ELAN3_L1_BASE(addr);
    
    tl1ptp = elan3_readptp (dev, l1ptp);
    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: l1ptbl %p l1ptp %lx l1base %x : tl1ptp %x\n", l1ptbl, l1ptp, l1base, tl1ptp);
    
    switch (ELAN3_PTP_TYPE(tl1ptp))
    {
    case ELAN3_ET_PTE:
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: level 1 page table for pte %x\n", tl1ptp);
    
	lXptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
	idx    = (PTP_TO_PT_PADDR(tl1ptp) - PTBL_TO_PTADDR(lXptbl))/ELAN3_PTE_SIZE;  
	
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: lXptbl %p idx %d\n",lXptbl, idx);

	tl1pte = elan3_readpte (dev,(PTBL_TO_PTADDR (lXptbl) + idx * ELAN3_PTE_SIZE));

	switch (elan3mmu_lock_ptbl (l1ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_1, &l1lock, &flags))
	{
	case LK_PTBL_OK:
	    elan3mmu_unlock_ptbl (l1ptbl, l1lock, flags);
	    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: lvl 1 l1pte matches value %llx\n", (long long) tl1pte);
	    break;
	    
	case LK_PTBL_FAILED:
	    panic ("elan3mmu_display: l1 lock failed");
	    /* NOTREACHED */
	    
	case LK_PTBL_MISMATCH:
	    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: PTBL_MISMATCH : lvl 1 ptbl %p flags %x elan3mmu %p base %x (%p %x) %llx\n",
			  l1ptbl, l1ptbl->ptbl_flags, l1ptbl->ptbl_elan3mmu, l1ptbl->ptbl_base, elan3mmu, addr, (long long)tl1pte);
	    
	    break;
	default:
	    panic ("elan3mmu_display: lvl 1 elan3mmu_lock_ptbl returned bad value");
	    /* NOTREACHED */
	}
	return;
	
    case ELAN3_ET_INVALID:
	return;
	
    case ELAN3_ET_PTP:
	break;
	
    default:
	panic ("elan3mmu_display: found bad entry in level 1 page table");
	/* NOTREACHED */
    }
    
    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: chain to level 2 ptbl from ptp %x\n", tl1ptp);
    
    l2ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl1ptp);
    l2ptp  = PTBL_TO_PTADDR(l2ptbl) + ELAN3_L2_INDEX(addr)*ELAN3_PTP_SIZE;
    l2base = ELAN3_L2_BASE(addr);
    
    tl2ptp = elan3_readptp (dev, l2ptp);
    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: l2ptbl %p l2ptp %lx l2base %x : tl2ptp %x\n",
		 l2ptbl, l2ptp, l2base, tl2ptp);
    
    switch (ELAN3_PTP_TYPE(tl2ptp))
    {
    case ELAN3_ET_PTE:
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: level 2 page table for pte %x\n", tl2ptp);
    
	lXptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
	idx    = (PTP_TO_PT_PADDR(tl2ptp) - PTBL_TO_PTADDR(lXptbl))/ELAN3_PTE_SIZE;  
	
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: lXptbl %p idx %d\n",lXptbl, idx);

	tl2pte = elan3_readpte (dev,(PTBL_TO_PTADDR (lXptbl) + idx * ELAN3_PTE_SIZE));

	switch (elan3mmu_lock_ptbl (l2ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_2, &l2lock, &flags))
	{
	case LK_PTBL_OK:
	    elan3mmu_unlock_ptbl (l2ptbl, l2lock, flags);
	    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: lvl 2 l1pte matches value %llx\n", (long long)tl2pte);
	    break;
	    
	case LK_PTBL_FAILED:
	    panic ("elan3mmu_display: l2 lock failed");
	    /* NOTREACHED */
	    
	case LK_PTBL_MISMATCH:
	    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: PTBL_MISMATCH : lvl 2 ptbl %p flags %x elan3mmu %p base %x (%p %x) %llx\n",
			  l2ptbl, l2ptbl->ptbl_flags, l2ptbl->ptbl_elan3mmu, l2ptbl->ptbl_base, elan3mmu, addr, (long long) tl2pte);
	    
	    break;
	default:
	    panic ("elan3mmu_display: lvl 2 elan3mmu_lock_ptbl returned bad value");
	    /* NOTREACHED */
	}
	return;
	
    case ELAN3_ET_INVALID:
	return;
	
    case ELAN3_ET_PTP:
	break;

    default:
	panic ("elan3mmu_display: found bad entry in level 2 page table");
	/* NOTREACHED */
    }
    
    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: chain to level 3 page table from ptp %x\n", tl2ptp);
    
    l3ptbl = elan3mmu_ta_to_ptbl (elan3mmu, &tl2ptp);
    l3pte  = PTBL_TO_PTADDR(l3ptbl) + ELAN3_L3_INDEX(addr)*ELAN3_PTE_SIZE;
    
    elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: l3ptbl %p l3pte %lx\n",l3ptbl, l3pte);
    
    tl3pte = elan3_readpte (dev, l3pte);
    switch (elan3mmu_lock_ptbl (l3ptbl, LK_PTBL_FAILOK, elan3mmu, addr, PTBL_LEVEL_3, &l3lock, &flags))
    {
    case LK_PTBL_OK:
	elan3mmu_unlock_ptbl (l3ptbl, l3lock, flags);
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: l3pte matches value %llx\n", (long long) tl3pte);
	break;
	
    case LK_PTBL_FAILED:
	panic ("elan3mmu_display: l3 lock failed");
	/* NOTREACHED */
	
    case LK_PTBL_MISMATCH:
	elan3_debugf (NULL, DBG_HAT, "elan3mmu_display: PTBL_MISMATCH : ptbl %p flags %x elan3mmu %p base %x (%p %x) %llx\n",
		     l3ptbl, l3ptbl->ptbl_flags, l3ptbl->ptbl_elan3mmu, l3ptbl->ptbl_base, elan3mmu, addr, (long long) tl3pte);
	
	break;
	
    default:
	panic ("elan3mmu_display: elan3mmu_lock_ptbl returned bad value");
	/* NOTREACHED */
    }
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
