/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elan3ops.c,v 1.4 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/elan3ops.c,v $*/

#include <qsnet/kernel.h>
#include <elan/elanmod.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elan3ops.h>

extern ELAN_STATS_OPS elan3_device_stats_ops;

ELAN_DEV_OPS elan3_dev_ops = {

	get_position,
	set_position,	

	ELAN_DEV_OPS_VERSION
};

ELAN_STATS_OPS elan3_device_stats_ops = {
	ELAN_STATS_OPS_VERSION,

	stats_get_index_name,
	stats_get_block,
	stats_clear_block
};

static char *elan3_device_stats_names[ELAN3_NUM_STATS] = 
{
		"version field",                 /* not cleared */
		"elan interrupts",
		"tlb flushes",
		"traps with invalid context",
		"interrupts com queue half full",
		"cproc traps",
		"dproc traps",
		"tproc traps",
		"iproc traps",
		"event interrupts",
		"elan page faults",
		"EopBadAcks",
		"EopResets",
		"InputterBadLength",
		"InputterCRCDiscards",
		"InputterCRCErrors",
		"InputterCRCBad",	
		"errors in dma data",
		"errors after dma identify",
		"errors after thread identify",
		"dma retries",
		"dma output timeouts",
		"dma packet ack errors",
		"forced tproc traps",
		"too many instruction traps",
		"output timeouts",
		"packet ack errors",
		"LockError",
		"DeskewError",
		"PhaseError",
		"DataError",
		"FifoOvFlow0",
		"FifoOvFlow1",
		"link error value on data error",
		"correctable ecc errors",
		"uncorrectable ecc errors",
		"multiple ecc errors",
		"sdram bytes free",              /* not cleared */
		"longest interrupt in ticks",
		"punts of event int's to thread",
		"reschedules of event int's thread"
};

int 
stats_get_index_name (void *arg, uint  index, caddr_t name)
{
	copyout (elan3_device_stats_names[index], name, strlen (elan3_device_stats_names[index]) + 1  /* with \0 */);

	return (0);
}

int
stats_get_block (void *arg, uint entries, ulong *value)
{
	ELAN3_DEV *dev = (ELAN3_DEV *) arg;

	if ( entries >  ELAN3_NUM_STATS ) /* if space too big only send valid portion */
		entries = ELAN3_NUM_STATS;
	
	copyout(&dev->Stats, value, sizeof(ulong) * entries);

	return (0);
}

int 
stats_clear_block (void *arg)
{
	ELAN3_DEV *dev = (ELAN3_DEV *) arg;
	u_long   *ptr = (u_long *) &dev->Stats;
	int                n;
	
	for (n = 0; n < ELAN3_NUM_STATS; n++)
	{
		switch (n) 
		{
		case offsetof (ELAN3_STATS, Version)/sizeof(u_long):
		case offsetof (ELAN3_STATS, SdramBytesFree)/sizeof(u_long):
			break;
		default:
			ptr[n] = (ulong)0;
		}
	}
	return (0);
}

int 
get_position (void *user_data, ELAN_POSITION *position)
{
	ELAN3_DEV *dev = (ELAN3_DEV *)user_data;

	copyout(&dev->Position, position, sizeof(ELAN_POSITION));

	return (0);	
}

int 
set_position (void *user_data, unsigned short nodeId, unsigned short numNodes)
{
	ELAN3_DEV *dev = (ELAN3_DEV *)user_data;

	if (ComputePosition (&dev->Position, nodeId, numNodes, dev->Devinfo.dev_num_down_links_value) != 0)
		return (EINVAL);
	
	return (0);	
}

int
elan3_register_dev_stats(ELAN3_DEV * dev) 
{
	char name[ELAN_STATS_NAME_MAX_LEN+1];

	sprintf (name, ELAN3_STATS_DEV_FMT, dev->Instance);

	elan_stats_register(&dev->StatsIndex,
			       name,
			       sizeof (elan3_device_stats_names)/sizeof (elan3_device_stats_names[0]),
			       &elan3_device_stats_ops,
			       (void *)dev);

	return (0);
}

void
elan3_deregister_dev_stats(ELAN3_DEV * dev) 
{
	elan_stats_deregister(dev->StatsIndex);
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
