/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elandebug.c,v 1.25 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/elandebug.c,v $*/

#include <qsnet/kernel.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>


void
elan3_debugf (void *p, unsigned int mode, char *fmt,...)
{
    char prefix[128];

#if defined (DIGITAL_UNIX)
#define PREFIX_FMT	"[%lx.%08x]"
#define PREFIX_VAL	(int)CURTHREAD()
#else
#define PREFIX_FMT	"[%lx.%04d]"
#define PREFIX_VAL	(current->pid)
#endif

    if ((unsigned long) p > DBG_NTYPES)
    {
	ELAN3_CTXT *ctxt = (ELAN3_CTXT *) p;

        if (elan3_debug_display_ctxt && (ctxt->Capability.cap_mycontext & MAX_ROOT_CONTEXT_MASK) != elan3_debug_display_ctxt)
            return;
        if (elan3_debug_ignore_ctxt  && (ctxt->Capability.cap_mycontext & MAX_ROOT_CONTEXT_MASK) == elan3_debug_ignore_ctxt)
            return;
 
	if (ctxt->Capability.cap_mycontext == ELAN_CAP_UNINITIALISED)
	    sprintf (prefix, PREFIX_FMT " (XXX) ", lbolt, PREFIX_VAL);
	else
	    sprintf (prefix, PREFIX_FMT " (%03x) ", lbolt, PREFIX_VAL,
		     ctxt->Capability.cap_mycontext & MAX_ROOT_CONTEXT_MASK);
    }
    else
    {
	char *what;

	if (elan3_debug_ignore_dev & (1 << ((unsigned long) p)))
	    return;

	switch ((unsigned long) p)
	{
	case (int) DBG_DEVICE: what = "dev"; break;
	case (int) DBG_KCOMM:  what = "kcm"; break;
	case (int) DBG_ICS:    what = "ics"; break;
	case (int) DBG_USER:   what = "usr"; break;
	default:               what = NULL; break;
	}
	    
	if (what)
	    sprintf (prefix, PREFIX_FMT " [%s] ", lbolt,  PREFIX_VAL, what);
	else
	    sprintf (prefix, PREFIX_FMT " [%3d] ", lbolt,  PREFIX_VAL, (int)(long)what);
    }

    {
	va_list       ap;

	va_start (ap, fmt);
	qsnet_vdebugf ((((mode & elan3_debug_buffer)?QSNET_DEBUG_BUFFER:0)|((mode & elan3_debug_console)?QSNET_DEBUG_CONSOLE:0)) , prefix, fmt, ap);
	va_end (ap);
    }
}


void
elan3_alloc_panicstate (ELAN3_DEV *dev, int allocsdram)
{
    register int bank;

    if (dev->PanicState.RegPtr == NULL)
	KMEM_ZALLOC (dev->PanicState.RegPtr, E3_Regs *, sizeof (E3_Regs), 1);

    if (allocsdram)
	for (bank = 0; bank < ELAN3_SDRAM_NUM_BANKS; bank++)
	    if (dev->PanicState.Sdram[bank] == NULL && dev->SdramBanks[bank].Size)
		KMEM_ZALLOC (dev->PanicState.Sdram[bank], char *, dev->SdramBanks[bank].Size, 1);
}

void
elan3_free_panicstate (ELAN3_DEV *dev)
{
    register int bank;

    if (dev->PanicState.RegPtr != NULL)
	KMEM_FREE (dev->PanicState.RegPtr, sizeof (E3_Regs));

    for (bank = 0; bank < ELAN3_SDRAM_NUM_BANKS; bank++)
	if (dev->PanicState.Sdram[bank] != NULL && dev->SdramBanks[bank].Size)
	    KMEM_FREE (dev->PanicState.Sdram[bank], dev->SdramBanks[bank].Size);

    bzero (&dev->PanicState, sizeof (dev->PanicState));
}

void
elan3_save_panicstate (ELAN3_DEV *dev)
{
    register int bank;
    
    if (dev->PanicState.RegPtr)
    {
	printk ("elan%d: saving state on panic .....\n", dev->Devinfo.dev_instance);

	bcopy ((void *) dev->RegPtr, (void *) dev->PanicState.RegPtr, sizeof (E3_Regs));
	
	for (bank = 0; bank < ELAN3_SDRAM_NUM_BANKS; bank++)
	    if (dev->SdramBanks[bank].Size && dev->PanicState.Sdram[bank])
		elan3_sdram_copyq_from_sdram (dev, (bank << ELAN3_SDRAM_BANK_SHIFT), dev->PanicState.Sdram[bank], dev->SdramBanks[bank].Size);
	
    }
}

int
elan3_assfail (ELAN3_DEV *dev, char *string, char *file, int line)
{
    if (panicstr)
	return (0);

    printk ("elan: assertion failed '%s' File '%s' Line %d\n", string, file, line);

#if defined(LINUX)
    elan3_save_panicstate (dev);

    panic ("elan: assertion failed '%s' File '%s' Line %d\n", string, file, line);
#else
    cmn_err (CE_PANIC, "elan: assertion failed '%s' File '%s' Line %d\n", string, file, line);
#endif
    /*NOTREACHED*/
    return (0);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
