/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elandev_generic.c,v 1.115.2.2 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/elandev_generic.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan3/dma.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/elansyscall.h>
#include <elan3/urom_addrs.h>
#include <elan3/elan3ops.h>

/*
 * Module globals, configurable from system file.
 */
u_int  elan3_debug                  = 0;
u_int  elan3_debug_console	  = 0;
u_int  elan3_debug_buffer           = -1;
u_int  elan3_debug_ignore_dev       = 0;
u_int  elan3_debug_ignore_kcomm     = 0;
u_int  elan3_debug_ignore_ctxt      = 0;
u_int  elan3_debug_display_ctxt     = 0;

int    eventint_punt_loops;
int    eventint_punt_ticks;
int    eventint_resched_ticks;

static void InitialiseDmaBuffers (ELAN3_DEV *dev, ioaddr_t CmdPort);
static int  ProbeSdram (ELAN3_DEV *dev);
static void InitialiseSdram (ELAN3_DEV *dev);
static void ReEnableErrorInterrupts (void *arg);
void        PollForDmaHungup (void *arg);
static void elan3_event_interrupt (ELAN3_DEV *dev);

/*
 * BaseAddr is ptr to the start of a table aligned on a power of two byte address.
 * SizePower must be in the range of 6 to 12. It defines the number of valid contexts as
 * shown below.
 *
 * SizePower	Valid Contexts	Table size in bytes.
 *     6	     64		      1k
 *     7	    128		      2k
 *     8	    256		      4K
 *     9	    512		      8k
 *    10	   1024		     16k
 *    11	   2048		     32k
 *    12	   4096		     64k
 */
#define GEN_CONTEXT_PTR(BaseAddr, SizePower) (((E3_uint32) BaseAddr) |	\
 					      (~((1 << ((SizePower) - 6)) - 1) & 0x3f))

int
InitialiseElan (ELAN3_DEV *dev, ioaddr_t CmdPort)
{
    E3_IprocTrapHeader_BE   TrapCleanup[4];
    E3_ContextControlBlock  ContextControlBlock;
    sdramaddr_t             ptr;
    int			    res;
    int		            i;

    eventint_punt_loops    = 100;
    eventint_punt_ticks    = (hz/100);
    eventint_resched_ticks = (hz/4);
    
    dev->Stats.Version     = ELAN3_STATS_VERSION;
    dev->Position.pos_mode = ELAN_POS_UNKNOWN;

    /*
     * The elan should have already been reset, so the interrupt mask
     * should be 0 and the schedule status register should be set to
     * its initial state
     */
    ASSERT (dev->InterruptMask == 0);
    ASSERT ((read_reg32 (dev, Exts.SchCntReg) & HaltStopAndExtTestMask) == Sched_Initial_Value);

    /*
     * Write any value here to clear out the half full and error bits of the command
     * overflow queues.
     */
    write_reg32 (dev, ComQueueStatus, 0);

    /* Initialise the cache tags before touching the SDRAM */
    /* we initialise them to "map" the bottom of SDRAM */
    for (i = 0; i < E3_NumCacheLines; i++)
    {
	write_cache_tag (dev, Tags[i][0].Value, 0x0000000000000000ULL);
	write_cache_tag (dev, Tags[i][1].Value, 0x0000080000000000ULL);
	write_cache_tag (dev, Tags[i][2].Value, 0x0000100000000000ULL);
	write_cache_tag (dev, Tags[i][3].Value, 0x0000180000000000ULL);
    }

#ifndef CONFIG_MPSAS
    for (i = 0; i < E3_NumCacheLines*(E3_CACHELINE_SIZE/sizeof(E3_uint64)); i++)
    {
	write_cache_set (dev, Set0[i], 0xcac1ecac1ecac1e0ULL);
	write_cache_set (dev, Set1[i], 0xcac1ecac1ecac1e1ULL);
	write_cache_set (dev, Set2[i], 0xcac1ecac1ecac1e2ULL);
	write_cache_set (dev, Set3[i], 0xcac1ecac1ecac1e3ULL);
    }
#endif

    if ((res = ProbeSdram(dev)) != ESUCCESS)
	return (res);

    /* Enable all cache sets before initialising the sdram allocators */
    write_reg32 (dev, Cache_Control_Reg.ContReg, (dev->Cache_Control_Reg |= CONT_EN_ALL_SETS));

    InitialiseSdram (dev);

    dev->TAndQBase              = elan3_sdram_alloc (dev, ELAN3_TANDQ_SIZE);
    dev->ContextTable           = elan3_sdram_alloc (dev, ELAN3_CONTEXT_SIZE);
    dev->ContextTableSize       = ELAN3_NUM_CONTEXTS;
    dev->CommandPortTraps[0]    = elan3_sdram_alloc (dev, ELAN3_COMMAND_TRAP_SIZE);
    dev->CommandPortTraps[1]    = elan3_sdram_alloc (dev, ELAN3_COMMAND_TRAP_SIZE);
    dev->CurrentCommandPortTrap = 0;

    PRINTF3 (DBG_DEVICE, DBG_CONFIG, "InitialiseElan: ContextTable %08lx TAndQ %08lx CommandPortTrap %08lx\n",
	     dev->ContextTable, dev->TAndQBase, dev->CommandPortTraps[0]);

    /* Allocate the thread amd dma trap areas */
    KMEM_ZALLOC (dev->ThreadTrap, THREAD_TRAP *, sizeof (THREAD_TRAP), TRUE);
    KMEM_ZALLOC (dev->DmaTrap, DMA_TRAP *, sizeof (DMA_TRAP), TRUE);

    /* Allocate the ctxt table */
    KMEM_ZALLOC (dev->CtxtTable,  ELAN3_CTXT **, dev->ContextTableSize * sizeof ( ELAN3_CTXT *), TRUE);

    /* Initialise halt queue list */
    dev->HaltOperationsTailpp   = &dev->HaltOperations;

    /* From elan3/code/harness/elanstuff.c */
    /* Init the clock. */
    write_ureg64 (dev, Clock.NanoSecClock, 0);
    
    /* Init the instruction count reg. */
    write_ureg32 (dev, InstCount.s.StatsCount, 0);
    
    /* Init the stats control reg. Must be done before the count regs.*/
    write_ureg32 (dev, StatCont.StatsControl, 0);
    
    /* Init the stats count regs. */
    write_ureg32 (dev, StatCounts[0].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[1].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[2].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[3].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[4].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[5].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[6].s.StatsCount, 0);
    write_ureg32 (dev, StatCounts[7].s.StatsCount, 0);
    
    /*
     * Initialise the Context_Ptr and Fault_Base_Ptr
     */
    write_reg32 (dev, Fault_Base_Ptr, dev->TAndQBase + offsetof(E3_TrapAndQueue, IProcSysCntx));
    write_reg32 (dev, Context_Ptr, GEN_CONTEXT_PTR (dev->ContextTable, ELAN3_LN2_NUM_CONTEXTS));

    /* scrub the TProc Registers */
    for (i = 0; i < 8; i++)
	write_reg32 (dev, Globals[i], 0xdeadbabe);
    for (i = 0; i < 8; i++)
	write_reg32 (dev, Outs[i], 0xdeadbabe);
    for (i = 0; i < 8; i++)
	write_reg32 (dev, Locals[i], 0xdeadbabe);
    for (i = 0; i < 8; i++)
	write_reg32 (dev, Ins[i], 0xdeadbabe);

    /*
     * Initialise the Queue pointers.  Arrange them so that the starting positions are
     * farthest apart in one set of the cache. Thus 512 bytes apart,  but with cntx0
     * thread the same as the interrupt queue.
     */
    write_reg32 (dev, TProc_NonSysCntx_FPtr, dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxThreadQueue[0xc0]));
    write_reg32 (dev, TProc_NonSysCntx_BPtr, dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxThreadQueue[0xc0]));
    write_reg32 (dev, TProc_SysCntx_FPtr,    dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxThreadQueue[0x80]));
    write_reg32 (dev, TProc_SysCntx_BPtr,    dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxThreadQueue[0x80]));
    
    write_reg32 (dev, DProc_NonSysCntx_FPtr, dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[0]));
    write_reg32 (dev, DProc_NonSysCntx_BPtr, dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[0]));
    write_reg32 (dev, DProc_SysCntx_FPtr,    dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0x10]));
    write_reg32 (dev, DProc_SysCntx_BPtr,    dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0x10]));
    
    dev->Event_Int_Queue_FPtr = dev->TAndQBase + offsetof (E3_TrapAndQueue, EventIntQueue[0x80]);
    write_reg32 (dev, Event_Int_Queue_FPtr, dev->Event_Int_Queue_FPtr);
    write_reg32 (dev, Event_Int_Queue_BPtr, dev->TAndQBase + offsetof (E3_TrapAndQueue, EventIntQueue[0x80]));
    
    
    /* Initialise Input_Trap_Base to last 8 Kbytes of trap area, uCode adds the right offset */
    write_reg32 (dev, Input_Trap_Base, dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxThreadQueue[0]));
    
    /* Ptr to word used to save the SP to when a thread deschedules */
    write_reg32 (dev, Thread_SP_Save_Ptr, dev->TAndQBase + offsetof (E3_TrapAndQueue, Thread_SP_Save));
    
    /* Initialise the command trap base */
    write_reg32 (dev, CProc_TrapSave_Addr, dev->CommandPortTraps[0]);
    
    /* Initialise the set event tracing registers */
    write_reg32 (dev, Event_Trace_Ptr, 0);
    write_reg32 (dev, Event_Trace_Mask, 0);
    
    /* Initialise Tlb_Line_Value to zero. The TLB cannot be read while either the */
    /* uCode or thread proc might be running. Must be set to 0. */
    write_reg64 (dev, Tlb_Line_Value, 0);

    /* Control register. Cache everything, Enable MMU, RefreshRate=3, CasLatency=1, StartSDR */
    dev->Cache_Control_Reg |= CONT_MMU_ENABLE | CONT_EN_ALL_SETS | CONT_CACHE_ALL | CONT_ENABLE_ECC;

#if ELAN3_PAGE_SHIFT == 13
    dev->Cache_Control_Reg |= CONT_ENABLE_8K_PAGES;
#endif

    write_reg32 (dev, Cache_Control_Reg.ContReg,  dev->Cache_Control_Reg);

    /*
     * Initialise the context table to be discard for all contexts
     */
    ContextControlBlock.rootPTP  = 0;
    ContextControlBlock.filter   = E3_CCB_DISCARD_ALL;
    ContextControlBlock.VPT_mask = 0;
    ContextControlBlock.VPT_ptr  = 0;

    for (i = 0, ptr = dev->ContextTable; i < ELAN3_NUM_CONTEXTS; i++, ptr += sizeof (E3_ContextControlBlock))
	elan3_sdram_copyl_to_sdram (dev, &ContextControlBlock, ptr, sizeof (E3_ContextControlBlock));

    /* From elan3/code/trap_handler/init.c */
    /*
     * Initialise the Trap And Queue area in Elan SDRAM.
     */
    TrapCleanup[0].s.TrTypeCntx.TypeContext = 0;
    TrapCleanup[0].s.TrAddr		    = 0;
    TrapCleanup[0].s.IProcTrapStatus.Status = CRC_STATUS_GOOD;
    TrapCleanup[0].s.TrData0		    = 0;
    TrapCleanup[1].s.TrTypeCntx.TypeContext = 0;
    TrapCleanup[1].s.TrAddr		    = 0;
    TrapCleanup[1].s.IProcTrapStatus.Status = CRC_STATUS_GOOD;
    TrapCleanup[1].s.TrData0		    = 0;
    TrapCleanup[2].s.TrTypeCntx.TypeContext = 0;
    TrapCleanup[2].s.TrAddr		    = 0;
    TrapCleanup[2].s.IProcTrapStatus.Status = CRC_STATUS_GOOD;
    TrapCleanup[2].s.TrData0		    = 0;
    TrapCleanup[3].s.TrTypeCntx.TypeContext = 0;
    TrapCleanup[3].s.TrAddr		    = 0;
    TrapCleanup[3].s.IProcTrapStatus.Status = CRC_STATUS_GOOD;
    TrapCleanup[3].s.TrData0		    = 0;

    elan3_sdram_writel (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcSysCntx.s.FaultContext),  0);
    elan3_sdram_writel (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcSysCntx.s.FSR.Status), 0);
    elan3_sdram_writel (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcNonSysCntx.s.FaultContext), 0);
    elan3_sdram_writel (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcNonSysCntx.s.FSR.Status), 0);
    
    /* Must now zero all the FSRs so that a subsequent Fault can be seen */ 
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, CProc), 16);

    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0), 64);
    
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, TProc), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcData), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcInst), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcOpen), 16);

    elan3_sdram_copyq_to_sdram (dev, TrapCleanup, dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_C0_TrHead[0]), 64);
    elan3_sdram_copyq_to_sdram (dev, TrapCleanup, dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_C0_TrHead[0]), 64);

    elan3_sdram_copyq_to_sdram (dev, TrapCleanup, dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_NonC0_TrHead[0]), 64);
    elan3_sdram_copyq_to_sdram (dev, TrapCleanup, dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_NonC0_TrHead[0]), 64);

    InitialiseDmaBuffers(dev, CmdPort);

    /* reserve a halt operation for flushing the context filter */
    ReserveHaltOperations (dev, 1, TRUE);

    /* Allow the Thread/Dma to run */
    CLEAR_SCHED_STATUS (dev, HaltThread | HaltDmas);

    /* Enable All Interrrupts */
    SET_INT_MASK (dev, (INT_PciMemErr | INT_SDRamInt | INT_EventInterrupt | INT_LinkError | INT_ComQueue |
			INT_TProc | INT_CProc | INT_DProc | INT_IProcCh1NonSysCntx | 
			INT_IProcCh1SysCntx | INT_IProcCh0NonSysCntx | INT_IProcCh0SysCntx));

    /* Take the link out of boundary scan */
    SET_SCHED_LINK_VALUE (dev, 0, 0);
    
    /* And clear any link errors */
    PULSE_SCHED_STATUS (dev, ClearLinkErrorInt);

    /* XXXX: clear discard context 0,  AFTER setting up the kernel comms */
    CLEAR_SCHED_STATUS (dev, DiscardSysCntxIn | DiscardNonSysCntxIn);

    /* Start a thread to handle excessive Event Interrrupts */
    if (kernel_thread_create (elan3_event_interrupt, (caddr_t) dev) == NULL)
    {
	panic ("InitialiseElan: cannot start elan3_event_interrupt\n");
	return (EFAIL);
    }
    dev->EventInterruptThreadStarted = 1;

    ReserveHaltOperations (dev, 1, TRUE);

    PollForDmaHungup (dev);

#if defined(IOPROC_PATCH_APPLIED) && !defined(NO_PTRACK)
    /* Register the device and stats with elanmod for RMS
     * but only if we've got the coproc patch applied */
    dev->DeviceIdx = elan_dev_register(&dev->Devinfo, &elan3_dev_ops, (void *) dev);
    
    elan3_register_dev_stats(dev);
#endif

    return (ESUCCESS);
}

static void
InitialiseDmaBuffers(ELAN3_DEV *dev, ioaddr_t CmdPort)
{
   register int i;

   /* GNAT sw-elan3/3908:
    * Clear down the power on state of the Dma_Desc registers to make sure we don't
    * try and interpret them when a trap happens.
    */
   write_reg32 (dev, Dma_Desc.dma_type,            0);
   write_reg32 (dev, Dma_Desc.dma_size,            0);
   write_reg32 (dev, Dma_Desc.dma_source,          0);
   write_reg32 (dev, Dma_Desc.dma_dest,            0);
   write_reg32 (dev, Dma_Desc.dma_destEvent,       0);
   write_reg32 (dev, Dma_Desc.dma_destCookieVProc, 0);
   write_reg32 (dev, Dma_Desc.dma_srcEvent,        0);
   write_reg32 (dev, Dma_Desc.dma_srcCookieVProc,  0);
   
   /*
    * The following is a sequence of writes to remove X's from the dma buffers and 
    * registers. It is only safe to write these registers after reset and before any
    * dma's have been issued. The chip will NOT function corectly if they are written at
    * any other time or in a different order.
    */
   write_reg64 (dev, Exts.Dmas.DmaWrs.LdAlignment, 0);
   write_reg64 (dev, Exts.Dmas.DmaWrs.LdDmaType, 0);
   write_reg64 (dev, Exts.Dmas.DmaWrs.ResetAckNLdBytesToWr, ((u_longlong_t)0x1000) << 32);
   write_reg64 (dev, Exts.Dmas.DmaWrs.LdBytesToRd, ((u_longlong_t)0x100) << 32);

   for (i=0;i<(4*8);i++)
       write_reg64 (dev, Dma_Alignment_Port[0], 0);

   /*
    * This is used to clear out X's from some of the trap registers. This is required to
    * prevent the first traps from possibly writting X's into the SDram and upsetting the
    * ECC value. It requires that the trap save area registers have been set up but does
    * not require any translations to be ready.
    */
   writel (-1, (void *)(CmdPort + offsetof (E3_CommandPort, SetEvent)));
   while ((read_reg32 (dev, Exts.InterruptReg) & INT_CProc) == 0)
   {
       mb();
       DELAY (1);
   }

   write_reg32 (dev, CProc_TrapSave_Addr, dev->CommandPortTraps[dev->CurrentCommandPortTrap]);
   
   PULSE_SCHED_STATUS(dev, RestartCProc);
}

void
FinaliseElan (ELAN3_DEV *dev)
{
    ELAN3_PTBL_GR *ptg;
    ELAN3_HALTOP  *op;
    ELAN3_HALTOP  *chain = NULL;
    int           bank;
    int		  indx;
    int		  size;
    unsigned long flags;
    int           level;

#if defined(IOPROC_PATCH_APPLIED) && !defined(NO_PTRACK)
    elan_stats_deregister (dev->StatsIndex);
    elan_dev_deregister(&dev->Devinfo);
#endif

    /* Cancel the dma poller */
    cancel_timer_fn (&dev->DmaPollTimeoutId);

    /* release it's halt operation */
    ReleaseHaltOperations (dev, 1);

    /* stop all kernel threads */
    dev->ThreadsShouldStop = 1;

    spin_lock_irqsave (&dev->IntrLock, flags);
    while (dev->EventInterruptThreadStarted && !dev->EventInterruptThreadStopped)
    {
	kcondvar_wakeupall (&dev->IntrWait, &dev->IntrLock);
	kcondvar_wait (&dev->IntrWait, &dev->IntrLock, &flags);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    /* Set the interrupt mask to 0 and the schedule control register to run nothing */
    SET_INT_MASK (dev, 0);
    SET_SCHED_STATUS (dev, DiscardNonSysCntxIn | DiscardSysCntxIn | HaltThread | HaltDmas);

    /* Cancel any link error timeout */
    if (timer_fn_queued(&dev->ErrorTimeoutId))
	cancel_timer_fn (&dev->ErrorTimeoutId);

    /* Free of and page tables that have been allocated */
    spin_lock (&dev->PtblGroupLock);
    for(level=0; level<4; level++) 
    {
	while ((ptg = dev->Level[level].PtblGroupList) != NULL)
	{
	    dev->Level[level].PtblGroupList = ptg->pg_next;

	    elan3_sdram_free (dev, ptg->pg_addr, PTBL_GROUP_SIZE);
	    FREE_PTBL_GR(ptg);
	}
    }
 
    spin_unlock (&dev->PtblGroupLock);

    /* Free of all halt operations */
    spin_lock_irqsave (&dev->FreeHaltLock, flags);
    while ((op = dev->FreeHaltOperations) != NULL)
    {
	dev->FreeHaltOperations = op->Next;

	/* Keep a list of 'freed' ops for later KMEM_FREE call */
	op->Next = chain;
	chain = op;
    }
    spin_unlock_irqrestore (&dev->FreeHaltLock, flags);

    /* Have now dropped the spinlock - can call KMEM_FREE */
    while ((op = chain) != NULL)
    {
	chain = op->Next;

	KMEM_FREE (op, sizeof (ELAN3_HALTOP));
    }
	
    /* Free of the ctxt table */
    KMEM_FREE (dev->CtxtTable,  dev->ContextTableSize * sizeof (ELAN3_CTXT *));

    /* Free of the thread and dma atrap areas */
    KMEM_FREE (dev->ThreadTrap, sizeof (THREAD_TRAP));
    KMEM_FREE (dev->DmaTrap, sizeof (DMA_TRAP));

    /* Free of the memsegs and pages */
    for (bank = 0; bank < ELAN3_SDRAM_NUM_BANKS; bank++)
    {
	if (dev->SdramBanks[bank].Size)
	{
	    UnmapDeviceRegister (dev, &dev->SdramBanks[bank].Handle);

	    KMEM_FREE (dev->SdramBanks[bank].PtblGroups, sizeof (ELAN3_PTBL_GR *) * (dev->SdramBanks[bank].Size / PTBL_GROUP_SIZE));

	    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size <= dev->SdramBanks[bank].Size; indx++, size <<= 1)
		KMEM_FREE (dev->SdramBanks[bank].Bitmaps[indx], sizeof (bitmap_t)*BT_BITOUL(dev->SdramBanks[bank].Size/size));
	}
    }
    elan3_sdram_fini (dev);
}

#define INIT_PATTERN(offset)	(0xBEEC000000000011ull | ((u_longlong_t)(offset)) << 16)
#define FREE_PATTERN(offset)	(0xBEEC000000000022ull | ((u_longlong_t)(offset)) << 16)

static int
ProbeSdram (ELAN3_DEV *dev)
{
    int			  Instance;
    u_int		  Bank;
    int		          MemSpaceSize;
    int			  BankMaxSize;
    int			  BankOffset;
    int			  BankSize;
    ioaddr_t		  BankBase;
    ioaddr_t		  PageBase;
    ioaddr_t		  PageBase1;
    ioaddr_t		  PageBase2;
    DeviceMappingHandle   BankHandle;
    DeviceMappingHandle   PageHandle;
    DeviceMappingHandle   PageHandle1;
    DeviceMappingHandle   PageHandle2;
    register int          i;
    u_longlong_t	  value;
    extern int            sdram_bank_limit;

    /* NOTE: The Cache control register is set to only enable cache set 0 */
    /*       and has ECC disabled */
    Instance = dev->Instance;
    
    /* Determine the size of the SDRAM from the BAR register */
    if (DeviceRegisterSize (dev, ELAN3_BAR_SDRAM, &MemSpaceSize) != ESUCCESS)
    {
	printk ("elan%d: cannot determine SDRAM size\n", Instance);
	return (EFAIL);
    }

    elan3_sdram_init (dev);

    BankMaxSize = MemSpaceSize / ELAN3_SDRAM_NUM_BANKS;

    for (Bank = 0; Bank < ELAN3_SDRAM_NUM_BANKS; Bank++)
    {
	BankOffset = Bank * BankMaxSize;
	
	PRINTF3 (DBG_DEVICE, DBG_CONFIG, "elan%d: Probing RAM Bank %d (max size %08x)\n", Instance, Bank, BankMaxSize);
	
	/* Probe the memory bank by mapping two pages that are the size of the cache apart */
	/* this guarantees that when we store the second pattern we displace the first pattern */
	/* from the cache, also store the second pattern again the size of the cache up again */
	/* to ensure that the SDRAM wires don't stay floating at pattern1 */

	if (MapDeviceRegister (dev, ELAN3_BAR_SDRAM, &BankBase, BankOffset, PAGESIZE, &BankHandle) != ESUCCESS)
	{
	    printk ("elan%d: Cannot probe memory bank %d\n", Instance, Bank);
	    continue;
	}
	
	if (MapDeviceRegister (dev, ELAN3_BAR_SDRAM, &PageBase1, BankOffset + ELAN3_MAX_CACHE_SIZE, PAGESIZE, &PageHandle1) != ESUCCESS)
	{
	    printk ("elan%d: Cannot probe memory bank %d\n", Instance, Bank);
	    UnmapDeviceRegister (dev, &BankHandle);
	    continue;
	}

	if (MapDeviceRegister (dev, ELAN3_BAR_SDRAM, &PageBase2, BankOffset + 2*ELAN3_MAX_CACHE_SIZE, PAGESIZE, &PageHandle2) != ESUCCESS)
	{
	    printk ("elan%d: Cannot probe memory bank %d\n", Instance, Bank);
	    UnmapDeviceRegister (dev, &BankHandle);
	    UnmapDeviceRegister (dev, &PageHandle1);
	    continue;
	}

#define PATTERN0	(0x5555555555555555ull)
#define PATTERN1	(0xAAAAAAAAAAAAAAAAull)
	writeq (PATTERN0, (u_longlong_t *) BankBase);
	writeq (PATTERN1, (u_longlong_t *) PageBase1);
	writeq (PATTERN1, (u_longlong_t *) PageBase2);

	mmiob();

	value = readq ((u_longlong_t *) BankBase);

	if (value != PATTERN0)
	{
	    UnmapDeviceRegister (dev, &BankHandle);
	    UnmapDeviceRegister (dev, &PageHandle1);
	    UnmapDeviceRegister (dev, &PageHandle2);
	    continue;
	}

	writeq (PATTERN1, (u_longlong_t *) BankBase);
	writeq (PATTERN0, (u_longlong_t *) PageBase1);
	writeq (PATTERN0, (u_longlong_t *) PageBase2);

	mmiob();
	
	value = readq ((u_longlong_t *) BankBase);
	if (value != PATTERN1)
	{
	    UnmapDeviceRegister (dev, &BankHandle);
	    UnmapDeviceRegister (dev, &PageHandle1);
	    UnmapDeviceRegister (dev, &PageHandle2);
	    continue;
	}
	UnmapDeviceRegister (dev, &PageHandle1);
	UnmapDeviceRegister (dev, &PageHandle2);

	/* Bank is present, so work out its size,  we store tha maximum size at the base */
	/* and then store the address at each address  on every power of two address until */
	/* we reach the minimum mappable size (PAGESIZE), we then read back the value at the */
	/* base to determine the bank size */
	writeq ((u_longlong_t) BankMaxSize, (u_longlong_t *) BankBase);

	for (BankSize = (BankMaxSize>>1); BankSize > PAGESIZE; BankSize >>= 1)
	{
	    if (MapDeviceRegister (dev, ELAN3_BAR_SDRAM, &PageBase, BankOffset + BankSize, PAGESIZE, &PageHandle) == ESUCCESS)
	    {
		writeq (BankSize, (u_longlong_t *) PageBase);
		UnmapDeviceRegister (dev, &PageHandle);
	    }
	}
	mmiob();

	BankSize = (u_long) readq ((u_longlong_t *) BankBase);
	
	if (sdram_bank_limit == 0 || BankSize <= (sdram_bank_limit * 1024 * 1024))
	    printk ("elan%d: memory bank %d is %dK\n", Instance, Bank, BankSize / 1024);
	else
	{
	    BankSize = (sdram_bank_limit * 1024 * 1024);
	    printk ("elan%d: limit memory bank %d to %dK\n", Instance, Bank, BankSize / 1024);
	}

	UnmapDeviceRegister (dev, &BankHandle);
	
	/* Now map all of this bank into the kernel */
	if (MapDeviceRegister (dev, ELAN3_BAR_SDRAM, &BankBase, BankOffset, BankSize, &BankHandle) != ESUCCESS)
	{
	    printk ("elan%d: Cannot initialise memory bank %d\n", Instance, Bank);
	    continue;
	}
	
	dev->SdramBanks[Bank].Size    = BankSize;
	dev->SdramBanks[Bank].Mapping = BankBase;
	dev->SdramBanks[Bank].Handle  = BankHandle;

#ifndef CONFIG_MPSAS
	/* Initialise it for ECC */
	preemptable_start {
	    for (i = 0; i < BankSize; i += 8)
	    {
		elan3_sdram_writeq (dev, (Bank << ELAN3_SDRAM_BANK_SHIFT) | i, INIT_PATTERN(BankOffset+i));

		preemptable_check();
	    }
	} preemptable_end;
#endif
    }
    
    return (ESUCCESS);
}

static void
InitialiseSdram (ELAN3_DEV *dev)
{
    int indx, size, b;

    for (b = 0; b < ELAN3_SDRAM_NUM_BANKS; b++)
    {
	ELAN3_SDRAM_BANK *bank = &dev->SdramBanks[b];

	if (bank->Size == 0)
	    continue;

	/* allocate a ptbl group pointer for each possible ptbl group in this bank */
	KMEM_ZALLOC (bank->PtblGroups, ELAN3_PTBL_GR **, sizeof (ELAN3_PTBL_GR *) * bank->Size/PTBL_GROUP_SIZE, TRUE);
	    
	/* allocate the buddy allocator bitmaps */
	for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size <= bank->Size; indx++, size <<= 1)
	    KMEM_ZALLOC (bank->Bitmaps[indx], bitmap_t *, sizeof (bitmap_t)*BT_BITOUL(bank->Size/size), TRUE);
	    
	/* and add it to the sdram buddy allocator */
	elan3_sdram_add (dev, (b << ELAN3_SDRAM_BANK_SHIFT), (b << ELAN3_SDRAM_BANK_SHIFT) + bank->Size);
    }
}

#include <elan3/vpd.h>

int
ReadVitalProductData (ELAN3_DEV *dev, int *CasLatency)
{
    DeviceMappingHandle RomHandle;
    unsigned char      *RomBase;
    unsigned char      *PCIDataPtr;
    unsigned char      *VPDPtr;
    unsigned char      *lim;
    int			type;
    int			i, len, len2;
    char		name[3] = "XX";
    char		value[256];
    int			finished = 0;

    
    /* default valud for CAS latency is 3 */
    (*CasLatency) = CAS_LATENCY_3;

    if (MapDeviceRegister (dev, ELAN3_BAR_EBUS, (ioaddr_t *) &RomBase, ELAN3_EBUS_ROM_OFFSET, ELAN3_EBUS_ROM_SIZE, &RomHandle) != ESUCCESS)
    {
	printk ("elan%d: Cannot map ROM\n", dev->Instance);
	return (EFAIL);
    }
    
    /* Check the ROM signature */
    if (RomBase[0] != 0x55 || RomBase[1] != 0xAA)
    {
	printk ("elan%d: Invalid ROM signature %02x %02x\n", dev->Instance, RomBase[0], RomBase[1]);
	return (ESUCCESS);
    }
    
    PCIDataPtr = RomBase + ((RomBase[0x19] << 8) | RomBase[0x18]);

    /* check the pci data structure */
    if (PCIDataPtr[0] != 'P' || PCIDataPtr[1] != 'C' || PCIDataPtr[2] != 'I' || PCIDataPtr[3] != 'R')
    {
	printk ("elan%d: Invalid PCI Data structure\n", dev->Instance);
	return (ESUCCESS);
    }
    
    /* Extract the VPD pointer */
    VPDPtr = RomBase + ((PCIDataPtr[9] << 8) | PCIDataPtr[8]);

    if (VPDPtr == RomBase)
    {
	printk ("elan%d: No Vital Product Data\n", dev->Instance);
	return (ESUCCESS);
    }

    while (! finished)
    {
	type = *VPDPtr++;
	
	if (type & LARGE_RESOURCE_BIT)
	{
	    len = *(VPDPtr++);
	    len += *(VPDPtr++) << 8;

	    switch (type & ~LARGE_RESOURCE_BIT)
	    {
	    case LARGE_RESOURCE_STRING:
		printk ("elan%d: ", dev->Instance);
		for (i = 0; i < len; i++)
		    printk ("%c", *VPDPtr++);
		printk ("\n");
		break;
		
	    case LARGE_RESOURCE_VENDOR_DEFINED:
		VPDPtr += len;
		break;
		
	    case LARGE_RESOURCE_VITAL_PRODUCT_DATA:
		for (lim = VPDPtr + len; VPDPtr < lim; )
		{
		    name[0] = *VPDPtr++;
		    name[1] = *VPDPtr++;
		    len2    = *VPDPtr++;

		    for (i = 0; i < len2 && VPDPtr < lim; i++)
			value[i] = *VPDPtr++;
		    value[i] = '\0';

		    if (! strcmp (name, "SN"))
			printk ("elan%d: Serial Number - %s\n", dev->Instance, value);

		    if (! strcmp (name, "Z0"))
			(*CasLatency) = (strcmp (value, "CAS_LATENCY_2") ? CAS_LATENCY_3 : CAS_LATENCY_2);
		}
		break;
		
	    default:
		printk ("elan%d: unknown large resource %x\n", dev->Instance, type);
		finished = 1;
		break;
	    }
	}
	else
	{
	    len = type & 0x7;

	    switch (type >> 3)
	    {
	    case SMALL_RESOURCE_COMPATIBLE_DEVICE_ID:
		VPDPtr += len;
		break;

	    case SMALL_RESOURCE_VENDOR_DEFINED:
		VPDPtr += len;
		break;
		
	    case SMALL_RESOURCE_END_TAG:
		finished = 1;
		break;
		
	    default:
		printk ("elan%d: unknown small resource %x\n", dev->Instance, type >> 3);
		finished = 1;
		break;
	    }
	}
    }
    
    UnmapDeviceRegister (dev, &RomHandle);
    return (ESUCCESS);
}

void
ElanSetPtblGr (ELAN3_DEV *dev, sdramaddr_t offset, ELAN3_PTBL_GR *ptg)
{
    int bank = offset >> ELAN3_SDRAM_BANK_SHIFT;
    
    dev->SdramBanks[bank].PtblGroups[(offset & (ELAN3_SDRAM_BANK_SIZE-1)) / PTBL_GROUP_SIZE] = ptg;
}

ELAN3_PTBL_GR *
ElanGetPtblGr (ELAN3_DEV *dev, sdramaddr_t offset)
{
    int bank = offset >> ELAN3_SDRAM_BANK_SHIFT;
    
    return (dev->SdramBanks[bank].PtblGroups[(offset & (ELAN3_SDRAM_BANK_SIZE-1)) / PTBL_GROUP_SIZE]);
}

void
ElanFlushTlb (ELAN3_DEV *dev)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->TlbLock, flags);
    BumpStat (dev, TlbFlushes);

    write_reg32 (dev, Cache_Control_Reg.ContReg, dev->Cache_Control_Reg | MMU_FLUSH);
    mmiob();
    spin_unlock_irqrestore (&dev->TlbLock, flags);

    while (! (read_reg32 (dev, Cache_Control_Reg.ContReg) & MMU_FLUSHED))
	mb();
}

void
KillNegativeDma (ELAN3_DEV *dev, void *arg)
{
    DMA_TRAP     *trap    = dev->DmaTrap;
    E3_Status_Reg status;
    sdramaddr_t   FPtr, BPtr;
    sdramaddr_t   Base, Top;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    ASSERT (read_reg32 (dev, Exts.InterruptReg) & INT_DProcHalted);

    /* Initialise the trap to deliver to the offending user process */
    trap->Status.Status   = read_reg32 (dev, Exts.DProcStatus.Status);
    trap->PacketInfo.Value = 0;
    
    bzero (&trap->FaultSave, sizeof (trap->FaultSave));
    bzero (&trap->Data0, sizeof (trap->Data0));
    bzero (&trap->Data1, sizeof (trap->Data1));
    bzero (&trap->Data2, sizeof (trap->Data2));
    bzero (&trap->Data3, sizeof (trap->Data3));

    /* run down the kernel dma run queue and panic on a -ve length dma */
    FPtr  = read_reg32 (dev, DProc_SysCntx_FPtr);
    BPtr  = read_reg32 (dev, DProc_SysCntx_BPtr);
    Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0]);
    Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[E3_SysCntxQueueSize-1]);
    
    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, &trap->Desc, sizeof (E3_DMA_BE));
	
	if (trap->Desc.s.dma_size > E3_MAX_DMA_SIZE)
	    panic ("KillNegativeDma: -ve sized kernel dma\n");

	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_DMA);
    }

    /* run down the user dma run queue and "remove" and -ve length dma's */
    FPtr  = read_reg32 (dev, DProc_NonSysCntx_FPtr);
    BPtr  = read_reg32 (dev, DProc_NonSysCntx_BPtr);
    Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[0]);
    Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, NonSysCntxDmaQueue[E3_NonSysCntxQueueSize-1]);
    
    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, &trap->Desc, sizeof (E3_DMA_BE));
	
	if (trap->Desc.s.dma_size > E3_MAX_DMA_SIZE)
	{
	    PRINTF3 (NULL, DBG_INTR, "KillNegativeDma: remove dma - context %d size %d SuspendAddr %x\n", 
		     trap->Desc.s.dma_u.s.Context, trap->Desc.s.dma_size, trap->Status.s.SuspendAddr);

	    trap->Status.s.TrapType = trap->Status.s.SuspendAddr;
	    trap->Status.s.Context  = trap->Desc.s.dma_u.s.Context;

	    DeliverDProcTrap (dev, trap, 0);

	    /*
	     * Remove the DMA from the queue by replacing it with one with
	     * zero size and no events.
	     *
	     * NOTE: we must preserve the SYS_CONTEXT_BIT since the Elan uses this
	     * to mark the approriate run queue as empty.
	     */
	    trap->Desc.s.dma_type            = 0;
	    trap->Desc.s.dma_size            = 0;
	    trap->Desc.s.dma_source          = (E3_Addr) 0;
	    trap->Desc.s.dma_dest            = (E3_Addr) 0;
	    trap->Desc.s.dma_destCookieVProc = (E3_Addr) 0;
	    trap->Desc.s.dma_srcEvent        = (E3_Addr) 0;
	    trap->Desc.s.dma_srcCookieVProc  = (E3_Addr) 0;

	    elan3_sdram_copyq_to_sdram (dev, &trap->Desc, FPtr, sizeof (E3_DMA_BE));
	}

	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_DMA);
    }

    status.Status = read_reg32 (dev, Exts.DProcStatus.Status);

    if (status.s.SuspendAddr == MI_DequeueNonSysCntxDma || 
	status.s.SuspendAddr == MI_DequeueSysCntxDma ||
	status.s.SuspendAddr == MI_DmaLoop)
    {
	PRINTF0 (NULL, DBG_INTR, "KillNegativeDma: unlock dma processor\n");
	write_reg32 (dev, Exts.Dmas.DmaWrs.LdAlignment, 0);
	write_reg32 (dev, Exts.Dmas.DmaWrs.LdDmaType,   0);
	mmiob();
	
	DELAY (10);
	
	write_reg32 (dev, Exts.Dmas.DmaWrs.LdAlignment, 0);
	write_reg32 (dev, Exts.Dmas.DmaWrs.LdDmaType,   0);
	mmiob();
    }

    PRINTF0 (NULL, DBG_INTR, "KillNegativeDma: dma processor restarted\n");

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    schedule_timer_fn (&dev->DmaPollTimeoutId, PollForDmaHungup, (void *) dev, 1);
}

void
ForceTProcTrap (ELAN3_DEV *dev, void *arg)
{
    printk ("elan%d: forced tproc trap .....\n", dev->Instance);

    schedule_timer_fn (&dev->DmaPollTimeoutId, PollForDmaHungup, (void *) dev, 1);
}

void
PollForDmaHungup (void *arg)
{
    ELAN3_DEV     *dev   = (ELAN3_DEV *) arg;
    unsigned long flags;
    E3_Status_Reg status;
    E3_uint32     insn1, insn3;
    register int  i;

    if (read_reg32 (dev, Dma_Desc.dma_size) > E3_MAX_DMA_SIZE)
    {
	status.Status = read_reg32 (dev, Exts.DProcStatus);

	PRINTF2 (NULL, DBG_INTR, "PollForDmaHungup: size %x SuspendAddr %x\n", read_reg32 (dev, Dma_Desc.dma_size), status.s.SuspendAddr);

	if (status.s.SuspendAddr == MI_DequeueNonSysCntxDma || 
	    status.s.SuspendAddr == MI_DequeueSysCntxDma ||
	    status.s.SuspendAddr == MI_DmaLoop)
	{
	    printk ("elan%d: PollForDmaHungup: size %x context %d SuspendAddr %x\n", 
		    dev->Instance, read_reg32 (dev, Dma_Desc.dma_size),
		    status.s.Context, status.s.SuspendAddr);
	
	    PRINTF2 (NULL, DBG_INTR, "PollForDmaHungup: dma_size %x status %x\n",
		     read_reg32 (dev, Dma_Desc.dma_size), status.Status);
	    
	    spin_lock_irqsave (&dev->IntrLock, flags);
	    QueueHaltOperation (dev, 0, NULL, INT_DProcHalted, KillNegativeDma, NULL);
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    
	    return;
	}
    }

    status.Status = read_reg32 (dev, Exts.TProcStatus);
    if (status.s.WakeupFunction == WakeupStopped)
    {
	E3_uint32 PC = read_reg32 (dev, ExecutePC);

	/* See if it's likely that the thread is really "stuck" on a waitevent/break 
	 * instruction ......... */
	for (i = 0; i < 10; i++)
	{
	    status.Status = read_reg32 (dev, Exts.TProcStatus);
	    insn1         = read_reg32 (dev, IBufferReg[1]);
	    insn3         = read_reg32 (dev, IBufferReg[3]);
	    
	    if (! (status.s.WakeupFunction == WakeupStopped && read_reg32 (dev, ExecutePC) == PC && 	/* stopping and it could be a break/waitevent */
		   (insn1 == 0x81a00000 || insn3 == 0x81a00000 || 					/* break instruction */
		    insn1 == 0x81b00000 || insn3 == 0x81b00000)))					/* waitevent instruction  */
		break;
	}

	if (i == 10)
	{
	    printk ("elan%d: forcing tproc trap from %s instruction at pc %x\n", dev->Instance, 
		    (insn1 == 0x81a00000 || insn3 == 0x81a00000) ? "break" : "waitevent", PC);

	    spin_lock_irqsave (&dev->IntrLock, flags);
	    QueueHaltOperation (dev, 0, NULL, INT_TProcHalted, ForceTProcTrap, NULL);
	    spin_unlock_irqrestore (&dev->IntrLock, flags);
	    return;
	}
    }

    schedule_timer_fn (&dev->DmaPollTimeoutId, PollForDmaHungup, (void *) dev, 10);
}

/*=======================================================================================*/
/*
 * Interrupt handler.
 */
static void
ReEnableErrorInterrupts (void *arg)
{
    ELAN3_DEV     *dev = (ELAN3_DEV *) arg;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);

    if ((dev->SchCntReg & LinkBoundaryScan) == 0)
	ENABLE_INT_MASK (dev, INT_ErrorInterrupts);

    PRINTF1 (DBG_DEVICE, DBG_INTR, "ReEnableErrorInterrupts: IntMask=%x\n", read_reg32 (dev, Exts.InterruptMask));

    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

void
CheckForExcessiveErrorRate (ELAN3_DEV *dev)
{
    if (dev->ErrorTime == (lbolt/hz))
    {
	if (dev->ErrorsPerTick++ > 100)
	{
	    PRINTF0 (DBG_DEVICE, DBG_INTR, "CheckForExcessiveErrorRate: too many links errors, disabling interrupt\n");

	    DISABLE_INT_MASK (dev, INT_ErrorInterrupts);

	    schedule_timer_fn (&dev->ErrorTimeoutId, ReEnableErrorInterrupts, (void *) dev, hz);
	}
    }
    else
    {
	dev->ErrorTime     = (lbolt/hz);
	dev->ErrorsPerTick = 0;
    }
}
/*=======================================================================================*/
/*
 * Interrupt handler.
 */
static void
HandlePciMemErr (ELAN3_DEV *dev)
{
    PRINTF0 (DBG_DEVICE, DBG_INTR, "HandlePciMemErr : masking out interrupt\n");
    
    ElanBusError (dev);
    panic ("elan pci memory error\n");
}

static void
HandleSDRamInterrupt (ELAN3_DEV *dev)
{
    E3_uint32     EccStatus0 = read_reg32 (dev, ECC_STATUS0);
    E3_uint32     EccStatus1 = read_reg32 (dev, ECC_STATUS1);
    unsigned long flags;

    PRINTF5 (DBG_DEVICE, DBG_INTR, "elan: ECC error - Addr=%x UE=%x CE=%x ME=%x Syn=%x\n",
	     EccStatus0 & ECC_ADDR_MASK, EccStatus0 & ECC_UE_MASK, 
	     EccStatus0 & ECC_CE_MASK, EccStatus0 & ECC_ME_MASK, 
	     EccStatus1 & ECC_SYN_MASK);

    if (EccStatus0 & (ECC_UE_MASK|ECC_CE_MASK))
    {
	printk ("elan%d: ECC memory error (Address=%08x Syndrome=%02x %s%s%s)\n",
		dev->Instance, 
		(EccStatus0 & ECC_ADDR_MASK), (EccStatus1 & ECC_SYN_MASK), 
		(EccStatus0 & ECC_UE_MASK) ? "Uncorrectable "   : "",
		(EccStatus0 & ECC_CE_MASK) ? "Correctable "     : "",
		(EccStatus0 & ECC_ME_MASK) ? "Multiple Errors " : "");
    }

    if (EccStatus0 & ECC_UE_MASK)
	panic ("elan: Uncorrectable ECC memory error");
    if (EccStatus0 & ECC_CE_MASK)
	BumpStat (dev, CorrectableErrors);
    if (EccStatus0 & ECC_ME_MASK)
	BumpStat (dev, MultipleErrors);

    /*
     * Clear the interrupt and reset the error flags.
     * Note. Might loose an UE or CE if it occurs between reading the status and
     *       clearing the interrupt. I don't think this matters very much as the
     *	  status reg will only be used to identify a bad simm.
     */

    spin_lock_irqsave (&dev->TlbLock, flags);
    write_reg32 (dev, Cache_Control_Reg.ContReg, dev->Cache_Control_Reg | CLEAR_SDRAM_ERROR);
    mmiob();
    spin_unlock_irqrestore (&dev->TlbLock, flags);

    CheckForExcessiveErrorRate (dev);
}

static int
HandleEventInterrupt (ELAN3_DEV *dev, int nticks, unsigned long *flags)
{
    E3_uint32 Fptr  = dev->Event_Int_Queue_FPtr;
    E3_uint32 Bptr  = read_reg32 (dev, Event_Int_Queue_BPtr); 						/* PCI read */
    long      tlim  = lbolt + nticks;
    long      count = 0;
    ELAN3_CTXT *ctxt;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    ASSERT ((dev->InterruptMask & INT_EventInterrupt) == 0);
	    
    while (Fptr != Bptr)
    {
	while (Fptr != Bptr)
	{
	    E3_EventInt_BE  EvInt;
	    E3_uint32       Context;

	    /* If we're running in the interrupt handler and have seen a high
	     * rate of event interrupts then punt to the thread  - however on 
	     * Linux the elan interrupt handler can block the timer interrupt,
	     * and so lbolt (jiffies) is not incremented, hence we punt after
	     a number of loops instead */
#if defined(LINUX)
	    if (in_interrupt() && ++count > eventint_punt_loops)
		return (EAGAIN);
#endif

	    if (nticks && ((int) (lbolt - tlim)) > 0)
	    {
		PRINTF2 (DBG_DEVICE, DBG_INTR, "HandleEventInterrupt: Fptr %x Bptr %x punting to thread\n", Fptr, Bptr);
		return (EAGAIN);
	    }

	    elan3_sdram_copyq_from_sdram (dev, Fptr, (void *) &EvInt, 8); 				/* PCI read */
	    
	    /* The context number is held in the top 16 bits of the EventContext */
	    Context = (EvInt.s.EventContext >> 16) & MAX_ROOT_CONTEXT_MASK;
	    
	    PRINTF2 (DBG_DEVICE, DBG_INTR, "HandleEventInterrupt: Context %d : Cookie %x\n", Context, EvInt.s.IntCookie);
	    
	    ctxt = ELAN3_DEV_CTX_TABLE(dev, Context);

	    /* Work out new fptr, and store it in the device, since we'll be dropping the IntrLock */
	    Fptr = E3_EVENT_INTQ_NEXT(Fptr);
	    dev->Event_Int_Queue_FPtr = Fptr;

	    if (ctxt == NULL)
	    {
		PRINTF3 (DBG_DEVICE, DBG_INTR, "HandleEventInterrupt: Fptr %x Bptr %x context %d invalid\n",
			 Fptr, Bptr, Context);
		BumpStat (dev, InvalidContext);
	    }
	    else
	    {
		BumpStat (dev, EventInterrupts);
		
		spin_unlock_irqrestore (&dev->IntrLock, *flags);
		QueueEventInterrupt (ctxt, EvInt.s.IntCookie);
		spin_lock_irqsave (&dev->IntrLock, *flags);
	    }
	    
	    /* Re-read the FPtr, since we've dropped the IntrLock */
	    Fptr = dev->Event_Int_Queue_FPtr;
	    
	    /* Store the new FPtr to the elan, this also clears the interrupt. */
	    write_reg32 (dev, Event_Int_Queue_FPtr, Fptr);					/* PCI write */

	    mmiob();
	}

	mb();
	Bptr = read_reg32 (dev, Event_Int_Queue_BPtr); 						/* PCI read */
    }

    return (ESUCCESS);
}

int
SetLinkBoundaryScan (ELAN3_DEV *dev)
{
    int           res = ESUCCESS;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    if ((dev->SchCntReg & LinkBoundaryScan) != 0)
	res = EAGAIN;
    else
    {
	PRINTF0 (DBG_DEVICE, DBG_BSCAN, "SetLinkBoundaryScan: setting link into boundary scan mode\n");

	/*
	 * We're going to set the link into boundary scan mode,  so firstly
	 * set the inputters to discard everything.
	 */
	if (dev->DiscardAllCount++ == 0)
	    SetSchedStatusRegister (dev, read_reg32 (dev, Exts.InterruptReg), NULL);

	/*
	 * Now disable the error interrupts
	 */
	DISABLE_INT_MASK (dev, INT_ErrorInterrupts);
	
	/*
	 * And set the link into boundary scan mode, and drive
	 * a reset token onto the link.
	 */
	SET_SCHED_LINK_VALUE (dev, 1, LinkResetToken);
    }

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

void
ClearLinkBoundaryScan (ELAN3_DEV *dev)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    if ((dev->SchCntReg & LinkBoundaryScan) != 0)
    {
	PRINTF0 (DBG_DEVICE, DBG_BSCAN, "ClearLinkBoundaryScan: taking link out of boundary scan mode\n");

	/*
	 * Take the link out of boundary scan 
	 */
	SET_SCHED_LINK_VALUE (dev, 0, 0);

	/*
	 * Clear any link errors.
	 */
	PULSE_SCHED_STATUS (dev, ClearLinkErrorInt);

	/*
	 * Re-enable the error interrupts.
	 */
	if (! timer_fn_queued(&dev->ErrorTimeoutId))
	    ENABLE_INT_MASK (dev, INT_ErrorInterrupts);

	/*
	 * And stop the inputter from discarding all packets.
	 */
	if (--dev->DiscardAllCount == 0)
	    SetSchedStatusRegister (dev, 0, NULL);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

int
WriteBoundaryScanValue (ELAN3_DEV *dev, int value)
{
    int           res = 0;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    if ((dev->SchCntReg & LinkBoundaryScan) != 0)
    {
	PRINTF1 (DBG_DEVICE, DBG_BSCAN, "WriteBoundaryScanValue: driving value 0x%x onto link\n", value);
	SET_SCHED_LINK_VALUE (dev, 1, value);

	res = read_reg32 (dev, Exts.LinkState);

	PRINTF1 (DBG_DEVICE, DBG_BSCAN, "WriteBoundaryScanValue: return 0x%x\n", res);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

int
ReadBoundaryScanValue(ELAN3_DEV *dev, int link)
{
    int           res;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    if ((dev->SchCntReg & LinkBoundaryScan) == 0)
    {
	PRINTF1 (DBG_DEVICE, DBG_BSCAN, "ReadBoundaryScanValue: set linkval 0x%x\n",  link);
	SET_SCHED_LINK_VALUE (dev, 0, link);
    }
    res = read_reg32 (dev, Exts.LinkState);
    PRINTF1 (DBG_DEVICE, DBG_BSCAN, "ReadBoundaryScanValue: return 0x%x\n", res);

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

static int
ReadLinkVal (ELAN3_DEV *dev, int link)
{
    if ((dev->SchCntReg & LinkBoundaryScan) == 0)
	SET_SCHED_LINK_VALUE (dev, 0, link);
    
    return (read_reg32 (dev, Exts.LinkState));
}

static void
HandleLinkError (ELAN3_DEV *dev)
{
    E3_uint32 value = read_reg32 (dev, Exts.LinkErrorTypes);

    PRINTF1 (DBG_DEVICE, DBG_LINKERR, "HandleLinkError: LinkErrorTypes %08x - clearing\n", value);
    
    if (value & LS_LockError)   BumpStat (dev, LockError);
    if (value & LS_DeskewError) BumpStat (dev, DeskewError);
    if (value & LS_PhaseError)  BumpStat (dev, PhaseError);
    if (value & LS_DataError)   BumpStat (dev, DataError);
    if (value & LS_FifoOvFlow0) BumpStat (dev, FifoOvFlow0);
    if (value & LS_FifoOvFlow1) BumpStat (dev, FifoOvFlow1);

    if (value & LS_DataError)
	dev->Stats.LinkErrorValue = ReadLinkVal (dev, 12) | (ReadLinkVal (dev, 13) << 9);

    PULSE_SCHED_STATUS (dev, ClearLinkErrorInt);

    CheckForExcessiveErrorRate (dev);
}

static void
HandleErrorInterrupt (ELAN3_DEV *dev, E3_uint32 Pend)
{
    if (Pend & INT_PciMemErr)
	HandlePciMemErr (dev);
    
    if (Pend & INT_SDRamInt)
	HandleSDRamInterrupt (dev);
    
    if (Pend & INT_LinkError)
	HandleLinkError (dev);
}
	
static void
HandleAnyIProcTraps (ELAN3_DEV *dev, E3_uint32 Pend)
{
    E3_uint32	     RestartBits = 0;
    
    if (Pend & INT_IProcCh0SysCntx)
    {
	HandleIProcTrap (dev, 0, Pend,
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcSysCntx),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_C0_TrHead[0]),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_C0_TrData[0]));
			 
	RestartBits |= RestartCh0SysCntx;
    }
    
    if (Pend & INT_IProcCh1SysCntx)
    {
	HandleIProcTrap (dev, 1, Pend,
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcSysCntx),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_C0_TrHead[0]),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_C0_TrData[0]));
			 
	RestartBits |= RestartCh1SysCntx;
    }

    if (Pend & INT_IProcCh0NonSysCntx)
    {
	HandleIProcTrap (dev, 0, Pend,
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcNonSysCntx),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_NonC0_TrHead[0]),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh0_NonC0_TrData[0]));

	RestartBits |= RestartCh0NonSysCntx;
    }
    

    if (Pend & INT_IProcCh1NonSysCntx)
    {
	HandleIProcTrap (dev, 1, Pend,
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, IProcNonSysCntx),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_NonC0_TrHead[0]),
			 dev->TAndQBase + offsetof (E3_TrapAndQueue, VCh1_NonC0_TrData[0]));
	RestartBits |= RestartCh1NonSysCntx;
    }

    PULSE_SCHED_STATUS (dev, RestartBits);
}

static void
elan3_event_interrupt (ELAN3_DEV *dev)
{
    unsigned long flags;

    kernel_thread_init("elan3_event_int");

    spin_lock_irqsave (&dev->IntrLock, flags);
    for (;;)
    {
	/* Make sure we never sleep with the EventInterrupt disabled */
	if (! (dev->InterruptMask & INT_EventInterrupt))
	{
	    if (HandleEventInterrupt (dev, eventint_resched_ticks, &flags) != ESUCCESS)
		BumpStat (dev, EventRescheds);
	    
	    ENABLE_INT_MASK (dev, INT_EventInterrupt);
	}

	if (dev->ThreadsShouldStop)
	    break;

	kcondvar_wait (&dev->IntrWait, &dev->IntrLock, &flags);
    }
    
    dev->EventInterruptThreadStopped = 1;
    kcondvar_wakeupall (&dev->IntrWait, &dev->IntrLock);

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    kernel_thread_exit ();
}

int
InterruptHandler (ELAN3_DEV *dev)
{
    E3_uint32     Mask;
    E3_uint32     Pend;
    E3_uint32     RestartBits;
    int           deliverDProcTrap;
    int		  deliverTProcTrap;
    static long   lboltsave;
    int           loop_count = 0; 
    unsigned long flags;
    int  tproc_delivered;

    spin_lock_irqsave (&dev->IntrLock, flags);

    BumpStat (dev, Interrupts);

    Mask = dev->InterruptMask;
    Pend = read_reg32 (dev, Exts.InterruptReg);						/* PCI read */

    /* Save the lbolt so we know how long in do loop or in event handling */
    lboltsave = lbolt;

    if ((Pend & Mask) == INT_EventInterrupt)
    {
	DISABLE_INT_MASK (dev, INT_EventInterrupt);

	if (HandleEventInterrupt (dev, eventint_punt_ticks, &flags) == ESUCCESS)
	    ENABLE_INT_MASK (dev, INT_EventInterrupt);
	else
	{
	    BumpStat (dev, EventPunts);

	    kcondvar_wakeupone (&dev->IntrWait, &dev->IntrLock);
	}

        if ((lbolt - lboltsave) > dev->Stats.LongestInterrupt)
            dev->Stats.LongestInterrupt = (lbolt - lboltsave);
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	return (ESUCCESS);
    }

    if ((Pend & Mask) == 0)
    {
	PRINTF3 (DBG_DEVICE, DBG_INTR, "InterruptHandler: Spurious Pend %x Mask %x SchedStatus %x\n", 
		 Pend, Mask, read_reg32 (dev, Exts.SchCntReg));

        if ((lbolt - lboltsave) > dev->Stats.LongestInterrupt)
            dev->Stats.LongestInterrupt = (lbolt - lboltsave);
	spin_unlock_irqrestore (&dev->IntrLock, flags);
	return (EFAIL);
    }

    PRINTF3 (DBG_DEVICE, DBG_INTR, "InterruptHandler: Pend %x Mask %08x SchedStatus %x\n", 
	     Pend, Mask, read_reg32 (dev, Exts.SchCntReg));

    do {
	loop_count++;
	RestartBits = 0;

	if (Pend & Mask & (INT_CProc | INT_ComQueue))
	    HandleCProcTrap (dev, Pend, &Mask);

	tproc_delivered = 0;

	if (Pend & Mask & INT_TProc) {
	    ELAN_REG_REC(Pend);
	    tproc_delivered = 1;
	    deliverTProcTrap = HandleTProcTrap (dev, &RestartBits);
	}
	else
	    deliverTProcTrap = 0;

	if (Pend & Mask & INT_DProc)
	    deliverDProcTrap = HandleDProcTrap (dev, &RestartBits);
	else
	    deliverDProcTrap = 0;

	ASSERT ((RestartBits & RestartDProc) == 0 || (read_reg32 (dev, Exts.DProcStatus.Status) >> 29) == 4);
	ASSERT ((RestartBits & RestartDProc) == 0 || elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc.s.FSR.Status))      == 0);
	ASSERT ((RestartBits & RestartDProc) == 0 || elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0.s.FSR.Status)) == 0);
	ASSERT ((RestartBits & RestartDProc) == 0 || elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData1.s.FSR.Status)) == 0);
	ASSERT ((RestartBits & RestartDProc) == 0 || elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData2.s.FSR.Status)) == 0);
	ASSERT ((RestartBits & RestartDProc) == 0 || elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData3.s.FSR.Status)) == 0);

	PULSE_SCHED_STATUS (dev, RestartBits);		/* Restart any processors which had trapped. */
	SET_INT_MASK (dev, Mask);			/* And install the new interrupt mask */

	if ((Pend & Mask & INT_TProc) && deliverTProcTrap)
	    DeliverTProcTrap (dev, dev->ThreadTrap, Pend);

	if ((Pend & Mask & INT_DProc) && deliverDProcTrap)
	    DeliverDProcTrap (dev, dev->DmaTrap, Pend);

	if (Pend & Mask & INT_Inputters)
	    HandleAnyIProcTraps (dev, Pend);
	
	if (Pend & Mask & INT_EventInterrupt)
	{
	    DISABLE_INT_MASK (dev, INT_EventInterrupt);
	    
	    if (loop_count == 1 && HandleEventInterrupt (dev, eventint_punt_ticks, &flags) == ESUCCESS)	/* always punt to the thread if we've */
		ENABLE_INT_MASK (dev, INT_EventInterrupt); 						/* been round the loop once */
	    else
	    {
		BumpStat (dev, EventPunts);

		kcondvar_wakeupone (&dev->IntrWait, &dev->IntrLock);
	    }
	}

	if (Pend & (INT_Halted | INT_Discarding))
	    ProcessHaltOperations (dev, Pend);

	if (Pend & Mask & INT_ErrorInterrupts)
	    HandleErrorInterrupt (dev, Pend);

	Mask = dev->InterruptMask;
	Pend = read_reg32 (dev, Exts.InterruptReg);	/* PCI read */
	
	if (tproc_delivered)
	    ELAN_REG_REC(Pend);

	PRINTF3 (DBG_DEVICE, DBG_INTR, "InterruptHandler: Pend %x Mask %08x SchedStatus %x\n", 
		 Pend, Mask, read_reg32 (dev, Exts.SchCntReg));
    }  while ((Pend & Mask) != 0);

    if ((lbolt - lboltsave) > dev->Stats.LongestInterrupt)
        dev->Stats.LongestInterrupt = (lbolt - lboltsave);
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    PRINTF2 (DBG_DEVICE, DBG_INTR, "InterruptHandler: lbolt is %lx; start lbolt is %lx\n", 
	     lbolt, lboltsave);

    return (ESUCCESS);
}

void
SetSchedStatusRegister (ELAN3_DEV *dev, E3_uint32 Pend, volatile E3_uint32 *Maskp)
{
    E3_uint32 HaltMask  = dev->HaltOperationsMask;
    E3_uint32 Mask      = Maskp ? *Maskp : dev->InterruptMask;
    E3_uint32 ClearBits = 0;
    E3_uint32 SetBits   = 0;

    PRINTF5 (DBG_DEVICE, DBG_INTR, "SetSchedStatusRegister: HaltOperationsMask=%x HaltAll=%d HaltDmaDequeue=%d HaltThread=%d DiscardAll=%d\n",
	     HaltMask, dev->HaltAllCount, dev->HaltDmaDequeueCount, dev->HaltThreadCount, dev->DiscardAllCount);

    if (dev->FlushCommandCount)
	SetBits |= FlushCommandQueues;
    
    if ((HaltMask & INT_DProcHalted) || dev->HaltAllCount)
    {
	SetBits |= HaltDmas | HaltDmaDequeue;
	if (Pend & INT_DProcHalted)
	    Mask &= ~INT_DProcHalted;
	else
	    Mask |= INT_DProcHalted;
    }

    if (dev->HaltDmaDequeueCount)
    {
	SetBits |= HaltDmaDequeue;
	if (Pend & INT_DProcHalted)
	    Mask &= ~INT_DProcHalted;
	else
	    Mask |= INT_DProcHalted;
    }

    if ((HaltMask & INT_TProcHalted) || dev->HaltAllCount || dev->HaltThreadCount)
    {
	SetBits |= HaltThread;
	if (Pend & INT_TProcHalted)
	    Mask &= ~INT_TProcHalted;
	else
	    Mask |= INT_TProcHalted;
    }

    if ((HaltMask & INT_DiscardingSysCntx) || dev->DiscardAllCount)
    {
	SetBits |= DiscardSysCntxIn;
	if (Pend & INT_DiscardingSysCntx)
	    Mask &= ~INT_DiscardingSysCntx;
	else
	    Mask |= INT_DiscardingSysCntx;
    }

    if ((HaltMask & INT_DiscardingNonSysCntx) || dev->DiscardNonContext0Count || dev->DiscardAllCount)
    {
	SetBits |= DiscardNonSysCntxIn;
	if (Pend & INT_DiscardingNonSysCntx)
	    Mask &= ~INT_DiscardingNonSysCntx;
	else
	    Mask |= INT_DiscardingNonSysCntx;
    }

    if (dev->HaltNonContext0Count)
	SetBits |= StopNonSysCntxs;

    ClearBits = SetBits ^ (FlushCommandQueues | HaltDmas | HaltDmaDequeue | HaltThread |
			   DiscardSysCntxIn | DiscardNonSysCntxIn | StopNonSysCntxs);

    PRINTF4 (DBG_DEVICE, DBG_INTR, "SetSchedStatusRegister: SetBits=%x InterruptMask=%x InterruptReg=%x Mask=%x\n",
	     SetBits, dev->InterruptMask, read_reg32 (dev, Exts.InterruptReg), Mask);

    MODIFY_SCHED_STATUS (dev, SetBits, ClearBits);

    if (Maskp)
	*Maskp = Mask;						/* copyback new interrupt mask */
    else
	SET_INT_MASK(dev, Mask);
}

void
FreeHaltOperation (ELAN3_DEV *dev, ELAN3_HALTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->FreeHaltLock, flags);
    op->Next = dev->FreeHaltOperations;
    dev->FreeHaltOperations = op;
    spin_unlock_irqrestore (&dev->FreeHaltLock, flags);
}

int
ReserveHaltOperations (ELAN3_DEV *dev, int count, int cansleep)
{
    ELAN3_HALTOP   *op;
    unsigned long flags;

    spin_lock_irqsave (&dev->FreeHaltLock, flags);
    while ((dev->NumHaltOperations - dev->ReservedHaltOperations) < count)
    {
	spin_unlock_irqrestore (&dev->FreeHaltLock, flags);

	KMEM_ZALLOC (op, ELAN3_HALTOP *, sizeof (ELAN3_HALTOP), cansleep);

	if (op == NULL)
	    return (FALSE);

	spin_lock_irqsave (&dev->FreeHaltLock, flags);

	dev->NumHaltOperations++;

	op->Next = dev->FreeHaltOperations;
	dev->FreeHaltOperations = op;
    }
		    
    dev->ReservedHaltOperations += count;
    
    spin_unlock_irqrestore (&dev->FreeHaltLock, flags);

    return (TRUE);
}

void
ReleaseHaltOperations (ELAN3_DEV *dev, int count)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->FreeHaltLock, flags);
    dev->ReservedHaltOperations -= count;
    spin_unlock_irqrestore (&dev->FreeHaltLock, flags);
}

void
QueueHaltOperation (ELAN3_DEV *dev, E3_uint32 Pend, volatile E3_uint32 *Maskp, 
		    E3_uint32 ReqMask, void (*Function)(ELAN3_DEV *, void *), void *Arguement)
{
    ELAN3_HALTOP *op;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    
    spin_lock (&dev->FreeHaltLock);
    op = dev->FreeHaltOperations;

    ASSERT (op != NULL);

    dev->FreeHaltOperations = op->Next;
    spin_unlock (&dev->FreeHaltLock);

    op->Mask      = ReqMask;
    op->Function  = (void (*)(void *, void *))Function;
    op->Arguement = Arguement;

    dev->HaltOperationsMask |= ReqMask;				/* Add our bits to the global bits needed. */
    SetSchedStatusRegister (dev, Pend, Maskp);			/* Set the control register and the interrupt mask */

    /*
     * If the condition is already satisfied, then SetSchedStatusRegister will
     * have masked out the interrupt, so re-enable it now to take it straight
     * away
     */
    if (Maskp == NULL)
    {
	if ((read_reg32 (dev, Exts.InterruptReg) & ReqMask) == ReqMask)
	    ENABLE_INT_MASK (dev, ReqMask);
    }
    else
    {
	if ((Pend & ReqMask) == ReqMask)
	    *Maskp |= ReqMask;
    }

    *dev->HaltOperationsTailpp = op;				/* Queue at end of list, since ProcessHaltOperations */
    dev->HaltOperationsTailpp = &op->Next; 			/* drops the IntrLock while running down the list */
    op->Next = NULL;
}
		    
void
ProcessHaltOperations (ELAN3_DEV *dev, E3_uint32 Pend)
{
    E3_uint32     Mask;
    ELAN3_HALTOP  *op;
    ELAN3_HALTOP **prevp;
    E3_uint32     haltMask;
    ELAN3_HALTOP  *next;

    PRINTF1 (DBG_DEVICE, DBG_INTR, "ProcessHaltOperations: Pend %x\n", Pend);

    for (;;)
    {
	ELAN3_HALTOP  *head = NULL;
	ELAN3_HALTOP **tailp = &head;

	/*
	 * Generate a list of halt operations which can be called now.
	 */
	for (haltMask = 0, prevp = &dev->HaltOperations; (op = *prevp) != NULL; )
	{
	    if ((Pend & op->Mask) != op->Mask)
	    {
		haltMask |= op->Mask;
		prevp = &op->Next;
	    }
	    else
	    {
		*prevp = op->Next;				/* remove from list */
		if (op->Next == NULL)
		    dev->HaltOperationsTailpp = prevp;
		
		*tailp = op;					/* add to local list */
		op->Next = NULL;
		tailp = &op->Next;
	    }
	}

	if (head == NULL)					/* nothing to do, so update */
	{							/* the schedule status register */
	    dev->HaltOperationsMask = haltMask;			/* and the interrupt mask */
	    SetSchedStatusRegister (dev, Pend, NULL);
	    return;
	}

	/*
	 * flush the command queues, before calling any operations
	 */
	Mask = dev->InterruptMask;
	
	if (dev->FlushCommandCount++ == 0)
	    SetSchedStatusRegister (dev, Pend, &Mask);
	
	if ((read_reg32 (dev, ComQueueStatus) & ComQueueNotEmpty) != 0)
	{
	    if (dev->HaltThreadCount++ == 0)
		SetSchedStatusRegister (dev, Pend, &Mask);

	    CAPTURE_CPUS();

	    while ((read_reg32 (dev, ComQueueStatus) & ComQueueNotEmpty) != 0)
		mb();

	    RELEASE_CPUS();
		    
	    if (--dev->HaltThreadCount == 0)
		SetSchedStatusRegister (dev, Pend, &Mask);
	}
		
	if (read_reg32 (dev, Exts.InterruptReg) & INT_CProc)
	{
	    PRINTF0 (DBG_DEVICE, DBG_INTR, "ProcessHaltOperations: command processor has trapped\n");
	    HandleCProcTrap (dev, Pend, &Mask);
	}
	
	if (--dev->FlushCommandCount == 0)
	    SetSchedStatusRegister (dev, Pend, &Mask);
	
	PRINTF2 (DBG_DEVICE, DBG_INTR, "ProcessHaltOperations: interrupt mask %08x -> %08x\n", 
		 dev->InterruptMask, Mask);
	
	SET_INT_MASK (dev, Mask);
	spin_unlock (&dev->IntrLock);

	/*
	 * now process the list of operations
	 * we have
	 */
	for (op = head; op != NULL; op = next)
	{
	    next = op->Next;

	    op->Function (dev, op->Arguement);
	    
	    FreeHaltOperation (dev, op);
	}

	spin_lock (&dev->IntrLock);
    }
}

int
ComputePosition (ELAN_POSITION *pos, unsigned nodeId, unsigned numNodes, unsigned numDownLinksVal)
{
    int i, lvl, n;
    char numDownLinks[ELAN_MAX_LEVELS];

    if (nodeId >= numNodes)
	return (EINVAL);

    for (i = 0; i < ELAN_MAX_LEVELS; i++, numDownLinksVal >>= 4)
	numDownLinks[i] = numDownLinksVal & 7;
    
    for (lvl = 0, n = numNodes; n > ((lvl % 3) == 2 ? 8 : 4) && lvl < ELAN_MAX_LEVELS; lvl++)
    {
	if (numDownLinks[lvl] == 0)
	    numDownLinks[lvl] = 4;
	
	if ((n % numDownLinks[lvl]) != 0)
	    return (EINVAL);
	
	n /= numDownLinks[lvl];
    }

    if (numDownLinks[lvl] == 0)
	numDownLinks[lvl] = n;

    if (numDownLinks[lvl] != n)
	return (EINVAL);

    for (i = 0; i <= lvl; i++)
	pos->pos_arity[i] = numDownLinks[lvl - i];

    pos->pos_nodes  = numNodes;
    pos->pos_levels = lvl + 1;
    pos->pos_nodeid = nodeId;
    pos->pos_mode   = ELAN_POS_MODE_SWITCHED;

    return (0);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
