/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: elandev_linux.c,v 1.112.2.7 2005/09/07 14:35:03 mike Exp $"
/* 	$Source: /cvs/master/quadrics/elan3mod/elan3/os/elandev_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <linux/config.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/reboot.h>
#include <linux/notifier.h>

#include <linux/init.h>
#include <qsnet/module.h>
#include <linux/pci.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgalloc.h>
#include <asm/pgtable.h>

#include <elan/devinfo.h>
#include <elan/elanmod.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elanio.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/elansyscall.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
#error please use a 2.2 series kernel or newer
#endif

/* Minor numbers encoded as :
 *   [5:0]	device number
 *   [15:6] 	function number
 */
#define ELAN3_DEVICE_MASK          0x3F

#define ELAN3_MINOR_CONTROL      0
#define ELAN3_MINOR_MEM          1
#define ELAN3_MINOR_USER	 2
#define ELAN3_MINOR_SHIFT        6

#define ELAN3_DEVICE(inode)	(MINOR(inode->i_rdev) & ELAN3_DEVICE_MASK)
#define ELAN3_MINOR(inode)	(MINOR(inode->i_rdev) >> ELAN3_MINOR_SHIFT)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#	define SetPageReserved(page)	set_bit(PG_reserved, &(page)->flags)
#	define ClearPageReserved(page)	clear_bit(PG_reserved, &(page)->flags)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
typedef void irqreturn_t;
#endif
#       define IRQ_NONE
#       define IRQ_HANDLED
#       define IRQ_RETVAL(x)
#endif

#if defined(LINUX_SPARC) || defined(LINUX_PPC64)
#define __io_remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#define __remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#elif defined(NO_RMAP)
#define __io_remap_page_range(from,offset,size,prot)	io_remap_page_range(from,offset,size,prot)
#define __remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#else
#define __io_remap_page_range(from,offset,size,prot)	io_remap_page_range(vma,from,offset,size,prot)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
#define __remap_page_range(from,offset,size,prot)	remap_pfn_range(vma,from,(offset)>>PAGE_SHIFT,size,prot)
#else
#define __remap_page_range(from,offset,size,prot)	remap_page_range(vma,from,offset,size,prot)
#endif
#endif

/*
 * Function prototypes.
 */
static int	elanattach(int instance, struct pci_dev *pcidev);
static int	elandetach(int instance);

static int 	elan3_open (struct inode *inode, struct file *file);
static int	elan3_ioctl (struct inode *inode, struct file *file, 
			     unsigned int cmd, unsigned long arg);
static int	elan3_mmap (struct file *file, struct vm_area_struct *vm_area);
static int	elan3_release (struct inode *inode, struct file *file);

static int      elan3_reboot_event (struct notifier_block *self, unsigned long event, void *buffer);
static int      elan3_panic_event (struct notifier_block *self, unsigned long event, void *buffer);

static irqreturn_t InterruptHandlerWrapper(int irq, void *dev_id, struct pt_regs *regs);

static int	ConfigurePci(ELAN3_DEV *dev);
static int	ResetElan(ELAN3_DEV *dev, ioaddr_t intPalAddr);

static void     elan3_shutdown_devices(int panicing);

/*
 * Globals. 
 */
static ELAN3_DEV	*elan3_devices[ELAN3_MAX_CONTROLLER];
static int       NodeId = ELAN3_INVALID_NODE;
static int       NumNodes;
static int       DownLinks;
static int       RandomRoutingDisabled;
int              BackToBackMaster;
int              BackToBackSlave;
int              enable_sdram_writecombining;
int		 sdram_bank_limit;
extern int       LwpNice;

char *    elan_reg_rec_file [ELAN_REG_REC_MAX];
int       elan_reg_rec_line [ELAN_REG_REC_MAX];
long      elan_reg_rec_lbolt[ELAN_REG_REC_MAX];
int       elan_reg_rec_cpu  [ELAN_REG_REC_MAX];
E3_uint32 elan_reg_rec_reg  [ELAN_REG_REC_MAX];
int       elan_reg_rec_index;

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("Elan3 Device Driver");

MODULE_LICENSE("GPL");

module_param(NodeId,uint, 0);
module_param(NumNodes,uint, 0);
module_param(RandomRoutingDisabled,uint, 0);
module_param(DownLinks,uint, 0);
module_param(BackToBackMaster,uint, 0);
module_param(BackToBackSlave,uint, 0);
module_param(LwpNice, uint, 0);
module_param(elan3_debug, uint, 0);
module_param(elan3_debug_console, uint, 0);
module_param(elan3_debug_buffer, uint, 0);
module_param(elan3mmu_debug, uint, 0);
module_param(sdram_bank_limit, uint, 0);

/* elan3/os/context.c */
EXPORT_SYMBOL(elan3_alloc);
EXPORT_SYMBOL(elan3_attach);
EXPORT_SYMBOL(elan3_doattach);
EXPORT_SYMBOL(elan3_free);
EXPORT_SYMBOL(elan3_detach);
EXPORT_SYMBOL(elan3_dodetach);
EXPORT_SYMBOL(elan3_block_inputter);
EXPORT_SYMBOL(CheckCommandQueueFlushed);

/* elan3/os/sdram.c */
EXPORT_SYMBOL(elan3_sdram_alloc);
EXPORT_SYMBOL(elan3_sdram_free);
EXPORT_SYMBOL(elan3_sdram_to_phys);
EXPORT_SYMBOL(elan3_sdram_writeb);
EXPORT_SYMBOL(elan3_sdram_writew);
EXPORT_SYMBOL(elan3_sdram_writel);
EXPORT_SYMBOL(elan3_sdram_writeq);
EXPORT_SYMBOL(elan3_sdram_readb);
EXPORT_SYMBOL(elan3_sdram_readw);
EXPORT_SYMBOL(elan3_sdram_readl);
EXPORT_SYMBOL(elan3_sdram_readq);
EXPORT_SYMBOL(elan3_sdram_zerob_sdram);
EXPORT_SYMBOL(elan3_sdram_zerow_sdram);
EXPORT_SYMBOL(elan3_sdram_zerol_sdram);
EXPORT_SYMBOL(elan3_sdram_zeroq_sdram);
EXPORT_SYMBOL(elan3_sdram_copyb_to_sdram);
EXPORT_SYMBOL(elan3_sdram_copyw_to_sdram);
EXPORT_SYMBOL(elan3_sdram_copyl_to_sdram);
EXPORT_SYMBOL(elan3_sdram_copyq_to_sdram);
EXPORT_SYMBOL(elan3_sdram_copyb_from_sdram);
EXPORT_SYMBOL(elan3_sdram_copyw_from_sdram);
EXPORT_SYMBOL(elan3_sdram_copyl_from_sdram);
EXPORT_SYMBOL(elan3_sdram_copyq_from_sdram);

/* elan3/os/tproc.c */
EXPORT_SYMBOL(DeliverTProcTrap);
EXPORT_SYMBOL(HandleTProcTrap);
EXPORT_SYMBOL(SaveThreadToStack);

/* elan3/os/tprocinsts.c */
EXPORT_SYMBOL(RollThreadToClose);

/* elan3/os/iproc.c */
EXPORT_SYMBOL(InspectIProcTrap);
EXPORT_SYMBOL(IProcTrapString);
EXPORT_SYMBOL(SimulateUnlockQueue);

/* elan3/os/cproc.c */
EXPORT_SYMBOL(HandleCProcTrap);

/* elan3/os/route_table.c */
EXPORT_SYMBOL(GenerateRoute);
EXPORT_SYMBOL(LoadRoute);
EXPORT_SYMBOL(InvalidateRoute);
EXPORT_SYMBOL(ValidateRoute);
EXPORT_SYMBOL(ClearRoute);
EXPORT_SYMBOL(GenerateProbeRoute);
EXPORT_SYMBOL(GenerateCheckRoute);

/* elan3/os/elandev_generic.c */
EXPORT_SYMBOL(elan3_debug);
EXPORT_SYMBOL(QueueHaltOperation);
EXPORT_SYMBOL(ReleaseHaltOperations);
EXPORT_SYMBOL(ReserveHaltOperations);

/* elan3/vm/elan3mmu_generic.c */
EXPORT_SYMBOL(elan3mmu_pteload);
EXPORT_SYMBOL(elan3mmu_unload);
EXPORT_SYMBOL(elan3mmu_set_context_filter);
EXPORT_SYMBOL(elan3mmu_reserve);
EXPORT_SYMBOL(elan3mmu_attach);
EXPORT_SYMBOL(elan3mmu_detach);
EXPORT_SYMBOL(elan3mmu_release);
/* elan3/vm/elan3mmu_linux.c */
EXPORT_SYMBOL(elan3mmu_phys_to_pte);
EXPORT_SYMBOL(elan3mmu_kernel_invalid_pte);

/* elan3/os/elan3_debug.c */
EXPORT_SYMBOL(elan3_debugf);

/* elan3/os/minames.c */
EXPORT_SYMBOL(MiToName);

/* elan3/os/elandev_generic.c */
EXPORT_SYMBOL(MapDeviceRegister);
EXPORT_SYMBOL(UnmapDeviceRegister);

EXPORT_SYMBOL(elan_reg_rec_lbolt);
EXPORT_SYMBOL(elan_reg_rec_file);
EXPORT_SYMBOL(elan_reg_rec_index);
EXPORT_SYMBOL(elan_reg_rec_cpu);
EXPORT_SYMBOL(elan_reg_rec_reg);
EXPORT_SYMBOL(elan_reg_rec_line);

/*
 * Standard device entry points.
 */
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)

#include <linux/dump.h>

static int      elan3_dump_event (struct notifier_block *self, unsigned long event, void *buffer);

static struct notifier_block elan3_dump_notifier = 
{
    notifier_call:	elan3_dump_event,
    priority:		0,
};

static int
elan3_dump_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if ( event == DUMP_BEGIN )
	elan3_shutdown_devices (FALSE);

    return (NOTIFY_DONE);
}

#endif

static struct file_operations elan3_fops = {
        ioctl:   elan3_ioctl,		/* ioctl */
        mmap:    elan3_mmap,		/* mmap */
        open:    elan3_open,		/* open */
        release: elan3_release,		/* release */
};

static struct notifier_block elan3_reboot_notifier = 
{
    notifier_call:	elan3_reboot_event,
    priority:		0,
};

#if !defined(NO_PANIC_NOTIFIER)

static int elan3_panic_event (struct notifier_block *self, unsigned long event, void *buffer);

static struct notifier_block elan3_panic_notifier = 
{
    notifier_call:	elan3_panic_event,
    priority:		0,
};

static int
elan3_panic_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    elan3_shutdown_devices (TRUE);

    return (NOTIFY_DONE);
}

#endif /* !defined(NO_PANIC_NOTIFIER) */

ELAN3_DEV *
elan3_device (int instance)
{
	if (instance < 0 || instance >= ELAN3_MAX_CONTROLLER)
	    return ((ELAN3_DEV *) NULL);
	return elan3_devices[instance];
}
EXPORT_SYMBOL(elan3_device);

/*
 * Called at rmmod time.  elandetach() for each card + general cleanup.
 */
#ifdef MODULE
static void __exit elan3_exit(void)
{
	int i;

	printk("elan: preparing to remove module\n");

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	unregister_dump_notifier (&elan3_dump_notifier);
#endif
	unregister_reboot_notifier (&elan3_reboot_notifier);
#if !defined(NO_PANIC_NOTIFIER)
	notifier_chain_unregister (&panic_notifier_list, &elan3_panic_notifier);
#endif

	/* call elandetach() for each device configured. */
	for (i = 0; i < ELAN3_MAX_CONTROLLER; i++)
		if (elan3_devices[i] != NULL)
			elandetach(i);

	FinaliseNetworkErrorResolver();
	elan3mmu_fini();

	cookie_fini();
	unregister_chrdev(ELAN3_MAJOR, ELAN3_NAME);

	elan3_procfs_fini();

	printk("elan: module removed\n");
}

/*
 * Called at insmod time.  First we perform general driver initialization,
 * then call elanattach() for each card.
 */
#ifdef MODULE
static int __init elan3_init(void)
#else
__initfunc(int elan3_init(void))
#endif
{
	int e;
	int boards;
	struct pci_dev *dev;
	char revid;

	elan_reg_rec_index=0;
	{
	    int i;
	    for(i=0;i<ELAN_REG_REC_MAX;i++)
		elan_reg_rec_file[i] = NULL;
	}	

	/* register major/minor num */
	e = register_chrdev(ELAN3_MAJOR, ELAN3_NAME, &elan3_fops);
	if (e < 0)
		return e;

	elan3_procfs_init ();

	cookie_init();
	elan3mmu_init();
	InitialiseNetworkErrorResolver();

	/* call elanattach() for each device found on PCI */
	memset(elan3_devices, 0, sizeof(elan3_devices));
	boards = 0;
	for (dev = NULL; (dev = pci_find_device(PCI_VENDOR_ID_QUADRICS, PCI_DEVICE_ID_ELAN3, dev)) != NULL ;) 
	{
	    pci_read_config_byte (dev, PCI_REVISION_ID, &revid);

	    if (revid == PCI_REVISION_ID_ELAN3_REVA)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		printk ("elan at pci %s - RevA device not supported\n", dev->slot_name);
#else
		printk ("elan at pci %s - RevA device not supported\n", pci_name(dev));
#endif
	    else
	    {
		if (boards < ELAN3_MAX_CONTROLLER)
			/* Count successfully attached devices */ 
			boards += ((elanattach(boards, dev) == 0) ? 1 : 0);
		else
		{
		    printk ("elan: max controllers = %d\n", ELAN3_MAX_CONTROLLER);
		    break;
		}
	    }
	}
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
	register_dump_notifier (&elan3_dump_notifier);
#endif
	register_reboot_notifier (&elan3_reboot_notifier);
#if !defined(NO_PANIC_NOTIFIER)
	notifier_chain_register (&panic_notifier_list, &elan3_panic_notifier);
#endif

	return 0;
}

/* Declare the module init and exit functions */
module_init(elan3_init);
module_exit(elan3_exit);

#endif

static void
elan3_shutdown_devices(int panicing)
{
    ELAN3_DEV *dev;
    unsigned long flags;
    register int i;

    local_irq_save (flags);
    for (i = 0; i < ELAN3_MAX_CONTROLLER; i++)
    {
	if ((dev = elan3_devices[i]) != NULL)
	{
	    if (! panicing) spin_lock (&dev->IntrLock);
	    
	    printk(KERN_INFO "elan%d: forcing link into reset\n", dev->Instance);

	    /*
	     * We're going to set the link into boundary scan mode,  so firstly
	     * set the inputters to discard everything.
	     */
	    if (dev->DiscardAllCount++ == 0)
		SetSchedStatusRegister (dev, read_reg32 (dev, Exts.InterruptReg), NULL);

	    dev->LinkShutdown = 1;
	    
	    /*
	     * Now disable the error interrupts
	     */
	    DISABLE_INT_MASK (dev, INT_ErrorInterrupts);
	    
	    /*
	     * And set the link into boundary scan mode, and drive
	     * a reset token onto the link.
	     */
	    SET_SCHED_LINK_VALUE (dev, 1, LinkResetToken);

	    if (! panicing) spin_unlock (&dev->IntrLock);
	}
    }
    local_irq_restore (flags);
}

static int
elan3_reboot_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if (! (event == SYS_RESTART || event == SYS_HALT || event == SYS_POWER_OFF))
	return (NOTIFY_DONE);

    elan3_shutdown_devices (FALSE);

    return (NOTIFY_DONE);
}

#include <elan3/elan3ops.h>
/*
 * Called by init_module() for each card discovered on PCI.
 */
static int
elanattach(int instance, struct pci_dev *pcidev)
{
	ELAN3_DEV *dev;
	int ramSize;
	int level;
	ioaddr_t sdramAddr, cmdPortAddr, intPalAddr;
	DeviceMappingHandle handle;

	printk("elan%d: attach, irq=%d\n", instance, pcidev->irq);

	/*
	 * Allocate the ELAN3_DEV structure.
	 */
	KMEM_ZALLOC(dev, ELAN3_DEV *, sizeof(ELAN3_DEV), TRUE);
	if (dev == NULL) {
		printk ("elan%d: KMEM_ALLOC failed\n", instance);
		return (-ENOMEM);
	}
	elan3_devices[instance] = dev;
	dev->Osdep.pci = pcidev;

	dev->Instance = instance;

	/* Initialise the device information */
	pci_read_config_word (pcidev, PCI_VENDOR_ID,   &dev->Devinfo.dev_vendor_id);
	pci_read_config_word (pcidev, PCI_DEVICE_ID,   &dev->Devinfo.dev_device_id);
	pci_read_config_byte (pcidev, PCI_REVISION_ID, &dev->Devinfo.dev_revision_id);

	dev->Devinfo.dev_instance             = instance;
	dev->Devinfo.dev_rail                 = instance;
	dev->Devinfo.dev_driver_version       = 0;
	dev->Devinfo.dev_num_down_links_value = DownLinks;

	dev->Position.pos_mode                = ELAN_POS_UNKNOWN;
	dev->Position.pos_random_disabled     = RandomRoutingDisabled;
	
	/*
	 * Set up PCI config regs.
	 */
	if (ConfigurePci(dev) != ESUCCESS)
	    goto fail0;

	/*
	 * Determine the PFnums of the SDRAM and command port
	 */
	if (MapDeviceRegister(dev, ELAN3_BAR_SDRAM, &sdramAddr, 0, PAGESIZE, &handle) != ESUCCESS)
	    goto fail1;

	DeviceRegisterSize(dev, ELAN3_BAR_SDRAM, &ramSize);
	
	dev->SdramPhysMask = ~((physaddr_t) ramSize - 1);
	dev->SdramPhysBase = kmem_to_phys((void *) sdramAddr);

	UnmapDeviceRegister (dev, &handle);

#if defined(LINUX_ALPHA)
	/*
	 * consider a physical address to be on the same pci bus
	 * as us if it's physical address is "close" to our sdram
	 * physical address.
	 * this is almost certainly incorrect for large memory (> 2Gb)
	 * i386 machines - and is only correct for alpha for 32 bit
	 * base address registers.
	 *
	 * Modified this to match the Tru64 driver value;
	 * i.e. PciPhysMask = 0xfffffffffffc0000
	 */
#  define PCI_ADDR_MASK (0x7FFFFFFFl)

	dev->PciPhysMask = ~PCI_ADDR_MASK;
	dev->PciPhysBase = dev->SdramPhysBase & dev->PciPhysMask;
#endif
	/*
	 * Now reset the elan chip.
	 */
	if (MapDeviceRegister(dev, ELAN3_BAR_REGISTERS, &dev->RegPtr, 0, 0, &dev->RegHandle) != ESUCCESS)
	    goto fail1;

	if (MapDeviceRegister(dev, ELAN3_BAR_EBUS, &intPalAddr, ELAN3_EBUS_INTPAL_OFFSET, PAGESIZE,
			      &handle) != ESUCCESS)
	    goto fail2;

	ResetElan(dev, intPalAddr);	

	UnmapDeviceRegister (dev, &handle);

	/* 
	 * Initialise the device mutex's which must be accessible from the 
	 * interrupt handler.  
	 */
	kcondvar_init (&dev->IntrWait);
	spin_lock_init (&dev->IntrLock);
	spin_lock_init (&dev->TlbLock);
	spin_lock_init (&dev->CProcLock);
	spin_lock_init (&dev->FreeHaltLock);
	for(level=0; level<4; level++)
	    spin_lock_init (&dev->Level[level].PtblLock);
	spin_lock_init (&dev->PtblGroupLock);

	/*
	 * Add the interrupt handler,  
	 */
	if (request_irq(dev->Osdep.pci->irq, InterruptHandlerWrapper, 
	    SA_SHIRQ, "elan3", dev) != 0) {
		printk ("elan%d: request_irq failed\n", instance);
		goto fail3;
	}

	if (MapDeviceRegister(dev, ELAN3_BAR_COMMAND_PORT, &cmdPortAddr, 0, PAGESIZE, &handle) != ESUCCESS)
	    goto fail4;
	
	if (InitialiseElan(dev, cmdPortAddr) == EFAIL) {
		printk ("elan%d: InitialiseElan failed\n", instance);
		UnmapDeviceRegister (dev, &handle);
		goto fail4;
	}
	UnmapDeviceRegister (dev, &handle);

	/* If our nodeid is defined, then set it now */
	if (NodeId != ELAN3_INVALID_NODE && ComputePosition (&dev->Position, NodeId, NumNodes, DownLinks) == 0)
	{
	    if (RandomRoutingDisabled & ((1 << (dev->Position.pos_levels-1))-1))
		printk ("elan%d: NodeId=%d NodeLevel=%d NumNodes=%d (random routing disabled 0x%x)\n", 
			dev->Instance, dev->Position.pos_nodeid, dev->Position.pos_levels, dev->Position.pos_nodes, RandomRoutingDisabled);
	    else
		printk ("elan%d: NodeId=%d NodeLevel=%d NumNodes=%d (random routing ok)\n",
			dev->Instance, dev->Position.pos_nodeid, dev->Position.pos_levels, dev->Position.pos_nodes);
	}

	if (BackToBackMaster || BackToBackSlave)
	{
	    dev->Position.pos_mode     = ELAN_POS_MODE_BACKTOBACK;
	    dev->Position.pos_nodeid   = (BackToBackMaster == 0);
	    dev->Position.pos_nodes    = 2;
	    dev->Position.pos_levels   = 1;
	    dev->Position.pos_arity[0] = 2;

	    printk ("elan%d: back-to-back %s - elan node %d\n", dev->Instance,
		    BackToBackMaster ? "master" : "slave", dev->Position.pos_nodeid);
	}

	elan3_procfs_device_init (dev);
	
	/* Success */
	return (0);

fail4:
	free_irq(dev->Osdep.pci->irq, dev);

fail3:
	kcondvar_destroy (&dev->IntrWait);
	spin_lock_destroy (&dev->IntrLock);
	spin_lock_destroy (&dev->InfoLock);
	spin_lock_destroy (&dev->TlbLock);
	spin_lock_destroy (&dev->CProcLock);
	spin_lock_destroy (&dev->FreeHaltLock);
	spin_lock_destroy (&dev->Level1PtblLock);
	spin_lock_destroy (&dev->Level2PtblLock);
	spin_lock_destroy (&dev->Level3PtblLock);
	spin_lock_destroy (&dev->PtblGroupLock);

fail2:
	UnmapDeviceRegister (dev, &dev->RegHandle);

fail1:
	pci_disable_device (dev->Osdep.pci);
fail0:
	KMEM_FREE(dev, sizeof(ELAN3_DEV));

	elan3_devices[instance] = NULL;
	
	/* Failure */
	return (-ENODEV);
}

/*
 * Called by elan3_exit() for each board found on PCI.
 */
static int
elandetach(int instance)
{
	ELAN3_DEV *dev = elan3_devices[instance];

	printk("elan%d: detach\n", instance);

	elan3_procfs_device_fini (dev);

	FinaliseElan (dev);

	UnmapDeviceRegister (dev, &dev->RegHandle);

	free_irq(dev->Osdep.pci->irq, dev);

	pci_disable_device(dev->Osdep.pci);

	kcondvar_destroy (&dev->IntrWait);
	spin_lock_destroy (&dev->IntrLock);
	spin_lock_destroy (&dev->InfoLock);
	spin_lock_destroy (&dev->TlbLock);
	spin_lock_destroy (&dev->CProcLock);
	spin_lock_destroy (&dev->FreeHaltLock);
	spin_lock_destroy (&dev->Level1PtblLock);
	spin_lock_destroy (&dev->Level2PtblLock);
	spin_lock_destroy (&dev->Level3PtblLock);
	spin_lock_destroy (&dev->PtblGroupLock);

	KMEM_FREE(dev, sizeof(ELAN3_DEV));
	elan3_devices[instance] = NULL;	

	return 0;
}

/*
 * generic ioctls - available on control and user devices.
 */

static int
device_stats_ioctl (ELAN3_DEV *dev, unsigned long arg)
{
    ELAN3IO_STATS_STRUCT *args;

    KMEM_ALLOC(args, ELAN3IO_STATS_STRUCT *, sizeof(ELAN3IO_STATS_STRUCT), TRUE);
	
    if (args == NULL)
	return (-ENOMEM);

    if (copy_from_user (args, (void *) arg, sizeof (ELAN3IO_STATS_STRUCT)))
    {
	KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	return (-EFAULT);
    }

    switch (args->which)
    {
    case ELAN3_SYS_STATS_DEVICE:
	if (copy_to_user (args->ptr, &dev->Stats, sizeof (ELAN3_STATS)))
	{
	    KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	    return (-EFAULT);
	}
	KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	return (0);

    case ELAN3_SYS_STATS_MMU:
	if (copy_to_user (args->ptr, &elan3mmu_global_stats, sizeof (ELAN3MMU_GLOBAL_STATS)))
	{
	    KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	    return (-EFAULT);
	}
	KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	return (0);
	    
    default:
	KMEM_FREE(args, sizeof(ELAN3IO_STATS_STRUCT));
	return (-EINVAL);
    }
}

/*
 * /dev/elan3/controlX - control device
 *
 */

typedef struct control_private
{
    u_int		pr_boundary_scan;
} CONTROL_PRIVATE;

static int
control_open (struct inode *inode, struct file *file)
{
    CONTROL_PRIVATE *pr;

    KMEM_ALLOC(pr, CONTROL_PRIVATE *, sizeof (CONTROL_PRIVATE), TRUE);

    if (pr == NULL)
	return (-ENOMEM);

    pr->pr_boundary_scan = 0;
    
    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;

    return (0);
}

static int
control_release (struct inode *inode, struct file *file)
{
    ELAN3_DEV        *dev = elan3_devices[ELAN3_DEVICE(inode)];
    CONTROL_PRIVATE *pr  = (CONTROL_PRIVATE *) file->private_data;

    if (pr->pr_boundary_scan)
	ClearLinkBoundaryScan(dev);

    KMEM_FREE (pr, sizeof(CONTROL_PRIVATE));

    MOD_DEC_USE_COUNT;
    return (0);
}

static int
control_ioctl (struct inode *inode, struct file *file, 
	       unsigned int cmd, unsigned long arg)
{
    ELAN3_DEV        *dev = elan3_devices[ELAN3_DEVICE(inode)];
    CONTROL_PRIVATE *pr  = (CONTROL_PRIVATE *) file->private_data;
    int	             res;

    switch (cmd) 
    {
    case ELAN3IO_SET_BOUNDARY_SCAN:
	if (SetLinkBoundaryScan (dev) == 0)
	    pr->pr_boundary_scan = 1;
	return (0);

    case ELAN3IO_CLEAR_BOUNDARY_SCAN:
	if (pr->pr_boundary_scan == 0)
	    return (-EINVAL);

	pr->pr_boundary_scan = 0;

	ClearLinkBoundaryScan (dev);
	return (0);

    case ELAN3IO_READ_LINKVAL:
    {
	E3_uint32 val;

	if (pr->pr_boundary_scan == 0)
	    return (-EINVAL);

	if (copy_from_user(&val, (E3_uint32 *)arg, sizeof(E3_uint32)))
	    return (-EFAULT);

	val = ReadBoundaryScanValue (dev, val);

	if (copy_to_user((E3_uint32 *)arg, &val, sizeof(E3_uint32)))
	    return (-EFAULT);
	return (0);
    }
	
    case ELAN3IO_WRITE_LINKVAL:
    {
	E3_uint32 val;

	if (pr->pr_boundary_scan == 0)
	    return (-EINVAL);

	if (copy_from_user(&val, (E3_uint32 *)arg, sizeof(E3_uint32)))
	    return (-EFAULT);

	val = WriteBoundaryScanValue (dev, val);

	if (copy_to_user((E3_uint32 *)arg, &val, sizeof(E3_uint32)))
	    return (-EFAULT);
	
	return (0);
    }

    case ELAN3IO_SET_POSITION:
    {
	ELAN3IO_SET_POSITION_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_SET_POSITION_STRUCT)))
	    return (-EFAULT);
	
	if (ComputePosition (&dev->Position, args.nodeId, args.numNodes, dev->Devinfo.dev_num_down_links_value) != 0)
	    return (-EINVAL);

	return (0);
    }

    case ELAN3IO_SET_DEBUG:
    {
	ELAN3IO_SET_DEBUG_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_SET_DEBUG_STRUCT)))
	    return (-EFAULT);

	if (! strcmp (args.what, "elan3_debug"))
	    elan3_debug = args.value;
	else if (! strcmp (args.what, "elan3_debug_console"))
	    elan3_debug_console = args.value;
	else if (! strcmp (args.what, "elan3_debug_buffer"))
	    elan3_debug_buffer = args.value;
	else if (! strcmp (args.what, "elan3_debug_ignore_dev"))
	    elan3_debug_ignore_dev = args.value;
	else if (! strcmp (args.what, "elan3_debug_ignore_ctxt"))
	    elan3_debug_ignore_ctxt = args.value;
	else if (! strcmp (args.what, "elan3mmu_debug"))
	    elan3mmu_debug = args.value;
	
	return (0);
    }

    case ELAN3IO_NETERR_SERVER:
    {
	ELAN3IO_NETERR_SERVER_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_NETERR_SERVER_STRUCT)))
	    return (-EFAULT);
	
	res = AddNeterrServerSyscall (args.elanid, args.addr, args.name, NULL);
	return (set_errno (res));
    }
    
    case ELAN3IO_NETERR_FIXUP:
    {
	NETERR_MSG *msg;

	KMEM_ALLOC(msg, NETERR_MSG *, sizeof (NETERR_MSG), TRUE);

	if (msg == NULL)
	    return (set_errno (ENOMEM));
	
	if (copy_from_user (msg, (void *) arg, sizeof (NETERR_MSG)))
	    res = EFAULT;
	else
	    res = ExecuteNetworkErrorFixup (msg);

	KMEM_FREE (msg, sizeof (NETERR_MSG));
	return (set_errno (res));
    }

    case ELAN3IO_STATS:
	return (device_stats_ioctl (dev, arg));

    case ELAN3IO_GET_DEVINFO:
    {
	if (copy_to_user ((void *) arg, &dev->Devinfo, sizeof (ELAN_DEVINFO)))
	    return (-EFAULT);
	return (0);
    }

    case ELAN3IO_GET_POSITION:
    {
	if (copy_to_user ((void *) arg, &dev->Position, sizeof (ELAN_POSITION)))
	    return (-EFAULT);
	return (0);
    }
    default:
	return (-EINVAL);
    }
}

static int
control_mmap (struct file *file, struct vm_area_struct *vma)
{
    ELAN3_DEV 	       *dev   = elan3_devices[ELAN3_DEVICE(file->f_dentry->d_inode)];
    int 		space = OFF_TO_SPACE(vma->vm_pgoff << PAGE_SHIFT);
    int 		off   = OFF_TO_OFFSET(vma->vm_pgoff << PAGE_SHIFT);
    int 		size;
    ioaddr_t 	        addr;
    DeviceMappingHandle handle;
    physaddr_t		phys;
    
    if (space < ELAN3_BAR_SDRAM || space > ELAN3_BAR_EBUS)
	return (-EINVAL);

    if (off < 0 || DeviceRegisterSize (dev, space, &size) != ESUCCESS || off > size)
	return (-EINVAL);

    if (MapDeviceRegister(dev, space, &addr, off, PAGESIZE, &handle) != ESUCCESS)
	return (-EINVAL);

    phys = kmem_to_phys((caddr_t) addr);
    UnmapDeviceRegister(dev, &handle);

    if (__remap_page_range(vma->vm_start, phys, vma->vm_end - vma->vm_start, vma->vm_page_prot))
	return (-EAGAIN);

    return (0);
}

/*
 * /dev/elan3/sdramX - sdram access device
 */
typedef struct mem_page
{
    struct mem_page *pg_next;
    sdramaddr_t      pg_addr;
    u_long	     pg_pgoff;
    u_int	     pg_ref;
} MEM_PAGE;

#define MEM_HASH_SIZE	32
#define MEM_HASH(pgoff)	((pgoff) & (MEM_HASH_SIZE-1))

typedef struct mem_private
{
    ELAN3_DEV   *pr_dev;
    MEM_PAGE   *pr_pages[MEM_HASH_SIZE];
    spinlock_t  pr_lock;
} MEM_PRIVATE;

static void 
mem_freepage (MEM_PRIVATE *pr, MEM_PAGE *pg)
{
    PRINTF (DBG_DEVICE, DBG_SEG, "mem_freepage: pr=%p pgoff=%lx pg=%p ref=%d\n", pr, pg->pg_pgoff, pg, pg->pg_ref);

    elan3_sdram_free (pr->pr_dev, pg->pg_addr, PAGE_SIZE);
    KMEM_FREE (pg, sizeof(MEM_PAGE));
}

static MEM_PAGE *
mem_getpage (MEM_PRIVATE *pr, u_long pgoff, virtaddr_t addr)
{
    int       hashval = MEM_HASH (pgoff);
    MEM_PAGE *npg = NULL;
    MEM_PAGE *pg;

    PRINTF (DBG_DEVICE, DBG_SEG, "mem_getpage: pr=%p pgoff=%lx addr=%lx\n", pr, pgoff, addr);
    
 again:
    spin_lock (&pr->pr_lock);
    for (pg = pr->pr_pages[hashval]; pg; pg = pg->pg_next)
	if (pg->pg_pgoff == pgoff)
	    break;
    
    if (pg != NULL)
    {
	PRINTF (DBG_DEVICE, DBG_SEG, "mem_getpage: pr=%p pgoff=%lx addr=%lx -> found %p addr=%lx\n", pr, pgoff, addr, pg, pg->pg_addr);

	pg->pg_ref++;
	spin_unlock (&pr->pr_lock);

	if (npg != NULL)					/* we'd raced and someone else had created */
	    mem_freepage (pr, npg);				/* this page - so free of our new one*/
	return (pg);
    }
    
    if (npg != NULL)						/* didn't find the page, so inset the */
    {								/* new one we've just created */
	npg->pg_next = pr->pr_pages[hashval];
	pr->pr_pages[hashval] = npg;
	
	spin_unlock (&pr->pr_lock);
	return (npg);
    }
    
    spin_unlock (&pr->pr_lock);					/* drop spinlock before creating a new page */
    
    KMEM_ALLOC(npg, MEM_PAGE *, sizeof (MEM_PAGE), TRUE);

    if (npg == NULL)
	return (NULL);

    if ((npg->pg_addr = elan3_sdram_alloc (pr->pr_dev, PAGE_SIZE)) == 0)
    {
	KMEM_FREE (npg, sizeof (MEM_PAGE));
	return (NULL);
    }

    /* zero the page before returning it to the user */
    elan3_sdram_zeroq_sdram (pr->pr_dev, npg->pg_addr, PAGE_SIZE);
    
    npg->pg_pgoff = pgoff;
    npg->pg_ref   = 1;
    
    /* created a new page - so have to rescan before inserting it */
    goto again;
}

static void
mem_droppage (MEM_PRIVATE *pr, u_long pgoff, int dontfree)
{
    MEM_PAGE **ppg;
    MEM_PAGE  *pg;

    spin_lock (&pr->pr_lock);
    for (ppg = &pr->pr_pages[MEM_HASH(pgoff)]; *ppg; ppg = &(*ppg)->pg_next)
	if ((*ppg)->pg_pgoff == pgoff)
	    break;

    pg = *ppg;

    ASSERT (*ppg != NULL);
    
    PRINTF (DBG_DEVICE, DBG_SEG, "mem_droppage: pr=%p pgoff=%lx pg=%p ref=%d dontfree=%d\n", pr, pgoff, (*ppg), (*ppg)->pg_ref, dontfree);

    if (--pg->pg_ref == 0 && !dontfree)
    {
	*ppg = pg->pg_next;

	mem_freepage (pr, pg);
    }

    spin_unlock (&pr->pr_lock);
}

static int
mem_open (struct inode *inode, struct file *file)
{
    ELAN3_DEV    *dev = elan3_devices[ELAN3_DEVICE(inode)];
    MEM_PRIVATE *pr;
    register int i;

    KMEM_ALLOC(pr, MEM_PRIVATE *, sizeof (MEM_PRIVATE), TRUE);

    if (pr == NULL)
	return (-ENOMEM);

    spin_lock_init (&pr->pr_lock);
    pr->pr_dev = dev;
    for (i = 0; i < MEM_HASH_SIZE; i++)
	pr->pr_pages[i] = NULL;

    file->private_data = (void *) pr;
    
    MOD_INC_USE_COUNT;
    return (0);
}

static int
mem_release (struct inode *node, struct file *file)
{
    MEM_PRIVATE *pr = (MEM_PRIVATE *) file->private_data;
    MEM_PAGE    *pg, *next;
    int          i;

    /* free off any pages that we'd allocated */
    spin_lock (&pr->pr_lock);
    for (i = 0; i < MEM_HASH_SIZE; i++)
    {
	for (pg = pr->pr_pages[i]; pg; pg = next)
	{
	    next = pg->pg_next;
	    mem_freepage (pr, pg);
	}
    }
    spin_unlock (&pr->pr_lock);

    KMEM_FREE (pr, sizeof (MEM_PRIVATE));

    MOD_DEC_USE_COUNT;
    return (0);
}

static int
mem_ioctl (struct inode *inode, struct file *file, 
		  unsigned int cmd, unsigned long arg)
{
    return (-EINVAL);
}

static void mem_vma_open(struct vm_area_struct *vma)
{
    MEM_PRIVATE   *pr = (MEM_PRIVATE *) vma->vm_private_data;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_DEVICE, DBG_SEG, "mem_vma_open: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    preemptable_start {
	for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++) {
	    mem_getpage (pr, pgoff, addr);
	    preemptable_check();
	}
    } preemptable_end;
}

static void mem_vma_close(struct vm_area_struct *vma)
{
    MEM_PRIVATE  *pr  = (MEM_PRIVATE *) vma->vm_private_data;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_DEVICE, DBG_SEG, "mem_vma_close: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    /* NOTE: the call to close may not have the same vm_start/vm_end values as 
     *       were passed into mmap()/open() - since if an partial unmap had occured
     *       then the vma could have been shrunk or even split.
     *
     *       if a the vma is split then an vma_open() will be called for the top
     *       portion - thus causing the reference counts to become incorrect.
     *
     * We drop the reference to any pages we're notified about - so they get freed
     * earlier than when the device is finally released.
     */
    for (pgoff = vma->vm_pgoff, addr = vma->vm_start; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	mem_droppage (pr, pgoff, 0);
}

static struct vm_operations_struct mem_vm_ops = {
    open:		mem_vma_open,
    close:		mem_vma_close,
};

static int
mem_mmap (struct file *file, struct vm_area_struct *vma)
{
    MEM_PRIVATE  *pr = (MEM_PRIVATE *) file->private_data;
    MEM_PAGE     *pg;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_DEVICE, DBG_SEG, "mem_mmap: vm_mm=%p start=%lx end=%lx pgoff=%lx prot=%llx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, (long long)vma->vm_page_prot.pgprot , file);

    preemptable_start {
	for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	{
	    if ((pg = mem_getpage (pr, pgoff, addr)) == NULL)
		goto failed;
	    
#ifdef LINUX_SPARC
	    pgprot_val(vma->vm_page_prot) &= ~(_PAGE_CACHE);
	    pgprot_val(vma->vm_page_prot) |= _PAGE_IE;
#elif defined(pgprot_noncached)
	    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
	    
#if defined(__ia64__)
	    if (enable_sdram_writecombining)
		vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
#endif
	    PRINTF (DBG_DEVICE, DBG_SEG, "mem_mmap: addr %lx -> pg=%p addr=%lx phys=%llx flags=%lx prot=%llx\n",
		    addr, pg, pg->pg_addr, (long long) elan3_sdram_to_phys (pr->pr_dev, pg->pg_addr), vma->vm_flags,  (long long)vma->vm_page_prot.pgprot);
	    
    	    if (__remap_page_range(addr, elan3_sdram_to_phys (pr->pr_dev, pg->pg_addr), PAGE_SIZE, vma->vm_page_prot))
	    {
		mem_droppage (pr, pgoff, 0);			/* drop our reference to this page */
		goto failed;
	    }

	    preemptable_check();
	}
    } preemptable_end;

    /* Don't try to swap out Elan SDRAM pages.. */
    vma->vm_flags |= VM_RESERVED;
    
    /*
     * Don't dump SDRAM pages to a core file 
     * (Pity I would really like to do this but it crashes in elf_core_dump() as
     * it can only handle pages that are in the mem_map area (addy 11/01/2002))
     */
    vma->vm_flags |= VM_IO;

    vma->vm_ops          = &mem_vm_ops;
    vma->vm_file         = file;
    vma->vm_private_data = (void *) pr;

    return (0);

 failed:
    PRINTF (DBG_DEVICE, DBG_SEG, "mem_mmap: failed\n");

    /* free of any pages we've already allocated/referenced */
    while (pgoff-- > vma->vm_pgoff)
	mem_droppage (pr, pgoff, 0);

    return (-ENOMEM);
}

#if !defined(NO_PTRACK) && defined(IOPROC_PATCH_APPLIED)

#include <linux/ptrack.h>

/*
 * /dev/elan3/userX - control device
 *
 * "user_private" can be referenced from a number of places
 *   1) the "file" structure.
 *   2) the "mm" ioproc ops
 *   3) the "mmap" of the command port.
 *
 */
typedef struct user_private
{
    spinlock_t        pr_lock;
    atomic_t 	      pr_mappings;
    atomic_t          pr_ref;
    ELAN3_CTXT        *pr_ctxt;
    struct mm_struct *pr_mm;
    struct ioproc_ops pr_ioproc;
} USER_PRIVATE;

static void
user_free (USER_PRIVATE *pr)
{
    /* Have to unreserve the FlagPage or else we leak memory like a sieve! */
    ClearPageReserved(pte_page(*find_pte_kernel((unsigned long) pr->pr_ctxt->FlagPage)));

    elan3_detach(pr->pr_ctxt);
    elan3_free (pr->pr_ctxt);

    KMEM_FREE (pr, sizeof(USER_PRIVATE));

    MOD_DEC_USE_COUNT;
}

static void
user_ioproc_release (void *arg, struct mm_struct *mm)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF3 (pr->pr_ctxt, DBG_SEG, "user_ioproc_release: ctxt=%p pr=%p ref=%d\n",
	     pr->pr_ctxt, pr, atomic_read (&pr->pr_ref));

    elan3mmu_pte_ctxt_unload (pr->pr_ctxt->Elan3mmu);

    pr->pr_mm = NULL;

    if (atomic_dec_and_test (&pr->pr_ref))
	user_free (pr);
}

/*
 * On 2.4 kernels we get passed a mm_struct, whereas on 2.6 kernels
 * we get the vma which is more usefull
 */
#if defined(IOPROC_MM_STRUCT_ARG)
static void
user_ioproc_sync_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_sync_range: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, mm, (caddr_t) start, end-start);
}

static void
user_ioproc_invalidate_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_invalidate_range: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, mm, (caddr_t) start, end-start);
}

static void
user_ioproc_update_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    ASSERT(start <= end && ((start & PAGEOFFSET) == 0) && ((end & PAGEOFFSET) == 0));

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_update_range: start=%lx end=%lx\n", start, end);

    elan3mmu_pte_range_update (pr->pr_ctxt->Elan3mmu, mm,(caddr_t) start, end-start);
}

static void
user_ioproc_change_protection (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end, pgprot_t newprot)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_change_protection: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, mm, (caddr_t) start, end-start);
}

#else

static void
user_ioproc_sync_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_sync_range: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) start, end-start);
}

static void
user_ioproc_invalidate_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_invalidate_range: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) start, end-start);
}

static void
user_ioproc_update_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    ASSERT(start <= end && ((start & PAGEOFFSET) == 0) && ((end & PAGEOFFSET) == 0));

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_update_range: start=%lx end=%lx\n", start, end);

    elan3mmu_pte_range_update (pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) start, end-start);
}

static void
user_ioproc_change_protection (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end, pgprot_t newprot)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF2 (pr->pr_ctxt, DBG_SEG, "user_ioproc_change_protection: start=%lx end=%lx\n", start, end);

    ASSERT(start <= end);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) start, end-start);
}
#endif /* defined(IOPROC_NO_VMA_RANGE) */

static void
user_ioproc_sync_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF1 (pr->pr_ctxt, DBG_SEG, "user_ioproc_sync_page: addr=%lx\n", addr);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) (addr & PAGE_MASK), PAGE_SIZE);
}

static void
user_ioproc_invalidate_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;
    
    PRINTF1 (pr->pr_ctxt, DBG_SEG, "user_ioproc_invalidate_page: addr=%lx\n", addr);

    elan3mmu_pte_range_unload(pr->pr_ctxt->Elan3mmu, vma->vm_mm, (caddr_t) (addr & PAGE_MASK), PAGE_SIZE);
}

static void
user_ioproc_update_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF1 (pr->pr_ctxt, DBG_SEG, "user_ioproc_update_page: addr=%lx\n", addr);

    elan3mmu_pte_range_update (pr->pr_ctxt->Elan3mmu,vma->vm_mm, (caddr_t) (addr & PAGE_MASK), PAGE_SIZE);
}

int
user_ptrack_handler (void *arg, int phase, struct task_struct *child)
{
    USER_PRIVATE *pr   = (USER_PRIVATE *) arg;
    ELAN3_CTXT    *ctxt = pr->pr_ctxt;

    PRINTF5 (pr->pr_ctxt, DBG_FN, "user_ptrack_handler: ctxt=%p pr=%p ref=%d phase %d mm->ref %d\n", 
	     pr->pr_ctxt, pr, atomic_read (&pr->pr_ref), phase, atomic_read (&current->mm->mm_count));

    if (phase == PTRACK_PHASE_EXIT)
    {
	/* this will force the helper thread to exit */
	elan3_swapout (ctxt, CTXT_EXITING);
	
	if (atomic_dec_and_test (&pr->pr_ref))
	    user_free (pr);
    }  
    return PTRACK_FINISHED;
}

static int
user_open (struct inode *inode, struct file *file)
{
    ELAN3_DEV     *dev = elan3_devices[ELAN3_DEVICE(inode)];
    USER_PRIVATE *pr;
    ELAN3_CTXT    *ctxt;

    if (dev == NULL)
	return (-ENXIO);

    KMEM_ALLOC(pr, USER_PRIVATE *, sizeof (USER_PRIVATE), TRUE);

    if (pr == NULL)
	return (-ENOMEM);
    
    if ((ctxt = elan3_alloc (dev, 0)) == NULL)
    {
	KMEM_FREE (pr, sizeof (USER_PRIVATE));
	return (-ENOMEM);
    }

    if (sys_init (ctxt) == NULL)
    {
	elan3_detach(ctxt);
	elan3_free (ctxt);
	KMEM_FREE (pr, sizeof (USER_PRIVATE));
	return (-ENOMEM);
    }

    /* initialise refcnt to 3 - one for "file", one for XA handler, one for the ioproc ops */
    atomic_set (&pr->pr_ref, 3);

    atomic_set (&pr->pr_mappings, 0);
    spin_lock_init (&pr->pr_lock);

    pr->pr_ctxt = ctxt;
    pr->pr_mm   = current->mm;

    /* register an ptrack handler to force the helper thread to exit when we do */
    if (ptrack_register (user_ptrack_handler, pr) < 0)
    {
	elan3_detach(ctxt);
	elan3_free (ctxt);
	KMEM_FREE (pr, sizeof (USER_PRIVATE));
	return (-ENOMEM);
    }

    /* register a ioproc callback to notify us of translation changes */
    
    pr->pr_ioproc.arg               = (void *) pr;
    pr->pr_ioproc.release           = user_ioproc_release;
    pr->pr_ioproc.sync_range        = user_ioproc_sync_range;
    pr->pr_ioproc.invalidate_range  = user_ioproc_invalidate_range;
    pr->pr_ioproc.update_range      = user_ioproc_update_range;
    pr->pr_ioproc.change_protection = user_ioproc_change_protection;
    pr->pr_ioproc.sync_page         = user_ioproc_sync_page;
    pr->pr_ioproc.invalidate_page   = user_ioproc_invalidate_page;
    pr->pr_ioproc.update_page       = user_ioproc_update_page;
    
    spin_lock (&current->mm->page_table_lock);
    ioproc_register_ops (current->mm, &pr->pr_ioproc);
    spin_unlock (&current->mm->page_table_lock);

    file->private_data = (void *) pr;

    PRINTF2 (pr->pr_ctxt, DBG_FN, "user_open: done ctxt=%p pr=%p\n", ctxt, pr);

    MOD_INC_USE_COUNT;
    return (0);
}

static int
user_release (struct inode *inode, struct file *file)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) file->private_data;
    
    PRINTF3 (pr->pr_ctxt, DBG_FN, "user_release: ctxt=%p pr=%p ref=%d\n", pr->pr_ctxt, pr,
	     atomic_read (&pr->pr_ref));

    if (atomic_dec_and_test (&pr->pr_ref))
	user_free (pr);

    return (0);
}

static int
user_ioctl (struct inode *inode, struct file *file, 
	    unsigned int cmd, unsigned long arg)
{
    USER_PRIVATE *pr   = (USER_PRIVATE *) file->private_data;
    ELAN3_CTXT    *ctxt = pr->pr_ctxt;
    SYS_CTXT     *sctx = (SYS_CTXT *) ctxt->Private;
    int           res  = 0;

    if (current->mm != pr->pr_mm)
	return (-EINVAL);
    
    PRINTF4 (ctxt, DBG_FN, "user_ioctl: ctxt=%p cmd=%x(%d) arg=%lx\n", ctxt, cmd, _IOC_NR(cmd), arg);

    switch (cmd)
    {
    case ELAN3IO_FREE:
	if (atomic_read (&pr->pr_mappings) > 0)
	    return (-EINVAL);
	
	spin_lock (&current->mm->page_table_lock);
	if (pr->pr_mm != current->mm)
	    spin_unlock (&current->mm->page_table_lock);
	else
	{
	    ioproc_unregister_ops (current->mm, &pr->pr_ioproc);
	    spin_unlock (&current->mm->page_table_lock);

	    user_ioproc_release (pr, current->mm);
	}

	if (ptrack_registered (user_ptrack_handler, pr))
	{
	    ptrack_deregister (user_ptrack_handler, pr);
	    user_ptrack_handler (pr, PTRACK_PHASE_EXIT, NULL);
	}
	break;
	
    case ELAN3IO_ATTACH:
    {
	ELAN_CAPABILITY *cap;

	KMEM_ALLOC(cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), TRUE);

	if (cap == NULL)
	    return (set_errno (EFAULT));

	if (copy_from_user (cap, (void *) arg, sizeof (ELAN_CAPABILITY)))
	    res = EFAULT;
	else
	{
	    if ((res = elan3_attach (ctxt, cap)) == 0)
	    {
		if (copy_to_user ((void *) arg, cap, sizeof (ELAN_CAPABILITY)))
		{
		    elan3_detach (ctxt);
		    res = EFAULT;
		}
	    }
	}
	KMEM_FREE (cap, sizeof(ELAN_CAPABILITY));
	break;
    }
    
    case ELAN3IO_DETACH:
	spin_lock (&pr->pr_lock);
	if (atomic_read (&pr->pr_mappings) > 0)
	    res = EINVAL;
	else
	    elan3_detach (ctxt);
	spin_unlock (&pr->pr_lock);
	break;

    case ELAN3IO_ADDVP:
    {
	ELAN3IO_ADDVP_STRUCT *args;

	KMEM_ALLOC(args, ELAN3IO_ADDVP_STRUCT *, sizeof (ELAN3IO_ADDVP_STRUCT), TRUE);

	if (args == NULL)
	    return (set_errno (ENOMEM));
	
	if (copy_from_user (args, (void *) arg, sizeof (ELAN3IO_ADDVP_STRUCT)))
	    res = EFAULT;
	else
	{
	    if ( (res=elan3_addvp (ctxt, args->process, &args->capability)) != 0)
		PRINTF0 (ctxt, DBG_FN, "ELAN3IO_ADDVP elan3_addvp failed \n");	
	}

	KMEM_FREE (args, sizeof (ELAN3IO_ADDVP_STRUCT));
	break;
    }

    case ELAN3IO_REMOVEVP:
	res = elan3_removevp (ctxt, arg);
	break;
	
    case ELAN3IO_BCASTVP:
    {
	ELAN3IO_BCASTVP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_BCASTVP_STRUCT)))
	    return (-EFAULT);
	
	res = elan3_addbcastvp (ctxt, args.process, args.lowvp, args.highvp);
	break;
    }

    case ELAN3IO_LOAD_ROUTE:
    {
	ELAN3IO_LOAD_ROUTE_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_LOAD_ROUTE_STRUCT)))
	    return (-EFAULT);
	
	res = elan3_load_route (ctxt, args.process, args.flits);
	break;
    }

    case ELAN3IO_CHECK_ROUTE:
    {
	ELAN3IO_CHECK_ROUTE_STRUCT args;

	args.routeError = 0;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_LOAD_ROUTE_STRUCT)))
	    return (-EFAULT);
	
	if ((res = elan3_check_route (ctxt, args.process, args.flits, & args.routeError)) ==  ESUCCESS)
	{
	    if (copy_to_user ( (void *) arg, &args,sizeof (ELAN3IO_LOAD_ROUTE_STRUCT)))
		return (-EFAULT);
	}
	break;
    }

    case ELAN3IO_PROCESS_2_LOCATION:
    {
	ELAN3IO_PROCESS_2_LOCATION_STRUCT args;
	ELAN_LOCATION                    loc;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_PROCESS_2_LOCATION_STRUCT)))
	    return (-EFAULT);

	krwlock_write (&ctxt->VpLock);
	loc = ProcessToLocation (ctxt, NULL, args.process , NULL);
	krwlock_done (&ctxt->VpLock);

	args.loc = loc;

	if (copy_to_user ( (void *) arg, &args,sizeof (ELAN3IO_PROCESS_2_LOCATION_STRUCT)))
	    return (-EFAULT);

	break;
    }

    case ELAN3IO_GET_ROUTE:
    {
	ELAN3IO_GET_ROUTE_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_GET_ROUTE_STRUCT)))
	    return (-EFAULT);
	
	if ((res = elan3_get_route (ctxt, args.process, args.flits)) ==  ESUCCESS)
	{
	    if (copy_to_user ( (void *) arg, &args,sizeof (ELAN3IO_GET_ROUTE_STRUCT)))
		return (-EFAULT);
	}
	break;
    }

    case ELAN3IO_RESET_ROUTE:
    {
	ELAN3IO_RESET_ROUTE_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_RESET_ROUTE_STRUCT)))
	    return (-EFAULT);
	
	res = elan3_reset_route (ctxt, args.process);
	break;
    }

    case ELAN3IO_VP2NODEID:
    {
	ELAN3IO_VP2NODEID_STRUCT *vp2nodeId;
	ELAN_LOCATION           location;

	KMEM_ALLOC (vp2nodeId, ELAN3IO_VP2NODEID_STRUCT *, sizeof(ELAN3IO_VP2NODEID_STRUCT), TRUE);
	if (vp2nodeId == NULL) 
	    return (set_errno (ENOMEM));

	if (copy_from_user (vp2nodeId, (void *) arg, sizeof (ELAN3IO_VP2NODEID_STRUCT))) {
	    KMEM_FREE (vp2nodeId, sizeof(ELAN3IO_VP2NODEID_STRUCT));
	    return (-EFAULT);
	}

	krwlock_write (&ctxt->VpLock);
	location = ProcessToLocation (ctxt, NULL, vp2nodeId->process , NULL);
	krwlock_done (&ctxt->VpLock);

	vp2nodeId->nodeId = location.loc_node;
	if (copy_to_user ( (void *) arg, vp2nodeId, sizeof (ELAN3IO_VP2NODEID_STRUCT))) {
	    KMEM_FREE (vp2nodeId, sizeof(ELAN3IO_VP2NODEID_STRUCT));
	    return (-EFAULT);
	}

	KMEM_FREE (vp2nodeId, sizeof(ELAN3IO_VP2NODEID_STRUCT));

	break;
    }

    case ELAN3IO_PROCESS:
	return (elan3_process (ctxt));

    case ELAN3IO_SETPERM:
    {
	ELAN3IO_SETPERM_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_SETPERM_STRUCT)))
	    return (-EFAULT);

	res = elan3mmu_setperm (ctxt->Elan3mmu, args.maddr, args.eaddr, args.len, args.perm);
	break;
    }

    case ELAN3IO_CLEARPERM:
    {
	ELAN3IO_CLEARPERM_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_CLEARPERM_STRUCT)))
	    return (-EFAULT);

	elan3mmu_clrperm (ctxt->Elan3mmu, args.eaddr, args.len);
	break;
    }

    case ELAN3IO_CHANGEPERM:
    {
	ELAN3IO_CHANGEPERM_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_CHANGEPERM_STRUCT)))
	    return (-EFAULT);

	res = EINVAL;
	break;
    }

    case ELAN3IO_HELPER_THREAD:
	res = elan3_lwp (ctxt);
	break;
	
    case ELAN3IO_WAITCOMMAND:
	res = WaitForCommandPort (ctxt);
	break;

    case ELAN3IO_BLOCK_INPUTTER:
	elan3_block_inputter (ctxt, arg);
	break;

    case ELAN3IO_SET_FLAGS:
	sctx->Flags = arg;
	break;

    case ELAN3IO_SET_SIGNAL:
	sctx->signal = arg;
	break;

    case ELAN3IO_WAITEVENT:
	res = sys_waitevent (ctxt, (E3_Event *) arg);
	break;

    case ELAN3IO_ALLOC_EVENTCOOKIE:
	res = cookie_alloc_cookie (sctx->Table, arg);
	break;

    case ELAN3IO_FREE_EVENTCOOKIE:
	res = cookie_free_cookie (sctx->Table, arg);
	break;

    case ELAN3IO_ARM_EVENTCOOKIE:
	res = cookie_arm_cookie (sctx->Table, arg);
	break;

    case ELAN3IO_WAIT_EVENTCOOKIE:
	res = cookie_wait_cookie (sctx->Table, arg);
	break;

    case ELAN3IO_SWAPSPACE:
	if (fuword (&((SYS_SWAP_SPACE *) arg)->Magic) != SYS_SWAP_MAGIC)
	    return (set_errno (EINVAL));
	
	((SYS_CTXT *) ctxt->Private)->Swap = (SYS_SWAP_SPACE *) arg;
	break;

    case ELAN3IO_EXCEPTION_SPACE:
	if (fuword (&((SYS_EXCEPTION_SPACE *) arg)->Magic) != SYS_EXCEPTION_MAGIC)
	    return (set_errno (EINVAL));

	((SYS_CTXT *) ctxt->Private)->Exceptions = (SYS_EXCEPTION_SPACE *) arg;
	break;

    case ELAN3IO_GET_EXCEPTION:
    {
	SYS_EXCEPTION *exception;

	if (((SYS_CTXT *) ctxt->Private)->Exceptions == NULL)
	    return (set_errno (EINVAL));
	
	KMEM_ALLOC(exception, SYS_EXCEPTION *, sizeof (SYS_EXCEPTION), TRUE);

	if (exception == NULL)
	    return (set_errno (ENOMEM));

	if ((res = sys_getException (((SYS_CTXT *) ctxt->Private), exception)) == 0 &&
	    copy_to_user ((void *) arg, exception, sizeof (SYS_EXCEPTION)))
	    res = EFAULT;
	
	KMEM_FREE (exception, sizeof (SYS_EXCEPTION));
	break;
    }
    
    case ELAN3IO_UNLOAD:
    {
	ELAN3MMU             *elan3mmu = ctxt->Elan3mmu;
	ELAN3IO_UNLOAD_STRUCT args;
	int	              span;
	unsigned long         flags;
	E3_Addr               eaddr;
	caddr_t		      addr;
	size_t		      len;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_UNLOAD_STRUCT)))
	    return (-EFAULT);

	addr = (caddr_t) args.addr;
	len  = args.len;

	if (((unsigned long) addr & PAGEMASK) || (len & PAGEMASK) || (len < 0))
	    return -EINVAL;

	spin_lock_irqsave (&elan3mmu->elan3mmu_lock, flags);
	for (; len; len -= span, addr += span)
	{
	    ELAN3MMU_RGN *rgn = elan3mmu_findrgn_main (elan3mmu, addr, 0);
	    
	    if (rgn == NULL || (rgn->rgn_mbase + rgn->rgn_len) < addr)
		span = len;
	    else if (rgn->rgn_mbase > addr)
		span = MIN(len, rgn->rgn_mbase - addr);
	    else
	    {
		span  = MIN(len, (rgn->rgn_mbase + rgn->rgn_len) - addr);
		eaddr = rgn->rgn_ebase + (addr - rgn->rgn_mbase);
		
		elan3mmu_unload (elan3mmu, eaddr, span, PTE_UNLOAD);
	    }
	}
	spin_unlock_irqrestore (&elan3mmu->elan3mmu_lock, flags);
	
	return 0;
    }

    case ELAN3IO_GET_DEVINFO:
    {
	ELAN3IO_GET_DEVINFO_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_GET_DEVINFO_STRUCT)))
	    return (-EFAULT);
	
	if (copy_to_user ((void *) args.devinfo, &ctxt->Device->Devinfo, sizeof (ELAN_DEVINFO))) 
	    res = EFAULT;
	break;
    }

    case ELAN3IO_GET_POSITION:
    {
	ELAN3IO_GET_POSITION_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN3IO_GET_POSITION_STRUCT)))
	    return (-EFAULT);	

	if (copy_to_user ((void *) args.position, &ctxt->Device->Position, sizeof (ELAN_POSITION)))
	    res = EFAULT;
	break;
    }

    default:
	return (-EINVAL);
    }

    return (res ? set_errno (res) : 0);
}

static void user_vma_open(struct vm_area_struct *vma)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) vma->vm_private_data;

    PRINTF (DBG_DEVICE, DBG_SEG, "user_vma_open: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    if (vma->vm_pgoff == ELAN3IO_OFF_COMMAND_PAGE)
	if (atomic_dec_and_test (&pr->pr_mappings))
	    pr->pr_ctxt->CommandPageMapping = NULL;
}

static void user_vma_close(struct vm_area_struct *vma)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) vma->vm_private_data;

    PRINTF (DBG_DEVICE, DBG_SEG, "user_vma_close: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    if (vma->vm_pgoff == ELAN3IO_OFF_COMMAND_PAGE)
	atomic_inc (&pr->pr_mappings);
}

static struct vm_operations_struct user_vm_ops = {
    open:		user_vma_open,
    close:		user_vma_close,
};

static int
user_mmap (struct file *file, struct vm_area_struct *vma)
{
    USER_PRIVATE  *pr   = (USER_PRIVATE *) file->private_data;
    ELAN3_CTXT     *ctxt = pr->pr_ctxt; 
    ioaddr_t       ioaddr;

    /* 
     * NOTE - since we need to maintain the reference count on
     *        the user_private we only permit single page 
     *        mmaps - this means that we will certainly see
     *        the correct number of closes to maintain the
     *        the reference count correctly.
     */
    
    if ((vma->vm_end - vma->vm_start) != PAGE_SIZE)
	return (-EINVAL);

    PRINTF (DBG_DEVICE, DBG_SEG, "user_mmap: vm_mm=%p start=%lx end=%lx pgoff=%lx flags=%lx prot=%llx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_flags,  (long long)vma->vm_page_prot.pgprot, vma->vm_file);

    switch (vma->vm_pgoff)
    {
    default:
	return (-EINVAL);
	
    case ELAN3IO_OFF_COMMAND_PAGE:
	spin_lock (&pr->pr_lock);
	if (ctxt->CommandPage == (ioaddr_t) 0 || atomic_read (&pr->pr_mappings) != 0)
	{
	    PRINTF (DBG_DEVICE, DBG_SEG, "user_mmap: command port - %s\n", ctxt->CommandPort ? "already mapped" : "not attached");
	    spin_unlock (&pr->pr_lock);
	    return (-EINVAL);
	}
#ifdef LINUX_SPARC
	pgprot_val(vma->vm_page_prot) &= ~(_PAGE_CACHE);
	pgprot_val(vma->vm_page_prot) |= _PAGE_IE;
#elif defined(pgprot_noncached)
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif

	PRINTF (DBG_DEVICE, DBG_SEG, "user_mmap: commandport at %lx phys %llx prot %llx\n", 
		vma->vm_start, (unsigned long long) kmem_to_phys ((void *) ctxt->CommandPort),  (long long)vma->vm_page_prot.pgprot);

	/* Don't try to swap out physical pages.. */
	vma->vm_flags |= VM_RESERVED;
    
	/*
	 * Don't dump addresses that are not real memory to a core file.
	 */
	vma->vm_flags |= VM_IO;

    	if (__remap_page_range(vma->vm_start, kmem_to_phys ((void *) ctxt->CommandPage), vma->vm_end - vma->vm_start, vma->vm_page_prot))
	{
	    spin_unlock (&pr->pr_lock);
	    return (-EAGAIN);
	}
	ctxt->CommandPageMapping = (void *) vma->vm_start;
	
	atomic_inc (&pr->pr_mappings);
	
	spin_unlock (&pr->pr_lock);
	break;

    case ELAN3IO_OFF_UREG_PAGE:
#ifdef LINUX_SPARC
	pgprot_val(vma->vm_page_prot) &= ~(_PAGE_CACHE);
	pgprot_val(vma->vm_page_prot) |= _PAGE_IE;
#elif defined(pgprot_noncached)
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
	ioaddr = ctxt->Device->RegPtr + (offsetof (E3_Regs, URegs) & PAGEMASK);

	PRINTF (DBG_DEVICE, DBG_SEG, "user_mmap: user_regs at %lx phys %llx prot %llx\n", vma->vm_start, 
		(unsigned long long) kmem_to_phys ((void *) ioaddr),  (long long)vma->vm_page_prot.pgprot);

	/* Don't try to swap out physical pages.. */
	vma->vm_flags |= VM_RESERVED;
    
	/*
	 * Don't dump addresses that are not real memory to a core file.
	 */
	vma->vm_flags |= VM_IO;
	if (__remap_page_range (vma->vm_start, kmem_to_phys ((void *) ioaddr),
				vma->vm_end - vma->vm_start, vma->vm_page_prot))
	    return (-EAGAIN);
	break;
	
    case ELAN3IO_OFF_FLAG_PAGE:
	PRINTF (DBG_DEVICE, DBG_SEG, "user_mmap: flag page at %lx phys %llx\n", vma->vm_start, 
		(unsigned long long) kmem_to_phys ((void *) ctxt->FlagPage));

	/* we do not want to have this area swapped out, lock it */
	vma->vm_flags |= VM_LOCKED;

	/* Mark the page as reserved or else the remap_page_range() doesn't remap it */
	SetPageReserved(pte_page(*find_pte_kernel((unsigned long) ctxt->FlagPage)));
	
	if (__remap_page_range (vma->vm_start, kmem_to_phys ((void *) ctxt->FlagPage),
			      vma->vm_end - vma->vm_start, vma->vm_page_prot))
	    return (-EAGAIN);
	break;
    }

    ASSERT (vma->vm_ops == NULL);
    
    vma->vm_ops          = &user_vm_ops;
    vma->vm_file         = file;
    vma->vm_private_data = (void *) pr;
    
    return (0);
}

#else /* defined(NO_PTRACK) || !defined(IOPROC_PATCH_APPLIED) */

static int
user_open (struct inode *inode, struct file *file)
{
    return -ENXIO;
}

static int
user_release (struct inode *inode, struct file *file)
{
    return 0;
}

static int
user_ioctl (struct inode *inode, struct file *file, 
	    unsigned int cmd, unsigned long arg)
{
    return 0;
}

static int
user_mmap (struct file *file, struct vm_area_struct *vma)
{
    return 0;
}
#endif

/* driver entry points */
static int
elan3_open (struct inode *inode, struct file *file)
{
    if (elan3_devices[ELAN3_DEVICE(inode)] == NULL)
	return (-ENXIO);

    PRINTF (DBG_DEVICE, DBG_FN, "elan3_open: device %d minor %d file=%p\n", ELAN3_DEVICE(inode), ELAN3_MINOR(inode), file);
    
    switch (ELAN3_MINOR (inode))
    {
    case ELAN3_MINOR_CONTROL:
	return (control_open (inode, file));
    case ELAN3_MINOR_MEM:
	return (mem_open (inode, file));
    case ELAN3_MINOR_USER:
	return (user_open (inode, file));
    default:
	return (-ENXIO);
    }
}

static int
elan3_release (struct inode *inode, struct file *file)
{
    PRINTF (DBG_DEVICE, DBG_FN, "elan3_release: device %d minor %d file=%p\n", ELAN3_DEVICE(inode), ELAN3_MINOR(inode), file);
    
    switch (ELAN3_MINOR (inode))
    {
    case ELAN3_MINOR_CONTROL:
	return (control_release (inode, file));
    case ELAN3_MINOR_MEM:
	return (mem_release (inode, file));
    case ELAN3_MINOR_USER:
	return (user_release (inode, file));
    default:
	return (-ENXIO);
    }
}

static int
elan3_ioctl (struct inode *inode, struct file *file, 
	     unsigned int cmd, unsigned long arg)
{
    switch (ELAN3_MINOR (inode))
    {
    case ELAN3_MINOR_CONTROL:
	return (control_ioctl (inode, file, cmd, arg));
    case ELAN3_MINOR_MEM:
	return (mem_ioctl (inode, file, cmd, arg));
    case ELAN3_MINOR_USER:
	return (user_ioctl (inode, file, cmd, arg));
    default:
	return (-ENXIO);
    }
}


static int
elan3_mmap (struct file *file, struct vm_area_struct *vma)
{
    PRINTF (DBG_DEVICE, DBG_SEG, "elan3_mmap: instance %d minor %d start=%lx end=%lx pgoff=%lx flags=%lx prot=%llx\n", 
	    ELAN3_DEVICE (file->f_dentry->d_inode), ELAN3_MINOR (file->f_dentry->d_inode),
	    vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_flags,  (long long)vma->vm_page_prot.pgprot);

    switch (ELAN3_MINOR (file->f_dentry->d_inode))
    {
    case ELAN3_MINOR_CONTROL:
	return (control_mmap (file, vma));
    case ELAN3_MINOR_MEM:
	return (mem_mmap (file, vma));
    case ELAN3_MINOR_USER:
	return (user_mmap (file, vma));
    default:
	return (-ENXIO);
    }
}

static irqreturn_t
InterruptHandlerWrapper(int irq, void *dev_id, struct pt_regs *regs)
{
    if (InterruptHandler ((ELAN3_DEV *)dev_id) == 0)
	return IRQ_HANDLED;
    else
	return IRQ_NONE;
}


/* 
 * Elan specific PCI configuration registers.
 */

#define PCI_CONF_PARITY_PHYS_LO         0x40
#define PCI_CONF_PARITY_PHYS_HI         0x44
#define PCI_CONF_PARITY_PHASE_ADDR      0x46
#define PCI_CONF_PARITY_MASTER_TYPE     0x47
#define PCI_CONF_ELAN3_CTRL              0x48
 
#define ECTRL_EXTEND_LATENCY            (1 << 0)
#define ECTRL_ENABLE_PREFETCH           (1 << 1)
#define ECTRL_SOFTWARE_INTERNAL_RESET   (1 << 2)
#define ECTRL_REDUCED_RETRY_RATE        (1 << 3)
#define ECTRL_CLOCK_DIVIDE_RATE_SHIFT   4
#define ECTRL_COMMS_DIVIDE_RATE_SHIFT   10
#define ECTRL_FORCE_COMMSCLK_LOCAL      (1 << 14)

/*
 * Configure PCI.
 */
static int
ConfigurePci(ELAN3_DEV *dev)
{
	struct pci_dev *pci = dev->Osdep.pci;
	u32 rom_address;

	if (pci_enable_device(pci))
	    return (ENXIO);

	/* disable ROM */
	pci_read_config_dword(pci, PCI_ROM_ADDRESS, &rom_address);
	rom_address &= ~PCI_ROM_ADDRESS_ENABLE;
	pci_write_config_dword(pci, PCI_ROM_ADDRESS, rom_address);
	mb();

	/* this is in 32-bit WORDS */
	pci_write_config_byte(pci, PCI_CACHE_LINE_SIZE, (64 >> 2));
	mb();

	/* allow 40 ticks to respond, 16 data phases */
	pci_write_config_byte(pci, PCI_LATENCY_TIMER, 255);
	mb();

	/* don't enable PCI_COMMAND_SERR--see note in elandev_dunix.c */
	pci_write_config_word(pci, PCI_COMMAND, PCI_COMMAND_MEMORY 
	    | PCI_COMMAND_MASTER | PCI_COMMAND_INVALIDATE | PCI_COMMAND_PARITY);
	mb();

	return ESUCCESS;
}

/* 
 * Reset chip to a known state.
 */
static int
ResetElan(ELAN3_DEV *dev, ioaddr_t intPalAddr)
{
	struct pci_dev *pci = dev->Osdep.pci;
	int instance = dev->Instance;
	u32 val;
	u8 revid;
	int CasLatency;
	int res;

	/* determine rev of board */
	pci_read_config_byte(pci, PCI_REVISION_ID, &revid);

	/* GNAT 2328 - don't set ECTRL_ENABLE_PREFETCH on Elan rev A */
	val = ECTRL_EXTEND_LATENCY | (39 << ECTRL_CLOCK_DIVIDE_RATE_SHIFT)
	    | (6 << ECTRL_COMMS_DIVIDE_RATE_SHIFT);
	switch (revid) 
	{
		case PCI_REVISION_ID_ELAN3_REVA:
			printk("elan%d: is an elan3 (revision a) - not supported\n", instance);
			return (EFAIL);

		case PCI_REVISION_ID_ELAN3_REVB:	
			val |= ECTRL_ENABLE_PREFETCH;
			if (BackToBackMaster)
				val |= ECTRL_FORCE_COMMSCLK_LOCAL;
			printk("elan%d: is an elan3 (revision b)\n", instance);
			break;
		default:
			printk("elan%d: unsupported elan3 revision %d\n", 
			    instance, revid);
			return EFAIL;
	}
	pci_write_config_dword(pci, PCI_CONF_ELAN3_CTRL, val);
	mb();

	/*
	 * GNAT: 2474
	 * Hit reset on the Elan, then we MUST initialise the schedule status
	 * register to drive reset on the link before the link can come out
	 * of reset (15 uS). We need to keep it like this until we've 
	 * initialised SDRAM
	 */
	pci_read_config_dword(pci, PCI_CONF_ELAN3_CTRL, &val);
	pci_write_config_dword(pci, PCI_CONF_ELAN3_CTRL, 
	    val | ECTRL_SOFTWARE_INTERNAL_RESET);
	mb();

	/* Read the Vital Product Data to determine the cas latency */
	if ((res = ReadVitalProductData (dev, &CasLatency)) != ESUCCESS)
	    return (res);

	/*
	 * Now clear the Software internal reset bit, and start the sdram
	 */
	pci_write_config_dword(pci, PCI_CONF_ELAN3_CTRL, val);
	mb();

	/* 
	 * Enable SDRAM before sizing and initalising it for ECC.
	 * NOTE: don't enable all sets of the cache (yet), nor ECC 
	 */
	dev->Cache_Control_Reg = (CasLatency | REFRESH_RATE_16US);

	write_reg32 (dev, Cache_Control_Reg.ContReg, (dev->Cache_Control_Reg | SETUP_SDRAM));
	mb();

	INIT_SCHED_STATUS(dev, Sched_Initial_Value);

	/*
	 * Set the interrupt mask to 0 and enable the interrupt PAL
	 * by writing any value to it.
	 */
	SET_INT_MASK (dev, 0);
	writeb (0, (void *) intPalAddr);
 
	return ESUCCESS;
}

/*
 * Determine the size of elan PCI address spaces.  EFAIL is returned if 
 * unused or invalid BAR is specified, or if board reports I/O mapped space.
 */
int
DeviceRegisterSize(ELAN3_DEV *dev, int rnumber, int *sizep)
{
	struct pci_dev *pdev = dev->Osdep.pci;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	*sizep = pci_resource_size(pdev, rnumber);
#else
	*sizep = pci_resource_end(pdev, rnumber) - pci_resource_start(pdev, rnumber) + 1;
#endif
	return ESUCCESS;
}

/*
 * Map PCI memory into kernel virtual address space.  On the alpha, 
 * we just return appropriate kseg address, and Unmap is a no-op.
 */
int
MapDeviceRegister(ELAN3_DEV *dev, int rnumber, ioaddr_t *addrp,
		  int off, int len, DeviceMappingHandle *handlep)
{	
	struct pci_dev *pdev = dev->Osdep.pci;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	u64 base = pci_get_base_address(pdev, rnumber);
	*addrp = (ioaddr_t) pci_base_to_kseg(base + off, pdev->bus->number);

#else
	if (len == 0)
	    len = pci_resource_end(pdev, rnumber) - pci_resource_start(pdev, rnumber) + 1;
	
	if (len == 0)
	    return (EINVAL);

	*addrp = (ioaddr_t) ioremap_nocache (pci_resource_start(pdev, rnumber) + off, len);
#endif

	*handlep = (void *) *addrp;

	return (*addrp ? ESUCCESS : ENOMEM);
}
void
UnmapDeviceRegister(ELAN3_DEV *dev, DeviceMappingHandle *handlep)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    iounmap (*handlep);
#endif
}

void
ElanBusError (ELAN3_DEV *dev)
{
	struct pci_dev	*pci = dev->Osdep.pci;	
	u8  phaseaddr, type;
	u16 status, cmd, physhi;
	u32 physlo;
 
	printk("elan%d: bus error occured\n", dev->Instance);

	pci_read_config_word (pci, PCI_STATUS,			&status);
	pci_read_config_word (pci, PCI_COMMAND,			&cmd);
	pci_read_config_dword(pci, PCI_CONF_PARITY_PHYS_LO,	&physlo);
	pci_read_config_word (pci, PCI_CONF_PARITY_PHYS_HI,	&physhi);
	pci_read_config_byte (pci, PCI_CONF_PARITY_PHASE_ADDR,	&phaseaddr); 
	pci_read_config_byte (pci, PCI_CONF_PARITY_MASTER_TYPE,	&type);

#define PCI_CONF_STAT_FORMAT	"\20" \
	"\6SIXTY_SIX_MHZ\7UDF\10FAST_BACK\11PARITY" \
	"\14SIG_TARGET_ABORT\15REC_TARGET_ABORT\16REC_MASTER_ABORT" \
	"\17SIG_SYSTEM_ERROR\20DETECTED_PARITY"

	printk ("elan%d: status %x cmd %4x physaddr %04x%08x phase %x type %x\n",
		dev->Instance, status, cmd, physhi, physlo, phaseaddr, type);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
