/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elansyscall.c,v 1.100 2004/11/01 18:01:51 robin Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/elansyscall.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <elan/elanmod.h>
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/elansyscall.h>
#include <elan/devinfo.h>

static int       sys_exception (ELAN3_CTXT *ctxt, int type, int proc, void *trap, va_list ap);
static int       sys_getWordItem (ELAN3_CTXT *ctxt, int list, void **itemp, E3_uint32 *valuep);
static int       sys_getBlockItem (ELAN3_CTXT *ctxt, int list, void **itemp, E3_Addr *valuep);
static void      sys_putWordItem (ELAN3_CTXT *ctxt, int list, E3_uint32 value);
static void      sys_putBlockItem (ELAN3_CTXT *ctxt, int list, E3_uint32 *ptr);
static void      sys_putbackItem (ELAN3_CTXT *ctxt, int list, void *item);
static void      sys_freeWordItem (ELAN3_CTXT *ctxt, void *item);
static void      sys_freeBlockItem (ELAN3_CTXT *ctxt, void *item);
static int       sys_countItems (ELAN3_CTXT *ctxt, int list);
static int       sys_event (ELAN3_CTXT *ctxt, E3_uint32 cookie, int flag);
static void      sys_swapin (ELAN3_CTXT *ctxt);
static void      sys_swapout (ELAN3_CTXT *ctxt);
static void      sys_freePrivate (ELAN3_CTXT *ctxt);
static int       sys_fixupNetworkError (ELAN3_CTXT *ctxt, NETERR_FIXUP *nef);
static int       sys_startFaultCheck (ELAN3_CTXT *ctxt);
static void      sys_endFaultCheck (ELAN3_CTXT *ctxt);
static E3_uint8  sys_load8 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void      sys_store8 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint8 val);
static E3_uint16 sys_load16 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void      sys_store16 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint16 val);
static E3_uint32 sys_load32 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void      sys_store32 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint32 val);
static E3_uint64 sys_load64 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void      sys_store64 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint64 val);

static ELAN3_OPS elan3_sys_ops = {
    ELAN3_OPS_VERSION,		/* Version */

    sys_exception,		/* Exception */
    sys_getWordItem,		/* GetWordItem */
    sys_getBlockItem,		/* GetBlockItem */
    sys_putWordItem,		/* PutWordItem */
    sys_putBlockItem,		/* PutBlockItem */
    sys_putbackItem,		/* PutbackItem */
    sys_freeWordItem,		/* FreeWordItem */
    sys_freeBlockItem,		/* FreeBlockItem */
    sys_countItems,		/* CountItems */
    sys_event,			/* Event */
    sys_swapin,			/* Swapin */
    sys_swapout,		/* Swapout */
    sys_freePrivate,		/* FreePrivate */
    sys_fixupNetworkError,	/* FixupNetworkError */
    NULL,			/* DProcTrap */
    NULL,			/* TProcTrap */
    NULL,			/* IProcTrap */
    NULL,			/* CProcTrap */
    NULL,			/* CProcReissue */
    sys_startFaultCheck,	/* StartFaultCheck */
    sys_endFaultCheck,          /* EndFaultCheck */
    sys_load8,			/* Load8 */
    sys_store8,			/* Store8 */
    sys_load16,			/* Load16 */
    sys_store16,		/* Store16 */
    sys_load32,			/* Load32 */
    sys_store32,		/* Store32 */
    sys_load64,			/* Load64 */
    sys_store64			/* Store64 */
};

va_list null_valist;

SYS_CTXT *
sys_init (ELAN3_CTXT *ctxt)
{
    SYS_CTXT *sctx;

    /* Allocate and initialise the context private data */
    KMEM_ZALLOC (sctx, SYS_CTXT *, sizeof  (SYS_CTXT), TRUE);

    if (sctx == NULL)
	return ((SYS_CTXT *) NULL);

    sctx->Swap    = NULL;
    sctx->Armed   = 0;
    sctx->Backoff = 1;
    sctx->Table   = cookie_alloc_table ((unsigned long) ELAN3_MY_TASK_HANDLE(), 0);
    sctx->signal  = SIGSEGV;

    if (sctx->Table == NULL)
    {
	KMEM_FREE (sctx, sizeof (SYS_CTXT));
	return ((SYS_CTXT *) NULL);
    }

    kmutex_init  (&sctx->Lock);
    spin_lock_init (&sctx->WaitLock);
    kcondvar_init (&sctx->NetworkErrorWait);
    
    /* Install my context operations and private data */
    ctxt->Operations = &elan3_sys_ops;
    ctxt->Private    = (void *) sctx;
    
    return (sctx);
}

/* returns -ve on error or ELAN_CAP_OK or ELAN_CAP_RMS */
/* use = ELAN_USER_ATTACH, ELAN_USER_P2P, ELAN_USER_BROADCAST */
int 
elan3_validate_cap(ELAN3_DEV *dev, ELAN_CAPABILITY *cap ,int use)
{
     /* Don't allow a user process to attach to system context */
    if (ELAN3_SYSTEM_CONTEXT (cap->cap_lowcontext) || ELAN3_SYSTEM_CONTEXT (cap->cap_highcontext)
	|| cap->cap_highcontext <= ELAN_USER_BASE_CONTEXT_NUM  || cap->cap_highcontext <= ELAN_USER_BASE_CONTEXT_NUM)
    {
	PRINTF2 (DBG_DEVICE, DBG_VP,"elan3_validate_cap: lctx %x hctx %x \n",cap->cap_lowcontext,  cap->cap_highcontext);
	PRINTF3 (DBG_DEVICE, DBG_VP,"elan3_validate_cap: bit %x  low %x high %x\n", ((cap->cap_lowcontext) & SYS_CONTEXT_BIT),
		 E3_NUM_CONTEXT_0, ELAN3_KCOMM_BASE_CONTEXT_NUM);


	PRINTF0 (DBG_DEVICE, DBG_VP,"elan3_validate_cap: user process cant attach to system cap\n");
	return (-EINVAL);
    }
 
    if (cap->cap_type & ELAN_CAP_TYPE_HWTEST)
    {
	if (!(cap->cap_type & ELAN_CAP_TYPE_NO_BITMAP)) /* cant have a bit map */
	{
	    PRINTF0 (DBG_DEVICE, DBG_VP, "elanmod_classify_cap: ELAN_CAP_TYPE_HWTEST must have ELAN_CAP_TYPE_NO_BITMAP\n");
	    return (-EINVAL);
	}
	
	if (cap->cap_lowcontext != cap->cap_highcontext) 
	{
	    PRINTF2 (DBG_DEVICE, DBG_VP, "elanmod_classify_cap: ELAN_CAP_TYPE_HWTEST (cap->cap_lowcontext != cap->cap_highcontext) %d %d\n",cap->cap_lowcontext , cap->cap_highcontext) ;
	    return (-EINVAL);
	}
	
	if ( ! (ELAN3_HWTEST_CONTEXT(cap->cap_lowcontext) && ELAN3_HWTEST_CONTEXT(cap->cap_highcontext)))
	{
	    PRINTF3 (DBG_DEVICE, DBG_VP, "elanmod_classify_cap: ELAN_CAP_TYPE_HWTEST HWTEST_BASE_CONTEXT %d %d %d \n" , ELAN3_HWTEST_BASE_CONTEXT_NUM,cap->cap_lowcontext ,ELAN3_HWTEST_TOP_CONTEXT_NUM);
	    return (-EINVAL);
	}
	
	if (cap->cap_lownode != ELAN_CAP_UNINITIALISED || cap->cap_highnode != ELAN_CAP_UNINITIALISED)
	{
	    PRINTF0 (DBG_DEVICE, DBG_VP, "elanmod_classify_cap: ELAN_CAP_TYPE_HWTEST nodes != ELAN_CAP_UNINITIALISED\n");
	    return (-EINVAL);
	}

	return ELAN_CAP_OK;
    }

    return elanmod_classify_cap(&dev->Position, cap, use);
}

int
sys_waitevent (ELAN3_CTXT *ctxt, E3_Event *event)
{
    SYS_CTXT    *sctx = (SYS_CTXT *) ctxt->Private;
    EVENT_COOKIE cookie;

    if (ctxt->Device->Devinfo.dev_revision_id == PCI_REVISION_ID_ELAN3_REVA)
	return (EINVAL);

    cookie = fuword ((int *) &event->ev_Type) & ~(EV_TYPE_MASK_EVIRQ | EV_TYPE_MASK_BCOPY);

    if (cookie_alloc_cookie (sctx->Table, cookie) != ESUCCESS)
	return (EINVAL);

    cookie_arm_cookie (sctx->Table, cookie);

    if (fuword ((int *) &event->ev_Count) > 0)
	cookie_wait_cookie (sctx->Table, cookie);
    
    cookie_free_cookie (sctx->Table, cookie);
    
    return (ESUCCESS);
}

static void *
sys_getItem (SYS_SWAP_SPACE *sp, int list)
{
    void *itemp = (void *) fuptr_noerr ((void **) &sp->ItemListsHead[list]);
    void *next;
    
    PRINTF4 (DBG_DEVICE, DBG_SYSCALL, "sys_getItem: sp=%p list=%d head=%p itemp=%p\n",
	     sp, list, &sp->ItemListsHead[list], itemp);
    
    if (itemp == NULL)
	return (NULL);

    next = (void *) fuptr_noerr ((void *) itemp);

    suptr_noerr ((void *) &sp->ItemListsHead[list], (void *) next);
    if (next == NULL)
	suptr_noerr ((void *) &sp->ItemListsTailp[list], (void *)&sp->ItemListsHead[list]);
    return (itemp);
}

static void
sys_putItemBack (SYS_SWAP_SPACE *sp, int list, void *itemp)
{
    PRINTF4 (DBG_DEVICE, DBG_SYSCALL, "sys_putItemBack: sp=%p list=%d itemp=%p value=%08x\n",
	     sp, list, itemp, fuword_noerr ((int *) &((SYS_WORD_ITEM *) itemp)->Value));

    suptr_noerr ((void **) itemp, NULL);							/* item->Next = NULL */
    suptr_noerr ((void **) fuptr_noerr ((void **) &sp->ItemListsTailp[list]), (void *)itemp); 	/* *Tailp = item */
    suptr_noerr ((void **) &sp->ItemListsTailp[list], (void *) itemp);				/* Tailp = &item->Next */
}

static void
sys_putItemFront (SYS_SWAP_SPACE *sp, int list, void *itemp)
{
    PRINTF4 (DBG_DEVICE, DBG_SYSCALL, "sys_putItemFront: sp=%p list=%d itemp=%p value=%08x\n",
	     sp, list, itemp, fuword_noerr ((int *) &((SYS_WORD_ITEM *) itemp)->Value));

    suptr_noerr ((void **) itemp, fuptr_noerr ((void **) &sp->ItemListsHead[list]));		/* item->Next = Head */
    suptr_noerr ((void **) &sp->ItemListsHead[list], (void *) itemp);				/* Head = item */

    if (fuptr_noerr ((void **) &sp->ItemListsTailp[list]) == (void *) &sp->ItemListsHead[list])	/* if (Tailp == &Head) */
	suptr_noerr ((void **) &sp->ItemListsTailp[list], (void *) itemp);			/*    Tailp = &Item->Next */
}


static int
sys_getWordItem (ELAN3_CTXT *ctxt, int list, void **itemp, E3_uint32 *valuep)
{
    SYS_CTXT	   *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    SYS_WORD_ITEM  *item;
    int		    res;
    label_t	    ljb;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljb))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return (0);
    }

    item = (SYS_WORD_ITEM *) sys_getItem (sp, list);

    if (item == NULL)
	res = 0;
    else
    {
	if (list == LIST_DMA_PTR)
	    sctx->Armed = TRUE;

	*itemp  = (void *) item;
	*valuep = (E3_Addr) fuword_noerr ((E3_int32 *) &item->Value);

	PRINTF3 (ctxt, DBG_SYSCALL, "sys_getWordItem: list=%d -> item=%p value=%08x\n", list, *itemp, *valuep);

	res = 1;
    }
    
    no_fault();
    kmutex_unlock (&sctx->Lock);

    return (res);
}

static int
sys_getBlockItem (ELAN3_CTXT *ctxt, int list, void **itemp, E3_Addr *valuep)
{
    SYS_CTXT	   *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    SYS_BLOCK_ITEM *item;
    int		    res;
    label_t	    ljb;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljb))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return (0);
    }

    item = sys_getItem (sp, list);

    if (item == NULL)
	res = 0;
    else
    {
	E3_uint32 *dest = fuptr_noerr ((void **) &item->Pointer);

	if (list == LIST_DMA_DESC)
	    sctx->Armed = TRUE;

	*itemp  = (void *) item;
	*valuep = elan3mmu_elanaddr (ctxt->Elan3mmu, (caddr_t) dest);

	PRINTF3 (ctxt, DBG_SYSCALL, "sys_getBlockItem: list=%d -> item=%p addr=%08x\n", list, *itemp, *valuep);
	PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
		 fuword_noerr ((int *) &dest[0]), fuword_noerr ((int *) &dest[1]), 
		 fuword_noerr ((int *) &dest[2]), fuword_noerr ((int *) &dest[3]));
	PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
		 fuword_noerr ((int *) &dest[4]), fuword_noerr ((int *) &dest[5]),
		 fuword_noerr ((int *) &dest[6]), fuword_noerr ((int *) &dest[7]));

	
	res = 1;
    }
    
    no_fault();
    kmutex_unlock (&sctx->Lock);

    return (res);
}

static void
sys_putWordItem (ELAN3_CTXT *ctxt, int list, E3_Addr value)
{
    SYS_CTXT	   *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    SYS_WORD_ITEM  *item;
    label_t	    ljp;

    kmutex_lock (&sctx->Lock);

    PRINTF2 (ctxt,DBG_SYSCALL, "sys_putWordItem: list=%x value=%x\n", list, value);

    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return;
    }

    item = sys_getItem (sp, LIST_FREE_WORD);

    PRINTF1 (ctxt, DBG_SYSCALL, "sys_putWordItem: item=%p\n", item);

    if (item == NULL)
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAILED, list, (void *) NULL, null_valist);
	return;
    }
    
    PRINTF2 (ctxt, DBG_SYSCALL, "sys_putWordItem: storing value=%08x at %p\n", value, &item->Value);

    PRINTF2 (ctxt, DBG_SYSCALL, "sys_putWordItem: item=%p value=%08x\n", item, value);

    suword_noerr ((E3_int32 *) &item->Value, value);						/* write "value" into item */

    sys_putItemBack (sp, list, item);

    no_fault();
    kmutex_unlock (&sctx->Lock);
}

static void
sys_putBlockItem (ELAN3_CTXT *ctxt, int list, E3_uint32 *ptr)
{
    SYS_CTXT	   *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    SYS_BLOCK_ITEM *item;
    label_t	    ljp;
    E3_uint32      *source;
    E3_uint32      *dest;

    PRINTF2 (ctxt, DBG_SYSCALL, "sys_putBlockItem: list=%x ptr=%p\n", list, ptr);

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return;
    }

    item = sys_getItem (sp, LIST_FREE_BLOCK);			/* get an item from the freelist. */

    if (item == NULL)
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAILED, list, (void *) NULL, null_valist);
	return;
    }

    /*
     * The block will have been read using 64 bit reads,  since we have
     * to write it to user memory using 32 bit writes, we need to perform
     * an endian swap on the Ultrasparc.
     */
    dest   = (E3_uint32 *) fuptr_noerr ((void **) &item->Pointer);
    source = (E3_uint32 *) ptr;

    PRINTF2 (ctxt, DBG_SYSCALL, "sys_putBlockItem: item=%p dest=%p\n",item, dest);
    PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
	    source[0^WordEndianFlip], source[1^WordEndianFlip], source[2^WordEndianFlip], source[3^WordEndianFlip]);
    PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
	     source[4^WordEndianFlip], source[5^WordEndianFlip], source[6^WordEndianFlip], source[7^WordEndianFlip]);

    suword_noerr ((E3_int32 *) &dest[7], (E3_int32) source[7^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[6], (E3_int32) source[6^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[5], (E3_int32) source[5^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[4], (E3_int32) source[4^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[3], (E3_int32) source[3^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[2], (E3_int32) source[2^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[1], (E3_int32) source[1^WordEndianFlip]);
    suword_noerr ((E3_int32 *) &dest[0], (E3_int32) source[0^WordEndianFlip]);

    sys_putItemBack (sp, list, item);				/* chain onto list of items. */

    no_fault();
    kmutex_unlock (&sctx->Lock);
}

static void
sys_freeWordItem (ELAN3_CTXT *ctxt, void *itemp)
{
    SYS_CTXT	   *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    label_t	    ljp;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, LIST_FREE_WORD, (void *) NULL, null_valist);
	return;
    }

    sys_putItemBack (sp, LIST_FREE_WORD, itemp);

    no_fault();
    kmutex_unlock (&sctx->Lock);
}

static void
sys_freeBlockItem (ELAN3_CTXT *ctxt, void *itemp)
{
    SYS_CTXT       *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    SYS_BLOCK_ITEM *item = (SYS_BLOCK_ITEM *)itemp;
    E3_uint32      *dest;
    label_t	    ljp;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, LIST_FREE_BLOCK, (void *) NULL, null_valist);
	return;
    }
#ifdef DEBUG_PRINTF
    dest = (E3_uint32 *) fuptr_noerr ((void **) &item->Pointer);

    PRINTF2 (ctxt, DBG_SYSCALL, "sys_freeBlockItem: item=%p dest=%p\n", item, dest);
    PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
	     fuword_noerr ((int *) &dest[0]), fuword_noerr ((int *) &dest[1]), 
	     fuword_noerr ((int *) &dest[2]), fuword_noerr ((int *) &dest[3]));
    PRINTF4 (ctxt, DBG_SYSCALL, "                  %08x %08x %08x %08x\n",
	     fuword_noerr ((int *) &dest[4]), fuword_noerr ((int *) &dest[5]),
	     fuword_noerr ((int *) &dest[6]), fuword_noerr ((int *) &dest[7]));
#endif

    sys_putItemBack (sp, LIST_FREE_BLOCK, itemp);

    no_fault();
    kmutex_unlock (&sctx->Lock);
}

static void
sys_putbackItem (ELAN3_CTXT *ctxt, int list, void *itemp)
{
    SYS_CTXT       *sctx = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp   = sctx->Swap;
    label_t	    ljp;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return;
    }

    sys_putItemFront (sp, list, itemp);

    no_fault();
    kmutex_unlock (&sctx->Lock);
}

static int
sys_countItems (ELAN3_CTXT *ctxt, int list)
{
    SYS_CTXT	   *sctx  = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp    = sctx->Swap;
    int		    count = 0;
    void	   *item;
    label_t	    ljb;

    kmutex_lock (&sctx->Lock);
    
    if (on_fault (&ljb))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	sys_exception (ctxt, EXCEPTION_SWAP_FAULT, list, (void *) NULL, null_valist);
	return (0);
    }

    for (item = (void *) fuptr_noerr ((void **) &sp->ItemListsHead[list]); 
	 item != NULL;
	 item = (void *) fuptr_noerr ((void **) item))
    {
	count++;
    }

    no_fault();
    kmutex_unlock (&sctx->Lock);

    return (count);
}


long sys_longTime;
long sys_shortTime;
int  sys_waitTicks;
int  sys_maxBackoff;

#define SYS_LONG_TIME		MAX((hz * 5) / 1000, 1)		/* 5 ms */
#define SYS_SHORT_TIME		MAX((hz * 2) / 1000, 1)		/* 2 ms */
#define SYS_WAIT_TICKS		MAX((hz * 1) / 1000, 1)		/* 1 ms  - backoff granularity */
#define SYS_MAX_BACKOFF		MAX((hz * 5) / 1000, 1)		/* 5 ms  - max backoff for "nacked" packets*/
#define SYS_TIMEOUT_BACKOFF	MAX((hz * 10) / 1000, 1)	/* 10 ms - backoff for output timeout (point to point) */
#define SYS_BCAST_BACKOFF	MAX((hz * 50) / 1000, 1)	/* 50 ms - backoff for output timeout (broadcast) */
#define SYS_NETERR_BACKOFF	MAX((hz * 10) / 1000, 1)	/* 10 ms - delay for network error in dma data */

static void
sys_backoffWait (ELAN3_CTXT *ctxt, int ticks)
{
    SYS_CTXT *sctx = (SYS_CTXT *) ctxt->Private;
    long      t;

    spin_lock (&sctx->WaitLock);

    t = lbolt - sctx->Time;

    if (sys_longTime   == 0) sys_longTime   = SYS_LONG_TIME;
    if (sys_shortTime  == 0) sys_shortTime  = SYS_SHORT_TIME;
    if (sys_waitTicks  == 0) sys_waitTicks  = SYS_WAIT_TICKS;
    if (sys_maxBackoff == 0) sys_maxBackoff = SYS_MAX_BACKOFF;

    if (t > sys_longTime)					/* It's a long time since the last trap */
	sctx->Backoff = 0;					/* so set the backoff back down to 0 */

    if (ticks)
    {
	PRINTF2 (ctxt, DBG_DPROC, "sys_backoffWait : Waiting - %d ticks [%lx]\n", ticks, t);
	kcondvar_timedwait (&sctx->NetworkErrorWait, &sctx->WaitLock, NULL, lbolt + ticks);
    }
    else if (sctx->Armed)
    {
	if (t < sys_shortTime)					/* It's been a short time since the last */
	{							/* trap, so increase the backoff */
	    sctx->Backoff++;
	    
	    if (sctx->Backoff > sys_maxBackoff)
		sctx->Backoff = sys_maxBackoff;
	}

	PRINTF2 (ctxt, DBG_DPROC, "sys_backoffWait : Waiting - %d [%lx]\n", sctx->Backoff, t);

       if (sctx->Backoff)
           kcondvar_timedwaitsig (&sctx->NetworkErrorWait, &sctx->WaitLock, NULL, lbolt + sctx->Backoff * sys_waitTicks);

	sctx->Armed = 0;
    }
    else
    {
	PRINTF1 (ctxt, DBG_DPROC, "sys_backoffWait : Not Waiting - %d\n", sctx->Backoff);

    }
    sctx->Time = lbolt;

    spin_unlock (&sctx->WaitLock);
}

static int
trapSize (int proc)
{
    switch (proc)
    {
    case DMA_PROC:	return (sizeof (DMA_TRAP));
    case THREAD_PROC:	return (sizeof (THREAD_TRAP));
    case COMMAND_PROC:	return (sizeof (COMMAND_TRAP));
    case INPUT_PROC:	return (sizeof (INPUT_TRAP));
    default:		return (0);
    }
}

static int
sys_exception (ELAN3_CTXT *ctxt, int type, int proc, void *trapp, va_list ap)
{
    SYS_CTXT *sctx  = (SYS_CTXT *) ctxt->Private;
    int	      res;

    PRINTF2 (ctxt, DBG_SYSCALL, "sys_exception: type %d proc %d\n", type, proc);

    switch (type)
    {
    case EXCEPTION_INVALID_ADDR:
    {
	E3_FaultSave_BE *faultSave = va_arg (ap, E3_FaultSave_BE *);
	int		 res       = va_arg (ap, int);
	
	sys_addException (sctx, type, proc, trapp, trapSize(proc), faultSave, res, 0);
	break;
    }
    
    case EXCEPTION_UNIMP_INSTR:
    {
	E3_uint32 instr = va_arg (ap, E3_uint32);
	
	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, instr);
	break;
    }
    
    case EXCEPTION_INVALID_PROCESS:
    {
	E3_uint32 vproc = va_arg (ap, E3_uint32);
	int       res  = va_arg (ap, int);

	switch (proc)
	{
	case DMA_PROC:
	    if (sctx->Flags & ELAN3_SYS_FLAG_DMA_BADVP)
	    {
		DMA_TRAP *trap = (DMA_TRAP *) trapp;

		if (trap->Desc.s.dma_direction != DMA_WRITE)
		    trap->Desc.s.dma_srcEvent = trap->Desc.s.dma_destEvent;

		trap->Desc.s.dma_direction       = DMA_WRITE;
		trap->Desc.s.dma_size            = 0;
		trap->Desc.s.dma_source          = (E3_Addr) 0;
		trap->Desc.s.dma_dest            = (E3_Addr) 0;
		trap->Desc.s.dma_destEvent       = (E3_Addr) 0;
		trap->Desc.s.dma_destCookieVProc = 0;
		trap->Desc.s.dma_srcCookieVProc  = 0;
		
		return (OP_IGNORE);
	    }
	    break;

	case THREAD_PROC:
	    if (sctx->Flags & ELAN3_SYS_FLAG_THREAD_BADVP)
	    {
		THREAD_TRAP *trap = (THREAD_TRAP *) trapp;

		trap->TrapBits.s.PacketAckValue = E3_PAckError;
		
		return (OP_IGNORE);
	    }
	    break;
	}
	    
	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, res, vproc);
	break;
    }
    
    case EXCEPTION_FAULTED:
    {
	E3_Addr addr = va_arg (ap, E3_Addr);

	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, addr);
	break;
    }
    
    case EXCEPTION_QUEUE_OVERFLOW:
    {
	E3_FaultSave_BE *faultSave = va_arg (ap, E3_FaultSave_BE *);
	int		 trapType  = va_arg (ap, int);
	
	sys_addException (sctx, type, proc, trapp, trapSize(proc), faultSave, 0, trapType);
	break;
    }
    
    case EXCEPTION_COMMAND_OVERFLOW:
    {
	int count = va_arg (ap, int);
	
	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, count);
	break;
    }
    
    case EXCEPTION_CHAINED_EVENT:
    {
	E3_Addr addr = va_arg (ap, E3_Addr);
	
	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, addr);
	break;
    }
    
    case EXCEPTION_DMA_RETRY_FAIL:
    case EXCEPTION_PACKET_TIMEOUT:
	if (proc != DMA_PROC)
	    sys_backoffWait (ctxt, SYS_TIMEOUT_BACKOFF);
	else
	{
	    DMA_TRAP *trap = (DMA_TRAP *) trapp;
	    
	    if (sctx->Flags & ELAN3_SYS_FLAG_DMAFAIL)
	    {
		E3_BlockCopyEvent *event;

		if (trap->Desc.s.dma_direction != DMA_WRITE)
		    trap->Desc.s.dma_srcEvent = trap->Desc.s.dma_destEvent;

		/* change the source word to be E3_EVENT_FAILED */
		if ((event = (E3_BlockCopyEvent *) elan3mmu_mainaddr (ctxt->Elan3mmu, trap->Desc.s.dma_srcEvent)) == NULL)
		{
		    sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, 0);
		    break;
		}

		suword (&event->ev_Source, E3_EVENT_FAILED);
		wmb(); mmiob();
		
		trap->Desc.s.dma_direction       = DMA_WRITE;
		trap->Desc.s.dma_size            = 0;
		trap->Desc.s.dma_source          = (E3_Addr) 0;
		trap->Desc.s.dma_dest            = (E3_Addr) 0;
		trap->Desc.s.dma_destEvent       = (E3_Addr) 0;
		trap->Desc.s.dma_destCookieVProc = 0;
		trap->Desc.s.dma_srcCookieVProc  = 0;
		
		return (OP_IGNORE);
	    }

	    if (type == EXCEPTION_DMA_RETRY_FAIL)
		sys_backoffWait (ctxt, 0);
	    else
	    {
		ELAN_LOCATION location;

		krwlock_read (&ctxt->VpLock);
		location = ProcessToLocation (ctxt, NULL, trap->Desc.s.dma_direction == DMA_WRITE ? 
					      trap->Desc.s.dma_destVProc : trap->Desc.s.dma_srcVProc, NULL);
		krwlock_done (&ctxt->VpLock);
		
		sys_backoffWait (ctxt, location.loc_node == ELAN3_INVALID_NODE ? SYS_BCAST_BACKOFF : SYS_TIMEOUT_BACKOFF);
	    }
	}
	return (OP_IGNORE);
	
    case EXCEPTION_NETWORK_ERROR:
    {
	INPUT_TRAP       *trap  = (INPUT_TRAP *) trapp;
	NETERR_RESOLVER **rvpp  = va_arg (ap, NETERR_RESOLVER **);

	ASSERT (trap->State == CTXT_STATE_NETWORK_ERROR);

	if (! (sctx->Flags & ELAN3_SYS_FLAG_NETERR) && (trap->DmaIdentifyTransaction || trap->ThreadIdentifyTransaction))
	{
	    if ((*rvpp) != (NETERR_RESOLVER *) NULL)
		res = (*rvpp)->Status;
	    else if ((res = QueueNetworkErrorResolver (ctxt, trap, rvpp)) == ESUCCESS)
	    {
		/* Successfully queued the network error resolver */
		return (OP_HANDLED);
	    }

	    /* network error resolution has failed - either a bad cookie or */
	    /* an rpc error has occured */
	    sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, res, 0);
	}
	else
	{
	    /* Must be an overlaped dma packet. Must wait long enough to
	     * ensure that the sending dma'er has tried to send the next
	     * packet and had it discarded. In the real world this should
	     * be greater than an output timeout. (About 8mSec) */
	    
	    sys_backoffWait (ctxt, SYS_NETERR_BACKOFF);
	    
	    /* set this inputter state to be ok, since we've been called 
	     * by the lwp it will lower the context filter for us, so 
	     * re-enabling the inputter,  note we don't need to execute
	     * any of the packet since the dma process will re-transmit
	     * it after receiving a nack for the next packet */
	    trap->State = CTXT_STATE_OK;
	    
	    return (OP_HANDLED);
	}
	break;
    }
    
    default:
	sys_addException (sctx, type, proc, trapp, trapSize(proc), NULL, 0, 0);
	break;
    }
    
    if (type != EXCEPTION_DEBUG)
#ifdef LINUX
#ifdef NO_NPTL
	psignal (CURPROC()->p_opptr, sctx->signal);
#else
	psignal (CURPROC()->parent, sctx->signal);
#endif
#else
	psignal (CURPROC(), sctx->signal);
#endif
    return (OP_HANDLED);
}

static int
sys_event (ELAN3_CTXT *ctxt, E3_uint32 cookie, int flag)
{
    SYS_CTXT *sctx = (SYS_CTXT *) ctxt->Private;

    cookie_fire_cookie (sctx->Table, cookie);

    return (OP_HANDLED); 
}

static void
sys_swapin (ELAN3_CTXT *ctxt)
{
    PRINTF0 (ctxt, DBG_SYSCALL, "sys_swapin\n");
}

static void
sys_swapout (ELAN3_CTXT *ctxt)
{
    PRINTF0 (ctxt, DBG_SYSCALL, "sys_swapout\n");
}

static void
sys_freePrivate (ELAN3_CTXT *ctxt)
{
    SYS_CTXT *sctx = (SYS_CTXT *) ctxt->Private;

    cookie_free_table (sctx->Table);

    kmutex_destroy (&sctx->Lock);
    spin_lock_destroy (&sctx->WaitLock);
    kcondvar_destroy (&sctx->NetworkErrorWait);

    KMEM_FREE (sctx, sizeof (SYS_CTXT));
    ctxt->Private = NULL;
}

static int
sys_checkThisDma (ELAN3_CTXT *ctxt, NETERR_FIXUP *nef, E3_DMA *dma)
{
    E3_DmaType type;
    E3_uint32  cookie;
    E3_uint32  cvproc;
    int	       ignore;
    int	       match;

    type.type = fuword_noerr ((int *) &dma->dma_type);

    if (type.s.direction == DMA_WRITE)
    {
	cookie = fuword_noerr ((int *) &dma->dma_srcCookieVProc);
	cvproc = fuword_noerr ((int *) &dma->dma_destCookieVProc);
    }
    else
    {
	cookie = fuword_noerr ((int *) &dma->dma_destCookieVProc);
	cvproc = fuword_noerr ((int *) &dma->dma_srcCookieVProc);
    }

    PRINTF5 (ctxt, DBG_NETERR, "sys_checkThisDma: dir = %d cookie = %08x cvproc = %08x CookieVProc %08x DstProcess %04x\n",
	     type.s.direction, cookie, cvproc, nef->Message.CookieVProc, nef->Message.DstProcess);

    /* A DMA matches a network errror fixup if it's going to the right place (or is a broadcast)
     * and the approriate cookie matches, except that we ignore DMA's which don't have a destEvent
     * since they don't have any atomic behaviour (though they still send the identify) */

    ignore = (type.s.direction == DMA_WRITE && cookie == 0 &&
	      fuword_noerr ((int *) &dma->dma_destEvent) == 0);
    match  = (nef->Message.CookieVProc == cookie &&
	      (nef->Message.DstProcess == (cvproc & DMA_PROCESS_MASK) || nef->Message.WaitForEop));

    PRINTF2 (ctxt, DBG_NETERR, "  -> %s %s\n", ignore ? "ignore" : match ? "matched" : "not-matched", nef->Message.WaitForEop ? "wait for eop" : "");

    if (match && !ignore && !nef->Message.WaitForEop)
    {
	PRINTF0 (ctxt, DBG_NETERR, "sys_checkThisDma: nuking the dma\n");

	/* NOTE - we access the dma descriptor backwards since it could exist in sdram */
	if (type.s.direction != DMA_WRITE)
	    suword_noerr ((int *) &dma->dma_srcEvent, 0);

	suword_noerr ((int *) &dma->dma_destEvent, 0);
	suword_noerr ((int *) &dma->dma_dest,      0);
	suword_noerr ((int *) &dma->dma_source,    0);
	suword_noerr ((int *) &dma->dma_size,      0);

	if (type.s.direction != DMA_WRITE)
	    suword_noerr ((int *) &dma->dma_type, fuword_noerr ((int *) &dma->dma_type) & E3_DMA_CONTEXT_MASK);

	wmb(); mmiob();
    }

    return (match && !ignore);
}

static int
sys_fixupNetworkError (ELAN3_CTXT *ctxt, NETERR_FIXUP *nef)
{
    SYS_CTXT       *sctx    = (SYS_CTXT *) ctxt->Private;
    SYS_SWAP_SPACE *sp      = sctx->Swap;
    int		    matched = 0;
    SYS_WORD_ITEM  *wordp;
    SYS_BLOCK_ITEM *blockp;
    label_t	    ljb;
    int		    res;

    PRINTF3 (ctxt, DBG_NETERR, "sys_fixupnetworkError %08x %08x %08x\n", 
	     nef->Message.CookieAddr, nef->Message.CookieVProc, nef->Message.NextCookie);

    if (nef->Message.CookieAddr == (E3_Addr) 0)			/* It's a DMA which requires fixing up */
    {
	kmutex_lock (&sctx->Lock);

	if (on_fault (&ljb))
	    res = EFAULT;
	else
	{
	    /* scan the dma ptr list */
	    for (wordp = (SYS_WORD_ITEM *) fuptr_noerr ((void **) &sp->ItemListsHead[LIST_DMA_PTR]);
		 wordp != NULL; 
		 wordp = (SYS_WORD_ITEM *) fuptr_noerr ((void **) &wordp->Next))
	    {
		E3_uint32 value = fuword_noerr ((int *) &wordp->Value);
		E3_DMA    *dma  = (E3_DMA *) elan3mmu_mainaddr (ctxt->Elan3mmu, value);

		PRINTF3 (ctxt, DBG_NETERR, "sys_fixupnetworkError: check block item %p Value %08x dma %p\n", wordp, value, dma);

		matched += sys_checkThisDma (ctxt, nef, dma);
	    }
	
	    /* scan the dma desc list */
	    for (blockp = (SYS_BLOCK_ITEM *) fuptr_noerr ((void **) &sp->ItemListsHead[LIST_DMA_DESC]);
		 blockp != NULL; 
		 blockp = (SYS_BLOCK_ITEM *) fuptr_noerr ((void **) &blockp->Next))
	    {
		E3_DMA *dma = (E3_DMA *) fuptr_noerr ((void *) &blockp->Pointer);
		
		PRINTF2 (ctxt, DBG_NETERR, "sys_fixupnetworkError: check block item %p Pointer %p\n", blockp, dma);

		matched += sys_checkThisDma (ctxt, nef, dma);
	    }
	    
	    /* If we've still not found it, then check the command port item */
	    /* it MUST be present as a command waiting to be executed, as */
	    /* otherwise it could have already happened and we will claim to */
	    /* have found it, but not realy */
	    if (ctxt->CommandPortItem != NULL)
	    {
		E3_DMA *dma = (E3_DMA *) fuptr_noerr ((void *) &((SYS_BLOCK_ITEM *) ctxt->CommandPortItem)->Pointer);

		if (sys_checkThisDma (ctxt, nef, dma))
		{
		    printk ("!!! it's the command port item - need to ensure that the command exists\n");
		    matched++;
		}
	    }

	    res = matched ? ESUCCESS : ESRCH;
	}
	no_fault();
	kmutex_unlock (&sctx->Lock);

	if (matched > 1)
	    ElanException (ctxt, EXCEPTION_COOKIE_ERROR, DMA_PROC, NULL, NULL, nef->Message.CookieVProc);
    }
    else							/* It's a thread which requires fixing up */
    {
	E3_int32  *cookiePtr = (E3_int32 *) elan3mmu_mainaddr (ctxt->Elan3mmu, nef->Message.CookieAddr);
	E3_uint32  curval    = fuword_noerr (cookiePtr);

	if (curval == nef->Message.CookieVProc)		/* thread doesn't think it's been done */
	{
	    if (! nef->Message.WaitForEop)
	    {
		suword_noerr (cookiePtr, nef->Message.NextCookie);
		mb(); mmiob();
	    }
	    
	    res = ESUCCESS;
	}
	else							/* thread thinks that it's been executed */
	{
	    res = ESRCH;
	}
    }
    
    CompleteNetworkErrorFixup (ctxt, nef, res);

    return (OP_HANDLED);
}


static int
sys_startFaultCheck (ELAN3_CTXT *ctxt)
{
    return (0);
}

static void
sys_endFaultCheck (ELAN3_CTXT *ctxt)
{
    wmb();
}

static E3_uint8
sys_load8 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    E3_uint8 *maddr = (E3_uint8 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    return (fubyte_noerr (maddr));
}

static void
sys_store8 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint8 val)
{
    E3_uint8 *maddr = (E3_uint8 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    subyte_noerr (maddr, val);
    wmb(); mmiob();
}

static E3_uint16
sys_load16 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    E3_uint16 *maddr = (E3_uint16 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    return (fusword_noerr (maddr));
}

static void
sys_store16 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint16 val)
{
    E3_uint16 *maddr = (E3_uint16 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    susword_noerr (maddr, val);
    wmb(); mmiob();
}

static E3_uint32
sys_load32 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    E3_uint32 *maddr = (E3_uint32 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    return (fuword_noerr (maddr));
}

static void
sys_store32 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint32 val)
{
    E3_uint32 *maddr = (E3_uint32 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    suword_noerr (maddr, val);
    wmb(); mmiob();
}

static E3_uint64
sys_load64 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    E3_uint64 *maddr = (E3_uint64 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    return (fulonglong_noerr ((long long *) maddr));
}

static void
sys_store64 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint64 val)
{
    E3_uint64 *maddr = (E3_uint64 *) elan3mmu_mainaddr (ctxt->Elan3mmu, addr);

    sulonglong_noerr ((long long *) maddr, val);
    wmb(); mmiob();
}


void
sys_addException (SYS_CTXT *sctx, int type, int proc, caddr_t trapp, int size,
		  E3_FaultSave_BE *faultSave, u_long res, u_long value)
{
    SYS_EXCEPTION      *ex_ptr;
    int			front;
    int			back;
    int			count;
    label_t		ljp;

    PRINTF4 (DBG_DEVICE, DBG_FN, "sys_addException: type %d proc %d res %ld value %ld\n",
	     type, proc, res, value);

    KMEM_ZALLOC (ex_ptr, SYS_EXCEPTION *, sizeof  (SYS_EXCEPTION), TRUE);

    if (ex_ptr != NULL)
    {
	bzero ((caddr_t) ex_ptr, sizeof (SYS_EXCEPTION));

	ex_ptr->Type  = type;
	ex_ptr->Proc  = proc;
	ex_ptr->Res   = res;
	ex_ptr->Value = value;
	
	if (trapp && size)
	    bcopy (trapp, (caddr_t) &ex_ptr->Union, size);
	if (faultSave)
	    bcopy ((caddr_t) faultSave, (caddr_t) &ex_ptr->FaultArea, sizeof (E3_FaultSave_BE));
    }

    kmutex_lock (&sctx->Lock);
    if (! on_fault (&ljp))
    {
	front = fuword_noerr (&sctx->Exceptions->Front);
	back  = fuword_noerr (&sctx->Exceptions->Back);
	count = fuword_noerr (&sctx->Exceptions->Count);

	if (count <= 0 || front < 0 || back < 0 || front >= count || back >= count)
	    suword_noerr (&sctx->Exceptions->Overflow, fuword_noerr (&sctx->Exceptions->Overflow) + 1);
	else if (((front+1) % count ) == back)
	    suword_noerr (&sctx->Exceptions->Overflow, fuword_noerr (&sctx->Exceptions->Overflow) + 1);
	else
	{
	    if (ex_ptr != NULL)
		copyout_noerr ((caddr_t) ex_ptr, (caddr_t) &sctx->Exceptions->Exceptions[front], sizeof (SYS_EXCEPTION));
	    else
	    {
		suword_noerr (&sctx->Exceptions->Exceptions[front].Type, EXCEPTION_ENOMEM);
		suword_noerr (&sctx->Exceptions->Exceptions[front].Proc, 0);
	    }
	    suword_noerr (&sctx->Exceptions->Front, (front + 1) % count);
	}

	/* always reset the magic number in case it's been overwritten */
	/* so that 'edb' can find the exception page in the core file */
	suword_noerr (&sctx->Exceptions->Magic, SYS_EXCEPTION_MAGIC);
    }
    no_fault();
    kmutex_unlock (&sctx->Lock);
    
    if (ex_ptr != NULL)
	KMEM_FREE (ex_ptr, sizeof  (SYS_EXCEPTION));
}

int
sys_getException (SYS_CTXT *sctx, SYS_EXCEPTION *ex)
{
    int     front;
    int     back;
    int     count;
    int     res;
    label_t ljp;

    if (sctx->Exceptions == NULL)
	return (EINVAL);

    kmutex_lock (&sctx->Lock);
    if (on_fault (&ljp))
    {
	no_fault();
	kmutex_unlock (&sctx->Lock);
	return (EFAULT);
    }
    
    front = fuword_noerr (&sctx->Exceptions->Front);
    back  = fuword_noerr (&sctx->Exceptions->Back);
    count = fuword_noerr (&sctx->Exceptions->Count);

    if (count <= 0 || front < 0 || back < 0 || front >= count || back >= count || back == front)
	res = EINVAL;
    else
    {
	copyin_noerr ((caddr_t) &sctx->Exceptions->Exceptions[back], (caddr_t) ex, sizeof (SYS_EXCEPTION));
	suword_noerr (&sctx->Exceptions->Back, (back+1) % count);

	res = ESUCCESS;
    }
    no_fault();
    kmutex_unlock (&sctx->Lock);

    return (res);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
