/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: iproc.c,v 1.47 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/iproc.c,v $ */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/trtype.h>
#include <elan3/vmseg.h>


static int TrSizeTable[] = {0, 8, 16, 32, 64};

static void  ConvertTransactionToSetEvent (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_Addr Addr);
static void  SimulateBlockWrite  (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap);
static void  SimulateWriteWord   (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap);
static void  SimulateWriteDWord  (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap);
static void  SimulateTraceRoute  (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap);
static void  BumpInputterStats (ELAN3_DEV *dev, E3_IprocTrapHeader_BE *hdrp);

void
HandleIProcTrap (ELAN3_DEV           *dev, 
		 int		     Channel,
		 E3_uint32	     Pend,
		 sdramaddr_t	     FaultSaveOff,
		 sdramaddr_t	     TransactionsOff,
		 sdramaddr_t	     DataOff)
{
    E3_IprocTrapHeader_BE Transaction0;
    ELAN3_CTXT	         *ctxt;
    INPUT_TRAP           *trap;
    register int          i;

    /*
     * Read the 1st set of transactions, so we can determine the 
     * context for the trap 
     */
    elan3_sdram_copyq_from_sdram (dev, TransactionsOff, (void *) &Transaction0, 16);
    
    BumpStat (dev, IProcTraps);
    BumpInputterStats (dev, &Transaction0);

    if (Transaction0.s.TrTypeCntx.s.TypeCntxInvalid)
    {
	/*
	 * The context is not valid. This will occur if the packet
	 * trapped for an EopError with no IdentTrans or an error corrupted the context
	 * giving a CRC error on the first transaction and the Ack had not been returned.
	 */
	if (Transaction0.s.TrTypeCntx.s.LastTrappedTrans)
	{
	    PRINTF0 (DBG_DEVICE, DBG_IPROC, "iproc: Error on EOP without a good context, ignoring trap\n");
	}
	else
	{
	    /* Check that only crap has been received.  If not then die. */
	    if (! Transaction0.s.IProcTrapStatus.s.BadLength &&
		(Transaction0.s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_GOOD)
	    {
		printk ("iproc: Did not have a valid context for the trap area.\n");
		printk ("iproc: TrTypeCntx=%x TrAddr=%x TrData0=%x IProcTrapStatus=%x\n",
			 Transaction0.s.TrTypeCntx.TypeContext, Transaction0.s.TrAddr,
			 Transaction0.s.TrData0, Transaction0.s.IProcTrapStatus.Status);
		panic ("elan3: iproc did not have a valid context");
		/* NOTREACHED */
	    }
	    PRINTF0 (DBG_DEVICE, DBG_IPROC, "iproc: First transaction is bad, ignoring trap\n");
	}
    }
    else
    {
	ctxt = ELAN3_DEV_CTX_TABLE(dev, Transaction0.s.TrTypeCntx.s.Context);
	
	if (ctxt == NULL)
	{
	    PRINTF1 (DBG_DEVICE, DBG_INTR, "HandleIProcTrap: context %x invalid\n", 
		     Transaction0.s.TrTypeCntx.s.Context);

	    BumpStat (dev, InvalidContext);
	}
	else
	{
	    trap = (Channel == 0) ? &ctxt->Input0Trap : &ctxt->Input1Trap;

	    ASSERT (trap->State == CTXT_STATE_OK);
	    
	    trap->Transactions[0] = Transaction0;

	    PRINTF1 (ctxt, DBG_INTR, "HandleIProcTrap: %s\n", IProcTrapString (&trap->Transactions[0], NULL));
	    /*
	     * Copy the rest of the transactions into the trap area.
	     */
	    for (i = 0; !(trap->Transactions[i].s.TrTypeCntx.s.LastTrappedTrans);)
	    {
		if (++i >= MAX_TRAPPED_TRANS)
		{
		    trap->Overflow = 1;
		    break;
		}

		elan3_sdram_copyq_from_sdram (dev, TransactionsOff + i*sizeof (E3_IprocTrapHeader), (void *) &trap->Transactions[i], 16);

		PRINTF1 (ctxt, DBG_INTR, "                 %s\n", IProcTrapString (&trap->Transactions[i], NULL));

		BumpInputterStats (dev, &trap->Transactions[i]);
	    }
	    
	    /*
	     * Remember the number of transactions we've copied.
	     */
	    trap->NumTransactions = i+1;

	    PRINTF1 (ctxt, DBG_INTR, "                 NumTransactions = %d\n", trap->NumTransactions);
	    
	    /*
	     * Copy all the data blocks in one go to let the Elan prefetcher work 
	     */
	    elan3_sdram_copyq_from_sdram (dev, DataOff, trap->DataBuffers, trap->NumTransactions*sizeof (E3_IprocTrapData));

	    /*
	     * Copy fault save area and clear out for next time round.
	     */
	    elan3_sdram_copyq_from_sdram (dev, FaultSaveOff, (void *) &trap->FaultSave, 16);
	    elan3_sdram_zeroq_sdram (dev, FaultSaveOff, 16);

	    if (ELAN3_OP_IPROC_TRAP (ctxt, trap, Channel) == OP_DEFER)
	    {
		/*
		 * Mark the trap as valid and set the inputter state to 
		 * raise the context filter.
		 */
		trap->State = CTXT_STATE_TRAPPED;
		kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
		
		SetInputterStateForContext (ctxt, Pend, NULL);
	    }
	}
    }
}

void
InspectIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap)
{
    int	      i;
    int	      StatusValid;

    trap->AckSent		    = 0;
    trap->BadTransaction            = 0;
    
    trap->TrappedTransaction        = NULL;
    trap->TrappedDataBuffer	    = NULL;
    trap->WaitForEopTransaction     = NULL;
    trap->WaitForEopDataBuffer      = NULL;
    trap->DmaIdentifyTransaction    = NULL;
    trap->ThreadIdentifyTransaction = NULL;
    trap->LockQueuePointer          = (E3_Addr) 0;
    trap->UnlockQueuePointer        = (E3_Addr) 0;

    /*
     * Now scan all the transactions received 
     */
    for (i = 0; i < trap->NumTransactions ; i++)
    {
	E3_IprocTrapHeader_BE *hdrp = &trap->Transactions[i];
	E3_IprocTrapData_BE   *datap = &trap->DataBuffers[i];

	StatusValid = hdrp->s.TrTypeCntx.s.StatusRegValid != 0;
	
	if (StatusValid && hdrp->s.IProcTrapStatus.s.AckSent)	/* Remember if we've sent the ack back */
	    trap->AckSent = 1;
	
	if (hdrp->s.TrTypeCntx.s.LastTrappedTrans)		/* Check for EOP */
	{
	    ASSERT (i == trap->NumTransactions - 1);

	    switch (hdrp->s.IProcTrapStatus.Status & E3_IPS_EopType)
	    {
	    case EOP_GOOD:
		/* if we get an EOP_GOOD then the outputer should have received a PAckOk. */  
		/* unless it was a flood, in which case someone must have sent an ack */
		/* but not necessarily us */
		break;

	    case EOP_BADACK:
		BumpUserStat (ctxt, EopBadAcks);

		/* if we get an EOP_BADACK then the outputer did not receive a PAckOk even if
		 * we sent a PAckOk. We can clear tinfo.AckSent. */
		if (trap->AckSent == 1)
		{
		    PRINTF0 (ctxt, DBG_IPROC, "InspectIProcTrap: Network error destroyed PAckOk\n");
		    trap->AckSent = 0;
		}
		break;

	    case EOP_ERROR_RESET:
		BumpUserStat (ctxt, EopResets);

		/* if we get an EOP_ERROR_RESET then the outputer may or may not have got a PAckOk. */
		trap->BadTransaction = 1;
		break;

	    default:
		panic ("InspectIProcTrap: invalid EOP type in status register\n");
		/* NOTREACHED */
	    }
	    continue;
	}

	PRINTF2 (ctxt, DBG_IPROC, "InspectIProcTrap: %2d: %s\n", i, IProcTrapString (hdrp, datap));
	
	if (! StatusValid)					/* We're looking at transactions stored before the trap */
	{							/* these should only be identifies and lock transactions */

	    if (hdrp->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT)
		panic ("InspectIProcTrap: writeblock transaction found in input trap header before trap occured\n");

	    switch (hdrp->s.TrTypeCntx.s.Type & TR_OPCODE_TYPE_MASK)
	    {
	    case TR_LOCKQUEUE & TR_OPCODE_TYPE_MASK:
		if (trap->LockQueuePointer)				/* Already seen a LOCKQUEUE transaction in this packet, */
		{						/* the user program should not have done this !! */
		    ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		    return;
		}

		trap->LockQueuePointer = (E3_Addr) hdrp->s.TrAddr;	/* Remember the queue pointer in case we need to unlock it */
		break;

	    case TR_DMAIDENTIFY & TR_OPCODE_TYPE_MASK:
		if (trap->DmaIdentifyTransaction ||		/* Already seen an identify transaction in this packet */
		    trap->ThreadIdentifyTransaction)		/* the user program should not have done this */
		{							
		    ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		    return;
		}
		trap->DmaIdentifyTransaction = hdrp;
		break;

	    case TR_THREADIDENTIFY & TR_OPCODE_TYPE_MASK:
		if (trap->DmaIdentifyTransaction ||		/* Already seen an identify transaction in this packet */
		    trap->ThreadIdentifyTransaction)		/* the user program should not have done this */
		{							
		    ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		    return;
		}
		trap->ThreadIdentifyTransaction = hdrp;
		break;
		
	    default:
		panic ("InspectIProcTrap: invalid transaction found in input trap header before trap occured\n");
		/* NOTREACHED */
	    }
	    continue;
	}

	if (StatusValid && trap->TrappedTransaction == NULL)	/* Remember the transaction which caused the */
	{							/* trap */
	    trap->TrappedTransaction = hdrp;
	    trap->TrappedDataBuffer  = datap;
	}

	if(hdrp->s.IProcTrapStatus.s.BadLength ||
	   ((hdrp->s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_ERROR) ||
	   ((hdrp->s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_BAD))
	{
	    int	j;
	    PRINTF0 (ctxt, DBG_IPROC, "InspectIProcTrap: transaction has a bad crc\n");
	    for (j=0; j<TRANS_DATA_WORDS; j+=4)
	       PRINTF5 (ctxt, DBG_IPROC, "InspectIProcTrap: Data %0d %8x %8x %8x %8x\n",
	       		j, datap->TrData[j], datap->TrData[j+1], datap->TrData[j+2], datap->TrData[j+3]);
	    trap->BadTransaction = 1;
	    continue;
	}
	
	/* No more to do if it's a writeblock transaction */
	if (hdrp->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT)
	    continue;

	
	if (GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus) == MI_InputDoTrap &&
	    (hdrp->s.TrTypeCntx.s.Type & TR_WAIT_FOR_EOP) != 0)
	{
	    /*
	     * This is a wait for eop transaction that has trapped because the inputer
	     * then received a EopError. The next transaction saved should always be an
	     * EopError.
	     */
	    PRINTF0 (ctxt, DBG_IPROC, "InspectIProcTrap: got a trapped WaitForEop transaction due to EopError\n");
	    
	    trap->WaitForEopTransaction = hdrp;
	    trap->WaitForEopDataBuffer  = datap;
	    continue;
	}

	switch (hdrp->s.TrTypeCntx.s.Type & TR_OPCODE_TYPE_MASK)
	{
	case TR_UNLOCKQUEUE & TR_OPCODE_TYPE_MASK:
	    if (trap->UnlockQueuePointer)
	    {
		ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		return;
	    }
	    trap->UnlockQueuePointer = (E3_Addr) hdrp->s.TrAddr;
	    break;
	}
    }
}

void
ResolveIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, NETERR_RESOLVER **rvpp)
{
    ELAN3_DEV     *dev = ctxt->Device;
    int           res;
    unsigned long flags;

    ASSERT (! CTXT_IS_KERNEL (ctxt));

    BumpUserStat (ctxt, IProcTraps);

    InspectIProcTrap (ctxt, trap);

    /*
     * fixup page fault if we've trapped because of one.
     */
    if (trap->FaultSave.s.FaultContext != 0)
    {
	/*
	 * If it's a WRITEBLOCK transaction, then see if we remember faulting
	 * before it, and try and prefault in a sensible amount past it.
	 */
	int                fixedFault = FALSE;
	INPUT_FAULT_SAVE  *entry;
	INPUT_FAULT_SAVE **predp;
	int                npages;

	if ((trap->TrappedTransaction->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT) != 0 &&	/* a DMA packet */
	    trap->LockQueuePointer == (E3_Addr) 0 &&					/* but not a queueing DMA */
	    trap->TrappedTransaction->s.TrAddr != 0)					/* and not a DMA to 0 */
	{
	    spin_lock (&ctxt->InputFaultLock);
	    
	    for (predp = &ctxt->InputFaultList; (entry = *predp)->Next != NULL ; predp = &entry->Next)
	    {
		if (entry->Addr == trap->TrappedTransaction->s.TrAddr)
		    break;
	    }
	    
	    *predp = entry->Next;
	    entry->Next = ctxt->InputFaultList;
	    ctxt->InputFaultList = entry;
	    
	    if (entry->Addr == trap->TrappedTransaction->s.TrAddr)
	    {
		if ((entry->Count <<= 1) > MAX_INPUT_FAULT_PAGES)
		    entry->Count = MAX_INPUT_FAULT_PAGES;
	    }
	    else
	    {
		entry->Count = MIN_INPUT_FAULT_PAGES;
	    }
	    
	    entry->Addr = trap->TrappedTransaction->s.TrAddr + (entry->Count * PAGESIZE);
	    npages = entry->Count;
	    
	    spin_unlock (&ctxt->InputFaultLock);
	    
	    if (elan3_pagefault (ctxt, &trap->FaultSave, npages) != ESUCCESS)
	    {
		PRINTF2 (ctxt, DBG_IPROC, "ResolveIProcTrap: pagefaulting %d pages at %08x - failed\n", 
			 npages, trap->TrappedTransaction->s.TrAddr);
	    }
	    else
	    {
		PRINTF2 (ctxt, DBG_IPROC, "ResolveIProcTrap: pagefaulting %d pages at %08x - succeeded\n", 
			 npages, trap->TrappedTransaction->s.TrAddr);
		
		fixedFault = TRUE;
	    }
	}

	/* Workaround WRITEBLOCK transaction executed when LOCKQUEUE transaction missed */
	/* the packet will have been nacked */
	if ((trap->TrappedTransaction->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT) && 	/* a DMA packet */
	    trap->LockQueuePointer == 0 && trap->UnlockQueuePointer && 			/* a queueing DMA */
	    trap->TrappedTransaction->s.TrAddr == trap->FaultSave.s.FaultAddress)	/* and missed lockqueue */
	{
	    fixedFault = TRUE;
	}

	if (! fixedFault)
	{
	    if ((res = elan3_pagefault (ctxt, &trap->FaultSave, 1)) != ESUCCESS)
	    {
		PRINTF1 (ctxt, DBG_IPROC, "ResolveIProcTrap: elan3_pagefault failed at %x\n", 
			 trap->FaultSave.s.FaultAddress);
		ElanException (ctxt, EXCEPTION_INVALID_ADDR, INPUT_PROC, trap, &trap->FaultSave, res);
		return;
	    }
	}
    }

    if (! trap->AckSent && trap->LockQueuePointer)			/* Queued DMA */
    {									/* The ack was not sent, so the queue will be locked. */
	SimulateUnlockQueue (ctxt, trap->LockQueuePointer, FALSE);	/* We must unlock it. */
    }

    if (trap->AckSent && trap->BadTransaction)
    {
	if (trap->DmaIdentifyTransaction)
	{
	    PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: Dma identify needs network resultion\n");

	    BumpStat (dev, DmaIdentifyNetworkErrors);
	    BumpUserStat (ctxt, DmaIdentifyNetworkErrors);

	    if (trap->WaitForEopTransaction)
		PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: have delayed wait for eop transaction\n");
	}
	else if (trap->ThreadIdentifyTransaction)
	{
	    PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: Thread identify needs network resolution\n");

	    BumpStat (dev, ThreadIdentifyNetworkErrors);
	    BumpUserStat (ctxt, ThreadIdentifyNetworkErrors);

	    if (trap->WaitForEopTransaction)
		PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: have delayed wait for eop transaction\n");
	}
	else
	{
	    BumpStat (dev, DmaNetworkErrors);
	    BumpUserStat (ctxt, DmaNetworkErrors);
	}
    }

    spin_lock_irqsave (&dev->IntrLock, flags);
    
    if (! trap->AckSent)
    {
	PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: ack not sent, lowering context filter\n");

	trap->State = CTXT_STATE_OK;
    }
    else
    {
	if (trap->BadTransaction)
	{
	    PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: ack sent, waiting on bad transaction\n");
	    trap->State = CTXT_STATE_NETWORK_ERROR;
	}
	else
	{
	    PRINTF0 (ctxt, DBG_IPROC, "ResolveIProcTrap: ack sent, waiting on packet to be re-executed\n");
	    trap->State = CTXT_STATE_NEEDS_RESTART;
	}
    }

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    if (trap->AckSent && trap->BadTransaction)
	ElanException (ctxt, EXCEPTION_NETWORK_ERROR, INPUT_PROC, trap, rvpp);
}

int
RestartIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap)
{
    PRINTF1 (ctxt, DBG_IPROC, "RestartIProc: %d transactions\n", trap->NumTransactions);

    if (trap->TrappedTransaction == NULL)			/* No transaction trapped - probably a network */
	return (ESUCCESS);					/* error */

    while (! trap->TrappedTransaction->s.TrTypeCntx.s.LastTrappedTrans)
    {
	E3_IprocTrapHeader_BE *hdrp = trap->TrappedTransaction;
	E3_IprocTrapData_BE   *datap = trap->TrappedDataBuffer;
	
	ASSERT (hdrp->s.TrTypeCntx.s.StatusRegValid != 0);

	PRINTF2 (ctxt, DBG_IPROC, "RestartIProc: TrType=0x%x Status=0x%x\n",
		 hdrp->s.TrTypeCntx.TypeContext, hdrp->s.IProcTrapStatus.Status);
	
	if ((hdrp->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT) != 0)
	{
	    PRINTF1 (ctxt, DBG_IPROC, "RestartIProc: WRITEBLOCK : Addr %x\n", hdrp->s.TrAddr);
	    SimulateBlockWrite (ctxt, hdrp, datap);
	}
	else
	{
	    switch (hdrp->s.TrTypeCntx.s.Type & TR_OPCODE_TYPE_MASK)
	    {
	    case TR_SETEVENT & TR_OPCODE_TYPE_MASK:
		PRINTF1 (ctxt, DBG_IPROC, "RestartIProc: SETEVENT : %x\n", hdrp->s.TrAddr);

		if (GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus) != MI_InputDoTrap)
		    FixupEventTrap (ctxt, INPUT_PROC, trap, GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus), &trap->FaultSave, FALSE);
		else if (hdrp->s.TrAddr)
		{
		    if (IssueCommand (ctxt, offsetof (E3_CommandPort, SetEvent), hdrp->s.TrAddr, FALSE) != ISSUE_COMMAND_OK)
			return (EAGAIN);
		}
		break;

	    case TR_WRITEWORD & TR_OPCODE_TYPE_MASK:
		SimulateWriteWord (ctxt, hdrp, datap);
		break;

	    case TR_WRITEDOUBLEWORD & TR_OPCODE_TYPE_MASK:
		SimulateWriteDWord (ctxt, hdrp, datap);
		break;
		
	    case TR_UNLOCKQUEUE & TR_OPCODE_TYPE_MASK:
		if (GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus) == MI_InputDoTrap)
		    ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		else
		{
		    switch (GET_STATUS_TRAPTYPE (hdrp->s.IProcTrapStatus))
		    {
		    case MI_WaitForUnLockDescRead:
			/*
			 * Fault occured on the read of the queue descriptor - since the ack
			 * has been sent we need to move the queue on one slot.
			 */
			PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: TR_UNLOCKQUEUE : desc read fault\n");

			SimulateUnlockQueue (ctxt, trap->LockQueuePointer, TRUE);
			
			if (IssueCommand (ctxt, offsetof (E3_CommandPort, SetEvent),
					  hdrp->s.TrAddr + E3_QUEUE_EVENT_OFFSET, FALSE) != ISSUE_COMMAND_OK)
			{
			    /* Failed to issue setevent to complete queue unlock, since we've already unlocked */
			    /* the queue, we should "convert" this transaction into a setevent transaction that */
			    /* hasn't trapped */
			    PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: could not issue setevent for SimulateUnlockQueue\n");

			    ConvertTransactionToSetEvent (ctxt, hdrp, hdrp->s.TrAddr + E3_QUEUE_EVENT_OFFSET);
			    return (EAGAIN);
			}
			break;
			
		    case MI_DoSetEvent:
			/*
			 * Fault occured on either the write to unlock the queue or during 
			 * processing of the event.  Test the fault address against the
			 * queue address to find out which - in this case, since the ack
			 * has been sent we need to move the queue on one slot.
			 */
			if (trap->FaultSave.s.FaultAddress == trap->LockQueuePointer)
			{
			    PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: fixed unlock queue write to unlock fault\n");

			    SimulateUnlockQueue (ctxt, trap->LockQueuePointer, TRUE);
			    
			    if (IssueCommand (ctxt, offsetof (E3_CommandPort, SetEvent),
					      hdrp->s.TrAddr + E3_QUEUE_EVENT_OFFSET, FALSE) != ISSUE_COMMAND_OK)
			    {
				/* Failed to issue setevent to complete queue unlock, since we've already unlocked */
				/* the queue, we should "convert" this transaction into a setevent transaction that */
				/* hasn't trapped */
				PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: could not issue setevent for SimulateUnlockQueue\n");
				
				ConvertTransactionToSetEvent (ctxt, hdrp, hdrp->s.TrAddr + E3_QUEUE_EVENT_OFFSET);
				return (EFAIL);
			    }
			    break;
			}
			/*DROPTHROUGH*/
			
		    default:
			FixupEventTrap (ctxt, INPUT_PROC, trap, GET_STATUS_TRAPTYPE (hdrp->s.IProcTrapStatus),
					&trap->FaultSave, FALSE);
			break;
		    }
		    trap->LockQueuePointer = trap->UnlockQueuePointer = 0;
		}
		break;

	    case TR_SENDDISCARD & TR_OPCODE_TYPE_MASK:
		/* Just ignore send-discard transactions */
		PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: ignore SENDDISCARD\n");
		break;

	    case TR_REMOTEDMA & TR_OPCODE_TYPE_MASK:
		PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: REMOTEDMA\n");		

		/* modify the dma type since it will still be a "read" dma */
		((E3_DMA_BE *) datap)->s.dma_type &= ~(DMA_TYPE_READ | E3_DMA_CONTEXT_MASK);
		((E3_DMA_BE *) datap)->s.dma_type |= DMA_TYPE_ISREMOTE;

		RestartDmaDesc (ctxt, (E3_DMA_BE *) datap);
		break;

	    case TR_TRACEROUTE & TR_OPCODE_TYPE_MASK:
		PRINTF0 (ctxt, DBG_IPROC, "RestartIProc: TRACEROUTE\n");
		SimulateTraceRoute (ctxt, hdrp, datap);
		break;

	    default:
		ElanException (ctxt, EXCEPTION_BAD_PACKET, INPUT_PROC, trap);
		break;
	    }
	}

	/*
	 * We've successfully processed this transaction, so move onto the 
	 * next one.
	 */
	trap->TrappedTransaction++;
	trap->TrappedDataBuffer++;
    }
    
    return (ESUCCESS);
}

static void
ConvertTransactionToSetEvent (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_Addr Addr)
{
    hdrp->s.TrTypeCntx.s.Type           = TR_SETEVENT;
    hdrp->s.TrTypeCntx.s.StatusRegValid = 0;
    hdrp->s.TrAddr                      = Addr;
}

void
SimulateBlockWrite (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap)
{
    void     *saddr  = (void *) ((unsigned long) datap + (hdrp->s.TrAddr & 0x3f));
    unsigned  nbytes = (hdrp->s.TrTypeCntx.s.Type) & TR_PARTSIZE_MASK;
    int       i;

    if (nbytes == 0)
	nbytes = sizeof (E3_IprocTrapData_BE);

    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	PRINTF1 (ctxt, DBG_IPROC, "SimulateBlockWrite: faulted at %x\n", hdrp->s.TrAddr);
	ElanException (ctxt, EXCEPTION_FAULTED, INPUT_PROC, NULL, hdrp->s.TrAddr);
	return;
    }

    /*
     * NOTE: since the block copy could be to sdram, we issue the writes backwards,
     *       except we MUST ensure that the last item in the block is written last.
     */
    switch (((hdrp->s.TrTypeCntx.s.Type) >> TR_TYPE_SHIFT) & TR_TYPE_MASK)
    {
    case TR_TYPE_BYTE:						/* 8 bit */
	for (i = nbytes - (2*sizeof (E3_uint8)); i >= 0; i -= sizeof (E3_uint8))
	    ELAN3_OP_STORE8 (ctxt, hdrp->s.TrAddr + i, ((E3_uint8 *) saddr)[i]);
	i = nbytes - sizeof (E3_uint8);
	ELAN3_OP_STORE8 (ctxt, hdrp->s.TrAddr + i, ((E3_uint8 *) saddr)[i]);
	break;
	
    case TR_TYPE_SHORT:						/* 16 bit */
	for (i = nbytes - (2*sizeof (E3_uint16)); i >= 0; i -= sizeof (E3_uint16))
	ELAN3_OP_STORE16 (ctxt, hdrp->s.TrAddr + i, ((E3_uint16 *) saddr)[i]);
	i = nbytes - sizeof (E3_uint16);
	ELAN3_OP_STORE16 (ctxt, hdrp->s.TrAddr + i, ((E3_uint16 *) saddr)[i]);
	break;
	
    case TR_TYPE_WORD:						/* 32 bit */
	for (i = nbytes - (2*sizeof (E3_uint32)); i >= 0; i -= sizeof (E3_uint32))
	    ELAN3_OP_STORE32 (ctxt, hdrp->s.TrAddr + i, ((E3_uint32 *) saddr)[i]);
	i = nbytes - sizeof (E3_uint32);
	ELAN3_OP_STORE32 (ctxt, hdrp->s.TrAddr + i, ((E3_uint32 *) saddr)[i]);
	break;
	
    case TR_TYPE_DWORD:						/* 64 bit  */
	for (i = nbytes - (2*sizeof (E3_uint64)); i >= 0; i -= sizeof (E3_uint64))
	    ELAN3_OP_STORE64 (ctxt, hdrp->s.TrAddr + i, ((E3_uint64 *) saddr)[i]);
	i = nbytes - sizeof (E3_uint64);
	ELAN3_OP_STORE64 (ctxt, hdrp->s.TrAddr + i, ((E3_uint64 *) saddr)[i]);
	break;
    }
    ELAN3_OP_END_FAULT_CHECK (ctxt);
}

void
SimulateWriteWord (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap)
{
    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	PRINTF1 (ctxt, DBG_IPROC, "SimulateWriteWord: faulted at %x\n", hdrp->s.TrAddr);
	ElanException (ctxt, EXCEPTION_FAULTED, INPUT_PROC, NULL, hdrp->s.TrAddr);
	return;
    }

    ELAN3_OP_STORE32 (ctxt, hdrp->s.TrAddr, ((E3_uint32 *) datap)[WordEndianFlip]);
    ELAN3_OP_END_FAULT_CHECK (ctxt);
}

void
SimulateWriteDWord (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap)
{
    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	PRINTF1 (ctxt, DBG_IPROC, "SimulateWriteDWord: faulted at %x\n", hdrp->s.TrAddr);
	ElanException (ctxt, EXCEPTION_FAULTED, INPUT_PROC, NULL, hdrp->s.TrAddr);
	return;
    }

    ELAN3_OP_STORE64 (ctxt, hdrp->s.TrAddr, ((E3_uint64 *) datap)[0]);
    ELAN3_OP_END_FAULT_CHECK (ctxt);
}

void
SimulateTraceRoute (ELAN3_CTXT *ctxt, E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap)
{
    E3_uint32 *saddr  = (E3_uint32 *) ((unsigned long) datap + (hdrp->s.TrAddr & 0x3f));
    unsigned   nwords = TrSizeTable[(hdrp->s.TrTypeCntx.s.Type >> TR_SIZE_SHIFT) & TR_SIZE_MASK] / sizeof (E3_uint32);
    int        i;

    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	PRINTF1 (ctxt, DBG_IPROC, "SimulateTraceRoute: faulted at %x\n", hdrp->s.TrAddr);
	ElanException (ctxt, EXCEPTION_FAULTED, INPUT_PROC, NULL, hdrp->s.TrAddr);
	return;
    }
    
    for (i = nwords-2; i >= 0; i--)
	ELAN3_OP_STORE32 (ctxt, hdrp->s.TrAddr + (i * sizeof (E3_uint32)), saddr[i ^ WordEndianFlip]);

    i = nwords-1;
    ELAN3_OP_STORE32 (ctxt, hdrp->s.TrAddr + (i * sizeof (E3_uint32)), saddr[i ^ WordEndianFlip]);

    ELAN3_OP_END_FAULT_CHECK (ctxt);
}

void
SimulateUnlockQueue (ELAN3_CTXT *ctxt, E3_Addr QueuePointer, int SentAck)
{
    E3_uint32 QueueLock;
    E3_Addr   QueueBPTR;
    E3_Addr   QueueFPTR;
    E3_uint64 QueueStateAndBPTR;

    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	PRINTF1 (ctxt, DBG_IPROC, "UnlockQueue: faulted with QueuePointer %x\n", QueuePointer);
	ElanException (ctxt, EXCEPTION_FAULTED, INPUT_PROC, NULL, QueuePointer);
	return;
    }
    
    if (SentAck)
    {
	QueueBPTR = ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_bptr));
	QueueFPTR = ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_fptr));

	if (QueueBPTR == ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_top)))	/* move on back pointer */
	    QueueBPTR = ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_base));
	else
	    QueueBPTR += ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_size));
	
	QueueLock = ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_state));

	if (QueueBPTR == QueueFPTR)				/* and set full bit if fptr == bptr */
	    QueueLock |= E3_QUEUE_FULL;
	
	QueueLock &= ~E3_QUEUE_LOCKED;
	
	QueueStateAndBPTR = (E3_uint64)QueueLock << 32 | QueueBPTR;

	ELAN3_OP_STORE64 (ctxt, QueuePointer + offsetof (E3_Queue, q_state), QueueStateAndBPTR);
    }
    else
    {
	QueueLock = ELAN3_OP_LOAD32 (ctxt, QueuePointer + offsetof (E3_Queue, q_state));

	QueueLock &= ~E3_QUEUE_LOCKED;
	
	ELAN3_OP_STORE32 (ctxt, QueuePointer + offsetof (E3_Queue, q_state), QueueLock);
    }

    no_fault();
}

static void
BumpInputterStats (ELAN3_DEV *dev, E3_IprocTrapHeader_BE *hdrp)
{
    if (hdrp->s.TrTypeCntx.s.LastTrappedTrans)			/* EOP */
    {
	switch (hdrp->s.IProcTrapStatus.Status & E3_IPS_EopType)
	{
	case EOP_BADACK:
	    BumpStat (dev, EopBadAcks);
	    break;
	case EOP_ERROR_RESET:
	    BumpStat (dev, EopResets);
	    break;
	}
    }
    else if (hdrp->s.TrTypeCntx.s.StatusRegValid)
    {
	/*
	 * Errors are tested in order of badness. i.e. badlength will prevent a BadCrc and so on...
	 */
	if (hdrp->s.IProcTrapStatus.s.BadLength)
	    BumpStat (dev, InputterBadLength);
	else if ((hdrp->s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_BAD)
	    BumpStat (dev, InputterCRCBad);
	else if ((hdrp->s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_ERROR)
	    BumpStat (dev, InputterCRCErrors);
	else if ((hdrp->s.IProcTrapStatus.Status & CRC_MASK) == CRC_STATUS_DISCARD)
	    BumpStat (dev, InputterCRCDiscards);
    }
}

char *
IProcTrapString (E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData_BE *datap)
{
    static char buffer[256];
    static char typeString[256];
    static char statusString[256];
    char *ptr;
    E3_Addr     Addr        = hdrp->s.TrAddr;
    E3_uint32   Type        = hdrp->s.TrTypeCntx.s.Type;
    E3_uint32   Context     = hdrp->s.TrTypeCntx.s.Context;
    E3_uint32   StatusValid = hdrp->s.TrTypeCntx.s.StatusRegValid;
    
    if (hdrp->s.TrTypeCntx.s.LastTrappedTrans)
    {
	switch (hdrp->s.IProcTrapStatus.Status & E3_IPS_EopType)
	{
	case EOP_GOOD:		sprintf (typeString, "EOP GOOD"); break;
	case EOP_BADACK:	sprintf (typeString, "EOP BADACK"); break;
	case EOP_ERROR_RESET:	sprintf (typeString, "EOP ERROR RESET"); break;
	default:		sprintf (typeString, "EOP - bad status"); break;
	}
	sprintf (buffer, "%15s Cntx=%08x", typeString, Context);
    }
    else
    {
	if (Type & TR_WRITEBLOCK_BIT)
	{
	    switch ((Type >> TR_TYPE_SHIFT) & TR_TYPE_MASK)
	    {
	    case TR_TYPE_BYTE:	ptr = "Byte";    break;
	    case TR_TYPE_SHORT:	ptr = "Short";   break;
	    case TR_TYPE_WORD:	ptr = "Word";    break;
	    case TR_TYPE_DWORD:	ptr = "Double";  break;
	    default:		ptr = "Unknown"; break;
	    }
	    
	    sprintf (typeString, "WriteBlock Type=%s Size=%2d", ptr, Type & TR_PARTSIZE_MASK);
	}
	else
	{
	    switch (Type & TR_OPCODE_TYPE_MASK)
	    {
	    case TR_SETEVENT & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Setevent"); break;
	    case TR_REMOTEDMA & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Remote DMA"); break;
	    case TR_LOCKQUEUE & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Lock Queue"); break;
	    case TR_UNLOCKQUEUE & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Unlock Queue"); break;
	    case TR_SENDDISCARD & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Send Discard"); break;
	    case TR_DMAIDENTIFY & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "DMA Identify"); break;
	    case TR_THREADIDENTIFY & TR_OPCODE_TYPE_MASK:	sprintf (typeString, "Thread Identify"); break;
	    case TR_GTE & TR_OPCODE_TYPE_MASK:			sprintf (typeString, "GTE"); break;
	    case TR_LT & TR_OPCODE_TYPE_MASK:			sprintf (typeString, "LT"); break;
	    case TR_EQ & TR_OPCODE_TYPE_MASK:			sprintf (typeString, "EQ"); break;
	    case TR_NEQ & TR_OPCODE_TYPE_MASK:			sprintf (typeString, "NEQ"); break;
	    case TR_WRITEWORD & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Write Word"); break;
	    case TR_WRITEDOUBLEWORD & TR_OPCODE_TYPE_MASK:	sprintf (typeString, "Write Double"); break;
	    case TR_ATOMICADDWORD & TR_OPCODE_TYPE_MASK:	sprintf (typeString, "Atomic Add"); break;
	    case TR_TESTANDWRITE & TR_OPCODE_TYPE_MASK:		sprintf (typeString, "Test and Write"); break;
	    default:						sprintf (typeString, "Type=%d", Type & TR_OPCODE_TYPE_MASK); break;
	    }
	}
	sprintf (buffer, "%15s Addr=%08x Cntx=%08x", typeString, Addr, Context);
	/*(Type & TR_SENDACK)      ? " Sendack" : "", */
	/*(Type & TR_LAST_TRANS)   ? " LastTrans" : "", */
	/*(Type & TR_WAIT_FOR_EOP) ? " WaitForEop" : ""); */
    }
    
    if (StatusValid)
    {
	sprintf (statusString, " Type=%s %x", MiToName (hdrp->s.IProcTrapStatus.s.TrapType), hdrp->s.IProcTrapStatus.Status);
	strcat (buffer, statusString);

	if (hdrp->s.IProcTrapStatus.s.BadLength)
	    strcat (buffer, " BadLength");
	switch (hdrp->s.IProcTrapStatus.Status & CRC_MASK)
	{
	case CRC_STATUS_DISCARD:
	    strcat (buffer, " CRC Discard");
	    break;
	case CRC_STATUS_ERROR:
	    strcat (buffer, " CRC Error");
	    break;

	case CRC_STATUS_BAD:
	    strcat (buffer, " CRC Bad");
	    break;
	}
    }

    return (buffer);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
