/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: procfs_linux.c,v 1.21.8.2 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/procfs_linux.c,v $*/

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elandebug.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanvp.h>

#include <qsnet/module.h>
#include <linux/ctype.h>

#include <qsnet/procfs_linux.h>

struct proc_dir_entry *elan3_procfs_root;
struct proc_dir_entry *elan3_config_root;

static int
proc_read_position (char *page, char **start, off_t off,
		    int count, int *eof, void *data)
{
    ELAN3_DEV *dev = (ELAN3_DEV *) data;
    int       len;

    if (dev->Position.pos_mode == ELAN_POS_UNKNOWN)
	len = sprintf (page, "<unknown>\n");
    else
	len = sprintf (page, 
		       "NodeId                 %d\n"
		       "NumLevels              %d\n"
		       "NumNodes               %d\n",
		       dev->Position.pos_nodeid, dev->Position.pos_levels, dev->Position.pos_nodes);

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_write_position (struct file *file, const char *buf, unsigned long count, void *data)
{
    ELAN3_DEV *dev      = (ELAN3_DEV *) data;
    unsigned  nodeid   = ELAN3_INVALID_NODE;
    unsigned  numnodes = 0;
    char     *page, *p;
    int       res;

    if (count == 0)
	return (0);

    if (count >= PAGE_SIZE)
	return (-EINVAL);

    if ((page = (char *) __get_free_page (GFP_KERNEL)) == NULL)
	return (-ENOMEM);

    MOD_INC_USE_COUNT;

    if (copy_from_user (page, buf, count))
	res = -EFAULT;
    else
    {
	page[count] = '\0';
	
	if (page[count-1] == '\n')
	    page[count-1] = '\0';

	if (! strcmp (page, "<unknown>"))
	{
	    dev->Position.pos_mode      = ELAN_POS_UNKNOWN;
	    dev->Position.pos_nodeid    = ELAN3_INVALID_NODE;
	    dev->Position.pos_nodes     = 0;
	    dev->Position.pos_levels    = 0;
	}
	else
	{
	    for (p = page; *p; )
	    {
		while (isspace (*p))
		    p++;
		
		if (! strncmp (p, "NodeId=", strlen("NodeId=")))
		    nodeid   = simple_strtoul (p + strlen ("NodeId="), NULL, 0);
		if (! strncmp (p, "NumNodes=", strlen ("NumNodes=")))
		    numnodes = simple_strtoul (p + strlen ("NumNodes="), NULL, 0);
		
		while (*p && !isspace(*p))
		    p++;
	    }

	    if (ComputePosition (&dev->Position, nodeid, numnodes, dev->Devinfo.dev_num_down_links_value) != 0)
		printk ("elan%d: invalid values for NodeId=%d NumNodes=%d\n", dev->Instance, nodeid, numnodes);
	    else
		printk ("elan%d: setting NodeId=%d NumNodes=%d NumLevels=%d\n", dev->Instance, dev->Position.pos_nodeid,
			dev->Position.pos_nodes, dev->Position.pos_levels);
	}
    }

    MOD_DEC_USE_COUNT;
    free_page ((unsigned long) page);

    return (count);
}


void
elan3_procfs_device_init (ELAN3_DEV *dev)
{
    struct proc_dir_entry *dir, *p;
    char name[NAME_MAX];

    sprintf (name, "device%d", dev->Instance);
    dir = dev->Osdep.procdir = proc_mkdir (name, elan3_procfs_root);

    if ((p = create_proc_entry ("position", 0, dir)) != NULL)
    {
	p->read_proc  = proc_read_position;
	p->write_proc = proc_write_position;
	p->data       = dev;
	p->owner      = THIS_MODULE;
    }

}

void
elan3_procfs_device_fini (ELAN3_DEV *dev)
{
    struct proc_dir_entry *dir = dev->Osdep.procdir;
    char name[NAME_MAX];

    remove_proc_entry ("position", dir);

    sprintf (name, "device%d", dev->Instance);
    remove_proc_entry (name, elan3_procfs_root);
}

void
elan3_procfs_init()
{
    extern int eventint_punt_loops;
    extern int ResolveRequestTimeout;

    elan3_procfs_root = proc_mkdir("elan3",  qsnet_procfs_root);

    elan3_config_root = proc_mkdir("config", elan3_procfs_root);

    qsnet_proc_register_hex (elan3_config_root, "elan3_debug",           &elan3_debug,           0);
    qsnet_proc_register_hex (elan3_config_root, "elan3_debug_console",   &elan3_debug_console,   0);
    qsnet_proc_register_hex (elan3_config_root, "elan3_debug_buffer",    &elan3_debug_buffer,    0);
    qsnet_proc_register_hex (elan3_config_root, "elan3mmu_debug",      &elan3mmu_debug,      0);
    qsnet_proc_register_int (elan3_config_root, "eventint_punt_loops", &eventint_punt_loops, 0);
    qsnet_proc_register_int (elan3_config_root, "neterr_timeout",      &ResolveRequestTimeout, 0);

#if defined(__ia64__)
    {
	extern int enable_sdram_writecombining;
	qsnet_proc_register_int (elan3_config_root, "enable_sdram_writecombining", &enable_sdram_writecombining, 0);
    }
#endif
}

void
elan3_procfs_fini()
{
#if defined(__ia64__)
    remove_proc_entry ("enable_sdram_writecombining", elan3_config_root);
#endif
    remove_proc_entry ("neterr_timeout",      elan3_config_root);
    remove_proc_entry ("eventint_punt_loops", elan3_config_root);
    remove_proc_entry ("elan3mmu_debug",      elan3_config_root);
    remove_proc_entry ("elan3_debug_buffer",    elan3_config_root);
    remove_proc_entry ("elan3_debug_console",   elan3_config_root);
    remove_proc_entry ("elan3_debug",           elan3_config_root);

    remove_proc_entry ("config",  elan3_procfs_root);
    remove_proc_entry ("version", elan3_procfs_root);
 
    remove_proc_entry ("elan3",  qsnet_procfs_root);
}

EXPORT_SYMBOL(elan3_procfs_root);
EXPORT_SYMBOL(elan3_config_root);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
