/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

/* ------------------------------------------------------------- */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/vmseg.h>

/* ---------------------------------------------------------------------- */
typedef struct elan3_net_location {
    int netid;
    int plane;
    int level;
} ELAN3_NET_LOCATION;
/* ---------------------------------------------------------------------- */
#define FLIT_LINK_ARRAY_MAX (ELAN3_MAX_LEVELS*2)
/* ---------------------------------------------------------------------- */
int 
elan3_route_follow_link( ELAN3_CTXT *ctxt, ELAN3_NET_LOCATION *loc, int link)
{
    ELAN_POSITION *pos = &ctxt->Position;

    if ((link<0) || (link>7)) 
    {
	PRINTF1 (ctxt, DBG_VP, "elan3_route_follow_link: link (%d) out of range \n",link);
	return (ELAN3_ROUTE_INVALID);
    }   

    /* going up or down ? */
    if ( link >= pos->pos_arity[loc->level] ) 
    {
	/* Up */
	if (loc->level >= pos->pos_levels)
	    loc->plane = 0;
	else
	{
	    if ((loc->level == 1) && (pos->pos_arity[0]  == 8)) /* oddness in some machines ie 512 */
		loc->plane = (16 * ( loc->plane / 8 )) + (4 * ( loc->plane % 4)) 
		    +(link - pos->pos_arity[loc->level]);
	    else
		loc->plane = (loc->plane * (8 - pos->pos_arity[loc->level]))
		    +(link - pos->pos_arity[loc->level]);
	}
	loc->level--; 
	if ( loc->level < 0 )
	{
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_follow_link: link goes off the top\n");
	    return (ELAN3_ROUTE_INVALID_LEVEL);
	}
	loc->netid = loc->netid / pos->pos_arity[loc->level];
    }
    else
    {
	/* going down */
	if ((loc->level == 0) && (pos->pos_arity[0] == 8)) /* oddness in some machines ie 512 */
	    loc->netid = link % 2;
	else
	    loc->netid =(loc->netid * pos->pos_arity[loc->level])+link;

	loc->level++;
	if (loc->level > pos->pos_levels)
	{
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_follow_link: link goes off the bottom\n");
	    return (ELAN3_ROUTE_INVALID_LEVEL); 
	}

	if ( loc->level >= (pos->pos_levels-1))
	    loc->plane = 0;
	else
	    if ((loc->level == 1) && (pos->pos_arity[0] == 8)) /* oddness in some machines ie 512 */
		loc->plane = (((loc->plane)>>2)*2) - ( ((loc->plane)>>2) & 3  ) + ((link<2)?0:4); /* ((p/4) % 4) */
	    else 
		loc->plane = loc->plane/(8-pos->pos_arity[loc->level]);
    }
    return (ELAN3_ROUTE_SUCCESS);
}
/* ---------------------------------------------------------------------- */
int /* assumes they are connected, really only used for finding the MyLink */
elan3_route_get_mylink (ELAN_POSITION *pos, ELAN3_NET_LOCATION *locA, ELAN3_NET_LOCATION *locB)
{
    /* whats the My Link for locA to LocB */
    if ( locA->level > locB->level ) 
	return locB->plane - (locA->plane * (8 - pos->pos_arity[locA->level])) + pos->pos_arity[locA->level];
    
    return  locB->netid - (locA->netid * pos->pos_arity[locA->level]);
}
/* ---------------------------------------------------------------------- */
#define FIRST_GET_HIGH_PRI(FLIT)    (FLIT & FIRST_HIGH_PRI)
#define FIRST_GET_AGE(FLIT)         ((FLIT & FIRST_AGE(15))>>11)
#define FIRST_GET_TIMEOUT(FLIT)     ((FLIT & FIRST_TIMEOUT(3))>>9)
#define FIRST_GET_NEXT(FLIT)        ((FLIT & FIRST_PACKED(3))>>7)
#define FIRST_GET_ROUTE(FLIT)       (FLIT & 0x7f)
#define FIRST_GET_BCAST(FLIT)       (FLIT & 0x40)
#define FIRST_GET_IS_INVALID(FLIT)  ((FLIT & 0x78) == 0x08)
#define FIRST_GET_TYPE(FLIT)        ((FLIT & 0x30)>>4)
#define PRF_GET_ROUTE(FLIT,N)       ((FLIT >> (N*4)) & 0x0F)
#define PRF_GET_IS_MYLINK(ROUTE)    (ROUTE == PACKED_MYLINK)
#define PRF_GET_IS_NORMAL(ROUTE)    (ROUTE & 0x8)
#define PRF_GET_NORMAL_LINK(ROUTE)  (ROUTE & 0x7)
#define PRF_MOVE_ON(INDEX,NEXT)     do { if (NEXT==3) {NEXT=0;INDEX++;} else {NEXT++; }} while (0);
/* ---------------------------------------------------------------------- */
int /* turn level needed or -1 if not possible */
elan3_route_get_min_turn_level( ELAN_POSITION *pos, int nodeId)
{
    int l,range = 1;

    for(l=pos->pos_levels-1;l>=0;l--)
    {
	range = range * pos->pos_arity[l];
	
	if ( ((pos->pos_nodeid - (pos->pos_nodeid % range)) <= nodeId ) 
	     && (nodeId <= (pos->pos_nodeid - (pos->pos_nodeid % range)+range -1))) 
	    return l;
    }
    return -1;
}
/* ---------------------------------------------------------------------- */
int  
elan3_route_check(ELAN3_CTXT *ctxt, E3_uint16 *flits, int destNodeId)
{
    ELAN3_NET_LOCATION lastLoc,currLoc;
    int               err;
    int               turnLevel;
    int               goingDown;
    int               lnk,index,next,val;
    ELAN_POSITION    *pos = &ctxt->Position;
   
    /* is the dest possible */
    if ( (destNodeId <0 ) || (destNodeId >= pos->pos_nodes))
	return  (ELAN3_ROUTE_PROC_RANGE);

    /* 
     * walk the route, 
     * - to see if we get there 
     * - checking we dont turn around 
     */
    currLoc.netid = pos->pos_nodeid;         /* the elan */
    currLoc.plane = 0;
    currLoc.level = pos->pos_levels;

    turnLevel = currLoc.level; /* track the how far the route goes in */
    goingDown = 0;             /* once set we cant go up again ie only one change of direction */

    /* move onto the network from the elan */
    if ((err=elan3_route_follow_link(ctxt,&currLoc,4)) != ELAN3_ROUTE_SUCCESS) 
    {
	PRINTF0 (ctxt, DBG_VP, "elan3_route_check: initial elan3_route_follow_link failed\n");
	return err;
    }
    /* do the first part of flit */
    switch ( FIRST_GET_TYPE(flits[0]) ) 
    {
    case 0  /* sent */   : { lnk = (flits[0] & 0x7);	                              break; }    
    case PACKED_MYLINK	 : { lnk = pos->pos_nodeid % pos->pos_arity[pos->pos_levels-1];    break; }
    case PACKED_ADAPTIVE : { lnk = 7; /* all routes are the same just check one */    break; }
    default : 
	PRINTF1 (ctxt, DBG_VP, "elan3_route_check: unexpected first flit (%d)\n",flits[0]);
	return (ELAN3_ROUTE_INVALID); 
    }
    
    /* move along this link and check new location */
    memcpy(&lastLoc,&currLoc,sizeof(ELAN3_NET_LOCATION)); /* keep track of last loc */
    if ((err=elan3_route_follow_link(ctxt,&currLoc,lnk)) != ELAN3_ROUTE_SUCCESS ) 
    {
	PRINTF0 (ctxt, DBG_VP, "elan3_route_check: elan3_route_follow_link failed\n");
	return err;
    }
    if ((currLoc.level > pos->pos_levels) || (currLoc.level < 0 )) 
    { 
	PRINTF0 (ctxt, DBG_VP, "elan3_route_check: route leaves machine\n");
	return (ELAN3_ROUTE_INVALID_LEVEL);
    }
    if ( lastLoc.level < currLoc.level ) 
    {
	turnLevel = lastLoc.level;
	goingDown = 1;
    }
    else 
    {
	if (turnLevel > currLoc.level)
	    turnLevel =  currLoc.level;
	if  (goingDown) 
	{
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_check: route ocilated\n");
	    return (ELAN3_ROUTE_OCILATES);
	}
    }   

    /* loop on doing the remaining flits */
    index = 1;
    next  = FIRST_GET_NEXT(flits[0]);
    val   = PRF_GET_ROUTE(flits[index],next);
    while(val)
    {
	if (PRF_GET_IS_NORMAL(val) ) 
	    lnk = PRF_GET_NORMAL_LINK(val);
	else
	{
	  switch ( val ) 
	  {
	  case  PACKED_MYLINK : 
	  {
	      lnk = elan3_route_get_mylink(pos, &currLoc,&lastLoc);
	      break;
	  }
	  default : 
	      PRINTF1 (ctxt, DBG_VP, "elan3_route_check: unexpected packed flit (%d)\n",val);
	      return (ELAN3_ROUTE_INVALID);
	  }
	}

	/* move along this link and check new location */
	memcpy(&lastLoc,&currLoc,sizeof(ELAN3_NET_LOCATION)); /* keep track of last loc */
	if ((err=elan3_route_follow_link(ctxt,&currLoc,lnk)) != ELAN3_ROUTE_SUCCESS) 
	    return err;
	
	if ((currLoc.level > pos->pos_levels ) || ( currLoc.level < 0 ))
	{ 
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_check: route leaves machine\n");
	    return (ELAN3_ROUTE_INVALID_LEVEL);
	}

	if ( lastLoc.level < currLoc.level ) 
	    goingDown = 1;
	else 
	{
	    if (turnLevel > currLoc.level)
		turnLevel =  currLoc.level;
	    if  (goingDown) 
	    {
		PRINTF0 (ctxt, DBG_VP, "elan3_route_check: route ocilated\n");
		return (ELAN3_ROUTE_OCILATES);
	    }
	}   

	/* move to next part of flit */
	PRF_MOVE_ON(index,next);
	if ( index >= MAX_FLITS)
	{
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_check: route too long\n");
	    return (ELAN3_ROUTE_TOO_LONG);
	}
	/* extract the new value */
	val = PRF_GET_ROUTE(flits[index],next);
    }

    /* have we got to where we want ? */
    if ((currLoc.level != pos->pos_levels) || (currLoc.netid != destNodeId))
    {
	PRINTF2 (ctxt, DBG_VP, "elan3_route_check: goes to %d instead of %d\n",currLoc.netid , destNodeId );
	return (ELAN3_ROUTE_WRONG_DEST);
    }

    /*
     * there is the case of src == dest 
     * getTurnLevel returns pos->pos_levels, and turnLevel is (pos->pos_levels -1) 
     * then we assume they really want to  go onto the network.
     * otherwise we check that the turn at the appriate level
     */
    if ( (pos->pos_nodeid != destNodeId) || ( turnLevel != (pos->pos_levels -1)) )
    {
	int lev;
	if ((lev = elan3_route_get_min_turn_level(pos,destNodeId)) == -1) 
	{
	    PRINTF0 (ctxt, DBG_VP, "elan3_route_check: cant calculate turn level\n");
	    return (ELAN3_ROUTE_INVALID); /* not sure this can happen here as checks above should protect me */
	}
	if (turnLevel != lev) 
	{
	    PRINTF2 (ctxt, DBG_VP, "elan3_route_check: turn level should be %d but is %d \n", lev, turnLevel);
	    return (ELAN3_ROUTE_TURN_LEVEL);
	}
    }
    return (ELAN3_ROUTE_SUCCESS);
}
/* ---------------------------------------------------------------------- */
int
elan3_route_broadcast_check(ELAN3_CTXT *ctxt , E3_uint16 *flits, int lowNode, int highNode ) 
{
    E3_uint16 flitsTmp[MAX_FLITS];
    int       nflits,i;
    
    nflits = GenerateRoute (&ctxt->Position, flitsTmp, lowNode, highNode, DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);
    
    for(i=0;i<nflits;i++)
	if ( flitsTmp[i] != flits[i] ) 
	{
	    PRINTF3 (ctxt, DBG_VP, "elan3_route_broadcast_check:  flit[%d] %d (should be %d)\n",i,flits[i],flitsTmp[i]);
	    return (ELAN3_ROUTE_INVALID);   
	}
    
    return (ELAN3_ROUTE_SUCCESS);
}
/* ---------------------------------------------------------------------- */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
