/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: sdram.c,v 1.17 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/sdram.c,v $*/


#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elandebug.h>

/* sdram access functions */
#define sdram_off_to_bank(dev,off)	(&dev->SdramBanks[(off) >> ELAN3_SDRAM_BANK_SHIFT])
#define sdram_off_to_offset(dev,off)	((off) & (ELAN3_SDRAM_BANK_SIZE-1))
#define sdram_off_to_bit(dev,indx,off)	(sdram_off_to_offset(dev,off) >> (SDRAM_MIN_BLOCK_SHIFT+(indx)))

#define sdram_off_to_mapping(dev,off)	(sdram_off_to_bank(dev,off)->Mapping + sdram_off_to_offset(dev,off))
    
unsigned char
elan3_sdram_readb (ELAN3_DEV *dev, sdramaddr_t off)
{
    return (readb ((unsigned char *) sdram_off_to_mapping(dev, off)));
}

unsigned short
elan3_sdram_readw (ELAN3_DEV *dev, sdramaddr_t off)
{
    return (readw ((unsigned short *) sdram_off_to_mapping(dev, off)));
}

unsigned int
elan3_sdram_readl (ELAN3_DEV *dev, sdramaddr_t off)
{
    return (readl ((unsigned int *) sdram_off_to_mapping(dev, off)));
}

unsigned long long
elan3_sdram_readq (ELAN3_DEV *dev, sdramaddr_t off)
{
    return (readq ((unsigned long long *) sdram_off_to_mapping(dev, off)));
}

void
elan3_sdram_writeb (ELAN3_DEV *dev, sdramaddr_t off, unsigned char val)
{
    writeb (val, (unsigned char *) sdram_off_to_mapping(dev, off));
    wmb();
}

void
elan3_sdram_writew (ELAN3_DEV *dev, sdramaddr_t off, unsigned short val)
{
    writew (val, (unsigned short *) sdram_off_to_mapping(dev, off));
    wmb();
}

void
elan3_sdram_writel (ELAN3_DEV *dev, sdramaddr_t off, unsigned int val)
{
    writel (val, (unsigned int *) sdram_off_to_mapping(dev, off));
    wmb();
}

void
elan3_sdram_writeq (ELAN3_DEV *dev, sdramaddr_t off, unsigned long long val)
{
    writeq (val, (unsigned long long *) sdram_off_to_mapping(dev, off));
    wmb();
}

void
elan3_sdram_copyb_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
    bcopy ((void *)sdram_off_to_mapping(dev, from), to, nbytes);
}

void
elan3_sdram_copyw_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
#ifdef __LITTLE_ENDIAN__
    bcopy ((void *)sdram_off_to_mapping(dev, from), to, nbytes);
#else
#error incorrect for big endian
#endif
}

void
elan3_sdram_copyl_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
#ifdef __LITTLE_ENDIAN__
    bcopy ((void *)sdram_off_to_mapping(dev, from), to, nbytes);
#else
#error incorrect for big endian
#endif
}

void
elan3_sdram_copyq_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
#ifdef __LITTLE_ENDIAN__
    bcopy ((void *)sdram_off_to_mapping(dev, from), to, nbytes);
#else
#error incorrect for big endian
#endif
}

#define E3_WRITEBUFFER_SIZE		16
#define E3_WRITEBUFFER_OFFSET(x)	(((unsigned long) x) & (E3_WRITEBUFFER_SIZE-1))
#define E3_WRITEBUFFER_BASE(x)		(((unsigned long) x) & ~((unsigned long) (E3_WRITEBUFFER_SIZE-1)))

void
elan3_sdram_copyb_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    virtaddr_t slim  = (virtaddr_t) from + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint8_t)) + sizeof (uint8_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint8_t); i++)
	    writeb (((uint8_t *) from)[i], &((uint8_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    slim -= ntop;
	    dlim -= ntop;
	    
	    for (i = 0; i < ntop/sizeof(uint8_t); i++)
		writeb (((uint8_t *) slim)[i], &((uint8_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;
	    slim -= E3_WRITEBUFFER_SIZE;

	    for (i = 0; i < E3_WRITEBUFFER_SIZE/sizeof (uint8_t); i++)
		writeb (((uint8_t *) slim)[i], &((uint8_t *) dlim)[i]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint8_t); i++)
		writeb (((uint8_t *) from)[i], &((uint8_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_zerob_sdram (ELAN3_DEV *dev, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint8_t)) + sizeof (uint8_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint8_t); i++)
	    writeb (0, &((uint8_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    dlim -= ntop;
	    
	    for (i = 0; i < ntop/sizeof(uint8_t); i++)
		writeb (0, &((uint8_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;

	    writeq (0, &((uint64_t *) dlim)[0]);
	    writeq (0, &((uint64_t *) dlim)[1]);

	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint8_t); i++)
		writeb (0, &((uint8_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_copyw_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    virtaddr_t slim  = (virtaddr_t) from + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint16_t)) + sizeof (uint16_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint16_t); i++)
	    writew (((uint16_t *) from)[i], &((uint16_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    slim -= ntop;
	    dlim -= ntop;

	    for (i = 0; i < ntop/sizeof(uint16_t); i++)
		writew (((uint16_t *) slim)[i], &((uint16_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;
	    slim -= E3_WRITEBUFFER_SIZE;

	    writew (((uint16_t *) slim)[0], &((uint16_t *) dlim)[0]);
	    writew (((uint16_t *) slim)[1], &((uint16_t *) dlim)[1]);
	    writew (((uint16_t *) slim)[2], &((uint16_t *) dlim)[2]);
	    writew (((uint16_t *) slim)[3], &((uint16_t *) dlim)[3]);
	    writew (((uint16_t *) slim)[4], &((uint16_t *) dlim)[4]);
	    writew (((uint16_t *) slim)[5], &((uint16_t *) dlim)[5]);
	    writew (((uint16_t *) slim)[6], &((uint16_t *) dlim)[6]);
	    writew (((uint16_t *) slim)[7], &((uint16_t *) dlim)[7]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint16_t); i++)
		writew (((uint16_t *) from)[i], &((uint16_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_zerow_sdram (ELAN3_DEV *dev, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint16_t)) + sizeof (uint16_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint16_t); i++)
	    writew (0, &((uint16_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    dlim -= ntop;
	    
	    for (i = 0; i < ntop/sizeof(uint16_t); i++)
		writew (0, &((uint16_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;

	    writeq (0, &((uint64_t *) dlim)[0]);
	    writeq (0, &((uint64_t *) dlim)[1]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint16_t); i++)
		writew (0, &((uint16_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_copyl_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    virtaddr_t slim  = (virtaddr_t) from + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint32_t)) + sizeof (uint32_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint32_t); i++)
	    writel (((uint32_t *) from)[i], &((uint32_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    slim -= ntop;
	    dlim -= ntop;

	    for (i = 0; i < ntop/sizeof(uint32_t); i++)
		writel (((uint32_t *) slim)[i], &((uint32_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;
	    slim -= E3_WRITEBUFFER_SIZE;

	    writel (((uint32_t *) slim)[0], &((uint32_t *) dlim)[0]);
	    writel (((uint32_t *) slim)[1], &((uint32_t *) dlim)[1]);
	    writel (((uint32_t *) slim)[2], &((uint32_t *) dlim)[2]);
	    writel (((uint32_t *) slim)[3], &((uint32_t *) dlim)[3]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint32_t); i++)
		writel (((uint32_t *) from)[i], &((uint32_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_zerol_sdram (ELAN3_DEV *dev, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint32_t)) + sizeof (uint32_t);
    int        i;

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	for (i = 0; i < nbytes/sizeof(uint32_t); i++)
	    writel (0, &((uint32_t *) dbase)[i]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    dlim -= ntop;

	    for (i = 0; i < ntop/sizeof(uint32_t); i++)
		writel (0, &((uint32_t *) dlim)[i]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;

	    writeq (0, &((uint64_t *) dlim)[0]);
	    writeq (0, &((uint64_t *) dlim)[1]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    for (i = 0; i < nbase/sizeof(uint32_t); i++)
		writel (0, &((uint32_t *) dbase)[i]);
	    wmb();
	}
    }
}

void
elan3_sdram_copyq_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    virtaddr_t slim  = (virtaddr_t) from + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint64_t)) + sizeof (uint64_t);

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	writeq (((uint64_t *) from)[0], &((uint64_t *) dbase)[0]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    slim -= ntop;
	    dlim -= ntop;

	    writeq (((uint64_t *) slim)[0], &((uint64_t *) dlim)[0]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;
	    slim -= E3_WRITEBUFFER_SIZE;

	    writeq (((uint64_t *) slim)[0], &((uint64_t *) dlim)[0]);
	    writeq (((uint64_t *) slim)[1], &((uint64_t *) dlim)[1]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    writeq (((uint64_t *) from)[0], &((uint64_t *) dbase)[0]);
	    wmb();
	}
    }
}

void
elan3_sdram_zeroq_sdram (ELAN3_DEV *dev, sdramaddr_t to, int nbytes)
{
    virtaddr_t dbase = (virtaddr_t) sdram_off_to_mapping (dev, to);
    virtaddr_t dlim  = (virtaddr_t) dbase + nbytes;
    unsigned   nbase = E3_WRITEBUFFER_SIZE - E3_WRITEBUFFER_OFFSET (dbase);
    unsigned   ntop  = E3_WRITEBUFFER_OFFSET (dlim - sizeof (uint64_t)) + sizeof (uint64_t);

    if (E3_WRITEBUFFER_BASE(dbase) == E3_WRITEBUFFER_BASE(dlim))
    {
	writeq (0, &((uint64_t *) dbase)[0]);
	wmb();
    }
    else
    {
	if (ntop < E3_WRITEBUFFER_SIZE)
	{
	    dlim -= ntop;

	    writeq (0, &((uint64_t *) dlim)[0]);
	    wmb();
	}
	
	while (dlim >= (dbase + E3_WRITEBUFFER_SIZE))
	{
	    dlim -= E3_WRITEBUFFER_SIZE;
	    
	    writeq (0, &((uint64_t *) dlim)[0]);
	    writeq (0, &((uint64_t *) dlim)[1]);
	    wmb();
	}
	
	if (nbase < E3_WRITEBUFFER_SIZE)
	{
	    writeq (0, &((uint64_t *) dbase)[0]);
	    wmb();
	}
    }
}

physaddr_t
elan3_sdram_to_phys (ELAN3_DEV *dev, sdramaddr_t off)
{
#if defined(DIGITAL_UNIX)
    return (KSEG_TO_PHYS (sdram_off_to_mapping (dev, off)));
#elif defined(LINUX)
    return (kmem_to_phys ((void *) sdram_off_to_mapping (dev, off)));
#endif    
}

/* sdram buddy allocator */
#define read_next(dev, block)		elan3_sdram_readl(dev, block + 0)
#define read_prev(dev, block)		elan3_sdram_readl(dev, block + 4)
#define write_next(dev, block, val)	(elan3_sdram_writel(dev, block + 0, val), val)
#define write_prev(dev, block, val)	(elan3_sdram_writel(dev, block + 4, val), val)

#define freelist_insert(dev,idx,block)\
do {\
    sdramaddr_t next = dev->SdramFreeLists[(idx)];\
\
    /*\
     * block->prev = NULL;\
     * block->next = next;\
     * if (next != NULL)\
     *    next->prev = block;\
     * freelist = block;\
     */\
    write_prev (dev, block, (sdramaddr_t) 0);\
    write_next (dev, block, next);\
    if (next != (sdramaddr_t) 0)\
	write_prev (dev, next, block);\
    dev->SdramFreeLists[idx] = block;\
\
    dev->SdramFreeCounts[idx]++;\
    dev->Stats.SdramBytesFree += (SDRAM_MIN_BLOCK_SIZE << idx);\
} while (0)

#define freelist_remove(dev,idx,block)\
do {\
    /*\
     * if (block->prev)\
     *     block->prev->next = block->next;\
     * else\
     *     dev->SdramFreeLists[idx] = block->next;\
     * if (block->next)\
     *     block->next->prev = block->prev;\
     */\
    sdramaddr_t blocknext = read_next (dev, block);\
    sdramaddr_t blockprev = read_prev (dev, block);\
\
    if (blockprev)\
	write_next (dev, blockprev, blocknext);\
    else\
	dev->SdramFreeLists[idx] = blocknext;\
    if (blocknext)\
	write_prev (dev, blocknext, blockprev);\
\
    dev->SdramFreeCounts[idx]--;\
    dev->Stats.SdramBytesFree -= (SDRAM_MIN_BLOCK_SIZE << idx);\
} while (0)

#define freelist_removehead(dev,idx,block)\
do {\
    sdramaddr_t blocknext = read_next (dev, block);\
\
    if ((dev->SdramFreeLists[idx] = blocknext) != 0)\
	write_prev (dev, blocknext, 0);\
\
    dev->SdramFreeCounts[idx]--;\
    dev->Stats.SdramBytesFree -= (SDRAM_MIN_BLOCK_SIZE << idx);\
} while (0)

#if defined(DEBUG)
static int
display_blocks (ELAN3_DEV *dev, int indx, char *string)
{
    sdramaddr_t block;
    int nbytes = 0;

    printk ("%s - indx %d\n", string, indx);
    for (block = dev->SdramFreeLists[indx]; block != (sdramaddr_t) 0; block = read_next (dev, block))
    {
	printk ("  %lx", block);
	nbytes += (SDRAM_MIN_BLOCK_SIZE << indx);
    }
    printk ("\n");

    return (nbytes);
}


void
elan3_sdram_display (ELAN3_DEV *dev, char *string)
{
    int indx;
    int nbytes = 0;

    printk ("elan3_sdram_display: dev=%p\n", dev);
    for (indx = 0; indx < SDRAM_NUM_FREE_LISTS; indx++)
	if (dev->SdramFreeLists[indx] != (sdramaddr_t) 0)
	    nbytes += display_blocks (dev, indx, string);
    printk ("\n%d bytes free\n", nbytes);
}

void
elan3_sdram_verify (ELAN3_DEV *dev)
{
    int indx, size, nbits, i, b;
    sdramaddr_t block;

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	unsigned count = 0;

	for (block = dev->SdramFreeLists[indx]; block; block = read_next (dev, block), count++)
	{
	    ELAN3_SDRAM_BANK *bank = sdram_off_to_bank (dev, block);
	    unsigned         off  = sdram_off_to_offset (dev, block);
	    int              bit  = sdram_off_to_bit (dev, indx, block);

	    if ((block & (size-1)) != 0)
		printk ("elan3_sdram_verify: block=%lx indx=%x - not aligned\n", block, indx);
	    
	    if (bank == NULL || off > bank->Size)
		printk ("elan3_sdram_verify: block=%lx indx=%x - outside bank\n", block, indx);
	    else if (BT_TEST (bank->Bitmaps[indx], bit) == 0)
		printk ("elan3_sdram_verify: block=%lx indx=%x - bit not set\n", block, indx);
	    else
	    {
		for (i = indx-1, nbits = 2; i >= 0; i--, nbits <<= 1)
		{
		    bit = sdram_off_to_bit (dev, i, block);

		    for (b = 0; b < nbits; b++)
			if (BT_TEST(bank->Bitmaps[i], bit + b))
			    printk ("elan3_sdram_verify: block=%lx indx=%x - also free i=%d bit=%x\n", block, indx, i, bit+b);
		}
	    }
	}

	if (dev->SdramFreeCounts[indx] != count)
	    printk ("elan3_sdram_verify: indx=%x expected %d got %d\n", indx, dev->SdramFreeCounts[indx], count);
    }
}

#endif /* defined(DEBUG) */

static void
free_block (ELAN3_DEV *dev, sdramaddr_t block, int indx)
{
    ELAN3_SDRAM_BANK *bank = sdram_off_to_bank (dev, block);
    unsigned	     bit  = sdram_off_to_bit(dev, indx, block);
    unsigned         size = SDRAM_MIN_BLOCK_SIZE << indx;

    PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: block=%lx indx=%d bit=%x\n", block, indx, bit);

    ASSERT ((block & (size-1)) == 0);
    ASSERT (BT_TEST (bank->Bitmaps[indx], bit) == 0);
    
    while (BT_TEST (bank->Bitmaps[indx], bit ^ 1))
    {
	sdramaddr_t buddy = block ^ size;
	
	PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: merge block=%lx buddy=%lx indx=%d\n", block, buddy, indx);

	BT_CLEAR (bank->Bitmaps[indx], bit ^ 1);

	freelist_remove (dev, indx, buddy);
	
	block = (block < buddy) ? block : buddy;
	indx++;
	size <<= 1;
	bit >>= 1;
    }

    PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: free block=%lx indx=%d bit=%x\n", block, indx, bit);

    freelist_insert (dev, indx, block);

    BT_SET (bank->Bitmaps[indx], bit);
}

void
elan3_sdram_init (ELAN3_DEV *dev)
{
    int indx;

    spin_lock_init (&dev->SdramLock);

    for (indx = 0; indx < SDRAM_NUM_FREE_LISTS; indx++)
    {
	dev->SdramFreeLists[indx]  = (sdramaddr_t) 0;
	dev->SdramFreeCounts[indx] = 0;
    }
}

void
elan3_sdram_fini (ELAN3_DEV *dev)
{
    spin_lock_destroy (&dev->SdramLock);
}

void
elan3_sdram_add (ELAN3_DEV *dev, sdramaddr_t base, sdramaddr_t top)
{
    register int indx;
    register unsigned long size;

    /* align to the minimum block size */
    base = (base + SDRAM_MIN_BLOCK_SIZE - 1) & ~((sdramaddr_t) SDRAM_MIN_BLOCK_SIZE-1);
    top &= ~((sdramaddr_t) SDRAM_MIN_BLOCK_SIZE-1);

    /* don't allow 0 as a valid "base" */
    if (base == 0)
	base = E3_CACHE_SIZE;

    /* carve the bottom to the biggest boundary */
    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	if ((base & size) == 0)
	    continue;

	if ((base + size) > top)
	    break;

	free_block (dev, base, indx);
	
	base += size;
    }

    /* carve the top down to the biggest boundary */
    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	if ((top & size) == 0)
	    continue;

	if ((top - size) < base)
	    break;

	free_block (dev, (top - size), indx);
	
	top -= size;
    }

    /* now free of the space in between */
    while (base < top)
    {
	free_block (dev, base, (SDRAM_NUM_FREE_LISTS-1));

	base += SDRAM_MAX_BLOCK_SIZE;
    }
}

sdramaddr_t
elan3_sdram_alloc (ELAN3_DEV *dev, int nbytes)
{
    sdramaddr_t block;
    register int i, indx;
    unsigned long size;
    unsigned long flags;

    spin_lock_irqsave (&dev->SdramLock, flags);

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size < nbytes; indx++, size <<= 1)
	;

    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan3_sdram_alloc: nbytes=%d indx=%d\n", nbytes, indx);

    /* find the smallest block which is big enough for this allocation */
    for (i = indx; i < SDRAM_NUM_FREE_LISTS; i++, size <<= 1)
	if (dev->SdramFreeLists[i])
	    break;
    
    if (i == SDRAM_NUM_FREE_LISTS)
    {
	spin_unlock_irqrestore (&dev->SdramLock, flags);
	return ((sdramaddr_t) 0);
    }
    
    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan3_sdram_alloc: use block=%lx indx=%d\n", dev->SdramFreeLists[i], i);

    /* remove the block from the free list */
    freelist_removehead (dev, i, (block = dev->SdramFreeLists[i]));

    /* clear the approriate bit in the bitmap */
    BT_CLEAR (sdram_off_to_bank (dev, block)->Bitmaps[i], sdram_off_to_bit (dev,i, block));

    /* and split it up as required */
    while (i-- > indx)
	free_block (dev, block + (size >>= 1), i);

    PRINTF1 (DBG_DEVICE, DBG_SDRAM, "elan3_sdram_alloc: return block=%lx\n", block);

    spin_unlock_irqrestore (&dev->SdramLock, flags);

    ASSERT ((block & ((SDRAM_MIN_BLOCK_SIZE << (indx))-1)) == 0);

    return ((sdramaddr_t) block);
}

void
elan3_sdram_free (ELAN3_DEV *dev, sdramaddr_t block, int nbytes)
{
    register int indx;
    unsigned long size;
    unsigned long flags;

    spin_lock_irqsave (&dev->SdramLock, flags);

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size < nbytes; indx++, size <<= 1)
	;

    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan3_sdram_free: indx=%d block=%lx\n", indx, block);
    
    free_block (dev, block, indx);

    spin_unlock_irqrestore (&dev->SdramLock, flags);
}



/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
