/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: tproc.c,v 1.52 2004/11/15 11:14:12 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/tproc.c,v $ */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/elansyscall.h>
#include <elan3/threadsyscall.h>
#include <elan3/intrinsics.h>
#include <elan3/vmseg.h>

int
HandleTProcTrap (ELAN3_DEV *dev, E3_uint32 *RestartBits)
{
    THREAD_TRAP  *trap  = dev->ThreadTrap;
    int           delay = 1;

    ASSERT(SPINLOCK_HELD (&dev->IntrLock));

    trap->Status.Status  = read_reg32 (dev, Exts.TProcStatus);
    trap->sp             = read_reg32 (dev, Thread_Desc_SP);
    trap->pc             = read_reg32 (dev, ExecutePC);
    trap->npc            = read_reg32 (dev, ExecuteNPC);
    trap->StartPC        = read_reg32 (dev, StartPC);
    trap->mi             = GET_STATUS_TRAPTYPE(trap->Status);
    trap->TrapBits.Bits  = read_reg32 (dev, TrapBits.Bits);
    trap->DirtyBits.Bits = read_reg32 (dev, DirtyBits.Bits);

    if ( ! (trap->Status.s.WakeupFunction == SleepOneTick) ) {
	int p,i;
	E3_uint32 reg = read_reg32 (dev, Exts.InterruptReg);	

	ELAN_REG_REC(reg);
	p = elan_reg_rec_index;
	for(i=0;i<ELAN_REG_REC_MAX;i++) {
	    if (elan_reg_rec_file[i] != NULL ) 
		printk("Elan Reg Record[%2d](%ld): cpu %d  reg %x [%d:%s]\n", p, elan_reg_rec_lbolt[p], elan_reg_rec_cpu[p], elan_reg_rec_reg[p],
		       elan_reg_rec_line[p], elan_reg_rec_file[p]);
	    p = ( (p+1) % ELAN_REG_REC_MAX);
	}
    }
    
    ASSERT(trap->Status.s.WakeupFunction == SleepOneTick);

    /* copy the four access fault areas */
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, TProc),          (void *) &trap->FaultSave, 16);
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcData), (void *) &trap->DataFaultSave, 16);
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcInst), (void *) &trap->InstFaultSave, 16);
    elan3_sdram_copyq_from_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcOpen), (void *) &trap->OpenFaultSave, 16);
    
    /* copy the registers,  note the endian swap flips the odd registers into the even registers
       and visa versa. */
    copy_thread_regs (dev, trap->Registers);

    /*
     * If the output was open then the ack may not have returned yet. Must wait for the
     * ack to become valid and update trap_dirty with the new value. Will simulate the
     * instructions later.
     */
    if (trap->TrapBits.s.OutputWasOpen)
    {
	trap->TrapBits.Bits = read_reg32 (dev, TrapBits.Bits);
	while (! trap->TrapBits.s.AckBufferValid)
	{
	    PRINTF0 (DBG_DEVICE, DBG_INTR, "tproc: waiting for ack to become valid\n");
	    trap->TrapBits.Bits = read_reg32 (dev, TrapBits.Bits);
	    DELAY (delay);

	    if ((delay <<= 1) == 0) delay = 1;
	}
    }
    
    /* update device statistics */
    BumpStat (dev, TProcTraps);
    switch (trap->mi)
    {
    case MI_UnimplementedError:
	if (trap->TrapBits.s.ForcedTProcTrap)
	    BumpStat (dev, ForcedTProcTraps);
	if (trap->TrapBits.s.ThreadTimeout)
	{
	    if (trap->TrapBits.s.PacketTimeout)
		BumpStat (dev, ThreadOutputTimeouts);
	    else if (trap->TrapBits.s.PacketAckValue == E3_PAckError)
		BumpStat (dev, ThreadPacketAckErrors);
	}
	if (trap->TrapBits.s.TrapForTooManyInsts)
	    BumpStat (dev, TrapForTooManyInsts);
	break;
    }

    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, TProc), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcData), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcInst), 16);
    elan3_sdram_zeroq_sdram (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, ThreadProcOpen), 16);

    *RestartBits |= RestartTProc;

    return (TRUE);
}

void
DeliverTProcTrap (ELAN3_DEV *dev, THREAD_TRAP *threadTrap, E3_uint32 Pend)
{
    ELAN3_CTXT	     *ctxt;
    THREAD_TRAP      *trap;

    ASSERT(SPINLOCK_HELD (&dev->IntrLock));

    ctxt = ELAN3_DEV_CTX_TABLE(dev, threadTrap->Status.s.Context);

    if (ctxt == NULL)
    {
	PRINTF1 (DBG_DEVICE, DBG_INTR, "DeliverTProcTrap: context %x invalid\n", threadTrap->Status.s.Context);
	BumpStat (dev, InvalidContext);
    }
    else
    {
	if (ELAN3_OP_TPROC_TRAP (ctxt, threadTrap) == OP_DEFER)
	{
	    if (ELAN3_QUEUE_REALLY_FULL (ctxt->ThreadTrapQ))
	    {
		ctxt->Status |= CTXT_COMMAND_OVERFLOW_ERROR;
		StartSwapoutContext (ctxt, Pend, NULL);
	    }
	    else
	    {
		trap = ELAN3_QUEUE_BACK (ctxt->ThreadTrapQ, ctxt->ThreadTraps);
		
		bcopy (threadTrap, trap, sizeof (THREAD_TRAP));
		
		PRINTF4 (ctxt, DBG_INTR, "DeliverTProcTrap: SP=%08x PC=%08x NPC=%08x StartPC %08x\n",
			 trap->sp, trap->pc, trap->npc, trap->StartPC);
		PRINTF3 (ctxt, DBG_INTR, "       mi=%s trap=%08x dirty=%08x\n",
			 MiToName (trap->mi), trap->TrapBits.Bits, trap->DirtyBits.Bits);
		PRINTF3 (ctxt, DBG_INTR, "       FaultSave : FaultAddress %08x EventAddress %08x FSR %08x\n",
			 trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress, trap->FaultSave.s.FSR.Status);
		PRINTF3 (ctxt, DBG_INTR, "       DataFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
			 trap->DataFaultSave.s.FaultAddress, trap->DataFaultSave.s.EventAddress, trap->DataFaultSave.s.FSR.Status);
		PRINTF3 (ctxt, DBG_INTR, "       InstFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
			 trap->InstFaultSave.s.FaultAddress, trap->InstFaultSave.s.EventAddress, trap->InstFaultSave.s.FSR.Status);
		PRINTF3 (ctxt, DBG_INTR, "       OpenFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
			 trap->OpenFaultSave.s.FaultAddress, trap->OpenFaultSave.s.EventAddress, trap->OpenFaultSave.s.FSR.Status);
		
		PRINTF4 (ctxt, DBG_INTR, "       g0=%08x g1=%08x g2=%08x g3=%08x\n", 
			 trap->Registers[REG_GLOBALS+(0^WordEndianFlip)], trap->Registers[REG_GLOBALS+(1^WordEndianFlip)], 
			 trap->Registers[REG_GLOBALS+(2^WordEndianFlip)], trap->Registers[REG_GLOBALS+(3^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       g4=%08x g5=%08x g6=%08x g7=%08x\n", 
			 trap->Registers[REG_GLOBALS+(4^WordEndianFlip)], trap->Registers[REG_GLOBALS+(5^WordEndianFlip)], 
			 trap->Registers[REG_GLOBALS+(6^WordEndianFlip)], trap->Registers[REG_GLOBALS+(7^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       o0=%08x o1=%08x o2=%08x o3=%08x\n", 
			 trap->Registers[REG_OUTS+(0^WordEndianFlip)], trap->Registers[REG_OUTS+(1^WordEndianFlip)], 
			 trap->Registers[REG_OUTS+(2^WordEndianFlip)], trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       o4=%08x o5=%08x o6=%08x o7=%08x\n", 
			 trap->Registers[REG_OUTS+(4^WordEndianFlip)], trap->Registers[REG_OUTS+(5^WordEndianFlip)], 
			 trap->Registers[REG_OUTS+(6^WordEndianFlip)], trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       l0=%08x l1=%08x l2=%08x l3=%08x\n", 
			 trap->Registers[REG_LOCALS+(0^WordEndianFlip)], trap->Registers[REG_LOCALS+(1^WordEndianFlip)],
			 trap->Registers[REG_LOCALS+(2^WordEndianFlip)], trap->Registers[REG_LOCALS+(3^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       l4=%08x l5=%08x l6=%08x l7=%08x\n", 
			 trap->Registers[REG_LOCALS+(4^WordEndianFlip)], trap->Registers[REG_LOCALS+(5^WordEndianFlip)],
			 trap->Registers[REG_LOCALS+(6^WordEndianFlip)], trap->Registers[REG_LOCALS+(7^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       i0=%08x i1=%08x i2=%08x i3=%08x\n", 
			 trap->Registers[REG_INS+(0^WordEndianFlip)], trap->Registers[REG_INS+(1^WordEndianFlip)],
			 trap->Registers[REG_INS+(2^WordEndianFlip)], trap->Registers[REG_INS+(3^WordEndianFlip)]);
		PRINTF4 (ctxt, DBG_INTR, "       i4=%08x i5=%08x i6=%08x i7=%08x\n", 
			 trap->Registers[REG_INS+(4^WordEndianFlip)], trap->Registers[REG_INS+(5^WordEndianFlip)],
			 trap->Registers[REG_INS+(6^WordEndianFlip)], trap->Registers[REG_INS+(7^WordEndianFlip)]);
		
		ELAN3_QUEUE_ADD (ctxt->ThreadTrapQ);
		kcondvar_wakeupone (&ctxt->Wait, &dev->IntrLock);
		
		if (ELAN3_QUEUE_FULL (ctxt->ThreadTrapQ))
		{
		    PRINTF0 (ctxt, DBG_INTR, "DeliverTProcTrap: thread queue full,  must swap out\n");
		    ctxt->Status |= CTXT_THREAD_QUEUE_FULL;
		    
		    StartSwapoutContext (ctxt, Pend, NULL);
		}
	    }
	}
    }
}

int
NextTProcTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap)
{
    ELAN3_DEV *dev = ctxt->Device;

    ASSERT (SPINLOCK_HELD (&dev->IntrLock));
    
    if (ELAN3_QUEUE_EMPTY (ctxt->ThreadTrapQ))
	return (0);

    *trap = *ELAN3_QUEUE_FRONT (ctxt->ThreadTrapQ, ctxt->ThreadTraps);
    ELAN3_QUEUE_REMOVE (ctxt->ThreadTrapQ);
    
    return (1);
}

void
ResolveTProcTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap)
{
    int       i;
    int       res;
    E3_Addr   StackPointer;

    PRINTF4 (ctxt, DBG_TPROC, "ResolveTProcTrap: SP=%08x PC=%08x NPC=%08x StartPC %08x\n",
	     trap->sp, trap->pc, trap->npc, trap->StartPC);
    PRINTF3 (ctxt, DBG_TPROC, "       mi=%s trap=%08x dirty=%08x\n",
	     MiToName (trap->mi), trap->TrapBits.Bits, trap->DirtyBits.Bits);
    PRINTF3 (ctxt, DBG_TPROC, "       FaultSave : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress, trap->FaultSave.s.FSR.Status);
    PRINTF3 (ctxt, DBG_TPROC, "       DataFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->DataFaultSave.s.FaultAddress, trap->DataFaultSave.s.EventAddress, trap->DataFaultSave.s.FSR.Status);
    PRINTF3 (ctxt, DBG_TPROC, "       InstFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->InstFaultSave.s.FaultAddress, trap->InstFaultSave.s.EventAddress, trap->InstFaultSave.s.FSR.Status);
    PRINTF3 (ctxt, DBG_TPROC, "       OpenFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->OpenFaultSave.s.FaultAddress, trap->OpenFaultSave.s.EventAddress, trap->OpenFaultSave.s.FSR.Status);
    
    PRINTF4 (ctxt, DBG_TPROC, "       g0=%08x g1=%08x g2=%08x g3=%08x\n", 
	     trap->Registers[REG_GLOBALS+(0^WordEndianFlip)], trap->Registers[REG_GLOBALS+(1^WordEndianFlip)], 
	     trap->Registers[REG_GLOBALS+(2^WordEndianFlip)], trap->Registers[REG_GLOBALS+(3^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       g4=%08x g5=%08x g6=%08x g7=%08x\n", 
	     trap->Registers[REG_GLOBALS+(4^WordEndianFlip)], trap->Registers[REG_GLOBALS+(5^WordEndianFlip)], 
	     trap->Registers[REG_GLOBALS+(6^WordEndianFlip)], trap->Registers[REG_GLOBALS+(7^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       o0=%08x o1=%08x o2=%08x o3=%08x\n", 
	     trap->Registers[REG_OUTS+(0^WordEndianFlip)], trap->Registers[REG_OUTS+(1^WordEndianFlip)], 
	     trap->Registers[REG_OUTS+(2^WordEndianFlip)], trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       o4=%08x o5=%08x o6=%08x o7=%08x\n", 
	     trap->Registers[REG_OUTS+(4^WordEndianFlip)], trap->Registers[REG_OUTS+(5^WordEndianFlip)], 
	     trap->Registers[REG_OUTS+(6^WordEndianFlip)], trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       l0=%08x l1=%08x l2=%08x l3=%08x\n", 
	     trap->Registers[REG_LOCALS+(0^WordEndianFlip)], trap->Registers[REG_LOCALS+(1^WordEndianFlip)],
	     trap->Registers[REG_LOCALS+(2^WordEndianFlip)], trap->Registers[REG_LOCALS+(3^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       l4=%08x l5=%08x l6=%08x l7=%08x\n", 
	     trap->Registers[REG_LOCALS+(4^WordEndianFlip)], trap->Registers[REG_LOCALS+(5^WordEndianFlip)],
	     trap->Registers[REG_LOCALS+(6^WordEndianFlip)], trap->Registers[REG_LOCALS+(7^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       i0=%08x i1=%08x i2=%08x i3=%08x\n", 
	     trap->Registers[REG_INS+(0^WordEndianFlip)], trap->Registers[REG_INS+(1^WordEndianFlip)],
	     trap->Registers[REG_INS+(2^WordEndianFlip)], trap->Registers[REG_INS+(3^WordEndianFlip)]);
    PRINTF4 (ctxt, DBG_TPROC, "       i4=%08x i5=%08x i6=%08x i7=%08x\n", 
	     trap->Registers[REG_INS+(4^WordEndianFlip)], trap->Registers[REG_INS+(5^WordEndianFlip)],
	     trap->Registers[REG_INS+(6^WordEndianFlip)], trap->Registers[REG_INS+(7^WordEndianFlip)]);
	    

    BumpUserStat (ctxt, TProcTraps);

    switch (trap->mi)
    {
    case MI_UnimplementedError:
    {
	/*
	 * This occurs if the threads processor trapped. All other cases will be for the ucode
	 * thread trapping.
	 */
	int restart = 1;
	int skip    = 0;
	
	PRINTF1 (ctxt, DBG_TPROC, "TProc: Mi=Unimp. Using trap->TrapBits=%x\n", trap->TrapBits.Bits);
	
	/*
	 * Data Access Exception.
	 */
	if (trap->TrapBits.s.DataAccessException)
	{
	    ASSERT (CTXT_IS_KERNEL(ctxt) || trap->DataFaultSave.s.FSR.Status == 0 ||
		    ctxt->Capability.cap_mycontext == trap->DataFaultSave.s.FaultContext);

	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: DataAccessException %08x\n", trap->DataFaultSave.s.FaultAddress);

	    if ((res = elan3_pagefault (ctxt, &trap->DataFaultSave, 1)) != ESUCCESS)
	    {
		PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: elan3_pagefault failed for data %08x\n",
			 trap->DataFaultSave.s.FaultAddress);

		if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, THREAD_PROC, trap, &trap->DataFaultSave, res) != OP_IGNORE)
		    restart = 0;
	    }
	}
	
	/* 
	 * Instruction Access Exception.
	 */
	if (trap->TrapBits.s.InstAccessException)
	{
	    ASSERT (CTXT_IS_KERNEL (ctxt) || trap->InstFaultSave.s.FSR.Status == 0 ||
		    ctxt->Capability.cap_mycontext == trap->InstFaultSave.s.FaultContext);
	    
	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: InstAccessException %08x\n", trap->InstFaultSave.s.FaultAddress);

	    if ((res = elan3_pagefault (ctxt, &trap->InstFaultSave, 1)) != ESUCCESS)
	    {
		PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: elan3_pagefault failed for inst %08x\n",
			 trap->InstFaultSave.s.FaultAddress);

		ElanException (ctxt, EXCEPTION_INVALID_ADDR, THREAD_PROC, trap, &trap->InstFaultSave, res);
		restart = 0;
	    }
	}
	
	/*
	 * Forced TProc trap/Unimplemented instruction
	 *
	 *  If there is a force tproc trap then don't look at 
	 *  the unimplemented instruction bit - since it can
	 *  be set in obscure circumstances.
	 */
	if (trap->TrapBits.s.ForcedTProcTrap)
	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: forced tproc trap, restarting\n");
	else if (trap->TrapBits.s.Unimplemented)
	{
	    E3_uint32 instr = ELAN3_OP_LOAD32 (ctxt, trap->pc & PC_MASK);

	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: unimplemented instruction %08x\n", instr);

	    if ((instr & OPCODE_MASK) == OPCODE_Ticc &&
		(instr & OPCODE_IMM)  == OPCODE_IMM &&
		(Ticc_COND(instr)     == Ticc_TA))
	    {
		switch (INSTR_IMM(instr))
		{
		case ELAN3_ELANCALL_TRAPNUM:
		    /*
		     * Since the thread cannot easily access the global variable which holds
		     * the elan system call number, we provide a different trap for the elan
		     * system call, and copy the system call number into %g1 before calling
		     * ThreadSyscall().
		     */
		    BumpUserStat (ctxt, ThreadElanCalls);

		    if (ThreadElancall (ctxt, trap, &skip) != ESUCCESS)
		    {
			ElanException (ctxt, EXCEPTION_BAD_SYSCALL, THREAD_PROC, trap);
			restart = 0;
		    }
		    break;

		case ELAN3_SYSCALL_TRAPNUM:
		    BumpUserStat (ctxt, ThreadSystemCalls);

		    if (ThreadSyscall (ctxt, trap, &skip) != ESUCCESS)
		    {
			ElanException (ctxt, EXCEPTION_BAD_SYSCALL, THREAD_PROC, trap);
			restart = 0;
		    }
		    break;

		case ELAN3_DEBUG_TRAPNUM:
		    ElanException (ctxt, EXCEPTION_DEBUG, THREAD_PROC, trap);
		    skip = 1;
		    break;
		    
		case ELAN3_ABORT_TRAPNUM:
		default:
		    ElanException (ctxt, EXCEPTION_UNIMP_INSTR, THREAD_PROC, trap, instr);
		    restart = 0;
		    break;
		}
		    
	    }
	    else
	    {
		ElanException (ctxt, EXCEPTION_UNIMP_INSTR, THREAD_PROC, trap, instr);
		restart = 0;
	    }
	}
	
	/*
	 * Faulted fetching routes.
	 */
	if (trap->TrapBits.s.OpenRouteFetch)
	{
	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: OpenRouteFetch %08x\n", trap->OpenFaultSave.s.FaultAddress);
	    
	    if ((res = ResolveVirtualProcess (ctxt, trap->OpenFaultSave.s.FaultAddress)) != ESUCCESS &&
		ElanException (ctxt, EXCEPTION_INVALID_PROCESS, THREAD_PROC, trap, trap->DataFaultSave.s.FaultAddress, res) != OP_IGNORE)
	    {
		restart = 0;
	    }
	    else if (RollThreadToClose (ctxt, trap, E3_PAckDiscard) != ESUCCESS)	/* Force a discard */
	    {
		restart = 0;
	    }
	}
	
	/*
	 * Thread Timeout
	 */
	if (trap->TrapBits.s.ThreadTimeout)
	{
	    if (ElanException (ctxt, EXCEPTION_PACKET_TIMEOUT, THREAD_PROC, trap) != OP_IGNORE)
		restart = 0;
	    else
	    {
		PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: timeout or PAckError!\n");
		
		/* Might deschedule the thread for a while or mark the link error here. */
		if (! trap->TrapBits.s.OutputWasOpen && RollThreadToClose (ctxt, trap, trap->TrapBits.s.PacketAckValue) != ESUCCESS)
		{
		    restart = 0;
		}
	    }
	}
	
	/*
	 * Open exception
	 */
	if (trap->TrapBits.s.OpenException)
	{
	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: open exception\n");
	    if (ElanException (ctxt, EXCEPTION_THREAD_KILLED, THREAD_PROC, trap) != OP_IGNORE)
		restart = 0;
	}
	
	/*
	 * Too many instructions.
	 */
	if (trap->TrapBits.s.TrapForTooManyInsts)
	{
	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: too many instructions\n");
	    if (ElanException (ctxt, EXCEPTION_THREAD_KILLED, THREAD_PROC, trap) != OP_IGNORE)
		restart = 0;
	}
	
	if (restart)
	{
	    /*
	     * If the output was open when the trap was taken then the trap code must move
	     * the PC on past the close instruction and simulate the effect of all the instructions
	     * that do not output onto the link. The value of the ack received is then used to
	     * simulate the close instruction.
	     */
	    if (trap->TrapBits.s.OutputWasOpen && RollThreadToClose(ctxt, trap, trap->TrapBits.s.PacketAckValue) != ESUCCESS)
	    {
		/*
		 * Don't restart if we couldn't roll it forweards 
		 * to a close instruction.
		 */
		break;
	    }

	    /*
	     * We must check back 3 instructions from the PC,  and if we see the
	     * c_close_cookie() sequence then we must execute the instructions to
	     * the end of it.
	     */
	    /* XXXX: code to be written */
	    
	    StackPointer = SaveThreadToStack (ctxt, trap, skip);
	    
	    ReissueStackPointer (ctxt, StackPointer);
	}
	
	break;
    }
    
    /*
     * This case is different from the others as %o6 has been overwritten with
     * the SP. The real PC can be read from StartPC and written back
     * into %o6 on the stack.
     */
    case MI_TProcNext:			/* Reading the outs block */
    {
	E3_Addr stack = (trap->sp & SP_MASK) - sizeof (E3_Stack);

	if (ELAN3_OP_START_FAULT_CHECK (ctxt))
	{
	    ELAN3_OP_END_FAULT_CHECK (ctxt);

	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: faulted writing StartPc to o6\n");
	    ElanException (ctxt, EXCEPTION_CANNOT_SAVE_THREAD, THREAD_PROC, NULL);
	    break;
	}
	ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Outs[6]), trap->StartPC & PC_MASK);
	ELAN3_OP_END_FAULT_CHECK (ctxt);
	/* DROPTHROUGH */
    }
    /*
     * all of these will be generated when starting up a thread.
     * Just re-issue the command after fixing the trap. The ucode keeps the startup
     * from trap information in Thread_Desc_SP while it is still loading the regs.
     */
    case MI_WaitForGlobalsRead:		/* Reading the globals block (trap restart) */
    case MI_WaitForNPCRead:		/* Reading the nPC, V and C (trap restart) */
    case MI_WaitForPCload:		/* Reading the PC, N and Z (trap restart) */
    case MI_WaitForInsRead:		/* Reading the ins block (trap restart) */
    case MI_WaitForLocals:		/* Reading the ins block (trap restart) */
    case MI_WaitForPCload2:		/* Reading the PC (normal thread start) */
    case MI_WaitForSpStore:		/* Writing the SP to the outs block */
	PRINTF2 (ctxt, DBG_TPROC, "ResolveTProcTrap: %s %08x\n", MiToName (trap->mi), trap->InstFaultSave.s.FaultAddress);

	if ((res = elan3_pagefault (ctxt, &trap->FaultSave, 1)) != ESUCCESS)
	{
	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: elan3_pagefault failed at %08x\n",
		     trap->FaultSave.s.FaultAddress);
	    if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, THREAD_PROC, &trap->FaultSave, trap, res) != OP_IGNORE)
		break;
	}

	ReissueStackPointer (ctxt, trap->sp);
	break;
	
	/*
	 * These traps could occur after the threads proc has stopped (either for a wait,
	 * break, or suspend, but not a trap). Must simulate the uCode's job.
	 */
    case MI_WaitForOutsWrite:		/* Writing the outs block */
    case MI_WaitForNPCWrite:		/* Writing the nPC block */
    {
	E3_uint32 DeschedBits = (trap->TrapBits.Bits & E3_TProcDescheduleMask);
	E3_Addr   stack       = (trap->sp & SP_MASK) - sizeof (E3_Stack);
	
	PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: trapped on %s while stopping a thread\n", MiToName(trap->mi));
	
	/*
	 * Copy npc into o6.
	 */
	trap->Registers[REG_OUTS+(6^WordEndianFlip)] = trap->npc;
	
	if (ELAN3_OP_START_FAULT_CHECK (ctxt))
	{
	    ELAN3_OP_END_FAULT_CHECK (ctxt);

	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: faulted writing outs to stack\n");
	    ElanException (ctxt, EXCEPTION_CANNOT_SAVE_THREAD, THREAD_PROC, NULL);
	    break;
	}
	
	/*
	 * Now write the outs back to the stack. NOTE then endian flip is undone.
	 */
	for (i = 0; i < 8; i++)
	    ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Outs[i]), trap->Registers[REG_OUTS+(i^WordEndianFlip)]);
	ELAN3_OP_END_FAULT_CHECK (ctxt);
	
	/*
	 * thread has been saved. Now find out why the thread proc stopped.
	 */
	if (DeschedBits == E3_TProcDescheduleSuspend)
	{
	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: suspend instruction executed\n");
	    break;
	}
	
	/*
	 * Break. Just reissue the command.
	 */
	if (DeschedBits == E3_TProcDescheduleBreak)
	{
	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: break instruction, reissue sp %08x\n", trap->sp);
	    ReissueStackPointer (ctxt, trap->sp);
	    break;
	}
	
	ASSERT (DeschedBits == E3_TProcDescheduleWait);
	 
	/* DROPTHROUGH to fix up a wait event */
    }
    
    /*
     * Trapped here trying to execute a wait instruction. All the thread state has already
     * been saved and the trap has been fixed so simplest thing to do is to start the
     * thread up at the wait instruction again.
     */
    case MI_WaitForEventWaitAddr:	/* Reading back the %o0,%o1 pair for a
					   wait event instr. */
    case MI_WaitForWaitEventAccess:	/* Locked dword read of the event location.
      					   Note that this read is done with write
					   permissions so we never get a trap on the write */
    {
	E3_Addr stack = (trap->sp & SP_MASK) - sizeof (E3_Stack);
	
	if ((res = elan3_pagefault (ctxt, &trap->FaultSave, 1)) != ESUCCESS)
	{
	    PRINTF1 (ctxt, DBG_TPROC, "ResolveTProcTrap: elan3_pagefault failed at %08x\n", 
		     trap->FaultSave.s.FaultAddress);
	    if (ElanException (ctxt, EXCEPTION_INVALID_ADDR, THREAD_PROC, trap, &trap->DataFaultSave, res) != OP_IGNORE)
		break;
	}

	if (ELAN3_OP_START_FAULT_CHECK (ctxt))
	{
	    ELAN3_OP_END_FAULT_CHECK (ctxt);

	    PRINTF0 (ctxt, DBG_TPROC, "ResolveTProcTrap: faulted writing pc to stack\n");
	    ElanException (ctxt, EXCEPTION_CANNOT_SAVE_THREAD, THREAD_PROC, NULL);
	    break;
	}

	ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Outs[6]), trap->pc);
	ELAN3_OP_END_FAULT_CHECK (ctxt);
	
	ReissueStackPointer (ctxt, trap->sp);
	break;
    }
    
    /*
     * Assume the fault will be fixed by FixupEventTrap.
     */
    default:
	FixupEventTrap (ctxt, THREAD_PROC, trap, trap->mi, &trap->FaultSave, 0);
	break;
    }
}

int
TProcNeedsRestart (ELAN3_CTXT *ctxt)
{
    return (ctxt->ItemCount[LIST_THREAD] != 0);
}

void
RestartTProcItems (ELAN3_CTXT *ctxt)
{
    void     *item;
    E3_uint32 StackPointer;

    kmutex_lock (&ctxt->SwapListsLock);
    
    while (ctxt->ItemCount[LIST_THREAD])
    {
	if (! ELAN3_OP_GET_WORD_ITEM (ctxt, LIST_THREAD, &item, &StackPointer))
	    ctxt->ItemCount[LIST_THREAD] = 0;
	else
	{
	    if (IssueCommand (ctxt, offsetof (E3_CommandPort, RunThread), StackPointer, 0) == ISSUE_COMMAND_RETRY)
	    {
		ELAN3_OP_PUTBACK_ITEM (ctxt, LIST_THREAD, item);
		kmutex_unlock (&ctxt->SwapListsLock);
		return;
	    }
	    
	    ctxt->ItemCount[LIST_THREAD]--;
	    ELAN3_OP_FREE_WORD_ITEM (ctxt, item);
	}
    }
    kmutex_unlock (&ctxt->SwapListsLock);
}

E3_Addr
SaveThreadToStack (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int SkipInstruction)
{
    E3_Addr      stack = (trap->sp & SP_MASK) - sizeof (E3_Stack);
    E3_Addr      orflag;
    register int i;

    /*
     * When the thread deschedules normally, the N & Z flags are written 
     * to the stack in o6, and the V & C flags are lost.
     * Since the Elan will store the NPC into o6 (to skip the instruction), 
     * the CC flags are visible to the trap handler in the trapped PC and NPC.   
     * If the instruction needs to be re-executed then the CC flags need to be
     * kept in the right place to be read in when the thread re-starts.
     *
     * PC  has N & Z from trapped NPC.
     * NPC has V & C from trapped PC.
     */
    if (SkipInstruction)
    {
	trap->Registers[REG_OUTS+(6^WordEndianFlip)]    = trap->npc;
	trap->Registers[REG_GLOBALS+(0^WordEndianFlip)] = ((trap->npc & PC_MASK) + 4) | (trap->pc & CC_MASK);
    }
    else
    {
	trap->Registers[REG_OUTS+(6^WordEndianFlip)]    = (trap->pc & PC_MASK) | (trap->npc & CC_MASK);
	trap->Registers[REG_GLOBALS+(0^WordEndianFlip)] = (trap->npc & PC_MASK) | (trap->pc & CC_MASK);
    }
    
    if (ELAN3_OP_START_FAULT_CHECK(ctxt))
    {
	PRINTF0 (ctxt, DBG_TPROC, "RestartThread: faulted writing out thread\n");
	ELAN3_OP_END_FAULT_CHECK(ctxt);

	ElanException (ctxt, EXCEPTION_CANNOT_SAVE_THREAD, THREAD_PROC, NULL);
	return ((E3_Addr) 0);
    }


#ifdef DEBUG_PRINTF
    PRINTF4 (ctxt, DBG_TPROC, "SaveThreadToStack: SP=%08x PC=%08x NPC=%08x DIRTY=%08x\n",
	     trap->sp, trap->pc, trap->npc, trap->DirtyBits.Bits);
    if (trap->DirtyBits.s.GlobalsDirty)
    {
	PRINTF4 (ctxt, DBG_TPROC, "       g0=%08x g1=%08x g2=%08x g3=%08x\n", 
		 trap->Registers[REG_GLOBALS+(0^WordEndianFlip)], trap->Registers[REG_GLOBALS+(1^WordEndianFlip)], 
		 trap->Registers[REG_GLOBALS+(2^WordEndianFlip)], trap->Registers[REG_GLOBALS+(3^WordEndianFlip)]);
	PRINTF4 (ctxt, DBG_TPROC, "       g4=%08x g5=%08x g6=%08x g7=%08x\n", 
		 trap->Registers[REG_GLOBALS+(4^WordEndianFlip)], trap->Registers[REG_GLOBALS+(5^WordEndianFlip)], 
		 trap->Registers[REG_GLOBALS+(6^WordEndianFlip)], trap->Registers[REG_GLOBALS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.OutsDirty)
    {
	PRINTF4 (ctxt, DBG_TPROC, "       o0=%08x o1=%08x o2=%08x o3=%08x\n", 
		 trap->Registers[REG_OUTS+(0^WordEndianFlip)], trap->Registers[REG_OUTS+(1^WordEndianFlip)], 
		 trap->Registers[REG_OUTS+(2^WordEndianFlip)], trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
	PRINTF4 (ctxt, DBG_TPROC, "       o4=%08x o5=%08x o6=%08x o7=%08x\n", 
		 trap->Registers[REG_OUTS+(4^WordEndianFlip)], trap->Registers[REG_OUTS+(5^WordEndianFlip)], 
		 trap->Registers[REG_OUTS+(6^WordEndianFlip)], trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.LocalsDirty)
    {
	PRINTF4 (ctxt, DBG_TPROC, "       l0=%08x l1=%08x l2=%08x l3=%08x\n", 
		 trap->Registers[REG_LOCALS+(0^WordEndianFlip)], trap->Registers[REG_LOCALS+(1^WordEndianFlip)], 
		 trap->Registers[REG_LOCALS+(2^WordEndianFlip)], trap->Registers[REG_LOCALS+(3^WordEndianFlip)]);
	PRINTF4 (ctxt, DBG_TPROC, "       l4=%08x l5=%08x l6=%08x l7=%08x\n", 
		 trap->Registers[REG_LOCALS+(4^WordEndianFlip)], trap->Registers[REG_LOCALS+(5^WordEndianFlip)], 
		 trap->Registers[REG_LOCALS+(6^WordEndianFlip)], trap->Registers[REG_LOCALS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.InsDirty)
    {
	PRINTF4 (ctxt, DBG_TPROC, "       i0=%08x i1=%08x i2=%08x i3=%08x\n", 
		 trap->Registers[REG_INS+(0^WordEndianFlip)], trap->Registers[REG_INS+(1^WordEndianFlip)], 
		 trap->Registers[REG_INS+(2^WordEndianFlip)], trap->Registers[REG_INS+(3^WordEndianFlip)]);
	PRINTF4 (ctxt, DBG_TPROC, "       i4=%08x i5=%08x i6=%08x i7=%08x\n", 
		 trap->Registers[REG_INS+(4^WordEndianFlip)], trap->Registers[REG_INS+(5^WordEndianFlip)], 
		 trap->Registers[REG_INS+(6^WordEndianFlip)], trap->Registers[REG_INS+(7^WordEndianFlip)]);
    }
#endif 
    
    PRINTF1 (ctxt, DBG_TPROC, "flushing registers to stack %08x\n", stack);

    /* 
     * NOTE - store the register to the stack in reverse order, since the stack 
     * will be allocated in sdram, and we cannot use the sdram accessing functions 
     * here, as it is "mapped" in user-space.
     */
    for (i = 0; i < 8; i++)
    {
	if (trap->DirtyBits.s.GlobalsDirty & (1 << i))
	    ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Globals[i]), trap->Registers[REG_GLOBALS+(i^WordEndianFlip)]);
	if (trap->DirtyBits.s.OutsDirty & (1 << i))
	    ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Outs[i]),    trap->Registers[REG_OUTS+(i^WordEndianFlip)]);
	if (trap->DirtyBits.s.LocalsDirty & (1 << i))
	    ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Locals[i]),  trap->Registers[REG_LOCALS+(i^WordEndianFlip)]);
	if (trap->DirtyBits.s.InsDirty & (1 << i))
	    ELAN3_OP_STORE32 (ctxt, stack + offsetof (E3_Stack, Ins[i]),     trap->Registers[REG_INS+(i^WordEndianFlip)]);
    }

    /* always restore all registers */
    orflag = ThreadRestartFromTrapBit | ThreadReloadAllRegs;
    
    ELAN3_OP_END_FAULT_CHECK (ctxt);

    return (trap->sp | orflag);
}

void
ReissueStackPointer (ELAN3_CTXT *ctxt, E3_Addr StackPointer)
{
    PRINTF1 (ctxt, DBG_TPROC, "ReissueStackPointer : Queue SP %08x\n", StackPointer);
    
    kmutex_lock (&ctxt->SwapListsLock);
    ctxt->ItemCount[LIST_THREAD]++;
    ELAN3_OP_PUT_WORD_ITEM (ctxt, LIST_THREAD, StackPointer);
    kmutex_unlock (&ctxt->SwapListsLock);
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
