/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: tproc_linux.c,v 1.22.2.1 2005/05/31 10:29:07 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/tproc_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <asm/mman.h>
#include <linux/file.h>

#ifdef NO_ABI
#include <asm/poll.h>
extern asmlinkage long sys_open(const char *, int, int);
extern asmlinkage ssize_t sys_write(unsigned int, const char *, size_t);
extern asmlinkage ssize_t sys_read(unsigned int, char *, size_t);
extern asmlinkage off_t sys_lseek(unsigned int, off_t, unsigned int);
extern asmlinkage long sys_poll(struct pollfd *, unsigned int, long);
extern asmlinkage long sys_kill(int, int); 
#elif  LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
# include <linux/syscall.h>
#else
# include <linux/syscalls.h>
#endif

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/elansyscall.h>
#include <elan3/threadsyscall.h>

/*
 * NOTE: system calls from kernel on Linux are different on alpha and i386 
 *       on alpha they return -errno on failure 
 *       on i386  they return -1 on failure and set errno 
 */

static void
ReturnSyscall (THREAD_TRAP *trap, unsigned long rc, int *skip)
{
    if (rc >= (unsigned long) (-130))
    {
	trap->pc |= PSR_C_BIT;  /* clear carry to indicate failure */

	trap->Registers[REG_OUTS+(0^WordEndianFlip)] = -rc;
    } 
    else 
    {
	trap->pc &= ~PSR_C_BIT; /* set carry to indicate success */
	trap->Registers[REG_OUTS+(0^WordEndianFlip)] = rc;
    }
    trap->Registers[REG_OUTS+(1^WordEndianFlip)] = 0;
    *skip = 1;
}

static void 
dump_regs(ELAN3_CTXT *ctxt, THREAD_TRAP *trap)
{
    PRINTF (ctxt, DBG_TPROC, "               OUTS %08x %08x %08x %08x\n",
      trap->Registers[REG_OUTS+(0^WordEndianFlip)], 
      trap->Registers[REG_OUTS+(1^WordEndianFlip)],
      trap->Registers[REG_OUTS+(2^WordEndianFlip)], 
      trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
    PRINTF (ctxt, DBG_TPROC, "                    %08x %08x %08x %08x\n",
      trap->Registers[REG_OUTS+(4^WordEndianFlip)], 
      trap->Registers[REG_OUTS+(5^WordEndianFlip)],
      trap->Registers[REG_OUTS+(6^WordEndianFlip)], 
      trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
}

int
ThreadSyscall (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int *skip)
{
    int	          code;
    caddr_t       maddr;
    struct file  *file;
    unsigned long rc;
    int           i;
    uintptr_t     av[6];
    uintptr_t     ptr;
   
    PRINTF (ctxt, DBG_TPROC, "ThreadSyscall: PC %08x G1 %08x\n", 
      trap->pc, trap->Registers[REG_GLOBALS+(1^WordEndianFlip)]);
    dump_regs(ctxt, trap);

    code = trap->Registers[REG_GLOBALS+(1^WordEndianFlip)];

    /* Copy the system call arguments from %o0-%o5 */
    for (i = 0; i < 6; i++)
	av[i] = trap->Registers[REG_OUTS+(i^WordEndianFlip)];
    
    rc = (unsigned long) -EINVAL;

    switch (code) {
#if defined(IOPROC_PATCH_APPLIED)
	case ELAN3_SYS_open:
	    maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[0]);
	    if (maddr != NULL)
	        rc = sys_open((const char *)maddr, av[1], av[2]);
	    break;

	case ELAN3_SYS_close:
	    rc = sys_close(av[0]);
	    break;

	case ELAN3_SYS_write:
	    maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[1]);
	    if (maddr != NULL)
		rc = sys_write(av[0], (const char *)maddr, av[2]);
	    break;

	case ELAN3_SYS_read:
	    maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[1]);
	    if (maddr != NULL)
		rc = sys_read(av[0], (char *)maddr, av[2]);
	    break;

	case ELAN3_SYS_poll:
	    maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[0]);
	    if (maddr != NULL)
		rc = sys_poll((struct pollfd *)maddr, av[1], av[2]);
	    break;
	
    	case ELAN3_SYS_lseek:
	    rc = sys_lseek(av[0], av[1], av[2]);
	    break;
	    
	case ELAN3_SYS_mmap:
	    if ((E3_Addr) av[0] == (E3_Addr) 0)
		maddr = NULL;
	    else if ((maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[0])) == NULL)
		break;
	
	    file = NULL;
	    /* GNAT 5515: If *not* anonymous memory need to do fget */
	    if ((av[3] & MAP_ANONYMOUS) == 0 && (file = fget (av[4])) == NULL)
	    {
		rc = -EBADF;
		break;
	    }

	    down_write (&current->mm->mmap_sem);
	    ptr = do_mmap_pgoff (file, (unsigned long) maddr, av[1], av[2], av[3], av[5] >>PAGE_SHIFT);
	    up_write (&current->mm->mmap_sem);

	    if (file)
		fput (file);
	    
	    if (IS_ERR((void *) ptr))
		rc = PTR_ERR((void *) ptr);
	    else
		rc = elan3mmu_elanaddr (ctxt->Elan3mmu, (caddr_t)ptr);

	    break;
	
	case ELAN3_SYS_munmap:
	    maddr = elan3mmu_mainaddr (ctxt->Elan3mmu, (E3_Addr) av[0]);

#ifdef AC
	    if (maddr != NULL)
	   	rc = do_munmap(current->mm, (unsigned long) maddr, av[1], 1);
#else
	    if (maddr != NULL)
	   	rc = do_munmap(current->mm, (unsigned long) maddr, av[1]);
#endif
	    break;

	case ELAN3_SYS_kill:
	    rc = sys_kill(av[0], av[1]);
	    break;

	case ELAN3_SYS_getpid:
	    rc = current->pid;
	    break;
#else

#warning "NO IOPROC patch applied - thread cannot perform system calls"

#endif /* defined(IOPROC_PATCH_APPLIED) */

	default:
	    return EINVAL;
    }
    ReturnSyscall(trap, rc, skip);
    return ESUCCESS;
}


int
ThreadElancall (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int *skip)
{
	int ret = ESUCCESS;

	PRINTF (ctxt, DBG_TPROC, "ThreadElancall: PC %08x\n", trap->pc);
	dump_regs(ctxt, trap);

	/*
	 * Elan system call 'type' is passed in o0
	 */
	switch (trap->Registers[REG_OUTS+(0^WordEndianFlip)]) 
	{
	default:
		ret = EINVAL;
		break;
	}
	return ret;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
