/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: tprocinsts.c,v 1.20 2003/09/24 13:57:25 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/os/tprocinsts.c,v $*/

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#include <elan3/urom_addrs.h>
#include <elan3/thread.h>
#include <elan3/vmseg.h>
#include <elan3/elan3mmu.h>

#define MAXINSTR	256		/* # Instructions to look at while looking for close */

static E3_uint32 ALU (ELAN3_CTXT *ctxt,
		      E3_uint32 fcode, E3_uint32 X, E3_uint32 Y,
		      E3_uint32 *Z, E3_uint32 *N, E3_uint32 *C, E3_uint32 *V);

char *OpcodeNames[] =
{
   "ADD   ",
   "AND   ",
   "OR    ",
   "XOR   ",
   "SUB   ",
   "ANDN  ",
   "ORN   ",
   "XNOR  ",
   "ADDX  ",
   "UNIP  ",
   "UMUL  ",
   "SMUL  ",
   "SUBX  ",
   "UNIP  ",
   "UDIV  ",
   "SDIV  ",
   "ADDcc ",
   "ANDcc ",
   "ORcc  ",
   "XORcc ",
   "SUBcc ",
   "ANDNcc",
   "ORNcc ",
   "XNORcc",
   "ADDXcc",
   "UNIPcc",
   "UMULcc",
   "SMULcc",
   "SUBXcc",
   "UNIPcc",
   "UDIVcc",
   "SDIVcc"
};

#define REGISTER_VALUE(trap, rN)		(((rN) == 0) ? 0 : (trap)->Registers[(rN)^WordEndianFlip])
#define ASSIGN_REGISTER(trap, rN, value)	((rN) != 0 ? trap->Registers[(rN)^WordEndianFlip] = (value) : 0)

int
RollThreadToClose (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, E3_uint32 PAckVal)
{
    E3_Addr   pc      = (trap->pc & PC_MASK);
    E3_Addr   npc     = (trap->npc & PC_MASK);
    E3_uint32 Z       = (trap->npc & PSR_Z_BIT) ? 1 : 0;
    E3_uint32 N       = (trap->npc & PSR_N_BIT) ? 1 : 0;
    E3_uint32 C       = (trap->pc  & PSR_C_BIT) ? 1 : 0;
    E3_uint32 V       = (trap->pc  & PSR_V_BIT) ? 1 : 0;
    E3_uint32 instr;
    E3_Addr   addr;

    if (ELAN3_OP_START_FAULT_CHECK (ctxt))
    {
    failed:
	ELAN3_OP_END_FAULT_CHECK (ctxt);

	ElanException (ctxt, EXCEPTION_SIMULATION_FAILED, THREAD_PROC, trap);
	return (EFAULT);
    }

    /*
     * Thread trapped with output open, or while closing,
     * so roll the PC forwards to the instruction after the
     * next c_close, and execute that with the register
     * specified in c_close set to the trap which occured.
     * (This is not 1 which means an ACK)
     */
    PRINTF1 (ctxt, DBG_TPROC, "RollThreadToClose: roll pc %x to c_close\n", pc);
    
    for (;;)
    {
	instr = ELAN3_OP_LOAD32 (ctxt, pc);

	PRINTF2 (ctxt, DBG_TPROC, "RollThreadToClose: PC=%x INSTR=%x\n", pc, instr);

	switch (OPCODE_CLASS(instr))
	{
	case OPCODE_CLASS_0:
	    switch ((instr) & OPCODE_CLASS0_MASK)
	    {
	    case OPCODE_SETHI:
		PRINTF3 (ctxt, DBG_TPROC, "PC %x : sethi r%d = %x\n", pc, INSTR_RD(instr), instr << 10);

		ASSIGN_REGISTER (trap, INSTR_RD(instr), instr << 10);
		break;

	    case OPCODE_SENDREG:
		PRINTF1 (ctxt, DBG_TPROC, "PC %x : sendreg\n", pc);
		break;
		
	    case OPCODE_SENDMEM:
		PRINTF1 (ctxt, DBG_TPROC, "PC %x : sendmem\n", pc);
		break;
		
	    case OPCODE_BICC:
	    {
		int     DoBranch   = (instr >> 28) & 1;
		int	CondBranch = 1;
		E3_Addr OldnPC     = npc;

		PRINTF5 (ctxt, DBG_TPROC, "PC %x : Bicc Z=%x N=%x C=%x V=%x ", pc, Z, N, C, V);
		switch (instr & OPCODE_BICC_MASK)
		{
		case OPCODE_BICC_BN:	CondBranch = 0; 		break;
		case OPCODE_BICC_BE:	DoBranch ^= Z;			break;
		case OPCODE_BICC_BLE:	DoBranch ^= Z | (N ^ V);	break;
		case OPCODE_BICC_BL:	DoBranch ^= N ^ V;		break;
		case OPCODE_BICC_BLEU:	DoBranch ^= C | Z; 		break;
		case OPCODE_BICC_BCS: 	DoBranch ^= C; 			break;
		case OPCODE_BICC_BNEG: 	DoBranch ^= N; 			break;
		case OPCODE_BICC_BVS: 	DoBranch ^= V; 			break;
		}

		/* Do the branch */
		if (DoBranch != 0)
		{
		    npc = pc + (((instr & 0x3fffff) << 2) |
				(((instr & 0x200000) != 0) ? 0xff000000 : 0));
		    
		    PRINTF2 (ctxt, DBG_TPROC, "PC %x : branch taken to %x\n", pc, npc);
		}
		else
		{
		    npc = npc + 4;
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : branch not taken\n", pc);
		}
		pc = OldnPC;

		/* Test if the next is annuled */
		if (((instr & OPCODE_BICC_ANNUL) != 0) &
		    ((DoBranch == 0) | (CondBranch == 0)))
		{
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : branch annulled\n", pc);

		    pc = npc;
		    npc += 4;
		}

		/*
		 * we've already consumed the instruction - so continue rather
		 * than break;
		 */
		continue;
	    }
	    
	    default:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : unknown class 0 instr %x\n", pc, instr);
		goto failed;
	    }
	    break;

	case OPCODE_CLASS_1:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : unknown class 1 instr %x\n", pc, instr);
		goto failed;
		
	case OPCODE_CLASS_2:
	{
	    E3_uint32 X = REGISTER_VALUE (trap, INSTR_RS1(instr));
	    E3_uint32 Y = (instr & OPCODE_IMM) ? INSTR_IMM(instr) : REGISTER_VALUE (trap, INSTR_RS2(instr));
	    
	    if ((instr & OPCODE_NOT_ALUOP) == 0)
	    {
		E3_uint32 fcode  = (instr >> OPCODE_FCODE_SHIFT) & OPCODE_FCODE_MASK;
		E3_uint32 result = ALU (ctxt, fcode, X, Y, &Z, &N, &C, &V);

		PRINTF5 (ctxt, DBG_TPROC, "PC %x : %s %x %x -> %x", pc, OpcodeNames[fcode], X, Y, result);
		PRINTF4 (ctxt, DBG_TPROC, "        Z=%x N=%x C=%x V=%x\n", Z, N, C, V);
		
		ASSIGN_REGISTER (trap, INSTR_RD(instr), result);
	    }
	    else
	    {
		switch (instr & OPCODE_MASK)
		{
		case OPCODE_OPEN:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : c_open\n", pc);
		    break;

		case OPCODE_CLOSE:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : c_close\n", pc);
		    goto found_close;

		case OPCODE_SLL:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : SLL\n", pc);

		    ASSIGN_REGISTER (trap, INSTR_RD(instr), X << Y);
		    break;
		    
		case OPCODE_SRL:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : SRL\n", pc);
		    
		    ASSIGN_REGISTER (trap, INSTR_RD(instr), X >> Y);
		    break;
		    
		case OPCODE_SRA:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : SRA\n", pc);
		    
		    ASSIGN_REGISTER (trap, INSTR_RD(instr), X >> Y);
		    break;
		    
		case OPCODE_BREAKTEST:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : BREAKTEST  not allowed while open\n", pc);
		    goto failed;
		    
		case OPCODE_BREAK:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : BREAK not allowed while open\n", pc);
		    goto failed;

		case OPCODE_SUSPEND:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : SUSPEND not allowed while open\n", pc);
		    goto failed;
		    
		case OPCODE_WAIT:
		    PRINTF1 (ctxt, DBG_TPROC, "PC %x : WAIT not allowed while open\n", pc);
		    goto failed;

		default:
		    PRINTF2 (ctxt, DBG_TPROC, "PC %x : unknown class 2 instr %x\n", pc, instr);
		    goto failed;
		}
	    }
	    break;
	}
	
	case OPCODE_CLASS_3:
	{
	    if ((instr & OPCODE_IMM) != 0)
		addr = REGISTER_VALUE (trap, INSTR_RS1(instr)) + INSTR_IMM(instr);
	    else
		addr = (REGISTER_VALUE (trap, INSTR_RS1(instr)) + 
			REGISTER_VALUE (trap, INSTR_RS2(instr)));

	    switch (instr & OPCODE_MASK)
	    {
	    case OPCODE_LD:
		PRINTF3 (ctxt, DBG_TPROC, "PC %x : LD [%x], r%d\n", pc, addr, INSTR_RD(instr));
		
		ASSIGN_REGISTER (trap, INSTR_RD(instr), ELAN3_OP_LOAD32 (ctxt, addr));
		break;
		
	    case OPCODE_LDD:
	    case OPCODE_LDBLOCK16:
	    case OPCODE_LDBLOCK32:
	    case OPCODE_LDBLOCK64:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : LDBLOCKx @ %x is not possible while output open\n", pc, addr);
		goto failed;
	    
	    case OPCODE_ST:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : ST @ %x\n", pc, addr);
		
		ELAN3_OP_STORE32 (ctxt, addr, REGISTER_VALUE (trap, INSTR_RD(instr)));
		break;
			      
	    case OPCODE_STD:
	    case OPCODE_STBLOCK16:
	    case OPCODE_STBLOCK32:
	    case OPCODE_STBLOCK64:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : STD @ %x is not posisble while output open\n", pc, addr);
		goto failed;

	    case OPCODE_SWAP:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : SWAP @ %x is not posible while output open\n", pc, addr);
		goto failed;
		
	    default:
		PRINTF2 (ctxt, DBG_TPROC, "PC %x : unknown class 3 instr %x\n", pc, instr);
		goto failed;
	    }
	    break;
	}}

	pc = npc;
	npc += 4;
    }
    
found_close:
    ELAN3_OP_END_FAULT_CHECK (ctxt);

    PRINTF1 (ctxt, DBG_TPROC, "PC %x : c_close\n", pc);
    
    /*
     * Found the new pc, and have the close instruction in *instr
     */
    ASSIGN_REGISTER (trap, INSTR_RD(instr), PAckVal);
    
    /*
     * Move to instruction after close.
    */
    trap->pc = npc;
    
    /* Insert the value of Z and N from the close inst */
    trap->npc = (npc + 4) | ((PAckVal == E3_PAckOk) ? 1 :
			     (PAckVal == E3_PAckTestFail) ? 2 : 0);

    return (ESUCCESS);
}

E3_uint32
ALU (ELAN3_CTXT *ctxt,
     E3_uint32 fcode, E3_uint32 X, E3_uint32 Y,
     E3_uint32 *Z, E3_uint32 *N, E3_uint32 *C, E3_uint32 *V)
{
    E3_uint32 XMSB, YMSB, ZMSB, Cprime;
    E3_uint32 Yprime;
    E3_uint32 Result=0;
    
    Yprime = ((fcode >> 2) & 1) ? ~Y : Y;
    Cprime = ((fcode >> 2) & 1) ^ (*C & ((fcode >> 3) & 1));
    XMSB             = (X >> 31) & 1;
    YMSB             = (Yprime >> 31) & 1;
    /* mul or div */
    if ((fcode & 0xa) == 0xa)
    {
	PRINTF0 (ctxt, DBG_TPROC, "ALU: tried a multiply or a divide\n");
	return (0);
    }

    switch (fcode & 3)
    {
	/*ADD */
    case 0:
	Result = X + Yprime + Cprime ;
	if ((fcode & 0x10) == 0)
	    return (Result);
	
	ZMSB   = Result >> 31;
	*V = ((XMSB & YMSB & ~ZMSB) | (~XMSB &~YMSB &  ZMSB));
	*C = ((fcode >> 2) & 1) ^ ( (XMSB & YMSB) | (~ZMSB & (XMSB | YMSB)));
	break;
	
	/*AND */
    case 1:
	Result = X & Yprime ;
	if ((fcode & 0x10) == 0)
	    return (Result);
	
	*V = 0;
	*C = 0;
	break;
	
	/*OR  */
    case 2:
	Result = X | Yprime ;
	if ((fcode & 0x10) == 0)
	    return (Result);
	
	*V = 0;
	*C = 0;
	break;
	
	/*XOR */
    case 3:
	Result = X ^ Yprime ;
	if ((fcode & 0x10) == 0)
	    return (Result);
	
	*V = 0;
	*C = 0;
	break;
    }
    
    *Z = (Result == 0) ? 1 : 0;
    *N = (Result >> 31) & 1;

    return (Result);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
