/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: debug.c,v 1.17 2005/03/23 11:06:15 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/debug.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>

unsigned       elan4_debug           = 0;
unsigned       elan4_debug_toconsole = 0;
unsigned       elan4_debug_tobuffer  = DBG_ALL;

unsigned       elan4_debug_display_ctxt;
unsigned       elan4_debug_ignore_ctxt;
unsigned       elan4_debug_ignore_type;

void
elan4_debug_init()
{
    if ((elan4_debug & elan4_debug_tobuffer) != 0)
	qsnet_debug_alloc();
}

void
elan4_debug_fini()
{
}

void
elan4_debugf (void *type, int mode, char *fmt,...)
{
    char    prefix[128];
    int     where = 0;
    va_list ap;

    if ((mode & elan4_debug_tobuffer) != 0 || type == DBG_BUFFER)
	where |= QSNET_DEBUG_BUFFER;
    if ((mode & elan4_debug_toconsole) != 0 || type == DBG_CONSOLE)
	where |= QSNET_DEBUG_CONSOLE;

    if (where == 0)
	return;
    
    if ((unsigned long) type > DBG_NTYPES)
    {
	ELAN4_CTXT *ctxt = (ELAN4_CTXT *) type;

        if (elan4_debug_display_ctxt && ctxt->ctxt_num != elan4_debug_display_ctxt)
            return;
        if (elan4_debug_ignore_ctxt  && ctxt->ctxt_num == elan4_debug_ignore_ctxt)
            return;

	sprintf (prefix, "[%08ld.%04d] elan4 (%03x) ", lbolt,  current->pid, ctxt->ctxt_num);
    }
    else if ((unsigned long) type == (int) DBG_CONSOLE)
	prefix[0] = '\0';
    else
    {
	char *what;

	if (elan4_debug_ignore_type & (1 << ((unsigned long) type)))
	    return;

	switch ((unsigned long) type)
	{
	case (int) DBG_DEVICE: what = "dev"; break;
	case (int) DBG_USER:   what = "usr"; break;
	default:               what = NULL; break;
	}
	    
	if (what)
	    sprintf (prefix, "[%08ld.%04d] elan4 [%s] ", lbolt, current->pid, what);
	else
	    sprintf (prefix, "[%08ld.%04d] elan4 [%3d] ", lbolt, current->pid, (int)(long)type);
    }

    va_start(ap,fmt);
    qsnet_vdebugf (where, prefix, fmt, ap);
    va_end (ap);
}

int
elan4_assfail (ELAN4_CTXT *ctxt, const char *ex, const char *func, const char *file, const int line)
{
    qsnet_debugf (QSNET_DEBUG_BUFFER, "elan%d: assertion failure: %s, function: %s, file %s, line: %d\n", 
		  ctxt->ctxt_dev->dev_instance, ex, func, file, line);

    printk (KERN_EMERG "elan%d: assertion failure: %s, function: %s, file %s, line: %d\n", 
		   ctxt->ctxt_dev->dev_instance, ex, func, file, line);

    if (panicstr)
	return 0;
    
    if (assfail_mode & 1)				/* return to BUG() */
	return 1;
    
    if (assfail_mode & 2)
	panic ("elan%d: assertion failure: %s, function: %s, file %s, line: %d\n", 
	       ctxt->ctxt_dev->dev_instance, ex, func, file, line);

    if (assfail_mode & 4)
	elan4_debug = 0;
    
    return 0;
    
}

int
elan4_debug_trigger (ELAN4_CTXT *ctxt, const char *func, const char *file, const int line, const char *fmt, ...)
{
    va_list ap;

    va_start (ap, fmt);
    qsnet_vdebugf (QSNET_DEBUG_CONSOLE|QSNET_DEBUG_BUFFER, "", fmt, ap);
    va_end (ap);

    printk (KERN_EMERG "elan%d: debug trigger: function: %s, file %s, line: %d\n", ctxt->ctxt_dev->dev_instance, func, file, line);

    if (panicstr)
	return 0;
    
    if (assfail_mode & 1)				/* return to BUG() */
	return 1;
    
    if (assfail_mode & 2)
	panic ("elan%d: debug trigger: function: %s, file %s, line: %d\n", ctxt->ctxt_dev->dev_instance, func, file, line);

    if (assfail_mode & 4)
	elan4_debug = 0;
    
    return 0;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
