/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: device.c,v 1.106.2.5 2005/08/09 09:57:04 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/device.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan4/sdram.h>
#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/commands.h>
#include <elan4/trtype.h>
#include <elan4/neterr.h>

#include <elan4/i2c.h>
#include <elan3/vpd.h>

/* allow this code to compile against an Eagle elanmod */
#ifdef __ELANMOD_DEVICE_H
#define ELAN_DEV_OPS		ELANMOD_DEV_OPS
#define ELAN_DEV_OPS_VERSION	ELANMOD_DEV_OPS_VERSION
#define elan_dev_register	elanmod_dev_register
#define elan_dev_deregister	elanmod_dev_deregister
#endif

/* XXXX configurational defines */

#if defined (CONFIG_MPSAS)
#define HASH_0_SIZE_VAL			(12 + 6)
#define HASH_1_SIZE_VAL			(2 + 6)
#define CTXT_TABLE_SHIFT		8
#define LN2_MAX_CQS			8		/* 256 */
#else
#define HASH_0_SIZE_VAL			(13 + 6)
#define HASH_1_SIZE_VAL			(2 + 6)
#define CTXT_TABLE_SHIFT		12
#define LN2_MAX_CQS			10		/* 1024 */
#endif

unsigned int elan4_hash_0_size_val       = HASH_0_SIZE_VAL;
unsigned int elan4_hash_1_size_val       = HASH_1_SIZE_VAL;
unsigned int elan4_ctxt_table_shift      = CTXT_TABLE_SHIFT;
unsigned int elan4_ln2_max_cqs           = LN2_MAX_CQS;
unsigned int elan4_dmaq_highpri_size     = 2;			/* 8192 entries */
unsigned int elan4_threadq_highpri_size  = 1;			/* 1024 entries */
unsigned int elan4_dmaq_lowpri_size      = 2;			/* 8192 entries */
unsigned int elan4_threadq_lowpri_size   = 1;			/* 1024 entries */
unsigned int elan4_interruptq_size       = 0;			/* 1024 entries */
unsigned int elan4_mainint_punt_loops    = 1;
unsigned int elan4_mainint_resched_ticks = 0;
unsigned int elan4_linkport_lock	 = 0xbe0fcafe;		/* default link port lock */
unsigned int elan4_eccerr_recheck        = 1;

static int 
elan4_op_get_position (void *arg, ELAN_POSITION *ptr)
{
    ELAN4_DEV     *dev = (ELAN4_DEV *)arg;
    ELAN_POSITION  pos;

    elan4_get_position (dev, &pos);

    return copyout (&pos, ptr, sizeof (ELAN_POSITION));
}

static int 
elan4_op_set_position (void *arg, unsigned short nodeid, unsigned short numnodes)
{
    /* XXXXX 

       ELAN4_DEV *dev = (ELAN4_DEV *) arg;

       compute_position (&pos, nodeid, numnode, num_down_links_value);

       return elan4_set_position (dev, pos);
    */
    return EINVAL;
}

ELAN_DEV_OPS elan4_dev_ops = 
{
    elan4_op_get_position,
    elan4_op_set_position,

    ELAN_DEV_OPS_VERSION
};

static E4_uint32
elan4_read_filter (ELAN4_DEV *dev, unsigned networkctx)
{
    return (elan4_sdram_readl (dev, dev->dev_ctxtable + (networkctx * sizeof (E4_ContextControlBlock)) + 
			       offsetof (E4_ContextControlBlock, Filter)));
}

static void
elan4_write_filter (ELAN4_DEV *dev, unsigned networkctx, E4_uint32 value)
{
    elan4_sdram_writel (dev, (dev->dev_ctxtable + (networkctx * sizeof (E4_ContextControlBlock)) +
			offsetof (E4_ContextControlBlock, Filter)), value);
    pioflush_sdram(dev);
}

void
elan4_set_schedstatus (ELAN4_DEV *dev, E4_uint32 intreg)
{
    E4_uint32 setbits  = 0;
    E4_uint32 intmask  = 0;
    E4_uint32 haltmask;
    E4_uint32 next_sched;
    E4_uint32 next_intmask;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_intmask_lock, flags);

    haltmask = (dev->dev_haltop_mask | dev->dev_haltop_active);

    if ((haltmask & INT_DProcHalted) || dev->dev_halt_all_count || dev->dev_halt_dproc_count)
	setbits |= SCH_DProcHalt;
    
    if ((haltmask & INT_TProcHalted) || dev->dev_halt_all_count || dev->dev_halt_tproc_count)
	setbits |= SCH_TProcHalt;

    if ((haltmask & INT_CProcHalted) || dev->dev_halt_all_count || dev->dev_halt_cproc_count)
	setbits |= SCH_CProcHalt;

    if ((haltmask & INT_DiscardingLowPri) || dev->dev_discard_all_count || dev->dev_discard_lowpri_count)
	setbits |= SCH_DiscardLowPriInput;
    
    if ((haltmask & INT_DiscardingHighPri) || dev->dev_discard_all_count || dev->dev_discard_highpri_count)
	setbits |= SCH_DiscardHighPriInput;
    
    if (dev->dev_halt_lowpri_count)
	setbits |= SCH_StopLowPriQueues;
    
    if (haltmask & INT_DProcHalted) intmask |= INT_DProcHalted;
    if (haltmask & INT_TProcHalted) intmask |= INT_TProcHalted;
    if (haltmask & INT_CProcHalted) intmask |= INT_CProcHalted;
    if (haltmask & INT_DiscardingLowPri) intmask |= INT_DiscardingLowPri;
    if (haltmask & INT_DiscardingHighPri) intmask |= INT_DiscardingHighPri;

    next_intmask = (dev->dev_intmask     & ~(INT_Halted | INT_Discarding)) | (intmask & ~intreg);
    next_sched   = (dev->dev_schedstatus & ~(SCH_Halt | SCH_Discard))      | setbits;

    PRINTF5 (DBG_DEVICE, DBG_REGISTER, "elan4_set_schedstatus: haltmask=%x setbits=%x intmask=%x next_sched=%x next_intmask=%x\n",
	     haltmask, setbits, intmask, next_sched, next_intmask);

    CHANGE_INT_MASK (dev, next_intmask);
    CHANGE_SCHED_STATUS (dev, next_sched);

    spin_unlock_irqrestore (&dev->dev_intmask_lock, flags);
}

int
elan4_route2str (E4_VirtualProcessEntry *route, char *routeStr)
{
    int        part = 0;
    int        shift;
    int        broadcast;
    E4_uint64  value;
    char      *ptr = routeStr;
    int        b;

    /* unpack first */
    value = route->Values[part] & 0x7f;
    if ( (value & 0x78) == 0) {
        /* empty route */
        strcpy(routeStr,"Invalid lead route");
        return (-EINVAL);
    }

    if ( value & 0x40 ) {
        /* broad cast */
	strcpy(routeStr,"Broadcast");
	return (-EINVAL);
    } else {
        switch ((value  & 0x30) >> 4) {
        case 0: { *ptr++ = '0' + (value & 0x7); break; }
        case 1: { *ptr++ = 'M';                 break; }
        case 2: { *ptr++ = 'U';                 break; }
        case 3: { *ptr++ = 'A';                 break; }
        }
    }

    shift = 16;
    broadcast = 0;
    while ( 1 ) {
        b =  (route->Values[part] >> shift) & 0xf;

        if ( broadcast ) {
            /* about to pick up the second byte of a broadcast pair */
            broadcast = 0;
        } else {
            if ( b & 0x8) {
                /*  output link */
                 *ptr++ = '0' + (b & 0x7);
            } else {
                if ( b & 0x4) {
                    /* broad cast */
                    broadcast = 1;
                } else {
                    switch ( b & 0x3 ) {
                    case 0: { *ptr++ =  0 ; return (0);     break; }
                    case 1: { *ptr++ = 'M';                 break; }
                    case 2: { *ptr++ = 'U';                 break; }
                    case 3: { *ptr++ = 'A';                 break; }
                    }
                }
            }
        }

        shift += 4; 
        if ( part != 0 ) {
            if ( shift > 36) {
                /* too far, now in the crc value */
                strcpy(routeStr,"Invalid route length");
                return (-EINVAL);
            }
        } else { 
            if ( shift >= 64) { 
                /* move to the next 64 bits */
                part = 1;
                shift = 2;
            }
        }
    }

    /* never reached */
    return (-EINVAL);
}

static int elan4_hardware_lock_count = 0;

void
elan4_hardware_lock_check(ELAN4_DEV *dev, char *from)
{

    int reg = read_reg32 (dev, CommandSchedDataPort[2]);

    /* dont spam too much */
    if ( elan4_hardware_lock_count++ > 10)  return;

    printk ("elan%d: %s timed out intmask=0x%x InterruptReg=0x%x (%d)\n", dev->dev_instance, from, dev->dev_intmask, read_reg32 (dev, InterruptReg), elan4_hardware_lock_count);

    /* an 0xF in either and we need to output more */
    if ((reg & 0xf0) || ( reg & 0x0f)) {
	ELAN4_ROUTE_RINGBUF *ringbuf;
	char                 routestr[33];
	
	printk ("elan%d: CommandSchedDataPort[0] 0x%016x 0x%016x 0x%016x 0x%016x\n", 
		dev->dev_instance, 
		read_reg32 (dev, CommandSchedDataPort[0]),
		read_reg32 (dev, CommandSchedDataPort[1]),
		reg,
		read_reg32 (dev, CommandSchedDataPort[3])
	    );
	/* dump out /proc/qsnet/elan4/deviceN/stats/cproctimeoutroutes */
	printk ("elan%d: cat of /proc/qsnet/elan4/device%d/stats/cproctimeoutroutes\n", dev->dev_instance, dev->dev_instance);

	ringbuf = &dev->dev_cproc_timeout_routes;
	
	if (!ringbuf) 
	    printk ("elan%d: No stats available\n", dev->dev_instance);
	else
	{
	    int start;
	    int end;
	    int i;
	    
	    memset(&routestr, 0, 33);
	    
	    start = ringbuf->start;
	    end = ringbuf->end;
	    
	    if (end < start)
		end = DEV_STASH_ROUTE_COUNT;
	    
	    for (i=start; i<end; i++) 
	    {
		elan4_route2str (&ringbuf->routes[i], routestr);
		printk ( "elan%d: Route %llx %llx->%s\n", dev->dev_instance, (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
	    }
	    
	    if (ringbuf->end < start)
	    {
		start = 0;
		end = ringbuf->end;
		for (i=start; i<end; i++)
		{
		    elan4_route2str (&ringbuf->routes[i], routestr);
		    printk ( "elan%d: Route %llx %llx->%s\n", dev->dev_instance, (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
		}
	    }
	    
	}    
    }
}
static void
dev_haltop_timer_func (unsigned long arg)
{
    ELAN4_DEV *dev  = (ELAN4_DEV *) arg;

    elan4_hardware_lock_check(dev,"haltop");
}

void
elan4_queue_haltop (ELAN4_DEV *dev, ELAN4_HALTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_haltop_lock, flags);

    /* add to the end of the halt operations list */
    list_add_tail (&op->op_link, &dev->dev_haltop_list);

    if ((dev->dev_haltop_mask & op->op_mask) != op->op_mask)
    {
	dev->dev_haltop_mask |= op->op_mask;
	
	elan4_set_schedstatus (dev, 0);
    }

    mod_timer (&dev->dev_haltop_timer, (jiffies + (HZ*10))); /* 10 seconds */

    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);
}

void
elan4_queue_intop (ELAN4_DEV *dev, ELAN4_CQ *cq, ELAN4_INTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_intop_lock, flags);

    op->op_cookie = INTOP_ONESHOT | ((dev->dev_intop_cookie++) & INTOP_VALUE_MASK);

    list_add_tail (&op->op_link, &dev->dev_intop_list);

    writeq ((op->op_cookie << E4_MAIN_INT_SHIFT) | INTERRUPT_CMD, (void *)(cq->cq_mapping));

    spin_unlock_irqrestore (&dev->dev_intop_lock, flags);
}

void
elan4_register_intop (ELAN4_DEV *dev, ELAN4_INTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_intop_lock, flags);

    op->op_cookie = INTOP_PERSISTENT | ((dev->dev_intop_cookie++) & INTOP_VALUE_MASK);

    list_add_tail (&op->op_link, &dev->dev_intop_list);

    spin_unlock_irqrestore (&dev->dev_intop_lock, flags);
}

void
elan4_deregister_intop (ELAN4_DEV *dev, ELAN4_INTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_intop_lock, flags);
    list_del (&op->op_link);
    spin_unlock_irqrestore (&dev->dev_intop_lock, flags);
}

static __inline__ void
__issue_dma_flushop_cmd (ELAN4_DEV *dev, ELAN4_CQ *cq)
{
    E4_uint64 eventaddr = dev->dev_tproc_space + 64;

    writeq (WAIT_EVENT_CMD | eventaddr,   (void *)(cq->cq_mapping));
    writeq (0,                            (void *)(cq->cq_mapping));
    writeq (0,                            (void *)(cq->cq_mapping));
    writeq (0,                            (void *)(cq->cq_mapping));

    writeq (DMA_ShMemWrite | RUN_DMA_CMD, (void *)(cq->cq_mapping));
    writeq (0 /* cookie */,               (void *)(cq->cq_mapping));
    writeq (0 /* vproc */,                (void *)(cq->cq_mapping));
    writeq (0 /* srcAddr */,              (void *)(cq->cq_mapping));
    writeq (0 /* dstAddr */,              (void *)(cq->cq_mapping));
    writeq (0 /* srcEvent */,             (void *)(cq->cq_mapping));
    writeq (0 /* dstEvent */,             (void *)(cq->cq_mapping));
    writeq (SET_EVENT_CMD,                (void *)(cq->cq_mapping));
}

static void
handle_dma_flushops_intop (ELAN4_DEV *dev, void *arg)
{
    unsigned int  hipri        = ((unsigned long) arg & 1);
    E4_uint64     status       = dev->dev_dma_flushop[hipri].status;
    ELAN4_CQ     *cq           = dev->dev_dma_flushop[hipri].cq;
    sdramaddr_t   cqdesc       = dev->dev_cqaddr + (elan4_cq2num(cq) * sizeof (E4_CommandQueueDesc));
    E4_uint64     queuePtrs    = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs));
    E4_uint32     completedPtr = CQ_CompletedPtr(queuePtrs);
    E4_uint32     size         = CQ_Size ((queuePtrs >> CQ_SizeShift) & CQ_SizeMask);
    unsigned long flags;

    /*
     * Since we're called from a main interrupt which was issued through the approriate
     * flushcq the command queue descriptor for dma flushing can no longer be in the 
     * insert cache, nor can it be in the extractor (as it's trapped), hence it is
     * safe to modify the completed pointer
     */

    spin_lock_irqsave (&dev->dev_haltop_lock, flags);

    ASSERT (status != 0);

    /* skip over either the DMA/SETEVENT or just the SETEVENT depending on the trap type */
    if (CPROC_TrapType (status) == CommandProcDmaQueueOverflow)
	completedPtr = (completedPtr & ~(size-1)) | ((completedPtr + 64) & (size - 1));
    else
	completedPtr = (completedPtr & ~(size-1)) | ((completedPtr + 8) & (size - 1));
    
    elan4_sdram_writel (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs) + 4,
			((queuePtrs >> 32) & ~CQ_PtrOffsetMask) | (completedPtr & CQ_PtrOffsetMask));
    
    elan4_restartcq (dev, dev->dev_dma_flushop[hipri].cq);

    if (! list_empty (&dev->dev_dma_flushop[hipri].list))
	__issue_dma_flushop_cmd (dev, dev->dev_dma_flushop[hipri].cq);

    dev->dev_dma_flushop[hipri].status = 0;
    
    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);

}

static void
handle_dma_flushops (ELAN4_DEV *dev, E4_uint64 status, int cqnum)
{
    unsigned int       hipri  = (cqnum == elan4_cq2num(dev->dev_dma_flushop[1].cq) ? 1 : 0);
    ELAN4_CQ          *cq     = dev->dev_dma_flushop[hipri].cq;
    ELAN4_CQ          *flushq = dev->dev_flush_cq[elan4_cq2num(cq) & (COMMAND_INSERTER_CACHE_ENTRIES-1)];
    struct list_head  *ops;
    unsigned long      flags;
    int		       qfull,count;
    E4_uint64	       queuePtrs;
    LIST_HEAD(list);
    
    spin_lock_irqsave (&dev->dev_haltop_lock, flags);
    
    ASSERT (cqnum == elan4_cq2num (dev->dev_dma_flushop[hipri].cq));
    ASSERT (! list_empty (&dev->dev_dma_flushop[hipri].list));
    ASSERT (dev->dev_dma_flushop[hipri].status == 0);
    
    /* remove the whole list */
    ops = dev->dev_dma_flushop[hipri].list.next;

    list_del_init (&dev->dev_dma_flushop[hipri].list);
    
    /* and add it to our local list */
    list_add_tail (&list, ops);
    
    /* now determine whether the queue was full - since it cannot be empty 
     * then if the front and back pointers are the same then it is full */
    queuePtrs = hipri ? read_reg64 (dev, DProcHighPriPtrs) : read_reg64 (dev, DProcLowPriPtrs);
    qfull     = (E4_QueueFrontPointer (queuePtrs) == E4_QueueBackPointer (queuePtrs));
    
    if (CPROC_TrapType(status) == CommandProcDmaQueueOverflow && !qfull)
	printk (" ******* queue overflow trap - but queue not full\n");

    if (qfull && CPROC_TrapType(status) != CommandProcDmaQueueOverflow)
	printk (" ****** queue full - but not overflow trap : %llx %llx %x\n", 
		read_reg64 (dev, DProcLowPriPtrs), read_reg64 (dev, DProcHighPriPtrs), CPROC_TrapType(status));

    /* Store the status register, this also indicates that the intop is pending */
    dev->dev_dma_flushop[hipri].status = status;

    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);

    /* Issue a main interrupt command to the approriate flush command queue,
     * which will then safely update the completed pointer to skip over the
     * command which has trapped, also prevent any new commands to be issued
     * to the command queue.
     */
    dev->dev_dma_flushop[hipri].intop.op_function = handle_dma_flushops_intop;
    dev->dev_dma_flushop[hipri].intop.op_arg      = (void *) (unsigned long) hipri;

    elan4_queue_intop (dev, flushq, &dev->dev_dma_flushop[hipri].intop);
    
    /* now execute all operations */
    for (count = 0; ! list_empty (&list); count++)
    {
	ELAN4_DMA_FLUSHOP *op = list_entry (list.next, ELAN4_DMA_FLUSHOP, op_link);
	
	list_del (&op->op_link);
	
	(*op->op_function) (dev, op->op_arg, qfull);
    }

    /* finally release the "reasons" for halting */
    spin_lock_irqsave (&dev->dev_haltop_lock, flags);
    if ((dev->dev_halt_dproc_count -= count) == 0)
	elan4_set_schedstatus (dev, 0);
    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);
	
    return;
}

void
elan4_queue_dma_flushop (ELAN4_DEV *dev, ELAN4_DMA_FLUSHOP *op, int hipri)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_haltop_lock, flags);

    if (dev->dev_halt_dproc_count++ == 0)			/* ensure that the DMA processor cannot */
	elan4_set_schedstatus (dev, 0);				/* execute the DMA we issue. */

    if (list_empty (&dev->dev_dma_flushop[hipri].list) && dev->dev_dma_flushop[hipri].status == 0)
	__issue_dma_flushop_cmd (dev, dev->dev_dma_flushop[hipri].cq);
	
    list_add_tail (&op->op_link, &dev->dev_dma_flushop[hipri].list);

    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);
}

static void
enable_elan_errors (void *arg)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) arg;

    ENABLE_INT_MASK (dev, INT_ErrorInterrupts);
}

#define ERROR_DISABLE_PERIOD	(hz/2)
#define ERROR_SAMPLE_PERIOD	(hz/10)
#define ERROR_LIMIT		(100)

static __inline__ void
check_error_rate (ELAN4_DEV *dev)
{
    if (dev->dev_error_time == (lbolt/ERROR_SAMPLE_PERIOD))
    {
        if (++dev->dev_errors_per_period >= ERROR_LIMIT && (dev->dev_intmask & INT_ErrorInterrupts))
	{
	    DISABLE_INT_MASK (dev, INT_ErrorInterrupts);
	    
	    schedule_timer_fn (&dev->dev_error_timeoutid, enable_elan_errors, (void *) dev, ERROR_DISABLE_PERIOD);
	}
    }
    else
    {
	dev->dev_error_time        = (lbolt/ERROR_SAMPLE_PERIOD);
	dev->dev_errors_per_period = 0;
    }
}

static __inline__ int
handle_mainints (ELAN4_DEV *dev, int nticks, int nintr)
{
    E4_uint32 nfptr = dev->dev_interruptq_nfptr;
    E4_uint32 bptr  = read_reg32 (dev, MainIntQueuePtrs.s.Back);
    E4_uint32 qsize = E4_QueueSize(elan4_interruptq_size);
    E4_uint32 qmask = qsize - 1;
    long      tlim  = lbolt + nticks;
    int       done = 0;
    unsigned long flags;

    do {
	int todo  = ((bptr - nfptr) & qmask) / E4_MainIntEntrySize;

	ASSERT (todo > 0);

	PRINTF4 (DBG_DEVICE, DBG_MAININT, "handle_mainints: fptr %x nfptr %x bptr %x : %d todo\n", 
		 read_reg32 (dev, MainIntQueuePtrs.s.Front), nfptr, bptr, todo);

	if (nintr >= 0 && (done + todo) > nintr)		/* punt because too may to do in interrupt */
	{
	    PRINTF4 (DBG_DEVICE, DBG_MAININT, "handle_mainints: punting (done %d todo %d) (bptr %x fptr %x)\n",
		     done, todo, bptr, read_reg32 (dev, MainIntQueuePtrs.s.Front));

	    return 1;
	}

	BucketDevStat (dev, s_mainints, todo, MainIntBuckets);

	/* consume all the entries in the queue which we think are there */
	do {
	    E4_uint64   value = elan4_sdram_readq (dev, nfptr);
	    ELAN4_CTXT *ctxt  = elan4_localctxt (dev, E4_MAIN_INT_CTX (value));
	    E4_uint32   fptr  = nfptr;

	    PRINTF2 (DBG_DEVICE, DBG_MAININT, "handle_mainints: process cookie %llx - write fptr=%x\n", value, nfptr);

	    if (ctxt == NULL)
		ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_mainints: context %d invalid\n", E4_MAIN_INT_CTX (value));
	    else
		ctxt->ctxt_ops->op_interrupt (ctxt, E4_MAIN_INT_COOKIE(value));

	    /* compute the next queue front pointer, before updating the front pointer
	     * since we need to ensure that elan4_queue_mainintop doesn't see the queue
	     * as being empty if an extra interrupt is queued in between */
	    dev->dev_interruptq_nfptr = nfptr = (nfptr & ~qmask) | ((nfptr + sizeof (E4_uint64)) & qmask);
    
	    /* update the queue front pointer, doing this will clear the
	     * interrupt for *all* interrupt cookies which have previously 
	     * been added to the queue */
	    write_reg32 (dev, MainIntQueuePtrs.s.Front, E4_QueueFrontValue (fptr, elan4_interruptq_size));
	    pioflush_reg (dev);
	} while (bptr != nfptr);
	
	/* re-sample the back pointer and if it's different from the previous
	 * queue front pointer, then the queue has something on it again */
	done += todo;
	
	if ((nticks > 0 && ((int) (lbolt - tlim)) > 0))		/* been executing for too long in thread */
	    return 1;

	bptr = read_reg32 (dev, MainIntQueuePtrs.s.Back);

	PRINTF3 (DBG_DEVICE, DBG_MAININT, "handle_mainints: resample : fptr %x nfptr %x bptr %x\n", 
		 read_reg32 (dev, MainIntQueuePtrs.s.Front), nfptr, bptr);

	/* at this point we've made some space in the interrupt queue,
	 * so check to see if we've got anything to restart */
	spin_lock_irqsave (&dev->dev_mainint_lock, flags);
	while (! list_empty (&dev->dev_interruptq_list))
	{
	    ELAN4_INTOP *op = list_entry (dev->dev_interruptq_list.next, ELAN4_INTOP, op_link);
	    
	    list_del (&op->op_link);

	    op->op_function (dev, op->op_arg);
	}
	spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);

    } while (bptr != nfptr);

    return 0;
}

static void
elan4_mainint_thread (ELAN4_DEV *dev)
{
    unsigned long flags;

    kernel_thread_init ("elan4_mainint");
    
    spin_lock_irqsave (&dev->dev_mainint_lock, flags);
    for (;;)
    {
	if (dev->dev_stop_threads)
	    break;
	
	if (! (dev->dev_intmask & INT_MainInterrupt))
	{
	    spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);
	    
	    if (handle_mainints (dev, elan4_mainint_resched_ticks, -1))
		BumpDevStat (dev, s_mainint_rescheds);

	    spin_lock_irqsave (&dev->dev_mainint_lock, flags);
	    ENABLE_INT_MASK (dev, INT_MainInterrupt);
	}
	
	kcondvar_wait (&dev->dev_mainint_wait, &dev->dev_mainint_lock, &flags);
    }

    dev->dev_mainint_stopped = 1;
    kcondvar_wakeupall (&dev->dev_mainint_wait, &dev->dev_mainint_lock);

    spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);

    kernel_thread_exit();
}

void
elan4_queue_mainintop (ELAN4_DEV *dev, ELAN4_INTOP *op)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_mainint_lock, flags);
    if (dev->dev_interruptq_nfptr == read_reg32 (dev, MainIntQueuePtrs.s.Back))
	op->op_function (dev, op->op_arg);
    else
	list_add_tail (&op->op_link, &dev->dev_interruptq_list);
    spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);
}

static __inline__ E4_uint32
handle_cproc_trap (ELAN4_DEV *dev)
{
    E4_uint32   cqptr   = read_reg32 (dev, CommandControl.CommandQueueDescsBase) & E4_QueueDescPtrMask;
    unsigned    cqnum   = ((cqptr - dev->dev_cqaddr) / sizeof (E4_CommandQueueDesc));
    sdramaddr_t cqdesc  = dev->dev_cqaddr + (cqnum * sizeof (E4_CommandQueueDesc));
    E4_uint64   control = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control));
    E4_uint64   status  = read_reg64 (dev, CProcStatus);
    ELAN4_CTXT *ctxt    = elan4_localctxt (dev, CQ_Context (control));

    PRINTF4 (DBG_DEVICE, DBG_INTR, "handle_cproc_trap: cqnum=%d status=%016llx control=%016llx TrapType\n", 
	     cqnum, status, control, CPROC_TrapType (status));
    PRINTF4 (DBG_DEVICE, DBG_INTR, "                   %016llx %016llx %016llx %016llx\n",
	     elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs)),
	     elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_HoldingValue)),
	     elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_AckBuffers)),
	     elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control)));

    BumpDevStat (dev, s_cproc_traps);

    if (ctxt == NULL)
	ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_cproc_trap: context %d is invalid\n", CQ_Context (control));
    else
	ctxt->ctxt_ops->op_cproc_trap (ctxt, status, cqnum);

    return (CPROC_TrapType (status) == CommandProcWaitTrap ? SCH_RestartCProc | SCH_RestartEProc : SCH_RestartCProc);
}

static __inline__ E4_uint32
handle_dproc_trap (ELAN4_DEV *dev, int unit)
{
    E4_uint64   status  = (unit == 0) ? read_reg64 (dev, DProc0Status) : read_reg64 (dev, DProc1Status);
    E4_uint32   restart = (unit == 0) ? SCH_RestartDma0Proc : SCH_RestartDma1Proc;
    ELAN4_CTXT *ctxt    = elan4_localctxt (dev, DPROC_Context (status));
    
    PRINTF3 (DBG_DEVICE, DBG_INTR, "handle_dproc_trap: unit %d context %d%s\n", unit, DPROC_Context(status),
	     DPROC_PrefetcherFault(status) ? " (prefetcher)" : "");

    if (DPROC_PrefetcherFault (status))
	restart |= SCH_RestartDmaPrefetchProc;
		      
    BumpDevStat (dev, s_dproc_traps);

    if (ctxt == NULL)
	ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_dproc_trap: context %d is invalid\n", DPROC_Context (status));
    else
	ctxt->ctxt_ops->op_dproc_trap (ctxt, status, unit);

    return (restart);
}

static __inline__ E4_uint32
handle_eproc_trap (ELAN4_DEV *dev)
{
    E4_uint64   status = read_reg64 (dev, EProcStatus);
    ELAN4_CTXT *ctxt   = elan4_localctxt (dev, EPROC_Context (status));

    BumpDevStat (dev, s_eproc_traps);

    if (ctxt == NULL)
	ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_eproc_trap: context %d is invalid\n", EPROC_Context (status));
    else
	ctxt->ctxt_ops->op_eproc_trap (ctxt, status);

    return (SCH_RestartEProc);
}

static __inline__ E4_uint32
handle_tproc_trap (ELAN4_DEV *dev)
{
    E4_uint64   status = read_reg64 (dev, TProcStatus);
    ELAN4_CTXT *ctxt   = elan4_localctxt (dev, TPROC_Context (status));

    BumpDevStat (dev, s_tproc_traps);

    if (ctxt == NULL)
	ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_tproc_trap: context %d is invalid\n", TPROC_Context (status));
    else
	ctxt->ctxt_ops->op_tproc_trap (ctxt, status);
    
    return (SCH_RestartTProc);
}

static __inline__ void
handle_haltints (ELAN4_DEV *dev, E4_uint32 intreg)
{
    struct list_head  list   = LIST_HEAD_INIT(list);
    E4_uint32         mask   = 0;
    E4_uint32         active = 0;
    struct list_head *entry;
    struct list_head *next;
    unsigned long     flags;

    BumpDevStat (dev, s_haltints);

    spin_lock_irqsave (&dev->dev_haltop_lock, flags);

    list_for_each_safe (entry, next, &dev->dev_haltop_list) {
	ELAN4_HALTOP *op = list_entry (entry, ELAN4_HALTOP, op_link);

	PRINTF (DBG_DEVICE, DBG_INTR, "handle_haltints: op=%p op_mask=%x intreg=%x\n", op, op->op_mask, intreg);

	if ((op->op_mask & intreg) != op->op_mask)
	    mask |= op->op_mask;
	else
	{
	    list_del (&op->op_link);				/* remove from list */
	    list_add_tail (&op->op_link, &list);		/* add to local list */

	    active |= op->op_mask;
	}
    }

    ASSERT (dev->dev_haltop_mask == (mask | active));

    dev->dev_haltop_mask = mask;

    if (list_empty (&dev->dev_haltop_list)) {
	del_timer(&dev->dev_haltop_timer);
    }

    if (list_empty (&list))
	elan4_set_schedstatus (dev, intreg);
    else
    {
	dev->dev_haltop_active = active;
	spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);

	while (! list_empty (&list)) 
	{
	    ELAN4_HALTOP *op = list_entry (list.next, ELAN4_HALTOP, op_link);
	    
	    list_del (&op->op_link);

	    (*op->op_function) (dev, op->op_arg);
	}

	spin_lock_irqsave (&dev->dev_haltop_lock, flags);
	dev->dev_haltop_active = 0;

	elan4_set_schedstatus (dev, 0);
    }

    spin_unlock_irqrestore (&dev->dev_haltop_lock, flags);
}

static __inline__ E4_uint32
handle_iproc_trap (ELAN4_DEV *dev, unsigned unit)
{
    sdramaddr_t hdroff = dev->dev_inputtraparea + offsetof (E4_IprocTrapState, TrHeader[0][unit]);
    E4_uint64   status = elan4_sdram_readq (dev, hdroff + offsetof (E4_IprocTrapHeader, IProcStatusCntxAndTrType));
    E4_uint32   filter = elan4_read_filter (dev, IPROC_NetworkContext (status));
    ELAN4_CTXT *ctxt   = elan4_localctxt (dev, filter & E4_FILTER_CONTEXT_MASK);

    /*
     * The context is not valid in the following case :
     *     ack not been sent AND bad CRC/bad length.
     *
     *  NOTE TransCRCStatus and BadLength only valid if NOT an EopTrap.
     */
    ASSERT ((IPROC_GoodAckSent (status) & (1 << IPROC_InputterChan (status))) || IPROC_EOPTrap (status) ||
	    (IPROC_TransCRCStatus (status) == CRC_STATUS_GOOD && !IPROC_BadLength (status)));
    
    BumpDevStat (dev, s_iproc_traps);

    if (ctxt == NULL)
    {
	ELAN4_DEBUG_TRIGGER (&dev->dev_ctxt, "elan4:handle_iproc_trap: network %d context %d (%x) is invalid\n", IPROC_NetworkContext (status), 
			     filter & E4_FILTER_CONTEXT_MASK, filter);
	
	elan4_write_filter (dev, IPROC_NetworkContext (status), E4_FILTER_DISCARD_ALL);
    }
    else
	ctxt->ctxt_ops->op_iproc_trap (ctxt, status, unit);

    return (SCH_RestartCh0LowPriInput << unit);
}

void
handle_pcimemerr (ELAN4_DEV *dev)
{
    elan4_pcierror (dev);

    check_error_rate (dev);
}

void
handle_sdramint (ELAN4_DEV *dev)
{
    E4_uint64 status    = read_reg64 (dev, SDRamECCStatus);
    E4_uint64 ConfigRegValue = read_reg64 (dev, SDRamConfigReg);
    char      errstr[200];
    int	      i;
    int	      Found = 0;

    PRINTF0 (DBG_DEVICE, DBG_INTR, "handle_sdramint\n");

    printk ("elan%d: ECC Error %s status=%llx\n",
    	    dev->dev_instance, elan4_sdramerr2str (dev, status, ConfigRegValue, errstr), (long long)status);

    if (!ECC_UncorrectableErr(status) && !ECC_MultUncorrectErrs(status))
	printk ("elan%d: ECC error data=%016llx\n", dev->dev_instance, elan4_sdram_readq (dev, ECC_Addr(status)));

    if (ECC_CorrectableErr (status))
	BumpDevStat (dev, s_correctable_errors);
    if (ECC_MultCorrectErrs (status))
	BumpDevStat (dev, s_multiple_errors);

    if (ECC_UncorrectableErr(status))
	panic ("elan%d: uncorrectable ECC error\n", dev->dev_instance);
    if (ECC_MultUncorrectErrs(status))
	panic ("elan%d: muliple uncorrectable ECC error\n", dev->dev_instance);
    
    PULSE_SYSCONTROL (dev, CONT_CLEAR_SDRAM_ERROR);

    /*
     * Now try to test for a read/write error type.
     * This can only be done if it was a correctable error as an uncorrectable error might lockup the node.
     * It should not be attempted if the data is in the dcache because fetching again would not generate an
     * error even if the problem was a read, and flushing the cache line would fix a write probelm.
     * Reading the same location again should cause a new error if the problem was caused by a bad write.
     */
    if (elan4_eccerr_recheck &&
	(dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA) &&
        ECC_CorrectableErr(status) && !ECC_UncorrectableErr(status))
    {
	E4_uint64 status2;
	E4_uint64 Addr = ECC_Addr(status) & ~(E4_CACHELINE_SIZE-1);
	E4_uint32 SetIndex = (Addr >> 6) & ~(E4_NumCacheLines-1);
	int	  InCache = 0;

	/* check the cache tags to see if the data has been read into a cache line. */
	for (i=0; i<E4_NumCacheSets; i++)
	   if (((E4_uint32)__elan4_readq (dev, dev->dev_regs + offsetof(E4_Registers, Tags.Tags[i][SetIndex].Value)) & 0x7fffe000) == (Addr & 0x7fffe000))
	   {
	       InCache = 1;
	       break;
	   }

	if (InCache == 0)
	{
	    printk ("elan%d: checking if ECC error was read or write\n", dev->dev_instance);

	    /* Now read and throw away the answer. A read of a word will schedule a block read of sdram */
	    elan4_sdram_readq (dev, Addr);
	    status2 = read_reg64 (dev, SDRamECCStatus);
	    if ((Addr == (ECC_Addr(status2) & ~(E4_CACHELINE_SIZE-1))) && ECC_CorrectableErr(status2))	// Write error.
	    {
	        status = (status & ~0x0030000000000000ULL) | 0x0010000000000000ULL;
		PULSE_SYSCONTROL (dev, CONT_CLEAR_SDRAM_ERROR);
	    }
	    else
	        status = (status & ~0x0030000000000000ULL) | 0x0020000000000000ULL;
	}
	else
	    status = status | 0x0030000000000000ULL;
    }
    else
	status &= ~0x0030000000000000ULL;

    /* search for this error already being logged */
    for (i = sizeof (dev->dev_sdramerrs)/sizeof (dev->dev_sdramerrs[0]) - 1; i >= 0; i--)
        if ((dev->dev_sdramerrs[i].EccStatus == status) && (dev->dev_sdramerrs[i].ConfigReg == ConfigRegValue))
	{
            Found = 1;
	    dev->dev_sdramerrs[i].ErrorCount += 1; // Keep a count.
	    break;
	}

    /* stash the status for /proc */
    if (!Found)
    {
	for (i = sizeof (dev->dev_sdramerrs)/sizeof (dev->dev_sdramerrs[0]) - 1; i > 0; i--)
	    dev->dev_sdramerrs[i] = dev->dev_sdramerrs[i-1];
	dev->dev_sdramerrs[0].EccStatus = status;
	dev->dev_sdramerrs[0].ConfigReg = ConfigRegValue;
	dev->dev_sdramerrs[0].ErrorCount = 1; // First error
    }
    
    check_error_rate (dev);
}

static void
clear_linkerr_led (void *arg)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) arg;

    write_i2c (dev, I2cStatus, read_i2c (dev, I2cStatus) | I2cCntl_ClearLinkError);
}

void
handle_linkerror (ELAN4_DEV *dev)
{
    E4_uint32 LinkState;
    E4_uint32 CurrState = read_reg32 (dev, LinkControlReg);

    /* Set for reading errors. */
    write_reg32 (dev, LinkControlReg,
                 (CurrState = CurrState & ~((LCONT_TEST_CONTROL_MASK << LCONT_TEST_CONTROL_SHIFT) |
					    (LCONT_TEST_VALUE_MASK << LCONT_TEST_VALUE_SHIFT))));
    LinkState = LCONT_LINK_STATE(CurrState = read_reg32 (dev, LinkControlReg));

#ifdef DEBUG
    {
	E4_uint8 ErrorMsg[256], DataErrorVal[64];

	strcpy (ErrorMsg, "handle_linkerror:");
	if (LinkState & LS_LockError)	 strcat (ErrorMsg, " LockError");
	if (LinkState & LS_DeskewError)	 strcat (ErrorMsg, " DeskewError");
	if (LinkState & LS_PhaseError)	 strcat (ErrorMsg, " PhaseError");
	if (LinkState & LS_DataError)
	{
	    E4_uint32 error[4];
	    E4_uint32 i;
	    strcat (ErrorMsg, " DataError");
	    /* Errors */
	    for(i = LRS_ErrorVal8to0; i <= LRS_ErrorVal35to27; i++)
	    {
		write_reg32 (dev, LinkControlReg,
			     CurrState | LCONT_TEST_VALUE(i) | (LCONT_READ_STATE << LCONT_TEST_CONTROL_SHIFT));
		error[i - LRS_ErrorVal8to0] = LCONT_LINK_STATE(read_reg32 (dev, LinkControlReg));
	    }
	    sprintf (DataErrorVal, " Link State Error Val: %09llx %03x %03x %03x %03x", 
		     (unsigned long long) ((error[0] & 0x1ffUL) | ((error[1] & 0x1ffUL) << 9)  |
				  ((error[2] & 0x1ffUL) << 18) | ((error[3] & 0x1ffUL) << 27)),
		     error[3], error[2], error[1], error[0]);
	    strcat (ErrorMsg, DataErrorVal);
	}
	if (LinkState & LS_FifoOvFlow0)	 strcat (ErrorMsg, " FifoOvFlow0");
	if (LinkState & LS_FifoOvFlow1)	 strcat (ErrorMsg, " FifoOvFlow1");
	if (LinkState & LS_Mod45Changed)	 strcat (ErrorMsg, " Mod45Changed");
	if (LinkState & LS_PAckNotSeenError) strcat (ErrorMsg, " PAckNotSeenError");
	strcat (ErrorMsg, "\n");
	PRINTF0 (DBG_DEVICE, DBG_INTR, ErrorMsg);
    }
#endif

    BumpDevStat (dev, s_link_errors);
    
    if (LinkState & LS_LockError)	 BumpDevStat (dev, s_lock_errors);
    if (LinkState & LS_DeskewError)	 BumpDevStat (dev, s_deskew_errors);
    if (LinkState & LS_PhaseError)	 BumpDevStat (dev, s_phase_errors);
    if (LinkState & LS_DataError)	 BumpDevStat (dev, s_data_errors);
    if (LinkState & LS_FifoOvFlow0)	 BumpDevStat (dev, s_fifo_overflow0);
    if (LinkState & LS_FifoOvFlow1)	 BumpDevStat (dev, s_fifo_overflow1);
    if (LinkState & LS_Mod45Changed)	 BumpDevStat (dev, s_mod45changed);
    if (LinkState & LS_PAckNotSeenError) BumpDevStat (dev, s_pack_not_seen);

    PULSE_SCHED_RESTART (dev, SCH_ClearLinkErrorInt);
    
    /* schedule a timer to clear the link error LED, so that it stays on 
     * for a second for every link error that occurs */
    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA && !timer_fn_queued (&dev->dev_linkerr_timeoutid))
	schedule_timer_fn (&dev->dev_linkerr_timeoutid, clear_linkerr_led, (void *) dev, HZ);

    /*
     * Signal the link error to the switch by
     * enabling the INT_LinkPortKeyFail bit.
     * Always clear the error bit as the switch
     * might have produced a spurious "ack" ...
     */
    PULSE_SYSCONTROL (dev, CONT_CLEAR_LINKPORT_INT);

    if (dev->dev_linkerr_signalled == 0)
	dev->dev_linkerr_signalled = 1;
    else
	dev->dev_linkerr_signalled = 2;
    
    ENABLE_INT_MASK (dev, INT_LinkPortKeyFail);
      
    check_error_rate (dev);
}

void
handle_linkportkeyfail (ELAN4_DEV *dev)
{
    PRINTF0 (DBG_DEVICE, DBG_INTR, "handle_linkportkeyfail\n");

    PULSE_SYSCONTROL (dev, CONT_CLEAR_LINKPORT_INT);

    if (! dev->dev_linkerr_signalled)
    {
	/* Hmmm - they're not playing ball */
	BumpDevStat (dev, s_linkport_keyfail);

	DISABLE_INT_MASK (dev, INT_LinkPortKeyFail);
    }
    else
    {
	/* If more link errors have occured since we 
	 * signalled the error, then leave it signalled. */
	if (--dev->dev_linkerr_signalled == 0)
	    DISABLE_INT_MASK (dev, INT_LinkPortKeyFail);
    }
}


static __inline__ void
__elan4_4msi0 (ELAN4_DEV *dev, E4_uint32 intreg, E4_uint32 intmask)
{
    unsigned long flags;

    if (intreg & intmask & INT_MainInterrupt)
    {
	DISABLE_INT_MASK (dev, INT_MainInterrupt);

	if (handle_mainints (dev, -1, elan4_mainint_punt_loops) == 0)
	    ENABLE_INT_MASK (dev, INT_MainInterrupt);
	else
	{
	    BumpDevStat (dev, s_mainint_punts);
	    
	    spin_lock_irqsave (&dev->dev_mainint_lock, flags);
	    kcondvar_wakeupone (&dev->dev_mainint_wait, &dev->dev_mainint_lock);
	    spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);
	}
    }
}

static __inline__ void
__elan4_4msi1 (ELAN4_DEV *dev, E4_uint32 intreg, E4_uint32 intmask)
{
    E4_uint32 restart = 0;

    PRINTF1 (DBG_DEVICE, DBG_INTR, "__elan4_4msi1: %x\n", intreg);
    
    spin_lock (&dev->dev_trap_lock);
    
    if (intreg & intmask & INT_CProc)
	restart |= handle_cproc_trap (dev);
    if (intreg & intmask & INT_EProc) 
	restart |= handle_eproc_trap (dev);
    if (intreg & intmask & INT_Dma0Proc) 
	restart |= handle_dproc_trap (dev, 0);
    if (intreg & intmask & INT_Dma1Proc) 
	restart |= handle_dproc_trap (dev, 1);
    if (intreg & intmask & INT_TProc)
	restart |= handle_tproc_trap (dev);
    
    PULSE_SCHED_RESTART (dev, restart);
    
    spin_unlock (&dev->dev_trap_lock);
    
    if (intreg & (INT_Halted|INT_Discarding))
	handle_haltints (dev, intreg);
}

static __inline__ void
__elan4_4msi2 (ELAN4_DEV *dev, E4_uint32 intreg, E4_uint32 intmask)
{
    E4_uint32 restart = 0;

    PRINTF1 (DBG_DEVICE, DBG_INTR, "__elan4_4msi2: %x\n", intreg);
    
    spin_lock (&dev->dev_trap_lock);
    if (intreg & intmask & INT_IProcCh0LowPri)
	restart |= handle_iproc_trap (dev, 0);
    
    if (intreg & intmask & INT_IProcCh1LowPri)
	restart |= handle_iproc_trap (dev, 1);
    
    if (intreg & intmask & INT_IProcCh0HighPri)
	restart |= handle_iproc_trap (dev, 2);
    
    if (intreg & intmask & INT_IProcCh1HighPri)
	restart |= handle_iproc_trap (dev, 3);
    
    PULSE_SCHED_RESTART (dev, restart);
    
    spin_unlock (&dev->dev_trap_lock);
}

static __inline__ void
__elan4_4msi3 (ELAN4_DEV *dev, E4_uint32 intreg, E4_uint32 intmask)
{
    PRINTF1 (DBG_DEVICE, DBG_INTR, "__elan4_4msi3: %x\n", intreg);
    
    if (intreg & intmask & INT_PciMemErr)
	handle_pcimemerr (dev);
    
    if (intreg & intmask & INT_SDRamInt)
	handle_sdramint (dev);
    
    if (intreg & intmask & INT_LinkError)
	handle_linkerror (dev);
    
    if (intreg & intmask & INT_LinkPortKeyFail)
	handle_linkportkeyfail (dev);
}

int
elan4_1msi0 (ELAN4_DEV *dev)
{
    E4_uint32 intmask = dev->dev_intmask;
    E4_uint32 intreg;

    if (intmask == 0 || ((intreg = read_reg32 (dev, InterruptReg)) & intmask) == 0)
	return (0);

    BumpDevStat (dev, s_interrupts);
    
    do {
	PRINTF1 (DBG_DEVICE, DBG_INTR, "elan4_1msi0: %x\n", intreg);

	if (intreg & intmask & INT_MSI0)
	    __elan4_4msi0(dev, intreg, intmask);
	if (intreg & intmask & INT_MSI1)
	    __elan4_4msi1(dev, intreg, intmask);
	if (intreg & intmask & INT_MSI2)
	    __elan4_4msi2(dev, intreg, intmask); 
	if (intreg & intmask & INT_MSI3)
	    __elan4_4msi3(dev, intreg, intmask);

	if (intreg & INT_LinkPortKeyFail)
	    handle_linkportkeyfail (dev);

	/* must ensure that the read of the interrupt mask
	 * completes before the read of the interrupt register
	 * since the main interrupt thread clears it's interrupt
	 * and then re-enables it in the interrupt mask. */
	intmask = dev->dev_intmask;
	mb();
	intreg = read_reg32 (dev, InterruptReg);

    } while ((intreg & intmask) != 0);

    return (1);
}

/* local context management */
int
elan4_insertctxt (ELAN4_DEV *dev, ELAN4_CTXT *ctxt, ELAN4_TRAP_OPS *ops)
{
    unsigned long flags;
    int tbl;
    int i;

    ctxt->ctxt_dev      = dev;
    ctxt->ctxt_ops      = ops;
    ctxt->ctxt_features = dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES];

    INIT_LIST_HEAD (&ctxt->ctxt_cqalist);
    spin_lock_init (&ctxt->ctxt_mmulock);

    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
    {
	ctxt->shuffle_needed[tbl] = 0;
	for(i=0; i<ELAN4_CTXT_MAX_SHUFFLE; i++)
	    ctxt->shuffle[tbl][i] = -1;

	KMEM_ZALLOC (ctxt->ctxt_mmuhash[tbl], ELAN4_HASH_ENTRY **,  dev->dev_hashsize[tbl] * sizeof (ELAN4_HASH_ENTRY *), 1);
	
	if (ctxt->ctxt_mmuhash[tbl] == NULL)
	{
	    if (tbl != 0)
		KMEM_FREE (ctxt->ctxt_mmuhash[0], dev->dev_hashsize[0] * sizeof (ELAN4_HASH_ENTRY *));
	    spin_lock_destroy (&ctxt->ctxt_mmulock);
	    return (-ENOMEM);
	}
    }

    spin_lock_irqsave (&dev->dev_ctxt_lock, flags);

    if ((ctxt->ctxt_num = bt_freebit (dev->dev_ctxmap, (1 << dev->dev_ctxtableshift))) >= 0)
    {
	/* chain onto the lists of all contexts */
	list_add (&ctxt->ctxt_link, &dev->dev_ctxt_list);

	BT_SET (dev->dev_ctxmap, ctxt->ctxt_num);
    }
    
    spin_unlock_irqrestore (&dev->dev_ctxt_lock, flags);

    if (ctxt->ctxt_num >= 0)
	proc_insertctxt(dev, ctxt);

    return (ctxt->ctxt_num < 0 ? -ENOMEM : 0);
}

void
elan4_removectxt (ELAN4_DEV *dev, ELAN4_CTXT *ctxt)
{
    unsigned long flags;
    int tbl;

    proc_removectxt(dev, ctxt);

    /* remove from list of contexts */
    spin_lock_irqsave (&dev->dev_ctxt_lock, flags);

    list_del (&ctxt->ctxt_link);

    BT_CLEAR (dev->dev_ctxmap, ctxt->ctxt_num);

    spin_unlock_irqrestore (&dev->dev_ctxt_lock, flags);

    spin_lock_destroy (&ctxt->ctxt_info_lock);

    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
	KMEM_FREE (ctxt->ctxt_mmuhash[tbl],  dev->dev_hashsize[tbl] * sizeof (ELAN4_HASH_ENTRY *));

    spin_lock_destroy (&ctxt->ctxt_mmulock);
}

ELAN4_CTXT *
elan4_localctxt (ELAN4_DEV *dev, unsigned num)
{
    struct list_head *entry;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_ctxt_lock, flags);

    list_for_each (entry, &dev->dev_ctxt_list) {
	ELAN4_CTXT *ctxt = list_entry (entry, ELAN4_CTXT, ctxt_link);
	
	if (ctxt->ctxt_num == num)
	{
	    spin_unlock_irqrestore (&dev->dev_ctxt_lock, flags);
	    return (ctxt);
	}
    }
    spin_unlock_irqrestore (&dev->dev_ctxt_lock, flags);

    return ((ELAN4_CTXT *) NULL);
}

ELAN4_CTXT *
elan4_networkctxt (ELAN4_DEV *dev, unsigned num)
{
    E4_uint32 filter = elan4_read_filter (dev, num);
    
    if ((filter & E4_FILTER_CONTEXT_MASK) == INVALID_CONTEXT)
	return NULL;
    else
	return elan4_localctxt (dev, filter & E4_FILTER_CONTEXT_MASK);
}

/* network context management */
int
elan4_attach_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum)
{
    ELAN4_DEV        *dev = ctxt->ctxt_dev;
    int               res = 0;
    E4_uint32         filter;
    unsigned long     flags;
    
    spin_lock_irqsave (&dev->dev_ctxt_lock, flags);
    
    filter = elan4_read_filter (dev, ctxnum);
    if ((filter & E4_FILTER_CONTEXT_MASK) != INVALID_CONTEXT)
    {
	PRINTF2 (ctxt, DBG_NETWORK_CTX, "elan4_attach_filter: ctx=%d filter=%x -> EBUSY\n", ctxnum, filter);
	res = -EBUSY;
    }
    else
    {
	PRINTF1 (ctxt, DBG_NETWORK_CTX, "elan4_attach_filter: ctx=%d - SUCCESS\n", ctxnum);

	elan4_write_filter (dev, ctxnum, ctxt->ctxt_num | E4_FILTER_DISCARD_ALL);
	PULSE_SCHED_RESTART (dev, SCH_ContextFilterFlush);
    }
    spin_unlock_irqrestore (&dev->dev_ctxt_lock, flags);
    
    return (res);
}

void
elan4_detach_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;

    PRINTF1 (ctxt, DBG_NETWORK_CTX, "elan4_detach_filter: detach from network context %d\n", ctxnum);
	    
    elan4_write_filter (dev, ctxnum, INVALID_CONTEXT | E4_FILTER_DISCARD_ALL);
    PULSE_SCHED_RESTART (dev, SCH_ContextFilterFlush);
}

void
elan4_set_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum, E4_uint32 state)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;

    PRINTF6 (ctxt, DBG_NETWORK_CTX, "elan4_set_filter: set filter state %x for network context %d <%s%s%s%s>\n", state, ctxnum,
	     (state & E4_FILTER_DISCARD_ALL) ? "discard,"  : "",
	     (state & E4_FILTER_ACKOK_ALL)   ? "ack-ok,"   : "",
	     (state & E4_FILTER_HIGH_PRI)    ? "high-pri," : "",
	     (state & E4_FILTER_STATS)       ? "stats,"    : "");
	    
    elan4_write_filter (dev, ctxnum, ctxt->ctxt_num | state);
    PULSE_SCHED_RESTART (dev, SCH_ContextFilterFlush);
}

void
elan4_set_routetable (ELAN4_CTXT *ctxt, ELAN4_ROUTE_TABLE *tbl)
{
    ELAN4_DEV *dev   = ctxt->ctxt_dev;
    E4_uint32  value = tbl ? (E4_VPT_VALID | E4_VPT_VALUE(tbl->tbl_entries, tbl->tbl_size)) : 0;
    
    /* and insert into the vp table */
    elan4_sdram_writel (dev, (dev->dev_ctxtable + (ctxt->ctxt_num * sizeof (E4_ContextControlBlock)) +
			offsetof (E4_ContextControlBlock, VirtualProcessTable)), value);
    pioflush_sdram(dev);

    PULSE_SYSCONTROL (dev, CONT_ROUTE_FLUSH);
}

/* command queue management */
ELAN4_CQA *
elan4_getcqa (ELAN4_CTXT *ctxt, unsigned int idx)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    struct list_head *el;

    spin_lock (&dev->dev_cqlock);
    list_for_each (el, &ctxt->ctxt_cqalist) {
	ELAN4_CQA *cqa = list_entry (el, ELAN4_CQA, cqa_link);
	
	if (cqa->cqa_idx == idx)
	{
	    cqa->cqa_ref++;

	    spin_unlock (&dev->dev_cqlock);
	    return cqa;
	}
    }
    spin_unlock (&dev->dev_cqlock);
    return NULL;
}

void
elan4_putcqa (ELAN4_CTXT *ctxt, unsigned int idx)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    struct list_head *el, *nel;

    spin_lock (&dev->dev_cqlock);
    list_for_each_safe (el, nel, &ctxt->ctxt_cqalist) {
	ELAN4_CQA *cqa = list_entry (el, ELAN4_CQA, cqa_link);
	
	if (cqa->cqa_idx == idx)
	{
	    if (--cqa->cqa_ref || bt_lowbit (cqa->cqa_bitmap, ELAN4_CQ_PER_CQA) != -1)
		spin_unlock (&dev->dev_cqlock);
	    else
	    {
		list_del (&cqa->cqa_link);
		
		BT_CLEAR (ctxt->ctxt_cqamap, cqa->cqa_idx);
		BT_CLEAR (dev->dev_cqamap, cqa->cqa_cqnum/ELAN4_CQ_PER_CQA);
		spin_unlock (&dev->dev_cqlock);
		
		KMEM_FREE (cqa, sizeof (ELAN4_CQA));
	    }
	    return;
	}
    }
    spin_unlock (&dev->dev_cqlock);

    printk ("elan4_putcqa: idx %d not found\n", idx);
    BUG();
}

static ELAN4_CQ *
elan4_getcq (ELAN4_CTXT *ctxt, unsigned int type)
{
    ELAN4_DEV        *dev = ctxt->ctxt_dev;
    ELAN4_CQA        *cqa;
    struct list_head *el;
    int		      cidx, didx;

    spin_lock (&dev->dev_cqlock);
    list_for_each (el, &ctxt->ctxt_cqalist) {
	cqa = list_entry (el, ELAN4_CQA, cqa_link);

	if (cqa->cqa_type == type && (cidx = bt_freebit (cqa->cqa_bitmap, ELAN4_CQ_PER_CQA)) >=0)
	{
	    BT_SET (cqa->cqa_bitmap, cidx);
	    
	    spin_unlock (&dev->dev_cqlock);
	    return &cqa->cqa_cq[cidx];
	}
    }
    spin_unlock (&dev->dev_cqlock);

    /* allocate a new cqa and it's chunk of command queue descriptors */
    KMEM_ZALLOC (cqa, ELAN4_CQA *, sizeof (ELAN4_CQA), 1);
    if (cqa == NULL)
	return NULL;

    spin_lock (&dev->dev_cqlock);
    cidx = bt_freebit (ctxt->ctxt_cqamap, ELAN4_MAX_CQA);

    /* On architectures which have MTRR registers for write-combinig
     * the top command queues from dev->dev_cqreorder upwards are
     * used for reordered queues.  Without MTRR registers any page
     * sized group can use write combinig through the ptes. */
    if (dev->dev_cqreorder == 0)
	didx = bt_freebit (dev->dev_cqamap, dev->dev_cqcount/ELAN4_CQ_PER_CQA);
    else
    {
	if ((type & CQ_Reorder) != 0)
	    didx = bt_nextbit (dev->dev_cqamap, dev->dev_cqcount/ELAN4_CQ_PER_CQA, (dev->dev_cqreorder/ELAN4_CQ_PER_CQA) - 1, 0);
	else
	    didx = bt_freebit (dev->dev_cqamap, dev->dev_cqreorder/ELAN4_CQ_PER_CQA);
    }

    if (cidx < 0 || didx < 0)
    {
	spin_unlock (&dev->dev_cqlock);
	KMEM_FREE (cqa, sizeof (ELAN4_CQA));
	return NULL;
    }

    BT_SET (ctxt->ctxt_cqamap, cidx);
    BT_SET (dev->dev_cqamap, didx);

    cqa->cqa_idx   = cidx;
    cqa->cqa_type  = type;
    cqa->cqa_cqnum = (didx * ELAN4_CQ_PER_CQA);
    
    list_add_tail (&cqa->cqa_link, &ctxt->ctxt_cqalist);
    
    /* initialise the cqa struct */
    for (cidx = 0; cidx < ELAN4_CQ_PER_CQA; cidx++)
    {
	cqa->cqa_cq[cidx].cq_idx = cidx;
	cqa->cqa_cq[cidx].cq_cqa = cqa;
    }

    /* no mappings yet */
    cqa->cqa_ref = 0;

    /* we're going to return entry zero */
    BT_SET (cqa->cqa_bitmap, 0);
    spin_unlock (&dev->dev_cqlock);
    
    return &cqa->cqa_cq[0];
}

static void
elan4_putcq (ELAN4_CTXT *ctxt, ELAN4_CQ *cq)
{
    ELAN4_DEV        *dev = ctxt->ctxt_dev;
    ELAN4_CQA        *cqa = cq->cq_cqa;

    spin_lock (&dev->dev_cqlock);

    BT_CLEAR (cqa->cqa_bitmap, cq->cq_idx);

    if (bt_lowbit (cqa->cqa_bitmap, ELAN4_CQ_PER_CQA) != -1 || cqa->cqa_ref)
	spin_unlock (&dev->dev_cqlock);
    else
    {
	list_del (&cqa->cqa_link);
	
	BT_CLEAR (ctxt->ctxt_cqamap, cqa->cqa_idx);
	BT_CLEAR (dev->dev_cqamap, cqa->cqa_cqnum/ELAN4_CQ_PER_CQA);
	spin_unlock (&dev->dev_cqlock);
	
	KMEM_FREE (cqa, sizeof (ELAN4_CQA));
    }
}

ELAN4_CQ *
elan4_alloccq (ELAN4_CTXT *ctxt, unsigned cqsize, unsigned perm, unsigned cqtype)
{
    ELAN4_DEV   *dev = ctxt->ctxt_dev;
    ELAN4_CQ    *cq;
    int         cqnum;
    sdramaddr_t cqdesc;
    unsigned    offset;
    E4_uint64   value;

    if ((cq = elan4_getcq (ctxt, cqtype)) == NULL)
	return NULL;

    cqnum = elan4_cq2num(cq);
    
    cq->cq_space = elan4_sdram_alloc (dev, CQ_Size(cqsize));
    if (cq->cq_space == (virtaddr_t) 0)
    {
	elan4_putcq (ctxt, cq);
	return (NULL);
    }

    cq->cq_size   = cqsize;
    cq->cq_perm   = perm;
    
    /* and finally initialise the command queue descriptor */
    cqdesc = dev->dev_cqaddr + (cqnum * sizeof (E4_CommandQueueDesc));

    value  = CQ_QueuePtrsValue (cqsize, cq->cq_space, cq->cq_space);
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	value |= ((cqtype & CQ_Priority) ? CQ_RevA_Priority : 0);
    else
	value |= (((cqtype & CQ_Priority) ? CQ_RevB_Priority : 0) | 
		  ((cqtype & CQ_Reorder)  ? CQ_RevB_ReorderingQueue : CQ_RevB_32bitWriteQueue));

    elan4_sdram_writeq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs), value);
    elan4_sdram_writeq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_HoldingValue), 0);
    elan4_sdram_writeq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_AckBuffers), 0);
    elan4_sdram_writeq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control), CQ_ControlValue (ctxt->ctxt_num, 2, perm));
    pioflush_sdram (dev);

    offset = (cqnum + dev->dev_cqoffset) * CQ_CommandMappingSize;

    cq->cq_mapping = elan4_map_device (dev, ELAN4_BAR_REGISTERS, (offset & ~(PAGE_SIZE-1)), 
				       PAGE_SIZE, &cq->cq_handle) + (offset & (PAGE_SIZE-1));
#ifdef CONFIG_MPSAS
    if (ctxt == &dev->dev_ctxt)
	return (cq);
#endif

    elan4_sdram_flushcache (dev, cq->cq_space, CQ_Size(cqsize));

    return (cq);
}
    
void
elan4_freecq (ELAN4_CTXT *ctxt, ELAN4_CQ *cq)
{
    ELAN4_DEV *dev    = ctxt->ctxt_dev;
    unsigned   offset = (elan4_cq2num(cq) + dev->dev_cqoffset) * CQ_CommandMappingSize;

    elan4_flushcq (dev, cq);

    elan4_unmap_device (dev, cq->cq_mapping - (offset & (PAGE_SIZE-1)), PAGE_SIZE, &cq->cq_handle);
    elan4_sdram_free (dev, cq->cq_space, CQ_Size (cq->cq_size));

    elan4_putcq (ctxt, cq);
}

void
elan4_restartcq (ELAN4_DEV *dev, ELAN4_CQ *cq)
{
    sdramaddr_t   cqdesc = dev->dev_cqaddr + (elan4_cq2num(cq) * sizeof (E4_CommandQueueDesc));
    int           hipri;
    unsigned long flags;
    
    PRINTF1 (DBG_DEVICE, DBG_CPROC, "restartcq: restarting cq %p\n", cq);
    
    spin_lock_irqsave (&dev->dev_requeue_lock, flags);

    while (read_reg32 (dev, CommandControl.CommandRequeuePtr) & E4_CommandRequeueBusy)
	;
    
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	hipri = (elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs)) & CQ_RevA_Priority) != 0;
    else
	hipri = (elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs)) & CQ_RevB_Priority) != 0;
    
    if (hipri)
    {
	PRINTF1 (DBG_DEVICE, DBG_CPROC, "restartcq: restart cq %d as high pri\n", elan4_cq2num(cq));
	write_reg32 (dev, CommandControl.CommandRequeuePtr, cqdesc | E4_CommandRequeueHighPri);
    }
    else
    {
	PRINTF1 (DBG_DEVICE, DBG_CPROC, "restartcq: restart cq %d as low pri\n", elan4_cq2num(cq));
	write_reg32 (dev, CommandControl.CommandRequeuePtr, cqdesc);
    }
    pioflush_reg (dev);
    
    spin_unlock_irqrestore (&dev->dev_requeue_lock, flags);
}

static void
flushcq_intop (ELAN4_DEV *dev, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_flush_lock, flags);
    dev->dev_flush_finished |= (1 << (unsigned long) arg);
    kcondvar_wakeupall (&dev->dev_flush_wait, &dev->dev_flush_lock);
    spin_unlock_irqrestore (&dev->dev_flush_lock, flags);
}
void
elan4_flushcq (ELAN4_DEV *dev, ELAN4_CQ *cq)
{
    int	          flushqnum = elan4_cq2num(cq) & (COMMAND_INSERTER_CACHE_ENTRIES-1);
    ELAN4_CQ     *flushq    = dev->dev_flush_cq[flushqnum];
    unsigned long flags;

    PRINTF (DBG_DEVICE, DBG_FLUSH, "elan4_flushcq: cqnum=%d\n", elan4_cq2num(cq));

    spin_lock_irqsave (&dev->dev_flush_lock, flags);

    while (! (dev->dev_flush_finished & (1 << flushqnum)))
	kcondvar_wait (&dev->dev_flush_wait, &dev->dev_flush_lock, &flags);
    
    dev->dev_flush_finished &= ~(1 << flushqnum);

    dev->dev_flush_op[flushqnum].op_function = flushcq_intop;
    dev->dev_flush_op[flushqnum].op_arg      = (void *) (unsigned long) flushqnum;
    
    elan4_queue_intop (dev, flushq, &dev->dev_flush_op[flushqnum]);

    while (! (dev->dev_flush_finished & (1 << flushqnum)))
	kcondvar_wait (&dev->dev_flush_wait, &dev->dev_flush_lock, &flags);
    
    spin_unlock_irqrestore (&dev->dev_flush_lock, flags);
}

void
elan4_updatecq (ELAN4_DEV *dev, ELAN4_CQ *cq, unsigned perm, unsigned restart)
{
    sdramaddr_t cqdesc  = dev->dev_cqaddr + (elan4_cq2num(cq) * sizeof (E4_CommandQueueDesc));
    E4_uint32   control = elan4_sdram_readl (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control));

    /* Write the command queues control word, but ensure that the ChannelNotCompleted fields
     * are not modified.   We use this to just alter the RestartCount/Permissions fields */

    elan4_sdram_writel (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control), 
			CQ_ControlValue (CQ_Context (control), restart ? restart : CQ_RestartCount (control), perm));
}

/* instruction cache flush */
static __inline__ void
elan4_flush_icache_locked (ELAN4_DEV *dev)
{
    int i, j;

    PRINTF0 (DBG_DEVICE, DBG_FLUSH, "elan4_flush_icache_locked: flushing icache\n");

    for (i = 0; i < (E4_ICacheLines/E4_ICachePortSize); i++)
    {
        write_reg64 (dev, ICachePort_Cntl_Addr, i << E4_ICacheTagAddrShift);
        for (j = 0; j < E4_ICachePortSize; j++)
	    write_reg64 (dev, ICachePort[j], E4_InvalidTagValue);
    }

    /*
     * Initialise the top of the ICache Set0 with a instruction which will
     * cause a know trap fingerprint so that the application can identify it
     * and ignore the trap.
     */
    write_reg64 (dev, ICachePort_Cntl_Addr, E4_ICacheFixupOffset | E4_AccessICacheRams);

    /* Errata 24: must ensure that the DCache is flushed after loading 
     *            code for the thread processor. */
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	elan4_sdram_flushcache (dev, 0, E4_CacheSize);

    pioflush_reg (dev);
}

static void
device_iflush_haltop (ELAN4_DEV *dev, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_flush_lock, flags);
    
    elan4_flush_icache_locked (dev);

    dev->dev_iflush_queued = 0;

    kcondvar_wakeupall (&dev->dev_flush_wait, &dev->dev_flush_lock);
    spin_unlock_irqrestore (&dev->dev_flush_lock, flags);
}

void
elan4_flush_icache_halted (ELAN4_CTXT *ctxt)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_flush_lock, flags);
    
    elan4_flush_icache_locked (dev);

    spin_unlock_irqrestore (&dev->dev_flush_lock, flags);
}

void
elan4_flush_icache (ELAN4_CTXT *ctxt)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    unsigned long flags;
    
    spin_lock_irqsave (&dev->dev_flush_lock, flags);

    PRINTF1 (DBG_DEVICE, DBG_FLUSH, "elan4_flush_icache: queued=%d\n", dev->dev_iflush_queued);

    if (! dev->dev_iflush_queued)
    {
	dev->dev_iflush_queued = 1;
	
	elan4_queue_haltop (dev, &dev->dev_iflush_haltop);
    }

    while (dev->dev_iflush_queued)
	kcondvar_wait (&dev->dev_flush_wait, &dev->dev_flush_lock, &flags);

    spin_unlock_irqrestore (&dev->dev_flush_lock, flags);
}

/* device context operations */
static void
device_cproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned cqnum)
{
    ELAN4_DEV        *dev  = ctxt->ctxt_dev;
    ELAN4_CPROC_TRAP *trap = &dev->dev_cproc_trap;

    elan4_extract_cproc_trap (dev, status, trap, cqnum);

    DBGCMD (DBG_DEVICE, DBG_FLUSH, elan4_display_cproc_trap (DBG_DEVICE, DBG_FLUSH, "device_cproc_trap", trap));

    switch (CPROC_TrapType (trap->tr_status))
    {
    case CommandProcInterruptQueueOverflow:
	PRINTF (ctxt, DBG_FLUSH, "device_cproc_trap: cqnum=%d\n", cqnum);

	/* XXXX: we could either just hit restart (and hope) - or we could extract
	 *       the event interrupt cookie out and "complete" the command before
	 *       restarting it */
	elan4_restartcq (dev, dev->dev_flush_cq[cqnum]);
	return;

    case CommandProcDmaQueueOverflow:
    case CommandProcPermissionTrap:
	handle_dma_flushops (dev, status, cqnum);
	return;
	
    default:
	printk ("device_cproc_trap: status=%llx control=%llx TrapType=%x cqnum=%d\n", (long long) trap->tr_status,
		elan4_sdram_readq (dev, dev->dev_cqaddr + cqnum * sizeof (E4_CommandQueueDesc) +
				   offsetof (E4_CommandQueueDesc, CQ_Control)),
		(int) CPROC_TrapType(trap->tr_status), cqnum);
	ELAN4_DEBUG_TRIGGER (ctxt, "elan4:device_cproc_trap\n");
    }
}

static void
device_tproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status)
{
    ELAN4_TPROC_TRAP trap;

    elan4_extract_tproc_trap (ctxt->ctxt_dev, status, &trap);

    elan4_display_tproc_trap (DBG_CONSOLE, DBG_TRAP, "device_tproc_trap", &trap);
    ELAN4_DEBUG_TRIGGER (ctxt, "elan4:device_tproc_trap\n");
}

static void
device_dproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    ELAN4_DPROC_TRAP trap;

    elan4_extract_dproc_trap (ctxt->ctxt_dev, status, &trap, unit);

    elan4_display_dproc_trap (DBG_CONSOLE, DBG_TRAP, "device_dproc_trap", &trap);

    ELAN4_DEBUG_TRIGGER (ctxt, "elan4:device_dproc_trap\n");
}

static void
device_interrupt (ELAN4_CTXT *ctxt, E4_uint64 cookie)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) ctxt;
    struct list_head *el,*nel;
    unsigned long flags;

    PRINTF (ctxt, DBG_FLUSH, "device_interrupt: cookie=%llx\n", cookie);

    spin_lock_irqsave (&dev->dev_intop_lock, flags);
    list_for_each_safe (el, nel, &dev->dev_intop_list) {
	ELAN4_INTOP *op = list_entry (el, ELAN4_INTOP, op_link);

	if (op->op_cookie == cookie)
	{
	    if ((op->op_cookie & INTOP_TYPE_MASK) == INTOP_ONESHOT)
		list_del (&op->op_link);

	    spin_unlock_irqrestore (&dev->dev_intop_lock, flags);
	    
	    (*op->op_function)(dev, op->op_arg);
	    return;
	}
    }
    spin_unlock_irqrestore (&dev->dev_intop_lock, flags);

    panic ("device_interrupt: interrupt cookie %llx not found\n", (long long)cookie);
}

static void
device_iproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    ELAN4_IPROC_TRAP *trap = &dev->dev_iproc_trap;

    elan4_extract_iproc_trap (dev, status, trap, unit);
    elan4_inspect_iproc_trap (trap);

    DBGCMD (ctxt, DBG_IPROC, elan4_display_iproc_trap (ctxt, DBG_IPROC, "device_iproc_trap", trap));

    if (elan4_neterr_iproc_trap (dev, trap))
	return;

    elan4_display_iproc_trap (DBG_CONSOLE, DBG_TRAP, "device_iproc_trap", trap);
    panic ("device_iproc_trap: unexpected trap\n");
}

static void
device_needs_shuffle (ELAN4_CTXT *ctxt, int tbl, int hashidx)
{   
    /* XXXX currently this doesnt need to do anything 
       as the chains have only 2 entries */
}

ELAN4_TRAP_OPS device_trap_ops = 
{
    NULL,
    device_cproc_trap,
    device_dproc_trap,
    device_tproc_trap,
    device_iproc_trap,
    device_interrupt,
    NULL,
    device_needs_shuffle,
};

/*
 * elan4_initialise_device
 *    initialise the ELAN4_DEV struct - spinlocks,cvs etc.
 *    map the registers, sdram etc
 */
int
elan4_initialise_device (ELAN4_DEV *dev)
{
    int i, bit;

    if (elan4_mainint_resched_ticks == 0)
	elan4_mainint_resched_ticks = (hz/4);

    /* map the registers */
    switch (dev->dev_devinfo.dev_revision_id)
    {
    case PCI_REVISION_ID_ELAN4_REVA:
	dev->dev_regs = elan4_map_device (dev, ELAN4_BAR_REGISTERS, ELAN4_REVA_REG_OFFSET, ELAN4_REG_SIZE, &dev->dev_regs_handle);
	
	dev->dev_rom  = elan4_map_device (dev, ELAN4_BAR_REGISTERS, ELAN4_REVA_EBUS_OFFSET + ELAN4_REVA_EBUS_ROM_OFFSET, 
					  ELAN4_REVA_EBUS_ROM_SIZE, &dev->dev_rom_handle);
	break;
	
    case PCI_REVISION_ID_ELAN4_REVB:
	dev->dev_regs = elan4_map_device (dev, ELAN4_BAR_REGISTERS, ELAN4_REVB_REG_OFFSET, ELAN4_REG_SIZE, &dev->dev_regs_handle);
	dev->dev_rom  = (ioaddr_t) 0;
	dev->dev_i2c  = elan4_map_device (dev, ELAN4_BAR_REGISTERS, ELAN4_REVB_I2C_OFFSET, ELAN4_REVB_I2C_SIZE, &dev->dev_i2c_handle);
	break;

    default:
	return -EINVAL;
    }

    /* XXXX: parse the ebus rom to determine the sdram configuration */
    {
	extern long long       sdram_cfg;

	if (sdram_cfg == 0)
	    dev->dev_sdram_cfg = SDRAM_STARTUP_VALUE;
	else
	    dev->dev_sdram_cfg = sdram_cfg;
    }

    for (bit = 0; ((1 << bit) & elan4_resource_len (dev, ELAN4_BAR_SDRAM)) == 0; bit++)
	;

    switch ((dev->dev_sdram_cfg >> SDRAM_RamSize_SH) & 3)
    {
    case 0:			/* 64Mbit, 128Mbit, 256Mbit, 512Mbit or 1Gbit (16-bit output) */
	dev->dev_sdram_numbanks = 4; bit -= 2;
	for (i = 0; i < dev->dev_sdram_numbanks; i++)
	{
	    dev->dev_sdram_banks[i].b_base = (i << bit);
	    dev->dev_sdram_banks[i].b_size = (1 << bit);
	}
	break;

    case 1:			/*  64Mbit, 128Mbit, 256Mbit or 512Mbit (8-bit output) */
	dev->dev_sdram_numbanks = 4; bit -= 2;
	for (i = 0; i < dev->dev_sdram_numbanks; i++)
	{
	    dev->dev_sdram_banks[i].b_base = ((i & 2) << (bit)) | ((i & 1) << (bit-1));
	    dev->dev_sdram_banks[i].b_size = (1 << bit);
	}
	break;
	
    case 2:			/* 2Gbit (16-bit output) or 1Gbit (8-bit output) */
	dev->dev_sdram_numbanks = 2; bit--;
	for (i = 0; i < dev->dev_sdram_numbanks; i++)
	{
	    dev->dev_sdram_banks[i].b_base = (i << bit);
	    dev->dev_sdram_banks[i].b_size = (1 << bit);
	}
	break;

    case 3:			/* 4Gbit (16-bit output) or 2Gbit (8-bit output) */
	dev->dev_sdram_numbanks = 1;
	dev->dev_sdram_banks[0].b_base = 0;
	dev->dev_sdram_banks[0].b_size = (1 << bit);
	break;
    }

    elan4_sdram_init (dev);

    /* initialise locks for classes of interrupts */
    spin_lock_init (&dev->dev_trap_lock);
    spin_lock_init (&dev->dev_intop_lock);
    spin_lock_init (&dev->dev_haltop_lock);
    spin_lock_init (&dev->dev_mainint_lock);

    init_timer (&dev->dev_haltop_timer);
    dev->dev_haltop_timer.function = dev_haltop_timer_func;
    dev->dev_haltop_timer.data     = (unsigned long) dev;

    /* initialise other locks */
    spin_lock_init (&dev->dev_i2c_lock);

    spin_lock_init (&dev->dev_mmulock);
    spin_lock_init (&dev->dev_cqlock);
    spin_lock_init (&dev->dev_ctxlock);

    spin_lock_init (&dev->dev_intmask_lock);
    spin_lock_init (&dev->dev_syscontrol_lock);

    spin_lock_init (&dev->dev_ctxt_lock);
    spin_lock_init (&dev->dev_flush_lock);
    spin_lock_init (&dev->dev_requeue_lock);

    kmutex_init (&dev->dev_lock);

    kcondvar_init (&dev->dev_mainint_wait);
    kcondvar_init (&dev->dev_flush_wait);

    /* initialsie lists */
    INIT_LIST_HEAD (&dev->dev_ctxt_list);
    INIT_LIST_HEAD (&dev->dev_intop_list);
    INIT_LIST_HEAD (&dev->dev_interruptq_list);
    INIT_LIST_HEAD (&dev->dev_hc_list);
    INIT_LIST_HEAD (&dev->dev_haltop_list);
    INIT_LIST_HEAD (&dev->dev_dma_flushop[0].list);
    INIT_LIST_HEAD (&dev->dev_dma_flushop[1].list);

    dev->dev_state = ELAN4_STATE_STOPPED;

    return (0);
}

void
elan4_finalise_device (ELAN4_DEV *dev)
{
    kcondvar_destroy (&dev->dev_flush_wait);
    kcondvar_destroy (&dev->dev_mainint_wait);

    kmutex_destroy (&dev->dev_lock);

    spin_lock_destroy (&dev->dev_requeue_lock);
    spin_lock_destroy (&dev->dev_flush_lock);
    spin_lock_destroy (&dev->dev_ctxt_lock);

    spin_lock_destroy (&dev->dev_syscontrol_lock);
    spin_lock_destroy (&dev->dev_intmask_lock);

    spin_lock_destroy (&dev->dev_ctxlock);
    spin_lock_destroy (&dev->dev_cqlock);
    spin_lock_destroy (&dev->dev_mmulock);

    spin_lock_destroy (&dev->dev_i2c_lock);

    spin_lock_destroy (&dev->dev_mainint_lock);
    spin_lock_destroy (&dev->dev_haltop_lock);
    spin_lock_destroy (&dev->dev_intop_lock);
    spin_lock_destroy (&dev->dev_trap_lock);

    del_timer_sync (&dev->dev_haltop_timer);

    while (! list_empty (&dev->dev_hc_list))
    {
	ELAN4_HASH_CHUNK *hc = list_entry (dev->dev_hc_list.next, ELAN4_HASH_CHUNK, hc_link);
	
	list_del (&hc->hc_link);

	KMEM_FREE(hc, sizeof (ELAN4_HASH_CHUNK));
    }
    
    elan4_sdram_fini (dev);
    
    switch (dev->dev_devinfo.dev_revision_id)
    {
    case PCI_REVISION_ID_ELAN4_REVA:
	elan4_unmap_device (dev, dev->dev_rom,  ELAN4_REVA_EBUS_ROM_SIZE, &dev->dev_rom_handle);
	elan4_unmap_device (dev, dev->dev_regs, ELAN4_REG_SIZE, &dev->dev_regs_handle);
	break;
    case PCI_REVISION_ID_ELAN4_REVB:
	elan4_unmap_device (dev, dev->dev_i2c,  ELAN4_REVB_I2C_SIZE, &dev->dev_i2c_handle);
	elan4_unmap_device (dev, dev->dev_regs, ELAN4_REG_SIZE, &dev->dev_regs_handle);
	break;
    }
}

static int
measure_sysclk (ELAN4_DEV *dev)	
{
    E4_uint64 val0, val1;
    E4_uint32 ticks, ns;
    
    write_ureg64 (dev, StatCont, STP_SYS_CLOCK_RATE0);
    
    val0 = read_ureg64 (dev, StatCounts[0]);
    udelay (1000);
    val1 = read_ureg64 (dev, StatCounts[0]);
    
    
    ticks = ((val1 >> 32) - (val0 >> 32));
    ns    = ((val1 & 0xffffffff) - (val0 & 0xffffffff));
    
    return (ticks / (ns / 1000));
}

static void
initialise_cache (ELAN4_DEV *dev)
{
    register int set, line;

    /* Initialise the cache to "map" the bottom of sdram - we will use
     * this space for cache flushing, so require the cache to be set
     * up so that cachelines for this are in the correct set.
     *
     * XXXX: for MPSAS we set bit 28, to ensure that any access to 
     *       sdram causes the line to be filled first to expunge any
     *       Xs. */
    for (set = 0; set < E4_NumCacheSets; set++)
	for (line = 0; line < E4_NumCacheLines; line++)
	    write_tag (dev, Tags[set][line], (((E4_uint64) set) << 29) | (1 << 28) | (line << 16));
}

#ifndef CONFIG_MPSAS
static void
initialise_cache_tags (ELAN4_DEV *dev, unsigned addr)
{
    register int set, line;

    /* Initialise the whole cache to hold sdram at "addr" as direct mapped */

    for (set = 0; set < E4_NumCacheSets; set++)
	for (line = 0; line < E4_NumCacheLines; line++)
	    write_tag (dev, Tags[set][line], addr | (set << 13) | (1 << 11));
}

static void
initialise_ecc (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    register int i, addr;

    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
    {
        initialise_cache_tags (dev, E4_CacheSize);
        for (addr = 0; addr < bank->b_size; addr += E4_CacheSize)
        {
	    for (i = 0; i < E4_CacheSize; i += sizeof (E4_uint64))
		writeq (0xbeec000000000000ull | addr | i, (void *)(bank->b_ioaddr + addr + i));
	    initialise_cache_tags (dev, addr);
        }
    }
    else
    {
	/* Write the whole of this bank of sdram. */
        for (addr = 0; addr < bank->b_size; addr += sizeof (E4_uint64))
	    writeq (0xbeec000000000000ull | addr, (void *)(bank->b_ioaddr + addr));

	/* Now flush out the top out of the cache */
        for (addr = 0; addr < E4_CacheSize; addr += sizeof (E4_uint64))
	    writeq (0xbeec000000000000ull | addr, (void *)(bank->b_ioaddr + addr));

	/* Now read the top value of sdram to guarantee the write has occured before the ecc is enabled */
	__elan4_readq (dev, bank->b_ioaddr + bank->b_size - sizeof (E4_uint64));
    }
}
#endif

#ifdef CONFIG_MPSAS
static void
do_initdma (ELAN4_DEV *dev)
{
#define VIRTUAL_ADDRESS	0x10000000ull
    ELAN4_CQ  *cq  = dev->dev_flush_cq[0];
    E4_uint64 value;
    E4_uint32 intreg;
    E4_uint64 status;

    PRINTF (DBG_DEVICE, DBG_CONFIG, "elan: performing initialising dma\n");
    
    DISABLE_INT_MASK (dev, INT_Dma0Proc | INT_Dma1Proc);
    
    /* initialise the context filter */
    elan4_attach_filter (&dev->dev_ctxt, 0);

    /* now issue a DMA - we expect this to trap */
    writeq (E4_DMA_TYPE_SIZE (128*4, DMA_DataTypeByte, 0, 0) | RUN_DMA_CMD, cq->cq_mapping + (0 << 3));
    writeq (0,                                                              cq->cq_mapping + (1 << 3));
    writeq (0,                                                              cq->cq_mapping + (2 << 3));
    writeq (dev->dev_tproc_space,                                           cq->cq_mapping + (3 << 3));
    writeq (dev->dev_tproc_space,                                           cq->cq_mapping + (4 << 3));
    writeq (0,                                                              cq->cq_mapping + (5 << 3));
    writeq (0,                                                              cq->cq_mapping + (6 << 3));
    
    /* spin waiting for it to trap - then restart the dma processor */
    do {
	value   = read_reg64 (dev, IntAndMaskReg);
	intreg  = (value >> E4_INTERRUPT_REG_SHIFT);
    } while ((intreg & (INT_Dma0Proc | INT_Dma1Proc)) == 0);
    
    /* check it trapped for the right reason */
    status = (intreg & INT_Dma0Proc) ? read_reg64 (dev, DProc0Status) : read_reg64 (dev, DProc1Status);
    
    if (DPROC_PrefetcherFault (status) || (DPROC_TrapType(status) != DmaProcFailCountError && DPROC_TrapType(status) != DmaProcPacketAckError))
    {
	printk ("elan: bad dma trap, status = %lx\n", (long)status);
	panic ("elan: bad dma trap\n");
    }
    
    PULSE_SCHED_RESTART (dev, SCH_RestartDma0Proc | SCH_RestartDma1Proc | SCH_RestartDmaPrefetchProc);

    elan4_detach _filter (&dev->dev_ctxt, 0);

    ENABLE_INT_MASK (dev, INT_Dma0Proc | INT_Dma1Proc);
#undef VIRTUAL_ADDRESS
}
#endif

static int
ebus_read_vpd (ELAN4_DEV *dev, unsigned char *data, unsigned int nob)
{
    unsigned int pci_data_ptr;
    unsigned int vpd_ptr;
    register int i;

    if (read_ebus_rom (dev, 0) != 0x55 || read_ebus_rom (dev, 1) != 0xaa)
    {
	printk ("elan%d: invalid rom signature in ebus rom\n", dev->dev_instance);
	return -EINVAL;
    }

    pci_data_ptr = (read_ebus_rom (dev, 0x19) << 8) | read_ebus_rom (dev, 0x18);

    /* check the pci data structure */
    if (read_ebus_rom (dev, pci_data_ptr + 0) != 'P' ||
	read_ebus_rom (dev, pci_data_ptr + 1) != 'C' ||
	read_ebus_rom (dev, pci_data_ptr + 2) != 'I' ||
	read_ebus_rom (dev, pci_data_ptr + 3) != 'R')
    {
	printk ("elan%d: invalid pci data structure in ebus rom\n", dev->dev_instance);
	return -EINVAL;
    }
    
    /* extract the VPD pointer */
    vpd_ptr = (read_ebus_rom (dev, pci_data_ptr + 9) << 8) | read_ebus_rom (dev, pci_data_ptr + 8);

    if (vpd_ptr == 0)
    {
	printk ("elan%d: no vital product data in ebus rom\n", dev->dev_instance);
	return -EINVAL;
    }
    
    /* read the vpd data */
    for (i = 0; i < nob; i++)
	data[i] = read_ebus_rom (dev, vpd_ptr + i);

    return 0;
}

int
elan4_read_vpd (ELAN4_DEV *dev, unsigned char *tag, unsigned char *result) 
{
    unsigned char vpd[I2C_ELAN_EEPROM_VPD_SIZE];
    unsigned char *ptr = vpd;
    unsigned int   finished = 0;
    unsigned char *lim;
    unsigned char  name[3];
    unsigned char  value[256];
    unsigned char  type;
    unsigned int   len, len2;
    register int   i;

    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
    {
	if (ebus_read_vpd (dev, vpd, I2C_ELAN_EEPROM_VPD_SIZE) < 0)
	{
	    PRINTF1 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, unable to read serial number from EBUS rom\n", dev->dev_instance);
	    return -EINVAL ;
	}	
    }
    else
    {
	if (i2c_read_rom (dev, I2C_ELAN_EEPROM_VPD_BASEADDR, I2C_ELAN_EEPROM_VPD_SIZE, vpd) < 0)
	{
	    PRINTF1 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, unable to read serial number from I2C rom\n", dev->dev_instance);
	    return  -EINVAL;
	}
    }

    result[0] = 0;
    while (! finished)
    {
	type = *ptr++;
	
	if (type & LARGE_RESOURCE_BIT)
	{
	    len = *(ptr++);
	    len += *(ptr++) << 8;
	    
	    switch (type & ~LARGE_RESOURCE_BIT)
	    {
	    case LARGE_RESOURCE_STRING:
	    case LARGE_RESOURCE_VENDOR_DEFINED:
		ptr += len;
		break;
		
	    case LARGE_RESOURCE_VITAL_PRODUCT_DATA:
		for (lim = ptr + len; ptr < lim; )
		{
		    name[0] = *ptr++;
		    name[1] = *ptr++;
		    name[2] = '\0';
		    len2    = *ptr++;
		    
		    for (i = 0; i < len2 && ptr < lim; i++)
			value[i] = *ptr++;
		    value[i] = '\0';
		    		    
		    PRINTF3 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, %s: $s\n", dev->dev_instance, name, value);

		    if (tag != NULL) 
		    { /* looking for just one tag */
			if (!strcmp (name, tag))
			    strcpy(result, value);
		    } 
		    else 
		    { /* get all tags */
			strcat(result,name);
			strcat(result,": ");
			strcat(result,value);
			strcat(result,"\n");
		    }
		}
		break;
		
	    default:
		PRINTF2 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, unknown large resource %x\n", dev->dev_instance, type);
		finished = 1;
		break;
	    }
	}
	else
	{
	    len = type & 0x7;
	    
	    switch (type >> 3)
	    {
	    case SMALL_RESOURCE_COMPATIBLE_DEVICE_ID:
		ptr += len;
		break;
		
	    case SMALL_RESOURCE_VENDOR_DEFINED:
		ptr += len;
		break;
		
	    case SMALL_RESOURCE_END_TAG:
		finished = 1;
		break;
		
	    default:
		PRINTF2 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, unknown small resource %x\n", dev->dev_instance, type >> 3);
		finished = 1;
		break;
	    }
	}
    }

    if ( result[0] == 0 ) {
	if ( tag != 0 ) 
	    PRINTF2 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, failed to find tag %s\n", dev->dev_instance, tag);
	else
	    PRINTF1 (DBG_DEVICE, DBG_CONFIG, "elan%d: elan4_read_vpd, failed to find any tags\n", dev->dev_instance);
	return -EINVAL;
    }

    return (0);
}

int
elan4_start_device (ELAN4_DEV *dev)
{
    E4_VirtualProcessEntry entry;
    unsigned	           pagesizeval[2];
    unsigned	           hashsizeval[2];
    register int           i, j, tbl, res;
    unsigned               attempts = 0;
    E4_PAGE_SIZE_TABLE;
    unsigned char          serial[256];
    unsigned int           sdram_factor = SDRAM_166_DLL_CORRECTION_FACTOR;

    PRINTF (DBG_DEVICE, DBG_ALL, "elan4_start_device: entered\n");

    dev->dev_state = ELAN4_STATE_STARTING;

 tryagain:
    /* Initialise the pci config space */
    if ((res = elan4_pciinit (dev)) < 0)
	return (res);

    /* Display the serial number */
    if (elan4_read_vpd (dev, "SN", serial))
	printk("elan%d: SN: failed to read\n", dev->dev_instance);
    else
	printk("elan%d: SN: %s\n", dev->dev_instance, serial);

    /* initialise the interrupt mask to zero */
    SET_INT_MASK (dev, 0);

    /* Initialise the device registers */
    write_reg64 (dev, TlbLineValue, 0);
    write_reg64 (dev, SysControlReg, 0);

    /* Initialise the SDRAM using the configuration value from the ROM */
    write_reg64 (dev, SDRamConfigReg, dev->dev_sdram_cfg | SDRAM_SETUP);

    /* Setup the linkport registers */
    write_reg64 (dev, LinkPortLock, elan4_linkport_lock);

    /* Setup the tick rates, start the clock, and init the stats registers */
    write_ureg32 (dev, ClockTickRate.s.TickRates, ELAN4_CLOCK_TICK_RATE);
    write_ureg64 (dev, Clock, 0);
    write_ureg32 (dev, InstCount.s.StatsCount, 0);
    for (i = 0; i < 8; i++)
	write_ureg32 (dev, StatCounts[i].s.StatsCount, 0);

    /* Initialise the Link Control register - disable the TLB prefetcher on RevB
     * as it can cause very occasional data corruption. */
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVB)
	write_reg32 (dev, LinkControlReg, LCONT_EN_SYS_READS | LCONT_REVB_DISABLE_TLB_PREFETCH);
    else
	write_reg32 (dev, LinkControlReg, LCONT_EN_SYS_READS);

    /* Initialise the Link Control Settings to set the PLL Reference Value */
    write_reg32 (dev, LinkContSettings, 
		 (elan4_mod45disable ? LCONT_MOD45_DISABLE : 0) |
		 (3 << LCONT_CONFIG_PHASE_SHIFT) |
		 ((elan4_pll_div & LCONT_PLL_REF_VAL_BITS_MASK) << LCONT_PLL_REF_VAL_BITS_SHIFT) |
		 (LCONT_VOD_360 << LCONT_LVDS_VOLTAGE_BITS_SHIFT) |
		 (LCONT_TERM_AUTO_OHM << LCONT_LVDS_TERMINATION_SHIFT));

    /* Clear the link error LED on RevB and above */
    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA)
	write_i2c (dev, I2cStatus, read_i2c (dev, I2cStatus) | I2cCntl_ClearLinkError);

    /* Compute the SysClk frequency and update the PLL if necessary */
    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA)
    {
	int mhz = measure_sysclk (dev);

	if (elan4_pll_cfg != 0 || mhz > 190 || mhz < 170)
	    printk ("elan%d: SysClk running at %d Mhz\n", dev->dev_instance, measure_sysclk (dev));
	else
	{
	    sdram_factor = SDRAM_150_DLL_CORRECTION_FACTOR;

	    elan4_updatepll (dev, ECTRL_SYS_CLOCK_RATIO_4_3);
	    
	    printk ("elan%d: SysClk now running at %d Mhz\n", dev->dev_instance, measure_sysclk (dev));
	}
    }
	
    initialise_cache (dev);

    /* Initialise the MMU hash table parameters */
    /* Select the largest elan pagesize which is spanned by the
     * system pagesize for mmu table 0*/
    for (i = 0; i < E4_PAGE_SIZE_TABLE_SIZE; i++)
	if (PageSizeTable[i] > PAGE_SHIFT)
	    break;

    pagesizeval[0] = i - 1;
    hashsizeval[0] = elan4_hash_0_size_val;
	
    /* Select a suitable elan pagesize to match any "large" page
     * support that the OS provides. */
    pagesizeval[1] = PAGE_SIZE_4M;
    hashsizeval[1] = elan4_hash_1_size_val;

    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
    {
	dev->dev_pagesizeval[tbl]   = pagesizeval[tbl];
	dev->dev_pageshift[tbl]     = PageSizeTable[pagesizeval[tbl]];
	dev->dev_hashsize[tbl]      = (1 << hashsizeval[tbl])/sizeof (E4_HashTableEntry);
	dev->dev_rsvd_hashmask[tbl] = ((1 << (27 - dev->dev_pageshift[tbl]))-1) & ~((1 << hashsizeval[tbl])-1);
	dev->dev_rsvd_hashval[tbl]  = 0xFFFFFFFF;
    }

    PRINTF2 (DBG_DEVICE, DBG_CONFIG, "elan4_start_device: pageshifts %d,%d\n", dev->dev_pageshift[0], 
	     NUM_HASH_TABLES == 2 ? dev->dev_pageshift[1] : 0);

    /* Initialise the control register to the desired value */
    dev->dev_syscontrol = (CONT_EN_ALL_SETS | CONT_MMU_ENABLE | CONT_CACHE_ALL | CONT_2K_NOT_1K_DMA_PACKETS |
			   (pagesizeval[0] << CONT_TABLE0_PAGE_SIZE_SHIFT) | (hashsizeval[0] << CONT_TABLE0_MASK_SIZE_SHIFT));

    if (NUM_HASH_TABLES == 2)
	dev->dev_syscontrol |= CONT_TWO_HASH_TABLES | (pagesizeval[1] << CONT_TABLE1_PAGE_SIZE_SHIFT) | (hashsizeval[1] << CONT_TABLE1_MASK_SIZE_SHIFT);

    write_reg64 (dev, SysControlReg, dev->dev_syscontrol);

    /* use direct mapped pci writes during sdram initialisation, since for 
     * cache flushing to work, we need to ensure that the cacheflush page
     * never gets lines into the incorrect cache set. */
    SET_SYSCONTROL (dev, dev_direct_map_pci_writes, CONT_DIRECT_MAP_PCI_WRITES);

    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVB)
	elan4_sdram_setup_delay_lines(dev, sdram_factor);

    for (i = res = 0; i < dev->dev_sdram_numbanks; i++)
	if (dev->dev_sdram_banks[i].b_size)
	    res |= elan4_sdram_init_bank (dev, &dev->dev_sdram_banks[i]);

    if (! res)
    {
	if (dev->dev_devinfo.dev_device_id == PCI_REVISION_ID_ELAN4_REVB && ++attempts < 5)
	{
	    printk ("elan%d: sdram not working, resetting\n", dev->dev_instance);
	    goto tryagain;
	}

	printk ("elan%d: could not find any sdram banks\n", dev->dev_instance);
	goto failed;
    }

#ifndef CONFIG_MPSAS
    PRINTF0 (DBG_DEVICE, DBG_CONFIG, "elan4_start_device: initialising for ECC\n");

    for (i = 0 ; i < dev->dev_sdram_numbanks; i++)
	if (dev->dev_sdram_banks[i].b_ioaddr)
	    initialise_ecc (dev, &dev->dev_sdram_banks[i]);
#endif

    dev->dev_sdram_initial_ecc_val = read_reg64 (dev, SDRamECCStatus);

    /* Now enable ECC after we've scrubbed the memory */
    write_reg64 (dev, SDRamConfigReg, dev->dev_sdram_cfg | SDRAM_ENABLE_ECC);

    /* clear any errors, and flush the tlb/route cache */
    PULSE_SYSCONTROL (dev, CONT_TLB_FLUSH | CONT_ROUTE_FLUSH | CONT_CLEAR_LINKPORT_INT | CONT_CLEAR_SDRAM_ERROR);

    write_ureg32 (dev, InstCount.s.StatsCount, 0);

    /* Initialise the thread processor's register file */
    for (i = 0; i < 64; i++)
	write_reg64 (dev, TProcRegs[i], 0);

    /* Initialise the thread processor's ICache tags */
    for (i = 0; i < (E4_ICacheLines/E4_ICachePortSize); i++)
    {
        write_reg64 (dev, ICachePort_Cntl_Addr, i << E4_ICacheTagAddrShift);
        for (j = 0; j < E4_ICachePortSize; j++)
	    write_reg64 (dev, ICachePort[j], E4_InvalidTagValue);
    }

    /*
     * Initialise the ICache with a sethi %hi(addr << 7), %r0
     * writing 8 64 bit values per loop of sethi %g0 values ending in 77 for something different??
     */
    for (i = 0; i < E4_ICacheSizeInBytes; i += (E4_ICachePortSize << 3))
    {
	write_reg64 (dev, ICachePort_Cntl_Addr, E4_AccessICacheRams | (i >> 3));

	for (j = 0; j < E4_ICachePortSize; j++)
	    write_reg64 (dev, ICachePort[j], 
			 (E4_uint64) (((E4_uint64)i << (4+7))    + ((E4_uint64)j << (1+7))    + (0x077)) |
		         (E4_uint64) (((E4_uint64)i << (4+7+32)) + ((E4_uint64)j << (1+7+32)) + (0x0e7)) << 32);
    }

    /*
     * Initialise the top of the ICache Set0 with a instruction which will
     * cause a know trap fingerprint so that the application can identify it
     * and ignore the trap.
     */
    write_reg64 (dev, ICachePort_Cntl_Addr, E4_ICacheFixupOffset | E4_AccessICacheRams);
    for (i = 0; i < E4_ICachePortSize; i++)
	write_reg64 (dev, ICachePort[i], E4_ICacheFixupInsn | (E4_ICacheFixupInsn << 32));

    /* create the buddy allocator for SDRAM */
    for (i = 0; i < dev->dev_sdram_numbanks; i++)
	if (dev->dev_sdram_banks[i].b_ioaddr)
	    elan4_sdram_add_bank (dev, &dev->dev_sdram_banks[i]);

    dev->dev_ctxtableshift        = elan4_ctxt_table_shift;
    dev->dev_cqcount              = (1 << elan4_ln2_max_cqs);
    dev->dev_cqreorder            = 0;

    /* allocate the sdram for cache flushing whilst still in direct mapped mode */
    dev->dev_cacheflush_space = elan4_sdram_alloc (dev, E4_CacheSize);

    /* and longer need direct mapped pci writes */
    CLEAR_SYSCONTROL (dev, dev_direct_map_pci_writes, CONT_DIRECT_MAP_PCI_WRITES);

    /* allocate the hash tables, command queues, context tables etc */
    PRINTF0 (DBG_DEVICE, DBG_CONFIG, "elan4_start_device: allocating hash tables, command queueus, context tables\n");

    dev->dev_comqlowpri       = elan4_sdram_alloc (dev, (1 << COMMAND_RUN_QUEUE_BITS));
    dev->dev_comqhighpri      = elan4_sdram_alloc (dev, (1 << COMMAND_RUN_QUEUE_BITS));
    dev->dev_cqaddr           = elan4_sdram_alloc (dev, sizeof (E4_CommandQueueDesc) * dev->dev_cqcount);
    dev->dev_dmaqhighpri      = elan4_sdram_alloc (dev, E4_QueueSize(elan4_dmaq_highpri_size));
    dev->dev_dmaqlowpri       = elan4_sdram_alloc (dev, E4_QueueSize(elan4_dmaq_lowpri_size));
    dev->dev_threadqhighpri   = elan4_sdram_alloc (dev, E4_QueueSize(elan4_threadq_highpri_size));
    dev->dev_threadqlowpri    = elan4_sdram_alloc (dev, E4_QueueSize(elan4_threadq_lowpri_size));
    dev->dev_interruptq       = elan4_sdram_alloc (dev, E4_QueueSize(elan4_interruptq_size));

    dev->dev_ctxtable         = elan4_sdram_alloc (dev, (1 << dev->dev_ctxtableshift) * sizeof (E4_ContextControlBlock));
    dev->dev_faultarea        = elan4_sdram_alloc (dev, CUN_Entries * sizeof (E4_FaultSave));
    dev->dev_inputtraparea    = elan4_sdram_alloc (dev, sizeof (E4_IprocTrapState));

    dev->dev_sdrampages[0]    = elan4_sdram_alloc (dev, SDRAM_PAGE_SIZE);
    dev->dev_sdrampages[1]    = elan4_sdram_alloc (dev, SDRAM_PAGE_SIZE);

    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
    {
	dev->dev_hashtable[tbl] = elan4_sdram_alloc (dev, dev->dev_hashsize[tbl] * sizeof (E4_HashTableEntry));
#ifndef CONFIG_MPSAS
	/* Initialise hash tables to invalid (zero) */
	elan4_sdram_zeroq_sdram (dev, dev->dev_hashtable[tbl], dev->dev_hashsize[tbl] * sizeof (E4_HashTableEntry));
#endif
    }

    /* Initialise all context filters to discard */
#ifdef CONFIG_MPSAS
    if (sas_memset_dev (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM, dev->dev_ctxtable, 
			E4_FILTER_DISCARD_ALL, (1 << (dev->dev_ctxtableshift-1))) < 0)
    {
	for (i = 0; i < (1 << dev->dev_ctxtableshift); i++)
	    elan4_write_filter (dev, i, E4_FILTER_DISCARD_ALL);
    }
#else
    for (i = 0; i < (1 << dev->dev_ctxtableshift); i++)
	elan4_write_filter (dev, i, E4_FILTER_DISCARD_ALL);
#endif

    PRINTF4 (DBG_DEVICE, DBG_CONFIG, "elan4_start_device: hashtables %x,%x, %x,%x\n", dev->dev_hashtable[0], 
	    dev->dev_hashsize[0], dev->dev_hashtable[1], dev->dev_hashsize[1]);

    /* install the hash table pointers */
    PRINTF0 (DBG_DEVICE, DBG_CONFIG, "elan4_start_device: initialise registers with table addresses\n");
    write_reg64 (dev, MmuTableBasePtrs, (((E4_uint64) dev->dev_hashtable[0]) | ((E4_uint64) dev->dev_hashtable[1]) << 32));
    write_reg64 (dev, MmuFaultAndRootCntxPtr, (((E4_uint64) dev->dev_ctxtableshift) | 
					       ((E4_uint64) dev->dev_ctxtable) |
					       ((E4_uint64) dev->dev_faultarea) << 32));
    write_reg64 (dev, InputTrapAndFilter, (((E4_uint64) dev->dev_ctxtableshift) | 
					   ((E4_uint64) dev->dev_ctxtable) |
					   ((E4_uint64) dev->dev_inputtraparea) << 32));
    /*
     * The run ptrs have this format: (Front << 32) | Back
     * The base for both the front and back is uses the high bits of the back pointer.
     * So writting just the base value is good enough.
     */
    write_reg64 (dev, CommandLowPriRunPtrs,  dev->dev_comqlowpri);
    write_reg64 (dev, CommandHighPriRunPtrs, dev->dev_comqhighpri);

    /* Initialise the run queues */
    write_reg64 (dev, DProcHighPriPtrs,       E4_QueueValue (dev->dev_dmaqhighpri,    elan4_dmaq_highpri_size));
    write_reg64 (dev, DProcLowPriPtrs,        E4_QueueValue (dev->dev_dmaqlowpri,     elan4_dmaq_lowpri_size));
    write_reg64 (dev, TProcHighPriPtrs,       E4_QueueValue (dev->dev_threadqhighpri, elan4_threadq_highpri_size));
    write_reg64 (dev, TProcLowPriPtrs,        E4_QueueValue (dev->dev_threadqlowpri,  elan4_threadq_lowpri_size));

    /* Initialise the interrupt queue as "empty" - this is actually with one entry on it */
    write_reg64 (dev, MainIntQueuePtrs.Value, (((E4_uint64) E4_QueueFrontValue (dev->dev_interruptq, elan4_interruptq_size) << 32) |
					       ((E4_uint64) E4_QueueBackPointer(dev->dev_interruptq + E4_MainIntEntrySize))));
    
    dev->dev_interruptq_nfptr = dev->dev_interruptq + E4_MainIntEntrySize;

    /*
     * Flush the context filter before dropping the Discard all bits in the schedule status register.
     * Also hit the SCH_RestartTProc to clear out X's from the trap state and
     * hit the SCH_RestartDmaPrefetchProc to clear out X's from the prev register.
     */
    PULSE_SCHED_RESTART (dev, SCH_ContextFilterFlush | SCH_RestartTProc | SCH_RestartDmaPrefetchProc);

    /* setup the schedule status register. */
    SET_SCHED_STATUS (dev, SCH_CProcTimeout6p2us | SCH_DProcTimeslice512us);

    /*
     * Now initialise the inserter cache.s
     * Bit 31 of the first word of the descriptor is a valid bit. This must be cleared.
     * Bit 31 becomes a used bit in the descriptors in memory.
     */
    for (i = 0; i < COMMAND_INSERTER_CACHE_ENTRIES; i++)
    {
	write_reg32 (dev, CommandControl.CommandQueueDescsBase, i); 	/* select a cache line */
	write_reg64 (dev, CommandCacheTestPort, 0);			/* Mark it invalid */
    }
    
    /* Setup the pointer to the command descriptors */
    /*   the table must be aligned on a CQ_CommandDescsAlignement boundary */
    /*   since we've allocated a small table - we work out the offset of the */
    /*   first entry in our table for mapping in the command ports later */
    dev->dev_cqoffset = (dev->dev_cqaddr & (CQ_CommandDescsAlignment-1)) / sizeof (E4_CommandQueueDesc);

    write_reg32 (dev, CommandControl.CommandQueueDescsBase, (dev->dev_cqaddr & ~(CQ_CommandDescsAlignment-1)) | COM_ENABLE_DEQUEUE);

    /* allocate the bitmaps for cq,ctxt allocation */
    KMEM_ZALLOC (dev->dev_cqamap, bitmap_t *, BT_BITOUL(dev->dev_cqcount/ELAN4_CQ_PER_CQA) * sizeof (bitmap_t), 1);
    KMEM_ZALLOC (dev->dev_ctxmap, bitmap_t *, BT_BITOUL(1 << dev->dev_ctxtableshift) * sizeof (bitmap_t), 1);

    if (dev->dev_cqamap == NULL || dev->dev_ctxmap == NULL)
	goto failed;

    /* Make every fourth context be invalid for ICache fixup.
     * context 0 is also invalid - since it is used to indicate 
     * an invalid tag. */
    for (i = 0; i < (1 << dev->dev_ctxtableshift); i += 4)
	BT_SET (dev->dev_ctxmap, i);
    
    /* initialise the halt operations */
    dev->dev_haltop_mask   = 0;
    dev->dev_haltop_active = 0;

    /* allocate the hash table shadow structures - and place all blocks on the free lists */
    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
    {
	KMEM_ZALLOC (dev->dev_mmuhash[tbl], ELAN4_HASH_ENTRY *,  dev->dev_hashsize[tbl] * sizeof (ELAN4_HASH_ENTRY), 1);

	if (dev->dev_mmuhash[tbl] == NULL)
	    goto failed;

	for (i = 0; i < dev->dev_hashsize[tbl]; i++)
	    dev->dev_mmuhash[tbl][i].he_entry = dev->dev_hashtable[tbl] + (i * sizeof (E4_HashTableEntry));
    }

    /* setup the interrupt mask register */
    SET_INT_MASK (dev, (INT_MSI0 | INT_MSI1 | INT_MSI2 | INT_MSI3) & ~(INT_Discarding | INT_Halted | INT_LinkPortKeyFail));

    /* start a thread to handle excessive main interrupts */
    if (kernel_thread_create (elan4_mainint_thread, (caddr_t) dev) == NULL)
	goto failed;
    dev->dev_mainint_started = 1;
    
    /* install the device context - and allocate the first 16 command queues */
    if (elan4_insertctxt (dev, &dev->dev_ctxt, &device_trap_ops) != 0)
	goto failed;

    /* Allocate command queues, one for each entry in the inserter cache, 
     * we'll use these queues to flush the insert cache */
    for (i = 0; i < COMMAND_INSERTER_CACHE_ENTRIES; i++)
    {
	if ((dev->dev_flush_cq[i] = elan4_alloccq (&dev->dev_ctxt, CQ_Size1K, CQ_DmaStartEnableBit | CQ_InterruptEnableBit, 
						   CQ_Priority)) == NULL)
	    goto failed;
	
	ASSERT (elan4_cq2num(dev->dev_flush_cq[i]) == i);

	dev->dev_flush_finished |= (1 << i);
    }

    /* Allocate command queues for dma halt operations */
    if ((dev->dev_dma_flushop[0].cq = elan4_alloccq (&dev->dev_ctxt, CQ_Size1K, CQ_DmaStartEnableBit|CQ_WaitEventEnableBit, 0)) == NULL ||
	(dev->dev_dma_flushop[1].cq = elan4_alloccq (&dev->dev_ctxt, CQ_Size1K, CQ_DmaStartEnableBit|CQ_WaitEventEnableBit, CQ_Priority)) == NULL)
	goto failed;

#ifdef CONFIG_MPSAS
    elan4_sdram_flushcache (dev, 0, E4_CacheSize);
#endif

    /* initialise halt operation for flushing the icache */
    dev->dev_iflush_haltop.op_function = device_iflush_haltop;
    dev->dev_iflush_haltop.op_arg      = dev;
    dev->dev_iflush_haltop.op_mask     = INT_TProcHalted;

    /* Allocate a route table, and create a valid route for vp==0, this is used
     * when a DMA is removed from the dma run queue */
    if ((dev->dev_routetable = elan4_alloc_routetable (dev, 0)) == NULL)
	goto failed;

    elan4_set_routetable (&dev->dev_ctxt, dev->dev_routetable);

    entry.Values[0] = FIRST_MYLINK;
    entry.Values[1] = 0;

    elan4_write_route (dev, dev->dev_routetable, 0, &entry);

    /* map the sdram pages into the elan */
    dev->dev_tproc_suspend = DEVICE_TPROC_SUSPEND_ADDR;
    dev->dev_tproc_space   = DEVICE_TPROC_SPACE_ADDR;


    elan4mmu_pteload (&dev->dev_ctxt, 0, dev->dev_tproc_suspend, HE_TYPE_SDRAM, (dev->dev_sdrampages[0] >> PTE_PADDR_SHIFT) | PTE_SetPerm(PERM_LocExecute));
    elan4mmu_pteload (&dev->dev_ctxt, 0, dev->dev_tproc_space,   HE_TYPE_SDRAM, (dev->dev_sdrampages[1] >> PTE_PADDR_SHIFT) | PTE_SetPerm(PERM_LocEventWrite));

    /* and store the thread suspend sequence in it for use when a thread is removed from the run queue */
    elan4_sdram_writel (dev, dev->dev_sdrampages[0], DEVICE_TPROC_SUSPEND_INSTR);

    /* and initialise the dma flush event in sdrampage[1] */
    elan4_sdram_writeq (dev, dev->dev_sdrampages[1] + 64, E4_EVENT_INIT_VALUE (-32, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));

#ifdef CONFIG_MPSAS
    do_initdma (dev);
#endif
    
    if (!elan4_neterr_init (dev))
	goto failed;

    elan4_configure_writecombining (dev);

    /* finally register the device with elanmod for rms */
    dev->dev_idx = elan_dev_register (&dev->dev_devinfo, &elan4_dev_ops, (void *) dev);

    dev->dev_state = ELAN4_STATE_STARTED;

    return (0);

 failed:
    printk ("elan%d: failed to start elan4 device - stopping\n", dev->dev_instance);

    elan4_stop_device (dev);
    return (-ENOMEM);
}

void
elan4_stop_device (ELAN4_DEV *dev)
{
    unsigned long flags;
    int i, tbl;

    dev->dev_state = ELAN4_STATE_STOPPING;

    elan_dev_deregister (&dev->dev_devinfo);

    elan4_unconfigure_writecombining (dev);

    elan4_neterr_destroy (dev);

    if (dev->dev_tproc_suspend)
	elan4mmu_unload_range (&dev->dev_ctxt, 0, dev->dev_tproc_suspend, 1 << dev->dev_pageshift[0]);

    if (dev->dev_tproc_space)
	elan4mmu_unload_range (&dev->dev_ctxt, 0, dev->dev_tproc_space,   1 << dev->dev_pageshift[0]);

    if (dev->dev_routetable)
    {
	elan4_set_routetable (&dev->dev_ctxt, NULL);
	elan4_free_routetable (dev, dev->dev_routetable);
    }

    for (i = 0; i < 2; i++)
	if (dev->dev_dma_flushop[i].cq)
	    elan4_freecq (&dev->dev_ctxt, dev->dev_dma_flushop[i].cq);

    /* free of the device context - and insert cache flushing command queues */
    for (i = 0; i < COMMAND_INSERTER_CACHE_ENTRIES; i++)
	if (dev->dev_flush_cq[i])
	    elan4_freecq (&dev->dev_ctxt, dev->dev_flush_cq[i]);

    if (dev->dev_ctxt.ctxt_dev)
	elan4_removectxt (dev, &dev->dev_ctxt);

    /* stop the mainint thread */
    spin_lock_irqsave (&dev->dev_mainint_lock, flags);
    dev->dev_stop_threads = 1;

    while (dev->dev_mainint_started && !dev->dev_mainint_stopped)
    {
	kcondvar_wakeupall (&dev->dev_mainint_wait, &dev->dev_mainint_lock);
	kcondvar_wait (&dev->dev_mainint_wait, &dev->dev_mainint_lock, &flags);
    }
    dev->dev_mainint_started = dev->dev_mainint_stopped = 0;
    spin_unlock_irqrestore (&dev->dev_mainint_lock, flags);

    /* cancel any error interrupt timeouts */
    if (timer_fn_queued (&dev->dev_error_timeoutid))
	cancel_timer_fn (&dev->dev_error_timeoutid);

    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA && timer_fn_queued (&dev->dev_linkerr_timeoutid))
	cancel_timer_fn (&dev->dev_linkerr_timeoutid);
    
    /* reset the interrupt mask register to zero */
    if (dev->dev_regs)
	SET_INT_MASK (dev, 0);

    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
    {
	if (dev->dev_mmuhash[tbl])
	    KMEM_FREE (dev->dev_mmuhash[tbl], dev->dev_hashsize[tbl] * sizeof (ELAN4_HASH_ENTRY));
	if (dev->dev_hashtable[tbl])
	    elan4_sdram_free (dev, dev->dev_hashtable[tbl], dev->dev_hashsize[tbl] * sizeof (E4_HashTableEntry));
    }

    if (dev->dev_cqamap)
	KMEM_FREE (dev->dev_cqamap, BT_BITOUL (dev->dev_cqcount/ELAN4_CQ_PER_CQA) * sizeof (bitmap_t));
    if (dev->dev_ctxmap)
	KMEM_FREE (dev->dev_ctxmap, BT_BITOUL(1 << dev->dev_ctxtableshift) * sizeof (bitmap_t));

    if (dev->dev_comqlowpri)
	elan4_sdram_free (dev, dev->dev_comqlowpri,     (1 << COMMAND_RUN_QUEUE_BITS));
    if (dev->dev_comqhighpri)
	elan4_sdram_free (dev, dev->dev_comqhighpri,    (1 << COMMAND_RUN_QUEUE_BITS));
    if (dev->dev_cqaddr)
	elan4_sdram_free (dev, dev->dev_cqaddr,         sizeof (E4_CommandQueueDesc) * dev->dev_cqcount);
    if (dev->dev_dmaqhighpri)
	elan4_sdram_free (dev, dev->dev_dmaqhighpri,    E4_QueueSize(elan4_dmaq_highpri_size));
    if (dev->dev_dmaqlowpri)
	elan4_sdram_free (dev, dev->dev_dmaqlowpri,     E4_QueueSize(elan4_dmaq_lowpri_size));
    if (dev->dev_threadqhighpri)
	elan4_sdram_free (dev, dev->dev_threadqhighpri, E4_QueueSize(elan4_threadq_highpri_size));
    if (dev->dev_threadqlowpri)
	elan4_sdram_free (dev, dev->dev_threadqlowpri,  E4_QueueSize(elan4_threadq_lowpri_size));
    if (dev->dev_interruptq)
	elan4_sdram_free (dev, dev->dev_interruptq,     E4_QueueSize(elan4_interruptq_size));
    
    if (dev->dev_ctxtable)
	elan4_sdram_free (dev, dev->dev_ctxtable,       (1 << dev->dev_ctxtableshift) * sizeof (E4_ContextControlBlock));
    if (dev->dev_faultarea)
	elan4_sdram_free (dev, dev->dev_faultarea,      CUN_Entries * sizeof (E4_FaultSave));
    if (dev->dev_inputtraparea)
	elan4_sdram_free (dev, dev->dev_inputtraparea,  sizeof (E4_IprocTrapState));

    if (dev->dev_sdrampages[0])
	elan4_sdram_free (dev, dev->dev_sdrampages[0],  SDRAM_PAGE_SIZE);
    if (dev->dev_sdrampages[1])
	elan4_sdram_free (dev, dev->dev_sdrampages[1],  SDRAM_PAGE_SIZE);

    for (i = 0; i < dev->dev_sdram_numbanks; i++)
	if (dev->dev_sdram_banks[i].b_ioaddr)
		elan4_sdram_fini_bank (dev, &dev->dev_sdram_banks[i]);

    elan4_pcifini (dev);

    dev->dev_state = ELAN4_STATE_STOPPED;

    if (dev->dev_ack_errors)
        kfree(dev->dev_ack_errors);
    if (dev->dev_dproc_timeout)
        kfree(dev->dev_dproc_timeout);
    if (dev->dev_cproc_timeout)
        kfree(dev->dev_cproc_timeout);
}

static __inline__ int
compute_arity (int lvl, unsigned n, char *arity)
{
    if (arity[lvl] == 0)
    {
	if (n <= 8)
	    arity[lvl] = n;
	else
	    arity[lvl] = 4;
    }

    return (arity[lvl]);
}

int
elan4_compute_position (ELAN_POSITION *pos, unsigned nodeid, unsigned numnodes, unsigned arityval)
{
    int i, lvl, n;
    char arity[ELAN_MAX_LEVELS];

    if (nodeid >= numnodes)
	return -EINVAL;

    for (i = 0; i < ELAN_MAX_LEVELS; i++, arityval >>= 4)
	arity[i] = arityval & 7;
    
    for (lvl = 0, n = numnodes; n > compute_arity(lvl, n, arity) && lvl < ELAN_MAX_LEVELS; lvl++)
    {
	if ((n % arity[lvl]) != 0)
	    return -EINVAL;
	
	n /= arity[lvl];
    }

    if (arity[lvl] != n)
	return -EINVAL;

    for (i = 0; i <= lvl; i++)
	pos->pos_arity[i] = arity[lvl - i];

    pos->pos_nodes  = numnodes;
    pos->pos_levels = lvl + 1;
    pos->pos_nodeid = nodeid;
    pos->pos_mode   = ELAN_POS_MODE_SWITCHED;

    return 0;
}

int
elan4_get_position (ELAN4_DEV *dev, ELAN_POSITION *pos)
{
    kmutex_lock (&dev->dev_lock);
    *pos = dev->dev_position;
    kmutex_unlock (&dev->dev_lock);

    return (pos->pos_mode);
}

int
elan4_set_position (ELAN4_DEV *dev, ELAN_POSITION *pos)
{
    int forceLocal = 0;
    int nnodes, i;
    unsigned int *ack_errors;
    unsigned int *dproc_timeout;
    unsigned int *cproc_timeout;

    switch (pos->pos_mode)
    {
    case ELAN_POS_UNKNOWN:
	break;
	
    case ELAN_POS_MODE_SWITCHED:
	if (pos->pos_levels > ELAN_MAX_LEVELS)
	    return (-EINVAL);
	
	for (i = 0, nnodes = 1; i < pos->pos_levels; i++)
	{

	    if (pos->pos_arity[i] <= 0 || (i == 0 ? pos->pos_arity[i] > 8 : pos->pos_arity[i] >= 8))  /* allow an 8 way top-switch */
		return (-EINVAL);
	    
	    nnodes *= pos->pos_arity[i];
	}

	if (pos->pos_nodes > nnodes || pos->pos_nodeid >= pos->pos_nodes)
	    return (-EINVAL);
	break;
	
    case ELAN_POS_MODE_LOOPBACK:
	if (pos->pos_levels != 1 || pos->pos_nodes != 1 || pos->pos_nodeid != 0 || pos->pos_arity[0] != 1)
	    return (-EINVAL);

	forceLocal = 1;
	break;

    case ELAN_POS_MODE_BACKTOBACK:
	if (pos->pos_levels != 1 || pos->pos_nodes != 2 || pos->pos_nodeid >= 2 || pos->pos_arity[0] != 2)
	    return (-EINVAL);

	forceLocal = (pos->pos_nodeid == 0);
	break;

    default:
	return (-EINVAL);
    }

    ack_errors = kmalloc(pos->pos_nodes * sizeof(unsigned int), GFP_KERNEL);
    if (!ack_errors)
	return (-EINVAL);
    memset(ack_errors, 0, pos->pos_nodes * sizeof(unsigned int));
    dproc_timeout = kmalloc(pos->pos_nodes * sizeof(unsigned int), GFP_KERNEL);
    if (!dproc_timeout) 
    {
        kfree(ack_errors);
        return (-EINVAL);
    }
    memset(dproc_timeout, 0, pos->pos_nodes * sizeof(unsigned int));
    cproc_timeout = kmalloc(pos->pos_nodes * sizeof(unsigned int), GFP_KERNEL);
    if (!cproc_timeout)
    {
        kfree(ack_errors);
        kfree(dproc_timeout);
        return (-EINVAL);
    }
    memset(cproc_timeout, 0, pos->pos_nodes * sizeof(unsigned int));
	
    kmutex_lock (&dev->dev_lock);
    dev->dev_position = *pos;
    dev->dev_ack_errors = ack_errors;
    dev->dev_dproc_timeout = dproc_timeout;
    dev->dev_cproc_timeout = cproc_timeout;

    if (forceLocal)
	write_reg32 (dev, LinkContSettings, read_reg32 (dev, LinkContSettings) | LCONT_FORCE_COMMSCLK_LOCAL);
    else
	write_reg32 (dev, LinkContSettings, read_reg32 (dev, LinkContSettings) & ~LCONT_FORCE_COMMSCLK_LOCAL);

    pioflush_reg (dev);
    kmutex_unlock (&dev->dev_lock);

    return (0);
}

void
elan4_get_params (ELAN4_DEV *dev, ELAN_PARAMS *params, unsigned short *mask)
{
    kmutex_lock (&dev->dev_lock);

    *mask = dev->dev_devinfo.dev_params_mask;
    memcpy (params, &dev->dev_devinfo.dev_params, sizeof (ELAN_PARAMS));
    
    kmutex_unlock (&dev->dev_lock);
}

void
elan4_set_params (ELAN4_DEV *dev, ELAN_PARAMS *params, unsigned short mask)
{	
    int i;

    kmutex_lock (&dev->dev_lock);
    for (i = 0; i < ELAN4_PARAM_COUNT; i++)
	if (mask & (1 << i))
	    dev->dev_devinfo.dev_params.values[i] = params->values[i];
    
    dev->dev_devinfo.dev_params_mask |= mask;
    kmutex_unlock (&dev->dev_lock);
}


EXPORT_SYMBOL(elan4_get_position);
EXPORT_SYMBOL(elan4_set_position);

EXPORT_SYMBOL(elan4_queue_haltop);
EXPORT_SYMBOL(elan4_queue_dma_flushop);
EXPORT_SYMBOL(elan4_queue_mainintop);

EXPORT_SYMBOL(elan4_insertctxt);
EXPORT_SYMBOL(elan4_removectxt);

EXPORT_SYMBOL(elan4_attach_filter);
EXPORT_SYMBOL(elan4_detach_filter);
EXPORT_SYMBOL(elan4_set_filter);
EXPORT_SYMBOL(elan4_set_routetable);

EXPORT_SYMBOL(elan4_alloccq);
EXPORT_SYMBOL(elan4_freecq);
EXPORT_SYMBOL(elan4_restartcq);

EXPORT_SYMBOL(elan4_flush_icache);
EXPORT_SYMBOL(elan4_hardware_lock_check);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
