/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: device_Linux.c,v 1.110.2.9 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/device_Linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>
#include <qsnet/kpte.h>

#include <asm/io.h>
#include <asm/irq.h>
#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#include <linux/init.h>
#include <linux/pci.h>
#include <qsnet/module.h>
#include <linux/reboot.h>
#include <linux/notifier.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#include <linux/wrapper.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
typedef void irqreturn_t;
#endif
#       define IRQ_NONE
#       define IRQ_HANDLED
#endif

#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>
#include <elan4/ioctl.h>
#include <elan4/intcookie.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#error please use a 2.4.0 series kernel or newer
#endif


#if defined(LINUX_SPARC) || defined(LINUX_PPC64)
#define __io_remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#define __remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#elif defined(NO_RMAP)
#define __io_remap_page_range(from,offset,size,prot)	io_remap_page_range(from,offset,size,prot)
#define __remap_page_range(from,offset,size,prot)	remap_page_range(from,offset,size,prot)
#else
#define __io_remap_page_range(from,offset,size,prot)	io_remap_page_range(vma,from,offset,size,prot)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
#define __remap_page_range(from,offset,size,prot)	remap_pfn_range(vma,from,(offset)>>PAGE_SHIFT,size,prot)
#else
#define __remap_page_range(from,offset,size,prot)	remap_page_range(vma,from,offset,size,prot)
#endif
#endif

#if defined (X86_FEATURE_PAT)
static unsigned int pat_pteval = -1;
#endif

#if defined(__alpha)
static inline physaddr_t bar2phys (unsigned long addr)
{
    return virt_to_phys((void *) ioremap(addr, PAGE_SIZE));
}
#elif defined(__ia64)
static inline physaddr_t bar2phys (unsigned long addr)
{
    return ((addr) & ~__IA64_UNCACHED_OFFSET);
}
#elif defined(__powerpc64__)

#ifdef CONFIG_PPC_PSERIES
#include <asm/eeh.h>

static inline physaddr_t bar2phys (unsigned long addr)
{
    return eeh_token_to_phys (addr);
}
#endif

#else
static inline physaddr_t bar2phys (unsigned long addr)
{
    return (addr);
}
#endif

#ifndef pgprot_noncached
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
	unsigned long prot = pgprot_val(_prot);
#if defined(__powerpc__)
	prot |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#elif defined(__sparc__)
	prot &= ~(_PAGE_CACHE);
	prot |= _PAGE_IE;
#endif
	return __pgprot(prot);
}
#endif

#ifndef pgprot_writecombine
static inline pgprot_t pgprot_writecombine (pgprot_t _prot)
{
    unsigned long prot = pgprot_val(_prot);

#if defined (X86_FEATURE_PAT)
    if (pat_pteval != -1)
	prot = (prot & ~(_PAGE_PCD | _PAGE_PWT | _PAGE_PSE)) | pat_pteval;
#endif
    return __pgprot (prot);
}
#endif

#define ELAN4_DRIVER_VERSION		0x103           /* 16 bit value */

/*
 * Function prototypes.
 */
static int	   elan4_attach_device (int instance, struct pci_dev *pdev);
static void	   elan4_detach_device (ELAN4_DEV *dev);

static int 	   elan4_open (struct inode *inode, struct file *file);
static int	   elan4_release(struct inode *inode, struct file *file);
static int	   elan4_ioctl (struct inode *inode, struct file *file, 
				unsigned int cmd, unsigned long arg);
static int	   elan4_mmap (struct file *file, struct vm_area_struct *vm_area);

static irqreturn_t elan4_irq (int irq, void *arg, struct pt_regs *regs);

static void        elan4_shutdown_devices(int panicing);

static int      disabled;					/* bitmask of which devices not to start */
unsigned int	elan4_pll_cfg      = 0;
int		elan4_pll_div      = 31;			/* RevC PCB */
int		elan4_mod45disable = 0;
static int      optimise_pci_bus   = 1;				/* 0 => don't, 1 => if ok, 2 => always */
static int      default_features   = 0;				/* default values for dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] */
int             assfail_mode       = 0;

long long       sdram_cfg = SDRAM_STARTUP_VALUE;
static int      sdram_cfg_lo;
static int	sdram_cfg_hi;
int		sdram_bank_limit;

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("Elan 4 Device Driver");
MODULE_LICENSE("GPL");

module_param(elan4_debug, uint, 0);
module_param(elan4_debug_toconsole, uint, 0);
module_param(elan4_debug_tobuffer, uint, 0);
module_param(elan4_debug_mmu, uint, 0);
module_param(elan4_pll_cfg, uint, 0);
module_param(elan4_pll_div, uint, 0);
module_param(elan4_mod45disable, uint, 0);
module_param(optimise_pci_bus, uint, 0);
module_param(default_features, uint, 0);
module_param(assfail_mode, uint, 0);

module_param(disabled, uint, 0);
module_param(sdram_cfg_lo, uint, 0);
module_param(sdram_cfg_hi, uint, 0);
module_param(sdram_bank_limit, uint, 0);

module_param(elan4_hash_0_size_val, uint, 0);
module_param(elan4_hash_1_size_val, uint, 0);
module_param(elan4_ctxt_table_shift, uint, 0);
module_param(elan4_ln2_max_cqs, uint, 0);
module_param(elan4_dmaq_highpri_size, uint, 0);
module_param(elan4_threadq_highpri_size, uint, 0);
module_param(elan4_dmaq_lowpri_size, uint, 0);
module_param(elan4_threadq_lowpri_size, uint, 0);
module_param(elan4_interruptq_size, uint, 0);

module_param(elan4_mainint_punt_loops, uint, 0);
module_param(elan4_mainint_resched_ticks, uint, 0);
module_param(elan4_linkport_lock, uint, 0);
module_param(elan4_eccerr_recheck, uint, 0);

module_param(user_p2p_route_options, uint, 0);
module_param(user_bcast_route_options, uint, 0);
module_param(user_dproc_retry_count, uint, 0);
module_param(user_cproc_retry_count, uint, 0);
module_param(user_ioproc_enabled, uint, 0);
module_param(user_pagefault_enabled, uint, 0);

/*
 * Standard device entry points.
 */
static struct file_operations elan4_fops = {
    ioctl:   elan4_ioctl,
    mmap:    elan4_mmap,
    open:    elan4_open,
    release: elan4_release,
};

ELAN4_DEV *elan4_devices[ELAN4_MAX_CONTROLLER];

#if defined(CONFIG_DEVFS_FS)
static devfs_handle_t devfs_handle;
#endif


#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
static int
elan4_ioctl32_cmds[] =
{	/* /dev/elan/control */
    ELAN4IO_DEVINFO,
    ELAN4IO_GET_POSITION,
    ELAN4IO_SET_POSITION,
    ELAN4IO_GET_PARAMS,
    ELAN4IO_SET_PARAMS,

    /* /dev/elan4/user */
    ELAN4IO_POSITION,
    ELAN4IO_FREE,
    ELAN4IO_ATTACH,
    ELAN4IO_DETACH,
    ELAN4IO_BLOCK_INPUTTER,

    ELAN4IO_ADD_P2PVP,
    ELAN4IO_ADD_BCASTVP,
    ELAN4IO_REMOVEVP,
    ELAN4IO_SET_ROUTE,
    ELAN4IO_RESET_ROUTE,
    ELAN4IO_GET_ROUTE,
    ELAN4IO_CHECK_ROUTE,

    ELAN4IO_ALLOCCQ,
    ELAN4IO_FREECQ,
    ELAN4IO_SETPERM32,
    ELAN4IO_CLRPERM32,
    ELAN4IO_TRAPSIG,
    ELAN4IO_TRAPHANDLER32,
    ELAN4IO_REQUIRED_MAPPINGS,
	
    ELAN4IO_RESUME_EPROC_TRAP,
    ELAN4IO_RESUME_CPROC_TRAP,
    ELAN4IO_RESUME_DPROC_TRAP,
    ELAN4IO_RESUME_TPROC_TRAP,
    ELAN4IO_RESUME_IPROC_TRAP,

    ELAN4IO_FLUSH_ICACHE,

    ELAN4IO_STOP_CTXT,

    ELAN4IO_ALLOC_INTCOOKIE,
    ELAN4IO_FREE_INTCOOKIE,
    ELAN4IO_ARM_INTCOOKIE,
    ELAN4IO_WAIT_INTCOOKIE,

    ELAN4IO_ALLOC_TRAP_QUEUES,
    ELAN4IO_NETERR_MSG,
    ELAN4IO_NETERR_TIMER,
    ELAN4IO_NETERR_FIXUP,

    ELAN4IO_DUMPCQ32,
};

static int      elan4_ioctl32 (unsigned int fd, unsigned int cmd, 
			       unsigned long arg, struct file *file);
#endif

/*
 * Standard device entry points.
 */
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)

#include <linux/dump.h>

static int
elan4_dump_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if (event == DUMP_BEGIN)
	elan4_shutdown_devices (FALSE);

    return (NOTIFY_DONE);
}
static struct notifier_block elan4_dump_notifier = 
{
    notifier_call:	elan4_dump_event,
    priority:		0,
};

#endif

static int
elan4_reboot_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if ((event == SYS_RESTART || event == SYS_HALT || event == SYS_POWER_OFF))
	elan4_shutdown_devices (0);

    return (NOTIFY_DONE);
}

static struct notifier_block elan4_reboot_notifier = 
{
    notifier_call:	elan4_reboot_event,
    priority:		0,
};

#if !defined(NO_PANIC_NOTIFIER)
static int
elan4_panic_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    elan4_shutdown_devices (1);

    return (NOTIFY_DONE);
}

static struct notifier_block elan4_panic_notifier = 
{
    notifier_call:	elan4_panic_event,
    priority:		0,
};
#endif 

static int __init
elan4_init (void)
{
    int             err;
    struct pci_dev *pdev;
    int		    count;
#if defined(__ia64)
    int             seenRevA = 0;
#endif
    
    if ((err = register_chrdev (ELAN4_MAJOR, ELAN4_NAME, &elan4_fops)) < 0)
	return (err);

#if defined(CONFIG_DEVFS_FS)
    devfs_handle = devfs_mk_dir (NULL, "elan4", NULL);
#endif

    intcookie_init();
    elan4_debug_init();
    elan4_procfs_init();
    
#ifdef CONFIG_MPSAS
    sas_init();
#endif

    if (sdram_cfg_lo != 0 && sdram_cfg_hi != 0)
	sdram_cfg = (((unsigned long long) sdram_cfg_hi) << 32) | ((unsigned long long) sdram_cfg_lo);

    for (count = 0, pdev = NULL; (pdev = pci_find_device(PCI_VENDOR_ID_QUADRICS, PCI_DEVICE_ID_ELAN4, pdev)) != NULL ; count++)
    {
#if defined(__ia64)
	unsigned char revid;
	
	pci_read_config_byte (pdev, PCI_REVISION_ID, &revid);

	if (revid == PCI_REVISION_ID_ELAN4_REVA && seenRevA++ != 0 && pci_find_device (PCI_VENDOR_ID_HP, 0x122e, NULL))
	{
	    printk ("elan: only a single elan4a supported on rx2600\n");
	    continue;
	}
#endif

	if (count < ELAN4_MAX_CONTROLLER)
	    elan4_attach_device (count, pdev);
    }

    if (count >= ELAN4_MAX_CONTROLLER)
	printk ("elan: found %d elan4 devices - only support %d\n", count, ELAN4_MAX_CONTROLLER);

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
    lock_kernel();
    {
	extern int register_ioctl32_conversion(unsigned int cmd, int (*handler)(unsigned int, unsigned int, unsigned long, struct file *));
	register int i;
	for (i = 0; i < sizeof (elan4_ioctl32_cmds)/sizeof(elan4_ioctl32_cmds[0]); i++)
	    register_ioctl32_conversion (elan4_ioctl32_cmds[i], elan4_ioctl32);
    }
    unlock_kernel();
#endif

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
    register_dump_notifier (&elan4_dump_notifier);
#endif
    register_reboot_notifier (&elan4_reboot_notifier);

#if !defined(NO_PANIC_NOTIFIER)
    notifier_chain_register (&panic_notifier_list, &elan4_panic_notifier);
#endif

    return (0);
}

#ifdef MODULE
static void __exit
elan4_exit (void)
{
    int i;

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
    lock_kernel();
    {
	extern void unregister_ioctl32_conversion(unsigned int cmd);

	for (i = 0; i < sizeof (elan4_ioctl32_cmds)/sizeof(elan4_ioctl32_cmds[0]); i++)
	    unregister_ioctl32_conversion (elan4_ioctl32_cmds[i]);
    }
    unlock_kernel();
#endif

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
    unregister_dump_notifier (&elan4_dump_notifier);
#endif
    unregister_reboot_notifier (&elan4_reboot_notifier);

#if !defined(NO_PANIC_NOTIFIER)
    notifier_chain_unregister (&panic_notifier_list, &elan4_panic_notifier);
#endif

    for (i = 0; i < ELAN4_MAX_CONTROLLER; i++)
	if (elan4_devices[i] != NULL)
	    elan4_detach_device (elan4_devices[i]);
    
    elan4_procfs_fini();
    elan4_debug_fini();
    intcookie_fini();

#if defined(CONFIG_DEVFS_FS)
    devfs_unregister (devfs_handle);
#endif

    unregister_chrdev(ELAN4_MAJOR, ELAN4_NAME);
}

module_init (elan4_init);
module_exit (elan4_exit);

#else
__initcall (elan4_init);
#endif

/*
 * Minor numbers encoded as :
 *   [5:0]	device number
 *   [15:6]	function number
 */
#define ELAN4_DEVICE_MASK	0x3F
#define ELAN4_DEVICE(inode)	(MINOR((inode)->i_rdev) & ELAN4_DEVICE_MASK)

#define ELAN4_MINOR_CONTROL	0
#define ELAN4_MINOR_MEM		1
#define ELAN4_MINOR_USER	2

#define ELAN4_MINOR_SHIFT	6
#define ELAN4_MINOR(inode)	(MINOR((inode)->i_rdev) >> ELAN4_MINOR_SHIFT)

/*
 * Called by init_module() for each card discovered on PCI.
 */
static int
elan4_attach_device (int instance, struct pci_dev *pdev)
{
    ELAN4_DEV *dev;
    int res;

    KMEM_ALLOC (dev, ELAN4_DEV *, sizeof (ELAN4_DEV), 1);
    if ((dev == NULL))
	return (-ENOMEM);
    memset (dev, 0, sizeof (ELAN4_DEV));

    /* setup os dependent section of ELAN4_DEV */
    dev->dev_instance   = instance;
    dev->dev_osdep.pdev = pdev;

#if !defined(IOPROC_PATCH_APPLIED)
    printk ("elan%d: pinning down pages as no ioproc patch\n", dev->dev_instance);

    default_features |= ELAN4_FEATURE_NO_IOPROC | ELAN4_FEATURE_PIN_DOWN;

    /* Also change this flag so as to make the /proc entry consistent */
    user_ioproc_enabled = 0;
#endif

    /* initialise the devinfo */
    pci_read_config_word (dev->dev_osdep.pdev, PCI_VENDOR_ID,   &dev->dev_devinfo.dev_vendor_id);
    pci_read_config_word (dev->dev_osdep.pdev, PCI_DEVICE_ID,   &dev->dev_devinfo.dev_device_id);
    pci_read_config_byte (dev->dev_osdep.pdev, PCI_REVISION_ID, &dev->dev_devinfo.dev_revision_id);

    dev->dev_devinfo.dev_rail                                        = instance;
    dev->dev_devinfo.dev_driver_version                              = ELAN4_DRIVER_VERSION;
    dev->dev_devinfo.dev_num_down_links_value                        = 0;
    dev->dev_devinfo.dev_params_mask                                 = (1 << ELAN4_PARAM_DRIVER_FEATURES);
    dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES]  = default_features;

    dev->dev_position.pos_mode = ELAN_POS_UNKNOWN;

    dev->dev_regs_phys  = bar2phys (pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS));
    dev->dev_sdram_phys = bar2phys (pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM));

    /* initialise the data structures and map the device */
    if ((res = elan4_initialise_device (dev)) != 0)
    {
	kfree (dev);
	return res;
    }

#if defined(CONFIG_DEVFS_FS)
    {
	char name[16];
	
	sprintf (name, "control%d", dev->dev_instance);
	dev->dev_osdep.devfs_control = devfs_register(devfs_handle, name, DEVFS_FL_NONE, ELAN4_MAJOR,
						      dev->dev_instance | (ELAN4_MINOR_CONTROL << ELAN4_MINOR_SHIFT), S_IFCHR | S_IRUSR | S_IWUSR, 
						      &elan4_fops, NULL);
	sprintf (name, "sdram%d", dev->dev_instance);
	dev->dev_osdep.devfs_sdram =  devfs_register(devfs_handle, name, DEVFS_FL_NONE, ELAN4_MAJOR,
						     dev->dev_instance | (ELAN4_MINOR_MEM << ELAN4_MINOR_SHIFT), S_IFCHR | S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH,
						     &elan4_fops, NULL);
	sprintf (name, "user%d", dev->dev_instance);
	dev->dev_osdep.devfs_user =  devfs_register(devfs_handle, name, DEVFS_FL_NONE, ELAN4_MAJOR,
						    dev->dev_instance | (ELAN4_MINOR_USER << ELAN4_MINOR_SHIFT), S_IFCHR | S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH|S_IWOTH,
						    &elan4_fops, NULL);
    }
#endif

    /* add the procfs entry */
    elan4_procfs_device_init (dev);

    /* allow the device to be referenced now */
    elan4_devices[instance] = dev;

    if ((disabled & (1 << instance)) == 0)
    {
	if (elan4_start_device (dev) != 0)
	{
	    printk ("elan%d: auto-start of device failed\n", dev->dev_instance);

	    elan4_detach_device (dev);
	    return (-ENXIO);
	}
	
	dev->dev_state = ELAN4_STATE_STARTED;
    }

#if defined (__sparc)
    printk ("elan%d: at pci %s (irq = %s)\n", instance, pdev->slot_name, __irq_itoa(pdev->irq));
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    printk ("elan%d: at pci %s (irq = %d)\n", instance, pdev->slot_name, pdev->irq);
#else
    printk ("elan%d: at pci %s (irq = %d)\n", instance, pci_name (pdev), pdev->irq);
#endif
#endif

    return (0);
}

/*
 * Called by cleanup_module() for each board found on PCI.
 */
static void
elan4_detach_device (ELAN4_DEV *dev)
{
    /* stop the chip and free of resources */
    if (dev->dev_state == ELAN4_STATE_STARTED)
	elan4_stop_device (dev);
    
    elan4_devices[dev->dev_instance] = NULL;

#if defined(CONFIG_DEVFS_FS)
    devfs_unregister (dev->dev_osdep.devfs_control);
    devfs_unregister (dev->dev_osdep.devfs_sdram);
    devfs_unregister (dev->dev_osdep.devfs_user);
#endif

    /* remove the procfs entry */
    elan4_procfs_device_fini (dev);

    /* unmap the device and finalise the data structures */
    elan4_finalise_device (dev);
    
    KMEM_FREE (dev, sizeof(*dev));
}

/*
 * Maintain reference counts on the device
 */
ELAN4_DEV *
elan4_reference_device (int instance, int state)
{
    ELAN4_DEV *dev = elan4_devices[instance];

    if (dev == NULL)
	return (NULL);

    kmutex_lock (&dev->dev_lock);

    if ((dev->dev_state & state) == 0)
    {
	kmutex_unlock (&dev->dev_lock);
	return (NULL);
    }

    dev->dev_references++;
    kmutex_unlock (&dev->dev_lock);

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

#ifdef CONFIG_MPSAS
    sas_set_position(dev);
#endif

    return (dev);
}

void
elan4_dereference_device (ELAN4_DEV *dev)
{
    kmutex_lock (&dev->dev_lock);
    dev->dev_references--;
    kmutex_unlock (&dev->dev_lock);

#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
}

static void
elan4_shutdown_devices(int panicing)
{
    ELAN4_DEV *dev;
    unsigned long flags;
    register int i;

    local_irq_save (flags);
    for (i = 0; i < ELAN4_MAX_CONTROLLER; i++)
    {
	if ((dev = elan4_devices[i]) != NULL)
	{
	    printk(KERN_INFO "elan%d: forcing link into reset\n", dev->dev_instance);

	    /* set the inputters to discard everything */
	    if (! panicing) spin_lock (&dev->dev_haltop_lock);

	    if (dev->dev_discard_lowpri_count++ == 0)
		elan4_set_schedstatus (dev, 0);
	    if (dev->dev_discard_highpri_count++ == 0)
		elan4_set_schedstatus (dev, 0);

	    if (! panicing) spin_unlock (&dev->dev_haltop_lock);

	    /* ideally we'd like to halt all the outputters too,
	     * however this will prevent the kernel comms flushing
	     * to work correctly .....
	     */
	}
    }
    local_irq_restore (flags);
}

/*
 * /dev/elan4/controlX - control device
 *
 */
static int
control_open (struct inode *inode, struct file *file)
{
    ELAN4_DEV       *dev = elan4_reference_device (ELAN4_DEVICE(inode), ELAN4_STATE_STOPPED | ELAN4_STATE_STARTED);
    CONTROL_PRIVATE *pr;
    
    if (dev == NULL)
	return (-ENXIO);
    
    KMEM_ALLOC (pr, CONTROL_PRIVATE *, sizeof (CONTROL_PRIVATE), 1);
    if ((pr == NULL))
    {
	elan4_dereference_device (dev);
	
	return (-ENOMEM);
    }

    PRINTF (DBG_USER, DBG_FILE, "control_open: dev=%p pr=%p\n", dev, pr);

    pr->pr_dev           = dev;
    pr->pr_boundary_scan = 0;

    file->private_data = (void *) pr;

    return (0);
}

static int
control_release (struct inode *inode, struct file *file)
{
    CONTROL_PRIVATE *pr  = (CONTROL_PRIVATE *) file->private_data;
    ELAN4_DEV       *dev = pr->pr_dev;

    PRINTF (DBG_DEVICE, DBG_FILE, "control_release: pr=%p\n", pr);

    //if (pr->pr_boundary_scan)
    //    elan4_clear_boundary_scan (dev, pr);

    elan4_dereference_device (dev);

    KMEM_FREE (pr, sizeof(*pr));

    return (0);
}

static int
control_ioctl (struct inode *inode, struct file *file, 
		     unsigned int cmd, unsigned long arg)
{
    CONTROL_PRIVATE *pr  = (CONTROL_PRIVATE *) file->private_data;

    PRINTF (DBG_DEVICE, DBG_FILE, "control_ioctl: cmd=%x arg=%lx\n", cmd, arg);

    switch (cmd)
    {
    case ELAN4IO_DEVINFO:
	if (copy_to_user ((void *) arg, &pr->pr_dev->dev_devinfo, sizeof (ELAN_DEVINFO)))
	    return (-EFAULT);
	return (0);

    case ELAN4IO_GET_POSITION:
    {
	ELAN_POSITION pos;

	elan4_get_position (pr->pr_dev, &pos);

	if (copy_to_user ((void *) arg, &pos, sizeof (ELAN_POSITION)))
	    return (-EFAULT);

	return (0);
    }

    case ELAN4IO_SET_POSITION:
    {
	ELAN_POSITION pos;

	if (copy_from_user (&pos, (void *) arg, sizeof (ELAN_POSITION)))
	    return (-EFAULT);
	
	return (elan4_set_position (pr->pr_dev, &pos));
    }

    case ELAN4IO_OLD_GET_PARAMS:
    {
	ELAN_PARAMS params;
	unsigned short mask;

	elan4_get_params (pr->pr_dev, &params, &mask);

	if (copy_to_user ((void *) arg, &params, sizeof (ELAN_PARAMS)))
	    return (-EFAULT);

	return (0);
    }

    case ELAN4IO_OLD_SET_PARAMS:
    {
	ELAN_PARAMS params;

	if (copy_from_user (&params, (void *) arg, sizeof (ELAN_PARAMS)))
	    return (-EFAULT);
	
	elan4_set_params (pr->pr_dev, &params, 3);
	
	return (0);
    }

    case ELAN4IO_SET_PARAMS:
    {
	ELAN4IO_PARAMS_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_PARAMS_STRUCT)))
	    return (-EFAULT);
	
	elan4_set_params (pr->pr_dev, &args.p_params, args.p_mask);
	
	return (0);
    }

    case ELAN4IO_GET_PARAMS:
    {
	ELAN4IO_PARAMS_STRUCT args;

	elan4_get_params (pr->pr_dev, &args.p_params, &args.p_mask);

	if (copy_to_user ((void *) arg, &args, sizeof (ELAN_PARAMS)))
	    return (-EFAULT);

	return (0);
    }
    }

    return (-EINVAL);
}

static int
control_mmap (struct file *file, struct vm_area_struct *vma)
{
    CONTROL_PRIVATE *pr  = (CONTROL_PRIVATE *) file->private_data;
    unsigned 	     bar = OFF_TO_BAR (vma->vm_pgoff << PAGE_SHIFT);
    unsigned long    off = OFF_TO_OFFSET (vma->vm_pgoff << PAGE_SHIFT);
    long	     len = vma->vm_end - vma->vm_start;

    PRINTF (DBG_USER, DBG_FILE, "control_mmap: pr=%p bar=%x off=%x\n", pr, bar, off);

    /* check bar number and translate the standard psuedo bars */
    switch (bar)
    {
    case ELAN4_BAR_SDRAM:
    case ELAN4_BAR_REGISTERS:
	break;

    default:
	return (-EINVAL);
    }

    if (off < 0 || (off + len) > pci_resource_len (pr->pr_dev->dev_osdep.pdev, bar))
	return (-EINVAL);

    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

    if (__io_remap_page_range (vma->vm_start, pci_resource_start (pr->pr_dev->dev_osdep.pdev, bar) + off, len, vma->vm_page_prot))
	return (-EAGAIN);

    return (0);
}

/*
 * /dev/elan4/sdramX - sdram access device
 */
static void 
mem_freepage (MEM_PRIVATE *pr, MEM_PAGE *pg)
{
    PRINTF (DBG_USER, DBG_MEM, "mem_freepage: pr=%p pgoff=%lx pg=%p ref=%d\n", pr, pg->pg_pgoff, pg, pg->pg_ref);

    elan4_sdram_free (pr->pr_dev, pg->pg_addr, SDRAM_PAGE_SIZE);

    KMEM_FREE(pg, sizeof(*pg));
}

static MEM_PAGE *
mem_getpage (MEM_PRIVATE *pr, unsigned long pgoff)
{
    int       hashval = MEM_HASH (pgoff);
    MEM_PAGE *npg = NULL;
    MEM_PAGE *pg;

    ASSERT ((pgoff & SDRAM_PGOFF_OFFSET) == 0);

    PRINTF (DBG_USER, DBG_MEM, "mem_getpage: pr=%p pgoff=%lx\n", pr, pgoff);
    
 again:
    spin_lock (&pr->pr_lock);
    for (pg = pr->pr_pages[hashval]; pg; pg = pg->pg_next)
	if (pg->pg_pgoff == pgoff)
	    break;
    
    if (pg != NULL)
    {
	PRINTF (DBG_USER, DBG_MEM, "mem_getpage: pr=%p pgoff=%lx -> found %p addr=%x\n", pr, pgoff, pg, pg->pg_addr);

	pg->pg_ref++;
	spin_unlock (&pr->pr_lock);

	if (npg != NULL)					/* we'd raced and someone else had created */
	    mem_freepage (pr, npg);				/* this page - so free of our new one*/
	return (pg);
    }
    
    if (npg != NULL)						/* didn't find the page, so inset the */
    {								/* new one we've just created */
	npg->pg_next = pr->pr_pages[hashval];
	pr->pr_pages[hashval] = npg;
	
	spin_unlock (&pr->pr_lock);
	return (npg);
    }
    
    spin_unlock (&pr->pr_lock);					/* drop spinlock before creating a new page */
    
    KMEM_ALLOC (npg, MEM_PAGE *, sizeof (MEM_PAGE), 1);
    if ((npg == NULL))
	return (NULL);
    
    if ((npg->pg_addr = elan4_sdram_alloc (pr->pr_dev, SDRAM_PAGE_SIZE)) == 0)
    {
	KMEM_FREE(npg, sizeof(*npg));
	return (NULL);
    }

#ifndef CONFIG_MPSAS
    /* zero the page before returning it to the user */
    elan4_sdram_zeroq_sdram (pr->pr_dev, npg->pg_addr, SDRAM_PAGE_SIZE);
#endif
    
    npg->pg_pgoff = pgoff;
    npg->pg_ref   = 1;
    
    /* created a new page - so have to rescan before inserting it */
    goto again;
}

static void
mem_droppage (MEM_PRIVATE *pr, unsigned long pgoff, int dontfree)
{
    MEM_PAGE **ppg;
    MEM_PAGE  *pg;

    spin_lock (&pr->pr_lock);
    for (ppg = &pr->pr_pages[MEM_HASH(pgoff)]; *ppg; ppg = &(*ppg)->pg_next)
	if ((*ppg)->pg_pgoff == pgoff)
	    break;

    pg = *ppg;

    ASSERT (*ppg != NULL);

    PRINTF (DBG_USER, DBG_MEM, "mem_droppage: pr=%p pgoff=%lx pg=%p ref=%d dontfree=%d\n", pr, pgoff, (*ppg), (*ppg)->pg_ref, dontfree);

    if (--pg->pg_ref == 0 && !dontfree)
    {
	*ppg = pg->pg_next;

	mem_freepage (pr, pg);
    }

    spin_unlock (&pr->pr_lock);
}

static int
mem_open (struct inode *inode, struct file *file)
{
    ELAN4_DEV   *dev = elan4_reference_device (ELAN4_DEVICE(inode), ELAN4_STATE_STARTED);
    MEM_PRIVATE *pr;
    register int i;

    if (dev == NULL)
	return (-ENXIO);

    KMEM_ALLOC (pr, MEM_PRIVATE *, sizeof (MEM_PRIVATE), 1);
    if ((pr == NULL))
    {
	elan4_dereference_device (dev);
	return (-ENOMEM);
    }

    spin_lock_init (&pr->pr_lock);
    pr->pr_dev = dev;
    for (i = 0; i < MEM_HASH_SIZE; i++)
	pr->pr_pages[i] = NULL;

    file->private_data = (void *) pr;
    
    return (0);
}

static int
mem_release (struct inode *node, struct file *file)
{
    MEM_PRIVATE *pr = (MEM_PRIVATE *) file->private_data;
    MEM_PAGE    *pg, *next;
    int          i;

    /* free off any pages that we'd allocated */
    spin_lock (&pr->pr_lock);
    for (i = 0; i < MEM_HASH_SIZE; i++)
    {
	for (pg = pr->pr_pages[i]; pg; pg = next)
	{
	    next = pg->pg_next;
	    mem_freepage (pr, pg);
	}
    }
    spin_unlock (&pr->pr_lock);

    elan4_dereference_device (pr->pr_dev);
    KMEM_FREE(pr, sizeof(*pr));

    return (0);
}

static int
mem_ioctl (struct inode *inode, struct file *file, 
		  unsigned int cmd, unsigned long arg)
{
    return (-EINVAL);
}

static void 
mem_vma_open (struct vm_area_struct *vma)
{
    MEM_PRIVATE   *pr = (MEM_PRIVATE *) vma->vm_private_data;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_USER, DBG_MEM, "mem_vma_open: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);
    
    for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	mem_getpage (pr, pgoff & SDRAM_PGOFF_MASK);
}

static void
mem_vma_close (struct vm_area_struct *vma)
{
    MEM_PRIVATE  *pr  = (MEM_PRIVATE *) vma->vm_private_data;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_USER, DBG_MEM, "mem_vma_close: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    /* NOTE: the call to close may not have the same vm_start/vm_end values as 
     *       were passed into mmap()/open() - since if an partial unmap had occurred
     *       then the vma could have been shrunk or even split.
     *
     *       if a the vma is split then an vma_open() will be called for the top
     *       portion - thus causing the reference counts to become incorrect.
     *
     * We drop the reference to any pages we're notified about - so they get freed
     * earlier than when the device is finally released.
     */
    for (pgoff = vma->vm_pgoff, addr = vma->vm_start; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	mem_droppage (pr, pgoff & SDRAM_PGOFF_MASK, 0);
}

struct vm_operations_struct mem_vm_ops = {
    open:		mem_vma_open,
    close:		mem_vma_close,
};

static int
mem_mmap (struct file *file, struct vm_area_struct *vma)
{
    MEM_PRIVATE  *pr = (MEM_PRIVATE *) file->private_data;
    MEM_PAGE     *pg;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (DBG_USER, DBG_MEM, "mem_mmap: vma=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma, vma->vm_start, vma->vm_end, vma->vm_pgoff, file);

    /* Don't allow these pages to be swapped out of dumped */
    vma->vm_flags |= (VM_RESERVED | VM_IO);

    vma->vm_ops          = &mem_vm_ops;
    vma->vm_file         = file;
    vma->vm_private_data = (void *) pr;

    for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
    {
	if ((pg = mem_getpage (pr, pgoff & SDRAM_PGOFF_MASK)) == NULL)
	    goto failed;

	PRINTF (DBG_USER, DBG_MEM, "mem_mmap: addr %lx -> pg=%p sdram=%x+%x bar=%lx\n",
		addr, pg, pg->pg_addr, (pgoff & SDRAM_PGOFF_OFFSET) * PAGE_SIZE,
		pci_resource_start (pr->pr_dev->dev_osdep.pdev, ELAN4_BAR_SDRAM));

	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	if (! (pr->pr_dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_WRITE_COMBINE))
	    vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	if (__io_remap_page_range (addr, pci_resource_start (pr->pr_dev->dev_osdep.pdev, ELAN4_BAR_SDRAM) +
				 pg->pg_addr + (pgoff & SDRAM_PGOFF_OFFSET) * PAGE_SIZE, PAGE_SIZE, vma->vm_page_prot))
	{
	    mem_droppage (pr, pgoff & SDRAM_PGOFF_MASK, 0); /* drop our reference to this page */
	    goto failed;
	}

#if defined(conditional_schedule)
	conditional_schedule();
#endif
    }

    return (0);

 failed:
    /* free of any pages we've already allocated/referenced */
    while (pgoff-- > vma->vm_pgoff)
	mem_droppage (pr, pgoff & SDRAM_PGOFF_MASK, 0);

    return (-ENOMEM);
}

int
mem_pteload (struct vm_area_struct *vma, unsigned long maddr, USER_CTXT *uctx, E4_Addr eaddr, int perm)
{
    MEM_PRIVATE  *pr    = (MEM_PRIVATE *) vma->vm_private_data;
    ELAN4_DEV    *dev   = uctx->uctx_ctxt.ctxt_dev;
    unsigned long pgoff = vma->vm_pgoff + ((maddr - vma->vm_start) >> PAGE_SHIFT);
    sdramaddr_t   pgaddr;
    MEM_PAGE      *pg;
    register int  i, res;

    if (pr->pr_dev != dev)
	return -EINVAL;

    if ((pg = mem_getpage (pr, pgoff & SDRAM_PGOFF_MASK)) == NULL)
	return -ENOMEM;
    
    pgaddr = pg->pg_addr + ((pgoff & SDRAM_PGOFF_OFFSET) << PAGE_SHIFT);

    if (! elan4mmu_sdram_aliascheck (&uctx->uctx_ctxt, eaddr, pgaddr))
	return -EINVAL;
	    
    for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
    {
	E4_uint64 newpte = ((pgaddr + i) >> PTE_PADDR_SHIFT) | PTE_SetPerm (perm);
	
	if ((res = elan4mmu_pteload (&uctx->uctx_ctxt, 0, eaddr + i, HE_TYPE_SDRAM, newpte)) < 0)
	    return res;
    }
    
    mem_droppage (pr, pgoff & SDRAM_PGOFF_MASK, 0);

    return 0;
}


/*
 * /dev/elan4/userX - control device
 *
 */
static inline void
user_private_free (USER_PRIVATE *pr)
{
    ELAN4_DEV *dev = pr->pr_uctx->uctx_ctxt.ctxt_dev;

    ASSERT (atomic_read (&pr->pr_ref) == 0);

    user_free (pr->pr_uctx);
    KMEM_FREE(pr, sizeof(*pr));

    elan4_dereference_device (dev);
}

#if defined(IOPROC_PATCH_APPLIED)
static void
user_ioproc_release (void *arg, struct mm_struct *mm)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_release: ref=%d\n", atomic_read (&pr->pr_ref));

    elan4mmu_invalidate_ctxt (&pr->pr_uctx->uctx_ctxt);

    pr->pr_mm = NULL;

    if (atomic_dec_and_test (&pr->pr_ref))
	user_private_free (pr);
}

/*
 * On 2.4 kernels we get passed a mm_struct, whereas on 2.6 kernels
 * we get the vma which is more usefull
 */
#if defined(IOPROC_MM_STRUCT_ARG)
static void
user_ioproc_sync_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_sync_range: start=%lx end=%lx\n", start, end);

    /* XXXX: this is intended to sync the modified bit from our page tables,
     *       into the main cpu's modified bits - however since we do not
     *       syncronize our modified bit on a ioproc_invalidate_page() call,
     *       then it could get lost if we modify the page after the last
     *       modification and writepage has occurred. Hence we invalidate
     *       all translations and allow it to refault.
     */

    user_unload_main (pr->pr_uctx, start, end - start);
}

static void
user_ioproc_invalidate_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_invalidate_range: start=%lx end=%lx\n", start, end);

    user_unload_main (pr->pr_uctx, start, end - start);
}

static void
user_ioproc_update_range (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;
    struct vm_area_struct *vma;

    if (pr->pr_uctx->uctx_ctxt.ctxt_dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_IOPROC_UPDATE)
	return;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_update_range: start=%lx end=%lx\n", start, end);

    vma = find_vma_intersection (mm, start, end);

    user_update_main (pr->pr_uctx, mm, vma, start, end - start);
}

static void
user_ioproc_change_protection (void *arg, struct mm_struct *mm, unsigned long start, unsigned long end, pgprot_t newprot)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_change_protection: start=%lx end=%lx\n", start, end);

    user_unload_main (pr->pr_uctx, start, end - start);
}


#else

static void
user_ioproc_sync_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_sync_range: start=%lx end=%lx\n", start, end);

    /* XXXX: this is intended to sync the modified bit from our page tables,
     *       into the main cpu's modified bits - however since we do not
     *       syncronize our modified bit on a ioproc_invalidate_page() call,
     *       then it could get lost if we modify the page after the last
     *       modification and writepage has occurred. Hence we invalidate
     *       all translations and allow it to refault.
     */

    user_unload_main (pr->pr_uctx, start, end - start);
}

static void
user_ioproc_invalidate_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_invalidate_range: start=%lx end=%lx\n", start, end);

    user_unload_main (pr->pr_uctx, start, end - start);
}

static void
user_ioproc_update_range (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    if (pr->pr_uctx->uctx_ctxt.ctxt_dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_IOPROC_UPDATE)
	return;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_update_range: start=%lx end=%lx\n", start, end);

    user_update_main (pr->pr_uctx, vma->vm_mm, vma, start, end - start);
}

static void
user_ioproc_change_protection (void *arg, struct vm_area_struct *vma, unsigned long start, unsigned long end, pgprot_t newprot)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_change_protection: start=%lx end=%lx\n", start, end);

    user_unload_main (pr->pr_uctx, start, end - start);
}
#endif /* defined(IOPROC_NO_VMA_RANGE) */

static void
user_ioproc_sync_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_sync_page: addr=%lx\n", addr);

    user_unload_main (pr->pr_uctx, addr & PAGE_MASK, PAGE_SIZE);
}

static void
user_ioproc_invalidate_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_invalidate_page: addr=%lx\n", addr);

    user_unload_main (pr->pr_uctx, addr & PAGE_MASK, PAGE_SIZE);
}

static void
user_ioproc_update_page (void *arg, struct vm_area_struct *vma, unsigned long addr)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) arg;

    if (pr->pr_uctx->uctx_ctxt.ctxt_dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_IOPROC_UPDATE)
	return;

    PRINTF (pr->pr_uctx, DBG_IOPROC, "user_ioproc_update_page: addr=%lx\n", addr);

    user_update_main (pr->pr_uctx, vma->vm_mm, vma, addr & PAGE_MASK, PAGE_SIZE);
}
#endif /* defined(IOPROC_PATCH_APPLIED) */

static int
user_open (struct inode *inode, struct file *file)
{
    ELAN4_DEV    *dev;
    USER_PRIVATE *pr;
    USER_CTXT    *uctx;
    
    PRINTF (DBG_USER, DBG_FILE, "user_open: mm=%p users=%d count=%d\n", current->mm,
	    atomic_read (&current->mm->mm_users), atomic_read (&current->mm->mm_count));

    if ((dev = elan4_reference_device (ELAN4_DEVICE(inode), ELAN4_STATE_STARTED)) == NULL)
	return (-ENXIO);
    
    KMEM_ALLOC (pr, USER_PRIVATE *, sizeof (USER_PRIVATE), 1);
    if ((pr == NULL))
    {
	elan4_dereference_device (dev);
	return (-ENOMEM);
    }

    uctx = user_alloc (dev);

    if (IS_ERR(uctx))
    {
	elan4_dereference_device (dev);
	KMEM_FREE(pr, sizeof(*pr));

	return PTR_ERR(uctx);
    }

    /* initialise refcnt to 1 - one for "file" */
    atomic_set (&pr->pr_ref, 1);

    pr->pr_uctx = uctx;
    pr->pr_mm   = current->mm;

#if defined(IOPROC_PATCH_APPLIED)
    if (! (uctx->uctx_ctxt.ctxt_features & ELAN4_FEATURE_NO_IOPROC))
    {
	/* register a ioproc callback to notify us of translation changes */
	pr->pr_ioproc.arg               = (void *) pr;
	pr->pr_ioproc.release           = user_ioproc_release;
	pr->pr_ioproc.sync_range        = user_ioproc_sync_range;
	pr->pr_ioproc.invalidate_range  = user_ioproc_invalidate_range;
	pr->pr_ioproc.update_range      = user_ioproc_update_range;
	pr->pr_ioproc.change_protection = user_ioproc_change_protection;
	pr->pr_ioproc.sync_page         = user_ioproc_sync_page;
	pr->pr_ioproc.invalidate_page   = user_ioproc_invalidate_page;
	pr->pr_ioproc.update_page       = user_ioproc_update_page;
	
	/* add an extra reference for the ioproc ops */
	atomic_inc (&pr->pr_ref);
	
	spin_lock (&current->mm->page_table_lock);
	ioproc_register_ops (current->mm, &pr->pr_ioproc);
	spin_unlock (&current->mm->page_table_lock);
    }
#endif

    file->private_data = (void *) pr;

    return (0);
}

static int
user_release (struct inode *inode, struct file *file)
{
    USER_PRIVATE *pr = (USER_PRIVATE *) file->private_data;

    PRINTF (pr->pr_uctx, DBG_FILE, "user_release: ref=%d\n", atomic_read (&pr->pr_ref));

    if (atomic_dec_and_test (&pr->pr_ref))
	user_private_free (pr);

    return (0);
}

static int
user_ioctl (struct inode *inode, struct file *file, 
	    unsigned int cmd, unsigned long arg)
{
    USER_PRIVATE *pr   = (USER_PRIVATE *) file->private_data;
    USER_CTXT    *uctx = pr->pr_uctx;
    int           res  = 0;

    PRINTF (uctx, DBG_FILE, "user_ioctl: cmd=%x arg=%lx\n", cmd, arg);

    if (current->mm != pr->pr_mm)
	return (-EINVAL);
    
    switch (cmd)
    {
    case ELAN4IO_DEVINFO:
    {
	ELAN_DEVINFO devinfo = uctx->uctx_ctxt.ctxt_dev->dev_devinfo;

	if ((devinfo.dev_params_mask & (1 << ELAN4_PARAM_DRIVER_FEATURES)) != 0)
	    devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] = uctx->uctx_ctxt.ctxt_features;

	if (copy_to_user ((void *) arg, &devinfo, sizeof (ELAN_DEVINFO)))
	    return (-EFAULT);
	return (0);
    }

    case ELAN4IO_POSITION:
    {
	ELAN4_DEV *dev = uctx->uctx_ctxt.ctxt_dev;

	if (copy_to_user ((void *) arg, &dev->dev_position, sizeof (ELAN_POSITION)))
	    return (-EFAULT);
	return (0);
    }

    case ELAN4IO_FREE:
#if defined(IOPROC_PATCH_APPLIED)
	if (! (uctx->uctx_ctxt.ctxt_features & ELAN4_FEATURE_NO_IOPROC))
	{
	    spin_lock (&current->mm->page_table_lock);
	    if (pr->pr_mm != current->mm)
		spin_unlock (&current->mm->page_table_lock);
	    else
	    {
		ioproc_unregister_ops (current->mm, &pr->pr_ioproc);
		spin_unlock (&current->mm->page_table_lock);
		
		user_ioproc_release (pr, current->mm);
	    }
	}
#endif
	return (0);

    case ELAN4IO_ATTACH:
    {
	ELAN_CAPABILITY *cap;

	KMEM_ALLOC (cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), 1);
	if ((cap == NULL))
	    return (-ENOMEM);

	if (copy_from_user (cap, (void *) arg, sizeof (ELAN_CAPABILITY)))
	    res = -EFAULT;
	else if ((res = user_attach (uctx, cap)) == 0 && 
		 copy_to_user ((void *) arg, cap, sizeof (ELAN_CAPABILITY)))
	{
	    user_detach (uctx, cap);
	    res = -EFAULT;
	}

	KMEM_FREE(cap, sizeof(*cap));
	return (res);
    }

    case ELAN4IO_DETACH:
    {
	ELAN_CAPABILITY *cap;

	KMEM_ALLOC (cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), 1);
	if ((cap == NULL))
	    return (-ENOMEM);

	if (copy_from_user (cap, (void *) arg, sizeof (ELAN_CAPABILITY)))
	    res = -EFAULT;
	else
	    user_detach (uctx, cap);

	KMEM_FREE(cap, sizeof(*cap));
	return (res);
    }

    case ELAN4IO_BLOCK_INPUTTER:
	user_block_inputter (uctx, arg);
	return (0);

    case ELAN4IO_ADD_P2PVP:
    {
	ELAN4IO_ADD_P2PVP_STRUCT *args;
	
	KMEM_ALLOC (args, ELAN4IO_ADD_P2PVP_STRUCT *, sizeof (ELAN4IO_ADD_P2PVP_STRUCT), 1);
	if ((args == NULL))
	    return (-ENOMEM);

	if (copy_from_user (args, (void *) arg, sizeof (ELAN4IO_ADD_P2PVP_STRUCT)))
	    res = -EFAULT;
	else 
	    res = user_add_p2pvp (uctx, args->vp_process, &args->vp_capability);
	
	KMEM_FREE(args, sizeof(*args));
	return (res);
    }

    case ELAN4IO_ADD_BCASTVP:
    {
	ELAN4IO_ADD_BCASTVP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ADD_BCASTVP_STRUCT)))
	    return (-EFAULT);

	return (user_add_bcastvp (uctx, args.vp_process, args.vp_lowvp, args.vp_highvp));
    }

    case ELAN4IO_REMOVEVP:
	return (user_removevp (uctx, arg));

    case ELAN4IO_SET_ROUTE:
    {
	ELAN4IO_ROUTE_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    return (-EFAULT);

	return (user_set_route (uctx, args.rt_process, &args.rt_route));
    }

    case ELAN4IO_RESET_ROUTE:
    {
	ELAN4IO_ROUTE_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    return (-EFAULT);

	return (user_reset_route (uctx, args.rt_process));
    }

    case ELAN4IO_GET_ROUTE:
    {
	ELAN4IO_ROUTE_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    return (-EFAULT);

	if ((res = user_get_route (uctx, args.rt_process, &args.rt_route)) == 0 &&
	    copy_to_user ((void *) arg, &args, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    res = -EFAULT;

	return (res);
    }

    case ELAN4IO_CHECK_ROUTE:
    {
	ELAN4IO_ROUTE_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    return (-EFAULT);

	if ((res = user_check_route (uctx, args.rt_process, &args.rt_route, &args.rt_error)) == 0 &&
	    copy_to_user ((void *) arg, &args, sizeof (ELAN4IO_ROUTE_STRUCT)))
	    res = -EFAULT;

	return (res);
    }
	
    case ELAN4IO_ALLOCCQ:
    {
	ELAN4IO_ALLOCCQ_STRUCT args;
	USER_CQ		     *ucq;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ALLOCCQ_STRUCT)))
	    return (-EFAULT);
	
	ucq = user_alloccq (uctx, args.cq_size & CQ_SizeMask, args.cq_perm & CQ_PermissionMask,
			    (args.cq_type & ELAN4IO_CQ_TYPE_REORDER) ? UCQ_REORDER : 0);
	if (IS_ERR (ucq))
	    return PTR_ERR (ucq);
	
	args.cq_indx = elan4_cq2idx (ucq->ucq_cq);
	
	if (copy_to_user ((void *) arg, &args, sizeof (ELAN4IO_ALLOCCQ_STRUCT)))
	{
	    user_dropcq (uctx, ucq);
	    return (-EFAULT);
	}
	
	/* don't drop the reference on the cq until the context is freed,
	 * or the caller explicitly frees the cq */
	return (0);
    }
	
    case ELAN4IO_FREECQ:
    {
	USER_CQ *ucq;
	unsigned indx;

	if (copy_from_user (&indx, (void *) arg, sizeof (unsigned)))
	    return (-EFAULT);

	if ((ucq = user_findcq (uctx, indx)) == NULL)		/* can't free unallocated cq */
	    return (-EINVAL);
	
	user_dropcq (uctx, ucq);				/* drop the reference we've just taken */

	if ((ucq->ucq_flags & UCQ_SYSTEM))			/* can't free device driver cq */
	    return (-EINVAL);

	user_dropcq (uctx, ucq);				/* and the one held from the alloccq call */

	return (0);
    }

    case ELAN4IO_DUMPCQ:
    {
	ELAN4IO_DUMPCQ_STRUCT args;
	ELAN4_DEV *dev = uctx->uctx_ctxt.ctxt_dev;
	USER_CQ *ucq;
	void *buf;
	int i;
	
	if (copy_from_user (&args, (void *) arg, sizeof(ELAN4IO_DUMPCQ_STRUCT)))
	    return (-EFAULT);

	if ((ucq = user_findcq (uctx, args.cq_indx)) == NULL)
	    return (-EINVAL);

	if (args.bufsize)
	{
	    E4_uint32 usedBufSize = min(args.cq_size, args.bufsize);

	    KMEM_ALLOC (buf, void *, usedBufSize, 1);

	    if (buf == NULL)
		return (-ENOMEM);

	    for (i=0; i<usedBufSize; i+=sizeof(int))
		((int *)buf)[i/sizeof(int)] = elan4_sdram_readl(dev, ucq->ucq_cq->cq_space + i);

	    if (copy_to_user((void *)args.buffer, buf, usedBufSize))
	    {
		KMEM_FREE(buf, args.bufsize);
		return (-EFAULT);
	    }
	    KMEM_FREE(buf, usedBufSize);
	    args.bufsize = usedBufSize;
	}

	args.cq_size = CQ_Size(ucq->ucq_cq->cq_size);
	args.cq_space = ucq->ucq_cq->cq_space;


	if (copy_to_user((void *)arg, &args, sizeof(ELAN4IO_DUMPCQ_STRUCT)))
	{
	    return (-EFAULT);
	}
	
	user_dropcq (uctx, ucq); /* drop the reference we've just taken */

	return (0);
    }

    case ELAN4IO_SETPERM:
    {
	ELAN4IO_PERM_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_PERM_STRUCT)))
	    return (-EFAULT);

	return (user_setperm (uctx, args.ps_maddr, args.ps_eaddr, args.ps_len, args.ps_perm));
    }

    case ELAN4IO_CLRPERM:
    {
	ELAN4IO_PERM_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_PERM_STRUCT)))
	    return (-EFAULT);

	user_clrperm (uctx, args.ps_eaddr, args.ps_len);
	return (0);
    }
    
    case ELAN4IO_TRAPSIG:
    {
	ELAN4IO_TRAPSIG_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_TRAPSIG_STRUCT)))
	    return (-EFAULT);

	pr->pr_uctx->uctx_trap_pid   = current->pid;
	pr->pr_uctx->uctx_trap_signo = args.ts_signo;
	
	return (0);
    }
    
    case ELAN4IO_TRAPHANDLER:
    {
	ELAN4IO_TRAPHANDLER_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_TRAPHANDLER_STRUCT)))
	    return (-EFAULT);

	return (user_trap_handler (pr->pr_uctx, (ELAN4_USER_TRAP *)args.th_trapp, args.th_nticks));
    }

    case ELAN4IO_REQUIRED_MAPPINGS:
    {
	ELAN4IO_REQUIRED_MAPPINGS_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_REQUIRED_MAPPINGS_STRUCT)))
	    return (-EFAULT);

	pr->pr_uctx->uctx_upage_addr    = args.rm_upage_addr;
	pr->pr_uctx->uctx_trestart_addr = args.rm_trestart_addr;

	return (0);
    }

    case ELAN4IO_ALLOC_TRAP_QUEUES:
    {
	ELAN4IO_ALLOC_TRAP_QUEUES_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_ALLOC_TRAP_QUEUES_STRUCT)))
	    return (-EFAULT);

	return (user_alloc_trap_queues (uctx, args.tq_ndproc_traps, args.tq_neproc_traps, 
					args.tq_ntproc_traps, args.tq_nthreads, args.tq_ndmas));
    }

    case ELAN4IO_RESUME_EPROC_TRAP:
    {
	ELAN4IO_RESUME_EPROC_TRAP_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_RESUME_EPROC_TRAP_STRUCT)))
	    return (-EFAULT);

	return (user_resume_eproc_trap (pr->pr_uctx, args.rs_addr));
    }

    case ELAN4IO_RESUME_CPROC_TRAP:
    {
	ELAN4IO_RESUME_CPROC_TRAP_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_RESUME_CPROC_TRAP_STRUCT)))
	    return (-EFAULT);

	return (user_resume_cproc_trap (pr->pr_uctx, args.rs_indx));
    }

    case ELAN4IO_RESUME_DPROC_TRAP:
    {
	ELAN4IO_RESUME_DPROC_TRAP_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_RESUME_DPROC_TRAP_STRUCT)))
	    return (-EFAULT);

	return (user_resume_dproc_trap (pr->pr_uctx, &args.rs_desc));
    }

    case ELAN4IO_RESUME_TPROC_TRAP:
    {
	ELAN4IO_RESUME_TPROC_TRAP_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_RESUME_TPROC_TRAP_STRUCT)))
	    return (-EFAULT);

	return (user_resume_tproc_trap (pr->pr_uctx, &args.rs_regs));
    }

    case ELAN4IO_RESUME_IPROC_TRAP:
    {
	ELAN4IO_RESUME_IPROC_TRAP_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_RESUME_IPROC_TRAP_STRUCT)))
	    return (-EFAULT);

	return (user_resume_iproc_trap (pr->pr_uctx, args.rs_channel, args.rs_trans, 
					&args.rs_header, &args.rs_data));
    }

    case ELAN4IO_FLUSH_ICACHE:
	elan4_flush_icache (&uctx->uctx_ctxt);
	return (0);

    case ELAN4IO_STOP_CTXT:
	if (arg)
	    user_swapout (uctx, UCTX_USER_STOPPED);
	else
	    user_swapin (uctx, UCTX_USER_STOPPED);
	return (0);

    case ELAN4IO_ALLOC_INTCOOKIE_TABLE:
    {
	ELAN_CAPABILITY *cap;
	INTCOOKIE_TABLE *tbl;

	KMEM_ALLOC (cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), 1);
	if ((cap == NULL))
	    return (-ENOMEM);

	if (copy_from_user (cap, (void *) arg, sizeof (ELAN_CAPABILITY)))
	    res = -EFAULT;
	else
	{
	    tbl = intcookie_alloc_table(cap);

	    if (tbl == NULL)
		res = -ENOMEM;
	    else
	    {
		/* Install the intcookie table we've just created */
		spin_lock (&uctx->uctx_spinlock);
		if (uctx->uctx_intcookie_table != NULL)
		    res = -EBUSY;
		else
		    uctx->uctx_intcookie_table = tbl;
		spin_unlock (&uctx->uctx_spinlock);
		
		/* drop the table we created if there already was one */
		if (res != 0)
		    intcookie_free_table (tbl);
	    }
	}

	KMEM_FREE(cap, sizeof(*cap));

	return (res);
    }

    case ELAN4IO_FREE_INTCOOKIE_TABLE:
    {
	INTCOOKIE_TABLE *tbl;

	spin_lock (&uctx->uctx_spinlock);
	tbl = uctx->uctx_intcookie_table;
	uctx->uctx_intcookie_table = NULL;
	spin_unlock (&uctx->uctx_spinlock);

	if (tbl != NULL)
	    intcookie_free_table (tbl);

	return (tbl == NULL ? -EINVAL : 0);
    }

    case ELAN4IO_ALLOC_INTCOOKIE:
    {
	/* For backwards compatibility with the old libs (pre 1.8.0)
	 * we allocate an intcookie table on the first cookie
	 * alloc if one hasn't be created already
	 */
	if (uctx->uctx_intcookie_table == NULL)
	{
	    ELAN_CAPABILITY *cap;
	    INTCOOKIE_TABLE *tbl;
	    
	    KMEM_ALLOC (cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), 1);
	    if ((cap == NULL))
		return (-ENOMEM);

	    /* Create a dummy capability */
	    elan_nullcap(cap);

	    /* Must be unique for each process on a node */
	    cap->cap_mycontext = (int) ELAN4_TASK_HANDLE();

	    /* Create a new intcookie table */
	    tbl = intcookie_alloc_table(cap);

	    /* Hang intcookie table off uctx */
	    spin_lock (&uctx->uctx_spinlock);
	    if (uctx->uctx_intcookie_table == NULL)
	    {
		uctx->uctx_intcookie_table = tbl;
		spin_unlock (&uctx->uctx_spinlock);
	    }
	    else
	    {
		spin_unlock (&uctx->uctx_spinlock);
		intcookie_free_table(tbl);
	    }

	    KMEM_FREE(cap, sizeof(*cap));
	}
	
	return (intcookie_alloc (uctx->uctx_intcookie_table, arg));
    }

    case ELAN4IO_FREE_INTCOOKIE:
	if (uctx->uctx_intcookie_table == NULL)
	    return -EINVAL;
	else
	    return (intcookie_free (uctx->uctx_intcookie_table, arg));

    case ELAN4IO_ARM_INTCOOKIE:
	if (uctx->uctx_intcookie_table == NULL)
	    return -EINVAL;
	else
	    return (intcookie_arm (uctx->uctx_intcookie_table, arg));

    case ELAN4IO_WAIT_INTCOOKIE:
	if (uctx->uctx_intcookie_table == NULL)
	    return -EINVAL;
	else
	    return (intcookie_wait (uctx->uctx_intcookie_table, arg));

    case ELAN4IO_FIRE_INTCOOKIE:
    {
	ELAN4IO_FIRECAP_STRUCT *args;

	KMEM_ALLOC (args, ELAN4IO_FIRECAP_STRUCT *, sizeof (ELAN4IO_FIRECAP_STRUCT), 1);
	if ((args == NULL))
	    return (-ENOMEM);

	if (copy_from_user (args, (void *) arg, sizeof (ELAN4IO_FIRECAP_STRUCT)))
	    res = -EFAULT;
	else
	    res = intcookie_fire_cap (&args->fc_capability, args->fc_cookie);
	
	KMEM_FREE(args, sizeof(*args));

	return (res);
    }

    case ELAN4IO_NETERR_MSG:
    {
	ELAN4IO_NETERR_MSG_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_NETERR_MSG_STRUCT)))
	    return (-EFAULT);

	return (user_send_neterr_msg (uctx, args.nm_vp, args.nm_nctx, args.nm_retries, &args.nm_msg));
    }

    case ELAN4IO_NETERR_TIMER:
    {
	unsigned long ticks = ((unsigned long) arg * HZ) / 1000;

	PRINTF (uctx, DBG_NETERR, "elan4_neterr_timer: arg %ld inc %ld\n", arg, ticks);

	mod_timer (&uctx->uctx_neterr_timer, (jiffies + (ticks > 0 ? ticks : 1)));
	return 0;
    }
		
    case ELAN4IO_NETERR_FIXUP:
    {
	ELAN4IO_NETERR_FIXUP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_NETERR_FIXUP_STRUCT)))
	    return (-EFAULT);

	if (args.nf_sten)
	    return (user_neterr_sten (uctx, args.nf_vp, args.nf_cookie, args.nf_waitforeop));
	else
	    return (user_neterr_dma  (uctx, args.nf_vp, args.nf_cookie, args.nf_waitforeop));
    }

    case ELAN4IO_LOAD_TRANSLATION:
    {
	ELAN4IO_TRANSLATION_STRUCT args;
	unsigned long              base, top;

	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_TRANSLATION_STRUCT)))
	    return (-EFAULT);
	
	top  = (args.tr_addr + args.tr_len - 1) | (PAGE_SIZE-1);
	base = args.tr_addr & PAGE_MASK;
	
	return user_load_range (uctx, base, top - base + 1, args.tr_access);
    }
    case ELAN4IO_UNLOAD_TRANSLATION:
    {
	ELAN4IO_TRANSLATION_STRUCT args;
	unsigned long              base, top;
	
	if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_TRANSLATION_STRUCT)))
	    return (-EFAULT);
	
	top  = (args.tr_addr + args.tr_len - 1) | (PAGE_SIZE-1);
	base = args.tr_addr & PAGE_MASK;

	elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* tbl */, base, top - base + 1);

	return 0;
    }

    default:
	PRINTF (uctx, DBG_FILE, "user_ioctl: invalid ioctl %x\n", cmd);
	return (-EINVAL);
    }
}

static void
user_vma_open (struct vm_area_struct *vma)
{
    USER_PRIVATE *pr   = (USER_PRIVATE *) vma->vm_private_data;
    USER_CTXT    *uctx = pr->pr_uctx;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (uctx, DBG_FILE, "user_vma_open: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	elan4_getcqa (&uctx->uctx_ctxt, pgoff);
}

static void 
user_vma_close (struct vm_area_struct *vma)
{
    USER_PRIVATE *pr   = (USER_PRIVATE *) vma->vm_private_data;
    USER_CTXT    *uctx = pr->pr_uctx;
    unsigned long addr;
    unsigned long pgoff;

    PRINTF (uctx, DBG_FILE, "user_vma_close: vm_mm=%p start=%lx end=%lx pgoff=%lx file=%p\n",
	    vma->vm_mm, vma->vm_start, vma->vm_end, vma->vm_pgoff, vma->vm_file);

    /* NOTE: the same comments apply as mem_vma_close */
    for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
	if (elan4_getcqa (&uctx->uctx_ctxt, pgoff) != NULL)
	{
	    elan4_putcqa (&uctx->uctx_ctxt, pgoff);			/* drop the reference we've just taken */
	    elan4_putcqa (&uctx->uctx_ctxt, pgoff);			/* and the one held by the mmap */
	}
}

struct vm_operations_struct user_vm_ops = {
    open:		user_vma_open,
    close:		user_vma_close,
};

static int
user_mmap (struct file *file, struct vm_area_struct *vma)
{
    USER_PRIVATE *pr    = (USER_PRIVATE *) file->private_data;
    USER_CTXT    *uctx  = pr->pr_uctx;
    ELAN4_DEV     *dev   = uctx->uctx_ctxt.ctxt_dev;
    ELAN4_CQA     *cqa;
    unsigned long addr;
    unsigned long pgoff;
    int           res;
    ioaddr_t      ioaddr;
    
    /* Don't allow these pages to be swapped out of dumped */
    vma->vm_flags |= (VM_RESERVED | VM_IO);

    vma->vm_ops          = &user_vm_ops;
    vma->vm_file         = file;
    vma->vm_private_data = (void *) pr;
    
    for (addr = vma->vm_start, pgoff = vma->vm_pgoff; addr < vma->vm_end; addr += PAGE_SIZE, pgoff++)
    {
	switch (pgoff)
	{
	default:
	    PRINTF (uctx, DBG_FILE, "user_mmap: command queue %ld mapping at %lx\n",  pgoff, addr);
	    
	    if ((cqa = elan4_getcqa (&uctx->uctx_ctxt, pgoff)) == NULL)
	    {
		res = -EINVAL;
		goto failed;
	    }

	    PRINTF (uctx, DBG_FILE, "user_mmap: cqa=%p idx=%d num=%d ref=%d\n", cqa, cqa->cqa_idx, cqa->cqa_cqnum, cqa->cqa_ref);
    
	    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	    if (! (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_WRITE_COMBINE) && (cqa->cqa_type & CQ_Reorder) != 0)
		vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);

	    PRINTF (uctx, DBG_FILE, "user_mmap: remap_page_range (%lx, %lx, %lx, %lx)\n",
		    addr, pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + 
		    (cqa->cqa_cqnum + dev->dev_cqoffset) * CQ_CommandMappingSize, PAGE_SIZE,
		    vma->vm_page_prot);

	    if (__io_remap_page_range (addr, 
				       pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + 
				       (cqa->cqa_cqnum + dev->dev_cqoffset) * CQ_CommandMappingSize,
				       PAGE_SIZE, vma->vm_page_prot))
	    {
		PRINTF (uctx, DBG_FILE, "user_mmap: remap_page_range failed\n");

		elan4_putcqa (&uctx->uctx_ctxt, pgoff);
		res = -ENOMEM;
		goto failed;
	    }
	    break;
	    
	case ELAN4_OFF_USER_REGS:
	    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	    switch (dev->dev_devinfo.dev_revision_id)
	    {
	    case PCI_REVISION_ID_ELAN4_REVA:
		ioaddr = pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + ELAN4_REVA_REG_OFFSET + offsetof(E4_Registers, uRegs);
		break;
		
	    case PCI_REVISION_ID_ELAN4_REVB:
		ioaddr = pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + ELAN4_REVB_REG_OFFSET + offsetof(E4_Registers, uRegs);
		break;

	    default:
		res = -EINVAL;
		goto failed;
	    }

	    PRINTF (uctx, DBG_FILE, "user_mmap: user_regs at %lx ioaddr %lx prot %lx\n",
		    addr, ioaddr, vma->vm_page_prot.pgprot);

	    if (__io_remap_page_range (addr,  (ioaddr & PAGEMASK), PAGE_SIZE, vma->vm_page_prot))
	    {			  
		res = -EAGAIN;
		goto failed;
	    }

	    break;
	    
	case ELAN4_OFF_USER_PAGE:
	    PRINTF (uctx, DBG_FILE, "user_mmap: shared user page - kaddr=%lx uaddr=%lx phys=%lx\n", 
		    uctx->uctx_upage, addr, kmem_to_phys (uctx->uctx_upage));

	    /* we do not want to have this area swapped out, lock it */
	    vma->vm_flags |= VM_LOCKED;
	    
	    /* Mark the page as reserved or else the remap_page_range() doesn't remap it */
	    SetPageReserved(pte_page(*find_pte_kernel((unsigned long) uctx->uctx_upage)));
	
	    if (__remap_page_range (addr, kmem_to_phys (uctx->uctx_upage), PAGE_SIZE, vma->vm_page_prot))
	    {
		PRINTF (uctx, DBG_FILE, "user_mmap: remap_page_range (user_page) failed\n");
		res = -ENOMEM;
		goto failed;
	    }
	    break;
	    
	case ELAN4_OFF_TPROC_TRAMPOLINE:
	    vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	    PRINTF (uctx, DBG_FILE, "user_mmap: tproc trampoline - kaddr=%lx uaddr=%lx phys=%lx\n", uctx->uctx_trampoline, addr, 
		    pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM) + uctx->uctx_trampoline + (addr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT)));

	    if (__io_remap_page_range (addr, pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM) + 
				       uctx->uctx_trampoline + (addr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT)),
				       PAGE_SIZE, vma->vm_page_prot))
	    {
		PRINTF (uctx, DBG_FILE, "user_mmap: remap_page_range (tproc_trampoline) failed\n");
		res = -ENOMEM;
		goto failed;
	    }
	    break;
	    
	case ELAN4_OFF_DEVICE_STATS:
	    printk ("user_mmap: device_stats\n");
	    break;
	}
	
    }

    return (0);

 failed:
    for (addr -= PAGE_SIZE, pgoff--; addr >= vma->vm_start; addr -= PAGE_SIZE, pgoff--)
	elan4_putcqa (&uctx->uctx_ctxt, pgoff);		/* drop the reference we've just taken */
    return (res);
}

int
user_pteload (struct vm_area_struct *vma, unsigned long maddr, USER_CTXT *uctx, E4_Addr eaddr, int perm)
{
    USER_PRIVATE *pr    = (USER_PRIVATE *) vma->vm_private_data;
    ELAN4_DEV    *dev   = uctx->uctx_ctxt.ctxt_dev;
    unsigned long pgoff = vma->vm_pgoff + ((maddr - vma->vm_start) >> PAGE_SHIFT);
    register int i, res;

    if (pr->pr_uctx != uctx)
	return -EINVAL;
    
    switch (pgoff)
    {
    default:
    {
	ELAN4_CQA    *cqa;
	unsigned long cqaddr;

	if ((cqa = elan4_getcqa (&uctx->uctx_ctxt, pgoff)) == NULL)
	    return -EINVAL;

	cqaddr = (cqa->cqa_cqnum + dev->dev_cqoffset) * CQ_CommandMappingSize;

	for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = ((cqaddr + i) >> PTE_PADDR_SHIFT) | PTE_SetPerm (perm) | PTE_CommandQueue;

	    if ((res = elan4mmu_pteload (&uctx->uctx_ctxt, 0, eaddr + i, HE_TYPE_COMMAND, newpte)) < 0)
		return res;
	}
	elan4_putcqa (&uctx->uctx_ctxt, pgoff);

	return 0;
    }

    case ELAN4_OFF_USER_REGS:
    {
	u32 blow, bhigh;
	physaddr_t ioaddr;

	/* compute a local pci address from our register BAR */
	pci_read_config_dword (dev->dev_osdep.pdev, PCI_BASE_ADDRESS_2, &blow);
	pci_read_config_dword (dev->dev_osdep.pdev, PCI_BASE_ADDRESS_3, &bhigh);

	ioaddr = (((physaddr_t) bhigh) << 32) | (blow & PCI_BASE_ADDRESS_MEM_MASK);

	switch (dev->dev_devinfo.dev_revision_id)
	{
	case PCI_REVISION_ID_ELAN4_REVA:
	    ioaddr |= ELAN4_REVA_REG_OFFSET + offsetof(E4_Registers, uRegs);
	    break;
	    
	case PCI_REVISION_ID_ELAN4_REVB:
	    ioaddr |= ELAN4_REVB_REG_OFFSET + offsetof(E4_Registers, uRegs);
	    break;
	    
	default:
	    return -EINVAL;
	}

	for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = (((ioaddr & PAGE_MASK) | i) >> PTE_PADDR_SHIFT) | PTE_SetPerm (perm) | PTE_PciNotLocal;

	    if ((res = elan4mmu_pteload (&uctx->uctx_ctxt, 0, eaddr + i, HE_TYPE_REGS, newpte)) < 0)
		return res;
	}

	return 0;
    }

    case ELAN4_OFF_USER_PAGE:
	for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
	    if ((res = elan4mmu_pteload_page (&uctx->uctx_ctxt, 0, eaddr, pte_page(*find_pte_kernel((unsigned long) uctx->uctx_upage)), perm)) < 0)
		return res;
	return 0;

    case ELAN4_OFF_TPROC_TRAMPOLINE:
    {
	sdramaddr_t trampoline = uctx->uctx_trampoline + (maddr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT));

	if (! elan4mmu_sdram_aliascheck (&uctx->uctx_ctxt, eaddr, trampoline))
	    return -EINVAL;
	    
	for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = ((trampoline + i) >> PTE_PADDR_SHIFT) | PTE_SetPerm (perm);

	    if ((res = elan4mmu_pteload (&uctx->uctx_ctxt, 0, eaddr + i, HE_TYPE_SDRAM, newpte)) < 0)
		return res;
	}
	return 0;
    }
    }
}

/* driver entry points */
static int
elan4_open (struct inode *inode, struct file *file)
{
    PRINTF (DBG_USER, DBG_FILE, "elan4_open: device %d minor %d file=%p\n", ELAN4_DEVICE(inode), ELAN4_MINOR(inode), file);
    
    switch (ELAN4_MINOR (inode))
    {
    case ELAN4_MINOR_CONTROL:
	return (control_open (inode, file));
    case ELAN4_MINOR_MEM:
	return (mem_open (inode, file));
    case ELAN4_MINOR_USER:
	return (user_open (inode, file));
    default:
	return (-ENXIO);
    }
}

static int
elan4_release (struct inode *inode, struct file *file)
{
    PRINTF (DBG_USER, DBG_FILE, "elan4_release: device %d minor %d file=%p\n", ELAN4_DEVICE(inode), ELAN4_MINOR(inode), file);
    
    switch (ELAN4_MINOR (inode))
    {
    case ELAN4_MINOR_CONTROL:
	return (control_release (inode, file));
    case ELAN4_MINOR_MEM:
	return (mem_release (inode, file));
    case ELAN4_MINOR_USER:
	return (user_release (inode, file));
    default:
	return (-ENXIO);
    }
}

static int
elan4_ioctl (struct inode *inode, struct file *file, 
	     unsigned int cmd, unsigned long arg)
{
    PRINTF (DBG_USER, DBG_FILE, "elan4_ioctl: device %d minor %d cmd %x\n", ELAN4_DEVICE(inode), ELAN4_MINOR(inode), cmd);
    
    switch (ELAN4_MINOR (inode))
    {
    case ELAN4_MINOR_CONTROL:
	return (control_ioctl (inode, file, cmd, arg));
    case ELAN4_MINOR_MEM:
	return (mem_ioctl (inode, file, cmd, arg));
    case ELAN4_MINOR_USER:
	return (user_ioctl (inode, file, cmd, arg));
    default:
	return (-ENXIO);
    }
}

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
static int
elan4_ioctl32 (unsigned int fd, unsigned int cmd, unsigned long arg, struct file *file)
{
    struct inode *inode = file->f_dentry->d_inode;
    extern int sys_ioctl (unsigned int fd, unsigned int cmd, unsigned long arg);

    PRINTF (DBG_USER, DBG_FILE, "elan4_ioctl32: device %d minor %d cmd %x\n", ELAN4_DEVICE(inode), ELAN4_MINOR(inode), cmd);
    
    if (ELAN4_MINOR (inode) == ELAN4_MINOR_USER)
    {
	USER_PRIVATE *pr    = (USER_PRIVATE *) file->private_data;
	USER_CTXT    *uctx  = pr->pr_uctx;

	if (current->mm != pr->pr_mm)
	    return -EINVAL;
	
	switch (cmd)
	{
	case ELAN4IO_SETPERM32:
	{
	    ELAN4IO_PERM_STRUCT32 args;
	    
	    if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_PERM_STRUCT32)))
		return (-EFAULT);
	    
	    PRINTF (DBG_USER, DBG_FILE, "user_ioctl32: setperm maddr=%x eaddr=%llx len=%llxx perm=%d\n",
		    args.ps_maddr, args.ps_eaddr,args.ps_len, args.ps_perm);

	    return (user_setperm (uctx, args.ps_maddr, args.ps_eaddr, args.ps_len, args.ps_perm));
	}
	
	case ELAN4IO_CLRPERM32:
	{
	    ELAN4IO_PERM_STRUCT32 args;
	    
	    if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_PERM_STRUCT32)))
		return (-EFAULT);
	    
	    PRINTF (DBG_USER, DBG_FILE, "user_ioctl32: clrperm eaddr=%llx len=%ll\n",
		    args.ps_eaddr, args.ps_len);

	    user_clrperm (uctx, args.ps_eaddr, args.ps_len);
	    return (0);
	}
    
	case ELAN4IO_TRAPHANDLER32:
	{
	    ELAN4IO_TRAPHANDLER_STRUCT32 args;
	    
	    if (copy_from_user (&args, (void *) arg, sizeof (ELAN4IO_TRAPHANDLER_STRUCT32)))
		return (-EFAULT);
	    
	    PRINTF (DBG_USER, DBG_FILE, "user_ioctl32: traphandler trapp=%x nticks=%d\n",
		    args.th_trapp, args.th_nticks);

	    return (user_trap_handler (pr->pr_uctx, (ELAN4_USER_TRAP *)(unsigned long)args.th_trapp, args.th_nticks));
	}
	}
    }

    PRINTF (DBG_USER, DBG_FILE, "elan4_ioctl32: fd=%d cmd=%x arg=%lx file=%p\n", fd, cmd, arg, file);
    return (sys_ioctl (fd, cmd, arg));
}
#endif



static int
elan4_mmap (struct file *file, struct vm_area_struct *vma)
{
    PRINTF (DBG_USER, DBG_FILE, "elan4_mmap: instance %d minor %d start=%lx end=%lx pgoff=%lx\n", 
	    ELAN4_DEVICE (file->f_dentry->d_inode), ELAN4_MINOR (file->f_dentry->d_inode),
	    vma->vm_start, vma->vm_end, vma->vm_pgoff);

    switch (ELAN4_MINOR (file->f_dentry->d_inode))
    {
    case ELAN4_MINOR_CONTROL:
	return (control_mmap (file, vma));
    case ELAN4_MINOR_MEM:
	return (mem_mmap (file, vma));
    case ELAN4_MINOR_USER:
	return (user_mmap (file, vma));
    default:
	return (-ENXIO);
    }
}

void
elan4_update_intel_p64h2 (ELAN4_DEV *dev, struct pci_dev *bridge)
{
    u16 cnf;
    
    pci_read_config_word (bridge, 0x40 /* CNF */, &cnf);
    
    /* We expect the CNF register to be configured as follows
     *
     * [8]   == 1	PMODE PCI Mode
     * [7:6] == 2/3	PFREQ PCI Frequency (100/133)
     * [5]   == 0	RSDIS Restreaming Disable
     * [4:3] == 0x	PP    Prefetch Policy
     * [2]   == 0       DTD   Delayed Transaction Depth
     * [1:0] == 10      MDT   MaximumDelaedTransactions
     */
    
    if ((cnf & (1 << 8)) == 0)
	printk ("elan%d: strangeness - elan reports PCI-X but P64H2 reports PCI mode !\n", dev->dev_instance);
    else if ((cnf & 0xb7) != 0x82 && (cnf & 0xb7) != 0x84 && optimise_pci_bus < 2)
	printk ("elan%d: P64H2 CNF is not configured as expected : RSDIS=%d PP=%d DTD=%d MDT=%d\n",
		dev->dev_instance, (cnf >> 5) & 1, (cnf >> 3) & 3, (cnf >> 2) & 1, cnf & 3);
    else
    {
	switch ((cnf >> 6) & 3)
	{
	case 2:						/* PCI-X 100 */
	    pci_write_config_word (bridge, 0xfc /* PC100 */, 0x7777);
	    
	    printk ("elan%d: optimise P64H2 : setting MDT=0, DTD=1, PFC=777 for PCI-X 100\n", dev->dev_instance);
	    
	    break;
	    
	case 3:						/* PCI-X 133 */
	    pci_write_config_word (bridge, 0xfe /* PC133 */, 0x7777);
	    
	    printk ("elan%d: optimise P64H2 : setting MDT=0, DTD=1, PFC=777 for PCI-X 133\n", dev->dev_instance);
	    break;
	}
	
	pci_write_config_word (bridge, 0x40 /* CNF */, (cnf & 0xfff8) | 0x4);	/* DTD=1 MDT=0 */
    }
}

int
elan4_optimise_intel_p64h2 (ELAN4_DEV *dev, struct pci_dev *pdev)
{
    struct pci_bus   *bus      = pdev->bus;
    struct pci_dev   *bridge   = bus->self;
    unsigned int      devcount = 0;
    u8                revision;
    u32               ectrl;
    struct list_head *el;
    
    pci_read_config_dword (pdev, PCI_ELAN_CONTROL, &ectrl);

    /* We can only run in PCI-Xmode with a B1 stepping P64H2 because of P64H2 Errata 3 */
    pci_read_config_byte (bridge, PCI_REVISION_ID, &revision);
    if (revision < 0x04)
    {
	if ((ectrl & ECTRL_INITIALISATION_MODE) != Pci2_2)
	{
	    static const char *p64h2_stepping[4] = {"UNKNOWN", "UNKNOWN", "UNKNOWN", "B0"};

	    printk ("elan%d: unable to use device because of P64H2 Errata 3 on\n"
		    "       %s stepping part and running in a PCI-X slot\n", 
		    dev->dev_instance, p64h2_stepping[revision]);
	    return -EINVAL;
	}
    }
    
    /* We can only alter the bus configuration registers if the Elan is the only device
     * on the bus ... */
    list_for_each (el, &bus->devices) {
	struct pci_dev *pcip = list_entry (el, struct pci_dev, bus_list);

	if (pcip == pdev || (pcip->vendor == PCI_VENDOR_ID_INTEL && pcip->device == 0x1462 /* P64H2 HOTPLUG */))
	    continue;
	    
	devcount++;
    }

    if (devcount > 0 || !list_empty (&bus->children))
    {
	printk ("elan%d: unable to optimise P64H2 settings as %s%s\n", dev->dev_instance,
		(devcount > 0) ? "more than one device on bus" :  "",
		! list_empty (&bus->children) ? "has child buses" : "");
	return 0;
    }

#ifdef __ia64
    if ((ectrl & ECTRL_INITIALISATION_MODE) == PciX100to133MHz)
    {
	struct pci_dev *pcip;
	unsigned int sioh_good      = 0;
	unsigned int sioh_downgrade = 0;
	unsigned int snc_good       = 0;
	unsigned int snc_downgrade  = 0;
	
	/* Search for the associated SIOH and SNC on ia64,
	 * if we have a C2 SIOH and a C0/C1 SNC, then we can
	 * reconfigure the P64H2 as follows:
	 *    CNF:MDT   = 0
	 *    CNF:DTD   = 1
	 *    CNF:PC133 = 7777
	 *
	 * if not, then issue a warning that down rev parts
	 * affect bandwidth.
	 */
	for (pcip = NULL; (pcip = pci_find_device (PCI_VENDOR_ID_INTEL, 0x500, pcip)); )
	{
	    pci_read_config_byte (pcip, PCI_REVISION_ID, &revision);
	    
	    if (revision >= 0x21)
		snc_good++;
	    else
	    {
		printk ("elan%d: SNC revision %x (%s)\n", dev->dev_instance, revision,
			revision == 0x00 ? "A0" : revision == 0x01 ? "A1" : 
			revision == 0x02 ? "A2" : revision == 0x03 ? "A3" :
			revision == 0x10 ? "B0" : revision == 0x20 ? "C0" : 
			revision == 0x21 ? "C1" : "UNKNOWN");
	    
		snc_downgrade++;
	    }
	}

	for (pcip = NULL; (pcip = pci_find_device (PCI_VENDOR_ID_INTEL, 0x510, pcip)) != NULL; )
	{
	    pci_read_config_byte (pcip, PCI_REVISION_ID, &revision);
	    
	    
	    if (revision >= 0x22)
		sioh_good++;
	    else
	    {
		printk ("elan%d: SIOH revsision %x (%s)\n", dev->dev_instance, revision,
			revision == 0x10 ? "C0" : revision == 0x20 ? "C0" : 
			revision == 0x21 ? "C1" : revision == 0x22 ? "C2" : "UNKNOWN");

		sioh_downgrade++;
	    }
	}

	if (optimise_pci_bus < 2 && (sioh_downgrade || snc_downgrade))
	    printk ("elan%d: unable to optimise as SNC/SIOH below required C1/C2 steppings\n", dev->dev_instance);
	else if (optimise_pci_bus < 2 && (sioh_good == 0 || snc_good == 0))
	    printk ("elan%d: unable to optimise as cannot determine SNC/SIOH revision\n", dev->dev_instance);
	else
	    elan4_update_intel_p64h2 (dev, bridge);
    }
#endif
    
#ifdef __i386
    if ((ectrl & ECTRL_INITIALISATION_MODE) == PciX100to133MHz)
	elan4_update_intel_p64h2 (dev, bridge);
#endif	    
    return 0;
}

int
elan4_optimise_intel_pxh (ELAN4_DEV *dev, struct pci_dev *pdev)
{
    dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] |= ELAN4_FEATURE_64BIT_READ;

    return 0;
}

void
elan4_optimise_serverworks_ciobx2 (ELAN4_DEV *dev)
{
    struct pci_dev *pdev = dev->dev_osdep.pdev;
    struct pci_dev *pcip;
    unsigned char   bus;
    unsigned int    dor;
    
    /* Find the CIOBX2 for our bus number */
    for (pcip = NULL; (pcip = pci_find_device (PCI_VENDOR_ID_SERVERWORKS, 0x0101, pcip)) != NULL;)
    {
	pci_read_config_byte (pcip, 0x44 /* BUSNUM */, &bus);
	
	if (pdev->bus->number == bus)
	{
	    printk ("elan%d: optimise CIOBX2 : setting DOR to disable read pipe lining\n", dev->dev_instance);

	    pci_read_config_dword (pcip, 0x78 /* DOR */, &dor);
	    pci_write_config_dword (pcip, 0x78 /* DOR */, dor | (1 << 16));

	    printk ("elan%d: disabling write-combining on ServerWorks chipset\n", dev->dev_instance);
	    dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] |= ELAN4_FEATURE_NO_WRITE_COMBINE;
	}
    }
}

#if defined(CONFIG_IA64_SGI_SN2) || defined(CONFIG_PPC_PSERIES) || defined(__alpha)
int
elan4_optimise_pci_map (ELAN4_DEV *dev, unsigned int features)
{
    dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] |= features;

    if (pci_set_dma_mask (dev->dev_osdep.pdev, (u64) 0XFFFFFFFFFFFFFFFFull) ||
	pci_set_consistent_dma_mask (dev->dev_osdep.pdev, (u64) 0XFFFFFFFFFFFFFFFFull))
    {
	printk (KERN_ERR "elan%d: unable to set DAC mode\n", dev->dev_instance);
	return -EINVAL;
    }

    return 0;
}
#endif

int
elan4_optimise_bus (ELAN4_DEV *dev)
{
    struct pci_dev *pdev = dev->dev_osdep.pdev;

    if (pdev->bus && pdev->bus->self) 
    {
	struct pci_dev *bridge = pdev->bus->self;
	
	if (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x1460 /* Intel P64H2 */)
	    return elan4_optimise_intel_p64h2 (dev, pdev);

	/* See http://pciids.sourceforge.net/iii/?i=8086 */

	if ((bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x0329) /* Intel 6700PXH Fn 0 */ ||
	    (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x032a) /* Intel 6700PXH Fn 2 */ ||
	    (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x032c) /* Intel 6702PXH */ ||
	    (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x0320) /* Intel PXH-D */ ||
	    (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x0330) /* Intel 80332 (A segment) */ ||
	    (bridge->vendor == PCI_VENDOR_ID_INTEL && bridge->device == 0x0332) /* Intel 80332 (B segment) */
	    )
	    return elan4_optimise_intel_pxh (dev, pdev);
    }

    if (pci_find_device (PCI_VENDOR_ID_HP, 0x122e, NULL) != NULL)		/* on HP ZX1 set the relaxed ordering  */
	dev->dev_pteval = PTE_RelaxedOrder;					/* bit to get better DMA bandwidth. */

    if (pci_find_device (PCI_VENDOR_ID_SERVERWORKS, 0x0101, NULL) != NULL) 	/* ServerWorks CIOBX2 */
	elan4_optimise_serverworks_ciobx2 (dev);

#if defined(CONFIG_IA64_SGI_SN2) || defined(CONFIG_PPC_PESERIES)
    return elan4_optimise_pci_map (dev, ELAN4_FEATURE_NO_WRITE_COMBINE | ELAN4_FEATURE_PCI_MAP | ELAN4_FEATURE_NO_PREFETCH);
#endif

#ifdef __alpha
    return elan4_optimise_pci_map (dev, ELAN4_FEATURE_PCI_MAP);
#endif

#ifdef __sparc
    if (! (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_PCI_MAP))
	dev->dev_pteval |= (0xfffe000000000000 >> PTE_PADDR_SHIFT);
#endif

    return 0;
}

int
elan4_pciinit (ELAN4_DEV *dev)
{
    int res;
    u32 value;
    u16 command;
    u8 cacheline;
    unsigned long flags;

    if (optimise_pci_bus && (res = elan4_optimise_bus (dev)) <0)
	return (res);

    if ((res = pci_enable_device (dev->dev_osdep.pdev)) < 0)
	return (res);

    pci_read_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, &value);
    if ((value & ECTRL_INITIALISATION_MODE) == Pci2_2)
	printk ("elan%d: is an elan4%c (PCI-2.2)\n", dev->dev_instance, 'a' + dev->dev_devinfo.dev_revision_id);
    else
    {
	switch (value & ECTRL_INITIALISATION_MODE)
	{
	case PciX50To66MHz:
	    printk ("elan%d: is an elan4%c (PCI-X 50-66)\n", dev->dev_instance, 'a' + dev->dev_devinfo.dev_revision_id);
	    break;
	    
	case PciX66to100MHz:
	    printk ("elan%d: is an elan4%c (PCI-X 66-100)\n", dev->dev_instance, 'a' + dev->dev_devinfo.dev_revision_id);
	    break;
	    
	case PciX100to133MHz:
	    printk ("elan%d: is an elan4%c (PCI-X 100-133)\n", dev->dev_instance, 'a' + dev->dev_devinfo.dev_revision_id);
	    break;
	    
	default:
	    printk ("elan%d: Invalid PCI-X mode\n", dev->dev_instance);
	    return (-EINVAL);
	}
    }

    /* initialise the elan pll control register */
    pci_read_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, &value);

    if (elan4_pll_cfg)
    {
	printk ("elan%d: setting pll control to %08x\n", dev->dev_instance, elan4_pll_cfg);

	pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, elan4_pll_cfg);
    }
    else
    {
	if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, 
				    (value & ~ECTRL_SYS_CLOCK_RATIO_MASK) | ECTRL_SYS_CLOCK_RATIO_4_3);
	else
	    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, 
				    (value & ~ECTRL_SYS_CLOCK_RATIO_MASK) | ECTRL_SYS_CLOCK_RATIO_6_5 | SysPll_FeedForwardISel0 | SysPll_FeedForwardISel1);
    }	

    /* initialise the elan control register */
    pci_read_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, &value);

    value = ((15 << ECTRL_IPROC_HIGH_PRI_TIME_SHIFT) |
	     (15 << ECTRL_OTHER_HIGH_PRI_TIME_SHIFT) |
	     (value & ECTRL_28_NOT_30_BIT_LOCAL_BAR) |
	     (dev->dev_topaddrmode ? ECTRL_ExtraMasterAddrBits : 0) |
	     ECTRL_ENABLE_LATENCY_RESET | 
	     ECTRL_ENABLE_WRITEBURSTS | 
	     ECTRL_ENABLE_2_2READBURSTS);

#ifdef LINUX_SPARC
    value &= ~(ECTRL_ENABLE_LATENCY_RESET | ECTRL_ENABLE_WRITEBURSTS);
#endif

    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, value | ECTRL_SOFTWARE_INTERNAL_RESET);

    switch (dev->dev_devinfo.dev_revision_id)
    {
    case PCI_REVISION_ID_ELAN4_REVA:
	/* Delay 10ms here if we've changed the sysclock ratio */
	/* to allow the PLL to stabalise before proceeding */
	udelay (10000);
	break;
	
    case PCI_REVISION_ID_ELAN4_REVB:
    {
	unsigned char val = read_i2c (dev, I2cLedsValue);

	/* On RevB we have to explicitly reset the PLLs */
	pci_read_config_word (dev->dev_osdep.pdev, PCI_COMMAND, &command);

	write_i2c (dev, I2cLedsValue, val | 0x80);
	udelay (1000);

	/* Issue the PLL counter reset and immediately inhibit all pci interaction 
	 * while the PLL is recovering. The write to the PCI_COMMAND register has 
	 * to occur within 50uS of the write to the i2c registers */
	local_irq_save (flags);
	write_i2c (dev, I2cLedsValue, val & ~0x80);
	pci_write_config_word (dev->dev_osdep.pdev, PCI_COMMAND, (1 << 10) /* PCI_COMMAND_DISABLE_INT */);
	local_irq_restore (flags);

	/* Wait for the write to occur and for the PLL to regain lock */
	udelay (20000); udelay (20000);

	/* Re-enable pci interaction and clear any spurious errors deteced */
	pci_write_config_word (dev->dev_osdep.pdev, PCI_STATUS, PCI_STATUS_DETECTED_PARITY | PCI_STATUS_SIG_SYSTEM_ERROR);
	pci_write_config_word (dev->dev_osdep.pdev, PCI_COMMAND, command);
	break;
    }
    }

    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, value);

    /* Enable master accesses */
    pci_set_master (dev->dev_osdep.pdev);

    /* Verify that the memWrInvalidate bit is set */
    pci_read_config_word (dev->dev_osdep.pdev, PCI_COMMAND, &command);
    pci_read_config_byte (dev->dev_osdep.pdev, PCI_CACHE_LINE_SIZE, &cacheline);

    if ((command & PCI_COMMAND_INVALIDATE) == 0)
    {
	printk ("elan%d: enable MemWrInvalidate (cacheline %d)\n",
		dev->dev_instance, cacheline * 4);

	pci_write_config_word (dev->dev_osdep.pdev, PCI_COMMAND, command | PCI_COMMAND_INVALIDATE);
    }

    if (pci_request_regions(dev->dev_osdep.pdev, "elan4"))
	return -ENODEV;

    /* add the interrupt handler */
    if (request_irq (dev->dev_osdep.pdev->irq, elan4_irq, SA_SHIRQ, "elan4", dev) != 0)
    {
	pci_release_regions (dev->dev_osdep.pdev);
	return -ENXIO;
    }

    return (0);
}

void
elan4_updatepll (ELAN4_DEV *dev, unsigned int val)
{
    u32 value;

    if (elan4_pll_cfg == 0)
    {
	pci_read_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, &value);

	pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_PLL_CONTROL, 
				(value & ~ECTRL_SYS_CLOCK_RATIO_MASK) | val);

	/* Delay 10ms here if we've changed the sysclock ratio */
	/* to allow the PLL to stabalise before proceeding */
	udelay (10000);
    }	
}

void
elan4_pcifini (ELAN4_DEV *dev)
{
    u32 value;

    /* release the interrupt handler */
    free_irq (dev->dev_osdep.pdev->irq, dev);

    /* release the address space */
    pci_release_regions (dev->dev_osdep.pdev);

    pci_read_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, &value);
    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, value | ECTRL_SOFTWARE_INTERNAL_RESET);
    pci_write_config_dword (dev->dev_osdep.pdev, PCI_ELAN_CONTROL, value);

    pci_disable_device (dev->dev_osdep.pdev);
}

void
elan4_pcierror (ELAN4_DEV *dev)
{
    struct pci_dev *pci = dev->dev_osdep.pdev;
    u8  type;
    u16 status, cmd;
    u32 physlo, physhi, control;
    
    printk("elan%d: pci error has occurred\n", dev->dev_instance);

    pci_read_config_word  (pci, PCI_STATUS,		 &status);
    pci_read_config_word  (pci, PCI_COMMAND,             &cmd);
    pci_read_config_dword (pci, PCI_ELAN_CONTROL,	 &control);

    if (control & ECTRL_REC_SPLIT_COMP_MESSAGE)
    {
	u32 message, attr;
	
	pci_write_config_dword (pci, PCI_ELAN_CONTROL, control & ~ECTRL_SELECT_SPLIT_MESS_ATTR);
	pci_read_config_dword (pci, PCI_ELAN_SPLIT_MESSAGE_VALUE, &message);
	pci_write_config_dword (pci, PCI_ELAN_CONTROL, control | ECTRL_SELECT_SPLIT_MESS_ATTR);
	pci_read_config_dword (pci, PCI_ELAN_SPLIT_MESSAGE_VALUE, &attr);

	printk ("elan%d: pcierror - received split completion message - attr=%08x, message=%08x\n", 
		dev->dev_instance, attr, message);

	pci_write_config_dword (pci, PCI_ELAN_CONTROL, control | ECTRL_REC_SPLIT_COMP_MESSAGE);	/* clear the error */
    }
    else
    {
	pci_read_config_dword (pci, PCI_ELAN_PARITY_ADDR_LO, &physlo);
	pci_read_config_dword (pci, PCI_ELAN_PARITY_ADDR_HI, &physhi);
	pci_read_config_byte  (pci, PCI_ELAN_PARITY_TYPE,    &type);
	
	printk ("elan%d: pcierror - status %x cmd %4x physaddr %08x%08x type %x\n", 
		dev->dev_instance, status, cmd, physhi, physlo, type);
	
	if (status & PCI_STATUS_PARITY)
	    printk ("elan%d: parity error signalled (PERR)\n", dev->dev_instance);
	if (status & PCI_STATUS_DETECTED_PARITY)
	    printk ("elan%d: detected parity error\n", dev->dev_instance);
	if (status & PCI_STATUS_REC_MASTER_ABORT)
	    printk ("elan%d: received master abort\n", dev->dev_instance);
	if (status & PCI_STATUS_REC_TARGET_ABORT)
	    printk ("elan%d: received target abort\n", dev->dev_instance);
	if (status & PCI_STATUS_SIG_SYSTEM_ERROR)
	    printk ("elan%d: signalled SERR\n", dev->dev_instance);
	if (status & PCI_STATUS_SIG_TARGET_ABORT)
	    printk ("elan%d: signalled target abort\n", dev->dev_instance);

	pci_write_config_word (pci, PCI_STATUS, status);	/* clear the errors */
    }

    DISABLE_INT_MASK (dev, INT_PciMemErr);

#ifdef notdef
    panic ("elan%d: pcierror\n", dev->dev_instance);		/* better panic ! */
#endif
}

static irqreturn_t
elan4_irq (int irq, void *arg, struct pt_regs *regs)
{
    if (elan4_1msi0 ((ELAN4_DEV *) arg))
	    return IRQ_HANDLED;
    else
	    return IRQ_NONE;
}

ioaddr_t
elan4_map_device (ELAN4_DEV *dev, unsigned bar, unsigned off, unsigned size, ELAN4_MAP_HANDLE *handle)
{
    return (ioaddr_t) ioremap_nocache (pci_resource_start (dev->dev_osdep.pdev, bar) + off, size);
}

void
elan4_unmap_device (ELAN4_DEV *dev, ioaddr_t ptr, unsigned size, ELAN4_MAP_HANDLE *handle)
{
    iounmap ((void *) ptr);
}

unsigned long
elan4_resource_len (ELAN4_DEV *dev, unsigned bar)
{
    return (pci_resource_len (dev->dev_osdep.pdev, bar));
}

void
elan4_configure_writecombining (ELAN4_DEV *dev)
{
#ifdef CONFIG_MTRR
    dev->dev_osdep.sdram_mtrr = dev->dev_osdep.regs_mtrr = -1;
#endif

    if ((dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_NO_WRITE_COMBINE))
	return;

#if (defined(__i386) || defined(__x86_64))

#if defined (X86_FEATURE_PAT)

#ifndef boot_cpu_has
#  define boot_cpu_has(bit)      test_bit(bit, boot_cpu_data.x86_capability)
#endif

    /* Try to utilise PAT entries which already exist */
    if (boot_cpu_has (X86_FEATURE_PAT))
    {
	unsigned int val0, val1, i;
	int slot = -1;

	/* Read the IA32CR_PAT MSR register and see if a slot is
	 * set for write-combinig.  Note we assume that all CPUs 
	 * are configured the same like they're supposed to. */
	rdmsr (0x277, val0, val1);
	
	/* Check for PAT write combining entry (value 0x01) */
	for (i = 0; i < 4; i++, val0 >>= 8)
	    if ((val0 & 0xff) == 0x01)
		slot = i;
	for (i = 4; i < 8; i++, val1 >>= 8)
	    if ((val1 & 0xff) == 0x01)
		slot = i;

	if (slot >= 0)
	{
	    printk ("elan%d: using PAT for write combining (slot %d)\n", dev->dev_instance, slot);

	    pat_pteval = ((slot & 4) ? _PAGE_PSE : 0) | ((slot & 2) ? _PAGE_PCD : 0) | ((slot & 1) ? _PAGE_PWT : 0);
	    return;
	}
    }
#endif

#ifdef CONFIG_MTRR
    /* try and initialise the MTRR registers to enable write-combining */
    dev->dev_osdep.sdram_mtrr = mtrr_add (pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM), 
					  pci_resource_len   (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM),
					  MTRR_TYPE_WRCOMB, 1);
    if (dev->dev_osdep.sdram_mtrr < 0)
	printk ("elan%d: cannot configure MTRR for sdram\n", dev->dev_instance);
    
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVB)
    {
	unsigned int cqreorder = dev->dev_cqcount >> 1;
	unsigned int cqcount   = dev->dev_cqcount  - cqreorder;

	dev->dev_osdep.regs_mtrr = mtrr_add (pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + 
					     (dev->dev_cqoffset + cqreorder) * CQ_CommandMappingSize,
					     CQ_CommandMappingSize * cqcount,
					     MTRR_TYPE_WRCOMB, 1);
	
	if (dev->dev_osdep.regs_mtrr < 0)
	    printk ("elan%d: cannot configure MTRR for command ports\n", dev->dev_instance);
	else
	{
	    dev->dev_cqreorder = cqreorder;
	    return;
	}
    }
#endif

    /* Set flag so that userspace knows write-combining is disabled */
    dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] |= ELAN4_FEATURE_NO_WRITE_COMBINE;
#endif

}

void
elan4_unconfigure_writecombining (ELAN4_DEV *dev)
{
#if defined (X86_FEATURE_PAT)
    if (pat_pteval != -1) return;
#endif

#ifdef CONFIG_MTRR
    if (dev->dev_osdep.sdram_mtrr >=0 )
	mtrr_del (dev->dev_osdep.sdram_mtrr, pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM), 
		  pci_resource_len   (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM));
    
    if (dev->dev_cqreorder && dev->dev_osdep.regs_mtrr >= 0)
	mtrr_del (dev->dev_osdep.regs_mtrr, 
		  pci_resource_start (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS) + 
		  (dev->dev_cqoffset + dev->dev_cqreorder) * CQ_CommandMappingSize,
		  CQ_CommandMappingSize * (dev->dev_cqcount >> 1));
#endif
}

EXPORT_SYMBOL(elan4_reference_device);
EXPORT_SYMBOL(elan4_dereference_device);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
