/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: mmu.c,v 1.47.2.3 2005/07/14 13:34:12 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/mmu.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <elan4/debug.h>
#include <elan4/device.h>
#include <linux/pagemap.h>

int elan4_debug_mmu;
int elan4_mmuhash_chain_reduction     = 1;
int elan4_mmuhash_chain_end_reduce    = 0;
int elan4_mmuhash_chain_middle_reduce = 0;
int elan4_mmuhash_chain_middle_fail   = 0;
int elan4_mmuhash_shuffle_attempts    = 0;
int elan4_mmuhash_shuffle_done        = 0;

/* Permission table - see ELAN4 MMU documentation */
u_char elan4_permtable[] =
{
   0x00, /* 0x000000 - Disable */
   0x00, /* 0x000000 - Unused  */
   0x01, /* 0x000001 - Local Data Read */
   0x03, /* 0x000011 - Local Data Write */
   0x11, /* 0x010001 - Local Read */
   0x10, /* 0x010000 - Local Execute */
   0x05, /* 0x000101 - Read Only */
   0x13, /* 0x010011 - Local Write */
   0x20, /* 0x100000 - Local Event Access */
   0x23, /* 0x100011 - Local Event Write Ac */
   0xa3, /* 1x100011 - Remote Ev Loc Write */
   0xaf, /* 1x101111 - Remote All */
   0x07, /* 0x000111 - Remote Read Only */
   0x0d, /* 0x001101 - Remote Write Only */
   0x0f, /* 0x001111 - Remote Read/Write */
   0xbf, /* 1x111111 - No Fault */
};

u_char elan4_permreadonly[] = 
{
    PERM_Disabled,		/* PERM_Disabled */
    PERM_Disabled,		/* PERM_Unused */
    PERM_LocDataRead,		/* PERM_LocDataRead */
    PERM_LocDataRead,		/* PERM_LocDataWrite */
    PERM_LocRead,		/* PERM_LocRead */
    PERM_LocExecute,		/* PERM_LocExecute */
    PERM_ReadOnly,		/* PERM_ReadOnly */
    PERM_LocRead,		/* PERM_LocWrite */
    PERM_LocEventOnly,		/* PERM_LocEventOnly */
    PERM_LocDataRead,		/* PERM_LocEventWrite */
    PERM_LocDataRead,		/* PERM_RemoteEvent */
    PERM_ReadOnly,		/* PERM_RemoteAll */
    PERM_RemoteReadOnly,	/* PERM_RemoteReadOnly */
    PERM_ReadOnly,		/* PERM_RemoteWriteLocRead */
    PERM_ReadOnly,		/* PERM_DataReadWrite */
    PERM_ReadOnly,		/* PERM_NoFault */
};

static void
elan4mmu_synctag (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *he, int tagidx)
{
    E4_uint64 value = (he->he_tag[tagidx] & HE_TAG_VALID) ? he->he_tag[tagidx] & (TAG_ADDRESS_MASK | TAG_CONTEXT_MASK) : INVALID_CONTEXT;
    
    if (he->he_next)
	value |= ((tagidx == 0) ? 
		  ((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) :
		  ((he->he_next->he_entry << TAG_CHAINPTR_LOW_SHIFT) & TAG_CHAINPTR_18TO6_MASK));
    else if (tagidx == 0)
	value |= TAG_CHAINPTR_30TO19_MASK;
    
    MPRINTF (DBG_DEVICE, 4, "elan4mmu_synctag: he=%p tagidx=%d he->he_tag=%llx -> value=%llx\n", he, tagidx, he->he_tag[tagidx], value);

    elan4_sdram_writeq (dev, he->he_entry + E4MMU_TAG_OFFSET(tagidx), value);
}

static void
elan4mmu_chain_hents (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *phe, ELAN4_HASH_ENTRY *he)
{
    ASSERT ((elan4_sdram_readq (dev, phe->he_entry + E4MMU_TAG_OFFSET(0)) & TAG_CHAINPTR_30TO19_MASK) == TAG_CHAINPTR_30TO19_MASK);

    elan4_sdram_writeq (dev, phe->he_entry + E4MMU_TAG_OFFSET(1),
			((phe->he_tag[1] & (TAG_ADDRESS_MASK | TAG_CONTEXT_MASK)) | ((he->he_entry << TAG_CHAINPTR_LOW_SHIFT) & TAG_CHAINPTR_18TO6_MASK)));
    elan4_sdram_writeq (dev, phe->he_entry + E4MMU_TAG_OFFSET(0),
			((phe->he_tag[0] & (TAG_ADDRESS_MASK | TAG_CONTEXT_MASK)) | ((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK)));
}

static void
elan4mmu_writepte (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *he, int tagidx, int pteidx, E4_uint64 value)
{
    /*
     * NOTE - we can only change a valid PTE if we're upgrading it's permissions,
     * any other changes should have invalidated it first. */

    MPRINTF (DBG_DEVICE, 4, "elan4mmu_writepte: he=%p tagidx=%d pteidx=%x value=%llx\n", he, tagidx, pteidx, (unsigned long long) value);

    if (pteidx == 3)
    {
	elan4_sdram_writew (dev, he->he_entry + E4MMU_PTE3_WORD1_OFFSET(tagidx), (value >> 16) & 0xFFFF);
	elan4_sdram_writew (dev, he->he_entry + E4MMU_PTE3_WORD2_OFFSET(tagidx), (value >> 32) & 0xFFFF);
	elan4_sdram_writew (dev, he->he_entry + E4MMU_PTE3_WORD0_OFFSET(tagidx), (value >> 0)  & 0xFFFF);
    }
    else
    {
	elan4_sdram_writew (dev, he->he_entry + E4MMU_PTE_HIGH_OFFSET(tagidx, pteidx), (value >> 32) & 0xFFFF);
	elan4_sdram_writel (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, pteidx), value & 0xFFFFFFFF);
    }
}

static void
elan4mmu_invalidatepte (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *he, int tagidx, int pteidx)
{
    if (pteidx == 3)
	elan4_sdram_writeb (dev, he->he_entry + E4MMU_PTE3_WORD0_OFFSET(tagidx), PTE_SetPerm (PERM_Disabled));
    else
	elan4_sdram_writeb (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, pteidx), PTE_SetPerm (PERM_Disabled));
}

static E4_uint64
elan4mmu_readpte (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *he, int tagidx, int pteidx)
{
    if (pteidx == 3)
	return (((E4_uint64) elan4_sdram_readw (dev, he->he_entry + E4MMU_PTE3_WORD0_OFFSET(tagidx)) << 0)  |
		((E4_uint64) elan4_sdram_readw (dev, he->he_entry + E4MMU_PTE3_WORD1_OFFSET(tagidx)) << 16) |
		((E4_uint64) elan4_sdram_readw (dev, he->he_entry + E4MMU_PTE3_WORD2_OFFSET(tagidx)) << 32));
    else
	return ((E4_uint64) elan4_sdram_readl (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, pteidx)) |
		((E4_uint64) elan4_sdram_readw (dev, he->he_entry + E4MMU_PTE_HIGH_OFFSET(tagidx, pteidx)) << 32));
}


void
elan4mmu_flush_tlb (ELAN4_DEV *dev)
{
    PULSE_SYSCONTROL (dev, CONT_TLB_FLUSH);

    while (read_reg64 (dev, SysControlReg) & CONT_TLB_FLUSH)
	DELAY (1);
}

/*
 * elanmmu_flush_tlb_hash - this flushes the hash copy entries and the elan
 * tlb.  However after the write to the hash copy entry if the elan was
 * in the process of walking, then it could write the hash copy with a valid
 * entry which we had just invalidated. However once we've seen the tlb flushed
 * then if the walk engine had done a write - then we need to invaldate the
 * hash copy entries again and reflush the tlb.
 *
 * If we're invalidating a lot of hash blocks, then the chances are that the
 * walk engine will perform a write - so we flush the tlb first, then invalidate
 * the hash copy entries, then flush the tlb again.
 */
static void
elan4mmu_flush_tlb_hash (ELAN4_DEV *dev, int tbl, unsigned baseidx, unsigned topidx)
{
    int       notmany = (abs(topidx - baseidx) < 5) ? 1 : 0;
    int       hashidx;
    E4_uint32 reg;

    if (notmany)
	PULSE_SYSCONTROL (dev, CONT_CLEAR_WALK_WROTE_TABLES);
    else
	elan4mmu_flush_tlb(dev);

    do {
	for (hashidx = baseidx; hashidx <= topidx; hashidx++)
	    if (dev->dev_mmuhash[tbl][hashidx].he_tag[0] & HE_TAG_COPY)
	    {
		ASSERT ((dev->dev_mmuhash[tbl][hashidx].he_tag[0] & HE_TAG_VALID) == 0);
		ASSERT ((dev->dev_mmuhash[tbl][hashidx].he_tag[1] & HE_TAG_VALID) == 0);

		elan4mmu_synctag (dev, &dev->dev_mmuhash[tbl][hashidx], 0);
		elan4mmu_synctag (dev, &dev->dev_mmuhash[tbl][hashidx], 1);
	    }
	
	PULSE_SYSCONTROL (dev, CONT_TLB_FLUSH);
	
	while ((reg = read_reg64 (dev, SysControlReg)) & CONT_TLB_FLUSH)
	    DELAY (1);
	
    } while (notmany-- && (reg & CONT_CLEAR_WALK_WROTE_TABLES) != 0);
}

void
elan4mmu_display_hent (ELAN4_DEV *dev, ELAN4_HASH_ENTRY *he, int hashidx)
{
    int tagidx;

    elan4_debugf (DBG_DEVICE, DBG_MMU, "elan4mmu_display_hent: hashidx=%d he=%p entry at %lx\n", hashidx, he, he->he_entry);
    elan4_debugf (DBG_DEVICE, DBG_MMU, "                       next=%p prev=%p chain=%p,%p\n", he->he_next, he->he_prev, he->he_chain[0], he->he_chain[1]);
    for (tagidx = 0; tagidx < 2; tagidx++)
    {
	E4_uint64 tag  = elan4_sdram_readq (dev, he->he_entry + E4MMU_TAG_OFFSET(tagidx));
	E4_uint64 pte0 = elan4_sdram_readq (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, 0));
	E4_uint64 pte1 = elan4_sdram_readq (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, 1));
	E4_uint64 pte2 = elan4_sdram_readq (dev, he->he_entry + E4MMU_PTE_LOW_OFFSET(tagidx, 2));
	E4_uint64 pte3 = ((pte0 >> 48) | (pte1 >> 32) | (pte2 >> 16));
	
	elan4_debugf (DBG_DEVICE, DBG_MMU, "                       Tag %d (%llx,%08x) context=%04x vaddr=%llx\n", tagidx, he->he_tag[tagidx], he->he_pte[tagidx], (int) (tag & TAG_CONTEXT_MASK), (tag & TAG_ADDRESS_MASK));
	elan4_debugf (DBG_DEVICE, DBG_MMU, "                       Pte 0 - PPN=%llx PERM=%x TYPE=%x%s%s\n", (pte0 & PTE_PPN_MASK) >> PTE_PPN_SHIFT, 
		      (int) (pte0 & PTE_PERM_MASK) >> PTE_PERM_SHIFT, (int)(pte0 & PTE_TYPE_MASK), (pte0 & PTE_MOD_MASK) ? " mod" : "", (pte0 & PTE_REF_MASK) ? " ref" : "");
	elan4_debugf (DBG_DEVICE, DBG_MMU, "                       Pte 1 - PPN=%llx PERM=%x TYPE=%x%s%s\n", (pte1 & PTE_PPN_MASK) >> PTE_PPN_SHIFT, 
		      (int) (pte1 & PTE_PERM_MASK) >> PTE_PERM_SHIFT, (int)(pte1 & PTE_TYPE_MASK), (pte1 & PTE_MOD_MASK) ? " mod" : "", (pte1 & PTE_REF_MASK) ? " ref" : "");
	elan4_debugf (DBG_DEVICE, DBG_MMU, "                       Pte 2 - PPN=%llx PERM=%x TYPE=%x%s%s\n", (pte2 & PTE_PPN_MASK) >> PTE_PPN_SHIFT, 
		      (int) (pte2 & PTE_PERM_MASK) >> PTE_PERM_SHIFT, (int)(pte2 & PTE_TYPE_MASK), (pte2 & PTE_MOD_MASK) ? " mod" : "", (pte2 & PTE_REF_MASK) ? " ref" : "");
	elan4_debugf (DBG_DEVICE, DBG_MMU, "                       Pte 3 - PPN=%llx PERM=%x TYPE=%x%s%s\n", (pte3 & PTE_PPN_MASK) >> PTE_PPN_SHIFT, 
		      (int) (pte3 & PTE_PERM_MASK) >> PTE_PERM_SHIFT, (int)(pte3 & PTE_TYPE_MASK), (pte3 & PTE_MOD_MASK) ? " mod" : "", (pte3 & PTE_REF_MASK) ? " ref" : "");
    }
}

static __inline__ ELAN4_HASH_ENTRY *
he_ctxt_next (ELAN4_HASH_ENTRY *he, int ctxnum)
{
    return ((he->he_tag[0] & TAG_CONTEXT_MASK) == ctxnum) ? he->he_chain[0] : he->he_chain[1];
}

static __inline__ ELAN4_HASH_ENTRY *
he_ctxt_unlink (ELAN4_CTXT *ctxt, int tbl, int hashidx, ELAN4_HASH_ENTRY *prevhe, ELAN4_HASH_ENTRY *he, ELAN4_HASH_ENTRY *next)
{
    /* Check whether either tag is in use by this context */
    if ((he->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num || (he->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num)
	return he;

    if (prevhe == NULL)
	ctxt->ctxt_mmuhash[tbl][hashidx] = next;
    else
    {
	/* previous he, ensure that both chain pointers are changed is this ctxt is using both tags */
	ASSERT ((prevhe->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num || (prevhe->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num);

	if ((prevhe->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num)
	    prevhe->he_chain[0] = next;
	if ((prevhe->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num)
	    prevhe->he_chain[1] = next;
    }

    return prevhe;
}

void
elan4mmu_display (ELAN4_CTXT *ctxt, int tbl, const char *tag)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    ELAN4_HASH_ENTRY *he;
    int hashidx;

    for (hashidx = 0; hashidx < dev->dev_hashsize[tbl]; hashidx++)
	for (he = ctxt->ctxt_mmuhash[tbl][hashidx]; he != NULL; he = he_ctxt_next (he, ctxt->ctxt_num))
	{
	    elan4_debugf (DBG_DEVICE, DBG_MMU, "%s: hashidx=%d he=%p tags <%llx,%llx>\n", tag, hashidx, he,
			  (he->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num ? E4MMU_TAG2VADDR (he->he_tag[0], hashidx, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1) : 0,
			  (he->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num ? E4MMU_TAG2VADDR (he->he_tag[1], hashidx, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1) : 0);
	    elan4mmu_display_hent (dev, he, hashidx);
	}
}
static ELAN4_HASH_ENTRY *
elan4mmu_find_next_free (ELAN4_HASH_ENTRY *he)
{
    /* the current one could be free */	
    /* return NULL if not free one   */
    while ( he ) 
    {
	if ( ((he->he_tag[0] & TAG_CONTEXT_MASK) == INVALID_CONTEXT) || ((he->he_tag[1] & TAG_CONTEXT_MASK) == INVALID_CONTEXT)) 
	    return he;
	he = he->he_next;
    }
    return (NULL);
}
static ELAN4_HASH_ENTRY *
elan4mmu_alloc_hent (ELAN4_DEV *dev, int tbl, int hashidx, E4_uint64 newtag, int *tagidx)
{
    ELAN4_HASH_ENTRY *he, *phe;
    unsigned long flags;
    int i;

    spin_lock_irqsave (&dev->dev_mmulock, flags);

    /* see if there are any partial free blocks */
    if ((he = elan4mmu_find_next_free (&dev->dev_mmuhash[tbl][hashidx])) != NULL)
    {
	*tagidx = ((he->he_tag[0] & TAG_CONTEXT_MASK) == INVALID_CONTEXT) ? 0 : 1;
	
	MPRINTF (DBG_DEVICE, 3, "elan4mmu_alloc_hent: allocate he=%p idx=%d%s\n", he, *tagidx, (he == &dev->dev_mmuhash[tbl][hashidx]) ? " hash-block" : "");
	
	he->he_tag[*tagidx] = newtag | HE_TAG_VALID;

	elan4mmu_synctag (dev, he, *tagidx);
		
	spin_unlock_irqrestore (&dev->dev_mmulock, flags);
	return (he);
    }
    
    if ((he = dev->dev_mmufreelist) != NULL)
	dev->dev_mmufreelist = he->he_next;
    else
    {
	ELAN4_HASH_CHUNK *hc;
	sdramaddr_t       entry;

	KMEM_ALLOC (hc, ELAN4_HASH_CHUNK *, sizeof (ELAN4_HASH_CHUNK), 0);
	
	if (hc == NULL)
	{
	    spin_unlock_irqrestore (&dev->dev_mmulock, flags);
	    return ((ELAN4_HASH_ENTRY *) NULL);
	}
	
	if ((entry = elan4_sdram_alloc (dev, sizeof (E4_HashTableEntry) * ELAN4_HENT_CHUNKS)) == (sdramaddr_t) 0)
	{
	    spin_unlock_irqrestore (&dev->dev_mmulock, flags);

	    KMEM_FREE (hc, sizeof (ELAN4_HASH_CHUNK));
	    return ((ELAN4_HASH_ENTRY *) NULL);
	}

	list_add_tail (&hc->hc_link, &dev->dev_hc_list);

	elan4_sdram_zeroq_sdram (dev, entry, sizeof (E4_HashTableEntry) * ELAN4_HENT_CHUNKS);

	/* no initialise all chunks and chain all but the first onto the freelist */
	for (i = 0; i < ELAN4_HENT_CHUNKS; i++, entry += sizeof (E4_HashTableEntry))
	{
	    hc->hc_hents[i].he_entry = entry;

	    if (i == 0)
		he = &hc->hc_hents[0];
	    else
	    {
		hc->hc_hents[i].he_next = dev->dev_mmufreelist;
		dev->dev_mmufreelist = &hc->hc_hents[i];
	    }
	}
    }

    /* Initialise hash entry, using slot 0 */
    *tagidx = 0;

    he->he_next     = NULL;
    he->he_prev     = NULL;
    he->he_chain[0] = NULL;
    he->he_chain[1] = NULL;
    he->he_tag[0]   = newtag | HE_TAG_VALID;
    he->he_tag[1]   = E4MMU_TAG(0, INVALID_CONTEXT);
    he->he_pte[0]   = 0;
    he->he_pte[1]   = 0;
    
    elan4mmu_synctag (dev, he, 0);
    
    /* add to mmuhash lists */
    for (phe = &dev->dev_mmuhash[tbl][hashidx]; phe->he_next; phe = phe->he_next)
	;
    phe->he_next = he;
    he->he_prev  = phe;
    he->he_next  = NULL;
    
    /* finally chain the hash block into the hash tables */
    elan4mmu_chain_hents (dev, phe, he);
    
    spin_unlock_irqrestore (&dev->dev_mmulock, flags);
    return (he);
}
void
elan4mmu_set_shuffle(ELAN4_CTXT *ctxt, int tbl, int hashidx)
{
    int i;
    
    for(i=0; (i<ELAN4_CTXT_MAX_SHUFFLE) && (ctxt->shuffle[tbl][i]>=0) && (ctxt->shuffle[tbl][i]!=hashidx); i++)
	;
    if (i<ELAN4_CTXT_MAX_SHUFFLE) {
	ctxt->shuffle_needed[tbl] = 1;
	ctxt->shuffle[tbl][i]     = hashidx;
    }
}
static int 
elan4mmm_try_to_free_hent(ELAN4_DEV *dev, int tbl, int hashidx, ELAN4_HASH_ENTRY *he)
{   
    ELAN4_HASH_ENTRY *prev;
    int               t;
    ELAN4_CTXT       *ctxt;

    
    while (he) {
	if ( ((he->he_tag[0] & TAG_CONTEXT_MASK) == INVALID_CONTEXT)
	     && ((he->he_tag[1] & TAG_CONTEXT_MASK) == INVALID_CONTEXT)) {
	    /* Both tags are now free */
	    
	    if (he != &dev->dev_mmuhash[tbl][hashidx]) {
		/* its not the hash entry block */
		
		if ( he->he_next == NULL ) {
		    /* its the end one so just remove it */
		    prev = he->he_prev;
		    
		    /* make the previous entry the end one and sync it */
		    prev->he_next = NULL;
		    elan4mmu_synctag (dev, prev, 0);
		    
		    /* make sure the elan had finished traversing the list */
		    elan4mmu_flush_tlb(dev);
		    
		    /* now we have a free he in our hands put it onto the free list */
		    he->he_next = dev->dev_mmufreelist;
		    dev->dev_mmufreelist = he;
		    
		    elan4_mmuhash_chain_end_reduce++;
		    
		    he = prev;
		} else {
		    /* can only remove if my he_entry high bits = next he_entry high bits. */

		    if (((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) == 
			((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK))
		    {
			prev = he->he_prev;
			
			/* make the previous entry jump over us and sync it */
			prev->he_next = he->he_next;
			elan4mmu_synctag (dev, prev, 1);
			he->he_next->he_prev = prev;
			
			/* make sure the elan had finished traversing the list */
			elan4mmu_flush_tlb(dev);
			
			/* now we have a free he in our hands put it onto the free list */
			he->he_next = dev->dev_mmufreelist;
			dev->dev_mmufreelist = he;
			
			elan4_mmuhash_chain_middle_reduce++;
			
			he = prev;	
			
		    } else {
			elan4_mmuhash_chain_middle_fail++;
			/* at this point we wanted to remove an entry but we cant because this would mean
			   chanaging the high bits of the perivious pointer.
			   
			   It is assumed that this is a fairly rare occurance.
			   
			   The plan is. to tell the ctxt's in the end entry (which can always be removed)
			   to shuffle down. They need to do this as its guarded by a ctxt lock i dont have.
			   
			   Note the ctxt entry might not exist by the time they get round to shuffling.
			   and/or the empty node we want to shuffle to might have gone. so there is no
			   value in storing info about what you want to shuffle. 
			   
			   just tell the ctxt to shuffle this hashidx. rather than allocate a block 
			   of memory the size of the number of hashidx's to handle this we will use 
			   a short array. assuming its rarely going to fill. if it does the all the ctxt's
			   hashidx's are shuffled (as its really unlikely to happen
			*/
			
			/* mark all up to the end as needing shuffle */
			while (he->he_next) {
			    for(t=0;t<2;t++) {
				if ((he->he_tag[t] & TAG_CONTEXT_MASK)!=INVALID_CONTEXT) {
				    ctxt = elan4_localctxt (dev, (he->he_tag[t] & TAG_CONTEXT_MASK));
				    if (ctxt) {
					ASSERT(ctxt->ctxt_ops);
					if (ctxt->ctxt_ops->op_need_shuffle)
					    ctxt->ctxt_ops->op_need_shuffle (ctxt, tbl, hashidx);
				    }
				}
			    }
			    he = he->he_next;
			}
			
			he = NULL;
		    }
		}
	    } else he = NULL;
	} else he = NULL;
    }
    return (0);
}
static void
elan4mmu_free_hent_nolock (ELAN4_DEV *dev, int tbl, int hashidx, ELAN4_HASH_ENTRY *he, int tagidx)
{ 
    /* assumes some one has the mmulock before this is called */
    int pteidx;

    /* Invalidate the tag, and zero all ptes */
    for (pteidx = 0; pteidx < 4; pteidx++)
	if (HE_GET_PTE(he, tagidx, pteidx) != HE_TYPE_INVALID)
	    elan4mmu_writepte (dev, he, tagidx, pteidx, 0);

    he->he_tag[tagidx] = E4MMU_TAG(0, INVALID_CONTEXT);
    he->he_pte[tagidx] = 0;

    elan4mmu_synctag (dev, he, tagidx);

    if ((he->he_tag[tagidx^1] & TAG_CONTEXT_MASK) == INVALID_CONTEXT) /* Both tags are now free */
    {
	if (he == &dev->dev_mmuhash[tbl][hashidx])		/* it's the hash block entry */
	{							/* so as it's already on the freelist */
	    he->he_chain[tagidx] = he->he_chain[tagidx^1];	/* just copy it's chain pointers */

	    MPRINTF (DBG_DEVICE, 3, "elan4mmu_free_hent: tbl=%d hashidx=%x tagidx=%d he=%p => all free but hashblk\n", tbl, hashidx, tagidx, he);
	}
	else
	{
	    MPRINTF (DBG_DEVICE, 3, "elan4mmu_free_hent: tbl=%d hashidx=%x tagidx=%d he=%p => all free\n", tbl, hashidx, tagidx, he);
	    
	    /* remove it from the hash table, and place back on the anonymous freelist */
	    he->he_chain[tagidx] = he->he_chain[tagidx^1];

	    if (elan4_mmuhash_chain_reduction) {
		elan4mmm_try_to_free_hent (dev, tbl, hashidx, he);
	    }
	}
    }
    else
    {
	/* Other tag still in use */
	MPRINTF (DBG_DEVICE, 3, "elan4mmu_free_hent: tbl=%d hashidx=%x tagidx=%d he=%p => other tag in use\n", tbl, hashidx, tagidx, he);
    }
}
static void
elan4mmu_free_hent (ELAN4_DEV *dev, int tbl, int hashidx, ELAN4_HASH_ENTRY *he, int tagidx)
{
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_mmulock, flags);
    elan4mmu_free_hent_nolock (dev, tbl, hashidx, he, tagidx);
    spin_unlock_irqrestore (&dev->dev_mmulock, flags);
}
void
print_dev(ELAN4_DEV *dev, int tbl, int index)
{
    ELAN4_HASH_ENTRY *he = &dev->dev_mmuhash[tbl][index];
    int count=0;

    while (he) {
	qsnet_debugf(1,"(dev) he%s %p  entry 0x%010lx he_(%p,%p) chain(%p,%p) tag(0x%016llx,0x%016llx) pte(0x%010x,0x%010x)%s\n",
	       (he==&dev->dev_mmuhash[tbl][index])?"*":" ", he, 
	       he->he_entry, he->he_next, he->he_prev, he->he_chain[0], he->he_chain[1],
	       (long long)he->he_tag[0], (long long)he->he_tag[1], he->he_pte[0], he->he_pte[1],
	       (he->he_next)? (( ((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) == 
				 ((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK))?" ":"*"):" ");
	he = he->he_next;

	if (count++ > 1000) {
	    qsnet_debugf(1,"List Failed\n");
	    he = NULL;  
	    elan4_mmuhash_chain_reduction = 0;
	}
    }
}
void
print_ctx(ELAN4_CTXT *ctxt,  int tbl, int index)
{
    ELAN4_HASH_ENTRY *he = ctxt->ctxt_mmuhash[tbl][index];
    ELAN4_DEV *dev = ctxt->ctxt_dev;
    int count=0;

    while (he) {    
	qsnet_debugf(1,"(%04d) he%s %p  entry 0x%010lx he_(%p,%p) chain(%p,%p) tag(0x%016llx,0x%016llx) pte(0x%010x,0x%010x)%s\n",
	       ctxt->ctxt_num, (he==&dev->dev_mmuhash[tbl][index])?"*":" ", he,
	       he->he_entry, he->he_next, he->he_prev, he->he_chain[0], he->he_chain[1],
	       (long long)he->he_tag[0], (long long)he->he_tag[1], he->he_pte[0], he->he_pte[1],  
	       (he->he_next)?(( ((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) == 
				((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK))?" ":"*"):" ");

	if (((he->he_tag[0] & TAG_CONTEXT_MASK) !=  ctxt->ctxt_num) && ((he->he_tag[1] & TAG_CONTEXT_MASK) != ctxt->ctxt_num)) {
	    qsnet_debugf(1,"(%04d) neither tag is us so stopping 0x%llx 0x%llx \n", ctxt->ctxt_num, (long long)(he->he_tag[0] & TAG_CONTEXT_MASK), (long long)(he->he_tag[1] & TAG_CONTEXT_MASK));
	    he = NULL;
	} else {
	    he =  he_ctxt_next (he, ctxt->ctxt_num);
	}
	if (count++ > 1000) {
	    qsnet_debugf(1,"List Failed\n"); 
	    he = NULL; 
	    elan4_mmuhash_chain_reduction = 0;
	}
    }
}
int
dev_count(ELAN4_DEV *dev, int tbl, int index, int ctxt_num)
{
    ELAN4_HASH_ENTRY *he = &dev->dev_mmuhash[tbl][index];
    int               count = 0;
    while (he) {

	if ((he->he_tag[0] & TAG_CONTEXT_MASK) == ctxt_num) count++;
	if ((he->he_tag[1] & TAG_CONTEXT_MASK) == ctxt_num) count++;

	he = he->he_next;
    }
    return (count);
}
int
ctx_count(ELAN4_CTXT *ctxt,  int tbl, int index)
{
    ELAN4_HASH_ENTRY *he = ctxt->ctxt_mmuhash[tbl][index];
    int count = 0;
    while (he) {    

	if ((he->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num) count++;
	if ((he->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num) count++;

	if (((he->he_tag[0] & TAG_CONTEXT_MASK) !=  ctxt->ctxt_num) && ((he->he_tag[1] & TAG_CONTEXT_MASK) != ctxt->ctxt_num)) {
	    he = NULL;
	} else {
	    he =  he_ctxt_next (he, ctxt->ctxt_num);
	}
    }
    return (count);
}
void 
elan4mmu_shuffle_up (ELAN4_CTXT *ctxt, int tbl, int hashidx)
{
    ELAN4_DEV        *dev       = ctxt->ctxt_dev;
    ELAN4_HASH_ENTRY *ctxt_prev = NULL;
    ELAN4_HASH_ENTRY *ctxt_he   = NULL;
    ELAN4_HASH_ENTRY *ctxt_next = NULL;
    ELAN4_HASH_ENTRY *hole;
    ELAN4_HASH_ENTRY *tmp;
    ELAN4_HASH_ENTRY *ctxt_remember;
    int               hole_tagidx;
    int               ctxt_tagidx;
    int               pteidx;
    E4_uint64         value;

    elan4_mmuhash_shuffle_attempts++;

    /* find the first hole */
    hole = elan4mmu_find_next_free ( &dev->dev_mmuhash[tbl][hashidx] ) ;
    if (hole == NULL) return;

    /* find the last ctx */
    /* 1 move tmp to the end */
    for(ctxt_he = hole; (ctxt_he->he_next != NULL); ctxt_he = ctxt_he->he_next)
	;
    /* 2 move tmp back looking for either hole or ctxt */
    while ((ctxt_he != hole)
	   && ((ctxt_he->he_tag[0] & TAG_CONTEXT_MASK) != ctxt->ctxt_num ) 
	   && ((ctxt_he->he_tag[1] & TAG_CONTEXT_MASK) != ctxt->ctxt_num ))
	ctxt_he = ctxt_he->he_prev;

    /* if we found hole then there is not usefull swap to do */
    if (ctxt_he == hole) return;

    while (ctxt_he != hole) {
	/***********/
	/* do swap */
	/***********/
	elan4_mmuhash_shuffle_done++;

	/* now we can move this ctxt's entry in ctxt_he to hole */
	if ( (hole->he_tag[0] & TAG_CONTEXT_MASK) == INVALID_CONTEXT ) hole_tagidx = 0;
	else                                                           hole_tagidx = 1;

	if ( (ctxt_he->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num ) ctxt_tagidx = 0;
	else                                                             ctxt_tagidx = 1;

	/* find us in list do this now before list is played with */
	ctxt_prev = NULL;
	tmp       = ctxt->ctxt_mmuhash[tbl][hashidx];
	ctxt_next = he_ctxt_next (tmp, ctxt->ctxt_num);
	while (tmp != ctxt_he) {
	    ctxt_prev = tmp;
	    tmp       = ctxt_next;
	    ctxt_next = he_ctxt_next (tmp, ctxt->ctxt_num);
	}

	/* copy over software chain and pte */
	hole->he_pte[hole_tagidx] = ctxt_he->he_pte[ctxt_tagidx];

	/* copy over the valid elan pte's */
	/* not preserving the modified and referene bits */
	for (pteidx = 0; pteidx <= 3; pteidx++)
	    if (HE_GET_PTE(hole, hole_tagidx, pteidx))
	    {
		/* copy the pg_page and pg_dma_addr */
		hole->he_pg[hole_tagidx][pteidx] = ctxt_he->he_pg[ctxt_tagidx][pteidx];

		value = elan4mmu_readpte (dev, ctxt_he, ctxt_tagidx, pteidx);
		elan4mmu_writepte (dev, hole, hole_tagidx, pteidx, value);
	    }	    

	/* copy over tag and sync it*/
	hole->he_tag[hole_tagidx] = ctxt_he->he_tag[ctxt_tagidx];
	elan4mmu_synctag (dev, hole, hole_tagidx);
	  
	/* before we remove it check if its going to get free'd */
	if ((ctxt_he->he_tag[ctxt_tagidx ^ 1] & TAG_CONTEXT_MASK) == INVALID_CONTEXT) {
	    /* this is ok as the existence of a hole guards agains falling off front of list */
	    ctxt_remember = ctxt_he->he_prev;
	} else ctxt_remember = ctxt_he;

	/* invalidate pte and tag */
	ctxt_he->he_tag[ctxt_tagidx] = E4MMU_TAG(0, INVALID_CONTEXT);
	elan4mmu_synctag (dev, ctxt_he, ctxt_tagidx);

	/* should ensure that any walk in progress has completed */
	elan4mmu_flush_tlb(dev);

	for (pteidx = 0; pteidx <= 3; pteidx++)
	    if (HE_GET_PTE(ctxt_he, ctxt_tagidx, pteidx))
		elan4mmu_invalidatepte (dev, ctxt_he, ctxt_tagidx, pteidx);

	/* remove from the source end */
	elan4mmu_free_hent_nolock (dev, tbl, hashidx, ctxt_he, ctxt_tagidx);


	/* sort out the ctxt links */
	/* first the hole */
	if ((hole->he_tag[hole_tagidx^1]& TAG_CONTEXT_MASK) ==  ctxt->ctxt_num) {
	    /* already in the list */
	    hole->he_chain[hole_tagidx] = hole->he_chain[hole_tagidx^1];	    
	} else {
	    /* hole not in list */
	    hole->he_chain[hole_tagidx] = ctxt->ctxt_mmuhash[tbl][hashidx];
	    ctxt->ctxt_mmuhash[tbl][hashidx] = hole;

	    /* this is one i missed for a bit                                */
	    /* if we put the hole onto the list it might become the previous */
	    if (ctxt_prev == NULL) ctxt_prev = hole;
	}

	/* second remove the old one */
	if ((ctxt_he->he_tag[ctxt_tagidx^1]& TAG_CONTEXT_MASK) ==  ctxt->ctxt_num) {
	    /* Nothing to do as still in list as other tag is ours */ 
	} else {
	    if (ctxt_prev == NULL) {
		ctxt->ctxt_mmuhash[tbl][hashidx] = ctxt_next;
	    } else {
		if ((ctxt_prev->he_tag[0] & TAG_CONTEXT_MASK) == ctxt->ctxt_num)
		    ctxt_prev->he_chain[0] = ctxt_next;
		if ((ctxt_prev->he_tag[1] & TAG_CONTEXT_MASK) == ctxt->ctxt_num)
		    ctxt_prev->he_chain[1] = ctxt_next;
	    }
	}

	/***********/
	/* move on */
	/***********/
	ctxt_he = ctxt_remember;

	/* the hole is still a valid place to start looking                        */
	/* cant use  elan4mmu_find_next_free as we need to stop if we pass ctxt_he */
	if (hole == ctxt_he) return;
	while ( hole
		&& ((hole->he_tag[0] & TAG_CONTEXT_MASK) != INVALID_CONTEXT)
		&& ((hole->he_tag[1] & TAG_CONTEXT_MASK) != INVALID_CONTEXT)) 
	{
	    hole = hole->he_next;
	    if (hole == ctxt_he) return;
	}
	if (hole == NULL) return;
	
	/* start looking for the next ctxt */
	while ((ctxt_he != hole)
	       && ((ctxt_he->he_tag[0] & TAG_CONTEXT_MASK) != ctxt->ctxt_num ) 
	       && ((ctxt_he->he_tag[1] & TAG_CONTEXT_MASK) != ctxt->ctxt_num )) 
	    ctxt_he = ctxt_he->he_prev;	
    }

    /* if we found hole then there is not usefull swap to do */
    return;
}
void
elan4mmu_do_shuffle (ELAN4_CTXT *ctxt, int tbl)
{
    int               i;
    ELAN4_DEV        *dev;
    unsigned long     flags;

    if (!ctxt) return;
    dev = ctxt->ctxt_dev; 

    spin_lock (&ctxt->ctxt_mmulock);

    for(i=0; i < ELAN4_CTXT_MAX_SHUFFLE ;i++) 
    {
	if (ctxt->shuffle[tbl][i] != -1) 
	{
	    spin_lock_irqsave (&dev->dev_mmulock, flags);

	    elan4mmu_shuffle_up(ctxt, tbl, ctxt->shuffle[tbl][i]);
	    ctxt->shuffle[tbl][i] = -1;
	    
	    spin_unlock_irqrestore (&dev->dev_mmulock, flags);
	}
    }

    ctxt->shuffle_needed[tbl] = 0;
    
    spin_unlock (&ctxt->ctxt_mmulock);
}

ELAN4_HASH_ENTRY *
elan4mmu_ptealloc (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, unsigned int *tagidxp)
{
    ELAN4_DEV        *dev     = ctxt->ctxt_dev;
    unsigned	      ctxnum  = ctxt->ctxt_num;
    unsigned          hashidx = E4MMU_HASH_INDEX (ctxnum, vaddr, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);
    E4_uint64         newtag  = E4MMU_TAG(vaddr, ctxnum);
    ELAN4_HASH_ENTRY *he      = &dev->dev_mmuhash[tbl][hashidx];
    unsigned	      tagidx;

    MPRINTF (ctxt, 2, "elan4mmu_ptealloc: tbl=%d ctxnum=%d vaddr=%llx -> hashidx %d\n", tbl, ctxnum, vaddr, hashidx);

    /* 1st) check whether we're reloading an existing entry */
    for (he = ctxt->ctxt_mmuhash[tbl][hashidx]; he != NULL; he = he_ctxt_next (he, ctxnum))
    {
	ASSERT ((he->he_tag[0] & TAG_CONTEXT_MASK) == ctxnum || (he->he_tag[1] & TAG_CONTEXT_MASK) == ctxnum);

	for (tagidx = 0; tagidx < 2; tagidx++)
	{
	    if ((he->he_tag[tagidx] & (TAG_ADDRESS_MASK | TAG_CONTEXT_MASK | HE_TAG_VALID)) == (newtag | HE_TAG_VALID))
	    {
		MPRINTF (ctxt, 2, "elan4mmu_ptealloc: return old he %p tagidx %d\n", he, tagidx);

		*tagidxp = tagidx;
		return he;
	    }
	}
    }

    if ((he = elan4mmu_alloc_hent (dev, tbl, hashidx, newtag, &tagidx)) == NULL)
	return NULL;

    /* chain onto context hash */
    if ((he->he_tag[tagidx ^ 1] & TAG_CONTEXT_MASK) == ctxnum)	/* already chained using other link */
    {					       			/* so ensure both slots are chained the same */
	he->he_chain[tagidx] = he->he_chain[tagidx^1];
    }
    else
    {
	he->he_chain[tagidx] = ctxt->ctxt_mmuhash[tbl][hashidx];
	ctxt->ctxt_mmuhash[tbl][hashidx] = he;
    }

    MPRINTF (ctxt, 2, "elan4mmu_ptealloc: return new he %p tagidx %d\n", he, tagidx);

    *tagidxp = tagidx;

    return he;
}

int
elan4mmu_pteload (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, int type, E4_uint64 newpte)
{
    ELAN4_DEV        *dev     = ctxt->ctxt_dev;
    unsigned          pteidx  = E4MMU_SHIFT_ADDR(vaddr, dev->dev_pageshift[tbl]) & 3;
    unsigned	      tagidx;
    ELAN4_HASH_ENTRY *he;

    MPRINTF (ctxt, 0, "elan4mmu_pteload: ctx=%d tbl=%d pteidx=%d vaddr=%llx type=%d pte=%llx\n", 
	    ctxt->ctxt_num, tbl, pteidx, (unsigned long long)vaddr, type, newpte);

    spin_lock (&ctxt->ctxt_mmulock);

    if ((he = elan4mmu_ptealloc (ctxt, tbl, vaddr, &tagidx)) == NULL)
    {
	spin_unlock (&ctxt->ctxt_mmulock);
	return -ENOMEM;
    }

    MPRINTF (ctxt, 1, "elan4mmu_pteload: %s he=%p tagidx=%d pteidx=%d\n", HE_GET_PTE(he,0,pteidx) ? "reloading" : "loading", he, tagidx, pteidx);

    if (HE_GET_PTE(he,tagidx,pteidx) != HE_TYPE_INVALID &&						/* invalid -> valid */
	(elan4mmu_readpte (dev, he, tagidx, pteidx) & PTE_PPN_MASK) != (newpte & PTE_PPN_MASK))	/* or same phys address */
    {
	spin_unlock (&ctxt->ctxt_mmulock);
	return -EINVAL;
    }
    
    elan4mmu_writepte (dev, he, tagidx, pteidx, newpte);
    
    HE_SET_PTE(he, tagidx, pteidx, type);

    spin_unlock (&ctxt->ctxt_mmulock);
    return 0;
}

int
elan4mmu_pteload_page (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, struct page *page, int perm)
{
    ELAN4_DEV        *dev    = ctxt->ctxt_dev;
    unsigned int      pteidx = E4MMU_SHIFT_ADDR(vaddr, dev->dev_pageshift[tbl]) & 3;
    unsigned int      tagidx;
    unsigned int      type;
    E4_uint64	      newpte;
    int		      topaddr;
    ELAN4_HASH_ENTRY *he;
    ELAN4_PTE_PAGE   *pg;

    MPRINTF (ctxt, 1, "elan4mmu_pteload_page: ctx=%d tbl=%d pteidx=%d vaddr=%llx page=%p\n",
	     ctxt->ctxt_num, tbl, pteidx, (unsigned long long)vaddr, page);

    spin_lock (&ctxt->ctxt_mmulock);
    if ((he = elan4mmu_ptealloc (ctxt, tbl, vaddr, &tagidx)) == NULL)
    {
	MPRINTF (ctxt, 1, "elan4mmu_pteload_page: ctx=%d failed ENOMEM\n", ctxt->ctxt_num);
	spin_unlock (&ctxt->ctxt_mmulock);
	return -ENOMEM;
    }
    
    pg = &he->he_pg[tagidx][pteidx];
    
    if (HE_GET_PTE(he,tagidx,pteidx) != HE_TYPE_INVALID && pg->pg_page != page)			/* invalid -> valid, or same page*/
    {
	MPRINTF (ctxt, 1, "elan4mmu_pteload_page: ctx=%d failed: pg_page=%p page=%p PTE=%x EINVAL\n",
		 ctxt->ctxt_num, pg->pg_page, page, HE_GET_PTE(he,tagidx,pteidx));

	spin_unlock (&ctxt->ctxt_mmulock);
	return -EINVAL;
    }

    if (HE_GET_PTE (he,tagidx,pteidx) == HE_TYPE_INVALID)
    {
	if ((ctxt->ctxt_features & ELAN4_FEATURE_PIN_DOWN) != 0)
	    page_cache_get (page);

	pg->pg_page = page;
	
	if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_PCI_MAP)
	{
	    struct scatterlist sg;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    sg.address = NULL;
#endif
	    sg.page    = page;
	    sg.offset  = 0;
	    sg.length  = PAGE_SIZE;

	    if (pci_map_sg (dev->dev_osdep.pdev, &sg, 1, PCI_DMA_BIDIRECTIONAL) == 0 || sg.length == 0)
	    {
		spin_unlock (&ctxt->ctxt_mmulock);
		return -EFAULT;
	    }

	    pg->pg_dma_addr = sg.dma_address | (vaddr & (PAGE_SIZE-1));

	    MPRINTF (ctxt, 1, "elan4mmu_pteload_page: pci_map_sg -> %lx\n", pg->pg_dma_addr);
	}
	else
	{
	    pg->pg_dma_addr = (page_to_pfn (page) << PAGE_SHIFT) | (vaddr & (PAGE_SIZE-1));

	    MPRINTF (ctxt, 1, "elan4mmu_pteload_page: directmap -> %lx\n", pg->pg_dma_addr);
	}
    }

#if defined(__BIG_ENDIAN__)
    type = PTE_SetPerm (perm) | PTE_PciNotLocal | PTE_BigEndian;
#else
    type = PTE_SetPerm (perm) | PTE_PciNotLocal;
#endif

    topaddr = elan4mmu_alloc_topaddr (dev, pg->pg_dma_addr, type);

    if (dev->dev_topaddrmode)
	newpte = dev->dev_pteval | (pg->pg_dma_addr >> PTE_PADDR_SHIFT) | (type & ~0xc) | (topaddr << 2);
    else
	newpte = dev->dev_pteval | ((pg->pg_dma_addr >> PTE_PADDR_SHIFT) & ~PTE_TOPADDR_MASK) | (((E4_uint64) topaddr) << 45) | type;

    ASSERT (HE_GET_PTE(he,tagidx,pteidx) == HE_TYPE_INVALID ||						/* invalid -> valid */
	    (elan4mmu_readpte (dev, he, tagidx, pteidx) & PTE_PPN_MASK) == (newpte & PTE_PPN_MASK));	/* or same phys address */
    
    elan4mmu_writepte (dev, he, tagidx, pteidx, newpte);
    
    HE_SET_PTE (he, tagidx, pteidx, HE_TYPE_PAGE);
    
    spin_unlock (&ctxt->ctxt_mmulock);

    return 0;
}

void
elan4mmu_pteunload (ELAN4_CTXT *ctxt, ELAN4_HASH_ENTRY *he, unsigned int tagidx, unsigned int pteidx)
{
    ELAN4_DEV     *dev = ctxt->ctxt_dev;
    ELAN4_PTE_PAGE *pg = &he->he_pg[tagidx][pteidx];

    switch (HE_GET_PTE(he,tagidx,pteidx))
    {
    case HE_TYPE_PAGE:
	MPRINTF (DBG_DEVICE, 1, "elan4mmu_pteunload: he=%p tagidx=%d pteidx=%d page=%p -> %lx\n",
		 he, tagidx, pteidx, pg->pg_page, pg->pg_dma_addr);

	if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_PCI_MAP)
	{
	    struct scatterlist sg;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    sg.address     = NULL;
#endif
	    sg.page        = pg->pg_page;
	    sg.offset      = 0;
	    sg.length      = PAGE_SIZE;
	    sg.dma_address = pg->pg_dma_addr;
	    
	    pci_unmap_sg (dev->dev_osdep.pdev, &sg, 1, PCI_DMA_BIDIRECTIONAL);
	}

	if ((ctxt->ctxt_features & ELAN4_FEATURE_PIN_DOWN) != 0)
	    page_cache_release (pg->pg_page);
	
	pg->pg_page     = NULL;
	pg->pg_dma_addr = 0;
	break;
    }

    HE_SET_PTE(he, tagidx, pteidx, HE_TYPE_INVALID);
				    
    elan4mmu_writepte (dev, he, tagidx, pteidx, 0);
}


void
elan4mmu_unload_range (ELAN4_CTXT *ctxt, int tbl, E4_Addr start, unsigned long len)
{
    ELAN4_DEV        *dev       = ctxt->ctxt_dev;
    unsigned          ctxnum    = ctxt->ctxt_num;
    unsigned long     tagspan   = (1 << (dev->dev_pageshift[tbl] + 2));
    E4_Addr           end       = start + len - 1;
    int		      needflush = 0;
    unsigned          baseidx, topidx;
    unsigned          hashidx, tagidx, pteidx;
    ELAN4_HASH_ENTRY *he, *prevhe, *next;
    
    MPRINTF (ctxt, 0, "elan4mmu_unload_range: tbl=%d start=%llx end=%llx len=%lx\n", tbl, start, end, len);

    /* determine how much of the hash table we've got to scan */
    
    /* GNAT 6760: When we have a Main page size which maps onto multiple Elan pages
     * we need to do something a bit more clever here or else it takes ms per page invalidate
     * This change helps in the meantime
     */
    /* if (len <= (1 << dev->dev_pageshift[tbl])) */
    if (len <= PAGE_SIZE)
    {
	baseidx = E4MMU_HASH_INDEX (ctxnum, start, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);
	topidx  = E4MMU_HASH_INDEX (ctxnum, end,   dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);

	if (baseidx != topidx)
	{
	    /* GNAT 6760: Need to search whole of the hash table (slow!) */
	    baseidx = 0;
	    topidx  = dev->dev_hashsize[tbl] - 1;
	}
    }
    else
    {
	baseidx = 0;
	topidx  = dev->dev_hashsize[tbl] - 1;
    }

    MPRINTF (ctxt, 1, "elan4mmu_unload_range: baseidx=%d topidx=%d\n", baseidx, topidx);

    spin_lock (&ctxt->ctxt_mmulock);

    /* 1st - invalidate the tag for all hash blocks which are completely invalidated,
     *       and remember the first/last hash blocks */
    for (hashidx = baseidx; hashidx <= topidx; hashidx++)
	for (he = ctxt->ctxt_mmuhash[tbl][hashidx]; he != NULL; he = he_ctxt_next (he, ctxnum))
	    for (tagidx = 0; tagidx < 2; tagidx++)
		if ((he->he_tag[tagidx] & TAG_CONTEXT_MASK) == ctxnum)
		{
		    E4_Addr base = E4MMU_TAG2VADDR (he->he_tag[tagidx], hashidx, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);
		    E4_Addr top  = base + (tagspan -1);
		    
		    if (start < top && end > base)
		    {
			unsigned bidx = (start <= base) ? 0 : (start & (tagspan-1)) >> dev->dev_pageshift[tbl];
			unsigned tidx = (end   >= top)  ? 3 : (end   & (tagspan-1)) >> dev->dev_pageshift[tbl];
			
			MPRINTF (ctxt, 1, "elan4mmu_unload_range: he=%p base=%llx top=%llx hashidx=%d bidx=%d tidx=%d\n", he, base, top, hashidx, bidx, tidx);
			
			for (pteidx = bidx; pteidx <= tidx; pteidx++)
			    if (HE_GET_PTE(he, tagidx, pteidx) != HE_TYPE_INVALID)
			    {
				elan4mmu_invalidatepte (dev, he, tagidx, pteidx);
				needflush = 1;
			    }
		    }
		    else if (base >= start && top <= end)		/* hash entry completely spanned */
		    {							/* so invalidate the tag */
			MPRINTF (ctxt, 1, "elan4mmu_unload_range: he=%p base=%llx top=%llx spanned\n", he, base, top);

			he->he_tag[tagidx] &= ~HE_TAG_VALID;
			
			elan4mmu_synctag (dev, he, tagidx);
			needflush = 1;
		    }
		}

    if (needflush)
    {
	/* 2nd invalidate the first/last hash blocks if they are partially invalidated
	 * and flush the tlb/hash copy blocks */
	elan4mmu_flush_tlb_hash (dev, tbl, baseidx, topidx);
	
	/* 3rd free off the hash entries which are completely invalidated */
	for (hashidx = baseidx; hashidx <= topidx; hashidx++)
	    for (prevhe = NULL, he = ctxt->ctxt_mmuhash[tbl][hashidx]; he != NULL; he = next)
	    {
		next = he_ctxt_next (he, ctxnum);
		
		for (tagidx = 0; tagidx < 2; tagidx++)
		    if ((he->he_tag[tagidx] & TAG_CONTEXT_MASK) == ctxnum)
		    {
			E4_Addr base = E4MMU_TAG2VADDR (he->he_tag[tagidx], hashidx, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);
			E4_Addr top  = base + (tagspan -1);
			
			if (start < top && end > base)
			{
			    unsigned bidx = (start <= base) ? 0 : (start & (tagspan-1)) >> dev->dev_pageshift[tbl];
			    unsigned tidx = (end   >= top)  ? 3 : (end   & (tagspan-1)) >> dev->dev_pageshift[tbl];
			    
			    MPRINTF (ctxt, 1, "elan4mmu_unload_range: he=%p base=%llx top=%llx bidx=%d tidx=%d\n", he, base, top, bidx, tidx);
			    
			    for (pteidx = bidx; pteidx <= tidx; pteidx++)
				if (HE_GET_PTE(he, tagidx, pteidx) != HE_TYPE_INVALID)
				    elan4mmu_pteunload (ctxt, he,tagidx, pteidx);
			}
			
			if ((base >= start && top <= end) || he->he_pte[tagidx] == 0) 	/* hash entry completely spanned or all pte's cleared */
			{									/* so invalidate the pte's and free it */
			    
			    MPRINTF (ctxt, 1, "elan4mmu_unload_range: he=%p base=%llx top=%llx spanned or empty\n", he, base, top);
			    
			    elan4mmu_free_hent (dev, tbl, hashidx, he, tagidx);
			}
		    }
		
		prevhe = he_ctxt_unlink (ctxt, tbl, hashidx, prevhe, he, next);
	    }
    }
    spin_unlock (&ctxt->ctxt_mmulock);
}

void
elan4mmu_invalidate_ctxt (ELAN4_CTXT *ctxt)
{
    ELAN4_DEV *dev    = ctxt->ctxt_dev;
    int        ctxnum = ctxt->ctxt_num;
    ELAN4_HASH_ENTRY *he;
    int tbl, hashidx, tagidx, pteidx;

    MPRINTF (ctxt, 0, "elan4mmu_invalidate_ctxt: invalidating ctxnum=%d\n", ctxnum);

    spin_lock (&ctxt->ctxt_mmulock);

    /* 1st invalidate all tags belonging to me */
    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
	for (hashidx = 0; hashidx < dev->dev_hashsize[tbl]; hashidx++)
	    for (he = ctxt->ctxt_mmuhash[tbl][hashidx]; he != NULL; he = he_ctxt_next (he, ctxnum))
		for (tagidx = 0; tagidx < 2; tagidx++)
		    if ((he->he_tag[tagidx] & TAG_CONTEXT_MASK) == ctxnum) /* own tag block */
		    {
			MPRINTF (ctxt, 1, "elan4mmu_invalidate_ctxt: he=%p addr=%llx hashidx=%d tagidx=%d\n", 
				 he, he->he_tag[tagidx] & TAG_ADDRESS_MASK, hashidx, tagidx);

			he->he_tag[tagidx] &= ~HE_TAG_VALID;
			
			elan4mmu_synctag (dev, he, tagidx);
		    }

    /* 2nd flush the tlb & cached hash block */
    elan4mmu_flush_tlb (dev);
    
    /* 3rd invalidate all pte's and free off the hash entries */
    for (tbl = 0; tbl < NUM_HASH_TABLES; tbl++)
	for (hashidx = 0; hashidx < dev->dev_hashsize[tbl]; hashidx++)
	    while ((he = ctxt->ctxt_mmuhash[tbl][hashidx]) != NULL)
	    {
		ctxt->ctxt_mmuhash[tbl][hashidx] = he_ctxt_next (he, ctxnum);

		for (tagidx = 0; tagidx < 2; tagidx++)
		    if ((he->he_tag[tagidx] & TAG_CONTEXT_MASK) == ctxnum)
		    {
			for (pteidx = 0; pteidx < 4; pteidx++)
			    if (HE_GET_PTE(he, tagidx, pteidx) != HE_TYPE_INVALID)
				elan4mmu_pteunload (ctxt, he, tagidx, pteidx);

			elan4mmu_free_hent (dev, tbl, hashidx, he, tagidx);
		    }
	    }
    spin_unlock (&ctxt->ctxt_mmulock);
}

ELAN4_HASH_CACHE *
elan4mmu_reserve (ELAN4_CTXT *ctxt, int tbl, E4_Addr start, unsigned int npages, int cansleep)
{
    ELAN4_DEV        *dev      = ctxt->ctxt_dev;
    E4_Addr           end      = start + (npages << dev->dev_pageshift[tbl]) - 1;
    unsigned long     tagshift = dev->dev_pageshift[tbl] + 2;
    E4_Addr           tagspan  = 1 << tagshift;
    E4_Addr           base     = (start & ~(tagspan-1));
    E4_Addr           top      = (end   & ~(tagspan-1)) + (tagspan-1);
    unsigned int      nhes     = (top - base + 1) >> tagshift;
    ELAN4_HASH_CACHE *hc;
    unsigned int      tagidx,  pteidx;
    E4_Addr           addr;
    int		      i;
    
    MPRINTF (ctxt, 0, "elan4mmu_reserve: start=%llx npages=%d\n", start, npages);
    MPRINTF (ctxt, 0, "         pageshift=%d tagspan=%lx base=%llx top=%llx end=%llx nhes=%d\n",
	     dev->dev_pageshift[tbl], tagspan, base, top, end, nhes);

    KMEM_ALLOC (hc, ELAN4_HASH_CACHE *, offsetof (ELAN4_HASH_CACHE, hc_hes[nhes]), cansleep);

    if (hc == NULL)
	return NULL;

    hc->hc_start = start;
    hc->hc_end   = end;
    hc->hc_tbl   = tbl;

    spin_lock (&ctxt->ctxt_mmulock);
    for (addr = base, i = 0; i < nhes; addr += tagspan, i++)
    {
	unsigned bidx = (i == 0)        ? (start & (tagspan-1)) >> dev->dev_pageshift[tbl] : 0;
	unsigned tidx = (i == (nhes-1)) ? (end   & (tagspan-1)) >> dev->dev_pageshift[tbl] : 3;

	
	if ((hc->hc_hes[i] = elan4mmu_ptealloc (ctxt, tbl, addr & ~(tagspan-1), &tagidx)) == NULL)
	    goto failed;


	MPRINTF (ctxt, 2, "elan4mmu_reserve: tbl=%d addr=%llx -> hashidx=%d tagidx=%d\n", tbl, addr & ~(tagspan-1), 
		 E4MMU_HASH_INDEX (ctxt->ctxt_num, (addr & ~(tagspan-1)), dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1), tagidx);
			 
	for (pteidx = bidx; pteidx <= tidx; pteidx++)
	{
	    ASSERT (HE_GET_PTE (hc->hc_hes[i], tagidx, pteidx) == HE_TYPE_INVALID);

	    MPRINTF (ctxt, 2, "elan4mmu_reserve: i=%d addr=%llx he=%p (tagidx=%d pteidx=%d)\n",
		     i, addr, hc->hc_hes[i], tagidx, pteidx);

	    HE_SET_PTE (hc->hc_hes[i], tagidx, pteidx, HE_TYPE_RESERVED);
	}
    }
    spin_unlock (&ctxt->ctxt_mmulock);

    return hc;

 failed:
    for (i--, addr -= tagspan; i >= 0; i--, addr -= tagspan)
    {
	unsigned bidx    = (i == 0) ? (start & (tagspan-1)) >> dev->dev_pageshift[tbl] : 0;
	unsigned tidx    = (i == (nhes-1)) ? (end   & (tagspan-1)) >> dev->dev_pageshift[tbl] : 3;
	unsigned hashidx = E4MMU_HASH_INDEX (ctxt->ctxt_num, addr, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1);
	unsigned tagidx  = (addr == E4MMU_TAG2VADDR (hc->hc_hes[i]->he_tag[0], hashidx, dev->dev_pageshift[tbl], dev->dev_hashsize[tbl]-1)) ? 0 : 1;
	    
	for (pteidx = bidx; pteidx <= tidx; pteidx++)
	    HE_SET_PTE(hc->hc_hes[i], tagidx, pteidx, HE_TYPE_INVALID);

	if (hc->hc_hes[i]->he_pte[tagidx] == 0)
	    elan4mmu_free_hent (dev, tbl, hashidx, hc->hc_hes[i], tagidx);
    }
    spin_unlock (&ctxt->ctxt_mmulock);

    KMEM_FREE (hc, offsetof (ELAN4_HASH_CACHE, hc_hes[nhes]));
    
    return NULL;
}

void
elan4mmu_release (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc)
{
    ELAN4_DEV        *dev      = ctxt->ctxt_dev;
    E4_Addr	      start    = hc->hc_start;
    E4_Addr           end      = hc->hc_end;
    unsigned long     tagshift = dev->dev_pageshift[hc->hc_tbl] + 2;
    E4_Addr           tagspan  = 1 << tagshift;
    E4_Addr           base     = (start & ~(tagspan-1));
    E4_Addr           top      = (end   & ~(tagspan-1)) + (tagspan-1);
    unsigned int      nhes     = (top - base + 1) >> tagshift;
    ELAN4_HASH_ENTRY *prevhe, *he, *next;
    E4_Addr           addr;
    unsigned int      pteidx;
    int		      i;

    spin_lock (&ctxt->ctxt_mmulock);

    MPRINTF (ctxt, 0, "elan4mmu_release: base=%llx top=%llx\n", base, top);

    for (addr = base, i = 0; i < nhes; addr += tagspan, i++)
    {
	unsigned bidx    = (i == 0)        ? (start & (tagspan-1)) >> dev->dev_pageshift[hc->hc_tbl] : 0;
	unsigned tidx    = (i == (nhes-1)) ? (end   & (tagspan-1)) >> dev->dev_pageshift[hc->hc_tbl] : 3;
	unsigned hashidx = E4MMU_HASH_INDEX (ctxt->ctxt_num, addr, dev->dev_pageshift[hc->hc_tbl], dev->dev_hashsize[hc->hc_tbl]-1);
	unsigned tagidx  = (addr == E4MMU_TAG2VADDR (hc->hc_hes[i]->he_tag[0], hashidx, dev->dev_pageshift[hc->hc_tbl], dev->dev_hashsize[hc->hc_tbl]-1)) ? 0 : 1;
	    
	for (pteidx = bidx; pteidx <= tidx; pteidx++)
	{
	    elan4mmu_invalidatepte (dev, hc->hc_hes[i], tagidx, pteidx);

	    HE_SET_PTE(hc->hc_hes[i], tagidx, pteidx, HE_TYPE_INVALID);
	}

	MPRINTF (ctxt, 2, "elan4mmu_release: i=%d addr=%llx he=%p (hashidx=%d tagidx=%d pteidx=%d) pte=%x\n",
		 i, addr, hc->hc_hes[i], hashidx, tagidx, pteidx, hc->hc_hes[i]->he_pte[tagidx]);

	/* remove from context hash */
	/* need to move to the  hc->hc_hes[i] in the ctxt list and set prevhe, he, next */
	prevhe = NULL;
	he = ctxt->ctxt_mmuhash[hc->hc_tbl][hashidx];
	next = he_ctxt_next (he, ctxt->ctxt_num);

	while(he != hc->hc_hes[i]) {
	    prevhe = he;
	    he = next;
	    next = he_ctxt_next (he, ctxt->ctxt_num);
	}

	if (he->he_pte[tagidx] == 0) 
	    elan4mmu_free_hent (dev, hc->hc_tbl, hashidx, he, tagidx);

	he_ctxt_unlink (ctxt, hc->hc_tbl, hashidx, prevhe, he, next);
    }
    spin_unlock (&ctxt->ctxt_mmulock);
}

void
elan4mmu_set_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx, E4_uint64 newpte)
{
    ELAN4_DEV        *dev      = ctxt->ctxt_dev;
    unsigned int      tbl      = hc->hc_tbl;
    unsigned int      tagshift = dev->dev_pageshift[tbl] + 2;
    E4_Addr           tagspan  = 1 << tagshift;
    E4_Addr           addr     = hc->hc_start + (idx << dev->dev_pageshift[tbl]);
    ELAN4_HASH_ENTRY *he       = hc->hc_hes[(addr - (hc->hc_start & ~(tagspan-1))) >> tagshift];
    unsigned          pteidx   = E4MMU_SHIFT_ADDR(addr, dev->dev_pageshift[tbl]) & 3;
    unsigned          tagidx   = he->he_tag[0] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID) ? 0 : 1;

    MPRINTF (ctxt, 2, "elan4mmu_set_pte: idx=%d addr=%llx he=%p (tagidx=%d pteidx=%d) newpte=%llx\n", idx, addr, he, tagidx, pteidx, newpte);

    ASSERT (he->he_tag[tagidx] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID));

    elan4mmu_writepte (dev, he, tagidx, pteidx, newpte);
}

E4_uint64
elan4mmu_get_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx)
{
    ELAN4_DEV        *dev      = ctxt->ctxt_dev;
    unsigned int      tbl      = hc->hc_tbl;
    unsigned int      tagshift = dev->dev_pageshift[tbl] + 2;
    E4_Addr           tagspan  = 1 << tagshift;
    E4_Addr           addr     = hc->hc_start + (idx << dev->dev_pageshift[tbl]);
    ELAN4_HASH_ENTRY *he       = hc->hc_hes[(addr - (hc->hc_start & ~(tagspan-1))) >> tagshift];
    unsigned          pteidx   = E4MMU_SHIFT_ADDR(addr, dev->dev_pageshift[tbl]) & 3;
    unsigned          tagidx   = he->he_tag[0] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID) ? 0 : 1;

    ASSERT (he->he_tag[tagidx] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID));

    return elan4mmu_readpte (dev, he, tagidx, pteidx);
}

void
elan4mmu_clear_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx)
{
    ELAN4_DEV        *dev      = ctxt->ctxt_dev;
    unsigned int      tbl      = hc->hc_tbl;
    unsigned int      tagshift = dev->dev_pageshift[tbl] + 2;
    E4_Addr           tagspan  = 1 << tagshift;
    E4_Addr           addr     = hc->hc_start + (idx << dev->dev_pageshift[tbl]);
    ELAN4_HASH_ENTRY *he       = hc->hc_hes[(addr - (hc->hc_start & ~(tagspan-1))) >> tagshift];
    unsigned          pteidx   = E4MMU_SHIFT_ADDR(addr, dev->dev_pageshift[tbl]) & 3;
    unsigned          tagidx   = he->he_tag[0] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID) ? 0 : 1;

    MPRINTF (ctxt, 2, "elan4mmu_clear_pte: idx=%d addr=%llx he=%p (tagidx=%d pteidx=%d)\n", idx, addr, he, tagidx, pteidx);

    ASSERT (he->he_tag[tagidx] == (E4MMU_TAG (addr, ctxt->ctxt_num) | HE_TAG_VALID));

    elan4mmu_invalidatepte (dev, he, tagidx, pteidx);
}

int 
elan4mmu_display_mmuhash(ELAN4_DEV *dev, int tbl, int *index_ptr, char *page, int count)
{
    char             *p = page;
    unsigned long     flags;
    ELAN4_HASH_ENTRY *he;
    int               index = *index_ptr;

    spin_lock_irqsave (&dev->dev_mmulock, flags);

    he = &dev->dev_mmuhash[tbl][index];

    /* move to the next entry that actually has contents in its chain */
    while ((he->he_next == NULL) && ((he->he_tag[0] & TAG_CONTEXT_MASK) == INVALID_CONTEXT)
	   && ((he->he_tag[1] & TAG_CONTEXT_MASK) == INVALID_CONTEXT))
    {
	index++;
	if ( index >= dev->dev_hashsize[tbl] ) {
	    /* didnt find anything and have looped */
	    spin_unlock_irqrestore (&dev->dev_mmulock, flags);
	    *index_ptr = dev->dev_hashsize[tbl];
	    return (p - page);
	}
	he = &dev->dev_mmuhash[tbl][index];
    }
    *index_ptr = index; /* the actual one we will print */


    while (he) {
	if ( ((p - page)+175) > count ) {
	    /* might not fit in */
	    p += sprintf( p , "...\n");
	    he = NULL;
	} else {
	    int ctxt0_bit = 0;
	    int ctxt1_bit = 0;
	    ELAN4_CTXT *ctxt0;
	    ELAN4_CTXT *ctxt1;

	    if ( (he->he_tag[0] &  TAG_CONTEXT_MASK) != INVALID_CONTEXT) {
		ctxt0 = elan4_localctxt (dev, (he->he_tag[0] &  TAG_CONTEXT_MASK));
		ctxt0_bit = ctxt0->shuffle_needed[0];
	    }
	    if ( (he->he_tag[1] &  TAG_CONTEXT_MASK) != INVALID_CONTEXT) {
		ctxt1 = elan4_localctxt (dev, (he->he_tag[1] &  TAG_CONTEXT_MASK));
		ctxt1_bit = ctxt1->shuffle_needed[0];
	    }

	    p += sprintf(p ,"(%4d,%1d) he%s %p  entry 0x%010lx he_(%p,%p) chain(%p,%p) tag(0x%016llx-%d,0x%016llx-%d) pte(0x%010x,0x%010x)%s\n",
			 index,tbl, (he==&dev->dev_mmuhash[tbl][index])?"*":" ", he, 
			 he->he_entry, he->he_next,  he->he_prev, he->he_chain[0], he->he_chain[1],
			 (long long)he->he_tag[0], ctxt0_bit, (long long)he->he_tag[1], ctxt1_bit, he->he_pte[0], he->he_pte[1],
			 (he->he_next)? (( ((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) == 
					   ((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK))?" ":"*"):" ");
	    he = he->he_next;
	}
    }

    spin_unlock_irqrestore (&dev->dev_mmulock, flags);

    return (p - page);
}

int 
elan4mmu_display_ctxt_mmuhash(ELAN4_CTXT *ctxt, int tbl, int *index_ptr, char *page, int count)
{
    ELAN4_HASH_ENTRY *he;
    char             *p      = page;
    ELAN4_DEV        *dev    = ctxt->ctxt_dev;
    int               index  = *index_ptr;

    spin_lock (&ctxt->ctxt_mmulock);

    he = ctxt->ctxt_mmuhash[tbl][index];
    while (! he ) { 
	index++;
	if ( index >= dev->dev_hashsize[tbl] ) {
	    /* didnt find anything and have looped */
	    spin_unlock (&ctxt->ctxt_mmulock);
	    *index_ptr = dev->dev_hashsize[tbl];
	    return (p - page);
	}
	he = ctxt->ctxt_mmuhash[tbl][index];	
    }
    *index_ptr = index; /* the actual one we will print */

    while (he) {
	if ( ((p - page)+175) > count ) {
	    /* might not fit in */
	    p += sprintf( p , "...\n");
	    he = NULL;
	} else {
	    p  += sprintf(p ,"(%4d,%1d) he%s %p  entry 0x%010lx he_(%p,%p) chain(%p,%p) tag(0x%016llx,0x%016llx) pte(0x%010x,0x%010x)%s\n",
			  index,tbl, (he==&dev->dev_mmuhash[tbl][index])?"*":" ", he,
			  he->he_entry, he->he_next, he->he_prev, he->he_chain[0], he->he_chain[1],
			  (long long)he->he_tag[0], (long long)he->he_tag[1], he->he_pte[0], he->he_pte[1],  
			  (he->he_next)?(( ((he->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK) == 
					   ((he->he_next->he_entry >> TAG_CHAINPTR_HIGH_SHIFT) & TAG_CHAINPTR_30TO19_MASK))?" ":"*"):" ");

	    he =  he_ctxt_next (he, ctxt->ctxt_num);
	}
    }

    spin_unlock (&ctxt->ctxt_mmulock);

    return (p - page);
}

int 
elan4mmu_display_bucket_mmuhash(ELAN4_DEV *dev, int tbl, int *buckets, int nBuckets, char *page, int c)
{   
    ELAN4_HASH_ENTRY *he;
    unsigned long     flags;
    char *p = page;
    int   b;
    int   index;
    int   count;
    int   totals[nBuckets];

    for (b=0;b<nBuckets;b++) 
	totals[b]=0;

    for (index=0; index < dev->dev_hashsize[tbl]; index++) {

	/* how long is this chain */
	spin_lock_irqsave (&dev->dev_mmulock, flags);

	he = &dev->dev_mmuhash[tbl][index];
	count = 0;
	while (he) {
	    count++;
	    ASSERT(count < 1000000); /* seems we have a loop */
	    he = he->he_next;
	}
	spin_unlock_irqrestore (&dev->dev_mmulock, flags);

	/* bucket the lenth */
	for(b=0;b<nBuckets;b++) 
	    if ( count <= buckets[b] ) {
		totals[b]++;
		b = nBuckets;
	    }
    }

    for (b=0; b < nBuckets; b++) 
	p += sprintf(p, " %5d ", buckets[b]);
    p += sprintf(p, "\n");
    for (b=0; b < nBuckets; b++) 
	p += sprintf(p, " %5d ", totals[b]);
    p += sprintf(p, "\n");

    return (p - page);
}

EXPORT_SYMBOL(elan4mmu_flush_tlb);
EXPORT_SYMBOL(elan4mmu_pteload);
EXPORT_SYMBOL(elan4mmu_unload_range);
EXPORT_SYMBOL(elan4mmu_reserve);
EXPORT_SYMBOL(elan4mmu_release);
EXPORT_SYMBOL(elan4mmu_set_pte);
EXPORT_SYMBOL(elan4mmu_get_pte);
EXPORT_SYMBOL(elan4mmu_clear_pte);
EXPORT_SYMBOL(elan4mmu_do_shuffle);
EXPORT_SYMBOL(elan4mmu_set_shuffle);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
