/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: mmu_Linux.c,v 1.17.2.1 2005/07/14 13:34:12 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/mmu_Linux.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>

#include <linux/pci.h>
#include <linux/version.h>

int
elan4mmu_sdram_aliascheck (ELAN4_CTXT *ctxt, E4_Addr addr, sdramaddr_t phys)
{
    ELAN4_DEV *dev = ctxt->ctxt_dev;

    /*
     * On MPSAS we don't allocate a large enough context table, so 
     * if we see an address/context pair which would "alias" because
     * they differ in unchecked hash bits to a previous pteload, 
     * then we kill the application.
     */
    unsigned hashval = (E4MMU_SHIFT_ADDR(addr, (dev->dev_pageshift[0]) + 2) ^ E4MMU_CONTEXT_SCRAMBLE(ctxt->ctxt_num));
    
    if (dev->dev_rsvd_hashval[0] == 0xFFFFFFFF)
	dev->dev_rsvd_hashval[0] = hashval & dev->dev_rsvd_hashmask[0];
    
    if ((hashval & dev->dev_rsvd_hashmask[0]) != dev->dev_rsvd_hashval[0])
    {
	printk ("elan4mmu_sdram_aliascheck: vaddr=%016llx ctxnum=%x -> [%x] overlaps %x - %x [hashidx=%x]\n", (unsigned long long) addr, 
		ctxt->ctxt_num, hashval, hashval & dev->dev_rsvd_hashmask[0], dev->dev_rsvd_hashval[0],
		E4MMU_HASH_INDEX (ctxt->ctxt_num, addr, dev->dev_pageshift[0], dev->dev_hashsize[0]-1));
	
	return 0;
    }

    if (((addr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT)) != (phys & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT))))
    {
	printk ("elan4mmu_sdram_aliascheck: vaddr=%016llx incorrectly alias sdram at %lx\n", (unsigned long long) addr, phys);
	return 0;
    }

    return 1;
}

int
elan4mmu_alloc_topaddr (ELAN4_DEV *dev, physaddr_t paddr, unsigned type)
{
#if defined(__i386) && !defined(CONFIG_X86_PAE)
    if (dev->dev_topaddrvalid == 0)
    {
	dev->dev_topaddrvalid = 1;

	pci_write_config_word (dev->dev_osdep.pdev, PCI_ELAN_TOPPHYSADDR(0), 0);
	pci_write_config_word (dev->dev_osdep.pdev, PCI_ELAN_TOPPHYSADDR(1), 0);
	pci_write_config_word (dev->dev_osdep.pdev, PCI_ELAN_TOPPHYSADDR(2), 0);
	pci_write_config_word (dev->dev_osdep.pdev, PCI_ELAN_TOPPHYSADDR(3), 0);
    }
    return (0);
#else
    register int i;
    E4_uint16 match;

    if (dev->dev_topaddrmode)					/* ExtraMasterAddrBits=1 => match {paddr[63:50],type[3:2]} */
	match = ((paddr >> 48) & ~3) | ((type >> 2) & 3);
    else							/* ExtraMasterAddrBits=0 => match {paddr[63:48]} */
	match = (paddr >> 48);
    
    MPRINTF (DBG_DEVICE, 2, "elan4mmu_alloc_topaddr: mode=%d paddr=%lx type=%x match=%x [%x %x.%x.%x.%x]\n",
	     dev->dev_topaddrmode, paddr, type, match, dev->dev_topaddrvalid,
	     dev->dev_topaddr[0], dev->dev_topaddr[1], dev->dev_topaddr[2], dev->dev_topaddr[3]);
    
    for (i = 0; i < 4; i++)
	if ((dev->dev_topaddrvalid & (1 << i)) && dev->dev_topaddr[i] == match)
	    return (i);
    
    for (i = 0; i < 4; i++)
    {
	if ((dev->dev_topaddrvalid & (1 << i)) == 0)
	{
	    MPRINTF (DBG_DEVICE, 2, "elan4mmu_alloc_topaddr: allocate slot %d for %x\n", i, match);

	    dev->dev_topaddrvalid |= (1 << i);
	    dev->dev_topaddr[i] = match;

	    pci_write_config_word (dev->dev_osdep.pdev, PCI_ELAN_TOPPHYSADDR(i), match);
	    return (i);
	}
    }

    panic ("elan4mmu_alloc_topaddr: all topaddrs in use\n");
    return (0);
#endif
}

/*
 * Convert a physical address into an pte.  This should generate a "local" pte for 
 * physical addresses which are elan4 sdram or elan4 command queues.  For elan4
 * registers and other addresses on the same bus, this should be the local pci 
 * bus address.  All other addresses should access the physical address via the
 * PCI bridge.
 */

int
elan4mmu_categorise_paddr (ELAN4_DEV *dev, physaddr_t *physp)
{
    physaddr_t sdram_base = dev->dev_sdram_phys;
    physaddr_t sdram_top  = dev->dev_sdram_phys + pci_resource_len (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM);
    physaddr_t regs_base  = dev->dev_regs_phys;
    physaddr_t regs_top   = dev->dev_regs_phys + pci_resource_len (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS);
    physaddr_t phys       = *physp;
    int        iscommand;

    if (phys >= sdram_base && phys <= sdram_top)
    {
	(*physp) = (phys ^ sdram_base);
	return HE_TYPE_SDRAM;
    }
    
    if (phys >= regs_base && phys < regs_top)
    {
	if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	    iscommand = (phys < (regs_base + ELAN4_REVA_REG_OFFSET));
	else
	    iscommand = (phys < (regs_base + ELAN4_REVB_I2C_OFFSET));
	
	if (iscommand)
	{
	    (*physp) = phys ^ regs_base;

	    return HE_TYPE_COMMAND;
	}
	else
	{
	    u32 blow, bhigh;

	    /* compute a local pci address from our register BAR */
	    pci_read_config_dword (dev->dev_osdep.pdev, PCI_BASE_ADDRESS_2, &blow);
	    pci_read_config_dword (dev->dev_osdep.pdev, PCI_BASE_ADDRESS_3, &bhigh);

	    (*physp) = (((physaddr_t) bhigh) << 32) | (blow & PCI_BASE_ADDRESS_MEM_MASK) | (phys ^ regs_base);

	    return HE_TYPE_REGS;
	}
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    if (VALID_PAGE (virt_to_page (phys_to_virt (phys))))
#else
    if (virt_addr_valid (phys_to_virt (phys)))
#endif
	return HE_TYPE_PAGE;
    
    return HE_TYPE_OTHER;
}

E4_uint64
elan4mmu_phys2pte (ELAN4_DEV *dev, physaddr_t phys, unsigned perm)
{
    unsigned int type = 0;
    E4_uint64    pte;

    switch (elan4mmu_categorise_paddr (dev, &phys))
    {
    case HE_TYPE_SDRAM:
	type = PTE_SetPerm (perm);
	break;
	
    case HE_TYPE_COMMAND:
	type = PTE_SetPerm (perm) | PTE_CommandQueue;
	break;
	
    case HE_TYPE_REGS:
	type = PTE_SetPerm (perm) | PTE_PciNotLocal;
	break;

    case HE_TYPE_PAGE:
	if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_PCI_MAP)
	{
	    struct scatterlist list;
	    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    list.address = NULL;
#endif
	    list.page    = virt_to_page (phys_to_virt (phys));;
	    list.offset  = (phys & (PAGE_SIZE-1));
	    list.length  = (1 << dev->dev_pageshift[0]);
	    
	    if (pci_map_sg (dev->dev_osdep.pdev, &list, 1, PCI_DMA_BIDIRECTIONAL) == 0)
	    {
		printk ("elan4mmu_phys2pte: pci_map_sg failed\n");
		return -EFAULT;
	    }

	    type = PTE_SetPerm (perm) | PTE_PciNotLocal | dev->dev_pteval;
	    phys = list.dma_address;
	    break;
	}
	/* DROPTHROUGH */
	
    case HE_TYPE_OTHER:
	if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_PCI_MAP)
	    return -EFAULT;

	type = PTE_SetPerm (perm) | PTE_PciNotLocal | dev->dev_pteval;
	break;
    }

    if ((type & PTE_PciNotLocal) == 0)
	pte = (phys >> PTE_PADDR_SHIFT) | type;
    else
    {
	unsigned topaddr = elan4mmu_alloc_topaddr (dev, phys, type);
	
	if (dev->dev_topaddrmode)
	    pte = (phys >> PTE_PADDR_SHIFT) | (type & ~0xc) | (topaddr << 2);
	else
	    pte = ((phys >> PTE_PADDR_SHIFT) & ~PTE_TOPADDR_MASK) | (((E4_uint64) topaddr) << 45) | type;
    }

    return pte;
}

physaddr_t
elan4mmu_pte2phys (ELAN4_DEV *dev, E4_uint64 pte)
{
    physaddr_t sdram_base = dev->dev_sdram_phys;
    physaddr_t regs_base  = dev->dev_regs_phys;
    physaddr_t phys;
    
    if (pte & PTE_PciNotLocal)
    {
	if (dev->dev_topaddrmode)
	    phys = ((physaddr_t)(dev->dev_topaddr[(pte >> 2) & 3] & 0xfffc) << 48) | ((pte & PTE_PPN_MASK) << PTE_PADDR_SHIFT);
	else
	    phys = ((physaddr_t)(dev->dev_topaddr[(pte >> 45) & 3] & 0xffff) << 48)| ((pte & PTE_PPN_MASK & ~PTE_TOPADDR_MASK) << PTE_PADDR_SHIFT);

#if defined(__alpha)
	    phys ^= alpha_mv.pci_dac_offset;
#elif defined(__sparc)
	    phys ^= 0xfffe000000000000;
#endif
	return phys;
    }
    
    if (pte & PTE_CommandQueue)
	return (regs_base | ((pte & PTE_PPN_MASK) << PTE_PADDR_SHIFT));
    
    /* sdram */
    return (sdram_base | ((pte & PTE_PPN_MASK) << PTE_PADDR_SHIFT));
}

EXPORT_SYMBOL(elan4mmu_phys2pte);
EXPORT_SYMBOL(elan4mmu_pte2phys);
