/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: neterr.c,v 1.8.2.1 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/neterr.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/sdram.h>
#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/commands.h>
#include <elan4/trtype.h>
#include <elan4/neterr.h>

typedef struct neterr_inputq
{
    E4_InputQueue	inputq;					/* input queue */
    E4_Event32		qevent;					/* input queue event */
    E4_uint64		sent;					/* # messages sent (cq flow control)*/
} NETERR_INPUTQ;

#define NETERR_NSLOTS	64					/* single page of queue space (4Kb) */

#define NETERR_RETRIES	16
#define NETERR_CQ_SIZE	CQ_Size8K
#define NETERR_CQ_MSGS	(CQ_Size(NETERR_CQ_SIZE) / (21*8))
#define NETERR_VP_COUNT	64					/* this *must* be > NETERR_CQ_MSGS */
#define NETERR_VP_BASE	1					/* use vp 1 upwards */

void
elan4_neterr_interrupt (ELAN4_DEV *dev, void *arg)
{
    E4_Addr          qfptr  = elan4_sdram_readq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_fptr));
    E4_Addr          qbptr  = elan4_sdram_readq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_bptr));
    E4_Addr          qfirst = DEVICE_NETERR_SLOTS_ADDR;
    E4_Addr          qlast  = qfirst + (NETERR_NSLOTS-1) * ELAN4_NETERR_MSG_SIZE;
    ELAN4_CQ        *cq     = dev->dev_neterr_intcq;
    int              count  = 0;
    ELAN4_CTXT      *ctxt;
    ELAN4_NETERR_MSG msg;

    while (qfptr != qbptr)
    {
	elan4_sdram_copyq_from_sdram (dev, dev->dev_neterr_slots + (qfptr - qfirst), &msg, ELAN4_NETERR_MSG_SIZE);

	ctxt = elan4_networkctxt (dev, msg.msg_context);

	if (ctxt != NULL && ctxt->ctxt_ops->op_neterrmsg)
	    ctxt->ctxt_ops->op_neterrmsg (ctxt, &msg);
	else
	    PRINTF (DBG_DEVICE, DBG_NETERR, "elan4_neterr_interrupt: no process - sender %d.%d\n", msg.msg_sender.loc_node, msg.msg_sender.loc_context);

	count++;

	/* move on the from pointer */
	qfptr = (qfptr == qlast) ? qfirst : qfptr + ELAN4_NETERR_MSG_SIZE;

	elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_fptr), qfptr);
    }

    if (count == 0)
    {
	printk ("elan4_neterr_interrupt: spurious\n");
	return;
    }

    /* Issue the waitevent to the interrupt queue */
    writeq (WAIT_EVENT_CMD | (DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, qevent)),	(void *)cq->cq_mapping);
    writeq (  E4_EVENT_INIT_VALUE (-32 * count, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0),	(void *)cq->cq_mapping);
    writeq (  DEVICE_NETERR_INTCQ_ADDR,								(void *)cq->cq_mapping);
    writeq (INTERRUPT_CMD | (dev->dev_neterr_intop.op_cookie << E4_MAIN_INT_SHIFT), 		(void *)cq->cq_mapping);

    pioflush_reg (dev);
}

int
elan4_neterr_init (ELAN4_DEV *dev)
{
    unsigned int intqaddr;
    E4_Addr	 qfirst, qlast;
    
    if ((dev->dev_neterr_inputq = elan4_sdram_alloc (dev, SDRAM_PAGE_SIZE)) == 0)
	return 0;

    if ((dev->dev_neterr_slots = elan4_sdram_alloc (dev, roundup (NETERR_NSLOTS * ELAN4_NETERR_MSG_SIZE, SDRAM_PAGE_SIZE))) == 0)
	return 0;

    if ((dev->dev_neterr_msgcq = elan4_alloccq (&dev->dev_ctxt, NETERR_CQ_SIZE, CQ_STENEnableBit | CQ_WriteEnableBit, CQ_Priority)) == NULL)
	return 0;

    if ((dev->dev_neterr_intcq = elan4_alloccq (&dev->dev_ctxt, CQ_Size1K, CQ_WaitEventEnableBit | CQ_InterruptEnableBit, CQ_Priority)) == NULL)
	return 0;

    intqaddr = (dev->dev_cqoffset + elan4_cq2num (dev->dev_neterr_intcq)) * CQ_CommandMappingSize;
    qfirst   = DEVICE_NETERR_SLOTS_ADDR;
    qlast    = qfirst + (NETERR_NSLOTS-1) * ELAN4_NETERR_MSG_SIZE;

    spin_lock_init (&dev->dev_neterr_lock);

    /* Register an interrupt operation */
    dev->dev_neterr_intop.op_function = elan4_neterr_interrupt;
    dev->dev_neterr_intop.op_arg      = NULL;

    elan4_register_intop (dev, &dev->dev_neterr_intop);

    /* Initialise the inputq descriptor and event */
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_fptr), qfirst);
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_bptr), qfirst);
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_control), E4_InputQueueControl (qfirst, qlast, ELAN4_NETERR_MSG_SIZE));
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, inputq.q_event), DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, qevent));
    
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, qevent.ev_CountAndType), E4_EVENT_INIT_VALUE (-32, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, qevent.ev_WritePtr), DEVICE_NETERR_INTCQ_ADDR);
    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, qevent.ev_WriteValue), (dev->dev_neterr_intop.op_cookie << E4_MAIN_INT_SHIFT) | INTERRUPT_CMD);

    elan4_sdram_writeq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, sent), 0);

    /* Map them all into the device context */
    elan4mmu_pteload (&dev->dev_ctxt, 0, DEVICE_NETERR_INPUTQ_ADDR, HE_TYPE_SDRAM, (dev->dev_neterr_inputq >> PTE_PADDR_SHIFT) | PTE_SetPerm(PERM_RemoteAll));
    elan4mmu_pteload (&dev->dev_ctxt, 0, DEVICE_NETERR_INTCQ_ADDR,  HE_TYPE_SDRAM, (intqaddr >> PTE_PADDR_SHIFT) | PTE_SetPerm(PERM_LocDataWrite) | PTE_CommandQueue);
    elan4mmu_pteload (&dev->dev_ctxt, 0, DEVICE_NETERR_SLOTS_ADDR,  HE_TYPE_SDRAM, (dev->dev_neterr_slots >> PTE_PADDR_SHIFT) | PTE_SetPerm(PERM_DataReadWrite));

    /* finally attach to the neterr context */
    if (elan4_attach_filter (&dev->dev_ctxt, ELAN4_NETERR_CONTEXT_NUM) != 0)
	panic ("elan4_neterr_init: failed to attach to neterr context\n");

    /* and drop the context filter */
    elan4_set_filter (&dev->dev_ctxt, ELAN4_NETERR_CONTEXT_NUM, E4_FILTER_HIGH_PRI);

    return 1;
}

void
elan4_neterr_destroy (ELAN4_DEV *dev)
{
    if (dev->dev_neterr_intcq)
    {
	elan4_detach_filter (&dev->dev_ctxt, ELAN4_NETERR_CONTEXT_NUM);
	
	elan4mmu_unload_range (&dev->dev_ctxt, 0, DEVICE_NETERR_SLOTS_ADDR,  1 << dev->dev_pageshift[0]);
	elan4mmu_unload_range (&dev->dev_ctxt, 0, DEVICE_NETERR_INTCQ_ADDR,  1 << dev->dev_pageshift[0]);
	elan4mmu_unload_range (&dev->dev_ctxt, 0, DEVICE_NETERR_INPUTQ_ADDR, 1 << dev->dev_pageshift[0]);

	spin_lock_destroy (&dev->dev_neterr_lock);
    }

    if (dev->dev_neterr_intcq)
	elan4_freecq (&dev->dev_ctxt, dev->dev_neterr_intcq);
    dev->dev_neterr_intcq = NULL;

    if (dev->dev_neterr_msgcq)
	elan4_freecq (&dev->dev_ctxt, dev->dev_neterr_msgcq);
    dev->dev_neterr_msgcq = NULL;

    if (dev->dev_neterr_slots)
	elan4_sdram_free (dev, dev->dev_neterr_slots, roundup (NETERR_NSLOTS * ELAN4_NETERR_MSG_SIZE, SDRAM_PAGE_SIZE));
    dev->dev_neterr_slots = 0;
    
    if (dev->dev_neterr_inputq)
	elan4_sdram_free (dev, dev->dev_neterr_inputq, SDRAM_PAGE_SIZE);
    dev->dev_neterr_inputq = 0;
}

int
elan4_neterr_sendmsg (ELAN4_DEV *dev, unsigned int nodeid, unsigned int retries, ELAN4_NETERR_MSG *msg)
{
    ELAN4_CQ  *cq = dev->dev_neterr_msgcq;
    E4_uint64  sent;
    E4_VirtualProcessEntry route;
    unsigned int vp;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_neterr_lock, flags);

    sent = elan4_sdram_readq (dev, dev->dev_neterr_inputq + offsetof (NETERR_INPUTQ, sent));

    PRINTF (DBG_DEVICE, DBG_NETERR, "elan4_neterr_sendmsg: nodeid=%d retries=%d cookie=%llx sender=%d,%d%s\n", 
	    nodeid, retries, msg->msg_cookies[0], msg->msg_sender.loc_node, msg->msg_sender.loc_context,
	    (dev->dev_neterr_queued - sent) >= NETERR_CQ_MSGS ? " - no cq space" : "");

    if ((dev->dev_neterr_queued - sent) >= NETERR_CQ_MSGS)
    {
	spin_unlock_irqrestore (&dev->dev_neterr_lock, flags);
	return 0;
    }

    vp = NETERR_VP_BASE + (dev->dev_neterr_queued % NETERR_VP_COUNT);

    if (elan4_generate_route (&dev->dev_position, &route, ELAN4_NETERR_CONTEXT_NUM, nodeid, nodeid, FIRST_SYSTEM_PACKET | FIRST_HIGH_PRI) < 0)
    {
	spin_unlock_irqrestore (&dev->dev_neterr_lock, flags);
	return 0;
    }

    elan4_write_route (dev, dev->dev_routetable, vp, &route);

    writeq ((GUARD_CMD | GUARD_CHANNEL(0) | GUARD_RESET(retries)),				(void *)cq->cq_mapping);
    writeq (NOP_CMD,										(void *)cq->cq_mapping);
    
    writeq (OPEN_STEN_PKT_CMD | OPEN_PACKET (0, PACK_OK | RESTART_COUNT_ZERO, vp),		(void *)cq->cq_mapping);
    writeq (SEND_TRANS_CMD | (TR_INPUT_Q_GETINDEX << 16),					(void *)cq->cq_mapping);
    writeq (  DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, inputq),			(void *)cq->cq_mapping);

    writeq (SEND_TRANS_CMD | (TR_WRITE (64 >> 3, 0, TR_DATATYPE_DWORD) << 16),			(void *)cq->cq_mapping);
    writeq (  0 /* address */,									(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[0],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[1],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[2],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[3],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[4],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[5],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[6],								(void *)cq->cq_mapping);
    writeq (  ((E4_uint64 *) msg)[7],								(void *)cq->cq_mapping);

    writeq (SEND_TRANS_CMD | (TR_INPUT_Q_COMMIT << 16),						(void *)cq->cq_mapping);
    writeq (  DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, inputq),			(void *)cq->cq_mapping);
    writeq (  0 /* cookie */,									(void *)cq->cq_mapping);
    
    writeq (GUARD_CMD | GUARD_CHANNEL(0) | GUARD_RESET(NETERR_RETRIES),				(void *)cq->cq_mapping);
    writeq (WRITE_DWORD_CMD | (DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, sent)),	(void *)cq->cq_mapping);
    writeq (  ++dev->dev_neterr_queued,								(void *)cq->cq_mapping);

    pioflush_reg (dev);

    spin_unlock_irqrestore (&dev->dev_neterr_lock, flags);

    return 1;
}

int
elan4_neterr_iproc_trap (ELAN4_DEV *dev, ELAN4_IPROC_TRAP *trap)
{
    E4_IprocTrapHeader *hdrp = &trap->tr_transactions[trap->tr_trappedTrans];
    unsigned long flags;

    switch (IPROC_TrapValue (hdrp->IProcStatusCntxAndTrType))
    {
    case InputEopErrorOnWaitForEop:
    case InputEopErrorTrap:
    case InputCrcErrorAfterPAckOk:
	return 1;

    case InputEventEngineTrapped:
	printk ("elan%d: device_iproc_trap: InputEventEngineTrapped - Trans=%x TrAddr=%llx\n", 
		dev->dev_instance, (int)IPROC_TransactionType (hdrp->IProcStatusCntxAndTrType), (long long) hdrp->TrAddr);

	if ((IPROC_TransactionType (hdrp->IProcStatusCntxAndTrType) & TR_OPCODE_MASK) == (TR_INPUT_Q_COMMIT & TR_OPCODE_MASK) &&
	    hdrp->TrAddr == DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, inputq))
	{
	    spin_lock_irqsave (&dev->dev_neterr_lock, flags);
	    writeq ((DEVICE_NETERR_INPUTQ_ADDR + offsetof (NETERR_INPUTQ, qevent)) | SET_EVENT_CMD, (void *)(dev->dev_neterr_msgcq->cq_mapping));
	    spin_unlock_irqrestore (&dev->dev_neterr_lock, flags);
	    return 1;
	}
	
    default:
	return 0;
    }
}
/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
