/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: procfs_Linux.c,v 1.43.2.6 2005/09/07 14:35:03 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/elan4mod/procfs_Linux.c,v $*/

#include <qsnet/kernel.h>

#include <qsnet/module.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>

#include <qsnet/procfs_linux.h>

#include <elan4/i2c.h>
#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>
#include <elan4/mmu.h>
/*
 *
 * procfs format for elan4:
 *
 * /proc/qsnet/elan4/config
 *    elan4_debug
 *    elan4_debug_toconsole
 *    elan4_debug_tobuffer
 *    elan4_debug_display_ctxt
 *    elan4_debug_ignore_ctxt
 *    elan4_debug_ignore_type
 *    elan4_debug_mmu
 *    elan4_mainint_punt_loops
 *    user_p2p_route_options
 *    user_bcast_route_options
 *
 * /proc/qsnet/elan4/deviceN
 *    stats
 *    position
 *    vpd
 */

struct proc_dir_entry *elan4_procfs_root;
struct proc_dir_entry *elan4_config_root;

/* borrowed from fs/proc/proc_misc - helper for proc_read_int */
static int 
proc_calc_metrics(char *page, char **start, off_t off, int count, int *eof, int len)
{
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;
    return len;
}

static int
proc_read_devinfo (char *page, char **start, off_t off,
		    int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    int        len = 0;

    if (! dev)
	len = sprintf (page, "<unknown>\n");
    else
    {
	len += sprintf (page + len, "dev_vendor_id            0x%x\n", dev->dev_devinfo.dev_vendor_id);
	len += sprintf (page + len, "dev_device_id            0x%x\n", dev->dev_devinfo.dev_vendor_id);
	len += sprintf (page + len, "dev_revision_id          0x%x\n", dev->dev_devinfo.dev_revision_id);
	len += sprintf (page + len, "dev_instance             0x%x\n", dev->dev_devinfo.dev_instance);
	len += sprintf (page + len, "dev_rail                 0x%x\n", dev->dev_devinfo.dev_rail);
	len += sprintf (page + len, "dev_driver_version       0x%x\n", dev->dev_devinfo.dev_driver_version);
	len += sprintf (page + len, "dev_params_mask          0x%x\n", dev->dev_devinfo.dev_params_mask);
	len += sprintf (page + len, "dev_params:                  \n");
	len += sprintf (page + len, " 0  - PciCmdQPadFlag     0x%x\n", dev->dev_devinfo.dev_params.values[0]);
	len += sprintf (page + len, " 1  - EventCopyWinPt     0x%x\n", dev->dev_devinfo.dev_params.values[1]);
	len += sprintf (page + len, " 2  - PciWriteCombining  0x%x\n", dev->dev_devinfo.dev_params.values[2]);
	len += sprintf (page + len, " 3  -                    0x%x\n", dev->dev_devinfo.dev_params.values[3]);
	len += sprintf (page + len, " 4  -                    0x%x\n", dev->dev_devinfo.dev_params.values[4]);
	len += sprintf (page + len, " 5  -                    0x%x\n", dev->dev_devinfo.dev_params.values[5]);
	len += sprintf (page + len, " 6  -                    0x%x\n", dev->dev_devinfo.dev_params.values[6]);
	len += sprintf (page + len, " 7  -                    0x%x\n", dev->dev_devinfo.dev_params.values[7]);
	len += sprintf (page + len, " 8  -                    0x%x\n", dev->dev_devinfo.dev_params.values[8]);
	len += sprintf (page + len, " 9  -                    0x%x\n", dev->dev_devinfo.dev_params.values[9]);
	len += sprintf (page + len, " 10 -                    0x%x\n", dev->dev_devinfo.dev_params.values[10]);
	len += sprintf (page + len, " 11 - features           0x%x\n", dev->dev_devinfo.dev_params.values[11]);
	len += sprintf (page + len, "dev_num_down_links_value 0x%x\n", dev->dev_devinfo.dev_num_down_links_value);
    }

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_read_position (char *page, char **start, off_t off,
		    int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    int        len;

    if (dev->dev_position.pos_mode == ELAN_POS_UNKNOWN)
	len = sprintf (page, "<unknown>\n");
    else
	len = sprintf (page, 
		       "NodeId                 %d\n"
		       "NumLevels              %d\n"
		       "NumNodes               %d\n",
		       dev->dev_position.pos_nodeid, 
		       dev->dev_position.pos_levels, 
		       dev->dev_position.pos_nodes);

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_write_position (struct file *file, const char *buf, unsigned long count, void *data)
{
    ELAN4_DEV *dev     = (ELAN4_DEV *) data;
    unsigned  nodeid   = ELAN_INVALID_NODE;
    unsigned  numnodes = 0;
    char     *page, *p;
    int       res;
    ELAN_POSITION pos;

    if (count == 0)
	return (0);

    if (count >= PAGE_SIZE)
	return (-EINVAL);

    if ((page = (char *) __get_free_page (GFP_KERNEL)) == NULL)
	return (-ENOMEM);

    MOD_INC_USE_COUNT;

    if (copy_from_user (page, buf, count))
	res = -EFAULT;
    else
    {
	page[count] = '\0';
	
	if (page[count-1] == '\n')
	    page[count-1] = '\0';

	if (! strcmp (page, "<unknown>"))
	{
	    pos.pos_mode      = ELAN_POS_UNKNOWN;
	    pos.pos_nodeid    = ELAN_INVALID_NODE;
	    pos.pos_nodes     = 0;
	    pos.pos_levels    = 0;
	}
	else
	{
	    for (p = page; *p; )
	    {
		while (isspace (*p))
		    p++;
		
		if (! strncmp (p, "NodeId=", strlen("NodeId=")))
		    nodeid   = simple_strtoul (p + strlen ("NodeId="), NULL, 0);
		if (! strncmp (p, "NumNodes=", strlen ("NumNodes=")))
		    numnodes = simple_strtoul (p + strlen ("NumNodes="), NULL, 0);
		
		while (*p && !isspace(*p))
		    p++;
	    }

	    if (elan4_compute_position (&pos, nodeid, numnodes, dev->dev_devinfo.dev_num_down_links_value) != 0)
		printk ("elan%d: invalid values for NodeId=%d NumNodes=%d\n", dev->dev_instance, nodeid, numnodes);
	    else
	    {
		printk ("elan%d: setting NodeId=%d NumNodes=%d NumLevels=%d\n", dev->dev_instance, pos.pos_nodeid,
			pos.pos_nodes, pos.pos_levels);

		if (elan4_set_position (dev, &pos) < 0)
		    printk ("elan%d: failed to set device position\n", dev->dev_instance);
	    }
	}
    }

    MOD_DEC_USE_COUNT;
    free_page ((unsigned long) page);

    return (count);
}

static int
proc_read_temp (char *page, char **start, off_t off,
		int count, int *eof, void *data)
{
    ELAN4_DEV    *dev = (ELAN4_DEV *) data;
    unsigned char values[2];
    int           len;

    if (i2c_disable_auto_led_update (dev) < 0)
	len = sprintf (page, "<unknown>");
    else
    {
	if (i2c_read (dev, I2C_TEMP_ADDR, 2, values) < 0)
	    len = sprintf (page, "<not-present>");
	else
	    len = sprintf (page, "%s%d%s\n", (values[0] & 0x80) ? "-" : "",
	    		   (values[0] & 0x80) ? -((signed char)values[0]) - 1 : values[0],
	    		   (values[1] & 0x80) ? ".5" : ".0");

	i2c_enable_auto_led_update (dev);
    }

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_read_eccerr (char *page, char **start, off_t off,
		  int count, int *eof, void *data)
{
    ELAN4_DEV    *dev = (ELAN4_DEV *) data;
    char          errstr[200];
    register int  i, len = 0;

    *page = '\0';

    for (i = 0; i < sizeof (dev->dev_sdramerrs)/sizeof(dev->dev_sdramerrs[0]); i++)
	if (dev->dev_sdramerrs[i].ErrorCount != 0)
	    len += sprintf (page + len, "%s occured %0d times\n",
	                    elan4_sdramerr2str (dev, dev->dev_sdramerrs[i].EccStatus, dev->dev_sdramerrs[i].ConfigReg, errstr),
			    dev->dev_sdramerrs[i].ErrorCount);

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_read_vpd (char *page, char **start, off_t off,
	       int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    int        len;

    if ( elan4_read_vpd (dev, NULL, page) )
	len = sprintf (page, "no vpd tags found\n");
    else
	len = strlen(page)+1;

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_read_linkportkey (char *page, char **start, off_t off,
		       int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    int        len;

    len = sprintf (page, "%llx\n", read_reg64 (dev, LinkPortLock));

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

static int
proc_write_linkportkey (struct file *file, const char *buf, unsigned long count, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    int	       res = 0;
    char       tmpbuf[30];

    if (count > sizeof (tmpbuf) - 1)
	return -EINVAL;

    MOD_INC_USE_COUNT;

    if (copy_from_user (tmpbuf, buf, count))
	res = -EFAULT;
    else
    {
	tmpbuf[count] = '\0';

	write_reg64 (dev, LinkPortLock, simple_strtoull (tmpbuf, NULL, 16));
    }

    MOD_DEC_USE_COUNT;

    return (count);
}

static int
proc_read_stats_translations (char *page, char **start, off_t off, int count, int *eof, void *data)
{	
    ELAN4_TRANS_STATS *pr  = (ELAN4_TRANS_STATS *)data;
    int                tbl = pr->tbl;
    ELAN4_DEV         *dev = list_entry(pr, ELAN4_DEV, trans_stats[tbl] );
    char              *p   = page;

    if (off) return (0);

    p += elan4mmu_display_bucket_mmuhash(dev, tbl, pr->buckets, ELAN4_TRANS_STATS_NUM_BUCKETS , p, count);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_write_stats_translations (struct file *file, const char *buf, unsigned long count, void *data)
{
    ELAN4_TRANS_STATS *pr  = (ELAN4_TRANS_STATS *)data;
    int                b0, b1, b2, b3, b4, b5, b6;

    int	       res = 0;
    char       tmpbuf[30];

    if (count > sizeof (tmpbuf) - 1)
	return -EINVAL;

    MOD_INC_USE_COUNT;

    if (copy_from_user (tmpbuf, buf, count))
	res = -EFAULT;
    else
    {
	tmpbuf[count] = '\0';
	sscanf(tmpbuf,"%d %d %d %d %d %d %d\n", &b0, &b1, &b2, &b3, &b4, &b5, &b6);

	pr->buckets[0] = b0;
	pr->buckets[1] = b1;
	pr->buckets[2] = b2;
	pr->buckets[3] = b3;
	pr->buckets[4] = b4;
	pr->buckets[5] = b5;
	pr->buckets[6] = b6;
	pr->buckets[6] = 99999999;

	b0 = 1;
	for(b1=0 ; b1 < ELAN4_TRANS_STATS_NUM_BUCKETS; b1++) {
	    if ( pr->buckets[b1] < b0) 
		pr->buckets[b1] = 99999999;
	    b0 =  pr->buckets[b1];
	}
    }

    MOD_DEC_USE_COUNT;

    return (count);
}
static int 
elan4_read_mmuhash_reduction_func (char *page, char **start, off_t off, int count, int *eof, void *data)
{	
    char              *p   = page;

    if (off) return (0);

    p += sprintf(p ,"elan4mmu hash reduction : %s\nend reductions    %d\nmiddle reductions %d\nmiddle failed     %d\n", 
		 elan4_mmuhash_chain_reduction?"On":"Off",
		 elan4_mmuhash_chain_end_reduce,
		 elan4_mmuhash_chain_middle_reduce,
		 elan4_mmuhash_chain_middle_fail);
    p += sprintf(p ,"shuffle attempts  %d\nshuffle done      %d\n", 
		 elan4_mmuhash_shuffle_attempts,
		 elan4_mmuhash_shuffle_done
		 );

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
elan4_write_mmuhash_reduction_func (struct file *file, const char *buf, unsigned long count, void *data)
{
    int	       res = 0;
    char       tmpbuf[30];

    if (count > sizeof (tmpbuf) - 1)
	return -EINVAL;

    MOD_INC_USE_COUNT;

    if (copy_from_user (tmpbuf, buf, count))
	res = -EFAULT;
    else
    {
	if (tmpbuf[0] == '0') elan4_mmuhash_chain_reduction = 0;
	if (tmpbuf[0] == '1') elan4_mmuhash_chain_reduction = 1;

	tmpbuf[count] = '\0';
    }

    MOD_DEC_USE_COUNT;

    return (count);
}

typedef struct elan4_trans_private
{
    ELAN4_DEV              *pr_dev;
    ELAN4_CTXT             *pr_ctxt;
    int                     pr_index;
    int                     pr_tbl;
    char		   *pr_page;
    unsigned		    pr_off;
    unsigned		    pr_len;

    int                     pr_changed;
} ELAN4_TRANS_PRIVATE;

static int
elan4_ctxt_trans_open (struct inode *inode, struct file *file)
{
    ELAN4_CTXT_TRANS_INDEX *trans  = (ELAN4_CTXT_TRANS_INDEX *)( PDE(inode)->data );
    ELAN4_TRANS_PRIVATE    *pr;

    if ((pr = kmalloc (sizeof (ELAN4_TRANS_PRIVATE), GFP_KERNEL)) == NULL)
	return (-ENOMEM);
    
    pr->pr_tbl   = trans->tbl;
    pr->pr_ctxt  = list_entry(trans, ELAN4_CTXT, trans_index[trans->tbl] );

    pr->pr_index = 0;

    pr->pr_len     = 0;
    pr->pr_off     = 0;
    pr->pr_changed = 1;
    pr->pr_page    = NULL;

    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;
    return (0);
}

static ssize_t
elan4_ctxt_trans_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    ELAN4_TRANS_PRIVATE *pr   = (ELAN4_TRANS_PRIVATE *) file->private_data;
    ELAN4_CTXT          *ctxt = pr->pr_ctxt;
    ELAN4_DEV           *dev  = ctxt->ctxt_dev;
    int                  error;

    if ( pr->pr_index >= dev->dev_hashsize[pr->pr_tbl] ) 
	return (0);

    if ((error = verify_area (VERIFY_WRITE, buf, count)) != 0)
	return (error);

    if (pr->pr_page == NULL && (pr->pr_page = (char *) __get_free_page (GFP_KERNEL)) == NULL)
	return (-ENOMEM);

    if (pr->pr_off >= pr->pr_len)
    {
	if (elan4mmu_display_ctxt_mmuhash(ctxt, pr->pr_tbl, &pr->pr_index, pr->pr_page, count)) 
	    pr->pr_len     = strlen (pr->pr_page);
	else
	    pr->pr_len     = 0;

	pr->pr_off     = 0;
	pr->pr_changed = 0;
	pr->pr_index++;	
    }

    if (count >= (pr->pr_len - pr->pr_off))
	count = pr->pr_len - pr->pr_off;

    copy_to_user (buf, pr->pr_page + pr->pr_off, count);

    pr->pr_off += count;
    *ppos      += count;

    if (pr->pr_off >= pr->pr_len)
    {
	free_page ((unsigned long) pr->pr_page);
	pr->pr_page = NULL;
    }

    return (count);
}

static int
elan4_ctxt_trans_release (struct inode *inode, struct file *file)
{
    ELAN4_TRANS_PRIVATE *pr = (ELAN4_TRANS_PRIVATE *) file->private_data;

    if (pr->pr_page)
	free_page ((unsigned long) pr->pr_page);
    kfree (pr);
    
    MOD_DEC_USE_COUNT;
    return (0);
}

static struct file_operations qsnet_ctxt_trans_fops = 
{
	open:	 elan4_ctxt_trans_open,
	release: elan4_ctxt_trans_release,
	read:	 elan4_ctxt_trans_read,
};

void
proc_insertctxt(ELAN4_DEV *dev, ELAN4_CTXT *ctxt)
{ 
    struct proc_dir_entry *p;
    char name[32];
    int  t;

    /* GNAT 7565: Need to hold kernel lock when adding/removing
     * procfs entries outside the module init/fini paths
     */
    lock_kernel();

    if (dev->dev_osdep.ctxtdir) 
    {
	sprintf(name,"%d", ctxt->ctxt_num);
	if ((ctxt->procdir = proc_mkdir (name, dev->dev_osdep.ctxtdir)) != NULL) 
	{
	    for (t = 0; t < NUM_HASH_TABLES; t++) 
	    {
		sprintf(name , "translations_%d", t);
		
		ctxt->trans_index[t].tbl   = t;
		ctxt->trans_index[t].index = 0;
		
		if ((p = create_proc_entry (name, 0, ctxt->procdir)) != NULL)
		{
		    p->proc_fops = &qsnet_ctxt_trans_fops;
		    p->data      = & ctxt->trans_index[t];
		    p->owner     = THIS_MODULE;
		}
	    }
	}
    }

    unlock_kernel();
}

void 
proc_removectxt(ELAN4_DEV *dev, ELAN4_CTXT *ctxt)
{
    int  t;
    char name[32];

    /* GNAT 7565: Need to hold kernel lock when adding/removing
     * procfs entries outside the module init/fini paths
     */
    lock_kernel();

    if (dev->dev_osdep.ctxtdir && ctxt->procdir != NULL) 
    {
	for (t = 0; t < NUM_HASH_TABLES; t++) 
	{
	    sprintf(name , "translations_%d", t);
	    remove_proc_entry (name, ctxt->procdir);
	}
	
	sprintf(name,"%d", ctxt->ctxt_num);
	remove_proc_entry (name, dev->dev_osdep.ctxtdir);
    }

    unlock_kernel();
}

static struct device_info 
{
    char *name;
    int (*read_func) (char *page, char **start, off_t off, int count, int *eof, void *data);
    int (*write_func) (struct file *file, const char *buf, unsigned long count, void *data);
    unsigned minrev;
} device_info[] = {
    {"devinfo",     proc_read_devinfo,     NULL,                   0},
    {"position",    proc_read_position,    proc_write_position,    0},
    {"temp",        proc_read_temp,        NULL,                   1},
    {"eccerr",      proc_read_eccerr,      NULL,                   0},
    {"vpd",         proc_read_vpd,         NULL,                   0},
    {"linkportkey", proc_read_linkportkey, proc_write_linkportkey, 0},
};

static int
proc_read_link_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;

    p += sprintf (p, "%20s %ld\n", "link_errors", dev->dev_stats.s_link_errors);
    p += sprintf (p, "%20s %ld\n", "lock_errors", dev->dev_stats.s_lock_errors);
    p += sprintf (p, "%20s %ld\n", "deskew_errors", dev->dev_stats.s_deskew_errors);
    p += sprintf (p, "%20s %ld\n", "phase_errors", dev->dev_stats.s_phase_errors);

    p += sprintf (p, "%20s %ld\n", "data_errors", dev->dev_stats.s_data_errors);
    p += sprintf (p, "%20s %ld\n", "fifo_overflow0", dev->dev_stats.s_fifo_overflow0);
    p += sprintf (p, "%20s %ld\n", "fifo_overflow1", dev->dev_stats.s_fifo_overflow1);
    p += sprintf (p, "%20s %ld\n", "mod45changed", dev->dev_stats.s_mod45changed);
    p += sprintf (p, "%20s %ld\n", "pack_not_seen", dev->dev_stats.s_pack_not_seen);

    p += sprintf (p, "%20s %ld\n", "linkport_keyfail", dev->dev_stats.s_linkport_keyfail);
    p += sprintf (p, "%20s %ld\n", "eop_reset", dev->dev_stats.s_eop_reset);
    p += sprintf (p, "%20s %ld\n", "bad_length", dev->dev_stats.s_bad_length);
    p += sprintf (p, "%20s %ld\n", "crc_error", dev->dev_stats.s_crc_error);
    p += sprintf (p, "%20s %ld\n", "crc_bad", dev->dev_stats.s_crc_bad);

    p += sprintf (p, "%20s %ld\n", "cproc_timeout", dev->dev_stats.s_cproc_timeout);
    p += sprintf (p, "%20s %ld\n", "dproc_timeout", dev->dev_stats.s_dproc_timeout);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static char *
proc_sprintf_bucket_stat (char *p, char *name, unsigned long *stats, int *buckets)
{
    int i;

    p += sprintf (p, "%20s ", name);

    for (i = 0; i < ELAN4_DEV_STATS_BUCKETS-1; i++)
	p += sprintf (p, "%ld(<=%d) ", stats[i], buckets[i]);
    p += sprintf (p, "%ld(>%d)\n", stats[i], buckets[i-1]);

    return p;
}

static int
proc_read_intr_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;

    p += sprintf (p, "%20s %ld\n", "interrupts",       dev->dev_stats.s_interrupts);
    p += sprintf (p, "%20s %ld\n", "haltints",         dev->dev_stats.s_haltints);

    p += sprintf (p, "%20s %ld\n", "mainint_punts",    dev->dev_stats.s_mainint_punts);
    p += sprintf (p, "%20s %ld\n", "mainint_rescheds", dev->dev_stats.s_mainint_rescheds);

    p  = proc_sprintf_bucket_stat (p, "mainints", dev->dev_stats.s_mainints, MainIntBuckets);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;

    p += sprintf (p, "%20s %ld\n", "cproc_traps", dev->dev_stats.s_cproc_traps);
    p += sprintf (p, "%20s %ld\n", "dproc_traps", dev->dev_stats.s_dproc_traps);
    p += sprintf (p, "%20s %ld\n", "eproc_traps", dev->dev_stats.s_eproc_traps);
    p += sprintf (p, "%20s %ld\n", "iproc_traps", dev->dev_stats.s_iproc_traps);
    p += sprintf (p, "%20s %ld\n", "tproc_traps", dev->dev_stats.s_tproc_traps);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_cproc_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;
    int        i;
    extern char *const CProcTrapNames[];

    for (i = 0; i < sizeof (dev->dev_stats.s_cproc_trap_types)/sizeof(dev->dev_stats.s_cproc_trap_types[0]); i++)
	p += sprintf (p, "%-40s %ld\n", CProcTrapNames[i], dev->dev_stats.s_cproc_trap_types[i]);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_dproc_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;
    int        i;
    extern char *const DProcTrapNames[];

    for (i = 0; i < sizeof (dev->dev_stats.s_dproc_trap_types)/sizeof(dev->dev_stats.s_dproc_trap_types[0]); i++)
	p += sprintf (p, "%-40s %ld\n", DProcTrapNames[i], dev->dev_stats.s_dproc_trap_types[i]);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_eproc_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;
    int        i;
    extern char *const EProcTrapNames[];

    for (i = 0; i < sizeof (dev->dev_stats.s_eproc_trap_types)/sizeof(dev->dev_stats.s_eproc_trap_types[0]); i++)
	p += sprintf (p, "%-40s %ld\n", EProcTrapNames[i], dev->dev_stats.s_eproc_trap_types[i]);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_iproc_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;
    int        i;
    extern char *const IProcTrapNames[];

    for (i = 0; i < sizeof (dev->dev_stats.s_iproc_trap_types)/sizeof(dev->dev_stats.s_iproc_trap_types[0]); i++)
	p += sprintf (p, "%-40s %ld\n", IProcTrapNames[i], dev->dev_stats.s_iproc_trap_types[i]);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_tproc_trap_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;
    int        i;
    extern char *const TProcTrapNames[];

    for (i = 0; i < sizeof (dev->dev_stats.s_tproc_trap_types)/sizeof(dev->dev_stats.s_tproc_trap_types[0]); i++)
	p += sprintf (p, "%-40s %ld\n", TProcTrapNames[i], dev->dev_stats.s_tproc_trap_types[i]);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_sdram_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev = (ELAN4_DEV *) data;
    char      *p   = page;

    p += sprintf (p, "%20s %ld\n",  "correctable_errors", dev->dev_stats.s_correctable_errors);
    p += sprintf (p, "%20s %ld\n",  "multiple_errors",    dev->dev_stats.s_multiple_errors);
    p += sprintf (p, "%20s %ldK\n", "sdram_bytes_free",   dev->dev_stats.s_sdram_bytes_free/1024);

    return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

void
elan4_ringbuf_store (ELAN4_ROUTE_RINGBUF *ringbuf, E4_VirtualProcessEntry *route, ELAN4_DEV *dev)
{
    int newend;
    
    ASSERT (kmutex_is_locked (&dev->dev_lock));

    memcpy(&ringbuf->routes[ringbuf->end], route, sizeof(E4_VirtualProcessEntry));
    newend = ringbuf->end + 1;
    if (newend >= DEV_STASH_ROUTE_COUNT)
        newend -= DEV_STASH_ROUTE_COUNT;
    if (newend == ringbuf->start)
        ringbuf->start += 1;
    if (ringbuf->start >= DEV_STASH_ROUTE_COUNT)
        ringbuf->start -= DEV_STASH_ROUTE_COUNT;
    ringbuf->end = newend;
}
	
static int
proc_read_dproc_timeout_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	unsigned int *dproc_timeout;

	dproc_timeout = dev->dev_dproc_timeout;

	if (!dproc_timeout) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int i;

		for (i=0; i<dev->dev_position.pos_nodes; i++) 
			if (dproc_timeout[i] != 0) 
				p += sprintf (p, "Node %d: %u errors\n", i, dproc_timeout[i]);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_dproc_timeout_routes (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	ELAN4_ROUTE_RINGBUF *ringbuf;
	char      routestr[33];

	ringbuf = &dev->dev_dproc_timeout_routes;

	if (!ringbuf) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int start;
		int end;
		int i;

		memset(&routestr, 0, 33);
		
		kmutex_lock(&dev->dev_lock);
		
		start = ringbuf->start;
		end = ringbuf->end;

		if (end < start)
			end = DEV_STASH_ROUTE_COUNT;

		for (i=start; i<end; i++) 
		{
			elan4_route2str (&ringbuf->routes[i], routestr);
			p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
		}

		if (ringbuf->end < start)
		{
			start = 0;
			end = ringbuf->end;
			for (i=start; i<end; i++)
			{
				elan4_route2str (&ringbuf->routes[i], routestr);
				p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
			}
		}

		kmutex_unlock(&dev->dev_lock);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}


static int
proc_read_cproc_timeout_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	unsigned int *cproc_timeout;

	cproc_timeout = dev->dev_cproc_timeout;

	if (!cproc_timeout) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int i;

		for (i=0; i<dev->dev_position.pos_nodes; i++) 
			if (cproc_timeout[i] != 0) 
				p += sprintf (p, "Node %d: %u errors\n", i, cproc_timeout[i]);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_cproc_timeout_routes (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	ELAN4_ROUTE_RINGBUF *ringbuf;
	char      routestr[33];

	ringbuf = &dev->dev_cproc_timeout_routes;

	if (!ringbuf) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int start;
		int end;
		int i;

		memset(&routestr, 0, 33);

		kmutex_lock(&dev->dev_lock);
		
		start = ringbuf->start;
		end = ringbuf->end;

		if (end < start)
			end = DEV_STASH_ROUTE_COUNT;

		for (i=start; i<end; i++) 
		{
			elan4_route2str (&ringbuf->routes[i], routestr);
			p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
		}

		if (ringbuf->end < start)
		{
			start = 0;
			end = ringbuf->end;
			for (i=start; i<end; i++)
			{
				elan4_route2str (&ringbuf->routes[i], routestr);
				p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
			}
		}

		kmutex_unlock(&dev->dev_lock);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_traperr_stats (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	unsigned int *ack_errors;

	ack_errors = dev->dev_ack_errors;

	if (!ack_errors) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int i;

		for (i=0; i<dev->dev_position.pos_nodes; i++) 
			if (ack_errors[i] != 0) 
				p += sprintf (p, "Node %d: %u errors\n", i, ack_errors[i]);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static int
proc_read_ackerror_routes (char *page, char **start, off_t off, int count, int *eof, void *data)
{
	ELAN4_DEV *dev = (ELAN4_DEV *) data;
	char      *p   = page;
	ELAN4_ROUTE_RINGBUF *ringbuf;
	char      routestr[33];

	ringbuf = &dev->dev_ack_error_routes;

	if (!ringbuf) 
		p += sprintf (p, "No stats available\n");
	else
	{
		int start;
		int end;
		int i;

		memset(&routestr, 0, 33);

		kmutex_lock(&dev->dev_lock);
		
		start = ringbuf->start;
		end = ringbuf->end;

		if (end < start)
			end = DEV_STASH_ROUTE_COUNT;

		for (i=start; i<end; i++) 
		{
			elan4_route2str (&ringbuf->routes[i], routestr);
			p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
		}

		if (ringbuf->end < start)
		{
			start = 0;
			end = ringbuf->end;
			for (i=start; i<end; i++)
			{
				elan4_route2str (&ringbuf->routes[i], routestr);
				p += sprintf (p, "Route %llx %llx->%s\n", (long long)ringbuf->routes[i].Values[0], (long long)ringbuf->routes[i].Values[1], routestr);
			}
		}

		kmutex_unlock(&dev->dev_lock);
	}

	return (proc_calc_metrics (page, start, off, count, eof, p - page));
}

static struct stats_info 
{
    char *name;
    int (*read_func) (char *page, char **start, off_t off, int count, int *eof, void *data);
    int (*write_func) (struct file *file, const char *buf, unsigned long count, void *data); 
} stats_info[] = {
    {"link",     proc_read_link_stats, NULL},
    {"intr",     proc_read_intr_stats, NULL},
    {"trap",     proc_read_trap_stats, NULL},
    {"cproc",    proc_read_cproc_trap_stats, NULL},
    {"dproc",    proc_read_dproc_trap_stats, NULL},
    {"eproc",    proc_read_eproc_trap_stats, NULL},
    {"iproc",    proc_read_iproc_trap_stats, NULL},
    {"tproc",    proc_read_tproc_trap_stats, NULL},
    {"sdram",    proc_read_sdram_stats, NULL},
    {"trapdmaerr", proc_read_traperr_stats, NULL},
    {"dproctimeout", proc_read_dproc_timeout_stats, NULL},
    {"cproctimeout", proc_read_cproc_timeout_stats, NULL},
    {"dproctimeoutroutes", proc_read_dproc_timeout_routes, NULL},
    {"cproctimeoutroutes", proc_read_cproc_timeout_routes, NULL},
    {"ackerrroutes", proc_read_ackerror_routes, NULL},
};

static int
proc_read_sysconfig (char *page, char **start, off_t off, int count, int *eof, void *data)
{
    ELAN4_DEV *dev        = (ELAN4_DEV *) data;
    E4_uint32  syscontrol = dev->dev_syscontrol;
    int	       len       = 0;

   *eof = 1;
   if (off != 0)
      return (0);

    if (syscontrol & CONT_EN_ALL_SETS)
	len += sprintf (page + len, "%sEN_ALL_SETS", len == 0 ? "" : " ");
    if (syscontrol & CONT_MMU_ENABLE)
	len += sprintf (page + len, "%sMMU_ENABLE", len == 0 ? "" : " ");
    if (syscontrol & CONT_CACHE_HASH_TABLE)
	len += sprintf (page + len, "%sCACHE_HASH_TABLE", len == 0 ? "" : " ");
    if (syscontrol & CONT_CACHE_CHAINS)
	len += sprintf (page + len, "%sCACHE_CHAINS", len == 0 ? "" : " ");
    if (syscontrol & CONT_CACHE_ROOT_CNTX)
	len += sprintf (page + len, "%sCACHE_ROOT_CNTX", len == 0 ? "" : " ");
    if (syscontrol & CONT_CACHE_STEN_ROUTES)
	len += sprintf (page + len, "%sCACHE_STEN_ROUTES", len == 0 ? "" : " ");
    if (syscontrol & CONT_CACHE_DMA_ROUTES)
	len += sprintf (page + len, "%sCACHE_DMA_ROUTES", len == 0 ? "" : " ");
    if (syscontrol & CONT_INHIBIT_MAX_CHAIN_ITEMS)
	len += sprintf (page + len, "%sINHIBIT_MAX_CHAIN_ITEMS", len == 0 ? "" : " ");

    len += sprintf (page + len, "%sTABLE0_MASK_SIZE=%d", len == 0 ? "" : " ", (syscontrol >> CONT_TABLE0_MASK_SIZE_SHIFT) & PAGE_MASK_MASK);
    len += sprintf (page + len, "%sTABLE0_PAGE_SIZE=%d", len == 0 ? "" : " ", (syscontrol >> CONT_TABLE0_PAGE_SIZE_SHIFT) & PAGE_SIZE_MASK);
    len += sprintf (page + len, "%sTABLE1_MASK_SIZE=%d", len == 0 ? "" : " ", (syscontrol >> CONT_TABLE1_MASK_SIZE_SHIFT) & PAGE_MASK_MASK);
    len += sprintf (page + len, "%sTABLE1_PAGE_SIZE=%d", len == 0 ? "" : " ", (syscontrol >> CONT_TABLE1_PAGE_SIZE_SHIFT) & PAGE_SIZE_MASK);

    if (syscontrol & CONT_2K_NOT_1K_DMA_PACKETS)
	len += sprintf (page + len, "%s2K_NOT_1K_DMA_PACKETS", len == 0 ? "" : " ");
    if (syscontrol & CONT_ALIGN_ALL_DMA_PACKETS)
	len += sprintf (page + len, "%sALIGN_ALL_DMA_PACKETS", len == 0 ? "" : " ");
    if (syscontrol & CONT_DIRECT_MAP_PCI_WRITES)
	len += sprintf (page + len, "%sDIRECT_MAP_PCI_WRITES", len == 0 ? "" : " ");

    len += sprintf (page + len, "\n");

   *start = page;
   return (len);
}

static int
proc_write_sysconfig (struct file *file, const char *ubuffer, unsigned long count, void *data)
{
    ELAN4_DEV *dev       = (ELAN4_DEV *) data;
    unsigned long page   = __get_free_page (GFP_KERNEL);
    char         *buffer = (char *)page;
    int            add   = 0;
    int            sub   = 0;
    
    count = MIN (count, PAGE_SIZE - 1);
    if (copy_from_user (buffer, ubuffer, count))
    {
	free_page (page);
	return (-EFAULT);
    }
   
    buffer[count] = 0;				/* terminate string */

    while (*buffer != 0)
    {
	char *ptr;
	char *end;
	int   ch;
	int   val;
	int   op;

	ch = *buffer;
	if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
	{
	    buffer++;
	    continue;
	}
      
	op = *buffer;
	if (op == '+' || op == '-')
	    buffer++;

	for (end = buffer; *end != 0; end++)
	    if (*end == ' ' || *end == '\t' ||
		*end == '\r' || *end == '\n')
		break;
      
	if (end == buffer)
	    break;
      
	ch = *end;
	*end = 0;

	for (ptr = buffer; *ptr != 0; ptr++)
	    if ('a' <= *ptr && *ptr <= 'z')
		*ptr = *ptr + 'A' - 'a';
      
	if (!strcmp (buffer, "EN_ALL_SETS"))
	    val = CONT_EN_ALL_SETS;
	if (!strcmp (buffer, "CACHE_HASH_TABLE"))
	    val = CONT_CACHE_HASH_TABLE;
	else if (!strcmp (buffer, "CACHE_CHAINS"))
	    val = CONT_CACHE_CHAINS;
	else if (!strcmp (buffer, "CACHE_ROOT_CNTX"))
	    val = CONT_CACHE_ROOT_CNTX;
	else if (!strcmp (buffer, "CACHE_STEN_ROUTES"))
	    val = CONT_CACHE_STEN_ROUTES;
	else if (!strcmp (buffer, "CACHE_DMA_ROUTES"))
	    val = CONT_CACHE_DMA_ROUTES;
	else if (!strcmp (buffer, "2K_NOT_1K_DMA_PACKETS"))
	    val = CONT_2K_NOT_1K_DMA_PACKETS;
	else if (!strcmp (buffer, "ALIGN_ALL_DMA_PACKETS"))
	    val = CONT_ALIGN_ALL_DMA_PACKETS;
	else
	    val = 0;

	if (op == '+')
	    add |= val;
	else if (op == '-')
	    sub |= val;

	*end = ch;
	buffer = end;
    }

    if ((add | sub) & CONT_EN_ALL_SETS)
	elan4_sdram_flushcache (dev, 0, E4_CacheSize);

    CHANGE_SYSCONTROL (dev, add, sub);
   
    if ((add | sub) & CONT_EN_ALL_SETS)
	elan4_sdram_flushcache (dev, 0, E4_CacheSize);

    free_page (page);
    return (count);
}

static struct config_info 
{
    char *name;
    int (*read_func) (char *page, char **start, off_t off, int count, int *eof, void *data);
    int (*write_func) (struct file *file, const char *buf, unsigned long count, void *data); 
} config_info[] = {
    {"sysconfig",  proc_read_sysconfig, proc_write_sysconfig},
};

static int
elan4_trans_open (struct inode *inode, struct file *file)
{
    ELAN4_TRANS_INDEX   *trans  = (ELAN4_TRANS_INDEX *)( PDE(inode)->data );
    ELAN4_TRANS_PRIVATE *pr;

    if ((pr = kmalloc (sizeof (ELAN4_TRANS_PRIVATE), GFP_KERNEL)) == NULL)
	return (-ENOMEM);
    
    pr->pr_tbl   = trans->tbl;
    pr->pr_dev   = list_entry(trans, ELAN4_DEV, trans_index[trans->tbl] );
    pr->pr_index = 0;

    pr->pr_len     = 0;
    pr->pr_off     = 0;
    pr->pr_changed = 1;
    pr->pr_page    = NULL;

    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;
    return (0);
}

static ssize_t
elan4_trans_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    ELAN4_TRANS_PRIVATE *pr = (ELAN4_TRANS_PRIVATE *) file->private_data;
    ELAN4_DEV           *dev = pr->pr_dev;
    int                  error;

    if ( pr->pr_index >= dev->dev_hashsize[pr->pr_tbl] ) 
	return (0);

    if ((error = verify_area (VERIFY_WRITE, buf, count)) != 0)
	return (error);

    if (pr->pr_page == NULL && (pr->pr_page = (char *) __get_free_page (GFP_KERNEL)) == NULL)
	return (-ENOMEM);

    if (pr->pr_off >= pr->pr_len)
    {
	if (elan4mmu_display_mmuhash(dev, pr->pr_tbl, &pr->pr_index, pr->pr_page, count)) 
	    pr->pr_len     = strlen (pr->pr_page);
	else
	    pr->pr_len     = 0; 

	pr->pr_off     = 0;
	pr->pr_changed = 0;
	pr->pr_index++;	
    }

    if (count >= (pr->pr_len - pr->pr_off))
	count = pr->pr_len - pr->pr_off;

    copy_to_user (buf, pr->pr_page + pr->pr_off, count);

    pr->pr_off += count;
    *ppos      += count;

    if (pr->pr_off >= pr->pr_len)
    {
	free_page ((unsigned long) pr->pr_page);
	pr->pr_page = NULL;
    }

    return (count);
}

static int
elan4_trans_release (struct inode *inode, struct file *file)
{
    ELAN4_TRANS_PRIVATE *pr = (ELAN4_TRANS_PRIVATE *) file->private_data;

    if (pr->pr_page)
	free_page ((unsigned long) pr->pr_page);
    kfree (pr);
    
    MOD_DEC_USE_COUNT;
    return (0);
}

static struct file_operations qsnet_trans_fops = 
{
	open:	 elan4_trans_open,
	release: elan4_trans_release,
	read:	 elan4_trans_read,
};

void
elan4_procfs_device_init (ELAN4_DEV *dev)
{
    struct proc_dir_entry *p;
    char name[NAME_MAX];
    int i;

    sprintf (name, "device%d", dev->dev_instance);
    dev->dev_osdep.procdir  = proc_mkdir (name, elan4_procfs_root);

    for (i = 0; i < sizeof (device_info)/sizeof (device_info[0]); i++)
    {
	if (dev->dev_devinfo.dev_revision_id < device_info[i].minrev)
	    continue;

	if ((p = create_proc_entry (device_info[i].name, 0, dev->dev_osdep.procdir)) != NULL)
	{
	    p->read_proc  = device_info[i].read_func;
	    p->write_proc = device_info[i].write_func;
	    p->data       = dev;
	    p->owner      = THIS_MODULE;
	}
    }

    for(i = 0; i < NUM_HASH_TABLES; i++) {
	sprintf (name, "translations_%d",i);

	dev->trans_index[i].tbl   = i;

	if ((p = create_proc_entry (name, 0, dev->dev_osdep.procdir)) != NULL)
	{
	    p->proc_fops = &qsnet_trans_fops;
	    p->data      = & dev->trans_index[i];
	    p->owner     = THIS_MODULE;
	}
    }

    dev->dev_osdep.configdir = proc_mkdir ("config", dev->dev_osdep.procdir);
    for (i = 0; i < sizeof (config_info)/sizeof (config_info[0]); i++)
    {
	if ((p = create_proc_entry (config_info[i].name, 0, dev->dev_osdep.configdir)) != NULL)
	{
	    p->read_proc  = config_info[i].read_func;
	    p->write_proc = config_info[i].write_func;
	    p->data       = dev;
	    p->owner      = THIS_MODULE;
	}
    }

    dev->dev_osdep.statsdir = proc_mkdir ("stats", dev->dev_osdep.procdir);
    for (i = 0; i < sizeof (stats_info)/sizeof (stats_info[0]); i++)
    {
	if ((p = create_proc_entry (stats_info[i].name, 0, dev->dev_osdep.statsdir)) != NULL)
	{
	    p->read_proc  = stats_info[i].read_func;
	    p->write_proc = stats_info[i].write_func;
	    p->data       = dev;
	    p->owner      = THIS_MODULE;
	}
    }
    for(i = 0; i < NUM_HASH_TABLES; i++) {
	sprintf (name, "translations_%d",i);

	dev->trans_stats[i].tbl   = i;
	dev->trans_stats[i].buckets[0] = 1;
	dev->trans_stats[i].buckets[1] = 5;
	dev->trans_stats[i].buckets[2] = 10;
	dev->trans_stats[i].buckets[3] = 50;
	dev->trans_stats[i].buckets[4] = 100;
	dev->trans_stats[i].buckets[5] = 200;
	dev->trans_stats[i].buckets[6] = 99999999;

	if ((p = create_proc_entry (name, 0, dev->dev_osdep.statsdir)) != NULL)
	{
	    p->read_proc  = proc_read_stats_translations;
	    p->write_proc = proc_write_stats_translations;
	    p->data       = & dev->trans_stats[i];
	    p->owner      = THIS_MODULE;
	}
    }

    dev->dev_osdep.ctxtdir = proc_mkdir ("ctxt", dev->dev_osdep.procdir);
}

void
elan4_procfs_device_fini (ELAN4_DEV *dev)
{
    char name[NAME_MAX];
    int i;

    if (dev->dev_osdep.ctxtdir)
	remove_proc_entry ("ctxt", dev->dev_osdep.procdir);

    for (i = 0; i < sizeof (stats_info)/sizeof (stats_info[0]); i++)
	remove_proc_entry (stats_info[i].name, dev->dev_osdep.statsdir);

    for (i = 0; i < NUM_HASH_TABLES; i++) {
	sprintf(name , "translations_%d", i);
	remove_proc_entry (name, dev->dev_osdep.statsdir);
    }
    remove_proc_entry ("stats", dev->dev_osdep.procdir);

    for (i = 0; i < sizeof (config_info)/sizeof (config_info[0]); i++)
	remove_proc_entry (config_info[i].name, dev->dev_osdep.configdir);
    remove_proc_entry ("config", dev->dev_osdep.procdir);

    for (i = 0; i < sizeof (device_info)/sizeof (device_info[0]); i++)
    {
	if (dev->dev_devinfo.dev_revision_id < device_info[i].minrev)
	    continue;
	
	remove_proc_entry (device_info[i].name, dev->dev_osdep.procdir);
    }
    
    for (i = 0; i < NUM_HASH_TABLES; i++) {
	sprintf(name , "translations_%d", i);
	remove_proc_entry (name, dev->dev_osdep.procdir);
    }

    sprintf (name, "device%d", dev->dev_instance);
    remove_proc_entry (name, elan4_procfs_root);
}

void
elan4_procfs_init(void)
{
    struct proc_dir_entry *p;
    
    elan4_procfs_root = proc_mkdir("elan4", qsnet_procfs_root);
    elan4_config_root = proc_mkdir("config", elan4_procfs_root);

    qsnet_proc_register_hex (elan4_config_root, "elan4_debug",              &elan4_debug,              0);
    qsnet_proc_register_hex (elan4_config_root, "elan4_debug_toconsole",    &elan4_debug_toconsole,    0);
    qsnet_proc_register_hex (elan4_config_root, "elan4_debug_tobuffer",     &elan4_debug_tobuffer,     0);
    qsnet_proc_register_int (elan4_config_root, "elan4_debug_mmu",          &elan4_debug_mmu,          0);
    qsnet_proc_register_int (elan4_config_root, "elan4_mainint_punt_loops", &elan4_mainint_punt_loops, 0);
    qsnet_proc_register_hex (elan4_config_root, "user_p2p_route_options",   &user_p2p_route_options,   0);
    qsnet_proc_register_hex (elan4_config_root, "user_bcast_route_options", &user_bcast_route_options, 0);
    qsnet_proc_register_int (elan4_config_root, "user_dproc_retry_count",   &user_dproc_retry_count,   0);
    qsnet_proc_register_int (elan4_config_root, "user_cproc_retry_count",   &user_cproc_retry_count,   0);
    qsnet_proc_register_int (elan4_config_root, "user_pagefault_enabled",   &user_pagefault_enabled,   0);
    qsnet_proc_register_int (elan4_config_root, "num_fault_save",           &num_fault_save,           0);
    qsnet_proc_register_int (elan4_config_root, "min_fault_pages",          &min_fault_pages,          0);
    qsnet_proc_register_int (elan4_config_root, "max_fault_pages",          &max_fault_pages,          0);
    qsnet_proc_register_int (elan4_config_root, "assfail_mode",             &assfail_mode,             0);
   
    if ((p = create_proc_entry ("mmuhash_reduction", 0, elan4_config_root)) != NULL)
    {
	p->read_proc  = elan4_read_mmuhash_reduction_func;
	p->write_proc = elan4_write_mmuhash_reduction_func;
	p->data       = NULL;
	p->owner      = THIS_MODULE;
    } 

#if defined(IOPROC_PATCH_APPLIED)
    qsnet_proc_register_int (elan4_config_root, "user_ioproc_enabled",      &user_ioproc_enabled,       0);
#endif
}

void
elan4_procfs_fini(void)
{
#if defined(IOPROC_PATCH_APPLIED)
    remove_proc_entry ("user_ioproc_enabled",      elan4_config_root);
#endif

    remove_proc_entry ("mmuhash_reduction",        elan4_config_root);

    remove_proc_entry ("assfail_mode",             elan4_config_root);
    remove_proc_entry ("max_fault_pages",          elan4_config_root);
    remove_proc_entry ("min_fault_pages",          elan4_config_root);
    remove_proc_entry ("num_fault_save",           elan4_config_root);
    remove_proc_entry ("user_pagefault_enabled",   elan4_config_root);
    remove_proc_entry ("user_cproc_retry_count",   elan4_config_root);
    remove_proc_entry ("user_dproc_retry_count",   elan4_config_root);
    remove_proc_entry ("user_bcast_route_options", elan4_config_root);
    remove_proc_entry ("user_p2p_route_options",   elan4_config_root);
    remove_proc_entry ("elan4_mainint_punt_loops", elan4_config_root);
    remove_proc_entry ("elan4_debug_mmu",          elan4_config_root);
    remove_proc_entry ("elan4_debug_tobuffer",     elan4_config_root);
    remove_proc_entry ("elan4_debug_toconsole",    elan4_config_root);
    remove_proc_entry ("elan4_debug",              elan4_config_root);

    remove_proc_entry ("config", elan4_procfs_root);
    remove_proc_entry ("elan4", qsnet_procfs_root);
}

EXPORT_SYMBOL(elan4_procfs_root);
EXPORT_SYMBOL(elan4_config_root);

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
