/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: regions.c,v 1.22 2004/10/21 15:31:12 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/regions.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>

/*================================================================================*/
/* elan address region management */
USER_RGN *
user_findrgn_elan (USER_CTXT *uctx, E4_Addr addr, int tail)
{
    USER_RGN *rgn;
    USER_RGN *hirgn;
    USER_RGN *lorgn;
    E4_Addr   base;
    E4_Addr   lastaddr;
    int	      forward;
    
    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) || kmutex_is_locked (&uctx->uctx_rgnmutex));

    if (uctx->uctx_ergns == NULL)
	return (NULL);

    rgn = uctx->uctx_ergnlast;
    if (rgn == NULL)
	rgn = uctx->uctx_ergns;

    forward = 0;
    if ((base = rgn->rgn_ebase) < addr)
    {
	if (addr <= (base + rgn->rgn_len - 1))
	    return (rgn);					/* ergnlast contained addr */
	
	hirgn = uctx->uctx_etail;
	
	if ((lastaddr = (hirgn->rgn_ebase + hirgn->rgn_len - 1)) < addr)
	    return (tail ? hirgn : NULL);			/* addr is out of range */
	
	if ((addr - base) > (lastaddr - addr))
	    rgn = hirgn;
	else
	{
	    rgn = rgn->rgn_enext;
	    forward++;
	}
    }
    else
    {
	lorgn = uctx->uctx_ergns;

	if (lorgn->rgn_ebase > addr)
	    return (lorgn);					/* lowest regions is higher than addr */
	if ((addr - lorgn->rgn_ebase) < (base - addr))
	{
	    rgn = lorgn;					/* search forward from head */
	    forward++;
	}
    }
    if (forward)
    {
	while ((rgn->rgn_ebase + rgn->rgn_len - 1) < addr)
	    rgn = rgn->rgn_enext;

	if (rgn->rgn_ebase <= addr)
	    uctx->uctx_ergnlast = rgn;
	return (rgn);
    }
    else
    {
	while (rgn->rgn_ebase > addr)
	    rgn = rgn->rgn_eprev;

	if ((rgn->rgn_ebase + rgn->rgn_len - 1) < addr)
	    return (rgn->rgn_enext);
	else
	{
	    uctx->uctx_ergnlast = rgn;
	    return (rgn);
	}
    }
}

static int
user_addrgn_elan (USER_CTXT *uctx, USER_RGN  *nrgn)
{
    USER_RGN *rgn   = user_findrgn_elan (uctx, nrgn->rgn_ebase, 1);
    E4_Addr   nbase = nrgn->rgn_ebase;
    E4_Addr   ntop  = nbase + nrgn->rgn_len - 1;
    E4_Addr   base;

    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) && kmutex_is_locked (&uctx->uctx_rgnmutex));

    if (rgn == NULL)
    {
	uctx->uctx_ergns = uctx->uctx_etail = nrgn;
	nrgn->rgn_enext = nrgn->rgn_eprev = NULL;
    }
    else
    {
	base = rgn->rgn_ebase;
	
	if ((base + rgn->rgn_len - 1) < nbase)			/* top of region below requested address */
	{							/* so insert after region (and hence at end */
	    nrgn->rgn_eprev = rgn;				/* of list */
	    nrgn->rgn_enext = NULL;
	    rgn->rgn_enext = uctx->uctx_etail = nrgn;
	}
	else
	{
	    if (nbase >= base || ntop >= base)			/* overlapping region */
		return (-1);
	    
	    nrgn->rgn_enext = rgn;				/* insert before region */
	    nrgn->rgn_eprev = rgn->rgn_eprev;
	    rgn->rgn_eprev  = nrgn;
	    if (uctx->uctx_ergns == rgn)
		uctx->uctx_ergns = nrgn;
	    else
		nrgn->rgn_eprev->rgn_enext = nrgn;
	}
    }
    uctx->uctx_ergnlast = nrgn;
    
    return (0);
}

static USER_RGN *
user_removergn_elan (USER_CTXT *uctx, USER_RGN  *rgn)
{
    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) && kmutex_is_locked (&uctx->uctx_rgnmutex));

    uctx->uctx_ergnlast = rgn->rgn_enext;
    if (rgn == uctx->uctx_etail)
	uctx->uctx_etail = rgn->rgn_eprev;
    else
	rgn->rgn_enext->rgn_eprev = rgn->rgn_eprev;
    
    if (rgn == uctx->uctx_ergns)
	uctx->uctx_ergns = rgn->rgn_enext;
    else
	rgn->rgn_eprev->rgn_enext = rgn->rgn_enext;

    return (rgn);
}

USER_RGN *
user_rgnat_elan (USER_CTXT *uctx, E4_Addr addr)
{
    USER_RGN *rgn = user_findrgn_elan (uctx, addr, 0);

    if (rgn != NULL && rgn->rgn_ebase <= addr && addr <= (rgn->rgn_ebase + rgn->rgn_len - 1))
	return (rgn);

    return (NULL);
}

/* main address region management */
USER_RGN *
user_findrgn_main (USER_CTXT *uctx, virtaddr_t addr, int tail)
{
    USER_RGN  *rgn;
    USER_RGN  *hirgn;
    USER_RGN  *lorgn;
    virtaddr_t lastaddr;
    virtaddr_t base;
    int	       forward;
    
    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) || kmutex_is_locked (&uctx->uctx_rgnmutex));

    if (uctx->uctx_mrgns == NULL)
	return (NULL);
    
    rgn = uctx->uctx_mrgnlast;
    if (rgn == NULL)
	rgn = uctx->uctx_mrgns;
    
    forward = 0;
    if ((base = rgn->rgn_mbase) < addr)
    {
	if (addr <= (base + rgn->rgn_len - 1))
	    return (rgn);					/* ergnlast contained addr */
	
	hirgn = uctx->uctx_mtail;
	if ((lastaddr = hirgn->rgn_mbase + hirgn->rgn_len - 1) < addr)
	    return (tail ? hirgn : NULL);			/* addr is out of range */
	
	if ((addr - base) > (lastaddr - addr))
	    rgn = hirgn;
	else
	{
	    rgn = rgn->rgn_mnext;
	    forward++;
	}
    }
    else
    {
	lorgn = uctx->uctx_mrgns;
	if (lorgn->rgn_mbase > addr)
	    return (lorgn);					/* lowest regions is higher than addr */
	if ((addr - lorgn->rgn_mbase) < (base - addr))
	{
	    rgn = lorgn;					/* search forward from head */
	    forward++;
	}
    }
    if (forward)
    {
	while ((rgn->rgn_mbase + rgn->rgn_len - 1) < addr)
	    rgn = rgn->rgn_mnext;

	if (rgn->rgn_mbase <= addr)
	    uctx->uctx_mrgnlast = rgn;
	return (rgn);
    }
    else
    {
	while (rgn->rgn_mbase > addr)
	    rgn = rgn->rgn_mprev;

	if ((rgn->rgn_mbase + rgn->rgn_len - 1) < addr)
	    return (rgn->rgn_mnext);
	else
	{
	    uctx->uctx_mrgnlast = rgn;
	    return (rgn);
	}
    }
}

static int
user_addrgn_main (USER_CTXT *uctx, USER_RGN *nrgn)
{
    USER_RGN  *rgn   = user_findrgn_main (uctx, nrgn->rgn_mbase, 1);
    virtaddr_t nbase = nrgn->rgn_mbase;
    virtaddr_t ntop  = nbase + nrgn->rgn_len - 1;
    virtaddr_t base;

    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) && kmutex_is_locked (&uctx->uctx_rgnmutex));

    if (rgn == NULL)
    {
	uctx->uctx_mrgns = uctx->uctx_mtail = nrgn;
	nrgn->rgn_mnext = nrgn->rgn_mprev = NULL;
    }
    else
    {
	base = rgn->rgn_mbase;

	if ((base + rgn->rgn_len - 1) < nbase)			/* top of region below requested address */
	{							/* so insert after region (and hence at end */
	    nrgn->rgn_mprev = rgn;				/* of list */
	    nrgn->rgn_mnext = NULL;
	    rgn->rgn_mnext = uctx->uctx_mtail = nrgn;
	}
	else
	{
	    if (nbase >= base || ntop >= base)			/* overlapping region */
		return (-1);

	    nrgn->rgn_mnext = rgn;				/* insert before region */
	    nrgn->rgn_mprev = rgn->rgn_mprev;
	    rgn->rgn_mprev  = nrgn;
	    if (uctx->uctx_mrgns == rgn)
		uctx->uctx_mrgns = nrgn;
	    else
		nrgn->rgn_mprev->rgn_mnext = nrgn;
	}
    }
    uctx->uctx_mrgnlast = nrgn;
    
    return (0);
}

static USER_RGN *
user_removergn_main (USER_CTXT *uctx, USER_RGN *rgn)
{
    ASSERT (SPINLOCK_HELD (&uctx->uctx_rgnlock) && kmutex_is_locked (&uctx->uctx_rgnmutex));

    uctx->uctx_mrgnlast = rgn->rgn_mnext;
    if (rgn == uctx->uctx_mtail)
	uctx->uctx_mtail = rgn->rgn_mprev;
    else
	rgn->rgn_mnext->rgn_mprev = rgn->rgn_mprev;
    
    if (rgn == uctx->uctx_mrgns)
	uctx->uctx_mrgns = rgn->rgn_mnext;
    else
	rgn->rgn_mprev->rgn_mnext = rgn->rgn_mnext;

    return (rgn);
}

/* Remove whole region from both lists */
static void
user_removergn (USER_CTXT *uctx, USER_RGN *rgn)
{
    spin_lock (&uctx->uctx_rgnlock);

    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* XXXX tbl */, rgn->rgn_ebase, rgn->rgn_len);
	    
    user_removergn_elan (uctx, rgn);
    user_removergn_main (uctx, rgn);
    
    spin_unlock (&uctx->uctx_rgnlock);
    
    KMEM_FREE (rgn, sizeof (USER_RGN));
}

/* Remove all allocated regions */
void
user_freergns (USER_CTXT *uctx)
{
    kmutex_lock (&uctx->uctx_rgnmutex);

    while (uctx->uctx_mrgns)
	user_removergn(uctx, uctx->uctx_mrgns);

    kmutex_unlock (&uctx->uctx_rgnmutex);

    ASSERT (uctx->uctx_ergns == NULL);
}

USER_RGN *
user_rgnat_main (USER_CTXT *uctx, virtaddr_t addr)
{
    USER_RGN *rgn = user_findrgn_main (uctx, addr, 0);
    
    if (rgn != NULL && rgn->rgn_mbase <= addr && addr <= (rgn->rgn_mbase + rgn->rgn_len - 1))
	return (rgn);
    return (NULL);
}

int
user_setperm (USER_CTXT *uctx, virtaddr_t maddr, E4_Addr eaddr, unsigned long len, unsigned perm)
{
    USER_RGN *nrgn;

    PRINTF4 (uctx, DBG_PERM, "user_setperm: user %lx elan %llx len %lx perm %x\n", maddr, (long long) eaddr, len, perm);

    if ((maddr & PAGEOFFSET) || (eaddr & PAGEOFFSET) || (len & PAGEOFFSET))
    {
        PRINTF0 (uctx, DBG_PERM, "user_setperm:  alignment failure\n");
	return (-EINVAL);
    }
    
    if ((maddr + len - 1) <= maddr || (eaddr + len - 1) <= eaddr) 
    {
	PRINTF0 (uctx, DBG_PERM, "user_setperm:  range failure\n");
	return (-EINVAL);
    }

    KMEM_ALLOC (nrgn, USER_RGN *, sizeof (USER_RGN), 1);
    
    if (nrgn == NULL)
	return (-ENOMEM);

    nrgn->rgn_mbase = maddr;
    nrgn->rgn_ebase = eaddr;
    nrgn->rgn_len   = len;
    nrgn->rgn_perm  = perm;

    kmutex_lock (&uctx->uctx_rgnmutex);
    spin_lock (&uctx->uctx_rgnlock);

    if (user_addrgn_elan (uctx, nrgn) < 0)
    {
	PRINTF0 (uctx, DBG_PERM, "user_setperm:  elan address exists\n");
	spin_unlock (&uctx->uctx_rgnlock);
	kmutex_unlock (&uctx->uctx_rgnmutex);

	KMEM_FREE (nrgn, sizeof (USER_RGN));
	return (-EINVAL);
    }
    
    if (user_addrgn_main (uctx, nrgn) < 0)
    {
	PRINTF0 (uctx, DBG_PERM, "user_setperm:  main address exists\n");
	user_removergn_elan (uctx, nrgn);
	
	spin_unlock (&uctx->uctx_rgnlock);
	kmutex_unlock (&uctx->uctx_rgnmutex);

	KMEM_FREE (nrgn, sizeof (USER_RGN));
	return (-EINVAL);
    }
    spin_unlock (&uctx->uctx_rgnlock);

    if ((perm & PERM_Preload))
	user_preload_main (uctx, maddr, len);

    kmutex_unlock (&uctx->uctx_rgnmutex);

    return (0);
}

void
user_clrperm (USER_CTXT *uctx, E4_Addr addr, unsigned long len)
{
    E4_Addr       raddr;
    E4_Addr       rtop;
    USER_RGN     *nrgn;
    USER_RGN     *rgn;
    USER_RGN     *rgn_next;
    unsigned long ssize;
    int	          res;

    PRINTF2 (uctx, DBG_PERM, "user_clrperm: elan %llx len %lx\n", addr, len);

    raddr = (addr & PAGEMASK);
    rtop = ((addr + len - 1) & PAGEMASK) + (PAGESIZE-1);

    kmutex_lock (&uctx->uctx_rgnmutex);
    
    for (rgn = user_findrgn_elan (uctx, addr, 0); rgn != NULL; rgn = rgn_next)
    {
	if (rtop < rgn->rgn_ebase)				/* rtop was in a gap */
	    break;
	
	rgn_next = rgn->rgn_enext;				/* Save next region pointer */
	
	PRINTF (uctx, DBG_PERM, "              elan %llx->%llx main %p->%p\n", 
		rgn->rgn_ebase, rgn->rgn_ebase + rgn->rgn_len-1,
		rgn->rgn_mbase, rgn->rgn_mbase + rgn->rgn_len-1);

	if (raddr <= rgn->rgn_ebase && rtop >= (rgn->rgn_ebase + rgn->rgn_len - 1))
	{
	    /* whole region is cleared */

	    PRINTF (uctx, DBG_PERM, "              whole region\n");
	    PRINTF (uctx, DBG_PERM, "              unload elan %llx->%llx\n", rgn->rgn_ebase, rgn->rgn_ebase + rgn->rgn_len-1);
	    user_removergn (uctx, rgn);
	}
	else if (raddr <= rgn->rgn_ebase)
	{
	    /* clearing at beginning, so shrink size and increment base ptrs */
	    ssize = rtop - rgn->rgn_ebase + 1;
	    
	    PRINTF (uctx, DBG_PERM, "              clear at beginning %x\n", ssize);

	    spin_lock (&uctx->uctx_rgnlock);

	    PRINTF (uctx, DBG_PERM, "              unload elan %llx->%llx\n", rgn->rgn_ebase, rgn->rgn_ebase + ssize-1);
	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* XXXX tbl */, rgn->rgn_ebase, ssize);

	    rgn->rgn_mbase += ssize;
	    rgn->rgn_ebase += ssize;
	    rgn->rgn_len   -= ssize;
	    
	    spin_unlock(&uctx->uctx_rgnlock);
	}
	else if (rtop >= (rgn->rgn_ebase + rgn->rgn_len - 1))
	{
	    /* clearing at end, so just shrink length of region */
	    ssize = (rgn->rgn_ebase + rgn->rgn_len - 1) - raddr + 1;

	    PRINTF (uctx, DBG_PERM, "              clear at end %x\n", ssize);

	    spin_lock (&uctx->uctx_rgnlock);

	    PRINTF (uctx, DBG_PERM, "              unload elan %llx->%llx\n", raddr, raddr+ssize-1);
	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* XXXX tbl */, raddr, ssize);

	    rgn->rgn_len -= ssize;

	    spin_unlock(&uctx->uctx_rgnlock);
	}
	else
	{
	    /* the section to go is in the middle,  so need to  */
	    /* split it into two regions */
	    KMEM_ALLOC (nrgn, USER_RGN *, sizeof (USER_RGN), 1);

	    spin_lock (&uctx->uctx_rgnlock);

	    PRINTF (uctx, DBG_PERM, "              unload elan %llx->%llx\n", raddr, rtop);
	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* XXXX tbl */, raddr, rtop - raddr + 1);

	    nrgn->rgn_mbase = rgn->rgn_mbase + (rtop - rgn->rgn_ebase + 1);
	    nrgn->rgn_ebase = rtop + 1;
	    nrgn->rgn_len   = (rgn->rgn_ebase + rgn->rgn_len - 1) - rtop;
	    nrgn->rgn_perm  = rgn->rgn_perm;

	    PRINTF (uctx, DBG_PERM, "              new elan %llx->%llx main %p->%p\n", 
		    nrgn->rgn_ebase, nrgn->rgn_ebase + nrgn->rgn_len-1,
		    nrgn->rgn_mbase, nrgn->rgn_mbase + nrgn->rgn_len-1);

	    rgn->rgn_len = (raddr - rgn->rgn_ebase);		/* shrink original region */

	    PRINTF (uctx, DBG_PERM, "              old elan %llx->%llx main %p->%p\n", 
		    rgn->rgn_ebase, rgn->rgn_ebase + rgn->rgn_len-1,
		    rgn->rgn_mbase, rgn->rgn_mbase + rgn->rgn_len-1);

	    res = user_addrgn_elan (uctx, nrgn);		/* insert new region */
	    ASSERT (res == 0);					/* which cannot fail */

	    res = user_addrgn_main (uctx, nrgn);	
	    ASSERT (res == 0);

	    spin_unlock(&uctx->uctx_rgnlock);
	}
    }
    kmutex_unlock (&uctx->uctx_rgnmutex);
}

int
user_checkperm (USER_CTXT *uctx, E4_Addr raddr, unsigned long rsize, unsigned access)
{
    USER_RGN *rgn;

    PRINTF3 (uctx, DBG_PERM, "user_checkperm: elan %lx len %lx access %x\n", raddr, rsize, access);
    
    if ((raddr + rsize - 1) < raddr)
	return (-ENOMEM);
    
    kmutex_lock (&uctx->uctx_rgnmutex);
    if ((rgn = user_rgnat_elan (uctx, raddr)) == (USER_RGN *) NULL)
    {
	kmutex_unlock (&uctx->uctx_rgnmutex);
	return (-ENOMEM);
    }
    else
    {
	register int ssize;
	
	for (; rsize != 0; rsize -= ssize, raddr += ssize)
	{
	    if (raddr > (rgn->rgn_ebase + rgn->rgn_len - 1))
	    {
		rgn  = rgn->rgn_enext;
		
		if (rgn == NULL || raddr != rgn->rgn_ebase)
		{
		    kmutex_unlock (&uctx->uctx_rgnmutex);
		    return (-ENOMEM);
		}
	    }
	    if ((raddr + rsize - 1) > (rgn->rgn_ebase + rgn->rgn_len - 1))
		ssize = ((rgn->rgn_ebase + rgn->rgn_len - 1) - raddr) + 1;
	    else
		ssize = rsize;
	    
	    PRINTF4 (uctx, DBG_PERM, "user_checkperm : rgn %lx -> %lx perm %x access %x\n",
		     rgn->rgn_ebase, rgn->rgn_ebase + (E4_Addr)rgn->rgn_len, rgn->rgn_perm, access);

	    if (ELAN4_INCOMPAT_ACCESS (rgn->rgn_perm, access))
	    {
		kmutex_unlock (&uctx->uctx_rgnmutex);
		return (-EACCES);
	    }
	}
    }
    
    kmutex_unlock (&uctx->uctx_rgnmutex);
    
    return (0);
}

virtaddr_t
user_elan2main (USER_CTXT *uctx, E4_Addr addr)
{
    USER_RGN  *rgn;
    virtaddr_t raddr;
    
    spin_lock (&uctx->uctx_rgnlock);
    
    if ((rgn = user_rgnat_elan (uctx, addr)) == (USER_RGN *) NULL)
	raddr = (virtaddr_t) 0;
    else
	raddr = rgn->rgn_mbase + (addr - rgn->rgn_ebase);

    spin_unlock (&uctx->uctx_rgnlock);

    return (raddr);
}

E4_Addr
user_main2elan (USER_CTXT *uctx, virtaddr_t addr)
{
    USER_RGN *rgn;
    E4_Addr   raddr;

    spin_lock (&uctx->uctx_rgnlock);
    
    if ((rgn = user_rgnat_main (uctx, addr)) == (USER_RGN *) NULL)
	raddr = (virtaddr_t) 0;
    else
	raddr = rgn->rgn_ebase + (addr - rgn->rgn_mbase);
    
    spin_unlock (&uctx->uctx_rgnlock);

    return (raddr);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
