/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: sdram.c,v 1.34.2.1 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/sdram.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>

EXPORT_SYMBOL_GPL(elan4_sdram_readb);
EXPORT_SYMBOL_GPL(elan4_sdram_readw);
EXPORT_SYMBOL_GPL(elan4_sdram_readl);
EXPORT_SYMBOL_GPL(elan4_sdram_readq);
EXPORT_SYMBOL_GPL(elan4_sdram_writeb);
EXPORT_SYMBOL_GPL(elan4_sdram_writew);
EXPORT_SYMBOL_GPL(elan4_sdram_writel);
EXPORT_SYMBOL_GPL(elan4_sdram_writeq);
EXPORT_SYMBOL_GPL(elan4_sdram_zerob_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_zerow_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_zerol_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_zeroq_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyb_from_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyw_from_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyl_from_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyq_from_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyb_to_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyw_to_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyl_to_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_copyq_to_sdram);
EXPORT_SYMBOL_GPL(elan4_sdram_alloc);
EXPORT_SYMBOL_GPL(elan4_sdram_free);
EXPORT_SYMBOL_GPL(elan4_sdram_flushcache);

#define SDRAM_MIN_BANK_SIZE		((1 << 15) * 8)		/* 256 Kbytes */

static inline ELAN4_SDRAM_BANK *
sdramaddr_to_bank (ELAN4_DEV *dev, sdramaddr_t saddr)
{
    register int i;
    
    for (i = 0; i < dev->dev_sdram_numbanks; i++)
    {
	ELAN4_SDRAM_BANK *bank = &dev->dev_sdram_banks[i];

	if (saddr >= bank->b_base && saddr < (bank->b_base + bank->b_size))
	    return (bank);
    }
    printk ("sdramaddr_to_bank: sdram address %lx not in a sdram bank\n", saddr);
    BUG();

    return (NULL);	/* NOTREACHED */
}

static inline int
sdramaddr_to_bankoffset (ELAN4_DEV *dev, sdramaddr_t saddr)
{
    return (saddr & (sdramaddr_to_bank (dev, saddr)->b_size-1));
}

static inline int
sdramaddr_to_bit(ELAN4_DEV *dev, int indx, sdramaddr_t saddr)
{
    return (sdramaddr_to_bankoffset(dev, saddr) >> (SDRAM_MIN_BLOCK_SHIFT+(indx)));
}

static inline ioaddr_t
sdramaddr_to_ioaddr (ELAN4_DEV *dev, sdramaddr_t saddr)
{
    ELAN4_SDRAM_BANK *bank = sdramaddr_to_bank (dev, saddr);

    return (bank->b_ioaddr + (saddr - bank->b_base));
}

unsigned char
elan4_sdram_readb (ELAN4_DEV *dev, sdramaddr_t off)
{
    return (__elan4_readb (dev, sdramaddr_to_ioaddr(dev, off)));
}

unsigned short
elan4_sdram_readw (ELAN4_DEV *dev, sdramaddr_t off)
{
    return (__elan4_readw (dev, sdramaddr_to_ioaddr(dev, off)));
}

unsigned int
elan4_sdram_readl (ELAN4_DEV *dev, sdramaddr_t off)
{
    return (__elan4_readl (dev, sdramaddr_to_ioaddr(dev, off)));
}

unsigned long long
elan4_sdram_readq (ELAN4_DEV *dev, sdramaddr_t off)
{
    return (__elan4_readq (dev, sdramaddr_to_ioaddr(dev, off)));
}

void
elan4_sdram_writeb (ELAN4_DEV *dev, sdramaddr_t off, unsigned char val)
{
    writeb (val, (void *) sdramaddr_to_ioaddr(dev, off));

    mb();
}

void
elan4_sdram_writew (ELAN4_DEV *dev, sdramaddr_t off, unsigned short val)
{
    writew (val, (void *) sdramaddr_to_ioaddr(dev, off));

    mb();
}

void
elan4_sdram_writel (ELAN4_DEV *dev, sdramaddr_t off, unsigned int val)
{
    writel (val, (void *) (sdramaddr_to_ioaddr(dev, off)));

    mb();
}

void
elan4_sdram_writeq (ELAN4_DEV *dev, sdramaddr_t off, unsigned long long val)
{
    writeq (val, (void *) (sdramaddr_to_ioaddr(dev, off)));

    mb();
}

void
elan4_sdram_zerob_sdram (ELAN4_DEV *dev, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u8))
	writeb (0, (void *) dest);
}

void
elan4_sdram_zerow_sdram (ELAN4_DEV *dev, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u8))
	writeb (0, (void *) dest);
}

void
elan4_sdram_zerol_sdram (ELAN4_DEV *dev, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u32))
	writel (0, (void *) dest);
}

void
elan4_sdram_zeroq_sdram (ELAN4_DEV *dev, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    ioaddr_t lim  = dest + nbytes;

#ifdef CONFIG_MPSAS
    if (sas_memset_dev (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM, to, 0, nbytes) == 0)
	return;
#endif

    for (; dest < lim; dest += sizeof (u64))
	writeq (0, (void *) dest);
}

void
elan4_sdram_copyb_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
    ioaddr_t src  = sdramaddr_to_ioaddr (dev, from);
    u8      *dest = (u8 *) to;
    ioaddr_t lim  = src + nbytes;

    for (; src < lim; src += sizeof (u8))
	*dest++ = __elan4_readb (dev, src);
}

void
elan4_sdram_copyw_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
    ioaddr_t src  = sdramaddr_to_ioaddr (dev, from);
    u16     *dest = (u16 *) to;
    ioaddr_t lim  = src + nbytes;

    for (; src < lim; src += sizeof (u16))
	*dest++ = __elan4_readw (dev, src);
}

void
elan4_sdram_copyl_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
    ioaddr_t src  = sdramaddr_to_ioaddr (dev, from);
    u32     *dest = (u32 *) to;
    ioaddr_t lim  = src + nbytes;

    for (; src < lim; src += sizeof (u32))
	*dest++ = __elan4_readl (dev, src);
}

void
elan4_sdram_copyq_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes)
{
    ioaddr_t src  = sdramaddr_to_ioaddr (dev, from);
    u64     *dest = (u64 *) to;
    ioaddr_t lim  = src + nbytes;

#ifdef CONFIG_MPSAS
    if (sas_copyfrom_dev (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM, from, (unsigned long) to, nbytes) == 0)
	return;
#endif

    for (; src < lim; src += sizeof (u64))
	*dest++ = __elan4_readq (dev, src);
}

void
elan4_sdram_copyb_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    u8      *src  = (u8 *) from;
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u8))
	writeb (*src++, (void *) (dest));

    mb();
}

void
elan4_sdram_copyw_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    u16     *src  = (u16 *) from;
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u16))
	writew (*src++, (void *) (dest));

    mb();
}

void
elan4_sdram_copyl_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    u32     *src  = (u32 *) from;
    ioaddr_t lim  = dest + nbytes;

    for (; dest < lim; dest += sizeof (u16))
	writew (*src++, (void *) (dest));

    mb();
}

void
elan4_sdram_copyq_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes)
{
    ioaddr_t dest = sdramaddr_to_ioaddr (dev, to);
    u64     *src  = (u64 *) from;
    ioaddr_t lim  = dest + nbytes;

#ifdef CONFIG_MPSAS
    if (sas_copyto_dev (dev->dev_osdep.pdev, ELAN4_BAR_SDRAM, to, (unsigned long) from, nbytes) == 0)
	return;
#endif

    for (; dest < lim; dest += sizeof (u64))
	writeq (*src++, (void *) (dest));

    mb();
}

/* sdram buddy allocator */
typedef struct sdramblock
{
    sdramaddr_t	next;
    sdramaddr_t prev;
} sdramblock_t;

static inline sdramaddr_t
read_next (ELAN4_DEV *dev, sdramaddr_t block)
{
    return __elan4_readl (dev, sdramaddr_to_ioaddr (dev, block + offsetof (sdramblock_t, next)));
}

static inline sdramaddr_t
read_prev (ELAN4_DEV *dev, sdramaddr_t block)
{
    return __elan4_readl (dev, sdramaddr_to_ioaddr (dev, block + offsetof (sdramblock_t, prev)));
}

static inline void
write_next (ELAN4_DEV *dev, sdramaddr_t block, sdramaddr_t val)
{
    writel (val, (void *) (sdramaddr_to_ioaddr (dev, block + offsetof (sdramblock_t, next))));
}

static inline void
write_prev (ELAN4_DEV *dev, sdramaddr_t block, sdramaddr_t val)
{
    writel (val, (void *) (sdramaddr_to_ioaddr (dev, block + offsetof (sdramblock_t, prev))));
}

static inline void
freelist_insert (ELAN4_DEV *dev, int idx, sdramaddr_t block)
{
    sdramaddr_t next = dev->dev_sdram_freelists[(idx)];

    /*
     * block->prev = NULL;
     * block->next = next;
     * if (next != NULL)
     *    next->prev = block;
     * freelist = block;
     */
    write_prev (dev, block, (sdramaddr_t) 0);
    write_next (dev, block, next);
    if (next != (sdramaddr_t) 0)
	write_prev (dev, next, block);
    dev->dev_sdram_freelists[idx] = block;

    dev->dev_sdram_freecounts[idx]++;
    dev->dev_stats.s_sdram_bytes_free += (SDRAM_MIN_BLOCK_SIZE << idx);

    mb();
}

static inline void
freelist_remove (ELAN4_DEV *dev,int idx, sdramaddr_t block)
{
    /*
     * if (block->prev)
     *     block->prev->next = block->next;
     * else
     *     dev->dev_sdram_freelists[idx] = block->next;
     * if (block->next)
     *     block->next->prev = block->prev;
     */
    sdramaddr_t blocknext = read_next (dev, block);
    sdramaddr_t blockprev = read_prev (dev, block);

    if (blockprev)
	write_next (dev, blockprev, blocknext);
    else
	dev->dev_sdram_freelists[idx] = blocknext;
    if (blocknext)
	write_prev (dev, blocknext, blockprev);

    dev->dev_sdram_freecounts[idx]--;
    dev->dev_stats.s_sdram_bytes_free -= (SDRAM_MIN_BLOCK_SIZE << idx);

    mb();
}

static inline void
freelist_removehead(ELAN4_DEV *dev, int idx, sdramaddr_t block)
{
    sdramaddr_t blocknext = read_next (dev, block);

    if ((dev->dev_sdram_freelists[idx] = blocknext) != 0)
	write_prev (dev, blocknext, 0);

    dev->dev_sdram_freecounts[idx]--;
    dev->dev_stats.s_sdram_bytes_free -= (SDRAM_MIN_BLOCK_SIZE << idx);

    mb();
}

#ifdef DEBUG
static int
display_blocks (ELAN4_DEV *dev, int indx, char *string)
{
    sdramaddr_t block;
    int nbytes = 0;

    PRINTF (DBG_DEVICE, DBG_SDRAM, "%s - indx %d\n", string, indx);
    for (block = dev->dev_sdram_freelists[indx]; block != (sdramaddr_t) 0; block = read_next (dev, block))
    {
	PRINTF (DBG_DEVICE, DBG_SDRAM, "  %x\n", block);
	nbytes += (SDRAM_MIN_BLOCK_SIZE << indx);
    }

    return (nbytes);
}

void
elan4_sdram_display (ELAN4_DEV *dev, char *string)
{
    int indx;
    int nbytes = 0;
    
    PRINTF (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_display: dev=%p\n", dev);
    for (indx = 0; indx < SDRAM_NUM_FREE_LISTS; indx++)
	if (dev->dev_sdram_freelists[indx] != (sdramaddr_t) 0)
	    nbytes += display_blocks (dev, indx, string);
    PRINTF (DBG_DEVICE, DBG_SDRAM, "\n%d bytes free - %d pages free\n", nbytes, nbytes/SDRAM_PAGE_SIZE);
}

void
elan4_sdram_verify (ELAN4_DEV *dev)
{
    int indx, size, nbits, i, b;
    sdramaddr_t block;

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	unsigned count = 0;

	for (block = dev->dev_sdram_freelists[indx]; block; block = read_next (dev, block), count++)
	{
	    ELAN4_SDRAM_BANK *bank = sdramaddr_to_bank (dev, block);
	    unsigned         off  = sdramaddr_to_bankoffset (dev, block);
	    int              bit  = sdramaddr_to_bit (dev, indx, block);

	    if ((block & (size-1)) != 0)
		printk ("elan4_sdram_verify: block=%lx indx=%x - not aligned\n", block, indx);
	    
	    if (bank == NULL || off > bank->b_size)
		printk ("elan4_sdram_verify: block=%lx indx=%x - outside bank\n", block, indx);
	    else if (BT_TEST (bank->b_bitmaps[indx], bit) == 0)
		printk ("elan4_sdram_verify: block=%lx indx=%x - bit not set\n", block, indx);
	    else
	    {
		for (i = indx-1, nbits = 2; i >= 0; i--, nbits <<= 1)
		{
		    bit = sdramaddr_to_bit (dev, i, block);

		    for (b = 0; b < nbits; b++)
			if (BT_TEST(bank->b_bitmaps[i], bit + b))
			    printk ("elan4_sdram_verify: block=%lx indx=%x - also free i=%d bit=%x\n", block, indx, i, bit+b);
		}
	    }
	}

	if (dev->dev_sdram_freecounts[indx] != count)
	    printk ("elan4_sdram_verify: indx=%x expected %d got %d\n", indx, dev->dev_sdram_freecounts[indx], count);
    }
}

#endif

static void
free_block (ELAN4_DEV *dev, sdramaddr_t block, int indx)
{
    ELAN4_SDRAM_BANK *bank = sdramaddr_to_bank (dev, block);
    unsigned         bit  = sdramaddr_to_bit (dev, indx, block);
    unsigned         size = SDRAM_MIN_BLOCK_SIZE << indx;

    PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: block=%x indx=%d bit=%x\n", block, indx, bit);

    ASSERT ((block & (size-1)) == 0);
    ASSERT (BT_TEST (bank->b_bitmaps[indx], bit) == 0);

    while (BT_TEST (bank->b_bitmaps[indx], bit ^ 1))
    {
	sdramaddr_t buddy = block ^ size;
	
	PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: merge block=%x buddy=%x indx=%d\n", block, buddy, indx);
	
	BT_CLEAR (bank->b_bitmaps[indx], bit ^ 1);
	
	freelist_remove (dev, indx, buddy);
	
	block = (block < buddy) ? block : buddy;
	indx++;
	size <<= 1;
	bit >>= 1;
    }
    
    PRINTF3 (DBG_DEVICE, DBG_SDRAM, "free_block: free block=%x indx=%d bit=%x\n", block, indx, bit);
    
    freelist_insert (dev, indx, block);
    
    BT_SET (bank->b_bitmaps[indx], bit);
}

void
elan4_sdram_init (ELAN4_DEV *dev)
{
    int indx;

    spin_lock_init (&dev->dev_sdram_lock);

    for (indx = 0; indx < SDRAM_NUM_FREE_LISTS; indx++)
    {
	dev->dev_sdram_freelists[indx]  = (sdramaddr_t) 0;
	dev->dev_sdram_freecounts[indx] = 0;
    }
}

void
elan4_sdram_fini (ELAN4_DEV *dev)
{
    spin_lock_destroy (&dev->dev_sdram_lock);
}

#ifdef CONFIG_MPSAS
/* size of Elan SDRAM in simulation */
#define SDRAM_used_addr_bits		(16)
#define SDRAM_SIMULATION_BANK_SIZE	((1 << SDRAM_used_addr_bits) * 8)	/* 128 kbytes */

static int
elan4_sdram_probe_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    printk ("elan%d: memory bank %d is %d Kb\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks), (int) (SDRAM_SIMULATION_BANK_SIZE / 1024));

    bank->b_size = SDRAM_SIMULATION_BANK_SIZE;

    return 1;
}

#else

static void
initialise_cache_tags (ELAN4_DEV *dev, unsigned addr)
{
    register int set, line;

    mb();

    /* Initialise the whole cache to hold sdram at "addr" as direct mapped */

    for (set = 0; set < E4_NumCacheSets; set++)
	for (line = 0; line < E4_NumCacheLines; line++)
	    write_tag (dev, Tags[set][line], addr | (set << 13) | (1 << 11));

    read_tag (dev, Tags[set][line]);	/* read it back to guarantee the memory system is quite again */
    mb();
}

static __inline__ int
sdram_GreyToBinary(int GreyVal, int NoOfBits)
{
    int Bit;
    int BinaryVal=0;
    for (Bit=(1 << (NoOfBits-1)); Bit != 0; Bit >>= 1)
	BinaryVal ^= (GreyVal & Bit) ^ ((BinaryVal >> 1) & Bit);
    return (BinaryVal);
}

static __inline__ int
sdram_BinaryToGrey(int BinaryVal)
{
    return (BinaryVal ^ (BinaryVal >> 1));
}

void
elan4_sdram_setup_delay_lines (ELAN4_DEV *dev, int factor)
{
    /* This is used to fix the SDRAM delay line values */
    int i, AutoGenDelayValue=0;
    int NewDelayValue;

    if (dev->dev_sdram_cfg & SDRAM_FIXED_DELAY_ENABLE)	  /* already setup. */
	return;

    /* now get an average of 10 dll values */
    for (i=0;i<10;i++)
       AutoGenDelayValue += sdram_GreyToBinary(SDRAM_GET_DLL_DELAY(read_reg64 (dev, SDRamConfigReg)),
					       SDRAM_FIXED_DLL_DELAY_BITS);

    NewDelayValue = factor + (AutoGenDelayValue / 10); /* Mean of 10 values */

    dev->dev_sdram_cfg = (dev->dev_sdram_cfg & ~(SDRAM_FIXED_DLL_DELAY_MASK << SDRAM_FIXED_DLL_DELAY_SHIFT)) |
			  SDRAM_FIXED_DELAY_ENABLE | SDRAM_FIXED_DLL_DELAY(sdram_BinaryToGrey(NewDelayValue));

    write_reg64 (dev, SDRamConfigReg, dev->dev_sdram_cfg);	/* Put back the new value */

    pioflush_reg (dev);
}

static int
elan4_sdram_probe_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    unsigned long      mappedsize = bank->b_size;
    ioaddr_t           ioaddr;
    unsigned long long value, size;
    register int       i;
    extern int         sdram_bank_limit;

    if (mappedsize > SDRAM_MAX_BLOCK_SIZE)
	mappedsize = SDRAM_MAX_BLOCK_SIZE;

    while ((ioaddr = elan4_map_device (dev, ELAN4_BAR_SDRAM, bank->b_base, mappedsize, &bank->b_handle)) == 0)
    {
	if (mappedsize <= (64*1024*1024))			/* boards normally populated with 64mb, so winge if we can't see this much */
	    printk ("elan%d: could not map bank %d size %dMb\n", dev->dev_instance, (int)(bank - dev->dev_sdram_banks), (int)mappedsize/(1024*1024));

	if ((mappedsize >>= 1) < (1024*1024))
	    return 0;
    }

    /* first probe to see if the memory bank is present */
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
       initialise_cache_tags (dev, E4_CacheSize);

    for (i = 0; i < 64; i++)
    {
	unsigned long long pattern = (1ull << i);

	writeq (pattern, (void *)ioaddr);					/* write pattern at base  */

        if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	    initialise_cache_tags (dev, 0);

	writeq (~pattern, (void *)(ioaddr + E4_CacheSize));			/* write ~pattern at cachesize */

        if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	   initialise_cache_tags (dev, E4_CacheSize);
	
	writeq (~pattern, (void *)(ioaddr + 2*E4_CacheSize));			/* write ~pattern at 2*cachesize */
        if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	    initialise_cache_tags (dev, 2*E4_CacheSize);
	
	value = __elan4_readq (dev, ioaddr);				/* read pattern back at 0 */
	
	if (value != pattern)
	{
	    printk ("elan%d: sdram bank %d not present\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks));
	    elan4_unmap_device (dev, ioaddr, mappedsize, &bank->b_handle);
	    return 0;
	}
    }
    
    /* sdram bank is present, so work out it's size.  We store the maximum size at the base
     * and then store the address at each address on every power of two address until
     * we reach the minimum mappable size (PAGESIZE), we then read back the value at the
     * base to determine the bank size */
    writeq (mappedsize, (void *)(ioaddr));
    if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
        initialise_cache_tags (dev, 0);

    for (size = mappedsize >> 1; size > PAGE_SIZE; size >>= 1)
    {
	writeq (size, (void *)(ioaddr + (long)size));
        if (dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA)
	    initialise_cache_tags (dev, size);
    }

    if ((size = __elan4_readq (dev, ioaddr)) < SDRAM_MIN_BANK_SIZE)
    {
	printk ("elan%d: memory bank %d dubious\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks));
	elan4_unmap_device (dev, ioaddr, mappedsize, &bank->b_handle);
	return 0;
    }

    if (sdram_bank_limit == 0 || size <= (sdram_bank_limit * 1024 * 1024))
	printk ("elan%d: memory bank %d is %d Mb\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks), (int) (size / (1024*1024)));
    else
    {
	size = (sdram_bank_limit * 1024 * 1024);
	printk ("elan%d: limit bank %d to %d Mb\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks), (int) (size / (1024*1024)));
    }

    bank->b_size = size;

    elan4_unmap_device (dev, ioaddr, mappedsize, &bank->b_handle);
    return 1;
}
#endif

int
elan4_sdram_init_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    int indx, size;

    bank->b_ioaddr = 0;

    if (! elan4_sdram_probe_bank (dev, bank))
	return 0;

    if ((bank->b_ioaddr = elan4_map_device (dev, ELAN4_BAR_SDRAM, bank->b_base, bank->b_size, &bank->b_handle)) == (ioaddr_t) 0)
    {
	printk ("elan%d: could not map sdrambank %d\n", dev->dev_instance, (int) (bank - dev->dev_sdram_banks));
	return 0;
    }

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size <= bank->b_size; indx++, size <<= 1) /* allocate the buddy allocator bitmaps */
	KMEM_ZALLOC (bank->b_bitmaps[indx], bitmap_t *, sizeof (bitmap_t) * BT_BITOUL(bank->b_size/size), 1);
    
    return 1;
}

void
elan4_sdram_fini_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    int indx, size;

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size <= bank->b_size; indx++, size <<= 1)
	KMEM_FREE (bank->b_bitmaps[indx], sizeof (bitmap_t) * BT_BITOUL(bank->b_size/size));
    
    elan4_unmap_device (dev, bank->b_ioaddr, bank->b_size, &bank->b_handle);
}

void
elan4_sdram_add_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank)
{
    sdramaddr_t base = bank->b_base;
    sdramaddr_t top  = bank->b_base + bank->b_size;
    register int indx;
    register unsigned long size;

    /* align to the minimum block size */
    base = (base + SDRAM_MIN_BLOCK_SIZE - 1) & ~((sdramaddr_t) SDRAM_MIN_BLOCK_SIZE-1);
    top &= ~((sdramaddr_t) SDRAM_MIN_BLOCK_SIZE-1);

    /* don't allow 0 as a valid "base" */
    if (base == 0)
	base = SDRAM_MIN_BLOCK_SIZE;

    /* carve the bottom to the biggest boundary */
    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	if ((base & size) == 0)
	    continue;

	if ((base + size) > top)
	    break;

	free_block (dev, base, indx);
	
	base += size;
    }

    /* carve the top down to the biggest boundary */
    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; indx < SDRAM_NUM_FREE_LISTS; indx++, size <<= 1)
    {
	if ((top & size) == 0)
	    continue;

	if ((top - size) < base)
	    break;

	free_block (dev, (top - size), indx);
	
	top -= size;
    }

    /* now free of the space in between */
    while (base < top)
    {
	free_block (dev, base, (SDRAM_NUM_FREE_LISTS-1));

	base += SDRAM_MAX_BLOCK_SIZE;
    }
}

sdramaddr_t
elan4_sdram_alloc (ELAN4_DEV *dev, int nbytes)
{
    sdramaddr_t block;
    register int i, indx;
    unsigned long size;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_sdram_lock, flags);

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size < nbytes; indx++, size <<= 1)
	;

    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_alloc: nbytes=%d indx=%d\n", nbytes, indx);

    /* need to split a bigger block up */
    for (i = indx; i < SDRAM_NUM_FREE_LISTS; i++, size <<= 1)
	if (dev->dev_sdram_freelists[i])
	    break;
    
    if (i == SDRAM_NUM_FREE_LISTS)
    {
	spin_unlock_irqrestore (&dev->dev_sdram_lock, flags);
	printk ("elan4_sdram_alloc: %d bytes failed\n", nbytes);
	return ((sdramaddr_t) 0);
    }
    
    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_alloc: use block=%x indx=%d\n", dev->dev_sdram_freelists[i], i);

    /* remove the block from the free list */
    freelist_removehead (dev, i, (block = dev->dev_sdram_freelists[i]));

    /* clear the approriate bit in the bitmap */
    BT_CLEAR (sdramaddr_to_bank (dev, block)->b_bitmaps[i], sdramaddr_to_bit (dev,i, block));

    /* and split it up as required */
    while (i-- > indx)
	free_block (dev, block + (size >>= 1), i);

    spin_unlock_irqrestore (&dev->dev_sdram_lock, flags);

    ASSERT ((block & ((SDRAM_MIN_BLOCK_SIZE << (indx))-1)) == 0);

#ifdef CONFIG_MPSAS
    elan4_sdram_zeroq_sdram (dev, block, sizeof (sdramblock_t));
#endif

    return ((sdramaddr_t) block);
}

void
elan4_sdram_free (ELAN4_DEV *dev, sdramaddr_t block, int nbytes)
{
    register int indx;
    unsigned long size;
    unsigned long flags;

    spin_lock_irqsave (&dev->dev_sdram_lock, flags);

    for (indx = 0, size = SDRAM_MIN_BLOCK_SIZE; size < nbytes; indx++, size <<= 1)
	;

    PRINTF2 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_free: indx=%d block=%x\n", indx, block);

    free_block (dev, block, indx);

    spin_unlock_irqrestore (&dev->dev_sdram_lock, flags);
}

void
elan4_sdram_flushcache (ELAN4_DEV *dev, sdramaddr_t addr, int len)
{
    int set, off;

    SET_SYSCONTROL (dev, dev_direct_map_pci_writes, CONT_DIRECT_MAP_PCI_WRITES);

    /*
     * if flushing more than a single set (8K), then you have to flush the whole cache.
     *   NOTE - in the real world we will probably want to generate a burst across
     *          the pci bus.
     */
    if (len >= E4_CacheSetSize)
    {
	PRINTF3 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_flushcache: addr=%x len=%x (%x) => whole cache\n", addr, len, addr + len);

#ifdef CONFIG_MPSAS
	elan4_sdram_zeroq_sdram (dev, dev->dev_cacheflush_space, E4_CacheSize);
#else
	for (set = 0; set < E4_NumCacheSets; set++)
	    for (off = 0; off < E4_CacheSetSize; off += E4_CacheLineSize)
		elan4_sdram_writeq (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + off, 0);
#endif
    }
    else
    {
	unsigned base    = addr & ~(E4_CACHELINE_SIZE-1);
	unsigned top     = (addr + len + (E4_CACHELINE_SIZE-1)) & ~(E4_CACHELINE_SIZE-1);
	unsigned baseoff = base & (E4_CacheSetSize-1);
	unsigned topoff  = top  & (E4_CacheSetSize-1);

	if ((base ^ top) & E4_CacheSetSize)			/* wraps */
	{
	    PRINTF7 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_flushcache: addr=%x len=%x (%x) => split cache (%x,%x %x,%x)\n", 
		     addr, len, addr + len, 0, topoff, baseoff, E4_CacheSetSize);

#ifdef CONFIG_MPSAS
	    for (set = 0; set < E4_NumCacheSets; set++)
	    {
		elan4_sdram_zeroq_sdram (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize), topoff);
		elan4_sdram_zeroq_sdram (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + baseoff, E4_CacheSetSize - baseoff);
	    }
#else
	    for (set = 0; set < E4_NumCacheSets; set++)
	    {
		for (off = 0; off < (top & (E4_CacheSetSize-1)); off += E4_CACHELINE_SIZE)
		    elan4_sdram_writeq (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + off, 0);
		
		for (off = (base & (E4_CacheSetSize-1)); off < E4_CacheSetSize; off += E4_CACHELINE_SIZE)
		    elan4_sdram_writeq (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + off, 0);
	    }
#endif
	}
	else
	{
	    PRINTF5 (DBG_DEVICE, DBG_SDRAM, "elan4_sdram_flushcache: addr=%x len=%x (%x) => part cache (%x,%x)\n", 
		     addr, len, addr + len, baseoff, topoff);

#ifdef CONFIG_MPSAS
	    for (set = 0; set < E4_NumCacheSets; set++)
		elan4_sdram_zeroq_sdram (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + baseoff, topoff - baseoff);
#else
	    for (set = 0; set < E4_NumCacheSets; set++)
		for (off = (base & (E4_CacheSetSize-1)); off < (top & (E4_CacheSetSize-1)); off += E4_CACHELINE_SIZE)
		    elan4_sdram_writeq (dev, dev->dev_cacheflush_space + (set * E4_CacheSetSize) + off, 0);
#endif
	}
    }
    pioflush_sdram (dev);
    
    CLEAR_SYSCONTROL (dev, dev_direct_map_pci_writes, CONT_DIRECT_MAP_PCI_WRITES);
}

static char *
get_correctableErr_bitpos(uint SyndromeBits)
{
    switch (SyndromeBits)
    {
    case 0x00: return ("NoErr");
    case 0x31: return ("00"); 
    case 0x32: return ("01"); 
    case 0xc4: return ("02"); 
    case 0xc8: return ("03"); 
    case 0x26: return ("04"); 
    case 0x91: return ("05"); 
    case 0x89: return ("06"); 
    case 0x64: return ("07"); 
    case 0xc1: return ("08"); 
    case 0xf2: return ("09"); 
    case 0x34: return ("10"); 
    case 0xf8: return ("11"); 
    case 0xf1: return ("12"); 
    case 0xc2: return ("13"); 
    case 0xf4: return ("14"); 
    case 0x38: return ("15"); 
    case 0xd6: return ("16"); 
    case 0xa1: return ("17"); 
    case 0x79: return ("18"); 
    case 0xa4: return ("19"); 
    case 0xd9: return ("20"); 
    case 0xa2: return ("21"); 
    case 0x76: return ("22"); 
    case 0xa8: return ("23"); 
    case 0xe6: return ("24"); 
    case 0x51: return ("25"); 
    case 0xb9: return ("26"); 
    case 0x54: return ("27"); 
    case 0xe9: return ("28"); 
    case 0x52: return ("29"); 
    case 0xb6: return ("30"); 
    case 0x58: return ("31"); 
    case 0x13: return ("32"); 
    case 0x23: return ("33"); 
    case 0x4c: return ("34"); 
    case 0x8c: return ("35"); 
    case 0x62: return ("36"); 
    case 0x19: return ("37"); 
    case 0x98: return ("38"); 
    case 0x46: return ("39"); 
    case 0x1c: return ("40"); 
    case 0x2f: return ("41"); 
    case 0x43: return ("42"); 
    case 0x8f: return ("43"); 
    case 0x1f: return ("44"); 
    case 0x2c: return ("45"); 
    case 0x4f: return ("46"); 
    case 0x83: return ("47"); 
    case 0x6d: return ("48"); 
    case 0x1a: return ("49"); 
    case 0x97: return ("50"); 
    case 0x4a: return ("51"); 
    case 0x9d: return ("52"); 
    case 0x2a: return ("53"); 
    case 0x67: return ("54"); 
    case 0x8a: return ("55"); 
    case 0x6e: return ("56"); 
    case 0x15: return ("57"); 
    case 0x9b: return ("58"); 
    case 0x45: return ("59"); 
    case 0x9e: return ("60"); 
    case 0x25: return ("61"); 
    case 0x6b: return ("62"); 
    case 0x85: return ("63"); 
    case 0x01: return ("C0"); 
    case 0x02: return ("C1"); 
    case 0x04: return ("C2"); 
    case 0x08: return ("C3"); 
    case 0x10: return ("C4"); 
    case 0x20: return ("C5"); 
    case 0x40: return ("C6"); 
    case 0x80: return ("C7"); 

    case 0x07: case 0x0b: case 0x0d: case 0x0e: case 0x3d: case 0x3e: case 0x70: case 0x7c: // T  
    case 0xb0: case 0xbc: case 0xc7: case 0xcb: case 0xd0: case 0xd3: case 0xe0: case 0xe3: // T  
       return ("triple");

    case 0x0f: case 0x55: case 0x5a: case 0xa5: case 0xaa: case 0xf0: case 0xff: // Q  
       return ("quadruple");

    case 0x16: case 0x29: case 0x37: case 0x3b: case 0x49: case 0x57: case 0x5b: case 0x5d: case 0x5e: case 0x61: // M  
    case 0x68: case 0x73: case 0x75: case 0x7a: case 0x7f: case 0x86: case 0x92: case 0x94: case 0xa7: case 0xab: // M  
    case 0xad: case 0xae: case 0xb3: case 0xb5: case 0xba: case 0xbf: case 0xcd: case 0xce: case 0xd5: case 0xda: // M  
    case 0xdc: case 0xdf: case 0xe5: case 0xea: case 0xec: case 0xef: case 0xf7: case 0xfb: case 0xfd: case 0xfe: // M  
       return ("multiple");

    default:	// all other cases
       return ("double");
    }
}

char *
elan4_sdramerr2str (ELAN4_DEV *dev, E4_uint64 status, E4_uint64 ConfigReg, char *str)
{
    E4_uint64 StartupSyndrome    = dev->dev_sdram_initial_ecc_val;
    int       RisingDQSsyndrome  = ((ECC_RisingDQSSyndrome(status) == ECC_RisingDQSSyndrome(StartupSyndrome)) ?
				    0 : ECC_RisingDQSSyndrome(status));
    int	      FallingDQSsyndrome = ((ECC_FallingDQSSyndrome(status) == ECC_FallingDQSSyndrome(StartupSyndrome)) ?
				    0 : ECC_FallingDQSSyndrome(status));
    E4_uint64 Addr = ECC_Addr(status);
    int       Bank = (Addr >> 6) & 3;
    int       Cas  = ((Addr >> 3) & 7) | ((Addr >> (8 - 3)) & 0xf8) | ((Addr >> (25 - 8)) & 0x100) |
    		     ((Addr >> (27 - 9)) & 0x200) | ((Addr >> (29 - 10)) & 0xc00);
    int       Ras  = ((Addr >> 13) & 0xfff) | ((Addr >> (26 - 12)) & 0x1000) | ((Addr >> (28 - 13)) & 0x2000) |
    		     ((Addr >> (30 - 14)) & 0x4000);

    sprintf (str, "Addr=%07llx Bank=%x Ras=%x Cas=%x Falling DQS=%s Rising DQS=%s Syndrome=%x%s%s%s%s Type=%s SDRamDelay=%s,%0d",		/* 41 + 16 + 8 + 15 + 24 + 13 + 22 + 10 + 10 == 151 */
	     (long long)Addr, Bank, Ras, Cas,
	     get_correctableErr_bitpos(FallingDQSsyndrome),
	     get_correctableErr_bitpos(RisingDQSsyndrome),
	     (int)ECC_Syndrome(status),
	     ECC_UncorrectableErr(status)   ? " Uncorrectable" : "",
	     ECC_MultUncorrectErrs(status)  ? " Multiple-Uncorrectable" : "",
	     ECC_CorrectableErr(status)     ? " Correctable" : "",
	     ECC_MultCorrectErrs(status)    ? " Multiple-Correctable" : "",
	     (status & 0x0010000000000000ull)  ? "W" :
	     (status & 0x0020000000000000ull)  ? "R" :
	     (status & 0x0030000000000000ull)  ? "C" : "-",
	     (ConfigReg & SDRAM_FIXED_DELAY_ENABLE)  ? "F" : "A",
	     sdram_GreyToBinary(SDRAM_GET_DLL_DELAY(ConfigReg), SDRAM_FIXED_DLL_DELAY_BITS));

    return str;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
