/*
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: trap.c,v 1.23.2.1 2005/07/20 11:35:36 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/trap.c,v $*/

#include <qsnet/kernel.h>

#include <elan4/debug.h>
#include <elan4/device.h>

#include <elan4/trtype.h>
#include <elan4/commands.h>

char * const PermTypes[16] = 
{
    "Disabled",       "Unused",          "LocalDataRead", "LocalDataWrite",
    "LocalRead",      "LocalExecute",    "ReadOnly",      "LocalWrite",
    "LocalEventOnly", "LocalEventWrite", "RemoteEvent",   "RemoteAll",
    "RemoteReadOnly", "RemoteWriteOnly", "DataReadWrite", "NoFault",
};

char * const AccTypes[] =
{
    "LocalDataRead ", "LocalDataWrite", "RemoteRead    ", "RemoteWrite   ",
    "Execute       ", "LocalEvent    ", "Unused        ", "RemoteEvent   "
};
char * const DataTypes[] = {"Byte ", "HWord", "Word ", "DWord"};
char * const PhysTypes[] = {"Special Read", "Special Write", "Physical Read", "Physical Write"};
    
char * const EProcTrapNames[] = {
    "EventProcNoFault",
    "EventProcAddressAlignment",
    "EventProcMemoryFault",
    "EventProcCountWrapError",
};

char * const CProcTrapNames[] = {
    "CommandProcNoFault",
    "CommandProcInserterError",
    "CommandProcPermissionTrap",
    "CommandProcSendTransInvalid",
    "CommandProcSendTransExpected",
    "CommandProcDmaQueueOverflow",
    "CommandProcInterruptQueueOverflow",
    "CommandProcMemoryFault",
    "CommandProcRouteFetchFault",
    "CommandProcFailCountZero",
    "CommandProcAddressAlignment",
    "CommandProcWaitTrap",
    "CommandProcMultipleGuards",
    "CommandProcOpenOnGuardedChan",
    "CommandProcThreadQueueOverflow",
    "CommandProcBadData",
};

char *const CProcInsertError[] = {
    "No Error",
    "Overflowed",
    "Invalid Write Size",
    "Invalid Write Order",
};

char * const DProcTrapNames[] = {
    "DmaProcNoFault",
    "DmaProcRouteFetchFault",
    "DmaProcFailCountError",
    "DmaProcPacketAckError",
    "DmaProcRunQueueReadFault",
    "DmaProcQueueOverFlow",
    "DmaProcPrefetcherFault",	/* addy: Added new trap type for Prefetcher faults */
};

char *const IProcTrapNames[] = {
    "InputNoFault",
    "InputAddressAlignment",
    "InputMemoryFault",
    "InputInvalidTransType",
    "InputDmaQueueOverflow",
    "InputEventEngineTrapped",
    "InputCrcErrorAfterPAckOk",
    "InputEopErrorOnWaitForEop",
    "InputEopErrorTrap",
    "InputDiscardAfterAckOk",
};

char *const TProcTrapNames[] = {
    "HaltThread",
    "TrapForTooManyInstructions",
    "InstAccessException",
    "Unimplemented",
    "DataAccessException",
    "DataAlignmentError",
    "TrapForUsingBadData",
};

#define declare_spaces(space, str)		char space[64]; do { int i; for (i = 0; i < strlen(str); i++) spaces[i] = ' '; space[i] = '\0'; } while (0)
#define declare_prefix(space, spaces, str)	char space[64]; do { strcpy (space, spaces); strcat (space, str); } while (0)

void
elan4_display_farea (void *type, int mode, char *str, E4_FaultSave *farea)
{
    E4_uint32 FSR = FaultSaveFSR(farea->FSRAndFaultContext);

    declare_spaces(spaces, str);
    
    elan4_debugf (type, mode, "%s Fault occurred at %016llx for context %4x\n", str,
		  farea->FaultAddress, FaultSaveContext(farea->FSRAndFaultContext));
    
    if (FSR & AT_VirtualWriteAccBit)				/* Virtual write access */
	elan4_debugf (type, mode, "%s FSR=%x: Virtual Write. DWSize=0x%x EndP=0x%x Access=%s DT=%s\n",
		      spaces, FSR, FSR & AT_VirtualWriteSizeMask,
		      (FSR >> AT_VirtualWriteEndPtrShift) & AT_VirtualWriteEndPtrMask,
		      AccTypes[(FSR >> AT_PermBitsShift) & AT_PermBitsMask],
		      DataTypes[(FSR >> AT_BlkDataTyShift) & AT_BlkDataTyMask]);
    else if (FSR & AT_VirtualReadAccBit)			/* Virtual read access */
	elan4_debugf (type, mode, "%s FSR=%x: Virtual Read. DWSize=0x%x Access=%s DT=%s\n",
		      spaces, FSR, FSR & AT_VirtualReadSizeMask,
		      AccTypes[(FSR >> AT_PermBitsShift) & AT_PermBitsMask],
		      DataTypes[(FSR >> AT_BlkDataTyShift) & AT_BlkDataTyMask]);
    else
	elan4_debugf (type, mode, "%s FSR=%x: %s. Size=0x%x\n", spaces,
		      FSR, PhysTypes[(FSR >> AT_SelBitsShift) & AT_SelBitsMask],
		      FSR & AT_OtherSizeMask);
    elan4_debugf (type, mode, "%s FSR: %s %s%s %sWalking\n", spaces,
		  (FSR & AT_NonAlloc) ? "NonAlloc" : "Alloc",
		  (FSR & AT_DmaData) ? "Dma " : "",
		  (FSR & FSR_WalkForThread) ? "ThreadAcc" : "UnitsAcc",
		  (FSR & FSR_Walking) ? "" : "Not");
    PRINTF (type, mode, "%s FSR: %s%sHashTable=%s\n", spaces,
	    (FSR & FSR_NoTranslationsFound) ? "NoTranslationsFound " : "",
	    (FSR & FSR_WalkingProtectionFault) ? "WalkingProtectionFault " : "",
	    (FSR & FSR_HashTable1) ? "1" : "0");
    if (FSR & (FSR_RouteVProcErr | FSR_FaultForBadData))
	elan4_debugf (type, mode, "%s FSR: %s%s\n", spaces,
		      (FSR & FSR_RouteVProcErr) ? "RouteVProcErr " : "",
		      (FSR & FSR_FaultForBadData) ? "FaultForBadData " : "");
}

void
elan4_display_eproc_trap (void *type, int mode, char *str, ELAN4_EPROC_TRAP *trap)
{
    declare_spaces (spaces, str);

    elan4_debugf (type, mode, "%s Status=%016llx %s EventAddr=%016llx CountAndType=%016llx\n", str,
		  trap->tr_status, EProcTrapNames[EPROC_TrapType(trap->tr_status)],
		  trap->tr_eventaddr, trap->tr_event.ev_CountAndType);
    elan4_debugf (type, mode, "%s Param=%016llx.%016llx\n", spaces,
		  trap->tr_event.ev_Params[0], trap->tr_event.ev_Params[1]);

    elan4_display_farea (type, mode, strcat (spaces, EPROC_Port0Fault(trap->tr_status) ? " EPROC0" : " EPROC1"), &trap->tr_faultarea);
}

void
elan4_display_cproc_trap (void *type, int mode, char *str, ELAN4_CPROC_TRAP *trap)
{
    declare_spaces(spaces, str);

    elan4_debugf (type, mode, "%s Status=%llx %s Command=%llx\n", str, trap->tr_status, 
		  CProcTrapNames[CPROC_TrapType(trap->tr_status)], trap->tr_command);
    elan4_debugf (type, mode, "%s Desc=%016llx %016llx %016llx %016llx\n", str,
		  trap->tr_qdesc.CQ_QueuePtrs, trap->tr_qdesc.CQ_HoldingValue,
		  trap->tr_qdesc.CQ_AckBuffers, trap->tr_qdesc.CQ_Control);

    switch (CPROC_TrapType (trap->tr_status))
    {
    case CommandProcInserterError:
	elan4_debugf (type, mode, "%s   %s\n", str, CProcInsertError[CQ_RevB_ErrorType(trap->tr_qdesc.CQ_QueuePtrs)]);
	break;

    case CommandProcWaitTrap:
	elan4_display_eproc_trap (type, mode, spaces, &trap->tr_eventtrap);
	break;

    default:
	elan4_display_farea (type, mode, spaces, &trap->tr_faultarea);
	break;
    }
}

void
elan4_display_dproc_trap (void *type, int mode, char *str, ELAN4_DPROC_TRAP *trap)
{
    declare_spaces (spaces, str);

    elan4_debugf (type, mode, "%s status %llx - %s\n", str,
		  trap->tr_status, DProcTrapNames[DPROC_TrapType(trap->tr_status)]);

    elan4_debugf (type, mode, "%s DESC %016llx %016llx %016llx %016llx\n", spaces, trap->tr_desc.dma_typeSize, 
		  trap->tr_desc.dma_cookie, trap->tr_desc.dma_vproc, trap->tr_desc.dma_srcAddr);
    elan4_debugf (type, mode, "%s      %016llx %016llx %016llx\n", spaces, trap->tr_desc.dma_dstAddr, 
		  trap->tr_desc.dma_srcEvent, trap->tr_desc.dma_dstEvent);

    if (DPROC_PrefetcherFault (trap->tr_status))
	elan4_display_farea (type, mode, spaces, &trap->tr_prefetchFault);
}

void
elan4_display_tproc_trap (void *type, int mode, char *str, ELAN4_TPROC_TRAP *trap)
{
    register int i;
    declare_spaces (spaces, str);

    elan4_debugf (type, mode, "%s PC=%016llx nPC=%016llx State=%016llx Status=%016llx -%s%s%s%s\n", str,
		  trap->tr_pc, trap->tr_npc, trap->tr_state, trap->tr_status, 
		  (trap->tr_state & TS_TrapForTooManyInstructions) ? " TrapForTooManyInstructions" : "",
		  (trap->tr_state & TS_Unimplemented)              ? " Unimplemented"              : "",
		  (trap->tr_state & TS_DataAlignmentError)         ? " DataAlignmentError"         : "",
		  (trap->tr_state & TS_InstAccessException)        ? " InstAccessException"        : "",
		  (trap->tr_state & TS_DataAccessException)        ? " DataAlignmentError"         : "");
    
    for (i = 0; i < 64; i += 4)
	elan4_debugf (type, mode, "%s r%d - %016llx %016llx %016llx %016llx\n", spaces, i,
		      trap->tr_regs[i], trap->tr_regs[i+1], trap->tr_regs[i+2], trap->tr_regs[i+3]);
    
    if (trap->tr_state & TS_InstAccessException)
    {
	declare_prefix (prefix, spaces, "Inst");

	elan4_display_farea (type, mode, prefix, &trap->tr_instFault);
    }

    if (trap->tr_state & TS_DataAccessException)
    {
	declare_prefix (prefix, spaces, "Data");
	elan4_display_farea (type, mode, prefix, &trap->tr_dataFault);
    }
}

void
elan4_display_iproc_trap (void *type, int mode, char *str, ELAN4_IPROC_TRAP *trap)
{
    register int i;
    declare_spaces (spaces, str);

    for (i = 0; i < trap->tr_numTransactions; i++)
    {
	E4_IprocTrapHeader *hdrp    = &trap->tr_transactions[i];
	E4_uint64           status  = hdrp->IProcStatusCntxAndTrType;
	E4_Addr             addr    = hdrp->TrAddr;
	char               *typeString;
	char		    buffer[256];
	char		   *ptr = buffer;
	
	if (IPROC_EOPTrap(status))
	{
	    switch (IPROC_EOPType(status))
	    {
	    case EOP_GOOD:	  typeString = "EopGood";   break;
	    case EOP_BADACK:	  typeString = "EopBadAck"; break;
	    case EOP_ERROR_RESET: typeString = "EopReset";  break;
	    default:		  typeString = "EopBad";    break;
	    }
	    
	    ptr += sprintf (ptr, "%15s Cntx=%-6d", typeString, IPROC_NetworkContext(status));
	}
	else
	{
	    if (IPROC_BadLength(status))
		typeString = "BadLength";
	    else if (IPROC_TransCRCStatus(status) == CRC_STATUS_DISCARD)
		typeString = "DiscardCrc";
	    else if (IPROC_TransCRCStatus(status) == CRC_STATUS_ERROR)
		typeString = "ErrorCrc Remote Network error";
	    else if (IPROC_TransCRCStatus(status) == CRC_STATUS_BAD)
		typeString = "BadCrc Cable error into this node.";
	    else
	    {
		if ((IPROC_TransactionType(status) & TR_BLOCK_OPCODE_MASK) == TR_WRITEBLOCK)
		    typeString = "WriteBlock";
		else
		{
		    switch (IPROC_TransactionType(status) & TR_OPCODE_MASK)
		    {
		    case TR_SETEVENT_IDENTIFY & TR_OPCODE_MASK:	typeString = "SetEvent";	break;
		    case TR_REMOTEDMA & TR_OPCODE_MASK:		typeString = "RemoteDma";	break;
		    case TR_SENDDISCARD & TR_OPCODE_MASK:	typeString = "SendDiscard";	break;
		    case TR_GTE & TR_OPCODE_MASK:		typeString = "GTE";		break;
		    case TR_LT & TR_OPCODE_MASK:		typeString = "LT";		break;
		    case TR_EQ & TR_OPCODE_MASK:		typeString = "EQ";		break;
		    case TR_NEQ & TR_OPCODE_MASK:		typeString = "NEQ";		break;
		    case TR_IDENTIFY & TR_OPCODE_MASK:		typeString = "Idenfity";	break;
		    case TR_ADDWORD & TR_OPCODE_MASK:		typeString = "AddWord";		break;
		    case TR_INPUT_Q_COMMIT & TR_OPCODE_MASK:	typeString = "InputQCommit";	break;
		    case TR_TESTANDWRITE & TR_OPCODE_MASK:	typeString = "TestAndWrite";	break;
		    case TR_INPUT_Q_GETINDEX & TR_OPCODE_MASK:	typeString = "InputQGetIndex";	break;
		    case TR_TRACEROUTE_TRANS & TR_OPCODE_MASK:	typeString = "TraceRoute";	break;
		    default:					typeString = "Unknown";		break;
		    }
		}
	    }

	    ptr += sprintf (ptr, "%15s Cntx=%-6d Addr=%016llx", typeString, IPROC_NetworkContext(status), (unsigned long long) addr);
	}
	
	
	if (IPROC_TrapValue(status) != InputNoFault)
	{
	    ptr += sprintf (ptr, " TrType=%2d ChanTrapped=%x GoodAck=%x BadAck=%x InputterChan=%d", IPROC_TrapValue(status),
			    IPROC_ChannelTrapped(status), IPROC_GoodAckSent(status), IPROC_BadAckSent(status),
			    IPROC_InputterChan(status));
	    if (IPROC_EOPTrap(status))
		ptr += sprintf (ptr, " EOPType=%d", IPROC_EOPType(status));
	    else
		ptr += sprintf (ptr, " %s%s%s%s", 
				IPROC_FirstTrans(status) ? " FirstTrans" : "",
				IPROC_LastTrans(status) ? " LastTrans" : "",
				(IPROC_TransactionType(status) & TR_WAIT_FOR_EOP) ? " WaitForEop" : "",
				(IPROC_GoodAckSent(status) &  (1 << IPROC_Channel(status))) ? " AckSent" : "");
	}
	
	elan4_debugf (type, mode, "%s %s\n", str, buffer);

	str = spaces;
    }

    elan4_display_farea (type, mode, spaces, &trap->tr_faultarea);
}

#define elan4_sdram_copy_faultarea(dev, unit, farea) \
    elan4_sdram_copyq_from_sdram ((dev), (dev)->dev_faultarea + (unit) * sizeof (E4_FaultSave), (E4_uint64 *) farea, sizeof (E4_FaultSave));

void
elan4_extract_eproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_EPROC_TRAP *trap, int iswaitevent)
{
    /* only one of the memory ports can fault at a time */
    ASSERT (EPROC_TrapType(status) != EventProcMemoryFault || (EPROC_Port0Fault(status) ^ EPROC_Port1Fault(status)) == 1);

    trap->tr_status = status;
    
    if (EPROC_Port0Fault(status))
	elan4_sdram_copy_faultarea (dev, CUN_EventProc0, &trap->tr_faultarea);
    if (EPROC_Port1Fault(status))
	elan4_sdram_copy_faultarea (dev, CUN_EventProc1, &trap->tr_faultarea);

    if (iswaitevent)
    {
	/*
	 * for waitevents the Event address is always taken from the command processor
	 * 
	 * if we trapped during the copy then we take the "Event" from the event processor
	 * since we need to complete the copy.  Otherwise we'll be reissuing the original
	 * command again
	 */
	E4_uint32 fsr = FaultSaveFSR(trap->tr_faultarea.FSRAndFaultContext);

	trap->tr_eventaddr = read_reg64 (dev, CommandHold) ^ WAIT_EVENT_CMD;

	if (EPROC_TrapType(trap->tr_status) == EventProcMemoryFault && 
	    (AT_Perm(fsr) == AT_PermLocalDataRead || AT_Perm(fsr) == AT_PermLocalDataWrite))
	{
	    trap->tr_event.ev_CountAndType = read_reg64 (dev, EventCountAndType);
	    trap->tr_event.ev_Params[0]    = read_reg64 (dev, EventParameters[0]);
	    trap->tr_event.ev_Params[1]    = read_reg64 (dev, EventParameters[1]);
	}
	else
	{
	    trap->tr_event.ev_Params[0]    = read_reg64 (dev, CommandCopy[5]);
	    trap->tr_event.ev_CountAndType = read_reg64 (dev, CommandCopy[4]);
	    trap->tr_event.ev_Params[1]    = read_reg64 (dev, CommandCopy[6]);

	}
    }
    else
    {
	trap->tr_eventaddr             = read_reg64 (dev, EventAddress);
	trap->tr_event.ev_CountAndType = read_reg64 (dev, EventCountAndType);
	trap->tr_event.ev_Params[0]    = read_reg64 (dev, EventParameters[0]);
	trap->tr_event.ev_Params[1]    = read_reg64 (dev, EventParameters[1]);
    }

    BumpDevStat (dev, s_eproc_trap_types[EPROC_TrapType(status)]);
}

int 
cproc_open_extract_vp (ELAN4_DEV *dev, ELAN4_CQ *cq, int chan)
{
	/* cq = ucq->ucq_cq */
	if ((cq->cq_perm & CQ_STENEnableBit) != 0)
	{
            sdramaddr_t   cqdesc       = dev->dev_cqaddr + (elan4_cq2num(cq) * sizeof (E4_CommandQueueDesc));
	    E4_uint64     queuePtrs    = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_QueuePtrs));
	    sdramaddr_t   insertPtr    = (queuePtrs & CQ_PtrMask);
	    sdramaddr_t   commandPtr   = CQ_CompletedPtr (queuePtrs);
	    unsigned int  cqSize       = CQ_Size ((queuePtrs >> CQ_SizeShift) & CQ_SizeMask);

	    if (dev->dev_devinfo.dev_revision_id != PCI_REVISION_ID_ELAN4_REVA && (queuePtrs & CQ_RevB_ReorderingQueue))
	    {
		E4_uint32 oooMask = elan4_sdram_readl (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_HoldingValue));

		for (; (oooMask & 1) != 0; oooMask >>= 1)
	            insertPtr = (insertPtr & ~(cqSize-1)) | ((insertPtr + sizeof (E4_uint64)) & (cqSize-1));
	    }

	    while (commandPtr != insertPtr)
	    {
		E4_uint64    command = elan4_sdram_readq (dev, commandPtr);
		unsigned int cmdSize;

                switch (__categorise_command (command, &cmdSize))
		{
		case 0:
		    (void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		    break;

		case 1: /* open */
		    if (((chan << 4) == (command & (1<<4))))
			/* Matches supplied channel */
			return (command >> 32);
		    else
			(void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		    break;

		case 2:
		    (void) __whole_command (&commandPtr, insertPtr, cqSize, cmdSize);
		case 3:
		    printk ("cproc_open_extract_vp: invalid command %llx\n", (long long)command);
		    return -1;
		}
	    } /* while */
	}

	return -1;
}

void
elan4_extract_cproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_CPROC_TRAP *trap, unsigned cqnum)
{
    /* extract the state from the device */
    elan4_sdram_copy_faultarea (dev, CUN_CommandProc, &trap->tr_faultarea);

    trap->tr_status  = status;
    trap->tr_command = read_reg64 (dev, CommandHold);
    
    elan4_sdram_copyq_from_sdram (dev, dev->dev_cqaddr + (cqnum * sizeof (E4_CommandQueueDesc)), &trap->tr_qdesc, sizeof (E4_CommandQueueDesc));

    if (CPROC_TrapType (status) == CommandProcWaitTrap)
	elan4_extract_eproc_trap (dev, read_reg64 (dev, EProcStatus), &trap->tr_eventtrap, 1);

    BumpDevStat (dev, s_cproc_trap_types[CPROC_TrapType(status)]);

    if (PackValue(trap->tr_qdesc.CQ_AckBuffers, 0) == PackTimeout || PackValue(trap->tr_qdesc.CQ_AckBuffers, 1) == PackTimeout)
	BumpDevStat (dev, s_cproc_timeout);
}

void
elan4_extract_dproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_DPROC_TRAP *trap, unsigned unit)
{
    trap->tr_status = status;
    
    if (unit == 0)
    {
	trap->tr_desc.dma_typeSize   = read_reg64 (dev, Dma0Desc.dma_typeSize);
	trap->tr_desc.dma_cookie     = read_reg64 (dev, Dma0Desc.dma_cookie);
	trap->tr_desc.dma_vproc      = read_reg64 (dev, Dma0Desc.dma_vproc);
	trap->tr_desc.dma_srcAddr    = read_reg64 (dev, Dma0Desc.dma_srcAddr);
	trap->tr_desc.dma_dstAddr    = read_reg64 (dev, Dma0Desc.dma_dstAddr);
	trap->tr_desc.dma_srcEvent   = read_reg64 (dev, Dma0Desc.dma_srcEvent);
	trap->tr_desc.dma_dstEvent   = read_reg64 (dev, Dma0Desc.dma_dstEvent);
	
	elan4_sdram_copy_faultarea (dev, CUN_DProcPA0, &trap->tr_packAssemFault);
    }
    else
    {
	trap->tr_desc.dma_typeSize   = read_reg64 (dev, Dma1Desc.dma_typeSize);
	trap->tr_desc.dma_cookie     = read_reg64 (dev, Dma1Desc.dma_cookie);
	trap->tr_desc.dma_vproc      = read_reg64 (dev, Dma1Desc.dma_vproc);
	trap->tr_desc.dma_srcAddr    = read_reg64 (dev, Dma1Desc.dma_srcAddr);
	trap->tr_desc.dma_dstAddr    = read_reg64 (dev, Dma1Desc.dma_dstAddr);
	trap->tr_desc.dma_srcEvent   = read_reg64 (dev, Dma1Desc.dma_srcEvent);
	trap->tr_desc.dma_dstEvent   = read_reg64 (dev, Dma1Desc.dma_dstEvent);
	
	elan4_sdram_copy_faultarea (dev, CUN_DProcPA1, &trap->tr_packAssemFault);
    }
    
    if (DPROC_PrefetcherFault (trap->tr_status))
    {
	elan4_sdram_copy_faultarea (dev, (CUN_DProcData0 | DPROC_FaultUnitNo(trap->tr_status)), &trap->tr_prefetchFault);
	/* addy: Added new trap type for Prefetcher faults */
	BumpDevStat (dev, s_dproc_trap_types[6]);
    }
    else if (DPROC_PacketTimeout (trap->tr_status))
	BumpDevStat (dev, s_dproc_timeout);
    else
	BumpDevStat (dev, s_dproc_trap_types[DPROC_TrapType(status)]);
}    

void
elan4_extract_tproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_TPROC_TRAP *trap)
{
    int i;

    trap->tr_status = status;
    trap->tr_state  = read_reg64 (dev, Thread_Trap_State);
    trap->tr_pc     = read_reg64 (dev, PC_W);
    trap->tr_npc    = read_reg64 (dev, nPC_W);
    trap->tr_dirty  = read_reg64 (dev, DirtyBits);
    trap->tr_bad    = read_reg64 (dev, BadBits);

#ifdef CONFIG_MPSAS
    if (sas_copyfrom_dev (dev->dev_osdep.pdev, ELAN4_BAR_REGISTERS, 
			  ((dev->dev_devinfo.dev_revision_id == PCI_REVISION_ID_ELAN4_REVA) ? ELAN4_REVA_REG_OFFSET : ELAN4_REVB_REG_OFFSET) +
			  offsetof (E4_Registers, Regs.TProcRegs), (unsigned long) &trap->tr_regs, 64*sizeof (E4_uint64)) < 0)
    {
	for (i = 0; i < 64; i++)
	    if (trap->tr_dirty & ((E4_uint64) 1 << i))
		trap->tr_regs[i] = read_reg64 (dev, TProcRegs[i]);
    }

    for (i = 0; i < 64; i++)
	if (! (trap->tr_dirty & ((E4_uint64) 1 << i)))
	    trap->tr_regs[i] = 0xdeadbabedeadbabeULL;
#else
    for (i = 0; i < 64; i++)
    {
	if (trap->tr_dirty & ((E4_uint64) 1 << i))
	    trap->tr_regs[i] = read_reg64 (dev, TProcRegs[i]);
	else
	    trap->tr_regs[i] = 0xdeadbabedeadbabeULL;
    }
#endif
    
    if (trap->tr_state & TS_DataAccessException)
	elan4_sdram_copy_faultarea (dev, CUN_TProcData0 | TS_DataPortNo (trap->tr_state), &trap->tr_dataFault);

    if (trap->tr_state & TS_InstAccessException)
	elan4_sdram_copy_faultarea (dev, CUN_TProcInst, &trap->tr_instFault);

    for (i = 0; i < 7; i++)
	if (trap->tr_state & (1 << i))
	    BumpDevStat (dev, s_tproc_trap_types[i]);
}

void
elan4_extract_iproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_IPROC_TRAP *trap, unsigned unit)
{
    sdramaddr_t hdroff  = dev->dev_inputtraparea + offsetof (E4_IprocTrapState, TrHeader[0][unit]);
    sdramaddr_t dataoff = dev->dev_inputtraparea + offsetof (E4_IprocTrapState, TrData[0][unit]);
    register int i, j;
    int		  CurrUnitNo    = (unit >= 2) ? CUN_IProcHighPri : CUN_IProcLowPri;
    sdramaddr_t CurrFaultArea = dev->dev_faultarea + (CurrUnitNo * sizeof (E4_FaultSave));

    /* Finally copy the fault area */
    elan4_sdram_copy_faultarea (dev, CurrUnitNo, &trap->tr_faultarea);

    /*
     * Clear out the fault save area after reading to allow a fault on the write of the back pointer of
     * an InputQCommit to be obsurved if a simultaneous event proc trap occurs.
     */
    elan4_sdram_writeq (dev, CurrFaultArea + offsetof(E4_FaultSave, FSRAndFaultContext), 0x0ULL);
    elan4_sdram_writeq (dev, CurrFaultArea + offsetof(E4_FaultSave, FaultAddress), 0x0ULL);

    /* copy the transaction headers */
    trap->tr_transactions[0].IProcStatusCntxAndTrType = status;
    trap->tr_transactions[0].TrAddr                   = elan4_sdram_readq (dev, hdroff + offsetof (E4_IprocTrapHeader, TrAddr));
    
    for (i = 0; !IPROC_EOPTrap(trap->tr_transactions[i].IProcStatusCntxAndTrType);)
    {
	if (IPROC_BadLength (trap->tr_transactions[i].IProcStatusCntxAndTrType))
	    BumpDevStat (dev, s_bad_length);
	else if (IPROC_TransCRCStatus (trap->tr_transactions[i].IProcStatusCntxAndTrType) == CRC_STATUS_BAD)
	    BumpDevStat (dev, s_crc_bad);
	else if (IPROC_TransCRCStatus (trap->tr_transactions[i].IProcStatusCntxAndTrType) == CRC_STATUS_ERROR)
	    BumpDevStat (dev, s_crc_error);

	BumpDevStat (dev, s_iproc_trap_types[IPROC_TrapValue (trap->tr_transactions[i].IProcStatusCntxAndTrType)]);

	hdroff += NO_OF_INPUT_CHANNELS*sizeof (E4_IprocTrapHeader);

	if (++i == MAX_TRAPPED_TRANS)
	    break;

	elan4_sdram_copyq_from_sdram (dev, hdroff, &trap->tr_transactions[i], sizeof (E4_IprocTrapHeader));
    }
    
    if (IPROC_EOPType (trap->tr_transactions[i].IProcStatusCntxAndTrType) == EOP_ERROR_RESET)
	BumpDevStat (dev, s_eop_reset);

    /* Remember the number of transactions we've copied */
    trap->tr_numTransactions = i + 1;
    
    /* Copy all the data blocks in one go */
    for (i = 0; i < MIN (trap->tr_numTransactions, MAX_TRAPPED_TRANS); i++, dataoff += NO_OF_INPUT_CHANNELS*sizeof (E4_IprocTrapData))
    {
	if (IPROC_BadLength(status) || IPROC_TransCRCStatus (status) != CRC_STATUS_GOOD)
	    elan4_sdram_copyq_from_sdram (dev, dataoff, trap->tr_dataBuffers[i].Data, TRANS_DATA_DWORDS*sizeof(E4_uint64));
	else
	{
	    int trtype  = IPROC_TransactionType(trap->tr_transactions[i].IProcStatusCntxAndTrType);
	    int ndwords = (trtype & TR_SIZE_MASK) >> TR_SIZE_SHIFT;

	    elan4_sdram_copyq_from_sdram (dev, dataoff, trap->tr_dataBuffers[i].Data, ndwords*sizeof(E4_uint64));

	    for (j = ndwords; j < TRANS_DATA_DWORDS; j++)
		trap->tr_dataBuffers[i].Data[j] = 0xbeec0f212345678ull;
	}
    }
    
}

void
elan4_inspect_iproc_trap (ELAN4_IPROC_TRAP *trap)
{
    int i;

    trap->tr_flags	     = 0;
    trap->tr_trappedTrans    = TR_TRANS_INVALID;
    trap->tr_waitForEopTrans = TR_TRANS_INVALID;
    trap->tr_identifyTrans   = TR_TRANS_INVALID;

    if (trap->tr_numTransactions > MAX_TRAPPED_TRANS)
	trap->tr_flags = TR_FLAG_TOOMANY_TRANS;

    /*
     * Now scan all the transactions received 
     */
    for (i = 0; i < MIN(trap->tr_numTransactions, MAX_TRAPPED_TRANS) ; i++)
    {
	E4_IprocTrapHeader *hdrp   = &trap->tr_transactions[i];
	E4_uint64           status = hdrp->IProcStatusCntxAndTrType;

	if (trap->tr_identifyTrans == TR_TRANS_INVALID)
	{
	    switch (IPROC_TransactionType (status) & (TR_OPCODE_MASK | TR_SIZE_MASK))
	    {
	    case TR_IDENTIFY          & (TR_OPCODE_MASK | TR_SIZE_MASK):
	    case TR_REMOTEDMA         & (TR_OPCODE_MASK | TR_SIZE_MASK):
	    case TR_SETEVENT_IDENTIFY & (TR_OPCODE_MASK | TR_SIZE_MASK):
	    case TR_INPUT_Q_COMMIT    & (TR_OPCODE_MASK | TR_SIZE_MASK):
	    case TR_ADDWORD           & (TR_OPCODE_MASK | TR_SIZE_MASK):
	    case TR_TESTANDWRITE      & (TR_OPCODE_MASK | TR_SIZE_MASK):
		trap->tr_identifyTrans = i;
		break;
	    }
	}

	if (IPROC_TrapValue(status) == InputNoFault)		/* We're looking at transactions stored before the trap */
	    continue;						/* these should only be identifies */
	
 	if (trap->tr_trappedTrans == TR_TRANS_INVALID)		/* Remember the transaction which caused the */
	    trap->tr_trappedTrans = i;				/* trap */

	if (IPROC_GoodAckSent (status) & (1 << IPROC_InputterChan (status)))
	    trap->tr_flags |= TR_FLAG_ACK_SENT;
	    
	if (IPROC_EOPTrap(status))				/* Check for EOP */
	{
	    ASSERT (i == trap->tr_numTransactions - 1);

	    switch (IPROC_EOPType(status))
	    {
	    case EOP_GOOD:
		/* if we get an EOP_GOOD then the outputer should have received a PAckOk. */  
		/* unless it was a flood, in which case someone must have sent an ack */
		/* but not necessarily us */
		break;

	    case EOP_BADACK:
		/* if we get an EOP_BADACK then the outputer did not receive a PAckOk even if
		 * we sent a PAckOk. WFlag this to ignore the AckSent. */
		trap->tr_flags |= TR_FLAG_EOP_BAD;
		break;

	    case EOP_ERROR_RESET:
		/* if we get an EOP_ERROR_RESET then the outputer may or may not have got a PAckOk. */
		trap->tr_flags |= TR_FLAG_EOP_ERROR;
		break;

	    default:
		printk ("elan4_inspect_iproc_trap: unknown eop type %d", IPROC_EOPType(status));
		BUG();
		/* NOTREACHED */
	    }
	    continue;
	}
	else
	{
	    if (IPROC_BadLength(status) || (IPROC_TransCRCStatus (status) == CRC_STATUS_ERROR ||
					    IPROC_TransCRCStatus (status) == CRC_STATUS_BAD))
	    {
		{
		    register int j;
		    if (IPROC_BadLength(status))
			PRINTF2 (DBG_DEVICE, DBG_INTR, "LinkError: Trapped on bad length data. status=%016llx Address=%016llx\n",
				 status, hdrp->TrAddr);
		    else
			PRINTF2 (DBG_DEVICE, DBG_INTR, "LinkError: Trapped with bad CRC. status=%016llx Address=%016llx\n",
				 status, hdrp->TrAddr);
		    for (j = 0; j < TRANS_DATA_DWORDS; j++)
			PRINTF2 (DBG_DEVICE, DBG_INTR, "LinkError: DataBuffers[%d] : %016llx\n", j, trap->tr_dataBuffers[i].Data[j]);
		}

		trap->tr_flags |= TR_FLAG_BAD_TRANS;
		continue;
	    }
	    
	    if (IPROC_TransCRCStatus (status) == CRC_STATUS_DISCARD)
		continue;

	    if ((((IPROC_TransactionType(status) & TR_BLOCK_OPCODE_MASK) == TR_WRITEBLOCK) ||
	         (IPROC_TransactionType(status) == TR_TRACEROUTE_TRANS)) &&
		(trap->tr_flags & TR_FLAG_ACK_SENT) && trap->tr_identifyTrans == TR_TRANS_INVALID)
	    {
		/* 
		 * Writeblock after the ack is sent without an identify transaction - this is 
		 * considered to be a DMA packet and requires the next packet to be nacked - since 
		 * the DMA processor will send this in a deterministic time and there's an upper 
		 * limit on the network latency (the output timeout) we just need to hold the context 
		 * filter up for a while.
		 */
		trap->tr_flags |= TR_FLAG_DMA_PACKET;
	    }
	    
	    if (IPROC_LastTrans(status) && (IPROC_TransactionType(status) & TR_WAIT_FOR_EOP))
	    {
		/*
		 * WaitForEop transactions - if we have to do network error fixup
		 * then we may need to execute/ignore this transaction dependant
		 * on whether the source will be resending it.
		 */
		trap->tr_waitForEopTrans = i;
	    }

	    /*
	     * This is a special case caused by a minor input processor bug.
	     * If simultaneous InputMemoryFault and InputEventEngineTrapped occur then the chip will probably return
	     * InputEventEngineTrapped even though the write of the back pointer has not occured and must be done by
	     * the trap handler.
	     * In this case the fault address will equal q->q_bptr. If there has been only EventEngineTrap then the
	     * the fault address should be zero as the trap handler now always zeros this after every input trap.
	     */
	    if ((IPROC_TransactionType (status) & TR_OPCODE_MASK) == (TR_INPUT_Q_COMMIT & TR_OPCODE_MASK) &&
		trap->tr_faultarea.FaultAddress == hdrp->TrAddr + offsetof(E4_InputQueue, q_bptr) &&
		IPROC_TrapValue(status) == InputEventEngineTrapped)
	    {
		hdrp->IProcStatusCntxAndTrType = (status & 0xFFFFFFF0FFFFFFFFull) | ((E4_uint64) InputMemoryFault << 32);
	    }
	}

	PRINTF (DBG_DEVICE, DBG_INTR, "inspect[%d] status=%llx TrapValue=%d -> flags %x\n", i, status, IPROC_TrapValue(status), trap->tr_flags);
    }
}

E4_uint64
elan4_trapped_open_command (ELAN4_DEV *dev, ELAN4_CQ *cq)
{
    sdramaddr_t cqdesc     = dev->dev_cqaddr + elan4_cq2num(cq) * sizeof (E4_CommandQueueDesc);
    E4_uint64   cqcontrol  = elan4_sdram_readq (dev, cqdesc + offsetof (E4_CommandQueueDesc, CQ_Control));
    E4_uint32   extractOff = CQ_ExtractPtr (cqcontrol) & (CQ_Size(cq->cq_size)-1);
    
    if (extractOff == 0)
	extractOff = CQ_Size(cq->cq_size) - sizeof (E4_uint64);
    else
	extractOff -= sizeof (E4_uint64);

    return (elan4_sdram_readq (dev, cq->cq_space + extractOff));
}

EXPORT_SYMBOL(elan4_extract_eproc_trap);
EXPORT_SYMBOL(elan4_display_eproc_trap);
EXPORT_SYMBOL(elan4_extract_cproc_trap);
EXPORT_SYMBOL(elan4_display_cproc_trap);
EXPORT_SYMBOL(elan4_extract_dproc_trap);
EXPORT_SYMBOL(elan4_display_dproc_trap);
EXPORT_SYMBOL(elan4_extract_tproc_trap);
EXPORT_SYMBOL(elan4_display_tproc_trap);
EXPORT_SYMBOL(elan4_extract_iproc_trap);
EXPORT_SYMBOL(elan4_inspect_iproc_trap);
EXPORT_SYMBOL(elan4_display_iproc_trap);


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
