/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: user_Linux.c,v 1.35.2.1 2005/06/09 14:46:55 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/user_Linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kpte.h>

#include <linux/pci.h>
#include <linux/pagemap.h>

#ifdef CONFIG_HUGETLB_PAGE
#include <linux/hugetlb.h>
#endif

#include <elan4/debug.h>
#include <elan4/device.h>
#include <elan4/user.h>

extern struct vm_operations_struct mem_vm_ops;
extern struct vm_operations_struct user_vm_ops;
extern int mem_pteload (struct vm_area_struct *vma, unsigned long pgoff, ELAN4_CTXT *ctxt, E4_Addr eaddr, int perm);
extern int user_pteload (struct vm_area_struct *vma, unsigned long pgoff, ELAN4_CTXT *ctxt, E4_Addr eaddr, int perm);

static inline int
user_load_page (USER_CTXT *uctx, struct vm_area_struct *vma, unsigned long maddr, E4_Addr eaddr, int perm, int writeable)
{
    ELAN4_DEV   *dev  = uctx->uctx_ctxt.ctxt_dev;
    struct page *page = NULL;
    int i, res = 0;
    
    if (get_user_pages (current, current->mm, maddr, 1, writeable, 0, &page, NULL) == 1)
    {
	/* NOTE - the page can't be paged out since we've pinned it down.
	 *        it also can't be munmap'd since we hold the mmap_sem */
	
	PRINTF (uctx, DBG_FAULT, "user_load_page: %lx %s page %p\n", maddr, writeable ? "writeable" : "readonly",  page);
	
	for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[0]))
	    if ((res = elan4mmu_pteload_page (&uctx->uctx_ctxt, 0, eaddr + i, page, perm)) < 0)
		break;
	
	page_cache_release (page);
    }
    else
    {
	if (vma && vma->vm_ops == &mem_vm_ops)
	    res = mem_pteload (vma, maddr, &uctx->uctx_ctxt, eaddr, perm);
	else if (vma && vma->vm_ops == &user_vm_ops)
	    res = user_pteload (vma, maddr, &uctx->uctx_ctxt, eaddr, perm);
	else
	    res = -EINVAL;
    }

    return res;
}

int
user_load_range (USER_CTXT *uctx, E4_Addr eaddr, unsigned long nbytes, E4_uint32 fsr)
{
    struct mm_struct      *mm        = current->mm;
    int                    writeable = (AT_Perm(fsr) == AT_PermLocalDataWrite ||
					AT_Perm(fsr) == AT_PermRemoteWrite    ||
					AT_Perm(fsr) == AT_PermLocalEvent     ||
					AT_Perm(fsr) == AT_PermRemoteEvent);
    struct vm_area_struct *vma;
    int                    perm;
    unsigned long          len;
    unsigned long          maddr;
    int		           res = 0;

    kmutex_lock (&uctx->uctx_rgnmutex);

    while (nbytes > 0)
    {
	USER_RGN *rgn = user_rgnat_elan (uctx, eaddr);

	if (rgn == NULL || ELAN4_INCOMPAT_ACCESS (rgn->rgn_perm, AT_Perm (fsr)))
	{
	    PRINTF (uctx, DBG_FAULT, "user_load_range: eaddr=%llx -> %s\n", eaddr, rgn == NULL ? "no mapping" : "no permission");

	    kmutex_unlock (&uctx->uctx_rgnmutex);
	    return (rgn == NULL ? -EFAULT : -EPERM);
	}

	if (writeable)
	    perm = rgn->rgn_perm;
	else if (AT_Perm(fsr) == AT_PermExecute)
	    perm = PERM_LocRead | (rgn->rgn_perm & ~PERM_Mask);
	else
	    perm = ELAN4_PERM_READONLY (rgn->rgn_perm & PERM_Mask) | (rgn->rgn_perm & ~PERM_Mask);

	PRINTF (uctx, DBG_FAULT, "user_load_range: rgn=%p [%llx.%lx.%x]\n", rgn, rgn->rgn_ebase, rgn->rgn_mbase, rgn->rgn_len);

	len = ((rgn->rgn_ebase + rgn->rgn_len) - eaddr);
	if (len > nbytes)
	    len = nbytes;
	nbytes -= len;
	
	maddr = rgn->rgn_mbase + (eaddr - rgn->rgn_ebase);

	PRINTF (uctx, DBG_FAULT, "user_load_range: eaddr=%llx->%llx -> %lx->%lx len=%x perm=%x\n", eaddr, 
		eaddr + len, maddr, maddr + len, len, perm);

	down_read (&mm->mmap_sem);
	while (len > 0)
	{
#if defined(conditional_schedule)
	    conditional_schedule();
#endif
	    if ((vma = find_vma_intersection (mm, maddr, maddr + PAGE_SIZE)) == NULL ||
		(writeable && !(vma->vm_flags & VM_WRITE)))
	    {
		PRINTF (DBG_USER, DBG_FAULT, "user_load_range: %s %lx\n", vma ? "not writeble at" : "no vma for", maddr);
		up_read (&mm->mmap_sem);
		kmutex_unlock (&uctx->uctx_rgnmutex);
		return (-EFAULT);
	    }

	    if ((res = user_load_page (uctx, vma, maddr, eaddr, perm, writeable)) < 0)
	    {
		PRINTF (DBG_USER, DBG_FAULT, "user_load_range: failed to load page res=%d\n", res);
		up_read (&mm->mmap_sem);
		kmutex_unlock (&uctx->uctx_rgnmutex);
		return res;
	    }
	    
	    eaddr += PAGE_SIZE;
	    maddr += PAGE_SIZE;
	    len   -= PAGE_SIZE;
	}
	up_read (&mm->mmap_sem);
    }
    kmutex_unlock (&uctx->uctx_rgnmutex);

    PRINTF (uctx, DBG_FAULT, "user_load_range: alldone\n");

    return (0);
}

void
user_preload_main (USER_CTXT *uctx, virtaddr_t addr, unsigned long len)
{
    virtaddr_t             lim = addr + len - 1;
    struct vm_area_struct *vma;

    down_read (&current->mm->mmap_sem);

    if ((vma = find_vma (current->mm, addr)) != NULL)
    {
	do {
	    unsigned long start = vma->vm_start;
	    unsigned long end   = vma->vm_end;

	    if ((start-1) >= lim)
		break;

	    if (start < addr) start = addr;
	    if ((end-1) > lim) end = lim+1;
		
	    if (vma->vm_flags & VM_IO)
		continue;

	    user_unload_main (uctx, start, end - start);

	    if (get_user_pages (current, current->mm, start, (end - start)/PAGE_SIZE, 
				(vma->vm_flags & VM_WRITE) != 0, 0, NULL, NULL) > 0)
		user_update_main (uctx, vma->vm_mm, vma, start, end - start);
	    
	    else if (vma->vm_ops == &mem_vm_ops)
		user_update_main (uctx, vma->vm_mm, vma, start, end - start);
	    else if (vma->vm_ops == &user_vm_ops)
		user_update_main (uctx, vma->vm_mm, vma, start, end - start);

	} while ((vma = find_vma (current->mm, vma->vm_end)) != NULL);
    }
    up_read (&current->mm->mmap_sem);
}

static void
user_update_range (USER_CTXT *uctx, int tbl, struct mm_struct *mm, struct vm_area_struct *vma, virtaddr_t maddr, E4_Addr eaddr, unsigned long len, int perm)
{
    ELAN4_DEV    *dev    = uctx->uctx_ctxt.ctxt_dev;
    int           roperm = ELAN4_PERM_READONLY(perm & PERM_Mask) | (perm & ~PERM_Mask);
    int           i, write;
    pte_t        *ptep;
    struct page  *page;
	
    if (vma && vma->vm_ops == &mem_vm_ops)
    {
	mem_pteload (vma, maddr, &uctx->uctx_ctxt, eaddr, perm);
	return;
    }

    if (vma && vma->vm_ops == &user_vm_ops)
    {
	user_pteload (vma, maddr, &uctx->uctx_ctxt, eaddr, perm);
	return;
    }
    
#ifdef CONFIG_HUGETLB_PAGE
    /* If the kernel has hugetlb support compiled in, then 
     * we can't walk the pte's unless we know for sure that
     * they're normal ptes. */

    if (vma == NULL || is_vm_hugetlb_page (vma))
	return;
#endif

    while (len > 0)
    {
	if ((ptep = find_pte_map (mm, maddr)) != NULL)
	{
	    write = (pte_write(*ptep) && pte_dirty(*ptep));

#if LINUX_VERSION_CODE < KERNEL_VERSION (2, 5, 0)
	    page = pte_page (*ptep);
	    if (! VALID_PAGE (page))
		page = NULL;
#else
	    {
		unsigned long pfn;

		pfn  = pte_pfn (*ptep);
		page = pfn_valid (pfn) ? pfn_to_page (pfn) : NULL;
	    }
#endif
	    pte_unmap (ptep);

	    PRINTF (uctx, DBG_IOPROC, "user_update_range: maddr=%lx eaddr=%llx -> page %p %lx %s\n", maddr, eaddr, page, page_to_pfn (page), write ? "writeable" : "read-only");

	    if (page != NULL)
		for (i = 0; i < PAGE_SIZE; i += (1 << dev->dev_pageshift[tbl]))
		    elan4mmu_pteload_page (&uctx->uctx_ctxt, tbl, eaddr + i, page, write ? perm : roperm);
	}

	eaddr += PAGE_SIZE;
	maddr += PAGE_SIZE;
	len   -= PAGE_SIZE;
    }
}

void
user_update_main (USER_CTXT *uctx, struct mm_struct *mm, struct vm_area_struct *vma, virtaddr_t start, unsigned long len)
{
    USER_RGN     *rgn;
    unsigned long ssize;
    virtaddr_t    end = start + len - 1;

    spin_lock (&uctx->uctx_rgnlock);

    PRINTF (uctx, DBG_IOPROC, "user_update_main: start=%lx end=%lx\n", start, end);

    for (rgn = user_findrgn_main (uctx, start, 0); rgn != NULL; rgn = rgn->rgn_mnext)
    {
	if (end < rgn->rgn_mbase)
	    break;
	
	if (start <= rgn->rgn_mbase && end >= (rgn->rgn_mbase + rgn->rgn_len - 1)) 
	{
	    PRINTF (uctx, DBG_IOPROC, "user_update_main: whole %lx -> %lx\n", rgn->rgn_mbase, rgn->rgn_mbase + rgn->rgn_len - 1);

	    user_update_range (uctx, 0  /* tbl */, mm, vma, rgn->rgn_mbase, rgn->rgn_ebase, rgn->rgn_len, rgn->rgn_perm);
	}
	else if (start <= rgn->rgn_mbase)
	{
	    ssize = end - rgn->rgn_mbase + 1;

	    PRINTF (uctx, DBG_IOPROC, "user_update_main: start %lx -> %lx\n", rgn->rgn_mbase, rgn->rgn_mbase + ssize);

	    user_update_range (uctx, 0 /* tbl */, mm, vma, rgn->rgn_mbase, rgn->rgn_ebase, ssize, rgn->rgn_perm);
	}
	else if (end >= (rgn->rgn_mbase + rgn->rgn_len - 1))
	{
	    ssize = (rgn->rgn_mbase + rgn->rgn_len) - start;

	    PRINTF (uctx, DBG_IOPROC, "user_update_main: end   %lx -> %lx\n", start, start + ssize);

	    user_update_range (uctx, 0 /* tbl */, mm, vma, start, rgn->rgn_ebase + (start - rgn->rgn_mbase), ssize, rgn->rgn_perm);
	}
	else
	{
	    PRINTF (uctx, DBG_IOPROC, "user_update_main: middle %lx -> %lx\n", start, end);

	    user_update_range (uctx, 0 /* tbl */, mm, vma, start, rgn->rgn_ebase + (start - rgn->rgn_mbase), len, rgn->rgn_perm);  
	}
    }
    spin_unlock (&uctx->uctx_rgnlock);
}

void
user_unload_main (USER_CTXT *uctx, virtaddr_t start, unsigned long len)
{
    USER_RGN     *rgn;
    unsigned long ssize;
    virtaddr_t    end = start + len - 1;

    spin_lock (&uctx->uctx_rgnlock);

    PRINTF (uctx, DBG_IOPROC, "user_unload_main: start=%lx end=%lx\n", start, end);

    for (rgn = user_findrgn_main (uctx, start, 0); rgn != NULL; rgn = rgn->rgn_mnext)
    {
	if (end < rgn->rgn_mbase)
	    break;
	
	if (start <= rgn->rgn_mbase && end >= (rgn->rgn_mbase + rgn->rgn_len - 1))
	{
	    PRINTF (uctx, DBG_IOPROC, "user_unload_main: whole %lx -> %lx\n", rgn->rgn_mbase, rgn->rgn_mbase + rgn->rgn_len - 1);

	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* tbl */, rgn->rgn_ebase, rgn->rgn_len);
	}
	else if (start <= rgn->rgn_mbase)
	{
	    ssize = end - rgn->rgn_mbase + 1;

	    PRINTF (uctx, DBG_IOPROC, "user_unload_main: start %lx -> %lx\n", rgn->rgn_mbase, rgn->rgn_mbase + ssize);

	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* tbl */, rgn->rgn_ebase, ssize);
	}
	else if (end >= (rgn->rgn_mbase + rgn->rgn_len - 1))
	{
	    ssize = (rgn->rgn_mbase + rgn->rgn_len) - start;
	    
	    PRINTF (uctx, DBG_IOPROC, "user_unload_main: end   %lx -> %lx\n", start, start + ssize);
	    
	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* tbl */, rgn->rgn_ebase + (start - rgn->rgn_mbase), ssize);
	}
	else
	{

	    PRINTF (uctx, DBG_IOPROC, "user_unload_main: middle %lx -> %lx\n", start, end);

	    elan4mmu_unload_range (&uctx->uctx_ctxt, 0 /* tbl */, rgn->rgn_ebase + (start - rgn->rgn_mbase), len);
	}
    }
    spin_unlock (&uctx->uctx_rgnlock);
}
/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
