/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2005 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: cm_procfs.c,v 1.5 2004/05/14 09:23:13 daniel Exp $"
/*      $Source: /cvs/master/quadrics/epmod/cm_procfs.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "debug.h"
#include "cm.h"
#include <elan/epsvc.h>

#include <qsnet/procfs_linux.h>

extern char *sprintClPeers (char *str, CM_RAIL *cmRail, int clvl);

static int
proc_read_cluster(char *page, char **start, off_t off,
		int count, int *eof, void *data)
{
    CM_RAIL *cmRail = (CM_RAIL *) data;
    char    *p      = page;

    page[0] = 0;

    if (cmRail->Rail->State != EP_RAIL_STATE_RUNNING)
	p += sprintf(p, "<not running>\n");
    else
    {
	CM_LEVEL *cmLevel;
	unsigned long flags;
	int  i, j;
	char clNodeStr[32]; /* [%d-%d][%d-%d] */
	char seperate_with;

	struct { int val; char *name; } bitvals[] = {
	    {CM_OFFLINE_BROADCAST, "Broadcast"},
	    {CM_OFFLINE_PROCFS,    "Offline"},
	    {CM_OFFLINE_MANAGER,   "Manager"}};
	
	spin_lock_irqsave (&cmRail->Lock, flags);
	
	for (i = 0; i < cmRail->NumLevels; i++)
	{
	    cmLevel = &cmRail->Levels[i];
	    
	    p += sprintf(p, "%23s %7s ", sprintClPeers (clNodeStr, cmRail, i), cmLevel->Online?"Online":"Offline");
	    
	    if ((cmLevel->Online ) | ( cmLevel->Connected > 0))
		p += sprintf(p, "Connected=%lu ", cmLevel->Connected);
	    
	    seperate_with = '<';
	    
	    if ( cmLevel->Restarting ) {
		p += sprintf(p, "%cRestarting", seperate_with);
		seperate_with = ',';
	    }
	    
	    if ( ! (cmLevel->GlobalMapValid & cmLevel->SubTreeMapValid )) {
		p += sprintf(p, "%cMap Not Valid", seperate_with);
		seperate_with = ',';
	    }
	    
	    if ( cmLevel->OfflineReasons ) {
		for (j = 0; j < sizeof (bitvals)/sizeof(bitvals[0]); j++)
		    if (cmLevel->OfflineReasons & bitvals[j].val) {
			p += sprintf(p, "%c%s", seperate_with, bitvals[j].name);
			seperate_with = ',';
		    }
	    }
	    if ( cmRail->OfflineReasons ) {
		for (j = 0; j < sizeof (bitvals)/sizeof(bitvals[0]); j++)
		    if (cmRail->OfflineReasons & bitvals[j].val) {
			p += sprintf(p, "%c%s", seperate_with, bitvals[j].name);
			seperate_with = ',';
		    }
	    }
	    
	    if ( seperate_with != '<' ) 
		p += sprintf(p,">\n");
	    else
		p += sprintf(p,"\n");
	}
	
	spin_unlock_irqrestore (&cmRail->Lock, flags);
    }

    return qsnet_proc_calc_metrics (page, start, off, count, eof, p - page);
}

static struct rail_info
{
    char *name;
    int (*read_func) (char *page, char **start, off_t off, int count, int *eof, void *data);
    int (*write_func) (struct file *file, const char *buf, unsigned long count, void *data);
} rail_info[] = {
    {"cluster", proc_read_cluster, NULL},
};

struct proc_dir_entry *svc_indicators_root;

typedef struct svc_indicator_data
{
    int       svc_indicator;
    EP_RAIL  *rail;
} SVC_INDICATOR_DATA;

static SVC_INDICATOR_DATA svc_indicator_data[EP_SVC_NUM_INDICATORS][EP_MAX_RAILS];
static char              *svc_indicator_names[EP_SVC_NUM_INDICATORS] = EP_SVC_NAMES;

static int
proc_read_svc_indicator_rail_bitmap (char *page, char **start, off_t off,
				     int count, int *eof, void *data)
{
    SVC_INDICATOR_DATA  *svc_data = (SVC_INDICATOR_DATA  *)data;
    unsigned int	 nnodes   = ep_numnodes (ep_system());
    bitmap_t		*bitmap;

    KMEM_ZALLOC (bitmap, bitmap_t *, (BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t)), 1);

    cm_svc_indicator_bitmap (svc_data->rail, svc_data->svc_indicator, bitmap, 0, nnodes);

    ep_sprintf_bitmap (page, PAGESIZE, bitmap, 0, 0, nnodes);
    
    KMEM_FREE (bitmap, (BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t)));
    
    strcat (page, "\n");

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

static int
proc_read_svc_indicator_bitmap(char *page, char **start, off_t off,
			       int count, int *eof, void *data)
{
    unsigned int         num      = (unsigned long) data;
    EP_SYS              *sys      = ep_system();
    unsigned int         nnodes   = ep_numnodes (sys);
    bitmap_t            *bitmap;

    KMEM_ALLOC(bitmap, bitmap_t *, (BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t)), 1);
     
    ep_svc_indicator_bitmap (sys, num, bitmap, 0, nnodes);

    ep_sprintf_bitmap (page, PAGESIZE, bitmap, 0, 0, nnodes);
    
    KMEM_FREE (bitmap, (BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t)));
    
    strcat (page, "\n");

    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

void
cm_procfs_rail_init (CM_RAIL *cmRail)
{
    EP_RAIL *rail = cmRail->Rail;
    struct proc_dir_entry *p;
    int i;
    
    for (i = 0; i < sizeof (rail_info)/sizeof (rail_info[0]); i++)
    {
	if ((p = create_proc_entry (rail_info[i].name, 0, cmRail->Rail->ProcDir)) != NULL)
	{
	    p->read_proc  = rail_info[i].read_func;
	    p->write_proc = rail_info[i].write_func;
	    p->data       = cmRail;
	    p->owner      = THIS_MODULE;
	}
    }

    if ((rail->SvcIndicatorDir = proc_mkdir ("svc_indicators", cmRail->Rail->ProcDir)) != NULL)
    {
	for (i = 0; i < EP_SVC_NUM_INDICATORS; i++)
	{
	    if ((p = create_proc_entry (svc_indicator_names[i], 0, rail->SvcIndicatorDir)) != NULL)
	    {
		svc_indicator_data[i][rail->Number].svc_indicator = i;
		svc_indicator_data[i][rail->Number].rail          = rail; 
		
		p->write_proc = NULL;
		p->read_proc  = proc_read_svc_indicator_rail_bitmap;
		p->data       = (void *)&svc_indicator_data[i][rail->Number];
		p->owner      = THIS_MODULE;
	    }
	}
    }
}

void
cm_procfs_rail_fini (CM_RAIL *cmRail)
{
    EP_RAIL *rail = cmRail->Rail;
    int i;

    if (rail->SvcIndicatorDir)
    {
	for (i = 0; i < EP_SVC_NUM_INDICATORS; i++)
	    remove_proc_entry (svc_indicator_names[i], rail->SvcIndicatorDir);

	remove_proc_entry ("svc_indicators", cmRail->Rail->ProcDir);
    }

    for (i = 0; i < sizeof (rail_info)/sizeof (rail_info[0]); i++)
	remove_proc_entry (rail_info[i].name, cmRail->Rail->ProcDir);
}

void
cm_procfs_init (CM_SUBSYS *subsys)
{
    struct proc_dir_entry *p;
    int i;

    qsnet_proc_register_hex (ep_config_root, "machine_id",      &MachineId,      0);

    if ((svc_indicators_root = proc_mkdir("svc_indicators", ep_procfs_root)) != NULL)
    {
	for (i = 0; i < EP_SVC_NUM_INDICATORS; i++)
	{
	    if ((p = create_proc_entry (svc_indicator_names[i], 0, svc_indicators_root)) != NULL)
	    {
		p->write_proc = NULL;
		p->read_proc  = proc_read_svc_indicator_bitmap;
		p->data       = (void *)(long) i;
		p->owner      = THIS_MODULE;
	    }
	}
	
    }
}

void
cm_procfs_fini (CM_SUBSYS *subsys)
{
    int i;

    if (svc_indicators_root)
    {
	for (i = 0; i < EP_SVC_NUM_INDICATORS; i++)
	    remove_proc_entry (svc_indicator_names[i], svc_indicators_root);
	
	remove_proc_entry ("svc_indicators",   ep_procfs_root);
    }

    remove_proc_entry ("machine_id",      ep_config_root);
}
