/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: commands_elan4.c,v 1.2.10.1 2005/07/20 12:01:33 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/commands_elan4.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "debug.h"

#include <elan4/trtype.h>

static __inline__ void
elan4_command_write (ELAN4_CQ *cq, E4_uint64 val, unsigned off)
{
    writeq (val, (void *)(cq->cq_mapping + offsetof (E4_CommandPort, Command[off])));
}

void
elan4_nop_cmd (ELAN4_CQ *cq, E4_uint64 tag)
{
    elan4_command_write (cq, tag | NOP_CMD, 0);
}

void
elan4_write_dword_cmd (ELAN4_CQ *cq, E4_Addr addr, E4_uint64 data)
{
    elan4_command_write (cq, addr | WRITE_DWORD_CMD, 0);
    elan4_command_write (cq, data, 1);
}

void
elan4_add_dword_cmd (ELAN4_CQ *cq, E4_Addr addr, E4_uint64 data)
{
    elan4_command_write (cq, addr | ADD_DWORD_CMD, 0);
    elan4_command_write (cq, data,                 1);
}

void
elan4_copy64_cmd (ELAN4_CQ *cq, E4_Addr from, E4_Addr to, E4_uint32 datatype)
{
    elan4_command_write (cq, from | (datatype << COPY64_DATA_TYPE_SHIFT) | COPY64_CMD, 0);
    elan4_command_write (cq, to   | (datatype << COPY64_DATA_TYPE_SHIFT),              1);
}

void
elan4_interrupt_cmd (ELAN4_CQ *cq, E4_uint64 cookie)
{
    elan4_command_write (cq, (cookie << E4_MAIN_INT_SHIFT) | INTERRUPT_CMD, 0);
}


void 
elan4_run_thread_cmd (ELAN4_CQ *cq, E4_ThreadRegs *regs)
{
    elan4_command_write (cq, regs->Registers[0] | RUN_THREAD_CMD, 0);
    elan4_command_write (cq, regs->Registers[1],                  1);
    elan4_command_write (cq, regs->Registers[2],                  2);
    elan4_command_write (cq, regs->Registers[3],                  3);
    elan4_command_write (cq, regs->Registers[4],                  4);
    elan4_command_write (cq, regs->Registers[5],                  5);
    elan4_command_write (cq, regs->Registers[6],                  6);
}

void
elan4_run_dma_cmd (ELAN4_CQ *cq, E4_DMA *dma)
{
    E4_uint64 *dmaptr = (E4_uint64 *) dma;

    elan4_command_write (cq, dmaptr[0] | RUN_DMA_CMD, 0);
    elan4_command_write (cq, dmaptr[1],               1);
    elan4_command_write (cq, dmaptr[2],               2);
    elan4_command_write (cq, dmaptr[3],               3);
    elan4_command_write (cq, dmaptr[4],               4);
    elan4_command_write (cq, dmaptr[5],               5);
    elan4_command_write (cq, dmaptr[6],               6);
}

void
elan4_set_event_cmd (ELAN4_CQ *cq, E4_Addr event)
{
    elan4_command_write (cq, event | SET_EVENT_CMD, 0);
}

void
elan4_set_eventn_cmd (ELAN4_CQ *cq, E4_Addr event, E4_uint32 count)
{
    elan4_command_write (cq, SET_EVENTN_CMD,0);
    elan4_command_write (cq, event | count, 1);
}
    
void
elan4_wait_event_cmd (ELAN4_CQ *cq, E4_Addr event, E4_uint64 candt, E4_uint64 param0, E4_uint64 param1)
{
    elan4_command_write (cq, event | WAIT_EVENT_CMD, 0);
    elan4_command_write (cq, candt,                  1);
    elan4_command_write (cq, param0,                 2);
    elan4_command_write (cq, param1,                 3);
}

void
elan4_open_packet (ELAN4_CQ *cq, E4_uint64 command)
{
    elan4_command_write (cq, command | OPEN_STEN_PKT_CMD, 0);
}

void
elan4_guard (ELAN4_CQ *cq, E4_uint64 command)
{
    elan4_command_write (cq, command | GUARD_CMD, 0);
}

void
elan4_sendtrans0 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr)
{
    elan4_command_write (cq, (trtype << 16) | SEND_TRANS_CMD, 0);
    elan4_command_write (cq, addr,			          1);
}

void
elan4_sendtrans1 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 p0)
{
    elan4_command_write (cq, (trtype << 16) | SEND_TRANS_CMD, 0);
    elan4_command_write (cq, addr,			          1);
    elan4_command_write (cq, p0, 			          2);
}

void
elan4_sendtrans2 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 p0, E4_uint64 p1)
{
    elan4_command_write (cq, (trtype << 16) | SEND_TRANS_CMD, 0);
    elan4_command_write (cq, addr,			          1);
    elan4_command_write (cq, p0, 			          2);
    elan4_command_write (cq, p1, 			          3);
}

void
elan4_sendtransn (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, ...)
{
    E4_uint32    ndword = ((trtype & TR_SIZE_MASK) >> TR_SIZE_SHIFT);
    va_list      ap;
    register int i;

    elan4_command_write (cq, (trtype << 16) | SEND_TRANS_CMD, 0);
    elan4_command_write (cq, addr,			          1);
    
    va_start (ap, addr);
    for (i = 2; i < ndword+2; i++) 
	elan4_command_write (cq, va_arg (ap, E4_uint64), i);
    va_end (ap);
}

void
elan4_sendtransp (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 *ptr)
{
    E4_uint32    ndword = ((trtype &TR_SIZE_MASK) >> TR_SIZE_SHIFT);
    register int i;

    elan4_command_write (cq, (trtype << 16) | SEND_TRANS_CMD, 0);
    elan4_command_write (cq, addr,                            1);
    for (i = 2; i < ndword+2; i++)
	elan4_command_write (cq, *ptr++, i);
}

