/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: conf_linux.c,v 1.40.2.3 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/conf_linux.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "cm.h"

#include "conf_linux.h"

#include <linux/init.h>
#include <qsnet/module.h>
#include <linux/reboot.h>
#include <linux/notifier.h>

/* Module parameters */
unsigned int epdebug	     = 0;
unsigned int epdebug_console = 0;
unsigned int epdebug_cmlevel = 0;
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
unsigned int epdebug_check_sum = 0;
#endif
int	     disabled        = 0;
int          sdram_assert    = 0;
int          assfail_mode    = 0;
int	     txd_stabilise   = 7;
int	     portals_envelopes = 0;

/* External module parameters */
extern int	MaxSwitchLevels;
extern int      RejoinCheck;
extern int      RejoinPanic;
extern int      PositionCheck;
extern int      MachineId;

/* Module globals */
EP_SYS          epsys;

#ifdef MODULE
MODULE_AUTHOR("Quadrics Ltd");
MODULE_DESCRIPTION("Elan Kernel Comms");

MODULE_LICENSE("GPL");

module_param(epdebug,         uint, 0);
module_param(epdebug_console, uint, 0);
module_param(epdebug_cmlevel, uint, 0);
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
module_param(epdebug_check_sum, uint, 0);
#endif
module_param(disabled,        uint, 0);

module_param(MachineId,       uint, 0);
module_param(RejoinPanic,     uint, 0);
module_param(RejoinCheck,     uint, 0);
module_param(PositionCheck,   uint, 0);
module_param(MaxSwitchLevels, uint, 0);

module_param(sdram_assert,    uint, 0);
module_param(assfail_mode,    uint, 0);
module_param(txd_stabilise,   uint, 0);
module_param(portals_envelopes,uint, 0);

/* epcomms.c large message service functions */
EXPORT_SYMBOL(ep_alloc_xmtr);
EXPORT_SYMBOL(ep_free_xmtr);
EXPORT_SYMBOL(ep_transmit_message);
EXPORT_SYMBOL(ep_multicast_message);
EXPORT_SYMBOL(ep_transmit_rpc);

EXPORT_SYMBOL(ep_alloc_rcvr);
EXPORT_SYMBOL(ep_free_rcvr);
EXPORT_SYMBOL(ep_queue_receive);
EXPORT_SYMBOL(ep_requeue_receive);
EXPORT_SYMBOL(ep_rpc_put);
EXPORT_SYMBOL(ep_rpc_get);
EXPORT_SYMBOL(ep_complete_rpc);
EXPORT_SYMBOL(ep_complete_receive);

EXPORT_SYMBOL(ep_poll_transmits);
EXPORT_SYMBOL(ep_enable_txcallbacks);
EXPORT_SYMBOL(ep_disable_txcallbacks);

/* epcomms.c functions for accessing fields of rxds/txds */
EXPORT_SYMBOL(ep_rxd_arg);
EXPORT_SYMBOL(ep_rxd_len);
EXPORT_SYMBOL(ep_rxd_isrpc);
EXPORT_SYMBOL(ep_rxd_envelope);
EXPORT_SYMBOL(ep_rxd_payload);
EXPORT_SYMBOL(ep_rxd_node);
EXPORT_SYMBOL(ep_rxd_status);
EXPORT_SYMBOL(ep_rxd_statusblk);
EXPORT_SYMBOL(ep_txd_node);
EXPORT_SYMBOL(ep_txd_statusblk);

/* kmap.c, nmh.c - handling mapping of pages into network memory */
EXPORT_SYMBOL(ep_dvma_reserve);
EXPORT_SYMBOL(ep_dvma_release);
EXPORT_SYMBOL(ep_dvma_load);
EXPORT_SYMBOL(ep_dvma_unload);
EXPORT_SYMBOL(ep_nmd_subset);
EXPORT_SYMBOL(ep_nmd_merge);

EXPORT_SYMBOL(ep_system);

/* kcomm.c */
EXPORT_SYMBOL(ep_nodeid);
EXPORT_SYMBOL(ep_numnodes);
EXPORT_SYMBOL(ep_waitfor_nodeid);

/* railhints.c */
EXPORT_SYMBOL(ep_pickRail);
EXPORT_SYMBOL(ep_xmtr_bcastrail);
EXPORT_SYMBOL(ep_xmtr_prefrail);
EXPORT_SYMBOL(ep_xmtr_availrails);
EXPORT_SYMBOL(ep_xmtr_noderails);
EXPORT_SYMBOL(ep_rcvr_prefrail);
EXPORT_SYMBOL(ep_rcvr_availrails);
EXPORT_SYMBOL(ep_rxd_railmask);

EXPORT_SYMBOL(ep_svc_indicator_bitmap);
EXPORT_SYMBOL(ep_svc_indicator_is_set);
EXPORT_SYMBOL(ep_svc_indicator_clear);
EXPORT_SYMBOL(ep_svc_indicator_set);

/* cm.c */
EXPORT_SYMBOL(cm_svc_indicator_clear);
EXPORT_SYMBOL(cm_svc_indicator_set);
EXPORT_SYMBOL(cm_svc_indicator_is_set);
EXPORT_SYMBOL(cm_svc_indicator_bitmap);

#endif

EP_SYS *
ep_system()
{
    return (&epsys);
}

void
ep_mod_inc_usecount()
{
    MOD_INC_USE_COUNT;
} 

void
ep_mod_dec_usecount()
{
    MOD_DEC_USE_COUNT;
}

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)

#include <linux/dump.h>

static int
ep_dump_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if (event == DUMP_BEGIN)
	ep_shutdown (&epsys);

    return (NOTIFY_DONE);
}
static struct notifier_block ep_dump_notifier = 
{
    notifier_call:	ep_dump_event,
    priority:		0,
};

#endif

static int
ep_reboot_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    if ((event == SYS_RESTART || event == SYS_HALT || event == SYS_POWER_OFF))
	ep_shutdown (&epsys);

    return (NOTIFY_DONE);
}

static struct notifier_block ep_reboot_notifier = 
{
    notifier_call:	ep_reboot_event,
    priority:		0,
};

#if !defined(NO_PANIC_NOTIFIER)
static int
ep_panic_event (struct notifier_block *self, unsigned long event, void *buffer)
{
    ep_shutdown (&epsys);

    return (NOTIFY_DONE);
}

static struct notifier_block ep_panic_notifier = 
{
    notifier_call:	ep_panic_event,
    priority:		0,
};
#endif

/*
 * Module configuration. 
 */
#ifdef MODULE
static int __init ep_init(void)
#else
__initfunc(int ep_init(void))
#endif
{
    register int rmask = 0;

    ep_procfs_init ();

    ep_sys_init (&epsys);

#if defined(CONFIG_ELAN4) || defined(CONFIG_ELAN4_MODULE)
    rmask = ep4_create_rails (&epsys, disabled);
#endif
    
    /* If we've brought up an elan4 rail, then disable all elan3 rails. */
    if ((rmask & ~disabled) != 0)
	disabled = ~rmask;

#if defined(CONFIG_ELAN3) || defined(CONFIG_ELAN3_MODULE)
    rmask = ep3_create_rails (&epsys, disabled);
#endif

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
    register_dump_notifier (&ep_dump_notifier);
#endif
    register_reboot_notifier (&ep_reboot_notifier);

#if !defined(NO_PANIC_NOTIFIER)
    notifier_chain_register (&panic_notifier_list, &ep_panic_notifier);
#endif

    return (0);
}

/*
 * Module removal.
 */
#ifdef MODULE
static void
__exit ep_exit(void)
{
    register int i;

#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
    unregister_dump_notifier (&ep_dump_notifier);
#endif
    unregister_reboot_notifier (&ep_reboot_notifier);

#if !defined(NO_PANIC_NOTIFIER)
    notifier_chain_unregister (&panic_notifier_list, &ep_panic_notifier);
#endif

    for (i = 0; i < EP_MAX_RAILS; i++)
    {
	if (epsys.Rails[i])
	{
	    switch (epsys.Rails[i]->State)
	    {
	    case EP_RAIL_STATE_UNINITIALISED:
		break;

	    case EP_RAIL_STATE_STARTED:
	    case EP_RAIL_STATE_RUNNING:
	    case EP_RAIL_STATE_INCOMPATIBLE:
		/* remove per-rail CM proc entries */
		ep_stop_rail (epsys.Rails[i]);
		break;
	    }

	    /* remove EP proc rail entries after per-rail CM entries */
	    ep_procfs_rail_fini (epsys.Rails[i]);
	    ep_destroy_rail (epsys.Rails[i]);
	}
    }

    ep_sys_fini (&epsys);

    ep_procfs_fini ();
}

/* Declare the module init and exit functions */
module_init(ep_init);
module_exit(ep_exit);

#endif


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
