/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: debug.c,v 1.30 2004/11/12 10:55:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/debug.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "debug.h"

DisplayInfo di_ep_debug = {ep_debugf, DBG_DEBUG};

/*
 * Generate a partial bitmap string, for the bitmap from offset "off" for "count" bits,
 * to allow for displaying of subsets, treat entry 0 of the bitmap as having value "base".
 */
int
ep_sprintf_bitmap (char *str, unsigned nbytes, bitmap_t *bitmap, int base, int off, int nbits)
{
    char entry[12];						/* space for N-N */
    register int i, j, len;
    register int notstart = off;
    register int notfirst = 0;
    char *p = str;
    
    for (i = off; i < nbits; i++)
    {
	if (BT_TEST (bitmap, i))
	{
	    for (j = i+1; j < nbits; j++)
		if (! BT_TEST (bitmap, j))
		    break;

	    if (j == (i+1))
		len = (int)sprintf (entry, "%d", base + i);
	    else
		len = (int)sprintf (entry, "%d-%d", base + i, base + j-1);
	    
	    /* NOTE the 2 is for: one for comma, one for (possible) closing bracket */
	    if ((p - str) <= (nbytes - (len+3)))
		p += (int)sprintf (p, "%c%s", notfirst++ ? ',' : notstart ? ' ' : '[', entry);
	    else
	    {
		/* no more space on this line, so move onto next */
		sprintf (p, "%c", notfirst++ ? ',' : '[');

		return (i);
	    }

	    i = j;
	}
    }
    
    if (!notfirst)
	sprintf (str, "<empty>");
    else
	strcpy (p, "]");

    return (-1);
}

void
ep_display_bitmap (char *prefix, char *tag, bitmap_t *bitmap, unsigned base, unsigned nbits)
{
    /* Tru64 kernel printf() truncates lines at 128 bytes - the man pages for printf (9)
     * do not mention this restriction, nor that it does not terminate the line with a
     * carriage return, this  is pretty naff. 
     * Linux has a similar limit though is much more generous at 1024 - and you can just 
     * look at the code to see why this has been done.
     *
     * Our nodeset information could well be longer than 128 characters,  so we're going to 
     * have to split it into a number of lines. */

#define LINEBUF_SIZE		128
    char *p, linebuf[LINEBUF_SIZE+1];				/* +1 for null termination */
    int i, noff, off = 0;

    do {
	if (off == 0)
	    p = linebuf + (int)sprintf (linebuf, "%s: %s ", prefix, tag);
	else
	{
	    p = linebuf + (int)sprintf (linebuf, "%s:  ", prefix);
	    for (i = 0; tag[i] != '\0'; i++)
		*p++ = ' ';
	}

	noff = ep_sprintf_bitmap (p, &linebuf[LINEBUF_SIZE-1]-p, bitmap, base, off, nbits);

	printk ("%s\n", linebuf);

    } while ((off = noff) != -1);

#undef LINEBUF_SIZE
}

void
ep_debugf (long mode, char *fmt, ...)
{
   va_list ap;
   char prefix[32];
   
   va_start (ap, fmt);
#if defined(LINUX)
   sprintf (prefix, "[%08d.%04d] ", (int) lbolt, current->pid);
#else
   sprintf (prefix, "[%08d.----] ", (int) lbolt);
#endif
   qsnet_vdebugf ((mode & epdebug_console ? QSNET_DEBUG_CONSOLE: 0) | QSNET_DEBUG_BUFFER, prefix, fmt, ap);
   va_end (ap);
}

int
ep_assfail (EP_RAIL *rail, const char *ex, const char *func, const char *file, const int line)
{
    qsnet_debugf (QSNET_DEBUG_BUFFER, "ep: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);
    
    printk (KERN_EMERG "ep: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);
    
    if (panicstr)
	return (0);
    
    if (assfail_mode & 1)				/* return to BUG() */
	return 1;
    
    if (assfail_mode & 2)
	panic ("ep: assertion failure: %s, function: %s, file %s, line: %d\n", ex, func, file, line);
    if (assfail_mode & 4)
	epdebug = 0;
    
    return 0;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
