
/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: ep_procfs.c,v 1.10 2005/03/30 14:06:34 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/ep_procfs.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "cm.h"
#include "debug.h"
#include "conf_linux.h"

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "epcomms_elan4.h"

#include <qsnet/procfs_linux.h>

struct proc_dir_entry *ep_procfs_xmtr_root;
struct proc_dir_entry *ep_procfs_rcvr_root;

static int
ep_proc_open (struct inode *inode, struct file *file)
{
    QSNET_PROC_PRIVATE *pr;
    int           pages = 4;

    if ((pr = kmalloc (sizeof (QSNET_PROC_PRIVATE), GFP_KERNEL)) == NULL)
	return (-ENOMEM);
    
    do {	
	pr->pr_data_len = PAGESIZE * pages;

	KMEM_ZALLOC (pr->pr_data, char *, pr->pr_data_len, 1);
	if (pr->pr_data == NULL) 
	{ 
	    pr->pr_len  = sprintf (pr->pr_data, "Out of Memory\n");
	    break;
	} 
	
	pr->pr_off     = 0;
	pr->pr_len     = 0;
	pr->pr_data[0] = 0;
	
	pr->pr_di.func  = qsnet_proc_character_fill;
	pr->pr_di.arg   = (long)pr;
	
	if (!strcmp("debug_xmtr", file->f_dentry->d_iname)) 
	{   
	    EP_XMTR *xmtr = (EP_XMTR *)(PDE(inode)->data);
	    ep_display_xmtr (&pr->pr_di, xmtr);
	}
	
	if (!strcmp("debug_rcvr", file->f_dentry->d_iname)) 
	{
	    EP_RCVR *rcvr = (EP_RCVR *)(PDE(inode)->data);
	    ep_display_rcvr (&pr->pr_di, rcvr, 0);
	}
	
	if (!strcmp("debug_full", file->f_dentry->d_iname)) 
	{
	    EP_RCVR *rcvr = (EP_RCVR *)(PDE(inode)->data);
	    ep_display_rcvr (&pr->pr_di, rcvr, 1);
	}

	if ( pr->pr_len < pr->pr_data_len) 
	    break; /* we managed to get all the output into the buffer */

	pages++;
	KMEM_FREE ( pr->pr_data,  pr->pr_data_len);
    } while (1);
	

    file->private_data = (void *) pr;

    MOD_INC_USE_COUNT;
    return (0);
}

struct file_operations ep_proc_operations = 
{
    read:	qsnet_proc_read,
    open:	ep_proc_open,
    release:	qsnet_proc_release,
};

static int
proc_read_rcvr_stats(char *page, char **start, off_t off,
		     int count, int *eof, void *data)
{
    EP_RCVR *rcvr = (EP_RCVR *)data;
    
    if (rcvr == NULL) 
	sprintf(page,"proc_read_rcvr_stats rcvr=NULL\n");
    else {
	page[0] = 0;
	ep_rcvr_fillout_stats(rcvr,page);
    }
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

static int
proc_read_rcvr_rail_stats(char *page, char **start, off_t off,
		     int count, int *eof, void *data)
{
    EP_RCVR_RAIL *rcvr_rail = (EP_RCVR_RAIL *)data;

    if (rcvr_rail == NULL) {
	strcpy(page,"proc_read_rcvr_rail_stats rcvr_rail=NULL");
    } else {
	page[0] = 0;
	ep_rcvr_rail_fillout_stats(rcvr_rail, page);
	EP_RCVR_OP(rcvr_rail,FillOutRailStats)(rcvr_rail,page);
    }
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

void
ep_procfs_rcvr_add(EP_RCVR *rcvr)
{ 
    /* ep/rcvr/service_number/stats       */
    /* ep/rcvr/service_number/debug_rcvr  */
    /* ep/rcvr/service_number/debug_full  */
    struct proc_dir_entry *p;
    char str[32];

    sprintf(str,"%d", rcvr->Service);

    rcvr->procfs_root = proc_mkdir (str, ep_procfs_rcvr_root);

    if ((p = create_proc_entry ("stats", 0,  rcvr->procfs_root)) != NULL)
    {
	p->write_proc = NULL;
	p->read_proc  = proc_read_rcvr_stats;
	p->data       = rcvr;
	p->owner      = THIS_MODULE;
    }

    if ((p = create_proc_entry ("debug_rcvr", 0, rcvr->procfs_root)) != NULL)
    {
	p->proc_fops = &ep_proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = rcvr;
    }

    if ((p = create_proc_entry ("debug_full", 0, rcvr->procfs_root)) != NULL)
    {
	p->proc_fops = &ep_proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = rcvr;
    }
}

void
ep_procfs_rcvr_del(EP_RCVR *rcvr)
{  
    char str[32];
    sprintf(str,"%d", rcvr->Service);

    remove_proc_entry ("debug_full", rcvr->procfs_root);
    remove_proc_entry ("debug_rcvr", rcvr->procfs_root);
    remove_proc_entry ("stats",      rcvr->procfs_root);

    remove_proc_entry (str, ep_procfs_rcvr_root);
}

void 
ep_procfs_rcvr_add_rail(EP_RCVR_RAIL *rcvrRail)
{
    /* ep/rcvr/service_number/railN/stats */

    struct proc_dir_entry *p;
    char str[32];
    sprintf(str,"rail%d",rcvrRail->CommsRail->Rail->Number);

    rcvrRail->procfs_root = proc_mkdir (str, rcvrRail->Rcvr->procfs_root);
    
    if ((p = create_proc_entry ("stats", 0,  rcvrRail->procfs_root)) != NULL)
    {
	p->write_proc = NULL;
	p->read_proc  = proc_read_rcvr_rail_stats;
	p->data       = rcvrRail;
	p->owner      = THIS_MODULE;
    } 
}

void 
ep_procfs_rcvr_del_rail(EP_RCVR_RAIL *rcvrRail)
{
    char str[32];
    sprintf(str,"rail%d",rcvrRail->CommsRail->Rail->Number);

    remove_proc_entry ("stats", rcvrRail->procfs_root);

    remove_proc_entry (str, rcvrRail->Rcvr->procfs_root);
}




static int
proc_read_xmtr_stats(char *page, char **start, off_t off,
		     int count, int *eof, void *data)
{
    EP_XMTR *xmtr = (EP_XMTR *)data;

    if (xmtr == NULL) 
	strcpy(page,"proc_read_xmtr_stats xmtr=NULL\n");
    else {
	page[0] = 0;
	ep_xmtr_fillout_stats(xmtr, page);
    }
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

static int
proc_read_xmtr_rail_stats(char *page, char **start, off_t off,
		     int count, int *eof, void *data)
{
    EP_XMTR_RAIL *xmtr_rail = (EP_XMTR_RAIL *)data;

    if (xmtr_rail == NULL) 
	strcpy(page,"proc_read_xmtr_rail_stats xmtr_rail=NULL\n");
    else {
	page[0] = 0;
	ep_xmtr_rail_fillout_stats(xmtr_rail, page);
	EP_XMTR_OP(xmtr_rail,FillOutRailStats)(xmtr_rail,page);
    }
    return (qsnet_proc_calc_metrics (page, start, off, count, eof, strlen(page)));
}

void
ep_procfs_xmtr_add(EP_XMTR *xmtr)
{ 
    /* ep/xmtr/service_number/stats       */
    /* ep/xmtr/service_number/debug_xmtr  */
    struct proc_dir_entry *p;
    char str[32];

    sprintf(str,"%llx", (unsigned long long) (unsigned long)xmtr);

    xmtr->procfs_root = proc_mkdir (str, ep_procfs_xmtr_root);

    if ((p = create_proc_entry ("stats", 0,  xmtr->procfs_root)) != NULL)
    {
	p->write_proc = NULL;
	p->read_proc  = proc_read_xmtr_stats;
	p->data       = xmtr;
	p->owner      = THIS_MODULE;
    } 

    if ((p = create_proc_entry ("debug_xmtr", 0, xmtr->procfs_root)) != NULL)
    {
	p->proc_fops = &ep_proc_operations;
	p->owner     = THIS_MODULE;
	p->data      = xmtr;
    }
}

void
ep_procfs_xmtr_del(EP_XMTR *xmtr)
{  
    char str[32];
    sprintf(str,"%llx", (unsigned long long) (unsigned long)xmtr);

    remove_proc_entry ("stats",      xmtr->procfs_root);
    remove_proc_entry ("debug_xmtr", xmtr->procfs_root);

    remove_proc_entry (str, ep_procfs_xmtr_root);
}

void 
ep_procfs_xmtr_add_rail(EP_XMTR_RAIL *xmtrRail)
{
    /* ep/xmtr/service_number/railN/stats */
    
    struct proc_dir_entry *p;
    char str[32];
    sprintf(str,"rail%d",xmtrRail->CommsRail->Rail->Number);

    xmtrRail->procfs_root = proc_mkdir (str, xmtrRail->Xmtr->procfs_root);

    if ((p = create_proc_entry ("stats", 0,  xmtrRail->procfs_root)) != NULL)
    {
	p->write_proc = NULL;
	p->read_proc  = proc_read_xmtr_rail_stats;
	p->data       = xmtrRail;
	p->owner      = THIS_MODULE;
    } 
}

void 
ep_procfs_xmtr_del_rail(EP_XMTR_RAIL *xmtrRail)
{
    char str[32];
    sprintf(str,"rail%d",xmtrRail->CommsRail->Rail->Number);

    remove_proc_entry ("stats", xmtrRail->procfs_root);

    remove_proc_entry (str, xmtrRail->Xmtr->procfs_root);
}

void
ep_procfs_rcvr_xmtr_init(void)
{
    ep_procfs_rcvr_root = proc_mkdir ("rcvr", ep_procfs_root);
    ep_procfs_xmtr_root = proc_mkdir ("xmtr", ep_procfs_root); 
}

void
ep_procfs_rcvr_xmtr_fini(void)
{
    remove_proc_entry ("rcvr", ep_procfs_root);
    remove_proc_entry ("xmtr", ep_procfs_root);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
