/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsFwd.c,v 1.12.4.1 2005/07/20 12:01:34 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsFwd.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"

unsigned int epcomms_forward_limit = 8;

static void
GenerateTree (unsigned nodeId, unsigned lowId, unsigned  highId, bitmap_t *bitmap, 
	      unsigned *parentp, unsigned *childrenp, int *nchildrenp)
{
    int i;
    int count;
    int branch;
    int nSub;
    int branchIndex;
    int parent;
    int nBranch;
    int rem;
    int self;
    int branchRatio;
    int node;
    int x, y, z;


#ifdef DEBUG_PRINTF
    {
#define OVERFLOW "...]"
#define LINESZ	 128
	char space[LINESZ+1];

	if (ep_sprintf_bitmap (space, LINESZ-strlen(OVERFLOW), bitmap, 0, 0, (highId - lowId)+1) != -1)
	    strcat (space, OVERFLOW);

	EPRINTF3 (DBG_FORWARD, "GenerateTree; elan node low=%d node high=%d bitmap=%s\n", lowId, highId, space);
#undef OVERFLOW
#undef LINESZ
    }
#endif

    /* Count the number of nodes in the partition */
    /* and work out which one I am */
    for (count = 0, self = ELAN_INVALID_NODE, i = lowId; i <= highId; i++)
    {
	if (BT_TEST (bitmap, i-lowId))
	{
	    if (i == nodeId)
		self = count;
	    count++;
	}
    }

    EPRINTF2 (DBG_FORWARD, "GenerateTree: count=%d self=%d\n", count, self);

    if (count == 0 || self == ELAN_INVALID_NODE)
    {
	*parentp    = ELAN_INVALID_NODE;
	*nchildrenp = 0;
	return;
    }

    /* search for position in tree */
    branchRatio = EP_TREE_ARITY;		/* branching ratio */
    branch      = 0;				/* start with process 0 */
    nSub        = count;			/* and whole tree */
    branchIndex = -1;				/* my branch # in parent */
    parent      = -1;				/* my parent's group index # */
    
    while (branch != self)			/* descend process tree */
    {						/* until I find myself */
	parent = branch;
	branch++;				/* parent + 1 = first born */
	nSub--;					/* set # descendents */
	
	rem  = nSub % branchRatio;
	nSub = nSub / branchRatio + 1;
	x = rem * nSub;
	y = self - branch;
	
	if (y < x)				/* my first 'rem' branches have */
	{					/* 1 more descendent... */
	    branchIndex = y / nSub;
	    branch += branchIndex * nSub;
	}
	else					/* than the rest of my branches */
	{
	    nSub--;
	    z = (y - x) / nSub;
	    branchIndex = rem + z;
	    branch += x + z * nSub;
	}
    }

    branch++;					/* my first born */
    nSub--;					/* total # of my descendents */
    /* leaves + their parents may have # children < branchRatio */
    nBranch = (nSub < branchRatio) ? nSub : branchRatio;	

    EPRINTF2 (DBG_FORWARD, "GenerateTree: parent=%d nBranch=%d\n", parent, nBranch);

    /* Now calculate the real elan id's of the parent and my children */
    if (parent == -1)
	*parentp = ELAN_INVALID_NODE;
    else
    {
	for (i = lowId, node = 0; i <= highId; i++)
	{
	    if (BT_TEST(bitmap, i-lowId))
		if (node++ == parent)
		    break;
	}
	*parentp = i;
    }

    for (i = lowId, branchIndex = 0, node = 0; branchIndex < nBranch && i <= highId; i++)
    {
	if (BT_TEST(bitmap, i-lowId))
	{
	    if (node == branch)
	    {
		branch = branch + nSub / branchRatio + ((branchIndex < (nSub % branchRatio)) ? 1 : 0);

		childrenp[branchIndex++] = i;
	    }
	    node++;
	}
    }

    *nchildrenp = branchIndex;
}

static void
ForwardTxDone (EP_TXD *txd, void *arg, EP_STATUS status)
{
    EP_FWD_DESC     *desc   = (EP_FWD_DESC *) arg;
    EP_RXD          *rxd    = desc->Rxd;
    EP_COMMS_SUBSYS *subsys = rxd->Rcvr->Subsys;
    unsigned long    flags;

    /* XXXX: if transmit fails, could step to next node in this subtree ? */

    spin_lock_irqsave (&subsys->ForwardDescLock, flags);

    if (--desc->NumChildren > 0)
	spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);
    else
    {
	rxd->Rcvr->ForwardRxdCount--;

	spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);

	KMEM_FREE (desc, sizeof (EP_FWD_DESC));

	rxd->Handler (rxd);
    }
}

long
ep_forward_rxds (EP_COMMS_SUBSYS *subsys, long nextRunTime)
{
    unsigned long flags;
    int i, res;

    spin_lock_irqsave (&subsys->ForwardDescLock, flags);
    while (! list_empty (&subsys->ForwardDescList)) 
    {
	EP_RXD      *rxd     = (EP_RXD *) list_entry (subsys->ForwardDescList.next, EP_RXD, Link);
	EP_RXD_MAIN *rxdMain = rxd->RxdMain;
	EP_ENVELOPE *env     = &rxdMain->Envelope;
	EP_FWD_DESC *desc;

	EPRINTF2 (DBG_FORWARD, "ep: forwarding rxd %p to range %x\n", rxd, env->Range);

	list_del (&rxd->Link);

	rxd->Rcvr->ForwardRxdCount++;

	spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);

	KMEM_ALLOC (desc, EP_FWD_DESC *, sizeof (EP_FWD_DESC), 1);

	if (desc == NULL)
	{
	    spin_lock_irqsave (&subsys->ForwardDescLock, flags);
	    rxd->Rcvr->ForwardRxdCount--;
	    spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);

	    rxd->Handler (rxd);
	}
	else
	{
	    /* compute the spanning tree for this message */
	    unsigned int destLo = EP_RANGE_LOW (env->Range);
	    unsigned int destHi = EP_RANGE_HIGH (env->Range);
	    unsigned int parent;

	    GenerateTree (subsys->Subsys.Sys->Position.pos_nodeid, destLo, destHi, rxdMain->Bitmap, &parent, desc->Children, &desc->NumChildren);
	    
	    if (desc->NumChildren == 0 || (epcomms_forward_limit && (rxd->Rcvr->ForwardRxdCount >= epcomms_forward_limit)))
	    {
		EPRINTF5 (DBG_FORWARD, "ep; don't forward rxd %p to /%d (%d children/ %d forwarding (%d))\n",
			  rxd, rxd->Rcvr->Service, desc->NumChildren, rxd->Rcvr->ForwardRxdCount, epcomms_forward_limit);

		spin_lock_irqsave (&subsys->ForwardDescLock, flags);
		rxd->Rcvr->ForwardRxdCount--;
		spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);

		KMEM_FREE (desc, sizeof (EP_FWD_DESC));
		
		rxd->Handler (rxd);
	    }
	    else
	    {
		ep_nmd_subset (&desc->Data, &rxd->Data, 0, ep_rxd_len (rxd));
		desc->Rxd = rxd;

		/* NOTE - cannot access 'desc' after last call to multicast, since it could complete
		 *        and free the desc before we access it again.  Hence the reverse loop. */
		for (i = desc->NumChildren-1; i >= 0; i--)
		{
		    ASSERT (desc->Children[i] < subsys->Subsys.Sys->Position.pos_nodes);

		    EPRINTF3 (DBG_FORWARD, "ep: forwarding rxd %p to node %d/%d\n", rxd, desc->Children[i], rxd->Rcvr->Service);

		    if ((res = ep_multicast_forward (subsys->ForwardXmtr, desc->Children[i], rxd->Rcvr->Service, 0, 
						     ForwardTxDone, desc, env, EP_HAS_PAYLOAD(env->Attr) ? &rxdMain->Payload : NULL,  
						     rxdMain->Bitmap, &desc->Data, 1)) != EP_SUCCESS)
		    {
			ep_debugf (DBG_FORWARD, "ep: ep_multicast_forward failed\n");
			ForwardTxDone (NULL, desc, res);
		    }
		}
		
	    }
	}

	spin_lock_irqsave (&subsys->ForwardDescLock, flags);
    }
    spin_unlock_irqrestore (&subsys->ForwardDescLock, flags);

    return (nextRunTime);
}

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
void
ep_csum_rxds (EP_COMMS_SUBSYS *subsys)
{
    unsigned long flags;
 
    spin_lock_irqsave (&subsys->CheckSumDescLock, flags);
    while (! list_empty (&subsys->CheckSumDescList)) 
    {
	EP_RXD      *rxd = (EP_RXD *) list_entry (subsys->CheckSumDescList.next, EP_RXD, CheckSumLink);
	EP_ENVELOPE *env = &rxd->RxdMain->Envelope;

	list_del_init (&rxd->CheckSumLink);
	spin_unlock_irqrestore (&subsys->CheckSumDescLock, flags);

	if (env->CheckSum) {
	    EP_NMD nmd;
	    uint32_t csum; 

	    ep_nmd_subset ( &nmd, &rxd->Data, 0, ep_rxd_len (rxd));

	    csum = ep_calc_check_sum(subsys->Subsys.Sys, env, &nmd, 1);
	    if ( env->CheckSum  != csum ) {
		int f;
	

		printk("Check Sum Error: env(0x%x,0x%x) data(0x%x,0x%x)\n", ((csum >> 16) & 0x7FFF), ((env->CheckSum >> 16) & 0x7FFF), 
		       (csum & 0xFFFF),  (env->CheckSum & 0xFFFF));
		printk("Check Sum Error: Sent : NodeId %u Range 0x%x Service %u Version 0x%x Attr 0x%x\n", env->NodeId, env->Range, rxd->Rcvr->Service, env->Version, env->Attr);
		printk("Check Sum Error: Sent : Xid Generation 0x%x Handle 0x%x Unique 0x%llx\n", env->Xid.Generation, env->Xid.Handle, (long long)env->Xid.Unique);
		printk("Check Sum Error: Sent : TxdRail 0x%x TxdMain nmd_addr 0x%x  nmd_len %u  nmd_attr 0x%x\n",  env->TxdRail, env->TxdMain.nmd_addr, env->TxdMain.nmd_len, env->TxdMain.nmd_attr ); 
		printk("Check Sum Error: Sent : nFrags %d \n", env->nFrags);
		for(f=0;f<env->nFrags;f++)
		    printk("Check Sum Error: Sent (%d): nmd_addr 0x%x   nmd_len %u   nmd_attr 0x%x\n", f,
			   env->Frags[f].nmd_addr, env->Frags[f].nmd_len, env->Frags[f].nmd_attr);
		printk("Check Sum Error: Recv : nmd_addr 0x%x   nmd_len %u   nmd_attr 0x%x\n",
		       nmd.nmd_addr, nmd.nmd_len, nmd.nmd_attr);

	    }
	}
	ep_rxd_received_now(rxd);

	spin_lock_irqsave (&subsys->CheckSumDescLock, flags);
    }
    spin_unlock_irqrestore (&subsys->CheckSumDescLock, flags);
}
#endif

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
