/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsRx.c,v 1.33 2004/11/30 12:02:06 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsRx.c,v $*/

#include <qsnet/kernel.h>
#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"

unsigned int ep_rxd_lowat = 5;

static int
AllocateRxdBlock (EP_RCVR *rcvr, EP_ATTRIBUTE attr, EP_RXD **rxdp)
{
    EP_RXD_BLOCK *blk;
    EP_RXD       *rxd;
    EP_RXD_MAIN  *pRxdMain;
    int		  i;
    unsigned long flags;

    KMEM_ZALLOC (blk, EP_RXD_BLOCK *, sizeof (EP_RXD_BLOCK), ! (attr & EP_NO_SLEEP));

    if (blk == NULL)
	return (ENOMEM);

    if ((pRxdMain = ep_shared_alloc_main (rcvr->Subsys->Subsys.Sys, EP_RXD_MAIN_SIZE * EP_NUM_RXD_PER_BLOCK, attr, &blk->NmdMain)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP_RXD_BLOCK));
	return (ENOMEM);
    }
    
    for (rxd = &blk->Rxd[0], i = 0; i < EP_NUM_RXD_PER_BLOCK; i++, rxd++)
    {
	rxd->Rcvr        = rcvr;
	rxd->RxdMain     = pRxdMain;

	ep_nmd_subset (&rxd->NmdMain, &blk->NmdMain, (i * EP_RXD_MAIN_SIZE), EP_RXD_MAIN_SIZE);

	/* move onto next descriptor */
	pRxdMain = (EP_RXD_MAIN *) ((unsigned long) pRxdMain + EP_RXD_MAIN_SIZE);
    }

    spin_lock_irqsave (&rcvr->FreeDescLock, flags);

    list_add  (&blk->Link, &rcvr->DescBlockList);

    rcvr->TotalDescCount += EP_NUM_RXD_PER_BLOCK;

    for (i = rxdp ? 1 : 0; i < EP_NUM_RXD_PER_BLOCK; i++)
    {
	
#if ! defined(CONFIG_EP_NO_CHECK_SUM)
	INIT_LIST_HEAD (&blk->Rxd[i].CheckSumLink);
#endif

	list_add (&blk->Rxd[i].Link, &rcvr->FreeDescList);
	
	rcvr->FreeDescCount++;

	if (rcvr->FreeDescWanted)
	{
	    rcvr->FreeDescWanted--;
	    kcondvar_wakeupone (&rcvr->FreeDescSleep, &rcvr->FreeDescLock);
	}
    }
    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);
    
    if (rxdp)
    {

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
	INIT_LIST_HEAD (&blk->Rxd[0].CheckSumLink);
#endif
	       
	*rxdp = &blk->Rxd[0];
    }
    return (ESUCCESS);
}

static void
FreeRxdBlock (EP_RCVR *rcvr, EP_RXD_BLOCK *blk)
{
    unsigned long flags;

    spin_lock_irqsave (&rcvr->FreeDescLock, flags);

    list_del (&blk->Link);

    rcvr->TotalDescCount -= EP_NUM_RXD_PER_BLOCK;
    rcvr->FreeDescCount -= EP_NUM_RXD_PER_BLOCK;

    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);

    ep_shared_free_main (rcvr->Subsys->Subsys.Sys, &blk->NmdMain);
    KMEM_FREE (blk, sizeof (EP_RXD_BLOCK));
}

static EP_RXD *
GetRxd (EP_RCVR *rcvr, EP_ATTRIBUTE attr)
{
    EP_RXD *rxd;
    unsigned long flags;
    int low_on_rxds;

    spin_lock_irqsave (&rcvr->FreeDescLock, flags);

    while (list_empty (&rcvr->FreeDescList))
    {
	if (! (attr & EP_NO_ALLOC))
	{
	    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);

	    if (AllocateRxdBlock (rcvr, attr, &rxd) == ESUCCESS)
		return (rxd);

	    spin_lock_irqsave (&rcvr->FreeDescLock, flags);
	}

	if (attr & EP_NO_SLEEP)
	{
	    IncrStat (rcvr->Subsys, NoFreeRxds);
	    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);

	    ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
	    return (NULL);
	}

	rcvr->FreeDescWanted++;
	kcondvar_wait (&rcvr->FreeDescSleep, &rcvr->FreeDescLock, &flags);
    }

    rxd = list_entry (rcvr->FreeDescList.next, EP_RXD, Link);

    list_del (&rxd->Link);

    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_rxds = (--rcvr->FreeDescCount < ep_rxd_lowat);

    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);

    if (low_on_rxds)
	ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);

    return (rxd);
}

static void
FreeRxd (EP_RCVR *rcvr, EP_RXD *rxd)
{
    unsigned long flags;

    ASSERT (EP_XID_INVALID(rxd->MsgXid));

    spin_lock_irqsave (&rcvr->FreeDescLock, flags);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    ASSERT(list_empty(&rxd->CheckSumLink));
#endif
   
    list_add (&rxd->Link, &rcvr->FreeDescList);

    rcvr->FreeDescCount++;

    if (rcvr->FreeDescWanted)					/* someone waiting for a receive */
    {								/* descriptor, so wake them up */
	rcvr->FreeDescWanted--;
	kcondvar_wakeupone (&rcvr->FreeDescSleep, &rcvr->FreeDescLock);
    }
    
    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);
}

int
ep_queue_receive (EP_RCVR *rcvr, EP_RXH *handler, void *arg, EP_NMD *nmd, EP_ATTRIBUTE attr)
{
    EP_RCVR_RAIL *rcvrRail;
    EP_RXD       *rxd;
    int           rnum;
    unsigned long flags;

    if ((rxd = GetRxd (rcvr, attr)) == NULL)
	return (ENOMEM);

    rxd->Handler      = handler;
    rxd->Arg          = arg;
    rxd->Data         = *nmd;
    rxd->RxdMain->Len = EP_RXD_PENDING;
    
    spin_lock_irqsave (&rcvr->Lock, flags);

    list_add_tail (&rxd->Link, &rcvr->ActiveDescList);
    
    if (EP_IS_PREFRAIL_SET(attr))
	rnum = EP_ATTR2PREFRAIL(attr);
    else 
	rnum = ep_rcvr_prefrail (rcvr, EP_NMD_RAILMASK(nmd));

    if (rnum < 0 || !(EP_NMD_RAILMASK(nmd) & EP_RAIL2RAILMASK(rnum) & rcvr->RailMask))
	rcvrRail = NULL;
    else
	rcvrRail = rcvr->Rails[rnum];

    EPRINTF7 (DBG_RCVR,"ep_queue_receive: rxd=%p svc %d nmd=%08x,%d,%x rnum=%d rcvrRail=%p\n",
	      rxd, rcvr->Service, nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr, rnum, rcvrRail);

    rxd->State = EP_RXD_RECEIVE_ACTIVE;

    if (rcvrRail == NULL || !EP_RCVR_OP (rcvrRail, QueueRxd) (rxd, rcvrRail))
    {
	rxd->State = EP_RXD_RECEIVE_UNBOUND;

	ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
    }

    spin_unlock_irqrestore (&rcvr->Lock, flags);

    return (ESUCCESS);
}

void
ep_requeue_receive (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *nmd, EP_ATTRIBUTE attr)
{
    EP_RCVR      *rcvr = rxd->Rcvr;
    EP_SYS       *sys  = rcvr->Subsys->Subsys.Sys;
    int           rnum = ep_pickRail(EP_NMD_RAILMASK(&rxd->Data));
    EP_RCVR_RAIL *rcvrRail;
    unsigned long flags;

    ASSERT (rxd->RxdRail == NULL);

    EPRINTF5 (DBG_RCVR,"ep_requeue_receive: rxd=%p svc %d nmd=%08x,%d,%x\n", 
	      rxd, rcvr->Service, nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr);

    rxd->Handler      = handler;
    rxd->Arg          = arg;
    rxd->Data         = *nmd;
    rxd->RxdMain->Len = EP_RXD_PENDING;
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    
    list_add_tail (&rxd->Link, &rcvr->ActiveDescList);

    /*
     * Rail selection: if they've asked for a particular rail, then use it, otherwise if
     *                 the rail it was last received on is mapped for the nmd and is available
     *                 then use that one, otherwise pick one that is mapped by the nmd.
     */
    if (EP_IS_PREFRAIL_SET(attr))
	rnum = EP_ATTR2PREFRAIL(attr);
    
    if (rnum < 0 || ! (EP_RAIL2RAILMASK (rnum) & EP_NMD_RAILMASK(nmd) & ep_rcvr_availrails (rcvr)))
	rnum = ep_rcvr_prefrail (rcvr, EP_NMD_RAILMASK(nmd));

    if (rnum < 0)
	rcvrRail = NULL;
    else
    {
	rcvrRail = rcvr->Rails[rnum];

	if (! (EP_NMD_RAILMASK(&rxd->Data) & EP_RAIL2RAILMASK(rnum)) && ep_nmd_map_rails (sys, &rxd->Data, EP_RAIL2RAILMASK(rnum)) < 0)
	    rcvrRail = NULL;
    }

    rxd->State = EP_RXD_RECEIVE_ACTIVE;

    if (rcvrRail == NULL || !EP_RCVR_OP(rcvrRail, QueueRxd) (rxd, rcvrRail))
    {
	EPRINTF1 (DBG_RCVR, "ep_requeue_receive: rcvrRail=%p - setting unbound\n", rcvrRail);

	rxd->State = EP_RXD_RECEIVE_UNBOUND;

	ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
    }

    if (rcvr->CleanupWaiting)
	kcondvar_wakeupall (&rcvr->CleanupSleep, &rcvr->Lock);
    rcvr->CleanupWaiting = 0;

    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void

ep_complete_receive (EP_RXD *rxd)
{
    EP_RCVR *rcvr = rxd->Rcvr;
    unsigned long flags;

    ASSERT (rxd->RxdRail == NULL && rxd->State == EP_RXD_COMPLETED);

    FreeRxd (rcvr, rxd);

    /* if we're waiting for cleanup, then wake them up */
    spin_lock_irqsave (&rcvr->Lock, flags);
    if (rcvr->CleanupWaiting)
	kcondvar_wakeupall (&rcvr->CleanupSleep, &rcvr->Lock);
    rcvr->CleanupWaiting = 0;
    spin_unlock_irqrestore (&rcvr->Lock, flags);   
}

int
ep_rpc_put (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *local, EP_NMD *remote, int nFrags)
{
    EP_RCVR      *rcvr = rxd->Rcvr;
    EP_SYS       *sys  = rcvr->Subsys->Subsys.Sys;
    EP_ENVELOPE  *env  = &rxd->RxdMain->Envelope;
    unsigned long flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    
    if (rxd->State == EP_RXD_BEEN_ABORTED)
    {
	EPRINTF2 (DBG_RCVR, "ep_rpc_put: rcvr %p rxd %p completed because no rails available\n", rcvr, rxd);
	
	/* rxd no longer on active list - just free it */
	/* off and return an error */
	spin_unlock_irqrestore (&rcvr->Lock, flags);
	
	return EP_CONN_RESET;
    }
    else
    {
	EP_RXD_RAIL      *rxdRail   = rxd->RxdRail;
	EP_RCVR_RAIL     *rcvrRail  = rxdRail->RcvrRail;
	EP_COMMS_RAIL    *commsRail = rcvrRail->CommsRail;
	EP_RAIL          *rail      = commsRail->Rail;
	EP_NODE_RAIL     *nodeRail  = &rail->Nodes[env->NodeId];
	int               i;
	
	/* Attempt to ensure that the local nmds are mapped */
	for (i = 0; i < nFrags; i++)
	    if (! (EP_NMD_RAILMASK(&local[i]) & EP_RAIL2RAILMASK(rail->Number)))
		ep_nmd_map_rails (sys, &local[i], EP_RAIL2RAILMASK(rail->Number));
    
	if (nodeRail->State == EP_NODE_CONNECTED &&									/* rail is connected */
	    (ep_nmd2railmask (local, nFrags) & ep_nmd2railmask (remote, nFrags) & EP_RAIL2RAILMASK (rail->Number))) 	/* and NMDs valid for it */
	{
	    rxd->State = EP_RXD_PUT_ACTIVE;

	    EP_RCVR_OP(rcvrRail, RpcPut) (rxd, local, remote, nFrags);
	}
	else
	{
	    /* RPC completion cannot progress - either node is no longer connected on this 
	     * rail or some of the source/destination NMDs are not mapped on this rail.
	     * Save the NMDs into the RXD and schedule the thread to request mappings */
	    EPRINTF4 (DBG_RCVR, "%s: ep_rpc_put: rcvr %p rxd %p %s\n", rail->Name, rcvr, rxd,
		      (nodeRail->State == EP_NODE_CONNECTED) ? "NMDs not valid on this rail" : "no longer connected on this rail");

	    rxd->State = EP_RXD_PUT_STALLED;

	    if (nodeRail->State == EP_NODE_CONNECTED)
		ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
	}

	/* install the handler */
	rxd->Handler = handler;
	rxd->Arg     = arg;
	
	/* store the arguements */
	rxd->nFrags = nFrags;
	for (i = 0; i < nFrags; i++)
	{
	    rxd->Local[i]  = local[i];
	    rxd->Remote[i] = remote[i];
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    return EP_SUCCESS;
}

int
ep_rpc_get (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *remote, EP_NMD *local, int nFrags)
{
    EP_RCVR      *rcvr = rxd->Rcvr;
    EP_SYS       *sys  = rcvr->Subsys->Subsys.Sys;
    EP_ENVELOPE  *env  = &rxd->RxdMain->Envelope;
    unsigned long flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    
    if (rxd->State == EP_RXD_BEEN_ABORTED)
    {
	EPRINTF2 (DBG_RCVR, "ep_rpc_get: rcvr %p rxd %p completed because no rails available\n", rcvr, rxd);
	
	spin_unlock_irqrestore (&rcvr->Lock, flags);

	return EP_CONN_RESET;
    }
    else
    {
	EP_RXD_RAIL      *rxdRail   = rxd->RxdRail;
	EP_RCVR_RAIL     *rcvrRail  = rxdRail->RcvrRail;
	EP_COMMS_RAIL    *commsRail = rcvrRail->CommsRail;
	EP_RAIL          *rail      = commsRail->Rail;
	EP_NODE_RAIL     *nodeRail  = &rail->Nodes[env->NodeId];
	int               i;
	
	/* Attempt to ensure that the local nmds are mapped */
	for (i = 0; i < nFrags; i++)
	    if (! (EP_NMD_RAILMASK(&local[i]) & EP_RAIL2RAILMASK(rail->Number)))
		ep_nmd_map_rails (sys, &local[i], EP_RAIL2RAILMASK(rail->Number));

	if (nodeRail->State == EP_NODE_CONNECTED &&									/* rail is connected */
	    (ep_nmd2railmask (local, nFrags) & ep_nmd2railmask (remote, nFrags) & EP_RAIL2RAILMASK (rail->Number))) 	/* and NMDs valid for it */
	{
	    rxd->State = EP_RXD_GET_ACTIVE;

	    EP_RCVR_OP (rcvrRail, RpcGet) (rxd, local, remote, nFrags);
	}
	else
	{
	    /* RPC completion cannot progress - either node is no longer connected on this 
	     * node or some of the source/destination NMDs are not mapped on this rail.
	     * Save the NMDs into the RXD and schedule the thread to request mappings */
	    EPRINTF4 (DBG_RCVR, "%s: ep_rpc_get: rcvr %p rxd %p %s\n", rail->Name, rcvr, rxd, 
		      (nodeRail->State == EP_NODE_CONNECTED) ? "NMDs not valid on this rail" : "no longer connected on this rail");
	    
	    rxd->State = EP_RXD_GET_STALLED;

	    if (nodeRail->State == EP_NODE_CONNECTED)
		ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
	}

	/* install the handler */
	rxd->Handler = handler;
	rxd->Arg     = arg;
	
	/* store the arguements */
	rxd->nFrags = nFrags;
	for (i = 0; i < nFrags; i++)
	{
	    rxd->Local[i]  = local[i];
	    rxd->Remote[i] = remote[i];
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
    
    return EP_SUCCESS;
}

int
ep_complete_rpc (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_STATUSBLK *blk, EP_NMD *local, EP_NMD *remote, int nFrags)
{
    EP_RCVR      *rcvr = rxd->Rcvr;
    EP_SYS       *sys  = rcvr->Subsys->Subsys.Sys;
    EP_ENVELOPE  *env  = &rxd->RxdMain->Envelope;
    unsigned long flags;

    spin_lock_irqsave (&rcvr->Lock, flags);

    if (rxd->State == EP_RXD_BEEN_ABORTED)
    {
	EPRINTF2 (DBG_RCVR, "ep_complete_rpc: rcvr %p rxd %p completed because no rails available\n", rcvr, rxd);
	
	spin_unlock_irqrestore (&rcvr->Lock, flags);
	return EP_CONN_RESET;
    }
    else
    {
	EP_RXD_RAIL      *rxdRail   = rxd->RxdRail;
	EP_RCVR_RAIL     *rcvrRail  = rxdRail->RcvrRail;
	EP_COMMS_RAIL    *commsRail = rcvrRail->CommsRail;
	EP_RAIL          *rail      = commsRail->Rail;
	EP_NODE_RAIL     *nodeRail  = &rail->Nodes[env->NodeId];
	int               i;

	if (blk == NULL)
	    bzero (&rxd->RxdMain->StatusBlk, sizeof (EP_STATUSBLK));
	else
	    bcopy (blk, &rxd->RxdMain->StatusBlk, sizeof (EP_STATUSBLK));

	/* Attempt to ensure that the local nmds are mapped */
	for (i = 0; i < nFrags; i++)
	    if (! (EP_NMD_RAILMASK(&local[i]) & EP_RAIL2RAILMASK(rail->Number)))
		ep_nmd_map_rails (sys, &local[i], EP_RAIL2RAILMASK(rail->Number));

	if (nodeRail->State == EP_NODE_CONNECTED &&									/* rail is connected */
	    (ep_nmd2railmask (local, nFrags) & ep_nmd2railmask (remote, nFrags) & EP_RAIL2RAILMASK (rail->Number))) 	/* and NMDs valid for it */
	{
	    rxd->State = EP_RXD_COMPLETE_ACTIVE;

	    EP_RCVR_OP (rcvrRail, RpcComplete) (rxd, local, remote, nFrags);
	}
	else
	{
	    /* RPC completion cannot progress - either node is no longer connected on this 
	     * node or some of the source/destination NMDs are not mapped on this rail.
	     * Save the NMDs into the RXD and schedule the thread to request mappings */
	    EPRINTF4 (DBG_RCVR, "%s: ep_complete_rpc: rcvr %p rxd %p %s\n", rail->Name, rcvr, rxd, 
		      (nodeRail->State == EP_NODE_CONNECTED) ? "NMDs not valid on this rail" : "no longer connected on this rail");

	    rxd->State = EP_RXD_COMPLETE_STALLED;

	    if (nodeRail->State == EP_NODE_CONNECTED)
		ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
	}

	/* install the handler */
	rxd->Handler = handler;
	rxd->Arg     = arg;
	
	/* store the arguements */
	rxd->nFrags = nFrags;
	for (i = 0; i < nFrags; i++)
	{
	    rxd->Local[i]  = local[i];
	    rxd->Remote[i] = remote[i];
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
	
    return (ESUCCESS);
}

/* functions for accessing fields of rxds */
void            *ep_rxd_arg(EP_RXD *rxd)		{ return (rxd->Arg); }
int              ep_rxd_len(EP_RXD *rxd) 		{ return (rxd->RxdMain->Len); }
EP_STATUS	 ep_rxd_status(EP_RXD *rxd) 		{ return (rxd->RxdMain->Len < 0 ? rxd->RxdMain->Len : EP_SUCCESS); }
int              ep_rxd_isrpc(EP_RXD *rxd) 		{ return (EP_IS_RPC(rxd->RxdMain->Envelope.Attr) != 0); }
EP_ENVELOPE     *ep_rxd_envelope(EP_RXD *rxd) 		{ return (&rxd->RxdMain->Envelope); }
EP_PAYLOAD      *ep_rxd_payload(EP_RXD *rxd) 		{ return (EP_HAS_PAYLOAD(rxd->RxdMain->Envelope.Attr) ? &rxd->RxdMain->Payload : NULL); }
int              ep_rxd_node(EP_RXD *rxd) 		{ return (rxd->RxdMain->Envelope.NodeId); }
EP_STATUSBLK    *ep_rxd_statusblk(EP_RXD *rxd) 		{ return (&rxd->RxdMain->StatusBlk); }
EP_RAILMASK      ep_rxd_railmask(EP_RXD *rxd)		{ return (rxd->Data.nmd_attr); }

static void
ProcessNmdMapResponse (EP_RCVR *rcvr, EP_RXD *rxd, EP_MANAGER_MSG *msg)
{
    EP_RXD_RAIL  *rxdRail  = rxd->RxdRail;
    EP_RCVR_RAIL *rcvrRail = rxdRail->RcvrRail;
    EP_RAIL      *rail     = rcvrRail->CommsRail->Rail;
    EP_NODE_RAIL *nodeRail = &rail->Nodes[rxd->RxdMain->Envelope.NodeId];
    int           i;

    ASSERT (msg->Body.MapNmd.nFrags == rxd->nFrags);
    
    for (i = 0; i < rxd->nFrags; i++)
	rxd->Remote[i] = msg->Body.MapNmd.Nmd[i];
    
    if (nodeRail->State == EP_NODE_CONNECTED &&	/* node is still connected on this rail */
	(ep_nmd2railmask (rxd->Local, rxd->nFrags) & ep_nmd2railmask (rxd->Remote, rxd->nFrags) & EP_RAIL2RAILMASK (rail->Number))) 	/* NMDs are now valid for this rail */
    {
	switch (rxd->State)
	{
	case EP_RXD_PUT_STALLED:
	    rxd->State = EP_RXD_PUT_ACTIVE;

	    EP_RCVR_OP(rcvrRail, RpcPut) (rxd, rxd->Local, rxd->Remote, rxd->nFrags);
	    break;

	case EP_RXD_GET_STALLED:
	    rxd->State = EP_RXD_GET_ACTIVE;

	    EP_RCVR_OP(rcvrRail, RpcGet) (rxd, rxd->Local, rxd->Remote, rxd->nFrags);
	    break;
	    
	case EP_RXD_COMPLETE_STALLED:
	    rxd->State = EP_RXD_COMPLETE_ACTIVE;

	    EP_RCVR_OP(rcvrRail, RpcComplete) (rxd, rxd->Local, rxd->Remote, rxd->nFrags);
	    break;

	default:
	    panic ("ProcessNmdMapResponse: XID match but rxd in invalid state\n");
	    break;
	}

	rxd->NextRunTime = 0;
    }
    else
	ep_debugf (DBG_MANAGER, "%s: ep_rcvr_xid_msg_handler: rcvr=%p rxd=%p - still cannot proceed\n", rail->Name, rcvr, rxd);
}

static void
ProcessFailoverResponse (EP_RCVR *rcvr, EP_RXD *rxd, EP_MANAGER_MSG *msg)
{
    /* XXXX - TBD */
#ifdef NOTYET
    EP_COMMS_SUBSYS *subsys   = rcvr->Subsys;
    EP_RXD_RAIL     *rxdRail  = rxd->RxdRail;
    EP_RCVR_RAIL    *rcvrRail = rxdRail->RcvrRail;
    EP_RAIL         *rail     = rcvrRail->CommsRail->Rail;
    EP_RCVR_RAIL    *nRcvrRail;
    EP_RXD_RAIL     *nRxdRail;

    ASSERT (rxd->RxdMain->Envelope.Attr & EP_RPC);

    EPRINTF6 (DBG_RCVR, "ep_rcvr_xid_msg_handler: rcvr=%p rxd=%p Xid=%016llx state %x.%x - txd on rail %d\n", rcvr, rxd, 
	      rxd->MsgXid.Unique, rxdRail->RxdMain->DataEvent, rxdRail->RxdMain->DoneEvent, msg->Body.FailoverTxd.Rail);

    if ((nRcvrRail = rcvr->Rails[msg->Body.FailoverTxd.Rail]) == NULL ||
	(nRcvrRail->Rcvr->RailMask & EP_RAIL2RAILMASK (rail->Number)) == NULL)
    {
	ep_debugf (DBG_MANAGER, "%s: ep_rcvr_xid_msg_handler: rcvr=%p rxd=%p - still cannot proceed\n", rail->Name, rcvr,rxd);
	return;
    }


    nRxdRail = EP_RCVR_OP (nrcvrRail, GetRxd) (rcvr, nRcvrRail);


    /* If the RPC was in progress, then rollback and mark it as flagged, 
     * this will then get treated as though the NMDs were not mapped
     * for the rail when the user initiated the operation.
     */
    switch (rxdRail->RxdMain->DataEvent)
    {
    case EP_EVENT_ACTIVE|EP_RXD_PHASE_PUT:
    case EP_EVENT_FLAGGED|EP_RXD_PHASE_PUT:
	ASSERT (rxdRail->RxdMain->DoneEvent == EP_EVENT_PRIVATE ||
		rxdRail->RxdMain->DoneEvent == EP_EVENT_PENDING);
	
	nRxdRail->RxdMain->DataEvent = EP_EVENT_FLAGGED|EP_RXD_PHASE_PUT;
	nRxdRail->RxdMain->DoneEvent = EP_EVENT_PENDING;
	break;

    case EP_EVENT_ACTIVE|EP_RXD_PHASE_GET:
    case EP_EVENT_FLAGGED|EP_RXD_PHASE_GET:
	ASSERT (rxdRail->RxdMain->DoneEvent == EP_EVENT_PRIVATE ||
		rxdRail->RxdMain->DoneEvent == EP_EVENT_PENDING);
	
	nRxdRail->RxdMain->DataEvent = EP_EVENT_FLAGGED|EP_RXD_PHASE_GET;
	nRxdRail->RxdMain->DoneEvent = EP_EVENT_PENDING;
	break;

    case EP_EVENT_PRIVATE:
	switch (rxdRail->RxdMain->DoneEvent)
	{
	case EP_EVENT_ACTIVE|EP_RXD_PHASE_COMPLETE:
	case EP_EVENT_FLAGGED|EP_RXD_PHASE_COMPLETE:
	    nRxdRail->RxdMain->DataEvent = EP_EVENT_PRIVATE;
	    nRxdRail->RxdMain->DoneEvent = EP_EVENT_FLAGGED|EP_RXD_PHASE_COMPLETE;
	    break;

	case EP_EVENT_PENDING:
	    break;

	default:
	    panic ("ep_rcvr_xid_msg_handler: rxd in invalid state\n");
	}
	break;

    default:
	panic ("ep_rcvr_xid_msg_handler: rxd in invalid staten");
    }
    
    UnbindRxdFromRail (rxd, rxdRail);

    /* Mark rxdRail as no longer active */
    rxdRail->RxdMain->DataEvent = EP_EVENT_PRIVATE;
    rxdRail->RxdMain->DoneEvent = EP_EVENT_PRIVATE;

    sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);
    sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP_RXD_RAIL_ELAN, DoneEvent.ev_Count), 0);
    
    FreeRxdRail (rcvrRail, rxdRail);

    BindRxdToRail (rxd, nRxdRail);

    ep_kthread_schedule (&subsys->Thread, lbolt);
#endif
}

void
ep_rcvr_xid_msg_handler (void *arg, EP_MANAGER_MSG *msg)
{
    EP_RCVR          *rcvr = (EP_RCVR *) arg;
    struct list_head *el;
    unsigned long     flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    list_for_each (el, &rcvr->ActiveDescList) {
	EP_RXD *rxd = list_entry (el,EP_RXD, Link);

	if (EP_XIDS_MATCH (msg->Hdr.Xid, rxd->MsgXid))
	{
	    EP_INVALIDATE_XID (rxd->MsgXid);

	    switch (msg->Hdr.Type)
	    {
	    case EP_MANAGER_MSG_TYPE_MAP_NMD_RESPONSE:
		ProcessNmdMapResponse (rcvr, rxd, msg);
		break;

	    case EP_MANAGER_MSG_TYPE_FAILOVER_RESPONSE:
		ProcessFailoverResponse (rcvr, rxd, msg);
		break;

	    default:
		panic ("ep_rcvr_xid_msg_handler: XID match but invalid message type\n");
	    }
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}


EP_RCVR *
ep_alloc_rcvr (EP_SYS *sys, EP_SERVICE svc, unsigned int nenvs)
{
    EP_COMMS_SUBSYS  *subsys;
    EP_RCVR          *rcvr;
    struct list_head *el;
    extern int portals_envelopes;

    if (portals_envelopes && (svc == EP_MSG_SVC_PORTALS_SMALL || svc == EP_MSG_SVC_PORTALS_LARGE))
    {
	printk ("ep: use %d envelopes rather than %d for portals %s message service\n", sys->Position.pos_nodes * 16, nenvs,
		svc == EP_MSG_SVC_PORTALS_SMALL ? "small" : "large");

	nenvs = portals_envelopes;
    }

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (sys, EPCOMMS_SUBSYS_NAME)) == NULL)
	return (NULL);

    KMEM_ZALLOC (rcvr, EP_RCVR *, sizeof (EP_RCVR), 1);

    if (rcvr == NULL)
	return (NULL);

    rcvr->Subsys            = subsys;
    rcvr->Service           = svc;
    rcvr->InputQueueEntries = nenvs;
    rcvr->FreeDescCount     = 0;
    rcvr->TotalDescCount    = 0;
    rcvr->ForwardRxdCount   = 0;

    spin_lock_init (&rcvr->Lock);
    INIT_LIST_HEAD (&rcvr->ActiveDescList);

    kcondvar_init (&rcvr->CleanupSleep);
    kcondvar_init (&rcvr->FreeDescSleep);
    spin_lock_init (&rcvr->FreeDescLock);
    INIT_LIST_HEAD (&rcvr->FreeDescList);
    INIT_LIST_HEAD (&rcvr->DescBlockList);

    ep_xid_cache_init (sys, &rcvr->XidCache);

    rcvr->XidCache.MessageHandler = ep_rcvr_xid_msg_handler;
    rcvr->XidCache.Arg            = rcvr;

    kmutex_lock (&subsys->Lock);
    /* See if this service is already in use */
    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);

	if (rcvr->Service == svc)
	{
	    KMEM_FREE (rcvr, sizeof (EP_RCVR));
	    kmutex_unlock (&subsys->Lock);   
	    return NULL;
	}
    }
    
    
    list_add_tail (&rcvr->Link, &subsys->Receivers);

    ep_procfs_rcvr_add(rcvr);

    /* Now add all rails which are already started */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	EP_RAIL_OP (commsRail, Rcvr.AddRail) (rcvr, commsRail);
    }
    kmutex_unlock (&subsys->Lock);   

    ep_mod_inc_usecount();

    return (rcvr);
}

void
ep_free_rcvr (EP_RCVR *rcvr)
{
    EP_COMMS_SUBSYS  *subsys = rcvr->Subsys;
    EP_SYS           *sys    = subsys->Subsys.Sys;
    struct list_head  list;
    struct list_head *el,*nel;
    unsigned long flags;
    
    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);
	
	EP_RAIL_OP (commsRail, Rcvr.DelRail) (rcvr, commsRail);
    }

    ep_procfs_rcvr_del(rcvr);

    list_del (&rcvr->Link);
    kmutex_unlock (&subsys->Lock);

    INIT_LIST_HEAD (&list);

    /* abort all rxds - should not be bound to a rail */
    spin_lock_irqsave (&rcvr->Lock, flags);   
    for (;;)
    {
	if (! list_empty (&rcvr->ActiveDescList))
	{
	    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
		EP_RXD *rxd = list_entry (el, EP_RXD, Link);
		
		ASSERT (rxd->RxdRail == NULL);
		ASSERT (rxd->RxdMain->Len == EP_RXD_PENDING);
		
		rxd->State = EP_RXD_COMPLETED;
		rxd->RxdMain->Len = EP_SHUTDOWN;
		
		list_del (&rxd->Link);
		list_add_tail (&rxd->Link, &list);
	    }
	    spin_unlock_irqrestore (&rcvr->Lock, flags);
	    
	    while (! list_empty (&list))
	    {
		EP_RXD *rxd = list_entry (list.next, EP_RXD, Link);
		
		list_del (&rxd->Link);
		
		if (rxd->Handler) 
		    rxd->Handler (rxd);
	    }
	    spin_lock_irqsave (&rcvr->Lock, flags);   
	    continue;
	}

	if (rcvr->FreeDescCount == rcvr->TotalDescCount)
	    break;

	rcvr->CleanupWaiting++;
	kcondvar_wait (&rcvr->CleanupSleep, &rcvr->Lock, &flags);
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    /* must all be in free list */
    ASSERT( rcvr->FreeDescCount ==  rcvr->TotalDescCount);

    while (! list_empty(& rcvr->DescBlockList) )
	FreeRxdBlock (rcvr, list_entry (rcvr->DescBlockList.next, EP_RXD_BLOCK, Link));

    /* had better be all gone now */
    ASSERT((rcvr->FreeDescCount == 0) && (rcvr->TotalDescCount == 0));

    ep_xid_cache_destroy (sys, &rcvr->XidCache);
 
    spin_lock_destroy (&rcvr->Lock);
    KMEM_FREE (rcvr, sizeof (EP_RCVR));

    ep_mod_dec_usecount();
}

EP_RXD *
StealRxdFromOtherRail (EP_RCVR *rcvr)
{
    EP_RXD	     *rxd;
    int               i;
	
    /* looking at the the rcvr railmask to find a rail to try to steal rxd from */
    for (i = 0; i < EP_MAX_RAILS; i++) 
	if (rcvr->RailMask & (1 << i) ) 
	    if ((rxd = EP_RCVR_OP (rcvr->Rails[i], StealRxd) (rcvr->Rails[i])) != NULL)
		return rxd;

    return NULL;
}

long
CheckUnboundRxd (EP_RCVR *rcvr, EP_RXD *rxd, long nextRunTime)
{
    EP_SYS       *sys = rcvr->Subsys->Subsys.Sys;
    EP_RCVR_RAIL *rcvrRail;
    int           rnum;
    
    if ((rnum = ep_rcvr_prefrail (rcvr, EP_NMD_RAILMASK(&rxd->Data))) < 0)
	rnum = ep_rcvr_prefrail (rcvr, ep_rcvr_availrails (rcvr));
    
    if ( rnum < 0 ) 	{
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
	
	return (nextRunTime);
    }

    ASSERT ( rnum >= 0 );

    rcvrRail = rcvr->Rails[rnum];

    ASSERT ( rcvrRail != NULL);

    rxd->State = EP_RXD_RECEIVE_ACTIVE;

    if ((!(EP_NMD_RAILMASK (&rxd->Data) & EP_RAIL2RAILMASK(rnum)) &&			/* not mapped already and */
	 ep_nmd_map_rails (sys, &rxd->Data, EP_RAIL2RAILMASK(rnum)) == 0) ||		/* failed mapping, or */
	!EP_RCVR_OP (rcvrRail, QueueRxd) (rxd, rcvrRail))				/* failed to queue */
    {
	ASSERT (rxd->RxdRail == NULL);

	EPRINTF4 (DBG_RCVR,"CheckUnboundRxd: rcvr=%p rxd=%p -> rnum=%d rcvrRail=%p (failed)\n", rcvr, rxd, rnum, rcvrRail);

	rxd->State = EP_RXD_RECEIVE_UNBOUND;
	
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }

    return (nextRunTime);
}

int
CheckRxdNmdsMapped (EP_RCVR *rcvr, EP_RXD *rxd)
{
    EP_RXD_RAIL *rxdRail = rxd->RxdRail;
    EP_RXD_MAIN *rxdMain = rxd->RxdMain;
    EP_ENVELOPE *env     = &rxdMain->Envelope;
    EP_SYS      *sys     = rcvr->Subsys->Subsys.Sys;
    EP_RAIL     *rail    = rxdRail->RcvrRail->CommsRail->Rail;
    int		 i;

    /* Try and map the local NMDs before checking to see if we can proceed */
    if (! (ep_nmd2railmask (rxd->Local, rxd->nFrags) & EP_RAIL2RAILMASK (rail->Number)))
    {
	EPRINTF3 (DBG_MAPNMD, "%s: rcvr=%p rxd=%p RPC Local NMDs not mapped\n", rail->Name, rcvr, rxd);
	
	for (i = 0; i < rxd->nFrags; i++)
	    if (! (EP_NMD_RAILMASK(&rxd->Local[i]) & EP_RAIL2RAILMASK(rail->Number)))
		if (ep_nmd_map_rails (sys, &rxd->Local[i], EP_RAIL2RAILMASK(rail->Number)))
		    rxd->NextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }
    
    /* Try and map remote NMDs if they are not valid for this rail */
    if (! (ep_nmd2railmask (rxd->Remote, rxd->nFrags) & EP_RAIL2RAILMASK (rail->Number)))
    {
	EP_MANAGER_MSG_BODY msgBody;

	EPRINTF3 (DBG_MAPNMD, "%s: rcvr=%p rxd=%p RPC Remote NMDs not mapped\n", rail->Name, rcvr, rxd);

	if (EP_XID_INVALID(rxd->MsgXid))
	    rxd->MsgXid = ep_xid_cache_alloc (sys, &rcvr->XidCache);

	msgBody.MapNmd.nFrags   = rxd->nFrags;
	msgBody.MapNmd.Railmask = EP_RAIL2RAILMASK (rail->Number);
	for (i = 0; i < rxd->nFrags; i++)
	    msgBody.MapNmd.Nmd[i] = rxd->Remote[i];

	if (ep_send_message (rail, env->NodeId, EP_MANAGER_MSG_TYPE_MAP_NMD_REQUEST, rxd->MsgXid, &msgBody) == 0)
	    rxd->NextRunTime = lbolt + MESSAGE_RETRY_TIME;
	else
	    rxd->NextRunTime = lbolt + MSGBUSY_RETRY_TIME;

	return 0;
    }

    if ((ep_nmd2railmask (rxd->Local, rxd->nFrags) & ep_nmd2railmask (rxd->Remote, rxd->nFrags) & EP_RAIL2RAILMASK (rail->Number)) != 0)
    {
	rxd->NextRunTime = 0;
	return 1;
    }

    return 0;
}

long
ep_check_rcvr (EP_RCVR *rcvr, long nextRunTime)
{
    struct list_head *el, *nel;
    unsigned long     flags;
    int               i;

    /* Check to see if we're low on rxds */
    if (rcvr->FreeDescCount < ep_rxd_lowat)
	AllocateRxdBlock (rcvr, 0, NULL);

    for (i = 0; i < EP_MAX_RAILS; i++) 
	if (rcvr->RailMask & (1 << i) )
	    nextRunTime = EP_RCVR_OP (rcvr->Rails[i], Check) (rcvr->Rails[i], nextRunTime);

    /* See if we have any rxd's which need to be handled */
    spin_lock_irqsave (&rcvr->Lock, flags);
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD      *rxd     = list_entry (el, EP_RXD, Link);
	EP_RXD_MAIN *rxdMain = rxd->RxdMain;
	EP_ENVELOPE *env     = &rxdMain->Envelope;
	EP_RXD_RAIL *rxdRail = rxd->RxdRail;

	if (rxdRail == NULL)
	    nextRunTime = CheckUnboundRxd (rcvr, rxd, nextRunTime);
	else
	{
	    EP_RCVR_RAIL *rcvrRail = rxdRail->RcvrRail;
	    EP_RAIL      *rail     = rcvrRail->CommsRail->Rail;

	    if (rxd->RxdMain->Len == EP_RXD_PENDING || 				/* envelope not received yet */
		rail->Nodes[env->NodeId].State != EP_NODE_CONNECTED) 		/* will be failing over */
		continue;

	    switch (rxd->State)
	    {
	    case EP_RXD_PUT_STALLED:
		if (CheckRxdNmdsMapped (rcvr, rxd))
		{
		    rxd->State = EP_RXD_PUT_ACTIVE;

		    EP_RCVR_OP (rcvrRail, RpcPut) (rxd, rxd->Local, rxd->Remote, rxd->nFrags);
		}
		break;

	    case EP_RXD_GET_STALLED:
		if (CheckRxdNmdsMapped (rcvr, rxd))
		{
		    rxd->State = EP_RXD_GET_ACTIVE;

		    EP_RCVR_OP (rcvrRail, RpcGet) (rxd, rxd->Local, rxd->Remote, rxd->nFrags);
		}
		break;
	    
	    case EP_RXD_COMPLETE_STALLED:
		if (CheckRxdNmdsMapped (rcvr, rxd))
		{
		    rxd->State = EP_RXD_COMPLETE_ACTIVE;

		    EP_RCVR_OP (rcvrRail, RpcComplete)(rxd, rxd->Local, rxd->Remote, rxd->nFrags);
		}
		break;
	    }
		
	    if (rxd->NextRunTime && (nextRunTime == 0 || AFTER (nextRunTime, rxd->NextRunTime)))
		nextRunTime = rxd->NextRunTime;
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
    
    return (nextRunTime);
}

void
ep_display_rxd (DisplayInfo *di, EP_RXD *rxd)
{
    EP_RXD_MAIN *rxdMain  = rxd->RxdMain;
    EP_ENVELOPE *env      = &rxdMain->Envelope;
    EP_RXD_RAIL *rxdRail  = rxd->RxdRail;
    
    (di->func)(di->arg, "  RXD: %p State=%x RxdMain=%p(%x.%x.%x) Data=%x.%x.%x %s\n", rxd,
	       rxd->State, rxd->RxdMain, rxd->NmdMain.nmd_addr, rxd->NmdMain.nmd_len,
	       rxd->NmdMain.nmd_attr, rxd->Data.nmd_addr, rxd->Data.nmd_len, rxd->Data.nmd_attr,
	       rxd->RxdMain->Len == EP_RXD_PENDING ? "Pending" : "Active");
    (di->func)(di->arg, "      NodeId=%d Range=%d.%d TxdRail=%x TxdMain=%x.%x.%x nFrags=%d XID=%08x.%08x.%016llx\n",
	       env->NodeId,  EP_RANGE_LOW(env->Range), EP_RANGE_HIGH(env->Range), env->TxdRail, env->TxdMain.nmd_addr,
	       env->TxdMain.nmd_len, env->TxdMain.nmd_attr, env->nFrags, env->Xid.Generation, env->Xid.Handle, env->Xid.Unique);;
    (di->func)(di->arg, "      Frag[0] %08x.%08x.%08x\n", env->Frags[0].nmd_addr, env->Frags[0].nmd_len, env->Frags[0].nmd_attr);
    (di->func)(di->arg, "      Frag[1] %08x.%08x.%08x\n", env->Frags[1].nmd_addr, env->Frags[1].nmd_len, env->Frags[1].nmd_attr);
    (di->func)(di->arg, "      Frag[2] %08x.%08x.%08x\n", env->Frags[2].nmd_addr, env->Frags[2].nmd_len, env->Frags[2].nmd_attr);
    (di->func)(di->arg, "      Frag[3] %08x.%08x.%08x\n", env->Frags[3].nmd_addr, env->Frags[3].nmd_len, env->Frags[3].nmd_attr);

    if (rxdRail) EP_RCVR_OP (rxdRail->RcvrRail, DisplayRxd) (di, rxdRail);
}

void
ep_display_rcvr (DisplayInfo *di, EP_RCVR *rcvr, int full)
{
    int               freeCount    = 0;
    int		      activeCount  = 0;
    int		      pendingCount = 0;
    int		      railCounts[EP_MAX_RAILS];
    struct list_head *el;
    int               i;
    unsigned long     flags;

    for (i = 0; i <EP_MAX_RAILS; i++)
	railCounts[i] = 0;

    spin_lock_irqsave (&rcvr->FreeDescLock, flags);
    list_for_each (el, &rcvr->FreeDescList)
	freeCount++;
    spin_unlock_irqrestore (&rcvr->FreeDescLock, flags);

    spin_lock_irqsave (&rcvr->Lock, flags);
    list_for_each (el, &rcvr->ActiveDescList) {
	EP_RXD      *rxd     = list_entry (el, EP_RXD, Link);
	EP_RXD_RAIL *rxdRail = rxd->RxdRail;

	if (rxd->RxdMain->Len == EP_RXD_PENDING)
	    pendingCount++;
	else
	    activeCount++;

	if (rxdRail)
	    railCounts[rxdRail->RcvrRail->CommsRail->Rail->Number]++;
    }

    (di->func)(di->arg, "RCVR: rcvr=%p number=%d\n", rcvr, rcvr->Service);
    (di->func)(di->arg, "      RXDS Free=%d (%d) Pending=%d Active=%d Rails=%d.%d.%d.%d\n",
	       freeCount, rcvr->FreeDescCount, pendingCount, activeCount, railCounts[0], railCounts[1],
	       railCounts[2], railCounts[3]);

    for (i = 0; i < EP_MAX_RAILS; i++)
	if (rcvr->Rails[i] != NULL)
	    EP_RCVR_OP (rcvr->Rails[i], DisplayRcvr) (di, rcvr->Rails[i]);

    list_for_each (el, &rcvr->ActiveDescList) {
	EP_RXD *rxd = list_entry (el, EP_RXD, Link);

	if (rxd->RxdMain->Len != EP_RXD_PENDING || full)
	    ep_display_rxd (di, rxd);
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep_rxd_received_now(EP_RXD *rxd)
{
    EP_ENVELOPE   *env  = &rxd->RxdMain->Envelope;
    EP_RCVR       *rcvr = rxd->Rcvr;
    unsigned long  flags;

    INC_STAT(rcvr->stats,rx);
    ADD_STAT(rcvr->stats,rx_len, rxd->RxdMain->Len);

    if (rxd->RxdMain->Len < 0 || !EP_IS_MULTICAST(env->Attr))
    {
	rxd->Handler (rxd);
    }
    else
    {
	EPRINTF5 (DBG_RCVR, "ep_rxd_received: forward rxd=%p Data=%08x.%08x.%08x len=%d\n", rxd, 
		  rxd->Data.nmd_addr, rxd->Data.nmd_len, rxd->Data.nmd_attr, ep_rxd_len(rxd));

	spin_lock_irqsave (&rcvr->Subsys->ForwardDescLock, flags);
	list_add_tail (&rxd->Link, &rcvr->Subsys->ForwardDescList);
	spin_unlock_irqrestore (&rcvr->Subsys->ForwardDescLock, flags);
	
	ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
    } 
}

#if defined(CONFIG_EP_NO_CHECK_SUM)
void
ep_rxd_received(EP_RXD *rxd) 
{
   ep_rxd_received_now(rxd);
}

#else

void
ep_rxd_received(EP_RXD *rxd) 
{
  EP_ENVELOPE   *env  = &rxd->RxdMain->Envelope;

  if (env->CheckSum) 
      ep_rxd_queue_csum(rxd);
  else 
      ep_rxd_received_now(rxd);
}

void
ep_rxd_queue_csum(EP_RXD *rxd)
{
    EP_RCVR       *rcvr = rxd->Rcvr;
    unsigned long flags;

    EPRINTF5 (DBG_RCVR, "ep_rxd_queue_csum: rxd=%p Data=%08x.%08x.%08x len=%d\n", rxd, 
	      rxd->Data.nmd_addr, rxd->Data.nmd_len, rxd->Data.nmd_attr, ep_rxd_len(rxd));
    
    spin_lock_irqsave (&rcvr->Subsys->CheckSumDescLock, flags);
    list_add_tail (&rxd->CheckSumLink, &rcvr->Subsys->CheckSumDescList);
    spin_unlock_irqrestore (&rcvr->Subsys->CheckSumDescLock, flags);
    
    ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
}
#endif

void
ep_rcvr_fillout_stats(EP_RCVR *rcvr, char *str)
{
    sprintf(str+strlen(str),"Rx     %lu  %lu /sec\n",       GET_STAT_TOTAL(rcvr->stats,rx),      GET_STAT_PER_SEC(rcvr->stats,rx) );
    sprintf(str+strlen(str),"MBytes %lu  %lu Mbytes/sec\n", GET_STAT_TOTAL(rcvr->stats,rx_len) / (1024*1024),  GET_STAT_PER_SEC(rcvr->stats,rx_len) / (1024*1024));
}

void
ep_rcvr_rail_fillout_stats(EP_RCVR_RAIL *rcvr_rail, char *str)
{
    sprintf(str+strlen(str),"Rx     %lu  %lu /sec\n",       GET_STAT_TOTAL(rcvr_rail->stats,rx),      GET_STAT_PER_SEC(rcvr_rail->stats,rx) );
    sprintf(str+strlen(str),"MBytes %lu  %lu Mbytes/sec\n", GET_STAT_TOTAL(rcvr_rail->stats,rx_len) / (1024*1024),  GET_STAT_PER_SEC(rcvr_rail->stats,rx_len) / (1024*1024));
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
