/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsRx_elan3.c,v 1.24 2005/03/10 15:25:52 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsRx_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "epcomms_elan3.h"
#include "debug.h"

#define RCVR_TO_RAIL(rcvrRail)		((EP3_RAIL *) ((EP_RCVR_RAIL *) rcvrRail)->CommsRail->Rail)
#define RCVR_TO_DEV(rcvrRail)		(RCVR_TO_RAIL(rcvrRail)->Device)
#define RCVR_TO_SUBSYS(rcvrRail)	(((EP_RCVR_RAIL *) rcvrRail)->Rcvr->Subsys)

static void RxDataEvent (EP3_RAIL *rail, void *arg);
static void RxDataRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status);
static void RxDataVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma);

static EP3_COOKIE_OPS RxDataCookieOps = 
{
    RxDataEvent,
    RxDataRetry,
    NULL, /* DmaCancelled */
    RxDataVerify,
};

static void RxDoneEvent (EP3_RAIL *rail, void *arg);
static void RxDoneRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status);
static void RxDoneVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma);

static EP3_COOKIE_OPS RxDoneCookieOps = 
{
    RxDoneEvent,
    RxDoneRetry,
    NULL, /* DmaCancelled */
    RxDoneVerify,
};

static int
AllocateRxdRailBlock (EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RAIL	      *rail = RCVR_TO_RAIL(rcvrRail);
    ELAN3_DEV         *dev  = rail->Device;
    EP3_RXD_RAIL_BLOCK *blk;
    EP3_RXD_RAIL       *rxdRail;
    sdramaddr_t        pRxdElan;
    EP3_RXD_RAIL_MAIN  *pRxdMain;
    E3_Addr	       pRxdElanAddr;
    E3_Addr	       pRxdMainAddr;
    E3_BlockCopyEvent  event;
    int		       i, j;
    unsigned long      flags;

    KMEM_ZALLOC (blk, EP3_RXD_RAIL_BLOCK *, sizeof (EP3_RXD_RAIL_BLOCK), 1);
    if (blk == NULL)
	return 0;

    if ((pRxdElan = ep_alloc_elan (&rail->Generic, EP3_RXD_RAIL_ELAN_SIZE * EP3_NUM_RXD_PER_BLOCK, 0, &pRxdElanAddr)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP3_RXD_RAIL_BLOCK));
	return 0;
    }

    if ((pRxdMain = ep_alloc_main (&rail->Generic, EP3_RXD_RAIL_MAIN_SIZE * EP3_NUM_RXD_PER_BLOCK, 0, &pRxdMainAddr)) == (sdramaddr_t) 0)
    {
	ep_free_elan (&rail->Generic, pRxdElanAddr, EP3_RXD_RAIL_ELAN_SIZE * EP3_NUM_RXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP3_RXD_RAIL_BLOCK));
	return 0;
    }
    
    if (ReserveDmaRetries (rail, EP3_NUM_RXD_PER_BLOCK, 0) != ESUCCESS)
    {
	ep_free_main (&rail->Generic, pRxdMainAddr, EP3_RXD_RAIL_MAIN_SIZE * EP3_NUM_RXD_PER_BLOCK);
	ep_free_elan (&rail->Generic, pRxdElanAddr, EP3_RXD_RAIL_ELAN_SIZE * EP3_NUM_RXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP3_RXD_RAIL_BLOCK));
	return 0;
    }

    for (rxdRail = &blk->Rxd[0], i = 0; i < EP3_NUM_RXD_PER_BLOCK; i++, rxdRail++)
    {
	rxdRail->Generic.RcvrRail = (EP_RCVR_RAIL *) rcvrRail;
	rxdRail->RxdElan          = pRxdElan;
	rxdRail->RxdElanAddr      = pRxdElanAddr;
	rxdRail->RxdMain          = pRxdMain;
	rxdRail->RxdMainAddr      = pRxdMainAddr;

	elan3_sdram_writel (dev, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, RxdMain),  0);
	elan3_sdram_writel (dev, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, Next),     0);
	elan3_sdram_writeq (dev, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, MainAddr), (long) rxdRail);

	for (j = 0; j < EP_MAXFRAG; j++)
	{
	    RegisterCookie (&rail->CookieTable, &rxdRail->ChainCookie[j], pRxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[j]), &RxDataCookieOps, (void *) rxdRail);

	    event.ev_Type  = EV_TYPE_DMA | (pRxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, Dmas[j+1]));
	    event.ev_Count = 0;

	    elan3_sdram_copyl_to_sdram (dev, &event, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[j]), sizeof (E3_BlockCopyEvent));
	}
	
	RegisterCookie (&rail->CookieTable, &rxdRail->DataCookie, pRxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, DataEvent), &RxDataCookieOps, (void *) rxdRail);
	RegisterCookie (&rail->CookieTable, &rxdRail->DoneCookie, pRxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent), &RxDoneCookieOps, (void *) rxdRail);

	EP3_INIT_COPY_EVENT (event, rxdRail->DataCookie, pRxdMainAddr + offsetof (EP3_RXD_RAIL_MAIN, DataEvent), 1);
	elan3_sdram_copyl_to_sdram (dev, &event, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent), sizeof (E3_BlockCopyEvent));

	EP3_INIT_COPY_EVENT (event, rxdRail->DoneCookie, pRxdMainAddr + offsetof (EP3_RXD_RAIL_MAIN, DoneEvent), 1);
	elan3_sdram_copyl_to_sdram (dev, &event, pRxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent), sizeof (E3_BlockCopyEvent));
	
	pRxdMain->DataEvent = EP3_EVENT_FREE;
	pRxdMain->DoneEvent = EP3_EVENT_FREE;

	/* move onto next descriptor */
	pRxdElan     += EP3_RXD_RAIL_ELAN_SIZE;
	pRxdElanAddr += EP3_RXD_RAIL_ELAN_SIZE;
	pRxdMain      = (EP3_RXD_RAIL_MAIN *) ((unsigned long) pRxdMain + EP3_RXD_RAIL_MAIN_SIZE);
	pRxdMainAddr += EP3_RXD_RAIL_MAIN_SIZE;
    }

    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);

    list_add  (&blk->Link, &rcvrRail->DescBlockList);
    rcvrRail->TotalDescCount += EP3_NUM_RXD_PER_BLOCK;
    rcvrRail->FreeDescCount  += EP3_NUM_RXD_PER_BLOCK;

    for (i = 0; i < EP3_NUM_RXD_PER_BLOCK; i++)
	list_add (&blk->Rxd[i].Generic.Link, &rcvrRail->FreeDescList);

    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);
    
    return 1;
}

static void
FreeRxdRailBlock (EP3_RCVR_RAIL *rcvrRail, EP3_RXD_RAIL_BLOCK *blk)
{
    EP3_RAIL     *rail = RCVR_TO_RAIL(rcvrRail);
    EP3_RXD_RAIL *rxdRail;
    unsigned long flags;
    int i, j;

    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);

    list_del (&blk->Link);

    rcvrRail->TotalDescCount -= EP3_NUM_RXD_PER_BLOCK;

    for (rxdRail = &blk->Rxd[0], i = 0; i < EP3_NUM_RXD_PER_BLOCK; i++, rxdRail++)
    {

	rcvrRail->FreeDescCount--;

	list_del (&rxdRail->Generic.Link);
	
	for (j = 0; j < EP_MAXFRAG; j++)
	    DeregisterCookie (&rail->CookieTable, &rxdRail->ChainCookie[j]);
	
	DeregisterCookie (&rail->CookieTable, &rxdRail->DataCookie);
	DeregisterCookie (&rail->CookieTable, &rxdRail->DoneCookie);
    }

    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);

    ReleaseDmaRetries (rail, EP3_NUM_RXD_PER_BLOCK);

    ep_free_main (&rail->Generic, blk->Rxd[0].RxdMainAddr, EP3_RXD_RAIL_MAIN_SIZE * EP3_NUM_RXD_PER_BLOCK);
    ep_free_elan (&rail->Generic, blk->Rxd[0].RxdElanAddr, EP3_RXD_RAIL_ELAN_SIZE * EP3_NUM_RXD_PER_BLOCK);

    KMEM_FREE (blk, sizeof (EP3_RXD_RAIL_BLOCK));
}

static EP3_RXD_RAIL *
GetRxdRail (EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RXD_RAIL *rxdRail;
    unsigned long flags;
    int low_on_rxds;

    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);

    if (list_empty (&rcvrRail->FreeDescList))
	rxdRail = NULL;
    else
    {
	rxdRail = list_entry (rcvrRail->FreeDescList.next, EP3_RXD_RAIL, Generic.Link);

	list_del (&rxdRail->Generic.Link);

	rcvrRail->FreeDescCount--;
    }

    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_rxds = (rcvrRail->FreeDescCount < ep_rxd_lowat);

    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);

    if (low_on_rxds)
	ep_kthread_schedule (&RCVR_TO_SUBSYS(rcvrRail)->Thread, lbolt);

    return (rxdRail);
}

static void
FreeRxdRail (EP3_RCVR_RAIL *rcvrRail, EP3_RXD_RAIL *rxdRail)
{
    unsigned long flags;

#if defined(DEBUG_ASSERT)
    {
	EP_RAIL  *rail = (EP_RAIL *) RCVR_TO_RAIL(rcvrRail);
	ELAN3_DEV *dev = RCVR_TO_DEV (rcvrRail);

	EP_ASSERT (rail, rxdRail->Generic.RcvrRail == &rcvrRail->Generic);
	
	EP_ASSERT (rail, rxdRail->RxdMain->DataEvent == EP3_EVENT_PRIVATE);
	EP_ASSERT (rail, rxdRail->RxdMain->DoneEvent == EP3_EVENT_PRIVATE);
	EP_ASSERT (rail, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));
	EP_ASSERT (rail, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));

	rxdRail->RxdMain->DataEvent = EP3_EVENT_FREE;
	rxdRail->RxdMain->DoneEvent = EP3_EVENT_FREE;
    }
#endif

    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);
    
    list_add (&rxdRail->Generic.Link, &rcvrRail->FreeDescList);

    rcvrRail->FreeDescCount++;

    if (rcvrRail->FreeDescWaiting)
    {
	rcvrRail->FreeDescWaiting--;
	kcondvar_wakeupall (&rcvrRail->FreeDescSleep, &rcvrRail->FreeDescLock);
    }

    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);
}

static void
BindRxdToRail (EP_RXD *rxd, EP3_RXD_RAIL *rxdRail)
{
    EP3_RAIL *rail = RCVR_TO_RAIL (rxdRail->Generic.RcvrRail);

    ASSERT (SPINLOCK_HELD (&rxd->Rcvr->Lock));

    EPRINTF3 (DBG_RCVR, "%s: BindRxdToRail: rxd=%p rxdRail=%p\n",  rail->Generic.Name, rxd, rxdRail);

    elan3_sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, RxdMain), rxd->NmdMain.nmd_addr);			/* PCI write */

    rxd->RxdRail         = &rxdRail->Generic;
    rxdRail->Generic.Rxd = rxd;
}

static void
UnbindRxdFromRail (EP_RXD *rxd, EP3_RXD_RAIL *rxdRail)
{
    EP3_RCVR_RAIL *rcvrRail = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;

    ASSERT (SPINLOCK_HELD (&rxd->Rcvr->Lock));
    ASSERT (rxd->RxdRail == &rxdRail->Generic && rxdRail->Generic.Rxd == rxd);
    
    EPRINTF3 (DBG_RCVR, "%s: UnbindRxdFromRail: rxd=%p rxdRail=%p\n",  RCVR_TO_RAIL(rxdRail->Generic.RcvrRail)->Generic.Name, rxd, rxdRail);

    rxd->RxdRail         = NULL;
    rxdRail->Generic.Rxd = NULL;

    if (rcvrRail->CleanupWaiting)
	kcondvar_wakeupall (&rcvrRail->CleanupSleep, &rxd->Rcvr->Lock);
    rcvrRail->CleanupWaiting = 0;
}

static void
LockRcvrThread (EP3_RCVR_RAIL *rcvrRail)
{
    EP_COMMS_RAIL     *commsRail   = rcvrRail->Generic.CommsRail;
    EP3_RAIL          *rail        = RCVR_TO_RAIL(rcvrRail);
    ELAN3_DEV	      *dev	   = rail->Device;
    sdramaddr_t        sle         = rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadLock);
    EP3_SPINLOCK_MAIN *sl          = &rcvrRail->RcvrMain->ThreadLock;
    E3_uint32          RestartBits = 0;
    int                delay       = 1;
    E3_uint32          seq;
    E3_uint32          reg;

    ASSERT (SPINLOCK_HELD (&rcvrRail->Generic.Rcvr->Lock));

    mb();
    elan3_sdram_writel (dev, sle + offsetof (EP3_SPINLOCK_ELAN, sl_lock), 1);
    mb();
    seq = elan3_sdram_readl (dev, sle + offsetof (EP3_SPINLOCK_ELAN, sl_seq));
    while (seq != sl->sl_seq)
    {
	while (sl->sl_seq == (seq - 1))
	{
	    mb();

	    if ((read_reg32 (dev, Exts.InterruptReg) & (INT_TProc | INT_TProcHalted)) != 0 && spin_trylock (&dev->IntrLock))
	    {
		reg=read_reg32 (dev, Exts.InterruptReg);
		ELAN_REG_REC(reg);

		if ((reg & (INT_TProc | INT_TProcHalted)) != 0&& 
		    elan3_sdram_readl (dev, sle + offsetof (EP3_SPINLOCK_ELAN, sl_seq)) != sl->sl_seq)
		{
		    EPRINTF1 (DBG_RCVR, "%s: LockRcvrThread - thread trapped\n", rail->Generic.Name);

		    /* The thread processor has *really* trapped, and the spinlock is still held.
		     * thus is must have trapped due to a network error - we need to complete the
		     * actions required for this envelope, since we may be spin-locking the receiver
		     * to search the dma retry lists for a particular dma.  So must ensure that
		     * if the thread had trapped then the dma has been queued onto the retry list
		     * *before* we inspect them.
		     */
		    IncrStat (commsRail, LockRcvrTrapped);

		    /* We're going to generate a spurious interrupt here - since we will
		     * handle the thread processor trap directly */
		    ELAN_REG_REC(reg);
		    if (HandleTProcTrap (dev, &RestartBits))
		    {
			/* NOTE - this is not an assert, since the "store" to unlock the lock could
			 *        be held up on the PCI interface, whilst the thread processor has
			 *        gone on and switched to a new thread, which has then trapped, and
			 *        our read of the InterruptReg can overtake the unlock write.
			 *
			 * ASSERT (dev->ThreadTrap->Registers[REG_GLOBALS + (1^WordEndianFlip)] == 
			 *         elan3_sdram_readl (dev, rcvr->RcvrElan + offsetof (EP_RCVR_ELAN, PendingRxDescsElan)));
			 */

			PULSE_SCHED_STATUS (dev, RestartBits);

			DeliverTProcTrap (dev, dev->ThreadTrap, INT_TProc);
		    }
		}
		spin_unlock (&dev->IntrLock);
	    }
	    
	    DELAY (delay); delay++;
	}
	seq = elan3_sdram_readl (dev, sle + offsetof (EP3_SPINLOCK_ELAN, sl_seq));
    }
}

static void
UnlockRcvrThread (EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RAIL   *rail = RCVR_TO_RAIL(rcvrRail);
    sdramaddr_t sle  = rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadLock);

    mb();
    elan3_sdram_writel (rail->Device, sle + offsetof (EP3_SPINLOCK_ELAN, sl_lock), 0);
    mmiob(); 
}

void
CompleteEnvelope (EP3_RAIL *rail, E3_Addr rxdElanAddr, E3_uint32 PAckVal)
{
    ELAN3_DEV         *dev       = rail->Device;
    sdramaddr_t        rxdElan   = ep_elan2sdram (&rail->Generic, rxdElanAddr);
    EP3_RXD_RAIL      *rxdRail   = (EP3_RXD_RAIL *) (unsigned long) elan3_sdram_readq (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, MainAddr));
    EP_RXD_MAIN       *rxdMain   = rxdRail->Generic.Rxd->RxdMain;
    EP_ENVELOPE       *env       = &rxdMain->Envelope;
    EP3_RCVR_RAIL     *rcvrRail  = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP_COMMS_RAIL     *commsRail = rcvrRail->Generic.CommsRail;
    EP_RCVR           *rcvr      = rcvrRail->Generic.Rcvr;
    sdramaddr_t        queue     = ((EP3_COMMS_RAIL *) commsRail)->QueueDescs + rcvr->Service * sizeof (EP3_InputQueue);
    sdramaddr_t        sle       = rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadLock);
    EP3_SPINLOCK_MAIN *sl        = &rcvrRail->RcvrMain->ThreadLock;
    int               nodeId;
    EP_NODE_RAIL     *nodeRail;
    E3_DMA_BE         dma;
    E3_Addr           nfptr;
    E3_Addr	      next;

    ASSERT (commsRail->Rail == &rail->Generic);
    ASSERT (rxdElanAddr == elan3_sdram_readl (dev, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs)));

    IncrStat (commsRail, CompleteEnvelope);

    /* We don't need to aquire the NodeLock here (however we might be holding it),
     * since this can only get called while the node is connected, or disconnecting.
     * If the node is disconnecting, then we can get called from FlushDisconnecting()
     * while holding the NodeLock - after we cannot get called again until the node 
     * has reconnected from scratch.
     */
    /* Copy the envelope information */
    nfptr = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_fptr));

    if (nfptr == elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_top)))
	nfptr = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_base));
    else
	nfptr += elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_size));

    /* Copy the envelope and payload (unconditionally) */
    elan3_sdram_copyl_from_sdram (dev, rcvrRail->InputQueueBase + (nfptr - rcvrRail->InputQueueAddr), env, EP_ENVELOPE_SIZE + EP_PAYLOAD_SIZE);

    ASSERT (env->Version == EP_ENVELOPE_VERSION);

    /* Copy the received message length */
    rxdMain->Len = elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_len));
	
    /* Remove the RXD from the pending desc list */
    if ((next = elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Next))) == 0)
	rcvrRail->RcvrMain->PendingDescsTailp = 0;
    elan3_sdram_writel (dev, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs), next);

    /* Copy the DMA descriptor to queue on the approriate retry list */
    elan3_sdram_copyq_from_sdram (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Dmas[0]), &dma, sizeof (E3_DMA));	/* PCI read block */
    
    EP_ASSERT (&rail->Generic, dma.s.dma_direction == DMA_READ);;

#if defined(DEBUG_ASSERT) && defined(DEBUG_SDRAM_ASSERT)
    /* NOTE: not an assertion, since the thread packet could have successfully
     *       transferred the "put" dma to the far side - which could then have
     *       completed - but the far side will see a network error which will
     *       cause the virtual circuit to be dropped by the far side and this 
     *       DMA will be removed */
    if (rxdRail->RxdMain->DataEvent != EP3_EVENT_ACTIVE ||
	elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) != 1)
    {
	printk ("CompleteEnvelope: suspicious dma : Node=%d DataBlock=%d Event=%d\n", 
		env->NodeId, rxdRail->RxdMain->DataEvent, 
		elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)));
    }
#endif
    
    EPRINTF6 (DBG_RCVR, "%s: CompleteEnvelope: rxd=%p NodeId=%d Xid=%llx Cookies=%08x,%08x\n", commsRail->Rail->Name,
	      rxdRail, env->NodeId, (long long) env->Xid.Unique, dma.s.dma_srcCookieVProc, dma.s.dma_destCookieVProc);

    /* we MUST convert this into a DMA_READ_REQUEUE dma as if we don't the DMA descriptor will
     * be read from the EP_RETRY_DMA rather than the original DMA - this can then get reused 
     * and an incorrect DMA descriptor sent */
    dma.s.dma_source    = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, Dmas[0]);
    dma.s.dma_direction = (dma.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
    
    nodeId   = EP_VP_TO_NODE(dma.s.dma_srcVProc);
    nodeRail = &rail->Generic.Nodes[nodeId];

    ASSERT (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE);

    if (PAckVal != E3_PAckOk)
    {
	if (nodeRail->State == EP_NODE_CONNECTED)
	    QueueDmaForRetry (rail, &dma, EP_RETRY_LOW_PRI_RETRY);
	else
	    QueueDmaOnStalledList (rail, &dma);
    }

    /* Finaly forcefully drop the spinlock for the thread */
    sl->sl_seq = elan3_sdram_readl (dev, sle + offsetof (EP3_SPINLOCK_ELAN, sl_seq));

    wmb();
}

void
StallThreadForNoDescs (EP3_RAIL *rail, E3_Addr rcvrElanAddr, E3_Addr sp)
{
    ELAN3_DEV      *dev       = rail->Device;
    sdramaddr_t    rcvrElan   = ep_elan2sdram (&rail->Generic, rcvrElanAddr);
    EP3_RCVR_RAIL  *rcvrRail  = (EP3_RCVR_RAIL *) (unsigned long) elan3_sdram_readq (dev, rcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, MainAddr));
    EP_RCVR        *rcvr      = rcvrRail->Generic.Rcvr;
    EP_COMMS_RAIL  *commsRail = rcvrRail->Generic.CommsRail;

    EPRINTF3 (DBG_RCVR, "%s: StallThreadForNoDescs - rcvrRail=%p sp=%x\n", commsRail->Rail->Name, rcvrRail, sp);
    
    IncrStat (commsRail, StallThread);

    /* NOTE: spin lock not required as thread is trapped */
    
    if (rcvrRail->RcvrMain->PendingDescsTailp != 0)
    {
	EPRINTF1 (DBG_RCVR, "%s: StallThreadForNoDescs - pending descriptors, wakeup thread\n", commsRail->Rail->Name);
	
	/*
	 * A receive buffer was queued after the thread had decided to go to 
	 * sleep, but before the event interrupt occured.  Just restart the
	 * thread to consume the envelope.
	 */
	IssueRunThread (rail, sp);
    }
    else
    {
	EPRINTF1 (DBG_RCVR, "%s: StallThreadForNoDescs - set ThreadWaiting\n", commsRail->Rail->Name);
	
	IncrStat (commsRail, ThrdWaiting);

	/* Mark the rcvr as waiting for a rxd, and schedule a call of ep_check_rcvr
	 * to attempt to "steal" a descriptor from a different rail */
	rcvrRail->ThreadWaiting = sp;

	ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
    }
}

void
StallThreadForHalted (EP3_RAIL *rail, E3_Addr rcvrElanAddr, E3_Addr sp)
{
    ELAN3_DEV     *dev       = rail->Device;
    sdramaddr_t    rcvrElan  = ep_elan2sdram (&rail->Generic, rcvrElanAddr);
    EP3_RCVR_RAIL *rcvrRail  = (EP3_RCVR_RAIL *) (unsigned long) elan3_sdram_readq (dev, rcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, MainAddr));
    EP_RCVR       *rcvr      = rcvrRail->Generic.Rcvr;
    unsigned long  flags     = 0;

    spin_lock_irqsave (&rcvr->Lock, flags);

    rcvrRail->ThreadHalted = sp;

    EPRINTF2 (DBG_EPTRAP, "%s: StallThreadForHalted: sp=%08x\n", rail->Generic.Name, sp);

    if (rcvrRail->CleanupWaiting)
	kcondvar_wakeupone (&rcvrRail->CleanupSleep, &rcvr->Lock);
    rcvrRail->CleanupWaiting = 0;

    spin_unlock_irqrestore (&rcvr->Lock, flags);
}
/*
 * RxDataEvent: arg == EP3_RXD_RAIL
 *   Called on completion of receiving data.
 */
static void
RxDataEvent (EP3_RAIL *rail, void *arg)
{
    EP3_RXD_RAIL  *rxdRail  = (EP3_RXD_RAIL *) arg;
    EP3_RCVR_RAIL *rcvrRail = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP_RXD        *rxd      = rxdRail->Generic.Rxd;
    EP_ENVELOPE   *env      = &rxd->RxdMain->Envelope;
    EP_RCVR       *rcvr     = rxd->Rcvr;
    ELAN3_DEV	  *dev      = rail->Device;
    unsigned long flags;
    int delay = 1;

    spin_lock_irqsave (&rcvr->Lock, flags);
    for (;;)
    {
	if (EP3_EVENT_FIRED (rxdRail->DataCookie, rxdRail->RxdMain->DataEvent))
	    break;

	if (EP3_EVENT_FIRING (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent), rxdRail->DataCookie, rxdRail->RxdMain->DataEvent))
	{
	    if (delay > EP3_EVENT_FIRING_TLIMIT)
		panic ("RxDataEvent: events set but block copy not completed\n");
	    DELAY(delay);
	    delay <<= 1;
	}
	else
	{
	    printk ("%s: RxDataEvent: rxd %p not complete [%x,%x,%x]\n", rail->Generic.Name, rxd, rxdRail->RxdMain->DataEvent,
		    elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)),
		    elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Type)));
	    
	    spin_unlock_irqrestore (&rcvr->Lock, flags);
	    return;
	}
	mb();
    }

    /*
     * Note, since the thread will have sent the "get" dma before copying the
     * envelope, we must check that it has completed doing this,  if not then
     * it might be that the thread trapped due to a network error, so we must
     * spinlock against the thread 
     */
    if (rxd->RxdMain->Len == EP_RXD_PENDING)
    {
	LockRcvrThread (rcvrRail);
	UnlockRcvrThread (rcvrRail);

	ASSERT (env->Version == EP_ENVELOPE_VERSION && rxd->RxdMain->Len != EP_RXD_PENDING);
    }

    EPRINTF7 (DBG_RCVR, "%s: RxDataEvent: rxd=%p rxdRail=%p completed from elan node %d [XID=%llx] Length %d State %x\n", 
	      rail->Generic.Name, rxd, rxdRail, env->NodeId, (long long) env->Xid.Unique, rxd->RxdMain->Len, rxd->State);

    EP_ASSERT (&rail->Generic, rxd->State == EP_RXD_RECEIVE_ACTIVE || rxd->State == EP_RXD_PUT_ACTIVE || rxd->State == EP_RXD_GET_ACTIVE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));	/* PCI read */
    EP_ASSERT (&rail->Generic, rxdRail->RxdMain->DoneEvent == EP3_EVENT_PRIVATE);

    rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
    rxd->Data.nmd_attr          = EP_RAIL2RAILMASK (rail->Generic.Number);

    if (rxd->RxdMain->Len >= 0 && EP_IS_RPC(env->Attr))
	rxd->State = EP_RXD_RPC_IN_PROGRESS;
    else
    {
	rxd->State = EP_RXD_COMPLETED;

	/* remove from active list */
	list_del (&rxd->Link);

	UnbindRxdFromRail (rxd, rxdRail);
	FreeRxdRail (rcvrRail, rxdRail);
    }

    spin_unlock_irqrestore (&rcvr->Lock, flags);
    ep_rxd_received (rxd);

}

/* 
 * RxDataRetry: arg == EP3_RXD_RAIL
 *   Called on retry of "get" dma of large transmit data
 *   and rpc_get/rpc_put and "put" of datavec of rpc completion.
 */
static void
RxDataRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status)
{
    EP3_RXD_RAIL  *rxdRail  = (EP3_RXD_RAIL *) arg;
    EP_COMMS_RAIL *commsRail = rxdRail->Generic.RcvrRail->CommsRail;
    EP_RXD        *rxd       = rxdRail->Generic.Rxd;

#if defined(DEBUG_ASSERT)
    RxDataVerify (rail, arg, dma);
#endif

    IncrStat (commsRail, RxDataRetry);

    EPRINTF4 (DBG_RCVR, "%s: RxDataRetry: rcvr %p rxd %p [XID=%llx]\n", rail->Generic.Name, rxd->Rcvr, rxd, (long long) rxd->RxdMain->Envelope.Xid.Unique);

    QueueDmaForRetry (rail, dma, EP_RETRY_LOW_PRI_RETRY + ep_backoff (&rxdRail->Backoff, EP_BACKOFF_DATA));
}

static void
RxDataVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma)
{
#if defined(DEBUG_ASSERT)
    EP3_RXD_RAIL   *rxdRail  = (EP3_RXD_RAIL *) arg;
    EP_RXD        *rxd       = rxdRail->Generic.Rxd;

    if (dma->s.dma_direction == DMA_WRITE)
    {
	EP_ASSERT (&rail->Generic, 
		   (rxd->State == EP_RXD_RECEIVE_ACTIVE  && rxdRail->RxdMain->DataEvent == EP3_EVENT_ACTIVE && rxdRail->RxdMain->DoneEvent == EP3_EVENT_PRIVATE) ||
		   (rxd->State == EP_RXD_PUT_ACTIVE      && rxdRail->RxdMain->DataEvent == EP3_EVENT_ACTIVE && rxdRail->RxdMain->DoneEvent == EP3_EVENT_PRIVATE) ||
		   (rxd->State == EP_RXD_COMPLETE_ACTIVE && rxdRail->RxdMain->DataEvent == EP3_EVENT_PRIVATE && rxdRail->RxdMain->DoneEvent == EP3_EVENT_ACTIVE));
	EP_ASSERT (&rail->Generic, SDRAM_ASSERT (rxd->State == EP_RXD_COMPLETE_ACTIVE ?
						 elan3_sdram_readl (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 1:		/* PCI read */
						 elan3_sdram_readl (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 1));		/* PCI read */
    }
    else
    {
	EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_READ_REQUEUE);
	
#if defined(DEBUG_SDRAM_ASSERT)
	/* NOTE: not an assertion, since the "get" DMA can still be running if
	 *       it's packet got a network error - and then the "put" from the
	 *       far side has completed - however the virtual circuit should
	 *       then be dropped by the far side and this DMA will be removed */
	if (EP_VP_TO_NODE(dma->s.dma_srcVProc) != ep_rxd_node(rxd) || 
	    (rxd->State != EP_RXD_RECEIVE_ACTIVE && rxd->State != EP_RXD_GET_ACTIVE) ||
	    rxdRail->RxdMain->DataEvent != EP3_EVENT_ACTIVE ||
	    elan3_sdram_readl (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) != 1)
	{
	    EPRINTF6 (DBG_RCVR, "%s: RxDataRetry: suspicious dma : VProc=%d NodeId=%d State=%d DataBlock=%x Event=%d\n",  
		      rail->Generic.Name, EP_VP_TO_NODE(dma->s.dma_srcVProc), ep_rxd_node(rxd), rxd->State, rxdRail->RxdMain->DataEvent, 
		      elan3_sdram_readl (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)));
	}
#endif /* defined(DEBUG_SDRAM_ASSERT) */
    }
#endif /* DEBUG_ASSERT */
}

/*
 * RxDoneEvent: arg == EP_RXD
 *   Called on completion of large receive.
 */
static void
RxDoneEvent (EP3_RAIL *rail, void *arg)
{
    EP3_RXD_RAIL  *rxdRail   = (EP3_RXD_RAIL *) arg;
    EP3_RCVR_RAIL *rcvrRail  = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP_COMMS_RAIL *commsRail = rcvrRail->Generic.CommsRail;
    EP_RXD        *rxd       = rxdRail->Generic.Rxd;
    EP_RCVR       *rcvr      = rxd->Rcvr;
    ELAN3_DEV	  *dev       = rail->Device;
    int            delay     = 1;
    unsigned long  flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    for (;;)
    {
	if (EP3_EVENT_FIRED (rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent))
	    break;
	
	if (EP3_EVENT_FIRING (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent), rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent))
	{
	    if (delay > EP3_EVENT_FIRING_TLIMIT)
		panic ("RxDoneEvent: events set but block copy not completed\n");
	    DELAY(delay);
	    delay <<= 1;
	}
	else
	{
	    printk ("RxDoneEvent: rxd %p not complete [%x,%x.%x]\n", rxd, rxdRail->RxdMain->DoneEvent,
		    elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)),
		    elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Type)));
	    
	    spin_unlock_irqrestore (&rcvr->Lock, flags);
	    return;
	}
	mb();
    }

    EPRINTF4 (DBG_RCVR, "%s: RxDoneEvent: rxd %p completed from elan node %d [XID=%llx]\n", 
	      commsRail->Rail->Name, rxd, rxd->RxdMain->Envelope.NodeId, (long long) rxd->RxdMain->Envelope.Xid.Unique);
    
    IncrStat (commsRail, RxDoneEvent);

    EP_ASSERT (&rail->Generic, rxdRail->RxdMain->DataEvent  == EP3_EVENT_PRIVATE);
    EP_ASSERT (&rail->Generic, EP3_EVENT_FIRED (rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent));
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));	/* PCI read */
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));	/* PCI read */

    /* mark rxd as private  */
    rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;

    /* remove from active list */
    list_del (&rxd->Link);

    UnbindRxdFromRail (rxd, rxdRail);
    FreeRxdRail (rcvrRail, rxdRail);
	
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    rxd->Handler (rxd);
}

/* 
 * RxDoneRetry: arg == EP_RXD
 *   Called on retry of "put" of RPC completion status block
 */
static void
RxDoneRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status)
{
    EP3_RXD_RAIL  *rxdRail   = (EP3_RXD_RAIL *) arg;
    EP_COMMS_RAIL *commsRail = rxdRail->Generic.RcvrRail->CommsRail;
    EP_RXD        *rxd       = rxdRail->Generic.Rxd;

#if defined(DEBUG_ASSERT)
    RxDoneVerify (rail, arg, dma);
#endif

    IncrStat (commsRail, RxDoneRetry);

    EPRINTF4 (DBG_RCVR, "%s: RxDoneRetry: rcvr %p rxd %p [XID=%llx]\n", commsRail->Rail->Name, rxd->Rcvr, rxd, (long long) rxd->RxdMain->Envelope.Xid.Unique);

    QueueDmaForRetry (rail, dma, EP_RETRY_LOW_PRI_RETRY + ep_backoff (&rxdRail->Backoff, EP_BACKOFF_DONE));
}

static void
RxDoneVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma)
{
#if defined(DEBUG_ASSERT)
    EP3_RXD_RAIL *rxdRail = (EP3_RXD_RAIL *) arg;
    EP_RXD       *rxd     = rxdRail->Generic.Rxd;

    EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_WRITE && EP_VP_TO_NODE(dma->s.dma_destVProc) == ep_rxd_node(rxd));
    EP_ASSERT (&rail->Generic, rxd->State == EP_RXD_COMPLETE_ACTIVE && rxdRail->RxdMain->DoneEvent  == EP3_EVENT_ACTIVE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 1));	/* PCI read */
#endif /* defined(DEBUG_ASSERT) */
}

int
ep3rcvr_queue_rxd (EP_RXD *rxd, EP_RCVR_RAIL *r)
{
    EP3_RCVR_RAIL *rcvrRail = (EP3_RCVR_RAIL *) r;
    EP3_RAIL      *rail     = RCVR_TO_RAIL(rcvrRail);
    ELAN3_DEV     *dev      = rail->Device;
    EP3_RXD_RAIL  *rxdRail;

    ASSERT ( SPINLOCK_HELD(&rxd->Rcvr->Lock));

    if ((rxdRail = GetRxdRail (rcvrRail)) == NULL)
	return 0;

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->Generic);

    elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_addr), rxd->Data.nmd_addr);		/* PCI write */
    elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_len),  rxd->Data.nmd_len);		/* PCI write */
    elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_attr), rxd->Data.nmd_attr);		/* PCI write */

    /* Bind the rxdRail and rxd together */
    BindRxdToRail (rxd, rxdRail);
    
    /* Mark as active */
    elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 1);

    rxdRail->RxdMain->DataEvent  = EP3_EVENT_ACTIVE;
    rxdRail->RxdMain->DoneEvent  = EP3_EVENT_PRIVATE;

    /* Interlock with StallThreadForNoDescs */
    spin_lock (&dev->IntrLock);

    EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_queue_rxd: rcvr %p rxd %p rxdRail %p\n", rail->Generic.Name, rxd->Rcvr, rxd, rxdRail);

    EP3_SPINENTER (dev, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingLock), &rcvrRail->RcvrMain->PendingLock);

    elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, Next), 0); 					/* PCI write */
    if (rcvrRail->RcvrMain->PendingDescsTailp == 0)
	elan3_sdram_writel (dev, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs), rxdRail->RxdElanAddr);	/* PCI write */
    else
	elan3_sdram_writel (dev, rcvrRail->RcvrMain->PendingDescsTailp, rxdRail->RxdElanAddr);				/* PCI write */
    rcvrRail->RcvrMain->PendingDescsTailp = rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, Next);
    
    EP3_SPINEXIT (dev, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingLock), &rcvrRail->RcvrMain->PendingLock);

    /* If the thread has paused because it was woken up with no receive buffer */
    /* ready, then wake it up to process the one we've just added */
    if (rcvrRail->ThreadWaiting)
    {
	EPRINTF1 (DBG_RCVR, "%s: DoReceive: ThreadWaiting - restart thread\n", rail->Generic.Name);

	IssueRunThread (rail, rcvrRail->ThreadWaiting);

	rcvrRail->ThreadWaiting = (E3_Addr) 0;
    }

    spin_unlock (&dev->IntrLock);

    return 1;
}

void
ep3rcvr_rpc_put (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP3_RXD_RAIL      *rxdRail   = (EP3_RXD_RAIL *) rxd->RxdRail;
    EP3_RCVR_RAIL     *rcvrRail  = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP3_RAIL          *rail      = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV         *dev       = rail->Device;

    EP3_RXD_RAIL_MAIN *rxdMain   = rxdRail->RxdMain;
    sdramaddr_t        rxdElan   = rxdRail->RxdElan;
    EP_ENVELOPE       *env       = &rxd->RxdMain->Envelope;
    E3_DMA_BE	       dmabe;
    int		       i, len;

    EP_ASSERT (&rail->Generic, rxd->State == EP_RXD_PUT_ACTIVE);
    EP_ASSERT (&rail->Generic, rxdMain->DataEvent == EP3_EVENT_PRIVATE && rxdMain->DoneEvent == EP3_EVENT_PRIVATE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));	/* PCI read */
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));	/* PCI read */

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->Generic);
    
    /* Generate the DMA chain to put the data in two loops to burst
     * the data across the PCI bus */
    for (len = 0, i = (nFrags-1), local += (nFrags-1), remote += (nFrags-1); i >= 0;   len += local->nmd_len, i--, local--, remote--)
    {
	dmabe.s.dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_WRITE, DMA_NORMAL, EP3_DMAFAILCOUNT);
	dmabe.s.dma_size            = local->nmd_len;
	dmabe.s.dma_source          = local->nmd_addr;
	dmabe.s.dma_dest            = remote->nmd_addr;
	dmabe.s.dma_destEvent       = (E3_Addr) 0;
	dmabe.s.dma_destCookieVProc = EP_VP_DATA (env->NodeId);
	if (i == (nFrags-1))
	    dmabe.s.dma_srcEvent    = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, DataEvent);
	else
	    dmabe.s.dma_srcEvent    = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i]);
	dmabe.s.dma_srcCookieVProc  = LocalCookie (rail, env->NodeId);
	
	EPRINTF9 (DBG_RCVR, "%s: ep3rcvr_rpc_put: rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x Cookies=%x.%x\n", rail->Generic.Name, rxd,
		  (long long) env->Xid.Unique, i, local->nmd_addr, remote->nmd_addr, local->nmd_len, dmabe.s.dma_destCookieVProc, dmabe.s.dma_srcCookieVProc);
	
	if (i != 0)
	    elan3_sdram_copyq_to_sdram (dev, &dmabe, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Dmas[i]), sizeof (E3_DMA)); /* PCI write block */
    }
    
    for (i = 0; i < nFrags; i++)
	elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i].ev_Count), 1);				/* PCI write */
    
    /* Initialise the data event */
    elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 1);			/* PCI write */
    rxdMain->DataEvent = EP3_EVENT_ACTIVE;
   
    ASSERT (rail->Generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->Generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    if (IssueDma (rail, &dmabe, EP_RETRY_LOW_PRI, FALSE) != ISSUE_COMMAND_OK)
    {
	/* Failed to issue the dma command, so copy the dma descriptor and queue it for retry */
	EPRINTF2 (DBG_RCVR, "%s: ep3rcvr_rpc_put: queue rxd %p on retry thread\n", rail->Generic.Name, rxd);
	
	QueueDmaForRetry (rail, &dmabe, EP_RETRY_LOW_PRI);
    }
    
    BucketStat (rxd->Rcvr->Subsys, RPCPut, len);
}

void
ep3rcvr_rpc_get (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP3_RXD_RAIL      *rxdRail   = (EP3_RXD_RAIL *) rxd->RxdRail;
    EP3_RCVR_RAIL     *rcvrRail  = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP3_RAIL          *rail      = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV         *dev       = rail->Device;

    EP3_RXD_RAIL_MAIN *rxdMain   = rxdRail->RxdMain;
    sdramaddr_t        rxdElan   = rxdRail->RxdElan;
    EP_ENVELOPE       *env       = &rxd->RxdMain->Envelope;
    E3_DMA_BE	       dmabe;
    int		       i, len;

    EP_ASSERT (&rail->Generic, rxd->State == EP_RXD_GET_ACTIVE);
    EP_ASSERT (&rail->Generic, rxdMain->DataEvent == EP3_EVENT_PRIVATE && rxdMain->DoneEvent == EP3_EVENT_PRIVATE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));	/* PCI read */
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));	/* PCI read */
	
    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->Generic);
    
    /* Generate the DMA chain to get the data in two loops to burst
     * the data across the PCI bus */
    for (len = 0, i = (nFrags-1), remote += (nFrags-1), local += (nFrags-1); i >= 0;   len += remote->nmd_len, i--, remote--, local--)
    {
	dmabe.s.dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_READ, DMA_NORMAL, EP3_DMAFAILCOUNT);
	dmabe.s.dma_size            = remote->nmd_len;
	dmabe.s.dma_source          = remote->nmd_addr;
	dmabe.s.dma_dest            = local->nmd_addr;
	if (i == (nFrags-1))
	    dmabe.s.dma_destEvent   = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, DataEvent);
	else
	    dmabe.s.dma_destEvent   = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i]);
	dmabe.s.dma_destCookieVProc = LocalCookie (rail, env->NodeId);
	dmabe.s.dma_srcEvent        = (E3_Addr) 0;
	dmabe.s.dma_srcCookieVProc  = RemoteCookie (rail, env->NodeId);
	
	EPRINTF9 (DBG_RCVR, "%s: ep3rcvr_rpc_get rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x Cookies=%x.%x\n", rail->Generic.Name, rxd,
		  (long long) env->Xid.Unique, i, remote->nmd_addr, local->nmd_addr, remote->nmd_len, dmabe.s.dma_destCookieVProc, 
		  dmabe.s.dma_srcCookieVProc);
	
	/* 
	 * Always copy down the dma descriptor, since we issue it as a READ_REQUEUE
	 * dma, and the elan will fetch the descriptor to send out of the link from
	 * the rxdElan->Dmas[i] location,  before issueing the DMA chain we modify
	 * the dma_source.
	 */
	elan3_sdram_copyq_to_sdram (dev, &dmabe, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Dmas[i]), sizeof (E3_DMA)); /* PCI write block */
    }
    
    for (i = 0; i < nFrags; i++)
	elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i].ev_Count), 1);				/* PCI write */
    
    /* Initialise the data event */
    elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 1);			/* PCI write */
    rxdMain->DataEvent  = EP3_EVENT_ACTIVE;
    
    ASSERT (rail->Generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->Generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    /* we MUST convert this into a DMA_READ_REQUEUE dma as if we don't the DMA descriptor will
     * be read from the EP_RETRY_DMA rather than the orignal DMA - this can then get reused 
     * and an incorrect DMA descriptor sent */
    dmabe.s.dma_source    = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, Dmas[0]);
    dmabe.s.dma_direction = (dmabe.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
    
    if (IssueDma (rail, &dmabe, EP_RETRY_LOW_PRI, FALSE) != ISSUE_COMMAND_OK)
    {
	/* Failed to issue the dma command, so copy the dma descriptor and queue it for retry */
	EPRINTF2 (DBG_RCVR, "%s: ep3rcvr_rpc_get: queue rxd %p on retry thread\n", rail->Generic.Name, rxd);
	
	QueueDmaForRetry (rail, &dmabe, EP_RETRY_LOW_PRI);
    }

    BucketStat (rxd->Rcvr->Subsys, RPCGet, len);
}
	
void
ep3rcvr_rpc_complete (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP3_RXD_RAIL      *rxdRail   = (EP3_RXD_RAIL *) rxd->RxdRail;
    EP3_RCVR_RAIL     *rcvrRail  = (EP3_RCVR_RAIL *) rxdRail->Generic.RcvrRail;
    EP3_RAIL          *rail      = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV         *dev       = rail->Device;

    EP3_RXD_RAIL_MAIN *rxdMain   = rxdRail->RxdMain;
    sdramaddr_t        rxdElan   = rxdRail->RxdElan;
    EP_ENVELOPE       *env       = &rxd->RxdMain->Envelope;
    E3_DMA_BE	       dmabe;
    int		       i, len;
    
    EP_ASSERT (&rail->Generic, rxd->State == EP_RXD_COMPLETE_ACTIVE);
    EP_ASSERT (&rail->Generic, rxdMain->DataEvent == EP3_EVENT_PRIVATE && rxdMain->DoneEvent == EP3_EVENT_PRIVATE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));	/* PCI read */
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));	/* PCI read */

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->Generic);
    
    /* Initialise the status block dma */
    dmabe.s.dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_WRITE, DMA_NORMAL, EP3_DMAFAILCOUNT);
    dmabe.s.dma_size            = sizeof (EP_STATUSBLK);
    dmabe.s.dma_source          = rxd->NmdMain.nmd_addr + offsetof (EP_RXD_MAIN, StatusBlk);
    dmabe.s.dma_dest            = env->TxdMain.nmd_addr + offsetof (EP_TXD_MAIN, StatusBlk);
    dmabe.s.dma_destEvent       = env->TxdRail + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent);
    dmabe.s.dma_destCookieVProc = EP_VP_DATA(env->NodeId);
    dmabe.s.dma_srcEvent        = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent);
    dmabe.s.dma_srcCookieVProc  = LocalCookie (rail, env->NodeId);
    
    EPRINTF8 (DBG_RCVR, "%s: ep3rcvr_rpc_complete: rxd %p [XID=%llx] statusblk source=%08x dest=%08x len=%x Cookies=%x.%x\n", rail->Generic.Name, rxd,
	      (long long) env->Xid.Unique, dmabe.s.dma_source, dmabe.s.dma_dest, dmabe.s.dma_size, dmabe.s.dma_destCookieVProc, 
	      dmabe.s.dma_srcCookieVProc);

    for (len = 0, i = EP_MAXFRAG, remote += (nFrags-1), local += (nFrags-1); i > EP_MAXFRAG-nFrags; len += local->nmd_len, i--, local--, remote--)
    {
	/* copy down previous dma */
	elan3_sdram_copyq_to_sdram (dev, &dmabe, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Dmas[i]),  sizeof (E3_DMA));	/* PCI write block */
	
	dmabe.s.dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_WRITE, DMA_NORMAL, EP3_DMAFAILCOUNT);
	dmabe.s.dma_size            = local->nmd_len;
	dmabe.s.dma_source          = local->nmd_addr;
	dmabe.s.dma_dest            = remote->nmd_addr;
	dmabe.s.dma_destEvent       = (E3_Addr) 0;
	dmabe.s.dma_destCookieVProc = EP_VP_DATA (env->NodeId);
	dmabe.s.dma_srcEvent        = rxdRail->RxdElanAddr + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i-1]);
	dmabe.s.dma_srcCookieVProc  = LocalCookie (rail, env->NodeId);
	
	EPRINTF9 (DBG_RCVR, "%s: ep3rcvr_rpc_complete: rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x Cookies=%x.%x\n", rail->Generic.Name, rxd,
		  (long long) env->Xid.Unique, i, local->nmd_addr, remote->nmd_addr, local->nmd_len, dmabe.s.dma_destCookieVProc, 
		  dmabe.s.dma_srcCookieVProc);
    }
    
    for (i = EP_MAXFRAG-nFrags; i < EP_MAXFRAG; i++)
	elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[i].ev_Count), 1);				/* PCI write */
    
    /* Initialise the done event */
    elan3_sdram_writel (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count), 1);					/* PCI write */
    rxdMain->DoneEvent  = EP3_EVENT_ACTIVE;

    ASSERT (rail->Generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->Generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    if (IssueDma (rail, &dmabe, EP_RETRY_LOW_PRI, FALSE) != ISSUE_COMMAND_OK)
    {
	/* Failed to issue the dma command, so copy the dma descriptor and queue it for retry */
	EPRINTF2 (DBG_RCVR, "%s: ep3rcvr_rpc_complete: queue rxd %p on retry thread\n", rail->Generic.Name, rxd);
	
	QueueDmaForRetry (rail, &dmabe, EP_RETRY_LOW_PRI);
    }

    BucketStat (rxd->Rcvr->Subsys, CompleteRPC, len);
}
	
void
ep3rcvr_add_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *commsRail)
{
    EP3_RAIL          *rail   = (EP3_RAIL *) commsRail->Rail;
    sdramaddr_t        qdescs = ((EP3_COMMS_RAIL *) commsRail)->QueueDescs;
    EP3_RCVR_RAIL     *rcvrRail;
    EP3_InputQueue     qdesc;
    sdramaddr_t        stack;
    unsigned long      flags;

    KMEM_ZALLOC (rcvrRail, EP3_RCVR_RAIL *, sizeof (EP3_RCVR_RAIL), TRUE);

    kcondvar_init (&rcvrRail->CleanupSleep);
    spin_lock_init (&rcvrRail->FreeDescLock);
    INIT_LIST_HEAD (&rcvrRail->FreeDescList);
    INIT_LIST_HEAD (&rcvrRail->DescBlockList);

    rcvrRail->Generic.CommsRail = commsRail;
    rcvrRail->Generic.Rcvr      = rcvr;

    rcvrRail->RcvrMain       = ep_alloc_main (&rail->Generic, sizeof (EP3_RCVR_RAIL_MAIN), 0, &rcvrRail->RcvrMainAddr);
    rcvrRail->RcvrElan       = ep_alloc_elan (&rail->Generic, sizeof (EP3_RCVR_RAIL_ELAN), 0, &rcvrRail->RcvrElanAddr);
    rcvrRail->InputQueueBase = ep_alloc_elan (&rail->Generic, EP_INPUTQ_SIZE * rcvr->InputQueueEntries, 0, &rcvrRail->InputQueueAddr);
    stack                    = ep_alloc_elan (&rail->Generic, EP3_STACK_SIZE, 0, &rcvrRail->ThreadStack);

    rcvrRail->TotalDescCount = 0;
    rcvrRail->FreeDescCount  = 0;

    /* Initialise the main/elan spin lock */
    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadLock.sl_lock), 0);
    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadLock.sl_seq),  0);

    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingLock.sl_lock), 0);
    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingLock.sl_seq), 0);
    
    /* Initialise the receive lists */
    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs), 0);
    
    /* Initialise the ThreadShould Halt */
    elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadShouldHalt), 0);

    /* Initialise pointer to the ep_rcvr_rail */
    elan3_sdram_writeq (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, MainAddr), (unsigned long) rcvrRail);

    /* Initialise elan visible main memory */
    rcvrRail->RcvrMain->ThreadLock.sl_seq  = 0;
    rcvrRail->RcvrMain->PendingLock.sl_seq = 0;
    rcvrRail->RcvrMain->PendingDescsTailp  = 0;

    /* initialise and copy down the input queue descriptor */
    qdesc.q_state          = E3_QUEUE_FULL;
    qdesc.q_base           = rcvrRail->InputQueueAddr;
    qdesc.q_top            = rcvrRail->InputQueueAddr + (rcvr->InputQueueEntries-1) * EP_INPUTQ_SIZE;
    qdesc.q_fptr           = rcvrRail->InputQueueAddr;
    qdesc.q_bptr           = rcvrRail->InputQueueAddr + EP_INPUTQ_SIZE;
    qdesc.q_size           = EP_INPUTQ_SIZE;
    qdesc.q_event.ev_Count = 0;
    qdesc.q_event.ev_Type  = 0;

    elan3_sdram_copyl_to_sdram (rail->Device, &qdesc, qdescs + rcvr->Service * sizeof (EP3_InputQueue), sizeof (EP3_InputQueue));

    spin_lock_irqsave (&rcvr->Lock, flags);
    rcvr->Rails[rail->Generic.Number] = &rcvrRail->Generic;
    rcvr->RailMask |= EP_RAIL2RAILMASK (rail->Generic.Number);
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    /* initialise and run the Elan thread to process the queue */
    IssueRunThread (rail, ep3_init_thread (rail->Device, ep_symbol (&rail->ThreadCode, "ep3comms_rcvr"),
					   rcvrRail->ThreadStack, stack, EP3_STACK_SIZE, 5,
					   rail->RailElanAddr, rcvrRail->RcvrElanAddr, rcvrRail->RcvrMainAddr,
					   EP_MSGQ_ADDR(rcvr->Service),
					   rail->ElanCookies));
}

void
ep3rcvr_del_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *commsRail)
{
    EP3_RAIL         *rail     = (EP3_RAIL *) commsRail->Rail;
    EP3_RCVR_RAIL    *rcvrRail = (EP3_RCVR_RAIL *) rcvr->Rails[rail->Generic.Number];  
    unsigned long     flags;
    struct list_head *el, *nel;

    EPRINTF1 (DBG_RCVR, "%s: ep3rcvr_del_rail: removing rail\n", rail->Generic.Name);

    /* flag the rail as no longer available */
    spin_lock_irqsave (&rcvr->Lock, flags);
    rcvr->RailMask &= ~EP_RAIL2RAILMASK (rail->Generic.Number);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
    
    /* mark the input queue descriptor as full */
    SetQueueLocked(rail, ((EP3_COMMS_RAIL *)commsRail)->QueueDescs + rcvr->Service * sizeof (EP3_InputQueue));

    /* need to halt the thread first         */
    /*   set ThreadShouldHalt in elan memory */
    /*   then trigger the event              */
    /*   and wait on haltWait                */
    elan3_sdram_writel  (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, ThreadShouldHalt), TRUE);

    IssueSetevent (rail,  EP_MSGQ_ADDR(rcvr->Service) + offsetof(EP3_InputQueue, q_event));

    spin_lock_irqsave (&rcvr->Lock, flags);

    while (rcvrRail->ThreadHalted == 0)
    {
	rcvrRail->CleanupWaiting++;
	kcondvar_wait (&rcvrRail->CleanupSleep, &rcvr->Lock, &flags);
    }

    /* at this point the thread is halted and it has no envelopes */
 
    /* we need to wait until all the rxd's in the list that are 
     * bound to the rail we are removing are not pending 
     */
    for (;;)
    {
	int mustWait = 0;
	
	list_for_each (el, &rcvr->ActiveDescList) {
	    EP_RXD       *rxd     = list_entry (el,EP_RXD, Link);
	    EP3_RXD_RAIL *rxdRail = (EP3_RXD_RAIL *) rxd->RxdRail;

	    if (rxdRail && RXD_BOUND2RAIL (rxdRail, rcvrRail) && rxd->RxdMain->Len != EP_RXD_PENDING)
	    {
		mustWait++;
		break;
	    }
	}
	
	if (! mustWait)
	    break;

	EPRINTF1 (DBG_RCVR, "%s: ep3rcvr_del_rail: waiting for active rxd's to be returned\n", rail->Generic.Name);

	rcvrRail->CleanupWaiting++;
	kcondvar_wait (&rcvrRail->CleanupSleep, &rcvr->Lock, &flags);
    }

    /* at this point all rxd's in the list that are bound to the deleting rail are not pending */
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd     = list_entry (el, EP_RXD, Link);
	EP3_RXD_RAIL *rxdRail = (EP3_RXD_RAIL *) rxd->RxdRail;
	
	if (rxdRail && RXD_BOUND2RAIL (rxdRail, rcvrRail))
	{
	    /* here we need to unbind the remaining rxd's */
	    rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
	    rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;
 
	    elan3_sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);	/* PCI write */
	    elan3_sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count), 0);	/* PCI write */

	    UnbindRxdFromRail (rxd, rxdRail);
	    FreeRxdRail(rcvrRail,  rxdRail );
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);
    
    /* wait for all rxd's for this rail to become free */
    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);
    while (rcvrRail->FreeDescCount != rcvrRail->TotalDescCount)
    {
	rcvrRail->FreeDescWaiting++;
	kcondvar_wait (&rcvrRail->FreeDescSleep, &rcvrRail->FreeDescLock, &flags);
    }
    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);

    /* can now remove the rail as it can no longer be used */
    spin_lock_irqsave (&rcvr->Lock, flags);
    rcvr->Rails[rail->Generic.Number] = NULL;
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    /* all the rxd's accociated with DescBlocks must be in the FreeDescList */
    ASSERT (rcvrRail->TotalDescCount == rcvrRail->FreeDescCount);

    /* run through the DescBlockList deleting them */
    while (!list_empty (&rcvrRail->DescBlockList))
	FreeRxdRailBlock (rcvrRail, list_entry(rcvrRail->DescBlockList.next, EP3_RXD_RAIL_BLOCK , Link));

    /* it had better be empty after that */
    ASSERT ((rcvrRail->TotalDescCount == 0) && (rcvrRail->TotalDescCount == rcvrRail->FreeDescCount));
    
    ep_free_elan (&rail->Generic, rcvrRail->ThreadStack, EP3_STACK_SIZE);
    ep_free_elan (&rail->Generic, rcvrRail->InputQueueAddr, EP_INPUTQ_SIZE * rcvr->InputQueueEntries);
    ep_free_elan (&rail->Generic, rcvrRail->RcvrElanAddr, sizeof (EP3_RCVR_RAIL_ELAN));
    ep_free_main (&rail->Generic, rcvrRail->RcvrMainAddr, sizeof (EP3_RCVR_RAIL_MAIN));

    KMEM_FREE (rcvrRail, sizeof (EP3_RCVR_RAIL));
}

EP_RXD *
ep3rcvr_steal_rxd (EP_RCVR_RAIL *r)
{
    EP3_RCVR_RAIL *rcvrRail = (EP3_RCVR_RAIL *) r;
    EP3_RAIL      *rail     = RCVR_TO_RAIL (rcvrRail);
    EP_RCVR       *rcvr     = rcvrRail->Generic.Rcvr;
    E3_Addr        rxdElanAddr;
    unsigned long flags;

    spin_lock_irqsave (&rcvr->Lock, flags);

    LockRcvrThread (rcvrRail);
    if ((rxdElanAddr = elan3_sdram_readl (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs))) != 0)
    {
	sdramaddr_t  rxdElan  = ep_elan2sdram (&rail->Generic, rxdElanAddr);
	EP3_RXD_RAIL *rxdRail = (EP3_RXD_RAIL *) (unsigned long) elan3_sdram_readq (rail->Device, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, MainAddr));
	EP_RXD      *rxd      = rxdRail->Generic.Rxd;
	sdramaddr_t  next;
	
	EPRINTF2 (DBG_RCVR, "%s: StealRxdFromOtherRail stealing rxd %p\n", rail->Generic.Name, rail);
	
	/* Remove the RXD from the pending desc list */
	if ((next = elan3_sdram_readl (rail->Device, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Next))) == 0)
	    rcvrRail->RcvrMain->PendingDescsTailp = 0;
	elan3_sdram_writel (rail->Device, rcvrRail->RcvrElan + offsetof (EP3_RCVR_RAIL_ELAN, PendingDescs), next);
	UnlockRcvrThread (rcvrRail);
	
	UnbindRxdFromRail (rxd, rxdRail);
	
	spin_unlock_irqrestore (&rcvr->Lock, flags);
	
	/* Mark rxdRail as no longer active */
	rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
	rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;
	elan3_sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);
	elan3_sdram_writel (rail->Device, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count), 0);
	
	FreeRxdRail (rcvrRail, rxdRail);

	return rxd;
    }

    UnlockRcvrThread (rcvrRail);
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    return NULL;
}

long
ep3rcvr_check (EP_RCVR_RAIL *r, long nextRunTime)
{
    EP3_RCVR_RAIL    *rcvrRail = (EP3_RCVR_RAIL *) r;
    EP3_RAIL         *rail     = RCVR_TO_RAIL (rcvrRail);
    EP_RCVR          *rcvr     = rcvrRail->Generic.Rcvr;
    EP_COMMS_SUBSYS *subsys    = rcvr->Subsys;
    EP_SYS           *sys       = subsys->Subsys.Sys;
    EP_RXD           *rxd;
    unsigned long     flags;

    if (rcvrRail->FreeDescCount < ep_rxd_lowat && !AllocateRxdRailBlock (rcvrRail))
    {
	EPRINTF1 (DBG_RCVR,"%s: failed to grow rxd rail pool\n", rail->Generic.Name);
		
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }
    
    if (rcvrRail->ThreadWaiting && (rxd = StealRxdFromOtherRail (rcvr)) != NULL)
    {
	/* Map the receive buffer into this rail as well */
	EPRINTF4 (DBG_RCVR, "%s: mapping rxd->Data (%08x.%08x.%08x) into this rails\n",
		  rail->Generic.Name, rxd->Data.nmd_addr,rxd->Data.nmd_len, rxd->Data.nmd_attr);

	spin_lock_irqsave (&rcvr->Lock, flags);
	if ((!(EP_NMD_RAILMASK (&rxd->Data) & EP_RAIL2RAILMASK(rail->Generic.Number)) &&		/* not already mapped and */
	     ep_nmd_map_rails (sys, &rxd->Data, EP_RAIL2RAILMASK(rail->Generic.Number)) == 0) ||	/* failed to map it */
	    ep3rcvr_queue_rxd (rxd, &rcvrRail->Generic))						/* or failed to queue it */
	{
	    EPRINTF5 (DBG_RCVR,"%s: stolen rcvr=%p rxd=%p -> rnum=%d rcvrRail=%p (failed)\n", 
		      rail->Generic.Name, rcvr, rxd, rail->Generic.Number, rcvrRail);
		
	    if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
		nextRunTime = lbolt + RESOURCE_RETRY_TIME;
	}
	spin_unlock_irqrestore (&rcvr->Lock, flags);
    }
    
    return nextRunTime;
}

static void
ep3rcvr_flush_filtering (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail)
{
    EP3_COMMS_RAIL *commsRail = (EP3_COMMS_RAIL *) rcvrRail->Generic.CommsRail;
    EP3_RAIL       *rail      = (EP3_RAIL *) commsRail->Generic.Rail;
    ELAN3_DEV      *dev       = rail->Device;
    sdramaddr_t    qdesc      = commsRail->QueueDescs + rcvr->Service*sizeof (EP3_InputQueue);
    E3_Addr        qTop       = elan3_sdram_readl (dev, qdesc + offsetof (EP3_InputQueue, q_top));
    E3_Addr        qBase      = elan3_sdram_readl (dev, qdesc + offsetof (EP3_InputQueue, q_base));
    E3_Addr        qSize      = elan3_sdram_readl (dev,qdesc + offsetof (EP3_InputQueue, q_size));
    E3_uint32      nfptr, qbptr;
    unsigned long  flags;
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    LockRcvrThread (rcvrRail);										/* PCI lock */

    nfptr = elan3_sdram_readl (dev, qdesc + offsetof (EP3_InputQueue, q_fptr));
    qbptr = elan3_sdram_readl (dev, qdesc + offsetof (EP3_InputQueue, q_bptr));
    
    if (nfptr == qTop)
	nfptr = qBase;
    else
	nfptr += qSize;
    
    while (nfptr != qbptr)
    {
	unsigned nodeId = elan3_sdram_readl (dev, rcvrRail->InputQueueBase + (nfptr - rcvrRail->InputQueueAddr) + 
				       offsetof (EP_ENVELOPE, NodeId));
	
	EPRINTF3 (DBG_DISCON, "%s: ep3rcvr_flush_filtering: nodeId=%d State=%d\n", rail->Generic.Name, nodeId, rail->Generic.Nodes[nodeId].State);
	
	if (rail->Generic.Nodes[nodeId].State == EP_NODE_LOCAL_PASSIVATE)
	    elan3_sdram_writel (dev, rcvrRail->InputQueueBase + (nfptr - rcvrRail->InputQueueAddr) + 
			  offsetof (EP_ENVELOPE, Version), 0);
	
	if (nfptr == qTop)
	    nfptr = qBase;
	else
	    nfptr += qSize;
    }
    
    UnlockRcvrThread (rcvrRail);										/* PCI unlock */
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

static void
ep3rcvr_flush_flushing (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RAIL         *rail = RCVR_TO_RAIL (rcvrRail);
    struct list_head *el, *nel;
    unsigned long     flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    LockRcvrThread (rcvrRail);										/* PCI lock */
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd      = list_entry (el, EP_RXD, Link);
	EP3_RXD_RAIL *rxdRail  = (EP3_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE  *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[env->NodeId];

	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || nodeRail->State != EP_NODE_LOCAL_PASSIVATE)
	    continue;
	
	EPRINTF6 (DBG_DISCON, "%s: ep3rcvr_flush_flushing: rcvr %p rxd %p state %x.%x elan node %d\n", rail->Generic.Name,
		  rcvr, rxd, rxdRail->RxdMain->DataEvent, rxdRail->RxdMain->DoneEvent, env->NodeId);
	
	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep3rcvr_flush_flushing: rxd state is free but bound to a fail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->RxdMain->DataEvent == EP3_EVENT_ACTIVE)		/* incomplete message receive */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - passive\n", 
			  rail->Generic.Name, rcvr, rxd, env->NodeId);
		
		nodeRail->MessageState |= EP_NODE_PASSIVE_MESSAGES;
		continue;
	    }
	    break;
	    
	default:
	    EP_ASSERT (&rail->Generic, EP_IS_RPC(env->Attr));

	    if (!EP3_EVENT_FIRED (rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent))	/* incomplete RPC */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - active\n", 
			  rail->Generic.Name, rcvr, rxd, env->NodeId);
		
		EP_INVALIDATE_XID (rxd->MsgXid);			/* Ignore any previous NMD map responses */
		
		nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		continue;
	    }
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep3rcvr_flush_flushing: rxd state is aborted but bound to a fail\n");
	    break;
	}

	EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - finished\n", 
		  rail->Generic.Name, rcvr, rxd, env->NodeId);
    }    

    UnlockRcvrThread (rcvrRail); 									/* PCI unlock */
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep3rcvr_flush_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RAIL *rail = RCVR_TO_RAIL(rcvrRail);

    switch (rail->Generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	ep3rcvr_flush_filtering (rcvr, rcvrRail);
	break;

    case EP_CB_FLUSH_FLUSHING:
	ep3rcvr_flush_flushing (rcvr, rcvrRail);
	break;
    }
}

void
ep3rcvr_failover_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail)
{
    EP_COMMS_SUBSYS  *subsys = rcvr->Subsys;
    EP3_RAIL         *rail   = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV        *dev    = rail->Device;
    struct list_head *el, *nel;
    unsigned long     flags;
#ifdef SUPPORT_RAIL_FAILOVER
    EP_SYS           *sys    = subsys->Subsys.Sys;
#endif
   
    spin_lock_irqsave (&rcvr->Lock, flags);
    LockRcvrThread (rcvrRail);										/* PCI lock */
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD             *rxd      = list_entry (el, EP_RXD, Link);
	EP3_RXD_RAIL       *rxdRail  = (EP3_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE        *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL       *nodeRail = &rail->Generic.Nodes[env->NodeId];
#ifdef SUPPORT_RAIL_FAILOVER
	EP_MANAGER_MSG_BODY msgBody;
	EP_NODE            *node     = &sys->Nodes[env->NodeId];
#endif
	
	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || nodeRail->State != EP_NODE_PASSIVATED)
	    continue;

	EPRINTF6 (DBG_FAILOVER, "%s: ep3rcvr_failover_callback: rcvr %p rxd %p elan node %d state %x.%x\n", rail->Generic.Name, rcvr, rxd, env->NodeId,
		  rxdRail->RxdMain->DataEvent, rxdRail->RxdMain->DoneEvent);

	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep4rcvr_failover_callback: rxd state is free but bound to a fail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->RxdMain->DataEvent == EP3_EVENT_ACTIVE)		/* incomplete message receive */
	    {
		EPRINTF4 (DBG_FAILOVER, "%s: ep3rcvr_failover_callback: rcvr %p rxd %p nodeId %d - unbind\n", rail->Generic.Name, rcvr, rxd, env->NodeId);
		
		UnbindRxdFromRail (rxd, rxdRail);
		
		/* clear the done flags - so that it will be ignored if an event interrupt is generated */
		rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
		rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;
		
		/* clear the data event - the done event should already be zero */
		elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);	/* PCI write */
		
		FreeRxdRail (rcvrRail, rxdRail);
		
		/* epcomms thread will requeue on different rail */
		ep_kthread_schedule (&subsys->Thread, lbolt);
		continue;
	    }
	    break;

	default:
	    EP_ASSERT (&rail->Generic, EP_IS_RPC(env->Attr));

#ifdef SUPPORT_RAIL_FAILOVER
	    if (!EP3_EVENT_FIRED (rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent) && !(EP_IS_NO_FAILOVER(env->Attr)))		/* incomplete RPC, which can be failed over  */
	    {
		EPRINTF7 (DBG_FAILOVER, "%s: ep3rcvr_failover_callback: rxd %p State %x.%x Xid %llxx MsgXid %llxx nodeId %d - failover\n", 
			  rail->Generic.Name, rxd, rxdRail->RxdMain->DataEvent, rxdRail->RxdMain->DoneEvent, 
			  (long long) env->Xid.Unique, (long long) rxd->MsgXid.Unique, env->NodeId);
		
		if (EP_XID_INVALID(rxd->MsgXid))
		    rxd->MsgXid = ep_xid_cache_alloc (sys, &rcvr->XidCache);
		
		/* XXXX maybe only send the message if the node failover retry is now ? */
		msgBody.Failover.Xid      = env->Xid;
		msgBody.Failover.Railmask = node->ConnectedRails;
		
		ep_send_message (&rail->Generic, env->NodeId, EP_MANAGER_MSG_TYPE_FAILOVER_REQUEST, rxd->MsgXid, &msgBody);
		
		nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		continue;
	    }
#endif
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep3rcvr_failover_callback: rxd state is aborted but bound to a rail\n");
	    break;
	}

	EPRINTF3 (DBG_FAILOVER, "%s: ep3rcvr_failover_callback: rxd %p nodeId %d - finished\n", rail->Generic.Name, rxd, env->NodeId);
    }
    
    UnlockRcvrThread (rcvrRail); 									/* PCI unlock */
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep3rcvr_disconnect_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail)
{
    EP3_RAIL         *rail = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV        *dev = rail->Device;
    struct list_head *el, *nel;
    struct list_head  rxdList;
    unsigned long     flags;

    INIT_LIST_HEAD (&rxdList);
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    LockRcvrThread (rcvrRail);										/* PCI lock */
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd      = list_entry (el, EP_RXD, Link);
	EP3_RXD_RAIL *rxdRail  = (EP3_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE  *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[env->NodeId];
	
	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || nodeRail->State != EP_NODE_DISCONNECTING)
	    continue;

	EPRINTF4 (DBG_DISCON, "%s: ep3rcvr_disconnect_callback: rcvr %p rxd %p elan node %d\n", rail->Generic.Name, rcvr, rxd, env->NodeId);

	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep3rcvr_disconnect_callback: rxd state is free but bound to a fail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->RxdMain->DataEvent == EP3_EVENT_ACTIVE) 			/* incomplete message receive */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - unbind\n", rail->Generic.Name, rcvr, rxd, env->NodeId);
		
		UnbindRxdFromRail (rxd, rxdRail);
		
		/* clear the done flags - so that it will be ignored if an event interrupt is generated */
		rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
		rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;
		
		/* clear the data event - the done event should already be zero */
		elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);	/* PCI write */
		
		FreeRxdRail (rcvrRail, rxdRail);

		/* remark it as pending if it was partially received */
		rxd->RxdMain->Len = EP_RXD_PENDING;
		
		/* epcomms thread will requeue on different rail */
		ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
		continue;
	    }
	    break;

	default:
	    EP_ASSERT (&rail->Generic, EP_IS_RPC(env->Attr));

	    if (!EP3_EVENT_FIRED (rxdRail->DoneCookie, rxdRail->RxdMain->DoneEvent)) 	/* incomplete RPC */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - not able to failover\n",
			  rail->Generic.Name, rcvr, rxd, env->NodeId);
	    
		/* Mark as no longer active */
		rxdRail->RxdMain->DataEvent = EP3_EVENT_PRIVATE;
		rxdRail->RxdMain->DoneEvent = EP3_EVENT_PRIVATE;
		
		elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count), 0);	/* PCI write */
		elan3_sdram_writel (dev, rxdRail->RxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count), 0);	/* PCI write */
		
		UnbindRxdFromRail (rxd, rxdRail);
		FreeRxdRail (rcvrRail, rxdRail);

		/* Ignore any previous NMD/failover responses */
		EP_INVALIDATE_XID (rxd->MsgXid);
		
		/* Remove from active list */
		list_del (&rxd->Link);
		
		if (rxd->State == EP_RXD_RPC_IN_PROGRESS)				/* ownder by user .... */
		    rxd->State = EP_RXD_BEEN_ABORTED;
		else									/* queue for completion */
		{
		    rxd->RxdMain->Len = EP_CONN_RESET;					/* ensure ep_rxd_status() fails */
		    list_add_tail (&rxd->Link, &rxdList);
		}
		continue;
	    }
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep4rcvr_failover_callback: rxd state is aborted but bound to a fail\n");
	    break;
	}
	    
	EPRINTF4 (DBG_RCVR, "%s: ep3rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - finished\n", 
		  rail->Generic.Name, rcvr, rxd, env->NodeId);
    }
    
    UnlockRcvrThread (rcvrRail); 									/* PCI unlock */
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    while (! list_empty (&rxdList)) 
    {
	EP_RXD *rxd = list_entry (rxdList.next, EP_RXD, Link);

	list_del (&rxd->Link);

	rxd->Handler (rxd);
    }
}

void
ep3rcvr_display_rxd (DisplayInfo *di, EP_RXD_RAIL *r)
{
    EP3_RXD_RAIL *rxdRail = (EP3_RXD_RAIL *) r;
    sdramaddr_t   rxdElan = rxdRail->RxdElan;
    EP3_RAIL     *rail    = RCVR_TO_RAIL (rxdRail->Generic.RcvrRail);
    ELAN3_DEV    *dev     = rail->Device;

    (di->func)(di->arg, "      ChainEvent=%x.%x %x.%x\n",
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[0].ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[0].ev_Type)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[1].ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[1].ev_Type)));
    (di->func)(di->arg, "      ChainEvent=%x.%x %x.%x\n",
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[2].ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[2].ev_Type)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[3].ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, ChainEvent[3].ev_Type)));
    (di->func)(di->arg, "      DataEvent=%x.%x DoneEvent=%x.%x\n",
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DataEvent.ev_Type)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Count)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, DoneEvent.ev_Type)));
    (di->func)(di->arg, "      Data=%x Len=%x\n",
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_addr)),
	       elan3_sdram_readl (dev, rxdElan + offsetof (EP3_RXD_RAIL_ELAN, Data.nmd_len)));
}

void
ep3rcvr_display_rcvr (DisplayInfo *di, EP_RCVR_RAIL *r)
{
    EP3_RCVR_RAIL  *rcvrRail  = (EP3_RCVR_RAIL *) r;
    EP3_COMMS_RAIL *commsRail = (EP3_COMMS_RAIL *) rcvrRail->Generic.CommsRail;
    EP3_RAIL       *rail      = RCVR_TO_RAIL (rcvrRail);
    ELAN3_DEV      *dev       = rail->Device;
    sdramaddr_t     queue     = commsRail->QueueDescs + rcvrRail->Generic.Rcvr->Service * sizeof (EP3_InputQueue);
    E3_Addr         qbase      = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_base));
    E3_Addr         qtop       = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_top));
    E3_uint32       qsize      = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_size));
    int             freeCount  = 0;
    int             blockCount = 0;
    unsigned long   flags;
    struct list_head *el;

    spin_lock_irqsave (&rcvrRail->FreeDescLock, flags);
    list_for_each (el, &rcvrRail->FreeDescList)
	freeCount++;
    list_for_each (el, &rcvrRail->DescBlockList)
	blockCount++;
    spin_unlock_irqrestore (&rcvrRail->FreeDescLock, flags);

    (di->func)(di->arg, "                 Rail %d FreeDesc %d (%d) Total %d Blocks %d %s\n",
	       rail->Generic.Number, rcvrRail->FreeDescCount, freeCount, rcvrRail->TotalDescCount, blockCount, 
	       rcvrRail->ThreadWaiting ? "ThreadWaiting" : "");
    
    (di->func)(di->arg, "                 InputQueue state=%x bptr=%x size=%x top=%x base=%x fptr=%x\n",
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_state)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_bptr)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_size)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_top)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_base)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_fptr)));
    (di->func)(di->arg, "                            event=%x.%x [%x.%x] wevent=%x.%x\n",
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_event.ev_Type)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_event.ev_Count)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_event.ev_Source)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_event.ev_Dest)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_wevent)),
	       elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_wcount)));
    
    LockRcvrThread (rcvrRail);
    {
	E3_Addr     nfptr = elan3_sdram_readl (dev, queue + offsetof (EP3_InputQueue, q_fptr));
	EP_ENVELOPE env;
	
	if (nfptr == qtop)
	    nfptr = qbase;
	else
	    nfptr += qsize;

	while (nfptr != elan3_sdram_readl (dev, queue + offsetof (E3_Queue, q_bptr)))
	{
	    elan3_sdram_copyl_from_sdram (dev, rcvrRail->InputQueueBase + (nfptr - rcvrRail->InputQueueAddr),
					  &env, sizeof (EP_ENVELOPE));
	    
	    (di->func)(di->arg, "                 ENVELOPE Version=%x Attr=%x Xid=%08x.%08x.%016llx\n",
		       env.Version, env.Attr, env.Xid.Generation, env.Xid.Handle, (long long) env.Xid.Unique);
	    (di->func)(di->arg, "                          NodeId=%x Range=%x TxdRail=%x TxdMain=%x.%x.%x\n",
		       env.NodeId, env.Range, env.TxdRail, env.TxdMain.nmd_addr,
		       env.TxdMain.nmd_len, env.TxdMain.nmd_attr);
	    
	    
	    if (nfptr == qtop)
		nfptr = qbase;
	    else
		nfptr += qsize;
	}
    }
    UnlockRcvrThread (rcvrRail);
}

void
ep3rcvr_fillout_rail_stats(EP_RCVR_RAIL *rcvr_rail, char *str) {
    /* no stats here yet */
    /* EP3_RCVR_RAIL * ep4rcvr_rail = (EP3_RCVR_RAIL *) rcvr_rail; */
}

