/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsRx_elan4.c,v 1.35.2.1 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsRx_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"
#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "epcomms_elan4.h"

#include <elan4/trtype.h>

#define RCVR_TO_COMMS(rcvrRail)		((EP4_COMMS_RAIL *) ((EP_RCVR_RAIL *) rcvrRail)->CommsRail)
#define RCVR_TO_RAIL(rcvrRail)		((EP4_RAIL *) ((EP_RCVR_RAIL *) rcvrRail)->CommsRail->Rail)
#define RCVR_TO_DEV(rcvrRail)		(RCVR_TO_RAIL(rcvrRail)->r_ctxt.ctxt_dev)
#define RCVR_TO_SUBSYS(rcvrRail)	(((EP_RCVR_RAIL *) rcvrRail)->Rcvr->Subsys)

#define RXD_TO_RCVR(txdRail)		((EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail)
#define RXD_TO_RAIL(txdRail)		RCVR_TO_RAIL(RXD_TO_RCVR(rxdRail))

static void rxd_interrupt (EP4_RAIL *rail, void *arg);

static __inline__ void 
__ep4_rxd_assert_free (EP4_RXD_RAIL *rxdRail, const char *file, const int line)
{
    EP4_RCVR_RAIL *rcvrRail = RXD_TO_RCVR(rxdRail);
    ELAN4_DEV     *dev      = RCVR_TO_DEV(rcvrRail);
    register int i, failed = 0;
    
    for (i = 0; i <= EP_MAXFRAG; i++)
	if (((rxdRail)->rxd_main->rxd_sent[i] != EP4_STATE_FREE)) 
	    failed |= (1 << i);
    
    if (((rxdRail)->rxd_main->rxd_failed != EP4_STATE_FREE))
	failed |= (1 << 5);
    if (((rxdRail)->rxd_main->rxd_done   != EP4_STATE_FREE)) 
	failed |= (1 << 6);
    
    if (sdram_assert)
    {
	if (((elan4_sdram_readq (RXD_TO_RAIL(rxdRail)->r_ctxt.ctxt_dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CountAndType)) >> 32) != 0)) 
	    failed |= (1 << 7);
	for (i = 0; i < EP_MAXFRAG; i++)
	    if (((elan4_sdram_readq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[i].ev_CountAndType)) >> 32) != 0)) 
		failed |= (1 << (8 + i));
	if (((elan4_sdram_readq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType)) >> 32) != 0)) 
	    failed |= (1 << 12);
	if (((int)(elan4_sdram_readq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType)) >> 32) != -32)) 
	    failed |= (1 << 13);
    }

    if (failed)
    {
	printk ("__ep4_rxd_assert_free: failed=%x rxdRail=%p %s - %d\n", failed, rxdRail, file, line);

	ep_debugf (DBG_DEBUG, "__ep4_rxd_assert_free: failed=%x rxdRail=%p %s - %d\n", failed, rxdRail, file, line);
	ep4rcvr_display_rxd (&di_ep_debug, &rxdRail->rxd_generic);

	for (i = 0; i <= EP_MAXFRAG; i++)
	    (rxdRail)->rxd_main->rxd_sent[i] = EP4_STATE_FREE;

	(rxdRail)->rxd_main->rxd_failed = EP4_STATE_FREE;
	(rxdRail)->rxd_main->rxd_done   = EP4_STATE_FREE;

	if (sdram_assert)
	{
	    elan4_sdram_writew (RXD_TO_RAIL(rxdRail)->r_ctxt.ctxt_dev,
				(rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CountAndType) + 4, 0);

	    for (i = 0; i < EP_MAXFRAG; i++)
		elan4_sdram_writew (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[i].ev_CountAndType) + 4, 0);
	    elan4_sdram_writew (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType) + 4, 0);
	    elan4_sdram_writew (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType) + 4, -32);
	}
	EP_ASSFAIL (RCVR_TO_RAIL(rcvrRail), "__ep4_rxd_assert_free");
    }
}

static __inline__ void
__ep4_rxd_assert_pending(EP4_RXD_RAIL *rxdRail, const char *file, const int line)
{ 
    EP4_RCVR_RAIL *rcvrRail = RXD_TO_RCVR(rcvrRail);
    register int failed = 0;

    failed |= ((rxdRail)->rxd_main->rxd_done != EP4_STATE_ACTIVE);

    if (failed)
    {
	printk ("__ep4_rxd_assert_pending: %s - %d\n", file, line);

	ep_debugf (DBG_DEBUG, "__ep4_rxd_assert_pending: %s - %d\n", file, line);
	ep4rcvr_display_rxd (&di_ep_debug, &rxdRail->rxd_generic);

	(rxdRail)->rxd_main->rxd_done = EP4_STATE_ACTIVE;

	EP_ASSFAIL (RCVR_TO_RAIL(rcvrRail), "__ep4_rxd_assert_pending");
    }
}

static __inline__ void
__ep4_rxd_assert_private(EP4_RXD_RAIL *rxdRail, const char *file, const int line)
{
    EP4_RCVR_RAIL *rcvrRail = RXD_TO_RCVR(rxdRail);
    ELAN4_DEV     *dev      = RCVR_TO_DEV(rcvrRail);
    register int failed = 0;

    if (((rxdRail)->rxd_main->rxd_failed != EP4_STATE_ACTIVE)) failed |= (1 << 0);
    if (((rxdRail)->rxd_main->rxd_done != EP4_STATE_PRIVATE))  failed |= (1 << 1);
    
    if (sdram_assert)
    {
	if (((elan4_sdram_readq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType)) >> 32) != 0))           failed |= (1 << 2);
	if (((int) (elan4_sdram_readq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType)) >> 32) != -32)) failed |= (1 << 3);
    }

    if (failed)
    {
	printk ("__ep4_rxd_assert_private: failed=%x rxdRail=%p %s - %d\n", failed, rxdRail, file, line);

	ep_debugf (DBG_DEBUG, "__ep4_rxd_assert_private: failed=%x rxdRail=%p %s - %d\n", failed, rxdRail, file, line);
	ep4rcvr_display_rxd (&di_ep_debug, &rxdRail->rxd_generic);

	(rxdRail)->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
	(rxdRail)->rxd_main->rxd_done   = EP4_STATE_PRIVATE;

	if (sdram_assert)
	{
	    elan4_sdram_writew (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType) + 4, 0);
	    elan4_sdram_writew (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType) + 4, -32);
	}

	EP_ASSFAIL (RCVR_TO_RAIL(rcvrRail), "__ep4_rxd_assert_private");
    }
}

static __inline__ void
__ep4_rxd_private_to_free (EP4_RXD_RAIL *rxdRail)
{
    register int i;

    for (i = 0; i <= EP_MAXFRAG; i++)
	rxdRail->rxd_main->rxd_sent[i] = EP4_STATE_FREE;
 
    rxdRail->rxd_main->rxd_failed = EP4_STATE_FREE;
    rxdRail->rxd_main->rxd_done   = EP4_STATE_FREE;
}

static __inline__ void
__ep4_rxd_force_private (EP4_RXD_RAIL *rxdRail)
{
    EP4_RAIL  *rail = RXD_TO_RAIL(rxdRail);
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;

    (rxdRail)->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
    (rxdRail)->rxd_main->rxd_done = EP4_STATE_PRIVATE;

    if (sdram_assert) 
	elan4_sdram_writeq (dev, (rxdRail)->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType),
			    E4_EVENT_INIT_VALUE(0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
}

#define EP4_RXD_ASSERT_FREE(rxdRail)		__ep4_rxd_assert_free(rxdRail, __FILE__, __LINE__)
#define EP4_RXD_ASSERT_PENDING(rxdRail)		__ep4_rxd_assert_pending(rxdRail, __FILE__, __LINE__)
#define EP4_RXD_ASSERT_PRIVATE(rxdRail)		__ep4_rxd_assert_private(rxdRail, __FILE__, __LINE__)
#define EP4_RXD_PRIVATE_TO_FREE(rxdRail) 	__ep4_rxd_private_to_free(rxdRail)
#define EP4_RXD_FORCE_PRIVATE(rxdRail)		__ep4_rxd_force_private(rxdRail)

static int
alloc_rxd_block (EP4_RCVR_RAIL *rcvrRail)
{
    EP4_RAIL           *rail = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV          *dev  = rail->r_ctxt.ctxt_dev;
    EP4_RXD_RAIL_BLOCK *blk;
    EP4_RXD_RAIL_MAIN  *rxdMain;
    EP_ADDR		rxdMainAddr;
    sdramaddr_t		rxdElan;
    EP_ADDR		rxdElanAddr;
    EP4_RXD_RAIL       *rxdRail;
    unsigned long       flags;
    int                 i, j;

    KMEM_ZALLOC (blk, EP4_RXD_RAIL_BLOCK *, sizeof (EP4_RXD_RAIL_BLOCK), 1);

    if (blk == NULL)
	return 0;

    if ((rxdElan = ep_alloc_elan (&rail->r_generic, EP4_RXD_RAIL_ELAN_SIZE * EP4_NUM_RXD_PER_BLOCK, 0, &rxdElanAddr)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP4_RXD_RAIL_BLOCK));
	return 0;
    }

    if ((rxdMain = ep_alloc_main (&rail->r_generic, EP4_RXD_RAIL_MAIN_SIZE * EP4_NUM_RXD_PER_BLOCK, 0, &rxdMainAddr)) == (EP4_RXD_RAIL_MAIN *) NULL)
    {
	ep_free_elan (&rail->r_generic, rxdElanAddr, EP4_RXD_RAIL_ELAN_SIZE * EP4_NUM_RXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP4_RXD_RAIL_BLOCK));
	return 0;
    }

    if (ep4_reserve_dma_retries (rail, EP4_NUM_RXD_PER_BLOCK, 0) != 0)
    {
	ep_free_main (&rail->r_generic, blk->blk_rxds[0].rxd_main_addr, EP4_RXD_RAIL_MAIN_SIZE * EP4_NUM_RXD_PER_BLOCK);
	ep_free_elan (&rail->r_generic, rxdElanAddr, EP4_RXD_RAIL_ELAN_SIZE * EP4_NUM_RXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP4_RXD_RAIL_BLOCK));

	return 0;
    }

    for (rxdRail = &blk->blk_rxds[0], i = 0; i < EP4_NUM_RXD_PER_BLOCK; i++, rxdRail++)
    {
	rxdRail->rxd_generic.RcvrRail = &rcvrRail->rcvr_generic;
	rxdRail->rxd_elan             = rxdElan;
	rxdRail->rxd_elan_addr        = rxdElanAddr;
	rxdRail->rxd_main             = rxdMain;
	rxdRail->rxd_main_addr        = rxdMainAddr;

	/* reserve 128 bytes of "event" cq space for the chained STEN packets */
	if ((rxdRail->rxd_ecq = ep4_get_ecq (rail, EP4_ECQ_EVENT, EP4_RXD_STEN_CMD_NDWORDS)) == NULL)
	    goto failed;

	/* allocate a single word of "setevent" command space */
	if ((rxdRail->rxd_scq = ep4_get_ecq (rail, EP4_ECQ_SINGLE, 1)) == NULL)
	{
	    ep4_put_ecq (rail, rxdRail->rxd_ecq, EP4_RXD_STEN_CMD_NDWORDS);
	    goto failed;
	}

	/* initialise the completion events */
	for (j = 0; j <= EP_MAXFRAG; j++)
	    rxdMain->rxd_sent[i] = EP4_STATE_FREE;

	rxdMain->rxd_done   = EP4_STATE_FREE;
	rxdMain->rxd_failed = EP4_STATE_FREE;

	/* initialise the scq for the thread */
	rxdMain->rxd_scq = rxdRail->rxd_scq->ecq_addr;

	/* initialise the "start" event to copy the first STEN packet into the command queue */
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CountAndType),
			    E4_EVENT_INIT_VALUE(0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_START_CMD_NDWORDS));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CopySource),
			    rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0]));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CopyDest),
			    rxdRail->rxd_ecq->ecq_addr);

	/* initialise the "chain" events to copy the next STEN packet into the command queue */
	for (j = 0; j < EP_MAXFRAG; j++)
	{
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[j].ev_CountAndType),
				E4_EVENT_INIT_VALUE(0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[j].ev_CopySource),
				rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j+1]));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[j].ev_CopyDest),
				rxdRail->rxd_ecq->ecq_addr);
	}

	/* initialise the portions of the sten packets which don't change */
	for (j = 0; j < EP_MAXFRAG+1; j++)
	{
	    if (j < EP_MAXFRAG)
		elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_dma_dstEvent),
				    rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[j]));
	    else
		elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_dma_dstEvent),
				    rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done));

	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_ok_guard),
				GUARD_CMD | GUARD_CHANNEL (1) | GUARD_TEST(0, PACK_OK) | GUARD_RESET (EP4_RXD_STEN_RETRYCOUNT));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_ok_write_cmd),
				WRITE_DWORD_CMD | (rxdMainAddr + offsetof (EP4_RXD_RAIL_MAIN, rxd_sent[j])));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_ok_write_value),
				EP4_STATE_FINISHED);
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_fail_guard),
				GUARD_CMD | GUARD_CHANNEL (1) | GUARD_TEST(0, RESTART_COUNT_ZERO) | GUARD_RESET (EP4_RXD_STEN_RETRYCOUNT));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_fail_setevent),
				SET_EVENT_CMD | (rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed)));
	    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[j].c_nop_cmd),
				NOP_CMD);
	}

	/* register a main interrupt cookie */
	ep4_register_intcookie (rail, &rxdRail->rxd_intcookie, rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_done),
				rxd_interrupt, rxdRail);

	/* initialise the command stream for the done event */
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done_cmd.c_write_cmd),
			    WRITE_DWORD_CMD | (rxdMainAddr + offsetof (EP4_RXD_RAIL_MAIN, rxd_done)));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done_cmd.c_write_value),
			    EP4_STATE_FINISHED);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done_cmd.c_intr_cmd),
			    INTERRUPT_CMD | (rxdRail->rxd_intcookie.int_val << E4_MAIN_INT_SHIFT));

	/* initialise the command stream for the fail event */
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed_cmd.c_write_cmd),
			    WRITE_DWORD_CMD | (rxdMainAddr + offsetof (EP4_RXD_RAIL_MAIN, rxd_failed)));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed_cmd.c_write_value),
			    EP4_STATE_FAILED);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed_cmd.c_intr_cmd),
			    INTERRUPT_CMD | (rxdRail->rxd_intcookie.int_val << E4_MAIN_INT_SHIFT));

	/* initialise the done and fail events */
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType),
			    E4_EVENT_INIT_VALUE(0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CopySource),
			    rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_done_cmd));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CopyDest),
			    rxdRail->rxd_ecq->ecq_addr);

	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType),
			    E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CopySource),
			    rxdElanAddr + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed_cmd));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CopyDest),
			    rxdRail->rxd_ecq->ecq_addr);
	
	/* initialise the pointer to the main memory portion */
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_main), 
			    rxdMainAddr);

	/* move onto next descriptor */
	rxdElan     += EP4_RXD_RAIL_ELAN_SIZE;
	rxdElanAddr += EP4_RXD_RAIL_ELAN_SIZE;
	rxdMain      = (EP4_RXD_RAIL_MAIN *) ((unsigned long) rxdMain + EP4_RXD_RAIL_MAIN_SIZE);
	rxdMainAddr += EP4_RXD_RAIL_MAIN_SIZE;
    }

    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);

    list_add  (&blk->blk_link, &rcvrRail->rcvr_blocklist);

    rcvrRail->rcvr_totalcount += EP4_NUM_RXD_PER_BLOCK;
    rcvrRail->rcvr_freecount  += EP4_NUM_RXD_PER_BLOCK;

    for (i = 0; i < EP4_NUM_RXD_PER_BLOCK; i++)
	list_add (&blk->blk_rxds[i].rxd_generic.Link, &rcvrRail->rcvr_freelist);

    spin_unlock_irqrestore (&rcvrRail->rcvr_freelock, flags);

    return 1;

 failed:
    while (--i >= 0)
    {
	rxdRail--;

	ep4_put_ecq (rail, rxdRail->rxd_ecq, EP4_RXD_STEN_CMD_NDWORDS);
	ep4_put_ecq (rail, rxdRail->rxd_scq, 1);

	ep4_deregister_intcookie (rail, &rxdRail->rxd_intcookie);
    }

    ep4_release_dma_retries (rail, EP4_NUM_RXD_PER_BLOCK);
    
    ep_free_main (&rail->r_generic, blk->blk_rxds[0].rxd_main_addr, EP4_RXD_RAIL_MAIN_SIZE * EP4_NUM_RXD_PER_BLOCK);
    ep_free_elan (&rail->r_generic, rxdElanAddr, EP4_RXD_RAIL_ELAN_SIZE * EP4_NUM_RXD_PER_BLOCK);
    KMEM_FREE (blk, sizeof (EP4_RXD_RAIL_BLOCK));

    return 0;
}


static void
free_rxd_block (EP4_RCVR_RAIL *rcvrRail, EP4_RXD_RAIL_BLOCK *blk)
{
    EP4_RAIL     *rail = RCVR_TO_RAIL (rcvrRail);
    EP4_RXD_RAIL *rxdRail;
    unsigned long flags;
    int           i;

    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);

    list_del (&blk->blk_link);

    rcvrRail->rcvr_totalcount -= EP4_NUM_RXD_PER_BLOCK;

    for (rxdRail = &blk->blk_rxds[0], i = 0; i < EP4_NUM_RXD_PER_BLOCK; i++, rxdRail++)
    {
	rcvrRail->rcvr_freecount--;

	ep4_put_ecq (rail, rxdRail->rxd_ecq, EP4_RXD_STEN_CMD_NDWORDS);
	ep4_put_ecq (rail, rxdRail->rxd_scq, 1);

	ep4_deregister_intcookie (rail, &rxdRail->rxd_intcookie);

	list_del (&rxdRail->rxd_generic.Link);
    }
    spin_unlock_irqrestore (&rcvrRail->rcvr_freelock, flags);

    ep4_release_dma_retries (rail, EP4_NUM_RXD_PER_BLOCK);

    ep_free_main (&rail->r_generic, blk->blk_rxds[0].rxd_main_addr, EP4_RXD_RAIL_MAIN_SIZE * EP4_NUM_RXD_PER_BLOCK);
    ep_free_elan (&rail->r_generic, blk->blk_rxds[0].rxd_elan_addr, EP4_RXD_RAIL_ELAN_SIZE * EP4_NUM_RXD_PER_BLOCK);

    KMEM_FREE (blk, sizeof (EP4_RXD_RAIL_BLOCK));
}

static EP4_RXD_RAIL *
get_rxd_rail (EP4_RCVR_RAIL *rcvrRail)
{
    EP_COMMS_SUBSYS  *subsys = RCVR_TO_SUBSYS(rcvrRail);
    EP4_RXD_RAIL     *rxdRail;
    unsigned long flags;
    int low_on_rxds;

    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);

    if (list_empty (&rcvrRail->rcvr_freelist))
	rxdRail = NULL;
    else
    {
	rxdRail = list_entry (rcvrRail->rcvr_freelist.next, EP4_RXD_RAIL, rxd_generic.Link);

	EP4_RXD_ASSERT_FREE(rxdRail);

	list_del (&rxdRail->rxd_generic.Link);

	rcvrRail->rcvr_freecount--;
    }
    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_rxds = (rcvrRail->rcvr_freecount < ep_rxd_lowat);

    spin_unlock_irqrestore (&rcvrRail->rcvr_freelock, flags);

    if (low_on_rxds)
	ep_kthread_schedule (&subsys->Thread, lbolt);

    return (rxdRail);
}

static void
free_rxd_rail (EP4_RCVR_RAIL *rcvrRail, EP4_RXD_RAIL *rxdRail)
{
    unsigned long flags;

    EP4_RXD_ASSERT_FREE(rxdRail);

    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);
    
    list_add (&rxdRail->rxd_generic.Link, &rcvrRail->rcvr_freelist);

    rcvrRail->rcvr_freecount++;

    if (rcvrRail->rcvr_freewaiting)
    {
	rcvrRail->rcvr_freewaiting--;
	kcondvar_wakeupall (&rcvrRail->rcvr_freesleep, &rcvrRail->rcvr_freelock);
    }

    spin_unlock_irqrestore (&rcvrRail->rcvr_freelock, flags);
}

static void
bind_rxd_rail (EP_RXD *rxd, EP4_RXD_RAIL *rxdRail)
{
    EP4_RAIL *rail = RCVR_TO_RAIL (rxdRail->rxd_generic.RcvrRail);

    ASSERT (SPINLOCK_HELD (&rxd->Rcvr->Lock));

    EPRINTF3 (DBG_RCVR, "%s: bind_rxd_rail: rxd=%p rxdRail=%p\n",  rail->r_generic.Name, rxd, rxdRail);

    elan4_sdram_writeq (rail->r_ctxt.ctxt_dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_rxd), rxd->NmdMain.nmd_addr);			/* PCI write */

    rxd->RxdRail             = &rxdRail->rxd_generic;
    rxdRail->rxd_generic.Rxd = rxd;
}

static void
unbind_rxd_rail (EP_RXD *rxd, EP4_RXD_RAIL *rxdRail)
{
    EP4_RCVR_RAIL *rcvrRail = (EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail;
    
    ASSERT (SPINLOCK_HELD (&rxd->Rcvr->Lock));
    ASSERT (rxd->RxdRail == &rxdRail->rxd_generic && rxdRail->rxd_generic.Rxd == rxd);

    EP4_RXD_ASSERT_PRIVATE (rxdRail);

    EPRINTF3 (DBG_RCVR, "%s: unbind_rxd_rail: rxd=%p rxdRail=%p\n",  RCVR_TO_RAIL(rcvrRail)->r_generic.Name, rxd, rxdRail);

    rxd->RxdRail             = NULL;
    rxdRail->rxd_generic.Rxd = NULL;

    if (rcvrRail->rcvr_cleanup_waiting)
	kcondvar_wakeupall (&rcvrRail->rcvr_cleanup_sleep, &rxd->Rcvr->Lock);
    rcvrRail->rcvr_cleanup_waiting = 0;

    EP4_RXD_PRIVATE_TO_FREE (rxdRail);
}


static void
rcvr_stall_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_RCVR_RAIL *rcvrRail = (EP4_RCVR_RAIL *) arg;
    EP_RCVR       *rcvr     = rcvrRail->rcvr_generic.Rcvr;
    unsigned long  flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    
    EPRINTF1 (DBG_RCVR, "rcvr_stall_interrupt: rcvrRail %p thread halted\n", rcvrRail);

    rcvrRail->rcvr_thread_halted = 1;

    kcondvar_wakeupall (&rcvrRail->rcvr_cleanup_sleep, &rcvr->Lock);

    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

static void
rcvr_stall_haltop (ELAN4_DEV *dev, void *arg)
{
    EP4_RCVR_RAIL  *rcvrRail  = (EP4_RCVR_RAIL *) arg;
    EP4_COMMS_RAIL *commsRail = RCVR_TO_COMMS(rcvrRail);
    EP_RCVR        *rcvr      = rcvrRail->rcvr_generic.Rcvr;
    sdramaddr_t     qdesc     = ((EP4_COMMS_RAIL *) commsRail)->r_descs + (rcvr->Service * EP_QUEUE_DESC_SIZE);
    E4_uint64       qbptr     = elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_bptr));

    /* Mark the queue as full by writing the fptr */
    if (qbptr == (rcvrRail->rcvr_slots_addr + EP_INPUTQ_SIZE * (rcvr->InputQueueEntries-1)))
	elan4_sdram_writeq (dev, qdesc + offsetof (E4_InputQueue, q_fptr), rcvrRail->rcvr_slots_addr);
    else
	elan4_sdram_writeq (dev, qdesc + offsetof (E4_InputQueue, q_fptr), qbptr + EP_INPUTQ_SIZE);

    /* Notify the thread that it should stall after processing any outstanding envelopes */
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_stall_intcookie),
			rcvrRail->rcvr_stall_intcookie.int_val);

    /* Issue a swtevent to the queue event to wake the thread up */
    ep4_set_event_cmd (rcvrRail->rcvr_resched, rcvrRail->rcvr_elan_addr + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qevent));
}

static void
rxd_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_RXD_RAIL      *rxdRail  = (EP4_RXD_RAIL *) arg;
    EP4_RCVR_RAIL     *rcvrRail = (EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail;
    EP_RCVR           *rcvr     = rcvrRail->rcvr_generic.Rcvr;
    EP4_RXD_RAIL_MAIN *rxdMain  = rxdRail->rxd_main;
    unsigned long      delay    = 1;
    EP_RXD            *rxd;
    EP_ENVELOPE       *env;
    unsigned long      flags;

    spin_lock_irqsave (&rcvr->Lock, flags);

    for (;;)
    {
	if (rxdMain->rxd_done == EP4_STATE_FINISHED || rxdMain->rxd_failed == EP4_STATE_FAILED)
	    break;

	/* The write to rxd_done could be held up in the PCI bridge even though
	 * we've seen the interrupt cookie.  Unlike elan3, there is no possibility
	 * of spurious interrupts since we flush the command queues on node 
	 * disconnection and the txcallback mechanism */
	mb();

	if (delay > EP4_EVENT_FIRING_TLIMIT)
	{
	    spin_unlock_irqrestore (&rcvr->Lock, flags);

	    EP_ASSFAIL (RCVR_TO_RAIL(rcvrRail), "rxd_interrupt - not finished\n");
	    return;
	}
	DELAY(delay);
	delay <<= 1;
    }

    if (rxdMain->rxd_done != EP4_STATE_FINISHED)
    {
	EPRINTF8 (DBG_RETRY, "%s: rxd_interrupt: rxdRail %p retry: done=%d failed=%d NodeId=%d XID=%08x.%08x.%016llx\n",
		  rail->r_generic.Name, rxdRail, (int)rxdMain->rxd_done, (int)rxdMain->rxd_failed, rxdRail->rxd_generic.Rxd->RxdMain->Envelope.NodeId,
		  rxdRail->rxd_generic.Rxd->RxdMain->Envelope.Xid.Generation, rxdRail->rxd_generic.Rxd->RxdMain->Envelope.Xid.Handle, 
		  (long long)rxdRail->rxd_generic.Rxd->RxdMain->Envelope.Xid.Unique);
    
	spin_lock (&rcvrRail->rcvr_retrylock);

	rxdRail->rxd_retry_time = lbolt + EP_RETRY_LOW_PRI_TIME;			/* XXXX backoff ? */

	list_add_tail (&rxdRail->rxd_retry_link, &rcvrRail->rcvr_retrylist);

	ep_kthread_schedule (&rail->r_retry_thread, rxdRail->rxd_retry_time);
	spin_unlock (&rcvrRail->rcvr_retrylock);

	spin_unlock_irqrestore (&rcvr->Lock, flags);
	return;
    }
    
    rxd = rxdRail->rxd_generic.Rxd;
    env = &rxd->RxdMain->Envelope;

    /*
     * Note, since the thread will have sent the remote dma packet before copying 
     * the envelope, we must check that it has completed doing this,  we do this
     * by acquiring the spinlock against the thread which it only drops once it's
     * completed.
     */
    if (rxd->RxdMain->Len == EP_RXD_PENDING)
    {
	EP4_SPINENTER (rail->r_ctxt.ctxt_dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock),
		       &rcvrRail->rcvr_main->rcvr_thread_lock);
	
	EP4_SPINEXIT (rail->r_ctxt.ctxt_dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock),
		      &rcvrRail->rcvr_main->rcvr_thread_lock);
	
	EP4_ASSERT (rail, env->Version == EP_ENVELOPE_VERSION && rxd->RxdMain->Len != EP_RXD_PENDING);
    }

    EPRINTF8 (DBG_RCVR, "%s: rxd_interrupt: rxd %p finished from %d XID %08x.%08x.%016llx len %d attr %x\n", rail->r_generic.Name, 
	      rxd, rxd->RxdMain->Envelope.NodeId, rxd->RxdMain->Envelope.Xid.Generation, rxd->RxdMain->Envelope.Xid.Handle, 
	      (long long)rxd->RxdMain->Envelope.Xid.Unique,  rxd->RxdMain->Len, rxd->RxdMain->Envelope.Attr);

    rxdMain->rxd_done  = EP4_STATE_PRIVATE;
    rxd->Data.nmd_attr = EP_RAIL2RAILMASK (rail->r_generic.Number);

    switch (rxd->State)
    {
    case EP_RXD_RECEIVE_ACTIVE:
	if (rxd->RxdMain->Len >= 0 && EP_IS_RPC(env->Attr))
	    rxd->State = EP_RXD_RPC_IN_PROGRESS;
	else
	{
	    rxd->State = EP_RXD_COMPLETED;

	    /* remove from active list */
	    list_del (&rxd->Link);

	    unbind_rxd_rail (rxd, rxdRail);
	    free_rxd_rail (rcvrRail, rxdRail);
	}

	if (rxd->RxdMain->Len >= 0) {
	    INC_STAT(rcvrRail->rcvr_generic.stats,rx);
	    ADD_STAT(rcvrRail->rcvr_generic.stats,rx_len,rxd->RxdMain->Len);
	    INC_STAT(rail->r_generic.Stats,rx);
	    ADD_STAT(rail->r_generic.Stats,rx_len,rxd->RxdMain->Len);
	}
	spin_unlock_irqrestore (&rcvr->Lock, flags);
	ep_rxd_received (rxd);

	break;

    case EP_RXD_PUT_ACTIVE:
    case EP_RXD_GET_ACTIVE:
	rxd->State = EP_RXD_RPC_IN_PROGRESS;
	spin_unlock_irqrestore (&rcvr->Lock, flags);
	
	rxd->Handler (rxd);
	break;

    case EP_RXD_COMPLETE_ACTIVE:
	rxd->State = EP_RXD_COMPLETED;

	/* remove from active list */
	list_del (&rxd->Link);

	unbind_rxd_rail (rxd, rxdRail);
	free_rxd_rail (rcvrRail, rxdRail);

	spin_unlock_irqrestore (&rcvr->Lock, flags);

	rxd->Handler(rxd);
	break;

    default:
	spin_unlock_irqrestore (&rcvr->Lock, flags);

	printk ("%s: rxd_interrupt: rxd %p in invalid state %d\n", rail->r_generic.Name, rxd, rxd->State);
	/* NOTREACHED */
    }
}

static void
ep4rcvr_flush_filtering (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail)
{
    EP4_COMMS_RAIL *commsRail = RCVR_TO_COMMS(rcvrRail);
    EP4_RAIL       *rail      = RCVR_TO_RAIL(rcvrRail);
    ELAN4_DEV      *dev       = rail->r_ctxt.ctxt_dev;
    sdramaddr_t    qdesc      = commsRail->r_descs + (rcvr->Service * EP_QUEUE_DESC_SIZE);
    E4_Addr        qbase      = rcvrRail->rcvr_slots_addr;
    E4_Addr        qlast      = qbase + EP_INPUTQ_SIZE * (rcvr->InputQueueEntries-1);
    E4_uint64      qfptr, qbptr;
    unsigned long  flags;
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    
    /* zip down the input queue and invalidate any envelope we find to a node which is locally passivated */
    qfptr = elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_fptr));
    qbptr = elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_bptr));

    while (qfptr != qbptr)
    {
	unsigned int nodeId = elan4_sdram_readl (dev, rcvrRail->rcvr_slots + (qfptr - qbase) + offsetof (EP_ENVELOPE, NodeId));

	EPRINTF3 (DBG_DISCON, "%s: ep4rcvr_flush_filtering: nodeId=%d State=%d\n", rail->r_generic.Name, nodeId, rail->r_generic.Nodes[nodeId].State);
	
	if (rail->r_generic.Nodes[nodeId].State == EP_NODE_LOCAL_PASSIVATE)
	    elan4_sdram_writel (dev,  rcvrRail->rcvr_slots + (qfptr - qbase) + offsetof (EP_ENVELOPE, Version), 0);
	
	if (qfptr != qlast)
	    qfptr += EP_INPUTQ_SIZE;
	else
	    qfptr = qbase;
    }

    /* Insert an setevent command into the thread's command queue
     * to ensure that all sten packets have completed */
    elan4_guard (rcvrRail->rcvr_ecq->ecq_cq, GUARD_ALL_CHANNELS);
    ep4comms_flush_setevent (commsRail, rcvrRail->rcvr_ecq->ecq_cq);
    
    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

static void
ep4rcvr_flush_flushing (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail)
{
    EP4_RAIL         *rail = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV	     *dev  = rail->r_ctxt.ctxt_dev;
    struct list_head *el, *nel;
    struct list_head  rxdList;
    unsigned long     flags;

    INIT_LIST_HEAD (&rxdList);
    
    /* remove any sten packates which are retrying to nodes which are being passivated */
    spin_lock_irqsave (&rcvrRail->rcvr_retrylock, flags);
    list_for_each_safe (el, nel, &rcvrRail->rcvr_retrylist) {
	EP4_RXD_RAIL *rxdRail  = list_entry (el, EP4_RXD_RAIL, rxd_retry_link);
	EP_ENVELOPE  *env      = &rxdRail->rxd_generic.Rxd->RxdMain->Envelope;
	EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[env->NodeId];

	if (nodeRail->State == EP_NODE_LOCAL_PASSIVATE)
	{
	    EPRINTF2 (DBG_XMTR, "%s; ep4rcvr_flush_flushing: removing rxdRail %p from retry list\n", rail->r_generic.Name, rxdRail);
	    
	    list_del (&rxdRail->rxd_retry_link);
	}
    }
    spin_unlock_irqrestore (&rcvrRail->rcvr_retrylock, flags);

    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd      = list_entry (el, EP_RXD, Link);
	EP4_RXD_RAIL *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE  *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[env->NodeId];

	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL (rxdRail, rcvrRail) || nodeRail->State != EP_NODE_LOCAL_PASSIVATE)
	    continue;
	
	EPRINTF6 (DBG_DISCON, "%s: ep4rcvr_flush_flushing: rcvr %p rxd %p state %d elan node %d state %d\n", 
		  rail->r_generic.Name, rcvr, rxd, (int)rxdRail->rxd_main->rxd_done, env->NodeId, rxd->State);
	
	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep4rcvr_flush_flushing: rxd state is free but bound to a fail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE)		/* incomplete message receive */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep4rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - passive\n", 
			  rail->r_generic.Name, rcvr, rxd, env->NodeId);
		
		nodeRail->MessageState |= EP_NODE_PASSIVE_MESSAGES;
		continue;
	    }
	    break;
	    
	default:
	    EP4_ASSERT (rail, EP_IS_RPC(env->Attr));

	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE)		/* incomplete RPC */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep4rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - active\n", 
			  rail->r_generic.Name, rcvr, rxd, env->NodeId);
		
		EP_INVALIDATE_XID (rxd->MsgXid);			/* Ignore any previous NMD map responses */
		
		nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		continue;
	    }
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep4rcvr_flush_flushing: rxd state is aborted but bound to a fail\n");
	    break;
	}

	EPRINTF4 (DBG_RCVR, "%s: ep4rcvr_flush_flushing: rcvr %p rxd %p nodeId %d - finished\n", 
		  rail->r_generic.Name, rcvr, rxd, env->NodeId);
    }    

    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep4rcvr_flush_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail)
{
    EP4_RAIL *rail = RCVR_TO_RAIL(rcvrRail);

    switch (rail->r_generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	ep4rcvr_flush_filtering (rcvr, rcvrRail);
	break;

    case EP_CB_FLUSH_FLUSHING:
	ep4rcvr_flush_flushing (rcvr, rcvrRail);
	break;
    }
}

void
ep4rcvr_failover_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail)
{
    EP_COMMS_SUBSYS  *subsys = rcvr->Subsys;
    EP4_RAIL         *rail   = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV	     *dev    = rail->r_ctxt.ctxt_dev;
    struct list_head *el, *nel;
    unsigned long     flags;
#if SUPPORT_RAIL_FAILOVER
    EP_SYS           *sys    = subsys->Subsys.Sys;
#endif
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD             *rxd      = list_entry (el, EP_RXD, Link);
	EP4_RXD_RAIL       *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE        *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL       *nodeRail = &rail->r_generic.Nodes[env->NodeId];
#if SUPPORT_RAIL_FAILOVER
	EP_NODE            *node     = &sys->Nodes[env->NodeId];
	EP_MANAGER_MSG_BODY msgBody;
#endif
	
	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || nodeRail->State != EP_NODE_PASSIVATED)
	    continue;

	EPRINTF5 (DBG_FAILOVER, "%s: ep4rcvr_failover_callback: rcvr %p rxd %p elan node %d state %d\n", 
		  rail->r_generic.Name, rcvr, rxd, env->NodeId, (int)rxdRail->rxd_main->rxd_done);

	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep4rcvr_failover_callback: rxd state is free but bound to a rail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE)			/* incomplete message receive */
	    {
		EPRINTF4 (DBG_FAILOVER, "%s: ep4rcvr_failover_callback: rcvr %p rxd %p nodeId %d - unbind\n", rail->r_generic.Name, rcvr, rxd, env->NodeId);

		EP4_RXD_FORCE_PRIVATE(rxdRail);
		
		unbind_rxd_rail (rxd, rxdRail);

		free_rxd_rail (rcvrRail, rxdRail);
	    
		/* epcomms thread will requeue on different rail */
		ep_kthread_schedule (&subsys->Thread, lbolt);
		continue;
	    }
	    break;

	default:
	    EP4_ASSERT (rail, EP_IS_RPC(env->Attr));

#if SUPPORT_RAIL_FAILOVER
	    /* XXXX - no rail failover for now .... */
	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE && !EP_IS_NO_FAILOVER(env->Attr))	/* incomplete RPC, which can be failed over */
	    {
		EPRINTF6 (DBG_FAILOVER, "%s: ep4rcvr_failover_callback: rxd %p State %d Xid %llxx MsgXid %llxx nodeId %d - failover\n", 
			  rail->r_generic.Name, rxd, rxd->State, (long long)env->Xid.Unique, (long long)rxd->MsgXid.Unique, env->NodeId);
		
		if (EP_XID_INVALID(rxd->MsgXid))
		    rxd->MsgXid = ep_xid_cache_alloc (sys, &rcvr->XidCache);
		
		/* XXXX maybe only send the message if the node failover retry is now ? */
		msgBody.Failover.Xid      = env->Xid;
		msgBody.Failover.Railmask = node->ConnectedRails;
		
		ep_send_message (&rail->r_generic, env->NodeId, EP_MANAGER_MSG_TYPE_FAILOVER_REQUEST, rxd->MsgXid, &msgBody);
		
		nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		continue;
	    }
#endif
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep4rcvr_failover_callback: rxd state is aborted but bound to a fail\n");
	    break;
	}
	EPRINTF3 (DBG_FAILOVER, "%s: ep4rcvr_failover_callback: rxd %p nodeId %d - finished\n", rail->r_generic.Name, rxd, env->NodeId);
    }
    
    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep4rcvr_disconnect_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail)
{
    EP4_RAIL         *rail = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV        *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el, *nel;
    struct list_head  rxdList;
    unsigned long     flags;

    INIT_LIST_HEAD (&rxdList);
    
    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd      = list_entry (el, EP_RXD, Link);
	EP4_RXD_RAIL *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE  *env      = &rxd->RxdMain->Envelope;
	EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[env->NodeId];
	
	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || nodeRail->State != EP_NODE_DISCONNECTING)
	    continue;

	EPRINTF5 (DBG_DISCON, "%s: ep4rcvr_disconnect_callback: rcvr %p rxd %p elan node %d state %x\n", rail->r_generic.Name, rcvr, rxd, env->NodeId, rxd->State);

	switch (rxd->State)
	{
	case EP_RXD_FREE:
	    printk ("ep4rcvr_disconnect_callback: rxd state is free but bound to a rail\n");
	    break;

	case EP_RXD_RECEIVE_ACTIVE:
	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE)		/* incomplete message receive */
	    {
		EPRINTF4 (DBG_RCVR, "%s: ep4rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - unbind\n", rail->r_generic.Name, rcvr, rxd, env->NodeId);

		EP4_RXD_FORCE_PRIVATE (rxdRail);
		
		unbind_rxd_rail (rxd, rxdRail);
		free_rxd_rail (rcvrRail, rxdRail);
		
		/* remark it as pending if it was partially received */
		rxd->RxdMain->Len = EP_RXD_PENDING;
		
		/* epcomms thread will requeue on different rail */
		ep_kthread_schedule (&rcvr->Subsys->Thread, lbolt);
		continue;
	    }
	    break;

	default:
	    if (rxdRail->rxd_main->rxd_done == EP4_STATE_ACTIVE || rxdRail->rxd_main->rxd_done == EP4_STATE_PRIVATE)		/* incomplete RPC */
	    {
		EPRINTF5 (DBG_RCVR, "%s: ep4rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d state %x - not able to failover\n",
			  rail->r_generic.Name, rcvr, rxd, env->NodeId, rxd->State);
	    
		EP4_RXD_FORCE_PRIVATE (rxdRail);

		unbind_rxd_rail (rxd, rxdRail);
		free_rxd_rail (rcvrRail, rxdRail);

		/* Ignore any previous NMD/failover responses */
		EP_INVALIDATE_XID (rxd->MsgXid);
		
		/* Remove from active list */
		list_del (&rxd->Link);
		
		if (rxd->State == EP_RXD_RPC_IN_PROGRESS)				/* ownder by user .... */
		    rxd->State = EP_RXD_BEEN_ABORTED;
		else									/* queue for completion */
		{
		    rxd->RxdMain->Len = EP_CONN_RESET;					/* ensure ep_rxd_status() fails */
		    list_add_tail (&rxd->Link, &rxdList);
		}
		continue;
	    }
	    break;

	case EP_RXD_BEEN_ABORTED:
	    printk ("ep4rcvr_disconnect_callback: rxd state is aborted but bound to a rail\n");
	    break;
	}

	printk ("%s: ep4rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - finished\n", 
		  rail->r_generic.Name, rcvr, rxd, env->NodeId);
	EPRINTF4 (DBG_RCVR, "%s: ep4rcvr_disconnect_callback: rcvr %p rxd %p nodeId %d - finished\n", 
		  rail->r_generic.Name, rcvr, rxd, env->NodeId);
	ep4rcvr_display_rxd (&di_ep_debug, &rxdRail->rxd_generic);
    }
    
    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    while (! list_empty (&rxdList)) 
    {
	EP_RXD *rxd = list_entry (rxdList.next, EP_RXD, Link);

	list_del (&rxd->Link);

	rxd->Handler (rxd);
    }
}

void
ep4rcvr_neterr_flush (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP4_COMMS_RAIL *commsRail = RCVR_TO_COMMS(rcvrRail);
    EP4_RAIL       *rail      = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV      *dev       = rail->r_ctxt.ctxt_dev;
    unsigned long   flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);

    /* Insert an setevent command into the thread's command queue
     * to ensure that all sten packets have completed */
    elan4_guard (rcvrRail->rcvr_ecq->ecq_cq, GUARD_ALL_CHANNELS);
    ep4comms_flush_setevent (commsRail, rcvrRail->rcvr_ecq->ecq_cq);
    
    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

void
ep4rcvr_neterr_check (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP4_RAIL         *rail = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV        *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el;
    unsigned long     flags;

    spin_lock_irqsave (&rcvr->Lock, flags);
    EP4_SPINENTER (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    
    list_for_each (el, &rcvr->ActiveDescList) {
	EP_RXD       *rxd      = list_entry (el, EP_RXD, Link);
	EP4_RXD_RAIL *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
	EP_ENVELOPE  *env      = &rxd->RxdMain->Envelope;

	if (rxd->RxdMain->Len == EP_RXD_PENDING || !RXD_BOUND2RAIL(rxdRail,rcvrRail) || env->NodeId != nodeId)
	    continue;

	if (rxd->State == EP_RXD_RECEIVE_ACTIVE || rxd->State == EP_RXD_GET_ACTIVE)
	{
	    EP_NETERR_COOKIE cookie;
	    unsigned int     first, this;

	    if (rxd->State == EP_RXD_RECEIVE_ACTIVE)
		first = (EP_MAXFRAG+1) - (( EP_IS_MULTICAST(env->Attr) ? 1 : 0) + (env->nFrags == 0 ? 1 : env->nFrags));
	    else
		first = (EP_MAXFRAG+1) - rxd->nFrags;

	    for (this = first; this < (EP_MAXFRAG+1); this++)
		if (rxdRail->rxd_main->rxd_sent[this] == EP4_STATE_ACTIVE)
		    break;
	    
	    if (this > first)
	    {
		/* Look at the last completed STEN packet and if it's neterr cookie matches, then change
		 * the rxd to look the same as if the sten packet had failed and then schedule it for retry */
		cookie = elan4_sdram_readq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[--this].c_cookie));
		
		if (cookie == cookies[0] || cookie == cookies[1])
		{
		    EP_NETERR_COOKIE ncookie = ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_STEN;

		    EPRINTF6 (DBG_NETWORK_ERROR, "%s: ep4rcvr_neterr_check: cookie <%lld%s%s%s%s> matches rxd %p rxdRail %p this %d new cookie <%lld%s%s%s%s>\n",
			      rail->r_generic.Name, EP4_COOKIE_STRING(cookie), rxd, rxdRail, this, EP4_COOKIE_STRING(ncookie));
		    
		    printk ("%s: ep4rcvr_neterr_check: cookie <%lld%s%s%s%s> matches rxd %p rxdRail %p this %d new cookie <%lld%s%s%s%s>\n",
			    rail->r_generic.Name, EP4_COOKIE_STRING(cookie), rxd, rxdRail, this, EP4_COOKIE_STRING(ncookie));
		    
		    /* Allocate a new cookie for this sten packet, since this message could be received more than once.  
		     * If the second arrives after we've sucessfully sent the response and the packet completes, then we
		     * could try and requeue it after the next sten packet got nacked. */
		    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[this].c_cookie), ncookie);

		    rxdRail->rxd_main->rxd_sent[this] = EP4_STATE_ACTIVE;
		    rxdRail->rxd_main->rxd_failed     = EP4_STATE_FAILED;
		    
		    spin_lock (&rcvrRail->rcvr_retrylock);

		    EP4_ASSERT (rail, rxdRail->rxd_retry_time == 0);

		    rxdRail->rxd_retry_time = lbolt + EP_RETRY_LOW_PRI_TIME;
			
		    list_add_tail (&rxdRail->rxd_retry_link, &rcvrRail->rcvr_retrylist);
			
		    ep_kthread_schedule (&rail->r_retry_thread, rxdRail->rxd_retry_time);
		    
		    spin_unlock (&rcvrRail->rcvr_retrylock);
		}
	    }
	}
    }
    EP4_SPINEXIT (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), &rcvrRail->rcvr_main->rcvr_thread_lock);
    spin_unlock_irqrestore (&rcvr->Lock, flags);
}

int
ep4rcvr_queue_rxd (EP_RXD *rxd, EP_RCVR_RAIL *r)
{
    EP4_RCVR_RAIL *rcvrRail = (EP4_RCVR_RAIL *) r;
    EP4_RAIL      *rail     = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV     *dev      = rail->r_ctxt.ctxt_dev;
    EP4_RXD_RAIL  *rxdRail;
    register int   i;

    ASSERT (SPINLOCK_HELD(&rxd->Rcvr->Lock));

    if ((rxdRail = get_rxd_rail (rcvrRail)) == NULL)
	return 0;
    
    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->r_generic);

    EPRINTF6 (DBG_RCVR, "%s: ep4rcvr_queue_rxd: rcvr %p rxd %p rxdRail %p buffer %x len %x\n", 
	      rail->r_generic.Name, rxd->Rcvr, rxd, rxdRail, rxd->Data.nmd_addr, rxd->Data.nmd_len);

    /* bind the rxdRail and rxd together */
    bind_rxd_rail (rxd, rxdRail);

    elan4_sdram_writel (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_buffer.nmd_addr), rxd->Data.nmd_addr);	/* PCI write */
    elan4_sdram_writel (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_buffer.nmd_len),  rxd->Data.nmd_len);		/* PCI write */
    elan4_sdram_writel (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_buffer.nmd_attr), rxd->Data.nmd_attr);	/* PCI write */

    /* Mark as active */
    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType), 
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
    
    for (i = 0; i <= EP_MAXFRAG; i++)
	rxdRail->rxd_main->rxd_sent[i] = EP4_STATE_ACTIVE;

    rxdRail->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
    rxdRail->rxd_main->rxd_done = EP4_STATE_ACTIVE;

    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0]) + 0x00, /* %r0 */
			ep_symbol (&rail->r_threadcode, "c_queue_rxd"));
    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0]) + 0x10, /* %r2 */
			rcvrRail->rcvr_elan_addr);
    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0]) + 0x18, /* %r3 */
			rxdRail->rxd_elan_addr);

    elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CountAndType),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_START_CMD_NDWORDS));

    ep4_set_event_cmd (rxdRail->rxd_scq, rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_start));

    return 1;
}

void
ep4rcvr_rpc_put (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP4_RXD_RAIL    *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
    EP4_RCVR_RAIL   *rcvrRail = (EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail;
    EP4_RAIL        *rail     = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV	    *dev      = RCVR_TO_DEV (rcvrRail);
    sdramaddr_t     rxdElan   = rxdRail->rxd_elan;
    EP_ENVELOPE    *env       = &rxd->RxdMain->Envelope;
    unsigned long   first     = (EP_MAXFRAG+1) - nFrags;
    EP4_RXD_DMA_CMD cmd;
    register int    i, len;

    EP4_ASSERT (rail, rxd->State == EP_RXD_PUT_ACTIVE);
    EP4_ASSERT (rail, rxdRail->rxd_main->rxd_done == EP4_STATE_PRIVATE);
    EP4_SDRAM_ASSERT (rail, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType),
		      E4_EVENT_INIT_VALUE (0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->r_generic);

    /* Generate the DMA chain to put the data */
    for (i = 0, len = 0; i < nFrags; i++, len += local->nmd_len, local++, remote++)
    {
	cmd.c_dma_typeSize     = RUN_DMA_CMD | E4_DMA_TYPE_SIZE(local->nmd_len, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
	cmd.c_dma_cookie       = ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_DMA;
	cmd.c_dma_vproc        = EP_VP_DATA(env->NodeId);
	cmd.c_dma_srcAddr      = local->nmd_addr;
	cmd.c_dma_dstAddr      = remote->nmd_addr;
	if (i == (nFrags-1))
	    cmd.c_dma_srcEvent = rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_done);
	else
	    cmd.c_dma_srcEvent = rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]);
	cmd.c_dma_dstEvent     = 0;
	cmd.c_nop_cmd          = NOP_CMD;

	EPRINTF7 (DBG_RCVR, "%s: ep4rcvr_rpc_put: rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x\n", 
		  rail->r_generic.Name, rxd, (long long)env->Xid.Unique, i, local->nmd_addr, remote->nmd_addr, local->nmd_len);
	
	elan4_sdram_copyq_to_sdram (dev, &cmd, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i]), sizeof (EP4_RXD_DMA_CMD));
    }

    /* Initialise the event chain */
    for (i = 0; i < nFrags-1; i++)
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]), 
			    E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_DMA_CMD_NDWORDS));

    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    for (i = 0; i <= EP_MAXFRAG; i++)
	rxdRail->rxd_main->rxd_sent[i] = EP4_STATE_ACTIVE;

    rxdRail->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
    rxdRail->rxd_main->rxd_done = EP4_STATE_ACTIVE;

    /* Initialise the previous event to start the whole chain off */
    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_DMA_CMD_NDWORDS));

    EP4_ASSERT (rail, rail->r_generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->r_generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    /* finally issue the setevent to start the whole chain */
    ep4_set_event_cmd (rxdRail->rxd_scq, rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]));

    BucketStat (rxd->Rcvr->Subsys, RPCPut, len);
}    

void
ep4rcvr_rpc_get (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP4_RXD_RAIL    *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
    EP4_RCVR_RAIL   *rcvrRail = (EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail;
    EP4_RAIL        *rail     = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV	    *dev      = RCVR_TO_DEV (rcvrRail);
    sdramaddr_t      rxdElan  = rxdRail->rxd_elan;
    EP_ENVELOPE     *env      = &rxd->RxdMain->Envelope;
    unsigned long    first    = (EP_MAXFRAG+1) - nFrags;
    register int    i, len;

    EP4_ASSERT (rail, rxd->State == EP_RXD_GET_ACTIVE);
    EP4_ASSERT (rail, rxdRail->rxd_main->rxd_done == EP4_STATE_PRIVATE);
    EP4_SDRAM_ASSERT (rail, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType),
		      E4_EVENT_INIT_VALUE (0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->r_generic);

    /* Generate the DMA chain to get the data */
    for (i = 0, len = 0; i < nFrags; i++, len += local->nmd_len, local++, remote++)
    {
	EPRINTF7 (DBG_RCVR, "%s: ep4rcvr_rpc_get rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x\n", 
		  rail->r_generic.Name, rxd, (long long)env->Xid.Unique, i, remote->nmd_addr, local->nmd_addr, remote->nmd_len);
	
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_open),
			    OPEN_STEN_PKT_CMD | OPEN_PACKET(0, PACK_OK | RESTART_COUNT_ZERO, EP_VP_DATA(env->NodeId)));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_trans),
			    SEND_TRANS_CMD | ((TR_REMOTEDMA | TR_WAIT_FOR_EOP) << 16));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_cookie),
			    ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_STEN);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_typeSize),
			    E4_DMA_TYPE_SIZE (local->nmd_len, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_cookie),
			    ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_DMA);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_vproc),
			    EP_VP_DATA (rail->r_generic.Position.pos_nodeid));
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_srcAddr),
			    remote->nmd_addr);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_dstAddr),
			    local->nmd_addr);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_srcEvent),
			    0);
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i].c_dma_dstEvent),
			    i == (nFrags-1) ? rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_done) : 
			                      rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]));
    }

    /* Initialise the event chain */
    for (i = 0; i < nFrags-1; i++)
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]), 
			    E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS));

    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    for (i = 0; i <= EP_MAXFRAG; i++)
	rxdRail->rxd_main->rxd_sent[i] = EP4_STATE_ACTIVE;

    rxdRail->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
    rxdRail->rxd_main->rxd_done = EP4_STATE_ACTIVE;

    /* Initialise the previous event to start the whole chain off */
    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS));

    EP4_ASSERT (rail, rail->r_generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->r_generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    /* finally issue the setevent to start the whole chain */
    ep4_set_event_cmd (rxdRail->rxd_scq, rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]));

    BucketStat (rxd->Rcvr->Subsys, RPCPut, len);
}

void
ep4rcvr_rpc_complete (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags)
{
    EP4_RXD_RAIL    *rxdRail  = (EP4_RXD_RAIL *) rxd->RxdRail;
    EP4_RCVR_RAIL   *rcvrRail = (EP4_RCVR_RAIL *) rxdRail->rxd_generic.RcvrRail;
    EP4_RAIL        *rail     = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV	    *dev      = RCVR_TO_DEV (rcvrRail);
    sdramaddr_t     rxdElan   = rxdRail->rxd_elan;
    EP_ENVELOPE    *env       = &rxd->RxdMain->Envelope;
    unsigned long   first     = (EP_MAXFRAG+1) - nFrags - 1;
    EP4_RXD_DMA_CMD cmd;
    register int    i, len;

    EP4_ASSERT (rail, rxd->State == EP_RXD_COMPLETE_ACTIVE);
    EP4_ASSERT (rail, rxdRail->rxd_main->rxd_done == EP4_STATE_PRIVATE);
    EP4_SDRAM_ASSERT (rail, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType),
		      E4_EVENT_INIT_VALUE (0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->r_generic);

    /* Generate the DMA chain to put the data */
    for (i = 0, len = 0; i < nFrags; i++, len += local->nmd_len, local++, remote++)
    {
	cmd.c_dma_typeSize = RUN_DMA_CMD | E4_DMA_TYPE_SIZE(local->nmd_len, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
	cmd.c_dma_cookie   = ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_DMA;
	cmd.c_dma_vproc    = EP_VP_DATA(env->NodeId);
	cmd.c_dma_srcAddr  = local->nmd_addr;
	cmd.c_dma_dstAddr  = remote->nmd_addr;
	cmd.c_dma_srcEvent = rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]);
	cmd.c_dma_dstEvent = 0;
	cmd.c_nop_cmd      = NOP_CMD;

	EPRINTF7 (DBG_RCVR, "%s: ep4rcvr_rpc_complete: rxd %p [XID=%llx] idx=%d Source=%08x Dest=%08x Len=%x\n", 
		  rail->r_generic.Name, rxd, (long long)env->Xid.Unique, i, local->nmd_addr, remote->nmd_addr, local->nmd_len);

	elan4_sdram_copyq_to_sdram (dev, &cmd, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[first + i]), sizeof (EP4_RXD_DMA_CMD));
    }
    
    /* Initialise the status block dma */
    cmd.c_dma_typeSize = RUN_DMA_CMD | E4_DMA_TYPE_SIZE(EP_STATUSBLK_SIZE, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
    cmd.c_dma_cookie   = ep4_neterr_cookie (rail, env->NodeId) | EP4_COOKIE_DMA;
    cmd.c_dma_vproc    = EP_VP_DATA(env->NodeId);
    cmd.c_dma_srcAddr  = rxd->NmdMain.nmd_addr + offsetof (EP_RXD_MAIN, StatusBlk);
    cmd.c_dma_dstAddr  = env->TxdMain.nmd_addr + offsetof (EP_TXD_MAIN, StatusBlk);
    cmd.c_dma_srcEvent = rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_done);
    cmd.c_dma_dstEvent = env->TxdRail + offsetof (EP4_TXD_RAIL_ELAN, txd_done);;
    cmd.c_nop_cmd      = NOP_CMD;

    EPRINTF6 (DBG_RCVR, "%s: ep4rcvr_rpc_complete: rxd %p [XID=%llx] statusblk source=%08x dest=%08x len=%llx\n", 
	      rail->r_generic.Name, rxd, (long long)env->Xid.Unique, (int) cmd.c_dma_srcAddr, (int) cmd.c_dma_dstAddr, (long long)EP_STATUSBLK_SIZE);

    elan4_sdram_copyq_to_sdram (dev, &cmd, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[EP_MAXFRAG]), sizeof (EP4_RXD_DMA_CMD));

    /* Initialise the event chain */
    for (i = 0; i < nFrags; i++)
	elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first + i]), 
			    E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_DMA_CMD_NDWORDS));

    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

    for (i = 0; i <= EP_MAXFRAG; i++)
	rxdRail->rxd_main->rxd_sent[i] = EP4_STATE_ACTIVE;

    rxdRail->rxd_main->rxd_failed = EP4_STATE_ACTIVE;
    rxdRail->rxd_main->rxd_done = EP4_STATE_ACTIVE;

    /* Initialise the previous event to start the whole chain off */
    elan4_sdram_writeq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]),
			E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_DMA_CMD_NDWORDS));

    EP4_ASSERT (rail, rail->r_generic.Nodes[env->NodeId].State >= EP_NODE_CONNECTED && rail->r_generic.Nodes[env->NodeId].State <= EP_NODE_LOCAL_PASSIVATE);

    /* finally issue the setevent to start the whole chain */
    ep4_set_event_cmd (rxdRail->rxd_scq, rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]));

    BucketStat (rxd->Rcvr->Subsys, CompleteRPC, len);
}

EP_RXD *
ep4rcvr_steal_rxd (EP_RCVR_RAIL *r)
{
    /* XXXX - TBD */
    return NULL;
}

long
ep4rcvr_check (EP_RCVR_RAIL *r, long nextRunTime)
{
    EP4_RCVR_RAIL *rcvrRail = (EP4_RCVR_RAIL *) r;
    EP4_RAIL      *rail     = RCVR_TO_RAIL (rcvrRail);

    if (rcvrRail->rcvr_freecount < ep_rxd_lowat && !alloc_rxd_block (rcvrRail))
    {
	EPRINTF1 (DBG_RCVR,"%s: failed to grow rxd rail pool\n", rail->r_generic.Name);
		
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }
    
    return nextRunTime;
}

unsigned long
ep4rcvr_retry (EP4_RAIL *rail, void *arg, unsigned long nextRunTime)
{
    EP4_RCVR_RAIL *rcvrRail = (EP4_RCVR_RAIL *) arg;
    ELAN4_DEV     *dev      = RCVR_TO_DEV(rcvrRail);
    unsigned long  flags;

    spin_lock_irqsave (&rcvrRail->rcvr_retrylock, flags);
    while (! list_empty (&rcvrRail->rcvr_retrylist))
    {
	EP4_RXD_RAIL *rxdRail = list_entry (rcvrRail->rcvr_retrylist.next, EP4_RXD_RAIL, rxd_retry_link);
	EP_ENVELOPE  *env     = &rxdRail->rxd_generic.Rxd->RxdMain->Envelope;
	unsigned int  first   = (EP_MAXFRAG+1) - ((env->Attr & EP_MULTICAST ? 1 : 0) + (env->nFrags == 0 ? 1 : env->nFrags));
	
	if (BEFORE (lbolt, rxdRail->rxd_retry_time))
	{
	    if (nextRunTime == 0 || AFTER (nextRunTime, rxdRail->rxd_retry_time))
		nextRunTime = rxdRail->rxd_retry_time;

	    break;
	}

	list_del (&rxdRail->rxd_retry_link);
	rxdRail->rxd_retry_time = 0;

	/* determine which sten packet to resubmit */
	for (; first < (EP_MAXFRAG+1); first++)
	    if (rxdRail->rxd_main->rxd_sent[first] == EP4_STATE_ACTIVE)
		break;

	EPRINTF3 (DBG_RETRY, "%s: ep4rcvr_retry: rxdRail %p, reissuing sten[%d]\n", rail->r_generic.Name, rxdRail, first);

	/* re-initialise the fail event */
	elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

	rxdRail->rxd_main->rxd_failed = EP4_STATE_ACTIVE;

	/* re-initialise the chain event to resubmit this sten packet */
	elan4_sdram_writeq (dev, rxdRail->rxd_elan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first-1].ev_CountAndType),
			    E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS));
	
	/* finally issue the setevent to start the chain again */
	ep4_set_event_cmd (rxdRail->rxd_scq, rxdRail->rxd_elan_addr + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[first - 1]));
    }
    spin_unlock_irqrestore (&rcvrRail->rcvr_retrylock, flags);
    
    return nextRunTime;
}

void
ep4rcvr_add_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *commsRail)
{
    EP4_RAIL          *rail   = (EP4_RAIL *) commsRail->Rail;
    ELAN4_DEV         *dev    = rail->r_ctxt.ctxt_dev;
    sdramaddr_t        qdescs = ((EP4_COMMS_RAIL *) commsRail)->r_descs;
    EP4_RCVR_RAIL     *rcvrRail;
    E4_InputQueue      qdesc;
    E4_ThreadRegs      tregs;
    sdramaddr_t        stack;
    unsigned long      flags;

    KMEM_ZALLOC (rcvrRail, EP4_RCVR_RAIL *, sizeof (EP4_RCVR_RAIL), 1);

    spin_lock_init (&rcvrRail->rcvr_freelock);
    INIT_LIST_HEAD (&rcvrRail->rcvr_freelist);
    INIT_LIST_HEAD (&rcvrRail->rcvr_blocklist);

    kcondvar_init (&rcvrRail->rcvr_cleanup_sleep);
    kcondvar_init (&rcvrRail->rcvr_freesleep);

    INIT_LIST_HEAD (&rcvrRail->rcvr_retrylist);
    spin_lock_init (&rcvrRail->rcvr_retrylock);

    rcvrRail->rcvr_generic.CommsRail = commsRail;
    rcvrRail->rcvr_generic.Rcvr      = rcvr;

    rcvrRail->rcvr_main  = ep_alloc_main (&rail->r_generic, sizeof (EP4_RCVR_RAIL_MAIN), 0, &rcvrRail->rcvr_main_addr);
    rcvrRail->rcvr_elan  = ep_alloc_elan (&rail->r_generic, sizeof (EP4_RCVR_RAIL_ELAN), 0, &rcvrRail->rcvr_elan_addr);
    rcvrRail->rcvr_slots = ep_alloc_elan (&rail->r_generic, EP_INPUTQ_SIZE * rcvr->InputQueueEntries, 0, &rcvrRail->rcvr_slots_addr);
    stack                = ep_alloc_elan (&rail->r_generic, EP4_STACK_SIZE, 0, &rcvrRail->rcvr_stack);

    /* allocate a command queue for the thread to use, plus space for it to wait/reschedule */
    rcvrRail->rcvr_ecq     = ep4_alloc_ecq (rail, CQ_Size64K);
    rcvrRail->rcvr_resched = ep4_get_ecq (rail, EP4_ECQ_ATOMIC, 8);

    ep4_register_intcookie (rail, &rcvrRail->rcvr_stall_intcookie, rcvrRail->rcvr_elan_addr + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_stall_intcookie),
			    rcvr_stall_interrupt, rcvrRail);

    /* Initialise the elan portion */
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qevent.ev_CountAndType), 0);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_halt.ev_CountAndType), 0);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock), 0);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_pending_tailp),
			rcvrRail->rcvr_elan_addr + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_pending_head));
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_pending_head), 0);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_stall_intcookie), 0);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qbase), rcvrRail->rcvr_slots_addr);
    elan4_sdram_writeq (dev, rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qlast), 
			rcvrRail->rcvr_slots_addr + EP_INPUTQ_SIZE * (rcvr->InputQueueEntries-1));

    /* Initialise the main memory portion */
    rcvrRail->rcvr_main->rcvr_thread_lock = 0;

    /* Install our retry handler */
    rcvrRail->rcvr_retryops.op_func = ep4rcvr_retry;
    rcvrRail->rcvr_retryops.op_arg  = rcvrRail;

    ep4_add_retry_ops (rail, &rcvrRail->rcvr_retryops);

    /* Update the queue desriptor */
    qdesc.q_bptr    = rcvrRail->rcvr_slots_addr;
    qdesc.q_fptr    = rcvrRail->rcvr_slots_addr;
    qdesc.q_control = E4_InputQueueControl (rcvrRail->rcvr_slots_addr, rcvrRail->rcvr_slots_addr + (EP_INPUTQ_SIZE * (rcvr->InputQueueEntries-1)), EP_INPUTQ_SIZE);
    qdesc.q_event   = rcvrRail->rcvr_elan_addr + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qevent);

    ep4_write_qdesc (rail, qdescs + (rcvr->Service * EP_QUEUE_DESC_SIZE), &qdesc);

    spin_lock_irqsave (&rcvr->Lock, flags);
    rcvr->Rails[rail->r_generic.Number] = &rcvrRail->rcvr_generic;
    rcvr->RailMask |= EP_RAIL2RAILMASK (rail->r_generic.Number);
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    {
	sdramaddr_t stackTop     = stack + EP4_STACK_SIZE;
	E4_Addr     stackTopAddr = rcvrRail->rcvr_stack + EP4_STACK_SIZE;

	ep4_init_thread (rail, &tregs, stackTop, stackTopAddr, ep_symbol (&rail->r_threadcode, "ep4comms_rcvr"), 6, 
			 (E4_uint64) rail->r_elan_addr, (E4_uint64) rcvrRail->rcvr_elan_addr, (E4_uint64) rcvrRail->rcvr_main_addr,
			 (E4_uint64) EP_MSGQ_ADDR(rcvr->Service), (E4_uint64) rcvrRail->rcvr_ecq->ecq_addr, (E4_uint64) rcvrRail->rcvr_resched->ecq_addr);
    }
    
    /* Issue the command to the threads private command queue */
    elan4_run_thread_cmd (rcvrRail->rcvr_ecq->ecq_cq, &tregs);

    ep_procfs_rcvr_add_rail(&(rcvrRail->rcvr_generic));
}

void
ep4rcvr_del_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *commsRail)
{
    EP4_RAIL         *rail     = (EP4_RAIL *) commsRail->Rail;
    EP4_RCVR_RAIL    *rcvrRail = (EP4_RCVR_RAIL *) rcvr->Rails[rail->r_generic.Number];  
    ELAN4_HALTOP      haltop;
    struct list_head *el, *nel;
    unsigned long     flags;

    ep_procfs_rcvr_del_rail(&(rcvrRail->rcvr_generic));

    /* Run a halt operation to mark the input queue as full and
     * request the thread to halt */
    haltop.op_mask     = INT_DiscardingHighPri | INT_TProcHalted;
    haltop.op_function = rcvr_stall_haltop;
    haltop.op_arg      = rcvrRail;

    elan4_queue_haltop (rail->r_ctxt.ctxt_dev, &haltop);

    /* Wait for the thread to tell us it's processed the input queue */
    spin_lock_irqsave (&rcvr->Lock, flags);
    while (! rcvrRail->rcvr_thread_halted)
	kcondvar_wait (&rcvrRail->rcvr_cleanup_sleep, &rcvr->Lock, &flags);
    rcvrRail->rcvr_thread_halted = 0;

    /* flag the rail as no longer available */
    rcvr->RailMask &= ~EP_RAIL2RAILMASK (rail->r_generic.Number);

    /* wait for all active communications to terminate */
    for (;;)
    {
	int mustWait = 0;

	list_for_each (el, &rcvr->ActiveDescList) {
	    EP_RXD       *rxd     = list_entry (el, EP_RXD, Link);
	    EP4_RXD_RAIL *rxdRail = (EP4_RXD_RAIL *) rxd->RxdRail;
	    
	    if (rxdRail && RXD_BOUND2RAIL (rxdRail, rcvrRail) && rxd->RxdMain->Len != EP_RXD_PENDING)
	    {
		mustWait++;
		break;
	    }
	}

	if (! mustWait)
	    break;

	rcvrRail->rcvr_cleanup_waiting++;
	kcondvar_wait (&rcvrRail->rcvr_cleanup_sleep, &rcvr->Lock, &flags);
    }

    /* at this point all rxd's in the list that are bound to the deleting rail are pending */
    list_for_each_safe (el, nel, &rcvr->ActiveDescList) {
	EP_RXD       *rxd     = list_entry (el, EP_RXD, Link);
	EP4_RXD_RAIL *rxdRail = (EP4_RXD_RAIL *) rxd->RxdRail;

	if (rxdRail && RXD_BOUND2RAIL (rxdRail, rcvrRail))
	{
	    EP4_RXD_ASSERT_PENDING (rxdRail);
	    EP4_RXD_FORCE_PRIVATE (rxdRail);

	    unbind_rxd_rail (rxd, rxdRail);
	    free_rxd_rail (rcvrRail, rxdRail);
	}
    }
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    /* wait for all rxd's for this rail to become free */
    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);
    while (rcvrRail->rcvr_freecount != rcvrRail->rcvr_totalcount)
    {
	rcvrRail->rcvr_freewaiting++;
	kcondvar_wait (&rcvrRail->rcvr_freesleep, &rcvrRail->rcvr_freelock, &flags);
    }
    spin_unlock_irqrestore (&rcvrRail->rcvr_freelock, flags);

    /* can now remove the rail as it can no longer be used */
    spin_lock_irqsave (&rcvr->Lock, flags);
    rcvr->Rails[rail->r_generic.Number] = NULL;
    spin_unlock_irqrestore (&rcvr->Lock, flags);

    /* all the rxd's accociated with DescBlocks must be in the FreeDescList */
    ASSERT (rcvrRail->rcvr_totalcount == rcvrRail->rcvr_freecount);

    /* run through the DescBlockList deleting them */
    while (!list_empty (&rcvrRail->rcvr_blocklist))
	free_rxd_block (rcvrRail, list_entry(rcvrRail->rcvr_blocklist.next, EP4_RXD_RAIL_BLOCK , blk_link));

    /* it had better be empty after that */
    ASSERT ((rcvrRail->rcvr_totalcount == 0) && (rcvrRail->rcvr_totalcount == rcvrRail->rcvr_freecount));

    ep4_remove_retry_ops (rail, &rcvrRail->rcvr_retryops);

    ep4_deregister_intcookie (rail, &rcvrRail->rcvr_stall_intcookie);

    ep4_put_ecq (rail, rcvrRail->rcvr_resched, 8);
    ep4_free_ecq (rail, rcvrRail->rcvr_ecq);

    ep_free_elan (&rail->r_generic, rcvrRail->rcvr_stack, EP4_STACK_SIZE);
    ep_free_elan (&rail->r_generic, rcvrRail->rcvr_slots_addr, EP_INPUTQ_SIZE * rcvr->InputQueueEntries);
    ep_free_elan (&rail->r_generic, rcvrRail->rcvr_elan_addr, sizeof (EP4_RCVR_RAIL_ELAN));
    ep_free_main (&rail->r_generic, rcvrRail->rcvr_main_addr, sizeof (EP4_RCVR_RAIL_MAIN));

    KMEM_FREE (rcvrRail, sizeof (EP4_RCVR_RAIL));
}

void
ep4rcvr_display_rxd (DisplayInfo *di, EP_RXD_RAIL *r)
{
    EP4_RXD_RAIL *rxdRail = (EP4_RXD_RAIL *) r;
    sdramaddr_t   rxdElan = rxdRail->rxd_elan;
    EP4_RAIL     *rail    = RCVR_TO_RAIL (rxdRail->rxd_generic.RcvrRail);
    ELAN4_DEV    *dev     = rail->r_ctxt.ctxt_dev;
    int i;

    (di->func)(di->arg, "    Rail %d rxd %p elan %lx(%x) main %p(%x) ecq %d scq %d debug %llx\n", rail->r_generic.Number,
	       rxdRail, rxdRail->rxd_elan, rxdRail->rxd_elan_addr, rxdRail->rxd_main, rxdRail->rxd_main_addr,
	       elan4_cq2num(rxdRail->rxd_ecq->ecq_cq), elan4_cq2num(rxdRail->rxd_scq->ecq_cq),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_debug)));
    (di->func)(di->arg, "          start    %016llx %016llx %016llx [%016llx %016llx]\n",
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_CountAndType)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_Params[0])),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_start.ev_Params[1])),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0].c_cookie)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[0].c_dma_cookie)));
	       
    for (i = 0; i < EP_MAXFRAG; i++)
	(di->func)(di->arg, "          chain[%d] %016llx %016llx %016llx [%016llx %016llx]\n", i,
		   elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[i].ev_CountAndType)),
		   elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[i].ev_Params[0])),
		   elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_chain[i].ev_Params[1])),
		   elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[i+1].c_cookie)),
		   elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_sten[i+1].c_dma_cookie)));
    (di->func)(di->arg, "          done    %016llx %016llx %016llx -> %016llx\n",
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_CountAndType)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_Params[0])),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_done.ev_Params[1])),
	       rxdRail->rxd_main->rxd_done);
    (di->func)(di->arg, "          fail    %016llx %016llx %016llx -> %016llx\n",
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_CountAndType)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_Params[0])),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_failed.ev_Params[1])),
	       rxdRail->rxd_main->rxd_failed);
    (di->func)(di->arg, "          next %016llx queued %016llx main %016llx\n",
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_next)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_queued)),
	       elan4_sdram_readq (dev, rxdElan + offsetof (EP4_RXD_RAIL_ELAN, rxd_main)));
    (di->func)(di->arg, "          sent %016llx %016llx %016llx %016llx %016llx\n",
	       rxdRail->rxd_main->rxd_sent[0], rxdRail->rxd_main->rxd_sent[1], rxdRail->rxd_main->rxd_sent[2],
	       rxdRail->rxd_main->rxd_sent[3], rxdRail->rxd_main->rxd_sent[4]);
}

void
ep4rcvr_display_rcvr (DisplayInfo *di, EP_RCVR_RAIL *r)
{
    EP_RCVR          *rcvr       = r->Rcvr;
    EP4_RCVR_RAIL    *rcvrRail   = (EP4_RCVR_RAIL *) r;
    EP4_COMMS_RAIL   *commsRail  = RCVR_TO_COMMS(rcvrRail);
    EP4_RAIL         *rail       = RCVR_TO_RAIL (rcvrRail);
    ELAN4_DEV        *dev        = rail->r_ctxt.ctxt_dev;
    sdramaddr_t       rcvrElan   = rcvrRail->rcvr_elan;
    sdramaddr_t       qdesc      = commsRail->r_descs + (rcvr->Service * EP_QUEUE_DESC_SIZE);
    sdramaddr_t       event      = rcvrRail->rcvr_elan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_qevent);
    unsigned int      freeCount  = 0;
    unsigned int      blockCount = 0;
    struct list_head *el;
    unsigned long     flags;
    
    spin_lock_irqsave (&rcvrRail->rcvr_freelock, flags);
    list_for_each (el, &rcvrRail->rcvr_freelist)
	freeCount++;
    list_for_each (el, &rcvrRail->rcvr_blocklist)
	blockCount++;
    spin_unlock_irqrestore(&rcvrRail->rcvr_freelock, flags);

    (di->func)(di->arg, "      Rail %d elan %lx(%x) main %p(%x) ecq %d resched %d debug %llx\n",
	       rail->r_generic.Number, rcvrRail->rcvr_elan, rcvrRail->rcvr_elan_addr,
	       rcvrRail->rcvr_main, rcvrRail->rcvr_main_addr, elan4_cq2num(rcvrRail->rcvr_ecq->ecq_cq),
	       elan4_cq2num (rcvrRail->rcvr_resched->ecq_cq),
	       elan4_sdram_readq (dev, rcvrElan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_debug)));
    (di->func)(di->arg, "        free %d (%d) total %d blocks %d\n",
	       rcvrRail->rcvr_freecount, freeCount, rcvrRail->rcvr_totalcount, blockCount);
    (di->func)(di->arg, "        spinlock %016llx %016llx\n", rcvrRail->rcvr_main->rcvr_thread_lock,
	       elan4_sdram_readq (dev, rcvrElan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_thread_lock)));
    (di->func)(di->arg, "        queue: bptr %016llx fptr %016llx control %016llx (base %lx %x)\n",
	       elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_bptr)),
	       elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_fptr)),
	       elan4_sdram_readq (dev, qdesc + offsetof (E4_InputQueue, q_control)),
	       rcvrRail->rcvr_slots, rcvrRail->rcvr_slots_addr);
    (di->func)(di->arg, "        event %016llx %016llx %016llx\n",
	       elan4_sdram_readq (dev, event + offsetof (E4_Event32, ev_CountAndType)),
	       elan4_sdram_readq (dev, event + offsetof (E4_Event32, ev_Params[0])),
	       elan4_sdram_readq (dev, event + offsetof (E4_Event32, ev_Params[1])));
    (di->func)(di->arg, "        pending_tailp %016llx pending_head %016llx\n", 
	       elan4_sdram_readq (dev, rcvrElan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_pending_tailp)),
	       elan4_sdram_readq (dev, rcvrElan + offsetof (EP4_RCVR_RAIL_ELAN, rcvr_pending_head)));
}

void
ep4rcvr_fillout_rail_stats(EP_RCVR_RAIL *rcvr_rail, char *str) {
    /* no stats here yet */
    /* EP4_RCVR_RAIL * ep4rcvr_rail = (EP4_RCVR_RAIL *) rcvr_rail; */
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
