/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsTx.c,v 1.30.2.2 2005/09/02 11:04:02 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsTx.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "cm.h"
#include "debug.h"

unsigned int ep_txd_lowat = 5;

static int
AllocateTxdBlock (EP_XMTR *xmtr, EP_ATTRIBUTE attr, EP_TXD **txdp)
{
    EP_TXD_BLOCK *blk;
    EP_TXD       *txd;
    EP_TXD_MAIN  *pTxdMain;
    int		  i;
    unsigned long flags;

    EPRINTF1 (DBG_XMTR, "AllocateTxdBlock: xmtr=%p\n", xmtr);

    KMEM_ZALLOC (blk, EP_TXD_BLOCK *, sizeof (EP_TXD_BLOCK), ! (attr & EP_NO_SLEEP));

    if (blk == NULL)
	return -ENOMEM;

    if ((pTxdMain = ep_shared_alloc_main (xmtr->Subsys->Subsys.Sys, EP_TXD_MAIN_SIZE * EP_NUM_TXD_PER_BLOCK, attr, &blk->NmdMain)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP_TXD_BLOCK));
	return -ENOMEM;
    }

    for (txd = &blk->Txd[0], i = 0; i < EP_NUM_TXD_PER_BLOCK; i++, txd++)
    {
	txd->Xmtr     = xmtr;
	txd->TxdMain = pTxdMain;

	ep_nmd_subset (&txd->NmdMain, &blk->NmdMain, (i * EP_TXD_MAIN_SIZE), EP_TXD_MAIN_SIZE);

	/* move onto next descriptor */
	pTxdMain = (EP_TXD_MAIN *) ((unsigned long) pTxdMain + EP_TXD_MAIN_SIZE);
    }

    spin_lock_irqsave (&xmtr->FreeDescLock, flags);

    list_add  (&blk->Link, &xmtr->DescBlockList);
    xmtr->TotalDescCount += EP_NUM_TXD_PER_BLOCK;

    for (i = txdp ? 1 : 0; i < EP_NUM_TXD_PER_BLOCK; i++)
    {
	list_add (&blk->Txd[i].Link, &xmtr->FreeDescList);

	xmtr->FreeDescCount++;

	if (xmtr->FreeDescWanted)
	{
	    xmtr->FreeDescWanted--;
	    kcondvar_wakeupone (&xmtr->FreeDescSleep, &xmtr->FreeDescLock);
	}
    }
    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);
    
    if (txdp)
	*txdp = &blk->Txd[0];

    return 0;
}

static void
FreeTxdBlock (EP_XMTR *xmtr, EP_TXD_BLOCK *blk)
{
    unsigned long flags;

    spin_lock_irqsave (&xmtr->FreeDescLock, flags);
    list_del (&blk->Link);

    xmtr->TotalDescCount -= EP_NUM_RXD_PER_BLOCK;
    xmtr->FreeDescCount -= EP_NUM_RXD_PER_BLOCK;
    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);

    ep_shared_free_main (xmtr->Subsys->Subsys.Sys, &blk->NmdMain);
    KMEM_FREE (blk, sizeof (EP_TXD_BLOCK));
}

static EP_TXD *
GetTxd (EP_XMTR *xmtr, EP_ATTRIBUTE attr)
{
    EP_COMMS_SUBSYS *subsys = xmtr->Subsys;
    EP_TXD          *txd;
    int low_on_txds;
    unsigned long flags;

    spin_lock_irqsave (&xmtr->FreeDescLock, flags);

    while (list_empty (&xmtr->FreeDescList))
    {
	if (! (attr & EP_NO_ALLOC))
	{
	    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);

	    if (AllocateTxdBlock (xmtr, attr, &txd) == ESUCCESS)
		return (txd);

	    spin_lock_irqsave (&xmtr->FreeDescLock, flags);
	}

	if (attr & EP_NO_SLEEP)
	{
	    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);

	    return (NULL);
	}

	xmtr->FreeDescWanted++;
	kcondvar_wait (&xmtr->FreeDescSleep, &xmtr->FreeDescLock, &flags);
    }

    txd = list_entry (xmtr->FreeDescList.next, EP_TXD, Link);

    list_del (&txd->Link);

    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_txds = (--xmtr->FreeDescCount < ep_txd_lowat);

    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);

    if (low_on_txds)
	ep_kthread_schedule (&subsys->Thread, lbolt);

    return (txd);
}

void
FreeTxd (EP_XMTR *xmtr, EP_TXD *txd)
{
    unsigned long flags;

    spin_lock_irqsave (&xmtr->FreeDescLock, flags);
    
    list_add (&txd->Link, &xmtr->FreeDescList);

    xmtr->FreeDescCount++;

    if (xmtr->FreeDescWanted)					/* someone waiting for a receive */
    {								/* descriptor, so wake them up */
	xmtr->FreeDescWanted--;
	kcondvar_wakeupone (&xmtr->FreeDescSleep, &xmtr->FreeDescLock);
    }
    
    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);
}

int
TxdShouldStabalise (EP_TXD_RAIL *txdRail, EP_RAIL *rail)
{
    EP_TXD      *txd  = txdRail->Txd;
    EP_XMTR     *xmtr = txd->Xmtr;
    EP_ATTRIBUTE attr = txd->Envelope.Attr;
    int		 stabilise;
    extern int   txd_stabilise;

    switch (EP_ATTR2TYPE (attr)) 
    {
    case EP_TYPE_SVC_INDICATOR:				/* is the rail in the current service indicator rail mask */
	if ((txd_stabilise & 4) == 0)
	    return 0;

	stabilise = (ep_xmtr_svc_indicator_railmask (xmtr, EP_ATTR2DATA (attr), txd->NodeId) & EP_RAIL2RAILMASK (rail->Number)) == 0;
	break;

    case EP_TYPE_TIMEOUT:
	if ((txd_stabilise & 2) == 0)
	    return 0;

	stabilise = AFTER(lbolt, txdRail->Txd->TimeStamp + EP_ATTR2DATA(attr));
	break;

    default:
	if ((txd_stabilise & 1) == 0)
	    return 0;

	stabilise = AFTER(lbolt, txdRail->Txd->TimeStamp + EP_DEFAULT_TIMEOUT);
	break;
    }

    if (stabilise)
    {
	txd->Envelope.Attr = EP_SET_TXD_STABALISING(txd->Envelope.Attr);
	txd->RetryTime     = lbolt;

	ep_kthread_schedule (&xmtr->Subsys->Thread, lbolt);    
    }

    return stabilise;
}

void ep_xmtr_txd_stat(EP_XMTR *xmtr, EP_TXD *txd) 
{
    int f;
    unsigned long size;
    EP_TXD_RAIL *txdRail = txd->TxdRail;

    size = 0;
    for (f=0; f < txd->Envelope.nFrags; f++)
	size += txd->Envelope.Frags[f].nmd_len;

    INC_STAT(xmtr->stats,tx);
    ADD_STAT(xmtr->stats,tx_len, size);  
    
    if ((txdRail != NULL) && (txdRail->XmtrRail != NULL)){
	INC_STAT(txdRail->XmtrRail->stats,tx);
	ADD_STAT(txdRail->XmtrRail->stats,tx_len, size); 
	
	if ((txdRail->XmtrRail->CommsRail != NULL) && ( txdRail->XmtrRail->CommsRail->Rail != NULL)) {
	    INC_STAT(txdRail->XmtrRail->CommsRail->Rail->Stats,tx);
	    ADD_STAT(txdRail->XmtrRail->CommsRail->Rail->Stats,tx_len, size);
	}
    }
}

static int
PollActiveTransmitList (EP_XMTR *xmtr, int flag)
{
    struct list_head *el, *nel;
    struct list_head list;
    unsigned long flags;
    int count;

    INIT_LIST_HEAD (&list);

    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD      *txd     = list_entry (el, EP_TXD, Link);
	EP_TXD_RAIL *txdRail = txd->TxdRail;
	
	if (txdRail == NULL)
	    continue;

	ASSERT (txdRail->Txd == txd);
	
	if (EP_XMTR_OP (txdRail->XmtrRail,PollTxd) (txdRail->XmtrRail, txdRail, flag))
	{
	    list_del (&txd->Link);				/* remove from active transmit list */
	    list_add_tail (&txd->Link, &list);			/* and add to list to call handlers */
	}
    }
    
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    for (count = 0; !list_empty (&list); count++)
    {
	EP_TXD *txd = list_entry (list.next, EP_TXD, Link);

	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_SUCCESS);

	FreeTxd (xmtr, txd);
    }
    return (count);
}

static inline void
DoTransmit (EP_XMTR *xmtr, EP_TXD *txd)
{
    EP_RAILMASK   nmdRailMask = ep_nmd2railmask (txd->Envelope.Frags, txd->Envelope.nFrags);
    EP_XMTR_RAIL *xmtrRail;
    unsigned long flags;
    int rnum;

    spin_lock_irqsave (&xmtr->Lock, flags);

    if (EP_IS_SVC_INDICATOR(txd->Envelope.Attr))
	nmdRailMask = nmdRailMask & ep_xmtr_svc_indicator_railmask(xmtr, EP_ATTR2DATA(txd->Envelope.Attr), txd->NodeId);

    if (EP_IS_PREFRAIL_SET(txd->Envelope.Attr))
	rnum = EP_ATTR2PREFRAIL(txd->Envelope.Attr);
    else 
	rnum = ep_xmtr_prefrail (xmtr, nmdRailMask, txd->NodeId);
    
    if (rnum < 0 || !(nmdRailMask & EP_RAIL2RAILMASK(rnum)))
	xmtrRail = NULL;
    else
	xmtrRail = xmtr->Rails[rnum];
    
    /* Allocate the XID while holding the xmtr->Lock from our XID cache */
    txd->Envelope.Xid = ep_xid_cache_alloc (xmtr->Subsys->Subsys.Sys, &xmtr->XidCache);
    
    EPRINTF7 (DBG_XMTR, "ep: transmit txd %p to %d/%d: Xid %llx nFrags %d [%08x.%d]\n",
	      txd, txd->NodeId, txd->Service, (long long) txd->Envelope.Xid.Unique, 
	      txd->Envelope.nFrags, txd->Envelope.Frags[0].nmd_addr, txd->Envelope.Frags[0].nmd_len);

    /* Store time transmit started to timeout if not received */
    txd->TimeStamp = lbolt;
    
    /* Initialise the retry backoff */
    txd->Backoff.type = EP_BACKOFF_FREE;

    list_add_tail (&txd->Link, &xmtr->ActiveDescList);

    if (xmtrRail == NULL || !EP_XMTR_OP(xmtrRail,BindTxd) (txd, xmtrRail, EP_TXD_PHASE_ACTIVE))
	ep_kthread_schedule (&xmtr->Subsys->Thread, lbolt);
    
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    if (EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	PollActiveTransmitList (xmtr, POLL_TX_LIST);
}

EP_STATUS
ep_transmit_message (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr, 
		     EP_TXH *handler, void *arg, EP_PAYLOAD *payload, EP_NMD *nmd, int nFrags)
{
    EP_TXD       *txd;
    int           i, len;

    if (nFrags > EP_MAXFRAG || service > EP_MSG_NSVC)
	return (EP_EINVAL);

    if ((txd = GetTxd (xmtr, attr)) == NULL)
	return (EP_ENOMEM);

    txd->Handler = handler;
    txd->Arg     = arg;
    txd->Service = service;
    txd->NodeId  = (unsigned short) dest;

    /* Initialise the envelope */
    txd->Envelope.Version   = EP_ENVELOPE_VERSION;
    txd->Envelope.Attr      = EP_CLEAR_LOCAL_ATTR(attr);
    txd->Envelope.Range     = EP_RANGE (dest, dest);
    txd->Envelope.TxdMain   = txd->NmdMain;
    txd->Envelope.nFrags    = nFrags;

    for (i = len = 0; i < nFrags; len += nmd[i].nmd_len, i++)
	txd->Envelope.Frags[i] = nmd[i];

    if (payload)
    {
	txd->Envelope.Attr = EP_SET_HAS_PAYLOAD(txd->Envelope.Attr);

	bcopy (payload, &txd->Payload, sizeof (EP_PAYLOAD));
    }

    DoTransmit (xmtr, txd);

    BucketStat (xmtr->Subsys, DataXmit, len);

    return (EP_SUCCESS);
}

EP_STATUS
ep_multicast_message (EP_XMTR *xmtr, unsigned int destLo, unsigned int destHi, bitmap_t *bitmap, EP_SERVICE service, 
		     EP_ATTRIBUTE attr, EP_TXH *handler, void *arg, EP_PAYLOAD *payload, EP_NMD *nmd, int nFrags)
{
    EP_SYS       *sys = xmtr->Subsys->Subsys.Sys;
    EP_TXD       *txd;
    int           nnodes;
    int           i, len;
    unsigned long flags;    

    if (nFrags > EP_MAXFRAG || service > EP_MSG_NSVC)
	return (EP_EINVAL);

    if (destLo == -1) 
	destLo = sys->Position.pos_nodeid & ~(EP_MAX_NODES-1);

    if (destHi == -1 && (destHi = ((sys->Position.pos_nodeid + EP_MAX_NODES) & ~(EP_MAX_NODES-1)) - 1) >= sys->Position.pos_nodes)
	destHi = sys->Position.pos_nodes-1;

    nnodes = (destHi-destLo+1);

    if ((txd = GetTxd (xmtr, attr)) == NULL)
	return (EP_ENOMEM);

    txd->Handler = handler;
    txd->Arg     = arg;
    txd->Service = service;

    /* Initialise the envelope */
    txd->Envelope.Version   = EP_ENVELOPE_VERSION;
    txd->Envelope.Attr      = EP_SET_MULTICAST(EP_CLEAR_LOCAL_ATTR(attr));
    txd->Envelope.Range     = EP_RANGE (destLo, destHi);
    txd->Envelope.TxdMain   = txd->NmdMain;
    txd->Envelope.nFrags    = nFrags;

    for (i = len = 0; i < nFrags; len += nmd[i].nmd_len, i++)
	txd->Envelope.Frags[i] = nmd[i];

    if (payload)
    {
	txd->Envelope.Attr = EP_SET_HAS_PAYLOAD(txd->Envelope.Attr);

	bcopy (payload, &txd->Payload, sizeof (EP_PAYLOAD));
    }

    spin_lock_irqsave (&sys->NodeLock, flags);
    if (EP_IS_SVC_INDICATOR(attr)) 
	ep_xmtr_svc_indicator_bitmap(xmtr, EP_ATTR2DATA(attr), txd->TxdMain->Bitmap, destLo, nnodes);
    else
	bt_subset (statemap_tobitmap(sys->NodeSet), txd->TxdMain->Bitmap, destLo, nnodes);
    spin_unlock_irqrestore (&sys->NodeLock, flags);

    if (bitmap != NULL)									/* bitmap supplied, so intersect it with */
	bt_intersect (txd->TxdMain->Bitmap, bitmap, nnodes);				/* the current node set map */
    
    if ((attr & EP_NOT_MYSELF) && destLo <= sys->Position.pos_nodeid && sys->Position.pos_nodeid <= destHi)
	BT_CLEAR (txd->TxdMain->Bitmap, (sys->Position.pos_nodeid-destLo));			/* clear myself if not wanted */

    if ((i = bt_lowbit (txd->TxdMain->Bitmap, nnodes)) < 0)
    {
	FreeTxd (xmtr, txd);
	return (EP_NODE_DOWN);
    }

    txd->NodeId = (unsigned short) i;

    DoTransmit (xmtr, txd);

    BucketStat (xmtr->Subsys, McastXmit, len);

    return (EP_SUCCESS);
}

EP_STATUS
ep_transmit_rpc (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr,
		 EP_TXH *handler, void *arg, EP_PAYLOAD *payload, EP_NMD *nmd, int nFrags)
{
    EP_TXD       *txd;
    int           i, len;

    if (nFrags > EP_MAXFRAG || service > EP_MSG_NSVC)
	return (EP_EINVAL);

    if ((txd = GetTxd (xmtr, attr)) == NULL)
	return (EP_ENOMEM);

    txd->Handler = handler;
    txd->Arg     = arg;
    txd->Service = service;
    txd->NodeId  = dest;

    /* Initialise the envelope */
    txd->Envelope.Version   = EP_ENVELOPE_VERSION;
    txd->Envelope.Attr      = EP_SET_RPC(EP_CLEAR_LOCAL_ATTR(attr));    
    txd->Envelope.Range     = EP_RANGE (dest, dest);
    txd->Envelope.TxdMain   = txd->NmdMain;
    txd->Envelope.nFrags    = nFrags;
     
    for (i = len = 0; i < nFrags; len += nmd[i].nmd_len, i++)
	txd->Envelope.Frags[i] = nmd[i];

    if (payload)
    {
	txd->Envelope.Attr = EP_SET_HAS_PAYLOAD(txd->Envelope.Attr);

	bcopy (payload, &txd->Payload, sizeof (EP_PAYLOAD));
    }

    DoTransmit (xmtr, txd);

    BucketStat (xmtr->Subsys, RPCXmit, len);

    return (EP_SUCCESS);
}

EP_STATUS
ep_multicast_forward (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr, EP_TXH *handler, void *arg,
		      EP_ENVELOPE *env,  EP_PAYLOAD *payload, bitmap_t *bitmap, EP_NMD *nmd, int nFrags)
{
    EP_TXD       *txd;
    int           i, len;

    if (nFrags > EP_MAXFRAG || service > EP_MSG_NSVC)
	return (EP_EINVAL);

    if ((txd = GetTxd (xmtr, attr)) == NULL)
	return (EP_ENOMEM);

    txd->Handler = handler;
    txd->Arg     = arg;
    txd->Service = service;
    txd->NodeId  = (unsigned short) dest;

    /* Initialise the envelope */
    txd->Envelope.Version   = EP_ENVELOPE_VERSION;
    txd->Envelope.Attr      = EP_SET_MULTICAST(EP_CLEAR_LOCAL_ATTR(attr));
    txd->Envelope.Range     = env->Range;
    txd->Envelope.TxdMain   = txd->NmdMain;
    txd->Envelope.nFrags    = nFrags;

    for (i = len = 0; i < nFrags; len += nmd[i].nmd_len, i++)
	txd->Envelope.Frags[i] = nmd[i];

    bt_copy (bitmap, txd->TxdMain->Bitmap, EP_RANGE_HIGH(env->Range) - EP_RANGE_LOW(env->Range) + 1);

    if (payload)
    {
	txd->Envelope.Attr = EP_SET_HAS_PAYLOAD(txd->Envelope.Attr);

	bcopy (payload, &txd->Payload, sizeof (EP_PAYLOAD));
    }

    DoTransmit (xmtr, txd);

    BucketStat (xmtr->Subsys, McastXmit, len);

    return (EP_SUCCESS);
}

int
ep_poll_transmits (EP_XMTR *xmtr)
{
    return (PollActiveTransmitList (xmtr, POLL_TX_LIST));
}

int
ep_enable_txcallbacks (EP_XMTR *xmtr)
{
    return (PollActiveTransmitList (xmtr, ENABLE_TX_CALLBACK));
}

int
ep_disable_txcallbacks (EP_XMTR *xmtr)
{
    return (PollActiveTransmitList (xmtr, DISABLE_TX_CALLBACK));
}

/* functions for accessing fields of txds */
int              ep_txd_node(EP_TXD *txd) 		{ return (txd->NodeId); }
EP_STATUSBLK    *ep_txd_statusblk(EP_TXD *txd) 		{ return (&txd->TxdMain->StatusBlk); }

void
ep_xmtr_xid_msg_handler (void *arg, EP_MANAGER_MSG *msg)
{
    EP_XMTR          *xmtr = (EP_XMTR *) arg;
    EP_SYS           *sys  = xmtr->Subsys->Subsys.Sys;
    struct list_head *el,*nel;
    unsigned long     flags;

    switch (msg->Hdr.Type)
    {
    case EP_MANAGER_MSG_TYPE_FAILOVER_REQUEST:
	spin_lock_irqsave (&xmtr->Lock, flags);
	list_for_each (el, &xmtr->ActiveDescList) {
	    EP_TXD      *txd     = list_entry (el, EP_TXD, Link);
	    EP_TXD_RAIL *txdRail = txd->TxdRail;

	    if (txdRail != NULL && EP_XIDS_MATCH (msg->Body.Failover.Xid, txd->Envelope.Xid))
	    {
		EP_XMTR_RAIL       *xmtrRail = txdRail->XmtrRail;
		EP_RAIL            *rail     = xmtrRail->CommsRail->Rail;
		EP_MANAGER_MSG_BODY msgBody;
		int	            rnum;

		if (! (msg->Body.Failover.Railmask & EP_RAIL2RAILMASK (rail->Number)))
		{
		    /* Need to failover this txd to a different rail, select a rail from
		     * the set that she has asked us to use and which is connected to her
		     * on this transmitter.   If there are no such rails, then in all probability
		     * we're offline on all common rails and eventually she will see we have no
		     * rails in common and abort the receive. */
		    if ((rnum = ep_xmtr_prefrail (xmtr, msg->Body.Failover.Railmask, txd->NodeId)) < 0)
			ep_debugf (DBG_XMTR, "%s: ep_xmtr_xid_msg_handler: FAILOVER_REQUEST but can't determine rail (%04x,%04x,%d,%04x)\n",
				   rail->Name, msg->Body.Failover.Railmask, xmtr->RailMask, txd->NodeId, sys->Nodes[txd->NodeId].ConnectedRails);
		    else
		    {
			EP_XMTR_RAIL *nXmtrRail = xmtr->Rails[rnum];

			EPRINTF4 (DBG_XMTR, "%s: ep_xmtr_xid_msg_handler: FAILOVER_REQUEST txd=%p XID=%llx-> rail %d\n", rail->Name, txd, (long long) txd->Envelope.Xid.Unique, rnum);

			/* Bind the txd rail onto the new rail - it doesn't matter if we fail
			 * as it will remain bound to the original rail */
			(void) EP_XMTR_OP (nXmtrRail, BindTxd) (txd, nXmtrRail, EP_TXD_PHASE_PASSIVE);
		    }
		}

		/* Send a failover response including an envelope update */
		msgBody.FailoverTxd.Rail     = rail->Number;
		msgBody.FailoverTxd.Xid      = txd->Envelope.Xid;
		msgBody.FailoverTxd.TxdRail  = txd->Envelope.TxdRail;

		ep_send_message (rail, msg->Hdr.NodeId, EP_MANAGER_MSG_TYPE_FAILOVER_RESPONSE, msg->Hdr.Xid, &msgBody);
	    }
	}
	spin_unlock_irqrestore (&xmtr->Lock, flags);
	break;

    case EP_MANAGER_MSG_TYPE_GET_NODE_STATE_RESPONSE: {
	int         txd_has_not_sent_envelope = 0;
	EP_TXD      *txd            = NULL;
	EP_TXD_RAIL *txdRail        = NULL;

 	if (msg->Body.NodeState.NetworkErrorState != 0)
	    ep_kthread_schedule (&xmtr->Subsys->Thread, lbolt + MESSAGE_RETRY_TIME);
	else
	{
	    spin_lock_irqsave (&xmtr->Lock, flags);
	    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
		
		txd     = list_entry (el, EP_TXD, Link);
		txdRail = txd->TxdRail;
		
		if (txdRail != NULL && EP_XIDS_MATCH (msg->Hdr.Xid, txd->Envelope.Xid)) {
		    txd_has_not_sent_envelope = EP_XMTR_OP(txdRail->XmtrRail,CheckTxdState)(txd);
		    break;
		}
	    }
	    
	    if (txd_has_not_sent_envelope) {
		EPRINTF2 (DBG_STABILISE, "ep_xmtr_xid_msg_handler: GET_NODE_STATE_RESPONSE txd=%p XID=%llx not sent envelope\n",
			  txd, (long long) txd->Envelope.Xid.Unique);

 		/* at this point it has finished stabalising */
		txd->Envelope.Attr = EP_CLEAR_TXD_STABALISING(txd->Envelope.Attr);

		/* store railmask into txd if not a service indicator or timeout */
		if (EP_IS_NO_TYPE(txd->Envelope.Attr))
		    txd->Envelope.Attr = EP_SET_DATA(txd->Envelope.Attr, EP_TYPE_RAILMASK, msg->Body.NodeState.Railmask);

		spin_unlock_irqrestore (&xmtr->Lock, flags);
		
		/* TXD is now no longer bound to a rail , so let ep_check_xmtr() handle it */
		ep_kthread_schedule (&xmtr->Subsys->Thread, lbolt);
	    }
	    else
		spin_unlock_irqrestore (&xmtr->Lock, flags);	
	}
	break;
    }
    default:
	panic ("ep_xmtr_xid_msg_handler: XID match but invalid message type\n");
    }
}

EP_XMTR *
ep_alloc_xmtr (EP_SYS *sys)
{
    EP_COMMS_SUBSYS   *subsys;
    EP_XMTR          *xmtr;
    struct list_head *el;

    if ((subsys = (EP_COMMS_SUBSYS *) ep_subsys_find (sys, EPCOMMS_SUBSYS_NAME)) == NULL)
	return (NULL);

    KMEM_ZALLOC (xmtr, EP_XMTR *, sizeof (EP_XMTR), 1);

    if (xmtr == NULL)
	return (NULL);
    
    xmtr->Subsys = subsys;

    spin_lock_init (&xmtr->Lock);
    INIT_LIST_HEAD (&xmtr->ActiveDescList);
    
    kcondvar_init (&xmtr->FreeDescSleep);
    spin_lock_init (&xmtr->FreeDescLock);
    INIT_LIST_HEAD (&xmtr->FreeDescList);
    INIT_LIST_HEAD (&xmtr->DescBlockList);

    ep_xid_cache_init (sys, &xmtr->XidCache);

    xmtr->XidCache.MessageHandler = ep_xmtr_xid_msg_handler;
    xmtr->XidCache.Arg            = xmtr;

    kmutex_lock (&subsys->Lock);
    list_add_tail (&xmtr->Link, &subsys->Transmitters);

    ep_procfs_xmtr_add(xmtr);

    /* Now add all rails which are already started */
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *commsRail = list_entry (el, EP_COMMS_RAIL, Link);

	EP_RAIL_OP(commsRail, Xmtr.AddRail) (xmtr, commsRail);
    }
    kmutex_unlock (&subsys->Lock);

    ep_mod_inc_usecount();

    return (xmtr);
}

void
ep_free_xmtr (EP_XMTR *xmtr)
{
    EP_COMMS_SUBSYS  *subsys = xmtr->Subsys;
    EP_SYS           *sys    = subsys->Subsys.Sys;
    struct list_head *el;
    
    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Rails) { 
	EP_COMMS_RAIL *rail = list_entry (el, EP_COMMS_RAIL, Link);

	EP_RAIL_OP(rail,Xmtr.DelRail) (xmtr, rail);
    }

    list_del (&xmtr->Link);
    kmutex_unlock (&subsys->Lock);

    /* all the desc's must be free */
    ASSERT(xmtr->FreeDescCount == xmtr->TotalDescCount);

    /* delete the descs */
    while (!list_empty (&xmtr->DescBlockList))
	FreeTxdBlock( xmtr, list_entry(xmtr->DescBlockList.next, EP_TXD_BLOCK , Link));

    /* they had better all be gone now */
    ASSERT((xmtr->FreeDescCount == 0) && (xmtr->TotalDescCount == 0));

    ep_procfs_xmtr_del(xmtr);

    ep_xid_cache_destroy (sys, &xmtr->XidCache);

    spin_lock_destroy (&xmtr->Lock);
    KMEM_FREE (xmtr, sizeof (EP_XMTR));

    ep_mod_dec_usecount();
}

long
ep_check_xmtr (EP_XMTR *xmtr, long nextRunTime)
{
    EP_COMMS_SUBSYS  *subsys = xmtr->Subsys;
    EP_SYS           *sys    = subsys->Subsys.Sys;
    struct list_head *el, *nel;
    struct list_head  txdList;
    unsigned long       flags;
    int                 timed_out=0;
    int                 i;
    EP_MANAGER_MSG_BODY body;

    INIT_LIST_HEAD (&txdList);

    /* See if we have any txd's which need to be bound to a rail */
    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD      *txd      = list_entry (el, EP_TXD, Link);
	EP_NODE     *node     = &sys->Nodes[txd->NodeId];
	EP_RAILMASK nodeRails = node->ConnectedRails & xmtr->RailMask;
	EP_ENVELOPE *env      = &txd->Envelope;

	if (EP_IS_TXD_STABALISING(txd->Envelope.Attr)) 
	{
	    ASSERT(txd->TxdRail != NULL);

	    if (AFTER (lbolt, txd->RetryTime))
	    {
		EPRINTF6 (DBG_STABILISE, "ep_check_xmtr txd=%p txdRail=%p send get node state to %d Xid=%08x.%08x.%016llx\n",
			  txd, txd->TxdRail, txd->NodeId, env->Xid.Generation, env->Xid.Handle, (long long)env->Xid.Unique);
		
		body.Service = txd->Service;
		if (ep_send_message ( txd->TxdRail->XmtrRail->CommsRail->Rail, txd->NodeId, EP_MANAGER_MSG_TYPE_GET_NODE_STATE, env->Xid, &body) == 0)
		    txd->RetryTime = lbolt + (MESSAGE_RETRY_TIME << ep_backoff (&txd->Backoff, EP_BACKOFF_STABILISE));
		else
		    txd->RetryTime = lbolt + MSGBUSY_RETRY_TIME;
	    }

	    ep_kthread_schedule (&subsys->Thread, txd->RetryTime);
	    continue;
	}

	if (txd->TxdRail != NULL)
	    continue;

	switch (EP_ATTR2TYPE(txd->Envelope.Attr)) 
	{
	case EP_TYPE_SVC_INDICATOR: 
	{
	    EP_RAILMASK       rmask=0;
	    struct list_head *tmp;

	    list_for_each (tmp, &subsys->Rails) { 
		EP_COMMS_RAIL *commsRail = list_entry (tmp, EP_COMMS_RAIL, Link);
		if ( cm_svc_indicator_is_set(commsRail->Rail, EP_ATTR2DATA(txd->Envelope.Attr), txd->NodeId))
		    rmask |= EP_RAIL2RAILMASK(commsRail->Rail->Number);
	    } 
	    nodeRails &= rmask;
	    break;
	}
	case EP_TYPE_TIMEOUT:
	    timed_out = AFTER(lbolt, txd->TimeStamp + EP_ATTR2DATA(txd->Envelope.Attr)) ? (1) : (0);
	    break;
	case EP_TYPE_RAILMASK:
	    nodeRails &= EP_ATTR2DATA(txd->Envelope.Attr);
	    break;
	default:
	    timed_out = AFTER(lbolt, txd->TimeStamp +  EP_DEFAULT_TIMEOUT) ? (1) : (0);
	    break;
	}

	if (nodeRails == 0 || timed_out || (EP_IS_NO_FAILOVER(env->Attr) && EP_IS_PREFRAIL_SET(env->Attr) && 
					    (nodeRails & EP_RAIL2RAILMASK(EP_ATTR2PREFRAIL(env->Attr))) == 0))
	{
	    EPRINTF5 (timed_out ? DBG_STABILISE : DBG_XMTR, "ep_check_xmtr: txd=%p XID=%llx to %d no rails connected or cannot failover (nodeRails=0x%x,timed_out=%d\n", 
		      txd, (long long) env->Xid.Unique, txd->NodeId, nodeRails, timed_out);

	    list_del  (&txd->Link);
	    list_add_tail (&txd->Link, &txdList);
	}
	else
	{
	    EP_XMTR_RAIL *xmtrRail;
	    int i, len, rnum;

	    if (EP_IS_PREFRAIL_SET(env->Attr) && (nodeRails & EP_RAIL2RAILMASK(EP_ATTR2PREFRAIL(env->Attr))))
		rnum = EP_ATTR2PREFRAIL(env->Attr);
	    else
		rnum = ep_pickRail (nodeRails);

	    EPRINTF3 (DBG_XMTR, "ep_check_xmtr: txd=%p XID=%llx mapping NMDs onto rail %d \n", txd, (long long) env->Xid.Unique, rnum);

	    for (i = len = 0; i < env->nFrags; i++, len += env->Frags[i].nmd_len)
		ep_nmd_map_rails (sys, &env->Frags[i], nodeRails);

	    if ((xmtrRail = xmtr->Rails[rnum]) == NULL || 
		!EP_XMTR_OP(xmtrRail,BindTxd) (txd, xmtrRail, EP_TXD_PHASE_ACTIVE))
		ep_kthread_schedule (&subsys->Thread, lbolt + RESOURCE_RETRY_TIME);
	}
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    while (! list_empty (&txdList))
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);
	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_NODE_DOWN);
	FreeTxd (xmtr, txd);
    }

    /* Check to see if we're low on txds */
    if (xmtr->FreeDescCount < ep_txd_lowat)
	AllocateTxdBlock (xmtr, 0, NULL);
    
    /* Then check each rail */
    for (i = 0; i < EP_MAX_RAILS; i++) 
	if (xmtr->RailMask & (1 << i) ) 
	    nextRunTime = EP_XMTR_OP (xmtr->Rails[i],Check) (xmtr->Rails[i], nextRunTime);
    return (nextRunTime);
}

void
ep_display_txd (DisplayInfo *di, EP_TXD *txd)
{
    EP_ENVELOPE *env     = &txd->Envelope;
    EP_TXD_RAIL *txdRail = txd->TxdRail;

    (di->func)(di->arg, "TXD: %p Version=%x Attr=%x Xid=%08x.%08x.%016llx\n", txd, 
	       env->Version, env->Attr, env->Xid.Generation, env->Xid.Handle, (long long) env->Xid.Unique);
    (di->func)(di->arg,  "     NodeId=%d Range=%d.%d TxdRail=%x TxdMain=%x.%x.%x nFrags=%d\n",
	       env->NodeId, EP_RANGE_LOW(env->Range), EP_RANGE_HIGH(env->Range), env->TxdRail,
	       env->TxdMain.nmd_addr, env->TxdMain.nmd_len, env->TxdMain.nmd_attr, env->nFrags);
    (di->func)(di->arg,  "       Frag[0] %08x.%08x.%08x\n", env->Frags[0].nmd_addr, env->Frags[0].nmd_len, env->Frags[0].nmd_attr);
    (di->func)(di->arg,  "       Frag[1] %08x.%08x.%08x\n", env->Frags[1].nmd_addr, env->Frags[1].nmd_len, env->Frags[1].nmd_attr);
    (di->func)(di->arg,  "       Frag[2] %08x.%08x.%08x\n", env->Frags[2].nmd_addr, env->Frags[2].nmd_len, env->Frags[2].nmd_attr);
    (di->func)(di->arg,  "       Frag[3] %08x.%08x.%08x\n", env->Frags[3].nmd_addr, env->Frags[3].nmd_len, env->Frags[3].nmd_attr);

    if (txdRail != NULL) EP_XMTR_OP (txdRail->XmtrRail, DisplayTxd) (di, txdRail);
}

void
ep_display_xmtr (DisplayInfo *di, EP_XMTR *xmtr)
{
    int               freeCount   = 0;
    int               activeCount = 0;
    struct list_head *el;
    int               i;
    unsigned long     flags;

    spin_lock_irqsave (&xmtr->FreeDescLock, flags);
    list_for_each (el, &xmtr->FreeDescList)
	freeCount++;
    spin_unlock_irqrestore (&xmtr->FreeDescLock, flags);

    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each (el, &xmtr->ActiveDescList)
	activeCount++;
    
    (di->func)(di->arg, "ep_display_xmtr: xmtr=%p Free=%d Active=%d\n", xmtr, freeCount, activeCount);
    for (i = 0; i < EP_MAX_RAILS; i++)
	if (xmtr->Rails[i]) EP_XMTR_OP (xmtr->Rails[i], DisplayXmtr) (di, xmtr->Rails[i]);

    list_for_each (el,&xmtr->ActiveDescList)
	ep_display_txd (di, list_entry (el, EP_TXD, Link));
    spin_unlock_irqrestore (&xmtr->Lock, flags);
}

void
ep_xmtr_fillout_stats(EP_XMTR *xmtr, char *str)
{
    sprintf(str+strlen(str),"Tx     %lu  %lu /sec\n",       GET_STAT_TOTAL(xmtr->stats,tx),      GET_STAT_PER_SEC(xmtr->stats,tx) );
    sprintf(str+strlen(str),"MBytes %lu  %lu Mbytes/sec\n", GET_STAT_TOTAL(xmtr->stats,tx_len) / (1024*1024),  GET_STAT_PER_SEC(xmtr->stats,tx_len) / (1024*1024));
}

void
ep_xmtr_rail_fillout_stats(EP_XMTR_RAIL *xmtr_rail, char *str)
{
    sprintf(str+strlen(str),"Tx     %lu  %lu /sec\n",       GET_STAT_TOTAL(xmtr_rail->stats,tx),      GET_STAT_PER_SEC(xmtr_rail->stats,tx) );
    sprintf(str+strlen(str),"MBytes %lu  %lu Mbytes/sec\n", GET_STAT_TOTAL(xmtr_rail->stats,tx_len) / (1024*1024),  GET_STAT_PER_SEC(xmtr_rail->stats,tx_len) / (1024*1024));
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
