/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsTx_elan3.c,v 1.19 2004/11/12 10:55:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsTx_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "epcomms_elan3.h"
#include "debug.h"

#define XMTR_TO_RAIL(xmtrRail)		((EP3_RAIL *) ((EP_XMTR_RAIL *) xmtrRail)->CommsRail->Rail)
#define XMTR_TO_DEV(xmtrRail)		(XMTR_TO_RAIL(xmtrRail)->Device)
#define XMTR_TO_SUBSYS(xmtrRail)	(((EP_XMTR_RAIL *) xmtrRail)->Xmtr->Subsys)

static void TxEnveEvent (EP3_RAIL *rail, void *arg);
static void TxEnveRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status);
static void TxEnveVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma);

static EP3_COOKIE_OPS EnveCookieOps =
{
    TxEnveEvent,
    TxEnveRetry,
    NULL, /* DmaCancelled */
    TxEnveVerify
};

static void TxDataEvent (EP3_RAIL *rail, void *arg);
static void TxDataRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status);
static void TxDataVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma);

static EP3_COOKIE_OPS DataCookieOps =
{
    TxDataEvent,
    TxDataRetry,
    NULL, /* DmaCancelled */
    TxDataVerify
};

static void TxDoneEvent (EP3_RAIL *dev, void *arg);
static void TxDoneRetry (EP3_RAIL *dev, void *arg, E3_DMA_BE *dma, int status);
static void TxDoneVerify (EP3_RAIL  *dev, void *arg, E3_DMA_BE *dma);

static EP3_COOKIE_OPS DoneCookieOps = 
{
    TxDoneEvent,
    TxDoneRetry,
    NULL, /* DmaCancelled */
    TxDoneVerify,
} ;

static int
AllocateTxdRailBlock (EP3_XMTR_RAIL *xmtrRail)
{
    EP3_RAIL          *rail = XMTR_TO_RAIL (xmtrRail);
    ELAN3_DEV         *dev = rail->Device;
    EP3_TXD_RAIL_BLOCK *blk;
    EP3_TXD_RAIL       *txdRail;
    sdramaddr_t        pTxdElan;
    EP3_TXD_RAIL_MAIN  *pTxdMain;
    E3_Addr	       pTxdElanAddr;
    E3_Addr	       pTxdMainAddr;
    E3_BlockCopyEvent  event;
    int		       i;
    unsigned long      flags;

    KMEM_ZALLOC (blk, EP3_TXD_RAIL_BLOCK *, sizeof (EP3_TXD_RAIL_BLOCK), 1);

    if (blk == NULL)
	return 0;

    if ((pTxdElan = ep_alloc_elan (&rail->Generic, EP3_TXD_RAIL_ELAN_SIZE * EP3_NUM_TXD_PER_BLOCK, 0, &pTxdElanAddr)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP3_TXD_RAIL_BLOCK));
	return 0;
    }

    if ((pTxdMain = ep_alloc_main (&rail->Generic, EP3_TXD_RAIL_MAIN_SIZE * EP3_NUM_TXD_PER_BLOCK, 0, &pTxdMainAddr)) == (EP3_TXD_RAIL_MAIN *) NULL)
    {
	ep_free_elan (&rail->Generic, pTxdElanAddr, EP3_TXD_RAIL_ELAN_SIZE * EP3_NUM_TXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP3_TXD_RAIL_BLOCK));
	return 0;
    }
    
    if (ReserveDmaRetries (rail, EP3_NUM_TXD_PER_BLOCK, 0) != ESUCCESS)
    {
	ep_free_main (&rail->Generic, pTxdMainAddr, EP3_TXD_RAIL_MAIN_SIZE * EP3_NUM_TXD_PER_BLOCK);
	ep_free_elan (&rail->Generic, pTxdElanAddr, EP3_TXD_RAIL_ELAN_SIZE * EP3_NUM_TXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP3_TXD_RAIL_BLOCK));
	return 0;
    }

    for (txdRail = &blk->Txd[0], i = 0; i < EP3_NUM_TXD_PER_BLOCK; i++, txdRail++)
    {
	txdRail->Generic.XmtrRail = &xmtrRail->Generic;
	txdRail->TxdElan          = pTxdElan;
	txdRail->TxdElanAddr      = pTxdElanAddr;
	txdRail->TxdMain          = pTxdMain;
	txdRail->TxdMainAddr      = pTxdMainAddr;

	RegisterCookie (&rail->CookieTable, &txdRail->EnveCookie, pTxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent), &EnveCookieOps, (void *) txdRail);
	RegisterCookie (&rail->CookieTable, &txdRail->DataCookie, pTxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, DataEvent), &DataCookieOps, (void *) txdRail);
	RegisterCookie (&rail->CookieTable, &txdRail->DoneCookie, pTxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent), &DoneCookieOps, (void *) txdRail);

	EP3_INIT_COPY_EVENT (event, txdRail->EnveCookie, pTxdMainAddr + offsetof (EP3_TXD_RAIL_MAIN, EnveEvent), 0);
	elan3_sdram_copyl_to_sdram (dev, &event, pTxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent), sizeof (E3_BlockCopyEvent));

	EP3_INIT_COPY_EVENT (event, txdRail->DataCookie, pTxdMainAddr + offsetof (EP3_TXD_RAIL_MAIN, DataEvent), 0);
	elan3_sdram_copyl_to_sdram (dev, &event, pTxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent), sizeof (E3_BlockCopyEvent));

	EP3_INIT_COPY_EVENT (event, txdRail->DoneCookie, pTxdMainAddr + offsetof (EP3_TXD_RAIL_MAIN, DoneEvent), 0);
	elan3_sdram_copyl_to_sdram (dev, &event, pTxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent), sizeof (E3_BlockCopyEvent));
	
	pTxdMain->EnveEvent = EP3_EVENT_FREE;
	pTxdMain->DataEvent = EP3_EVENT_FREE;
	pTxdMain->DoneEvent = EP3_EVENT_FREE;

	/* move onto next descriptor */
	pTxdElan     += EP3_TXD_RAIL_ELAN_SIZE;
	pTxdElanAddr += EP3_TXD_RAIL_ELAN_SIZE;
	pTxdMain      = (EP3_TXD_RAIL_MAIN *) ((unsigned long) pTxdMain + EP3_TXD_RAIL_MAIN_SIZE);
	pTxdMainAddr += EP3_TXD_RAIL_MAIN_SIZE;
    }

    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);

    list_add  (&blk->Link, &xmtrRail->DescBlockList);
    xmtrRail->TotalDescCount += EP3_NUM_TXD_PER_BLOCK;
    xmtrRail->FreeDescCount  += EP3_NUM_TXD_PER_BLOCK;

    for (i = 0; i < EP3_NUM_TXD_PER_BLOCK; i++)
	list_add (&blk->Txd[i].Generic.Link, &xmtrRail->FreeDescList);

    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);
    
    return 1;
}

static void
FreeTxdRailBlock (EP3_XMTR_RAIL *xmtrRail, EP3_TXD_RAIL_BLOCK *blk)
{
    EP3_RAIL     *rail = XMTR_TO_RAIL(xmtrRail);
    EP3_TXD_RAIL *txdRail;
    unsigned long flags;
    int i;

    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);

    list_del (&blk->Link);
    
    xmtrRail->TotalDescCount -= EP3_NUM_TXD_PER_BLOCK;
    
    for (txdRail = &blk->Txd[0], i = 0; i < EP3_NUM_TXD_PER_BLOCK; i++, txdRail++)
    {
	xmtrRail->FreeDescCount--;
	
	list_del (&txdRail->Generic.Link);
	
	DeregisterCookie (&rail->CookieTable, &txdRail->EnveCookie);
	DeregisterCookie (&rail->CookieTable, &txdRail->DataCookie);
	DeregisterCookie (&rail->CookieTable, &txdRail->DoneCookie);
    }

    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);

    ReleaseDmaRetries (rail, EP3_NUM_TXD_PER_BLOCK);

    ep_free_main (&rail->Generic, blk->Txd[0].TxdMainAddr, EP3_TXD_RAIL_MAIN_SIZE * EP3_NUM_TXD_PER_BLOCK);
    ep_free_elan (&rail->Generic, blk->Txd[0].TxdElanAddr, EP3_TXD_RAIL_ELAN_SIZE * EP3_NUM_TXD_PER_BLOCK);
    KMEM_FREE (blk, sizeof (EP3_TXD_RAIL_BLOCK));
}

static EP3_TXD_RAIL *
GetTxdRail (EP3_XMTR_RAIL *xmtrRail)
{
    EP_COMMS_SUBSYS  *subsys = xmtrRail->Generic.Xmtr->Subsys;
    EP3_TXD_RAIL     *txdRail;
    int low_on_txds;
    unsigned long flags;

    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);

    if (list_empty (&xmtrRail->FreeDescList))
	txdRail = NULL;
    else
    {
	txdRail = list_entry (xmtrRail->FreeDescList.next, EP3_TXD_RAIL, Generic.Link);

#if defined(DEBUG)
	{
	    EP_RAIL   *rail = xmtrRail->Generic.CommsRail->Rail;
	    ELAN3_DEV *dev  = ((EP3_RAIL *) rail)->Device;
	    
	    EP_ASSERT (rail, txdRail->TxdMain->EnveEvent == EP3_EVENT_FREE);
	    EP_ASSERT (rail, txdRail->TxdMain->DataEvent == EP3_EVENT_FREE);
	    EP_ASSERT (rail, txdRail->TxdMain->DoneEvent == EP3_EVENT_FREE);
	    EP_ASSERT (rail, SDRAM_ASSERT(elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)) == 0));
	    EP_ASSERT (rail, SDRAM_ASSERT(elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));
	    EP_ASSERT (rail, SDRAM_ASSERT(elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));
	}
#endif
	
	list_del (&txdRail->Generic.Link);

	xmtrRail->FreeDescCount--;
    }
    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_txds = (xmtrRail->FreeDescCount < ep_txd_lowat);

    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);

    if (low_on_txds)
	ep_kthread_schedule (&subsys->Thread, lbolt);

    return (txdRail);
}

static void
FreeTxdRail (EP3_XMTR_RAIL *xmtrRail, EP3_TXD_RAIL *txdRail)
{
    unsigned long flags;

#if defined(DEBUG_ASSERT)
    {
	EP_RAIL   *rail = xmtrRail->Generic.CommsRail->Rail;
	ELAN3_DEV *dev  = ((EP3_RAIL *) rail)->Device;

	EP_ASSERT (rail, txdRail->Generic.XmtrRail == &xmtrRail->Generic);
	
	EP_ASSERT (rail, txdRail->TxdMain->EnveEvent == EP3_EVENT_PRIVATE);
	EP_ASSERT (rail, txdRail->TxdMain->DataEvent == EP3_EVENT_PRIVATE);
	EP_ASSERT (rail, txdRail->TxdMain->DoneEvent == EP3_EVENT_PRIVATE);
	EP_ASSERT (rail, SDRAM_ASSERT (elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)) == 0));
	EP_ASSERT (rail, SDRAM_ASSERT (elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));
	EP_ASSERT (rail, SDRAM_ASSERT (elan3_sdram_readl (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count)) == 0));
	
	txdRail->TxdMain->EnveEvent = EP3_EVENT_FREE;
	txdRail->TxdMain->DataEvent = EP3_EVENT_FREE;
	txdRail->TxdMain->DoneEvent = EP3_EVENT_FREE;
    }
#endif

    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);
    
    list_add (&txdRail->Generic.Link, &xmtrRail->FreeDescList);

    xmtrRail->FreeDescCount++;

    if (xmtrRail->FreeDescWaiting)
    {
	xmtrRail->FreeDescWaiting--;
	kcondvar_wakeupall (&xmtrRail->FreeDescSleep, &xmtrRail->FreeDescLock);
    }

    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);
}

static void
BindTxdToRail (EP_TXD *txd, EP3_TXD_RAIL *txdRail)
{
    ASSERT (SPINLOCK_HELD (&txd->Xmtr->Lock));

    EPRINTF6 (DBG_XMTR, "%s: BindTxdToRail: txd=%p txdRail=%p XID=%08x.%08x.%016llx\n", 
	      XMTR_TO_RAIL(txdRail->Generic.XmtrRail)->Generic.Name, txd, txdRail, 
	      txd->Envelope.Xid.Generation, txd->Envelope.Xid.Handle, (long long) txd->Envelope.Xid.Unique);

    txd->TxdRail = &txdRail->Generic;
    txdRail->Generic.Txd = txd;
}

static void
UnbindTxdFromRail (EP_TXD *txd, EP3_TXD_RAIL *txdRail)
{
    ASSERT (SPINLOCK_HELD (&txd->Xmtr->Lock));
    ASSERT (txd->TxdRail == &txdRail->Generic && txdRail->Generic.Txd == txd);

    EPRINTF6 (DBG_XMTR, "%s: UnbindTxdToRail: txd=%p txdRail=%p XID=%08x.%08x.%016llx\n", 
	      XMTR_TO_RAIL(txdRail->Generic.XmtrRail)->Generic.Name, txd, txdRail, 
	      txd->Envelope.Xid.Generation, txd->Envelope.Xid.Handle, (long long) txd->Envelope.Xid.Unique);
    txd->TxdRail = NULL;
    txdRail->Generic.Txd = NULL; 
}

/*
 * TxEnveEvent: arg == EP_TXD
 *    Called when envelope delivered
 */
static void
TxEnveEvent (EP3_RAIL *rail, void *arg)
{
    panic ("TxEnveEvent");
}

/*
 * TxEnveRetry: arg == EP3_TXD_RAIL
 *    Called on retry of dma of large message envelope.
 */
static void
TxEnveRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status)
{
    EP3_TXD_RAIL  *txdRail  = (EP3_TXD_RAIL *) arg;
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    
    EPRINTF3 (DBG_XMTR, "%s: TxEnveRetry: xmtr %p txd %p\n",  rail->Generic.Name, xmtrRail, txdRail);
    
    EP_ASSERT (&rail->Generic, txdRail->TxdMain->EnveEvent == EP3_EVENT_ACTIVE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)) == 1)); /* PCI read */
    EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_WRITE && EP_VP_TO_NODE(dma->s.dma_destVProc) == txdRail->Generic.Txd->NodeId);

    if (! TxdShouldStabalise (&txdRail->Generic, &rail->Generic))
	QueueDmaForRetry (rail, dma, EP_RETRY_LOW_PRI_RETRY + ep_backoff (&txdRail->Backoff, EP_BACKOFF_ENVELOPE));
    else
	QueueDmaForRetry (rail, dma, EP_RETRY_STABALISING);	/* place dma on stabilising list for neterr fixup */
}

static void
TxEnveVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma)
{
    EP3_TXD_RAIL *txdRail = (EP3_TXD_RAIL *) arg;
    
    EP_ASSERT (&rail->Generic, txdRail->TxdMain->EnveEvent == EP3_EVENT_ACTIVE);
    EP_ASSERT (&rail->Generic, SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)) == 1)); /* PCI read */
    EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_WRITE && EP_VP_TO_NODE(dma->s.dma_destVProc) == txdRail->Generic.Txd->NodeId);
}

/*
 * TxDataEvent: arg == EP3_TXD
 *    Called on completion of a large transmit.
 */
static void
TxDataEvent (EP3_RAIL *rail, void *arg)
{
    EP3_TXD_RAIL      *txdRail  = (EP3_TXD_RAIL *) arg;
    EP3_XMTR_RAIL     *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP_XMTR	      *xmtr     = xmtrRail->Generic.Xmtr;
    EP3_TXD_RAIL_MAIN *txdMain  = txdRail->TxdMain;
    sdramaddr_t        txdElan  = txdRail->TxdElan;
    int                delay    = 1;
    EP_TXD            *txd;
    unsigned long      flags;

    spin_lock_irqsave (&xmtr->Lock, flags);
    for (;;)
    {
	if (EP3_EVENT_FIRED (txdRail->DataCookie, txdMain->DataEvent))
	    break;

	if (EP3_EVENT_FIRING (rail->Device, txdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent), txdRail->DataCookie, txdMain->DataEvent))		/* PCI read */
	{
	    if (delay > EP3_EVENT_FIRING_TLIMIT)
		panic ("TxDataEvent: events set but block copy not completed\n");
	    DELAY(delay);
	    delay <<= 1;
	}
	else
	{
	    EPRINTF3 (DBG_XMTR, "%s: TxDataEvent: xmtr %p txd %p previously collecting by polling\n", 
		      rail->Generic.Name, xmtrRail, txdRail);
	    spin_unlock_irqrestore (&xmtr->Lock, flags);
	    return;
	}
	mb();
    }

    if ((txd = txdRail->Generic.Txd) == NULL ||			/* If there is no txd, or if the descriptor is marked */
	!(EP_IS_INTERRUPT_ENABLED(txd->Envelope.Attr)) ||	/* as no interrupt, or been reused as an RPC, */
	(EP_IS_RPC(txd->Envelope.Attr)))       			/* then we were either called as a result of a previous */
    {								/* tx which was completed by polling or as a result */
	spin_unlock_irqrestore (&xmtr->Lock, flags);		/* of a EnableTxCallBack/DisableTxCallback */

	EPRINTF4 (DBG_XMTR, "%s: TxDataEvent: xmtr %p txd %p recyled (%x)\n", 
		  rail->Generic.Name, xmtr, txd, txd ? txd->Envelope.Attr : 0);
	return;
    }

    ASSERT (EP3_EVENT_FIRED (txdRail->EnveCookie, txdMain->EnveEvent));

    EPRINTF5 (DBG_XMTR, "%s: TxDataEvent : xmtrRail=%p txdRail=%p tx=%p XID=%llx\n", 
	      rail->Generic.Name, xmtrRail, txdRail, txd, (long long) txd->Envelope.Xid.Unique);
    
    ep_xmtr_txd_stat(xmtr,txd);
    
    /* remove from active transmit lists */
    list_del (&txd->Link);

    UnbindTxdFromRail (txd, txdRail);
    
    /* clear the done flags for next time round */
    txdMain->EnveEvent = EP3_EVENT_PRIVATE;
    txdMain->DataEvent = EP3_EVENT_PRIVATE;
    txdMain->DoneEvent = EP3_EVENT_PRIVATE;
    
    FreeTxdRail (xmtrRail, txdRail);

    spin_unlock_irqrestore (&xmtr->Lock, flags);
    
    txd->Handler (txd, txd->Arg, EP_SUCCESS);
    
    FreeTxd (xmtr, txd);
}

/*
 * TxDataRetry: arg == EP3_TXD
 *    Called on retry of remote "put" dma of large transmit data.
 */
static void
TxDataRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status)
{
    EP3_TXD_RAIL  *txdRail  = (EP3_TXD_RAIL *) arg;
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP_TXD        *txd      = txdRail->Generic.Txd;

    EP_ASSERT (&rail->Generic, ((txdRail->TxdMain->DataEvent == EP3_EVENT_ACTIVE && 
				 SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) >= 1)) ||	/* PCI read */
				(EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent) &&
				 SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) == 0))));	/* PCI read */
    EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_WRITE && EP_VP_TO_NODE(dma->s.dma_destVProc) == txd->NodeId);

    EPRINTF5 (DBG_XMTR, "%s: TxDataRetry: xmtrRail=%p txdRail=%p txd=%p XID=%llx\n", 
	      rail->Generic.Name, xmtrRail, txdRail, txd, (long long) txd->Envelope.Xid.Unique);
    
    QueueDmaForRetry (rail, dma, EP_RETRY_LOW_PRI_RETRY + ep_backoff (&txdRail->Backoff, EP_BACKOFF_DATA));
}

static void
TxDataVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma)
{
    EP3_TXD_RAIL *txdRail = (EP3_TXD_RAIL *) arg;
    EP_TXD       *txd     = txdRail->Generic.Txd;

    EP_ASSERT (&rail->Generic, ((txdRail->TxdMain->DataEvent == EP3_EVENT_ACTIVE && 
				 SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) >= 1)) || 	/* PCI read */
				(EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent) &&
				 SDRAM_ASSERT (elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) == 0))));	/* PCI read */
    EP_ASSERT (&rail->Generic, dma->s.dma_direction == DMA_WRITE && EP_VP_TO_NODE(dma->s.dma_destVProc) == txd->NodeId);
}

/*
 * TxDoneEvent: arg == EP3_TXD
 *    Called on completion of a RPC.
 */
static void
TxDoneEvent (EP3_RAIL *rail, void *arg)
{
    EP3_TXD_RAIL      *txdRail  = (EP3_TXD_RAIL *) arg;
    EP3_XMTR_RAIL     *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP_XMTR	      *xmtr     = xmtrRail->Generic.Xmtr;
    int                delay   = 1;
    EP_TXD	      *txd;
    unsigned long      flags;

    spin_lock_irqsave (&xmtr->Lock, flags);

    for (;;)
    {
	if (EP3_EVENT_FIRED (txdRail->DoneCookie, txdRail->TxdMain->DoneEvent) && 
	    EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent))
	    break;
	
	if (EP3_EVENT_FIRING (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent), txdRail->DoneCookie, txdRail->TxdMain->DoneEvent) && 
	    EP3_EVENT_FIRING (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent), txdRail->DataCookie, txdRail->TxdMain->DataEvent))
	{
	    if (delay > EP3_EVENT_FIRING_TLIMIT)
		panic ("TxDoneEvent: events set but block copy not completed\n");
	    DELAY(delay);
	    delay <<= 1;
	}
	else
	{
	    EPRINTF3 (DBG_XMTR, "%s: TxDoneEvent: xmtr %p txdRail %p previously collecting by polling\n", 
		      rail->Generic.Name, xmtr, txdRail);
	    spin_unlock_irqrestore (&xmtr->Lock, flags);
	    return;
	}
	mb();
    }

    if ((txd = txdRail->Generic.Txd) == NULL ||		                         	 /* If there is no txd, or if the descriptor is marked */
	!(EP_IS_INTERRUPT_ENABLED(txd->Envelope.Attr) || EP_IS_RPC(txd->Envelope.Attr))) /* marked as no interrupt, or been reused as an transmit, */
    {							                    	         /* then we were either called as a result of a previous */
	spin_unlock_irqrestore (&xmtr->Lock, flags);		                         /* tx which was completed by polling or as a result */
								                         /* of a EnableTxCallBack/DisableTxCallback */

	EPRINTF4 (DBG_XMTR, "%s: TxDoneEvent: xmtr %p txd %p recyled (%x)\n", 
		  rail->Generic.Name, xmtr, txd, txd ? txd->Envelope.Attr : 0);
	return;	
    }

    EPRINTF5 (DBG_XMTR, "%s: TxDoneEvent: xmtrRail=%p txdRail=%p txd=%p XID=%llx\n", 
	      rail->Generic.Name, xmtrRail, txdRail, txd, (long long) txd->Envelope.Xid.Unique);

    ep_xmtr_txd_stat(xmtr,txd);

    /* remove from active transmit list */
    list_del (&txd->Link);
    
    UnbindTxdFromRail (txd, txdRail);
    
    /* clear the done flags for next time round */
    txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
    
    FreeTxdRail (xmtrRail, txdRail);

    spin_unlock_irqrestore (&xmtr->Lock, flags);
	    
    if (txd->Handler)
	txd->Handler (txd, txd->Arg, EP_SUCCESS);
	
    FreeTxd (xmtr, txd);
}

/*
 * TxDoneRetry: arg == EP3_TXD
 */
static void
TxDoneRetry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int status)
{
    panic ("TxDoneRetry");
}

static void
TxDoneVerify (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma)
{
    panic ("TxDoneVerify");
}

static void
EnableTransmitCallback (EP_TXD *txd, EP3_TXD_RAIL *txdRail)
{
    ELAN3_DEV *dev = XMTR_TO_RAIL(txdRail->Generic.XmtrRail)->Device;

    EPRINTF3 (DBG_XMTR, "%s: EnableTransmitCallback: txd %p txdRail %p\n", XMTR_TO_RAIL (txdRail->Generic.XmtrRail)->Generic.Name, txd, txdRail);

    txd->Envelope.Attr = EP_SET_INTERRUPT_ENABLED(txd->Envelope.Attr);
		
    elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Type), EV_TYPE_BCOPY);
	
    if (EP_IS_RPC(txd->Envelope.Attr))
    {
	elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Type), EV_TYPE_BCOPY);
	elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Type),  EV_TYPE_BCOPY | EV_TYPE_EVIRQ | txdRail->DoneCookie.Cookie);
    }
    else
    {
	elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Type), EV_TYPE_BCOPY | EV_TYPE_EVIRQ | txdRail->DataCookie.Cookie);
	elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Type), EV_TYPE_BCOPY);
    }
}

static void
DisableTransmitCallback (EP_TXD *txd, EP3_TXD_RAIL *txdRail)
{
    ELAN3_DEV *dev = XMTR_TO_RAIL(txdRail->Generic.XmtrRail)->Device;

    EPRINTF3 (DBG_XMTR, "%s: DisableTransmitCallback: txd %p txdRail %p\n", XMTR_TO_RAIL (txdRail->Generic.XmtrRail)->Generic.Name, txd, txdRail);

    txd->Envelope.Attr = EP_CLEAR_INTERRUPT_ENABLED(txd->Envelope.Attr);

    elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Type), EV_TYPE_BCOPY);
    elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Type), EV_TYPE_BCOPY);
    elan3_sdram_writel (dev, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Type), EV_TYPE_BCOPY);
}

static void
InitialiseTxdRail (EP_TXD *txd, EP3_TXD_RAIL *txdRail, int phase)
{
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP3_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);

    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->Generic);

    /* Initialise the per-rail fields in the envelope */
    txd->Envelope.TxdRail = txdRail->TxdElanAddr;
    txd->Envelope.NodeId  = rail->Generic.Position.pos_nodeid;

    /* Initialise the dma backoff */
    txdRail->Backoff.type = EP_BACKOFF_FREE;

    /* Initialise the per-rail events */
    switch (phase)
    {
    case EP_TXD_PHASE_ACTIVE:
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 1);
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 
			    (txd->Envelope.nFrags ? txd->Envelope.nFrags : 1) + (EP_IS_MULTICAST(txd->Envelope.Attr) ? 1 : 0));
	
	txdRail->TxdMain->EnveEvent = EP3_EVENT_ACTIVE;
	txdRail->TxdMain->DataEvent = EP3_EVENT_ACTIVE;
	break;
	
    case EP_TXD_PHASE_PASSIVE:
	ASSERT (EP_IS_RPC(txd->Envelope.Attr));

	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0);
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0);

	txdRail->TxdMain->EnveEvent = txdRail->EnveCookie.Cookie;
	txdRail->TxdMain->DataEvent = txdRail->DataCookie.Cookie;
	break;
    }

    if (! EP_IS_RPC(txd->Envelope.Attr))
	txdRail->TxdMain->DoneEvent = txdRail->DoneCookie.Cookie;
    else
    {
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 1);
	txdRail->TxdMain->DoneEvent = EP3_EVENT_ACTIVE;
    }

    if (EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	DisableTransmitCallback (txd, txdRail);
    else
	EnableTransmitCallback (txd, txdRail);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    if ( epdebug_check_sum ) 
	txd->Envelope.CheckSum = ep_calc_check_sum( txd->Xmtr->Subsys->Subsys.Sys, &txd->Envelope, txd->Envelope.Frags, txd->Envelope.nFrags);
    else
#endif
	txd->Envelope.CheckSum = 0;  

    /* copy the envelope and payload if present down to sdram */
    elan3_sdram_copyl_to_sdram (rail->Device, &txd->Envelope, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, Envelope), EP_ENVELOPE_SIZE);
    
    if (EP_HAS_PAYLOAD(txd->Envelope.Attr))
	elan3_sdram_copyl_to_sdram (rail->Device, &txd->Payload, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, Payload), EP_PAYLOAD_SIZE);
}

void
ep3xmtr_flush_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail)
{
    EP3_RAIL *rail = XMTR_TO_RAIL (xmtrRail);
    struct list_head *el;
    unsigned long flags;

    switch (rail->Generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	/* only need to acquire/release the Lock to ensure that
	 * the node state transition has been noticed. */
	spin_lock_irqsave (&xmtr->Lock, flags);
	spin_unlock_irqrestore (&xmtr->Lock, flags);
	break;

    case EP_CB_FLUSH_FLUSHING:
	spin_lock_irqsave (&xmtr->Lock, flags);
	
	list_for_each (el, &xmtr->ActiveDescList) {
	    EP_TXD       *txd      = list_entry (el, EP_TXD, Link);
	    EP3_TXD_RAIL *txdRail  = (EP3_TXD_RAIL *) txd->TxdRail;
	    EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[txd->NodeId];
	    
	    if (!TXD_BOUND2RAIL(txdRail, xmtrRail) || nodeRail->State != EP_NODE_LOCAL_PASSIVATE)
		continue;
	    
	    if (EP_IS_RPC(txd->Envelope.Attr))
	    {
		if (! EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent))
		    nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		else if (! EP3_EVENT_FIRED (txdRail->DoneCookie, txdRail->TxdMain->DoneEvent))
		    nodeRail->MessageState |= EP_NODE_PASSIVE_MESSAGES;
	    }
	    else
	    {
		if (! EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent))
		    nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
	    }
	}
	spin_unlock_irqrestore (&xmtr->Lock, flags);
	break;

    default:
	panic ("ep3xmtr_flush_callback: invalid callback step\n");
	break;
    }
}

void
ep3xmtr_failover_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail)
{
    EP3_RAIL         *rail   = XMTR_TO_RAIL (xmtrRail);
    struct list_head  txdList;
    struct list_head *el, *nel;
    unsigned long flags;
#ifdef SUPPORT_RAIL_FAILOVER
    EP_COMMS_SUBSYS  *subsys = xmtr->Subsys;
#endif

    INIT_LIST_HEAD (&txdList);

    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD       *txd       = list_entry (el, EP_TXD, Link);
	EP3_TXD_RAIL *txdRail   = (EP3_TXD_RAIL *) txd->TxdRail;
	EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[txd->NodeId];
	    
	/* Only progress relocation of txd's bound to this rail */
	if (!TXD_BOUND2RAIL(txdRail, xmtrRail) || nodeRail->State != EP_NODE_PASSIVATED)
	    continue;
	
#ifdef SUPPORT_RAIL_FAILOVER
	/* Transmit data not been sent, so just restart on different rail */
	if (! EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent))
	{
	    EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_failover_callback - xmtr %p txd %p node %d unbind an retry\n", rail->Generic.Name, xmtr, txd, txd->NodeId);
	    
	    UnbindTxdFromRail (txd, txdRail);
	    
	    /* clear the done flags - so that it will be ignored if an event interrupt is generated */
	    txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
	    txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
	    txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;

	    /* reset all events, since non of them could have been set */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0); 				/* PCI write */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0); 				/* PCI write */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0); 				/* PCI write */
	    
	    FreeTxdRail (xmtrRail, txdRail);
	    
	    /* epcomms thread will restart on different rail */
	    ep_kthread_schedule (&subsys->Thread, lbolt);
	    continue;
	}

	if (EP_IS_RPC(txd->Envelope.Attr) && !EP3_EVENT_FIRED (txdRail->DoneCookie, txdRail->TxdMain->DoneEvent))
	{
	    if (EP_IS_NO_FAILOVER(txd->Envelope.Attr))
	    {
		EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_failover_callback - xmtr %p txd %p node %d - not able to failover\n",
			  rail->Generic.Name, xmtr, txd, txd->NodeId);

		list_del (&txd->Link);
		UnbindTxdFromRail (txd, txdRail);
		
		/* clear the done flags - so that it will be ignored if an event interrupt is generated */
		txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
		txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
		txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
		
		/* envelope and data events must have been set, so only clear the done event */
		EP_ASSERT (&rail->Generic, SDRAM_ASSERT(elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)) == 0));
		EP_ASSERT (&rail->Generic, SDRAM_ASSERT(elan3_sdram_readl (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)) == 0));

		elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0); 				/* PCI write */
		
		FreeTxdRail (xmtrRail, txdRail);
	    
		list_add_tail (&txd->Link, &txdList);
		continue;
	    }
	    EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_failover_callback - xmtr %p txd %p node %d passive\n", rail->Generic.Name, xmtr, txd, txd->NodeId);
	    
	    nodeRail->MessageState |= EP_NODE_PASSIVE_MESSAGES;
	    continue;
	}

	EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_failover_callback - xmtr %p txd %p node %d completed\n", rail->Generic.Name, xmtr, txd, txd->NodeId);
#endif

    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    while (! list_empty (&txdList)) 
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);

	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_CONN_RESET);
	
	FreeTxd (xmtr, txd);
    }
}


void
ep3xmtr_disconnect_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail)
{
    EP3_RAIL         *rail = XMTR_TO_RAIL (xmtrRail);
    struct list_head *el, *nel;
    struct list_head  txdList;
    unsigned long flags;
    
    INIT_LIST_HEAD (&txdList);

    spin_lock_irqsave (&xmtr->Lock, flags);

    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD       *txd       = list_entry (el, EP_TXD, Link);
	EP3_TXD_RAIL *txdRail   = (EP3_TXD_RAIL *) txd->TxdRail;
	EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[txd->NodeId];
	    
	if (!TXD_BOUND2RAIL(txdRail, xmtrRail) || nodeRail->State != EP_NODE_DISCONNECTING)
	    continue;
	
	if (EP3_EVENT_FIRED (txdRail->EnveCookie, txdRail->TxdMain->EnveEvent) &&
	    EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent) &&
	    EP3_EVENT_FIRED (txdRail->DoneCookie, txdRail->TxdMain->DoneEvent))
	{
	    EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_disconnect_callback - xmtr %p txd %p completed to node %d\n", rail->Generic.Name, xmtr, txd, txd->NodeId);
	    continue;
	}

	/* Remove from active list */
	list_del (&txd->Link);
	
	UnbindTxdFromRail (txd, txdRail);
	
	/* clear the done flags - so that it will be ignored if an event interrupt is generated */
	txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
	txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
	txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
	
	/* reset the envelope and data events, since only they could have been set */
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0);				/* PCI write */
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0);				/* PCI write */
	elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0);				/* PCI write */
	
	FreeTxdRail (xmtrRail, txdRail);
	    
	EPRINTF4 (DBG_XMTR, "%s: ep3xmtr_disconnect_callback - xmtr %p txd %p node %d not conected\n", rail->Generic.Name, xmtr, txd, txd->NodeId);

	/* add to the list of txd's which are to be completed */
	list_add_tail (&txd->Link, &txdList);
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);
    
    while (! list_empty (&txdList)) 
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);

	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_CONN_RESET);
	
	FreeTxd (xmtr, txd);
    }
}

int
ep3xmtr_poll_txd (EP_XMTR_RAIL *x, EP_TXD_RAIL *t, int how)
{
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) x;
    EP3_TXD_RAIL  *txdRail  = (EP3_TXD_RAIL *) t;
    EP_TXD        *txd      = txdRail->Generic.Txd;

    switch (how)
    {
    case ENABLE_TX_CALLBACK:
	if (EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	    EnableTransmitCallback (txd, txdRail);
	break;

    case DISABLE_TX_CALLBACK:
	if (EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	    DisableTransmitCallback (txd, txdRail);
	break;
    }

    if (EP3_EVENT_FIRED (txdRail->EnveCookie, txdRail->TxdMain->EnveEvent) &&
	EP3_EVENT_FIRED (txdRail->DataCookie, txdRail->TxdMain->DataEvent) &&
	EP3_EVENT_FIRED (txdRail->DoneCookie, txdRail->TxdMain->DoneEvent))
    {
	EPRINTF3 (DBG_XMTR, "%s: ep3xmtr_poll_txd: txd=%p XID=%llx completed\n", 
		  XMTR_TO_RAIL (xmtrRail)->Generic.Name, txd, (long long) txd->Envelope.Xid.Unique);

	ep_xmtr_txd_stat(xmtrRail->Generic.Xmtr,txd);

	UnbindTxdFromRail (txd, txdRail);
	
	/* clear the done flags - so that it will be ignored if an event interrupt is generated */
	txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
	txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
	txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
	
	FreeTxdRail (xmtrRail, txdRail);

	return 1;
    }

    return 0;
}

int
ep3xmtr_bind_txd (EP_TXD *txd, EP_XMTR_RAIL *x, unsigned int phase)
{
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) x;
    EP3_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);
    EP3_TXD_RAIL  *txdRail;
    E3_DMA_BE      dmabe;

    if ((txdRail = GetTxdRail (xmtrRail)) == NULL)
	return 0;

    switch (phase)
    {
    case EP_TXD_PHASE_ACTIVE:
	if (rail->Generic.Nodes[txd->NodeId].State != EP_NODE_CONNECTED)
	{
	    EPRINTF2 (DBG_XMTR, "%s: TransmitTxdOnRail: node %u not connected on this rail\n", rail->Generic.Name, txd->NodeId);

	    /* clear the done flags - so that it will be ignored if an event interrupt is generated */
	    txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
	    txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
	    txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;

	    /* reset all events, since non of them could have been set */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0); 				/* PCI write */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0); 				/* PCI write */
	    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0); 				/* PCI write */

	    FreeTxdRail (xmtrRail, txdRail);
	    return 0;
	}

	InitialiseTxdRail (txd, txdRail, phase);

	/* Initialise the dma descriptor */
	dmabe.s.dma_type            = E3_DMA_TYPE (DMA_BYTE, DMA_WRITE, DMA_QUEUED, EP3_DMAFAILCOUNT);
	dmabe.s.dma_size            = (EP_HAS_PAYLOAD(txd->Envelope.Attr) ? EP_INPUTQ_SIZE : EP_ENVELOPE_SIZE);
	dmabe.s.dma_source          = txdRail->TxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, Envelope);
	dmabe.s.dma_dest            = (E3_Addr) 0;
	dmabe.s.dma_destEvent       = EP_MSGQ_ADDR(txd->Service);
	dmabe.s.dma_destCookieVProc = EP_VP_DATA (txd->NodeId);
	dmabe.s.dma_srcEvent        = txdRail->TxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent);
	dmabe.s.dma_srcCookieVProc  = LocalCookie (rail, txd->NodeId);

	EPRINTF8 (DBG_XMTR, "%s: TransmitTxdOnRail: txd=%p txdRail=%p @ %x XID=%llx dest=%u srcEvent=%x srcCookie=%x\n", rail->Generic.Name, 
		  txd, txdRail, txdRail->TxdElanAddr, (long long) txd->Envelope.Xid.Unique, txd->NodeId, dmabe.s.dma_srcEvent, dmabe.s.dma_srcCookieVProc);
	
	BindTxdToRail (txd, txdRail);
	
	if (IssueDma (rail, &dmabe, EP_RETRY_LOW_PRI, FALSE) != ISSUE_COMMAND_OK)
	    QueueDmaForRetry (rail, &dmabe, EP_RETRY_LOW_PRI);
	break;

    case EP_TXD_PHASE_PASSIVE:
	InitialiseTxdRail (txd, txdRail, EP_TXD_PHASE_PASSIVE);				/* initialise as passive (updated envelope) */
	
	EP_XMTR_OP (txd->TxdRail->XmtrRail, UnbindTxd) (txd, EP_TXD_PHASE_PASSIVE);	/* unbind from existing rail */

	BindTxdToRail (txd, txdRail);							/* and bind it to our new rail */
	break;
    }

    return 1;
}

void
ep3xmtr_unbind_txd (EP_TXD *txd, unsigned int phase)
{
    EP3_TXD_RAIL  *txdRail  = (EP3_TXD_RAIL *) txd->TxdRail;
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP3_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);

    /* XXXX - TBD assertions on phase */

    UnbindTxdFromRail (txd, txdRail);
    
    /* clear the done flags - so that it will be ignored if an event interrupt is generated */
    txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
    
    /* reset the envelope and data events, since only they could have been set */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0);				/* PCI write */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0);				/* PCI write */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0);				/* PCI write */		
    
    FreeTxdRail (xmtrRail, txdRail);
}

long
ep3xmtr_check (EP_XMTR_RAIL *x, long nextRunTime)
{
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) x;

    if (xmtrRail->FreeDescCount < ep_txd_lowat && !AllocateTxdRailBlock(xmtrRail))
    {
	EPRINTF1 (DBG_RCVR,"%s: failed to grow txd rail pool\n", XMTR_TO_RAIL(xmtrRail)->Generic.Name);
		
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }
    
    return nextRunTime;
}

void
ep3xmtr_add_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail)
{
    EP3_XMTR_RAIL *xmtrRail;
    unsigned long  flags;

    KMEM_ZALLOC (xmtrRail, EP3_XMTR_RAIL *, sizeof (EP3_XMTR_RAIL), 1);

    spin_lock_init (&xmtrRail->FreeDescLock);
    kcondvar_init  (&xmtrRail->FreeDescSleep);
    INIT_LIST_HEAD (&xmtrRail->FreeDescList);
    INIT_LIST_HEAD (&xmtrRail->DescBlockList);

    xmtrRail->Generic.CommsRail = commsRail;
    xmtrRail->Generic.Xmtr      = xmtr;

    spin_lock_irqsave (&xmtr->Lock, flags);

    xmtr->Rails[commsRail->Rail->Number] = &xmtrRail->Generic;
    xmtr->RailMask |= EP_RAIL2RAILMASK(commsRail->Rail->Number);

    spin_unlock_irqrestore (&xmtr->Lock, flags);
}

void
ep3xmtr_del_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail)
{
    EP3_RAIL         *rail     = (EP3_RAIL *) commsRail->Rail;
    EP3_XMTR_RAIL    *xmtrRail = (EP3_XMTR_RAIL *) xmtr->Rails[commsRail->Rail->Number];
    unsigned long     flags;

    /* rail mask set as not usable */
    spin_lock_irqsave (&xmtr->Lock, flags);
    xmtr->RailMask &= ~EP_RAIL2RAILMASK (rail->Generic.Number);
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    /* wait for all txd's for this rail to become free */
    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);
    while (xmtrRail->FreeDescCount != xmtrRail->TotalDescCount)
    {
	xmtrRail->FreeDescWaiting++;
	kcondvar_wait (&xmtrRail->FreeDescSleep, &xmtrRail->FreeDescLock, &flags);
    }
    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);

    spin_lock_irqsave (&xmtr->Lock, flags);
    xmtr->Rails[commsRail->Rail->Number] = NULL;
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    /* need to free up the txd's and blocks */
    /* all the txd's accociated with DescBlocks must be in the FreeDescList */
    ASSERT (xmtrRail->TotalDescCount == xmtrRail->FreeDescCount);

    /* run through the DescBlockList deleting them */
    while (!list_empty (&xmtrRail->DescBlockList))
	FreeTxdRailBlock (xmtrRail, list_entry(xmtrRail->DescBlockList.next, EP3_TXD_RAIL_BLOCK , Link));
    
    /* it had better be empty after that */
    ASSERT ((xmtrRail->FreeDescCount == 0) && (xmtrRail->TotalDescCount == 0));

    spin_lock_destroy (&xmtrRail->FreeDescLock);
    kcondvar_destroy (&xmtrRail->FreeDescSleep);

    KMEM_FREE (xmtrRail, sizeof (EP3_XMTR_RAIL));
}

void
ep3xmtr_display_xmtr (DisplayInfo *di, EP_XMTR_RAIL *x)
{
    EP3_XMTR_RAIL *xmtrRail = (EP3_XMTR_RAIL *) x;
    EP3_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);
    struct list_head *el;
    unsigned long flags;
    int freeCount = 0;

    spin_lock_irqsave (&xmtrRail->FreeDescLock, flags);
    list_for_each (el, &xmtrRail->FreeDescList)
	freeCount++;
    spin_unlock_irqrestore (&xmtrRail->FreeDescLock, flags);

    (di->func)(di->arg, "                 Rail=%d Free=%d Total=%d (%d)\n",
		rail->Generic.Number, xmtrRail->FreeDescCount, xmtrRail->TotalDescCount, freeCount);
}

void
ep3xmtr_display_txd (DisplayInfo *di, EP_TXD_RAIL *t)
{
    EP3_TXD_RAIL      *txdRail   = (EP3_TXD_RAIL *) t;
    EP3_XMTR_RAIL     *xmtrRail  = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP3_TXD_RAIL_MAIN *txdMain   = txdRail->TxdMain;
    sdramaddr_t        txdElan   = txdRail->TxdElan;
    EP3_RAIL          *rail      = (EP3_RAIL *) xmtrRail->Generic.CommsRail->Rail;
    ELAN3_DEV         *dev       = rail->Device;
    
    (di->func)(di->arg, "     EnveEvent=%x DataEvent=%x DoneEvent=%x Rail=%s\n", 
	       txdMain->EnveEvent, txdMain->DataEvent, txdMain->DoneEvent, rail->Generic.Name);
    (di->func)(di->arg, "     EnveEvent=%x.%x DataEvent=%x.%x DoneEvent=%x.%x\n",
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count)),
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Type)),
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count)),
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Type)),
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count)),
	       elan3_sdram_readl (dev, txdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Type)));
}

int
ep3xmtr_check_txd_state (EP_TXD *txd)
{
    EP3_TXD_RAIL  *txdRail   = (EP3_TXD_RAIL *) txd->TxdRail;
    EP3_XMTR_RAIL *xmtrRail  = (EP3_XMTR_RAIL *) txdRail->Generic.XmtrRail;
    EP3_RAIL      *rail      = XMTR_TO_RAIL (xmtrRail);
    E3_Addr        enveEvent = txdRail->TxdElanAddr + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent);
    EP3_RETRY_DMA *retry = NULL;

    struct list_head *el;
    struct list_head *nel;
    unsigned long     flags;

    /*  is enevelope event is really not set */
    if (EP3_EVENT_FIRED (txdRail->EnveCookie, txdRail->TxdMain->EnveEvent )) 
	return (0);
    
    /* remove matching dma from stalled list */	    
    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    
    list_for_each_safe(el, nel,  &rail->DmaRetries[EP_RETRY_STABALISING]) {
	retry = list_entry (el, EP3_RETRY_DMA, Link);
	
	if ( retry->Dma.s.dma_srcEvent == enveEvent ) {
	    /* remove from retry list */
	    list_del (&retry->Link);
	    break; /* there can only be one */
	} 
    }
    ASSERT ( retry != NULL); /* must find one in list */
    ASSERT ( retry->Dma.s.dma_srcEvent == enveEvent ); /* better still be the right type then */    

    /* add to free list */
    list_add (&retry->Link, &rail->DmaRetryFreeList);

    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);    
       		
    UnbindTxdFromRail (txd, txdRail);
	
    /* clear the done flags - so that it will be ignored if an event interrupt is generated */
    txdRail->TxdMain->EnveEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DataEvent = EP3_EVENT_PRIVATE;
    txdRail->TxdMain->DoneEvent = EP3_EVENT_PRIVATE;
    
    /* reset the envelope and data events, since only they could have been set */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, EnveEvent.ev_Count), 0);				/* PCI write */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DataEvent.ev_Count), 0);				/* PCI write */
    elan3_sdram_writel (rail->Device, txdRail->TxdElan + offsetof (EP3_TXD_RAIL_ELAN, DoneEvent.ev_Count), 0);				/* PCI write */		
    
    FreeTxdRail (xmtrRail, txdRail);

    return (1);
}

void
ep3xmtr_fillout_rail_stats(EP_XMTR_RAIL *xmtr_rail, char *str) {
    /* no stats here yet */
    /* EP3_XMTR_RAIL * ep3xmtr_rail = (EP3_XMTR_RAIL *) xmtr_rail; */
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
