/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcommsTx_elan4.c,v 1.32.2.1 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcommsTx_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"
#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "epcomms_elan4.h"

#include <elan4/trtype.h>

#define XMTR_TO_COMMS(xmtrRail)		((EP4_COMMS_RAIL *) ((EP_XMTR_RAIL *) xmtrRail)->CommsRail)
#define XMTR_TO_RAIL(xmtrRail)		((EP4_RAIL *) ((EP_XMTR_RAIL *) xmtrRail)->CommsRail->Rail)
#define XMTR_TO_DEV(xmtrRail)		(XMTR_TO_RAIL(xmtrRail)->r_ctxt.ctxt_dev)
#define XMTR_TO_SUBSYS(xmtrRail)	(((EP_XMTR_RAIL *) xmtrRail)->Xmtr->Subsys)

#define TXD_TO_XMTR(txdRail)		((EP4_XMTR_RAIL *) txdRail->txd_generic.XmtrRail)
#define TXD_TO_RAIL(txdRail)		XMTR_TO_RAIL(TXD_TO_XMTR(txdRail))

static void txd_interrupt (EP4_RAIL *rail, void *arg);
static void poll_interrupt (EP4_RAIL *rail, void *arg);

static __inline__ int
on_list (struct list_head *ent, struct list_head *list)
{
    struct list_head *el;
    unsigned int count = 0;
    list_for_each (el, list) {
	if (el == ent)
	    count++;
    }
    return count;
}

static __inline__ void
__ep4_txd_assert_free (EP4_TXD_RAIL *txdRail, const char *file, const int line)
{
    EP4_XMTR_RAIL *xmtrRail = TXD_TO_XMTR (txdRail);
    ELAN4_DEV     *dev      = XMTR_TO_DEV (xmtrRail);
    register int   failed   = 0;
    
    if ((txdRail)->txd_retry_time     != 0)              failed |= (1 << 0);
    if ((txdRail)->txd_main->txd_env  != EP4_STATE_FREE) failed |= (1 << 1);
    if ((txdRail)->txd_main->txd_data != EP4_STATE_FREE) failed |= (1 << 2);
    if ((txdRail)->txd_main->txd_done != EP4_STATE_FREE) failed |= (1 << 3);

    if (sdram_assert)
    {
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)) >> 32)  != -32) failed |= (1 << 4);
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType)) >> 32) != 0)   failed |= (1 << 5);
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType)) >> 32) != 0)   failed |= (1 << 6);
    }

    if (failed)
    {
	printk ("__ep4_txd_assert_free: failed=%x txdRail=%p at %s:%d\n", failed, txdRail, file, line);

	ep_debugf (DBG_DEBUG, "__ep4_txd_assert_free: failed=%x txdRail=%p at %s:%d\n", failed, txdRail, file, line);
	ep4xmtr_display_txd (&di_ep_debug, &txdRail->txd_generic);

	(txdRail)->txd_retry_time     = 0;
	(txdRail)->txd_main->txd_env  = EP4_STATE_FREE;
	(txdRail)->txd_main->txd_data = EP4_STATE_FREE;
	(txdRail)->txd_main->txd_done = EP4_STATE_FREE;

	if (sdram_assert)
	{
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)  + 4, -32);
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType) + 4, 0);
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType) + 4, 0);
	}
	EP_ASSFAIL (XMTR_TO_RAIL(xmtrRail), "__ep4_txd_assert_free");
    }
}

static __inline__ void
__ep4_txd_assert_finished (EP4_TXD_RAIL *txdRail, const char *file, const int line)
{
    EP4_XMTR_RAIL *xmtrRail = TXD_TO_XMTR (txdRail);
    ELAN4_DEV     *dev      = XMTR_TO_DEV (xmtrRail);
    register int   failed   = 0;
    
    if ((txdRail)->txd_retry_time     != 0)                  failed |= (1 << 0);
    if ((txdRail)->txd_main->txd_env  != EP4_STATE_FINISHED) failed |= (1 << 1);
    if ((txdRail)->txd_main->txd_data != EP4_STATE_FINISHED) failed |= (1 << 2);
    if ((txdRail)->txd_main->txd_done != EP4_STATE_FINISHED) failed |= (1 << 3);
    
    if (sdram_assert)
    {
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)) >> 32)  != -32) failed |= (1 << 4);
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType)) >> 32) != 0)   failed |= (1 << 5);
	if ((int)(elan4_sdram_readq (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType)) >> 32) != 0)   failed |= (1 << 6);
    }

    if (failed)
    {
	printk ("__ep4_txd_assert_finished: failed=%x txdRail=%p at %s:%d\n", failed, txdRail, file, line);

	ep_debugf (DBG_DEBUG, "__ep4_txd_assert_finished: failed=%x txdRail=%p at %s:%d\n", failed, txdRail, file, line);
	ep4xmtr_display_txd (&di_ep_debug, &txdRail->txd_generic);

	(txdRail)->txd_retry_time     = 0;
	(txdRail)->txd_main->txd_env  = EP4_STATE_FINISHED;
	(txdRail)->txd_main->txd_data = EP4_STATE_FINISHED;
	(txdRail)->txd_main->txd_done = EP4_STATE_FINISHED;

	if (sdram_assert)
	{
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)  + 4, -32);
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType) + 4, 0);
	    elan4_sdram_writel (dev, (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType) + 4, 0);
	}
	EP_ASSFAIL (XMTR_TO_RAIL(xmtrRail), "__ep4_txd_assert_finished");
    }
}

static __inline__ int
__ep4_txd_assfail (EP4_TXD_RAIL *txdRail, const char *expr, const char *file, const int line)
{
    EP4_XMTR_RAIL *xmtrRail = TXD_TO_XMTR (txdRail);

    printk ("__ep4_txd_assfail: %s:%d '%s'\n", file, line, expr);

    ep_debugf (DBG_DEBUG, "__ep4_txd_assfail: %s:%d '%s'\n", file, line, expr);
    ep4xmtr_display_txd (&di_ep_debug, &txdRail->txd_generic);

    EP_ASSFAIL (XMTR_TO_RAIL (xmtrRail), "__ep4_txd_assfail");

    return 0;
}

#define EP4_TXD_ASSERT(txdRail, EX)		((void) ((EX) || (__ep4_txd_assfail(txdRail, #EX, __FILE__, __LINE__))))
#define EP4_TXD_ASSERT_FREE(txdRail)		__ep4_txd_assert_free(txdRail, __FILE__, __LINE__)
#define EP4_TXD_ASSERT_FINISHED(txdRail)	__ep4_txd_assert_finished(txdRail, __FILE__, __LINE__)

static int
alloc_txd_block (EP4_XMTR_RAIL *xmtrRail)
{
    EP4_RAIL           *rail = XMTR_TO_RAIL(xmtrRail);
    ELAN4_DEV          *dev  = XMTR_TO_DEV(xmtrRail);
    EP4_TXD_RAIL_BLOCK *blk;
    EP4_TXD_RAIL_MAIN  *txdMain;
    EP_ADDR		txdMainAddr;
    sdramaddr_t		txdElan;
    EP_ADDR		txdElanAddr;
    EP4_TXD_RAIL       *txdRail;
    unsigned long       flags;
    int                 i;

    KMEM_ZALLOC (blk, EP4_TXD_RAIL_BLOCK *, sizeof (EP4_TXD_RAIL_BLOCK), 1);

    if (blk == NULL)
	return 0;

    if ((txdElan = ep_alloc_elan (&rail->r_generic, EP4_TXD_RAIL_ELAN_SIZE * EP4_NUM_TXD_PER_BLOCK, 0, &txdElanAddr)) == (sdramaddr_t) 0)
    {
	KMEM_FREE (blk, sizeof (EP4_TXD_RAIL_BLOCK));
	return 0;
    }

    if ((txdMain = ep_alloc_main (&rail->r_generic, EP4_TXD_RAIL_MAIN_SIZE * EP4_NUM_TXD_PER_BLOCK, 0, &txdMainAddr)) == (EP4_TXD_RAIL_MAIN *) NULL)
    {
	ep_free_elan (&rail->r_generic, txdElanAddr, EP4_TXD_RAIL_ELAN_SIZE * EP4_NUM_TXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP4_TXD_RAIL_BLOCK));
	return 0;
    }

    if (ep4_reserve_dma_retries (rail, EP4_NUM_TXD_PER_BLOCK, 0) != 0)
    {
	ep_free_main (&rail->r_generic, blk->blk_txds[0].txd_main_addr, EP4_TXD_RAIL_MAIN_SIZE * EP4_NUM_TXD_PER_BLOCK);
	ep_free_elan (&rail->r_generic, txdElanAddr, EP4_TXD_RAIL_ELAN_SIZE * EP4_NUM_TXD_PER_BLOCK);
	KMEM_FREE (blk, sizeof (EP4_TXD_RAIL_BLOCK));
	return 0;
    }

    for (txdRail = &blk->blk_txds[0], i = 0; i < EP4_NUM_TXD_PER_BLOCK; i++, txdRail++)
    {
	txdRail->txd_generic.XmtrRail = &xmtrRail->xmtr_generic;
	txdRail->txd_elan             = txdElan;
	txdRail->txd_elan_addr        = txdElanAddr;
	txdRail->txd_main             = txdMain;
	txdRail->txd_main_addr        = txdMainAddr;

	/* We only need to reserve space for one command stream, since the sten packet
	 * can only be retrying *before* the dma source event is set.
	 * reserve bytes of "event" cq space for the completion write + interrupt */
	if ((txdRail->txd_ecq = ep4_get_ecq (rail, EP4_ECQ_EVENT, EP4_INTR_CMD_NDWORDS)) == NULL)
	    goto failed;

	/* register the main interrupt cookies */
	ep4_register_intcookie (rail, &txdRail->txd_intcookie, txdElanAddr + offsetof (EP4_TXD_RAIL_ELAN, txd_done), txd_interrupt, txdRail);

	/* initialise the events */
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CopySource),
			    txdElanAddr + offsetof (EP4_TXD_RAIL_ELAN, txd_env_cmd));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CopyDest),
			    txdRail->txd_ecq->ecq_addr);

	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_WritePtr),
			    txdMainAddr + offsetof (EP4_TXD_RAIL_MAIN, txd_data));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_WriteValue),
			    EP4_STATE_FINISHED);

	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CopySource),
			    txdElanAddr + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CopyDest),
			    txdRail->txd_ecq->ecq_addr);

	/* Initialise the command streams */
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env_cmd.c_write_cmd),
			    WRITE_DWORD_CMD | (txdMainAddr + offsetof (EP4_TXD_RAIL_MAIN, txd_env)));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env_cmd.c_write_value),
			    EP4_STATE_FAILED);
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env_cmd.c_intr_cmd),
			    INTERRUPT_CMD | (txdRail->txd_intcookie.int_val << E4_MAIN_INT_SHIFT));

	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_write_cmd),
			    WRITE_DWORD_CMD | (txdMainAddr + offsetof (EP4_TXD_RAIL_MAIN, txd_done)));
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_write_value),
			    EP4_STATE_FINISHED);
	elan4_sdram_writeq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_intr_cmd),
			    INTERRUPT_CMD | (txdRail->txd_intcookie.int_val << E4_MAIN_INT_SHIFT));

	txdMain->txd_env  = EP4_STATE_FREE;
	txdMain->txd_data = EP4_STATE_FREE;
	txdMain->txd_done = EP4_STATE_FREE;

	/* move onto next descriptor */
	txdElan     += EP4_TXD_RAIL_ELAN_SIZE;
	txdElanAddr += EP4_TXD_RAIL_ELAN_SIZE;
	txdMain      = (EP4_TXD_RAIL_MAIN *) ((unsigned long) txdMain + EP4_TXD_RAIL_MAIN_SIZE);
	txdMainAddr += EP4_TXD_RAIL_MAIN_SIZE;
    }

    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);

    list_add  (&blk->blk_link, &xmtrRail->xmtr_blocklist);

    xmtrRail->xmtr_totalcount += EP4_NUM_TXD_PER_BLOCK;
    xmtrRail->xmtr_freecount  += EP4_NUM_TXD_PER_BLOCK;

    for (i = 0; i < EP4_NUM_TXD_PER_BLOCK; i++)
	list_add (&blk->blk_txds[i].txd_generic.Link, &xmtrRail->xmtr_freelist);

    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);

    return 1;

 failed:
    while (--i >= 0)
    {
	ep4_put_ecq (rail, txdRail->txd_ecq, EP4_INTR_CMD_NDWORDS);
	ep4_deregister_intcookie (rail, &txdRail->txd_intcookie);
    }
    ep4_release_dma_retries (rail, EP4_NUM_TXD_PER_BLOCK);

    ep_free_main (&rail->r_generic, blk->blk_txds[0].txd_main_addr, EP4_TXD_RAIL_MAIN_SIZE * EP4_NUM_TXD_PER_BLOCK);
    ep_free_elan (&rail->r_generic, blk->blk_txds[0].txd_elan_addr, EP4_TXD_RAIL_ELAN_SIZE * EP4_NUM_TXD_PER_BLOCK);

    KMEM_FREE (blk, sizeof (EP4_TXD_RAIL_BLOCK));

    return 0;
}

static void
free_txd_block (EP4_XMTR_RAIL *xmtrRail, EP4_TXD_RAIL_BLOCK *blk)
{
    EP4_RAIL     *rail = XMTR_TO_RAIL (xmtrRail);
    EP4_TXD_RAIL *txdRail;
    unsigned long flags;
    int           i;

    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);

    list_del (&blk->blk_link);

    xmtrRail->xmtr_totalcount -= EP4_NUM_TXD_PER_BLOCK;

    for (txdRail = &blk->blk_txds[0], i = 0; i < EP4_NUM_TXD_PER_BLOCK; i++, txdRail++)
    {
	xmtrRail->xmtr_freecount--;

	ep4_put_ecq (rail, txdRail->txd_ecq, EP4_INTR_CMD_NDWORDS);

	ep4_deregister_intcookie (rail, &txdRail->txd_intcookie);

	list_del (&txdRail->txd_generic.Link);
    }
    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);

    ep4_release_dma_retries (rail, EP4_NUM_TXD_PER_BLOCK);

    ep_free_main (&rail->r_generic, blk->blk_txds[0].txd_main_addr, EP4_TXD_RAIL_MAIN_SIZE * EP4_NUM_TXD_PER_BLOCK);
    ep_free_elan (&rail->r_generic, blk->blk_txds[0].txd_elan_addr, EP4_TXD_RAIL_ELAN_SIZE * EP4_NUM_TXD_PER_BLOCK);

    KMEM_FREE (blk, sizeof (EP4_TXD_RAIL_BLOCK));
}

static EP4_TXD_RAIL *
get_txd_rail (EP4_XMTR_RAIL *xmtrRail)
{
    EP_COMMS_SUBSYS  *subsys = XMTR_TO_SUBSYS(xmtrRail);
    EP4_TXD_RAIL     *txdRail;
    unsigned long flags;
    int low_on_txds;

    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);

    if (list_empty (&xmtrRail->xmtr_freelist))
	txdRail = NULL;
    else
    {
	txdRail = list_entry (xmtrRail->xmtr_freelist.next, EP4_TXD_RAIL, txd_generic.Link);

	EP4_TXD_ASSERT_FREE(txdRail);

	list_del (&txdRail->txd_generic.Link);

	xmtrRail->xmtr_freecount--;
    }
    /* Wakeup the descriptor primer thread if there's not many left */
    low_on_txds = (xmtrRail->xmtr_freecount < ep_txd_lowat);

    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);

    if (low_on_txds)
	ep_kthread_schedule (&subsys->Thread, lbolt);


    return (txdRail);
}

static void
free_txd_rail (EP4_XMTR_RAIL *xmtrRail, EP4_TXD_RAIL *txdRail)
{
    unsigned long flags;

    EP4_TXD_ASSERT_FREE(txdRail);

    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);
    
    list_add (&txdRail->txd_generic.Link, &xmtrRail->xmtr_freelist);

    xmtrRail->xmtr_freecount++;

    if (xmtrRail->xmtr_freewaiting)
    {
	xmtrRail->xmtr_freewaiting--;
	kcondvar_wakeupall (&xmtrRail->xmtr_freesleep, &xmtrRail->xmtr_freelock);
    }

    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);
}

static void
bind_txd_rail (EP_TXD *txd, EP4_TXD_RAIL *txdRail)
{
    EPRINTF6 (DBG_XMTR, "%s: bind_txd_rail: txd=%p txdRail=%p XID=%08x.%08x.%016llx\n", 
	      XMTR_TO_RAIL(txdRail->txd_generic.XmtrRail)->r_generic.Name, txd, txdRail, 
	      txd->Envelope.Xid.Generation, txd->Envelope.Xid.Handle, (long long)txd->Envelope.Xid.Unique);

    txd->TxdRail = &txdRail->txd_generic;
    txdRail->txd_generic.Txd = txd;
}

static void
unbind_txd_rail (EP_TXD *txd, EP4_TXD_RAIL *txdRail)
{
    EP4_TXD_ASSERT (txdRail, txd->TxdRail == &txdRail->txd_generic && txdRail->txd_generic.Txd == txd);

    EPRINTF6 (DBG_XMTR, "%s: unbind_txd_rail: txd=%p txdRail=%p XID=%08x.%08x.%016llx\n", 
	      XMTR_TO_RAIL(txdRail->txd_generic.XmtrRail)->r_generic.Name, txd, txdRail, 
	      txd->Envelope.Xid.Generation, txd->Envelope.Xid.Handle, (long long)txd->Envelope.Xid.Unique);


    txdRail->txd_generic.Txd = NULL; 
    txd->TxdRail = NULL;
}

static void
initialise_txd (EP_TXD *txd, EP4_TXD_RAIL *txdRail, unsigned int phase)
{
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) txdRail->txd_generic.XmtrRail;
    EP4_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);
    ELAN4_DEV     *dev      = rail->r_ctxt.ctxt_dev;
    
    /* Flush the Elan TLB if mappings have changed */
    ep_perrail_dvma_sync (&rail->r_generic);
    
    /* Initialise the per-rail fields in the envelope */
    txd->Envelope.TxdRail = txdRail->txd_elan_addr;
    txd->Envelope.NodeId  = rail->r_generic.Position.pos_nodeid;

    /* Allocate a network error fixup cookie */
    txdRail->txd_cookie = ep4_neterr_cookie (rail, txd->NodeId) | EP4_COOKIE_STEN;

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    if ( epdebug_check_sum ) 
	txd->Envelope.CheckSum = ep_calc_check_sum( txd->Xmtr->Subsys->Subsys.Sys, &txd->Envelope, txd->Envelope.Frags, txd->Envelope.nFrags);
    else
#endif
	txd->Envelope.CheckSum = 0;  

    /* Initialise the per-rail events */
    switch (phase)
    {
    case EP_TXD_PHASE_ACTIVE:
    {
	unsigned int nsets = (txd->Envelope.nFrags ? txd->Envelope.nFrags : 1) + ( EP_IS_MULTICAST(txd->Envelope.Attr) ? 1 : 0);

	if (! EP_IS_RPC(txd->Envelope.Attr))
	{
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType),
				E4_EVENT_INIT_VALUE (-32 * nsets, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

	    txdRail->txd_main->txd_data = EP4_STATE_FINISHED;
	}
	else
	{
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType),
				E4_EVENT_INIT_VALUE(-32 * nsets , E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType),
				E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));

	    txdRail->txd_main->txd_data = EP4_STATE_ACTIVE;
	}
		   
	txdRail->txd_main->txd_env  = EP4_STATE_ACTIVE;
	txdRail->txd_main->txd_done = EP4_STATE_ACTIVE;
	break;
    }

    case EP_TXD_PHASE_PASSIVE:
	EP4_TXD_ASSERT (txdRail, EP_IS_RPC(txd->Envelope.Attr));
	
	txdRail->txd_main->txd_env  = EP4_STATE_FINISHED;
	txdRail->txd_main->txd_data = EP4_STATE_FINISHED;
	txdRail->txd_main->txd_done = EP4_STATE_ACTIVE;

	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	break;
    }

   if (EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_intr_cmd), NOP_CMD);
}

static void
terminate_txd_rail (EP4_XMTR_RAIL *xmtrRail, EP4_TXD_RAIL *txdRail)
{
    EP4_SDRAM_ASSERT (TXD_TO_RAIL(txdRail),\
		      (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),\
		      E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));\

    /* clear the done flags - so that it will be ignored if an event interrupt is generated */
    txdRail->txd_main->txd_env  = EP4_STATE_FREE;
    txdRail->txd_main->txd_data = EP4_STATE_FREE;
    txdRail->txd_main->txd_done = EP4_STATE_FREE;

#if defined(DEBUG_ASSERT)
    if (sdram_assert)
    {
	ELAN4_DEV *dev = XMTR_TO_RAIL (xmtrRail)->r_ctxt.ctxt_dev;

	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));
	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
    }
#endif 
}

static void
defer_txd_rail (EP4_TXD_RAIL *txdRail)
{
    EP4_XMTR_RAIL   *xmtrRail = TXD_TO_XMTR(txdRail);
    EP4_RAIL        *rail     = XMTR_TO_RAIL(xmtrRail);
    ELAN4_DEV       *dev      = rail->r_ctxt.ctxt_dev;
    EP_COMMS_SUBSYS *subsys   = XMTR_TO_SUBSYS(xmtrRail);

    EPRINTF5 (DBG_XMTR, "%s: defer_txd_rail: xmtrRail=%p txdRail=%p env/data (%d,%d) not finished\n",
	      rail->r_generic.Name, xmtrRail, txdRail, (int)txdRail->txd_main->txd_env, (int)txdRail->txd_main->txd_data);
		    
    /* transmit has completed, but the data dma has not completed
     * (because of network error fixup), we queue the txdRail onto a list
     * to be polled for completion later.
     */
    if (txdRail->txd_retry_time)
    {
	EP4_TXD_ASSERT (txdRail, (on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]) == 1 ||
				  on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]) == 1));

	list_del (&txdRail->txd_retry_link);

	txdRail->txd_main->txd_env = EP4_STATE_FINISHED;

	/* re-initialise the envelope event */
	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
    }
    
    txdRail->txd_retry_time = lbolt;
	
    list_add_tail (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_POLL]);
	
    ep_kthread_schedule (&subsys->Thread, lbolt);
}

static void
finalise_txd (EP_TXD *txd, EP4_TXD_RAIL *txdRail)
{
    EP4_XMTR_RAIL   *xmtrRail = TXD_TO_XMTR(txdRail);

    EP4_TXD_ASSERT_FINISHED (txdRail);

    unbind_txd_rail (txd, txdRail);
    
    terminate_txd_rail (xmtrRail, txdRail);
    free_txd_rail (xmtrRail, txdRail);
}

static void
txd_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_TXD_RAIL    *txdRail  = (EP4_TXD_RAIL *) arg;
    EP4_XMTR_RAIL   *xmtrRail = TXD_TO_XMTR(txdRail);
    EP_XMTR         *xmtr     = xmtrRail->xmtr_generic.Xmtr;
    int              delay    = 1;
    EP_TXD          *txd;
    unsigned long    flags;

    spin_lock_irqsave (&xmtr->Lock, flags);
    for (;;)
    {
	if (txdRail->txd_main->txd_done == EP4_STATE_FINISHED || txdRail->txd_main->txd_env == EP4_STATE_FAILED)
	    break;
	
	/* The write to txd_done could be held up in the PCI bridge even though
	 * we've seen the interrupt cookie.  Unlike elan3, there is no possibility
	 * of spurious interrupts since we flush the command queues on node 
	 * disconnection and the txcallback mechanism */
	mb();

	if (delay > EP4_EVENT_FIRING_TLIMIT)
	{
	    spin_unlock_irqrestore (&xmtr->Lock, flags);

	    EP_ASSFAIL (XMTR_TO_RAIL(xmtrRail), "txd_interrupt - not finished\n");
	    return;
	}
	DELAY (delay);
	delay <<= 1;
    }

    txd = txdRail->txd_generic.Txd;

    if (txdRail->txd_main->txd_env == EP4_STATE_FAILED)
    {
	spin_lock (&xmtrRail->xmtr_retrylock);

	EP4_TXD_ASSERT (txdRail, txdRail->txd_retry_time == 0); 			/* cannot be on retry/poll list */
	EP4_TXD_ASSERT (txdRail, txdRail->txd_main->txd_done != EP4_STATE_FINISHED); 	/* data xfer cannot have finished */

	if (TxdShouldStabalise (&txdRail->txd_generic, &rail->r_generic))
	{
	    EPRINTF6 (DBG_STABILISE, "%s: txd_interrupt: stablise xmtrRail=%p txdRail=%p txd=%p XID=%llx dest=%u\n", rail->r_generic.Name,
		      xmtrRail, txdRail, txd, (long long)txd->Envelope.Xid.Unique, txd->NodeId);

	    txdRail->txd_retry_time = lbolt;			/* indicate on retry list */
	    
	    list_add_tail (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]);
	}
	else
	{
	    EPRINTF6 (DBG_RETRY, "%s: txd_interrupt: retry xmtrRail=%p txdRail=%p txd=%p XID=%llx dest=%u\n", rail->r_generic.Name,
		      xmtrRail, txdRail, txd, (long long)txd->Envelope.Xid.Unique, txd->NodeId);

	    txdRail->txd_retry_time = lbolt + EP_RETRY_LOW_PRI_TIME;		/* XXXX: backoff ? */
	    
	    list_add_tail (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]);
	    
	    ep_kthread_schedule (&rail->r_retry_thread, txdRail->txd_retry_time);
	}
	spin_unlock (&xmtrRail->xmtr_retrylock);

	spin_unlock_irqrestore (&xmtr->Lock, flags);
	return;
    }

    EP4_TXD_ASSERT (txdRail, txd != NULL && !(EP_IS_NO_INTERRUPT(txd->Envelope.Attr)));

    EPRINTF6 (DBG_XMTR, "%s: txd_interrupt: xmtrRail=%p txdRail=%p txd=%p XID=%llx dest=%u\n", rail->r_generic.Name,
	      xmtrRail, txdRail, txd, (long long)txd->Envelope.Xid.Unique, txd->NodeId);
	     
    if (txdRail->txd_main->txd_env != EP4_STATE_FINISHED || txdRail->txd_main->txd_data != EP4_STATE_FINISHED)
    {
	defer_txd_rail (txdRail);

	spin_unlock_irqrestore (&xmtr->Lock, flags);
    }
    else
    {
	/* remove from active transmit list */
	list_del (&txd->Link);

	ep_xmtr_txd_stat(xmtr,txd);

	finalise_txd (txd, txdRail);
	
	spin_unlock_irqrestore (&xmtr->Lock, flags);
	
	txd->Handler (txd, txd->Arg, EP_SUCCESS);
	
	FreeTxd (xmtr, txd);
    }
}

static void
poll_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_XMTR_RAIL   *xmtrRail = (EP4_XMTR_RAIL *) arg;

    ep_poll_transmits (xmtrRail->xmtr_generic.Xmtr);
}

void
issue_envelope_packet (EP4_XMTR_RAIL *xmtrRail, EP4_TXD_RAIL *txdRail)
{
    EP_TXD    *txd    = txdRail->txd_generic.Txd;
    ELAN4_CQ  *cq     = xmtrRail->xmtr_cq;
    E4_uint64 *blk0   = (E4_uint64 *) &txd->Envelope;
    E4_uint64 *blk1   = EP_HAS_PAYLOAD(txd->Envelope.Attr) ? (E4_uint64 *) &txd->Payload : NULL;
    E4_Addr    qaddr  = EP_MSGQ_ADDR(txd->Service);

    EP4_SDRAM_ASSERT (TXD_TO_RAIL(txdRail),\
		      (txdRail)->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),\
		      E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));\

    elan4_open_packet (cq, OPEN_PACKET (0, PACK_OK | RESTART_COUNT_ZERO, EP_VP_DATA(txd->NodeId)));
    elan4_sendtrans0 (cq, TR_INPUT_Q_GETINDEX, EP_MSGQ_ADDR(txd->Service));
	    
    /* send the payload if present */
    if (blk0) elan4_sendtransp (cq, TR_WRITE(128 >> 3, 0, TR_DATATYPE_BYTE), 0,   blk0);
    if (blk1) elan4_sendtransp (cq, TR_WRITE(128 >> 3, 0, TR_DATATYPE_BYTE), 128, blk1);

    elan4_sendtrans1 (cq, TR_INPUT_Q_COMMIT, qaddr, txdRail->txd_cookie);

    elan4_guard (cq, GUARD_CHANNEL (1) | GUARD_TEST(0, PACK_OK) | GUARD_RESET (EP4_TXD_STEN_RETRYCOUNT));
    elan4_write_dword_cmd (cq, txdRail->txd_main_addr + offsetof (EP4_TXD_RAIL_MAIN, txd_env), EP4_STATE_FINISHED);
	    
    elan4_guard (cq, GUARD_CHANNEL (1) | GUARD_TEST(0, RESTART_COUNT_ZERO) | GUARD_RESET (EP4_TXD_STEN_RETRYCOUNT));
    elan4_set_event_cmd (cq, txdRail->txd_elan_addr + offsetof (EP4_TXD_RAIL_ELAN, txd_env));
    
    elan4_write_dword_cmd (cq, xmtrRail->xmtr_main_addr + offsetof (EP4_XMTR_RAIL_MAIN, xmtr_flowcnt), ++xmtrRail->xmtr_flowcnt);
}

void
ep4xmtr_flush_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail)
{
    EP4_RAIL       *rail      = XMTR_TO_RAIL (xmtrRail);
    EP4_COMMS_RAIL *commsRail = XMTR_TO_COMMS (xmtrRail);
    struct list_head *el, *nel;
    unsigned long flags;

    switch (rail->r_generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	/* need to acquire/release the Lock to ensure that the node state
	 * transition has been noticed and no new envelopes are queued to 
	 * nodes which are passivating. */
	spin_lock_irqsave (&xmtr->Lock, flags);

	/* Then we insert a "setevent" into the command queue to flush
	 * through the envelopes which have already been submitted */
	ep4comms_flush_setevent (commsRail, xmtrRail->xmtr_cq);

	spin_unlock_irqrestore (&xmtr->Lock, flags);

	break;

    case EP_CB_FLUSH_FLUSHING:
	/* remove any envelopes which are retrying to nodes which are going down */
	spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
	list_for_each_safe (el, nel, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]) {
	    EP4_TXD_RAIL *txdRail  = list_entry (el, EP4_TXD_RAIL, txd_retry_link);
	    EP_TXD       *txd      = txdRail->txd_generic.Txd;
	    EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[txd->NodeId];
	    
	    EP4_TXD_ASSERT (txdRail, txdRail->txd_main->txd_env == EP4_STATE_FAILED);
	    
	    if (nodeRail->State == EP_NODE_LOCAL_PASSIVATE)
	    {
		EPRINTF2 (DBG_XMTR, "%s; ep4xmtr_flush_callback: removing txdRail %p from retry list\n", rail->r_generic.Name, txdRail);
		
		EP4_TXD_ASSERT (txdRail, txdRail->txd_retry_time != 0);

		list_del (&txdRail->txd_retry_link);
		list_add_tail (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]);
	    }
	}
	spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);

	/* Determine whether we have active or passive messages to 
	 * any node which is passivating */
	spin_lock_irqsave (&xmtr->Lock, flags);
	list_for_each (el, &xmtr->ActiveDescList) {
	    EP_TXD       *txd      = list_entry (el, EP_TXD, Link);
	    EP4_TXD_RAIL *txdRail  = (EP4_TXD_RAIL *) txd->TxdRail;
	    EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[txd->NodeId];
	    
	    if (txdRail == NULL || txdRail->txd_generic.XmtrRail != &xmtrRail->xmtr_generic || nodeRail->State != EP_NODE_LOCAL_PASSIVATE)
		continue;
	    
	    EPRINTF5 (DBG_XMTR, "%s: flush txd=%p txdRail=%p data=%llx done=%llx\n", rail->r_generic.Name,
		      txd, txdRail, (long long)txdRail->txd_main->txd_data, (long long)txdRail->txd_main->txd_done);

	    if (EP_IS_RPC(txd->Envelope.Attr))
	    {
		if (txdRail->txd_main->txd_data == EP4_STATE_ACTIVE)
		    nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
		else if (txdRail->txd_main->txd_data == EP4_STATE_ACTIVE)
		    nodeRail->MessageState |= EP_NODE_PASSIVE_MESSAGES;
	    }
	    else
	    {
		if (txdRail->txd_main->txd_data == EP4_STATE_ACTIVE)
		    nodeRail->MessageState |= EP_NODE_ACTIVE_MESSAGES;
	    }
	}
	spin_unlock_irqrestore (&xmtr->Lock, flags);
	break;

    default:
	panic ("ep4xmtr_flush_callback: invalid callback step\n");
	break;
    }
}

void
ep4xmtr_failover_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail)
{
    EP4_RAIL         *rail   = XMTR_TO_RAIL (xmtrRail);
    struct list_head  txdList;
    struct list_head *el, *nel;
    unsigned long flags;

    INIT_LIST_HEAD (&txdList);

    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD       *txd       = list_entry (el, EP_TXD, Link);
	EP4_TXD_RAIL *txdRail   = (EP4_TXD_RAIL *) txd->TxdRail;
	EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[txd->NodeId];
	    
	/* Only progress relocation of txd's bound to this rail */
	if (! TXD_BOUND2RAIL (txdRail, xmtrRail) || nodeRail->State != EP_NODE_PASSIVATED)
	    continue;
	
	/* XXXX - no rail failover for now ....*/

	EPRINTF4 (DBG_XMTR, "%s: ep4xmtr_failover_callback - xmtr %p txd %p node %d completed\n", rail->r_generic.Name, xmtr, txd, txd->NodeId);
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    while (! list_empty (&txdList)) 
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);

	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_CONN_RESET);
	
	FreeTxd (xmtr, txd);
    }
}


void
ep4xmtr_disconnect_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail)
{
    EP4_RAIL         *rail = XMTR_TO_RAIL (xmtrRail);
    ELAN4_DEV        *dev  = rail->r_ctxt.ctxt_dev;
    struct list_head *el, *nel;
    struct list_head  txdList;
    unsigned long flags;
    
    INIT_LIST_HEAD (&txdList);

    spin_lock_irqsave (&xmtr->Lock, flags);

    list_for_each_safe (el, nel, &xmtr->ActiveDescList) {
	EP_TXD       *txd       = list_entry (el, EP_TXD, Link);
	EP4_TXD_RAIL *txdRail   = (EP4_TXD_RAIL *) txd->TxdRail;
	EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[txd->NodeId];
	    
	if ( ! TXD_BOUND2RAIL (txdRail, xmtrRail) || nodeRail->State != EP_NODE_DISCONNECTING)
	    continue;
	
	if (txdRail->txd_main->txd_done == EP4_STATE_ACTIVE)
	{

	    EPRINTF8 (DBG_DISCON, "ep4xmtr_disconnect_callback: txdRail=%p : events %llx,%llx,%llx done %llx,%llx,%llx retry %lx\n",txdRail,
		      elan4_sdram_readq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)),
		      elan4_sdram_readq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType)),
		      elan4_sdram_readq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType)),
		      (long long)txdRail->txd_main->txd_env, (long long)txdRail->txd_main->txd_data, (long long)txdRail->txd_main->txd_done,
		      txdRail->txd_retry_time);
		       
	    if (txdRail->txd_retry_time)
	    {
		/* re-initialise the envelope event */
		elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
				    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
		
		EP4_TXD_ASSERT (txdRail, on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]) == 1);

		txdRail->txd_retry_time  = 0;

		list_del (&txdRail->txd_retry_link);
	    }

	    /* Remove from active list */
	    list_del (&txd->Link);
	
	    unbind_txd_rail (txd, txdRail);

	    terminate_txd_rail (xmtrRail, txdRail);
	    free_txd_rail (xmtrRail, txdRail);
	    
	    EPRINTF4 (DBG_XMTR, "%s: ep4xmtr_disconnect_callback - xmtr %p txd %p node %d not conected\n", rail->r_generic.Name, xmtr, txd, txd->NodeId);

	    /* add to the list of txd's which are to be completed */
	    list_add_tail (&txd->Link, &txdList);
	}
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);
    
    while (! list_empty (&txdList)) 
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);

	list_del (&txd->Link);

	txd->Handler (txd, txd->Arg, EP_CONN_RESET);
	
	FreeTxd (xmtr, txd);
    }
}

void
ep4xmtr_neterr_flush (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP4_COMMS_RAIL   *commsRail = XMTR_TO_COMMS (xmtrRail);
    unsigned long flags;
    
    spin_lock_irqsave (&xmtr->Lock, flags);

    /* insert a "setevent" into the command queue to flush
     * through the envelopes which have already been submitted */
    ep4comms_flush_setevent (commsRail, xmtrRail->xmtr_cq);

    spin_unlock_irqrestore (&xmtr->Lock, flags);
}

void
ep4xmtr_neterr_check (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP4_RAIL *rail = XMTR_TO_RAIL (xmtrRail);
    ELAN4_DEV *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el;
    unsigned long flags;
    
    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each (el, &xmtr->ActiveDescList) {
	EP_TXD       *txd     = list_entry (el, EP_TXD, Link);
	EP4_TXD_RAIL *txdRail = (EP4_TXD_RAIL *) txd->TxdRail;
	    
	if ( ! TXD_BOUND2RAIL (txdRail, xmtrRail) || txd->NodeId != nodeId)
	    continue;
	
	/* The only non-dma associated with a txd is the initial sten packet, if it has been acked 
	 * and the neterr cookie matches, then change it to look like it's been acked since the
	 * INPUT_Q_COMMIT transaction has already been executed */
	if (txdRail->txd_main->txd_env == EP4_STATE_FAILED && (txdRail->txd_cookie == cookies[0] || txdRail->txd_cookie == cookies[1]))
	{
	    EPRINTF4 (DBG_NETWORK_ERROR, "%s: ep4xmtr_neterr_callback: cookie <%lld%s%s%s%s> matches txd %p txdRail %p\n", 
		     rail->r_generic.Name, (long long)EP4_COOKIE_STRING(txdRail->txd_cookie), txd, txdRail);

	    EP4_TXD_ASSERT (txdRail, txdRail->txd_retry_time != 0);

	    txdRail->txd_main->txd_env = EP4_STATE_FINISHED;

	    /* re-initialise the envelope event */
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
				E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	    
	    spin_lock (&xmtrRail->xmtr_retrylock);

	    EP4_TXD_ASSERT (txdRail, (on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]) == 1 ||
				      on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]) == 1));

	    txdRail->txd_retry_time = 0;

	    list_del (&txdRail->txd_retry_link);

	    spin_unlock (&xmtrRail->xmtr_retrylock);
	}
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);
}

int
ep4xmtr_poll_txd (EP_XMTR_RAIL *x, EP_TXD_RAIL *t, int how)
{
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) x;
    ELAN4_DEV     *dev      = XMTR_TO_DEV (xmtrRail);
    EP4_TXD_RAIL  *txdRail  = (EP4_TXD_RAIL *) t;
    EP_TXD        *txd      = txdRail->txd_generic.Txd;

    if (! EP_IS_NO_INTERRUPT(txd->Envelope.Attr))
	return 0;

    switch (how)
    {
    case ENABLE_TX_CALLBACK:
	if (!EP_IS_INTERRUPT_ENABLED(txd->Envelope.Attr))
	{
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_intr_cmd),
				INTERRUPT_CMD | (xmtrRail->xmtr_intcookie.int_val << E4_MAIN_INT_SHIFT));

	    txd->Envelope.Attr |= EP_INTERRUPT_ENABLED;
	}
	break;

    case DISABLE_TX_CALLBACK:
	if (EP_IS_INTERRUPT_ENABLED(txd->Envelope.Attr & EP_INTERRUPT_ENABLED))
	{
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_intr_cmd), NOP_CMD);

	    txd->Envelope.Attr &= ~EP_INTERRUPT_ENABLED;
	}
    }
    
    if (txdRail->txd_main->txd_env == EP4_STATE_FINISHED && txdRail->txd_main->txd_data == EP4_STATE_FINISHED && txdRail->txd_main->txd_done == EP4_STATE_FINISHED)
    {
	EPRINTF3 (DBG_XMTR, "%s: ep4xmtr_poll_txd: txd=%p XID=%llx completed\n",
		  XMTR_TO_RAIL (xmtrRail)->r_generic.Name, txd, (long long)txd->Envelope.Xid.Unique);
	
	elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_done_cmd.c_intr_cmd),
			    INTERRUPT_CMD | (txdRail->txd_intcookie.int_val << E4_MAIN_INT_SHIFT));


	ep_xmtr_txd_stat(xmtrRail->xmtr_generic.Xmtr,txd);

	finalise_txd (txd, txdRail);

	return 1;
    }

    return 0;
}

int
ep4xmtr_bind_txd (EP_TXD *txd, EP_XMTR_RAIL *x, unsigned int phase)
{
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) x;
    EP4_RAIL      *rail     = XMTR_TO_RAIL (xmtrRail);
    EP4_TXD_RAIL  *txdRail;
    unsigned long  flags;

    if ((txdRail = get_txd_rail (xmtrRail)) == NULL)
	return 0;
    
    switch (phase)
    {
    case EP_TXD_PHASE_ACTIVE:
	if (rail->r_generic.Nodes[txd->NodeId].State != EP_NODE_CONNECTED)
	{
	    EPRINTF2 (DBG_XMTR, "%s: ep4xmtr_bind_txd: node %u not connected on this rail\n", rail->r_generic.Name, txd->NodeId);

	    free_txd_rail (xmtrRail, txdRail);
	    return 0;
	}

	initialise_txd (txd, txdRail, EP_TXD_PHASE_ACTIVE);

	bind_txd_rail (txd, txdRail);
	
	/* generate the STEN packet to transfer the envelope */
	spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
	if (((int) (xmtrRail->xmtr_flowcnt - xmtrRail->xmtr_main->xmtr_flowcnt)) < EP4_XMTR_FLOWCNT)
	    issue_envelope_packet (xmtrRail, txdRail);
	else
	{
	    txdRail->txd_retry_time = lbolt;

	    list_add_tail (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]);

	    ep_kthread_schedule (&rail->r_retry_thread, txdRail->txd_retry_time);
	}
	spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);
	break;

    case EP_TXD_PHASE_PASSIVE:
	initialise_txd (txd, txdRail, EP_TXD_PHASE_PASSIVE);
	
	EP_XMTR_OP (txd->TxdRail->XmtrRail, UnbindTxd) (txd, EP_TXD_PHASE_PASSIVE);	/* unbind from existing rail */

	bind_txd_rail (txd, txdRail);							/* and bind it to our new rail */
	break;
    }

    return 1;
}

void
ep4xmtr_unbind_txd (EP_TXD *txd, unsigned int phase)
{
    /* XXXX - TBD */
}

long
ep4xmtr_check (EP_XMTR_RAIL *x, long nextRunTime)
{
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) x;
    EP_XMTR       *xmtr     = xmtrRail->xmtr_generic.Xmtr;
    struct list_head  txdList;
    struct list_head *el, *nel;
    unsigned long flags;

    INIT_LIST_HEAD (&txdList);

    if (xmtrRail->xmtr_freecount < ep_txd_lowat && !alloc_txd_block (xmtrRail))
    {
	EPRINTF1 (DBG_RCVR,"%s: failed to grow txd rail pool\n", XMTR_TO_RAIL(xmtrRail)->r_generic.Name);
		
	if (nextRunTime == 0 || AFTER (nextRunTime, lbolt + RESOURCE_RETRY_TIME))
	    nextRunTime = lbolt + RESOURCE_RETRY_TIME;
    }

    spin_lock_irqsave (&xmtr->Lock, flags);
    list_for_each_safe (el, nel, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_POLL]) {
	EP4_TXD_RAIL *txdRail = list_entry (el, EP4_TXD_RAIL, txd_retry_link);

	if (txdRail->txd_main->txd_env != EP4_STATE_FINISHED || txdRail->txd_main->txd_data != EP4_STATE_FINISHED)
	{
	    ep_debugf (DBG_XMTR, "%s: ep4xmtr_check: xmtrRail=%p txdRail=%p env/data (%d,%d) not finished\n",
		       XMTR_TO_RAIL(xmtrRail)->r_generic.Name, xmtrRail, txdRail, (int)txdRail->txd_main->txd_env, (int)txdRail->txd_main->txd_data);
		    
	    nextRunTime = lbolt + HZ;
	}
	else
	{
	    EP_TXD *txd = txdRail->txd_generic.Txd;

	    ep_debugf (DBG_XMTR, "%s: ep4xmtr_check: xmtrRail=%p txdRail=%p env/data (%d,%d) finished\n",
		       XMTR_TO_RAIL(xmtrRail)->r_generic.Name, xmtrRail, txdRail, (int)txdRail->txd_main->txd_env, (int)txdRail->txd_main->txd_data);

	    EPRINTF5 (DBG_XMTR, "%s: ep4xmtr_check: xmtrRail=%p txdRail=%p env/data (%d,%d) finished\n",
		      XMTR_TO_RAIL(xmtrRail)->r_generic.Name, xmtrRail, txdRail, (int)txdRail->txd_main->txd_env, (int)txdRail->txd_main->txd_data);
	    EPRINTF3  (DBG_XMTR, "%s:    done %x data %x\n", XMTR_TO_RAIL(xmtrRail)->r_generic.Name,
		       (uint)(txdRail->txd_elan_addr + offsetof (EP4_TXD_RAIL_ELAN, txd_done)),
		       (uint)(txdRail->txd_elan_addr + offsetof (EP4_TXD_RAIL_ELAN, txd_data)));

	    EP4_TXD_ASSERT (txdRail, txdRail->txd_retry_time != 0);

	    /* remove txd from active list and add to list to call handlers */
	    list_del (&txd->Link);
	    list_add_tail (&txd->Link, &txdList);

	    /* remove and free of txdRail */
	    txdRail->txd_retry_time = 0;
	    list_del (&txdRail->txd_retry_link);

	    finalise_txd (txd, txdRail);

	}
    }
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    while (! list_empty (&txdList))
    {
	EP_TXD *txd = list_entry (txdList.next, EP_TXD, Link);

	list_del (&txd->Link);

	ep_xmtr_txd_stat (xmtr,txd);

	txd->Handler (txd, txd->Arg, EP_SUCCESS);

	FreeTxd (xmtr, txd);
    }

    return nextRunTime;
}

unsigned long
ep4xmtr_retry (EP4_RAIL *rail, void *arg, unsigned long nextRunTime)
{
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) arg;
    ELAN4_DEV     *dev      = XMTR_TO_DEV(xmtrRail);
    unsigned long  flags;

    spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
    while (! list_empty (&xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]))
    {
	EP4_TXD_RAIL *txdRail = list_entry (xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY].next, EP4_TXD_RAIL, txd_retry_link);

	if (BEFORE (lbolt, txdRail->txd_retry_time))
	{
	    if (nextRunTime == 0 || AFTER (nextRunTime, txdRail->txd_retry_time))
		nextRunTime = txdRail->txd_retry_time;

	    break;
	}

	if (((int) (xmtrRail->xmtr_flowcnt - xmtrRail->xmtr_main->xmtr_flowcnt)) < EP4_XMTR_FLOWCNT)
	{
	    txdRail->txd_retry_time = 0;

	    list_del (&txdRail->txd_retry_link);
	    
	    /* re-initialise the envelope event */
	    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
				E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	    
	    EPRINTF3 (DBG_RETRY, "%s: ep4xmtr_retry: re-issue envelope packet to %d for txdRail=%p\n", 
		      rail->r_generic.Name, txdRail->txd_generic.Txd->Envelope.NodeId, txdRail);
	    
	    txdRail->txd_main->txd_env = EP4_STATE_ACTIVE;
	    
	    issue_envelope_packet (xmtrRail, txdRail);
	}
	else
	{
	    EPRINTF2 (DBG_RETRY, "%s: ep4xmtr_retry: cannot re-issue envelope packet to %d\n", rail->r_generic.Name, txdRail->txd_generic.Txd->Envelope.NodeId);

	    if (nextRunTime == 0 || AFTER (nextRunTime, txdRail->txd_retry_time))
		nextRunTime = txdRail->txd_retry_time;

	    break;
	}
    }
    spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);
    
    return nextRunTime;
}

void
ep4xmtr_add_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail)
{
    EP4_RAIL         *rail   = (EP4_RAIL *) commsRail->Rail;
    EP_COMMS_SUBSYS  *subsys = xmtr->Subsys;
    EP4_XMTR_RAIL    *xmtrRail;
    unsigned long     flags;
    int		      i;

    KMEM_ZALLOC (xmtrRail, EP4_XMTR_RAIL *, sizeof (EP4_XMTR_RAIL), 1);

    spin_lock_init (&xmtrRail->xmtr_freelock);
    kcondvar_init  (&xmtrRail->xmtr_freesleep);
    INIT_LIST_HEAD (&xmtrRail->xmtr_freelist);
    INIT_LIST_HEAD (&xmtrRail->xmtr_blocklist);

    for (i = 0; i < EP4_TXD_NUM_LISTS; i++)
	INIT_LIST_HEAD (&xmtrRail->xmtr_retrylist[i]);
    spin_lock_init (&xmtrRail->xmtr_retrylock);

    xmtrRail->xmtr_generic.CommsRail = commsRail;
    xmtrRail->xmtr_generic.Xmtr      = xmtr;

    xmtrRail->xmtr_main = ep_alloc_main (&rail->r_generic, sizeof (EP4_XMTR_RAIL_MAIN), 0, &xmtrRail->xmtr_main_addr);
    xmtrRail->xmtr_cq   = elan4_alloccq (&rail->r_ctxt, EP4_XMTR_CQSIZE, CQ_EnableAllBits, CQ_Priority);

    xmtrRail->xmtr_retryops.op_func = ep4xmtr_retry;
    xmtrRail->xmtr_retryops.op_arg  = xmtrRail;

    ep4_add_retry_ops (rail, &xmtrRail->xmtr_retryops);

    ep4_register_intcookie (rail, &xmtrRail->xmtr_intcookie, xmtrRail->xmtr_main_addr,
			    poll_interrupt, xmtrRail);

    spin_lock_irqsave (&xmtr->Lock, flags);

    xmtr->Rails[commsRail->Rail->Number] = &xmtrRail->xmtr_generic;
    xmtr->RailMask |= EP_RAIL2RAILMASK(commsRail->Rail->Number);

    spin_unlock_irqrestore (&xmtr->Lock, flags);

    ep_kthread_schedule (&subsys->Thread, lbolt);

    ep_procfs_xmtr_add_rail(&(xmtrRail->xmtr_generic));
}

void
ep4xmtr_del_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail)
{
    EP4_RAIL         *rail     = (EP4_RAIL *) commsRail->Rail;
    EP4_XMTR_RAIL    *xmtrRail = (EP4_XMTR_RAIL *) xmtr->Rails[commsRail->Rail->Number];
    unsigned long     flags;

    /* rail mask set as not usable */
    spin_lock_irqsave (&xmtr->Lock, flags);
    xmtr->RailMask &= ~EP_RAIL2RAILMASK (rail->r_generic.Number);
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    ep_procfs_xmtr_del_rail(&(xmtrRail->xmtr_generic));

    /* wait for all txd's for this rail to become free */
    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);
    while (xmtrRail->xmtr_freecount != xmtrRail->xmtr_totalcount)
    {
	xmtrRail->xmtr_freewaiting++;
	kcondvar_wait (&xmtrRail->xmtr_freesleep, &xmtrRail->xmtr_freelock, &flags);
    }
    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);

    spin_lock_irqsave (&xmtr->Lock, flags);
    xmtr->Rails[commsRail->Rail->Number] = NULL;
    spin_unlock_irqrestore (&xmtr->Lock, flags);

    /* all the txd's accociated with DescBlocks must be in the freelist */
    ASSERT (xmtrRail->xmtr_totalcount == xmtrRail->xmtr_freecount);

    /* run through the DescBlockList deleting them */
    while (!list_empty (&xmtrRail->xmtr_blocklist))
	free_txd_block (xmtrRail, list_entry(xmtrRail->xmtr_blocklist.next, EP4_TXD_RAIL_BLOCK , blk_link));
    
    /* it had better be empty after that */
    ASSERT ((xmtrRail->xmtr_freecount == 0) && (xmtrRail->xmtr_totalcount == 0));

    ep4_deregister_intcookie (rail, &xmtrRail->xmtr_intcookie);

    ep4_remove_retry_ops (rail, &xmtrRail->xmtr_retryops);

    elan4_freecq (&rail->r_ctxt, xmtrRail->xmtr_cq);
    ep_free_main (&rail->r_generic, xmtrRail->xmtr_main_addr, sizeof (EP4_XMTR_RAIL_MAIN));

    spin_lock_destroy (&xmtrRail->xmtr_retrylock);

    spin_lock_destroy (&xmtrRail->xmtr_freelock);
    kcondvar_destroy (&xmtrRail->xmtr_freesleep);

    KMEM_FREE (xmtrRail, sizeof (EP4_XMTR_RAIL));
}

void
ep4xmtr_display_xmtr (DisplayInfo *di, EP_XMTR_RAIL *x)
{
    EP4_XMTR_RAIL    *xmtrRail     = (EP4_XMTR_RAIL *) x;
    EP4_RAIL         *rail         = XMTR_TO_RAIL (xmtrRail);
    unsigned int      freeCount    = 0;
    unsigned int      pollCount    = 0;
    unsigned int      stalledCount = 0;
    unsigned int      retryCount   = 0;
    struct list_head *el;
    unsigned long     flags;

    spin_lock_irqsave (&xmtrRail->xmtr_freelock, flags);
    list_for_each (el, &xmtrRail->xmtr_freelist)
	freeCount++;
    spin_unlock_irqrestore (&xmtrRail->xmtr_freelock, flags);

    spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
    list_for_each (el, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_POLL])
	pollCount++;
    list_for_each (el, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED])
	stalledCount++;
    list_for_each (el, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY])
	retryCount++;
    spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);

    (di->func)(di->arg, "        rail=%d free=%d total=%d (%d) (retry %d,%d,%d)\n",
	       rail->r_generic.Number, xmtrRail->xmtr_freecount, xmtrRail->xmtr_totalcount, 
	       freeCount, pollCount, stalledCount, retryCount);
    (di->func)(di->arg, "        cq %d flowcnt %lld,%lld\n", elan4_cq2num (xmtrRail->xmtr_cq), xmtrRail->xmtr_flowcnt, xmtrRail->xmtr_main->xmtr_flowcnt);
}

void
ep4xmtr_display_txd (DisplayInfo *di, EP_TXD_RAIL *t)
{
    EP4_TXD_RAIL      *txdRail  = (EP4_TXD_RAIL *) t;
    EP4_XMTR_RAIL     *xmtrRail = TXD_TO_XMTR(txdRail);
    EP4_TXD_RAIL_MAIN *txdMain  = txdRail->txd_main;
    sdramaddr_t        txdElan  = txdRail->txd_elan;
    EP4_RAIL          *rail     = XMTR_TO_RAIL (xmtrRail);
    ELAN4_DEV         *dev      = XMTR_TO_DEV (xmtrRail);
    char	      *list     = "";
    unsigned long      flags;

    spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
    if (txdRail->txd_retry_time)
    {
	if (on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_POLL]))
	    list = " poll";
	else if (on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]))
	    list = " stalled";
	else if (on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_RETRY]))
	    list = " retry";
	else
	    list = " ERROR";
    }
    spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);

    (di->func)(di->arg, "      Rail %d txd %p elan %lx (%x) main %p (%x) cookie <%lld%s%s%s%s> ecq %d %s\n", rail->r_generic.Number,
	       txdRail, txdRail->txd_elan, txdRail->txd_elan_addr, txdRail->txd_main, txdRail->txd_main_addr, 
	       EP4_COOKIE_STRING(txdRail->txd_cookie), elan4_cq2num (txdRail->txd_ecq->ecq_cq), list);
    
    (di->func)(di->arg, "        env  %016llx %016llx %016llx -> %016llx\n",
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType)),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_Params[0])),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_Params[1])),
	       txdMain->txd_env);
    (di->func)(di->arg, "        data %016llx %016llx %016llx -> %016llx\n",
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_CountAndType)),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_Params[0])),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_data.ev_Params[1])),
	       txdMain->txd_data);
    (di->func)(di->arg, "        done %016llx %016llx %016llx -> %016llx\n",
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_CountAndType)),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_Params[0])),
	       elan4_sdram_readq (dev, txdElan + offsetof (EP4_TXD_RAIL_ELAN, txd_done.ev_Params[1])),
	       txdMain->txd_done);
}

int
ep4xmtr_check_txd_state (EP_TXD *txd) 
{
    EP4_TXD_RAIL  *txdRail  = (EP4_TXD_RAIL *) txd->TxdRail;
    EP4_XMTR_RAIL *xmtrRail = (EP4_XMTR_RAIL *) txdRail->txd_generic.XmtrRail;
    ELAN4_DEV     *dev      = XMTR_TO_DEV (xmtrRail);
    unsigned long  flags;

    if (txdRail->txd_main->txd_env == EP4_STATE_FINISHED)
	return 0;

    EP4_TXD_ASSERT (txdRail, txdRail->txd_retry_time != 0);

    spin_lock_irqsave (&xmtrRail->xmtr_retrylock, flags);
    EP4_TXD_ASSERT (txdRail, on_list (&txdRail->txd_retry_link, &xmtrRail->xmtr_retrylist[EP4_TXD_LIST_STALLED]) == 1);

    list_del (&txdRail->txd_retry_link);
    txdRail->txd_retry_time  = 0;
    spin_unlock_irqrestore (&xmtrRail->xmtr_retrylock, flags);
    
    /* re-initialise the envelope event */
    elan4_sdram_writeq (dev, txdRail->txd_elan + offsetof (EP4_TXD_RAIL_ELAN, txd_env.ev_CountAndType),
			E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS));
	    
    unbind_txd_rail (txd, txdRail);

    terminate_txd_rail (xmtrRail, txdRail);
    free_txd_rail (xmtrRail, txdRail);

    return 1;
}

void
ep4xmtr_fillout_rail_stats(EP_XMTR_RAIL *xmtr_rail, char *str) {
    /* no stats here yet */
    /* EP4_XMTR_RAIL * ep4xmtr_rail = (EP4_XMTR_RAIL *) xmtr_rail; */
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
