/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcomms_elan3.c,v 1.60 2004/08/03 11:34:34 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "kcomm_elan3.h"
#include "epcomms_elan3.h"

void
ep3comms_flush_callback (void *arg, statemap_t *map)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[commsRail->Rail->Number])
	    ep3xmtr_flush_callback (xmtr, (EP3_XMTR_RAIL *) xmtr->Rails[commsRail->Rail->Number]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[commsRail->Rail->Number])
	    ep3rcvr_flush_callback (rcvr, (EP3_RCVR_RAIL *) rcvr->Rails[commsRail->Rail->Number]);
    }
    kmutex_unlock (&subsys->Lock);
}

void
ep3comms_failover_callback (void *arg, statemap_t *map)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[commsRail->Rail->Number])
	    ep3xmtr_failover_callback (xmtr, (EP3_XMTR_RAIL *) xmtr->Rails[commsRail->Rail->Number]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[commsRail->Rail->Number])
	    ep3rcvr_failover_callback (rcvr, (EP3_RCVR_RAIL *) rcvr->Rails[commsRail->Rail->Number]);
    }
    kmutex_unlock (&subsys->Lock);
}

void
ep3comms_disconnect_callback (void *arg, statemap_t *map)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[commsRail->Rail->Number])
	    ep3xmtr_disconnect_callback (xmtr, (EP3_XMTR_RAIL *) xmtr->Rails[commsRail->Rail->Number]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[commsRail->Rail->Number])
	    ep3rcvr_disconnect_callback (rcvr, (EP3_RCVR_RAIL *) rcvr->Rails[commsRail->Rail->Number]);
    }
    kmutex_unlock (&subsys->Lock);
}

EP_COMMS_RAIL *
ep3comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *r)
{
    EP3_RAIL         *rail   = (EP3_RAIL *) r;
    ELAN3_DEV        *dev    = rail->Device;
    EP3_COMMS_RAIL   *commsRail;
    EP3_InputQueue    qdesc;
    int i;

    KMEM_ZALLOC (commsRail, EP3_COMMS_RAIL *, sizeof (EP3_COMMS_RAIL), TRUE);

    if (commsRail == NULL)
	return NULL;
    
    commsRail->Generic.Ops.DelRail          = ep3comms_del_rail;
    commsRail->Generic.Ops.DisplayRail      = ep3comms_display_rail;
    commsRail->Generic.Ops.Rcvr.AddRail     = ep3rcvr_add_rail;
    commsRail->Generic.Ops.Rcvr.DelRail     = ep3rcvr_del_rail;
    commsRail->Generic.Ops.Rcvr.Check       = ep3rcvr_check;
    commsRail->Generic.Ops.Rcvr.QueueRxd    = ep3rcvr_queue_rxd;
    commsRail->Generic.Ops.Rcvr.RpcPut      = ep3rcvr_rpc_put;
    commsRail->Generic.Ops.Rcvr.RpcGet      = ep3rcvr_rpc_get;
    commsRail->Generic.Ops.Rcvr.RpcComplete = ep3rcvr_rpc_complete;

    commsRail->Generic.Ops.Rcvr.StealRxd    = ep3rcvr_steal_rxd;

    commsRail->Generic.Ops.Rcvr.FillOutRailStats = ep3rcvr_fillout_rail_stats;

    commsRail->Generic.Ops.Rcvr.DisplayRcvr = ep3rcvr_display_rcvr;
    commsRail->Generic.Ops.Rcvr.DisplayRxd  = ep3rcvr_display_rxd;

    commsRail->Generic.Ops.Xmtr.AddRail     = ep3xmtr_add_rail;
    commsRail->Generic.Ops.Xmtr.DelRail     = ep3xmtr_del_rail;
    commsRail->Generic.Ops.Xmtr.Check       = ep3xmtr_check;
    commsRail->Generic.Ops.Xmtr.BindTxd     = ep3xmtr_bind_txd;
    commsRail->Generic.Ops.Xmtr.UnbindTxd   = ep3xmtr_unbind_txd;
    commsRail->Generic.Ops.Xmtr.PollTxd     = ep3xmtr_poll_txd;
    commsRail->Generic.Ops.Xmtr.CheckTxdState = ep3xmtr_check_txd_state;

    commsRail->Generic.Ops.Xmtr.DisplayXmtr = ep3xmtr_display_xmtr;
    commsRail->Generic.Ops.Xmtr.DisplayTxd  = ep3xmtr_display_txd;

    commsRail->Generic.Ops.Xmtr.FillOutRailStats = ep3xmtr_fillout_rail_stats;

    /* Allocate the input queues at their fixed elan address */
    if (! (commsRail->QueueDescs = ep_alloc_memory_elan (r, EP_EPCOMMS_QUEUE_BASE, roundup (EP_MSG_NSVC * sizeof (EP3_InputQueue), PAGESIZE), EP_PERM_ALL, 0)))
    {
	KMEM_FREE (commsRail, sizeof (EP3_COMMS_RAIL));
	return NULL;
    }

    qdesc.q_state          = E3_QUEUE_FULL;
    qdesc.q_base           = 0;
    qdesc.q_top            = 0;
    qdesc.q_fptr           = 0;
    qdesc.q_bptr           = 0;
    qdesc.q_size           = 0;
    qdesc.q_event.ev_Count = 0;
    qdesc.q_event.ev_Type  = 0;

    /* Initialise all queue entries to be full */
    for (i = 0; i < EP_MSG_NSVC; i++)
	elan3_sdram_copyl_to_sdram (dev, &qdesc, commsRail->QueueDescs + (i * sizeof (EP3_InputQueue)), sizeof (EP3_InputQueue));

    ep_register_callback (r, EP_CB_FLUSH_FILTERING, ep3comms_flush_callback,      commsRail);
    ep_register_callback (r, EP_CB_FLUSH_FLUSHING,  ep3comms_flush_callback,      commsRail);
    ep_register_callback (r, EP_CB_FAILOVER,        ep3comms_failover_callback,   commsRail);
    ep_register_callback (r, EP_CB_DISCONNECTING,   ep3comms_disconnect_callback, commsRail);

    return (EP_COMMS_RAIL *) commsRail;
}

void
ep3comms_del_rail (EP_COMMS_RAIL *r)
{
    EP3_COMMS_RAIL *commsRail = (EP3_COMMS_RAIL *) r;
    EP_RAIL        *rail      = commsRail->Generic.Rail;

    ep_remove_callback (rail, EP_CB_FLUSH_FILTERING, ep3comms_flush_callback,      commsRail);
    ep_remove_callback (rail, EP_CB_FLUSH_FLUSHING,  ep3comms_flush_callback,      commsRail);
    ep_remove_callback (rail, EP_CB_FAILOVER,        ep3comms_failover_callback,   commsRail);
    ep_remove_callback (rail, EP_CB_DISCONNECTING,   ep3comms_disconnect_callback, commsRail);

    ep_free_memory_elan (rail, EP_EPCOMMS_QUEUE_BASE);

    KMEM_FREE (commsRail, sizeof (EP3_COMMS_RAIL));
}

void
ep3comms_display_rail (EP_COMMS_RAIL *r)
{
    
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
