/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __EPCOMMS_ELAN3_H
#define __EPCOMMS_ELAN3_H

#ident "@(#)$Id: epcomms_elan3.h,v 1.28 2004/11/12 10:55:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan3.h,v $ */

#define EP3_DMAFAILCOUNT		3


/* Main/Elan spinlock */
typedef struct ep3_spinlock_elan 
{
    volatile E3_uint32	sl_lock;		/* main wants a lock */
    volatile E3_uint32	sl_seq;			/* thread owns this word */
    /* NOTE: The lock/seq words must be within the same 32-byte Elan cache-line */
    E3_uint64		sl_pad[14];		/* pad to 64-bytes */
} EP3_SPINLOCK_ELAN;

/* Declare this as a main memory cache block for efficiency */
typedef struct ep3_spinlock_main {
    volatile E3_uint32	sl_seq;			/* copy of seq number updated by Elan */
    volatile E3_uint32	sl_pad[15];		/* pad to 64-bytes */
} EP3_SPINLOCK_MAIN;

#if defined (__ELAN3__)

extern void ep3_spinblock (EP3_SPINLOCK_ELAN *, EP3_SPINLOCK_MAIN *);

#define EP3_SPINENTER(SLE,SL) \
do {\
	(SLE)->sl_seq++; \
	if ((SLE)->sl_lock) \
		ep3_spinblock(SLE, SL);\
} while (0)

#define EP3_SPINEXIT(SLE,SL) \
do {\
	(SL)->sl_seq = (SLE)->sl_seq;\
} while (0)

#else

#define EP3_SPINENTER(DEV,SLE,SL) do { \
    E3_uint32 seq; \
\
    mb();\
    elan3_sdram_writel (DEV, (SLE) + offsetof (EP3_SPINLOCK_ELAN, sl_lock), 1);\
    mb();\
    seq = elan3_sdram_readl (DEV, (SLE) + offsetof (EP3_SPINLOCK_ELAN, sl_seq));\
    while (seq != (SL)->sl_seq)\
    {\
	while ((SL)->sl_seq == (seq - 1))\
	{\
	    mb();\
\
	    DELAY (1); \
	}\
	seq = elan3_sdram_readl (DEV, (SLE) + offsetof (EP3_SPINLOCK_ELAN, sl_seq));\
    }\
} while (0)

#define EP3_SPINEXIT(DEV,SLE,SL) do { \
	wmb(); \
    	elan3_sdram_writel (DEV, (SLE) + offsetof (EP3_SPINLOCK_ELAN, sl_lock), 0);\
	mmiob(); \
} while (0)

#endif /* ! __ELAN3__ */

/* per-rail elan memory portion receive descriptor */
typedef struct ep3_rxd_rail_elan
{
    E3_DMA		Dmas[EP_MAXFRAG+1];			/* Dma's for fetching data/putting data & status blk */
    E3_Event		ChainEvent[EP_MAXFRAG];			/* Events to chain dmas */
    E3_BlockCopyEvent	DataEvent;				/* message received block event */
    E3_BlockCopyEvent	DoneEvent;				/* RPC status block event */
    
    EP_NMD		Data;					/* Network mapping handle for receive data */

    E3_Addr		RxdMain;				/* pointer to main memory portion */

    E3_Addr		Next;					/* linked list when on pending list (elan address) */

    E3_uint64		MainAddr;				/* kernel address of ep_rxd_main */
} EP3_RXD_RAIL_ELAN;

#define EP3_RXD_RAIL_ELAN_SIZE	roundup (sizeof (EP3_RXD_RAIL_ELAN), E3_DMA_ALIGN)

/* per-rail main memory portion of receive descriptor */
typedef struct ep3_rxd_rail_main
{
    E3_uint32		DataEvent;				/* dest for done event */
    E3_uint32		DoneEvent;				/* dest for done event */
} EP3_RXD_RAIL_MAIN;

#define EP3_RXD_RAIL_MAIN_SIZE	roundup (sizeof(EP3_RXD_RAIL_MAIN), sizeof (E3_uint32))

#if !defined(__ELAN3__)
/* Kernel memory portion of per-rail receive descriptor */
typedef struct ep3_rxd_rail
{
    EP_RXD_RAIL		Generic;				/* generic rxd rail */

    EP3_COOKIE		DataCookie;				/* Event cookie */
    EP3_COOKIE		DoneCookie;				/* Event cookie */
    EP3_COOKIE		ChainCookie[EP_MAXFRAG];		/* Event cookie */

    sdramaddr_t		RxdElan;				/* per-rail elan receive descriptor */
    E3_Addr		RxdElanAddr;				/*   and elan address */

    EP3_RXD_RAIL_MAIN  *RxdMain;				/* per-rail main receive descriptor */
    E3_Addr		RxdMainAddr;				/*   and elan address */

    EP_BACKOFF	        Backoff;				/* dma backoff */
} EP3_RXD_RAIL;

#define EP3_NUM_RXD_PER_BLOCK	16

typedef struct ep3_rxd_rail_block
{
    struct list_head 	Link;
    
    EP3_RXD_RAIL        Rxd[EP3_NUM_RXD_PER_BLOCK];
} EP3_RXD_RAIL_BLOCK;

#endif /* ! __ELAN3__ */

typedef struct ep3_rcvr_rail_elan				/* Elan memory service structure */
{
    EP3_SPINLOCK_ELAN  ThreadLock;				/* elan memory portion of spin lock */
    EP3_SPINLOCK_ELAN  PendingLock;				/* spin lock for pending rx list */

    E3_Addr	       PendingDescs;				/* list of pending receive descriptors */
    E3_uint32          ThreadShouldHalt;                        /* marks that the thread should halt */

    E3_uint64	       MainAddr;				/* kernel address of ep_rcvr (for StallThreadForNoDescs)*/
} EP3_RCVR_RAIL_ELAN;

typedef struct ep3_rcvr_rail_main				/* Main memory service strucure */
{
    EP3_SPINLOCK_MAIN	ThreadLock;				/* main memory portion of spin lock */
    EP3_SPINLOCK_MAIN	PendingLock;				/* spinlock for pending rx list */

    volatile unsigned   PendingDescsTailp;			/* next pointer of last receive descriptor on pending list */
} EP3_RCVR_RAIL_MAIN;

#if !defined(__ELAN3__)

typedef struct ep3_rcvr_rail_stats
{
    unsigned long some_stat;
} EP3_RCVR_RAIL_STATS;

typedef struct ep3_rcvr_rail
{
    EP_RCVR_RAIL	Generic;				/* generic portion */
    
    EP3_RCVR_RAIL_MAIN *RcvrMain;
    E3_Addr		RcvrMainAddr;
    sdramaddr_t         RcvrElan;
    E3_Addr             RcvrElanAddr;

    sdramaddr_t		InputQueueBase;				/* base of receive queue */
    E3_Addr		InputQueueAddr;				/* elan address of receive queue */

    E3_Addr	        ThreadStack;				/* Thread processor stack */
    E3_Addr	        ThreadWaiting;				/* Elan thread is waiting as no receive descriptors pending (sp stored here ) */
    E3_Addr		ThreadHalted;				/* Elan thread is waiting as it was requested to halt */

    struct list_head	FreeDescList;				/* freelist of per-rail receive descriptors */
    unsigned int	FreeDescCount;				/*   and number on free list */
    unsigned int        TotalDescCount;				/*   total number created */
    spinlock_t		FreeDescLock;				/*   and lock for free list */
    struct list_head    DescBlockList;				/* list of receive descriptor blocks */

    unsigned int        FreeDescWaiting;			/* waiting for descriptors to be freed */
    kcondvar_t		FreeDescSleep;				/*   and sleep here */

    unsigned int	CleanupWaiting;				/* waiting for cleanup */
    kcondvar_t		CleanupSleep;				/*   and sleep here */

    EP3_RCVR_RAIL_STATS stats;                                  /* elan3 specific rcvr_rail stats */
} EP3_RCVR_RAIL;

#endif /* ! __ELAN3__ */

/* per-rail portion of transmit descriptor */
typedef struct ep3_txd_rail_elan
{
    EP_ENVELOPE	       Envelope;				/* message envelope */
    EP_PAYLOAD	       Payload;					/* message payload */

    E3_BlockCopyEvent EnveEvent;				/* envelope event */
    E3_BlockCopyEvent DataEvent;				/* data transfer event */
    E3_BlockCopyEvent DoneEvent;				/* rpc done event */
} EP3_TXD_RAIL_ELAN;

#define EP3_TXD_RAIL_ELAN_SIZE	roundup (sizeof (EP3_TXD_RAIL_ELAN), E3_BLK_ALIGN)

typedef struct ep3_txd_rail_main
{
    E3_uint32         EnveEvent;				/* dest for envelope event */
    E3_uint32         DataEvent;				/* dest for data transfer event */
    E3_uint32	      DoneEvent;				/* dest for rpc done event */
} EP3_TXD_RAIL_MAIN;

#define EP3_TXD_RAIL_MAIN_SIZE	roundup (sizeof(EP3_TXD_RAIL_MAIN), E3_BLK_ALIGN)

#if !defined(__ELAN3__)

typedef struct ep3_txd_rail
{
    EP_TXD_RAIL	       Generic;					/* generic txd rail */

    EP3_COOKIE	       EnveCookie;				/* Event cookies */
    EP3_COOKIE	       DataCookie;
    EP3_COOKIE	       DoneCookie;

    sdramaddr_t	       TxdElan;					/* Elan TX descriptor */
    E3_Addr	       TxdElanAddr;				/*  and elan address */

    EP3_TXD_RAIL_MAIN *TxdMain;					/* Elan Main memory tx descriptor */
    E3_Addr	       TxdMainAddr;				/*  and elan address */

    EP_BACKOFF	       Backoff;					/* dma backoff */
} EP3_TXD_RAIL;


#define EP3_NUM_TXD_PER_BLOCK	16

typedef struct ep3_txd_rail_block
{
    struct list_head 	Link;
    
    EP3_TXD_RAIL 	Txd[EP3_NUM_TXD_PER_BLOCK];
} EP3_TXD_RAIL_BLOCK;

typedef struct ep3_xmtr_rail_stats
{
    unsigned long some_stat;
} EP3_XMTR_RAIL_STATS;

typedef struct ep3_xmtr_rail
{
    EP_XMTR_RAIL	Generic;				/* generic portion */

    struct list_head	FreeDescList;				/* freelist of per-rail receive descriptors */
    unsigned int	FreeDescCount;				/*   and number on free list */
    unsigned int        TotalDescCount;
    spinlock_t		FreeDescLock;				/*   and lock for free list */
    struct list_head    DescBlockList;				/* list of receive descriptor blocks */

    unsigned int        FreeDescWaiting;			/* waiting for descriptors to be freed */
    kcondvar_t          FreeDescSleep;				/*   and sleep here */

    EP3_XMTR_RAIL_STATS stats;                                  /* elan3 specific xmtr rail stats */
} EP3_XMTR_RAIL;

typedef struct ep3_comms_rail
{
    EP_COMMS_RAIL	Generic;				/* generic comms rail */
    sdramaddr_t		QueueDescs;				/* input queue descriptors */
} EP3_COMMS_RAIL;

/* epcommxTx_elan3.c */
extern void           ep3xmtr_flush_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail);
extern void           ep3xmtr_failover_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail);
extern void           ep3xmtr_disconnect_callback (EP_XMTR *xmtr, EP3_XMTR_RAIL *xmtrRail);

/* epcommsRx_elan3.c */
extern void	      CompleteEnvelope (EP3_RAIL *rail, E3_Addr rxdMainAddr, E3_uint32 PAckVal);
extern void           StallThreadForNoDescs (EP3_RAIL *rail, E3_Addr rcvrElanAddr, E3_Addr sp);
extern void           StallThreadForHalted  (EP3_RAIL *rail, E3_Addr rcvrElanAddr, E3_Addr sp);

extern void           ep3rcvr_flush_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail);
extern void           ep3rcvr_failover_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail);
extern void           ep3rcvr_disconnect_callback (EP_RCVR *rcvr, EP3_RCVR_RAIL *rcvrRail);

/* epcomms_elan3.c */
extern EP_COMMS_RAIL *ep3comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *r);
extern void           ep3comms_del_rail (EP_COMMS_RAIL *r);
extern void           ep3comms_display_rail (EP_COMMS_RAIL *r);

/* epcommsTx_elan3.c */
extern int            ep3xmtr_bind_txd (EP_TXD *txd, EP_XMTR_RAIL *xmtrRail, unsigned int phase);
extern void           ep3xmtr_unbind_txd (EP_TXD *txd, unsigned int phase);
extern int            ep3xmtr_poll_txd (EP_XMTR_RAIL *xmtrRail, EP_TXD_RAIL *txdRail, int how);
extern long           ep3xmtr_check (EP_XMTR_RAIL *xmtrRail, long nextRunTime);
extern void           ep3xmtr_add_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail);
extern void           ep3xmtr_del_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail);
extern int            ep3xmtr_check_txd_state(EP_TXD *txd);

extern void           ep3xmtr_display_xmtr (DisplayInfo *di, EP_XMTR_RAIL *xmtrRail);
extern void           ep3xmtr_display_txd  (DisplayInfo *di, EP_TXD_RAIL *txdRail);

extern void           ep3xmtr_fillout_rail_stats (EP_XMTR_RAIL *xmtr_rail, char *str);

/* epcommsRx_elan3.c */
extern int	      ep3rcvr_queue_rxd (EP_RXD *rxd, EP_RCVR_RAIL *rcvrRail);
extern void	      ep3rcvr_rpc_put (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
extern void	      ep3rcvr_rpc_get (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
extern void	      ep3rcvr_rpc_complete (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);

extern EP_RXD	     *ep3rcvr_steal_rxd (EP_RCVR_RAIL *rcvrRail);

extern long	      ep3rcvr_check (EP_RCVR_RAIL *rcvrRail, long nextRunTime);
extern void           ep3rcvr_add_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);
extern void           ep3rcvr_del_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);

extern void           ep3rcvr_display_rcvr (DisplayInfo *di, EP_RCVR_RAIL *rcvrRail);
extern void           ep3rcvr_display_rxd  (DisplayInfo *di, EP_RXD_RAIL *rxdRail);

extern void           ep3rcvr_fillout_rail_stats (EP_RCVR_RAIL *rcvr_rail, char *str);

#endif /* !defined(__ELAN3__) */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __EPCOMMS_ELAN3_H */
