/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcomms_elan3_thread.c,v 1.4 2004/01/20 11:03:15 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan3_thread.c,v $ */

//#include <qsnet/types.h>

typedef char               int8_t;
typedef unsigned char      uint8_t;
typedef short              int16_t;
typedef unsigned short     uint16_t;
typedef int                int32_t;
typedef unsigned int       uint32_t;
typedef long long          int64_t;
typedef unsigned long long uint64_t;

#include <elan3/e3types.h>
#include <elan3/events.h>
#include <elan3/elanregs.h>
#include <elan3/intrinsics.h>

#include <elan/nmh.h>
#include <elan/kcomm.h>
#include <elan/epcomms.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "epcomms_elan3.h"

#ifndef offsetof
#define offsetof(s, m)			(unsigned long)(&(((s *)0)->m))
#endif

EP3_RAIL_ELAN *rail;
EP3_RCVR_RAIL_ELAN *r;
EP3_RCVR_RAIL_MAIN *rm;

void
ep3comms_rcvr (EP3_RAIL_ELAN *rail, EP3_RCVR_RAIL_ELAN *rcvrElan, EP3_RCVR_RAIL_MAIN *rcvrMain, 
	      EP3_InputQueue *q, unsigned int *cookies)
{
    int           count = 1;
    E3_Addr       nfptr = q->q_fptr + q->q_size;
    E3_uint32     tmp;
    int           i;
    E3_Addr       buffer;
    int	          len;
    E3_DMA       *dma;
    E3_Event     *event;

    /* clear the queue state to allow envelopes to arrive */
    q->q_state = 0;

    for (;;)
    {
	if (! rcvrElan->ThreadShouldHalt)
	    c_waitevent ((E3_Event *) &q->q_event, count); 						/* HALT POINT */

	if (rcvrElan->ThreadShouldHalt && nfptr == q->q_bptr)
	{
	    asm volatile ("mov %0, %%g1" : /* no outputs */ : "r" (rcvrElan));
	    asm volatile ("ta %0"        : /* no outputs */ : "i" (EP3_UNIMP_THREAD_HALTED));		/* HALT POINT */
	    continue;
	}

	count = 0;
	do {
	    /* Process the message at nfptr */
	    EP_ENVELOPE      *env = (EP_ENVELOPE *) nfptr;
	    EP3_RXD_RAIL_ELAN *rxd;
	    int ack;
	    
	    EP3_SPINENTER(&rcvrElan->ThreadLock, &rcvrMain->ThreadLock);					/* HALT POINT */
	    
	    while ((rxd = (EP3_RXD_RAIL_ELAN *)rcvrElan->PendingDescs) == 0)
	    {
		/* no receive descriptors, so trap to the kernel to wait
		 * for receive descriptor to be queued, we pass the rcvr
		 * in %g1, so that the trap handler can restart us. */
		EP3_SPINEXIT(&rcvrElan->ThreadLock, &rcvrMain->ThreadLock);
		asm volatile ("mov %0, %%g1" : /* no outputs */ : "r" (rcvrElan));
		asm volatile ("ta %0"        : /* no outputs */ : "i" (EP3_UNIMP_TRAP_NO_DESCS));	/* HALT POINT */
		EP3_SPINENTER(&rcvrElan->ThreadLock, &rcvrMain->ThreadLock);				/* HALT POINT */
	    }

	    if (env->Version != EP_ENVELOPE_VERSION)
	    {
		/* This envelope has been cancelled - so just consume it */
		EP3_SPINEXIT(&rcvrElan->ThreadLock, &rcvrMain->ThreadLock);
		goto consume_envelope;
	    }

	    dma   = rxd->Dmas;
	    event = rxd->ChainEvent;

	    if (EP_IS_MULTICAST(env->Attr))
	    {
		dma->dma_type            = E3_DMA_TYPE (DMA_BYTE, DMA_READ, DMA_NORMAL, EP3_DMAFAILCOUNT);
		dma->dma_size            = BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t);
		dma->dma_source          = env->TxdMain.nmd_addr + offsetof (EP_TXD_MAIN, Bitmap);
		dma->dma_dest            = (E3_Addr) &((EP_RXD_MAIN *) rxd->RxdMain)->Bitmap;
		dma->dma_destEvent       = (E3_Addr) event;
		dma->dma_destCookieVProc = DMA_COOKIE_THREAD | DMA_COOKIE (cookies[env->NodeId], EP_VP_DATA (rail->NodeId));
		dma->dma_srcEvent        = env->TxdRail + offsetof (EP3_TXD_RAIL_ELAN, DataEvent);
		dma->dma_srcCookieVProc  = DMA_COOKIE_THREAD | DMA_REMOTE_COOKIE (cookies[env->NodeId], EP_VP_DATA (env->NodeId));
		
		event->ev_Count = 1;

		dma++; event++;
	    }

	    if (env->nFrags == 0)
	    {
		/* Generate a "get" DMA to accept the envelope and fire the rx handler */
		dma->dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_READ, DMA_NORMAL, EP3_DMAFAILCOUNT);
		dma->dma_size            = 0;
		dma->dma_destEvent       = (E3_Addr) &rxd->DataEvent;
		dma->dma_destCookieVProc = DMA_COOKIE_THREAD | DMA_COOKIE (cookies[env->NodeId], EP_VP_DATA (rail->NodeId));
		dma->dma_srcEvent        = env->TxdRail + offsetof (EP3_TXD_RAIL_ELAN, DataEvent);
		dma->dma_srcCookieVProc  = DMA_COOKIE_THREAD | DMA_REMOTE_COOKIE (cookies[env->NodeId], EP_VP_DATA (env->NodeId));
		len = 0;
	    }
	    else
	    {
		/* Generate the DMA chain to fetch the data */
		for (i = 0, buffer = rxd->Data.nmd_addr, len = 0; i < env->nFrags; i++, dma++, event++)
		{
		    dma->dma_type            = E3_DMA_TYPE(DMA_BYTE, DMA_READ, DMA_NORMAL, EP3_DMAFAILCOUNT);
		    dma->dma_size            = env->Frags[i].nmd_len;
		    dma->dma_source          = env->Frags[i].nmd_addr;
		    dma->dma_dest            = buffer;
		    dma->dma_destEvent       = (E3_Addr) event;
		    dma->dma_destCookieVProc = DMA_COOKIE_THREAD | DMA_COOKIE (cookies[env->NodeId], EP_VP_DATA (rail->NodeId));
		    dma->dma_srcEvent        = env->TxdRail + offsetof (EP3_TXD_RAIL_ELAN, DataEvent);
		    dma->dma_srcCookieVProc  = DMA_COOKIE_THREAD | DMA_REMOTE_COOKIE (cookies[env->NodeId], EP_VP_DATA (env->NodeId));
		    
		    event->ev_Count = 1;
		    
		    buffer += dma->dma_size;
		    len    += dma->dma_size;
		}
		
		/* Point the last dma at the done event */
		(--dma)->dma_destEvent = (E3_Addr) &rxd->DataEvent;
		
		if (rxd->Data.nmd_len < len)
		{
		    /* The receive descriptor was too small for the message */
		    /* complete the message anyway,  but don't transfer any */
		    /* data,  we set the length to EP_MSG_TOO_BIG */
		    for (i = 0, dma = rxd->Dmas; i < env->nFrags; i++, dma++)
			dma->dma_size = 0;
		    
		    len = EP_MSG_TOO_BIG;
		}
	    }
	    
	    /* Store the received message length in the rxdElan for CompleteEnvelope */
	    rxd->Data.nmd_len = len;

	    /* Initialise %g1 with the  "rxd" so the trap handler can
	     * complete the envelope processing if we trap while sending the
	     * packet */
	    asm volatile ("mov %0, %%g1" : /* no outputs */ : "r" (rxd));

	    /* Generate a packet to start the data transfer */
	    c_open (EP_VP_DATA (env->NodeId));
	    c_sendtrans2 (TR_THREADIDENTIFY, rxd->Dmas->dma_destCookieVProc, 0, 0);
	    c_sendmem (TR_SENDACK | TR_REMOTEDMA, 0, rxd->Dmas); 
	    ack = c_close();
	    
	    /*
	     * If we trapped for an output timeout, then the trap handler will have
	     * completed processing this envelope and cleared the spinlock, so we just
	     * need to update the queue descriptor.
	     */
	    if (ack == EP3_PAckStolen)
		goto consume_envelope;
	    
	    if (ack != E3_PAckOk)
	    {
		/* our packet got nacked, so trap into the kernel so that
		 * it can complete processing of this envelope.
		 */
		asm volatile ("ta %0" : /* no outputs */ : "i" (EP3_UNIMP_TRAP_PACKET_NACKED));		/* HALT POINT */
		goto consume_envelope;
	    }

	    /* remove the RXD from the pending list */
	    EP3_SPINENTER (&rcvrElan->PendingLock, &rcvrMain->PendingLock);
	    if ((rcvrElan->PendingDescs = rxd->Next) == 0)
		rcvrMain->PendingDescsTailp = 0;
	    EP3_SPINEXIT (&rcvrElan->PendingLock, &rcvrMain->PendingLock);

	    /* Copy the envelope information - as 5 64 byte chunks.
	     * We force the parameters in g5, g6 so that they aren't
	     * trashed by the loadblk32 into the locals/ins
	     */
	    if (EP_HAS_PAYLOAD(env->Attr))
	    { 
		register void *src asm ("g5") = (void *) env;
		register void *dst asm ("g6") = (void *)  &((EP_RXD_MAIN *) rxd->RxdMain)->Envelope;

		asm volatile (
		    "and     %%sp,63,%%g7		! Calculate stack alignment\n"
		    "add     %%g7,64,%%g7		! Space to save the registers\n"
		    "sub     %%sp,%%g7,%%sp		! align stack\n" 
		    "stblock64 %%l0,[%%sp]		! save the locals and ins\n"

		    "ldblock64 [%0 + 0],%%l0		! load 64-byte block into locals/ins\n"		/* copy envelope */
		    "stblock64 %%l0,[%1 + 0]		! store 64-byte block from local/ins\n"
		    "ldblock64 [%0 + 64],%%l0		! load 64-byte block into locals/ins\n"
		    "stblock64 %%l0,[%1 + 64]		! store 64-byte block from local/ins\n"

		    "ldblock64 [%0 + 128],%%l0		! load 64-byte block into locals/ins\n"		/* copy payload */
		    "stblock64 %%l0,[%1 + 128]		! store 64-byte block from local/ins\n"
		    "ldblock64 [%0 + 192],%%l0		! load 64-byte block into locals/ins\n"
		    "stblock64 %%l0,[%1 + 192]		! store 64-byte block from local/ins\n"

		    "ldblock64 [%%sp],%%l0		! restore locals and ins\n"
		    "add     %%sp,%%g7,%%sp		! restore stack pointer\n"
		    : /* outputs */
		    : /* inputs */ "r" (src), "r" (dst)
		    : /* clobbered */ "g5", "g6", "g7" );
	    }
	    else
	    { 
		register void *src asm ("g5") = (void *) env;
		register void *dst asm ("g6") = (void *)  &((EP_RXD_MAIN *) rxd->RxdMain)->Envelope;

		asm volatile (
		    "and     %%sp,63,%%g7		! Calculate stack alignment\n"
		    "add     %%g7,64,%%g7		! Space to save the registers\n"
		    "sub     %%sp,%%g7,%%sp		! align stack\n" 
		    "stblock64 %%l0,[%%sp]		! save the locals and ins\n"

		    "ldblock64 [%0 + 0],%%l0		! load 64-byte block into locals/ins\n"
		    "stblock64 %%l0,[%1 + 0]		! store 64-byte block from local/ins\n"
		    "ldblock64 [%0 + 64],%%l0		! load 64-byte block into locals/ins\n"
		    "stblock64 %%l0,[%1 + 64]		! store 64-byte block from local/ins\n"

		    "ldblock64 [%%sp],%%l0		! restore locals and ins\n"
		    "add     %%sp,%%g7,%%sp		! restore stack pointer\n"
		    : /* outputs */
		    : /* inputs */ "r" (src), "r" (dst)
		    : /* clobbered */ "g5", "g6", "g7" );
	    }

	    /* Store the message length to indicate that I've finished */
	    ((EP_RXD_MAIN *) rxd->RxdMain)->Len = rxd->Data.nmd_len;					/* PCI write  */
	    
	    EP3_SPINEXIT(&rcvrElan->ThreadLock, &rcvrMain->ThreadLock);

	consume_envelope:
	    /* Sample the queue full bit *BEFORE* moving the fptr.
	     * Then only clear it if it was full before, otherwise,
	     * as soon as the fptr is moved on the queue could fill 
	     * up, and so clearing it could mark a full queue as 
	     * empty.
	     *
	     * While the full bit is set, the queue is in a 'steady
	     * state', so it is safe to set the q_state
	     * 
	     */
	    if (((tmp = q->q_state) & E3_QUEUE_FULL) == 0)
		q->q_fptr = nfptr;				/* update queue */
	    else
	    {
		q->q_fptr = nfptr;				/* update queue */
		q->q_state = tmp &~E3_QUEUE_FULL;		/* and clear full flag */
	    }

	    count++;						/* bump message count */
	    if (nfptr == q->q_top)				/* queue wrap */
		nfptr = q->q_base;
	    else
		nfptr += q->q_size;

	    c_break_busywait();					/* be nice 		HALT POINT */

	} while (nfptr != q->q_bptr);				/* loop until Fptr == Bptr */
    }
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
