/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcomms_elan4.c,v 1.12.2.1 2005/08/09 09:57:14 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"
#include "kcomm_elan4.h"
#include "epcomms_elan4.h"

static void
ep4comms_flush_interrupt (EP4_RAIL *rail, void *arg)
{
    EP4_COMMS_RAIL *commsRail = (EP4_COMMS_RAIL *) arg;
    unsigned long  flags;

    spin_lock_irqsave (&commsRail->r_flush_lock, flags);
    commsRail->r_flush_count = 0;
    kcondvar_wakeupall (&commsRail->r_flush_sleep, &commsRail->r_flush_lock);
    spin_unlock_irqrestore  (&commsRail->r_flush_lock, flags);
}

void
ep4comms_flush_start (EP4_COMMS_RAIL *commsRail)
{
    kmutex_lock (&commsRail->r_flush_mutex);
}

void
ep4comms_flush_wait (EP4_COMMS_RAIL *commsRail)
{
    unsigned long flags;

    ep4_wait_event_cmd (commsRail->r_flush_mcq, 
			commsRail->r_elan_addr + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event),
			E4_EVENT_INIT_VALUE (-32 * commsRail->r_flush_count, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0),
			commsRail->r_flush_ecq->ecq_addr, 
			INTERRUPT_CMD | (commsRail->r_flush_intcookie.int_val << E4_MAIN_INT_SHIFT));

    spin_lock_irqsave (&commsRail->r_flush_lock, flags); 
    while (commsRail->r_flush_count != 0) 
	if (kcondvar_timedwait (&commsRail->r_flush_sleep, &commsRail->r_flush_lock, &flags, (lbolt + (HZ*10))) == -1) 
	    elan4_hardware_lock_check(((EP4_RAIL *)(commsRail->r_generic.Rail))->r_ctxt.ctxt_dev, "flush_wait");
    spin_unlock_irqrestore (&commsRail->r_flush_lock, flags);
    
    kmutex_unlock (&commsRail->r_flush_mutex);
}

void
ep4comms_flush_setevent (EP4_COMMS_RAIL *commsRail, ELAN4_CQ *cq)
{
    unsigned long flags;

    spin_lock_irqsave (&commsRail->r_flush_lock, flags);

    elan4_set_event_cmd (cq, commsRail->r_elan_addr + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event));
    
    commsRail->r_flush_count++;
    
    spin_unlock_irqrestore (&commsRail->r_flush_lock, flags);
}

void
ep4comms_flush_callback (void *arg, statemap_t *map)
{
    EP4_COMMS_RAIL   *commsRail = (EP4_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->r_generic.Subsys;
    EP4_RAIL	     *rail      = (EP4_RAIL *) commsRail->r_generic.Rail;
    unsigned int     rnum       = rail->r_generic.Number;
    struct list_head *el;

    /*
     * We stall the retry thread from CB_FLUSH_FILTERING until
     * we've finished CB_FLUSH_FLUSHING to ensure that sten 
     * packets can not be being retried while we flush them
     * through.
     */
    switch (rail->r_generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	ep_kthread_stall (&rail->r_retry_thread);

	ep4comms_flush_start (commsRail);
	break;

    case EP_CB_FLUSH_FLUSHING:
	break;
    }

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[rnum])
	    ep4xmtr_flush_callback (xmtr, (EP4_XMTR_RAIL *) xmtr->Rails[rnum]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[rnum])
	    ep4rcvr_flush_callback (rcvr, (EP4_RCVR_RAIL *) rcvr->Rails[rnum]);
    }
    kmutex_unlock (&subsys->Lock);

    switch (rail->r_generic.CallbackStep)
    {
    case EP_CB_FLUSH_FILTERING:
	ep4comms_flush_wait (commsRail);
	break;

    case EP_CB_FLUSH_FLUSHING:
	ep_kthread_resume (&rail->r_retry_thread);
	break;
    }
}

void
ep4comms_failover_callback (void *arg, statemap_t *map)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    unsigned int     rnum       = commsRail->Rail->Number;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[rnum])
	    ep4xmtr_failover_callback (xmtr, (EP4_XMTR_RAIL *) xmtr->Rails[rnum]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[rnum])
	    ep4rcvr_failover_callback (rcvr, (EP4_RCVR_RAIL *) rcvr->Rails[rnum]);
    }
    kmutex_unlock (&subsys->Lock);
}

void
ep4comms_disconnect_callback (void *arg, statemap_t *map)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    unsigned int     rnum       = commsRail->Rail->Number;
    struct list_head *el;

    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[rnum])
	    ep4xmtr_disconnect_callback (xmtr, (EP4_XMTR_RAIL *) xmtr->Rails[rnum]);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[rnum])
	    ep4rcvr_disconnect_callback (rcvr, (EP4_RCVR_RAIL *) rcvr->Rails[rnum]);
    }
    kmutex_unlock (&subsys->Lock);
}

void
ep4comms_neterr_callback (EP4_RAIL *rail, void *arg, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP_COMMS_RAIL    *commsRail = (EP_COMMS_RAIL *) arg;
    EP_COMMS_SUBSYS  *subsys    = commsRail->Subsys;
    unsigned int     rnum       = commsRail->Rail->Number;
    struct list_head *el;
    
    /* First - stall the retry thread, so that it will no longer restart 
     *         any sten packets from the retry lists */
    ep_kthread_stall (&rail->r_retry_thread);

    ep4comms_flush_start ((EP4_COMMS_RAIL *) commsRail);

    /* Second - flush through all command queues for xmtrs and rcvrs */
    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[rnum])
	    ep4xmtr_neterr_flush (xmtr, (EP4_XMTR_RAIL *) xmtr->Rails[rnum], nodeId, cookies);
    }
    
    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[rnum])
	    ep4rcvr_neterr_flush (rcvr, (EP4_RCVR_RAIL *) rcvr->Rails[rnum], nodeId, cookies);
    }
    kmutex_unlock (&subsys->Lock);

    /* Third - wait for flush to complete */
    ep4comms_flush_wait ((EP4_COMMS_RAIL *) commsRail);
    
    /* Fourth - flush through all command queues */
    ep4_flush_ecqs (rail);
    
    /* Fifth - search all the retry lists for the network error cookies */
    kmutex_lock (&subsys->Lock);
    list_for_each (el, &subsys->Transmitters) {
	EP_XMTR *xmtr = list_entry (el, EP_XMTR, Link);
	
	if (xmtr->Rails[rnum])
	    ep4xmtr_neterr_check (xmtr, (EP4_XMTR_RAIL *) xmtr->Rails[rnum], nodeId, cookies);
    }

    list_for_each (el, &subsys->Receivers) {
	EP_RCVR *rcvr = list_entry (el, EP_RCVR, Link);
	
	if (rcvr->Rails[rnum])
	    ep4rcvr_neterr_check (rcvr, (EP4_RCVR_RAIL *) rcvr->Rails[rnum], nodeId, cookies);
    }
    kmutex_unlock (&subsys->Lock);

    ep_kthread_resume (&rail->r_retry_thread);
}


EP_COMMS_RAIL *
ep4comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *r)
{
    EP4_RAIL       *rail = (EP4_RAIL *)r;
    ELAN4_DEV      *dev  = rail->r_ctxt.ctxt_dev;
    EP4_COMMS_RAIL *commsRail;
    E4_InputQueue   qdesc;
    int i;

    KMEM_ZALLOC (commsRail, EP4_COMMS_RAIL *,sizeof (EP4_COMMS_RAIL), 1);

    if (commsRail == NULL)
	return NULL;
    
    commsRail->r_generic.Ops.DelRail          = ep4comms_del_rail;
    commsRail->r_generic.Ops.DisplayRail      = ep4comms_display_rail;
    commsRail->r_generic.Ops.Rcvr.AddRail     = ep4rcvr_add_rail;
    commsRail->r_generic.Ops.Rcvr.DelRail     = ep4rcvr_del_rail;
    commsRail->r_generic.Ops.Rcvr.Check       = ep4rcvr_check;
    commsRail->r_generic.Ops.Rcvr.QueueRxd    = ep4rcvr_queue_rxd;
    commsRail->r_generic.Ops.Rcvr.RpcPut      = ep4rcvr_rpc_put;
    commsRail->r_generic.Ops.Rcvr.RpcGet      = ep4rcvr_rpc_get;
    commsRail->r_generic.Ops.Rcvr.RpcComplete = ep4rcvr_rpc_complete;

    commsRail->r_generic.Ops.Rcvr.StealRxd    = ep4rcvr_steal_rxd;

    commsRail->r_generic.Ops.Rcvr.DisplayRcvr = ep4rcvr_display_rcvr;
    commsRail->r_generic.Ops.Rcvr.DisplayRxd  = ep4rcvr_display_rxd;

    commsRail->r_generic.Ops.Rcvr.FillOutRailStats = ep4rcvr_fillout_rail_stats;

    commsRail->r_generic.Ops.Xmtr.AddRail     = ep4xmtr_add_rail;
    commsRail->r_generic.Ops.Xmtr.DelRail     = ep4xmtr_del_rail;
    commsRail->r_generic.Ops.Xmtr.Check       = ep4xmtr_check;
    commsRail->r_generic.Ops.Xmtr.BindTxd     = ep4xmtr_bind_txd;
    commsRail->r_generic.Ops.Xmtr.UnbindTxd   = ep4xmtr_unbind_txd;
    commsRail->r_generic.Ops.Xmtr.PollTxd     = ep4xmtr_poll_txd;
    commsRail->r_generic.Ops.Xmtr.CheckTxdState = ep4xmtr_check_txd_state;

    commsRail->r_generic.Ops.Xmtr.DisplayXmtr = ep4xmtr_display_xmtr;
    commsRail->r_generic.Ops.Xmtr.DisplayTxd  = ep4xmtr_display_txd;

    commsRail->r_generic.Ops.Xmtr.FillOutRailStats = ep4xmtr_fillout_rail_stats;

    /* Allocate command queue space for flushing (1 dword for interrupt + 4 dwords for waitevent) */
    if ((commsRail->r_flush_ecq = ep4_get_ecq (rail, EP4_ECQ_EVENT, 1)) == NULL)
    {
	KMEM_FREE (commsRail, sizeof (EP4_COMMS_RAIL));
	return NULL;
    }

    if ((commsRail->r_flush_mcq = ep4_get_ecq (rail, EP4_ECQ_MAIN, 4)) == NULL)
    {
	ep4_put_ecq (rail, commsRail->r_flush_ecq, 1);
	KMEM_FREE (commsRail, sizeof (EP4_COMMS_RAIL));
	return NULL;
    }

    /* Allocate and initialise the elan memory part */
    if ((commsRail->r_elan = ep_alloc_elan (r, EP4_COMMS_RAIL_ELAN_SIZE, 0, &commsRail->r_elan_addr)) == (sdramaddr_t) 0)
    {
	ep4_put_ecq (rail, commsRail->r_flush_mcq, 4);
	ep4_put_ecq (rail, commsRail->r_flush_ecq, 1);
	KMEM_FREE (commsRail, sizeof (EP4_COMMS_RAIL));
	return NULL;
    }

    ep4_register_intcookie (rail, &commsRail->r_flush_intcookie, commsRail->r_elan_addr + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event),
			    ep4comms_flush_interrupt, commsRail);

    elan4_sdram_writeq (dev, commsRail->r_elan + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event.ev_CountAndType),
			E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));


    /* Allocate and initialise all the queue desriptors as "full" with no event */
    if ((commsRail->r_descs = ep_alloc_memory_elan (r, EP_EPCOMMS_QUEUE_BASE, roundup (EP_MSG_NSVC * EP_QUEUE_DESC_SIZE, SDRAM_PAGE_SIZE), EP_PERM_ALL, 0)) == (sdramaddr_t) 0)
    {
	ep_free_elan (r, commsRail->r_elan_addr, EP4_COMMS_RAIL_ELAN_SIZE);
	ep4_put_ecq (rail, commsRail->r_flush_mcq, 4);
	ep4_put_ecq (rail, commsRail->r_flush_ecq, 1);
	KMEM_FREE (commsRail, sizeof (EP4_COMMS_RAIL));
	return NULL;
    }

    qdesc.q_bptr    = 0;
    qdesc.q_fptr    = 8;
    qdesc.q_control = E4_InputQueueControl (qdesc.q_bptr,qdesc.q_fptr, 8);
    qdesc.q_event   = 0;

    for (i = 0; i < EP_MSG_NSVC; i++)
	elan4_sdram_copyq_to_sdram (rail->r_ctxt.ctxt_dev, &qdesc, commsRail->r_descs + (i * EP_QUEUE_DESC_SIZE),
				    sizeof (E4_InputQueue));

    kmutex_init (&commsRail->r_flush_mutex);
    spin_lock_init (&commsRail->r_flush_lock);
    kcondvar_init (&commsRail->r_flush_sleep);

    ep_register_callback (r, EP_CB_FLUSH_FILTERING, ep4comms_flush_callback,      commsRail);
    ep_register_callback (r, EP_CB_FLUSH_FLUSHING,  ep4comms_flush_callback,      commsRail);
    ep_register_callback (r, EP_CB_FAILOVER,        ep4comms_failover_callback,   commsRail);
    ep_register_callback (r, EP_CB_DISCONNECTING,   ep4comms_disconnect_callback, commsRail);

    commsRail->r_neterr_ops.op_func = ep4comms_neterr_callback;
    commsRail->r_neterr_ops.op_arg  = commsRail;
    
    ep4_add_neterr_ops (rail, &commsRail->r_neterr_ops);

    return (EP_COMMS_RAIL *) commsRail;
}

void
ep4comms_del_rail (EP_COMMS_RAIL *r)
{
    EP4_COMMS_RAIL *commsRail = (EP4_COMMS_RAIL *) r;
    EP4_RAIL       *rail      = (EP4_RAIL *) commsRail->r_generic.Rail;

    ep_remove_callback (&rail->r_generic, EP_CB_FLUSH_FILTERING, ep4comms_flush_callback,      commsRail);
    ep_remove_callback (&rail->r_generic, EP_CB_FLUSH_FLUSHING,  ep4comms_flush_callback,      commsRail);
    ep_remove_callback (&rail->r_generic, EP_CB_FAILOVER,        ep4comms_failover_callback,   commsRail);
    ep_remove_callback (&rail->r_generic, EP_CB_DISCONNECTING,   ep4comms_disconnect_callback, commsRail);

    kcondvar_destroy (&commsRail->r_flush_sleep);
    spin_lock_destroy (&commsRail->r_flush_lock);
    kmutex_destroy (&commsRail->r_flush_mutex);

    ep_free_memory_elan (&rail->r_generic, EP_EPCOMMS_QUEUE_BASE);
    ep_free_elan (&rail->r_generic, commsRail->r_elan_addr, EP4_COMMS_RAIL_ELAN_SIZE);

    ep4_deregister_intcookie (rail, &commsRail->r_flush_intcookie);

    ep4_put_ecq (rail, commsRail->r_flush_mcq, 4);
    ep4_put_ecq (rail, commsRail->r_flush_ecq, 1);

    KMEM_FREE (commsRail, sizeof (EP4_COMMS_RAIL));
}

void
ep4comms_display_rail (EP_COMMS_RAIL *r)
{
    EP4_COMMS_RAIL *commsRail = (EP4_COMMS_RAIL *) r;
    EP4_RAIL       *rail      = (EP4_RAIL *) commsRail->r_generic.Rail;
    ELAN4_DEV      *dev       = rail->r_ctxt.ctxt_dev;
    
    ep4_display_rail (rail);

    ep_debugf (DBG_DEBUG, "   flush count=%d mcq=%p ecq=%p event %llx.%llx.%llx\n", 
	       commsRail->r_flush_count, commsRail->r_flush_mcq, commsRail->r_flush_ecq,
	       elan4_sdram_readq (dev, commsRail->r_elan + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event.ev_CountAndType)),
	       elan4_sdram_readq (dev, commsRail->r_elan + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event.ev_WritePtr)),
	       elan4_sdram_readq (dev, commsRail->r_elan + offsetof (EP4_COMMS_RAIL_ELAN, r_flush_event.ev_WriteValue)));
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
