/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __EPCOMMS_ELAN4_H
#define __EPCOMMS_ELAN4_H

#ident "@(#)$Id: epcomms_elan4.h,v 1.15 2005/03/22 16:47:36 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan4.h,v $ */


#include <elan4/types.h>

/*
 * Elan4 spinlocks are a pair of 64 bit words, one in elan sdram and one in main memory
 * the sdram word holds the thread sequence number in the bottom 32 bits and the main
 * lock in the top 32 bits.  The main memory word holds the sequence number only in
 * it's bottom 32 bits */

typedef volatile E4_uint64 EP4_SPINLOCK_MAIN;
typedef volatile E4_uint64 EP4_SPINLOCK_ELAN;

#define EP4_SPINLOCK_SEQ	0
#define EP4_SPINLOCK_MLOCK	4

#if defined(__elan4__)

#define EP4_SPINENTER(CPORT,SLE,SLM) \
do { \
    register long tmp; \
\
    asm volatile ("ld4		[%1], %0\n" \
		  "inc		%0\n" \
		  "st4		%0, [%1]\n" \
		  "ld4		[%1 + 4], %0\n" \
		  "srl8,byte	%0, 4, %0\n" \
		  : /* outputs */ "=r" (tmp)  \
		  : /* inputs */ "r" (SLE), "r" (SLM)); \
\
    if (tmp) \
	ep4_spinblock (CPORT,SLE, SLM); \
} while (0)

extern void ep4_spinblock(E4_uint64 *cport, EP4_SPINLOCK_ELAN *sle, EP4_SPINLOCK_MAIN *slm);

#define EP4_SPINEXIT(CPORT,SLE,SLM) \
do { \
    register long tmp; \
\
    asm volatile ("ld4		[%1], %0\n" \
		  "st4		%0, [%2]\n" \
		  : /* outputs */ "=r" (tmp) \
		  : /* inputs */ "r" (SLE), "r" (SLM)); \
} while (0)

#else

#define EP4_SPINENTER(DEV,SLE,SLM) \
do { \
    uint32_t seq; \
\
    mb(); \
    elan4_sdram_writel (DEV, (SLE) + EP4_SPINLOCK_MLOCK, 1); \
    mb(); \
    while ((seq = elan4_sdram_readl (DEV, (SLE) + EP4_SPINLOCK_SEQ)) != *((uint32_t *) (SLM))) \
    { \
	while (*((uint32_t *) (SLM)) == (seq - 1)) \
	{ \
	    mb(); \
	    DELAY(1); \
	} \
    } \
} while (0)

#define EP4_SPINEXIT(DEV,SLE,SLM) \
do { \
    wmb(); \
    elan4_sdram_writel (DEV, (SLE) + EP4_SPINLOCK_MLOCK, 0); \
} while (0)

#endif /* !defined(__elan4__) */

#define EP4_TXD_STEN_RETRYCOUNT	16
#define EP4_RXD_STEN_RETRYCOUNT	1
#define EP4_DMA_RETRYCOUNT	16

typedef struct ep4_intr_cmd
{
    E4_uint64		c_write_cmd;
    E4_uint64		c_write_value;
    E4_uint64		c_intr_cmd;
} EP4_INTR_CMD;

#define	EP4_INTR_CMD_NDWORDS	(sizeof (EP4_INTR_CMD) / 8)

typedef struct ep4_rxd_sten_cmd
{
    E4_uint64		c_open;

    E4_uint64		c_trans;
    E4_uint64		c_cookie;
    E4_uint64		c_dma_typeSize;
    E4_uint64		c_dma_cookie;
    E4_uint64		c_dma_vproc;
    E4_uint64		c_dma_srcAddr;
    E4_uint64		c_dma_dstAddr;
    E4_uint64		c_dma_srcEvent;
    E4_uint64		c_dma_dstEvent;

    E4_uint64		c_ok_guard;
    E4_uint64		c_ok_write_cmd;
    E4_uint64		c_ok_write_value;
    
    E4_uint64		c_fail_guard;
    E4_uint64		c_fail_setevent;

    E4_uint64		c_nop_cmd;
} EP4_RXD_STEN_CMD;

#define EP4_RXD_STEN_CMD_NDWORDS	(sizeof (EP4_RXD_STEN_CMD) / 8)

typedef struct ep4_rxd_dma_cmd
{
    E4_uint64		c_dma_typeSize;
    E4_uint64		c_dma_cookie;
    E4_uint64		c_dma_vproc;
    E4_uint64		c_dma_srcAddr;
    E4_uint64		c_dma_dstAddr;
    E4_uint64		c_dma_srcEvent;
    E4_uint64		c_dma_dstEvent;
    E4_uint64		c_nop_cmd;
} EP4_RXD_DMA_CMD;

#define EP4_RXD_DMA_CMD_NDWORDS		(sizeof (EP4_RXD_DMA_CMD) / 8)
#define EP4_RXD_START_CMD_NDWORDS	(sizeof (E4_ThreadRegs) / 8)

typedef struct ep4_rxd_rail_elan
{
    EP4_RXD_STEN_CMD    rxd_sten[EP_MAXFRAG+1];

    EP4_INTR_CMD	rxd_done_cmd;				/* command stream issued by done event (aligned to 64 bytes) */
    E4_Addr		rxd_next;				/* linked list when on pending list (pad to 32 bytes)*/
    E4_Event32		rxd_failed;				/* event set when sten packet fails */

    EP4_INTR_CMD        rxd_failed_cmd;				/* command stream issued by fail event (aligned to 64 bytes) */
    E4_uint64		rxd_queued;				/* rxd queuing thread has executed (pad to 32 bytes)*/

    E4_Event32		rxd_start;				/* event to set to fire off and event chain (used as chain[0]) */
    E4_Event32		rxd_chain[EP_MAXFRAG];			/* chained events (aligned to 32 bytes) */
    E4_Event32		rxd_done;				/* event to fire done command stream causing interrupt (used as chain[EP_MAXFRAG]) */

    E4_Addr		rxd_rxd;				/* elan address of EP4_RXD_MAIN */
    E4_Addr		rxd_main;				/* elan address of EP4_RXD_RAIL_MAIN */
    E4_uint64		rxd_debug;				/* thread debug value */

    EP_NMD		rxd_buffer;				/* Network mapping descriptor for receive data */
} EP4_RXD_RAIL_ELAN;

#define EP4_RXD_RAIL_ELAN_SIZE	roundup(sizeof (EP4_RXD_RAIL_ELAN), 64)

typedef struct ep4_rxd_rail_main
{
    E4_uint64		rxd_sent[EP_MAXFRAG+1];			/* sten packet sent */
    E4_uint64		rxd_failed;				/* sten packet failed */
    E4_uint64		rxd_done;				/* operation complete */

    E4_Addr		rxd_scq;				/* command port for scq */
} EP4_RXD_RAIL_MAIN;

#define EP4_RXD_RAIL_MAIN_SIZE	roundup(sizeof (EP4_RXD_RAIL_MAIN), 8)

#if !defined(__elan4__)
typedef struct ep4_rxd_rail
{
    EP_RXD_RAIL		rxd_generic;

    struct list_head    rxd_retry_link;
    unsigned long       rxd_retry_time;

    EP4_INTCOOKIE	rxd_intcookie;

    sdramaddr_t		rxd_elan;
    EP_ADDR		rxd_elan_addr;
    
    EP4_RXD_RAIL_MAIN  *rxd_main;
    EP_ADDR		rxd_main_addr;

    EP4_ECQ	       *rxd_ecq;				/* cq with 128 bytes targetted by event */
    EP4_ECQ	       *rxd_scq;				/* cq with 8 bytes targetted by main/thread store */
} EP4_RXD_RAIL;

#define EP4_NUM_RXD_PER_BLOCK	16

typedef struct ep4_rxd_rail_block
{
    struct list_head	blk_link;
    EP4_RXD_RAIL	blk_rxds[EP4_NUM_RXD_PER_BLOCK];
} EP4_RXD_RAIL_BLOCK;

#endif /* !defined(__elan4__) */

typedef struct ep4_rcvr_rail_elan
{
    E4_uint64		rcvr_thread_stall[8];			/* place for thread to stall */
    E4_Event32		rcvr_qevent;				/* Input queue event */
    E4_Event32		rcvr_thread_halt;			/* place for thread to halt */

    volatile E4_Addr    rcvr_pending_tailp;			/* list of pending rxd's (elan addr) */
    volatile E4_Addr	rcvr_pending_head;			/*   -- this pair aligned to 16 bytes */

    EP4_SPINLOCK_ELAN	rcvr_thread_lock;			/* spinlock for thread processing loop */

    E4_uint64		rcvr_stall_intcookie;			/* interrupt cookie to use when requseted to halt */

    E4_uint64		rcvr_qbase;				/* base of input queue */
    E4_uint64		rcvr_qlast;				/* last item in input queue */

    E4_uint64		rcvr_debug;				/* thread debug value */
} EP4_RCVR_RAIL_ELAN;

typedef struct ep4_rcvr_rail_main
{
    EP4_SPINLOCK_MAIN   rcvr_thread_lock;			/* spinlock for thread processing loop */
} EP4_RCVR_RAIL_MAIN;

#if !defined(__elan4__)

typedef struct ep4_rcvr_rail_stats
{
    unsigned long some_stat;
} EP4_RCVR_RAIL_STATS;

typedef struct ep4_rcvr_rail
{
    EP_RCVR_RAIL	rcvr_generic;				/* generic portion */
    
    sdramaddr_t		rcvr_elan;
    EP_ADDR		rcvr_elan_addr;

    EP4_RCVR_RAIL_MAIN *rcvr_main;
    EP_ADDR		rcvr_main_addr;

    sdramaddr_t		rcvr_slots;				/* input queue slots */
    EP_ADDR		rcvr_slots_addr;			/*   and elan address */

    EP_ADDR		rcvr_stack;				/* stack for thread */

    EP4_ECQ	       *rcvr_ecq;				/* command queue space for thread STEN packets */
    EP4_ECQ	       *rcvr_resched;				/* command queue space to reschedule the thread */

    struct list_head    rcvr_freelist;				/* freelist of per-rail receive descriptors */
    unsigned int        rcvr_freecount;				/*   and number on free list */
    unsigned int        rcvr_totalcount;				/*   total number created */
    spinlock_t          rcvr_freelock;				/*   and lock for free list */
    struct list_head    rcvr_blocklist;				/* list of receive descriptor blocks */

    unsigned int        rcvr_freewaiting;			/* waiting for descriptors to be freed */
    kcondvar_t		rcvr_freesleep;				/*   and sleep here */

    EP4_INTCOOKIE	rcvr_stall_intcookie;			/* interrupt cookie for thread halt */
    unsigned char	rcvr_thread_halted;			/* thread has been halted */
    unsigned char       rcvr_cleanup_waiting;			/* waiting for cleanup */
    kcondvar_t          rcvr_cleanup_sleep;			/*   and sleep here */

    EP4_RETRY_OPS	rcvr_retryops;

    struct list_head    rcvr_retrylist;				/* list of txd's to retry envelopes for */
    struct list_head    rcvr_polllist;				/* list of txd's to poll for completion */
    spinlock_t          rcvr_retrylock;
    
    EP4_RCVR_RAIL_STATS rcvr_stats;                             /* elan4 specific rcvr_rail stats */

} EP4_RCVR_RAIL;

#endif /* !defined(__elan4__) */

typedef struct ep4_txd_rail_elan
{
    EP4_INTR_CMD        txd_env_cmd;				/* command stream for envelope event (64 byte aligned) */
    E4_uint64		txd_pad0;				/*  pad to 32 bytes */
    E4_Event32		txd_env;				/* event set when STEN packet fails */

    EP4_INTR_CMD	txd_done_cmd;				/* command stream for done event (64 byte aligned) */
    E4_uint64		txd_pad1;				/*  pad to 32 bytes */
    E4_Event32		txd_done;				/* event set when transmit complete */

    E4_Event32		txd_data;				/* event set when xmit completes (=> phase becomes passive) */
} EP4_TXD_RAIL_ELAN;

#define EP4_TXD_RAIL_ELAN_SIZE		roundup(sizeof(EP4_TXD_RAIL_ELAN), 64)

typedef struct ep4_txd_rail_main
{
    E4_uint64		txd_env;
    E4_uint64		txd_data;
    E4_uint64		txd_done;
} EP4_TXD_RAIL_MAIN;

#define EP4_TXD_RAIL_MAIN_SIZE		roundup(sizeof(EP4_TXD_RAIL_MAIN), 8)

#if !defined (__elan4__)
typedef struct ep4_txd_rail
{
    EP_TXD_RAIL		txd_generic;

    struct list_head    txd_retry_link;
    unsigned long	txd_retry_time;

    EP4_INTCOOKIE	txd_intcookie;

    sdramaddr_t		txd_elan;
    EP_ADDR		txd_elan_addr;
    
    EP4_TXD_RAIL_MAIN  *txd_main;
    EP_ADDR		txd_main_addr;

    EP4_ECQ	       *txd_ecq;

    E4_uint64		txd_cookie;
} EP4_TXD_RAIL;

#define EP4_NUM_TXD_PER_BLOCK	21

typedef struct ep4_txd_rail_block
{
    struct list_head	blk_link;
    EP4_TXD_RAIL	blk_txds[EP4_NUM_TXD_PER_BLOCK];
} EP4_TXD_RAIL_BLOCK;

typedef struct ep4_xmtr_rail_main
{
    E4_int64		xmtr_flowcnt;
} EP4_XMTR_RAIL_MAIN;

typedef struct ep4_xmtr_rail_stats
{
    unsigned long some_stat;
} EP4_XMTR_RAIL_STATS;

#define EP4_TXD_LIST_POLL	0
#define EP4_TXD_LIST_STALLED	1
#define EP4_TXD_LIST_RETRY	2
#define EP4_TXD_NUM_LISTS	3
typedef struct ep4_xmtr_rail
{
    EP_XMTR_RAIL	xmtr_generic;

    EP4_XMTR_RAIL_MAIN *xmtr_main;
    EP_ADDR		xmtr_main_addr;

    struct list_head    xmtr_freelist;
    unsigned int        xmtr_freecount;
    unsigned int        xmtr_totalcount;
    spinlock_t          xmtr_freelock;
    struct list_head    xmtr_blocklist;
    unsigned int        xmtr_freewaiting;
    kcondvar_t		xmtr_freesleep;

    EP4_INTCOOKIE	xmtr_intcookie;				/* interrupt cookie for "polled" descriptors */

    ELAN4_CQ           *xmtr_cq;
    E4_int64		xmtr_flowcnt;

    EP4_RETRY_OPS	xmtr_retryops;

    struct list_head    xmtr_retrylist[EP4_TXD_NUM_LISTS];	/* list of txd's to retry envelopes for */
    struct list_head    xmtr_polllist;				/* list of txd's to poll for completion */
    spinlock_t          xmtr_retrylock;

    EP4_XMTR_RAIL_STATS stats;                                  /* elan4 specific xmtr rail stats */
} EP4_XMTR_RAIL;

#define EP4_XMTR_CQSIZE		CQ_Size64K				/* size of command queue for xmtr */
#define EP4_XMTR_FLOWCNT	(CQ_Size(EP4_XMTR_CQSIZE) / 512)	/* # of STEN packets which can fit in */

typedef struct ep4_comms_rail_elan
{
    E4_Event32		r_flush_event;
} EP4_COMMS_RAIL_ELAN;

#define EP4_COMMS_RAIL_ELAN_SIZE	roundup(sizeof (EP4_COMMS_RAIL_ELAN), 32)

typedef struct ep4_comms_rail
{
    EP_COMMS_RAIL	r_generic;				/* generic comms rail */
    sdramaddr_t		r_descs;				/* input queue descriptors */

    sdramaddr_t		r_elan;					/* elan portion */
    EP_ADDR		r_elan_addr;

    kmutex_t		r_flush_mutex;				/* sequentialise flush usage */
    EP4_INTCOOKIE       r_flush_intcookie;			/* interrupt cookie to generate */

    kcondvar_t		r_flush_sleep;				/* place to sleep waiting */
    spinlock_t		r_flush_lock;				/*   and spinlock to use */

    unsigned int	r_flush_count;				/* # setevents issued */
    EP4_ECQ	       *r_flush_ecq;				/* command queue for interrupt */
    EP4_ECQ	       *r_flush_mcq;				/* command queeu to issue waitevent */

    EP4_NETERR_OPS      r_neterr_ops;				/* network error fixup ops */
} EP4_COMMS_RAIL;

/* epcommsTx_elan4.c */
extern void           ep4xmtr_flush_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail);
extern void           ep4xmtr_failover_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail);
extern void           ep4xmtr_disconnect_callback (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail);

extern void	      ep4xmtr_neterr_flush (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies);
extern void	      ep4xmtr_neterr_check (EP_XMTR *xmtr, EP4_XMTR_RAIL *xmtrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies);

/* epcommsRx_elan4.c */
extern void           ep4rcvr_flush_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail);
extern void           ep4rcvr_failover_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail);
extern void           ep4rcvr_disconnect_callback (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail);

extern void	      ep4rcvr_neterr_flush (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies);
extern void	      ep4rcvr_neterr_check (EP_RCVR *rcvr, EP4_RCVR_RAIL *rcvrRail, unsigned int nodeId, EP_NETERR_COOKIE *cookies);

/* epcomms_elan4.c */
extern void           ep4comms_flush_start (EP4_COMMS_RAIL *commsRail);
extern void           ep4comms_flush_wait (EP4_COMMS_RAIL *commsRail);
extern void           ep4comms_flush_setevent (EP4_COMMS_RAIL *commsRail, ELAN4_CQ *cq);

extern EP_COMMS_RAIL *ep4comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *r);
extern void           ep4comms_del_rail (EP_COMMS_RAIL *r);
extern void	      ep4comms_display_rail (EP_COMMS_RAIL *r);

/* epcommsTx_elan4.c */
extern int            ep4xmtr_bind_txd (EP_TXD *txd, EP_XMTR_RAIL *xmtrRail, unsigned int phase);
extern void           ep4xmtr_unbind_txd (EP_TXD *txd, unsigned int phase);
extern int            ep4xmtr_poll_txd (EP_XMTR_RAIL *xmtrRail, EP_TXD_RAIL *txdRail, int how);
extern long           ep4xmtr_check (EP_XMTR_RAIL *xmtrRail, long nextRunTime);
extern void           ep4xmtr_add_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail);
extern void           ep4xmtr_del_rail (EP_XMTR *xmtr, EP_COMMS_RAIL *commsRail);
extern int            ep4xmtr_check_txd_state(EP_TXD *txd);

extern void           ep4xmtr_display_xmtr (DisplayInfo *di, EP_XMTR_RAIL *xmtrRail);
extern void           ep4xmtr_display_txd  (DisplayInfo *di, EP_TXD_RAIL *txdRail);

extern void           ep4xmtr_fillout_rail_stats (EP_XMTR_RAIL *xmtr_rail, char *str);

/* epcommsRx_elan4.c */
extern int	      ep4rcvr_queue_rxd (EP_RXD *rxd, EP_RCVR_RAIL *rcvrRail);
extern void	      ep4rcvr_rpc_put (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
extern void	      ep4rcvr_rpc_get (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
extern void	      ep4rcvr_rpc_complete (EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);

extern EP_RXD	     *ep4rcvr_steal_rxd (EP_RCVR_RAIL *rcvrRail);

extern long	      ep4rcvr_check (EP_RCVR_RAIL *rcvrRail, long nextRunTime);
extern void           ep4rcvr_add_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);
extern void           ep4rcvr_del_rail (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);

extern void           ep4rcvr_display_rcvr (DisplayInfo *di, EP_RCVR_RAIL *rcvrRail);
extern void           ep4rcvr_display_rxd  (DisplayInfo *di, EP_RXD_RAIL *rxdRail);

extern void           ep4rcvr_fillout_rail_stats (EP_RCVR_RAIL *rcvr_rail, char *str);

#endif /* !defined(__elan4__) */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __EPCOMMS_ELAN4_H */
