/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: epcomms_elan4_thread.c,v 1.13 2005/03/22 14:41:55 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms_elan4_thread.c,v $*/

//#include <qsnet/types.h>

typedef char           int8_t;
typedef unsigned char  uint8_t;
typedef short          int16_t;
typedef unsigned short uint16_t;
typedef int            int32_t;
typedef unsigned int   uint32_t;
typedef long           int64_t;
typedef unsigned long  uint64_t;

#include <elan/nmh.h>
#include <elan/kcomm.h>
#include <elan/epcomms.h>

#include <elan4/registers.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "epcomms_elan4.h"

#include <elan4/trtype.h>

/* assembler in epcomms_asm_elan4_thread.S */
extern void               c_waitevent_interrupt (E4_uint64 *cport, E4_Event32 *event, E4_uint64 count, E4_uint64 intcookie);
extern EP4_RXD_RAIL_ELAN *c_stall_thread (EP4_RCVR_RAIL_ELAN *rcvrRail);

#define R32_to_R47		"%r32", "%r33", "%r34", "%r35", "%r36", "%r37", "%r38", "%r39", \
				"%r40", "%r41", "%r42", "%r43", "%r44", "%r45", "%r46", "%r47"
#define R48_to_R63		"%r48", "%r49", "%r50", "%r51", "%r52", "%r53", "%r54", "%r55", \
	      			"%r56", "%r57", "%r58", "%r59", "%r60", "%r61", "%r62", "%r63"

/* proto types for code in asm_elan4_thread.S */
extern void c_waitevent (E4_uint64 *commandport, E4_Addr event, E4_uint64 count);
extern void c_reschedule(E4_uint64 *commandport);

static inline unsigned long
c_load_u16(unsigned short *ptr)
{
    unsigned long value;

    asm volatile ("ld2		[%1], %%r2\n"
		  "srl8,byte	%%r2, %1, %0\n"
		  "sll8		%0, 48, %0\n"
		  "srl8		%0, 48, %0\n"
		  : /* outputs */ "=r" (value) 
		  : /* inputs */ "r" (ptr)
		  : /* clobbered */ "%r2");
    return value;
}

static inline unsigned long
c_load_u32(unsigned int *ptr)
{
    unsigned long value;

    asm volatile ("ld4		[%1], %%r2\n"
		  "srl8,byte	%%r2, %1, %0\n"
		  "sll8		%0, 32, %0\n"
		  "srl8		%0, 32, %0\n"
		  : /* outputs */ "=r" (value) 
		  : /* inputs */ "r" (ptr)
		  : /* clobbered */ "%r2");
    return value;
}

static inline void
c_store_u32(unsigned int *ptr, unsigned long value)
{
    asm volatile ("sll8,byte	%0, %1, %%r2\n"
		  "st4		%%r2, [%1]\n"
		  : /* no outputs */ 
		  : /* inputs */ "r" (value), "r" (ptr)
		  : /* clobbered */ "%r2");
}

/* Reschedule the current Elan thread to the back of the run queue 
 * if there is another one ready to run */
static inline void
c_yield (E4_uint64 *commandport)
{
    unsigned long rval;

    asm volatile ("breaktest %0" : /* outputs */ "=r" (rval) : /* inputs */);

    if (rval  & ICC_SIGNED_BIT)
	c_reschedule(commandport);
}

/* Reschedule the current thread if we're in danger of exceeding the 
 * thread instruction count */
static inline void
c_insn_check(E4_uint64 *commandport)
{
    unsigned long rval;

    asm volatile ("breaktest %0" : /* outputs */ "=r" (rval) : /* inputs */);

    if (rval & ICC_ZERO_BIT)
	c_reschedule(commandport);
}

void
ep4_spinblock (E4_uint64 *cport, EP4_SPINLOCK_ELAN *sle, EP4_SPINLOCK_MAIN *slm)
{
    do {
	unsigned long val = *sle & 0xfffffffff;

	*slm = val;				 	/* Release my lock */
	
	while (*sle >> 32)				/* Wait until the main */
	    c_yield(cport);				/* releases the lock */
	
	c_store_u32 ((unsigned int *) sle, val + 1);	/* and try and relock */
    } while (*sle >> 32);
}

#define RESCHED_AFTER_PKTS	((CQ_Size(CQ_Size64K) / 128) - 1)

void
ep4comms_rcvr (EP4_RAIL_ELAN *rail, EP4_RCVR_RAIL_ELAN *rcvrElan, EP4_RCVR_RAIL_MAIN *rcvrMain,
	       E4_InputQueue *inputq, E4_uint64 *cport, E4_uint64 *resched)
{
    long count = 1;
    long fptr  = inputq->q_fptr;

    for (;;)
    {
	c_waitevent (cport, inputq->q_event, -count << 5);

	count = 0;

	while (fptr != inputq->q_bptr)
	{
	    EP_ENVELOPE        *env      = (EP_ENVELOPE *) fptr;
	    unsigned long       nodeid   = c_load_u32 (&env->NodeId);
	    unsigned long       opencmd  = OPEN_STEN_PKT_CMD | OPEN_PACKET(0, PACK_OK | RESTART_COUNT_ZERO, EP_VP_DATA(nodeid));
	    unsigned long       vproc    = EP_VP_DATA(rail->r_nodeid);
	    EP_ATTRIBUTE        attr     = c_load_u32 (&env->Attr);
	    unsigned long       txdRail  = c_load_u32 (&env->TxdRail);
	    unsigned long       nFrags   = c_load_u32 (&env->nFrags);
	    unsigned long       srcevent = (EP_IS_RPC(attr) ? txdRail + offsetof (EP4_TXD_RAIL_ELAN, txd_data) :
					    txdRail + offsetof (EP4_TXD_RAIL_ELAN, txd_done));
	    E4_uint64           cookie;
	    EP4_RXD_RAIL_ELAN  *rxdElan;
	    EP4_RXD_RAIL_MAIN  *rxdMain;
	    EP_RXD_MAIN        *rxd;
	    EP4_RXD_STEN_CMD   *sten;
	    E4_Event32         *event;
	    unsigned long       first;
	    unsigned long       buffer;
	    unsigned long       len;
	    unsigned long       i;

	    EP4_SPINENTER(resched, &rcvrElan->rcvr_thread_lock, &rcvrMain->rcvr_thread_lock);

	    if ((rxdElan = (EP4_RXD_RAIL_ELAN *) rcvrElan->rcvr_pending_head) == 0)
	    {
		EP4_SPINEXIT (resched, &rcvrElan->rcvr_thread_lock, &rcvrMain->rcvr_thread_lock);

		rxdElan = c_stall_thread (rcvrElan);

		EP4_SPINENTER(resched, &rcvrElan->rcvr_thread_lock, &rcvrMain->rcvr_thread_lock);
	    }
	    
	    if (c_load_u32 (&env->Version) != EP_ENVELOPE_VERSION)		/* envelope has been cancelled */
	    {
		EP4_SPINEXIT (resched, &rcvrElan->rcvr_thread_lock, &rcvrMain->rcvr_thread_lock);
		goto consume_envelope;
	    }

	    rxd     = (EP_RXD_MAIN *) rxdElan->rxd_rxd;
	    rxdMain = (EP4_RXD_RAIL_MAIN *) rxdElan->rxd_main;
	    first   = (EP_MAXFRAG+1) - (( EP_IS_MULTICAST(attr) ? 1 : 0) + (nFrags == 0 ? 1 : nFrags));
	    sten    = &rxdElan->rxd_sten[first];
	    event   = &rxdElan->rxd_chain[first];
	    cookie  = rail->r_cookies[nodeid];

	    if (EP_IS_MULTICAST(attr))				/* need to fetch broadcast bitmap */
	    {
		sten->c_open          = opencmd;
		sten->c_trans         = SEND_TRANS_CMD | ((TR_REMOTEDMA | TR_WAIT_FOR_EOP) << 16);
		sten->c_cookie        = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_STEN;
		sten->c_dma_typeSize  = E4_DMA_TYPE_SIZE(BT_BITOUL(EP_MAX_NODES) * sizeof (bitmap_t), DMA_DataTypeWord, 0, EP4_DMA_RETRYCOUNT);
		sten->c_dma_cookie    = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_REMOTE | EP4_COOKIE_DMA | EP4_COOKIE_INC;
		sten->c_dma_vproc     = vproc;
		sten->c_dma_srcAddr   = c_load_u32 (&env->TxdMain.nmd_addr) + offsetof(EP_TXD_MAIN, Bitmap);
		sten->c_dma_dstAddr   = (E4_Addr) &rxd->Bitmap;
		sten->c_dma_srcEvent  = srcevent;
		sten->c_dma_dstEvent  = (E4_Addr) event;

		event->ev_CountAndType = E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS);

		cookie += (EP4_COOKIE_INC << 1);

		sten++; event++;
	    }

	    if (nFrags == 0)
	    {
		/* Generate an empty "get" DMA to accept the envelope and fire the rx handler */
		sten->c_open          = opencmd;
		sten->c_trans         = SEND_TRANS_CMD | ((TR_REMOTEDMA | TR_WAIT_FOR_EOP) << 16);
		sten->c_cookie        = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_STEN;
		sten->c_dma_typeSize  = E4_DMA_TYPE_SIZE(0, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
		sten->c_dma_cookie    = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_REMOTE | EP4_COOKIE_DMA | EP4_COOKIE_INC;
		sten->c_dma_vproc     = vproc;
		sten->c_dma_srcEvent  = srcevent;
		sten->c_dma_dstEvent  = (E4_Addr) event;

		event->ev_CountAndType = E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS);

		len = 0;

		cookie += (EP4_COOKIE_INC << 1);
	    }
	    else
	    {
		/* Generate the DMA chain to fetch the data */
		for (i = 0, buffer = c_load_u32 (&rxdElan->rxd_buffer.nmd_addr), len = 0; i < nFrags; i++)
		{
		    unsigned long fragLen = c_load_u32 (&env->Frags[i].nmd_len);

		    sten->c_open          = opencmd;
		    sten->c_trans         = SEND_TRANS_CMD | ((TR_REMOTEDMA | TR_WAIT_FOR_EOP) << 16);
		    sten->c_cookie        = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_STEN;
		    sten->c_dma_typeSize  = E4_DMA_TYPE_SIZE(fragLen, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
		    sten->c_dma_cookie    = cookie | EP4_COOKIE_THREAD | EP4_COOKIE_REMOTE | EP4_COOKIE_DMA | EP4_COOKIE_INC;
		    sten->c_dma_vproc     = vproc;
		    sten->c_dma_srcAddr   = c_load_u32 (&env->Frags[i].nmd_addr);
		    sten->c_dma_dstAddr   = buffer;
		    sten->c_dma_srcEvent  = srcevent;
		    sten->c_dma_dstEvent  = (E4_Addr) event;
		    
		    event->ev_CountAndType = E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_RXD_STEN_CMD_NDWORDS);
		    
		    buffer += fragLen;
		    len    += fragLen;

		    cookie += (EP4_COOKIE_INC << 1);

		    sten++; event++;
		}
		
		(--event)->ev_CountAndType = E4_EVENT_INIT_VALUE(-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_INTR_CMD_NDWORDS);

		if (c_load_u32 (&rxdElan->rxd_buffer.nmd_len) < len)
		{
		    /* The receive descriptor was too small for the message */
		    /* complete the message anyway,  but don't transfer any */
		    /* data,  we set the length to EP_MSG_TOO_BIG */
		    for (i = first, sten = &rxdElan->rxd_sten[first]; i <= EP_MAXFRAG; i++, sten++)
			sten->c_dma_typeSize = E4_DMA_TYPE_SIZE(0, DMA_DataTypeByte, 0, EP4_DMA_RETRYCOUNT);
		    
		    len = EP_MSG_TOO_BIG;
		}
	    }
	    
	    /* Stuff the first STEN packet into the command queue, there's always enough space, 
	     * since we will insert a waitevent at least once for the queue size */
	    asm volatile ("ld64		[%0], %%r32\n"
			  "ld64		[%0 + 64], %%r48\n"
			  "st64		%%r32, [%1]\n"
			  "st64		%%r48, [%1]\n"
			  : /* no outputs */
			  : /* inputs */ "r" (&rxdElan->rxd_sten[first]), "r" (cport)
			  : /* clobbered */ R32_to_R47, R48_to_R63);

	    /* remove the RXD from the pending list */
	    if ((rcvrElan->rcvr_pending_head = rxdElan->rxd_next) == 0)
		rcvrElan->rcvr_pending_tailp = (E4_Addr)&rcvrElan->rcvr_pending_head;

	    /* mark as not queued */
	    rxdElan->rxd_queued = 0;

	    /* copy down the envelope */
	    if (EP_HAS_PAYLOAD(attr))
		asm volatile ("ld64	[%0],    %%r32\n"
			      "ld64	[%0+64], %%r48\n"
			      "st64	%%r32, [%1]\n"
			      "ld64	[%0+128], %%r32\n"
			      "st64	%%r48, [%1+64]\n"
			      "ld64	[%0+192], %%r48\n"
			      "st64	%%r32, [%1 + 128]\n"
			      "st64	%%r48, [%1 + 192]\n"
			      : /* no outputs */
			      : /* inputs */	"r" (env), "r" (&rxd->Envelope)
			      : /* clobbered */ R32_to_R47, R48_to_R63);
 
	    else
		asm volatile ("ld64	[%0],    %%r32\n"
			      "ld64	[%0+64], %%r48\n"
			      "st64	%%r32, [%1]\n"
			      "st64	%%r48, [%1+64]\n"
			      : /* no outputs */
			      : /* inputs */	"r" (env), "r" (&rxd->Envelope)
			      : /* clobbered */ R32_to_R47, R48_to_R63);

	    /* Store the message length to indicate that I've finished */
	    c_store_u32 (&rxd->Len, len);
	    
	    /* Finally update the network error cookie */
	    rail->r_cookies[nodeid] = cookie;

	    EP4_SPINEXIT (resched, &rcvrElan->rcvr_thread_lock, &rcvrMain->rcvr_thread_lock);

	consume_envelope:
	    if (fptr != rcvrElan->rcvr_qlast)
		fptr += EP_INPUTQ_SIZE;
	    else
		fptr = rcvrElan->rcvr_qbase;

	    if (! rcvrElan->rcvr_stall_intcookie)
		inputq->q_fptr = fptr;

	    if (++count >= RESCHED_AFTER_PKTS)
		break;

	    c_insn_check (cport);
	}
	
	if (rcvrElan->rcvr_stall_intcookie)
	{
	    c_waitevent_interrupt (cport, &rcvrElan->rcvr_thread_halt, -(1 << 5), rcvrElan->rcvr_stall_intcookie);
	    inputq->q_fptr = fptr;

	    count++;						/* one extra as we were given an extra set to wake us up */
	}
    }
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
