/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kalloc.c,v 1.19 2004/12/14 10:19:23 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kalloc.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "debug.h"

static void
HashInPool (EP_ALLOC *alloc, EP_POOL *pool)
{
    int idx0 = HASH (pool->Handle.nmh_nmd.nmd_addr);
    int idx1 = HASH (pool->Handle.nmh_nmd.nmd_addr + pool->Handle.nmh_nmd.nmd_len);

    list_add (&pool->HashBase, &alloc->HashBase[idx0]);
    list_add (&pool->HashTop, &alloc->HashTop[idx1]);
}

static void
HashOutPool (EP_ALLOC *alloc, EP_POOL *pool)
{
    list_del (&pool->HashBase);
    list_del (&pool->HashTop);
}

static EP_POOL *
LookupPool (EP_ALLOC *alloc, EP_ADDR addr)
{
    struct list_head *el;
    
    list_for_each (el, &alloc->HashBase[HASH(addr)]) {
	EP_POOL *pool = list_entry (el, EP_POOL, HashBase);
	
	if (pool->Handle.nmh_nmd.nmd_addr <= addr && addr < (pool->Handle.nmh_nmd.nmd_addr + pool->Handle.nmh_nmd.nmd_len))
	    return (pool);
    }
    
    list_for_each (el, &alloc->HashTop[HASH(addr)]) {
	EP_POOL *pool = list_entry (el, EP_POOL, HashTop);
	
	if (pool->Handle.nmh_nmd.nmd_addr <= addr && addr < (pool->Handle.nmh_nmd.nmd_addr + pool->Handle.nmh_nmd.nmd_len))
	    return (pool);
    }
    
    return (NULL);
}

static EP_POOL *
AllocatePool (EP_ALLOC *alloc, EP_ADDR addr, unsigned size, unsigned int perm, EP_ATTRIBUTE attr)
{
    EP_ADDR base = 0;
    EP_POOL *pool;
    EP_RAIL *rail;
    int i, railmask = 0;
    struct list_head *el;

    KMEM_ZALLOC (pool, EP_POOL *, sizeof (EP_POOL), !(attr & EP_NO_SLEEP));
    
    if (pool == NULL)
	return (NULL);
    
    if (addr != 0)
	base = addr;
    else
    {
	for (i = LN2_MIN_SIZE; i <= LN2_MAX_SIZE; i ++)
	{
	    KMEM_ZALLOC (pool->Bitmaps[i - LN2_MIN_SIZE], bitmap_t *, BT_BITOUL(1 << (LN2_MAX_SIZE-i)) * sizeof (bitmap_t), !(attr & EP_NO_SLEEP));
	    if (pool->Bitmaps[i - LN2_MIN_SIZE] == NULL)
		goto failed;
	}
    
	if ((base = ep_rmalloc (alloc->ResourceMap, size, !(attr & EP_NO_SLEEP))) == 0)
	    goto failed;
    }

    switch (alloc->Type)
    {
    case EP_ALLOC_TYPE_PRIVATE_SDRAM:
	rail = alloc->Data.Private.Rail;

	if ((pool->Buffer.Sdram = rail->Operations.SdramAlloc (rail, base, size)) == 0)
	    goto failed;

	ep_perrail_sdram_map (rail, base, pool->Buffer.Sdram, size, perm, attr);

	pool->Handle.nmh_nmd.nmd_addr = base;
	pool->Handle.nmh_nmd.nmd_len  = size;
	break;
	
    case EP_ALLOC_TYPE_PRIVATE_MAIN:
        KMEM_GETPAGES(pool->Buffer.Ptr, unsigned long, btop (size), !(attr & EP_NO_SLEEP));
	if (pool->Buffer.Ptr == 0)
	    goto failed;

	ep_perrail_kaddr_map (alloc->Data.Private.Rail, base, pool->Buffer.Ptr, size, perm, attr);

	pool->Handle.nmh_nmd.nmd_addr = base;
	pool->Handle.nmh_nmd.nmd_len  = size;
	break;

    case EP_ALLOC_TYPE_SHARED_MAIN:
        KMEM_GETPAGES(pool->Buffer.Ptr, unsigned long, btop (size), !(attr & EP_NO_SLEEP));
	if (pool->Buffer.Ptr == 0)
	    goto failed;

	list_for_each (el, &alloc->Data.Shared.Rails) {
	    EP_RAIL *rail = list_entry (el, EP_RAIL_ENTRY, Link)->Rail;

	    ep_perrail_kaddr_map (rail, base, pool->Buffer.Ptr, size, perm, attr);

	    railmask |= (1 << rail->Number);
	}
	pool->Handle.nmh_nmd.nmd_addr = base;
	pool->Handle.nmh_nmd.nmd_len  = size;
	pool->Handle.nmh_nmd.nmd_attr = EP_NMD_ATTR (alloc->Data.Shared.System->Position.pos_nodeid, railmask);

	ep_nmh_insert (&alloc->Data.Shared.System->MappingTable, &pool->Handle);
	break;

    default:
	goto failed;
    }

    return (pool);
    
 failed:
    if (addr == 0 && base)
	ep_rmfree (alloc->ResourceMap, size, base);

    for (i = LN2_MIN_SIZE; i <= LN2_MAX_SIZE; i ++)
	if (pool->Bitmaps[i - LN2_MIN_SIZE] != NULL)
	    KMEM_FREE (pool->Bitmaps[i - LN2_MIN_SIZE], BT_BITOUL(1 << (LN2_MAX_SIZE - i)) * sizeof (bitmap_t));
    
    KMEM_FREE (pool, sizeof (EP_POOL));
    return (NULL);
}

static void
FreePool (EP_ALLOC *alloc, EP_POOL *pool)
{
    struct list_head *el;
    int i;

    switch (alloc->Type)
    {
    case EP_ALLOC_TYPE_PRIVATE_SDRAM:
	ep_perrail_unmap (alloc->Data.Private.Rail, pool->Handle.nmh_nmd.nmd_addr, pool->Handle.nmh_nmd.nmd_len);

	alloc->Data.Private.Rail->Operations.SdramFree (alloc->Data.Private.Rail, pool->Buffer.Sdram, pool->Handle.nmh_nmd.nmd_len);
	break;
	
    case EP_ALLOC_TYPE_PRIVATE_MAIN:
	ep_perrail_unmap (alloc->Data.Private.Rail, pool->Handle.nmh_nmd.nmd_addr, pool->Handle.nmh_nmd.nmd_len);

	KMEM_FREEPAGES (pool->Buffer.Ptr, btop (pool->Handle.nmh_nmd.nmd_len));
	break;

    case EP_ALLOC_TYPE_SHARED_MAIN:
	ep_nmh_remove (&alloc->Data.Shared.System->MappingTable, &pool->Handle);

	list_for_each (el, &alloc->Data.Shared.Rails) {
	    EP_RAIL *rail = list_entry (el, EP_RAIL_ENTRY, Link)->Rail;

	    ep_perrail_unmap (rail, pool->Handle.nmh_nmd.nmd_addr, pool->Handle.nmh_nmd.nmd_len);
	}

	KMEM_FREEPAGES (pool->Buffer.Ptr, btop (pool->Handle.nmh_nmd.nmd_len));
	break;
    }
    
    if (pool->Bitmaps[0])
    {
	ep_rmfree (alloc->ResourceMap, pool->Handle.nmh_nmd.nmd_len, pool->Handle.nmh_nmd.nmd_addr);
    
	for (i = LN2_MIN_SIZE; i <= LN2_MAX_SIZE; i ++)
	    KMEM_FREE (pool->Bitmaps[i - LN2_MIN_SIZE], BT_BITOUL(1 << (LN2_MAX_SIZE - i)) * sizeof (bitmap_t));
    }
    
    KMEM_FREE (pool, sizeof (EP_POOL));
}

static int
AddRail (EP_ALLOC *alloc, EP_RAIL *rail)
{
    struct list_head *el;
    EP_RAIL_ENTRY *l;
    unsigned long flags;
    int i;

    ASSERT (alloc->Type == EP_ALLOC_TYPE_SHARED_MAIN);

    KMEM_ZALLOC (l, EP_RAIL_ENTRY *, sizeof (EP_RAIL_ENTRY), 1);

    if (l == NULL)
	return (ENOMEM);

    l->Rail = rail;

    spin_lock_irqsave (&alloc->Lock, flags);
    for (i = 0; i < NHASH; i++)
    {
	list_for_each (el, &alloc->HashBase[i]) {
	    EP_POOL *pool = list_entry (el, EP_POOL, HashBase);

	    ep_perrail_kaddr_map (rail, pool->Handle.nmh_nmd.nmd_addr, pool->Buffer.Ptr, 
				  pool->Handle.nmh_nmd.nmd_len, EP_PERM_WRITE, EP_NO_SLEEP);

	    pool->Handle.nmh_nmd.nmd_attr |= EP_NMD_ATTR (0, 1 << rail->Number);
	}
    }

    list_add (&l->Link, &alloc->Data.Shared.Rails);

    spin_unlock_irqrestore (&alloc->Lock, flags); 
    return (0);
}

static void
RemoveRail (EP_ALLOC *alloc, EP_RAIL *rail)
{
    struct list_head *el;
    unsigned long flags;
    int i;

    spin_lock_irqsave (&alloc->Lock, flags);
    for (i = 0; i < NHASH; i++)
    {
	list_for_each (el, &alloc->HashBase[i]) {
	    EP_POOL *pool = list_entry (el, EP_POOL, HashBase);

	    ep_perrail_unmap (rail, pool->Handle.nmh_nmd.nmd_addr, pool->Handle.nmh_nmd.nmd_len);

	    pool->Handle.nmh_nmd.nmd_attr &= ~EP_NMD_ATTR (0, 1 << rail->Number);
	}
    }

    list_for_each (el, &alloc->Data.Shared.Rails) {
	EP_RAIL_ENTRY *tmp = list_entry (el, EP_RAIL_ENTRY, Link);
	if (tmp->Rail == rail)
	{
	    list_del (el);
	    KMEM_FREE(tmp, sizeof (EP_RAIL_ENTRY));
	    break;
	}
    }

    spin_unlock_irqrestore (&alloc->Lock, flags);
}

static EP_POOL *
AllocateBlock (EP_ALLOC *alloc, unsigned size, EP_ATTRIBUTE attr, int *offset)
{
    int block, j, k;
    unsigned long flags;
    EP_POOL *pool;


    if (size > MAX_SIZE)
    {
	if ((attr & EP_NO_ALLOC) || (pool  = AllocatePool (alloc, 0, size, alloc->Perm, attr)) == NULL)
	    return (NULL);

	spin_lock_irqsave (&alloc->Lock, flags);
	HashInPool (alloc, pool);
	spin_unlock_irqrestore (&alloc->Lock, flags);

	*offset = 0;

	return pool;
    }

    spin_lock_irqsave (&alloc->Lock, flags);

    /* Round up size to next power of 2 */
    for (k = LN2_MIN_SIZE; (1 << k) < size; k++)
	;
    
    /* k now has ln2 of the size to allocate. */
    /* find the free list with the smallest block we can use*/
    for (j = k; j <= LN2_MAX_SIZE && list_empty (&alloc->Freelists[j - LN2_MIN_SIZE]); j++)
	;
    
    /* j has ln2 of the smallest size block we can use */
    if (j < LN2_MAX_SIZE)
    {
	int nbits = 1 << (LN2_MAX_SIZE-j);
	
	pool  = list_entry (alloc->Freelists[j - LN2_MIN_SIZE].next, EP_POOL, Link[j - LN2_MIN_SIZE]);
	block = (bt_lowbit (pool->Bitmaps[j - LN2_MIN_SIZE], nbits) << j);
	
	BT_CLEAR (pool->Bitmaps[j - LN2_MIN_SIZE], block >> j);
	
	if (bt_lowbit (pool->Bitmaps[j - LN2_MIN_SIZE], nbits) == -1)
	    list_del (&pool->Link[j - LN2_MIN_SIZE]);
    }
    else
    {
	spin_unlock_irqrestore (&alloc->Lock, flags);
	
	if ((attr & EP_NO_ALLOC) || (pool  = AllocatePool (alloc, 0, MAX_SIZE, alloc->Perm, attr)) == NULL)
	    return (NULL);

	block = 0;
	j = LN2_MAX_SIZE;
	
	spin_lock_irqsave (&alloc->Lock, flags);
	
	HashInPool (alloc, pool);
    }
    
    /* Split it until the buddies are the correct size, putting one
     * buddy back on the free list and continuing to split the other */
    while (--j >= k)
    {
	list_add (&pool->Link[j - LN2_MIN_SIZE], &alloc->Freelists[j - LN2_MIN_SIZE]);
	
	BT_SET (pool->Bitmaps[j - LN2_MIN_SIZE], block >> j);
	
	block += (1 << j);
    }
    spin_unlock_irqrestore (&alloc->Lock, flags);

    *offset = block;

    return (pool);
}

static void
FreeBlock (EP_ALLOC *alloc, EP_ADDR addr, unsigned size)
{
    EP_POOL *pool;
    int  k, block = 0;
    unsigned long flags;
    
    spin_lock_irqsave (&alloc->Lock, flags);
    /* Round up size to next power of 2 */
    for (k = LN2_MIN_SIZE; (1 << k) < size; k++)
	;

    /* Find the pool containing this block */
    pool = LookupPool (alloc, addr);

    /* It must exist */
    ASSERT (pool != NULL);

    /* If we're freeing a subset of it, then update the bitmaps */
    if (size <= MAX_SIZE)
    {
	ASSERT (BT_TEST (pool->Bitmaps[k - LN2_MIN_SIZE], (addr - pool->Handle.nmh_nmd.nmd_addr) >> k) == 0);
	
	block = addr - pool->Handle.nmh_nmd.nmd_addr;
	
	while (k < LN2_MAX_SIZE && BT_TEST (pool->Bitmaps[k - LN2_MIN_SIZE], (block >> k) ^ 1))
	{
	    BT_CLEAR (pool->Bitmaps[k - LN2_MIN_SIZE], (block >> k) ^ 1);
	    
	    if (bt_lowbit (pool->Bitmaps[k - LN2_MIN_SIZE], (1 << (LN2_MAX_SIZE - k))) == -1)
		list_del (&pool->Link[k - LN2_MIN_SIZE]);
	    
	    k++;
	}
    }

    if (k >= LN2_MAX_SIZE)
    {
	HashOutPool (alloc, pool);
	spin_unlock_irqrestore (&alloc->Lock, flags);

	FreePool (alloc, pool);
    }
    else
    {
	if (bt_lowbit (pool->Bitmaps[k - LN2_MIN_SIZE], (1 << (LN2_MAX_SIZE - k))) == -1)
	    list_add (&pool->Link[k - LN2_MIN_SIZE], &alloc->Freelists[k - LN2_MIN_SIZE]);

	BT_SET (pool->Bitmaps[k - LN2_MIN_SIZE], block >> k);

	spin_unlock_irqrestore (&alloc->Lock, flags);
    }
}

static void
InitialiseAllocator (EP_ALLOC *alloc, EP_ALLOC_TYPE type, unsigned int perm, EP_RMAP *rmap)
{
    int i;

    spin_lock_init (&alloc->Lock);

    alloc->Type        = type;
    alloc->ResourceMap = rmap;
    alloc->Perm        = perm;

    for (i = 0; i < NHASH; i++)
    {
	(&alloc->HashBase[i])->next = &alloc->HashBase[i];

	INIT_LIST_HEAD (&alloc->HashBase[i]);
	INIT_LIST_HEAD (&alloc->HashTop[i]);
    }
    
    for (i = 0; i < NUM_FREELISTS; i++)
	INIT_LIST_HEAD (&alloc->Freelists[i]);
}

static void
DestroyAllocator (EP_ALLOC *alloc)
{
    struct list_head *el, *next;
    int i;

    for (i = 0; i < NHASH; i++)
    {
	list_for_each_safe (el, next, &alloc->HashBase[i]) { 
	    EP_POOL *pool = list_entry (el, EP_POOL, HashBase);

	    printk ("!!DestroyAllocator: pool=%p type=%d addr=%x len=%x\n", pool, alloc->Type,
		    pool->Handle.nmh_nmd.nmd_addr, pool->Handle.nmh_nmd.nmd_len);

	    list_del (&pool->HashBase);
	    list_del (&pool->HashTop);

	    // XXXX: FreePool (alloc, pool);
	}
    }

    spin_lock_destroy (&alloc->Lock);
}

void
ep_display_alloc (EP_ALLOC *alloc)
{
    struct list_head *el;
    int i;
    int npools = 0;
    int nbytes = 0;
    int nfree = 0;
    unsigned long flags;

    spin_lock_irqsave (&alloc->Lock, flags);

    ep_debugf (DBG_DEBUG, "Kernel comms memory allocator %p type %d\n", alloc, alloc->Type);
    for (i = 0; i < NHASH; i++)
    {
	list_for_each (el, &alloc->HashBase[i]) {
	    EP_POOL *pool = list_entry (el, EP_POOL, HashBase);

	    ep_debugf (DBG_DEBUG, "  POOL %4x: %p -> %x.%x\n", i, pool, pool->Handle.nmh_nmd.nmd_addr,
		       pool->Handle.nmh_nmd.nmd_addr + pool->Handle.nmh_nmd.nmd_len);

	    npools++;
	    nbytes += pool->Handle.nmh_nmd.nmd_len;
	}
    }
    
    for (i = LN2_MIN_SIZE; i <= LN2_MAX_SIZE; i++)
    {
	int n = 0;

	list_for_each (el, &alloc->Freelists[i - LN2_MIN_SIZE]) {
	    EP_POOL *pool  = list_entry (el, EP_POOL, Link[i - LN2_MIN_SIZE]);
	    int      nbits = bt_nbits (pool->Bitmaps[i - LN2_MIN_SIZE], 1 << (LN2_MAX_SIZE - i));

	    n += nbits;
	    nfree += (nbits << i);
	}
	
	if (n != 0)
	    ep_debugf (DBG_DEBUG, "  SIZE %5d : num %d\n", (1 << i), n);
    }
    ep_debugf (DBG_DEBUG, "%d pools with %d bytes and %d bytes free\n", npools, nbytes, nfree);

    spin_unlock_irqrestore (&alloc->Lock, flags);
}

/* per-rail allocators */
void
ep_alloc_init (EP_RAIL *rail)
{
    EP_RMAP *rmap = ep_rmallocmap (EP_PRIVATE_RMAP_SIZE, "PrivateMap", 1);

    ep_rmfree (rmap, EP_PRIVATE_TOP-EP_PRIVATE_BASE, EP_PRIVATE_BASE);

    InitialiseAllocator (&rail->ElanAllocator, EP_ALLOC_TYPE_PRIVATE_SDRAM, EP_PERM_ALL, rmap);
    InitialiseAllocator (&rail->MainAllocator, EP_ALLOC_TYPE_PRIVATE_MAIN, EP_PERM_WRITE, rmap);

    rail->ElanAllocator.Data.Private.Rail = rail;
    rail->MainAllocator.Data.Private.Rail = rail;
}

void
ep_alloc_fini (EP_RAIL *rail)
{
    EP_RMAP *rmap = rail->ElanAllocator.ResourceMap;

    DestroyAllocator (&rail->ElanAllocator);
    DestroyAllocator (&rail->MainAllocator);
    
    ep_rmfreemap (rmap);
}

sdramaddr_t
ep_alloc_memory_elan (EP_RAIL *rail, EP_ADDR addr, unsigned size, unsigned int perm, EP_ATTRIBUTE attr)
{
    EP_POOL *pool = AllocatePool (&rail->ElanAllocator, addr, size, perm, attr);
    unsigned long flags;

    if (pool == NULL)
	return (0);

    spin_lock_irqsave (&rail->ElanAllocator.Lock, flags);
    HashInPool (&rail->ElanAllocator, pool);
    spin_unlock_irqrestore (&rail->ElanAllocator.Lock, flags);

    return (pool->Buffer.Sdram);
}

void
ep_free_memory_elan (EP_RAIL *rail, EP_ADDR addr)
{
    EP_POOL *pool;
    unsigned long flags;

    spin_lock_irqsave (&rail->ElanAllocator.Lock, flags);
    pool = LookupPool (&rail->ElanAllocator, addr);
    
    HashOutPool (&rail->ElanAllocator, pool);
    spin_unlock_irqrestore (&rail->ElanAllocator.Lock, flags);
    
    FreePool (&rail->ElanAllocator, pool);
}

sdramaddr_t
ep_alloc_elan (EP_RAIL *rail, unsigned size, EP_ATTRIBUTE attr, EP_ADDR *addrp)
{
    int	     offset;
    EP_POOL *pool;

    if ((pool = AllocateBlock (&rail->ElanAllocator, size, attr, &offset)) == NULL)
	return (0);
    
    *addrp  = pool->Handle.nmh_nmd.nmd_addr + offset;

    return (pool->Buffer.Sdram + offset);
}

void
ep_free_elan (EP_RAIL *rail, EP_ADDR addr, unsigned size)
{
    FreeBlock (&rail->ElanAllocator, addr, size);
}

void *
ep_alloc_main (EP_RAIL *rail, unsigned size, EP_ATTRIBUTE attr, EP_ADDR *addrp)
{
    int	     offset;
    EP_POOL *pool;

    if ((pool = AllocateBlock (&rail->MainAllocator, size, attr, &offset)) == NULL)
	return (NULL);
    
    *addrp  = pool->Handle.nmh_nmd.nmd_addr + offset;

    return ((void *) ((unsigned long) pool->Buffer.Ptr + offset));
}

void
ep_free_main (EP_RAIL *rail, EP_ADDR addr, unsigned size)
{
    FreeBlock (&rail->MainAllocator, addr, size);
}

sdramaddr_t
ep_elan2sdram (EP_RAIL *rail, EP_ADDR addr)
{
    EP_POOL    *pool;
    sdramaddr_t res;
    unsigned long flags;

    spin_lock_irqsave (&rail->ElanAllocator.Lock, flags);
    if ((pool = LookupPool (&rail->ElanAllocator, addr)) == NULL)
	res = 0;
    else
	res = pool->Buffer.Sdram + (addr - pool->Handle.nmh_nmd.nmd_addr);
    spin_unlock_irqrestore (&rail->ElanAllocator.Lock, flags);

    return (res);
}

void *
ep_elan2main (EP_RAIL *rail, EP_ADDR addr)
{
    EP_POOL *pool;
    void *res;
    unsigned long flags;

    spin_lock_irqsave (&rail->MainAllocator.Lock, flags);
    if ((pool = LookupPool (&rail->MainAllocator, addr)) == NULL)
	res = NULL;
    else
	res = (void *) ((unsigned long) pool->Buffer.Ptr + (addr - pool->Handle.nmh_nmd.nmd_addr));
    spin_unlock_irqrestore (&rail->MainAllocator.Lock, flags);

    return (res);
}

/* shared allocators */
int
ep_shared_alloc_add_rail (EP_SYS *sys, EP_RAIL *rail)
{
    return (AddRail (&sys->Allocator, rail));
}

void
ep_shared_alloc_remove_rail (EP_SYS *sys, EP_RAIL *rail)
{
    RemoveRail (&sys->Allocator, rail);
}

void
ep_shared_alloc_init (EP_SYS *sys)
{
    EP_RMAP *rmap = ep_rmallocmap (EP_SHARED_RMAP_SIZE, "shared_alloc_map", 1);

    ep_rmfree (rmap, EP_SHARED_TOP - EP_SHARED_BASE, EP_SHARED_BASE);

    InitialiseAllocator (&sys->Allocator, EP_ALLOC_TYPE_SHARED_MAIN, EP_PERM_WRITE, rmap);

    INIT_LIST_HEAD (&sys->Allocator.Data.Shared.Rails);

    sys->Allocator.Data.Shared.System = sys;
}

void
ep_shared_alloc_fini (EP_SYS *sys)
{
    EP_RMAP *rmap = sys->Allocator.ResourceMap;

    DestroyAllocator (&sys->Allocator);

    ep_rmfreemap (rmap);
}

void *
ep_shared_alloc_main (EP_SYS *sys, unsigned size, EP_ATTRIBUTE attr, EP_NMD *nmd)
{
    int offset;
    EP_POOL *pool;

    if ((pool = AllocateBlock (&sys->Allocator, size, attr, &offset)) == NULL)
	return (NULL);

    ep_nmd_subset (nmd, &pool->Handle.nmh_nmd, offset, size);

    return ((void *) ((unsigned long) pool->Buffer.Ptr + offset));
}

void
ep_shared_free_main (EP_SYS *sys, EP_NMD *nmd)
{
    FreeBlock (&sys->Allocator, nmd->nmd_addr, nmd->nmd_len);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
