
/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kcomm_elan3.c,v 1.34 2004/11/30 12:02:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "conf_linux.h"

extern EP_CODE threadcode_elan3;

unsigned int
ep3_create_rails (EP_SYS *sys, unsigned int disabled)
{
    unsigned int rmask = 0;
    ELAN3_DEV   *dev;
    EP_RAIL     *rail;
    int          i;

    for (i = 0; i < EP_MAX_RAILS; i++)
    {
	if ((dev = elan3_device (i)) != NULL)
	{
	    if ((rail = ep3_create_rail (sys, dev)) != NULL)
	    {
		if (disabled & (1 << rail->Number))
		    printk ("%s: auto-start of device disabled by configuration\n", rail->Name);
		else
		    ep_start_rail (rail);
		
		ep_procfs_rail_init(rail);

		rmask |= (1 << rail->Number);
	    }
	}
    }

    return rmask;
}

EP_RAIL *
ep3_create_rail (EP_SYS *sys, ELAN3_DEV *dev)
{
    EP3_RAIL *rail;
    int       res;

    KMEM_ZALLOC (rail, EP3_RAIL *, sizeof (EP3_RAIL), TRUE);

    if (rail == NULL)
	return (EP_RAIL *) NULL;

    if ((res = ep_init_rail (sys, &rail->Generic)) != 0)
    {
	KMEM_FREE (rail, sizeof (EP3_RAIL));
	return (EP_RAIL *) NULL;
    }

    rail->Device = dev;

    /* Install our rail operations */
    rail->Generic.Operations.DestroyRail      = ep3_destroy_rail;
    rail->Generic.Operations.StartRail        = ep3_start_rail;
    rail->Generic.Operations.StallRail        = ep3_stall_rail;
    rail->Generic.Operations.StopRail         = ep3_stop_rail;

    rail->Generic.Operations.SdramAlloc       = ep3_sdram_alloc;
    rail->Generic.Operations.SdramFree        = ep3_sdram_free;
    rail->Generic.Operations.SdramWriteb      = ep3_sdram_writeb;

    rail->Generic.Operations.KaddrMap         = ep3_kaddr_map;
    rail->Generic.Operations.SdramMap         = ep3_sdram_map;
    rail->Generic.Operations.Unmap            = ep3_unmap;

    rail->Generic.Operations.DvmaReserve      = ep3_dvma_reserve;
    rail->Generic.Operations.DvmaRelease      = ep3_dvma_release;
    rail->Generic.Operations.DvmaSetPte       = ep3_dvma_set_pte;
    rail->Generic.Operations.DvmaReadPte      = ep3_dvma_read_pte;
    rail->Generic.Operations.DvmaUnload       = ep3_dvma_unload;
    rail->Generic.Operations.FlushTlb         = ep3_flush_tlb;

    rail->Generic.Operations.ProbeRoute       = ep3_probe_route;
    rail->Generic.Operations.PositionFound    = ep3_position_found;
    rail->Generic.Operations.CheckPosition    = ep3_check_position;
    rail->Generic.Operations.NeterrFixup      = ep3_neterr_fixup;

    rail->Generic.Operations.LoadSystemRoute  = ep3_load_system_route;

    rail->Generic.Operations.LoadNodeRoute    = ep3_load_node_route;
    rail->Generic.Operations.UnloadNodeRoute  = ep3_unload_node_route;
    rail->Generic.Operations.LowerFilter      = ep3_lower_filter;
    rail->Generic.Operations.RaiseFilter      = ep3_raise_filter;
    rail->Generic.Operations.NodeDisconnected = ep3_node_disconnected;

    rail->Generic.Operations.FlushFilters     = ep3_flush_filters;
    rail->Generic.Operations.FlushQueues      = ep3_flush_queues;

    rail->Generic.Operations.AllocInputQ      = ep3_alloc_inputq;
    rail->Generic.Operations.FreeInputQ       = ep3_free_inputq;
    rail->Generic.Operations.EnableInputQ     = ep3_enable_inputq;
    rail->Generic.Operations.DisableInputQ    = ep3_disable_inputq;
    rail->Generic.Operations.PollInputQ       = ep3_poll_inputq;

    rail->Generic.Operations.AllocOutputQ     = ep3_alloc_outputq;
    rail->Generic.Operations.FreeOutputQ      = ep3_free_outputq;
    rail->Generic.Operations.OutputQMsg       = ep3_outputq_msg;
    rail->Generic.Operations.OutputQState     = ep3_outputq_state;
    rail->Generic.Operations.OutputQSend      = ep3_outputq_send;

    rail->Generic.Operations.FillOutStats     = ep3_fillout_stats;

    rail->Generic.Devinfo = dev->Devinfo;

    printk ("%s: connected via elan3 rev%c device %d\n", rail->Generic.Name,
	    'a' + dev->Devinfo.dev_revision_id, dev->Instance);

    return (EP_RAIL *) rail;
}

void
ep3_destroy_rail (EP_RAIL *r)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    
    KMEM_FREE (rail, sizeof (EP3_RAIL));
}

static int
ep3_attach_rail (EP3_RAIL *rail)
{
    ELAN3_DEV        *dev = rail->Device;
    ELAN3_CTXT       *ctxt;
    ELAN_CAPABILITY  *cap;
    int               ctx;
    unsigned long     flags;

    if ((ctxt = elan3_alloc (dev, TRUE)) == (ELAN3_CTXT *) NULL)
    {
	printk ("%s: cannot allocate elan context\n", rail->Generic.Name);
	return -ENXIO;
    }
    
    ctxt->Operations = &ep3_elan3_ops;
    ctxt->Private    = (void *) rail;
    
    /* Initialise a capability and attach to the elan*/
    KMEM_ALLOC (cap, ELAN_CAPABILITY *, sizeof (ELAN_CAPABILITY), TRUE);
    
    elan_nullcap (cap);
    
    cap->cap_type        = ELAN_CAP_TYPE_KERNEL;
    cap->cap_version     = ELAN_CAP_VERSION_NUMBER;
    cap->cap_mycontext   = ELAN3_MRF_CONTEXT_NUM | SYS_CONTEXT_BIT;
    cap->cap_lowcontext  = ELAN3_MRF_CONTEXT_NUM | SYS_CONTEXT_BIT;
    cap->cap_highcontext = ELAN3_MRF_CONTEXT_NUM | SYS_CONTEXT_BIT;
    cap->cap_railmask    = 1 << dev->Devinfo.dev_rail;
    
    /* Ensure the context filter is raised while we initialise */
    elan3_block_inputter (ctxt, TRUE);

    if (elan3_doattach (ctxt, cap) != 0)
    {
	printk ("%s: cannot attach to kernel context\n", rail->Generic.Name);

	KMEM_FREE (cap, sizeof (ELAN_CAPABILITY));
	elan3_free (ctxt);
	return -ENXIO;
    }
    KMEM_FREE (cap, sizeof (ELAN_CAPABILITY));

    /* now attach to all the kernel comms input/dmaring/data contexts */
    spin_lock_irqsave (&dev->IntrLock, flags);

    for (ctx = ELAN3_DMARING_BASE_CONTEXT_NUM; ctx <= ELAN3_DMARING_TOP_CONTEXT_NUM; ctx++)
    {
	/* place it in the info table.  NOTE: don't call elan3mmu_set_info, as this */
	/* will queue the info again on the devices info list */
	dev->CtxtTable[ctx] = ctxt;
	
	elan3mmu_set_context_filter (dev, ctx|SYS_CONTEXT_BIT, TRUE, 0, NULL);
	elan3mmu_attach (dev, ctx, ctxt->Elan3mmu, ctxt->RouteTable->Table, ctxt->RouteTable->Size-1);
    }

    for (ctx = ELAN3_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN3_KCOMM_TOP_CONTEXT_NUM; ctx++)
    {
	/* place it in the info table.  NOTE: don't call elan3mmu_set_info, as this */
	/* will queue the info again on the devices info list */
	dev->CtxtTable[ctx] = ctxt;
	
	elan3mmu_set_context_filter (dev, ctx|SYS_CONTEXT_BIT, TRUE, 0, NULL);
	elan3mmu_attach (dev, ctx, ctxt->Elan3mmu, ctxt->RouteTable->Table, ctxt->RouteTable->Size-1);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    /* Stash the ctxt,commandport, mmu and route table */
    rail->Ctxt        = ctxt;
    rail->CommandPort = ctxt->CommandPort;
    rail->Elan3mmu    = ctxt->Elan3mmu;
    rail->RouteTable  = ctxt->RouteTable;

    return 0;
}

static void
ep3_detach_rail (EP3_RAIL *rail)
{
    ELAN3_DEV *dev = rail->Device;
    unsigned long flags;
    int ctx;

    /* detach from the elan */
    spin_lock_irqsave (&dev->IntrLock, flags);

    for (ctx = ELAN3_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN3_KCOMM_TOP_CONTEXT_NUM; ctx++)
    {
	dev->CtxtTable[ctx] = NULL;
	elan3mmu_detach (dev, ctx);
    }

    for (ctx = ELAN3_DMARING_BASE_CONTEXT_NUM; ctx <= ELAN3_DMARING_TOP_CONTEXT_NUM; ctx++)
    {
	dev->CtxtTable[ctx] = NULL;
	elan3mmu_detach (dev, ctx);
    }
    spin_unlock_irqrestore (&dev->IntrLock, flags);

    elan3_dodetach(rail->Ctxt);
    elan3_free (rail->Ctxt);

    rail->Ctxt        = NULL;
    rail->CommandPort = 0;
    rail->Elan3mmu    = NULL;
    rail->RouteTable  = NULL;
}

int
ep3_start_rail (EP_RAIL *r)
{
    EP3_RAIL     *rail = (EP3_RAIL *) r;
    int           i, res;
    unsigned long flags;

    if ((res = ep3_attach_rail (rail)) != 0)
	return res;

    spin_lock_init (&rail->CookieLock);
    kmutex_init (&rail->HaltOpMutex);
    kcondvar_init (&rail->HaltOpSleep);

    /* Initialise event interrupt cookie table */
    InitialiseCookieTable (&rail->CookieTable);

    /* Load and map the thread code */
    rail->ThreadCode = threadcode_elan3;
    if (ep_loadcode (&rail->Generic, &rail->ThreadCode) != ESUCCESS)
	goto failed;

    /* Map the command port to be visible to the Elan */
    ep3_ioaddr_map (&rail->Generic, EP3_COMMANDPORT_ADDR, rail->Ctxt->CommandPage, PAGESIZE, EP_PERM_WRITE);
    rail->CommandPortAddr = EP3_COMMANDPORT_ADDR + (rail->Ctxt->CommandPort - rail->Ctxt->CommandPage);

    /* Allocate the elan visible sdram/main memory */
    if ((rail->RailElan = ep_alloc_elan (&rail->Generic, sizeof (EP3_RAIL_ELAN), 0, &rail->RailElanAddr)) == 0 ||
	(rail->RailMain = ep_alloc_main (&rail->Generic, sizeof (EP3_RAIL_MAIN), 0, &rail->RailMainAddr)) == 0)
    {
	goto failed;
    }

    /* Allocate the system input queues at their fixed elan address */
    if (! (rail->QueueDescs = ep_alloc_memory_elan (&rail->Generic, EP_SYSTEM_QUEUE_BASE, PAGESIZE, EP_PERM_ALL, 0)))
	goto failed;

    /* Initialise all queue entries to be full */
    for (i = 0; i < EP_NUM_SYSTEMQ; i++)
	elan3_sdram_writel (rail->Device, EP_SYSTEMQ_DESC(rail->QueueDescs, i) + offsetof (EP3_InputQueue, q_state), E3_QUEUE_FULL);

    /* initialise the dma rings */
    if (DmaRingsCreate (rail))
	goto failed;
    
    if (InitialiseDmaRetries (rail))
	goto failed;

    if (ep3_init_probenetwork (rail))
	goto failed;

    /* can now drop the context filter for the system context */
    spin_lock_irqsave (&rail->Device->IntrLock, flags);
    elan3mmu_set_context_filter (rail->Device, ELAN3_MRF_CONTEXT_NUM|SYS_CONTEXT_BIT, FALSE, 0, NULL);
    spin_unlock_irqrestore (&rail->Device->IntrLock, flags);

    return 0;

 failed:
    printk ("ep3_start_rail: failed for rail %d\n", rail->Generic.Number);
    ep3_stop_rail (&rail->Generic);

    return -ENOMEM;
}

void
ep3_stall_rail (EP_RAIL *r)
{
    EP3_RAIL     *rail = (EP3_RAIL *) r;
    int           ctx;
    unsigned long flags;

    /* raise all the context filters */
    spin_lock_irqsave (&rail->Device->IntrLock, flags);

    for (ctx = ELAN3_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN3_KCOMM_TOP_CONTEXT_NUM; ctx++)
	elan3mmu_set_context_filter (rail->Device, ctx|SYS_CONTEXT_BIT, TRUE, 0, NULL);

    for (ctx = ELAN3_DMARING_BASE_CONTEXT_NUM; ctx <= ELAN3_DMARING_TOP_CONTEXT_NUM; ctx++)
	elan3mmu_set_context_filter (rail->Device, ctx|SYS_CONTEXT_BIT, TRUE, 0, NULL);

    elan3mmu_set_context_filter (rail->Device, ELAN3_MRF_CONTEXT_NUM|SYS_CONTEXT_BIT, TRUE, 0, NULL);

    spin_unlock_irqrestore (&rail->Device->IntrLock, flags);
}

void
ep3_stop_rail (EP_RAIL *r)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;

    ep3_destroy_probenetwork (rail);

    if (rail->DmaRetryInitialised)
	DestroyDmaRetries (rail);

    DmaRingsRelease(rail);

    if (rail->Generic.State == EP_RAIL_STATE_RUNNING)
    {
	KMEM_FREE (rail->MainCookies, rail->Generic.Position.pos_nodes * sizeof (E3_uint32));

	ep_free_elan (&rail->Generic, rail->ElanCookies, rail->Generic.Position.pos_nodes * sizeof (E3_uint32));
    }

    if (rail->QueueDescs)
	ep_free_memory_elan (&rail->Generic, EP_SYSTEM_QUEUE_BASE);
    rail->QueueDescs = 0;

    if (rail->RailMain)
	ep_free_main (&rail->Generic, rail->RailMainAddr, sizeof (EP3_RAIL_MAIN));
    rail->RailMain = 0;

    if (rail->RailElan)
	ep_free_elan (&rail->Generic, rail->RailElanAddr, sizeof (EP3_RAIL_ELAN));
    rail->RailElan = 0;

    ep_unloadcode (&rail->Generic, &rail->ThreadCode);

    DestroyCookieTable (&rail->CookieTable);

    ep_perrail_unmap (&rail->Generic, rail->Ctxt->CommandPage, PAGESIZE);

    kcondvar_destroy (&rail->HaltOpSleep);
    kmutex_destroy (&rail->HaltOpMutex);
    spin_lock_destroy (&rail->CookieLock);

    ep3_detach_rail (rail);
}

void
ep3_position_found (EP_RAIL *r, ELAN_POSITION *pos)
{
    EP3_RAIL   *rail = (EP3_RAIL *) r;
    sdramaddr_t addr;

    rail->SwitchBroadcastLevelTick = lbolt;

    elan3_sdram_writel (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, NodeId), pos->pos_nodeid);

    /* Allocate Network Identify cookie state */
    KMEM_ZALLOC (rail->MainCookies, E3_uint32 *, pos->pos_nodes * sizeof (E3_uint32), 1);
    
    if (! (addr = ep_alloc_elan (&rail->Generic, pos->pos_nodes * sizeof (E3_uint32), 0, &rail->ElanCookies)))
	panic ("ep: PositionFound: cannot allocate elan cookies array\n");

    elan3_sdram_zeroq_sdram (rail->Device, addr, pos->pos_nodes * sizeof (E3_uint32));

    ep3_probe_position_found (rail, pos);
}

sdramaddr_t
ep3_sdram_alloc (EP_RAIL *r, EP_ADDR addr, unsigned size)
{
    return elan3_sdram_alloc (((EP3_RAIL *) r)->Device, size);
}

void
ep3_sdram_free (EP_RAIL *r, sdramaddr_t addr, unsigned size)
{
    elan3_sdram_free (((EP3_RAIL *) r)->Device, addr, size);
}

void
ep3_sdram_writeb (EP_RAIL *r, sdramaddr_t addr, unsigned char val)
{
    elan3_sdram_writeb (((EP3_RAIL *) r)->Device, addr, val);
}

void
ep3_flush_tlb (EP_RAIL *r)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    ELAN3_DEV *dev = rail->Device;
    unsigned long flags;

    spin_lock_irqsave (&dev->TlbLock, flags);
    
    IncrStat (dev, TlbFlushes);
    
    write_reg32 (dev, Cache_Control_Reg.ContReg, dev->Cache_Control_Reg | MMU_FLUSH);
    mmiob ();
    spin_unlock_irqrestore (&dev->TlbLock, flags);

    while (! (read_reg32 (dev, Cache_Control_Reg.ContReg) & MMU_FLUSHED))
	mb();
}

void
ep3_load_system_route (EP_RAIL *r, unsigned vp, unsigned lowNode, unsigned highNode)
{
    EP3_RAIL  *rail = (EP3_RAIL *) r;
    E3_uint16  flits[MAX_FLITS];
    int        nflits;
    
    nflits = GenerateRoute (&rail->Generic.Position, flits, lowNode, highNode, DEFAULT_ROUTE_TIMEOUT, HIGH_ROUTE_PRIORITY);
	
    if (LoadRoute (rail->Device, rail->RouteTable, vp, ELAN3_MRF_CONTEXT_NUM|SYS_CONTEXT_BIT, nflits, flits) != 0)
    {
	/* XXXX: whilst LoadRoute() can fail - it is not likely. */
	panic ("ep3_load_system_route: cannot load p2p route entry\n");
    }
}

void
ep3_load_node_route (EP_RAIL *r, unsigned nodeId)
{
    EP3_RAIL     *rail = (EP3_RAIL *) r;
    E3_uint16     flits[MAX_FLITS];
    int           nflits;

    nflits = GenerateRoute (&rail->Generic.Position, flits, nodeId, nodeId, DEFAULT_ROUTE_TIMEOUT, DEFAULT_ROUTE_PRIORITY);

    if (LoadRoute (rail->Device, rail->RouteTable, EP_VP_DATA(nodeId), EP3_CONTEXT_NUM(rail->Generic.Position.pos_nodeid), nflits, flits) != 0)
	panic ("ep3_load_node_route: cannot load p2p data route entry\n");
}

void
ep3_unload_node_route (EP_RAIL *r, unsigned nodeId)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;

    ClearRoute (rail->Device, rail->RouteTable, EP_VP_DATA(nodeId));
}

void
ep3_lower_filter (EP_RAIL *r, unsigned nodeId)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    unsigned long flags;

    spin_lock_irqsave (&rail->Device->IntrLock, flags);
    elan3mmu_set_context_filter (rail->Device, EP3_CONTEXT_NUM(nodeId), 0, 0, NULL);
    spin_unlock_irqrestore (&rail->Device->IntrLock, flags);
}

void
ep3_raise_filter (EP_RAIL *r, unsigned nodeId)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    unsigned long flags;

    spin_lock_irqsave (&rail->Device->IntrLock, flags);
    elan3mmu_set_context_filter (rail->Device, EP3_CONTEXT_NUM(nodeId), 1, 0, NULL);
    spin_unlock_irqrestore (&rail->Device->IntrLock, flags);
}

void
ep3_node_disconnected (EP_RAIL *r, unsigned nodeId)
{
    FreeStalledDmas ((EP3_RAIL *) r, nodeId);
}

void
ep3_fillout_stats(EP_RAIL *r, char *str) 
{
    /* no stats here yet */
    /* EP3_RAIL *ep3rail = (EP3_RAIL *)r; */
}
