/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __EP_KCOMM_ELAN3_H
#define __EP_KCOMM_ELAN3_H

#ident "@(#)$Id: kcomm_elan3.h,v 1.53 2004/12/14 10:19:23 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm_elan3.h,v $*/

#if !defined(__ELAN3__)
#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>
#endif /* !defined(__ELAN3__) */

#include <elan3/trtype.h>

/* private address allocation */
#define EP3_TEXT_BASE			0xFF000000		/* base address for thread code (defined in makerules.elan3) */
#define EP3_COMMANDPORT_ADDR		0xFFF00000		/* mapping address for elan command port */

#define EP3_STACK_SIZE			1024			/* default thread code stack size */

#define EP3_PACEMAKER_EVENTADDR		0xfeedbeef		/* mis-aligned address used by heartbeat pacemaker */

/* context number allocation */
#define EP3_CONTEXT_NUM(nodeId)		((ELAN3_KCOMM_BASE_CONTEXT_NUM + (nodeId)) | SYS_CONTEXT_BIT)
#define EP3_CONTEXT_ISDATA(ctx)		(((ctx) & MAX_ROOT_CONTEXT_MASK) >= ELAN3_KCOMM_BASE_CONTEXT_NUM && \
					 ((ctx) & MAX_ROOT_CONTEXT_MASK) <= ELAN3_KCOMM_TOP_CONTEXT_NUM)
#define EP3_CONTEXT_TO_NODE(ctx)	(((ctx) & MAX_ROOT_CONTEXT_MASK) - ELAN3_KCOMM_BASE_CONTEXT_NUM)

/* DMA issueing rings */
#define EP3_RING_CRITICAL		0
#define EP3_RING_CRITICAL_LEN		128
#define EP3_RING_HIGH_PRI		1
#define EP3_RING_HIGH_PRI_LEN		64
#define EP3_RING_LOW_PRI		2
#define EP3_RING_LOW_PRI_LEN		32
#define EP3_NUM_RINGS			3

/* Value to "return" from c_close() when envelope handled  by the trap handler */
#define EP3_PAckStolen			4

/* unimplemented instruction trap types for thread code */
#define EP3_UNIMP_TRAP_NO_DESCS		0
#define EP3_UNIMP_TRAP_PACKET_NACKED	1
#define EP3_UNIMP_THREAD_HALTED		2
#define EP3_NUM_UNIMP_TRAPS		3

/* forward declarations */
typedef struct ep3_rail	EP3_RAIL;

/* block copy elan3 inputter queue - with waitvent0 */
typedef struct ep3_inputqueue
{
    volatile E3_uint32	q_state;	/* queue is full=bit0, queue is locked=bit8 */
    volatile E3_Addr	q_bptr;		/* block aligned ptr to current back item */
    E3_uint32		q_size;		/* size of queue item; 0x1 <= size <= (0x40 * 5) */
    E3_Addr		q_top;		/* block aligned ptr to last queue item */
    E3_Addr		q_base;		/* block aligned ptr to first queue item */
    volatile E3_Addr	q_fptr;		/* block aligned ptr to current front item */
    E3_BlockCopyEvent	q_event;	/* queue block copy event */
    E3_uint32		q_pad[4];	/* pad to 64 bytes */
    E3_Addr		q_wevent;	/* WaitEvent0 struct */
    E3_int32		q_wcount;
} EP3_InputQueue;


#if !defined(__ELAN3__)

/* dma retries types and retry times */
typedef struct ep3_retry_dma
{
    struct list_head    Link;					/* chained on free/retry list */
    long		RetryTime;				/* "lbolt" to retry at */
    E3_DMA_BE		Dma;					/* DMA (in main memory) */
} EP3_RETRY_DMA;

typedef struct ep3_dma_ring
{
    sdramaddr_t		pEvent;  
    E3_Addr		epEvent;
    
    sdramaddr_t		pDma;     
    E3_Addr		epDma; 
    
    E3_uint32	       *pDoneBlk; 
    E3_Addr		epDoneBlk; 
    
    int			Entries;				/* number of slots in array  */
    int			Position;				/* current position in array */

    ioaddr_t            CommandPort;
    ioaddr_t		CommandPage;
    DeviceMappingHandle CommandPageHandle;
} EP3_DMA_RING;

#define DMA_RING_EVENT(ring,n)		((ring)->pEvent + (n)*sizeof (E3_BlockCopyEvent))
#define DMA_RING_EVENT_ELAN(ring,n)	((ring)->epEvent + (n)*sizeof (E3_BlockCopyEvent))

#define DMA_RING_DMA(ring,n)		((ring)->pDma   + (n)*sizeof (E3_DMA))
#define DMA_RING_DMA_ELAN(ring,n)	((ring)->epDma   + (n)*sizeof (E3_DMA))

#define DMA_RING_DONE_ELAN(ring,n)	((ring)->epDoneBlk + (n)*sizeof (E3_uint32))

/* Event interrupt cookie operations and lookup table */
typedef struct ep3_cookie_ops
{
    void	(*Event)       (EP3_RAIL *rail, void *arg);				/* called from the interrupt handler when an event is "set" */
    void	(*DmaRetry)    (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int error);	/* called from the interrupt handler when a DMA is "nacked" */
    void	(*DmaCancelled)(EP3_RAIL *rail, void *arg, E3_DMA_BE *dma); 		/* called from the interrupt handler/flush disconnecting when cancelled. */
    void	(*DmaVerify)   (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma); 		/* called from multiple places, to check dma is consistent with state. */
} EP3_COOKIE_OPS;

typedef struct ep3_cookie
{
    struct ep3_cookie  *Next;					/* Cookies are chained in hash table. */
    E3_uint32		Cookie;					/* Cooke store in ev_Type */
    EP3_COOKIE_OPS     *Operations;				/* Cookie operations */
    void	       *Arg;					/* Users arguement. */
} EP3_COOKIE;

#define EP3_COOKIE_HASH_SIZE		(256)
#define EP3_HASH_COOKIE(a)		((((a) >> 3) ^ ((a) >> 7) ^ ((a) >> 11)) & (EP3_COOKIE_HASH_SIZE-1))

typedef struct ep3_cookie_table
{
    spinlock_t		Lock;
    EP3_COOKIE	       *Entries[EP3_COOKIE_HASH_SIZE];
} EP3_COOKIE_TABLE;

#endif /* !defined(__ELAN3__) */

#define EP3_EVENT_FREE			((1 << 4) | EV_WCOPY)
#define EP3_EVENT_ACTIVE		((2 << 4) | EV_WCOPY)
/* DONE == Cookie */
#define EP3_EVENT_FAILED		((3 << 4) | EV_WCOPY)
#define EP3_EVENT_PRIVATE		((4 << 4) | EV_WCOPY)

/* The event cookie can get posted (and seen) before the write has */
/* hit main memory - in this case the event count is <= 0 and the block */
/* will be marked as ACTIVE - but could transition to DONE at any time */
/* Also for a word copy event, the value written into the "done" word */
/* can be the event interrupt cookie rather than the "source" value */
/* this happens since the uCode does not wait for the write to have */
/* occured before overwriting TMP_0 with the cookie */
#define EP3_EVENT_FIRING(edev, event, cookie, done) \
	(((((done) & ~(EV_TYPE_BCOPY | EV_TYPE_MASK_EVIRQ)) == (cookie).Cookie) || (done) == EP3_EVENT_ACTIVE) && \
	 (int) elan3_sdram_readl (edev, (event) + offsetof (E3_BlockCopyEvent, ev_Count)) <= 0)
#define EP3_EVENT_FIRED(cookie, done) \
	(((done) & ~(EV_TYPE_BCOPY | EV_TYPE_MASK_EVIRQ)) == (cookie).Cookie)


/* Time limit to wait while event is firing and block write has not occured */
#define EP3_EVENT_FIRING_TLIMIT		16384			/* 1023 uS */

#define EP3_INIT_COPY_EVENT(event, cookie, dest, intr)							\
{													\
	(event).ev_Count  = 0;										\
	(event).ev_Type   = (intr) ? EV_TYPE_BCOPY | EV_TYPE_EVIRQ | (cookie).Cookie : EV_TYPE_BCOPY;	\
	(event).ev_Source = (cookie).Cookie | EV_WCOPY;							\
	(event).ev_Dest   = (dest) | EV_TYPE_BCOPY_WORD;						\
}

#if !defined(__ELAN3__)

/* Generic input queues which can be polled */
typedef struct ep3_inputq
{
    EP3_COOKIE		q_cookie;
    unsigned int	q_slotSize;
    unsigned int	q_slotCount;

    void	       *q_slots;
    EP_ADDR		q_slotsAddr;

    EP_INPUTQ_CALLBACK *q_callback;
    void	       *q_arg;

    sdramaddr_t		q_desc;
    E3_Addr		q_descAddr;

    E3_Addr		q_base;
    E3_Addr		q_top;
    E3_Addr		q_fptr;

    E3_uint32		q_waitCount;
} EP3_INPUTQ;

typedef struct ep3_outputq
{
    EP3_COOKIE		q_cookie;

    unsigned int	q_slotCount;				/* # slots allocated */
    unsigned int	q_slotSize;				/* size of each slot (rounded up) */

    sdramaddr_t		q_elan;
    E3_Addr		q_elanAddr;

    void	       *q_main;
    E3_Addr		q_mainAddr;
} EP3_OUTPUTQ;

#endif /* !defined(__ELAN3__) */

/* per-rail elan memory portion of device */
typedef struct ep3_rail_elan
{
    E3_uint16		 ProbeSource0[TR_TRACEROUTE_ENTRIES];	/* 32 byte aligned */
    E3_uint16		 ProbeSource1[TR_TRACEROUTE_ENTRIES];

    E3_BlockCopyEvent	 ProbeDone;				/* 16 byte aligned */
    E3_Event		 ProbeStart;				/* 8 byte aligned */
    
    E3_uint32		 ProbeType;				/* 4 byte aligned */
    E3_uint32		 ProbeLevel;

    E3_uint32		 NodeId;
} EP3_RAIL_ELAN;

/* values for ProbeType */
#define PROBE_SINGLE			0
#define PROBE_MULTIPLE			1
/* number of attempts for each type */
#define PROBE_SINGLE_ATTEMPTS		10
#define PROBE_SINGLE_TIMEOUTS		5
#define PROBE_MULTIPLE_ATTEMPTS		20
#define PROBE_MULTIPLE_TIMEOUTS		10

/* per-rail elan memory portsion of device */
typedef struct ep3_rail_main
{
    E3_uint16		ProbeDest0[TR_TRACEROUTE_ENTRIES];	/* 32 byte aligned */
    E3_uint16		ProbeDest1[TR_TRACEROUTE_ENTRIES];
    
    E3_uint32		ProbeDone;				/* 4 byte aligned */
    E3_uint32		ProbeResult;
    E3_uint32		ProbeLevel;
} EP3_RAIL_MAIN;

#if !defined(__ELAN3__)

struct ep3_rail
{
    EP_RAIL		Generic;				/* Generic rail */

    ELAN3_DEV          *Device;					/* Elan device we're using */
    ELAN3_CTXT	       *Ctxt;					/* Elan context struct */
    ioaddr_t            CommandPort;				/* commandport from context */
    E3_Addr		CommandPortAddr;			/*  and address mapped into elan */

    ELAN3_ROUTE_TABLE  *RouteTable;				/* routetable from context */
    ELAN3MMU	       *Elan3mmu;				/* elanmmu from context */

    EP3_COOKIE_TABLE    CookieTable;				/* Event cookie table */

    EP_CODE		ThreadCode;				/* copy of thread code */
    unsigned int        CommandPortEventTrap;			/* flag to indicate command port eventint queue overflow trap */

    sdramaddr_t         RailElan;				/* Elan visible main/sdram portions of */
    E3_Addr             RailElanAddr;				/* device structure */
    EP3_RAIL_MAIN      *RailMain;
    E3_Addr	        RailMainAddr;

    /* small system message queues */
    sdramaddr_t	        QueueDescs;				/* Input Queue descriptors */

    /* Network position prober */
    E3_Addr	        ProbeStack;				/* Network position thread command structure */
    EP3_COOKIE	        ProbeCookie;				/*   event cookie for Done event */
    kcondvar_t	        ProbeWait;				/*   place to wait on probe thread */
    spinlock_t	        ProbeLock;				/*     and lock */
    volatile int        ProbeDone;				/*     and flag to indicate it's done */

    E3_uint16	        ProbeDest0[TR_TRACEROUTE_ENTRIES];	/* last result of CheckNetworkPosition */
    E3_uint16	        ProbeDest1[TR_TRACEROUTE_ENTRIES];
    E3_uint32	        ProbeResult;

    long	        ProbeLevelTick[ELAN_MAX_LEVELS];
    long	        SwitchBroadcastLevelTick;

    /* rings for issueing dmas */
    EP3_DMA_RING        DmaRings[EP3_NUM_RINGS];

    /* retry lists for dmas */
    struct list_head    DmaRetries[EP_NUM_RETRIES];		/* Dma retry lists */
    struct list_head    DmaRetryFreeList;			/*   and free list */
    u_int	        DmaRetryCount;				/*   and total retry count */
    u_int	        DmaRetryReserved;			/*   and number reserved */
    u_int	        DmaRetryThreadShouldStall;		/*   count of reasons to stall retries */
    u_int	        DmaRetryThreadStarted:1;		/*   dma retry thread running */
    u_int	        DmaRetryThreadShouldStop:1;		/*     but should stop */
    u_int	        DmaRetryThreadStopped:1;		/*     and now it's stopped */
    u_int		DmaRetryInitialised:1;			/* have initialise dma retries */

    spinlock_t	        DmaRetryLock;				/*   spinlock protecting lists */
    kcondvar_t	        DmaRetryWait;				/*   place retry thread sleeps */
    long	        DmaRetryTime;				/*   and when it will next wakeup */
    unsigned int        DmaRetrySleeping;			/*   and it's sleeping there */

    /* Network Identify Cookies */
    E3_uint32	       *MainCookies;				/* One cookie allocator per-node for main*/
    E3_Addr		ElanCookies;				/*   and one for elan */
    spinlock_t		CookieLock;				/* spinlock to protect main cookies */

    /* Halt operation flags for flushing. */
    kmutex_t            HaltOpMutex;				/* serialize access to halt operations */
    unsigned int	HaltOpCompleted;			/* flag to indicate halt operation completed */
    kcondvar_t		HaltOpSleep;				/*   place to wait for it to complete */

    /* Network error state */
    kcondvar_t	        NetworkErrorSleep;			/* place to sleep for network error halt operation */
    u_int	        NetworkErrorFlushed;			/*   and flag to indicate flushed */


    EP3_RAIL_STATS	Stats;					/* statistics */
};

/* support.c */

extern ELAN3_OPS  ep3_elan3_ops;

extern E3_uint32    LocalCookie (EP3_RAIL *rail, unsigned int remoteNode);
extern E3_uint32    RemoteCookie (EP3_RAIL *rail, unsigned int remoteNode);

extern void         InitialiseCookieTable (EP3_COOKIE_TABLE *table);
extern void         DestroyCookieTable (EP3_COOKIE_TABLE *table);
extern void         RegisterCookie (EP3_COOKIE_TABLE *table, EP3_COOKIE *cookie, 
				    E3_Addr event, EP3_COOKIE_OPS *ops, void *arg);
extern void         DeregisterCookie (EP3_COOKIE_TABLE *table, EP3_COOKIE *cookie);
extern EP3_COOKIE   *LookupCookie (EP3_COOKIE_TABLE *table, uint32_t cookie);
extern EP3_COOKIE   *LookupEventCookie (EP3_RAIL *rail, EP3_COOKIE_TABLE *table, E3_Addr);

extern int          DmaRingsCreate (EP3_RAIL *rail);
extern void         DmaRingsRelease (EP3_RAIL *rail);
extern int          IssueDma (EP3_RAIL *rail, E3_DMA_BE *dma, int type, int retryThread);

extern int          IssueWaitevent (EP3_RAIL *rail, E3_Addr value);
extern void         IssueSetevent (EP3_RAIL *rail, E3_Addr value);
extern void         IssueRunThread (EP3_RAIL *rail, E3_Addr value);
extern long         DmaRetryTime (int type);
extern int          InitialiseDmaRetries (EP3_RAIL *rail);
extern void         DestroyDmaRetries (EP3_RAIL *rail);
extern int          ReserveDmaRetries (EP3_RAIL *rail, int count, EP_ATTRIBUTE attr);
extern void         ReleaseDmaRetries (EP3_RAIL *rail, int count);
extern void         StallDmaRetryThread (EP3_RAIL *rail);
extern void         ResumeDmaRetryThread (EP3_RAIL *rail);
extern void         QueueDmaForRetry (EP3_RAIL *rail, E3_DMA_BE *dma, int interval);
extern void         QueueDmaOnStalledList (EP3_RAIL *rail, E3_DMA_BE *dma);
extern void         FreeStalledDmas (EP3_RAIL *rail, unsigned int nodeId);

extern void         SetQueueLocked(EP3_RAIL *rail, sdramaddr_t qaddr);

/* threadcode_elan3.c */
extern E3_Addr    ep3_init_thread (ELAN3_DEV *dev, E3_Addr fn, E3_Addr addr, sdramaddr_t stack,
				   int stackSize, int nargs, ...);

/* probenetwork.c */
extern int        ep3_init_probenetwork (EP3_RAIL *rail);
extern void       ep3_destroy_probenetwork (EP3_RAIL *rail);
extern void       ep3_probe_position_found (EP3_RAIL *rail, ELAN_POSITION *pos);
extern int        ep3_probe_route (EP_RAIL *r, int level, int sw, int nodeid, int *linkup, int *linkdown, int attempts, EP_SWITCH *lsw);
extern int        ep3_check_position (EP_RAIL *rail);

/* neterr_elan3.c */
extern void       ep3_neterr_fixup (EP_RAIL *r, unsigned int nodeId, EP_NETERR_COOKIE *cookies);

/* kcomm_elan3.c */
extern EP_RAIL    *ep3_create_rail (EP_SYS *sys, ELAN3_DEV *dev);
extern void        ep3_destroy_rail (EP_RAIL *rail);

extern int         ep3_start_rail (EP_RAIL *rail);
extern void        ep3_stall_rail (EP_RAIL *rail);
extern void        ep3_stop_rail (EP_RAIL *rail);

extern void	   ep3_position_found (EP_RAIL *rail, ELAN_POSITION *pos);

extern sdramaddr_t ep3_sdram_alloc (EP_RAIL *rail, EP_ADDR addr, unsigned int size);
extern void        ep3_sdram_free (EP_RAIL *rail, sdramaddr_t addr, unsigned int size);
extern void        ep3_sdram_writeb (EP_RAIL *rail, sdramaddr_t addr, unsigned char val);

extern void        ep3_flush_tlb (EP_RAIL *r);
extern void        ep3_load_system_route (EP_RAIL *r, unsigned int vp, unsigned int lowNode, unsigned int highNode);
extern void        ep3_load_node_route (EP_RAIL *r, unsigned int nodeId);
extern void        ep3_unload_node_route (EP_RAIL *r, unsigned int nodeId);
extern void        ep3_lower_filter (EP_RAIL *r, unsigned int nodeId);
extern void        ep3_raise_filter (EP_RAIL *rail, unsigned int nodeId);
extern void        ep3_node_disconnected (EP_RAIL *r, unsigned int nodeId);

extern void        ep3_fillout_stats(EP_RAIL *rail, char *str);

/* kmap_elan3.c */
extern void        ep3_kaddr_map (EP_RAIL *r, EP_ADDR eaddr, virtaddr_t kaddr, unsigned int len, unsigned int perm, int ep_attr);
extern void        ep3_sdram_map (EP_RAIL *r, EP_ADDR eaddr, sdramaddr_t saddr, unsigned int len, unsigned int perm, int ep_attr);
extern void        ep3_ioaddr_map (EP_RAIL *r, EP_ADDR eaddr, ioaddr_t ioaddr, unsigned int len, unsigned int perm);
extern void        ep3_unmap (EP_RAIL *r, EP_ADDR eaddr, unsigned int len);
extern void       *ep3_dvma_reserve (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages);
extern void        ep3_dvma_release (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages, void *private);
extern void        ep3_dvma_set_pte (EP_RAIL *r, void *private, unsigned int index, physaddr_t paddr, unsigned int perm);
extern physaddr_t  ep3_dvma_read_pte (EP_RAIL *r, void *private, unsigned int index);
extern void        ep3_dvma_unload (EP_RAIL *r, void *private, unsigned int index, unsigned int npages);

/* kmsg_elan3.c */
extern EP_INPUTQ  *ep3_alloc_inputq (EP_RAIL *r, unsigned int qnum, unsigned int slotSize, unsigned int slotCount,
				     EP_INPUTQ_CALLBACK *callback, void *arg);
extern void        ep3_free_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern void        ep3_enable_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern void        ep3_disable_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern int         ep3_poll_inputq (EP_RAIL *r, EP_INPUTQ *q, int maxCount, EP_INPUTQ_HANDLER *handler, void *arg);
extern EP_OUTPUTQ *ep3_alloc_outputq (EP_RAIL *r, unsigned int slotSize, unsigned int slotCount);
extern void        ep3_free_outputq (EP_RAIL *r, EP_OUTPUTQ *q);
extern void       *ep3_outputq_msg (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum);
extern int         ep3_outputq_state (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum);
extern int         ep3_outputq_send (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum, unsigned int size,
				     unsigned int nodeId, unsigned int qnum, unsigned int retries);

/* support_elan3.c */
extern void        ep3_flush_filters (EP_RAIL *r);
extern void        ep3_flush_queues (EP_RAIL *r);

#endif /* !defined(__ELAN3__) */

#endif /* __EP_KCOMM_ELAN3_H */
