/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kcomm_elan4.c,v 1.19 2004/11/30 12:02:06 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm_elan4.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "conf_linux.h"

extern EP_CODE threadcode_elan4;

unsigned int
ep4_create_rails (EP_SYS *sys, unsigned int disabled)
{
    unsigned int rmask = 0;
    ELAN4_DEV   *dev;
    EP_RAIL     *rail;
    int          i;

    for (i = 0; i < EP_MAX_RAILS; i++)
    {
	if ((dev = elan4_reference_device (i, ELAN4_STATE_STARTED)) != NULL)
	{
	    if ((rail = ep4_create_rail (sys, dev)) == NULL)
		elan4_dereference_device (dev);
	    else
	    {
		if (disabled & (1 << rail->Number))
		    printk ("%s: auto-start of device disabled by configuration\n", rail->Name);
		else
		    ep_start_rail (rail);
		
		ep_procfs_rail_init(rail);

		rmask |= (1 << rail->Number);
	    }
	}
    }

    if (rmask)
	qsnet_debug_alloc();

    return rmask;
}

EP_RAIL *
ep4_create_rail (EP_SYS *sys, ELAN4_DEV *dev)
{
    EP4_RAIL *rail;
    int res;

    KMEM_ZALLOC (rail, EP4_RAIL *, sizeof (EP4_RAIL), 1);

    if (rail == NULL)
	return (EP_RAIL *) NULL;
    
    if ((res = ep_init_rail (sys, &rail->r_generic)) != 0)
    {
	KMEM_FREE (rail, sizeof (EP4_RAIL));
	return (EP_RAIL *) NULL;
    }

    rail->r_ctxt.ctxt_dev = dev;

    /* install our rail operations */
    rail->r_generic.Operations.DestroyRail      = ep4_destroy_rail;
    rail->r_generic.Operations.StartRail        = ep4_start_rail;
    rail->r_generic.Operations.StallRail        = ep4_stall_rail;
    rail->r_generic.Operations.StopRail         = ep4_stop_rail;    

    rail->r_generic.Operations.SdramAlloc       = ep4_sdram_alloc;
    rail->r_generic.Operations.SdramFree        = ep4_sdram_free;
    rail->r_generic.Operations.SdramWriteb      = ep4_sdram_writeb;

    rail->r_generic.Operations.KaddrMap         = ep4_kaddr_map;
    rail->r_generic.Operations.SdramMap         = ep4_sdram_map;
    rail->r_generic.Operations.Unmap            = ep4_unmap;

    rail->r_generic.Operations.DvmaReserve	= ep4_dvma_reserve;
    rail->r_generic.Operations.DvmaRelease	= ep4_dvma_release;
    rail->r_generic.Operations.DvmaSetPte	= ep4_dvma_set_pte;
    rail->r_generic.Operations.DvmaReadPte	= ep4_dvma_read_pte;
    rail->r_generic.Operations.DvmaUnload	= ep4_dvma_unload;
    rail->r_generic.Operations.FlushTlb	        = ep4_flush_tlb;

    rail->r_generic.Operations.ProbeRoute       = ep4_probe_route;

    rail->r_generic.Operations.PositionFound    = ep4_position_found;
    rail->r_generic.Operations.CheckPosition    = ep4_check_position;
    rail->r_generic.Operations.NeterrFixup      = ep4_neterr_fixup;

    rail->r_generic.Operations.LoadSystemRoute  = ep4_load_system_route;

    rail->r_generic.Operations.LoadNodeRoute    = ep4_load_node_route;
    rail->r_generic.Operations.UnloadNodeRoute  = ep4_unload_node_route;
    rail->r_generic.Operations.LowerFilter	= ep4_lower_filter;
    rail->r_generic.Operations.RaiseFilter	= ep4_raise_filter;
    rail->r_generic.Operations.NodeDisconnected = ep4_node_disconnected;

    rail->r_generic.Operations.FlushFilters     = ep4_flush_filters;
    rail->r_generic.Operations.FlushQueues	= ep4_flush_queues;

    rail->r_generic.Operations.AllocInputQ	= ep4_alloc_inputq;
    rail->r_generic.Operations.FreeInputQ	= ep4_free_inputq;
    rail->r_generic.Operations.EnableInputQ     = ep4_enable_inputq;
    rail->r_generic.Operations.DisableInputQ    = ep4_disable_inputq;
    rail->r_generic.Operations.PollInputQ	= ep4_poll_inputq;

    rail->r_generic.Operations.AllocOutputQ     = ep4_alloc_outputq;
    rail->r_generic.Operations.FreeOutputQ	= ep4_free_outputq;
    rail->r_generic.Operations.OutputQMsg	= ep4_outputq_msg;
    rail->r_generic.Operations.OutputQState     = ep4_outputq_state;
    rail->r_generic.Operations.OutputQSend	= ep4_outputq_send;

    rail->r_generic.Operations.FillOutStats     = ep4_fillout_stats;
    rail->r_generic.Operations.Debug		= ep4_debug_rail;

    rail->r_generic.Devinfo = dev->dev_devinfo;

    printk ("%s: connected via elan4 rev%c device %d\n", rail->r_generic.Name,
	    'a' + dev->dev_devinfo.dev_revision_id, dev->dev_instance);

    return (EP_RAIL *) rail;
}

void
ep4_destroy_rail (EP_RAIL *r)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    elan4_dereference_device (rail->r_ctxt.ctxt_dev);

    KMEM_FREE (rail, sizeof (EP4_RAIL));
}

static int
ep4_attach_rail (EP4_RAIL *r)
{
    EP4_RAIL  *rail = (EP4_RAIL *) r;
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;
    unsigned   ctx;

    if (elan4_insertctxt (dev, &rail->r_ctxt, &ep4_trap_ops) != 0)
	return -ENOMEM;
    
    if ((rail->r_routetable = elan4_alloc_routetable (dev, 4)) == NULL)	/* 512 << 4 == 8192 entries */
    {
	elan4_removectxt (dev, &rail->r_ctxt);
	return -ENOMEM;
    }
    elan4_set_routetable (&rail->r_ctxt, rail->r_routetable);

    /* Attach to the kernel comms nextwork context */
    if (elan4_attach_filter (&rail->r_ctxt, ELAN4_KCOMM_CONTEXT_NUM) < 0)
    {
	elan4_free_routetable (dev, rail->r_routetable);
	elan4_removectxt (dev, &rail->r_ctxt);

	return -EBUSY;
    }

    for (ctx = ELAN4_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN4_KCOMM_TOP_CONTEXT_NUM; ctx++)
	elan4_attach_filter (&rail->r_ctxt, ctx);

    return 0;
}

static void
ep4_detach_rail (EP4_RAIL *rail)
{
    ELAN4_DEV *dev = rail->r_ctxt.ctxt_dev;
    unsigned   ctx;

    elan4_detach_filter (&rail->r_ctxt, ELAN4_KCOMM_CONTEXT_NUM);

    for (ctx = ELAN4_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN4_KCOMM_TOP_CONTEXT_NUM; ctx++)
	elan4_detach_filter (&rail->r_ctxt, ctx);

    if (rail->r_routetable)
    {
	elan4_set_routetable (&rail->r_ctxt, NULL);
	elan4_free_routetable (dev, rail->r_routetable);
    }

    elan4_removectxt (dev, &rail->r_ctxt);
}

int
ep4_start_rail (EP_RAIL *r)
{
    EP4_RAIL     *rail = (EP4_RAIL *) r;
    ELAN4_DEV    *dev  = rail->r_ctxt.ctxt_dev;
    E4_InputQueue qdesc;
    int           i, res;

    if ((res = ep4_attach_rail (rail)) < 0)
	return res;

    /* Initialise main interrupt cookie table */
    spin_lock_init (&rail->r_intcookie_lock);
    for (i = 0; i < EP4_INTCOOKIE_HASH_SIZE; i++)
	INIT_LIST_HEAD (&rail->r_intcookie_hash[i]);

    kmutex_init (&rail->r_haltop_mutex);
    kcondvar_init (&rail->r_haltop_sleep);
    spin_lock_init (&rail->r_haltop_lock);

    spin_lock_init (&rail->r_cookie_lock);

    INIT_LIST_HEAD (&rail->r_ecq_list[EP4_ECQ_EVENT]);
    INIT_LIST_HEAD (&rail->r_ecq_list[EP4_ECQ_ATOMIC]);
    INIT_LIST_HEAD (&rail->r_ecq_list[EP4_ECQ_SINGLE]);
    INIT_LIST_HEAD (&rail->r_ecq_list[EP4_ECQ_MAIN]);
    spin_lock_init (&rail->r_ecq_lock);

    ep_kthread_init (&rail->r_retry_thread);
    INIT_LIST_HEAD (&rail->r_retry_ops);

    INIT_LIST_HEAD (&rail->r_neterr_ops);

    kmutex_init (&rail->r_flush_mutex);
    kcondvar_init (&rail->r_flush_sleep);

    /* Allocate the elan visible sdram/main memory */
    if ((rail->r_elan = ep_alloc_elan (&rail->r_generic, sizeof (EP4_RAIL_ELAN), 0, &rail->r_elan_addr)) == 0 ||
	(rail->r_main = ep_alloc_main (&rail->r_generic, sizeof (EP4_RAIL_MAIN), 0, &rail->r_main_addr)) == 0)
    {
	goto failed;
    }

    for (i = 0; i < EP_NUM_SYSTEMQ; i++)
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_qevents[i].ev_CountAndType), 0);

    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event.ev_CountAndType), E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));

    /* Allocate the system input queues at their fixed elan address */
    /*   avoid sdram address aliasing by allocating the min sdram pagesize */
    if (! (rail->r_queuedescs= ep_alloc_memory_elan (&rail->r_generic, EP_SYSTEM_QUEUE_BASE, SDRAM_PAGE_SIZE, EP_PERM_ALL, 0)))
	goto failed;

    /* Initialise the input queue descriptor as "full" with no event */
    qdesc.q_bptr    = 0;
    qdesc.q_fptr    = 8;
    qdesc.q_control = E4_InputQueueControl(qdesc.q_bptr, qdesc.q_fptr, 8);
    qdesc.q_event   = 0;

    for (i = 0; i < EP_NUM_SYSTEMQ; i++)
	elan4_sdram_copyq_to_sdram (dev, &qdesc, EP_SYSTEMQ_DESC (rail->r_queuedescs, i), sizeof (E4_InputQueue));

    /* Allocate the resource map for command queue mappings */
    if ((rail->r_ecq_rmap = ep_rmallocmap (EP4_ECQ_RMAPSIZE, "r_ecq_rmap", 1)) == NULL)
	goto failed;
    
    ep_rmfree (rail->r_ecq_rmap, EP4_ECQ_TOP - EP4_ECQ_BASE, EP4_ECQ_BASE);

    /* register an interrupt cookie & allocate command queues for command queue flushing */
    rail->r_flush_mcq = ep4_get_ecq (rail, EP4_ECQ_MAIN, 4);
    rail->r_flush_ecq = ep4_get_ecq (rail, EP4_ECQ_EVENT, 1);

    if (rail->r_flush_mcq == NULL || rail->r_flush_ecq == NULL)
	goto failed;

    ep4_register_intcookie (rail, &rail->r_flush_intcookie, rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_flush_event), ep4_flush_interrupt, rail);

    /* startup the retry thread */
    if (kernel_thread_create (ep4_retry_thread, (void *) rail) == 0)
	goto failed;
    ep_kthread_started (&rail->r_retry_thread);
    
    ep4_initialise_dma_retries (rail);

    if ((rail->r_event_ecq = ep4_alloc_ecq (rail, CQ_Size1K)) == NULL)
	goto failed;
    
    rail->r_threadcode = threadcode_elan4;
    if (ep_loadcode (&rail->r_generic, &rail->r_threadcode))
	goto failed;

    elan4_flush_icache (&rail->r_ctxt);

    if (ep4_probe_init (rail))
	goto failed;

    /* can now drop the context filter for the system context */
    elan4_set_filter (&rail->r_ctxt, ELAN4_KCOMM_CONTEXT_NUM, E4_FILTER_HIGH_PRI);

    return 0;

 failed:
    printk ("ep4_start_rail: failed for rail '%s'\n", rail->r_generic.Name);
    ep4_stop_rail (&rail->r_generic);

    return -ENOMEM;
}

void
ep4_stall_rail (EP_RAIL *r)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;
    unsigned  ctx;

    /* Raise all the context filters */
    elan4_set_filter (&rail->r_ctxt, ELAN4_KCOMM_CONTEXT_NUM, E4_FILTER_DISCARD_ALL);

    for (ctx = ELAN4_KCOMM_BASE_CONTEXT_NUM; ctx <= ELAN4_KCOMM_TOP_CONTEXT_NUM; ctx++)
	elan4_set_filter (&rail->r_ctxt, ctx, E4_FILTER_DISCARD_ALL);
}

void
ep4_stop_rail (EP_RAIL *r)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    if (rail->r_generic.State == EP_RAIL_STATE_RUNNING) /* undo ep4_position_found() */
    {
	ELAN_POSITION *pos  = &rail->r_generic.Position;
	EP_ADDR        addr = elan4_sdram_readq (rail->r_ctxt.ctxt_dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_cookies));

	ep_free_elan (&rail->r_generic, addr, pos->pos_nodes * sizeof (E4_uint64));

	KMEM_FREE (rail->r_cookies, pos->pos_nodes * sizeof (E4_uint64));
    }

    ep4_probe_destroy (rail);

    ep_unloadcode (&rail->r_generic, &rail->r_threadcode);

    if (rail->r_event_ecq)
	ep4_free_ecq (rail, rail->r_event_ecq);
    rail->r_event_ecq = NULL;

    ep4_finalise_dma_retries (rail);

    ep_kthread_stop (&rail->r_retry_thread);
    ep_kthread_destroy (&rail->r_retry_thread);

    if (rail->r_flush_intcookie.int_arg)
	ep4_deregister_intcookie (rail, &rail->r_flush_intcookie);
    rail->r_flush_intcookie.int_arg = NULL;

    if (rail->r_flush_mcq)
	ep4_put_ecq (rail, rail->r_flush_mcq, 4);
    rail->r_flush_mcq = NULL;

    if (rail->r_flush_ecq)
	ep4_put_ecq (rail, rail->r_flush_ecq, 1);
    rail->r_flush_ecq = NULL;

    if (rail->r_ecq_rmap)
	ep_rmfreemap (rail->r_ecq_rmap);
    
    if (rail->r_queuedescs)
	ep_free_memory_elan (&rail->r_generic, EP_SYSTEM_QUEUE_BASE);
    rail->r_queuedescs = 0;

    if (rail->r_elan)
	ep_free_elan (&rail->r_generic, rail->r_elan_addr, sizeof (EP4_RAIL_ELAN));
    rail->r_elan = 0;

    if (rail->r_main)
	ep_free_main (&rail->r_generic, rail->r_main_addr, sizeof (EP4_RAIL_MAIN));
    rail->r_main = NULL;

    kcondvar_destroy (&rail->r_flush_sleep);
    kmutex_destroy (&rail->r_flush_mutex);

    spin_lock_destroy (&rail->r_ecq_lock);
    spin_lock_destroy (&rail->r_cookie_lock);

    spin_lock_destroy (&rail->r_haltop_lock);
    kcondvar_destroy(&rail->r_haltop_sleep);
    kmutex_destroy (&rail->r_haltop_mutex);
    spin_lock_destroy (&rail->r_intcookie_lock);

    ep4_detach_rail (rail);
}

void
ep4_position_found (EP_RAIL *r, ELAN_POSITION *pos)
{
    EP4_RAIL   *rail = (EP4_RAIL *) r;
    sdramaddr_t cookies;
    EP_ADDR     addr;
    int         i;

    KMEM_ZALLOC (rail->r_cookies, E4_uint64 *, pos->pos_nodes * sizeof (E4_uint64), 1);

    if (! (cookies = ep_alloc_elan (&rail->r_generic, pos->pos_nodes * sizeof (E4_uint64), 0, &addr)))
	panic ("ep4_position_found: cannot allocate elan cookies array\n");

    for (i = 0; i < pos->pos_nodes; i++)
	elan4_sdram_writeq (rail->r_ctxt.ctxt_dev, cookies + (i * sizeof (E4_uint64)), 0);
    
    for (i = 0; i < pos->pos_nodes; i++)
	rail->r_cookies[i] = 0;

    elan4_sdram_writeq (rail->r_ctxt.ctxt_dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_nodeid), pos->pos_nodeid);
    elan4_sdram_writeq (rail->r_ctxt.ctxt_dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_cookies), addr);

    ep4_probe_position_found (rail, pos);
}

sdramaddr_t
ep4_sdram_alloc (EP_RAIL *r, EP_ADDR addr, unsigned size)
{
    ELAN4_DEV *dev = ((EP4_RAIL *) r)->r_ctxt.ctxt_dev;

    if (size >= SDRAM_PAGE_SIZE)
	return elan4_sdram_alloc (dev, size);
    else
    {
	sdramaddr_t block = elan4_sdram_alloc (dev, SDRAM_PAGE_SIZE);
	sdramaddr_t sdram = block + (addr & (SDRAM_PAGE_SIZE-1));

	/* free of the portion before sdram */
	if (sdram > block)
	    elan4_sdram_free (dev, block, sdram - block);

	/* free of the portion after sdram */
	if ((block + SDRAM_PAGE_SIZE) > (sdram + size))
	    elan4_sdram_free (dev, sdram + size, block + SDRAM_PAGE_SIZE - (sdram + size));

	return sdram;
    }
}

void
ep4_sdram_free (EP_RAIL *r, sdramaddr_t addr, unsigned size)
{
    elan4_sdram_free (((EP4_RAIL *) r)->r_ctxt.ctxt_dev, addr, size);
}

void
ep4_sdram_writeb (EP_RAIL *r, sdramaddr_t addr, unsigned char val)
{
    elan4_sdram_writeb (((EP4_RAIL *) r)->r_ctxt.ctxt_dev, addr, val);
}

void
ep4_flush_tlb (EP_RAIL *r)
{
    elan4mmu_flush_tlb (((EP4_RAIL *) r)->r_ctxt.ctxt_dev);
}

void
ep4_load_system_route (EP_RAIL *r, unsigned vp, unsigned lowNode, unsigned highNode)
{
    EP4_RAIL  *rail = (EP4_RAIL *) r;
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;
    E4_VirtualProcessEntry route;

    if (elan4_generate_route (&rail->r_generic.Position, &route, ELAN4_KCOMM_CONTEXT_NUM, 
			      lowNode, highNode, FIRST_SYSTEM_PACKET | FIRST_HIGH_PRI | FIRST_TIMEOUT(3)) < 0)
    {
	panic ("ep4_load_system_route: generate route failed\n");
	/* NOTREACHED */
    }

    elan4_write_route (dev, rail->r_routetable, vp, &route);
}

void
ep4_load_node_route (EP_RAIL *r, unsigned nodeId)
{
    EP4_RAIL  *rail = (EP4_RAIL *) r;
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;
    E4_VirtualProcessEntry route;

    if (elan4_generate_route (&rail->r_generic.Position, &route, EP4_CONTEXT_NUM(rail->r_generic.Position.pos_nodeid),
			      nodeId, nodeId, FIRST_SYSTEM_PACKET | FIRST_TIMEOUT(3)) < 0)
    {
	panic ("ep4_load_node_route: generate route failed\n");
	/* NOTREACHED */
    }

    elan4_write_route (dev, rail->r_routetable, EP_VP_DATA(nodeId), &route);
}

void
ep4_unload_node_route (EP_RAIL *r, unsigned nodeId)
{
    EP4_RAIL  *rail = (EP4_RAIL *) r;
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;
    
    elan4_invalidate_route (dev, rail->r_routetable, EP_VP_DATA(nodeId));
}

void
ep4_lower_filter (EP_RAIL *r, unsigned nodeId)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    elan4_set_filter (&rail->r_ctxt, EP4_CONTEXT_NUM(nodeId), E4_FILTER_HIGH_PRI);
}

void
ep4_raise_filter (EP_RAIL *r, unsigned nodeId)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    elan4_set_filter (&rail->r_ctxt, EP4_CONTEXT_NUM(nodeId), E4_FILTER_DISCARD_ALL);
}

void
ep4_node_disconnected (EP_RAIL *r, unsigned nodeId)
{
    ep4_free_stalled_dmas ((EP4_RAIL *) r, nodeId);
}

void
ep4_fillout_stats(EP_RAIL *r, char *str) 
{
    /* no stats here yet */
    /* EP4_RAIL *ep4rail = (EP4_RAIL *)r; */
}
