/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __EP_KCOMM_ELAN4_H
#define __EP_KCOMM_ELAN4_H

#ident "@(#)$Id: kcomm_elan4.h,v 1.19.2.1 2005/07/20 12:01:34 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm_elan4.h,v $*/

#include <elan4/types.h>

#include <elan4/dma.h>
#include <elan4/events.h>
#include <elan4/commands.h>

#if !defined(__elan4__)
#include <elan4/device.h>
#endif /* ! defined(__elan4__) */

/* private address allocation */
#define EP4_TEXT_BASE			0xF8000000		/* base address for thread code (defined in makerules.elan4) */
#define EP4_ECQ_BASE			0xFF000000		/* address space for mapping command queues */
#define EP4_ECQ_TOP			0xFF800000

#define EP4_ECQ_RMAPSIZE		128
#define EP4_STACK_SIZE			1024			/* default thread code stack size */
#define EP4_MAX_LEVELS			8			/* same as ELAN_MAX_LEVELS */

/* context number allocation */
#define EP4_CONTEXT_NUM(nodeId)		(ELAN4_KCOMM_BASE_CONTEXT_NUM + (nodeId))
#define EP4_CONTEXT_ISDATA(ctx)		((ctx) >= ELAN4_KCOMM_BASE_CONTEXT_NUM && \
					 (ctx) <= ELAN4_KCOMM_TOP_CONTEXT_NUM)
#define EP4_CONTEXT_TO_NODE(ctx)	((ctx) - ELAN4_KCOMM_BASE_CONTEXT_NUM)

/*
 * network error cookie format:
 *  -------------------------------------------------
 *  | unique cookie value | Remote | DMA | Location |
 *  -------------------------------------------------
 * [63:4] Cookie   - unique cookie number
 * [3]    Thread   - cookie generated by thread code
 * [2]	  Remote   - cookie generated by remote end
 * [1]    STEN     - cookie is for a STEN packet
 * [0]    DMA      - cookie is for a DMA
 */
#define EP4_COOKIE_DMA		(1    << 0)
#define EP4_COOKIE_STEN		(1    << 1)
#define EP4_COOKIE_REMOTE	(1    << 2)
#define EP4_COOKIE_THREAD	(1    << 3)
#define EP4_COOKIE_INC		(1ull << 4)

#define EP4_COOKIE_STRING(val)	(long long)(((val) & ~(EP4_COOKIE_INC-1)) >> 4), \
				((val) & EP4_COOKIE_DMA)    ? ",dma" : "", \
				((val) & EP4_COOKIE_REMOTE) ? ",remote" : "", \
				((val) & EP4_COOKIE_THREAD) ? ",thread" : "", \
				((val) & EP4_COOKIE_STEN)   ? ",sten" : ""
/*
 * Done "word" values 
 */
#define EP4_STATE_FREE		0
#define EP4_STATE_ACTIVE	1
#define EP4_STATE_FINISHED	2
#define EP4_STATE_FAILED	3
#define EP4_STATE_PRIVATE	4

#define EP4_EVENT_FIRING_TLIMIT	16384			/* 1023 uS */

/* forward declarations */
typedef struct ep4_rail	EP4_RAIL;

#if !defined(__elan4__)

typedef struct ep4_intcookie
{
    struct list_head		 int_link;
    E4_uint64			 int_val;
    void		       (*int_callback)(EP4_RAIL *rail, void *arg);
    void			*int_arg;
} EP4_INTCOOKIE;

#define EP4_INTCOOKIE_HASH_SIZE	256
#define EP4_INTCOOKIE_HASH(a)		((((a) >> 3) ^ ((a) >> 7) ^ ((a) >> 11)) & (EP4_INTCOOKIE_HASH_SIZE-1))

typedef struct ep4_ecq
{
    struct list_head	ecq_link;				/* linked on r_ecq_list */
    ELAN4_INTOP		ecq_intop;				/* main interrupt op space */
    ELAN4_CQ	       *ecq_cq;					/* command queue */
    E4_Addr		ecq_addr;				/* address mapped into elan */
    unsigned int	ecq_avail;				/* # dwords still available */

    spinlock_t	        ecq_lock;				/* spinlock for main accesses */
    sdramaddr_t		ecq_event;				/* event for flushing "event" queues */
    EP_ADDR		ecq_event_addr;
    struct ep4_ecq     *ecq_flushcq;				/*  and command port to issue setevent to */
} EP4_ECQ;

#define EP4_ECQ_EVENT		0				/* command queues targetted by multi-blocks events */
#define EP4_ECQ_ATOMIC		1				/* command queues targetted by atomic store operations */
#define EP4_ECQ_SINGLE		2				/* command queues targetted by single word commands from main */
#define EP4_ECQ_MAIN		3				/* command queues targetted by multi word commands from main */
#define EP4_NUM_ECQ		4

#define EP4_ECQ_Size(which)		((which) == EP4_ECQ_EVENT  ? CQ_Size64K : \
					 (which) == EP4_ECQ_ATOMIC ? CQ_Size8K  : \
					 (which) == EP4_ECQ_SINGLE ? CQ_Size1K  : \
					 (which) == EP4_ECQ_MAIN   ? CQ_Size8K  : \
					 CQ_Size1K)

typedef struct ep4_dma_retry
{
    struct list_head    retry_link;				/* chained on free/retry list */
    unsigned long	retry_time;				/* "lbolt" to retry at */
    E4_DMA		retry_dma;				/* DMA (in main memory) */
} EP4_DMA_RETRY;

#define EP4_DMA_RETRY_CQSIZE		CQ_Size8K				/* size of command queue for dma retry */
#define EP4_DMA_RETRY_FLOWCNT		(CQ_Size(EP4_DMA_RETRY_CQSIZE)/72)	/* # of reissued DMA's which can fit in */

typedef struct ep4_inputq
{
    EP4_INTCOOKIE	q_intcookie;
    unsigned int	q_slotSize;
    unsigned int	q_slotCount;

    void	       *q_slots;
    EP_ADDR		q_slotsAddr;
    
    EP_INPUTQ_CALLBACK *q_callback;
    void	       *q_arg;

    sdramaddr_t		q_desc;
    EP_ADDR		q_descAddr;
    EP_ADDR		q_eventAddr;
    EP4_ECQ	       *q_wcq;					/* command queue to issue waitevent to */
    EP4_ECQ	       *q_ecq;					/* command queue targetted by event to generate interrupt */

    EP_ADDR		q_fptr;					/* cached current front pointer */
    EP_ADDR		q_last;					/* elan addr for last queue slot  */

    atomic_t		q_fired;				/* atomic flag that interrupt received */
    unsigned int	q_count;				/* count of slots consumed */
} EP4_INPUTQ;

typedef struct ep4_outputq
{
    spinlock_t		q_lock;
    unsigned int	q_slotCount;
    unsigned int	q_slotSize;
    unsigned int        q_dwords;
    ELAN4_CQ	       *q_cq;
    void	       *q_main;
    EP_ADDR		q_mainAddr;
    unsigned int	q_retries;
} EP4_OUTPUTQ;

#endif /* ! defined(__elan4__) */

typedef struct ep4_check_sten
{
    E4_uint64		c_reset_event_cmd;			/* WRITEDWORD to reset start event */
    E4_uint64 		c_reset_event_value;

    E4_uint64		c_open;					/* OPEN VP_PROBE(lvl) */
    E4_uint64		c_trans_traceroute0;			/* SENDTRANS TR_TRACEROUTE 0s */
    E4_uint64		c_addr_traceroute0;
    E4_uint64		c_data_traceroute0[8];
    E4_uint64		c_trans_traceroute1;			/* SENDTRANS TR_TRACEROUTE 1s */
    E4_uint64		c_addr_traceroute1;
    E4_uint64		c_data_traceroute1[8];
    E4_uint64		c_trans_sendack;			/* SENDTRANS SENDACK */
    E4_uint64		c_addr_sendack;
    
    E4_uint64		c_guard_ok;				/* GUARD OK - write level */
    E4_uint64		c_writedword_ok;
    E4_uint64		c_value_ok;
    
    E4_uint64		c_guard_fail;				/* GUARD FAIL - chain setevent/write fail */
    E4_uint64		c_setevent_fail;
    E4_uint64		c_setevent_nop;
    E4_uint64		c_nop_pad;
} EP4_CHECK_STEN;

#define EP4_CHECK_STEN_NDWORDS	(sizeof (EP4_CHECK_STEN) >> 3)

typedef struct ep4_rail_elan
{
    EP4_CHECK_STEN	r_check_sten[EP4_MAX_LEVELS];
    E4_Event32		r_check_fail;					/* Check failed (== r_check_start[-1]) */
    E4_Event32          r_check_start[EP4_MAX_LEVELS];

    E4_Event32		r_qevents[EP_NUM_SYSTEMQ];
    E4_Event32		r_flush_event;

    E4_uint64		r_nodeid;
#ifdef __elan4__
    E4_uint64	       *r_cookies;
#else
    E4_Addr		r_cookies;
#endif
} EP4_RAIL_ELAN;

#define TRACEROUTE_ENTRIES	16			/* 2 * ELAN_MAX_LEVELS */
#define TRACEROUTE_NDWORDS	(TRACEROUTE_ENTRIES/2)

typedef struct ep4_rail_main
{
    E4_uint32		r_probe_dest0[TRACEROUTE_ENTRIES];
    E4_uint32		r_probe_dest1[TRACEROUTE_ENTRIES];
    E4_uint64		r_probe_result;
    E4_uint64		r_probe_level;

    E4_uint64           r_dma_flowcnt;				/*  count of dma's queued */
} EP4_RAIL_MAIN;

#define EP4_PROBE_ACTIVE	(0xffff)
#define EP4_PROBE_FAILED	(0xfffe)

#if !defined(__elan4__)

typedef struct ep4_retry_ops
{
    struct list_head	op_link;
    unsigned long     (*op_func)(EP4_RAIL *rail, void *arg, unsigned long nextRunTime);
    void	       *op_arg;
} EP4_RETRY_OPS;

typedef struct ep4_neterr_ops
{
    struct list_head	op_link;
    void	      (*op_func) (EP4_RAIL *rail, void *arg, unsigned int nodeId, EP_NETERR_COOKIE *cookies);
    void	       *op_arg;
} EP4_NETERR_OPS;

struct ep4_rail
{
    EP_RAIL		r_generic;
    ELAN4_CTXT		r_ctxt;
    ELAN4_ROUTE_TABLE  *r_routetable;
    
    spinlock_t		r_intcookie_lock;
    struct list_head    r_intcookie_hash[EP4_INTCOOKIE_HASH_SIZE];

    sdramaddr_t		r_elan;
    EP_ADDR		r_elan_addr;
    EP4_RAIL_MAIN      *r_main;
    EP_ADDR		r_main_addr;
    
    EP_CODE		r_threadcode;				/* copy of thread code */

    sdramaddr_t		r_queuedescs;				/* systemq queue descriptors */

    E4_uint64	       *r_cookies;				/* network error cookies */
    spinlock_t          r_cookie_lock;				/*    and spin lock */

    kcondvar_t		r_probe_wait;				/* network position probing */
    spinlock_t		r_probe_lock;
    volatile int	r_probe_done;
    EP4_INTCOOKIE	r_probe_intcookie;
    EP4_ECQ           *r_probe_cq;
    E4_uint32		r_probe_source0[TRACEROUTE_ENTRIES];
    E4_uint32		r_probe_source1[TRACEROUTE_ENTRIES];

    kmutex_t		r_haltop_mutex;				/* halt/flush operations */
    ELAN4_HALTOP	r_haltop;
    ELAN4_DMA_FLUSHOP   r_flushop;
    kcondvar_t		r_haltop_sleep;
    spinlock_t		r_haltop_lock;

    struct list_head    r_ecq_list[EP4_NUM_ECQ];		/* list of statically allocated command queues */
    EP_RMAP	       *r_ecq_rmap;				/* resource map for command queue mappings */
    spinlock_t          r_ecq_lock;				/* spinlock for list/space management */

    kmutex_t		r_flush_mutex;				/* serialize command queue flushing */
    unsigned long	r_flush_count;				/* # setevents issued for flushing */
    EP4_ECQ	       *r_flush_mcq;				/*   and command queue for waitevent */
    EP4_ECQ            *r_flush_ecq;				/*   and command queue for interrupt */
    EP4_INTCOOKIE       r_flush_intcookie;			/*   and interrupt cookie */
    kcondvar_t          r_flush_sleep;				/*   and place to sleep ... */

    EP_KTHREAD		r_retry_thread;				/* retry thread */
    struct list_head    r_retry_ops;				/*  list of retry operations */

    EP4_RETRY_OPS       r_dma_ops;				/* dma retry operations */
    EP4_ECQ 	       *r_dma_ecq;				/*   command queue to reissue DMAs */
    E4_uint64           r_dma_flowcnt;				/*   count of dma's reissued */
    struct list_head    r_dma_retrylist[EP_NUM_RETRIES];	/*   retry lists  */
    struct list_head    r_dma_freelist;				/*   and free list */
    spinlock_t          r_dma_lock;				/*   and spinlock to protect lists */
    unsigned long       r_dma_allocated;			/*   # retries allocated*/
    unsigned long       r_dma_reserved;				/*   # retries reserved */

    EP4_ECQ	       *r_event_ecq;				/* command queue for occasional setevents */

    struct list_head    r_neterr_ops;				/* list of neterr fixup operations */

    ELAN4_IPROC_TRAP    r_iproc_trap;
    ELAN4_TPROC_TRAP    r_tproc_trap;
} ;

#define EP4_CTXT_TO_RAIL(ctxt)	((EP4_RAIL *) (((unsigned long) (ctxt)) - offsetof (EP4_RAIL, r_ctxt)))

#if defined(DEBUG_ASSERT)
#define EP4_ASSERT(rail,EXPR)			EP_ASSERT(&((rail)->r_generic), EXPR)
#define EP4_SDRAM_ASSERT(rail,off,value)	EP4_ASSERT(rail, (sdram_assert ? elan4_sdram_readq ((rail)->r_ctxt.ctxt_dev, (off)) == (value) : 1))
#else
#define EP4_ASSERT(rail,EXPR)
#define EP4_SDRAM_ASSERT(rail,off,value)
#endif

/* kcomm_elan4.c */
extern EP_RAIL    *ep4_create_rail (EP_SYS *sys, ELAN4_DEV *dev);
extern void        ep4_destroy_rail (EP_RAIL *rail);

extern int         ep4_start_rail (EP_RAIL *rail);
extern void        ep4_stall_rail (EP_RAIL *rail);
extern void        ep4_stop_rail (EP_RAIL *rail);

extern void	   ep4_debug_rail (EP_RAIL *rail);

extern void        ep4_position_found (EP_RAIL *rail, ELAN_POSITION *pos);

extern sdramaddr_t ep4_sdram_alloc (EP_RAIL *rail, EP_ADDR addr, unsigned int size);
extern void        ep4_sdram_free (EP_RAIL *rail, sdramaddr_t addr, unsigned int size);
extern void        ep4_sdram_writeb (EP_RAIL *rail, sdramaddr_t addr, unsigned char val);

extern void        ep4_flush_tlb (EP_RAIL *r);
extern void        ep4_load_system_route (EP_RAIL *r, unsigned int vp, unsigned int lowNode, unsigned int highNode);
extern void        ep4_load_node_route (EP_RAIL *r, unsigned int nodeId);
extern void        ep4_unload_node_route (EP_RAIL *r, unsigned int nodeId);
extern void        ep4_lower_filter (EP_RAIL *r, unsigned int nodeId);
extern void        ep4_raise_filter (EP_RAIL *rail, unsigned int nodeId);
extern void        ep4_node_disconnected (EP_RAIL *r, unsigned int nodeId);

/* kmap_elan4.c */
extern void        ep4_kaddr_map (EP_RAIL *r, EP_ADDR eaddr, virtaddr_t kaddr, unsigned int len, unsigned int perm, int ep_attr);
extern void        ep4_sdram_map (EP_RAIL *r, EP_ADDR eaddr, sdramaddr_t saddr, unsigned int len, unsigned int perm, int ep_attr);
extern void        ep4_cport_map (EP_RAIL *r, EP_ADDR eaddr, unsigned long cqaddr, unsigned int len, unsigned int perm);
extern void        ep4_unmap (EP_RAIL *r, EP_ADDR eaddr, unsigned int len);
extern void       *ep4_dvma_reserve (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages);
extern void        ep4_dvma_release (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages, void *private);
extern void        ep4_dvma_set_pte (EP_RAIL *r, void *private, unsigned int index, physaddr_t paddr, unsigned int perm);
extern physaddr_t  ep4_dvma_read_pte (EP_RAIL *r, void *private, unsigned int index);
extern void        ep4_dvma_unload (EP_RAIL *r, void *private, unsigned int index, unsigned int npages);

/* kmsg_elan4.c */
extern EP_INPUTQ  *ep4_alloc_inputq (EP_RAIL *r, unsigned int qnum, unsigned int slotSize, unsigned int slotCount,
				     EP_INPUTQ_CALLBACK *callback, void *arg);
extern void        ep4_free_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern void        ep4_enable_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern void        ep4_disable_inputq (EP_RAIL *r, EP_INPUTQ *q);
extern int         ep4_poll_inputq (EP_RAIL *r, EP_INPUTQ *q, int maxCount, EP_INPUTQ_HANDLER *handler, void *arg);
extern EP_OUTPUTQ *ep4_alloc_outputq (EP_RAIL *r, unsigned int slotSize, unsigned int slotCount);
extern void        ep4_free_outputq (EP_RAIL *r, EP_OUTPUTQ *q);
extern void       *ep4_outputq_msg (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum);
extern int         ep4_outputq_state (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum);
extern int         ep4_outputq_send (EP_RAIL *r, EP_OUTPUTQ *q, unsigned int slotNum, unsigned int size,
				     unsigned int nodeId, unsigned int qnum, unsigned int retries);

/* probenetwork_elan4.c */
extern int         ep4_probe_init (EP4_RAIL *r);
extern void        ep4_probe_destroy (EP4_RAIL *r);
extern void        ep4_probe_position_found (EP4_RAIL *rail, ELAN_POSITION *pos);
extern int         ep4_probe_route (EP_RAIL *r, int level, int sw, int nodeid, int *linkup, int *linkdown, int attempts, EP_SWITCH *lsw);
extern int         ep4_check_position (EP_RAIL *rail);

/* support_elan4.c */
extern ELAN4_TRAP_OPS ep4_trap_ops;
extern void           ep4_register_intcookie (EP4_RAIL *rail, EP4_INTCOOKIE *cp, E4_uint64 cookie, void (*callback)(EP4_RAIL *r, void *arg), void *arg);
extern void           ep4_deregister_intcookie (EP4_RAIL *rail, EP4_INTCOOKIE *cp);
extern EP4_INTCOOKIE *ep4_lookup_intcookie (EP4_RAIL *rail, E4_uint64 cookie);
extern E4_uint64      ep4_neterr_cookie (EP4_RAIL *rail, unsigned int node);

extern void           ep4_flush_filters (EP_RAIL *r);
extern void           ep4_flush_queues (EP_RAIL *r);
extern void	      ep4_write_qdesc (EP4_RAIL *rail, sdramaddr_t qaddr, E4_InputQueue *qdesc);

extern EP4_ECQ       *ep4_alloc_ecq (EP4_RAIL *rail, unsigned int cqsize);
extern void           ep4_free_ecq (EP4_RAIL *rail, EP4_ECQ *ecq);
extern EP4_ECQ	     *ep4_get_ecq (EP4_RAIL *rail, unsigned int which, unsigned int ndwords);
extern void           ep4_put_ecq (EP4_RAIL *rail, EP4_ECQ *ecq, unsigned int ndwords);

extern void           ep4_nop_cmd (EP4_ECQ *ecq, E4_uint64 tag);
extern void           ep4_set_event_cmd (EP4_ECQ *ecq, E4_Addr event);
extern void           ep4_wait_event_cmd (EP4_ECQ *ecq, E4_Addr event, E4_uint64 candt, E4_uint64 param0, E4_uint64 param1);

extern void           ep4_flush_interrupt (EP4_RAIL *rail, void *arg);
extern void           ep4_flush_ecqs (EP4_RAIL *rail);

extern void           ep4_init_thread (EP4_RAIL *rail, E4_ThreadRegs *regs, sdramaddr_t stackTop, 
				       EP_ADDR stackAddr, E4_Addr startpc, int nargs,...);

extern void           ep4_initialise_dma_retries (EP4_RAIL *rail);
extern void           ep4_finalise_dma_retries (EP4_RAIL *rail);
extern int            ep4_reserve_dma_retries (EP4_RAIL *rail, unsigned int count, unsigned int attr);
extern void	      ep4_release_dma_retries(EP4_RAIL *rail, unsigned int count);
extern void           ep4_queue_dma_retry (EP4_RAIL *rail, E4_DMA *dma, int interval);
extern void           ep4_queue_dma_stalled (EP4_RAIL *rail, E4_DMA *dma);
extern void           ep4_free_stalled_dmas (EP4_RAIL *rail, unsigned int nodeId);
extern void           ep4_display_rail (EP4_RAIL *rail);

extern void           ep4_add_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops);
extern void           ep4_remove_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops);
extern void           ep4_retry_thread (EP4_RAIL *rail);

/* neterr_elan4.c */
extern void           ep4_add_neterr_ops (EP4_RAIL *rail, EP4_NETERR_OPS *ops);
extern void           ep4_remove_neterr_ops (EP4_RAIL *rail, EP4_NETERR_OPS *ops);
extern void           ep4_neterr_fixup (EP_RAIL *r, unsigned int nodeId, EP_NETERR_COOKIE *cookies);

/* commands_elan4.c */
extern void           elan4_nop_cmd (ELAN4_CQ *cq, E4_uint64 tag);
extern void           elan4_write_dword_cmd (ELAN4_CQ *cq, E4_Addr addr, E4_uint64 data);
extern void           elan4_add_dword_cmd (ELAN4_CQ *cq, E4_Addr addr, E4_uint64 data);
extern void           elan4_copy64_cmd (ELAN4_CQ *cq, E4_Addr from, E4_Addr to, E4_uint32 datatype);
extern void           elan4_interrupt_cmd (ELAN4_CQ *cq, E4_uint64 cookie);
extern void           elan4_run_thread_cmd (ELAN4_CQ *cq, E4_ThreadRegs *regs);
extern void           elan4_run_dma_cmd (ELAN4_CQ *cq, E4_DMA *dma);
extern void           elan4_set_event_cmd (ELAN4_CQ *cq, E4_Addr event);
extern void           elan4_set_eventn_cmd (ELAN4_CQ *cq, E4_Addr event, E4_uint32 count);
extern void           elan4_wait_event_cmd (ELAN4_CQ *cq, E4_Addr event, E4_uint64 candt, E4_uint64 param0, E4_uint64 param1);
extern void           elan4_open_packet (ELAN4_CQ *cq, E4_uint64 command);
extern void           elan4_guard (ELAN4_CQ *cq, E4_uint64 command);
extern void           elan4_sendtrans0 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr);
extern void           elan4_sendtrans1 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 p0);
extern void           elan4_sendtrans2 (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 p0, E4_uint64 p1);
extern void           elan4_sendtransn (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, ...);
extern void           elan4_sendtransp (ELAN4_CQ *cq, E4_uint16 trtype, E4_uint64 addr, E4_uint64 *ptr);

extern void           ep4_add_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops);
extern void	      ep4_remove_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops);
extern void           ep4_retry_thread (EP4_RAIL *rail);

extern void           ep4_fillout_stats(EP_RAIL *rail, char *str);

#endif /* ! defined(__elan4__) */

#endif /* __EP_KCOMM_ELAN4_H */
