/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kmap_elan3.c,v 1.4 2004/12/14 10:19:23 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kmap_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan3/elanregs.h>
#include <elan3/elandev.h>
#include <elan3/elanvp.h>
#include <elan3/elan3mmu.h>
#include <elan3/elanctxt.h>
#include <elan3/elandebug.h>

#include <elan/kcomm.h>

#include "kcomm_elan3.h"

#if defined(DIGITAL_UNIX)
#  define kernel_map			(first_task->map)
#  define vaddr_to_phys(map, addr)	(pmap_extract (vm_map_pmap ((vm_map_t) map), (unsigned long) addr))
#elif defined(LINUX)
#  define kernel_map			get_kern_mm()
#  define vaddr_to_phys(map, addr)	(kmem_to_phys(addr))
#elif defined(SOLARIS)
#  define kernel_map			&kas
#  define vaddr_to_phys(map,addr)	ptob(hat_getpfnum (((struct as *) map)->a_hat, (caddr_t) addr))
#endif

#define ELAN3_PTES_PER_PAGE		(PAGESIZE/ELAN3_PAGE_SIZE)

#if defined(__LITTLE_ENDIAN__)
#define PERM_ENDIAN	0
#else
#define PERM_ENDIAN	ELAN3_PTE_BIG_ENDIAN
#endif

static unsigned int main_permtable[] = 
{
    ELAN3_PERM_REMOTEALL,		/* EP_PERM_EXECUTE */
    ELAN3_PERM_REMOTEREAD,		/* EP_PERM_READ */
    ELAN3_PERM_REMOTEWRITE,		/* EP_PERM_WRITE */
    ELAN3_PERM_REMOTEWRITE,		/* EP_PERM_ALL */
};

static unsigned int sdram_permtable[] = 
{
    ELAN3_PERM_REMOTEREAD,		/* EP_PERM_EXECUTE */
    ELAN3_PERM_REMOTEREAD,		/* EP_PERM_READ */
    ELAN3_PERM_REMOTEWRITE,		/* EP_PERM_WRITE */
    ELAN3_PERM_REMOTEALL,		/* EP_PERM_ALL */
};

static unsigned int io_permtable[] = 
{
    ELAN3_PERM_LOCAL_READ,		/* EP_PERM_EXECUTE */
    ELAN3_PERM_REMOTEREAD,		/* EP_PERM_READ */
    ELAN3_PERM_REMOTEWRITE,		/* EP_PERM_WRITE */
    ELAN3_PERM_REMOTEWRITE,		/* EP_PERM_ALL */
};

void
ep3_kaddr_map (EP_RAIL *r, EP_ADDR eaddr, virtaddr_t kaddr, unsigned len, unsigned int perm, int ep_attr)
{
    EP3_RAIL    *rail   = (EP3_RAIL *) r;
    unsigned     npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (kaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    for (i = 0; i < npages; i++)
    {
	physaddr_t paddr = vaddr_to_phys (kernel_map, (void *) kaddr);

	for (off = 0; off < PAGESIZE; off += ELAN3_PAGE_SIZE)
	    elan3mmu_pteload (rail->Elan3mmu, PTBL_LEVEL_3, eaddr + off, paddr + off, 
			      main_permtable[perm], PTE_LOAD_LOCK | PTE_LOAD_NOSYNC | ((ep_attr & EP_NO_SLEEP) ? PTE_NO_SLEEP : 0));

	eaddr += PAGESIZE;
	kaddr += PAGESIZE;
    }
}

void
ep3_sdram_map (EP_RAIL *r, EP_ADDR eaddr, sdramaddr_t saddr, unsigned len, unsigned int perm, int ep_attr)
{
    EP3_RAIL    *rail   = (EP3_RAIL *) r;
    unsigned     npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (saddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    for (i = 0; i < npages; i++)
    {
	physaddr_t paddr = elan3_sdram_to_phys (rail->Device, saddr);

	for (off = 0; off < PAGESIZE; off += ELAN3_PAGE_SIZE)
	    elan3mmu_pteload (rail->Elan3mmu, PTBL_LEVEL_3, eaddr+off, paddr+off, 
			      sdram_permtable[perm], PTE_LOAD_LOCK | PTE_LOAD_NOSYNC | ((ep_attr & EP_NO_SLEEP) ? PTE_NO_SLEEP : 0) );

	eaddr += PAGESIZE;
	saddr += PAGESIZE;
    }
}

void
ep3_ioaddr_map (EP_RAIL *r, EP_ADDR eaddr, ioaddr_t ioaddr, unsigned len, unsigned int perm)
{
    EP3_RAIL    *rail   = (EP3_RAIL *) r;
    unsigned     npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (ioaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    for (i = 0; i < npages; i++)
    {
	physaddr_t paddr = vaddr_to_phys (kernel_map, (void *) ioaddr);

	for (off = 0; off < PAGESIZE; off += ELAN3_PAGE_SIZE)
	    elan3mmu_pteload (rail->Elan3mmu, PTBL_LEVEL_3, eaddr + off, paddr + off, 
			      io_permtable[perm], PTE_LOAD_LOCK | PTE_LOAD_NOSYNC);

	eaddr += PAGESIZE;
	ioaddr += PAGESIZE;
    }
}
void
ep3_unmap (EP_RAIL *r, EP_ADDR eaddr, unsigned len)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    elan3mmu_unload (rail->Elan3mmu, eaddr, len, PTE_UNLOAD_UNLOCK | PTE_UNLOAD_NOSYNC);
}

void *
ep3_dvma_reserve (EP_RAIL *r, EP_ADDR eaddr, unsigned npages)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    void     *private;

    KMEM_ALLOC (private, void *, npages * ELAN3_PTES_PER_PAGE * sizeof (sdramaddr_t), 1);
    
    if (private == NULL)
	return NULL;
    
    elan3mmu_reserve (rail->Elan3mmu, eaddr, npages * ELAN3_PTES_PER_PAGE, (sdramaddr_t *) private);

    return private;
}

void
ep3_dvma_release (EP_RAIL *r, EP_ADDR eaddr, unsigned npages, void *private)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;

    elan3mmu_release (rail->Elan3mmu, eaddr, npages * ELAN3_PTES_PER_PAGE, (sdramaddr_t *) private);

    KMEM_FREE (private, npages * ELAN3_PTES_PER_PAGE * sizeof (sdramaddr_t));
}

void
ep3_dvma_set_pte (EP_RAIL *r, void *private, unsigned index, physaddr_t paddr, unsigned int perm)
{
    ELAN3_DEV    *dev  = ((EP3_RAIL *) r)->Device;
    sdramaddr_t *ptep = &((sdramaddr_t *) private)[index * ELAN3_PTES_PER_PAGE];
    int          off;

    for (off =0 ; off < PAGESIZE; off += ELAN3_PAGE_SIZE)
    {	
	ELAN3_PTE newpte = elan3mmu_phys_to_pte (dev, paddr + off, main_permtable[perm]) | ELAN3_PTE_REF | ELAN3_PTE_MOD;

	elan3_writepte (dev, *ptep, newpte);

	ptep++;
    }
}

physaddr_t
ep3_dvma_read_pte (EP_RAIL *r, void *private, unsigned index)
{
    EP3_RAIL    *rail = (EP3_RAIL *) r;
    sdramaddr_t *ptep = &((sdramaddr_t *) private)[index * ELAN3_PTES_PER_PAGE];
    ELAN3_PTE     pte  = elan3_readpte (rail->Device, *ptep);

    return pte & ELAN3_PTE_PFN_MASK;
}

void
ep3_dvma_unload (EP_RAIL *r, void *private, unsigned index, unsigned npages)
{
    EP3_RAIL    *rail = (EP3_RAIL *) r;
    sdramaddr_t *ptep = &((sdramaddr_t *) private)[index * ELAN3_PTES_PER_PAGE];
    ELAN3_PTE     tpte = elan3mmu_kernel_invalid_pte (rail->Elan3mmu);
    int i;

    for (i = (npages * ELAN3_PTES_PER_PAGE) - 1; i >= 0; i--)
	elan3_writepte (rail->Device, ptep[i], tpte);
}
