/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kmap_elan4.c,v 1.12.2.1 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kmap_elan4.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "debug.h"
#include "kcomm_elan4.h"

#if defined(DIGITAL_UNIX)
#  define kernel_map			(first_task->map)
#  define vaddr_to_phys(map, addr)	(pmap_extract (vm_map_pmap ((vm_map_t) map), (unsigned long) addr))
#elif defined(LINUX)
#  define kernel_map			get_kern_mm()
#  define vaddr_to_phys(map, addr)	(kmem_to_phys(addr))
#elif defined(SOLARIS)
#  define kernel_map			&kas
#  define vaddr_to_phys(map,addr)	ptob(hat_getpfnum (((struct as *) map)->a_hat, (caddr_t) addr))
#endif

static unsigned int main_permtable[] = 
{
    PERM_Unused,			/* EP_PERM_EXECUTE */
    PERM_RemoteReadOnly,		/* EP_PERM_READ */
    PERM_DataReadWrite,			/* EP_PERM_WRITE */
    PERM_DataReadWrite,			/* EP_PERM_ALL */
};

static unsigned int sdram_permtable[] = 
{
    PERM_LocExecute,			/* EP_PERM_EXECUTE */
    PERM_RemoteReadOnly,		/* EP_PERM_READ */
    PERM_DataReadWrite,			/* EP_PERM_WRITE */
    PERM_RemoteAll,			/* EP_PERM_ALL */
};

static unsigned int cport_permtable[] = 
{
    PERM_Unused,			/* EP_PERM_EXECUTE */
    PERM_RemoteReadOnly,		/* EP_PERM_READ */
    PERM_DataReadWrite,			/* EP_PERM_WRITE */
    PERM_Unused,			/* EP_PERM_ALL */
};

void
ep4_kaddr_map (EP_RAIL *r, EP_ADDR eaddr, virtaddr_t kaddr, unsigned int len, unsigned int perm, int ep_attr)
{
    EP4_RAIL    *rail   = (EP4_RAIL *) r;
    ELAN4_DEV   *dev    = rail->r_ctxt.ctxt_dev;
    unsigned int npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (kaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    for (i = 0; i < npages; i++)
    {
	physaddr_t paddr = vaddr_to_phys (kernel_map, (void *) kaddr);

	for (off = 0; off < PAGESIZE; off += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = elan4mmu_phys2pte (dev, paddr + off, main_permtable[perm]);

	    elan4mmu_pteload (&rail->r_ctxt, 0, eaddr + off, HE_TYPE_OTHER, newpte);
	}

	eaddr += PAGESIZE;
	kaddr += PAGESIZE;
    }
}

void
ep4_sdram_map (EP_RAIL *r, EP_ADDR eaddr, sdramaddr_t saddr, unsigned int len, unsigned int perm, int ep_attr)
{
    EP4_RAIL    *rail   = (EP4_RAIL *) r;
    ELAN4_DEV   *dev    = rail->r_ctxt.ctxt_dev;
    unsigned int npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (saddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    if ((eaddr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT)) != (saddr & (SDRAM_PGOFF_OFFSET << PAGE_SHIFT)))
	printk ("ep4_sdram_map: eaddr=%x saddr=%lx - incorrectly alised\n", eaddr, saddr);

    for (i = 0; i < npages; i++)
    {
	for (off = 0; off < PAGESIZE; off += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = ((saddr + off) >> PTE_PADDR_SHIFT) | PTE_SetPerm (sdram_permtable[perm]);

	    elan4mmu_pteload (&rail->r_ctxt, 0, eaddr + off, HE_TYPE_OTHER, newpte);
	}

	eaddr += PAGESIZE;
	saddr += PAGESIZE;
    }
}

void
ep4_cport_map (EP_RAIL *r, EP_ADDR eaddr, unsigned long cqaddr, unsigned int len, unsigned int perm)
{
    EP4_RAIL    *rail   = (EP4_RAIL *) r;
    ELAN4_DEV   *dev    = rail->r_ctxt.ctxt_dev;
    unsigned int npages = len >> PAGESHIFT;
    int          i;
    unsigned int off;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (cqaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    for (i = 0; i < npages; i++)
    {
	for (off = 0; off < PAGESIZE; off += (1 << dev->dev_pageshift[0]))
	{
	    E4_uint64 newpte = ((cqaddr + off) >> PTE_PADDR_SHIFT) | PTE_SetPerm(cport_permtable[perm]) | PTE_CommandQueue;

	    elan4mmu_pteload (&rail->r_ctxt, 0, eaddr + off, HE_TYPE_OTHER, newpte);
	}

	eaddr += PAGESIZE;
	cqaddr += PAGESIZE;
    }
}
void
ep4_unmap (EP_RAIL *r, EP_ADDR eaddr, unsigned int len)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    ASSERT ((eaddr & PAGEOFFSET) == 0 && (len & PAGEOFFSET) == 0);

    elan4mmu_unload_range (&rail->r_ctxt, 0, eaddr, len);
}

void *
ep4_dvma_reserve (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages)
{
    EP4_RAIL  *rail = (EP4_RAIL *) r;
    ELAN4_DEV *dev  = rail->r_ctxt.ctxt_dev;

    EPRINTF3 (DBG_KMAP, "ep4_dvma_reserve: eaddr=%x npages=%d (=> %d)\n", eaddr, npages, (npages << (PAGE_SHIFT - dev->dev_pageshift[0])));

    return elan4mmu_reserve (&rail->r_ctxt, 0, (E4_Addr) eaddr, (npages << (PAGE_SHIFT - dev->dev_pageshift[0])), 1);
}

void
ep4_dvma_release (EP_RAIL *r, EP_ADDR eaddr, unsigned int npages, void *private)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;

    EPRINTF3 (DBG_KMAP, "ep4_dvma_release: eaddr=%x npages=%d private=%p\n", eaddr, npages, private);

    elan4mmu_release (&rail->r_ctxt, (ELAN4_HASH_CACHE *) private);
}

void
ep4_dvma_set_pte (EP_RAIL *r, void *private, unsigned int index, physaddr_t paddr, unsigned int perm)
{
    EP4_RAIL     *rail  = (EP4_RAIL *) r;
    ELAN4_DEV    *dev   = rail->r_ctxt.ctxt_dev;
    unsigned int  off;
    unsigned long flags;

    EPRINTF3 (DBG_KMAP, "ep4_dvma_set_pte: index %x -> eaddr %llx paddr %llx\n", 
	      index, (long long)(((ELAN4_HASH_CACHE *) private)->hc_start + (index * PAGE_SIZE)), (long long) paddr);

    local_irq_save (flags);
    for (off = 0; off < PAGESIZE; off += (1 << dev->dev_pageshift[0]))
    {
	E4_uint64 newpte = elan4mmu_phys2pte (dev, paddr + off, main_permtable[perm]);

	elan4mmu_set_pte (&rail->r_ctxt, (ELAN4_HASH_CACHE *) private, (index << (PAGE_SHIFT - dev->dev_pageshift[0])) +
			  (off >> dev->dev_pageshift[0]), newpte);
    }
    local_irq_restore (flags);
}

physaddr_t
ep4_dvma_read_pte (EP_RAIL *r, void *private, unsigned int index)
{
    EP4_RAIL     *rail  = (EP4_RAIL *) r;
    ELAN4_DEV    *dev   = rail->r_ctxt.ctxt_dev;
    E4_uint64     pte;
    unsigned long flags;

    local_irq_save (flags);
    pte = elan4mmu_get_pte (&rail->r_ctxt, (ELAN4_HASH_CACHE *) private, index << (PAGE_SHIFT - dev->dev_pageshift[0]));
    local_irq_restore (flags);

    return elan4mmu_pte2phys (dev, pte);
}

void
ep4_dvma_unload (EP_RAIL *r, void *private, unsigned int index, unsigned int npages)
{
    EP4_RAIL  *rail  = (EP4_RAIL *) r;
    ELAN4_DEV *dev   = rail->r_ctxt.ctxt_dev;
    EP_ADDR    eaddr = ((ELAN4_HASH_CACHE *) private)->hc_start + (index * PAGE_SIZE);
    unsigned long idx = (index << (PAGE_SHIFT - dev->dev_pageshift[0]));
    unsigned long lim = idx + (npages << (PAGE_SHIFT - dev->dev_pageshift[0]));
    unsigned long flags;

    EPRINTF5 (DBG_KMAP, "ep4_dvma_unload: eaddr %x -> %lx : index=%d idx=%ld lim=%ld\n", 
	      eaddr, (unsigned long)(eaddr + (npages * PAGE_SIZE)), index, idx, lim);

    local_irq_save (flags);
    for (; idx < lim; idx++)
	elan4mmu_clear_pte (&rail->r_ctxt, (ELAN4_HASH_CACHE *) private, idx);
    local_irq_restore (flags);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
