/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kmsg_elan3.c,v 1.4.2.1 2005/07/19 14:26:48 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kmsg_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "debug.h"

static void
ep3_inputq_event (EP3_RAIL *rail, void *arg)
{
    EP3_INPUTQ *inputq = (EP3_INPUTQ *) arg;
    
    (*inputq->q_callback)((EP_RAIL *)rail, inputq->q_arg);
}

static EP3_COOKIE_OPS ep3_inputq_cookie_ops = 
{
    ep3_inputq_event,
};

EP_INPUTQ *
ep3_alloc_inputq (EP_RAIL *r, unsigned qnum, unsigned slotSize, unsigned slotCount,
		  EP_INPUTQ_CALLBACK *callback, void *arg)
{
    EP3_RAIL      *rail = (EP3_RAIL *) r;
    EP3_INPUTQ    *inputq;
    EP3_InputQueue qdesc;
    void          *slots;
    int            i;

    ASSERT ((slotSize & (EP_SYSTEMQ_MSG_ALIGN-1)) == 0);

    KMEM_ALLOC (inputq, EP3_INPUTQ *, sizeof (EP3_INPUTQ), TRUE);

    if (inputq == NULL)
	return (EP_INPUTQ *) NULL;
    
    if ((slots = ep_alloc_main (&rail->Generic, slotSize * slotCount, 0, &inputq->q_slotsAddr)) == NULL)
    {
	KMEM_FREE (inputq, sizeof (EP3_INPUTQ));
	return (EP_INPUTQ *) NULL;
    }

    inputq->q_slotSize  = slotSize;
    inputq->q_slotCount = slotCount;
    inputq->q_callback  = callback;
    inputq->q_arg       = arg;
    inputq->q_slots     = slots;

    /* Initialise all the slots to be "unreceived" */
    for (i = 0; i < slotCount; i++)
	((uint32_t *) ((unsigned long) slots + (i+1) * slotSize))[-1] = EP_SYSTEMQ_UNRECEIVED;
    
    inputq->q_base     = inputq->q_slotsAddr;
    inputq->q_top      = inputq->q_base + (slotCount-1) * slotSize;
    inputq->q_fptr     = inputq->q_base;
    inputq->q_desc     = EP_SYSTEMQ_DESC(rail->QueueDescs, qnum);
    inputq->q_descAddr = EP_SYSTEMQ_ADDR (qnum);

    if (callback)
	RegisterCookie (&rail->CookieTable, &inputq->q_cookie, inputq->q_descAddr, &ep3_inputq_cookie_ops, inputq);

    /* Initialise the input queue descriptor */
    qdesc.q_state          = E3_QUEUE_FULL;
    qdesc.q_bptr           = inputq->q_base + slotSize;
    qdesc.q_fptr           = inputq->q_fptr;
    qdesc.q_base           = inputq->q_base;
    qdesc.q_top            = inputq->q_top;
    qdesc.q_size           = slotSize;
    qdesc.q_event.ev_Count = 1;
    qdesc.q_event.ev_Type  = callback ? EV_TYPE_EVIRQ | inputq->q_cookie.Cookie : 0;
    qdesc.q_wevent         = inputq->q_descAddr + offsetof (EP3_InputQueue, q_event);
    qdesc.q_wcount         = 0;

    /* copy the queue descriptor down to sdram */
    elan3_sdram_copyl_to_sdram (rail->Device, &qdesc, inputq->q_desc, sizeof (EP3_InputQueue));

    return (EP_INPUTQ *) inputq;
}

void
ep3_free_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP3_RAIL   *rail   = (EP3_RAIL *) r;
    EP3_INPUTQ *inputq = (EP3_INPUTQ *) q;

    ep_free_main (&rail->Generic, inputq->q_slotsAddr, inputq->q_slotSize * inputq->q_slotCount);

    if (inputq->q_callback)
	DeregisterCookie (&rail->CookieTable, &inputq->q_cookie);

    KMEM_FREE (inputq, sizeof (EP3_INPUTQ));
}

void
ep3_enable_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP3_RAIL   *rail   = (EP3_RAIL *) r;
    EP3_INPUTQ *inputq = (EP3_INPUTQ *) q;

    elan3_sdram_writel (rail->Device, inputq->q_desc + offsetof (EP3_InputQueue, q_state), 0);
}

void
ep3_disable_inputq (EP_RAIL *r, EP_INPUTQ *q)
{
    EP3_RAIL      *rail   = (EP3_RAIL *) r;
    EP3_INPUTQ    *inputq = (EP3_INPUTQ *) q;
    EP3_InputQueue qdesc;

    /* mark the queue as locked */
    SetQueueLocked (rail, inputq->q_desc);

    /* re-initialise the queue as empty */
    qdesc.q_state          = E3_QUEUE_FULL;
    qdesc.q_bptr           = (E3_Addr) inputq->q_base + inputq->q_slotSize;
    qdesc.q_fptr           = inputq->q_fptr;
    qdesc.q_base           = inputq->q_base;
    qdesc.q_top            = inputq->q_top;
    qdesc.q_size           = inputq->q_slotSize;
    qdesc.q_event.ev_Count = 1;
    qdesc.q_event.ev_Type  = inputq->q_callback ? EV_TYPE_EVIRQ | inputq->q_cookie.Cookie : 0;
    qdesc.q_wevent         = inputq->q_descAddr + offsetof (EP3_InputQueue, q_event);
    qdesc.q_wcount         = 0;

    /* copy the queue descriptor down to sdram */
    elan3_sdram_copyl_to_sdram (rail->Device, &qdesc, inputq->q_desc, sizeof (EP3_InputQueue));
}

int
ep3_poll_inputq (EP_RAIL *r, EP_INPUTQ *q, int maxCount, EP_INPUTQ_HANDLER *handler, void *arg)
{
    EP3_RAIL          *rail   = (EP3_RAIL *) r;
    EP3_INPUTQ        *inputq = (EP3_INPUTQ *) q;
    sdramaddr_t        qdesc  = inputq->q_desc;
    E3_Addr            nfptr;
    int                count = 0;
    E3_uint32          state;
    int		       delay;

 run_again_because_of_eventqueue_overflow:
    nfptr = inputq->q_fptr + inputq->q_slotSize;
    if (nfptr > inputq->q_top)					
	nfptr = inputq->q_base;

    while (nfptr != elan3_sdram_readl (rail->Device, qdesc + offsetof (EP3_InputQueue, q_bptr)))			/* PCI read */
    {
	unsigned long slot = (unsigned long) inputq->q_slots + (nfptr - inputq->q_base);

	/* Poll the final word of the message until the message has completely
	 * arrived in main memory. */
	for (delay = 1; ((uint32_t *) (slot + inputq->q_slotSize))[-1] == EP_SYSTEMQ_UNRECEIVED && delay < EP_SYSTEMQ_UNRECEIVED_TLIMIT; delay <<= 1)
	    DELAY (delay);

	/* Call the message handler */
	(*handler) (r, arg, (void *) slot);
	
	/* reset the last word of the slot to "unreceived" */
	((uint32_t *) (slot + inputq->q_slotSize))[-1] = EP_SYSTEMQ_UNRECEIVED;
	    
	state = elan3_sdram_readl (rail->Device, qdesc + offsetof (EP3_InputQueue, q_state)); 				/* PCI read */
	if ((state & E3_QUEUE_FULL) == 0)
	    elan3_sdram_writel (rail->Device, qdesc + offsetof (EP3_InputQueue, q_fptr), nfptr);			/* PCI write */
	else
	{
	    elan3_sdram_writel (rail->Device, qdesc + offsetof (EP3_InputQueue, q_fptr), nfptr);			/* PCI write */
	    elan3_sdram_writel (rail->Device, qdesc + offsetof (EP3_InputQueue, q_state), (state & ~E3_QUEUE_FULL));	/* PCI write */
	}
	inputq->q_fptr = nfptr;
	
	nfptr += roundup (inputq->q_slotSize, E3_BLK_ALIGN);
	if (nfptr > inputq->q_top)
	    nfptr = inputq->q_base;

	if (++count >= maxCount && maxCount)
	    break;
    }
    
    if (inputq->q_callback && count != 0)
    {
	if (count != inputq->q_waitCount)
	    elan3_sdram_writel (rail->Device, qdesc + offsetof (EP3_InputQueue, q_wcount), inputq->q_waitCount = count);

	if (IssueWaitevent (rail, inputq->q_descAddr + offsetof (EP3_InputQueue, q_wevent)) == ISSUE_COMMAND_TRAPPED)
	    goto run_again_because_of_eventqueue_overflow;
    }

    return count;
}

#define Q_EVENT(q,slotNum)		((q)->q_elan      + (slotNum) * sizeof (E3_BlockCopyEvent))
#define Q_EVENT_ADDR(q,slotNum)		((q)->q_elanAddr  + (slotNum) * sizeof (E3_BlockCopyEvent))
#define Q_MSG(q,slotNum)	(void *)((q)->q_main      + (slotNum) * (q)->q_slotSize)
#define Q_MSG_ADDR(q,slotNum)		((q)->q_mainAddr  + (slotNum) * (q)->q_slotSize)
#define Q_DONE(q,slotNum)     (*((int *)((q)->q_main      + (q)->q_slotCount * (q)->q_slotSize + (slotNum) * sizeof (E3_uint32))))
#define Q_DONE_ADDR(q,slotNum)		((q)->q_mainAddr  + (q)->q_slotCount * (q)->q_slotSize + (slotNum) * sizeof (E3_uint32))

#define Q_ELAN_SIZE(q)			((q)->q_slotCount * sizeof (E3_BlockCopyEvent))
#define Q_MAIN_SIZE(q)			((q)->q_slotCount * ((q)->q_slotSize + sizeof (E3_uint32)))

static void
ep3_outputq_retry (EP3_RAIL *rail, void *arg, E3_DMA_BE *dma, int error)
{
    E3_DMA_BE    *dmabe = (E3_DMA_BE *) dma;
    sdramaddr_t   event = ep_elan2sdram (&rail->Generic, dmabe->s.dma_srcEvent);
    E3_Addr       done  = elan3_sdram_readl (rail->Device, event + offsetof (E3_BlockCopyEvent, ev_Dest));
    E3_uint32    *donep = ep_elan2main (&rail->Generic, done & ~EV_BCOPY_DTYPE_MASK);

    EPRINTF1 (DBG_KMSG, "ep3_ouputq_retry: donep at %p -> FAILED\n", donep);
    
    *donep = EP3_EVENT_FAILED;
}

static EP3_COOKIE_OPS ep3_outputq_cookie_ops =
{
    NULL, /* Event */
    ep3_outputq_retry,
    NULL, /* DmaCancelled */
    NULL, /* DmaVerify */
};

EP_OUTPUTQ *
ep3_alloc_outputq (EP_RAIL *r, unsigned slotSize, unsigned slotCount)
{
    EP3_RAIL         *rail = (EP3_RAIL *) r;
    EP3_OUTPUTQ      *outputq;
    int               i;
    E3_BlockCopyEvent event;

    ASSERT ((slotSize & (EP_SYSTEMQ_MSG_ALIGN-1)) == 0);

    KMEM_ALLOC (outputq, EP3_OUTPUTQ *, sizeof (EP3_OUTPUTQ), 1);

    if (outputq == NULL)
	return NULL;

    outputq->q_slotCount = slotCount;
    outputq->q_slotSize  = slotSize;

    outputq->q_elan = ep_alloc_elan (r, Q_ELAN_SIZE(outputq), 0, &outputq->q_elanAddr);

    if (outputq->q_elan == (sdramaddr_t) 0)
    {
	KMEM_FREE (outputq, sizeof (EP3_OUTPUTQ));
	return NULL;
    }

    outputq->q_main = ep_alloc_main (r, Q_MAIN_SIZE(outputq), 0, &outputq->q_mainAddr);

    if (outputq->q_main == (void *) NULL)
    {
	ep_free_elan (r, outputq->q_elanAddr, Q_ELAN_SIZE(outputq));
	KMEM_FREE (outputq, sizeof (EP3_OUTPUTQ));
	return NULL;
    }

    RegisterCookie (&rail->CookieTable, &outputq->q_cookie, outputq->q_elanAddr, &ep3_outputq_cookie_ops, outputq);

    for (i = 0; i < slotCount; i++)
    {
	EP3_INIT_COPY_EVENT (event, outputq->q_cookie, Q_DONE_ADDR(outputq, i), 0);

	Q_DONE(outputq, i) = outputq->q_cookie.Cookie;
	
	elan3_sdram_copyl_to_sdram (rail->Device, &event, Q_EVENT(outputq, i), sizeof (E3_BlockCopyEvent));
    }

    return (EP_OUTPUTQ *) outputq;
}

void
ep3_free_outputq (EP_RAIL *r, EP_OUTPUTQ *q)
{
    EP3_RAIL    *rail    = (EP3_RAIL *) r;
    EP3_OUTPUTQ *outputq = (EP3_OUTPUTQ *) q;

    DeregisterCookie (&rail->CookieTable, &outputq->q_cookie);
    
    ep_free_main (r, outputq->q_mainAddr, Q_MAIN_SIZE(outputq));
    ep_free_elan (r, outputq->q_elanAddr, Q_ELAN_SIZE(outputq));
 
    KMEM_FREE (outputq, sizeof (EP3_OUTPUTQ));
}

void *
ep3_outputq_msg (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum)
{
    return Q_MSG ((EP3_OUTPUTQ *) q, slotNum);
}

int
ep3_outputq_state (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum)
{
    switch (Q_DONE((EP3_OUTPUTQ *) q, slotNum))
    {
    case EP3_EVENT_ACTIVE:
	return EP_OUTPUTQ_BUSY;
	
    case EP3_EVENT_FAILED:
	return EP_OUTPUTQ_FAILED;
	
    default:
	return EP_OUTPUTQ_FINISHED;
    }
}

int
ep3_outputq_send (EP_RAIL *r, EP_OUTPUTQ *q, unsigned slotNum, unsigned size, 
		  unsigned vp, unsigned qnum, unsigned retries)
{
    EP3_RAIL    *rail    = (EP3_RAIL *) r;
    EP3_OUTPUTQ *outputq = (EP3_OUTPUTQ *) q;
    unsigned     base    = outputq->q_slotSize - roundup (size, E3_BLK_ALIGN);
    E3_DMA_BE    dmabe;

    dmabe.s.dma_type	        = E3_DMA_TYPE(DMA_BYTE, DMA_WRITE, DMA_QUEUED, retries);
    dmabe.s.dma_size            = roundup (size, E3_BLK_ALIGN);
    dmabe.s.dma_source          = Q_MSG_ADDR(outputq, slotNum) + base;
    dmabe.s.dma_dest            = base;
    dmabe.s.dma_destEvent       = EP_SYSTEMQ_ADDR(qnum);
    dmabe.s.dma_destCookieVProc = vp;
    dmabe.s.dma_srcEvent        = Q_EVENT_ADDR(outputq, slotNum);
    dmabe.s.dma_srcCookieVProc  = 0;

    Q_DONE(outputq, slotNum) = EP3_EVENT_ACTIVE;
    
    elan3_sdram_writel (rail->Device, Q_EVENT(outputq, slotNum), 1);

    if (IssueDma (rail, &dmabe, EP_RETRY_CRITICAL, FALSE) != ISSUE_COMMAND_OK)
    {
	Q_DONE(outputq, slotNum) = EP3_EVENT_FAILED;
	return FALSE;
    }

    return TRUE;
}
