/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: neterr.c,v 1.27.2.1 2005/07/20 12:01:34 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/neterr.c,v $ */

#include <qsnet/kernel.h>
#include <elan/kcomm.h>

#include "debug.h"

void
ep_queue_network_error (EP_RAIL *rail, int nodeId, int what, int channel, EP_NETERR_COOKIE cookie)
{
    EP_SYS       *sys      = rail->System;
    EP_NODE_RAIL *nodeRail = &rail->Nodes[nodeId];
    unsigned long flags;

    spin_lock_irqsave (&sys->NodeLock, flags);

    ASSERT (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE);
    
    if (nodeRail->NetworkErrorState == 0)
    {
	EPRINTF2 (DBG_NETWORK_ERROR, "%s: raise context filter for node %d due to network error\n", rail->Name, nodeId);
	
	rail->Operations.RaiseFilter (rail, nodeId);
	
	if (nodeRail->State == EP_NODE_LOCAL_PASSIVATE)
	    printk ("%s: node %d is flushing - deferring network error fixup\n", rail->Name, nodeId);
	else
	    list_add_tail (&nodeRail->Link, &rail->NetworkErrorList);
    }
    
    switch (what)
    {
    case EP_NODE_NETERR_ATOMIC_PACKET:
	ASSERT (nodeRail->NetworkErrorCookies[channel] == 0);
	
	/* Need to raise the approriate context filter for this node,
	 * and periodically send a neterr fixup message to it until 
	 * we receive an ack from it
	 */
	IncrStat (rail, NeterrAtomicPacket);
	
	nodeRail->NetworkErrorCookies[channel] = cookie;
	
	nodeRail->NetworkErrorState |= EP_NODE_NETERR_ATOMIC_PACKET;
	nodeRail->MsgXid             = ep_xid_cache_alloc (sys, &rail->XidCache);
	
	EPRINTF3 (DBG_NETWORK_ERROR, "%s: atomic packet destroyed - node %d cookie %llx\n", rail->Name, nodeId, (long long)cookie);
	break;

    case EP_NODE_NETERR_DMA_PACKET:
	/* Must be an overlapped dma packet, raise the context filter,
	 * and hold it up for a NETWORK_ERROR_TIMEOUT */
	IncrStat (rail, NeterrDmaPacket);
	
	nodeRail->NetworkErrorState |= EP_NODE_NETERR_DMA_PACKET;
	break;
    }

    nodeRail->NextRunTime = lbolt + NETWORK_ERROR_TIMEOUT;
    
    spin_unlock_irqrestore (&sys->NodeLock, flags);

    ep_kthread_schedule (&sys->ManagerThread, nodeRail->NextRunTime);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */

