/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: neterr_elan3.c,v 1.24 2003/11/17 13:26:45 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/neterr_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "debug.h"

typedef struct neterr_halt_args
{
    EP3_RAIL	     *Rail;
    unsigned int      NodeId;
    EP_NETERR_COOKIE *Cookies;
} NETERR_HALT_ARGS;

static int
DmaMatchesCookie (EP3_RAIL *rail, E3_DMA_BE *dma, int nodeId, EP_NETERR_COOKIE *cookies, char *where)
{
    E3_uint32     cvproc;
    E3_uint32     cookie;
    
    if (dma->s.dma_direction == DMA_WRITE)
    {
	cvproc = dma->s.dma_destCookieVProc;
	cookie = dma->s.dma_srcCookieVProc;
    }
    else
    {
	cvproc = dma->s.dma_srcCookieVProc;
	cookie = dma->s.dma_destCookieVProc;
    }
    
    EPRINTF6 (DBG_NETWORK_ERROR, "%s: Neterr - %s: DMA %08x %08x %08x %08x\n", rail->Generic.Name, where,
	      dma->s.dma_type, dma->s.dma_size, dma->s.dma_source, dma->s.dma_dest);
    EPRINTF5 (DBG_NETWORK_ERROR, "%s:                     %08x %08x %08x %08x\n", rail->Generic.Name,
	      dma->s.dma_destEvent, dma->s.dma_destCookieVProc, dma->s.dma_srcEvent, dma->s.dma_srcCookieVProc);

    if (EP_VP_ISDATA((cvproc & DMA_PROCESS_MASK)) && EP_VP_TO_NODE(cvproc & DMA_PROCESS_MASK) == nodeId)
    {
	/*
	 * This is a DMA going to the node which has a network fixup
	 * request pending, so check if the cookie matches.
	 */
	if ((cookie == cookies[0] || cookie == cookies[1]) /* && !WaitForEop */)
	{
	    EPRINTF3 (DBG_NETWORK_ERROR, "%s: match cookie %08x on %s\n", rail->Generic.Name, cookie, where);
	    
	    return (TRUE);
	}
    }

    return (FALSE);
}


static void
NetworkErrorHaltOperation (ELAN3_DEV *dev, void *arg)
{
    NETERR_HALT_ARGS *args = (NETERR_HALT_ARGS *) arg;
    EP3_RAIL         *rail = args->Rail;
    EP_SYS           *sys  = rail->Generic.System;
    sdramaddr_t       FPtr, BPtr;
    sdramaddr_t       Base, Top;
    E3_DMA_BE         dma;
    unsigned long     flags;

    spin_lock_irqsave (&sys->NodeLock, flags);

    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc.s.FSR)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData1.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData2.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData3.s.FSR.Status)) == 0);
    
    FPtr  = read_reg32 (dev, DProc_SysCntx_FPtr);
    BPtr =  read_reg32 (dev, DProc_SysCntx_BPtr);
    Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0]);
    Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[E3_SysCntxQueueSize-1]);
    
    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, &dma, sizeof (E3_DMA_BE));
	
	if (DmaMatchesCookie (rail, &dma, args->NodeId, args->Cookies, "runq "))
	{
	    /*
	     * Transfer the DMA to the node, it's source event will 
	     * get executed later.
	     */
	    QueueDmaOnStalledList (rail, &dma);
	    
	    /*
	     * Remove the DMA from the queue by replacing it with one with
	     * zero size and no events.
	     *
	     * NOTE: we must preserve the SYS_CONTEXT_BIT since the Elan uses this
	     * to mark the approriate run queue as empty.
	     */
	    dma.s.dma_type            = (SYS_CONTEXT_BIT << 16);
	    dma.s.dma_size            = 0;
	    dma.s.dma_source          = (E3_Addr) 0;
	    dma.s.dma_dest            = (E3_Addr) 0;
	    dma.s.dma_destEvent       = (E3_Addr) 0;
	    dma.s.dma_destCookieVProc = 0;
	    dma.s.dma_srcEvent        = (E3_Addr) 0;
	    dma.s.dma_srcCookieVProc  = 0;
	    
	    elan3_sdram_copyq_to_sdram (dev, &dma, FPtr, sizeof (E3_DMA_BE));
	}

	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_DMA);
    }

    rail->NetworkErrorFlushed = TRUE;
    kcondvar_wakeupall (&rail->NetworkErrorSleep, &sys->NodeLock);

    spin_unlock_irqrestore (&sys->NodeLock, flags);
}

void
ep3_neterr_fixup (EP_RAIL *r, unsigned int nodeId, EP_NETERR_COOKIE *cookies)
{
    EP3_RAIL        *rail        = (EP3_RAIL *) r;
    EP_SYS          *sys         = rail->Generic.System;
    ELAN3_DEV       *dev         = rail->Device;
    EP_NODE_RAIL    *nodeRail    = &rail->Generic.Nodes[nodeId];
    E3_DMA_BE        dmabe;
    EP3_COOKIE      *cp;
    E3_uint32        vp;
    NETERR_HALT_ARGS args;
    struct list_head *el, *nel, matchedList;
    int              i;
    unsigned long    flags;

    INIT_LIST_HEAD (&matchedList);

    StallDmaRetryThread (rail);

    args.Rail       = rail;
    args.NodeId     = nodeId;
    args.Cookies    = cookies;

    spin_lock_irqsave (&rail->Device->IntrLock, flags);
    QueueHaltOperation (rail->Device, 0, NULL, INT_TProcHalted | INT_DProcHalted, NetworkErrorHaltOperation, &args);
    spin_unlock_irqrestore (&rail->Device->IntrLock, flags);
    
    spin_lock_irqsave (&sys->NodeLock, flags);
    while (! rail->NetworkErrorFlushed)
	kcondvar_wait (&rail->NetworkErrorSleep, &sys->NodeLock, &flags);
    rail->NetworkErrorFlushed = FALSE;
    
    spin_lock (&rail->DmaRetryLock);
    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	list_for_each_safe (el, nel, &rail->DmaRetries[i]) {
	    EP3_RETRY_DMA *retry = list_entry (el, EP3_RETRY_DMA, Link);

	    if (DmaMatchesCookie (rail, &retry->Dma, nodeId, cookies, "retry"))
	    {
		/* remove from retry list */
		list_del (&retry->Link);

		/* add to list of dmas which matched */
		list_add_tail (&retry->Link, &matchedList);
	    }
	}
    }
    
    list_for_each_safe (el, nel, &nodeRail->StalledDmas) {
	EP3_RETRY_DMA *retry = list_entry (el, EP3_RETRY_DMA, Link);
	
	if (DmaMatchesCookie (rail, &retry->Dma, nodeId, cookies, "stalled"))
	{
	    /* remove from retry list */
	    list_del (&retry->Link);
	    
	    /* add to list of dmas which matched */
	    list_add_tail (&retry->Link, &matchedList);
	}
    }
    
    spin_unlock (&rail->DmaRetryLock);
    spin_unlock_irqrestore (&sys->NodeLock, flags);
    
    ResumeDmaRetryThread (rail);

    /* Now "set" the source event of any write DMA's */
    while (! list_empty (&matchedList))
    {
	EP3_RETRY_DMA *retry = list_entry (matchedList.next, EP3_RETRY_DMA, Link);
	
	list_del (&retry->Link);

	if (retry->Dma.s.dma_direction == DMA_WRITE && retry->Dma.s.dma_srcEvent)
	{
	    sdramaddr_t event = ep_elan2sdram (&rail->Generic, retry->Dma.s.dma_srcEvent);

	    /* Block local interrupts, since we need to atomically
	     * decrement the event count and perform the word write
	     */
	    local_irq_save (flags);
	    {
		E3_uint32 type  = elan3_sdram_readl (dev, event + offsetof (E3_Event, ev_Type));
		E3_uint32 count = elan3_sdram_readl (dev, event + offsetof (E3_Event, ev_Count));

		elan3_sdram_writel (dev, event + offsetof (E3_Event, ev_Count), count - 1);

		if (count == 1)
		{
		    if (type & EV_TYPE_MASK_BCOPY)
		    {
			E3_Addr srcVal  = elan3_sdram_readl (dev, event + offsetof (E3_BlockCopyEvent, ev_Source));
			E3_Addr dstAddr = elan3_sdram_readl (dev, event + offsetof (E3_BlockCopyEvent, ev_Dest)) & ~EV_BCOPY_DTYPE_MASK;

			ASSERT ((srcVal & EV_WCOPY) != 0);
			
			EPRINTF3 (DBG_NETWORK_ERROR, "%s: neterr perform event word write at %08x with %08x\n", rail->Generic.Name, dstAddr, srcVal);

			ELAN3_OP_STORE32 (rail->Ctxt, dstAddr, srcVal);
		    }

		    if ((type & ~EV_TYPE_MASK_BCOPY) != 0)
		    {
			if ((type & EV_TYPE_MASK_CHAIN) == EV_TYPE_CHAIN)
			{
			    printk ("%s: event at %08x - chained event %x is invalid\n", rail->Generic.Name, retry->Dma.s.dma_srcEvent, type);
			    panic ("ep: neterr invalid event type\n");
			}
			else if ((type & EV_TYPE_MASK_EVIRQ) == EV_TYPE_EVIRQ)
			{
			    EPRINTF2 (DBG_NETWORK_ERROR, "%s: neterr event interrupt - cookie %08x\n", rail->Generic.Name, (type & ~(EV_TYPE_MASK_EVIRQ|EV_TYPE_MASK_BCOPY)));
			    
			    cp = LookupCookie (&rail->CookieTable, (type & ~(EV_TYPE_MASK_EVIRQ|EV_TYPE_MASK_BCOPY)));
			    
			    if (cp->Operations->Event)
				cp->Operations->Event(rail, cp->Arg);
			}
			else if ((type & EV_TYPE_MASK_DMA) == EV_TYPE_DMA)
			{
			    sdramaddr_t dma = ep_elan2sdram (&rail->Generic, (type & ~EV_TYPE_MASK2));
			    
			    EPRINTF2 (DBG_NETWORK_ERROR, "%s: neterr chained dma - %08x\n", rail->Generic.Name, (type & ~EV_TYPE_MASK2));
			    
			    elan3_sdram_copyq_from_sdram (dev, dma, &dmabe, sizeof (E3_DMA));
			    
			    if (dmabe.s.dma_direction == DMA_WRITE)
			    {
				vp = dmabe.s.dma_destVProc;
				cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_srcEvent);
			    }
			    else
			    {
				vp = dmabe.s.dma_srcVProc;
				cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_destEvent);
				
				/* we MUST convert this into a DMA_READ_REQUEUE dma as if we don't the 
				 * DMA descriptor will be read from the EP_RETRY_DMA rather than the 
				 * original DMA - this can then get reused and an incorrect DMA 
				 * descriptor sent 
				 * eventp->ev_Type contains the dma address with type in the lower bits 
				 */ 
			    
				dmabe.s.dma_source    = (type & ~EV_TYPE_MASK2);
				dmabe.s.dma_direction = (dmabe.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
			    }
			
			    ASSERT (EP_VP_ISDATA(vp));
			
			    nodeRail = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];

			    switch (nodeRail->State)
			    {
			    case EP_NODE_CONNECTED:
			    case EP_NODE_LEAVING_CONNECTED:
				if (cp != NULL)
				    cp->Operations->DmaRetry (rail, cp->Arg, &dmabe, EAGAIN);
				else
				{
				    ASSERT (dmabe.s.dma_direction == DMA_WRITE && dmabe.s.dma_srcEvent == 0 && dmabe.s.dma_isRemote);
				
				    QueueDmaForRetry (rail, &dmabe, EP_RETRY_ANONYMOUS);
				}
				break;

			    case EP_NODE_LOCAL_PASSIVATE:
				QueueDmaOnStalledList (rail, &dmabe);
				break;

			    default:
				panic ("ep: neterr incorrect state for node\n");
			    }
			}
			else if ((type & EV_TYPE_MASK_THREAD) == EV_TYPE_THREAD)
			{
			    printk ("%s: event at %08x - thread waiting %x is invalid\n", rail->Generic.Name, retry->Dma.s.dma_srcEvent, type);
			    panic ("ep: neterr invalid event type\n");
			}
		    }
		}
	    }
	    local_irq_restore(flags);
	}
	
	/* add to free list */
	spin_lock_irqsave (&rail->DmaRetryLock, flags);
	list_add (&retry->Link, &rail->DmaRetryFreeList);
	spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
    }
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */

