/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */
#ident "@(#)$Id: nmh.c,v 1.6 2004/01/05 13:48:08 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/nmh.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#define EP_NMD_SPANS(nmd, base, top)	((nmd)->nmd_addr <= (base) &&  \
					 ((nmd)->nmd_addr + (nmd)->nmd_len - 1) >= (top))

#define EP_NMD_OVERLAPS(nmd, addr, len)	((nmd)->nmd_addr <= ((addr) + (len)) && \
					 ((nmd)->nmd_addr + (nmd)->nmd_len - 1) >= (addr))

#define EP_NMH_HASH(tbl,idx,addr)	((addr) % (tbl)->tbl_size[idx])

int
ep_nmh_init (EP_NMH_TABLE *tbl)
{
    int i, idx, hsize = 1;

    for (idx = EP_NMH_NUMHASH-1; idx >= 0; idx--, hsize <<= 1)
    {
	tbl->tbl_size[idx] = (hsize < EP_NMH_HASHSIZE) ? hsize : EP_NMH_HASHSIZE;

	KMEM_ZALLOC (tbl->tbl_hash[idx], struct list_head *, sizeof (struct list_head) * tbl->tbl_size[idx], 1);
	
	if (tbl->tbl_hash == NULL)
	{
	    while (++idx < EP_NMH_NUMHASH)
		KMEM_FREE (tbl->tbl_hash[idx], sizeof (struct list_head) * tbl->tbl_size[idx]);
	    return (ENOMEM);
	}

	for (i = 0; i < tbl->tbl_size[idx]; i++)
	    INIT_LIST_HEAD (&tbl->tbl_hash[idx][i]);
    }

    return (0);
}

void
ep_nmh_fini (EP_NMH_TABLE *tbl)
{
    int idx;

    for (idx = 0; idx < EP_NMH_NUMHASH; idx++)
	if (tbl->tbl_hash[idx])
	    KMEM_FREE (tbl->tbl_hash[idx], sizeof (struct list_head) * tbl->tbl_size[idx]);
    
    bzero (tbl, sizeof (EP_NMH_TABLE));
}

void
ep_nmh_insert (EP_NMH_TABLE *tbl, EP_NMH *nmh)
{
    EP_ADDR base = nmh->nmh_nmd.nmd_addr;
    EP_ADDR top  = base + nmh->nmh_nmd.nmd_len - 1;
    int     idx;

    for (idx = 0, base >>= 12, top >>= 12; base != top && idx < EP_NMH_NUMHASH; idx++, base >>= 1, top >>= 1)
	;

    list_add_tail (&nmh->nmh_link, &tbl->tbl_hash[idx][EP_NMH_HASH(tbl, idx, base)]);
}

void
ep_nmh_remove (EP_NMH_TABLE *tbl, EP_NMH *nmh)
{
    list_del (&nmh->nmh_link);
}

EP_NMH *
ep_nmh_find (EP_NMH_TABLE *tbl, EP_NMD *nmd)
{
    EP_ADDR           base = nmd->nmd_addr;
    EP_ADDR           top  = base + nmd->nmd_len - 1;
    int               idx;
    struct list_head *le;
    
    for (idx = 0, base >>= 12, top >>= 12; base != top && idx < EP_NMH_NUMHASH; idx++, base >>= 1, top >>= 1)
	;
    
    for (; idx < EP_NMH_NUMHASH; idx++, base >>= 1, top >>= 1) {

	list_for_each (le, &tbl->tbl_hash[idx][EP_NMH_HASH(tbl, idx, base)]) {
	    EP_NMH *nmh = list_entry (le, EP_NMH, nmh_link);

	    if (EP_NMD_SPANS (&nmh->nmh_nmd, nmd->nmd_addr, nmd->nmd_addr + nmd->nmd_len - 1))
		return (nmh);
	}
    }

    return (0);
}

void
ep_nmd_subset (EP_NMD *subset, EP_NMD *nmd, unsigned off, unsigned len)
{
    ASSERT ((off + len - 1) <= nmd->nmd_len);

    subset->nmd_addr = nmd->nmd_addr + off;
    subset->nmd_len  = len;
    subset->nmd_attr = nmd->nmd_attr;
}

int
ep_nmd_merge (EP_NMD *merged, EP_NMD *a, EP_NMD *b)
{
    if (EP_NMD_NODEID (a) != EP_NMD_NODEID (b))			/* not generated on the same node */
	return 0;
    
    if ((EP_NMD_RAILMASK (a) & EP_NMD_RAILMASK (b)) == 0)	/* no common rails */
	return 0;
    
    if (b->nmd_addr == (a->nmd_addr + a->nmd_len))
    {
	if (merged != NULL)
	{
	    merged->nmd_addr = a->nmd_addr;
	    merged->nmd_len  = a->nmd_len + b->nmd_len;
	    merged->nmd_attr = EP_NMD_ATTR(EP_NMD_NODEID(a), EP_NMD_RAILMASK(a) & EP_NMD_RAILMASK(b));
	}
	return 1;
    }
    
    if (a->nmd_addr == (b->nmd_addr + b->nmd_len))
    {
	if (merged != NULL)
	{
	    merged->nmd_addr = b->nmd_addr;
	    merged->nmd_len  = b->nmd_len + a->nmd_len;
	    merged->nmd_attr = EP_NMD_ATTR(EP_NMD_NODEID(b), EP_NMD_RAILMASK(a) & EP_NMD_RAILMASK(b));
	}
	
	return 1;
    }

    return 0;
}

int
ep_nmd_map_rails (EP_SYS *sys, EP_NMD *nmd, unsigned railmask)
{
    EP_NMH *nmh = ep_nmh_find (&sys->MappingTable, nmd);

    if (nmh == NULL)
    {
	printk ("ep_nmd_map_rails: nmd=%08x.%08x.%08x cannot be found\n",
		nmd->nmd_addr, nmd->nmd_len, nmd->nmd_attr);
	return (-1);
    }

    return (nmh->nmh_ops->op_map_rails (sys, nmh, nmd, railmask));
}

EP_RAILMASK
ep_nmd2railmask (EP_NMD *frags, int nFrags)
{
    EP_RAILMASK mask;

    if (nFrags == 0)
	return ((EP_RAILMASK)-1);
    
    for (mask = EP_NMD_RAILMASK(frags); --nFrags; )
	mask &= EP_NMD_RAILMASK(++frags);

    return (mask);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
