/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: probenetwork.c,v 1.43 2004/04/19 15:43:15 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/probenetwork.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include "debug.h"

int PositionCheck = 1;

#define NUM_DOWN_FROM_VAL(NumDownLinksVal, level)	(((NumDownLinksVal) >> ((level) << 2)) & 0xF)

int
ProbeNetwork (EP_RAIL *rail, ELAN_POSITION *pos)
{
    int               lvl, i;
    int               level;
    int               nodeid;
    int               numnodes;
    int		      randomRoutingDisabled;
    int               sw;
    int               nacks;
    int               nowayup;
    int		      nalias;
    int		      upmask;
    int		      partial;
    int		      link;
    int		      invalid;
    int		      linkdown[ELAN_MAX_LEVELS];
    int		      linkup[ELAN_MAX_LEVELS];
    EP_SWITCH        *switches[ELAN_MAX_LEVELS];
    int               switchCount[ELAN_MAX_LEVELS+1];
    int               lowestBcast;
    int               numUpLinks[ELAN_MAX_LEVELS];
    int               routedown [ELAN_MAX_LEVELS];

    EPRINTF1 (DBG_PROBE, "%s: ProbeNetwork started\n", rail->Name);

    switchCount[0] = 1;
    numUpLinks [0] = 4;

    for (level = 0; level < ELAN_MAX_LEVELS; level++)
    {
	int ndown  = NUM_DOWN_FROM_VAL (rail->Devinfo.dev_num_down_links_value, level);

	KMEM_ZALLOC (switches[level], EP_SWITCH *, sizeof (EP_SWITCH) * switchCount[level], 1);

	for (sw = 0, nacks = 0, nowayup = 0, lowestBcast=7; sw < switchCount[level]; sw++)
	{
	    EP_SWITCH *lsw  = &switches[level][sw];
	    int        good = 1;
	    int        tsw;

	    for (nodeid = 0,tsw = sw, lvl = level-1 ; lvl >= 0 ; lvl--)
	    {
		EP_SWITCH *lsw;
		int        link = (8-numUpLinks[lvl]) + (tsw % numUpLinks[lvl]);

		tsw  = tsw / numUpLinks[lvl];
		lsw  = &switches[lvl][tsw];

		if (lsw->present == 0 || (lsw->lnr & (1 << link)))
		{
		    EPRINTF4 (DBG_PROBE, "lvl %d sw %d present=%d lnr=%x\n", lvl, sw, lsw->present, lsw->lnr);
		    good = 0;
		}
		
		linkup[lvl]   = link;
		linkdown[lvl] = lsw->link;

		if ( lvl ) nodeid = ((nodeid + linkdown[lvl]) * (8-numUpLinks[lvl-1]));
		else       nodeid += linkdown[0];

	    }
	    
	    /* 
	     * don't bother probing routes which we we've already seen are unreachable 
	     * because a link upwards was in reset or the switch previously nacked us.
	     */
	    if (! good)
	    {
		lsw->present = 0;

		nacks++;
		nowayup++;

		continue;
	    }

	    lsw->present = rail->Operations.ProbeRoute (rail, level, sw, nodeid, linkup, linkdown, 5, lsw);

	    if (! lsw->present)
	    {
		EPRINTF3 (DBG_PROBE, "%s: level %d switch %d - unexpected nack\n", rail->Name, level, sw);

		nacks++;
		nowayup++;
	    }
	    else
	    {
		EPRINTF5 (DBG_PROBE, "%s: level %d switch %d - link %d bcast %d\n", rail->Name, level, sw, lsw->link, lsw->bcast);

		if (level == 2 && rail->Devinfo.dev_device_id == PCI_DEVICE_ID_ELAN3)
		{
		    /* If we see broadcast top as 7, and we came in on a low link, then we can't
		     * determine whether we're in a 128 way or a un-configured 64u64d switch, so
		     * we treat it as a 64u64d and detect the 128 way case by "going over the top" 
		     * below. Unless we've been told what it really is by NumDownLinksVal.
		     */
		    if (lsw->bcast == 7 && lsw->link < 4)
			lsw->bcast = ndown ? (ndown - 1) : 3;
		}

		if ( lowestBcast > lsw->bcast ) 
		    lowestBcast = lsw->bcast;

		if (lsw->link > (ndown ? (ndown-1) : (lowestBcast == 7 ? 3 : lowestBcast)))
		{
		    /* We've arrived on a "up-link" - this could be either
		     * we're in the top half of a x8 top-switch - or we're
		     * in the bottom half and have gone "over the top". We
		     * differentiate these cases since the switches below
		     * a x8 top-switch will have broadcast top set to 3, 
		     * and the x8 topswitch have broadcast top set to 7.
		     */
		    if (lsw->bcast == 7)
			nowayup++;
		    else
		    {
			EPRINTF2 (DBG_PROBE, "%s: level %d - gone over the top\n",
				  rail->Name, level);

			if (level > 0)
			{
			    KMEM_FREE (switches[level], sizeof (EP_SWITCH) * switchCount[level] );
			    level--;
			}
			
			numUpLinks[level] = 0;
			goto finished;
		    }
		}

	    }
	}

	numUpLinks[level]    = ndown ? (8 - ndown) : (7 - lowestBcast);
	switchCount[level+1] = switchCount[level] *  numUpLinks[level];
	
	/* Now we know which links are uplinks, we can see whether there is
	 * any possible ways up */
	upmask = (ndown ? (0xFF << ndown) & 0xFF : (0xFF << (8 - numUpLinks[level])) & 0xFF);

	for (sw = 0; sw < switchCount[level]; sw++)
	{
	    EP_SWITCH *lsw  = &switches[level][sw];

	    if (lsw->present && lsw->link <= (ndown ? (ndown-1) : (lowestBcast == 7 ? 3 : lowestBcast)) && (switches[level][sw].lnr & upmask) == upmask)
		nowayup++;
	}

	EPRINTF7 (DBG_PROBE, "%s: level %d - sw=%d nacks=%d nowayup=%d bcast=%d numup=%d\n", 
		  rail->Name, level, sw, nacks, nowayup, lowestBcast, numUpLinks[level]);

	if (nacks == sw)
	{
	    static bitmap_t printed[BT_BITOUL(EP_MAX_RAILS)];

	    if (! BT_TEST (printed, rail->Number))
		printk ("%s: cannot determine network position\n", rail->Name);
	    BT_SET (printed, rail->Number);
	    goto failed;
	}

	if (nowayup == sw)
	    goto finished;
    }
    
    printk ("%s: exceeded number of levels\n", rail->Name);
    level = ELAN_MAX_LEVELS - 1;

 failed:
    
    for (lvl = 0; lvl <= level; lvl++)
	KMEM_FREE (switches[lvl], sizeof (EP_SWITCH) * switchCount[lvl] );

    return -EAGAIN;

 finished:
    /* we've successfully probed the network - now calculate our node 
     * positon and what level of random routing is possible */
    nalias = 1;
    for (lvl = 0, invalid = 0, partial = 0, randomRoutingDisabled = 0; lvl <= level; lvl++)
    {
	int ndown  = NUM_DOWN_FROM_VAL (rail->Devinfo.dev_num_down_links_value, lvl);
	int upmask = ndown ? (0xFF << ndown) & 0xFF : 0xF0;

	for (sw = 0, nalias = 0; sw < switchCount[lvl]; sw++)
	{
	    EP_SWITCH *lsw = &switches[lvl][sw];
	    
	    /* You can only use adaptive routing if links 4-7 are uplinks, and at least one of them is
	     * not in reset.   Otherwise you can randomly select an "uplink" if all the uplinks are not
	     * in reset. */
	    if (lsw->present && ((upmask == 0xF0) ? (lsw->lnr & upmask) == upmask : (lsw->lnr & upmask) != 0))
		randomRoutingDisabled |= (1 << lvl);
	    
	    if (!lsw->present)
		partial++;
	    else
	    {
		if (lsw->invalid)
		{
		    printk ("%s: invalid switch detected (level %d switch %d)\n", rail->Name, lvl, sw);
		    invalid++;
		}
		
		for (i = 0; i < nalias; i++)
		    if (linkdown[i] == lsw->link)
			break;
		if (i == nalias)
		    linkdown[nalias++] = lsw->link;
	    }
	}
	
	link = linkdown[0];
	for (i = 1; i < nalias; i++)
	    if (linkdown[i] < link)
		link = linkdown[i];

	if (nalias > 1 && lvl != level)
	{
	    printk ("%s: switch aliased below top level (level %d)\n", rail->Name, lvl);
	    invalid++;
	}
	
	routedown[lvl] = link;
   }

    for (lvl = 0; lvl <= level; lvl++) 
	KMEM_FREE (switches[lvl], sizeof (EP_SWITCH) * switchCount[lvl] );

    if (invalid)
    {
	printk ("%s: invalid switch configuration\n", rail->Name);
	return (EINVAL);
    }

    /* Handle the aliasing case where a 16 way is used as multiple smaller switches */
    if (nalias == 1)
	level++;
    else if (nalias == 2)					/* a 16 way as 2x8 ways */
	numUpLinks[level++] = 6;				/*   only 2 down links */
    else if (nalias > 4)					/* a 16 way as 8x2 ways */
	numUpLinks[level-1] = 6;
    
    /* 
     * Compute my nodeid and number of nodes in the machine
     * from the routedown and the number of downlinks at each level.
     */
    for(nodeid=0, lvl = level - 1; lvl >= 0; lvl--)
    {
	if (lvl) nodeid = ((nodeid + routedown[lvl]) * (8-numUpLinks[lvl-1]));	
	else     nodeid += routedown[0];
    }

    for (numnodes = 1, lvl = 0; lvl < level; lvl++)
	numnodes *= (8 - numUpLinks[lvl]);

    sprintf (rail->Name, "ep%d[%d]", rail->Number, nodeid);

    if (randomRoutingDisabled & ((1 << (level-1))-1))
	printk ("%s: nodeid=%d level=%d numnodes=%d (random routing disabled 0x%x)\n", 
		rail->Name, nodeid, level, numnodes, randomRoutingDisabled);
    else if (partial)
	printk ("%s: nodeid=%d level=%d numnodes=%d (random routing ok)\n",
		rail->Name, nodeid, level, numnodes);
    else
	printk ("%s: nodeid=%d level=%d numnodes=%d\n",
		rail->Name, nodeid, level, numnodes);

    pos->pos_mode		 = ELAN_POS_MODE_SWITCHED;
    pos->pos_nodeid              = nodeid;
    pos->pos_levels              = level;
    pos->pos_nodes               = numnodes;
    pos->pos_random_disabled     = randomRoutingDisabled;

    for(lvl = 0; lvl < level; lvl++)
	pos->pos_arity[level -lvl - 1] = (8-numUpLinks[lvl]);
    pos->pos_arity[level] = 1;				/* XXXX why does this need to be 1 ? */
    
    return 0;
}

/*
 * broadcast top is invalid if it is not set to the number of downlinks-1,
 * or at the topmost level it is less than ndown-1.
 */
#define BCAST_TOP_INVALID(lvl, bcast, ndown)	((lvl) == 0 ? (bcast) < ((ndown)-1) : (bcast) != ((ndown) - 1))

void
CheckPosition (EP_RAIL *rail)
{
    ELAN_POSITION *pos     = &rail->Position;
    unsigned int   nodeid  = pos->pos_nodeid;
    unsigned int   invalid = 0;
    unsigned int   changed = 0;
    int lvl, slvl;

    if (! PositionCheck)
	return;

    if (rail->Operations.CheckPosition(rail))		/* is update ready for this rail */
    {
	EPRINTF2 (DBG_ROUTETABLE, "%s: check position: SwitchProbeLevel=%d\n", rail->Name, rail->SwitchProbeLevel);

	for (lvl = 0, slvl = pos->pos_levels-1; lvl <= rail->SwitchProbeLevel; lvl++, slvl--)
	{
	    EP_SWITCHSTATE *state  = &rail->SwitchState[lvl];
	    EP_SWITCHSTATE *lstate = &rail->SwitchLast[lvl];
	    unsigned int    ndown  = pos->pos_arity[slvl];
	    unsigned int    upmask = (0xFF << ndown) & 0xFF;
	    unsigned int    mylink = nodeid % ndown;
	    unsigned int    error  = 0;
	    unsigned int    binval = 0;

	    nodeid /= ndown;

	    /*
	     * broadcast top is invalid if it is not set to the number of downlinks-1,
	     * or at the topmost level it is less than ndown-1.
	     */
	    if (BCAST_TOP_INVALID(lvl, state->bcast, ndown) || (state->LNR & upmask) == upmask)
	    {
		/* no way up from here - we'd better be at the top */
		if (lvl != (pos->pos_levels-1))
		{
		    if (state->bcast != (ndown-1))
			printk ("%s: invalid broadcast top %d at level %d\n", rail->Name, state->bcast, lvl);
		    else if ((state->LNR & upmask) == upmask && (lstate->LNR & upmask) == upmask)
			printk ("%s: no way up to switch at level %d (turned off ?)\n", rail->Name, lvl+1);
		}
		else
		{
		    if (state->linkid != mylink)
			printk ("%s: moved at top level was connected to link %d now connected to %d\n", rail->Name, mylink, state->linkid);
		}

		if (state->linkid != mylink)
		    error++;
		
		if (BCAST_TOP_INVALID (lvl, state->bcast, ndown))
		    binval++;
	    }
	    else
	    {
		if (state->linkid != mylink)
		{
		    if (state->linkid != rail->SwitchLast[lvl].linkid)
			printk ("%s: moved at lvl %d was connected to link %d now connected to %d\n", rail->Name, lvl, mylink, state->linkid);
			
		    error++;
		}
	    }

	    if (error == 0 && invalid == 0)
		rail->SwitchProbeTick[lvl] = lbolt;
	    
	    EPRINTF10 (DBG_ROUTETABLE, "%s:   lvl=%d (slvl=%d) linkid=%d bcast=%d lnr=%02x uplink=%d : error=%d binval=%d invalid=%d\n", 
		       rail->Name, lvl, slvl, state->linkid, state->bcast, state->LNR, state->uplink, error, binval, invalid);

	    invalid |= (error | binval);
	}
	
	for (lvl = 0; lvl < rail->SwitchProbeLevel; lvl++)
	    if (rail->SwitchState[lvl].uplink != rail->SwitchLast[lvl].uplink)
		changed++;

	if (changed)
	{
	    printk ("%s: broadcast tree has changed from", rail->Name);
	    for (lvl = 0; lvl < rail->SwitchProbeLevel; lvl++)
		printk ("%c%d", lvl == 0 ? ' ' : ',', rail->SwitchLast[lvl].uplink);

	    for (lvl = 0; lvl < rail->SwitchProbeLevel; lvl++)
		printk ("%s%d", lvl == 0 ? " to " : ",", rail->SwitchState[lvl].uplink);
	    printk ("\n");
	}

	if (rail->SwitchProbeLevel > 0)
	    bcopy (rail->SwitchState, rail->SwitchLast, rail->SwitchProbeLevel * sizeof (EP_SWITCHSTATE));
    }

    for (lvl = 0; lvl < pos->pos_levels; lvl++)
    {
	EPRINTF4 (DBG_ROUTETABLE, "%s: level %d lbolt=%lx ProbeLevelTick=%lx\n",
		  rail->Name, lvl, lbolt, rail->SwitchProbeTick[lvl]);
	
	if (AFTER (lbolt, rail->SwitchProbeTick[lvl] + EP_POSITION_TIMEOUT))
	{
	    if (lvl < rail->SwitchBroadcastLevel+1)
	    {
		if (lvl == 0)
		    printk ("%s: cable disconnected\n", rail->Name);
		else
		    printk ("%s: broadcast level has dropped to %d (should be %d)\n",
			    rail->Name, lvl, rail->Position.pos_levels);
	    }
	    break;
	}
    }
    
    if (lvl > rail->SwitchBroadcastLevel+1)
    {
	if (rail->SwitchBroadcastLevel < 0)
	    printk ("%s: cable reconnected\n", rail->Name);
	if (lvl == rail->Position.pos_levels)
	    printk ("%s: broadcast level has recovered\n", rail->Name);
	else
	    printk ("%s: broadcast level has recovered to %d (should be %d)\n", 
		    rail->Name, lvl, rail->Position.pos_levels);
    }
    
    if (rail->SwitchBroadcastLevel != (lvl - 1))
    {
	EPRINTF2 (DBG_ROUTETABLE, "%s: setting SwitchBroadcastLevel to %d\n", rail->Name, lvl-1);
	
	rail->SwitchBroadcastLevel     = lvl - 1;
	rail->SwitchBroadcastLevelTick = lbolt;
    }
}


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
