/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: probenetwork_elan3.c,v 1.41 2005/04/26 09:36:19 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/probenetwork_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "debug.h"

#include <elan3/intrinsics.h>

static void ep3_probe_event (EP3_RAIL *rail, void *arg);
static EP3_COOKIE_OPS ep3_probe_ops = 
{
    ep3_probe_event
} ;

int
ep3_init_probenetwork (EP3_RAIL *rail)
{
    sdramaddr_t	      stack;
    E3_Addr           sp;
    E3_BlockCopyEvent event;
    int               i;

    if (! (stack = ep_alloc_elan (&rail->Generic, EP3_STACK_SIZE, 0, &rail->ProbeStack)))
	return -ENOMEM;

    spin_lock_init (&rail->ProbeLock);
    kcondvar_init (&rail->ProbeWait);

    /* Initialise the probe command structure */
    for (i = 0; i < TR_TRACEROUTE_ENTRIES; i++)
	elan3_sdram_writew (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeSource0[i]), 0);
    for (i = 0; i < TR_TRACEROUTE_ENTRIES; i++)
	elan3_sdram_writew (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeSource1[i]), 1);
    
    RegisterCookie (&rail->CookieTable, &rail->ProbeCookie, rail->RailElanAddr + offsetof (EP3_RAIL_ELAN, ProbeDone), &ep3_probe_ops, rail);
    
    elan3_sdram_writel (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeStart.ev_Type), 0);
    elan3_sdram_writel (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeStart.ev_Count), 0);

    EP3_INIT_COPY_EVENT (event, rail->ProbeCookie, rail->RailMainAddr + offsetof (EP3_RAIL_MAIN, ProbeDone), 1);
    elan3_sdram_copyl_to_sdram (rail->Device, &event, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeDone), sizeof (E3_BlockCopyEvent));

    rail->RailMain->ProbeDone = EP3_EVENT_FREE;

    sp = ep3_init_thread (rail->Device, ep_symbol (&rail->ThreadCode, "kcomm_probe"),
			  rail->ProbeStack, stack, EP3_STACK_SIZE,
			  3, rail->CommandPortAddr, rail->RailElanAddr, rail->RailMainAddr);
    
    IssueRunThread (rail, sp);

    return 0;
}

void
ep3_destroy_probenetwork (EP3_RAIL *rail)
{
    if (rail->ProbeStack == (sdramaddr_t) 0)
	return;

    /* XXXX: ensure that the network probe thread is stopped */

    DeregisterCookie (&rail->CookieTable, &rail->ProbeCookie);

    kcondvar_destroy (&rail->ProbeWait);
    spin_lock_destroy (&rail->ProbeLock);
    
    ep_free_elan (&rail->Generic, rail->ProbeStack, EP3_STACK_SIZE);
}

static void
ep3_probe_event (EP3_RAIL *rail, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->ProbeLock, flags);
    rail->ProbeDone = 1;
    kcondvar_wakeupone (&rail->ProbeWait, &rail->ProbeLock);
    spin_unlock_irqrestore (&rail->ProbeLock, flags);
}

int
ep3_probe_route (EP_RAIL *r, int level, int sw, int nodeid, int *linkup, int *linkdown, int attempts, EP_SWITCH *lsw)
{
    EP3_RAIL      *rail     = (EP3_RAIL *) r;
    EP3_RAIL_MAIN *railMain = rail->RailMain;
    sdramaddr_t    railElan = rail->RailElan;
    E3_uint16      flits[MAX_FLITS];
    E3_uint32      result;
    int		   nflits;
    unsigned long  flags;

    spin_lock_irqsave (&rail->ProbeLock, flags);

    nflits = GenerateProbeRoute ( flits, nodeid, level, linkup, linkdown, 0);
	    
    if (LoadRoute (rail->Device, rail->RouteTable, EP_VP_PROBE(level), ELAN3_MRF_CONTEXT_NUM|SYS_CONTEXT_BIT, nflits, flits) != 0)
    {
	EPRINTF0 (DBG_ROUTETABLE, "ProbeRoute: cannot load route entry\n");
	spin_unlock_irqrestore (&rail->ProbeLock, flags);
	return (EINVAL);
    }

    do {
	/* Initialise the probe source to include our partially computed nodeid */
	elan3_sdram_writew (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeSource0[TR_TRACEROUTE_ENTRIES-1]), nodeid);
	elan3_sdram_writew (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeSource1[TR_TRACEROUTE_ENTRIES-1]), nodeid);

	/* Initialise the count result etc */
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeType), PROBE_SINGLE);
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeLevel), level);

	railMain->ProbeResult  = -1;
	    
	/* Clear the receive area */
	bzero (railMain->ProbeDest0, sizeof (railMain->ProbeDest0));
	bzero (railMain->ProbeDest1, sizeof (railMain->ProbeDest1));
    
	/* Re-arm the completion event */
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeDone.ev_Count), 1);
	railMain->ProbeDone = EP3_EVENT_ACTIVE;
	rail->ProbeDone = 0;

	/* And wakeup the thread to do the probe */
	IssueSetevent (rail, rail->RailElanAddr + offsetof (EP3_RAIL_ELAN, ProbeStart));

	/* Now wait for it to complete */
	while (! rail->ProbeDone)
	    kcondvar_wait (&rail->ProbeWait, &rail->ProbeLock, &flags);

	/* wait for block copy event to flush write buffers */
	while (! EP3_EVENT_FIRED (rail->ProbeCookie, railMain->ProbeDone))
	    if (! EP3_EVENT_FIRING(rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeDone), rail->ProbeCookie, railMain->ProbeDone))
		panic ("ProbeRoute: network probe event failure\n");

	result = railMain->ProbeResult;

	if (result == C_ACK_ERROR)
	    kcondvar_timedwait (&rail->ProbeWait, &rail->ProbeLock, &flags, lbolt + (hz/8));
	
	railMain->ProbeDone = EP3_EVENT_FREE;

    } while (result != C_ACK_OK && --attempts);

    if (result == C_ACK_OK)
    {
	if (railMain->ProbeDest0[TR_TRACEROUTE_ENTRIES - ((2*level)+1) - 1] != nodeid ||
	    railMain->ProbeDest1[TR_TRACEROUTE_ENTRIES - ((2*level)+1) - 1] != nodeid)
	{
	    static unsigned long printed = 0;
	    if ((lbolt - printed) > (HZ*10))
	    {
		printk ("%s: lost nodeid at level %d switch %d - %d != %x\n", rail->Generic.Name, level, sw,
			railMain->ProbeDest0[TR_TRACEROUTE_ENTRIES - ((2*level)+1) - 1], nodeid);
		printed = lbolt;
	    }
	    result = C_ACK_ERROR;
	}
	else
	{
	    E3_uint16 val0 = railMain->ProbeDest0[TR_TRACEROUTE_ENTRIES - level - 1];
	    E3_uint16 val1 = railMain->ProbeDest1[TR_TRACEROUTE_ENTRIES - level - 1];
		
	    EPRINTF7 (DBG_PROBE, "%s: level %d switch %d - linkid=%d bcast=%d LNR=%02x%s\n", 
		      rail->Generic.Name, level, sw, TR_TRACEROUTE0_LINKID(val0),
		      TR_TRACEROUTE1_BCAST_TOP(val1), TR_TRACEROUTE0_LNR(val0),
		      TR_TRACEROUTE0_REVID(val0) ? "" : " RevA Part");
	    
	    lsw->lnr     = TR_TRACEROUTE0_LNR(val0);
	    lsw->link    = TR_TRACEROUTE0_LINKID(val0);
	    lsw->bcast   = TR_TRACEROUTE1_BCAST_TOP(val1);
	    lsw->invalid = (TR_TRACEROUTE0_REVID(val0) == 0);
	}
    }
    spin_unlock_irqrestore (&rail->ProbeLock, flags);
    
    return (result == C_ACK_OK);
}

void
ep3_probe_position_found (EP3_RAIL *rail, ELAN_POSITION *pos)
{
    E3_uint16  flits[MAX_FLITS];
    int        lvl, nflits;
    
    for (lvl = 0; lvl < pos->pos_levels; lvl++)
    {
	nflits = GenerateCheckRoute (pos, flits, pos->pos_levels - lvl - 1, 0);

	if (LoadRoute (rail->Device, rail->Ctxt->RouteTable, EP_VP_PROBE(lvl), ELAN3_MRF_CONTEXT_NUM|SYS_CONTEXT_BIT, nflits, flits) != 0)
	    panic ("ep3_probe_position_found: cannot load probe route entry\n");
    }
    
    /* Initialise the traceroute source data with our nodeid */
    elan3_sdram_writew (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeSource0[TR_TRACEROUTE_ENTRIES-1]), pos->pos_nodeid);
    elan3_sdram_writew (rail->Device, rail->RailElan + offsetof (EP3_RAIL_ELAN, ProbeSource1[TR_TRACEROUTE_ENTRIES-1]), pos->pos_nodeid);
}

int
ep3_check_position (EP_RAIL *r)
{
    EP3_RAIL      *rail     = (EP3_RAIL *) r;
    EP3_RAIL_MAIN *railMain = rail->RailMain;
    sdramaddr_t    railElan = rail->RailElan;
    ELAN_POSITION *pos      = &rail->Generic.Position;
    unsigned int   level    = rail->RailMain->ProbeLevel;
    unsigned int   updated  = EP3_EVENT_FIRED (rail->ProbeCookie, railMain->ProbeDone);
    unsigned int   lvl;

    if (updated)
    {
	if (railMain->ProbeResult != C_ACK_OK)
	{
	    EPRINTF2 (DBG_PROBE, "%s: CheckNetworkPosition: packet nacked result=%d\n", rail->Generic.Name, railMain->ProbeResult); 
	    
	    rail->Generic.SwitchProbeLevel = -1;
	}
	else
	{
	    E3_uint16 val0 = railMain->ProbeDest0[TR_TRACEROUTE_ENTRIES - 2*(level+1)];
	    E3_uint16 val1 = railMain->ProbeDest1[TR_TRACEROUTE_ENTRIES - 2*(level+1)];

	    if (val0 != pos->pos_nodeid || val1 != pos->pos_nodeid)
	    {
		static unsigned long printed = 0;

		/* We've received a packet from another node - this probably means
		 * that we've moved */
		if ((lbolt - printed) > (HZ*10))
		{
		    printk ("%s: ep3_check_position - level %d lost nodeid\n", rail->Generic.Name, level);
		    printed = lbolt;
		}

		rail->Generic.SwitchProbeLevel = -1;
	    }
	    else
	    {
		for (lvl = 0; lvl <= level; lvl++)
		{
		    E3_uint16 val0 = railMain->ProbeDest0[TR_TRACEROUTE_ENTRIES - ((2*level) - lvl + 1)];
		    E3_uint16 val1 = railMain->ProbeDest1[TR_TRACEROUTE_ENTRIES - ((2*level) - lvl + 1)];

		    rail->Generic.SwitchState[lvl].linkid = TR_TRACEROUTE0_LINKID(val0);
		    rail->Generic.SwitchState[lvl].LNR    = TR_TRACEROUTE0_LNR(val0);
		    rail->Generic.SwitchState[lvl].bcast  = TR_TRACEROUTE1_BCAST_TOP(val1);
		    rail->Generic.SwitchState[lvl].uplink = 4;

		    EPRINTF5 (DBG_PROBE, " --- lvl %d: linkid=%d LNR=%x bcast=%d uplink=%d\n", lvl, rail->Generic.SwitchState[lvl].linkid,
			      rail->Generic.SwitchState[lvl].LNR, rail->Generic.SwitchState[lvl].bcast ,rail->Generic.SwitchState[lvl].uplink);
		}
		rail->Generic.SwitchProbeLevel = level;
	    }
	}

	railMain->ProbeDone = EP3_EVENT_FREE;
    }

    if (railMain->ProbeDone == EP3_EVENT_FREE)
    {
	if (rail->Generic.SwitchBroadcastLevel == rail->Generic.Position.pos_levels-1)
	    level = rail->Generic.Position.pos_levels - 1;
	else
	    level = rail->Generic.SwitchBroadcastLevel + 1;

	EPRINTF2 (DBG_PROBE, "%s: ep3_check_postiion: level %d\n", rail->Generic.Name, level);

	/* Initialise the count result etc */
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeType), PROBE_MULTIPLE);
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeLevel), level);

	railMain->ProbeResult = -1;
	railMain->ProbeLevel  = -1;
	
	/* Clear the receive area */
	bzero (railMain->ProbeDest0, sizeof (railMain->ProbeDest0));
	bzero (railMain->ProbeDest1, sizeof (railMain->ProbeDest1));
	
	/* Re-arm the completion event */
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeDone.ev_Type), EV_TYPE_BCOPY);
	elan3_sdram_writel (rail->Device, railElan + offsetof (EP3_RAIL_ELAN, ProbeDone.ev_Count), 1);

	railMain->ProbeDone = EP3_EVENT_ACTIVE;
	
	IssueSetevent (rail, rail->RailElanAddr + offsetof (EP3_RAIL_ELAN, ProbeStart));
    }

    return updated;
}

