/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: probenetwork_elan3_thread.c,v 1.19 2004/03/24 11:32:56 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/probenetwork_elan3_thread.c,v $*/

#include <elan3/e3types.h>
#include <elan3/events.h>
#include <elan3/elanregs.h>
#include <elan3/intrinsics.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"

static int
kcomm_probe_vp (EP3_RAIL_ELAN *railElan, EP3_RAIL_MAIN *railMain, int vp, int attempts, int timeouts)
{
    int rc;

    /* Since we use %g1 to hold the "rxd" so the trap handler can
     * complete the envelope processing - we pass zero to indicate we're
     * not a receiver thread */
    asm volatile ("mov %g0, %g1");

    while (attempts && timeouts)
    {
	c_open (vp);
	c_sendmem (TR_TRACEROUTE, &railMain->ProbeDest0, &railElan->ProbeSource0);
	c_sendmem (TR_TRACEROUTE, &railMain->ProbeDest1, &railElan->ProbeSource1);
	c_sendtrans0 (TR_SENDACK | TR_SETEVENT, (E3_Addr) 0);
	
	switch (rc = c_close())
	{
	case C_ACK_OK:
	    return (C_ACK_OK);
	    
	case C_ACK_DISCARD:
	    attempts--;
	    break;

	default:					/* output timeout */
	    timeouts--;
	}

	c_break_busywait();
    }

    return (timeouts == 0 ? C_ACK_ERROR : C_ACK_DISCARD);
}

void
kcomm_probe (E3_CommandPort *cport, EP3_RAIL_ELAN *railElan, EP3_RAIL_MAIN *railMain)
{
    int level;

    for (;;)
    {
	c_waitevent (&railElan->ProbeStart, 1);

	switch (railElan->ProbeType)
	{
	case PROBE_SINGLE:
	    railMain->ProbeResult = kcomm_probe_vp (railElan, railMain, EP_VP_PROBE(railElan->ProbeLevel),
						    PROBE_SINGLE_ATTEMPTS, PROBE_SINGLE_TIMEOUTS);

	    cport->SetEvent = (E3_Addr) &railElan->ProbeDone;
	    break;

	case PROBE_MULTIPLE:
	    for (level = railElan->ProbeLevel; level >= 0; level--)
	    {
		if (kcomm_probe_vp (railElan, railMain, EP_VP_PROBE(level),
				    PROBE_MULTIPLE_ATTEMPTS, PROBE_MULTIPLE_TIMEOUTS) == C_ACK_OK)
		{
		    railMain->ProbeLevel  = level;
		    railMain->ProbeResult = C_ACK_OK;
		    break;
		}

		c_break_busywait();
	    }
	    cport->SetEvent = (E3_Addr) &railElan->ProbeDone;
	    break;
	}

    }
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
