/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: probenetwork_elan4.c,v 1.10.2.1 2005/07/20 11:35:37 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/probenetwork_elan4.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "debug.h"

#include <elan4/trtype.h>
#include <elan4/commands.h>

static void
probe_interrupt (EP4_RAIL *rail, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->r_probe_lock, flags);
    rail->r_probe_done = 1;
    kcondvar_wakeupone (&rail->r_probe_wait, &rail->r_probe_lock);
    spin_unlock_irqrestore (&rail->r_probe_lock, flags);
}

int
ep4_probe_init (EP4_RAIL *rail)
{
    spin_lock_init (&rail->r_probe_lock);
    kcondvar_init (&rail->r_probe_wait);

    rail->r_probe_cq = ep4_alloc_ecq (rail, CQ_Size1K);

    if (rail->r_probe_cq == NULL)
	return -ENOMEM;

    ep4_register_intcookie (rail, &rail->r_probe_intcookie, rail->r_elan_addr, probe_interrupt, rail);

    return 0;
}

void
ep4_probe_destroy (EP4_RAIL *rail)
{
    if (rail->r_probe_cq)
	ep4_free_ecq (rail, rail->r_probe_cq);

    if (rail->r_probe_intcookie.int_arg == NULL)
	return;
    ep4_deregister_intcookie (rail, &rail->r_probe_intcookie);

    kcondvar_destroy (&rail->r_probe_wait);
    spin_lock_destroy (&rail->r_probe_lock);
}

#define LINKDOWN(nodeid, level)	((nodeid >> (level << 1)) & 3)
#define PROBE_PATTERN0(nodeid)	(0xaddebabe ^ nodeid)
#define PROBE_PATTERN1(nodeid)  (0xfeedbeef ^ nodeid)

#define EP4_PROBE_RETRIES	4

int
ep4_probe_route (EP_RAIL *r, int level, int sw, int nodeid, int *linkup, int *linkdown, int attempts, EP_SWITCH *lsw)
{
    EP4_RAIL      *rail  = (EP4_RAIL *) r;
    EP4_RAIL_MAIN *rmain = rail->r_main;
    E4_uint16      first = 0;
    int	           rb    = 0;

    E4_uint8  packed[ROUTE_NUM_PACKED];
    E4_VirtualProcessEntry route;
    unsigned long flags;
    int i;

    for (i = 0; i < ROUTE_NUM_PACKED; i++)
	packed[i] = 0;

    /* Generate "up" routes */
    for (i = 0; i < level; i++)
	if (first == 0)
	    first = linkup ? FIRST_ROUTE(linkup[i]) : FIRST_ADAPTIVE;
	else
	    packed[rb++] = linkup ? PACKED_ROUTE(linkup[i]) : PACKED_ADAPTIVE;
    
    /* Generate a "to-me" route down */
    if (first == 0)
	first = FIRST_MYLINK;
    else
	packed[rb++] = PACKED_MYLINK;
    
    /* Generate the "down" routes */
    for (i = level-1; i >= 0; i--)
	packed[rb++] = linkdown ? PACKED_ROUTE(linkdown[i]) : PACKED_ROUTE(LINKDOWN(nodeid, i));
    
    /* Pack up the routes into the virtual process entry */
    route.Values[0] = first | FIRST_HIGH_PRI | FIRST_SYSTEM_PACKET | FIRST_TIMEOUT(3);
    route.Values[1] = ROUTE_CTXT_VALUE(ELAN4_KCOMM_CONTEXT_NUM);

    for (i = 0; i < (ROUTE_NUM_PACKED >> 1); i++)
    {
	route.Values[0] |= ((E4_uint64) packed[i]) << ((i << 2) + ROUTE_PACKED_OFFSET);
	route.Values[1] |= ((E4_uint64) packed[i+(ROUTE_NUM_PACKED >> 1)]) << ((i << 2));
    }

    elan4_write_route (rail->r_ctxt.ctxt_dev, rail->r_routetable, EP_VP_PROBE(level), &route);
    
    while (attempts--)
    {
	rail->r_probe_done = 0;

	/* generate the STEN packet - note we use a datatype of dword as we're copying to elan in dwords
	 *   NB - no flow control is required, since the max packet size is less than the command queue
	 *        size and it's dedicated for network probing.
	 */
	
	elan4_guard   (rail->r_probe_cq->ecq_cq, GUARD_CHANNEL(1) | GUARD_RESET(EP4_PROBE_RETRIES));
	elan4_nop_cmd (rail->r_probe_cq->ecq_cq, 0);
	
	elan4_open_packet (rail->r_probe_cq->ecq_cq, OPEN_STEN_PKT_CMD | OPEN_PACKET(0, PACK_OK | RESTART_COUNT_ZERO, EP_VP_PROBE(level)));
	elan4_sendtransn  (rail->r_probe_cq->ecq_cq, TR_TRACEROUTE(TRACEROUTE_NDWORDS),
			   rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_dest0),
			   0x0000000000000000ull, 0x0000000000000000ull, 0x0000000000000000ull, 0x0000000000000000ull, 
			   0x0000000000000000ull, 0x0000000000000000ull, 0x0000000000000000ull, 0x0000000000000000ull | ((E4_uint64)PROBE_PATTERN0(nodeid) << 32));
	elan4_sendtransn  (rail->r_probe_cq->ecq_cq, TR_TRACEROUTE(TRACEROUTE_NDWORDS),
			   rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_dest1),
			   0x0000000100000001ull, 0x0000000100000001ull, 0x0000000100000001ull, 0x0000000100000001ull, 
			   0x0000000100000001ull, 0x0000000100000001ull, 0x0000000100000001ull, 0x0000000000000001ull | ((E4_uint64)PROBE_PATTERN1(nodeid) << 32));
	elan4_sendtrans0  (rail->r_probe_cq->ecq_cq, TR_NOP_TRANS | TR_LAST_AND_SEND_ACK, 0);

	elan4_guard           (rail->r_probe_cq->ecq_cq, GUARD_CHANNEL(1) | GUARD_TEST(0, PACK_OK) | GUARD_RESET(EP4_PROBE_RETRIES));
	elan4_write_dword_cmd (rail->r_probe_cq->ecq_cq, rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_result), EP4_STATE_FINISHED);

	elan4_guard            (rail->r_probe_cq->ecq_cq, GUARD_CHANNEL(1) | GUARD_TEST(0, RESTART_COUNT_ZERO) | GUARD_RESET(EP4_PROBE_RETRIES));
	elan4_write_dword_cmd  (rail->r_probe_cq->ecq_cq, rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_result), EP4_STATE_FAILED);

	elan4_interrupt_cmd   (rail->r_probe_cq->ecq_cq,  rail->r_probe_intcookie.int_val);

	spin_lock_irqsave (&rail->r_probe_lock, flags);
	while (! rail->r_probe_done)
	    kcondvar_wait (&rail->r_probe_wait, &rail->r_probe_lock, &flags);
	spin_unlock_irqrestore (&rail->r_probe_lock, flags);

	if (rmain->r_probe_result == EP4_STATE_FINISHED)
	{
	    if (rmain->r_probe_dest0[TRACEROUTE_ENTRIES - ((2*level)+1) - 1] != PROBE_PATTERN0(nodeid) ||
		rmain->r_probe_dest1[TRACEROUTE_ENTRIES - ((2*level)+1) - 1] != PROBE_PATTERN1(nodeid))
	    {
		static unsigned long printed = 0;
		if ((lbolt - printed) > (HZ*10))
		{
		    printk ("%s: lost nodeid at level %d switch %d - %d != %x\n", rail->r_generic.Name, level, sw,
			    rmain->r_probe_dest0[TRACEROUTE_ENTRIES - ((2*level)+1) - 1], PROBE_PATTERN0(nodeid));
		    printed = lbolt;
		}
	    }
	    else
	    {
		E4_uint32 val0 = rmain->r_probe_dest0[TRACEROUTE_ENTRIES - level - 1];
		E4_uint32 val1 = rmain->r_probe_dest1[TRACEROUTE_ENTRIES - level - 1];
		
		lsw->lnr     = TR_TRACEROUTE0_LNR(val0);
		lsw->link    = TR_TRACEROUTE0_LINKID(val0);
		lsw->bcast   = TR_TRACEROUTE1_BCAST_TOP(val1);
		lsw->invalid = 0;

		return 1;
	    }
	}

	rmain->r_probe_result = EP4_STATE_FREE;
    }

    return 0;
}


void
ep4_probe_position_found (EP4_RAIL *rail, ELAN_POSITION *pos)
{
    ELAN4_DEV  *dev  = rail->r_ctxt.ctxt_dev;
    int         lvl;

    for (lvl = 0; lvl < pos->pos_levels; lvl++)
    {
	/* Initialise the "probe" route to use the broadcast tree */
	ELAN_POSITION *pos     = &rail->r_generic.Position;
	unsigned char *arityp  = &pos->pos_arity[pos->pos_levels - 1];
	unsigned int   spanned = *arityp;
	E4_uint16      first   = 0;
	int	       rb      = 0;
	
	E4_uint8  packed[ROUTE_NUM_PACKED];
	E4_VirtualProcessEntry route;
	int i;
	
	for (i = 0; i < ROUTE_NUM_PACKED; i++)
	    packed[i] = 0;

	/* Generate "up" routes */
	for (i = 0; i < lvl; i++, spanned *= *(--arityp))
	{
	    if (first == 0)
		first = FIRST_BCAST_TREE;
	    else
		packed[rb++] = PACKED_BCAST_TREE;
	}

	/* Generate a "to-me" route down */
	if (first == 0)
	    first = FIRST_MYLINK;
	else
	    packed[rb++] = PACKED_MYLINK;

	spanned /= *arityp++;

	/* Generate the "down" routes */
	for (i = lvl-1; i >= 0; i--)
	{
	    spanned /= *arityp;
	    packed[rb++] = PACKED_ROUTE((pos->pos_nodeid / spanned) % *arityp);
	    arityp++;
	}

    
	/* Pack up the routes into the virtual process entry */
	route.Values[0] = first | FIRST_HIGH_PRI | FIRST_SYSTEM_PACKET | FIRST_TIMEOUT(3);
	route.Values[1] = ROUTE_CTXT_VALUE(ELAN4_KCOMM_CONTEXT_NUM);
	
	for (i = 0; i < (ROUTE_NUM_PACKED >> 1); i++)
	{
	    route.Values[0] |= ((E4_uint64) packed[i]) << ((i << 2) + ROUTE_PACKED_OFFSET);
	    route.Values[1] |= ((E4_uint64) packed[i+(ROUTE_NUM_PACKED >> 1)]) << ((i << 2));
	}
	
	elan4_write_route (rail->r_ctxt.ctxt_dev, rail->r_routetable, EP_VP_PROBE(lvl), &route);
	
	/* Initialise "start" event for this level */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_start[lvl].ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_CHECK_STEN_NDWORDS));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_start[lvl].ev_CopySource),
			    rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl]));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_start[lvl].ev_CopyDest),
			    rail->r_probe_cq->ecq_addr);

	/* Initiailise command stream - reset the start event */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_reset_event_cmd),
			    WRITE_DWORD_CMD | (rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_check_start[lvl])));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_reset_event_value),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, EP4_CHECK_STEN_NDWORDS));

	/* Initiailise command stream - sten traceroute packet */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_open),
			    OPEN_STEN_PKT_CMD | OPEN_PACKET (0, PACK_OK | RESTART_COUNT_ZERO, EP_VP_PROBE(lvl)));

	/* Initiailise command stream - traceroute 0 */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_trans_traceroute0),
			    SEND_TRANS_CMD | (TR_TRACEROUTE(TRACEROUTE_NDWORDS) << 16));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_addr_traceroute0),
			    rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_dest0));
	for (i = 0; i < (TRACEROUTE_NDWORDS-1); i++)
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_data_traceroute0[i]),
				0x0000000000000000ull);
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_data_traceroute0[i]),
			    0x0000000000000000ull | ((E4_uint64) PROBE_PATTERN0(pos->pos_nodeid) << 32));

	/* Initiailise command stream - traceroute 1 */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_trans_traceroute1),
			    SEND_TRANS_CMD | (TR_TRACEROUTE(TRACEROUTE_NDWORDS) << 16));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_addr_traceroute1),
			    rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_dest1));
	for (i = 0; i < (TRACEROUTE_NDWORDS-1); i++)
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_data_traceroute1[i]),
				0x0000000100000001ull);
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_data_traceroute1[i]),
			    0x0000000000000001ull | ((E4_uint64) PROBE_PATTERN1(pos->pos_nodeid) << 32));

	/* Initiailise command stream - null sendack */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_trans_sendack),
			    SEND_TRANS_CMD | ((TR_NOP_TRANS | TR_LAST_AND_SEND_ACK) << 16));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_addr_sendack),
			    0);
	
	/* Initiailise command stream - guard ok, write done */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_guard_ok),
			    GUARD_CMD | GUARD_CHANNEL(1) | GUARD_TEST(0, PACK_OK) | GUARD_RESET(EP4_PROBE_RETRIES));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_writedword_ok),
			    WRITE_DWORD_CMD | (rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_level)));
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_value_ok),
			    lvl);

	/* Initiailise command stream - guard fail, chain to next or write done */
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_guard_fail),
			    GUARD_CMD | GUARD_CHANNEL(1) | GUARD_TEST(0, RESTART_COUNT_ZERO) | GUARD_RESET(EP4_PROBE_RETRIES));

	if (lvl > 0)
	{
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_setevent_fail),
				SET_EVENT_CMD | (rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_check_start[lvl-1])));
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_setevent_nop),
				NOP_CMD);
	}
	else
	{
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_setevent_fail),
				WRITE_DWORD_CMD | (rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_probe_level)));
	    elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_setevent_nop),
				EP4_PROBE_FAILED);
	}
	elan4_sdram_writeq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_check_sten[lvl].c_nop_pad),
			    NOP_CMD);
    }

    
    rail->r_main->r_probe_level = EP4_PROBE_ACTIVE;

    mb();
    ep4_set_event_cmd (rail->r_probe_cq, rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_check_start[pos->pos_levels-1]));
}

int
ep4_check_position (EP_RAIL *r)
{
    EP4_RAIL      *rail = (EP4_RAIL *) r;
    ELAN_POSITION *pos  = &rail->r_generic.Position;
    unsigned int level  = rail->r_main->r_probe_level;
    unsigned int lvl;

    EPRINTF2 (DBG_PROBE, "%s: ep4_check_position: level=%lld\n", rail->r_generic.Name, (long long)rail->r_main->r_probe_level);

    if (rail->r_main->r_probe_level != EP4_PROBE_ACTIVE)
    {
	if (rail->r_main->r_probe_level == EP4_PROBE_FAILED)
	{
	    EPRINTF1 (DBG_PROBE, "%s: ep4_check_position: packets all nacked\n", rail->r_generic.Name);

	    rail->r_generic.SwitchProbeLevel = -1;
	}
	else
	{
	    E4_uint32 val0  = rail->r_main->r_probe_dest0[TRACEROUTE_ENTRIES - 2*(level+1)];
	    E4_uint32 val1  = rail->r_main->r_probe_dest1[TRACEROUTE_ENTRIES - 2*(level+1)];

	    if (val0 != PROBE_PATTERN0 (pos->pos_nodeid) || val1 != PROBE_PATTERN1 (pos->pos_nodeid))
	    {
		static unsigned long printed = 0;

		/* We've received a packet from another node - this probably means
		 * that we've moved */
		if ((lbolt - printed) > (HZ*10))
		{
		    printk ("%s: ep4_check_position - level %d lost nodeid\n", rail->r_generic.Name, level);
		    printed = lbolt;
		}

		rail->r_generic.SwitchProbeLevel = -1;
	    }
	    else
	    {
		for (lvl = 0 ; lvl <= level; lvl++)
		{
		    E4_uint32 uval0  = rail->r_main->r_probe_dest0[TRACEROUTE_ENTRIES - lvl - 1];
		    E4_uint32 dval0  = rail->r_main->r_probe_dest0[TRACEROUTE_ENTRIES - ((2*level) - lvl + 1)];
		    E4_uint32 dval1  = rail->r_main->r_probe_dest1[TRACEROUTE_ENTRIES - ((2*level) - lvl + 1)];

		    rail->r_generic.SwitchState[lvl].linkid = TR_TRACEROUTE0_LINKID (dval0);
		    rail->r_generic.SwitchState[lvl].LNR    = TR_TRACEROUTE0_LNR(dval0);
		    rail->r_generic.SwitchState[lvl].bcast  = TR_TRACEROUTE1_BCAST_TOP (dval1);
		    rail->r_generic.SwitchState[lvl].uplink = TR_TRACEROUTE0_LINKID (uval0);

		    EPRINTF5 (DBG_PROBE, " --- lvl %d: linkid=%d LNR=%x bcast=%d uplink=%d\n", lvl, rail->r_generic.SwitchState[lvl].linkid,
			      rail->r_generic.SwitchState[lvl].LNR, rail->r_generic.SwitchState[lvl].bcast ,rail->r_generic.SwitchState[lvl].uplink);

		}

		rail->r_generic.SwitchProbeLevel = level;
	    }
	}

	rail->r_main->r_probe_level = EP4_PROBE_ACTIVE;
	mb();

	if (rail->r_generic.SwitchBroadcastLevel == rail->r_generic.Position.pos_levels-1)
	    level = rail->r_generic.Position.pos_levels - 1;
	else
	    level = rail->r_generic.SwitchBroadcastLevel + 1;

	ep4_set_event_cmd (rail->r_probe_cq, rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_check_start[level]));

	return 1;
    }

    return 0;
}
