/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: railhints.c,v 1.5 2004/02/06 22:37:06 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/railhints.c,v $*/

#include <qsnet/kernel.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "debug.h"

int
ep_pickRail(EP_RAILMASK railmask)
{
    static volatile int lastGlobal;
    int i, rnum, last = lastGlobal;

    /* Pick a single rail out of the railmask */
    for (i = 0; i < EP_MAX_RAILS; i++)
	if (railmask & (1 << ((last + i) % EP_MAX_RAILS)))
	    break;

    if (i == EP_MAX_RAILS)
	return (-1);

    rnum = (last + i) % EP_MAX_RAILS;

    lastGlobal = (rnum + 1) % EP_MAX_RAILS;

    ASSERT (railmask & (1 << rnum));

    return (rnum);
}

int
ep_xmtr_bcastrail (EP_XMTR *xmtr, EP_RAILMASK allowedRails)
{
    /* Retrun a single rail out of allowed mask with the best connectivity for broadcast. */
    return (ep_pickRail (allowedRails & xmtr->RailMask));
}

int
ep_xmtr_prefrail (EP_XMTR *xmtr, EP_RAILMASK allowedRails, unsigned nodeId)
{
    EP_NODE *node = &xmtr->Subsys->Subsys.Sys->Nodes[nodeId];

    EPRINTF5 (DBG_XMTR, "ep_xmtr_prefrail: xmtr=%p allowedRails=%x nodeId=%d xmtr->RailMaks=%x Connected=%x\n", 
	      xmtr, allowedRails, nodeId, xmtr->RailMask, node->ConnectedRails);

    /* Return a single rail which is currently connected to nodeId (limited to rails
     * in allowedmask) - if more than one rail is possible, then round-robin between 
     * them */
    return (ep_pickRail (allowedRails & xmtr->RailMask & node->ConnectedRails));
}

EP_RAILMASK
ep_xmtr_availrails (EP_XMTR *xmtr)
{
    /* Return which rails can be used to transmit one. */

    return (xmtr->RailMask);
}

EP_RAILMASK
ep_xmtr_noderails (EP_XMTR *xmtr, unsigned nodeId)
{
    EP_NODE *node = &xmtr->Subsys->Subsys.Sys->Nodes[nodeId];

    /* Return which rails can be used to transmit to this node. */

    return (xmtr->RailMask & node->ConnectedRails);
}

int
ep_rcvr_prefrail (EP_RCVR *rcvr, EP_RAILMASK allowedRails)
{
    /* Return the "best" rail for queueing a receive buffer out on - this will be a
     * rail with ThreadWaiting set or the rail with the least descriptors queued
     * on it. */
    
    return (ep_pickRail (allowedRails & rcvr->RailMask));
}

EP_RAILMASK
ep_rcvr_availrails (EP_RCVR *rcvr)
{
    /* Return which rails can be used to queue receive buffers. */
    return (rcvr->RailMask);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
