/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: statemap.c,v 1.12 2004/09/01 16:13:43 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/statemap.c,v $ */

#include <qsnet/kernel.h>
#include <elan/statemap.h>

/******************************** global state bitmap stuff **********************************/
static int
statemap_setmapbit (bitmap_t *map, int offset, int bit)
{
   bitmap_t *e    = &map[offset >> BT_ULSHIFT];
   bitmap_t  mask = ((bitmap_t)1) << (offset & BT_ULMASK);
   int       rc = ((*e) & mask) != 0;
   
   if (bit)
   {
      *e |= mask;
      return (!rc);
   }

   *e &= ~mask;
   return (rc);
}

static int
statemap_firstsegbit (bitmap_t seg)
{
   int            bit = 0;
   
   if (seg == 0)
      return (-1);

#if (BT_ULSHIFT == 6)
   if ((seg & 0xffffffffL) == 0)
   {
      seg >>= 32;
      bit += 32;
   }
#elif (BT_ULSHIFT != 5)
# error "Unexpected value of BT_ULSHIFT"
#endif

   if ((seg & 0xffff) == 0)
   {
      seg >>= 16;
      bit += 16;
   }
      
   if ((seg & 0xff) == 0)
   {
      seg >>= 8;
      bit += 8;
   }
      
   if ((seg & 0xf) == 0)
   {
      seg >>= 4;
      bit += 4;
   }
      
   if ((seg & 0x3) == 0)
   {
      seg >>= 2;
      bit += 2;
   }

   return (((seg & 0x1) == 0) ? bit + 1 : bit);
}

bitmap_t
statemap_getseg (statemap_t *map, unsigned int offset)
{
   ASSERT (offset < map->size);
   ASSERT ((offset & BT_ULMASK) == 0);

   return (map->bitmap[offset >> BT_ULSHIFT]);
}

void
statemap_setseg (statemap_t *map, unsigned int offset, bitmap_t seg)
{
   ASSERT (offset < map->size);
   ASSERT ((offset & BT_ULMASK) == 0);

   offset >>= BT_ULSHIFT;
   if (map->bitmap[offset] == seg)
      return;

   map->bitmap[offset] = seg;

   if (statemap_setmapbit (map->changemap2, offset,       1) &&
       statemap_setmapbit (map->changemap1, offset >>= BT_ULSHIFT, 1))
      statemap_setmapbit (map->changemap0, offset >>= BT_ULSHIFT, 1);
}

bitmap_t
statemap_getbits (statemap_t *map, unsigned int offset, int nbits)
{
   int      index = offset >> BT_ULSHIFT;
   bitmap_t mask  = (nbits == BT_NBIPUL) ? (bitmap_t) -1 : (((bitmap_t)1) << nbits) - 1;
   
   ASSERT (nbits <= BT_NBIPUL);
   ASSERT (offset + nbits <= map->size);

   offset &= BT_ULMASK;
   if (offset + nbits <= BT_NBIPUL)
      return ((map->bitmap[index] >> offset) & mask);
   
   return (((map->bitmap[index] >> offset) |
	    (map->bitmap[index + 1] << (BT_NBIPUL - offset))) & mask);
}

void
statemap_setbits (statemap_t *map, unsigned int offset, bitmap_t bits, int nbits)
{
   int      index = offset >> BT_ULSHIFT;
   bitmap_t mask;
   bitmap_t seg;
   bitmap_t newseg;

   ASSERT (nbits <= BT_NBIPUL);
   ASSERT (offset + nbits <= map->size);

   offset &= BT_ULMASK;
   if (offset + nbits <= BT_NBIPUL)
   {
      mask = ((nbits == BT_NBIPUL) ? -1 : ((((bitmap_t)1) << nbits) - 1)) << offset;
      seg = map->bitmap[index];
      newseg = ((bits << offset) & mask) | (seg & ~mask);
      
      if (seg == newseg)
	 return;
   
      map->bitmap[index] = newseg;
      
      if (statemap_setmapbit (map->changemap2, index,       1) &&
	  statemap_setmapbit (map->changemap1, index >>= BT_ULSHIFT, 1))
	 statemap_setmapbit (map->changemap0, index >>= BT_ULSHIFT, 1);
      return;
   }
   
   mask = ((bitmap_t)-1) << offset;
   seg = map->bitmap[index];
   newseg = ((bits << offset) & mask) | (seg & ~mask);

   if (seg != newseg)
   {
      map->bitmap[index] = newseg;
      
      if (statemap_setmapbit (map->changemap2, index,       1) &&
	  statemap_setmapbit (map->changemap1, index >> BT_ULSHIFT, 1))
	 statemap_setmapbit (map->changemap0, index >> (2 * BT_ULSHIFT), 1);
   }
   
   index++;
   offset = BT_NBIPUL - offset;
   mask = (((bitmap_t)1) << (nbits - offset)) - 1;
   seg = map->bitmap[index];
   newseg = ((bits >> offset) & mask) | (seg & ~mask);
   
   if (seg == newseg)
      return;
   
   map->bitmap[index] = newseg;
   
   if (statemap_setmapbit (map->changemap2, index,       1) &&
       statemap_setmapbit (map->changemap1, index >>= BT_ULSHIFT, 1))
      statemap_setmapbit (map->changemap0, index >>= BT_ULSHIFT, 1);
}

void
statemap_zero (statemap_t *dst)
{
   int       size       = dst->size;
   int       offset     = 0;
   bitmap_t *changemap0 = dst->changemap0;
   bitmap_t *changemap1 = dst->changemap1;
   bitmap_t *changemap2 = dst->changemap2;
   bitmap_t *dstmap     = dst->bitmap;
   bitmap_t  bit0;
   bitmap_t  bit1;
   bitmap_t  bit2;

   for (bit0 = 1; offset < size; bit0 <<= 1, changemap1++)
   {
      for (bit1 = 1; bit1 != 0 && offset < size; bit1 <<= 1, changemap2++)
      {
	 for (bit2 = 1; bit2 != 0 && offset < size; bit2 <<= 1, dstmap++, offset += BT_NBIPUL)
	 {
	     *dstmap = 0;
	     *changemap2 |= bit2;
	 }
	 *changemap1 |= bit1;
      }
      *changemap0 |= bit0;
   }
}
   
void
statemap_setmap (statemap_t *dst, statemap_t *src)
{
   int       size       = dst->size;
   int       offset     = 0;
   bitmap_t *changemap0 = dst->changemap0;
   bitmap_t *changemap1 = dst->changemap1;
   bitmap_t *changemap2 = dst->changemap2;
   bitmap_t *dstmap     = dst->bitmap;
   bitmap_t *srcmap     = src->bitmap;
   bitmap_t  bit0;
   bitmap_t  bit1;
   bitmap_t  bit2;

   ASSERT (src->size == size);
   
   for (bit0 = 1; offset < size; bit0 <<= 1, changemap1++)
   {
      for (bit1 = 1; bit1 != 0 && offset < size; bit1 <<= 1, changemap2++)
      {
	 for (bit2 = 1; bit2 != 0 && offset < size; bit2 <<= 1, dstmap++, srcmap++, offset += BT_NBIPUL)
	    if (*dstmap != *srcmap)
	    {
	       *dstmap = *srcmap;
	       *changemap2 |= bit2;
	    }
	 if (*changemap2 != 0)
	    *changemap1 |= bit1;
      }
      if (*changemap1 != 0)
	 *changemap0 |= bit0;
   }
}

void
statemap_ormap (statemap_t *dst, statemap_t *src)
{
   int       size       = dst->size;
   int       offset     = 0;
   bitmap_t *changemap0 = dst->changemap0;
   bitmap_t *changemap1 = dst->changemap1;
   bitmap_t *changemap2 = dst->changemap2;
   bitmap_t *dstmap     = dst->bitmap;
   bitmap_t *srcmap     = src->bitmap;
   bitmap_t  bit0;
   bitmap_t  bit1;
   bitmap_t  bit2;
   bitmap_t  seg;

   ASSERT (src->size == size);
   
   for (bit0 = 1; offset < size; bit0 <<= 1, changemap1++)
   {
      for (bit1 = 1; bit1 != 0 && offset < size; bit1 <<= 1, changemap2++)
      {
	 for (bit2 = 1; bit2 != 0 && offset < size; bit2 <<= 1, dstmap++, srcmap++, offset += BT_NBIPUL)
	 {
	    seg = *dstmap | *srcmap;
	    if (*dstmap != seg)
	    {
	       *dstmap = seg;
	       *changemap2 |= bit2;
	    }
	 }
	 if (*changemap2 != 0)
	    *changemap1 |= bit1;
      }
      if (*changemap1 != 0)
	 *changemap0 |= bit0;
   }
}

int
statemap_findchange (statemap_t *map, bitmap_t *newseg, int clearchange)
{
   int          bit0;
   bitmap_t    *cm1;
   int          bit1;
   bitmap_t    *cm2;
   int          bit2;
   unsigned int offset;

   bit0 = statemap_firstsegbit (*(map->changemap0));
   if (bit0 < 0)
      return (-1);

   offset = bit0;
   cm1 = map->changemap1 + offset;
   bit1 = statemap_firstsegbit (*cm1);
   ASSERT (bit1 >= 0);

   offset = (offset << BT_ULSHIFT) + bit1;
   cm2 = map->changemap2 + offset;
   bit2 = statemap_firstsegbit (*cm2);
   ASSERT (bit2 >= 0);
   
   offset = (offset << BT_ULSHIFT) + bit2;
   *newseg = map->bitmap[offset];

   if (clearchange &&
       (*cm2 &= ~(((bitmap_t)1) << bit2)) == 0 &&
       (*cm1 &= ~(((bitmap_t)1) << bit1)) == 0)
      map->changemap0[0] &= ~(((bitmap_t)1) << bit0);

   return (offset << BT_ULSHIFT);
}

int
statemap_changed (statemap_t *map)
{
   return ((*(map->changemap0) != 0));
}

void
statemap_reset (statemap_t *map)
{
   bzero (map->changemap0, map->changemap_nob + map->bitmap_nob);
}

void
statemap_copy (statemap_t *dst, statemap_t *src)
{
   ASSERT (dst->size == src->size);
   bcopy (src->changemap0, dst->changemap0, src->changemap_nob + src->bitmap_nob);
}

void
statemap_clearchanges (statemap_t *map)
{
   if (statemap_changed (map))
      bzero (map->changemap0, map->changemap_nob);
}

bitmap_t *
statemap_tobitmap (statemap_t *map)
{
    return (map->bitmap);
}

statemap_t *
statemap_create (int size)
{
   int   struct_entries     = (sizeof (statemap_t) * 8 + (BT_NBIPUL-1)) >> BT_ULSHIFT;
   int   bitmap_entries     = (size + (BT_NBIPUL-1)) >> BT_ULSHIFT;
   int   changemap2_entries = (bitmap_entries + (BT_NBIPUL-1)) >> BT_ULSHIFT;
   int   changemap1_entries = (changemap2_entries + (BT_NBIPUL-1)) >> BT_ULSHIFT;
   int   changemap0_entries = (changemap1_entries + (BT_NBIPUL-1)) >> BT_ULSHIFT;
   int   changemap_entries  = changemap0_entries + changemap1_entries + changemap2_entries;
   int   nob                = (struct_entries + bitmap_entries + changemap_entries) * sizeof (bitmap_t);
   statemap_t *map;

   ASSERT ((1 << BT_ULSHIFT) == BT_NBIPUL);
   ASSERT (changemap0_entries == 1);

   KMEM_ZALLOC (map, statemap_t *, nob, 1);

   map->size = size;
   map->nob  = nob;
   map->changemap_nob = changemap_entries * sizeof (bitmap_t);
   map->bitmap_nob = bitmap_entries * sizeof (bitmap_t);
   map->changemap0 = ((bitmap_t *)map) + struct_entries;
   map->changemap1 = map->changemap0 + changemap0_entries;
   map->changemap2 = map->changemap1 + changemap1_entries;
   map->bitmap     = map->changemap2 + changemap2_entries;

   return (map);
}

void
statemap_destroy (statemap_t *map)
{
   KMEM_FREE (map, map->nob);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
