/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: support.c,v 1.39 2004/09/30 14:59:15 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/support.c,v $ */

#include <qsnet/kernel.h>
#include <elan/kcomm.h>

/****************************************************************************************/
/*
 * Nodeset/flush callbacks.
 */
int
ep_register_callback (EP_RAIL *rail, unsigned idx, void (*routine)(void *, statemap_t *), void *arg)
{
    EP_CALLBACK *cb;
    
    KMEM_ALLOC (cb, EP_CALLBACK *, sizeof (EP_CALLBACK), 1);
    
    cb->Routine = routine;
    cb->Arg     = arg;

    kmutex_lock (&rail->CallbackLock);
    cb->Next = rail->CallbackList[idx];
    rail->CallbackList[idx] = cb;
    kmutex_unlock (&rail->CallbackLock);
    
    return (ESUCCESS);
}

void
ep_remove_callback (EP_RAIL *rail, unsigned idx, void (*routine)(void *, statemap_t *), void *arg)
{
    EP_CALLBACK  *cb;
    EP_CALLBACK **predp;

    kmutex_lock (&rail->CallbackLock);
    for (predp = &rail->CallbackList[idx]; (cb = *predp); predp = &cb->Next)
	if (cb->Routine == routine && cb->Arg == arg)
	    break;

    if (cb == NULL)
	panic ("ep_remove_member_callback");
    
    *predp = cb->Next;
    kmutex_unlock (&rail->CallbackLock);
    
    KMEM_FREE (cb, sizeof (EP_CALLBACK));
}

void
ep_call_callbacks (EP_RAIL *rail, unsigned idx, statemap_t *map)
{
    EP_CALLBACK *cb;

    kmutex_lock (&rail->CallbackLock);

    rail->CallbackStep = idx;

    for (cb = rail->CallbackList[idx]; cb; cb = cb->Next) {
	(cb->Routine) (cb->Arg, map);
    }
    kmutex_unlock (&rail->CallbackLock);
}

unsigned int
ep_backoff (EP_BACKOFF *backoff, int type)
{
    static int bcount[EP_NUM_BACKOFF] = {1, 16, 32, 64, 128, 256, 512, 1024};
    
    if (backoff->type != type)
    {
	backoff->type  = type;
	backoff->indx  = 0;
	backoff->count = 0;
    }

    if (++backoff->count > bcount[backoff->indx] && backoff->indx < (EP_NUM_BACKOFF-1))
    {
	backoff->indx++;
	backoff->count = 0;
    }

    return (backoff->indx);
}

/* Generic checksum algorithm */
uint16_t
CheckSum (char *msg, int nob)
{
    uint16_t sum = 0;
   
    while (nob-- > 0)
	sum = sum * 13 + *msg++;

    return (sum);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
