/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: support_elan3.c,v 1.47.2.1 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/support_elan3.c,v $ */

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan/kcomm.h>
#include <elan/epsvc.h>
#include <elan/epcomms.h>

#include "kcomm_vp.h"
#include "kcomm_elan3.h"
#include "epcomms_elan3.h"
#include "debug.h"

#include <elan3/thread.h>
#include <elan3/urom_addrs.h>

/****************************************************************************************/
#define DMA_RING_NEXT_POS(ring)      ((ring)->Position+1 == ring->Entries ? 0 : ((ring)->Position+1))
#define DMA_RING_PREV_POS(ring,pos)  ((pos) == 0 ? (ring)->Entries-1 : (pos) - 1)

static int 
DmaRingCreate (EP3_RAIL *rail, EP3_DMA_RING *ring, int ctxnum, int entries)
{
    unsigned long pgnum = (ctxnum * sizeof (E3_CommandPort)) / PAGE_SIZE;
    unsigned long pgoff = (ctxnum * sizeof (E3_CommandPort)) & (PAGE_SIZE-1);
    int           s;    
        
    /* set up the initial position */
    ring->Entries  = entries;
    ring->Position = 0;
    
    if (! (ring->pEvent = ep_alloc_elan (&rail->Generic, entries * sizeof (E3_BlockCopyEvent), 0, &ring->epEvent)))
    {
	ring->CommandPort = (ioaddr_t) NULL;
	return (ENOMEM);
    }
    
    if (! (ring->pDma = ep_alloc_elan (&rail->Generic, entries * sizeof (E3_DMA), 0, &ring->epDma)))
    {
	ep_free_elan (&rail->Generic, ring->epEvent, entries * sizeof (E3_BlockCopyEvent));

	ring->CommandPort = (ioaddr_t) NULL;
	return (ENOMEM);
    }
    
    if (! (ring->pDoneBlk = ep_alloc_main (&rail->Generic, entries * sizeof (E3_uint32), 0, &ring->epDoneBlk)))
    {
	ep_free_elan (&rail->Generic, ring->epEvent, entries * sizeof (E3_BlockCopyEvent));
	ep_free_elan (&rail->Generic, ring->epDma,   entries * sizeof (E3_DMA));

	ring->CommandPort = (ioaddr_t) NULL;
	return (ENOMEM);
    }
    
    if (MapDeviceRegister (rail->Device, ELAN3_BAR_COMMAND_PORT, &ring->CommandPage, pgnum * PAGE_SIZE, PAGE_SIZE, &ring->CommandPageHandle) != ESUCCESS)
    {
	ep_free_elan (&rail->Generic, ring->epEvent,   entries * sizeof (E3_BlockCopyEvent));
	ep_free_elan (&rail->Generic, ring->epDma,     entries * sizeof (E3_DMA));
	ep_free_main (&rail->Generic, ring->epDoneBlk, entries * sizeof (E3_uint32));

	ring->CommandPort = (ioaddr_t) NULL;
	return (ENOMEM);
    }
    ring->CommandPort = ring->CommandPage + pgoff;
	
    for (s = 0; s < entries; s++)
    {
	/* setup the event */
	elan3_sdram_writel(rail->Device, DMA_RING_EVENT(ring,s) + offsetof(E3_BlockCopyEvent,ev_Type),   
			   EV_TYPE_BCOPY | EV_TYPE_DMA | DMA_RING_DMA_ELAN(ring, s));
	elan3_sdram_writel(rail->Device, DMA_RING_EVENT(ring,s) + offsetof(E3_BlockCopyEvent,ev_Source), DMA_RING_DMA_ELAN(ring,s)  | EV_WCOPY);
	elan3_sdram_writel(rail->Device, DMA_RING_EVENT(ring,s) + offsetof(E3_BlockCopyEvent,ev_Dest),   DMA_RING_DONE_ELAN(ring,s) | EV_TYPE_BCOPY_WORD );         

	/* need to set all the doneBlks to appear that they have completed */
	ring->pDoneBlk[s] = DMA_RING_DMA_ELAN(ring,s)  | EV_WCOPY;
    }

    return 0; /* success */
}

static void
DmaRingRelease(EP3_RAIL *rail, EP3_DMA_RING *ring)
{
    if (ring->CommandPage != (ioaddr_t) 0)
    {
	UnmapDeviceRegister(rail->Device, &ring->CommandPageHandle);

	ep_free_elan (&rail->Generic, ring->epEvent,   ring->Entries * sizeof (E3_BlockCopyEvent));
	ep_free_elan (&rail->Generic, ring->epDma,     ring->Entries * sizeof (E3_DMA));
	ep_free_main (&rail->Generic, ring->epDoneBlk, ring->Entries * sizeof (E3_uint32));
    }
    ring->CommandPage = (ioaddr_t) 0;
}

void 
DmaRingsRelease (EP3_RAIL *rail)
{
    DmaRingRelease (rail, &rail->DmaRings[EP3_RING_CRITICAL]);
    DmaRingRelease (rail, &rail->DmaRings[EP3_RING_HIGH_PRI]);
    DmaRingRelease (rail, &rail->DmaRings[EP3_RING_LOW_PRI]);
}

int 
DmaRingsCreate (EP3_RAIL *rail)
{
    if (DmaRingCreate (rail, &rail->DmaRings[EP3_RING_CRITICAL], ELAN3_DMARING_BASE_CONTEXT_NUM + EP3_RING_CRITICAL, EP3_RING_CRITICAL_LEN) ||
	DmaRingCreate (rail, &rail->DmaRings[EP3_RING_HIGH_PRI], ELAN3_DMARING_BASE_CONTEXT_NUM + EP3_RING_HIGH_PRI, EP3_RING_HIGH_PRI_LEN) ||
	DmaRingCreate (rail, &rail->DmaRings[EP3_RING_LOW_PRI],  ELAN3_DMARING_BASE_CONTEXT_NUM + EP3_RING_LOW_PRI,  EP3_RING_LOW_PRI_LEN))
    {
	DmaRingsRelease (rail);
	return (ENOMEM);
    }
  
    return 0;
}

static int 
DmaRingNextSlot (EP3_DMA_RING *ring)
{
    int pos  = ring->Position;
    int npos = DMA_RING_NEXT_POS(ring);

    if (ring->pDoneBlk[npos] == EP3_EVENT_ACTIVE)
	return (-1);
    
    ring->pDoneBlk[pos] = EP3_EVENT_ACTIVE;

    ring->Position = npos; /* move on one */

    return (pos);
}


/****************************************************************************************/
/*
 * Dma/event command issueing - these handle cproc queue overflow traps.
 */
static int
DmaRunQueueSizeCheck (EP3_RAIL *rail, E3_uint32 len)
{
    E3_uint64  FandBPtr = read_reg64 (rail->Device, DProc_SysCntx_FPtr);
    E3_uint32  FPtr, BPtr;
    E3_uint32  qlen;

#if (BYTE_ORDER == LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__)
    FPtr = (FandBPtr & 0xFFFFFFFFull);
    BPtr = (FandBPtr >> 32);
#else
    FPtr = (FandBPtr >> 32);
    BPtr = (FandBPtr & 0xFFFFFFFFull);
#endif
    
    qlen = (((BPtr - FPtr)/sizeof (E3_DMA)) & (E3_SysCntxQueueSize-1));
    
    if      (qlen < 4)   IncrStat (rail, DmaQueueLength[0]);
    else if (qlen < 8)   IncrStat (rail, DmaQueueLength[1]);
    else if (qlen < 16)  IncrStat (rail, DmaQueueLength[2]);
    else if (qlen < 32)  IncrStat (rail, DmaQueueLength[3]);
    else if (qlen < 64)  IncrStat (rail, DmaQueueLength[4]);
    else if (qlen < 128) IncrStat (rail, DmaQueueLength[5]);
    else if (qlen < 240) IncrStat (rail, DmaQueueLength[6]);
    else                 IncrStat (rail, DmaQueueLength[7]);
	
    return (qlen < len);
}

int
IssueDma (EP3_RAIL *rail, E3_DMA_BE * dmabe, int type, int retryThread)
{
    ELAN3_DEV     *dev = rail->Device;
    EP3_RETRY_DMA *retry;
    EP3_DMA_RING  *ring;
    int           slot;
    int           i, res;
    unsigned long flags;

    ASSERT (dmabe->s.dma_direction == DMA_WRITE || dmabe->s.dma_direction == DMA_READ_REQUEUE);

    ASSERT (! EP_VP_ISDATA(dmabe->s.dma_destVProc) ||
	    (dmabe->s.dma_direction == DMA_WRITE ? 
	     EP_VP_TO_NODE(dmabe->s.dma_srcVProc) == rail->Generic.Position.pos_nodeid :
	     EP_VP_TO_NODE(dmabe->s.dma_destVProc) == rail->Generic.Position.pos_nodeid));
    
    /*
     * If we're not the retry thread - then don't issue this DMA
     * if there are any already queued on the retry lists with
     * higher or equal priority than this one that are ready to
     * retry.
     */
    if (! retryThread)
    {
	for (i = EP_RETRY_BASE; i < type; i++)
	{
	    if (list_empty (&rail->DmaRetries[i]))
		continue;

	    retry = list_entry (rail->DmaRetries[i].next, EP3_RETRY_DMA, Link);
		
	    if (AFTER (lbolt, retry->RetryTime))
	    {
		IncrStat (rail, IssueDmaFail[type]);
		return (ISSUE_COMMAND_RETRY);
	    }
	}
    }

    /*
     * Depending on the type of DMA we're issuing - throttle back
     * issueing of it if the DMA run queue is too full.  This then
     * prioritises the "special" messages and completing data 
     * transfers which have matched a receive buffer.
     */

    if (type >= EP_RETRY_LOW_PRI_RETRY)
    {
	if (! DmaRunQueueSizeCheck (rail, E3_SysCntxQueueSize / 2))
	{
	    IncrStat (rail, IssueDmaFail[type]);
	    return (ISSUE_COMMAND_RETRY);
	}
	ring = &rail->DmaRings[EP3_RING_LOW_PRI];
    } 
    else if (type == EP_RETRY_LOW_PRI)
    {
	if (! DmaRunQueueSizeCheck (rail, E3_SysCntxQueueSize / 3))
	{
	    IncrStat (rail, IssueDmaFail[type]);
	    return (ISSUE_COMMAND_RETRY);
	}
	ring = &rail->DmaRings[EP3_RING_LOW_PRI];
    }
    else if (type >= EP_RETRY_HIGH_PRI)
	ring = &rail->DmaRings[EP3_RING_HIGH_PRI];
    else
	ring = &rail->DmaRings[EP3_RING_CRITICAL];

    local_irq_save (flags);
    if (! spin_trylock (&dev->CProcLock))
    {
	IncrStat (rail, IssueDmaFail[type]);

	res = ISSUE_COMMAND_RETRY;
    }
    else
    {
	if ((slot = DmaRingNextSlot (ring)) == -1)
	{
	    IncrStat (rail, IssueDmaFail[type]);
	    
	    res = ISSUE_COMMAND_RETRY;
	}
	else
	{
	    EPRINTF4 (DBG_COMMAND, "IssueDma: type %08x size %08x Elan source %08x Elan dest %08x\n",
		      dmabe->s.dma_type, dmabe->s.dma_size, dmabe->s.dma_source, dmabe->s.dma_dest);
	    EPRINTF2 (DBG_COMMAND, "          dst event %08x cookie/proc %08x\n",
		      dmabe->s.dma_destEvent, dmabe->s.dma_destCookieVProc);
	    EPRINTF2 (DBG_COMMAND, "          src event %08x cookie/proc %08x\n",
		      dmabe->s.dma_srcEvent, dmabe->s.dma_srcCookieVProc);

	    elan3_sdram_copyq_to_sdram (dev,  dmabe,  DMA_RING_DMA(ring, slot), sizeof (E3_DMA)); 			/* PCI write block */
	    elan3_sdram_writel (dev, DMA_RING_EVENT(ring, slot) + offsetof (E3_BlockCopyEvent, ev_Count), 1);	/* PCI write */
	    
	    mb();								/* ensure writes to main memory completed */
	    writel (DMA_RING_EVENT_ELAN(ring,slot), (void *)(ring->CommandPort + offsetof (E3_CommandPort, SetEvent)));
	    mmiob();		     						/* and flush through IO writes */
	    
	    res = ISSUE_COMMAND_OK;
	}
	spin_unlock (&dev->CProcLock);
    }
    local_irq_restore (flags);

    return (res);
}

int
IssueWaitevent (EP3_RAIL *rail, E3_Addr value)
{
    ELAN3_DEV     *dev   = rail->Device;
    int           res;
    unsigned long flags;
    
    spin_lock_irqsave (&dev->IntrLock, flags);

    ASSERT (rail->CommandPortEventTrap == FALSE);

    /*
     * Disable the command processor interrupts, so that we don't see
     * spurious interrupts appearing.
     */
    DISABLE_INT_MASK (dev, INT_CProc | INT_ComQueue);

    EPRINTF1 (DBG_COMMAND, "IssueWaitevent: %08x\n", value);

    mb();								/* ensure writes to main memory completed */
    writel (value, (void *)(rail->CommandPort + offsetof (E3_CommandPort, WaitEvent0)));
    mmiob();		     						/* and flush through IO writes */
    
    do {
	res = CheckCommandQueueFlushed (rail->Ctxt, EventComQueueNotEmpty, ISSUE_COMMAND_CANT_WAIT, &flags);

	EPRINTF1 (DBG_COMMAND, "IssueWaitevent: CheckCommandQueueFlushed -> %d\n", res);

	if (res == ISSUE_COMMAND_WAIT)
	    HandleCProcTrap (dev, 0, NULL);
    } while (res != ISSUE_COMMAND_OK);

    if (! rail->CommandPortEventTrap)
	res = ISSUE_COMMAND_OK;
    else
    {
	rail->CommandPortEventTrap = FALSE;
	res = ISSUE_COMMAND_TRAPPED;
    }

    EPRINTF1 (DBG_COMMAND, "IssueWaitevent: -> %d\n", res);

    /*
     * Re-enable the command processor interrupt as we've finished 
     * polling it.
     */
    ENABLE_INT_MASK (dev, INT_CProc | INT_ComQueue);

    spin_unlock_irqrestore (&dev->IntrLock, flags);

    return (res);
}

void
IssueSetevent (EP3_RAIL *rail, E3_Addr value)
{
    EPRINTF1 (DBG_COMMAND, "IssueSetevent: %08x\n", value);

    mb();								/* ensure writes to main memory completed */
    writel (value, (void *)(rail->CommandPort + offsetof (E3_CommandPort, SetEvent)));
    mmiob();		     						/* and flush through IO writes */
}

void
IssueRunThread (EP3_RAIL *rail, E3_Addr value)
{
    EPRINTF1 (DBG_COMMAND, "IssueRunThread: %08x\n", value);

    mb();								/* ensure writes to main memory completed */
    writel (value, (void *)(rail->CommandPort + offsetof (E3_CommandPort, RunThread)));
    mmiob();		     						/* and flush through IO writes */
}

/****************************************************************************************/
/*
 * DMA retry list management
 */
static unsigned DmaRetryTimes[EP_NUM_RETRIES]; 

static void
ep3_dma_retry (EP3_RAIL *rail)
{
    EP3_COOKIE    *cp;
    int            res;
    int		   vp;
    unsigned long  flags;
    int            i;

    kernel_thread_init("ep3_dma_retry");

    spin_lock_irqsave (&rail->DmaRetryLock, flags);

    for (;;)
    {
	long yieldAt   = lbolt + (hz/10);
	long retryTime = 0;

	if (rail->DmaRetryThreadShouldStop)
	    break;
	
	for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
	{
	    while (! list_empty (&rail->DmaRetries[i]))
	    {
		EP3_RETRY_DMA *retry = list_entry (rail->DmaRetries[i].next, EP3_RETRY_DMA, Link);

		if (! AFTER (lbolt, retry->RetryTime))
		    break;
		
		if (rail->DmaRetryThreadShouldStall || AFTER (lbolt, yieldAt))
		    goto cant_do_more;

		EPRINTF2 (DBG_RETRY, "%s: DmaRetryThread: retry %p\n", rail->Generic.Name, retry);
		EPRINTF5 (DBG_RETRY, "%s:                 %08x %08x %08x %08x\n",
			  rail->Generic.Name, retry->Dma.s.dma_type, retry->Dma.s.dma_size, retry->Dma.s.dma_source, retry->Dma.s.dma_dest);
		EPRINTF5 (DBG_RETRY, "%s:                 %08x %08x %08x %08x\n",
			  rail->Generic.Name, retry->Dma.s.dma_destEvent, retry->Dma.s.dma_destCookieVProc,
			  retry->Dma.s.dma_srcEvent, retry->Dma.s.dma_srcCookieVProc);
#if defined(DEBUG)
		if (retry->Dma.s.dma_direction == DMA_WRITE)
		    cp = LookupEventCookie (rail, &rail->CookieTable, retry->Dma.s.dma_srcEvent);
		else
		    cp = LookupEventCookie (rail, &rail->CookieTable, retry->Dma.s.dma_destEvent);

		ASSERT (cp != NULL || (retry->Dma.s.dma_srcEvent == 0 && retry->Dma.s.dma_direction == DMA_WRITE && retry->Dma.s.dma_isRemote));
		
		if (cp && cp->Operations->DmaVerify)
		    cp->Operations->DmaVerify (rail, cp->Arg, &retry->Dma);
#endif

#if defined(DEBUG_ASSERT)
		if (retry->Dma.s.dma_direction == DMA_WRITE)
		    vp = retry->Dma.s.dma_destVProc;
		else
		    vp = retry->Dma.s.dma_srcVProc;

		ASSERT (!EP_VP_ISDATA(vp) || 
			(rail->Generic.Nodes[EP_VP_TO_NODE(vp)].State >= EP_NODE_CONNECTED &&
			 rail->Generic.Nodes[EP_VP_TO_NODE(vp)].State <= EP_NODE_LOCAL_PASSIVATE));
#endif
		spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
		res = IssueDma (rail, &(retry->Dma), i, TRUE);
		spin_lock_irqsave (&rail->DmaRetryLock, flags);
		
		if (res != ISSUE_COMMAND_OK)
		    goto cant_do_more;
		
		/* Command issued, so remove from list, and add to free list */
		list_del (&retry->Link);
		list_add (&retry->Link, &rail->DmaRetryFreeList);
	    }
	}
    cant_do_more:
	
	for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
	{
	    if (!list_empty (&rail->DmaRetries[i]))
	    {
		EP3_RETRY_DMA *retry = list_entry (rail->DmaRetries[i].next, EP3_RETRY_DMA, Link);

		retryTime = retryTime ? MIN(retryTime, retry->RetryTime) : retry->RetryTime;
	    }
	}

	if (retryTime && !AFTER (retryTime, lbolt))
	    retryTime = lbolt + 1;

	do {
	    EPRINTF3 (DBG_RETRY, "%s: ep_cm_retry: %s %lx\n", rail->Generic.Name, rail->DmaRetryThreadShouldStall ? "stalled" : "sleeping", retryTime);
	    
	    if (rail->DmaRetryTime == 0 || (retryTime != 0 && retryTime < rail->DmaRetryTime))
		rail->DmaRetryTime = retryTime;
	    
	    rail->DmaRetrySleeping = TRUE;
	    
	    if (rail->DmaRetryThreadShouldStall)					/* wakeup threads waiting in StallDmaRetryThread */
		kcondvar_wakeupall (&rail->DmaRetryWait, &rail->DmaRetryLock); 	/* for us to really go to sleep for good. */

	    if (rail->DmaRetryTime == 0 || rail->DmaRetryThreadShouldStall)
		kcondvar_wait (&rail->DmaRetryWait, &rail->DmaRetryLock, &flags);
	    else
		kcondvar_timedwait (&rail->DmaRetryWait, &rail->DmaRetryLock, &flags, rail->DmaRetryTime);

	    rail->DmaRetrySleeping = FALSE;

	} while (rail->DmaRetryThreadShouldStall);

	rail->DmaRetryTime = 0;
    }

    rail->DmaRetryThreadStopped = 1;
    kcondvar_wakeupall (&rail->DmaRetryWait, &rail->DmaRetryLock);
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);

    kernel_thread_exit();
}

void
StallDmaRetryThread (EP3_RAIL *rail)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    rail->DmaRetryThreadShouldStall++;

    while (! rail->DmaRetrySleeping)
	kcondvar_wait (&rail->DmaRetryWait, &rail->DmaRetryLock, &flags);
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

void 
ResumeDmaRetryThread (EP3_RAIL *rail)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);

    ASSERT (rail->DmaRetrySleeping);

    if (--rail->DmaRetryThreadShouldStall == 0)
    {
	rail->DmaRetrySleeping = 0;
	kcondvar_wakeupone (&rail->DmaRetryWait, &rail->DmaRetryLock);
    }
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

int
InitialiseDmaRetries (EP3_RAIL *rail)
{
    int i;

    spin_lock_init (&rail->DmaRetryLock);
    kcondvar_init (&rail->DmaRetryWait);

    for (i = 0; i < EP_NUM_RETRIES; i++)
	INIT_LIST_HEAD (&rail->DmaRetries[i]);

    INIT_LIST_HEAD (&rail->DmaRetryFreeList);

    DmaRetryTimes[EP_RETRY_HIGH_PRI]  = EP_RETRY_HIGH_PRI_TIME;

    for (i =0 ; i < EP_NUM_BACKOFF; i++)
	DmaRetryTimes[EP_RETRY_HIGH_PRI_RETRY+i] = EP_RETRY_HIGH_PRI_TIME << i;
    
    DmaRetryTimes[EP_RETRY_LOW_PRI] = EP_RETRY_LOW_PRI_TIME;

    for (i =0 ; i < EP_NUM_BACKOFF; i++)
	DmaRetryTimes[EP_RETRY_LOW_PRI_RETRY+i] = EP_RETRY_LOW_PRI_TIME << i;
    
    DmaRetryTimes[EP_RETRY_ANONYMOUS] = EP_RETRY_ANONYMOUS_TIME;
    DmaRetryTimes[EP_RETRY_NETERR]    = EP_RETRY_NETERR_TIME;

    rail->DmaRetryInitialised = 1;

    if (kernel_thread_create (ep3_dma_retry, (void *) rail) == 0)
    {
	spin_lock_destroy (&rail->DmaRetryLock);
	return (ENOMEM);
    }

    rail->DmaRetryThreadStarted = 1;

    return (ESUCCESS);
}

void
DestroyDmaRetries (EP3_RAIL *rail)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    rail->DmaRetryThreadShouldStop = 1;
    while (rail->DmaRetryThreadStarted && !rail->DmaRetryThreadStopped)
    {
	kcondvar_wakeupall (&rail->DmaRetryWait, &rail->DmaRetryLock);
	kcondvar_wait (&rail->DmaRetryWait, &rail->DmaRetryLock, &flags);
    }
    rail->DmaRetryThreadStarted = 0;
    rail->DmaRetryThreadStopped = 0;
    rail->DmaRetryThreadShouldStop = 0;
    rail->DmaRetryInitialised = 0;

    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);

    /* Everyone should have given back their retry dma's by now */
    ASSERT (rail->DmaRetryReserved == 0);

    while (! list_empty (&rail->DmaRetryFreeList))
    {
	EP3_RETRY_DMA *retry = list_entry (rail->DmaRetryFreeList.next, EP3_RETRY_DMA, Link);
	
	list_del (&retry->Link);

	KMEM_FREE (retry, sizeof (EP3_RETRY_DMA));
    }

    kcondvar_destroy (&rail->DmaRetryWait);
    spin_lock_destroy (&rail->DmaRetryLock);
}

int
ReserveDmaRetries (EP3_RAIL *rail, int count, EP_ATTRIBUTE attr)
{
    EP3_RETRY_DMA *retry;
    int		  remaining = count;
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    
    if (remaining <= (rail->DmaRetryCount - rail->DmaRetryReserved))
    {
	rail->DmaRetryReserved += remaining;

	spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
	return (ESUCCESS);
    }

    remaining -= (rail->DmaRetryCount - rail->DmaRetryReserved);

    rail->DmaRetryReserved = rail->DmaRetryCount;

    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);

    while (remaining)
    {
	KMEM_ALLOC (retry, EP3_RETRY_DMA *, sizeof (EP3_RETRY_DMA), !(attr & EP_NO_SLEEP));
	
	if (retry == NULL)
	    goto failed;

	/* clear E3_DMA */
	bzero((char *)(&(retry->Dma.s)), sizeof(E3_DMA));

	remaining--; 

	spin_lock_irqsave (&rail->DmaRetryLock, flags);

	list_add (&retry->Link, &rail->DmaRetryFreeList);

	rail->DmaRetryCount++;
	rail->DmaRetryReserved++;

	spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
    }
    return (ESUCCESS);

 failed:
    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    rail->DmaRetryReserved -= (count - remaining);
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
    return (ENOMEM);
}

void
ReleaseDmaRetries (EP3_RAIL *rail, int count)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    rail->DmaRetryReserved -= count;
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

void
QueueDmaForRetry (EP3_RAIL *rail, E3_DMA_BE *dma, int interval)
{
    EP3_RETRY_DMA *retry;
    unsigned long flags;

    /*
     * When requeueing DMAs they must never be "READ" dma's since
     * these would fetch the DMA descriptor from the retryn descriptor
     */
    ASSERT (dma->s.dma_direction == DMA_WRITE || dma->s.dma_direction == DMA_READ_REQUEUE);
    ASSERT (dma->s.dma_direction == DMA_WRITE ? 
	    EP_VP_TO_NODE(dma->s.dma_srcVProc) == rail->Generic.Position.pos_nodeid :
	    EP_VP_TO_NODE(dma->s.dma_destVProc) == rail->Generic.Position.pos_nodeid);

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    
    EP_ASSERT (&rail->Generic, !list_empty (&rail->DmaRetryFreeList));

    /* take an item of the free list */
    retry = list_entry (rail->DmaRetryFreeList.next, EP3_RETRY_DMA, Link);

    list_del (&retry->Link);
    
    EPRINTF5 (DBG_RETRY, "%s: QueueDmaForRetry: %08x %08x %08x %08x\n", rail->Generic.Name,
	      dma->s.dma_type, dma->s.dma_size, dma->s.dma_source, dma->s.dma_dest);
    EPRINTF5 (DBG_RETRY, "%s:                   %08x %08x %08x %08x\n",rail->Generic.Name,
	     dma->s.dma_destEvent, dma->s.dma_destCookieVProc,
	     dma->s.dma_srcEvent, dma->s.dma_srcCookieVProc);

    /* copy the DMA into the retry descriptor */
    retry->Dma.s.dma_type            = dma->s.dma_type;
    retry->Dma.s.dma_size            = dma->s.dma_size;
    retry->Dma.s.dma_source          = dma->s.dma_source;
    retry->Dma.s.dma_dest            = dma->s.dma_dest;
    retry->Dma.s.dma_destEvent       = dma->s.dma_destEvent;
    retry->Dma.s.dma_destCookieVProc = dma->s.dma_destCookieVProc;
    retry->Dma.s.dma_srcEvent        = dma->s.dma_srcEvent;
    retry->Dma.s.dma_srcCookieVProc  = dma->s.dma_srcCookieVProc;

    retry->RetryTime = lbolt + DmaRetryTimes[interval];

    /* chain onto the end of the approriate retry list */
    list_add_tail (&retry->Link, &rail->DmaRetries[interval]);

    /* now wakeup the retry thread */
    if (rail->DmaRetryTime == 0 || retry->RetryTime < rail->DmaRetryTime)
	rail->DmaRetryTime = retry->RetryTime;
    
    if (rail->DmaRetrySleeping && !rail->DmaRetryThreadShouldStall)
    {
	rail->DmaRetrySleeping = 0;
	kcondvar_wakeupone (&rail->DmaRetryWait, &rail->DmaRetryLock);
    }

    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

void
QueueDmaOnStalledList (EP3_RAIL *rail, E3_DMA_BE *dma)
{
    EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[dma->s.dma_direction == DMA_WRITE ? 
						  EP_VP_TO_NODE(dma->s.dma_srcVProc) :
						  EP_VP_TO_NODE(dma->s.dma_destVProc)];
    EP3_RETRY_DMA *retry;
    unsigned long flags;

    /*
     * When requeueing DMAs they must never be "READ" dma's since
     * these would fetch the DMA descriptor from the retryn descriptor
     */
    ASSERT (dma->s.dma_direction == DMA_WRITE || dma->s.dma_direction == DMA_READ_REQUEUE);
    ASSERT (dma->s.dma_direction == DMA_WRITE ? 
	    EP_VP_TO_NODE(dma->s.dma_srcVProc) == rail->Generic.Position.pos_nodeid :
	    EP_VP_TO_NODE(dma->s.dma_destVProc) == rail->Generic.Position.pos_nodeid);

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    
    EP_ASSERT (&rail->Generic, !list_empty (&rail->DmaRetryFreeList));

    /* take an item of the free list */
    retry = list_entry (rail->DmaRetryFreeList.next, EP3_RETRY_DMA, Link);

    list_del (&retry->Link);
    
    EPRINTF5 (DBG_RETRY, "%s: QueueDmaOnStalledList: %08x %08x %08x %08x\n", rail->Generic.Name,
	      dma->s.dma_type, dma->s.dma_size, dma->s.dma_source, dma->s.dma_dest);
    EPRINTF5 (DBG_RETRY, "%s:                        %08x %08x %08x %08x\n", rail->Generic.Name,
	      dma->s.dma_destEvent, dma->s.dma_destCookieVProc,
	      dma->s.dma_srcEvent, dma->s.dma_srcCookieVProc);

    /* copy the DMA into the retry descriptor */
    retry->Dma.s.dma_type            = dma->s.dma_type;
    retry->Dma.s.dma_size            = dma->s.dma_size;
    retry->Dma.s.dma_source          = dma->s.dma_source;
    retry->Dma.s.dma_dest            = dma->s.dma_dest;
    retry->Dma.s.dma_destEvent       = dma->s.dma_destEvent;
    retry->Dma.s.dma_destCookieVProc = dma->s.dma_destCookieVProc;
    retry->Dma.s.dma_srcEvent        = dma->s.dma_srcEvent;
    retry->Dma.s.dma_srcCookieVProc  = dma->s.dma_srcCookieVProc;

    /* chain onto the node cancelled dma list */
    list_add_tail (&retry->Link, &nodeRail->StalledDmas);

    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

void
FreeStalledDmas (EP3_RAIL *rail, unsigned int nodeId)
{
    EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[nodeId];
    struct list_head *el, *nel;
    unsigned long flags;

    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    list_for_each_safe (el, nel, &nodeRail->StalledDmas) {
	list_del (el);
	list_add (el, &rail->DmaRetryFreeList);
    }
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);
}

/****************************************************************************************/
/*
 * Connection management.
 */
static void
DiscardingHaltOperation (ELAN3_DEV *dev, void *arg)
{
    EP3_RAIL *rail = (EP3_RAIL *) arg;
    unsigned long flags;

    spin_lock_irqsave (&dev->IntrLock, flags);
    rail->HaltOpCompleted = 1;
    kcondvar_wakeupall (&rail->HaltOpSleep, &dev->IntrLock);
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}
 
typedef struct {
     EP3_RAIL  *rail;
    sdramaddr_t qaddr;
} SetQueueFullData;
 
static void
SetQueueLockedOperation (ELAN3_DEV *dev, void *arg)
{
    SetQueueFullData *data =  (SetQueueFullData *) arg;
    unsigned long     flags;     

    spin_lock_irqsave (&dev->IntrLock, flags);

    elan3_sdram_writel  (dev, data->qaddr, E3_QUEUE_LOCKED | elan3_sdram_readl(dev, data->qaddr));
   
    data->rail->HaltOpCompleted = 1;
    kcondvar_wakeupall (&data->rail->HaltOpSleep, &dev->IntrLock);

    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

static void
FlushDmaQueuesHaltOperation (ELAN3_DEV *dev, void *arg)
{
    EP3_RAIL      *rail    = (EP3_RAIL *) arg;
    sdramaddr_t    FPtr, BPtr;
    sdramaddr_t	   Base, Top;
    E3_DMA_BE      dma;
    EP_NODE_RAIL  *node;
    int            vp;
    unsigned long  flags;

    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProc.s.FSR)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData0.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData1.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData2.s.FSR.Status)) == 0);
    ASSERT (elan3_sdram_readl (dev, dev->TAndQBase + offsetof (E3_TrapAndQueue, DProcData3.s.FSR.Status)) == 0);
    
    FPtr  = read_reg32 (dev, DProc_SysCntx_FPtr);
    BPtr =  read_reg32 (dev, DProc_SysCntx_BPtr);
    Base  = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[0]);
    Top   = dev->TAndQBase + offsetof (E3_TrapAndQueue, SysCntxDmaQueue[E3_SysCntxQueueSize-1]);
    
    while (FPtr != BPtr)
    {
	elan3_sdram_copyq_from_sdram (dev, FPtr, &dma, sizeof (E3_DMA_BE));
	
	EPRINTF5 (DBG_DISCON, "%s: FlushDmaQueuesHaltOperation: %08x %08x %08x %08x\n", rail->Generic.Name,
		  dma.s.dma_type, dma.s.dma_size, dma.s.dma_source, dma.s.dma_dest);
	EPRINTF5 (DBG_DISCON, "%s:                              %08x %08x %08x %08x\n", rail->Generic.Name,
		  dma.s.dma_destEvent, dma.s.dma_destCookieVProc,
		 dma.s.dma_srcEvent, dma.s.dma_srcCookieVProc);
	
	ASSERT ((dma.s.dma_u.s.Context & SYS_CONTEXT_BIT) != 0);

	if (dma.s.dma_direction == DMA_WRITE)
	    vp = dma.s.dma_destVProc;
	else
	    vp = dma.s.dma_srcVProc;
	
	node = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];

	ASSERT (!EP_VP_ISDATA(vp) || (node->State >= EP_NODE_CONNECTED && node->State <= EP_NODE_LOCAL_PASSIVATE));

	if (EP_VP_ISDATA(vp) && node->State == EP_NODE_LOCAL_PASSIVATE)
	{
	    /*
	     * This is a DMA going to the node which is being removed, 
	     * so move it onto the node dma list where it will get
	     * handled later.
	     */
	    EPRINTF1 (DBG_DISCON, "%s: FlushDmaQueuesHaltOperation: move dma to cancelled list\n", rail->Generic.Name);
	   
	    if (dma.s.dma_direction != DMA_WRITE)
	    {
		/* for read dma's set the DMA_READ_REQUEUE bits as the dma_source has been 
		 * modified by the elan to point at the dma in the rxd where it was issued
		 * from */
		dma.s.dma_direction = (dma.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
	    }
	    
	    QueueDmaOnStalledList (rail, &dma);
	    
	    /*
	     * Remove the DMA from the queue by replacing it with one with
	     * zero size and no events.
	     *
	     * NOTE: we must preserve the SYS_CONTEXT_BIT since the Elan uses this
	     * to mark the approriate run queue as empty.
	     */
	    dma.s.dma_type            = (SYS_CONTEXT_BIT << 16);
	    dma.s.dma_size            = 0;
	    dma.s.dma_source          = (E3_Addr) 0;
	    dma.s.dma_dest            = (E3_Addr) 0;
	    dma.s.dma_destEvent       = (E3_Addr) 0;
	    dma.s.dma_destCookieVProc = 0;
	    dma.s.dma_srcEvent        = (E3_Addr) 0;
	    dma.s.dma_srcCookieVProc  = 0;
	    
	    elan3_sdram_copyq_to_sdram (dev, &dma, FPtr, sizeof (E3_DMA_BE));
	}

	FPtr = (FPtr == Top) ? Base : FPtr + sizeof (E3_DMA);
    }

    spin_lock_irqsave (&dev->IntrLock, flags);
    rail->HaltOpCompleted = 1;
    kcondvar_wakeupall (&rail->HaltOpSleep, &dev->IntrLock);
    spin_unlock_irqrestore (&dev->IntrLock, flags);
}

void
SetQueueLocked (EP3_RAIL *rail, sdramaddr_t qaddr)
{
    ELAN3_DEV        *dev = rail->Device;
    SetQueueFullData  data;
    unsigned long     flags;
    
    /* Ensure that the context filter changes have been seen by halting
     * then restarting the inputters - this also ensures that any setevent
     * commands used to issue dma's have completed and any trap has been
     * handled. */
    data.rail  = rail;
    data.qaddr = qaddr;

    kmutex_lock (&rail->HaltOpMutex);
    spin_lock_irqsave (&dev->IntrLock, flags);
    QueueHaltOperation (dev, 0, NULL, INT_DiscardingSysCntx | INT_TProcHalted, SetQueueLockedOperation, &data);

    while (! rail->HaltOpCompleted)
	kcondvar_wait (&rail->HaltOpSleep, &dev->IntrLock, &flags);
    rail->HaltOpCompleted = 0;

    spin_unlock_irqrestore (&dev->IntrLock, flags);
    kmutex_unlock (&rail->HaltOpMutex);
}

void
ep3_flush_filters (EP_RAIL *r)
{
    EP3_RAIL *rail = (EP3_RAIL *) r;
    ELAN3_DEV *dev  = rail->Device;
    unsigned long flags;

    /* Ensure that the context filter changes have been seen by halting
     * then restarting the inputters - this also ensures that any setevent
     * commands used to issue dma's have completed and any trap has been
     * handled. */
    kmutex_lock (&rail->HaltOpMutex);
    spin_lock_irqsave (&dev->IntrLock, flags);
    QueueHaltOperation (dev, 0, NULL, INT_DiscardingSysCntx, DiscardingHaltOperation, rail);
    
    while (! rail->HaltOpCompleted)
	kcondvar_wait (&rail->HaltOpSleep, &dev->IntrLock, &flags);
    rail->HaltOpCompleted = 0;
    spin_unlock_irqrestore (&dev->IntrLock, flags);
    kmutex_unlock (&rail->HaltOpMutex);
}

void
ep3_flush_queues (EP_RAIL *r)
{
    EP3_RAIL         *rail = (EP3_RAIL *) r;
    ELAN3_DEV         *dev  = rail->Device;
    struct list_head *el;
    struct list_head *nel;
    EP_NODE_RAIL     *node;
    unsigned long flags;
    int vp, i;

    ASSERT (NO_LOCKS_HELD);
    
    /* First - stall the dma retry thread, so that it will no longer
     *         restart any dma's from the rety lists. */
    StallDmaRetryThread (rail);

    /* Second - queue a halt operation to flush through all DMA's which are executing
     *          or on the run queue. */
    kmutex_lock (&rail->HaltOpMutex);
    spin_lock_irqsave (&dev->IntrLock, flags);
    QueueHaltOperation (dev, 0, NULL, INT_DProcHalted | INT_TProcHalted, FlushDmaQueuesHaltOperation, rail);
    while (! rail->HaltOpCompleted)
	kcondvar_wait (&rail->HaltOpSleep, &dev->IntrLock, &flags);
    rail->HaltOpCompleted = 0;
    spin_unlock_irqrestore (&dev->IntrLock, flags);
    kmutex_unlock (&rail->HaltOpMutex);

    /* Third - run down the dma retry lists and move all entries to the cancelled
     *         list.  Any dma's which were on the run queues have already been
     *         moved there */
    spin_lock_irqsave (&rail->DmaRetryLock, flags);
    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	list_for_each_safe (el, nel, &rail->DmaRetries[i]) {
	    EP3_RETRY_DMA *retry = list_entry (el, EP3_RETRY_DMA, Link);

	    if (retry->Dma.s.dma_direction == DMA_WRITE)
		vp = retry->Dma.s.dma_destVProc;
	    else
		vp = retry->Dma.s.dma_srcVProc;
	    
	    node = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];
	    
	    ASSERT (!EP_VP_ISDATA(vp) || (node->State >= EP_NODE_CONNECTED && node->State <= EP_NODE_LOCAL_PASSIVATE));

	    if (EP_VP_ISDATA(vp) && node->State == EP_NODE_LOCAL_PASSIVATE)
	    {
		EPRINTF5 (DBG_DISCON, "%s: FlushDmaQueues: %08x %08x %08x %08x\n",rail->Generic.Name,
			  retry->Dma.s.dma_type, retry->Dma.s.dma_size, retry->Dma.s.dma_source, retry->Dma.s.dma_dest);
		EPRINTF5 (DBG_DISCON, "%s:                 %08x %08x %08x %08x\n", rail->Generic.Name,
			  retry->Dma.s.dma_destEvent, retry->Dma.s.dma_destCookieVProc,
			  retry->Dma.s.dma_srcEvent, retry->Dma.s.dma_srcCookieVProc);

		list_del (&retry->Link);

		list_add_tail (&retry->Link, &node->StalledDmas);
	    }
	}
    }
    spin_unlock_irqrestore (&rail->DmaRetryLock, flags);

    /* Finally - allow the dma retry thread to run again */
    ResumeDmaRetryThread (rail);
}

/****************************************************************************************/
/* NOTE - we require that all cookies are non-zero, which is 
 *        achieved because EP_VP_DATA() is non-zero for all
 *        nodes */
E3_uint32
LocalCookie (EP3_RAIL *rail, unsigned remoteNode)
{
    E3_uint32     cookie;
    unsigned long flags;

    spin_lock_irqsave (&rail->CookieLock, flags);
    cookie = DMA_COOKIE (rail->MainCookies[remoteNode], EP_VP_DATA(rail->Generic.Position.pos_nodeid));
    spin_unlock_irqrestore (&rail->CookieLock, flags);

    /* Main processor cookie for srcCookie - this is what is sent
     * to the remote node along with the setevent from the put
     * or the dma descriptor for a get */
    return (cookie);
}

E3_uint32
RemoteCookie (EP3_RAIL *rail, u_int remoteNode)
{
    uint32_t      cookie;
    unsigned long flags;

    spin_lock_irqsave (&rail->CookieLock, flags);
    cookie = DMA_REMOTE_COOKIE (rail->MainCookies[remoteNode], EP_VP_DATA(remoteNode));
    spin_unlock_irqrestore (&rail->CookieLock, flags);

    /* Main processor cookie for dstCookie - this is the cookie
     * that the "remote put" dma uses for it's setevent packets for
     * a get dma */
    
    return (cookie);
}

/****************************************************************************************/
/*
 * Event Cookie management.
 *
 *   We find the ep_cookie in one of two ways:
 *     1) for block copy events
 *          the cookie value is stored in the ev_Source - for EVIRQ events
 *          it is also stored in the ev_Type
 *     2) for normal events
 *          we just use the event address.
 */
void 
InitialiseCookieTable (EP3_COOKIE_TABLE *table)
{
    register int i;
    
    spin_lock_init (&table->Lock);
    
    for (i = 0; i < EP3_COOKIE_HASH_SIZE; i++)
	table->Entries[i] = NULL;
}

void
DestroyCookieTable (EP3_COOKIE_TABLE *table)
{
    register int i;

    for (i = 0; i < EP3_COOKIE_HASH_SIZE; i++)
	if (table->Entries[i])
	    printk ("DestroyCookieTable: entry %d not empty\n", i);

    spin_lock_destroy (&table->Lock);
}

void
RegisterCookie (EP3_COOKIE_TABLE *table, EP3_COOKIE *cp, E3_uint32 cookie, EP3_COOKIE_OPS *ops, void *arg)
{
    EP3_COOKIE *tcp;
    int hashval = EP3_HASH_COOKIE(cookie);
    unsigned long flags;

    spin_lock_irqsave (&table->Lock, flags);
    
    cp->Operations = ops;
    cp->Arg        = arg;
    cp->Cookie     = cookie;
    
#if defined(DEBUG)
    /* Check that the cookie is unique */
    for (tcp = table->Entries[hashval]; tcp; tcp = tcp->Next)
	if (tcp->Cookie == cookie)
	    panic ("RegisterEventCookie: non unique cookie\n");
#endif
    cp->Next = table->Entries[hashval];
    
    table->Entries[hashval] = cp;
    
    spin_unlock_irqrestore (&table->Lock, flags);
}

void
DeregisterCookie (EP3_COOKIE_TABLE *table, EP3_COOKIE *cp)
{
    EP3_COOKIE **predCookiep;
    unsigned long flags;

    spin_lock_irqsave (&table->Lock, flags);
    
    for (predCookiep = &table->Entries[EP3_HASH_COOKIE (cp->Cookie)]; *predCookiep; predCookiep = &(*predCookiep)->Next)
    {
	if (*predCookiep == cp)
	{
	    *predCookiep = cp->Next;
	    break;
	}
    }

    spin_unlock_irqrestore (&table->Lock, flags);

    cp->Operations = NULL;
    cp->Arg        = NULL;
    cp->Cookie     = 0;
    cp->Next       = NULL;
}

EP3_COOKIE *
LookupCookie (EP3_COOKIE_TABLE *table, E3_Addr cookie)
{
    EP3_COOKIE *cp;
    unsigned long flags;

    spin_lock_irqsave (&table->Lock, flags);
    
    for (cp = table->Entries[EP3_HASH_COOKIE(cookie)]; cp; cp = cp->Next)
	if (cp->Cookie == cookie)
	    break;
    
    spin_unlock_irqrestore (&table->Lock, flags);
    return (cp);
}

EP3_COOKIE *
LookupEventCookie (EP3_RAIL *rail, EP3_COOKIE_TABLE *table, E3_Addr eaddr)
{
    sdramaddr_t event;
    E3_uint32 type;

    if ((event = ep_elan2sdram (&rail->Generic, eaddr)) != (sdramaddr_t) 0)
    {
	type = elan3_sdram_readl (rail->Device, event + offsetof (E3_BlockCopyEvent, ev_Type));

	if (type & EV_TYPE_BCOPY)
	    return (LookupCookie (table, elan3_sdram_readl (rail->Device, event + offsetof (E3_BlockCopyEvent, ev_Source)) & ~EV_WCOPY));
	else
	    return (LookupCookie (table, eaddr));
    }

    return (NULL);
}

/****************************************************************************************/
/*
 * Elan context operations - note only support interrupt ops.
 */
static int        ep3_event     (ELAN3_CTXT *ctxt, E3_uint32 cookie, int flag);
static int        ep3_dprocTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap);
static int        ep3_tprocTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap);
static int        ep3_iprocTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, int chan);
static int        ep3_cprocTrap (ELAN3_CTXT *ctxt, COMMAND_TRAP *trap);
static int        ep3_cprocReissue (ELAN3_CTXT *ctxt, CProcTrapBuf_BE *tbuf);

static E3_uint8   ep3_load8 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void       ep3_store8 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint8 val);
static E3_uint16  ep3_load16 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void       ep3_store16 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint16 val);
static E3_uint32  ep3_load32 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void       ep3_store32 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint32 val);
static E3_uint64  ep3_load64 (ELAN3_CTXT *ctxt, E3_Addr addr);
static void       ep3_store64 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint64 val);

ELAN3_OPS ep3_elan3_ops = 
{
    ELAN3_OPS_VERSION,		/* Version */
    
    NULL,			/* Exception */
    NULL,			/* GetWordItem */
    NULL,			/* GetBlockItem */
    NULL,			/* PutWordItem */
    NULL,			/* PutBlockItem */
    NULL,			/* PutbackItem */
    NULL,			/* FreeWordItem */
    NULL,			/* FreeBlockItem */
    NULL,			/* CountItems */
    ep3_event,			/* Event */
    NULL,			/* SwapIn */
    NULL,			/* SwapOut */
    NULL,			/* FreePrivate */
    NULL,			/* FixupNetworkError */
    ep3_dprocTrap,		/* DProcTrap */
    ep3_tprocTrap,		/* TProcTrap */
    ep3_iprocTrap,		/* IProcTrap */
    ep3_cprocTrap,		/* CProcTrap */
    ep3_cprocReissue,		/* CProcReissue */
    NULL,			/* StartFaultCheck */
    NULL,			/* EndFaulCheck */
    ep3_load8,			/* Load8 */
    ep3_store8,			/* Store8 */
    ep3_load16,			/* Load16 */
    ep3_store16,		/* Store16 */
    ep3_load32,			/* Load32 */
    ep3_store32,		/* Store32 */
    ep3_load64,			/* Load64 */
    ep3_store64,		/* Store64 */
};

static int
ep3_event (ELAN3_CTXT *ctxt, E3_uint32 cookie, int flag)
{
    EP3_RAIL  *rail = (EP3_RAIL *) ctxt->Private;
    EP3_COOKIE *cp   = LookupCookie (&rail->CookieTable, cookie);
    
    if (cp == NULL)
    {
	printk ("ep3_event: cannot find event cookie for %x\n", cookie);
	return (OP_HANDLED);
    }
    
    if (cp->Operations->Event)
	cp->Operations->Event(rail, cp->Arg);
    
    return (OP_HANDLED);
}

/* Trap interface */
int
ep3_dprocTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap)
{
    EP3_RAIL        *rail = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV        *dev = rail->Device;
    EP3_COOKIE       *cp;
    E3_FaultSave_BE *FaultArea;
    E3_uint16        vp;
    int		     validTrap;
    int		     numFaults;
    int	             i;
    sdramaddr_t      event;
    E3_uint32        type;
    sdramaddr_t      dma;
    E3_DMA_BE        dmabe;
    int              status = EAGAIN;

    EPRINTF4 (DBG_EPTRAP, "ep3_dprocTrap: WakeupFnt=%x Cntx=%x SuspAddr=%x TrapType=%s\n",
	      trap->Status.s.WakeupFunction, trap->Status.s.Context, 
	      trap->Status.s.SuspendAddr, MiToName (trap->Status.s.TrapType));
    EPRINTF4 (DBG_EPTRAP, "              type %08x size %08x source %08x dest %08x\n",
	      trap->Desc.s.dma_type, trap->Desc.s.dma_size, trap->Desc.s.dma_source, trap->Desc.s.dma_dest);
    EPRINTF2 (DBG_EPTRAP, "              Dest event %08x cookie/proc %08x\n",
	      trap->Desc.s.dma_destEvent, trap->Desc.s.dma_destCookieVProc);
    EPRINTF2 (DBG_EPTRAP, "              Source event %08x cookie/proc %08x\n",
	      trap->Desc.s.dma_srcEvent, trap->Desc.s.dma_srcCookieVProc);

    ASSERT (trap->Status.s.Context & SYS_CONTEXT_BIT);

    switch (trap->Status.s.TrapType)
    {
    case MI_DmaPacketTimedOutOrPacketError:
	if (trap->Desc.s.dma_direction == DMA_WRITE)
	    vp = trap->Desc.s.dma_destVProc;
	else
	    vp = trap->Desc.s.dma_srcVProc;
	
	if (! trap->PacketInfo.s.PacketTimeout)
	    status = ETIMEDOUT;
	else
	{
	    status = EHOSTDOWN;

	    /* XXXX: dma timedout - might want to "restart" tree ? */
	}
	goto retry_dma;

    case MI_DmaFailCountError:
	goto retry_dma;

    case MI_TimesliceDmaQueueOverflow:
	IncrStat (rail, DprocDmaQueueOverflow);

	goto retry_dma;

    case MI_RemoteDmaCommand:
    case MI_RunDmaCommand:
    case MI_DequeueNonSysCntxDma:
    case MI_DequeueSysCntxDma:
	/*
	 * The DMA processor has trapped due to outstanding prefetches from the previous 
	 * dma.  The "current" dma has not been consumed, so we just ignore the trap
	 */
	return (OP_HANDLED);
	
    case MI_EventQueueOverflow:
	IncrStat (rail, DprocEventQueueOverflow);

	if ((event = ep_elan2sdram (&rail->Generic, trap->Desc.s.dma_srcEvent)) != (sdramaddr_t) 0 &&
	    ((type  = elan3_sdram_readl (dev, event + offsetof(E3_Event,ev_Type))) & EV_TYPE_MASK_EVIRQ) == EV_TYPE_EVIRQ)
	{
	    spin_unlock (&ctxt->Device->IntrLock);
	    ep3_event (ctxt, (type & ~(EV_TYPE_MASK_EVIRQ | EV_TYPE_MASK_BCOPY)), OP_LWP);
	    spin_lock (&ctxt->Device->IntrLock);
	}
	return (OP_HANDLED);
	
    case MI_DmaQueueOverflow:
	IncrStat (rail, DprocDmaQueueOverflow);

	if ((event = ep_elan2sdram (&rail->Generic, trap->Desc.s.dma_srcEvent)) != (sdramaddr_t) 0 &&
	    ((type = elan3_sdram_readl (dev, event + offsetof (E3_Event, ev_Type))) & EV_TYPE_MASK_DMA) == EV_TYPE_DMA &&
	    (dma  = ep_elan2sdram (&rail->Generic, (type & ~EV_TYPE_MASK2))) != (sdramaddr_t) 0)
	{
	    elan3_sdram_copyq_from_sdram (dev, dma, &dmabe, sizeof (E3_DMA));
	    
	    /* We only chain together DMA's of the same direction, so since
	     * we took a DmaQueueOverflow trap - this means that DMA which
	     * trapped was a WRITE dma - hence the one we chain to must also
	     * be a WRITE dma.
	     */
	    ASSERT (dmabe.s.dma_direction == DMA_WRITE);
	    
	    cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_srcEvent);

#ifdef DEBUG_ASSERT
	    {
		E3_uint16     vp       = dmabe.s.dma_destVProc;
		EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];

		ASSERT (cp != NULL && (!EP_VP_ISDATA(vp) || (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE)));
	    }
#endif
	    cp->Operations->DmaRetry (rail, cp->Arg, &dmabe, EAGAIN);
	    
	    return (OP_HANDLED);
	}

	panic ("ep3_dprocTrap\n");
	return (OP_HANDLED);

    default:
	break;
    }

    /* If it's a dma which traps past the end of the source, then */
    /* just re-issue it */
    numFaults = validTrap = (trap->FaultSave.s.FSR.Status != 0);
    for (i = 0, FaultArea = &trap->Data0; i < 4; i++, FaultArea++)
    {
	if (FaultArea->s.FSR.Status != 0)
	{
	    numFaults++;

	    /* XXXX: Rev B Elans can prefetch data past the end of the dma descriptor */
	    /*       if the fault relates to this, then just ignore it */
	    if (FaultArea->s.FaultAddress >= (trap->Desc.s.dma_source+trap->Desc.s.dma_size))
	    {
		static int i;
		if (i < 10 && i++ < 10)
		    printk ("ep3_dprocTrap: Rev B prefetch trap error %08x %08x\n",
			     FaultArea->s.FaultAddress, (trap->Desc.s.dma_source+trap->Desc.s.dma_size));
		continue;
	    }

	    validTrap++;
	}
    }

    /*
     * NOTE: for physical errors (uncorrectable ECC/PCI parity errors) the FSR will
     *       be zero - hence we will not see any faults - and none will be valid, 
     *       so only ignore a Rev B prefetch trap if we've seen some faults. Otherwise
     *       we can reissue a DMA which has already sent it's remote event !
     */
    if (numFaults != 0 && validTrap == 0)
    {
    retry_dma:
	if (trap->Desc.s.dma_direction == DMA_WRITE)
	{
	    vp = trap->Desc.s.dma_destVProc;
	    cp = LookupEventCookie (rail, &rail->CookieTable, trap->Desc.s.dma_srcEvent);
	}
	else
	{
	    ASSERT (EP3_CONTEXT_ISDATA(trap->Desc.s.dma_queueContext) || trap->Desc.s.dma_direction == DMA_READ_REQUEUE);

	    vp = trap->Desc.s.dma_srcVProc;
	    cp = LookupEventCookie (rail, &rail->CookieTable, trap->Desc.s.dma_destEvent);

	    /* for read dma's set the DMA_READ_REQUEUE bits as the dma_source has been 
	     * modified by the elan to point at the dma in the rxd where it was issued
	     * from */
	    trap->Desc.s.dma_direction = (trap->Desc.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
	}

#ifdef DEBUG_ASSERT
	{
	    EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];

	    ASSERT (!EP_VP_ISDATA(vp) || (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE));
	}
#endif
	
	if (cp != NULL)
	    cp->Operations->DmaRetry (rail, cp->Arg, &trap->Desc, status);
	else
	{
	    ASSERT (trap->Desc.s.dma_direction == DMA_WRITE && trap->Desc.s.dma_srcEvent == 0 && trap->Desc.s.dma_isRemote);

	    QueueDmaForRetry (rail, &trap->Desc, EP_RETRY_ANONYMOUS);
	}

	return (OP_HANDLED);
    }
    
    printk ("ep3_dprocTrap: WakeupFnt=%x Cntx=%x SuspAddr=%x TrapType=%s\n",
	     trap->Status.s.WakeupFunction, trap->Status.s.Context, 
	     trap->Status.s.SuspendAddr, MiToName (trap->Status.s.TrapType));
    printk ("                    FaultAddr=%x EventAddr=%x FSR=%x\n",
	     trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress,
	     trap->FaultSave.s.FSR.Status);
    for (i = 0, FaultArea = &trap->Data0; i < 4; i++, FaultArea++)
	printk ("                  %d FaultAddr=%x EventAddr=%x FSR=%x\n", i,
		 FaultArea->s.FaultAddress, FaultArea->s.EventAddress, FaultArea->s.FSR.Status);
    
    printk ("                  type %08x size %08x source %08x dest %08x\n",
	     trap->Desc.s.dma_type, trap->Desc.s.dma_size, trap->Desc.s.dma_source, trap->Desc.s.dma_dest);
    printk ("                  Dest event %08x cookie/proc %08x\n",
	     trap->Desc.s.dma_destEvent, trap->Desc.s.dma_destCookieVProc);
    printk ("                  Source event %08x cookie/proc %08x\n",
	     trap->Desc.s.dma_srcEvent, trap->Desc.s.dma_srcCookieVProc);

//    panic ("ep3_dprocTrap");

    return (OP_HANDLED);
}

int
ep3_tprocTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap)
{
    EP3_RAIL *rail = (EP3_RAIL *) ctxt->Private;

    EPRINTF6 (DBG_EPTRAP, "ep3_tprocTrap: SP=%08x PC=%08x NPC=%08x DIRTY=%08x TRAP=%08x MI=%s\n",
	      trap->sp, trap->pc, trap->npc, trap->DirtyBits.Bits, trap->TrapBits.Bits, MiToName (trap->mi));
    EPRINTF4 (DBG_EPTRAP, "              g0=%08x g1=%08x g2=%08x g3=%08x\n", 
	      trap->Registers[REG_GLOBALS+(0^WordEndianFlip)], trap->Registers[REG_GLOBALS+(1^WordEndianFlip)], 
	      trap->Registers[REG_GLOBALS+(2^WordEndianFlip)], trap->Registers[REG_GLOBALS+(3^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              g4=%08x g5=%08x g6=%08x g7=%08x\n", 
	      trap->Registers[REG_GLOBALS+(4^WordEndianFlip)], trap->Registers[REG_GLOBALS+(5^WordEndianFlip)], 
	      trap->Registers[REG_GLOBALS+(6^WordEndianFlip)], trap->Registers[REG_GLOBALS+(7^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              o0=%08x o1=%08x o2=%08x o3=%08x\n", 
	      trap->Registers[REG_OUTS+(0^WordEndianFlip)], trap->Registers[REG_OUTS+(1^WordEndianFlip)], 
	      trap->Registers[REG_OUTS+(2^WordEndianFlip)], trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              o4=%08x o5=%08x o6=%08x o7=%08x\n", 
	      trap->Registers[REG_OUTS+(4^WordEndianFlip)], trap->Registers[REG_OUTS+(5^WordEndianFlip)], 
	      trap->Registers[REG_OUTS+(6^WordEndianFlip)], trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              l0=%08x l1=%08x l2=%08x l3=%08x\n", 
	      trap->Registers[REG_LOCALS+(0^WordEndianFlip)], trap->Registers[REG_LOCALS+(1^WordEndianFlip)], 
	      trap->Registers[REG_LOCALS+(2^WordEndianFlip)], trap->Registers[REG_LOCALS+(3^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              l4=%08x l5=%08x l6=%08x l7=%08x\n", 
	      trap->Registers[REG_LOCALS+(4^WordEndianFlip)], trap->Registers[REG_LOCALS+(5^WordEndianFlip)], 
	      trap->Registers[REG_LOCALS+(6^WordEndianFlip)], trap->Registers[REG_LOCALS+(7^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              i0=%08x i1=%08x i2=%08x i3=%08x\n", 
	      trap->Registers[REG_INS+(0^WordEndianFlip)], trap->Registers[REG_INS+(1^WordEndianFlip)], 
	      trap->Registers[REG_INS+(2^WordEndianFlip)], trap->Registers[REG_INS+(3^WordEndianFlip)]);
    EPRINTF4 (DBG_EPTRAP, "              i4=%08x i5=%08x i6=%08x i7=%08x\n", 
	      trap->Registers[REG_INS+(4^WordEndianFlip)], trap->Registers[REG_INS+(5^WordEndianFlip)], 
	      trap->Registers[REG_INS+(6^WordEndianFlip)], trap->Registers[REG_INS+(7^WordEndianFlip)]);
    
    ASSERT (trap->Status.s.Context & SYS_CONTEXT_BIT);

    switch (trap->mi)
    {
    case MI_UnimplementedError:
	if (trap->TrapBits.s.ForcedTProcTrap)
	{
	    ASSERT (trap->TrapBits.s.OutputWasOpen == 0);
	    
	    EPRINTF0 (DBG_EPTRAP, "ep3_tprocTrap: ForcedTProcTrap\n");

	    IssueRunThread (rail, SaveThreadToStack (ctxt, trap, FALSE));
	    return (OP_HANDLED);
	}

	if (trap->TrapBits.s.ThreadTimeout)
	{
	    EPRINTF0 (DBG_EPTRAP, "ep3_tprocTrap: ThreadTimeout\n");

	    if (trap->Registers[REG_GLOBALS + (1^WordEndianFlip)] == 0)
		RollThreadToClose (ctxt, trap, trap->TrapBits.s.PacketAckValue);
	    else
	    {
		CompleteEnvelope (rail, trap->Registers[REG_GLOBALS + (1^WordEndianFlip)], trap->TrapBits.s.PacketAckValue);

		RollThreadToClose (ctxt, trap, EP3_PAckStolen);
	    }
		
	    IssueRunThread (rail, SaveThreadToStack (ctxt, trap, FALSE));
	    return (OP_HANDLED);
	}

	if (trap->TrapBits.s.Unimplemented)
	{
	    E3_uint32 instr = ELAN3_OP_LOAD32 (ctxt, trap->pc & PC_MASK);

	    PRINTF1 (ctxt, DBG_EPTRAP, "ep3_tprocTrap: unimplemented instruction %08x\n", instr);

	    if ((instr & OPCODE_MASK) == OPCODE_Ticc &&
		(instr & OPCODE_IMM)  == OPCODE_IMM &&
		(Ticc_COND(instr)     == Ticc_TA))
	    {
		switch (INSTR_IMM(instr))
		{
		case EP3_UNIMP_TRAP_NO_DESCS:
		    StallThreadForNoDescs (rail, trap->Registers[REG_GLOBALS + (1^WordEndianFlip)], 
					   SaveThreadToStack (ctxt, trap, TRUE));
		    return (OP_HANDLED);

		case EP3_UNIMP_TRAP_PACKET_NACKED:
		    CompleteEnvelope (rail, trap->Registers[REG_GLOBALS + (1^WordEndianFlip)], E3_PAckDiscard);

		    IssueRunThread (rail, SaveThreadToStack (ctxt, trap, TRUE));
		    return (OP_HANDLED);

		case EP3_UNIMP_THREAD_HALTED: 
		    StallThreadForHalted (rail, trap->Registers[REG_GLOBALS + (1^WordEndianFlip)], 
					  SaveThreadToStack (ctxt, trap, TRUE));
		    return (OP_HANDLED);

		default:
		    break;
		    
		}
	    }
	}
	break;

    default:
	break;
    }

    /* All other traps should not happen for kernel comms */
    printk ("ep3_tprocTrap: SP=%08x PC=%08x NPC=%08x DIRTY=%08x TRAP=%08x MI=%s\n",
	     trap->sp, trap->pc, trap->npc, trap->DirtyBits.Bits,
	     trap->TrapBits.Bits, MiToName (trap->mi));
    printk ("              FaultSave : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress, trap->FaultSave.s.FSR.Status);
    printk ("              DataFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->DataFaultSave.s.FaultAddress, trap->DataFaultSave.s.EventAddress, trap->DataFaultSave.s.FSR.Status);
    printk ("              InstFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->InstFaultSave.s.FaultAddress, trap->InstFaultSave.s.EventAddress, trap->InstFaultSave.s.FSR.Status);
    printk ("              OpenFault : FaultAddress %08x EventAddress %08x FSR %08x\n",
	     trap->OpenFaultSave.s.FaultAddress, trap->OpenFaultSave.s.EventAddress, trap->OpenFaultSave.s.FSR.Status);

    if (trap->DirtyBits.s.GlobalsDirty)
    {
	printk ("              g0=%08x g1=%08x g2=%08x g3=%08x\n", 
		 trap->Registers[REG_GLOBALS+(0^WordEndianFlip)], trap->Registers[REG_GLOBALS+(1^WordEndianFlip)], 
		 trap->Registers[REG_GLOBALS+(2^WordEndianFlip)], trap->Registers[REG_GLOBALS+(3^WordEndianFlip)]);
	printk ("              g4=%08x g5=%08x g6=%08x g7=%08x\n", 
		 trap->Registers[REG_GLOBALS+(4^WordEndianFlip)], trap->Registers[REG_GLOBALS+(5^WordEndianFlip)], 
		 trap->Registers[REG_GLOBALS+(6^WordEndianFlip)], trap->Registers[REG_GLOBALS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.OutsDirty)
    {
	printk ("              o0=%08x o1=%08x o2=%08x o3=%08x\n", 
		 trap->Registers[REG_OUTS+(0^WordEndianFlip)], trap->Registers[REG_OUTS+(1^WordEndianFlip)], 
		 trap->Registers[REG_OUTS+(2^WordEndianFlip)], trap->Registers[REG_OUTS+(3^WordEndianFlip)]);
	printk ("              o4=%08x o5=%08x o6=%08x o7=%08x\n", 
		 trap->Registers[REG_OUTS+(4^WordEndianFlip)], trap->Registers[REG_OUTS+(5^WordEndianFlip)], 
		 trap->Registers[REG_OUTS+(6^WordEndianFlip)], trap->Registers[REG_OUTS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.LocalsDirty)
    {
	printk ("              l0=%08x l1=%08x l2=%08x l3=%08x\n", 
		 trap->Registers[REG_LOCALS+(0^WordEndianFlip)], trap->Registers[REG_LOCALS+(1^WordEndianFlip)], 
		 trap->Registers[REG_LOCALS+(2^WordEndianFlip)], trap->Registers[REG_LOCALS+(3^WordEndianFlip)]);
	printk ("              l4=%08x l5=%08x l6=%08x l7=%08x\n", 
		 trap->Registers[REG_LOCALS+(4^WordEndianFlip)], trap->Registers[REG_LOCALS+(5^WordEndianFlip)], 
		 trap->Registers[REG_LOCALS+(6^WordEndianFlip)], trap->Registers[REG_LOCALS+(7^WordEndianFlip)]);
    }
    if (trap->DirtyBits.s.InsDirty)
    {
	printk ("              i0=%08x i1=%08x i2=%08x i3=%08x\n", 
		 trap->Registers[REG_INS+(0^WordEndianFlip)], trap->Registers[REG_INS+(1^WordEndianFlip)], 
		 trap->Registers[REG_INS+(2^WordEndianFlip)], trap->Registers[REG_INS+(3^WordEndianFlip)]);
	printk ("              i4=%08x i5=%08x i6=%08x i7=%08x\n", 
		 trap->Registers[REG_INS+(4^WordEndianFlip)], trap->Registers[REG_INS+(5^WordEndianFlip)], 
		 trap->Registers[REG_INS+(6^WordEndianFlip)], trap->Registers[REG_INS+(7^WordEndianFlip)]);
    }
    
//    panic ("ep3_tprocTrap");

    return (OP_HANDLED);
}

int
ep3_iprocTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, int channel)
{
    EP3_RAIL      *rail = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV      *dev = ctxt->Device;
    EP3_COOKIE    *cp;
    sdramaddr_t    event;
    E3_uint32      type;
    sdramaddr_t    dma;
    E3_DMA_BE      dmabe;

    ASSERT (trap->Transactions[0].s.TrTypeCntx.s.Context & SYS_CONTEXT_BIT);

    /*
     * first process the trap to determine the cause
     */
    InspectIProcTrap (ctxt, trap);

    if (! trap->AckSent && trap->LockQueuePointer)		/* Must be a network error in a queueing DMA */
    {								/* packet - unlock the queue */
	IncrStat (rail, QueueingPacketTrap);

	SimulateUnlockQueue (ctxt, trap->LockQueuePointer, FALSE);
	return (OP_HANDLED);
    }

    if (trap->AckSent && trap->BadTransaction)
    {
	spin_unlock (&dev->IntrLock);

	/* NOTE - no network error fixup is necessary for system context
	 *        messages since they are idempotent and are single packet 
	 *        dmas
	 */
	if (EP3_CONTEXT_ISDATA (trap->Transactions[0].s.TrTypeCntx.s.Context))
	{
	    int nodeId = EP3_CONTEXT_TO_NODE(trap->Transactions[0].s.TrTypeCntx.s.Context);
	    
	    if (trap->DmaIdentifyTransaction)
	    {
		printk ("%s: network error on dma identify <%x> from node %d\n", rail->Generic.Name, trap->DmaIdentifyTransaction->s.TrAddr, nodeId);

		ep_queue_network_error (&rail->Generic, nodeId, EP_NODE_NETERR_ATOMIC_PACKET, channel, trap->DmaIdentifyTransaction->s.TrAddr);
	    }
	    else if (trap->ThreadIdentifyTransaction)
	    {
		printk ("%s: network error on thread identify <%x> from node %d\n", rail->Generic.Name, trap->ThreadIdentifyTransaction->s.TrAddr, nodeId);

		ep_queue_network_error (&rail->Generic, nodeId, EP_NODE_NETERR_ATOMIC_PACKET, channel, trap->ThreadIdentifyTransaction->s.TrAddr);
	    }
	    else
	    {
		printk ("%s: network error on dma packet from node %d\n", rail->Generic.Name, nodeId);

		ep_queue_network_error (&rail->Generic, nodeId, EP_NODE_NETERR_DMA_PACKET, channel, 0);
	    }
	}

	spin_lock (&dev->IntrLock);
	return (OP_HANDLED);
    }
    
    if (trap->AckSent)
    {
	if (trap->TrappedTransaction == NULL)
	    return (OP_HANDLED);
	
	while (! trap->TrappedTransaction->s.TrTypeCntx.s.LastTrappedTrans)
	{
	    E3_IprocTrapHeader_BE *hdrp  = trap->TrappedTransaction;
	    E3_IprocTrapData_BE   *datap = trap->TrappedDataBuffer;
	    
	    ASSERT (hdrp->s.TrTypeCntx.s.StatusRegValid != 0);
	    
	    if ((hdrp->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT) != 0)
	    {
		printk ("ep3_iprocTrap: WRITEBLOCK : Addr %x\n", hdrp->s.TrAddr);
//		panic ("ep3_iprocTrap\n");
	    }
	    else
	    {
		switch (hdrp->s.TrTypeCntx.s.Type & TR_OPCODE_TYPE_MASK)
		{
		case TR_SETEVENT & TR_OPCODE_TYPE_MASK:
		    switch (GET_STATUS_TRAPTYPE (hdrp->s.IProcTrapStatus))
		    {
		    case MI_DmaQueueOverflow:
			IncrStat (rail, IprocDmaQueueOverflow);

			if ((event = ep_elan2sdram (&rail->Generic, hdrp->s.TrAddr)) != (sdramaddr_t) 0 &&
			    ((type = elan3_sdram_readl (dev, event + offsetof (E3_Event, ev_Type))) & EV_TYPE_MASK_DMA) == EV_TYPE_DMA &&
			    (dma  = ep_elan2sdram (&rail->Generic, (type & ~EV_TYPE_MASK2))) != (sdramaddr_t) 0)
			{
			    elan3_sdram_copyq_from_sdram (dev, dma, &dmabe, sizeof (E3_DMA));
			    
			    if (dmabe.s.dma_direction == DMA_WRITE)
				cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_srcEvent);
			    else
			    {
				cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_destEvent);
				
				/* we MUST convert this into a DMA_READ_REQUEUE dma as if we don't the 
				 * DMA descriptor will be read from the EP3_RETRY_DMA rather than the 
				 * original DMA - this can then get reused and an incorrect DMA 
				 * descriptor sent 
				 * eventp->ev_Type contains the dma address with type in the lower bits 
				 */ 
				
				dmabe.s.dma_source    = (type & ~EV_TYPE_MASK2);
				dmabe.s.dma_direction = (dmabe.s.dma_direction & ~DMA_READ) | DMA_READ_REQUEUE;
			    }

#ifdef DEBUG_ASSERT
			    {
				E3_uint16     vp       = (dmabe.s.dma_direction == DMA_WRITE ? dmabe.s.dma_destVProc : dmabe.s.dma_srcVProc);
				EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];

				ASSERT (!EP_VP_ISDATA(vp) || (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE));
			    }
#endif

			    if (cp != NULL)
				cp->Operations->DmaRetry (rail, cp->Arg, &dmabe, EAGAIN);
			    else
			    {
				ASSERT (dmabe.s.dma_direction == DMA_WRITE && dmabe.s.dma_srcEvent == 0 && dmabe.s.dma_isRemote);
				
				QueueDmaForRetry (rail, &dmabe, EP_RETRY_ANONYMOUS);
			    }
			    break;
			}

			printk ("ep3_iprocTrap: SETEVENT : %x - cannot find dma to restart\n", hdrp->s.TrAddr);
//			panic ("ep3_iprocTrap\n");
			break;

		    case MI_EventQueueOverflow:
		    {
			sdramaddr_t event;
			E3_uint32   type;

			IncrStat (rail, IprocEventQueueOverflow);

			if ((event = ep_elan2sdram (&rail->Generic, hdrp->s.TrAddr)) != (sdramaddr_t) 0 &&
			    ((type = elan3_sdram_readl (dev, event + offsetof (E3_Event, ev_Type))) & EV_TYPE_MASK_EVIRQ) == EV_TYPE_EVIRQ)
			{
			    spin_unlock (&dev->IntrLock);
			    ep3_event (ctxt, (type & ~(EV_TYPE_MASK_EVIRQ|EV_TYPE_MASK_BCOPY)), OP_LWP);
			    spin_lock (&dev->IntrLock);

			    break;
			}

			printk ("ep3_iprocTrap: SETEVENT : %x - cannot find event\n", hdrp->s.TrAddr);
//			panic ("ep3_iprocTrap\n");
			break;
		    }

		    default:
			printk ("ep3_iprocTrap: SETEVENT : %x MI=%x\n", hdrp->s.TrAddr, GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus));
//			panic ("ep3_iprocTrap\n");
			break;
		    }
		    break;
		    
		case TR_SENDDISCARD & TR_OPCODE_TYPE_MASK:
		    /* Just ignore send-discard transactions */
		    break;
		    
		case TR_REMOTEDMA & TR_OPCODE_TYPE_MASK:
		{
		    E3_DMA_BE *dmap = (E3_DMA_BE *) datap;

		    if (GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus) != MI_DmaQueueOverflow)
		    {
			printk ("ep3_iprocTrap: MI=%x\n", GET_STATUS_TRAPTYPE(hdrp->s.IProcTrapStatus));
			break;
		    }

		    IncrStat (rail, IprocDmaQueueOverflow);

		    cp = LookupEventCookie (rail, &rail->CookieTable, dmap->s.dma_srcEvent);

		    /* modify the dma type since it will still be a "read" dma */
		    dmap->s.dma_type = (dmap->s.dma_type & ~DMA_TYPE_READ) | DMA_TYPE_ISREMOTE;

#ifdef DEBUG_ASSERT
		    {
			E3_uint16     vp       = dmap->s.dma_destVProc;
			EP_NODE_RAIL *nodeRail = &rail->Generic.Nodes[EP_VP_TO_NODE(vp)];
			
			ASSERT (!EP_VP_ISDATA(vp) || (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE));
		    }
#endif
		    if (cp != NULL)
			cp->Operations->DmaRetry (rail, cp->Arg, dmap, EAGAIN);
		    else
		    {
			ASSERT (dmap->s.dma_direction == DMA_WRITE && dmap->s.dma_srcEvent == 0 && dmap->s.dma_isRemote);
			
			QueueDmaForRetry (rail, dmap, EP_RETRY_ANONYMOUS);
		    }
		    break;
		}   
		default:
		    printk ("ep3_iprocTrap: %s\n", IProcTrapString (hdrp, datap));
		    break;
		}
	    }
	    
	    /*
	     * We've successfully processed this transaction, so move onto the 
	     * next one.
	     */
	    trap->TrappedTransaction++;
	    trap->TrappedDataBuffer++;
	}

	return (OP_HANDLED);
    }
    
    /* Workaround WRITEBLOCK transaction executed when LOCKQUEUE transaction missed */
    if ((trap->TrappedTransaction->s.TrTypeCntx.s.Type & TR_WRITEBLOCK_BIT) && 	/* a DMA packet */
	trap->LockQueuePointer == 0 && trap->UnlockQueuePointer && 		/* a queueing DMA */
	trap->TrappedTransaction->s.TrAddr == trap->FaultSave.s.FaultAddress)	/* and missed lockqueue */
    {
	printk ("ep3_iprocTrap: missed lockqueue transaction for queue %x\n", trap->UnlockQueuePointer);
	return (OP_HANDLED);
    }

    if (trap->FaultSave.s.FaultContext != 0)
	printk ("ep3_iprocTrap: pagefault at %08x in context %x\n", 
		trap->FaultSave.s.FaultAddress, trap->FaultSave.s.FaultContext);

//    panic ("ep3_iprocTrap: unexpected inputter trap\n");
    
    return (OP_HANDLED);
}

/*
 * Command processor trap
 *   kernel comms should only be able to generate
 *   queue overflow traps
 */
int
ep3_cprocTrap (ELAN3_CTXT *ctxt, COMMAND_TRAP *trap)
{
    EP3_RAIL     *rail   = (EP3_RAIL *) ctxt->Private;
    int           ctxnum = (trap->TrapBuf.r.Breg >> 16) & MAX_ROOT_CONTEXT_MASK;
    ELAN3_DEV     *dev    = rail->Device;
    EP3_DMA_RING  *ring;
    EP3_COOKIE   *cp;
    E3_DMA_BE     dmabe;
    int           vp, slot;
    unsigned long flags;

    switch (trap->Status.s.TrapType)
    {
    case MI_DmaQueueOverflow:
	IncrStat (rail, CprocDmaQueueOverflow);

	/* Use the context number that the setevent was issued in,
	 * to find the appropriate dma ring, then since they are guaranteed
	 * to be issued in order, we just search backwards till we find the
	 * last one which has completed its word copy - this must be the
	 * one which had caused the DmaQueueOverflow trap ! */

	ASSERT (ctxnum >= ELAN3_DMARING_BASE_CONTEXT_NUM && ctxnum < (ELAN3_DMARING_BASE_CONTEXT_NUM+EP3_NUM_RINGS));

	spin_lock_irqsave (&dev->CProcLock, flags);

	ring = &rail->DmaRings[ctxnum - ELAN3_DMARING_BASE_CONTEXT_NUM];
	slot = DMA_RING_PREV_POS(ring, ring->Position);
	
	while (ring->pDoneBlk[slot] == EP3_EVENT_ACTIVE)
	    slot = DMA_RING_PREV_POS(ring, slot);
	
	elan3_sdram_copyq_from_sdram (rail->Device , DMA_RING_DMA(ring,slot), &dmabe, sizeof (E3_DMA));

#if defined(DEBUG_ASSERT)
	while (slot != DMA_RING_PREV_POS(ring, ring->Position))
	{
	    ASSERT (ring->pDoneBlk[slot] != EP3_EVENT_ACTIVE);
	    
	    slot = DMA_RING_PREV_POS(ring, slot);
	}
#endif
	spin_unlock_irqrestore (&dev->CProcLock, flags);

	if (dmabe.s.dma_direction == DMA_WRITE)
	    cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_srcEvent);
	else
	{
	    ASSERT (dmabe.s.dma_direction = DMA_READ_REQUEUE);

	    cp = LookupEventCookie (rail, &rail->CookieTable, dmabe.s.dma_destEvent);
	}

#if defined(DEBUG_ASSERT)
	if (dmabe.s.dma_direction == DMA_WRITE)
	    vp = dmabe.s.dma_destVProc;
	else
	    vp = dmabe.s.dma_srcVProc;
	
	ASSERT (!EP_VP_ISDATA(vp) || (rail->Generic.Nodes[EP_VP_TO_NODE(vp)].State >= EP_NODE_CONNECTED &&
				      rail->Generic.Nodes[EP_VP_TO_NODE(vp)].State <= EP_NODE_LOCAL_PASSIVATE));
#endif

	if (cp != NULL)
	    cp->Operations->DmaRetry (rail, cp->Arg, &dmabe, EAGAIN);
	else
	{
	    ASSERT (dmabe.s.dma_direction == DMA_WRITE && dmabe.s.dma_srcEvent == 0 && dmabe.s.dma_isRemote);
	    
	    QueueDmaForRetry (rail, &dmabe, EP_RETRY_ANONYMOUS);
	}
	
	return (OP_HANDLED);

    case MI_EventQueueOverflow:
	ASSERT (ctxnum == ELAN3_MRF_CONTEXT_NUM);

	IncrStat (rail, CprocEventQueueOverflow);
	
	rail->CommandPortEventTrap = TRUE;
	return (OP_HANDLED);
	
#if defined(PER_CPU_TIMEOUT)
    case MI_SetEventReadWait:
	if (ctxnum == ELAN3_MRF_CONTEXT_NUM && trap->FaultSave.s.EventAddress == EP_PACEMAKER_EVENTADDR)
	{
	    HeartbeatPacemaker (rail);
	    return (OP_HANDLED);
	}
#endif

    default:
	printk ("ep3_cprocTrap : Context=%x Status=%x TrapType=%x\n", ctxnum, trap->Status.Status, trap->Status.s.TrapType);
	printk ("               FaultAddr=%x EventAddr=%x FSR=%x\n",
		 trap->FaultSave.s.FaultAddress, trap->FaultSave.s.EventAddress,
		 trap->FaultSave.s.FSR.Status);
	break;
    }

//    panic ("ep3_cprocTrap");

    return (OP_HANDLED);
}

static int
ep3_cprocReissue (ELAN3_CTXT *ctxt, CProcTrapBuf_BE *tbuf)
{
    EP3_RAIL   *rail    = (EP3_RAIL *) ctxt->Private;
    unsigned  cmdoff = (tbuf->s.ContextType >> 5) & 0xFF;
    int       ctxnum = (tbuf->s.ContextType >> 16) & MAX_ROOT_CONTEXT_MASK;
    
    if (ctxnum >= ELAN3_DMARING_BASE_CONTEXT_NUM && ctxnum < (ELAN3_DMARING_BASE_CONTEXT_NUM+EP3_NUM_RINGS))
    {
	EP3_DMA_RING *ring = &rail->DmaRings[ctxnum - ELAN3_DMARING_BASE_CONTEXT_NUM];

	ASSERT ((cmdoff << 2) == offsetof (E3_CommandPort, SetEvent)); /* can only be setevent commands! */
	ASSERT (tbuf->s.Addr >= DMA_RING_EVENT_ELAN(ring,0) && tbuf->s.Addr < DMA_RING_EVENT_ELAN(ring, ring->Entries));
	
	writel (tbuf->s.Addr, (void *)(ring->CommandPort + (cmdoff << 2)));
    }
    else
    {
	ASSERT (ctxnum == ELAN3_MRF_CONTEXT_NUM);

	writel (tbuf->s.Addr, (void *)(ctxt->CommandPort + (cmdoff << 2)));
    }
    
    return (OP_HANDLED);
}

static E3_uint8
ep3_load8 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    EP3_RAIL    *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV    *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint8   *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	return (elan3_sdram_readb (dev, offset));
    if ((ptr = ep_elan2main (&rail->Generic, addr)) != NULL)
	return (*ptr);

    printk ("ep3_load8: %08x\n", addr);
    return (0);
}

static void
ep3_store8 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint8 val)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint8   *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	elan3_sdram_writeb (dev, offset, val);
    else if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	*ptr = val;
    else
	printk ("ep3_store8 %08x\n", addr);
}

static E3_uint16
ep3_load16 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint16  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	return (elan3_sdram_readw (dev, offset));
    if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	return (*ptr);

    printk ("ep3_load16 %08x\n", addr);
    return (0);
}

static void
ep3_store16 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint16 val)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint16  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	elan3_sdram_writew (dev, offset, val);
    else if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	*ptr = val;
    else
	printk ("ep3_store16 %08x\n", addr);
}

static E3_uint32
ep3_load32 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint32  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	return (elan3_sdram_readl(dev, offset));
    if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	return (*ptr);
    
    printk ("ep3_load32 %08x\n", addr);
    return (0);
}

static void
ep3_store32 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint32 val)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint32  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	elan3_sdram_writel (dev, offset, val);
    else if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	*ptr = val;
    else
	printk ("ep3_store32 %08x\n", addr);
}

static E3_uint64
ep3_load64 (ELAN3_CTXT *ctxt, E3_Addr addr)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint64  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	return (elan3_sdram_readq (dev, offset));
    if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	return (*ptr);

    printk ("ep3_load64 %08x\n", addr);
    return (0);
}

static void
ep3_store64 (ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint64 val)
{
    EP3_RAIL   *rail  = (EP3_RAIL *) ctxt->Private;
    ELAN3_DEV   *dev = ctxt->Device;
    sdramaddr_t offset;
    E3_uint64  *ptr;

    if ((offset = ep_elan2sdram (&rail->Generic, addr)) != 0)
	elan3_sdram_writeq (dev, offset, val);
    else if ((ptr = ep_elan2main (&rail->Generic, addr)) != 0)
	*ptr = val;
    else
	printk ("ep3_store64 %08x\n", addr);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
