/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: support_elan4.c,v 1.24.2.2 2005/08/09 09:57:14 mike Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/epmod/support_elan4.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/kthread.h>

#include <elan/kcomm.h>

#include "kcomm_vp.h"
#include "kcomm_elan4.h"
#include "debug.h"

#include <elan4/trtype.h>
#include <elan4/debug.h>

void
ep4_register_intcookie (EP4_RAIL *rail, EP4_INTCOOKIE *cp, E4_uint64 cookie, void (*callback)(EP4_RAIL *r, void *arg), void *arg)
{
    unsigned long flags;
    
    cp->int_val      = cookie;
    cp->int_callback = callback;
    cp->int_arg      = arg;
	
    spin_lock_irqsave (&rail->r_intcookie_lock, flags);
    list_add_tail (&cp->int_link, &rail->r_intcookie_hash[EP4_INTCOOKIE_HASH(cookie)]);
    spin_unlock_irqrestore (&rail->r_intcookie_lock, flags);
}

void
ep4_deregister_intcookie (EP4_RAIL *rail, EP4_INTCOOKIE *cp)
{
    unsigned long flags;
    
    spin_lock_irqsave (&rail->r_intcookie_lock, flags);
    list_del (&cp->int_link);
    spin_unlock_irqrestore (&rail->r_intcookie_lock, flags);
}


EP4_INTCOOKIE *
ep4_lookup_intcookie (EP4_RAIL *rail, E4_uint64 cookie)
{
    struct list_head *el;
    unsigned long flags;

    spin_lock_irqsave (&rail->r_intcookie_lock, flags);
    list_for_each (el, &rail->r_intcookie_hash[EP4_INTCOOKIE_HASH(cookie)]) {
	EP4_INTCOOKIE *cp = list_entry (el, EP4_INTCOOKIE, int_link);

	if (cp->int_val == cookie)
	{
	    spin_unlock_irqrestore (&rail->r_intcookie_lock, flags);
	    return cp;
	}
    }
    spin_unlock_irqrestore (&rail->r_intcookie_lock, flags);
    return NULL;
}

E4_uint64
ep4_neterr_cookie (EP4_RAIL *rail, unsigned int node)
{
    E4_uint64      cookie;
    unsigned long  flags;

    spin_lock_irqsave (&rail->r_cookie_lock, flags);
    cookie = rail->r_cookies[node];

    rail->r_cookies[node] += EP4_COOKIE_INC;
    
    spin_unlock_irqrestore (&rail->r_cookie_lock, flags);

    return cookie;
}

void
ep4_eproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status)
{
    EP4_RAIL        *rail = EP4_CTXT_TO_RAIL (ctxt);
    ELAN4_EPROC_TRAP trap;

    elan4_extract_eproc_trap (ctxt->ctxt_dev, status, &trap, 0);

    if (epdebug & DBG_EPTRAP)
	elan4_display_eproc_trap (DBG_BUFFER, 0, "ep4_eproc_trap", &trap);

    switch (EPROC_TrapType (status))
    {
    case EventProcNoFault:
	EPRINTF1 (DBG_EPTRAP, "%s: EventProcNoFault\n", rail->r_generic.Name);
	return;

    default:
	printk ("%s: unhandled eproc trap %d\n", rail->r_generic.Name, EPROC_TrapType (status));
	elan4_display_eproc_trap (DBG_CONSOLE, 0, "ep4_eproc_trap", &trap);
    }
}

void
ep4_cproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned cqnum)
{
    EP4_RAIL        *rail = EP4_CTXT_TO_RAIL (ctxt);
    ELAN4_CPROC_TRAP trap;
    struct list_head *el;
    register int      i;

    elan4_extract_cproc_trap (ctxt->ctxt_dev, status, &trap, cqnum);

    if (epdebug & DBG_EPTRAP)
	elan4_display_cproc_trap (DBG_BUFFER, 0, "ep4_cproc_trap", &trap);
	
    switch (CPROC_TrapType (status))
    {
    case CommandProcInterruptQueueOverflow:
	/*
	 * Try and handle a bunch of elan main interrupts
	 */
	for (i = 0; i <EP4_NUM_ECQ; i++) {
	    list_for_each (el, &rail->r_ecq_list[i]) {
	        EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);
	    
		if (elan4_cq2num (ecq->ecq_cq) == cqnum)
		{
		    printk ("%s: defer command queue %d after trap %x\n",
			    rail->r_generic.Name, cqnum, CPROC_TrapType (status));
	
		    elan4_queue_mainintop (ctxt->ctxt_dev, &ecq->ecq_intop);
		    return;
		}
	    }
	}
	break;

    case CommandProcDmaQueueOverflow:
    case CommandProcThreadQueueOverflow:
	for (i = 0; i <EP4_NUM_ECQ; i++) {
	    list_for_each (el, &rail->r_ecq_list[i]) {
	        EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);
	    
		if (elan4_cq2num (ecq->ecq_cq) == cqnum)
		{
		    printk ("%s: restart command queue %d after trap %x\n",
			    rail->r_generic.Name, cqnum, CPROC_TrapType (status));

		    elan4_restartcq (ctxt->ctxt_dev, ecq->ecq_cq);
		    return;
		}
	    }
	}
	break;
    }

    printk ("%s: unhandled cproc trap %d for cqnum %d\n", rail->r_generic.Name, CPROC_TrapType (status), cqnum);
    elan4_display_cproc_trap (DBG_CONSOLE, 0, "ep4_cproc_trap", &trap);
}

void
ep4_dproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    EP4_RAIL        *rail = EP4_CTXT_TO_RAIL (ctxt);
    ELAN4_DPROC_TRAP trap;

    elan4_extract_dproc_trap (ctxt->ctxt_dev, status, &trap, unit);

    if (epdebug & DBG_EPTRAP)
	elan4_display_dproc_trap (DBG_BUFFER, 0, "ep4_dproc_trap", &trap);

    if (! DPROC_PrefetcherFault (trap.tr_status))
    {
	switch (DPROC_TrapType (trap.tr_status))
	{
	case DmaProcFailCountError:
	    goto retry_this_dma;

	case DmaProcPacketAckError:
	    goto retry_this_dma;

	case DmaProcQueueOverflow:
	    goto retry_this_dma;
	}
    }

    printk ("%s: unhandled dproc trap\n", rail->r_generic.Name);
    elan4_display_dproc_trap (DBG_CONSOLE, 0, "ep4_dproc_trap", &trap);
    return;

 retry_this_dma:
    /*XXXX implement backoff .... */

    ep4_queue_dma_retry (rail, &trap.tr_desc, EP_RETRY_LOW_PRI);
}

void
ep4_tproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status)
{
    EP4_RAIL         *rail = EP4_CTXT_TO_RAIL (ctxt);
    ELAN4_TPROC_TRAP *trap = &rail->r_tproc_trap;

    elan4_extract_tproc_trap (ctxt->ctxt_dev, status, trap);

    if (epdebug & DBG_EPTRAP)
	elan4_display_tproc_trap (DBG_BUFFER, 0, "ep4_tproc_trap", trap);
	
    printk ("%s: unhandled tproc trap\n", rail->r_generic.Name);
    elan4_display_tproc_trap (DBG_CONSOLE, 0, "ep4_tproc_trap", trap);
}

void
ep4_iproc_trap (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit)
{
    EP4_RAIL         *rail = EP4_CTXT_TO_RAIL (ctxt);
    ELAN4_IPROC_TRAP *trap = &rail->r_iproc_trap;

    elan4_extract_iproc_trap (ctxt->ctxt_dev, status, trap, unit);

    if (epdebug & DBG_EPTRAP)
	elan4_display_iproc_trap (DBG_BUFFER, 0, "ep4_iproc_trap", trap);
	
    elan4_inspect_iproc_trap (trap);

    switch (IPROC_TrapValue (trap->tr_transactions[trap->tr_trappedTrans].IProcStatusCntxAndTrType))
    {
    case InputDmaQueueOverflow:
	ep4_queue_dma_retry (rail, (E4_DMA *) &trap->tr_dataBuffers[trap->tr_trappedTrans], EP_RETRY_LOW_PRI);
	return;

    case InputEventEngineTrapped:
    {
	E4_IprocTrapHeader *hdrp = &trap->tr_transactions[trap->tr_trappedTrans];
	sdramaddr_t         inputq;
	E4_Addr		    event;

	/* XXXX: flow control on the command queue which we issue to is 
	 * rather difficult, we don't want to have space for an event 
	 * for each possible context, nor the mechanism to hold the 
	 * context filter up until the event has been executed.  Given
	 * that the event engine will be restarted by this same interrupt
	 * and we're using high priority command queues, then we just use
	 * a single small command queue for this.
	 */
	switch (IPROC_TransactionType(hdrp->IProcStatusCntxAndTrType) & TR_OPCODE_MASK)
	{
	case TR_SETEVENT & TR_OPCODE_MASK:
	    if (hdrp->TrAddr != 0)
		ep4_set_event_cmd (rail->r_event_ecq, hdrp->TrAddr);
	    return;

	case TR_INPUT_Q_COMMIT & TR_OPCODE_MASK:
	    if ((inputq = ep_elan2sdram (&rail->r_generic, hdrp->TrAddr)) == 0)
		printk ("%s: TR_INPUT_Q_COMMIT at %llx is not sdram\n", rail->r_generic.Name, (long long)hdrp->TrAddr);
	    else
	    {
		if ((event = elan4_sdram_readq (rail->r_ctxt.ctxt_dev, inputq + offsetof (E4_InputQueue, q_event))) != 0)
		    ep4_set_event_cmd (rail->r_event_ecq, event);
		return;
	    }
	}
	break;
    }

    case InputEopErrorOnWaitForEop:
    case InputEopErrorTrap:
    case InputCrcErrorAfterPAckOk:
	if (! (trap->tr_flags & TR_FLAG_ACK_SENT) || (trap->tr_flags & TR_FLAG_EOP_BAD))
	    return;
	
	if (EP4_CONTEXT_ISDATA (IPROC_NetworkContext (status)))
	{
	    unsigned int nodeId = EP4_CONTEXT_TO_NODE (IPROC_NetworkContext (status));

	    if ((trap->tr_flags & (TR_FLAG_DMA_PACKET | TR_FLAG_BAD_TRANS)) || 
		((trap->tr_flags & TR_FLAG_EOP_ERROR) && (trap->tr_identifyTrans == TR_TRANS_INVALID)))
	    {
		EPRINTF2  (DBG_NETWORK_ERROR, "%s: network error on dma packet from node %d\n", rail->r_generic.Name, nodeId);
		printk ("%s: network error on dma packet from node %d\n", rail->r_generic.Name, nodeId);

		ep_queue_network_error (&rail->r_generic, EP4_CONTEXT_TO_NODE(IPROC_NetworkContext (status)), EP_NODE_NETERR_DMA_PACKET, unit & 1, 0);
		return;
	    }
	    
	    if (trap->tr_flags & TR_FLAG_EOP_ERROR)
	    {
		E4_uint64        status = trap->tr_transactions[trap->tr_identifyTrans].IProcStatusCntxAndTrType;
		EP_NETERR_COOKIE cookie = 0;

		switch (IPROC_TransactionType (status) & TR_OPCODE_MASK)
		{
		case TR_SETEVENT_IDENTIFY & TR_OPCODE_MASK:
		    if (IPROC_TrapValue(status) == InputNoFault)
			cookie = trap->tr_transactions[trap->tr_identifyTrans].TrAddr;
		    else
			cookie = trap->tr_dataBuffers[trap->tr_identifyTrans].Data[0];
		    EPRINTF3(DBG_NETWORK_ERROR, "%s: network error on setevent <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    printk ("%s: network error on setevent <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    break;

		case TR_INPUT_Q_COMMIT & TR_OPCODE_MASK:
		    if (IPROC_TrapValue(status) == InputNoFault)
			cookie = trap->tr_transactions[trap->tr_identifyTrans].TrAddr;
		    else
			cookie = trap->tr_dataBuffers[trap->tr_identifyTrans].Data[0];
		    EPRINTF3 (DBG_NETWORK_ERROR, "%s: network error on queue commit <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    printk ("%s: network error on queue commit <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    break;
		    
		case TR_REMOTEDMA & TR_OPCODE_MASK:
		    cookie = trap->tr_transactions[trap->tr_identifyTrans].TrAddr;
		    EPRINTF3 (DBG_NETWORK_ERROR, "%s: network error on remote dma <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    printk ("%s: network error on remote dma <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    break;

		case TR_IDENTIFY & TR_OPCODE_MASK:
		    cookie = trap->tr_transactions[trap->tr_identifyTrans].TrAddr;
		    EPRINTF3 (DBG_NETWORK_ERROR, "%s: network error on identify <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    printk ("%s: network error on identify <%lld%s%s%s%s> from node %d\n", rail->r_generic.Name, EP4_COOKIE_STRING(cookie), nodeId);
		    break;

		default:
		    panic ("%s: unknown identify transaction type %x for eop error from node %d\n", rail->r_generic.Name,
			    IPROC_TransactionType (trap->tr_transactions[trap->tr_identifyTrans].IProcStatusCntxAndTrType), nodeId);
		    break;
		}

		ep_queue_network_error (&rail->r_generic, nodeId, EP_NODE_NETERR_ATOMIC_PACKET, unit & 1, cookie);
	    }
	}
	return;
    }

    printk ("%s: unhandled iproc trap\n", rail->r_generic.Name);
    elan4_display_iproc_trap (DBG_CONSOLE, 0, "ep4_iproc_trap", trap);
}

void
ep4_interrupt (ELAN4_CTXT *ctxt, E4_uint64 cookie)
{
    EP4_RAIL      *rail = EP4_CTXT_TO_RAIL (ctxt);
    EP4_INTCOOKIE *cp  = ep4_lookup_intcookie (rail, cookie);

    if (cp == NULL)
    {
	printk ("ep4_interrupt: cannot find event cookie for %016llx\n", (long long) cookie);
	return;
    }

    cp->int_callback (rail, cp->int_arg);
}

ELAN4_TRAP_OPS ep4_trap_ops = 
{
    ep4_eproc_trap,
    ep4_cproc_trap,
    ep4_dproc_trap,
    ep4_tproc_trap,
    ep4_iproc_trap,
    ep4_interrupt,
};

void
ep4_flush_filters (EP_RAIL *r)
{
    /* nothing to do here as elan4_set_filter() flushes the context filter */
}

struct flush_queues_desc
{
    EP4_RAIL      *rail;
    volatile int   done;
} ;

static void
ep4_flush_queues_flushop (ELAN4_DEV *dev, void *arg, int qfull)
{
    struct flush_queues_desc *desc  = (struct flush_queues_desc *) arg;
    EP4_RAIL                 *rail  = desc->rail;
    E4_uint64		      qptrs = read_reg64 (dev, DProcHighPriPtrs);
    E4_uint32                 qsize = E4_QueueSize (E4_QueueSizeValue (qptrs));
    E4_uint32                 qfptr = E4_QueueFrontPointer (qptrs);
    E4_uint32                 qbptr = E4_QueueBackPointer (qptrs);
    E4_DProcQueueEntry        qentry;
    unsigned long             flags;

    while ((qfptr != qbptr) || qfull)
    {
	E4_uint64 typeSize = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_typeSize));
	
	if (DMA_Context (qentry.Desc.dma_typeSize) == rail->r_ctxt.ctxt_num)
	{
	    E4_uint64     vp       = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_vproc));
	    EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[EP_VP_TO_NODE(vp)];
	    
	    EP4_ASSERT (rail, !EP_VP_ISDATA(vp) || (nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE));
	    
	    if (EP_VP_ISDATA(vp) && nodeRail->State == EP_NODE_LOCAL_PASSIVATE)
	    {
		/*
		 * This is a DMA going to the node which is being removed, 
		 * so move it onto the node dma list where it will get
		 * handled later.
		 */
		qentry.Desc.dma_typeSize = typeSize;
		qentry.Desc.dma_cookie   = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_cookie));
		qentry.Desc.dma_vproc    = vp;
		qentry.Desc.dma_srcAddr  = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_srcAddr));
		qentry.Desc.dma_dstAddr  = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_dstAddr));
		qentry.Desc.dma_srcEvent = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_srcEvent));
		qentry.Desc.dma_dstEvent = elan4_sdram_readq (dev, qfptr + offsetof (E4_DProcQueueEntry, Desc.dma_dstEvent));
		
		EPRINTF4 (DBG_RETRY, "ep4_flush_dmas: %016llx %016llx %016llx %016llx\n", (long long)qentry.Desc.dma_typeSize, 
			  (long long)qentry.Desc.dma_cookie,  (long long)qentry.Desc.dma_vproc,  (long long)qentry.Desc.dma_srcAddr);
		EPRINTF3 (DBG_RETRY, "                %016llx %016llx %016llx\n",  (long long)qentry.Desc.dma_dstAddr, 
			  (long long)qentry.Desc.dma_srcEvent,  (long long)qentry.Desc.dma_dstEvent);
		
		ep4_queue_dma_stalled (rail, &qentry.Desc);
		
		qentry.Desc.dma_typeSize = DMA_ShMemWrite | dev->dev_ctxt.ctxt_num;
		qentry.Desc.dma_cookie   = 0;
		qentry.Desc.dma_vproc    = 0;
		qentry.Desc.dma_srcAddr  = 0;
		qentry.Desc.dma_dstAddr  = 0;
		qentry.Desc.dma_srcEvent = 0;
		qentry.Desc.dma_dstEvent = 0;
		
		elan4_sdram_copyq_to_sdram (dev, &qentry, qfptr, sizeof (E4_DProcQueueEntry));
	    }
	}

	qfptr = (qfptr & ~(qsize-1)) | ((qfptr + sizeof (E4_DProcQueueEntry)) & (qsize-1));
	qfull = 0;
    }

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    desc->done = 1;
    kcondvar_wakeupall (&rail->r_haltop_sleep, &rail->r_haltop_lock);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
}

static void
ep4_flush_queues_haltop (ELAN4_DEV *dev, void *arg)
{
    struct flush_queues_desc *desc = (struct flush_queues_desc *) arg;

    elan4_queue_dma_flushop (dev, &desc->rail->r_flushop, 1);
}

void
ep4_flush_queues (EP_RAIL *r)
{
    EP4_RAIL *rail = (EP4_RAIL *) r;
    struct flush_queues_desc desc;
    struct list_head *el, *nel;
    unsigned long flags;
    int i;

    /* initialise descriptor */
    desc.rail  = rail;
    desc.done  = 0;

    /* First -  stall the dma retry thread, so that it will no longer restart
     *          any dma's from the retry list */
    ep_kthread_stall (&rail->r_retry_thread);

    /* Second - flush through all command queues targetted by events, thread etc */
    ep4_flush_ecqs (rail);

    /* Third - queue a halt operation to flush through all DMA's which are executing
     *         or on the run queues */
    kmutex_lock (&rail->r_haltop_mutex);

    rail->r_haltop.op_mask      = INT_DProcHalted;
    rail->r_haltop.op_function  = ep4_flush_queues_haltop;
    rail->r_haltop.op_arg       = &desc;

    rail->r_flushop.op_function = ep4_flush_queues_flushop;
    rail->r_flushop.op_arg      = &desc;
    
    elan4_queue_haltop (rail->r_ctxt.ctxt_dev, &rail->r_haltop);

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    while (! desc.done)
	kcondvar_wait (&rail->r_haltop_sleep, &rail->r_haltop_lock, &flags);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
    kmutex_unlock (&rail->r_haltop_mutex);

    /* Fourth - run down the dma retry lists and move all entries to the cancelled
     *          list.  Any dma's which were on the run queues have already been
     *          moved there */
    spin_lock_irqsave (&rail->r_dma_lock, flags);
    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	list_for_each_safe (el,nel, &rail->r_dma_retrylist[i]) {
	    EP4_DMA_RETRY *retry    = list_entry (el, EP4_DMA_RETRY, retry_link);
	    EP_NODE_RAIL  *nodeRail = &rail->r_generic.Nodes[EP_VP_TO_NODE(retry->retry_dma.dma_vproc)];

	    EP4_ASSERT (rail, nodeRail->State >= EP_NODE_CONNECTED && nodeRail->State <= EP_NODE_LOCAL_PASSIVATE);

	    if (nodeRail->State == EP_NODE_LOCAL_PASSIVATE)
	    {
		list_del (&retry->retry_link);
		list_add_tail (&retry->retry_link, &nodeRail->StalledDmas);
	    }
	}
    }
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
    
    /* Finally - allow the retry thread to run again */
    ep_kthread_resume (&rail->r_retry_thread);
}

struct write_qdesc_desc
{
    EP4_RAIL      *rail;
    sdramaddr_t    qaddr;
    E4_InputQueue *qdesc;
    volatile int   done;
} ;

static void
ep4_write_qdesc_haltop (ELAN4_DEV *dev, void *arg)
{
    struct write_qdesc_desc *desc = (struct write_qdesc_desc *) arg;
    EP4_RAIL                *rail = desc->rail;
    unsigned long            flags;

    elan4_sdram_copyq_to_sdram (dev, desc->qdesc, desc->qaddr, sizeof (E4_InputQueue));

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    desc->done = 1;
    kcondvar_wakeupall (&rail->r_haltop_sleep, &rail->r_haltop_lock);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
}

void
ep4_write_qdesc (EP4_RAIL *rail, sdramaddr_t qaddr, E4_InputQueue *qdesc)
{
    struct write_qdesc_desc desc;
    unsigned long flags;

    /* initialise descriptor */
    desc.rail  = rail;
    desc.qaddr = qaddr;
    desc.qdesc = qdesc;
    desc.done  = 0;

    kmutex_lock (&rail->r_haltop_mutex);

    rail->r_haltop.op_mask     = INT_DiscardingHighPri;
    rail->r_haltop.op_function = ep4_write_qdesc_haltop;
    rail->r_haltop.op_arg      = &desc;
    
    elan4_queue_haltop (rail->r_ctxt.ctxt_dev, &rail->r_haltop);

    spin_lock_irqsave (&rail->r_haltop_lock, flags);
    while (! desc.done)
	kcondvar_wait (&rail->r_haltop_sleep, &rail->r_haltop_lock, &flags);
    spin_unlock_irqrestore (&rail->r_haltop_lock, flags);
    
    kmutex_unlock (&rail->r_haltop_mutex);
}
#define CQ_SIZE_NWORDS ((CQ_Size (ecq->ecq_cq->cq_size) >> 3) - 8)  	/* available number of dwords (less enough to flush) */
EP4_ECQ *
ep4_alloc_ecq (EP4_RAIL *rail, unsigned cqsize)
{
    ELAN4_DEV *dev = rail->r_ctxt.ctxt_dev;
    EP4_ECQ *ecq;
    unsigned long pgoff, cqaddr;

    /* no space available, so allocate a new entry */
    KMEM_ZALLOC (ecq, EP4_ECQ *, sizeof (EP4_ECQ), 1);

    if (ecq == NULL)
	return 0;

    if ((ecq->ecq_cq = elan4_alloccq (&rail->r_ctxt, cqsize, CQ_EnableAllBits, CQ_Priority)) == NULL)
    {
	KMEM_FREE (ecq, sizeof (EP4_ECQ));
	return 0;
    }

    pgoff  = (ecq->ecq_cq->cq_mapping & (PAGE_SIZE-1));
    cqaddr = (ecq->ecq_cq->cq_cqa->cqa_cqnum + ecq->ecq_cq->cq_idx + dev->dev_cqoffset) * CQ_CommandMappingSize;

    ecq->ecq_addr  = ep_rmalloc (rail->r_ecq_rmap, PAGESIZE, 0) + pgoff;
    ecq->ecq_avail = CQ_SIZE_NWORDS;			/* available number of dwords (less enough to flush) */

    ecq->ecq_intop.op_function = (ELAN4_HALTFN *) elan4_restartcq;
    ecq->ecq_intop.op_arg      = ecq->ecq_cq;

    ep4_cport_map (&rail->r_generic, ecq->ecq_addr - pgoff, cqaddr - pgoff, PAGESIZE, EP_PERM_WRITE);

    spin_lock_init (&ecq->ecq_lock);

    return ecq;
}

void
ep4_free_ecq (EP4_RAIL *rail, EP4_ECQ *ecq)
{
    unsigned long pgoff = (ecq->ecq_cq->cq_mapping & (PAGE_SIZE-1));

    spin_lock_destroy (&ecq->ecq_lock);

    ep4_unmap (&rail->r_generic, ecq->ecq_addr - pgoff, PAGESIZE);
    ep_rmfree (rail->r_ecq_rmap, PAGESIZE, ecq->ecq_addr - pgoff);

    elan4_freecq (&rail->r_ctxt, ecq->ecq_cq);
    
    KMEM_FREE (ecq, sizeof (EP4_ECQ));
}

EP4_ECQ *
ep4_get_ecq (EP4_RAIL *rail, unsigned which, unsigned ndwords)
{
    ELAN4_DEV *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el;
    unsigned long flags;
    EP4_ECQ *ecq;
    
    spin_lock_irqsave (&rail->r_ecq_lock, flags);
    list_for_each (el, &rail->r_ecq_list[which]) {
	EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);

	if (ecq->ecq_avail >= ndwords)
	{
	    ecq->ecq_avail -= ndwords;

	    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);

	    return ecq;
	}
    }
    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);

    if ((ecq = ep4_alloc_ecq (rail, EP4_ECQ_Size (which))) == NULL)
	return NULL;

    if (which == EP4_ECQ_EVENT)
    {
	if ((ecq->ecq_event = ep_alloc_elan (&rail->r_generic, sizeof (E4_Event32), 0, &ecq->ecq_event_addr)) == 0)
	{
	    ep4_free_ecq (rail, ecq);
	    return NULL;
	}
	
	elan4_sdram_writeq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_CountAndType),
			    E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));
	elan4_sdram_writeq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_WritePtr),
			    ecq->ecq_addr);
	elan4_sdram_writeq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_WriteValue),
			    SET_EVENT_CMD | (rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_flush_event)));
	
	if ((ecq->ecq_flushcq = ep4_get_ecq (rail, EP4_ECQ_SINGLE, 1)) == NULL)
	{
	    ep_free_elan (&rail->r_generic, ecq->ecq_event_addr, sizeof (E4_Event32));
	    ep4_free_ecq (rail, ecq);
	    return NULL;
	}
    }

    spin_lock_irqsave (&rail->r_ecq_lock, flags);
    list_add (&ecq->ecq_link, &rail->r_ecq_list[which]);

    ecq->ecq_avail -= ndwords;
    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);

    return ecq;
}

void
ep4_put_ecq (EP4_RAIL *rail, EP4_ECQ *ecq, unsigned ndwords)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->r_ecq_lock, flags);

    ecq->ecq_avail += ndwords;
    
    if (ecq->ecq_avail !=  CQ_SIZE_NWORDS) 
	spin_unlock_irqrestore (&rail->r_ecq_lock, flags);
    else
    {
	list_del (&ecq->ecq_link);
	spin_unlock_irqrestore (&rail->r_ecq_lock, flags);
	
	if (ecq->ecq_flushcq)
	    ep4_put_ecq (rail, ecq->ecq_flushcq, 1);
	if (ecq->ecq_event_addr)
	    ep_free_elan (&rail->r_generic, ecq->ecq_event_addr, sizeof (E4_Event32));

	ep4_free_ecq (rail, ecq);
    }
}

void
ep4_nop_cmd (EP4_ECQ *ecq, E4_uint64 tag)
{
    unsigned long flags;

    spin_lock_irqsave (&ecq->ecq_lock, flags);
    elan4_nop_cmd (ecq->ecq_cq, tag);
    spin_unlock_irqrestore (&ecq->ecq_lock, flags);
    
}

void
ep4_set_event_cmd (EP4_ECQ *ecq, E4_Addr event)
{
    unsigned long flags;

    spin_lock_irqsave (&ecq->ecq_lock, flags);
    elan4_set_event_cmd (ecq->ecq_cq, event);
    spin_unlock_irqrestore (&ecq->ecq_lock, flags);
}

void
ep4_wait_event_cmd (EP4_ECQ *ecq, E4_Addr event, E4_uint64 candt, E4_uint64 param0, E4_uint64 param1)
{
    unsigned long flags;

    spin_lock_irqsave (&ecq->ecq_lock, flags);
    elan4_wait_event_cmd (ecq->ecq_cq, event, candt, param0, param1);
    spin_unlock_irqrestore (&ecq->ecq_lock, flags);
}

void
ep4_flush_interrupt (EP4_RAIL *rail, void *arg)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->r_ecq_lock, flags);
    rail->r_flush_count = 0;
    kcondvar_wakeupone (&rail->r_flush_sleep, &rail->r_ecq_lock);
    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);
}

void
ep4_flush_ecqs (EP4_RAIL *rail)
{
    ELAN4_DEV *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el;
    unsigned long flags;
    int i;

    kmutex_lock (&rail->r_flush_mutex);

    EP4_SDRAM_ASSERT (rail, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event), E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG,0));

    spin_lock_irqsave (&rail->r_ecq_lock, flags);
    /* first flush all the "event" queues */
    list_for_each (el, &rail->r_ecq_list[EP4_ECQ_EVENT]) {
	EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);

	elan4_sdram_writeq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_CountAndType),
			    E4_EVENT_INIT_VALUE (-32, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0));

	ep4_set_event_cmd (ecq->ecq_flushcq, ecq->ecq_event_addr);

	rail->r_flush_count++;
    }

    /* next issue the setevents to all the other queues */
    for (i = EP4_ECQ_ATOMIC; i <EP4_NUM_ECQ; i++)
    {
	list_for_each (el,&rail->r_ecq_list[i]) {
	    EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);

	    ep4_set_event_cmd (ecq, rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_flush_event));

	    rail->r_flush_count++;
	}
    }

    /* issue the waitevent command */
    ep4_wait_event_cmd (rail->r_flush_mcq,  rail->r_elan_addr + offsetof (EP4_RAIL_ELAN, r_flush_event),
			E4_EVENT_INIT_VALUE (-32 * rail->r_flush_count, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG,0),
			rail->r_flush_ecq->ecq_addr,
			INTERRUPT_CMD | (rail->r_flush_intcookie.int_val << E4_MAIN_INT_SHIFT));
    
    while (rail->r_flush_count)
	if (kcondvar_timedwait (&rail->r_flush_sleep, &rail->r_ecq_lock, &flags, (lbolt + (HZ*10))) == -1)
	     elan4_hardware_lock_check(dev, "flush_ecqs");
    
    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);

    EP4_SDRAM_ASSERT (rail, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event), E4_EVENT_INIT_VALUE (0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG,0));

    kmutex_unlock (&rail->r_flush_mutex);
}

void
ep4_init_thread (EP4_RAIL *rail, E4_ThreadRegs *regs, sdramaddr_t stackTop, 
		 EP_ADDR stackAddr, E4_Addr startpc, int nargs,...)
{
    sdramaddr_t   sp = stackTop - roundup (nargs * sizeof (E4_uint64), E4_STACK_ALIGN);
    int           i;
    va_list       ap;
    
    /*
     * the thread start code expects the following :
     *   %r1 = stack pointer
     *   %r6 = frame pointer
     *   %r2 = function to call
     *
     *   function args are store on stack above %sp
     */

    va_start(ap, nargs);
    for (i = 0; i < nargs; i++)
	elan4_sdram_writeq (rail->r_ctxt.ctxt_dev, sp + (i * sizeof (E4_uint64)), va_arg (ap, E4_uint64));
    va_end (ap);
    
    regs->Registers[0] = ep_symbol (&rail->r_threadcode, ".thread_start");		/* %r0 - PC */
    regs->Registers[1] = stackAddr - (stackTop - sp);					/* %r1 - stack pointer */
    regs->Registers[2] = startpc;							/* %r2 - start pc */
    regs->Registers[3] = 0;
    regs->Registers[4] = 0;
    regs->Registers[5] = 0;
    regs->Registers[6] = stackTop;							/* %r6 - frame pointer */ 
}

/* retransmission thread */

void
ep4_add_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops)
{
    ep_kthread_stall (&rail->r_retry_thread);
    list_add_tail (&ops->op_link, &rail->r_retry_ops);
    ep_kthread_resume (&rail->r_retry_thread);
}

void
ep4_remove_retry_ops (EP4_RAIL *rail, EP4_RETRY_OPS *ops)
{
    ep_kthread_stall (&rail->r_retry_thread);
    list_del (&ops->op_link);
    ep_kthread_resume (&rail->r_retry_thread);
}

void
ep4_retry_thread (EP4_RAIL *rail)
{
    struct list_head *el;

    kernel_thread_init ("ep4_retry");
    
    for (;;)
    {
	long nextRunTime = 0;

	list_for_each (el, &rail->r_retry_ops) {
	    EP4_RETRY_OPS *ops = list_entry (el, EP4_RETRY_OPS, op_link);

	    nextRunTime = ops->op_func (rail, ops->op_arg, nextRunTime);
	}

	if (ep_kthread_sleep (&rail->r_retry_thread, nextRunTime) < 0)
	    break;
    }

    ep_kthread_stopped (&rail->r_retry_thread);

    kernel_thread_exit();
}

/* DMA retransmission */
static unsigned ep4_dma_retry_times[EP_NUM_RETRIES];

static unsigned long
ep4_retry_dmas (EP4_RAIL *rail, void *arg, unsigned long nextRunTime)
{
    unsigned long yieldAt = lbolt + (hz/10);
    unsigned long flags;
    int           i;

    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	while (! list_empty (&rail->r_dma_retrylist[i]))
	{
	    EP4_DMA_RETRY *retry = list_entry (rail->r_dma_retrylist[i].next, EP4_DMA_RETRY, retry_link);
	    
	    if (! AFTER(lbolt, retry->retry_time))
		break;

	    if (ep_kthread_should_stall (&rail->r_retry_thread) || AFTER (lbolt, yieldAt))
		goto cant_do_more;
	    
	    EPRINTF3 (DBG_RETRY, "%s: ep4_retry_dmas: flowcnt %llx %llx\n", rail->r_generic.Name,  (long long)rail->r_dma_flowcnt,  (long long)rail->r_main->r_dma_flowcnt);

	    if ((rail->r_dma_flowcnt - rail->r_main->r_dma_flowcnt) > EP4_DMA_RETRY_FLOWCNT)
	    {
		printk ("ep4_retry_dmas: flowcnt %llx %llx\n",  (long long)rail->r_dma_flowcnt, (long long)rail->r_main->r_dma_flowcnt);

		goto cant_do_more;
	    }

	    EPRINTF4 (DBG_RETRY, "%s: ep4_retry_dmas: %016llx %016llx %016llx\n", rail->r_generic.Name,
		      (long long)retry->retry_dma.dma_typeSize,  (long long)retry->retry_dma.dma_cookie,  (long long)retry->retry_dma.dma_vproc);
	    EPRINTF5 (DBG_RETRY, "%s:                  %016llx %016llx %016llx %016llx\n", rail->r_generic.Name,
		      (long long)retry->retry_dma.dma_srcAddr,  (long long)retry->retry_dma.dma_dstAddr,  (long long)retry->retry_dma.dma_srcEvent, 
		      (long long)retry->retry_dma.dma_dstEvent);

	    elan4_run_dma_cmd (rail->r_dma_ecq->ecq_cq, &retry->retry_dma);
	    elan4_write_dword_cmd (rail->r_dma_ecq->ecq_cq, rail->r_main_addr + offsetof (EP4_RAIL_MAIN, r_dma_flowcnt), ++rail->r_dma_flowcnt);

	    spin_lock_irqsave (&rail->r_dma_lock, flags);
	    list_del (&retry->retry_link);
	    list_add (&retry->retry_link, &rail->r_dma_freelist);
	    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
	}
    }
 cant_do_more:

    /* re-compute the next retry time */
    for (i = EP_RETRY_BASE; i < EP_NUM_RETRIES; i++)
    {
	if (! list_empty (&rail->r_dma_retrylist[i]))
	{
	    EP4_DMA_RETRY *retry = list_entry (rail->r_dma_retrylist[i].next, EP4_DMA_RETRY, retry_link);

	    SET_NEXT_RUN_TIME (nextRunTime, retry->retry_time);
	}
    }

    return nextRunTime;
}

void
ep4_initialise_dma_retries (EP4_RAIL *rail)
{
    int i;

    spin_lock_init (&rail->r_dma_lock);
    
    for (i = 0; i < EP_NUM_RETRIES; i++)
	INIT_LIST_HEAD (&rail->r_dma_retrylist[i]);
    
    INIT_LIST_HEAD (&rail->r_dma_freelist);
    
    rail->r_dma_ecq = ep4_alloc_ecq (rail, EP4_DMA_RETRY_CQSIZE);
    
    rail->r_dma_allocated = 0;
    rail->r_dma_reserved  = 0;

    ep4_dma_retry_times[EP_RETRY_HIGH_PRI] = EP_RETRY_HIGH_PRI_TIME;

    for (i =0 ; i < EP_NUM_BACKOFF; i++)
	ep4_dma_retry_times[EP_RETRY_HIGH_PRI_RETRY+i] = EP_RETRY_HIGH_PRI_TIME << i;
    
    ep4_dma_retry_times[EP_RETRY_LOW_PRI] = EP_RETRY_LOW_PRI_TIME;

    for (i =0 ; i < EP_NUM_BACKOFF; i++)
	ep4_dma_retry_times[EP_RETRY_LOW_PRI_RETRY+i] = EP_RETRY_LOW_PRI_TIME << i;
    
    ep4_dma_retry_times[EP_RETRY_ANONYMOUS] = EP_RETRY_ANONYMOUS_TIME;
    ep4_dma_retry_times[EP_RETRY_NETERR]    = EP_RETRY_NETERR_TIME;

    rail->r_dma_ops.op_func = ep4_retry_dmas;
    rail->r_dma_ops.op_arg  = NULL;

    ep4_add_retry_ops (rail, &rail->r_dma_ops);
}

void
ep4_finalise_dma_retries (EP4_RAIL *rail)
{
    ep4_remove_retry_ops (rail, &rail->r_dma_ops);

    /* Everyone should have given back their retry dma's by now */
    EP4_ASSERT (rail, rail->r_dma_reserved == 0);

    while (! list_empty (&rail->r_dma_freelist))
    {
	EP4_DMA_RETRY *retry = list_entry (rail->r_dma_freelist.next, EP4_DMA_RETRY, retry_link);

	list_del (&retry->retry_link);

	KMEM_FREE (retry, sizeof (EP4_DMA_RETRY));
    }

    ep4_free_ecq (rail, rail->r_dma_ecq);

    spin_lock_destroy (&rail->r_dma_lock);
}

int
ep4_reserve_dma_retries (EP4_RAIL *rail, unsigned int count, EP_ATTRIBUTE attr)
{
    EP4_DMA_RETRY *retry;
    unsigned int   remaining = count;
    unsigned long  flags;

    spin_lock_irqsave (&rail->r_dma_lock, flags);

    if (remaining <= (rail->r_dma_allocated - rail->r_dma_reserved))
    {
	rail->r_dma_reserved += remaining;

	spin_unlock_irqrestore (&rail->r_dma_lock, flags);

	return 0;
    }
    
    remaining -= (rail->r_dma_allocated - rail->r_dma_reserved);

    rail->r_dma_reserved = rail->r_dma_allocated;

    spin_unlock_irqrestore (&rail->r_dma_lock, flags);

    while (remaining > 0)
    {
	KMEM_ALLOC (retry, EP4_DMA_RETRY *, sizeof (EP4_DMA_RETRY), !(attr & EP_NO_SLEEP));

	if (retry == NULL)
	    goto failed;
	
	remaining--;

	spin_lock_irqsave (&rail->r_dma_lock, flags);
	list_add (&retry->retry_link, &rail->r_dma_freelist);

	rail->r_dma_allocated++;
	rail->r_dma_reserved++;
	spin_unlock_irqrestore (&rail->r_dma_lock, flags);
    }

    return 0;

 failed:
    spin_lock_irqsave (&rail->r_dma_lock, flags);
    rail->r_dma_reserved -= (count - remaining);
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);

    return 1;
}

void
ep4_release_dma_retries (EP4_RAIL *rail, unsigned int count)
{
    unsigned long flags;

    spin_lock_irqsave (&rail->r_dma_lock, flags);
    rail->r_dma_reserved -= count;
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
}

void
ep4_queue_dma_retry (EP4_RAIL *rail, E4_DMA *dma, int interval)
{
    EP4_DMA_RETRY *retry;
    unsigned long  flags;
    
    spin_lock_irqsave (&rail->r_dma_lock, flags);

    EP4_ASSERT (rail, !list_empty (&rail->r_dma_freelist));
    
    /* take an item of the free list */
    retry = list_entry (rail->r_dma_freelist.next, EP4_DMA_RETRY, retry_link);

    list_del (&retry->retry_link);
    
    EPRINTF5 (DBG_RETRY, "%s: ep4_queue_dma_retry: %016llx %016llx %016llx %016llx\n", rail->r_generic.Name,
	      (long long)dma->dma_typeSize,  (long long)dma->dma_cookie,  (long long)dma->dma_vproc, (long long)dma->dma_srcAddr);
    EPRINTF5 (DBG_RETRY, "%s:                      %016llx %016llx %016llx (%d)\n", rail->r_generic.Name,
	      (long long)dma->dma_dstAddr,  (long long)dma->dma_srcEvent, (long long)dma->dma_dstEvent, interval);

    retry->retry_dma.dma_typeSize = dma->dma_typeSize;
    retry->retry_dma.dma_cookie   = dma->dma_cookie;
    retry->retry_dma.dma_vproc    = dma->dma_vproc;
    retry->retry_dma.dma_srcAddr  = dma->dma_srcAddr;
    retry->retry_dma.dma_dstAddr  = dma->dma_dstAddr;
    retry->retry_dma.dma_srcEvent = dma->dma_srcEvent;
    retry->retry_dma.dma_dstEvent = dma->dma_dstEvent;

    retry->retry_time             = lbolt + ep4_dma_retry_times[interval];

    /* chain onto the end of the approriate retry list */
    list_add_tail (&retry->retry_link, &rail->r_dma_retrylist[interval]);

    ep_kthread_schedule (&rail->r_retry_thread, retry->retry_time);

    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
}

void
ep4_queue_dma_stalled (EP4_RAIL *rail, E4_DMA *dma)
{
    EP_NODE_RAIL  *nodeRail = &rail->r_generic.Nodes[EP_VP_TO_NODE(dma->dma_vproc)];
    EP4_DMA_RETRY *retry;
    unsigned long  flags;
    
    spin_lock_irqsave (&rail->r_dma_lock, flags);

    EP4_ASSERT (rail, !list_empty (&rail->r_dma_freelist));
    
    /* take an item of the free list */
    retry = list_entry (rail->r_dma_freelist.next, EP4_DMA_RETRY, retry_link);

    list_del (&retry->retry_link);
    
    EPRINTF5 (DBG_RETRY, "%s: ep4_queue_dma_stalled: %016llx %016llx %016llx %016llx\n", rail->r_generic.Name,
	      (long long)dma->dma_typeSize, (long long)dma->dma_cookie, (long long)dma->dma_vproc, (long long)dma->dma_srcAddr);
    EPRINTF4 (DBG_RETRY, "%s:                        %016llx %016llx %016llx\n", rail->r_generic.Name,
	      (long long)dma->dma_dstAddr, (long long)dma->dma_srcEvent, (long long)dma->dma_dstEvent);

    retry->retry_dma.dma_typeSize = dma->dma_typeSize;
    retry->retry_dma.dma_cookie   = dma->dma_cookie;
    retry->retry_dma.dma_vproc    = dma->dma_vproc;
    retry->retry_dma.dma_srcAddr  = dma->dma_srcAddr;
    retry->retry_dma.dma_dstAddr  = dma->dma_dstAddr;
    retry->retry_dma.dma_srcEvent = dma->dma_srcEvent;
    retry->retry_dma.dma_dstEvent = dma->dma_dstEvent;

    /* chain onto the node cancelled dma list */
    list_add_tail (&retry->retry_link, &nodeRail->StalledDmas);

    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
}

void
ep4_free_stalled_dmas (EP4_RAIL *rail, unsigned int nodeId)
{
    EP_NODE_RAIL *nodeRail = &rail->r_generic.Nodes[nodeId];
    struct list_head *el, *nel;
    unsigned long flags;

    spin_lock_irqsave (&rail->r_dma_lock, flags);
    list_for_each_safe (el, nel, &nodeRail->StalledDmas) {
	list_del (el);
	list_add (el, &rail->r_dma_freelist);
    }
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
}

void
ep4_display_rail (EP4_RAIL *rail)
{
    ELAN4_DEV        *dev = rail->r_ctxt.ctxt_dev;
    struct list_head *el;
    register int      i;
    unsigned long     flags;

    ep_debugf (DBG_DEBUG, "%s: vendorid=%x deviceid=%x\n", rail->r_generic.Name, 
	       rail->r_generic.Devinfo.dev_vendor_id, rail->r_generic.Devinfo.dev_device_id);

    spin_lock_irqsave (&rail->r_ecq_lock, flags);
    for (i = 0; i < EP4_NUM_ECQ; i++)
    {
	list_for_each (el, &rail->r_ecq_list[i]) {
	    EP4_ECQ *ecq = list_entry (el, EP4_ECQ, ecq_link);
	    
	    if (i == EP4_ECQ_EVENT)
		ep_debugf (DBG_DEBUG, "   ECQ[%d] ecq=%p cqnum=%d addr=%llx avail=%d event=%llx,%llx,%llx\n",
			   i, ecq, elan4_cq2num (ecq->ecq_cq), (long long)ecq->ecq_addr, ecq->ecq_avail,
			   elan4_sdram_readq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_CountAndType)),
			   elan4_sdram_readq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_WriteValue)),
			   elan4_sdram_readq (dev, ecq->ecq_event + offsetof (E4_Event32, ev_WritePtr)));

	    else
		ep_debugf (DBG_DEBUG, "   ECQ[%d] ecq=%p cqnum=%d addr=%llx avail=%d\n",
			   i, ecq, elan4_cq2num (ecq->ecq_cq), (long long)ecq->ecq_addr, ecq->ecq_avail);
	}
    }
    spin_unlock_irqrestore (&rail->r_ecq_lock, flags);

    ep_debugf (DBG_DEBUG, "   flush count=%ld mcq=%p ecq=%p event %llx.%llx.%llx\n", 
	       rail->r_flush_count, rail->r_flush_mcq, rail->r_flush_ecq,
	       elan4_sdram_readq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event.ev_CountAndType)),
	       elan4_sdram_readq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event.ev_WritePtr)),
	       elan4_sdram_readq (dev, rail->r_elan + offsetof (EP4_RAIL_ELAN, r_flush_event.ev_WriteValue)));
    
    spin_lock_irqsave (&rail->r_dma_lock, flags);
    for (i = 0; i < EP_NUM_RETRIES; i++)
    {
	list_for_each (el, &rail->r_dma_retrylist[i]) {
	    EP4_DMA_RETRY *retry = list_entry (el, EP4_DMA_RETRY, retry_link);
	    
	    ep_debugf (DBG_DEBUG, "    RETRY[%d] typeSize %llx cookie %llx vproc %llx events %llx %llx\n",
		       i, (long long)retry->retry_dma.dma_typeSize, (long long)retry->retry_dma.dma_cookie,
		       (long long)retry->retry_dma.dma_vproc, (long long)retry->retry_dma.dma_srcEvent, (long long)retry->retry_dma.dma_dstEvent);
	}
    }
    spin_unlock_irqrestore (&rail->r_dma_lock, flags);
}
