/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: threadcode.c,v 1.11 2003/10/07 13:22:38 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/threadcode.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

EP_ADDR
ep_symbol (EP_CODE *code, char *name)
{
    EP_SYMBOL *s = code->symbols;
    
    while (s->name && strcmp (s->name, name))
	s++;
    
    return (s->name ? s->value : (EP_ADDR) 0);
}

int
ep_loadcode (EP_RAIL *rail, EP_CODE *code)
{
    register int i;

    EP_ADDR  _stext  = ep_symbol (code, "_stext");
    EP_ADDR  _etext  = ep_symbol (code, "_etext");
    EP_ADDR  _sdata  = ep_symbol (code, "_sdata");
    EP_ADDR  _edata  = ep_symbol (code, "_edata");
    EP_ADDR  _end    = ep_symbol (code, "_end");
    EP_ADDR  _rodata = roundup (_etext, sizeof (uint64_t));

    if (_stext == (EP_ADDR) 0 || _etext == (EP_ADDR) 0 ||
	_sdata == (EP_ADDR) 0 || _edata == (EP_ADDR) 0 ||
	_end == (EP_ADDR) 0)
    {
	printk ("ep_loadcode: symbols not defined correctly for code at %p\n", code);
	return (EINVAL);
    }

    /*
     * Include the rodata in the text segment
     */
    _etext = _rodata + code->rodata_size;

    /*
     * If _etext is in the same page as _sdata,  then allocate a contiguous
     * chunk of memory and map it as read/write. otherwise allocate two chunks
     * and map the code in as read-only.
     */
    if ((_etext & PAGEMASK) == (_sdata & PAGEMASK))
    {
	code->ntext  = btopr (_end - (_stext & PAGEMASK));
	code->pptext = ep_alloc_memory_elan (rail, _stext & PAGEMASK, ptob (code->ntext), EP_PERM_EXECUTE, 0);

	if (code->pptext == (sdramaddr_t) 0)
	    return (ENOMEM);
	
	code->_stext  = code->pptext + (_stext & PAGEOFFSET);
	code->_rodata = code->_stext + (_rodata - _stext);
	code->_sdata  = code->_stext + (_sdata - _stext);
    }
    else
    {
	code->ntext  = btopr (_etext - (_stext & PAGEMASK));
	code->ndata  = btopr (_end - (_sdata & PAGEMASK));

	if (code->ntext)
	{
	    code->pptext = ep_alloc_memory_elan (rail, _stext & PAGEMASK, ptob (code->ntext), EP_PERM_EXECUTE, 0);

	    if (code->pptext == (sdramaddr_t) 0)
		return (ENOMEM);

	    code->_stext  = code->pptext + (_stext & PAGEOFFSET);
	    code->_rodata = code->_stext + (_rodata - _stext);
	}
	
	if (code->ndata)
	{
	    code->ppdata = ep_alloc_memory_elan (rail, _sdata & PAGEMASK, ptob (code->ndata), EP_PERM_WRITE, 0);

	    if (code->ppdata == (sdramaddr_t) 0)
	    {
		if (code->ntext) ep_free_memory_elan (rail, _sdata & PAGEMASK);
		code->ntext = 0;

		return (ENOMEM);
	    }
	    
	    code->_sdata = code->ppdata + (_sdata & PAGEOFFSET);
	}
    }
    
#ifdef __LITTLE_ENDIAN__
#  define Flip	3
#else
#  define Flip  0
#endif

    /*
     * Now copy the text and rodata into the SDRAM
     * this is linked into the module to be byte 
     * copied to the SDRAM, since we want to copy
     * with word accesses we have to do the byte
     * assembly correctly.
     */
    for (i = 0; i < code->text_size; i++)
	rail->Operations.SdramWriteb (rail, code->_stext + i, code->text[i^Flip]);

    for (i = 0; i < code->rodata_size; i++)
	rail->Operations.SdramWriteb (rail, code->_rodata + i, code->rodata[i^Flip]);
    
    /*
     * And the initialised data segment.
     */
    for (i = 0; i < code->data_size; i++)
	rail->Operations.SdramWriteb (rail, code->_sdata + i, code->data[i^Flip]);

    return (ESUCCESS);
}

void
ep_unloadcode (EP_RAIL *rail, EP_CODE *code)
{
    EP_ADDR  _stext = ep_symbol (code, "_stext");
    EP_ADDR  _sdata = ep_symbol (code, "_sdata");

    if (code->pptext)
	ep_free_memory_elan (rail, _stext & PAGEMASK);
    if (code->ppdata)
	ep_free_memory_elan (rail, _sdata & PAGEMASK);
    code->pptext = code->ppdata = 0;
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
