/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: threadcode_elan3.c,v 1.11 2003/10/07 13:22:38 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/threadcode_elan3.c,v $ */

#include <qsnet/kernel.h>

#include <elan/kcomm.h>

#include "kcomm_elan3.h"
#include "debug.h"

#include <elan3/thread.h>

E3_Addr
ep3_init_thread (ELAN3_DEV  *dev,
		 E3_Addr     fn,				/* Elan address of function */
		 E3_Addr     addr,				/* Elan address of stack */
		 sdramaddr_t stack,				/* sdram address of stack */
		 int	       stackSize,			/* stack size (in bytes) */
		 int	       nargs,
		 ...)
{
    sdramaddr_t  frame;
    sdramaddr_t  regs;
    sdramaddr_t  argsp;
    int		 i;
    va_list      ap;

    /*
     * Align the stack pointer at the top of the stack and leave space for a stack frame
     */
    stack = ((stack + stackSize) & ~(E3_STACK_ALIGN-1)) - sizeof (E3_Frame);
    addr  = ((addr  + stackSize) & ~(E3_STACK_ALIGN-1)) - sizeof (E3_Frame);

    va_start (ap, nargs);

    if (nargs > 6)
    {
	stack -= (((nargs*sizeof (E3_uint32))+E3_STACK_ALIGN-1) & ~(E3_STACK_ALIGN-1));
	addr  -= (((nargs*sizeof (E3_uint32))+E3_STACK_ALIGN-1) & ~(E3_STACK_ALIGN-1));
    }
    
    frame  = stack;
    regs   = stack - sizeof (E3_OutsRegs);

    /*
     * Initialise the registers, and stack frame.
     */
    elan3_sdram_writel (dev, regs + offsetof (E3_OutsRegs, o[6]), fn);
    elan3_sdram_writel (dev, regs + offsetof (E3_OutsRegs, o[7]), 0);
    
    if (nargs <= 6)
    {
	for (i = 0; i < nargs; i++)
	    elan3_sdram_writel (dev, regs + offsetof (E3_OutsRegs, o[i]), va_arg (ap, E3_uint32));
    }
    else
    {
	for (i = 0; i < 6; i++)
	    elan3_sdram_writel (dev, regs + offsetof (E3_OutsRegs, o[i]), va_arg (ap, E3_uint32));
	
	for (argsp = frame + offsetof (E3_Frame, fr_argx[0]); i < nargs; i++, argsp += sizeof (E3_uint32))
	    elan3_sdram_writel (dev, argsp, va_arg (ap, int));
    }

    elan3_sdram_writel (dev, frame + offsetof (E3_Frame, fr_savefp), 0);
    elan3_sdram_writel (dev, frame + offsetof (E3_Frame, fr_savepc), 0);

    va_end (ap);

    return (addr);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
