/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: jtagdrv.c,v 1.12 2003/06/07 16:02:35 david Exp $"
/*      $Source: /cvs/master/quadrics/jtagmod/jtagdrv.c,v $*/

#include <qsnet/types.h>

#include "jtagdrv.h"
#include <jtag/jtagio.h>

int
jtagdrv_strobe_data (JTAG_DEV *dev, u_char data)
{
    u_char dsr;

    PRINTF (DBG_ECPP, ("jtagdrv_strobe_data: %s %s %s -> ", (data & LPT_DATA_TRST) ? "TRST" : "trst", 
		       (data & LPT_DATA_TDI) ? "TDI" : "tdi", (data & LPT_DATA_TMS) ? "TMS" : "tms"));


    LPT_WRITE_DATA (dev, data); DELAY(5);			/* Drive NEW values on data wires */
    LPT_WRITE_CTRL (dev, LPT_CTRL_TCLK); DELAY(5);		/* Drive strobe low */
    LPT_READ_STAT  (dev, dsr); DELAY(5);			/* Sample TDI from ring */
    LPT_WRITE_CTRL (dev, 0); DELAY(5);				/* Drive strobe high */

    PRINTF (DBG_ECPP, ("%s\n", (dsr & LPT_STAT_PE) ? "TDO" : "tdo"));

    return ((dsr & LPT_STAT_PE) ? 1 : 0);
}

void
jtagdrv_select_ring (JTAG_DEV *dev, u_int ring)
{
    PRINTF (DBG_ECPP, ("jtagdrv_select_ring: ring=0x%x\n", ring));

    LPT_WRITE_CTRL (dev, 0); DELAY(5);				/* Drive strobe and TCLK high */
    LPT_WRITE_DATA (dev, ring);	DELAY(5);			/* Drive ring address */
    LPT_WRITE_CTRL (dev, LPT_CTRL_RCLK); DELAY(5);		/* Drive strobe low */
    LPT_WRITE_CTRL (dev, 0); DELAY(5);				/* Drive strobe high */
}

void
jtagdrv_reset (JTAG_DEV *dev)
{
    register int i;

    for (i = 0; i < 5; i++)
	jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);		/* 5 clocks to Reset from any state */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Run-Test/Idle */
}

void
jtagdrv_shift_ir (JTAG_DEV *dev, u_char *value, int nbits)
{
    register int i;
    register int bit;

    jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);			/* to Select DR-Scan */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);			/* to Select IR-Scan */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Capture-IR */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Shift-IR */
    
    for (i = 0; i < nbits; i++)
    {
	/* strobe through the instruction bits,  asserting TMS on the last bit */

	if (i == (nbits-1))
	    bit = jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS | (JTAG_BIT(value, i) ? LPT_DATA_TDI : 0));
	else
	    bit = jtagdrv_strobe_data (dev, LPT_DATA_TRST | (JTAG_BIT(value, i) ? LPT_DATA_TDI : 0));
	
	if (bit)
	    JTAG_SET_BIT(value, i);
	else
	    JTAG_CLR_BIT(value, i);
    }
    
    jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);			/* to Update-IR */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Run-Test/Idle */
}


void
jtagdrv_shift_dr (JTAG_DEV *dev, u_char *value, int nbits)
{
    register int i;
    register int bit;

    jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);			/* to Select DR-Scan */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Capture-DR */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Shift-DR */
    
    for (i = 0; i < nbits; i++)
    {
	/* strobe through the data bits,  asserting TMS on the last bit */

	if (i == (nbits-1))
	    bit = jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS | (JTAG_BIT(value, i) ? LPT_DATA_TDI : 0));
	else
	    bit = jtagdrv_strobe_data (dev, LPT_DATA_TRST | (JTAG_BIT(value, i) ? LPT_DATA_TDI : 0));
	
	if (bit)
	    JTAG_SET_BIT(value, i);
	else
	    JTAG_CLR_BIT(value, i);
    }
    
    jtagdrv_strobe_data (dev, LPT_DATA_TRST | LPT_DATA_TMS);			/* to Update-DR */
    jtagdrv_strobe_data (dev, LPT_DATA_TRST);					/* to Run-Test/Idle */
}

static int
jtagdrv_i2c_start (JTAG_DEV *dev)
{
    u_char dsr;
    int i;

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_start\n"));
    
    /* Issue a stop sequence */
    LPT_WRITE_CTRL (dev,  LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
    LPT_WRITE_DATA (dev, 0); DELAY(5);				/* SDA low */
    LPT_WRITE_CTRL (dev, 0); DELAY(5);				/* SCLK high */
    LPT_WRITE_DATA (dev, LPT_DATA_SDA); DELAY(5);		/* SDA high */
    
    /* sample the line to see if we're idle */
    LPT_READ_STAT (dev, dsr);					/* sample SDA */
    if ((dsr & LPT_STAT_SDA) == 0)				/* Cannot start if SDA already driven */
    {
	PRINTF (DBG_ECPP, ("jtagdrv_i2c_start: cannot start - sda driven low\n"));

	for (i = 0; i < 16 ; i++)
	{
	    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(5);	/* SCLK low */
	    LPT_WRITE_CTRL (dev, 0); DELAY(5);			/* SCLK high */
	    LPT_READ_STAT  (dev, dsr);
	    
	    if (dsr & LPT_STAT_SDA)
	    {
		PRINTF (DBG_ECPP, ("jtagdrv_i2c_start - stopped after %d clocks\n", i));
		break;
	    }
	}

	if ((dsr & LPT_STAT_SDA) == 0)
	{
	    PRINTF (DBG_ECPP, ("jtagdrv_i2c_start - cannot start - not idle\n"));
	    return (0);
	}

	/* seen SDA float high, so issue a stop sequence */
	LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
	LPT_WRITE_DATA (dev, 0); DELAY(5);			/* SDA low */
	LPT_WRITE_CTRL (dev, 0); DELAY(5);			/* SCLK high */
	LPT_WRITE_DATA (dev, LPT_DATA_SDA); DELAY(5);		/* SDA high */
    }

    LPT_WRITE_DATA (dev, 0); DELAY(4);				/* drive SDA low */
    return (1);
}

static void
jtagdrv_i2c_stop (JTAG_DEV *dev)
{
    u_char dsr;
    int    i;

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_stop\n"));

    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
    LPT_WRITE_DATA (dev, 0); DELAY(5);				/* SDA low */
    LPT_WRITE_CTRL (dev, 0); DELAY(5);				/* SCLK high */
    LPT_WRITE_DATA (dev, LPT_DATA_SDA); DELAY(5);		/* SDA high */

    /* 
     * bug fix for temperature sensor chip
     * if it's still driving SDA, then clock
     * it until it stops driving it 
     */
    LPT_READ_STAT (dev, dsr);
    if ((dsr & LPT_STAT_SDA) == 0)
    {
	PRINTF (DBG_ECPP, ("jtagdrv_i2c_stop - slave not stodeved\n"));
	for (i = 0; i < 16 ; i++)
	{
	    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(5);	/* SCLK low */
	    LPT_WRITE_CTRL (dev, 0); DELAY(5);			/* SCLK high */
	    LPT_READ_STAT  (dev, dsr);
	    
	    if (dsr & LPT_STAT_SDA)
		break;
	}
	PRINTF (DBG_ECPP, ("jtagdrv_i2c_stop - stodeved after %d clocks\n", i));
    }
}

static int
jtagdrv_i2c_strobe (JTAG_DEV *dev, u_char data)
{
    u_char dsr;
    
    PRINTF (DBG_ECPP, ("jtagdrv_i2c_strobe : %s", (data & LPT_DATA_SDA) ? "SDA" : "sda"));

    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
    LPT_WRITE_DATA (dev, data);	DELAY(5);			/* write data */
    LPT_WRITE_CTRL (dev, 0);					/* SCLK high */
    LPT_READ_STAT  (dev, dsr); DELAY(4);			/* Sample SDA */

    PRINTF (DBG_ECPP, (" -> %s\n", (dsr & LPT_STAT_SDA) ? "SDA" : "sda"));

    return ((dsr & LPT_STAT_SDA) ? 1 : 0);
}

static int
jtagdrv_i2c_get_ack (JTAG_DEV *dev)
{
    u_char dsr;

    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
    LPT_WRITE_DATA (dev, LPT_DATA_SDA);	DELAY(5);		/* SDA high */
    LPT_WRITE_CTRL (dev, 0);					/* SCLK high */
    LPT_READ_STAT  (dev, dsr); DELAY(4);			/* Sample SDA */

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_get_ack -> %s\n", (dsr & LPT_STAT_SDA) ? "no ack" : "ack"));
    
    return ((dsr & LPT_STAT_SDA) ? 0 : 1);
}

static int
jtagdrv_i2c_drive_ack (JTAG_DEV *dev, int nack)
{
    u_char dsr;

    LPT_WRITE_CTRL (dev, LPT_CTRL_SCLK); DELAY(1);		/* SCLK low */
    LPT_WRITE_DATA (dev, nack ? LPT_DATA_SDA : 0); DELAY(5);	/* SDA low for ack, high for nack */
    LPT_WRITE_CTRL (dev, 0);					/* SCLK high */
    LPT_READ_STAT  (dev, dsr); DELAY(4);			/* Sample SDA for ack */

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_drive_ack %d -> %s\n", nack, (dsr & LPT_STAT_SDA) ? "done" : "more"));
    
    return ((dsr & LPT_STAT_SDA) ? 1 : 0);
}

static void
jtagdrv_i2c_shift_addr (JTAG_DEV *dev, u_int address, int readNotWrite)
{
    register int i;

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_shift_addr: %x\n", address));

    for (i = I2C_ADDR_LEN-1; i >= 0; i--)
	jtagdrv_i2c_strobe (dev, (address & (1 << i)) ? LPT_DATA_SDA : 0);
    
    jtagdrv_i2c_strobe (dev, readNotWrite ? LPT_DATA_SDA : 0);
}

static u_char
jtagdrv_i2c_shift_data (JTAG_DEV *dev, u_char data)
{
    register int i;
    u_char val = 0;
    
    PRINTF (DBG_ECPP, ("jtagdrv_i2c_shift_data : %02x\n", data));

    for (i = I2C_DATA_LEN-1; i >= 0; i--)
	if (jtagdrv_i2c_strobe (dev, data & (1 << i) ? LPT_DATA_SDA : 0))
	    val |= (1 << i);

    PRINTF (DBG_ECPP, ("jtagdrv_i2c_shift_data : -> %02x\n", val));

    return (val);
}

int
jtagdrv_i2c_write (JTAG_DEV *dev, u_int address, u_int count, u_char *data)
{
    register int i;

    PRINTF (DBG_FN, ("jtagdrv_i2c_write: address=%x count=%d data=%02x\n", address, count, data[0]));

    if (! jtagdrv_i2c_start (dev))
	return (I2C_OP_NOT_IDLE);

    jtagdrv_i2c_shift_addr (dev, address, 0);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_write: no ack on address phase\n"));

	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }
    
    for (i = 0; i < count; i++)
    {
	jtagdrv_i2c_shift_data (dev, data[i]);
	
	if (! jtagdrv_i2c_get_ack (dev))
	{
	    PRINTF (DBG_FN, ("jtagdrv_i2c_write: no ack on data phase %d\n", i));

	    jtagdrv_i2c_stop (dev);
	    return (I2C_OP_WRITE_TO_BIG);
	}
    }

    jtagdrv_i2c_stop (dev);
    return (I2C_OP_SUCCESS);
}

int
jtagdrv_i2c_read (JTAG_DEV *dev, u_int address, u_int count, u_char *data)
{
    register int i;

    PRINTF (DBG_FN, ("jtagdrv_i2c_read: address=%x count=%d\n", address, count));

    if (! jtagdrv_i2c_start (dev))
	return (I2C_OP_NOT_IDLE);

    jtagdrv_i2c_shift_addr (dev, address, 1);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_read: no ack on address phase\n"));

	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }
    
    for (i = 0; i < count; i++)
    {
	data[i] = jtagdrv_i2c_shift_data (dev, 0xff);

	jtagdrv_i2c_drive_ack (dev, (i == (count-1) ? 1 : 0));
    }

    jtagdrv_i2c_stop (dev);
    
    return (I2C_OP_SUCCESS);
}

int
jtagdrv_i2c_writereg (JTAG_DEV *dev, u_int address, u_int intaddress, u_int count, u_char *data)
{
    register int i;

    PRINTF (DBG_FN, ("jtagdrv_i2c_writereg: address=%x count=%d\n", address, count));

    if (! jtagdrv_i2c_start (dev))
	return (I2C_OP_NOT_IDLE);

    jtagdrv_i2c_shift_addr (dev, address, 0);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_writereg: no ack on address phase\n"));

	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }
    
    jtagdrv_i2c_shift_data (dev, intaddress);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_writereg: no ack on intaddress phase\n"));
	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }
    
    for (i = 0; i < count; i++)
    {
	jtagdrv_i2c_shift_data (dev, data[i]);
	if (! jtagdrv_i2c_get_ack (dev))
	{
	    PRINTF (DBG_FN, ("jtagdrv_i2c_writedate: no ack on byte %d\n", i));
	    jtagdrv_i2c_stop (dev);
	    return (I2C_OP_WRITE_TO_BIG);
	}
    }
    
    jtagdrv_i2c_stop (dev);
    return (I2C_OP_SUCCESS);
}

int
jtagdrv_i2c_readreg (JTAG_DEV *dev, u_int address, u_int intaddress, u_int count, u_char *data)
{
    PRINTF (DBG_FN, ("jtagdrv_i2c_readreg: address=%x count=%d\n", address, count));

    if (! jtagdrv_i2c_start (dev))
	return (I2C_OP_NOT_IDLE);

    jtagdrv_i2c_shift_addr (dev, address, 0);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_readreg: no ack on address phase\n"));

	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }
    
    jtagdrv_i2c_shift_data (dev, intaddress);
    
    if (! jtagdrv_i2c_get_ack (dev))
    {
	PRINTF (DBG_FN, ("jtagdrv_i2c_readreg: no ack on intaddress phase\n"));
	jtagdrv_i2c_stop (dev);
	return (I2C_OP_NO_DEVICE);
    }

    jtagdrv_i2c_stop (dev);

    return (jtagdrv_i2c_read (dev, address, count, data));
}

void
jtagdrv_i2c_clock_shift (JTAG_DEV *dev, u_int t, u_int n, u_int m)
{
    int i;

    for (i = 2; i >= 0; i--)
    {
	LPT_WRITE_DATA (dev, ((t & (1 << i)) ? LPT_DATA_TDI : 0)); DELAY(1);			/* clock low  | data */
	LPT_WRITE_DATA (dev, ((t & (1 << i)) ? LPT_DATA_TDI : 0) | LPT_DATA_TMS); DELAY(1);	/* clock high | data */
    }

    for (i = 1; i >= 0; i--)
    {
	LPT_WRITE_DATA (dev, ((n & (1 << i)) ? LPT_DATA_TDI : 0)); DELAY(1);			/* clock low  | data */
	LPT_WRITE_DATA (dev, ((n & (1 << i)) ? LPT_DATA_TDI : 0)| LPT_DATA_TMS); DELAY(1);	/* clock high | data */
    }    

    for (i = 6; i >= 0; i--)
    {
	LPT_WRITE_DATA (dev, ((m & (1 << i)) ? LPT_DATA_TDI : 0)); DELAY(1);			/* clock low  | data */
	LPT_WRITE_DATA (dev, ((m & (1 << i)) ? LPT_DATA_TDI : 0) | LPT_DATA_TMS); DELAY(1);	/* clock high | data */
    }    

    LPT_WRITE_DATA (dev, 0); DELAY(1);								/* clock low  | 0 */

    LPT_WRITE_CTRL (dev, LPT_CTRL_TCLK); DELAY(1);						/* strobe low */
    LPT_WRITE_CTRL (dev, 0); DELAY(1);								/* strobe low */
}

