/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

/*
 * $Id: jtagdrv_Linux.c,v 1.19.2.3 2005/09/07 14:35:03 mike Exp $
 * $Source: /cvs/master/quadrics/jtagmod/jtagdrv_Linux.c,v $
 */

#include "jtagdrv.h"
#include <linux/slab.h>
#include <jtag/jtagio.h>

#include <qsnet/module.h>
#include <linux/ioport.h>

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("JTAG Parallel port QsNet switch interface");

MODULE_LICENSE("GPL");

#define MAJOR_INSTANCE	0	/* 0 is dynamic assign of device major 	*/ 
#define MAX_JTAG_DEV 	4

int jtag_major = MAJOR_INSTANCE;
int jtagdebug  = 0;
module_param(jtag_major, uint, 0);
module_param(jtagdebug, uint, 0);

JTAG_DEV	jtag_devs[MAX_JTAG_DEV];

int io[MAX_JTAG_DEV]= { 0, };
MODULE_PARM(io, "1-4i");


/* The fops functions */
int jtag_open(struct inode *, struct file *);
int jtag_close(struct inode *, struct file *);
int jtag_ioctl(struct inode *, struct file *, unsigned int, unsigned long );

struct file_operations jtag_fops = {
    ioctl:   jtag_ioctl,
    open:    jtag_open,
    release: jtag_close,
};

int
jtag_probe(void)
{
	int i=0;	
	int default_io = 1;
	JTAG_DEV *dev;
	unsigned char value=0xff;


	/* see if there are any user supplied io addr */
	for ( i = 0; i < MAX_JTAG_DEV; i++) {
		if ( io[i] != 0x00)
			default_io = 0;
		jtag_devs[i].base = io[i];
	}
	
	if ( default_io ) {
		jtag_devs[0].base = 0x3bc;
		jtag_devs[1].base = 0x378;
		jtag_devs[2].base = 0x278;
		jtag_devs[3].base = 0x268;
	}

	for ( i = 0 ; i < MAX_JTAG_DEV; i++) {
		if ( jtag_devs[i].base == 0x3bc ) 
			jtag_devs[i].region = 3;
		else
			jtag_devs[i].region = 8;
		jtag_devs[i].present = 0;
	}	


	if( default_io )
	{
		for( i = 0 ; i < MAX_JTAG_DEV; i++) {
			dev=&(jtag_devs[i]);
			if(dev->base && request_region(dev->base, dev->region, "jtag")) {
				LPT_WRITE(dev, 0,0);
				LPT_READ(dev, 0,value);
				if ( value != 0xff) {
					PRINTMSG("(%d , %d) present, io=0x%04lx\n",jtag_major,i,dev->base);
			
					dev->present=1; 
				}
				else
				    release_region(dev->base, dev->region);
			}
			else
			{
			    PRINTMSG("failed to request_region (%d , %d), io=0x%04lx\n",jtag_major,i,dev->base);
			    return -1;
			}
		}
		return 0;
       }     
       else /* Force the region to be present, this makes the PCI parallel cards work */
       {
		for( i = 0 ; i < MAX_JTAG_DEV; i++) 
		{
                        dev=&(jtag_devs[i]);
                        if(dev->base && request_region(dev->base, dev->region, "jtag") && (dev->base != 0)) 
			{
                                PRINTMSG("(%d , %d) forced by user, io=0x%04lx\n",jtag_major,i,dev->base);
                                        dev->present=1;
			}	
                        else	
			{
                                if( dev->base != 0)
					release_region(dev->base, dev->region);
                	}
		}
                return 0;
       }
}

int init_module(void)
{
	int result,i;
	result = register_chrdev(jtag_major, DRIVER_NAME, &jtag_fops);
	if (result < 0) {
		PRINTMSG("Couldn't register char device err == %d\n",jtag_major);
		return -1;
	}

	if ( jtag_major == 0 ) 
		jtag_major = result;

	for ( i = 0; i < MAX_JTAG_DEV; i++) {
		jtag_devs[i].base=io[i];	
	}

	jtag_probe();

	PRINTMSG("Registered character device, major == %d\n",jtag_major);	
	return 0;
}	

void cleanup_module(void)
{
	int i=0;

	for( i = 0; i < MAX_JTAG_DEV; i++) {
		if( jtag_devs[i].present)	
			release_region(jtag_devs[i].base, jtag_devs[i].region);
	}
			
	unregister_chrdev(jtag_major, DRIVER_NAME);
	PRINTMSG("Unloaded char device\n");
}


int
jtag_open (struct inode *inode, struct file *filp)
{
    int unit = MINOR(inode->i_rdev);
    JTAG_DEV *dev = &jtag_devs[unit];
    
    if (unit < 0 || unit > MAX_JTAG_DEV || !dev->present)
	return (-ENXIO);

    /*
     * Only allow a single open at a time 
     */
    if (dev->open)
	return (-EBUSY);
    dev->open = 1;
    
    /*
     * Initialise the hardware registers
     */
   
    LPT_WRITE (dev, LPT_CTRL, 0);
    DELAY(50);
    LPT_WRITE (dev, LPT_CTRL, LPT_CTRL_INIT);

    MOD_INC_USE_COUNT;

    return (0);
}

int
jtag_close(struct inode *inode, struct file *filp)
{
  
    int unit = MINOR(inode->i_rdev);
    JTAG_DEV *dev = &jtag_devs[unit];
    
    if (unit < 0 || unit > MAX_JTAG_DEV || !dev->present)
	return (-ENXIO);
    
    dev->open = 0;

    MOD_DEC_USE_COUNT;

    return (0);
}

int
jtag_ioctl (struct inode *inode, struct file *filp, unsigned int io_cmd, unsigned long io_data)
{
    int                  unit = MINOR(inode->i_rdev);
    JTAG_DEV             *dev = &jtag_devs[unit];
    JTAG_RESET_ARGS      *resetargs;
    JTAG_SHIFT_ARGS      *shiftargs;
    I2C_ARGS	         *i2cargs;
    I2C_CLOCK_SHIFT_ARGS *clockargs;
    u_char		 *buf;
    int			  freq;

    if (unit < 0 || unit > MAX_JTAG_DEV || !dev->present)
	return (-ENXIO);
    
    PRINTF (DBG_IOCTL, ("jtag_ioctl: device %d cmd=%x\n", unit, io_cmd));

    switch (io_cmd)
    {
    case JTAG_RESET:
	resetargs = (JTAG_RESET_ARGS *) io_data;

	if (! VALID_JTAG_RING (resetargs->ring))
	    return (-EINVAL);
	
	jtagdrv_select_ring (dev, resetargs->ring);
	jtagdrv_reset (dev);
	return (0);
	
    case JTAG_SHIFT_IR:
    case JTAG_SHIFT_DR:
	shiftargs = (JTAG_SHIFT_ARGS *) io_data;
	
	if (! VALID_JTAG_RING (shiftargs->ring) || shiftargs->nbits > (JTAG_MAX_DATA_LEN*JTAG_MAX_CHIPS)) {
	    return (-EFAULT);
		}

	buf = (u_char *) kmalloc (JTAG_NBYTES(shiftargs->nbits), GFP_KERNEL);

	if (buf == (u_char *) NULL)
	    return (-ENOMEM);
	
	if (copy_from_user (buf, shiftargs->value, JTAG_NBYTES(shiftargs->nbits)))
	{
	    kfree(buf);
	    return (-EFAULT);
	}


	jtagdrv_select_ring (dev, shiftargs->ring);

	if (io_cmd == JTAG_SHIFT_IR)
	    jtagdrv_shift_ir (dev, buf, shiftargs->nbits);
	else
	    jtagdrv_shift_dr (dev, buf, shiftargs->nbits);
	
	if (copy_to_user (shiftargs->value, buf, JTAG_NBYTES (shiftargs->nbits)))
	{
	    kfree (buf);
	    return (-EFAULT);
	}

	kfree (buf);
	return (0);

    case I2C_WRITE:
    case I2C_READ:
    case I2C_WRITEREG:
    case I2C_READREG:
	i2cargs = (I2C_ARGS *) io_data;

	if (! VALID_I2C_RING(i2cargs->ring) || i2cargs->count > I2C_MAX_DATA_LEN)
	    return (-EFAULT);

	jtagdrv_select_ring (dev, RING_I2C_BIT | i2cargs->ring);
	switch (io_cmd)
	{
	case I2C_WRITE:
	    i2cargs->ok = jtagdrv_i2c_write (dev, i2cargs->device, i2cargs->count, i2cargs->data);
	    break;

	case I2C_READ:
	    i2cargs->ok = jtagdrv_i2c_read (dev, i2cargs->device, i2cargs->count, i2cargs->data);
	    break;

	case I2C_WRITEREG:
	    i2cargs->ok = jtagdrv_i2c_writereg (dev, i2cargs->device, i2cargs->reg, i2cargs->count, i2cargs->data);
	    break;

	case I2C_READREG:
	    i2cargs->ok = jtagdrv_i2c_readreg (dev, i2cargs->device, i2cargs->reg, i2cargs->count, i2cargs->data);
	    break;
	}
	return (0);

    case I2C_CLOCK_SHIFT:
	clockargs = (I2C_CLOCK_SHIFT_ARGS *) io_data;

	freq = (10 * clockargs->m / (1 << (((clockargs->n + 1) & 3))));
	
	/* validate the value, and initialise the ring */
	if (clockargs->t != 0 || clockargs->n > 3 || clockargs->m > 127)
	    return (-EINVAL);
	
	jtagdrv_select_ring (dev, RING_I2C_BIT | RING_CLOCK_SHIFT);
	jtagdrv_i2c_clock_shift (dev, clockargs->t, clockargs->n, clockargs->m);
	jtagdrv_select_ring (dev, 0);
	return (0);

    default:
	return (-EINVAL);
    }
    return (-EINVAL);
}

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
