/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: jtagdrv_Linux.h,v 1.3 2002/08/09 11:18:37 addy Exp $"
/*      $Source: /cvs/master/quadrics/jtagmod/jtagdrv_Linux.h,v $*/

#ifndef __JTAGDRV_LINUX_H
#define __JTAGDRV_LINUX_H

#include <qsnet/kernel.h>
#include <asm/io.h>

typedef struct jtag_dev
{
    unsigned long	base;
    int                 region;

    u_int		present:1;
    u_int		open:1;
} JTAG_DEV;

/*
**
**			Hardware Defines
**
*/

/*
 * Assume that bit 4 of the Control Register is set to 1 (by default) 
 * to enable the printer port (CS3).
 *
 * The default base address is 3BC-3BF. 
 */

#define LPT0	0x3BC		 	/* CSR Base Address - note this can
					 * change depending on the setting
					 * in the Control Register 0.
					 *
					 * LPT1	0x378
					 * LPT2	0x278
					 * LPT3	0x268
					*/

/*
 *	Register offsets from the port base address
 */

#define LPT_REGISTER_0	0
#define LPT_REGISTER_1	1
#define LPT_REGISTER_2	2
#define LPT_REGISTER_3	0x400
#define LPT_REGISTER_4	0x401
#define LPT_REGISTER_5	0x402

/*
 *	Chip control registers
 */
					/* Base address for Super I/O National*/

#define SIO_BASE_ADDR	0x26e		/* Semiconductor PC87332VLJ combo-chip*/
#define CR4_REG		0x04 		/* index 4, printer control reg 4 */

#define LPT_EPP		0x01		/* Enable bit for epp */
#define LPT_ECP		0x04		/* Enable bit for ecp */

/*
 *	Registers for use with centronics, nibble and byte modes.
 */

#define LPT_DATA	LPT_REGISTER_0		/* line printer port data */
#define LPT_STAT	LPT_REGISTER_1		/* LPT port status 	  */
#define LPT_CTRL	LPT_REGISTER_2		/* LPT port control	  */

/*
 *	Registers for use with ECP mode.
 */ 

#define LPT_DFIFO	LPT_REGISTER_3		/* r/w fifo register    */
#define LPT_CFGB	LPT_REGISTER_4		/* Configuration B	*/
#define LPT_ECR		LPT_REGISTER_5		/* Exteded control 	*/

/*
 * Bit assignments for ECR register.
 */

	/* Bits 0-4 */

#define LPT_ECR_EMPTY	0x01		/* FIFO is empty */
#define LPT_ECR_FULL	0x02		/* FIFO is full */
#define LPT_ECR_SERV	0x04		/* Service bit */
#define LPT_ECR_DMA	0x08		/* DMA enable */
#define LPT_ECR_nINTR	0x10		/* Interrupt disable */

	/*
	 * Bits 5-7 are ECR modes.
	 */

#define LPT_ECR_PAR	0x20		/* Parallel port FIFO mode */
#define LPT_ECR_ECP	0x60		/* ECP mode */
#define LPT_ECR_CFG	0xE0		/* Configuration mode */
#define LPT_ECR_CLEAR	~0xE0		/* Cear mode bits */

/*
 * Bit assignments for the parallel port STATUS register:
 */

#define LPT_STAT_BIT0	0X1	/* Reserved. Bit always set.		*/
#define LPT_STAT_BIT1	0X2	/* Reserved. Bit always set.		*/
#define LPT_STAT_IRQ	0x4	/* interrupt status bit			*/
#define LPT_STAT_ERROR	0x8	/* set to 0 to indicate error		*/
#define LPT_STAT_SLCT	0x10	/* status of SLCT lead from printer	*/
#define LPT_STAT_PE	0x20	/* set to 1 when out of paper		*/
#define LPT_STAT_ACK	0x40	/* acknowledge - set to 0 when ready	*/
#define LPT_STAT_nBUSY	0x80	/* busy status bit, 0=busy, 1=ready	*/

/*
 * Bit assignments for the parallel port CONTROL register:
 */

#define LPT_CTRL_nSTROBE	0x1	/* Printer Strobe Control	*/
#define LPT_CTRL_nAUTOFD	0x2	/* Auto Feed Control		*/
#define LPT_CTRL_INIT		0x4	/* Initialize Printer Control	*/
#define LPT_CTRL_nSLCTIN	0x8	/* 0=select printer, 1=not selected */
#define LPT_CTRL_IRQ		0x10	/* Interrupt Request Enable Control */
#define LPT_CTRL_DIR		0x20	/* Direction control		*/
#define LPT_CTRL_BIT6		0X40	/* Reserved. Bit always set.	*/
#define LPT_CTRL_BIT7		0X80	/* Reserved. Bit always set.	*/


#define LPT_WRITE(dev, regname, value)	do { outb(value, (dev)->base + regname); } while (0)
#define LPT_READ(dev, regname,value)	do { value = inb((dev)->base + regname); } while (0)



/* Standard register access macros */
#define LPT_WRITE_CTRL(dev, value)	LPT_WRITE(dev, LPT_CTRL, LPT_CTRL_INIT | value)
#define LPT_WRITE_DATA(dev, value)	LPT_WRITE(dev, LPT_DATA, value)
#define LPT_READ_STAT(dev, value)	LPT_READ(dev, LPT_STAT, value)

/*
 * The jtag signals are connected to the parallel port as follows :
 *
 *  TRST	bit 0
 *  TDI		bit 1
 *  TMS		bit 2
 *  TCLK	AFX
 *  TDO		PE
 */
#define LPT_DATA_TRST	1
#define LPT_DATA_TDI	2
#define LPT_DATA_TMS	4
#define LPT_CTRL_TCLK	LPT_CTRL_nAUTOFD
#define LPT_STAT_TDO	LPT_STAT_PE

/*
 * The I2C signals are connected as follows :
 */
#define LPT_DATA_SDA	2
#define LPT_CTRL_SCLK	LPT_CTRL_nAUTOFD
#define LPT_STAT_SDA	LPT_STAT_PE

/*
 * The ring selection signals are as follows :
 *  addr	bit 0-7
 *  clock	nSLCTIN
 */
#define LPT_CTRL_RCLK	LPT_CTRL_nSLCTIN


#endif /* __JTAGDRV_LINUX_H */
