/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: i686_mmx.c,v 1.11 2004/01/05 12:08:25 mike Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/i686_mmx.c,v $*/

#include <qsnet/kernel.h>

#if defined(LINUX_I386)

#include <linux/config.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/i387.h>

int mmx_disabled = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
/* These functions are lifted from arch/i386/kernel/i387.c
 * and MUST be kept in step with the kernel (currently 2.4.17)
 * alternatively we should export the kernel_fpu_begin() function
 */
static inline void __save_init_fpu( struct task_struct *tsk )
{
	if ( cpu_has_fxsr ) {
		asm volatile( "fxsave %0 ; fnclex"
			      : "=m" (tsk->thread.i387.fxsave) );
	} else {
		asm volatile( "fnsave %0 ; fwait"
			      : "=m" (tsk->thread.i387.fsave) );
	}
	tsk->flags &= ~PF_USEDFPU;
}
#if defined(MODULE)
void kernel_fpu_begin(void)
{
	struct task_struct *tsk = current;

	if (tsk->flags & PF_USEDFPU) {
		__save_init_fpu(tsk);
		return;
	}
	clts();
}
#endif
#endif

extern inline int
mmx_preamble(void)
{
    if (mmx_disabled || in_interrupt())
	return (0);

    kernel_fpu_begin();

    return (1);
}

extern inline void
mmx_postamble(void)
{
    kernel_fpu_end();
}

extern u64
qsnet_readq (volatile u64 *ptr)
{
    u64 value;

    if (! mmx_preamble())
	value = *ptr;
    else
    {
	asm volatile ("movq (%0), %%mm0\n"
		      "movq %%mm0, (%1)\n"
		      : : "r" (ptr), "r" (&value) : "memory");
	mmx_postamble();
    }
    return (value);
}

void
qsnet_writeq(u64 value, volatile u64 *ptr)
{
    if (! mmx_preamble())
	*ptr = value;
    else
    {
	asm volatile ("movq (%0), %%mm0\n"
		      "movq %%mm0, (%1)\n"
		      : : "r" (&value), "r" (ptr) : "memory");
	mmx_postamble();
    }
}
#endif
