/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: kernel_linux.c,v 1.74.2.5 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/kernel_linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/ctrl_linux.h>
#include <qsnet/kpte.h>

#include <linux/sysctl.h>
#include <linux/init.h>
#include <qsnet/module.h>
#include <linux/vmalloc.h>

#include <qsnet/procfs_linux.h>

#include <linux/smp.h>		/* for smp_call_function() prototype */
#include <linux/smp_lock.h>
#include <linux/mm.h>

#include <linux/highmem.h>

extern int mmx_disabled;
extern int qsnet_debug_line_size;
extern int qsnet_debug_num_lines;

gid_t		       qsnet_procfs_gid;
struct proc_dir_entry *qsnet_procfs_root;
struct proc_dir_entry *qsnet_procfs_config;

MODULE_AUTHOR("Quadrics Ltd.");
MODULE_DESCRIPTION("QsNet Kernel support code");

MODULE_LICENSE("GPL");

#if defined(LINUX_I386)
module_param(mmx_disabled, uint, 0);
#endif

module_param(qsnet_debug_line_size, uint, 0);
module_param(qsnet_debug_num_lines, uint, 0);

module_param(qsnet_procfs_gid, uint, 0);

#ifdef KMEM_DEBUG
EXPORT_SYMBOL(qsnet_kmem_alloc_debug);
EXPORT_SYMBOL(qsnet_kmem_free_debug);
#else
EXPORT_SYMBOL(qsnet_kmem_alloc);
EXPORT_SYMBOL(qsnet_kmem_free);
#endif

EXPORT_SYMBOL(qsnet_kmem_display);
EXPORT_SYMBOL(kmem_to_phys);

EXPORT_SYMBOL(cpu_hold_all);
EXPORT_SYMBOL(cpu_release_all);

#if defined(LINUX_I386)
EXPORT_SYMBOL(qsnet_readq);
EXPORT_SYMBOL(qsnet_writeq);
#endif

/* debug.c */
EXPORT_SYMBOL(qsnet_debugf);
EXPORT_SYMBOL(kqsnet_debugf);
EXPORT_SYMBOL(qsnet_vdebugf);
EXPORT_SYMBOL(qsnet_debug_buffer);
EXPORT_SYMBOL(qsnet_debug_alloc);
EXPORT_SYMBOL(qsnet_debug_dump);
EXPORT_SYMBOL(qsnet_debug_kmem);
EXPORT_SYMBOL(qsnet_debug_disable);

EXPORT_SYMBOL(qsnet_assfail);

EXPORT_SYMBOL(qsnet_procfs_gid);
EXPORT_SYMBOL(qsnet_procfs_root);

static int qsnet_open    (struct inode *ino, struct file *fp);
static int qsnet_release (struct inode *ino, struct file *fp);
static int qsnet_ioctl   (struct inode *ino, struct file *fp, unsigned int cmd, unsigned long arg);

static struct file_operations qsnet_ioctl_fops = 
{
	ioctl:	 qsnet_ioctl,
	open:	 qsnet_open,
	release: qsnet_release,
};

static int
qsnet_open (struct inode *inode, struct file *fp)
{
	MOD_INC_USE_COUNT;
	fp->private_data = NULL;
	return (0);
}

static int
qsnet_release (struct inode *inode, struct file *fp)
{
	MOD_DEC_USE_COUNT;
	return (0);
}

static int 
qsnet_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
	int res=0;

	switch (cmd) 
	{
	case QSNETIO_DEBUG_KMEM:
	{
		QSNETIO_DEBUG_KMEM_STRUCT args;

		if (copy_from_user (&args, (void *) arg, sizeof (QSNETIO_DEBUG_KMEM_STRUCT)))
			return (-EFAULT);

		/* doesnt use handle as a pointer */
		qsnet_kmem_display(args.handle);
		break;
	}

	case QSNETIO_DEBUG_DUMP : 
	{
		res = qsnet_debug_dump();
		break;
	}

	case QSNETIO_DEBUG_BUFFER :
	{
		QSNETIO_DEBUG_BUFFER_STRUCT args;

		if (copy_from_user (&args, (void *) arg, sizeof (QSNETIO_DEBUG_BUFFER_STRUCT)))
			return (-EFAULT);

		/* qsnet_debug_buffer uses copyout */
		if ((res = qsnet_debug_buffer (args.addr, args.len)) != -1)
		{
			args.len = res;
			if (copy_to_user ((void *) arg, &args, sizeof (QSNETIO_DEBUG_BUFFER_STRUCT)))
				return (-EFAULT);
			res = 0;
		}
		break;
	}
	default:
		res = EINVAL;
		break;
	}

	return ((res == 0) ? 0 : -res);
}

#ifdef KMEM_DEBUG
static int qsnet_kmem_open    (struct inode *ino, struct file *fp);
static int qsnet_kmem_release (struct inode *ino, struct file *fp);
static ssize_t qsnet_kmem_read (struct file *file, char *buf, size_t count, loff_t *ppos);

static struct file_operations qsnet_kmem_fops = 
{
	open:	 qsnet_kmem_open,
	release: qsnet_kmem_release,
	read:	 qsnet_kmem_read,
};

typedef struct qsnet_private_space
{
	char * space;
	int    size;
	struct qsnet_private_space *next;
} QSNET_PRIVATE_SPACE;

typedef struct qsnet_private  
{
	QSNET_PRIVATE_SPACE *space_chain;
        QSNET_PRIVATE_SPACE *current_space;
	int                  current_pos;

} QSNET_PRIVATE;

#define QSNET_KMEM_DEBUG_LINE_SIZE ((int)512)
#define QSNET_PRIVATE_PAGE_SIZE    ((int)(4*1024))

static int qsnet_kmem_fill(QSNET_PRIVATE *pd);

void
destroy_chain(QSNET_PRIVATE * pd)
{
	QSNET_PRIVATE_SPACE *mem, *next;
	
	if (pd == NULL) return;

	for(mem = pd->space_chain ; mem != NULL; )
	{
		next = mem->next; 
		if ( mem->space ) 
			kfree ( mem->space);
		kfree(mem);
		mem = next;
	}
	kfree (pd);
}

QSNET_PRIVATE *
make_chain(int len)
{
	QSNET_PRIVATE       * pd;
	QSNET_PRIVATE_SPACE * mem;
	int                   i;

	/* make the private data block */
	if ((pd = kmalloc (sizeof (QSNET_PRIVATE), GFP_KERNEL)) == NULL)
		return NULL;
	pd->space_chain = NULL;

	/* first make the holders */
	for(i=0;i<len;i++)
	{
		if ((mem = kmalloc (sizeof (QSNET_PRIVATE_SPACE), GFP_KERNEL)) == NULL)
		{
			destroy_chain(pd);
			return (NULL);
		}
		mem->next  = pd->space_chain;
		mem->size  = 0;
		mem->space = 0;
		pd->space_chain = mem;

		/* now add the space */
		if ((mem->space = kmalloc (QSNET_PRIVATE_PAGE_SIZE, GFP_KERNEL)) == NULL)
		{
			destroy_chain(pd);
			return (NULL);
		}			

		mem->space[0] = 0;

	}

	pd->current_space = pd->space_chain;
	pd->current_pos   = 0;

	return pd;
}

static int
qsnet_kmem_open (struct inode *inode, struct file *fp)
{
	MOD_INC_USE_COUNT;
	fp->private_data = NULL;
	return (0);
}

static int
qsnet_kmem_release (struct inode *inode, struct file *fp)
{
	if ( fp->private_data )
	{
		QSNET_PRIVATE * pd = (QSNET_PRIVATE *) fp->private_data;

		/* free the space */
		if (pd->space_chain)
			kfree (pd->space_chain);	

		/* free struct */
		kfree (pd);
	}
	MOD_DEC_USE_COUNT;
	return (0);
}

static ssize_t
qsnet_kmem_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	QSNET_PRIVATE * pd = (QSNET_PRIVATE *) file->private_data;
	int             error;
	int             output_count;
	int             num_of_links=10;

	/* make a buffer to output count bytes in */
	if ((error = verify_area (VERIFY_WRITE, buf, count)) != 0)
		return (error);

	if ( pd == NULL) 
	{
		/* first time */

		/* ok we have to guess at how much space we are going to need  */
		/* if it fails we up the space and carry try again             */
		/* we have to do it this way as we cant get more memory whilst */
		/* holding the lock                                            */
		if ((pd = make_chain(num_of_links)) == NULL)
			return (-ENOMEM);	

		while ( qsnet_kmem_fill(pd) ) 
		{
			destroy_chain(pd);
			num_of_links += 10;
			if ((pd = make_chain(num_of_links)) == NULL)
				return (-ENOMEM);	
		}

		/* we have the space and filled it */
		file->private_data = (void *)pd; 	
	}
		
	/* output buffer */
	if ( pd->current_pos >= pd->current_space->size )
		return (0); /* finished */

	output_count = pd->current_space->size - pd->current_pos;
	if ( output_count > count ) 
		output_count = count;

	copy_to_user(buf, (pd->current_space->space + pd->current_pos), output_count);

	pd->current_pos += output_count;
	ppos            += output_count;

	/* just check to see if we have finished the current space */
	if ( pd->current_pos >= pd->current_space->size )
	{
		if ( pd->current_space->next )
		{
			pd->current_space = pd->current_space->next;
			pd->current_pos   = 0;
		}
	}	

	return (output_count);
}
#endif /* KMEM_DEBUG */

static int
proc_write_qsnetdebug(struct file *file, const char *buffer,
		      unsigned long count, void *data)
{
	char    tmpbuf[128];
	int     res;
	
	if (count > sizeof (tmpbuf)-1)
		return (-EINVAL);
	
	MOD_INC_USE_COUNT;
	
	if (copy_from_user (tmpbuf, buffer, count))
		res = -EFAULT;
	else 
	{
		tmpbuf[count] = '\0'; 	
		
		if (tmpbuf[count-1] == '\n')
			tmpbuf[count-1] = '\0';
		
		if (! strcmp (tmpbuf, "on"))
			qsnet_debug_buffer_on();

		if (! strcmp (tmpbuf, "clear"))
			qsnet_debug_buffer_clear();

		if (! strncmp (tmpbuf, "mark",4))
			qsnet_debug_buffer_mark( &tmpbuf[4] );
		
		res = count;
	}
	
	MOD_DEC_USE_COUNT;
	
	return (res);
}

static int
proc_read_qsnetdebug(char *page, char **start, off_t off,
		     int count, int *eof, void *data)
{
	int len = sprintf (page, "echo command > /proc/qsnet/config/qsnetdebug\ncommand = on | off | clear | mark text\n");
	return (qsnet_proc_calc_metrics (page, start, off, count, eof, len));
}

#include "quadrics_version.h"
extern int kqsnet_debug_running;
static char       quadrics_version[] = QUADRICS_VERSION;

static int __init qsnet_init(void)
{
	struct proc_dir_entry *p;

	printk ("qsnet Module (version %s)\n", quadrics_version);
	if ((qsnet_procfs_root = proc_mkdir ("qsnet", 0)) == NULL)
	{
		printk ("qsnet: failed to create /proc/qsnet \n");
		return (-ENXIO);
	}
	
	if ((p = create_proc_entry ("ioctl", S_IRUGO|S_IWUSR|S_IWGRP, qsnet_procfs_root)) == NULL)
	{
		printk ("qsnet: failed to register /proc/qsnet/ioctl\n");
		return (-ENXIO);
	}
	p->proc_fops = &qsnet_ioctl_fops;
	p->owner     = THIS_MODULE;
	p->data      = NULL;
	p->gid       = qsnet_procfs_gid;

	qsnet_proc_register_str (qsnet_procfs_root, "version", quadrics_version, S_IRUGO);

	if ((qsnet_procfs_config = proc_mkdir ("config", qsnet_procfs_root)) == NULL)
	{
		printk ("qsnet: failed to create /proc/qsnet/config \n");
		return (-ENXIO);
	}

#ifdef KMEM_DEBUG
	if ((p = create_proc_entry ("kmem_debug", S_IRUGO|S_IWUSR|S_IWGRP, qsnet_procfs_config)) == NULL)
	{
		printk ("qsnet: failed to register /proc/qsnet/config/kmem_debug\n");
		return (-ENXIO);
	}
	p->proc_fops = &qsnet_kmem_fops;
	p->owner     = THIS_MODULE;
	p->data      = NULL;
	p->gid       = qsnet_procfs_gid;
#endif		
      
	qsnet_debug_init(); 

	qsnet_proc_register_int (qsnet_procfs_config, "kqsnet_debug_running", &kqsnet_debug_running, 0);

	if ((p = create_proc_entry ("qsnetdebug", S_IRUGO|S_IWUSR|S_IWGRP, qsnet_procfs_config)) == NULL)
	{
		printk ("qsnet: failed to register /proc/qsnet/config/qsnetdebug\n");
		return (-ENXIO);
	}
	p->read_proc  = proc_read_qsnetdebug;
	p->write_proc = proc_write_qsnetdebug;
	p->owner      = THIS_MODULE;
	p->data       = NULL;
	p->gid        = qsnet_procfs_gid;
	
	return (0);
}

static void __exit qsnet_exit(void)
{
#ifdef KMEM_DEBUG
	qsnet_kmem_display(0);
#endif
	qsnet_debug_fini();

	remove_proc_entry ("qsnetdebug",           qsnet_procfs_config);
	remove_proc_entry ("kqsnet_debug_running", qsnet_procfs_config);
#ifdef KMEM_DEBUG
	remove_proc_entry ("kmem_debug", 	   qsnet_procfs_config);
#endif
	remove_proc_entry ("config",               qsnet_procfs_root);

	remove_proc_entry ("version", qsnet_procfs_root);
	remove_proc_entry ("ioctl",   qsnet_procfs_root);

	remove_proc_entry ("qsnet", 0);
}

/* Declare the module init and exit functions */
module_init(qsnet_init);
module_exit(qsnet_exit);

#ifdef KMEM_DEBUG
/*
 * Kernel memory allocation.  We maintain our own list of allocated mem
 * segments so we can free them on module cleanup.
 * 
 * We use kmalloc for allocations less than one page in size; vmalloc for
 * larger sizes.
 */

typedef struct {
	struct list_head list;
	void 		*ptr;
	int 		len;
	int 		used_vmalloc;
	void            *owner;
	void            *caller;
	unsigned int     time;
	int              line;
	char             filename[20];
} kmalloc_t;

static LIST_HEAD(kmalloc_head);

static spinlock_t	kmalloc_lock = SPIN_LOCK_UNLOCKED;

/*
 * Kernel memory allocation.  We use kmalloc for allocations less 
 * than one page in size; vmalloc for larger sizes.
 */

static int
qsnet_kmem_fill(QSNET_PRIVATE *pd)
{
	kmalloc_t *kp;
	struct list_head *lp;
	unsigned long flags;
	char str[QSNET_KMEM_DEBUG_LINE_SIZE];
	QSNET_PRIVATE_SPACE * current_space;
	int                   current_pos;
	int                   len;
	current_space = pd->space_chain;
	current_pos   = 0;
	
	
	current_space->space[0] = 0;	
	spin_lock_irqsave(&kmalloc_lock, flags);
	for (lp = kmalloc_head.next; lp != &kmalloc_head;  lp = lp->next) {
		kp = list_entry(lp, kmalloc_t, list);
		
		/* make the next line */
		sprintf(str,"%p %d %d %p %p %u %d %s\n",
			kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->time, kp->line, kp->filename);
		len = strlen(str);
		
		/* does it fit on the current page */
		if ( (current_pos + len + 1) >=  QSNET_PRIVATE_PAGE_SIZE)
		{
			/* move onto next page */
			if ((current_space = current_space->next) == NULL)
			{
				/* run out of space !!!! */
				spin_unlock_irqrestore(&kmalloc_lock, flags);
				return (1);
			}
			current_space->space[0] = 0;	
			current_pos = 0;
		}
		strcat( current_space->space + current_pos, str);
		current_pos += len;

		/* remember how much we wrote to this page */
		current_space->size = current_pos;

	}
	spin_unlock_irqrestore(&kmalloc_lock, flags);

	return (0);
}

void * 
qsnet_kmem_alloc_debug(int len, int cansleep, int zerofill, char *file, int line)
{
	void *new;
	unsigned long flags;
	kmalloc_t *kp;

	if (len < PAGE_SIZE || !cansleep)
		new = kmalloc(len, cansleep ? GFP_KERNEL : GFP_ATOMIC);
	else
		new = vmalloc(len);

	if (len >= PAGE_SIZE)
		ASSERT(PAGE_ALIGNED((uintptr_t) new));

	if (new && zerofill)
		memset(new,0,len);

	/* record allocation */
	kp = kmalloc(sizeof(kmalloc_t), cansleep ? GFP_KERNEL : GFP_ATOMIC);
	ASSERT(kp != NULL);
	kp->len = len;
	kp->ptr = new;
	kp->used_vmalloc = (len >= PAGE_SIZE || cansleep);
	kp->owner  = current;
	kp->caller = __builtin_return_address(0);
	kp->time = lbolt;
	kp->line = line;
	len = strlen(file);

	if (len > 18) 
		strcpy(kp->filename,&file[len-18]);
	else
		strcpy(kp->filename,file);

	spin_lock_irqsave(&kmalloc_lock, flags);
	list_add(&kp->list, &kmalloc_head);
	spin_unlock_irqrestore(&kmalloc_lock, flags);

	return new;
}

void 
qsnet_kmem_free_debug(void *ptr, int len, char *file, int line)
{
	unsigned long flags;
	kmalloc_t *kp;
	struct list_head *lp;

	spin_lock_irqsave(&kmalloc_lock, flags);
	for (lp = kmalloc_head.next; lp != &kmalloc_head; lp = lp->next) {
		kp = list_entry(lp, kmalloc_t, list);
		if (kp->ptr == ptr) {
			if (kp->len != len)
				printk("qsnet_kmem_free_debug(%p) ptr %p len %d mismatch: expected %d caller %p owner %p (%s:%d)\n",
				       current, ptr, len, kp->len, __builtin_return_address(0), kp->caller, file, line);
			list_del(lp);
			kfree(kp); /* free off descriptor */
			break;
		}
	}
	spin_unlock_irqrestore(&kmalloc_lock, flags);

	if (lp == &kmalloc_head) /* segment must be found */
	{
		printk( "qsnet_kmem_free_debug(%p) ptr %p len %d not found: caller %p (%s:%d)\n",
			current, ptr, len, __builtin_return_address(0), file, line);
	}

	if ((((unsigned long) ptr) >= VMALLOC_START && ((unsigned long) ptr) < VMALLOC_END)) 
		vfree (ptr);
	else
		kfree (ptr);
}

#else /* !KMEM_DEBUG */

void * 
qsnet_kmem_alloc(int len, int cansleep, int zerofill)
{
	void *new;

	if (len < PAGE_SIZE || !cansleep)
		new = kmalloc(len, cansleep ? GFP_KERNEL : GFP_ATOMIC);
	else
		new = vmalloc(len);

	if (len >= PAGE_SIZE)
		ASSERT(PAGE_ALIGNED((unsigned long) new));

	if (new && zerofill)
		memset(new,0,len);

	return new;
}

void 
qsnet_kmem_free(void *ptr, int len)
{
	if ((((unsigned long) ptr) >= VMALLOC_START && ((unsigned long) ptr) < VMALLOC_END)) 
		vfree (ptr);
	else
		kfree (ptr);
}
#endif /* !KMEM_DEBUG */

void
qsnet_kmem_display(void *handle)
{
#ifdef KMEM_DEBUG
	kmalloc_t *kp;
	struct list_head *lp;
	unsigned long flags;
	int count = 0, totsize = 0;

	spin_lock_irqsave(&kmalloc_lock, flags);
	for (lp = kmalloc_head.next; lp != &kmalloc_head;  lp = lp->next) {
		kp = list_entry(lp, kmalloc_t, list);

		if (!handle || handle == kp->owner)
		{
			printk("qsnet_kmem_display(%p): mem %p len %d unfreed caller %p (%p) \n",
			       handle, kp->ptr, kp->len, kp->caller, kp->owner);
		    
			count++;
			totsize += kp->len;
		}
	}
	spin_unlock_irqrestore(&kmalloc_lock, flags);

	printk("qsnet_kmem_display(%p): %d bytes left in %d objects\n", handle, totsize, count);
#endif
}

physaddr_t
kmem_to_phys(void *ptr)
{
	virtaddr_t virt = (virtaddr_t) ptr;
	physaddr_t phys;
	pte_t     *pte;

	if ((virt >= VMALLOC_START && virt < VMALLOC_END))
	{
		pte = find_pte_kernel(virt);
		ASSERT(pte && !pte_none(*pte));
		phys = pte_phys(*pte) + (virt & (PAGE_SIZE-1));
	}
#if defined(PKMAP_BASE)
	else if (virt >= PKMAP_BASE && virt < (PKMAP_BASE + LAST_PKMAP * PAGE_SIZE))
	{
		pte = find_pte_kernel(virt);
		ASSERT(pte && !pte_none(*pte));
		phys = pte_phys(*pte) + (virt & (PAGE_SIZE-1));
	}
#endif
#if defined(__ia64)
	else if (virt >= __IA64_UNCACHED_OFFSET && virt < PAGE_OFFSET)
	{
		/* ia64 non-cached KSEG */
		phys = ((physaddr_t) ptr - __IA64_UNCACHED_OFFSET);
	}
#endif
	else /* otherwise it's KSEG */
	{
		phys = __pa(virt);
	}
	    
#if defined(CONFIG_ALPHA_GENERIC) || (defined(CONFIG_ALPHA_EV6) && !defined(USE_48_BIT_KSEG))
	/* 
	 * with TS_BIAS as bit 40 - the tsunami pci space is mapped into
	 * the kernel at 0xfffff500.00000000 however we need to convert
	 * this to the true physical address 0x00000800.00000000.
	 *
	 * there is no need for PHYS_TWIDDLE since we knew we'd get a kernel
	 * virtual address already and handled this with __pa().
	 */
	if (phys & (1ul << 40)) {
		phys &= ~(1ul << 40);   /*   clear bit 40 (kseg I/O select) */
		phys |= (1ul << 43);    /*   set   bit 43 (phys I/O select) */
	}
#endif
	return phys;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)

EXPORT_SYMBOL(pci_resource_size);
EXPORT_SYMBOL(pci_get_base_address);
EXPORT_SYMBOL(pci_base_to_kseg);


/*
 * PCI stuff.  
 *
 * XXX pci_base_to_kseg() and pci_kseg_to_phys() are problematic
 * in that they may not work on non-Tsunami (DS20, ES40, etc) 
 * architectures, and may not work in non-zero PCI bus numbers.
 */

unsigned long 
pci_get_base_address(struct pci_dev *pdev, int index)
{
	unsigned long base;

	ASSERT(index >= 0 && index <= 5);
	/* borrowed in part from drivers/scsi/sym53c8xx.c */
	base = pdev->base_address[index++];

#if BITS_PER_LONG > 32
	if ((base & 0x7) == 0x4)
		base |= (((unsigned long)pdev->base_address[index]) << 32);
#endif
	return base;
}

unsigned long 
pci_resource_size(struct pci_dev *pdev, int index)
{
	u32 addr, mask, size;

	static u32 bar_addr[] = {
		PCI_BASE_ADDRESS_0, 
		PCI_BASE_ADDRESS_1, 
		PCI_BASE_ADDRESS_2,
		PCI_BASE_ADDRESS_3, 
		PCI_BASE_ADDRESS_4, 
		PCI_BASE_ADDRESS_5, 
	};
	ASSERT(index >= 0 && index <= 5);

	/* algorithm from Rubini book */
	pci_read_config_dword (pdev,	bar_addr[index], &addr);
	pci_write_config_dword(pdev,	bar_addr[index], ~0);
	pci_read_config_dword (pdev,	bar_addr[index], &mask);
	pci_write_config_dword(pdev,	bar_addr[index], addr);

	mask &= PCI_BASE_ADDRESS_MEM_MASK;
	size = ~mask + 1;
	return size;
}

/*
 * Convert BAR register value to KSEG address.
 */
void *
pci_base_to_kseg(u64 baddr, int bus)
{
	u64 kseg;

	/* XXX tsunami specific */
	baddr &= ~(u64)0x100000000;  /* mask out hose bit */
	kseg = TSUNAMI_MEM(bus) + baddr;
	return (void *)kseg; 
}

#endif /* LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,0) */

/*
 * Spin the other CPU's in an SMP system.
 * smp_call_function() needed to be exported to modules.  It will be
 * papered over in <linux/smp.h> if running on a non-SMP box.
 */
static spinlock_t hold_lock = SPIN_LOCK_UNLOCKED;

#if 0
static void cpu_hold(void *unused)
{
	spin_lock(&hold_lock);
	spin_unlock(&hold_lock);
}
#endif

void cpu_hold_all(void)
{
	spin_lock(&hold_lock);

#if 0
	{
		int res;
		int retries = 10; 
	    
		/* XXXXX: cannot call smp_call_function() from interrupt context */
	    
		do {
			/* only request blocking retry if not in interrupt context */
			res = smp_call_function(cpu_hold, NULL, !in_interrupt(), 0);
			if (res)
				mdelay(5);
		} while (res && retries--);
	    
		if (res)
			printk("cpu_hold_all: IPI timeout\n");
	}
#endif
}

void cpu_release_all(void)
{
	spin_unlock(&hold_lock);
}

void
qsnet_proc_character_fill (long mode, char *fmt, ...)
{
    int len;
    va_list ap;
    QSNET_PROC_PRIVATE *private = (QSNET_PROC_PRIVATE *)mode;
    
    /* is the buffer already full */
    if (private->pr_len >= private->pr_data_len) 
	return;
    
    /* attempt to fill up to the remaining space */
    va_start (ap, fmt);
    len = vsnprintf ( & private->pr_data[private->pr_len], (private->pr_data_len - private->pr_len), fmt, ap);
    va_end (ap);
    
    if (len < 0 ) 
    {
	/* we have reached the end of buffer and need to fail all future writes
	 * the caller can check (pr_len >= pr_data_len) and recall with more space 
	 */
	private->pr_len = private->pr_data_len;
	return;
    }
    
    /* move the length along */
    private->pr_len += len;   
}

int
qsnet_proc_release (struct inode *inode, struct file *file)
{
    QSNET_PROC_PRIVATE *pr = (QSNET_PROC_PRIVATE *) file->private_data;
    
    if (pr->pr_data)
	KMEM_FREE (pr->pr_data, pr->pr_data_len);
    kfree (pr);
    
    MOD_DEC_USE_COUNT;
    return (0);
}

EXPORT_SYMBOL(qsnet_proc_character_fill);
EXPORT_SYMBOL(qsnet_proc_release);

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
