/*
 *    Copyright (c) 2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: qsnetkmem_linux.c,v 1.3 2003/08/13 10:03:27 fabien Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/qsnetkmem_linux.c,v $*/

/* macro macros */
#define MACRO_BEGIN     do {
#define MACRO_END       } while (0)
#define offsetof(T,F) ((int )&(((T *)0)->F))

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <qsnet/config.h>
#include <qsnet/list.h>
#include <qsnet/procfs_linux.h>
#include <signal.h>
#include <sys/wait.h>

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)

typedef struct {
	struct list_head list;
	void 		*ptr;
	int 		len;
	int 		used_vmalloc;
	void            *owner;
	void            *caller;
	unsigned int     time;
	int              mark;
	int              line;
	char             file[256];
	
} kmalloc_t;


static LIST_HEAD(current_kmem);
static LIST_HEAD(stored_kmem);

void
count_kmem(struct list_head * list, long * count, long * size )
{
	long              c,s;
	struct list_head *tmp;
	kmalloc_t        *kmem_ptr = NULL;


	c = s = 0L;

	list_for_each(tmp, list) {
		kmem_ptr = list_entry(tmp, kmalloc_t , list);
		c++;
		s += kmem_ptr->len;
	}	

	*count = c;
	*size  = s;
}

void
clear_kmem(struct list_head * list)
{
	struct list_head *tmp,*tmp2;
	kmalloc_t        *kmem_ptr = NULL;

	list_for_each_safe(tmp, tmp2, list) {
		kmem_ptr = list_entry(tmp, kmalloc_t , list);
		list_del_init(&kmem_ptr->list);
		free( kmem_ptr );
	}
}

void 
move_kmem(struct list_head * dest, struct list_head *src)
{
	struct list_head *tmp,*tmp2;
	kmalloc_t        *kp= NULL;

	list_for_each_safe(tmp, tmp2, src) {
		kp = list_entry(tmp, kmalloc_t , list);
		list_del_init(&kp->list);

/*
		printf("mem %p len %d (vm=%d)  caller %p owner %p (%s:%d)\n",
		       kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->file, kp->line);
*/		    

		list_add_tail(&kp->list, dest);
	}
}

void
read_kmem(struct list_head * list)
{
	FILE      * fd;
	char        line[1024];
	int         line_size = 100;
	char      * rep;
	kmalloc_t * kp;

	clear_kmem(list);

	fd = fopen(QSNET_PROCFS_KMEM_DEBUG,"r");
	if ( fd == NULL) 
	{
		printf("No Kmem Debug\n");
		return;
	}

	rep = fgets(line,line_size, fd);

	while ( rep != NULL ) 
	{
		kp = malloc(sizeof(kmalloc_t));

		sscanf(line,"%p %d %d %p %p %u %d %s\n",
		       &kp->ptr, &kp->len, &kp->used_vmalloc, &kp->caller, &kp->owner, &kp->time, &kp->line, &kp->file[0]);

/*
		printf(">>%s<<\n",line);
		printf("%p %d %d %p %p %u %d %s\n",
		       kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->time, kp->line, kp->file);
*/

		list_add_tail(&kp->list, list);

		rep = fgets(line,line_size, fd);
	}
	fclose(fd);
}

void
mark_kmem(struct list_head * list, int mark)
{
	struct list_head *tmp;
	kmalloc_t        *kp = NULL;

	list_for_each(tmp, list) {
		kp = list_entry(tmp, kmalloc_t , list);

		kp->mark = mark;
	}
}

kmalloc_t *
find_kmem(kmalloc_t * value, struct list_head * list)
{
	struct list_head *tmp;
	kmalloc_t        *kp = NULL;

	
	list_for_each(tmp, list) {
		kp = list_entry(tmp, kmalloc_t , list);
		if ( (kp->ptr == value->ptr)
		     && (kp->len == value->len)
		     && (kp->used_vmalloc  == value->used_vmalloc )
		     && (kp->owner  == value->owner )
		     && (kp->caller  == value->caller )
		     && (kp->time  == value->time )
		     && (kp->line  == value->line )
		     && !(strcmp(kp->file,value->file) ))
			return kp;
	}	
	return NULL;
}

void 
diff_kmem(struct list_head *curr, struct list_head *stored)
{
	struct list_head *tmp;
	kmalloc_t        *kp = NULL;
	long              c,s;

	mark_kmem(stored,  0);
	mark_kmem(curr,    0);
	
	list_for_each(tmp, stored) {
		kp = list_entry(tmp, kmalloc_t , list);
		if (find_kmem( kp, curr) != NULL) 
			kp->mark = 1;
	}
	
	list_for_each(tmp, curr) {
		kp = list_entry(tmp, kmalloc_t , list);
		if (find_kmem( kp, stored) != NULL) 
			kp->mark = 1;
	}		

	c=s=0L;
	list_for_each(tmp, stored) {
		kp = list_entry(tmp, kmalloc_t , list);
		if (kp->mark != 1)
		{
			printf("-- mem %p len %d (vm=%d)  caller %p owner %p (%s:%d)\n",
			       kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->file, kp->line);
			c++;
			s+= kp->len;
		}
	}
	printf("-- %4ld %10ld \n",c,s);
	
	c=s=0L;
	list_for_each(tmp, curr) {
		kp = list_entry(tmp, kmalloc_t , list);
		if (kp->mark != 1)
		{
			printf("++ mem %p len %d (vm=%d)  caller %p owner %p (%s:%d)\n",
			       kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->file, kp->line);
			c++;
			s+= kp->len;
		}
	}		
	printf("++ %4ld %10ld \n",c,s);
}


void
print_kmem(struct list_head * list)
{
	struct list_head *tmp;
	kmalloc_t        *kp = NULL;

	list_for_each(tmp, list) {
		kp = list_entry(tmp, kmalloc_t , list);

		printf("mem %p len %d (vm=%d)  caller %p owner %p (%s:%d)\n",
		       kp->ptr, kp->len, kp->used_vmalloc, kp->caller, kp->owner, kp->file, kp->line);
		    
	}
}

void 
print_cmds()
{
	long c,s;

	printf("q : quits \n");
	printf("r : read\n");
	printf("c : print current\n");
	printf("o : print stored\n");
	printf("s : store\n");

	count_kmem(&current_kmem, &c, &s );
	printf("\ncurrent : %4ld %10ld\n", c , s);
 
	count_kmem(&stored_kmem, &c, &s );
	printf("store   : %4ld %10ld\n", c , s);
 
}

int
main()
{
	char            line[128];
	int             line_size=127;
	int             len;


	while (1)
	{
		
		printf(">> ");
		fgets(line,line_size, stdin);
	
		
		len = strlen( line ) -1;
		if ( len ) 
		{
			switch ( tolower(line[0]) ) 
			{
			case 'q':
				exit(0);

			case 'r' :
				read_kmem(&current_kmem);
				break;

			case 'c' :
				print_kmem(&current_kmem);
				break;

			case 'o' :
				print_kmem(&stored_kmem);
				break;

			case 's' :
				clear_kmem(&stored_kmem);
				move_kmem(&stored_kmem, &current_kmem);
				break;

			case 'd' :
				diff_kmem(&current_kmem, &stored_kmem);
				break;

			default:
				print_cmds();	
			}

		
			
		}
		else
			print_cmds();
	}

}


/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
