/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: rms_kern_Linux.c,v 1.25.2.3 2005/09/07 14:35:04 mike Exp $"
/*      $Source: /cvs/master/quadrics/rmsmod/rms_kern_Linux.c,v $*/

#include <qsnet/kernel.h>
#include <qsnet/autoconf.h>

#include <linux/sysctl.h>
#include <linux/init.h>
#include <qsnet/module.h>
#include <linux/proc_fs.h>
#ifndef NO_PTRACK
#include <linux/ptrack.h>
#endif

#include <rms/rmscall.h>
#include <rms/rmsio.h>

extern int ptrack_enabled;

MODULE_AUTHOR("Quadrics Ltd");
MODULE_DESCRIPTION("RMS support module");
MODULE_LICENSE("GPL");

#ifndef NO_PTRACK
module_param(ptrack_enabled, uint, 0);
#endif

int rms_debug = 0;

ctl_table rms_table[] = {
    {
	.ctl_name = 1,
	.procname = "rms_debug",
	.data     = &rms_debug,
	.maxlen   = sizeof(int),
	.mode     = 0644,
	.child    = NULL,
	.proc_handler = &proc_dointvec,
    },
    {0}
};

ctl_table rms_root_table[] = {
    {
	.ctl_name = CTL_DEBUG,
	.procname = "rms",
	.data     = NULL,
	.maxlen   = 0,
	.mode     = 0555,
	.child    = rms_table,
    },
    {0}
};

static struct ctl_table_header *rms_sysctl_header;

static int rms_open (struct inode *ino, struct file *fp);
static int rms_release (struct inode *ino, struct file *fp);
static int rms_ioctl (struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg);

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
static int
rms_ioctl32_cmds[] =
{
    RMSIO_GETPRGID32,
    RMSIO_GETCAP32
};

static int      rms_ioctl32 (unsigned int fd, unsigned int cmd, 
			     unsigned long arg, struct file *file);
#endif

static struct file_operations rms_fops =
{
    .owner   = THIS_MODULE,
    .ioctl   = rms_ioctl,
    .open    = rms_open,
    .release = rms_release,
};

struct proc_dir_entry *rms_procfs_programs;
static struct proc_dir_entry *rms_procfs_root;

int version_callback(char* page, char** start, off_t off, int count, int* eof, void* data)
{
    return(sprintf(page, "$Id: rms_kern_Linux.c,v 1.25.2.3 2005/09/07 14:35:04 mike Exp $\n"));
}

int ptrack_callback(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    if (ptrack_enabled)
	return(sprintf(page, "enabled\n"));
    else
	return(sprintf(page, "disabled\n"));
}


static int __init rms_start(void)
{
    struct proc_dir_entry *p;
    int res;

    if ((rms_sysctl_header = register_sysctl_table(rms_root_table, 1)) == 0)
    {
	printk ("rmsmod: failed to register sysctl table\n");
	return (-ENXIO);
    }
    
    if ((rms_procfs_root = proc_mkdir("rms",  NULL)) == NULL ||
	(rms_procfs_programs = proc_mkdir("programs",  rms_procfs_root)) == NULL ||
	(p = create_proc_entry ("control", S_IRUGO, rms_procfs_root)) == NULL)
    {
	unregister_sysctl_table (rms_sysctl_header);
	printk ("rmsmod: failed to register /proc/rms\n");
	return (-ENXIO);
    }
    p->proc_fops = &rms_fops;
    p->owner     = THIS_MODULE;
    p->data      = NULL;

    if ((p = create_proc_entry ("version", S_IRUGO, rms_procfs_root)) != NULL)
    {
	p->owner = THIS_MODULE;
	p->data = NULL;
	p->read_proc = version_callback;
    }

#ifndef NO_PTRACK
    if ((p = create_proc_entry ("ptrack", S_IRUGO, rms_procfs_root)) != NULL)
    {
	p->owner = THIS_MODULE;
	p->data = NULL;
	p->read_proc = ptrack_callback;
    }
#endif

    if ((res = rms_init()) != ESUCCESS)
    {
#ifndef NO_PTRACK
	remove_proc_entry ("ptrack", rms_procfs_root);
#endif
	remove_proc_entry ("version", rms_procfs_root);
	remove_proc_entry ("programs", rms_procfs_root);
	remove_proc_entry ("control", rms_procfs_root);
	remove_proc_entry ("rms", NULL);
	unregister_sysctl_table (rms_sysctl_header);
	return (-res);
    }

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
    lock_kernel();
    {
	extern int register_ioctl32_conversion(unsigned int cmd, int (*handler)(unsigned int, unsigned int, unsigned long, struct file *));
	register int i;
	for (i = 0; i < sizeof (rms_ioctl32_cmds)/sizeof(rms_ioctl32_cmds[0]); i++)
	    register_ioctl32_conversion (rms_ioctl32_cmds[i], rms_ioctl32);
    }
    unlock_kernel();
#endif
    return (0);
}

static void __exit rms_exit(void)
{
    rms_fini();

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
    lock_kernel();
    {
	extern void unregister_ioctl32_conversion(unsigned int cmd);
	register int i;

	for (i = 0; i < sizeof (rms_ioctl32_cmds)/sizeof(rms_ioctl32_cmds[0]); i++)
	    unregister_ioctl32_conversion (rms_ioctl32_cmds[i]);
    }
    unlock_kernel();
#endif

#ifndef NO_PTRACK
    remove_proc_entry ("ptrack", rms_procfs_root);
#endif
    remove_proc_entry ("version", rms_procfs_root);
    remove_proc_entry ("programs", rms_procfs_root);
    remove_proc_entry ("control", rms_procfs_root);
    remove_proc_entry ("rms", NULL);
    unregister_sysctl_table(rms_sysctl_header);
}

/* Declare the module init and exit functions */
module_init(rms_start);
module_exit(rms_exit);

static int
rms_open (struct inode *inode, struct file *fp)
{
    MOD_INC_USE_COUNT;
    fp->private_data = NULL;

    return (0);
}

static int
rms_release (struct inode *inode, struct file *fp)
{
    MOD_DEC_USE_COUNT;
    return (0);
}

static int 
rms_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
    int res;

    /* printk ("rmsmod: ioctl %x\n", cmd); */

    switch (cmd) 
    {
/* no corepath support in Linux yet */
#if 0
    case RMSIO_SETCOREPATH:
	res = rms_setcorepath((caddr_t)arg);
	break;
	
    case RMSIO_GETCOREPATH:
    {
	RMSIO_GETCOREPATH_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_getcorepath(args.pid, args.corepath, args.maxlen);
	break;
    }
#endif
	
    case RMSIO_PRGCREATE:
    {
	RMSIO_PRGCREATE_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prgcreate(args.id, args.uid, args.cpus);
	break;
    }

    case RMSIO_PRGDESTROY:
	res = rms_prgdestroy(arg);
	break;
	
    case RMSIO_PRGIDS:
    {
	RMSIO_PRGIDS_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prgids(args.maxids, args.prgids, args.nprgs);
	break;
    }

    case RMSIO_PRGINFO:
    {
	RMSIO_PRGINFO_STRUCT args;
	
	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prginfo(args.id, args.maxpids, args.pids, args.nprocs);
	break;
    }
	
    case RMSIO_PRGSIGNAL:
    {
	RMSIO_PRGSIGNAL_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prgsignal(args.id, args.signo);
	break;
    }
	
    case RMSIO_PRGADDCAP:
    {
	RMSIO_PRGADDCAP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prgaddcap(args.id, args.index, args.cap);
	break;
    }

    case RMSIO_SETCAP:
    {
	RMSIO_SETCAP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_setcap(args.index, args.ctx);
	break;
    }
	
    case RMSIO_NCAPS:
	res = rms_ncaps((int *)arg);
	break;
	
    case RMSIO_GETPRGID:
    {
	RMSIO_GETPRGID_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_getprgid(args.pid, args.id);
	break;
    }

    case RMSIO_GETMYCAP:
	res = rms_mycap((int *)arg);
	break;
	
    case RMSIO_GETCAP:
    {
	RMSIO_GETCAP_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_getcap(args.index, args.cap);
	break;
    }

    case RMSIO_PRGGETSTATS:
    {
	/* no longer supported */
	res = EINVAL;
	break;
    }
    
    case RMSIO_PRGGETSTATS2:
    {
	RMSIO_PRGGETSTATS2_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prggetstats(args.id, args.stats);
	break;
    }

    case RMSIO_PRGSUSPEND:
	res = rms_prgsuspend(arg);
	break;
	
    case RMSIO_PRGRESUME:
	res = rms_prgresume(arg);
	break;

    case RMSIO_ELANINITDONE:
	res = rms_elaninitdone(arg);
	break;

    case RMSIO_PRGELANPIDS:
    {
	RMSIO_PRGELANPIDS_STRUCT args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_prgelanpids(args.id, args.maxpids, args.vps, args.pids, args.npids);
	break;
    }

    case RMSIO_SETELANSTATS:
    {
	RMSIO_SETELANSTATS_STRUCT args;
	elanstats_t estats;

	if (copy_from_user(&args, (void *)arg, sizeof(args)) ||
	    copy_from_user(&estats, (void *)args.estats, sizeof(estats)))
	    return(-EFAULT);
	
	res = rms_setelanstats(args.id, estats.ebytes, estats.exfers);
	break;
    }

    case RMSIO_MODVERSION:
    {
	RMSIO_MODVERSION_STRUCT args;
	int version = rms_modversion();
	
	if (copy_from_user (&args, (void *)arg, sizeof (args)))
	    return (-EFAULT);
	
	if (copyout(&version, args.version, sizeof(int)))
	    res = EFAULT;
	else
	    res = ESUCCESS;

	break;
    }

    /* 
     * Patch free kernel support, proc entries added manually
     */
    case RMSIO_ADDPROC:
    {
	RMSIO_PROC_STRUCT args;
	
	if (copy_from_user (&args, (void *)arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_addproc(args.id, args.pid);
	break;
    }
    case RMSIO_REMOVEPROC:
    {
	RMSIO_PROC_STRUCT args;
	
	if (copy_from_user (&args, (void *)arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_removeproc(args.id, args.pid);
	break;
    }

    default:
	res = EINVAL;
	break;
    }
    
    /* printk ("rmsmod: ioctl %x res %d\n", cmd, res); */
    
    return ((res == 0) ? 0 : -res);
}

#if defined(CONFIG_PPC64) || defined(CONFIG_SPARC64) || defined(CONFIG_X86_64)
static int
rms_ioctl32 (unsigned int fd, unsigned int cmd, unsigned long arg, struct file *file)
{
    int res;

    switch (cmd)
    {
    case RMSIO_GETPRGID32:
    {
	RMSIO_GETPRGID_STRUCT32 args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_getprgid(args.pid, (int *)(unsigned long) args.idptr);
	break;
    }
	
    case RMSIO_GETCAP32:
    {
	RMSIO_GETCAP_STRUCT32 args;

	if (copy_from_user (&args, (void *) arg, sizeof (args)))
	    return (-EFAULT);

	res = rms_getcap(args.index, (ELAN_CAPABILITY *)(unsigned long) args.capptr);
	break;
    }

    default:
	return (sys_ioctl (fd, cmd, arg));
    }

    return ((res == 0) ? 0 : -res);
}
#endif

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
