/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __QSNET_BITMAP_H
#define __QSNET_BITMAP_H

#ident "$Id: bitmap.h,v 1.5 2004/01/20 17:32:15 david Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/bitmap.h,v $ */

typedef unsigned int			bitmap_t;

#define BT_NBIPUL			32			/* n bits per bitmap_t */
#define BT_ULSHIFT			5			/* log 2 BT_NBIPUL to extract word index */
#define BT_ULMASK			0x1f			/* to extract bit index */

#define BT_WIM(bitmap,bitindex)		((bitmap)[(bitindex) >> BT_ULSHIFT])		/* word in map */
#define BT_BIW(bitindex)		(1 << ((bitindex) & BT_ULMASK))		/* bit in word */

/* BT_BITOUL -- n bits to n words */
#define BT_BITOUL(nbits)		(((nbits) + BT_NBIPUL -1) / BT_NBIPUL)

#define BT_TEST(bitmap,bitindex)	((BT_WIM((bitmap), (bitindex)) & BT_BIW(bitindex)) ? 1 : 0)
#define BT_SET(bitmap,bitindex)		do { BT_WIM((bitmap), (bitindex)) |= BT_BIW(bitindex); } while (0)
#define BT_CLEAR(bitmap,bitindex)	do { BT_WIM((bitmap), (bitindex)) &= ~BT_BIW(bitindex); } while (0)

/* return first free bit in the bitmap, or -1 for failure */
extern int  bt_freebit (bitmap_t *bitmap, int nbits);

/* return the index of the lowest set bit in the bitmap or -1 for failure */
extern int bt_lowbit (bitmap_t *bitmap, int nbits);

/* return the index of the next set/clear bit in the bitmap or -1 for failure */
extern int bt_nextbit (bitmap_t *bitmap, int nbits, int last, int isset);

/* copy/zero/fill/compare a bit map */
extern void bt_copy (bitmap_t *a, bitmap_t *b, int nbits);
extern void bt_zero (bitmap_t *a, int nbits);
extern void bt_fill (bitmap_t *a, int nbits);
extern int  bt_cmp (bitmap_t *a, bitmap_t *b, int nbits);

/* intersect bitmap 'a' with bitmap 'b' and return in 'a' */
extern void bt_intersect (bitmap_t *a, bitmap_t *b, int nbits);

/* remove/add bitmap 'b' from bitmap 'a' */
extern void bt_remove (bitmap_t *a, bitmap_t *b, int nbits);
extern void bt_add (bitmap_t *a, bitmap_t *b, int nbits);

/* check whether bitmap 'a' spans bitmap 'b' */
extern int  bt_spans (bitmap_t *a, bitmap_t *b, int nbits);

/* copy [base,base+nbits-1] from 'a' to 'b' */
extern void bt_subset (bitmap_t *a, bitmap_t *b, int base, int nbits);

/* find bits clear in 'a' and set in 'b', put result in 'c' */
extern void bt_up (bitmap_t *a, bitmap_t *b, bitmap_t *c, int nbits);

/* find bits set in 'a' and clear in 'b', put result in 'c' */
extern void bt_down (bitmap_t *a, bitmap_t *b, bitmap_t *c, int nbits);

/* return number of bits set in bitmap */
extern int  bt_nbits (bitmap_t *a, int nbits);


#endif /* __QSNET_BITMAP_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
