/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: capability.h,v 1.18 2005/05/17 09:52:53 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/capability.h,v $*/

#ifndef __ELAN_CAPABILITY_H
#define __ELAN_CAPABILITY_H

#include <elan/bitmap.h>

/* Maximum number of rails */
#define ELAN_MAX_RAILS          (31)
/* Maximum number of virtual processes we support */
#define ELAN_MAX_VPS		(16384)

/* Number of words in a bitmap capability */
#define ELAN_BITMAPSIZE		BT_BITOUL(ELAN_MAX_VPS)

/* Guaranteed invalid values */
#define ELAN_INVALID_PROCESS	(0x7fffffff)		/* A GUARANTEED invalid process # */
#define ELAN_INVALID_NODE	(0xFFFF)
#define ELAN_INVALID_CONTEXT	(0xFFFF)

/* Number of values in a user key */
#define ELAN_USERKEY_ENTRIES	4

typedef void * ELAN_CAP_OWNER;

/* 
 * When used in userspace this is relative to the base of
 * the capabality but is an absolute location for kernel space.
 */
typedef struct elan_location
{
	unsigned short loc_node;
	unsigned short loc_context;
} ELAN_LOCATION;

typedef struct elan_userkey
{
	unsigned	key_values[ELAN_USERKEY_ENTRIES];
} ELAN_USERKEY;

typedef struct elan_capability
{
	ELAN_USERKEY	cap_userkey;				/* User defined protection */

	int		cap_version;				/* Version number */
	unsigned short	cap_type;				/* Capability Type */
	unsigned short  cap_spare;				/* spare was cap_elan_type */

	int		cap_lowcontext;				/* low context number in block */
	int		cap_highcontext;			/* high context number in block */
	int		cap_mycontext;				/* my context number */
    
	int		cap_lownode;				/* low elan id of group */
	int		cap_highnode;				/* high elan id of group */

	unsigned int	cap_railmask;				/* which rails this capability is valid for */
	
	bitmap_t	cap_bitmap[ELAN_BITMAPSIZE];		/* Bitmap of process to processor translation */
} ELAN_CAPABILITY;

#define ELAN_CAP_UNINITIALISED		(-1)

#define ELAN_CAP_VERSION_NUMBER		(0x00010002)

#define ELAN_CAP_NUM_NODES(cap)		((cap)->cap_highnode - (cap)->cap_lownode + 1)
#define ELAN_CAP_NUM_CONTEXTS(cap)	((cap)->cap_highcontext - (cap)->cap_lowcontext + 1)

/* using or defining our own MIN/MAX had confilicts with dunix so we define ELAN_ ones */
#define ELAN_MIN(a,b)	((a) > (b) ? (b) : (a))
#define ELAN_MAX(a,b)	((a) > (b) ? (a) : (b))
#define ELAN_CAP_BITMAPSIZE(cap)	(ELAN_MAX (ELAN_MIN (ELAN_CAP_NUM_NODES(cap) * ELAN_CAP_NUM_CONTEXTS(cap), ELAN_MAX_VPS), 0))

#define ELAN_CAP_SIZE(cap)		(offsetof (ELAN_CAPABILITY, cap_bitmap[BT_BITOUL(ELAN_CAP_BITMAPSIZE(cap))]))
#define ELAN_CAP_ENTRIES(cap)           (((cap)->cap_type & ELAN_CAP_TYPE_NO_BITMAP) ? ELAN_CAP_BITMAPSIZE((cap)) : bt_nbits((cap)->cap_bitmap, ELAN_CAP_BITMAPSIZE((cap))))

#define ELAN_CAP_IS_RAIL_SET(cap,rail)  ((cap)->cap_railmask & (1<<rail))

#define ELAN_CAP_KEY_MATCH(cap1,cap2)	((cap1)->cap_userkey.key_values[0] == (cap2)->cap_userkey.key_values[0] && \
					 (cap1)->cap_userkey.key_values[1] == (cap2)->cap_userkey.key_values[1] && \
					 (cap1)->cap_userkey.key_values[2] == (cap2)->cap_userkey.key_values[2] && \
					 (cap1)->cap_userkey.key_values[3] == (cap2)->cap_userkey.key_values[3])

#define ELAN_CAP_TYPE_MATCH(cap1,cap2)  ((cap1)->cap_version           == (cap2)->cap_version           && \
					 (cap1)->cap_type              == (cap2)->cap_type)

#define ELAN_CAP_GEOM_MATCH(cap1,cap2)	((cap1)->cap_lowcontext        == (cap2)->cap_lowcontext        && \
					 (cap1)->cap_highcontext       == (cap2)->cap_highcontext       && \
					 (cap1)->cap_lownode           == (cap2)->cap_lownode           && \
					 (cap1)->cap_highnode          == (cap2)->cap_highnode		&& \
                                         (cap1)->cap_railmask          == (cap2)->cap_railmask          && \
					 !bcmp (&(cap1)->cap_bitmap[0], &(cap2)->cap_bitmap[0], 	   \
					        BT_BITOUL(ELAN_CAP_BITMAPSIZE(cap1)*sizeof(bitmap_t))))

#define ELAN_CAP_MATCH(cap1,cap2)	(ELAN_CAP_KEY_MATCH (cap1, cap2)  && \
					 ELAN_CAP_TYPE_MATCH (cap1, cap2) && \
					 ELAN_CAP_GEOM_MATCH (cap1, cap2))

#define ELAN_CAP_VALID_MYCONTEXT(cap)   (    ((cap)->cap_lowcontext  != ELAN_CAP_UNINITIALISED)     \
					  && ((cap)->cap_mycontext   != ELAN_CAP_UNINITIALISED)     \
					  && ((cap)->cap_highcontext != ELAN_CAP_UNINITIALISED)     \
					  && ((cap)->cap_lowcontext <= (cap)->cap_mycontext)        \
					  && ((cap)->cap_mycontext <= (cap)->cap_highcontext)) 

/*
 * Definitions for type 
 */
#define ELAN_CAP_TYPE_BLOCK		1		/* Block distribution */
#define ELAN_CAP_TYPE_CYCLIC		2		/* Cyclic distribution */
#define ELAN_CAP_TYPE_KERNEL		3		/* Kernel capability */

#define ELAN_CAP_TYPE_MASK		(0xFFF)		/* Mask for type */

/* OR these bits in for extra features */
#define ELAN_CAP_TYPE_HWTEST		(1 << 12)	/* Hardware test capability type */
#define ELAN_CAP_TYPE_MULTI_RAIL	(1 << 13)	/* "new" multi rail capability */
#define ELAN_CAP_TYPE_NO_BITMAP		(1 << 14)	/* don't use bit map */
#define ELAN_CAP_TYPE_BROADCASTABLE	(1 << 15)	/* broadcastable */


extern void          elan_nullcap     (ELAN_CAPABILITY *cap);
extern char         *elan_capability_string (ELAN_CAPABILITY *cap, char *str);
extern ELAN_LOCATION elan_vp2location (unsigned process, ELAN_CAPABILITY *cap);
extern int           elan_location2vp (ELAN_LOCATION location, ELAN_CAPABILITY *cap);
extern int           elan_nvps        (ELAN_CAPABILITY *cap);
extern int           elan_nlocal      (int node, ELAN_CAPABILITY *cap);
extern int           elan_maxlocal    (ELAN_CAPABILITY *cap);
extern int           elan_localvps    (int node, ELAN_CAPABILITY *cap, int *vps, int size);
extern int           elan_nrails      (ELAN_CAPABILITY *cap);
extern int           elan_rails       (ELAN_CAPABILITY *cap, int *rails);
extern int           elan_cap_overlap (ELAN_CAPABILITY *cap1, ELAN_CAPABILITY *cap2);

/*
 * capability creation/access fns provide for running
 * new libelan code on old OS releases
 */
extern int elan_lowcontext(ELAN_CAPABILITY *cap);
extern int elan_mycontext(ELAN_CAPABILITY *cap);
extern int elan_highcontext(ELAN_CAPABILITY *cap);
extern int elan_lownode(ELAN_CAPABILITY *cap);
extern int elan_highnode(ELAN_CAPABILITY *cap);
extern int elan_captype(ELAN_CAPABILITY *cap);
extern int elan_railmask(ELAN_CAPABILITY *cap);

extern int elan_getenvCap (ELAN_CAPABILITY *cap, int index);
extern ELAN_CAPABILITY *elan_createCapability(void);
extern ELAN_CAPABILITY *elan_copyCapability(ELAN_CAPABILITY *from, int ctxShift);
extern int elan_generateCapability(char *string);
extern int elan_getMachinesCap (char *filename, ELAN_CAPABILITY *cap);

typedef struct elan_cap_struct
{
	ELAN_CAP_OWNER   owner;
	ELAN_CAPABILITY  cap;

	int              attached; /* count of people attached */
	unsigned int     active;   /* ie not being destroyed   */
} ELAN_CAP_STRUCT;

#if ! defined(__KERNEL__)
extern void          elan_get_random_key(ELAN_USERKEY *key);
extern int           elan_prefrails(ELAN_CAPABILITY *cap, int *pref, int nvp);
#endif

#if defined(__KERNEL__)
/* capability.c */
extern int elan_validate_cap  (ELAN_CAPABILITY *cap);
extern int elan_validate_map  (ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map);

extern int elan_create_cap  (ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap);
extern int elan_destroy_cap (ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap);
extern int elan_create_vp   (ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map);
extern int elan_destroy_vp  (ELAN_CAP_OWNER owner, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map);

typedef	void (*ELAN_DESTROY_CB)(void *args, ELAN_CAPABILITY *cap, ELAN_CAPABILITY *map);

extern int elan_attach_cap  (ELAN_CAPABILITY *cap, unsigned int rail, void *args, ELAN_DESTROY_CB callback);
extern int elan_detach_cap  (ELAN_CAPABILITY *cap, unsigned int rail);

extern int elan_get_caps    (uint *number_of_results, uint array_size, ELAN_CAP_STRUCT *caps);
extern int elan_cap_dump    (void);
#endif /* __KERNEL__ */


#endif /* __ELAN_CAPABILITY_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
