/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: devinfo.h,v 1.16 2005/02/01 12:35:53 david Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/devinfo.h,v $*/

#ifndef __ELAN_DEVINFO_H
#define __ELAN_DEVINFO_H

#define ELAN_MAX_LEVELS			8			/* maximum number of levels in switch network */

typedef struct elan_position
{
	unsigned	pos_mode;				/* mode we're operating in */
	unsigned	pos_nodeid;				/* port this device connected to */
	unsigned	pos_levels;				/* number of levels to top switch */
	unsigned	pos_nodes;				/* number of nodes in the machine */
	unsigned	pos_random_disabled;			/* levels at which "random" routing is not possible */
	unsigned char	pos_arity[ELAN_MAX_LEVELS];		/* number of downlinks per switch level */
} ELAN_POSITION;

#define ELAN4_PARAM_PCI_PADDING_FLAGS		0		/* A bit field, representing good places to burst across the pci                      */
#define ELAN4_PARAM_EVENT_COPY_WIN		1		/* The num of cmds when it becomes quicker to send via event copy than write directly */
#define ELAN4_PARAM_WRITE_COMBINING		2		/* If set the device supports bursts accesses across the pci bus                      */
#define ELAN4_PARAM_DRIVER_FEATURES		11		/* device driver features */
#define ELAN4_PARAM_COUNT			12

/* values for ELAN4_PARAM_DRIVER_FEATURES, dev_features */
#define ELAN4_FEATURE_PCI_MAP		(1 << 0)			/* must use pci mapping functions */
#define ELAN4_FEATURE_64BIT_READ	(1 << 1)			/* must perform 64 bit PIO reads */
#define ELAN4_FEATURE_PIN_DOWN		(1 << 2)			/* must pin down pages */
#define ELAN4_FEATURE_NO_WRITE_COMBINE	(1 << 3)			/* don't allow write combinig at all */
#define ELAN4_FEATURE_NO_IOPROC		(1 << 4)			/* unpatched kernel or disabled by procfs */
#define ELAN4_FEATURE_NO_IOPROC_UPDATE	(1 << 5)			/* don't do coproc update xlation loading */
#define ELAN4_FEATURE_NO_PAGEFAULT	(1 << 6)			/* don't do pagefaulting */
#define ELAN4_FEATURE_NO_PREFETCH	(1 << 7)			/* don't allow prefetching of elan sdram/cports */

typedef struct elan_params
{
	unsigned        values[ELAN4_PARAM_COUNT];
} ELAN_PARAMS;

/* values for pos_mode */
#define ELAN_POS_UNKNOWN		0			/* network position unknown */
#define ELAN_POS_MODE_SWITCHED		1			/* connected to a switch */
#define ELAN_POS_MODE_LOOPBACK		2			/* loopback connector */
#define ELAN_POS_MODE_BACKTOBACK	3			/* cabled back-to-back to another node */

typedef struct elan_devinfo
{
	unsigned short	dev_vendor_id;				/* pci vendor id */
	unsigned short	dev_device_id;				/* pci device id */
	unsigned char	dev_revision_id;			/* pci revision id */
	unsigned char   dev_instance;				/* device instance number */
	unsigned char   dev_rail;				/* device rail number */

	unsigned short  dev_driver_version;			/* device driver version */
	unsigned short  dev_params_mask;                        /* mask for valid entries in dev_params array */
	ELAN_PARAMS     dev_params;				/* device parametization */

	unsigned        dev_num_down_links_value;               /* hint to machine size */
} ELAN_DEVINFO;

#define PCI_VENDOR_ID_QUADRICS		0x14fc
#define PCI_DEVICE_ID_ELAN3		0x0000
#define   PCI_REVISION_ID_ELAN3_REVA	0x0000
#define   PCI_REVISION_ID_ELAN3_REVB	0x0001
#define PCI_DEVICE_ID_ELAN4		0x0001
#define   PCI_REVISION_ID_ELAN4_REVA	0x0000
#define   PCI_REVISION_ID_ELAN4_REVB	0x0001

#if defined(__KERNEL__)
/* devinfo.c */
#include <elan/capability.h>
#include <elan/device.h>
extern int elan_get_devinfo  (ELAN_DEV_IDX devidx, ELAN_DEVINFO  *devinfo);
extern int elan_get_position (ELAN_DEV_IDX devidx, ELAN_POSITION *position);
extern int elan_set_position (ELAN_DEV_IDX devidx, unsigned short nodeId, unsigned short numNodes);
#endif /* __KERNEL__ */


#endif /* __ELAN_DEVINFO_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
