/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: elanmod.h,v 1.13 2005/05/26 16:14:21 addy Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/elanmod.h,v $*/

#ifndef __ELAN_MOD_H
#define __ELAN_MOD_H

#include <elan/devinfo.h>
#include <elan/device.h>
#include <elan/capability.h>
#include <elan/stats.h>

#if defined(__KERNEL__)

#include <elan/elanmoddebug.h>

/* Linux RW semaphores */
#include <asm/semaphore.h>
#include <linux/rwsem.h>

#define ELANMOD_RWLOCK			struct rw_semaphore
#define ELANMOD_RWLOCK_INIT(l)		init_rwsem(l)
#define ELANMOD_RWLOCK_DESTROY(l)	
#define ELANMOD_RWLOCK_READ(l)		down_read(l)
#define ELANMOD_RWLOCK_WRITE(l)		down_write(l)
#define ELANMOD_RWLOCK_READ_UNLOCK(l)	up_read(l)
#define ELANMOD_RWLOCK_WRITE_UNLOCK(l)	up_write(l)

extern ELANMOD_RWLOCK elan_rwlock;

/* elan_general.c */
extern int elan_init(void);
extern int elan_fini(void);

/* return codes, -ve => errno, +ve => success */
#define ELAN_CAP_OK  (0)
#define ELAN_CAP_RMS (1)

#define ELAN_USER_ATTACH    (1)
#define ELAN_USER_DETACH    (2)
#define ELAN_USER_P2P       (3)
#define ELAN_USER_BROADCAST (4)

extern int elanmod_classify_cap (ELAN_POSITION *position, ELAN_CAPABILITY *cap, unsigned use);

#define ELAN_USER_BASE_CONTEXT_NUM 	0x000			/* first user allowable context */
#define ELAN_USER_TOP_CONTEXT_NUM	0x7FF			/* last user allowable context */

#define ELAN_RMS_BASE_CONTEXT_NUM	0x400			/* reserved for RMS allocation */
#define ELAN_RMS_TOP_CONTEXT_NUM	0x7FF

#define ELAN_USER_CONTEXT(ctx)     	((ctx) >= ELAN_USER_BASE_CONTEXT_NUM && \
					 (ctx) <= ELAN_USER_TOP_CONTEXT_NUM)    

#define ELAN_RMS_CONTEXT(ctx)          ((ctx) >= ELAN_RMS_BASE_CONTEXT_NUM && \
					 (ctx) <= ELAN_RMS_TOP_CONTEXT_NUM)    


/* capability.c */
struct elan_cap_node_struct;
extern int elan_usercopy_attach (ELAN_CAPABILITY *cap, struct elan_cap_node_struct **node_ptr, void *handle, void *owner);
extern int elan_usercopy_detach (struct elan_cap_node_struct *cap_ptr, void *owner);
extern int elan_usercopy_handle (struct elan_cap_node_struct *cap_ptr, int ctxId, void **handlep);

/* usercopy.c */
extern int elan_usercopy (void *remote, void *local, size_t len, int write,
			  int ctxId, struct elan_cap_node_struct *cap_ptr);

#endif /* __KERNEL__ */

#endif /* __ELAN_MOD_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
