/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN_SUBSYSTEM_H
#define __ELAN_SUBSYSTEM_H

#include <sys/types.h>
#include <sys/param.h>

#if defined( __KERNEL__) 
int elan_configure(
    cfg_op_t op,
    caddr_t  indata,
    ulong    indata_size,
    caddr_t  outdata,
    ulong    outdata_size);
#endif

#define ELAN_KMOD_CODE(x) 	((x)+CFG_OP_SUBSYS_MIN)
#define ELAN_MAX_KMOD_CODES 100

#define ELAN_SUBSYS "elan"

#define ELAN_STATS_GET_NEXT	0x01
typedef struct {
	ELAN_STATS_IDX statidx;
	ELAN_STATS_IDX *next_statidx;	
} elan_stats_get_next_struct;


#define ELAN_STATS_FIND_INDEX   0x02
typedef struct {
	caddr_t          block_name;
	ELAN_STATS_IDX  *statidx; /* return value */
	uint        *num_entries; /* return value */
} elan_stats_find_index_struct;

#define ELAN_STATS_GET_BLOCK_INFO  0x03
typedef struct {
	ELAN_STATS_IDX  statidx; 
	caddr_t       block_name; /* return value */
	uint        *num_entries; /* return value */
} elan_stats_get_block_info_struct;

#define ELAN_STATS_GET_INDEX_NAME  0x04
typedef struct {
	ELAN_STATS_IDX statidx; 
	uint           index;
	caddr_t        name; /* return value */
} elan_stats_get_index_name_struct;

#define ELAN_STATS_CLEAR_BLOCK  0x05
typedef struct {
	ELAN_STATS_IDX statidx; 
} elan_stats_clear_block_struct;

#define ELAN_STATS_GET_BLOCK     0x06
typedef struct 
{
	ELAN_STATS_IDX statidx; 
	uint           entries;  
	ulong         *values; /* return values */
} elan_stats_get_block_struct;

#define ELAN_GET_DEVINFO     0x07
typedef struct 
{
	ELAN_DEV_IDX  devidx; 
	ELAN_DEVINFO *devinfo; /* return values */
} elan_get_devinfo_struct;

#define ELAN_GET_POSITION  0x08
typedef struct {
	ELAN_DEV_IDX   devidx; 
	ELAN_POSITION *position; /* return values */
} elan_get_position_struct;

#define ELAN_SET_POSITION   0x09
typedef struct {
	ELAN_DEV_IDX   devidx; 
	unsigned short nodeId;
	unsigned short numNodes;
} elan_set_position_struct;

#define ELAN_CREATE_CAP  0x0a
typedef struct {
	ELAN_CAPABILITY cap;
} elan_create_cap_struct;

#define ELAN_DESTROY_CAP    0x0b
typedef struct {
	ELAN_CAPABILITY cap;
} elan_destroy_cap_struct;

#define ELAN_CREATE_VP   0x0c
typedef struct {
	ELAN_CAPABILITY cap;
	ELAN_CAPABILITY map;
} elan_create_vp_struct;

#define ELAN_DESTROY_VP    0x0d
typedef struct {
	ELAN_CAPABILITY cap;
	ELAN_CAPABILITY map;
} elan_destroy_vp_struct;


#define ELAN_DEBUG_DUMP   0x0e

#define ELAN_GET_CAPS    0x0f
typedef struct {
	uint            *number_of_results;
	uint             array_size;
	ELAN_CAP_STRUCT *caps;
} elan_get_caps_struct;

#define ELAN_DEBUG_BUFFER 0x10
typedef struct {
	caddr_t addr;
	int     len;
} elan_debug_buffer_struct;

#define ELANMOD_PROCFS_IOCTL      "/proc/qsnet/elan/ioctl"
#define ELANMOD_PROCFS_VERSION    "/proc/qsnet/elan/version"
#define ELANMOD_PROCFS_DEBUG_MASK "/proc/qsnet/elan/debug_mask"
#define ELANMOD_PROCFS_DEBUG_MODE "/proc/qsnet/elan/debug_mode"

#endif /* __ELAN_SUBSYSTEM_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
