/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN_EPCOMMS_H
#define __ELAN_EPCOMMS_H

#ident "$Id: epcomms.h,v 1.46 2004/11/12 10:55:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/epcomms.h,v $ */

#include <elan/kcomm.h>
#include <elan/bitmap.h>

#define EPCOMMS_SUBSYS_NAME	"epcomms"

/* message service numbers */
#define EP_MSG_SVC_EIP512		0x00				/* Quadrics EIP services */
#define EP_MSG_SVC_EIP1K		0x01
#define EP_MSG_SVC_EIP2K		0x02
#define EP_MSG_SVC_EIP4K		0x03
#define EP_MSG_SVC_EIP8K		0x04
#define EP_MSG_SVC_EIP16K		0x05
#define EP_MSG_SVC_EIP32K		0x06
#define EP_MSG_SVC_EIP64K		0x07
#define EP_MSG_SVC_EIP128K		0x08

#define EP_MSG_SVC_PFS			0x09				/* Quadrics PFS rpc service */

#define EP_MSG_SVC_PORTALS_SMALL	0x10				/* Lustre Portals */
#define EP_MSG_SVC_PORTALS_LARGE	0x11

#define EP_MSG_NSVC			0x40				/* Max number of services */

#define EP_MSGQ_ADDR(qnum)		(EP_EPCOMMS_QUEUE_BASE + (qnum) * EP_QUEUE_DESC_SIZE)

/*
 * EP_ENVELOPE
 *   Messages are sent by sending an envelope to the destination
 *   describing the source buffers to transfer.  The receiving thread
 *   then allocates a receive buffer and fetches the data by issuing
 *   "get" dmas.
 *
 * NOTE:  envelopes are not explicitly converted to network byte order
 *        since they are always transferred little endian as they are
 *        copied to/from elan memory using word operations.
 */
typedef struct ep_envelope
{
    uint32_t	  Version;					/* Protocol version field */

    EP_ATTRIBUTE  Attr; 					/* Attributes */

    EP_XID	  Xid;						/* transaction id */

    uint32_t	  NodeId;					/* Source processor */
    uint32_t	  Range;					/* range we're sending to (high << 16 | low) */

    EP_ADDR	  TxdRail;					/* address of per-rail txd */
    EP_NMD	  TxdMain;					/* address of main memory portion of txd */

    uint32_t      nFrags;					/* # fragments */
    EP_NMD	  Frags[EP_MAXFRAG];				/* network mapping handles of source data */

    uint32_t      CheckSum;                                     /* holds the check sum value when active 
								 * must be after all members to be checksum'd
								 */

    uint32_t	  Pad[6];					/* Pad to 128 bytes */
} EP_ENVELOPE;

#define EP_ENVELOPE_VERSION		0xdac10001
#define EP_ENVELOPE_SIZE		roundup (sizeof (EP_ENVELOPE), EP_BLK_SIZE)

/*
 * RPC payload - this small amount of data is transfered in
 * the envelope for RPCs
 */
typedef struct ep_payload
{
    uint32_t	Data[128/sizeof(uint32_t)];
} EP_PAYLOAD;

#define EP_PAYLOAD_SIZE			roundup (sizeof (EP_PAYLOAD), EP_BLK_SIZE)

#define EP_INPUTQ_SIZE			(EP_ENVELOPE_SIZE + EP_PAYLOAD_SIZE)

/*
 * EP_STATUSBLK
 *   RPC completion transfers a status block to the client.
 */
typedef struct ep_statusblk
{
    uint32_t	Data[128/sizeof(uint32_t)];
} EP_STATUSBLK;

#define EP_STATUSBLK_SIZE		roundup (sizeof(EP_STATUSBLK), EP_BLK_SIZE)

#define EP_RANGE(low,high)		((high) << 16 | (low))
#define EP_RANGE_LOW(range)		((range) & 0xFFFF)
#define EP_RANGE_HIGH(range)		(((range) >> 16) & 0xFFFF)

/* return codes from functions, + 'res' parameter to txd callback, ep_rxd_status() */
typedef enum
{
    EP_SUCCESS		= 0,					/* message sent/received successfully */
    EP_RXD_PENDING	= -1,					/* rxd not completed by thread */
    EP_CONN_RESET	= -2,					/* virtual circuit reset */
    EP_NODE_DOWN	= -3,					/* node down - transmit not attempted */
    EP_MSG_TOO_BIG      = -4,					/* received message larger than buffer */
    EP_ENOMEM		= -5,					/* memory alloc failed */
    EP_EINVAL		= -6,					/* invalid parameters */
    EP_SHUTDOWN		= -7,					/* receiver is being shut down */
} EP_STATUS;

/* forward declarations */
typedef struct ep_rxd 		EP_RXD;
typedef struct ep_txd 		EP_TXD;
typedef struct ep_rcvr_rail	EP_RCVR_RAIL;
typedef struct ep_rcvr		EP_RCVR;
typedef struct ep_xmtr_rail	EP_XMTR_RAIL;
typedef struct ep_xmtr		EP_XMTR;
typedef struct ep_comms_rail    EP_COMMS_RAIL;
typedef struct ep_comms_subsys  EP_COMMS_SUBSYS;

typedef struct ep_rcvr_stats		EP_RCVR_STATS;
typedef struct ep_xmtr_stats		EP_XMTR_STATS;
typedef struct ep_rcvr_rail_stats	EP_RCVR_RAIL_STATS;
typedef struct ep_xmtr_rail_stats	EP_XMTR_RAIL_STATS;

typedef void (EP_RXH)(EP_RXD *rxd);				/* callback function from receive completion */
typedef void (EP_TXH)(EP_TXD *txd, void *arg, EP_STATUS res);	/* callback function from transmit completion  */

/* Main memory portion shared descriptor */
typedef struct ep_rxd_main
{
    EP_ENVELOPE		Envelope;				/* 128 byte aligned envelope */
    EP_PAYLOAD		Payload;				/* 128 byte aligned payload */
    bitmap_t	        Bitmap[BT_BITOUL(EP_MAX_NODES)];	/* broadcast bitmap */
    EP_STATUSBLK	StatusBlk;				/* RPC status block to return */
    uint64_t		Next;					/* linked list when on active list (main address) */
    int32_t		Len;					/* Length of message received */
} EP_RXD_MAIN;

#define EP_RXD_MAIN_SIZE	roundup (sizeof (EP_RXD_MAIN), EP_BLK_SIZE)

/* Phases for message/rpc */
#ifndef __ELAN__

/* Kernel memory portion of per-rail receive descriptor */
typedef struct ep_rxd_rail
{
    struct list_head    Link;					/* linked on freelist */
    EP_RCVR_RAIL       *RcvrRail;				/* rvcr we're associated with */
    
    EP_RXD	       *Rxd;					/* receive descriptor we're bound to */
} EP_RXD_RAIL;

#define RXD_BOUND2RAIL(rxdRail,rcvrRail)	((rxdRail) != NULL && ((EP_RXD_RAIL *) (rxdRail))->RcvrRail == (EP_RCVR_RAIL *) rcvrRail)

struct ep_rxd
{
    struct list_head	Link;					/* linked on free/active list */
    EP_RCVR	       *Rcvr;					/* owning receiver */

    EP_RXD_MAIN	       *RxdMain;				/* shared main memory portion. */
    EP_NMD	        NmdMain;				/*  and network mapping descriptor */

    EP_RXD_RAIL	       *RxdRail;				/* per-rail rxd we're bound to */
    
    EP_RXH	       *Handler;				/* completion function */
    void	       *Arg;					/*    and arguement */

    unsigned int	State;					/* RXD status (active,stalled,failed) */

    EP_NMD		Data;					/* network mapping descriptor for user buffer */

    int			nFrags;					/* network mapping descriptor for put/get/complete */
    EP_NMD		Local[EP_MAXFRAG];
    EP_NMD		Remote[EP_MAXFRAG];

    long		NextRunTime;				/* time to resend failover/map requests */
    EP_XID		MsgXid;					/*   and transaction id */

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    struct list_head	CheckSumLink; 			        /* linked on check sum list */
#endif
};

#define EP_NUM_RXD_PER_BLOCK	16

/* rxd->State */
#define EP_RXD_FREE		0

#define EP_RXD_RECEIVE_UNBOUND	1
#define EP_RXD_RECEIVE_ACTIVE	2

#define EP_RXD_PUT_ACTIVE	3
#define EP_RXD_PUT_STALLED	4
#define EP_RXD_GET_ACTIVE	5
#define EP_RXD_GET_STALLED	6

#define EP_RXD_COMPLETE_ACTIVE	7
#define EP_RXD_COMPLETE_STALLED	8

#define EP_RXD_RPC_IN_PROGRESS	9
#define EP_RXD_COMPLETED	10	

#define EP_RXD_BEEN_ABORTED	11				/* rxd was aborted while in a private state */

typedef struct ep_rxd_block
{
    struct list_head	Link;

    EP_NMD	        NmdMain;

    EP_RXD	        Rxd[EP_NUM_RXD_PER_BLOCK];
} EP_RXD_BLOCK;

struct ep_rcvr_rail_stats 
{
    EP_STATS_COUNT rx;
    EP_STATS_COUNT rx_len;
};

struct ep_rcvr_rail
{
    EP_RCVR	       *Rcvr;					/* associated receiver */
    EP_COMMS_RAIL      *CommsRail;				/* comms rail */

    struct proc_dir_entry *procfs_root;                         /* root of this rcvr_rail's procfs entry */
    EP_RCVR_RAIL_STATS     stats;                               /* generic rcvr_rail stats */
};

struct ep_rcvr_stats
{
    EP_STATS_COUNT rx;
    EP_STATS_COUNT rx_len;
};

struct ep_rcvr
{
    struct list_head  Link;					/* queued on subsystem */
    EP_COMMS_SUBSYS  *Subsys;					/* kernel comms subsystem */
    EP_SERVICE        Service;  				/* service number */

    unsigned int      InputQueueEntries;			/* # entries on receive queue */

    EP_RAILMASK	      RailMask;					/* bitmap of which rails are available */
    EP_RCVR_RAIL     *Rails[EP_MAX_RAILS];

    spinlock_t	      Lock;					/* spinlock for rails/receive lists */

    struct list_head  ActiveDescList;				/* List of pending/active receive descriptors */

    EP_XID_CACHE      XidCache;					/* XID cache (protected by Lock) */

    struct list_head  FreeDescList;				/* List of free receive descriptors */
    unsigned int      FreeDescCount;				/*   and number on free list */
    unsigned int      TotalDescCount;                           /*   total number created */
    spinlock_t	      FreeDescLock;				/*   and lock for free list */
    kcondvar_t	      FreeDescSleep;				/*   with place to sleep for rx desc */
    int		      FreeDescWanted;				/*   and flag */
    struct list_head  DescBlockList;

    unsigned int      ForwardRxdCount;				/* count of rxd's being forwarded */
    unsigned int      CleanupWaiting;				/* waiting for cleanup */
    kcondvar_t	      CleanupSleep;				/*   and place to sleep */

    struct proc_dir_entry *procfs_root;                         /* place where this rcvr's proc entry is */
    EP_RCVR_STATS          stats;                                    
};

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
#define EP_ENVELOPE_CHECK_SUM      (1<<31)
extern uint32_t ep_calc_check_sum (EP_SYS *sys, EP_ENVELOPE *env, EP_NMD *nmd, int nFrags);
#endif

#endif /* ! __ELAN__ */

typedef struct ep_txd_main
{
    EP_STATUSBLK      StatusBlk;				/* RPC status block */
    bitmap_t          Bitmap[BT_BITOUL(EP_MAX_NODES)];		/* broadcast bitmap */
} EP_TXD_MAIN;

#define EP_TXD_MAIN_SIZE	roundup (sizeof (EP_TXD_MAIN), EP_BLK_SIZE)

#ifndef __ELAN__
typedef struct ep_txd_rail
{
    struct list_head  Link;					/* linked on freelist */
    EP_XMTR_RAIL     *XmtrRail;					/* xmtr we're associated with */

    EP_TXD	     *Txd;					/* txd we're bound to */
} EP_TXD_RAIL;

#define TXD_BOUND2RAIL(rxdRail,xmtrRail)	((txdRail) != NULL && ((EP_TXD_RAIL *) (txdRail))->XmtrRail == (EP_XMTR_RAIL *) xmtrRail)

struct ep_txd
{
    struct list_head  Link;					/* linked on free/active list */
    EP_XMTR	     *Xmtr;					/* service we're associated with */

    EP_TXD_MAIN	     *TxdMain;					/* shared main memory portion */
    EP_NMD	      NmdMain;					/*   and network mapping descriptor */

    EP_TXD_RAIL      *TxdRail;					/* per-rail txd for this phase */

    EP_TXH	     *Handler;					/* completion function */
    void	     *Arg;					/*    and arguement */
    
    unsigned short    NodeId;					/* node transmit is to. */
    EP_SERVICE        Service;					/*    and seervice */

    long              TimeStamp;                                 /* time we where created at, to find sends taking too long */
    long	      RetryTime;
    EP_BACKOFF	      Backoff;

    EP_ENVELOPE	      Envelope;					/* envelope for transmit */
    EP_PAYLOAD	      Payload;					/* payload for transmit */
};

#define EP_NUM_TXD_PER_BLOCK	16

/* "phase" parameter to BindTxd */
#define EP_TXD_PHASE_ACTIVE		1
#define EP_TXD_PHASE_PASSIVE		2

typedef struct ep_txd_block
{
    struct list_head	Link;
    EP_NMD		NmdMain;
    EP_TXD	        Txd[EP_NUM_TXD_PER_BLOCK];		/* transmit descriptors */
} EP_TXD_BLOCK;

struct ep_xmtr_rail_stats
{
    EP_STATS_COUNT tx;
    EP_STATS_COUNT tx_len;
};

struct ep_xmtr_rail
{
    EP_COMMS_RAIL      *CommsRail;				/* associated comms rail */
    EP_XMTR	       *Xmtr;					/* associated transmitter */

    struct proc_dir_entry *procfs_root;                         /* place where this xmtr's proc entry is */

    EP_XMTR_RAIL_STATS     stats;
};

struct ep_xmtr_stats
{
    EP_STATS_COUNT tx;
    EP_STATS_COUNT tx_len;
};

struct ep_xmtr
{
    struct list_head  Link;					/* Linked on subsys */
    EP_COMMS_SUBSYS  *Subsys;					/* kernel comms subsystem */

    EP_RAILMASK	      RailMask;					/* bitmap of which rails are available */
    EP_XMTR_RAIL     *Rails[EP_MAX_RAILS];			/* per-rail state */

    spinlock_t	      Lock;					/* lock for active descriptor list */

    struct list_head  ActiveDescList;				/* list of active transmit descriptors */

    EP_XID_CACHE      XidCache;					/* XID cache (protected by Lock) */

    struct list_head  FreeDescList;				/* List of free receive descriptors */
    unsigned int      FreeDescCount;				/*   and number on free list */
    unsigned int      TotalDescCount;
    spinlock_t	      FreeDescLock;				/*   and lock for free list */
    kcondvar_t	      FreeDescSleep;				/*   with place to sleep for rx desc */
    int		      FreeDescWanted;				/*   and flag */
    struct list_head  DescBlockList;

    struct proc_dir_entry *procfs_root;                         /* place where this rcvr's proc entry is */
    EP_XMTR_STATS          stats;   
};

/* forward descriptor */
#define EP_TREE_ARITY		3

typedef struct ep_fwd_desc
{
    struct list_head    Link;					/* linked on forward/free lists */
    EP_RXD	       *Rxd;					/* rxd to forward */
    EP_NMD		Data;					/* nmd of subset of receive buffer */
    unsigned		NumChildren;				/*   number of places we're forwarding */
    unsigned		Children[EP_TREE_ARITY];
} EP_FWD_DESC;

typedef struct ep_comms_ops
{
    void	     (*DelRail) (EP_COMMS_RAIL *rail);
    void	     (*DisplayRail) (EP_COMMS_RAIL *rail);

    struct {
	void	     (*AddRail) (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);
	void	     (*DelRail) (EP_RCVR *rcvr, EP_COMMS_RAIL *rail);

	long	     (*Check) (EP_RCVR_RAIL *rcvrRail, long nextRunTime);

	int	     (*QueueRxd) (EP_RXD *rxd, EP_RCVR_RAIL *rcvrRail);
	void	     (*RpcPut)(EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
	void	     (*RpcGet)(EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);
	void	     (*RpcComplete)(EP_RXD *rxd, EP_NMD *local, EP_NMD *remote, unsigned nFrags);

	EP_RXD      *(*StealRxd)(EP_RCVR_RAIL *rcvrRail);

	void	     (*DisplayRcvr) (DisplayInfo *di, EP_RCVR_RAIL *rcvrRail);
	void	     (*DisplayRxd)  (DisplayInfo *di, EP_RXD_RAIL *rxdRail);

	void         (*FillOutRailStats) (EP_RCVR_RAIL *rcvr_rail, char *str);

    } Rcvr;

    struct {
	void         (*AddRail) (EP_XMTR *xmtr, EP_COMMS_RAIL *rail);
	void         (*DelRail) (EP_XMTR *xmtr, EP_COMMS_RAIL *rail);

	long	     (*Check) (EP_XMTR_RAIL *xmtrRail, long nextRunTime);
	
	int	     (*BindTxd) (EP_TXD *txd, EP_XMTR_RAIL *xmtrRail, unsigned int phase);
	void	     (*UnbindTxd) (EP_TXD *txd, unsigned int phase);
	int	     (*PollTxd) (EP_XMTR_RAIL *xmtrRail, EP_TXD_RAIL *txdRail, int how);

	void	     (*DisplayXmtr) (DisplayInfo *di, EP_XMTR_RAIL *xmtrRail);
	void	     (*DisplayTxd)  (DisplayInfo *di, EP_TXD_RAIL *txdRail);

	int          (*CheckTxdState) (EP_TXD *txd);

	void         (*FillOutRailStats) (EP_XMTR_RAIL *xmtr_rail, char *str);

    } Xmtr;
} EP_COMMS_OPS;

#define EP_RAIL_OP(commsRail, Which)	(commsRail)->Ops.Which
#define EP_RCVR_OP(rcvrRail, Which)	(rcvrRail)->CommsRail->Ops.Rcvr.Which
#define EP_XMTR_OP(xmtrRail, Which)	(xmtrRail)->CommsRail->Ops.Xmtr.Which

/* "how" parameter to PollTxd */
#define POLL_TX_LIST		0
#define ENABLE_TX_CALLBACK	1
#define DISABLE_TX_CALLBACK	2

struct ep_comms_rail
{
    struct list_head	Link;					/* Linked on subsys */
    EP_RAIL	       *Rail;					/* kernel comms rail */
    EP_COMMS_SUBSYS    *Subsys;
    EP_COMMS_OPS        Ops;

    EP_COMMS_RAIL_STATS Stats;					/* statistics */
};

struct ep_comms_subsys
{
    EP_SUBSYS		Subsys;					/* is a kernel comms subsystem */

    kmutex_t		Lock;					/* global lock */

    EP_COMMS_STATS	Stats;					/* statistics */

    struct list_head	Rails;					/* list of all rails */

    struct list_head    Receivers;				/* list of receivers */
    struct list_head	Transmitters;				/* and transmitters */

    /* forward/allocator thread */
    EP_KTHREAD		Thread;					/* place thread sleeps */

    /* message passing "broadcast" forward lists */
    spinlock_t	        ForwardDescLock;			/* Lock for broadcast forwarding */
    struct list_head    ForwardDescList;			/* List of rxd's to forward */

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    spinlock_t	        CheckSumDescLock;			/* Lock for CheckSums */
    struct list_head    CheckSumDescList;			/* List of rxd's to be CheckSumed */
#endif

    EP_XMTR	       *ForwardXmtr;				/* and transmitter to forward with */
};

/* epcomms.c subsystem initialisation */
extern unsigned int   epcomms_forward_limit;

extern int	      ep_comms_init (EP_SYS *sys);
extern void           ep_comms_display (EP_SYS *sys, char *how);
extern EP_RAILMASK    ep_rcvr_railmask (EP_SYS *epsys, EP_SERVICE service);

/* epcomms_elan3.c */
extern EP_COMMS_RAIL *ep3comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *rail);

/* epcomms_elan4.c */
extern EP_COMMS_RAIL *ep4comms_add_rail (EP_SUBSYS *s, EP_SYS *sys, EP_RAIL *rail);

/* epcommsTx.c */
extern int            TxdShouldStabalise (EP_TXD_RAIL *txdRail, EP_RAIL *rail);
extern void	      FreeTxd (EP_XMTR *xmtr, EP_TXD *txd);

extern unsigned int   ep_txd_lowat;
extern long           ep_check_xmtr (EP_XMTR *xmtr, long nextRunTime);
extern void           ep_display_xmtr (DisplayInfo *di, EP_XMTR *xmtr);
extern void           ep_xmtr_flush_callback (EP_XMTR *xmtr, EP_XMTR_RAIL *xmtrRail);
extern void           ep_xmtr_reloc_callback (EP_XMTR *xmtr, EP_XMTR_RAIL *xmtrRail);

extern void           ep_xmtr_fillout_stats      (EP_XMTR      *xmtr,      char *str);
extern void           ep_xmtr_rail_fillout_stats (EP_XMTR_RAIL *xmtr_rail, char *str);

extern void           ep_xmtr_txd_stat (EP_XMTR *xmtr, EP_TXD *txd);

/* epcommsRx.c */
extern EP_RXD        *StealRxdFromOtherRail (EP_RCVR *rcvr);

extern unsigned int   ep_rxd_lowat;
extern long	      ep_check_rcvr (EP_RCVR *rcvr, long nextRunTime);
extern void           ep_rcvr_flush_callback (EP_RCVR *rcvr, EP_RCVR_RAIL *rcvrRail);
extern void           ep_rcvr_reloc_callback (EP_RCVR *rcvr, EP_RCVR_RAIL *rcvrRail);
extern void           ep_display_rcvr (DisplayInfo *di, EP_RCVR *rcvr, int full);

extern long           ep_forward_rxds (EP_COMMS_SUBSYS *subsys, long nextRunTime);

extern void           ep_rcvr_fillout_stats      (EP_RCVR      *rcvr,      char *str);
extern void           ep_rcvr_rail_fillout_stats (EP_RCVR_RAIL *rcvr_rail, char *str);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
extern void           ep_csum_rxds    (EP_COMMS_SUBSYS *subsys);
extern void           ep_rxd_queue_csum (EP_RXD *rxd);
#endif

extern void           ep_rxd_received     (EP_RXD *rxd);
extern void           ep_rxd_received_now (EP_RXD *rxd);

/* ep_procfs.c */
extern struct proc_dir_entry *ep_procfs_root;

extern void ep_procfs_rcvr_xmtr_init(void);
extern void ep_procfs_rcvr_xmtr_fini(void);

extern void ep_procfs_rcvr_add(EP_RCVR *rcvr);
extern void ep_procfs_rcvr_del(EP_RCVR *rcvr);

extern void ep_procfs_rcvr_add_rail(EP_RCVR_RAIL *rcvrRail);
extern void ep_procfs_rcvr_del_rail(EP_RCVR_RAIL *rcvrRail);

extern void ep_procfs_xmtr_add(EP_XMTR *xmtr);
extern void ep_procfs_xmtr_del(EP_XMTR *xmtr);

extern void ep_procfs_xmtr_add_rail(EP_XMTR_RAIL *xmtrRail);
extern void ep_procfs_xmtr_del_rail(EP_XMTR_RAIL *xmtrRail);


/* Public Interface */


/* epcomms.c message xmtr functions */
extern EP_XMTR       *ep_alloc_xmtr (EP_SYS *sys);
extern void           ep_free_xmtr (EP_XMTR *xmtr);

extern EP_STATUS      ep_transmit_message (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr, 
					   EP_TXH *handler, void *arg, EP_PAYLOAD *payload,
					   EP_NMD *nmd, int nFrag);
extern EP_STATUS      ep_multicast_message (EP_XMTR *xmtr, unsigned int destLo, unsigned int destHi, bitmap_t *bitmap, 
					    EP_SERVICE service, EP_ATTRIBUTE attr, EP_TXH *handler, void *arg, 
					    EP_PAYLOAD *payload, EP_NMD *nmd, int nFrag);
extern EP_STATUS      ep_transmit_rpc (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr, 
				       EP_TXH *handler, void *arg, EP_PAYLOAD *payload,
				       EP_NMD *nmd, int nFrag);
extern EP_STATUS      ep_multicast_forward (EP_XMTR *xmtr, unsigned int dest, EP_SERVICE service, EP_ATTRIBUTE attr, 
					    EP_TXH *handler, void *arg, EP_ENVELOPE *env, EP_PAYLOAD *payload, 
					    bitmap_t *bitmap, EP_NMD *nmd, int nFrags);

/* epcomms.c functions for use with polled transmits */
extern int            ep_poll_transmits (EP_XMTR *xmtr);
extern int            ep_enable_txcallbacks (EP_XMTR *xmtr);
extern int            ep_disable_txcallbacks (EP_XMTR *xmtr);

/* epcomms.c message rcvr functions */
extern EP_RCVR       *ep_alloc_rcvr (EP_SYS *sys, EP_SERVICE svc, unsigned int nenvelopes);
extern void	      ep_free_rcvr (EP_RCVR *rcvr);

extern EP_STATUS      ep_queue_receive (EP_RCVR *rcvr, EP_RXH *handler, void *arg, EP_NMD *nmd, EP_ATTRIBUTE attr);
extern void	      ep_requeue_receive (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *nmd, EP_ATTRIBUTE attr);
extern EP_STATUS      ep_rpc_put (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *from, EP_NMD *to, int nFrags);
extern EP_STATUS      ep_rpc_get (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_NMD *from, EP_NMD *to, int nFrags);
extern EP_STATUS      ep_complete_rpc (EP_RXD *rxd, EP_RXH *handler, void *arg, EP_STATUSBLK *blk, 
				       EP_NMD *from, EP_NMD *to, int nFrags);
extern void	      ep_complete_receive (EP_RXD *rxd);

/* railhints.c */
extern int            ep_xmtr_bcastrail (EP_XMTR *xmtr, EP_RAILMASK allowedRails);
extern int            ep_xmtr_prefrail (EP_XMTR *xmtr, EP_RAILMASK allowedRails, unsigned nodeId);
extern EP_RAILMASK    ep_xmtr_availrails (EP_XMTR *xmtr);
extern EP_RAILMASK    ep_xmtr_noderails (EP_XMTR *xmtr, unsigned nodeId);
extern int            ep_rcvr_prefrail (EP_RCVR *rcvr, EP_RAILMASK allowedRails);
extern EP_RAILMASK    ep_rcvr_availrails (EP_RCVR *rcvr);
extern EP_RAILMASK    ep_rxd_railmask (EP_RXD *rxd);

/* epcomms.c functions for accessing fields of rxds */
extern void          *ep_rxd_arg(EP_RXD *rxd);
extern int            ep_rxd_len(EP_RXD *rxd);
extern EP_STATUS      ep_rxd_status(EP_RXD *rxd);
extern int            ep_rxd_isrpc(EP_RXD *rxd);
extern EP_ENVELOPE   *ep_rxd_envelope(EP_RXD *rxd);
extern EP_PAYLOAD    *ep_rxd_payload(EP_RXD *rxd);
extern int            ep_rxd_node(EP_RXD *rxd);
extern EP_STATUSBLK  *ep_rxd_statusblk(EP_RXD *rxd);

/* functions for accessing fields of txds */
extern int            ep_txd_node(EP_TXD *txd);
extern EP_STATUSBLK  *ep_txd_statusblk(EP_TXD *txd);

/* functions for controlling how many processes are using module */
extern void              ep_mod_dec_usecount (void);
extern void              ep_mod_inc_usecount (void);

extern EP_RAILMASK ep_xmtr_svc_indicator_railmask (EP_XMTR *xmtr, int svc_indicator, int nodeId);
extern int ep_xmtr_svc_indicator_bitmap (EP_XMTR *xmtr, int svc_indicator, bitmap_t * bitmap, int low, int nnodes);

#endif /* ! __ELAN__ */
/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN_EPCOMMS_H */

