/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_KALLOC_H
#define __ELAN3_KALLOC_H

#ident "$Id: kalloc.h,v 1.11 2004/05/19 10:23:59 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kalloc.h,v $ */

#include <elan/rmap.h>

/*
 * Memory allocator
 */
#define LN2_MIN_SIZE	6					/* 64 bytes */
#define LN2_MAX_SIZE	16					/* 64k bytes */
#define NUM_FREELISTS	(LN2_MAX_SIZE-LN2_MIN_SIZE + 1)
#define MIN_SIZE	(1 << LN2_MIN_SIZE)
#define MAX_SIZE	(1 << LN2_MAX_SIZE)

#define HASHSHIFT	LN2_MAX_SIZE
#define NHASH		32
#define HASH(addr)	(((addr) >> HASHSHIFT) & (NHASH-1))

typedef enum
{
    EP_ALLOC_TYPE_PRIVATE_SDRAM,
    EP_ALLOC_TYPE_PRIVATE_MAIN,
    EP_ALLOC_TYPE_SHARED_MAIN,
} EP_ALLOC_TYPE;

typedef struct ep_pool
{
    EP_NMH		  Handle;				/* network mapping handle */

    struct list_head	  HashBase;				/* linked on hash lists */
    struct list_head	  HashTop;				/* linked on hash lists */

    struct list_head	  Link[NUM_FREELISTS];			/* linked on free lists */
    bitmap_t		 *Bitmaps[NUM_FREELISTS];		/* bitmaps for each size */

    union {
	sdramaddr_t	Sdram;
	unsigned long   Ptr;
    } Buffer;
} EP_POOL;

typedef struct ep_alloc
{
    spinlock_t	     Lock;
    
    EP_ALLOC_TYPE    Type;
    unsigned int     Perm;

    EP_RMAP	     *ResourceMap;

    struct list_head HashBase[NHASH];
    struct list_head HashTop[NHASH];
    struct list_head Freelists[NUM_FREELISTS];

    union {
	struct {
	    EP_SYS	       *System;
	    struct list_head	Rails;
	} Shared;
	
	struct {
	    EP_RAIL	       *Rail;
	} Private;
    } Data;
} EP_ALLOC;

extern void		ep_display_alloc (EP_ALLOC *alloc);

extern void        	ep_alloc_init (EP_RAIL *rail);
extern void        	ep_alloc_fini (EP_RAIL *rail);

extern sdramaddr_t 	ep_alloc_memory_elan (EP_RAIL *rail, EP_ADDR addr, unsigned size, unsigned int perm, EP_ATTRIBUTE attr);
extern void        	ep_free_memory_elan (EP_RAIL *rail, EP_ADDR addr);

extern sdramaddr_t 	ep_alloc_elan (EP_RAIL *rail, unsigned size, EP_ATTRIBUTE attr, EP_ADDR *addrp);
extern void        	ep_free_elan (EP_RAIL *rail, EP_ADDR addr, unsigned size);
extern void	       *ep_alloc_main (EP_RAIL *rail, unsigned size, EP_ATTRIBUTE attr, EP_ADDR *addr);
extern void       	ep_free_main (EP_RAIL *rail, EP_ADDR addr, unsigned size);

extern sdramaddr_t 	ep_elan2sdram (EP_RAIL *rail, EP_ADDR addr);
extern void            *ep_elan2main (EP_RAIL *rail, EP_ADDR addr);

extern void		ep_shared_alloc_init (EP_SYS *sys);
extern void        	ep_shared_alloc_fini (EP_SYS *sys);
extern int	   	ep_shared_alloc_add_rail (EP_SYS *sys, EP_RAIL *rail);
extern void        	ep_shared_alloc_remove_rail (EP_SYS *sys, EP_RAIL *rail);

extern void	       *ep_shared_alloc_main (EP_SYS *sys, unsigned size, EP_ATTRIBUTE attr, EP_NMD *nmd);
extern void        	ep_shared_free_main (EP_SYS *sys, EP_NMD *nmd);

#endif /* __ELAN_KALLOC_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
