/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN_KCOMM_H
#define __ELAN_KCOMM_H

#ident "$Id: kcomm.h,v 1.82 2005/04/05 16:36:28 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm.h,v $*/
#define EP_KCOMM_MAJOR_VERSION		3
#define EP_KCOMM_MINOR_VERSION		1

#define EP_PROTOCOL_VERSION		1			/* CM/KCOMM protocol revision */

#define EP_MAX_NODES			2048			/* Max nodes we support */
#define EP_MAX_RAILS			16			/* max number of rails (we use an unsigned short for bitmaps !) */
#define EP_MAXFRAG			4			/* max number of fragments */

#define EP_BLK_SIZE			64			/* align objects for elan access */

/* Elan virtual address address space */
#define EP_SYSTEM_QUEUE_BASE		0x00010000		/* Base address for system queues */
#define EP_MSGSYS_QUEUE_BASE		0x00020000		/* Base address for msgsys queues */
#define EP_EPCOMMS_QUEUE_BASE		0x00030000		/* Base address for message queues */
#define EP_DVMA_BASE			0x10000000		/* elan address range for dvma mapping. */
#define EP_DVMA_TOP			0xE0000000

#define EP_SHARED_BASE			0xE0000000		/* shared main/elan allocators */
#define EP_SHARED_TOP			0xF0000000

#define EP_PRIVATE_BASE			0xF0000000		/* private main/elan allocators */
#define EP_PRIVATE_TOP			0xF8000000

#define EP_DVMA_RMAP_SIZE		1024			/* size of resource map for dvma address space */
#define EP_SHARED_RMAP_SIZE		1024			/* size of resource map for shared address space */
#define EP_PRIVATE_RMAP_SIZE		1024			/* size of resource map for private address space */

/* Input queue descriptors fit into 64 bytes */
#define EP_QUEUE_DESC_SIZE		64

/* Timeouts for checking network position */
#define EP_POSITION_TIMEOUT		(4*HZ)		/* 1s   time to notice CheckNetworkPosition changes */
#define EP_WITHDRAW_TIMEOUT		(2*HZ)          /* 2s   time before withdrawing from unreachable nodes */

/* Time to try again due to resource failue (eg malloc etc) */
#define RESOURCE_RETRY_TIME		(HZ/20)

/* Time to retransmit message when send failed */
#define MSGBUSY_RETRY_TIME		(HZ/20)

/* Time between retransmits of messages network flush requests */
#define MESSAGE_RETRY_TIME		(HZ/5)

/* time to hold the context filter up to ensure that the
 * next packet of a dma is guaranteed to get nacked (8mS) */
#define NETWORK_ERROR_TIMEOUT		(1 + roundup (HZ * 8 / 1000, 1))

/* Time between retransmits of message failover requests */
#define FAILOVER_RETRY_TIME		(HZ/5)

/* compute earliest time */
#define SET_NEXT_RUN_TIME(nextRunTime, time) \
do { \
    if ((nextRunTime) == 0 || AFTER(nextRunTime, (time)))\
	(nextRunTime) = (time);\
} while (0)

/* DMA retry backoff/priorities/issue rings */
#define EP_NUM_BACKOFF			8
#define EP_RETRY_STABALISING            0
#define EP_RETRY_BASE			1

#define EP_RETRY_CRITICAL		EP_RETRY_BASE
#define EP_RETRY_HIGH_PRI		(EP_RETRY_CRITICAL + 1)
#define EP_RETRY_HIGH_PRI_TIME		(1)
#define EP_RETRY_HIGH_PRI_RETRY		(EP_RETRY_HIGH_PRI + 1)
#define EP_RETRY_HIGH_PRI_RETRY_TIME	(2)
#define EP_RETRY_LOW_PRI		(EP_RETRY_HIGH_PRI_RETRY + EP_NUM_BACKOFF)
#define EP_RETRY_LOW_PRI_TIME		(2)
#define EP_RETRY_LOW_PRI_RETRY		(EP_RETRY_LOW_PRI + 1)
#define EP_RETRY_LOW_PRI_RETRY_TIME	(4)
#define EP_RETRY_ANONYMOUS		(EP_RETRY_LOW_PRI_RETRY + EP_NUM_BACKOFF)
#define EP_RETRY_ANONYMOUS_TIME		(10)
#define EP_RETRY_NETERR			(EP_RETRY_ANONYMOUS + EP_NUM_BACKOFF)
#define EP_RETRY_NETERR_TIME		(10)
#define EP_NUM_RETRIES			(EP_RETRY_NETERR + 1)

typedef unsigned short EP_SERVICE;

/* EP_ATTRIBUTE 32 bits 
 *
 * 0-2
 *   for initial call :-
 *     0 (0x1) EP_NO_ALLOC				used once
 *     1 (0x2) EP_NO_SLEEP				used once
 *     2 (0x4) EP_NOT_MYSELF				used once
 *
 *   when stored and transmited :-
 *     0 (0x0) EP_MULTICAST				envelope
 *     1 (0x2) EP_RPC					envelope
 *     2 (0x4) EP_HAS_PAYLOAD				envelope
 *
 * 3-11
 *     3   (0x08) EP_PREFRAIL_SET		        preserved
 *     4-7 (0xf0) Pref Rail
 *     8  (0x100) EP_NO_INTERUPT
 *     9  (0x200) EP_NO_FAILOVER
 *
 *    10 (0x400) EP_INTERRUPT_ENABLED			internal
 *    11 (0x800) EP_TXD_STABALISING			internal
 *
 * 12-13 Not Used.
 * 
 * 14-15 (0xC000) Data Type.                    	passed in
 *    00 none. 
 *    01 Service Indicator.
 *    10 TimeOut.
 *    11 RailMask
 *          
 * 16-31 (0x10000)  Data.  Service Indicator, TimeOut, RailMask, Pref Rail.
 *         
*/

typedef uint32_t EP_ATTRIBUTE;

#define EP_LOCAL_ATTR_MASK 0x07
#define EP_CLEAR_LOCAL_ATTR(ATTR)  ( (ATTR) & ~EP_LOCAL_ATTR_MASK )

#define EP_NO_ALLOC      0x01	/* Don't call allocators if no free descriptors */
#define EP_NO_SLEEP      0x02	/* Don't sleep if no free descriptors */
#define EP_NOT_MYSELF    0x04	/* Don't send multicast to me */

#define EP_MULTICAST         0x01	/* Message is a multicast */
#define EP_RPC               0x02	/* Wait for RPC reply */
#define EP_HAS_PAYLOAD_BIT   0x04	/* transfer payload */


#define EP_PREFRAIL_SET  0x08	/* preferred rail is set (otherwise pick one from the NMDs) */

#define EP_PREFRAIL_SHIFT  (4)
#define EP_PREFRAIL_MASK   0xf0
#define EP_IS_PREFRAIL_SET(ATTR)      (((ATTR) &  EP_PREFRAIL_SET) != 0)
#define EP_CLEAR_PREFRAIL(ATTR)       (((ATTR) & ~EP_PREFRAIL_SET) & ~EP_PREFRAIL_MASK) 
#define EP_SET_PREFRAIL(ATTR,RAIL)    (EP_CLEAR_PREFRAIL(ATTR) | (((RAIL) <<  EP_PREFRAIL_SHIFT ) &  EP_PREFRAIL_MASK ) |  EP_PREFRAIL_SET)


#define EP_ATTR2PREFRAIL(ATTR)            (((ATTR) & EP_PREFRAIL_MASK) >> EP_PREFRAIL_SHIFT)


#define EP_INTERRUPT_ENABLED 0x400	/* event interrupt enabled on EP_NO_INTERRUPT */
#define EP_TXD_STABALISING   0x800      /* flag to indicate this is attempting to stabalise */

#define EP_IS_MULTICAST(ATTR)                 (((ATTR) &  EP_MULTICAST) != 0)
#define EP_SET_MULTICAST(ATTR)                ( (ATTR) |  EP_MULTICAST)
#define EP_CLEAR_MULTICAST(ATTR)              ( (ATTR) & ~EP_MULTICAST)

#define EP_IS_RPC(ATTR)                       (((ATTR) &  EP_RPC) != 0)
#define EP_SET_RPC(ATTR)                      ( (ATTR) |  EP_RPC)
#define EP_CLEAR_RPC(ATTR)                    ( (ATTR) & ~EP_RPC)

#define EP_HAS_PAYLOAD(ATTR)                  (((ATTR) &  EP_HAS_PAYLOAD_BIT) != 0)
#define EP_SET_HAS_PAYLOAD(ATTR)              ( (ATTR) |  EP_HAS_PAYLOAD_BIT)
#define EP_CLEAR_HAS_PAYLOAD(ATTR)            ( (ATTR) & ~EP_HAS_PAYLOAD_BIT)

#define EP_IS_INTERRUPT_ENABLED(ATTR)         (((ATTR) &  EP_INTERRUPT_ENABLED) != 0)
#define EP_SET_INTERRUPT_ENABLED(ATTR)        ( (ATTR) |  EP_INTERRUPT_ENABLED)
#define EP_CLEAR_INTERRUPT_ENABLED(ATTR)      ( (ATTR) & ~EP_INTERRUPT_ENABLED)

#define EP_IS_TXD_STABALISING(ATTR)           (((ATTR) &  EP_TXD_STABALISING) != 0)
#define EP_SET_TXD_STABALISING(ATTR)          ( (ATTR) |  EP_TXD_STABALISING)
#define EP_CLEAR_TXD_STABALISING(ATTR)        ( (ATTR) & ~EP_TXD_STABALISING)

#define EP_NO_INTERRUPT      0x100	/* Don't generate completion interrupt (tx) */
#define EP_NO_FAILOVER       0x200	/* don't attempt rail failover, just abort */

#define EP_IS_NO_INTERRUPT(ATTR)    (((ATTR) &  EP_NO_INTERRUPT) != 0)
#define EP_SET_NO_INTERRUPT(ATTR)   ( (ATTR) |  EP_NO_INTERRUPT)
#define EP_CLEAR_NO_INTERRUPT(ATTR) ( (ATTR) & ~EP_NO_INTERRUPT)

#define EP_IS_NO_FAILOVER(ATTR)    (((ATTR) &  EP_NO_FAILOVER) != 0)
#define EP_SET_NO_FAILOVER(ATTR)   ( (ATTR) |  EP_NO_FAILOVER)
#define EP_CLEAR_NO_FAILOVER(ATTR) ( (ATTR) & ~EP_NO_FAILOVER)

#define EP_TYPE_MASK           0xC000
#define EP_TYPE_SVC_INDICATOR  0x4000
#define EP_TYPE_TIMEOUT        0x8000
#define EP_TYPE_RAILMASK       0xC000

#define EP_ATTR2TYPE(ATTR)        ( (ATTR) & EP_TYPE_MASK )

#define EP_IS_SVC_INDICATOR(ATTR) (EP_ATTR2TYPE(ATTR) == EP_TYPE_SVC_INDICATOR)
#define EP_IS_TIMEOUT(ATTR)       (EP_ATTR2TYPE(ATTR) == EP_TYPE_TIMEOUT)
#define EP_IS_RAILMASK(ATTR)      (EP_ATTR2TYPE(ATTR) == EP_TYPE_RAILMASK)
#define EP_IS_NO_TYPE(ATTR)       (EP_ATTR2TYPE(ATTR) == 0)

#define EP_DATA_SHIFT          (16)
#define EP_DATA_MASK           0xffff0000

#define EP_ATTR2DATA(ATTR)     (((ATTR) & EP_DATA_MASK) >> EP_DATA_SHIFT)
#define EP_DATA2ATTR(DATA)     (((DATA) <<  EP_DATA_SHIFT) & EP_DATA_MASK)

#define EP_CLEAR_DATA(ATTR)    (((ATTR) & ~EP_TYPE_MASK) & ~EP_DATA_MASK)
#define EP_SET_DATA(ATTR,TYPE,DATA)   (EP_CLEAR_DATA(ATTR) | ((TYPE) & EP_TYPE_MASK) | (((DATA) <<  EP_DATA_SHIFT) & EP_DATA_MASK))

#define EP_DEFAULT_TIMEOUT     (HZ*30)

#if !defined(offsetof)
#define offsetof(s, m)		(unsigned long)(&(((s *)0)->m))
#endif
#if !defined(roundup)
#define roundup(x, y) 		((((x)+((y)-1))/(y))*(y))
#endif

/* 
 * Message transaction ID's - these are unique 64 bts 
 * numbers which include the initial rail number.
 */
typedef struct ep_xid
{
    uint32_t	Generation;
    uint32_t	Handle;
    uint64_t	Unique;
} EP_XID;

#define EP_INVALIDATE_XID(xid)	((xid).Generation = (xid).Handle = (xid).Unique = 0)

#define EP_XID_INVALID(xid)	((xid).Generation == 0 && (xid).Handle == 0 && (xid).Unique == 0)
#define EP_XIDS_MATCH(a,b)	((a).Generation == (b).Generation && (a).Handle == (b).Handle && (a).Unique == (b).Unique)

typedef struct ep_backoff
{
    unsigned char	type;
    unsigned char	indx;
    unsigned short	count;
} EP_BACKOFF;

/* values for "type" */
#define EP_BACKOFF_FREE		0
#define EP_BACKOFF_ENVELOPE	1
#define EP_BACKOFF_FETCH	2
#define EP_BACKOFF_DATA		3
#define EP_BACKOFF_DONE		4
#define EP_BACKOFF_STABILISE	5

#ifndef __ELAN__

/* forward declaration of types */
typedef struct ep_rail	EP_RAIL;
typedef struct ep_sys	EP_SYS;

#include <elan/nmh.h>
#include <elan/kmap.h>
#include <elan/statemap.h>
#include <elan/kalloc.h>
#include <elan/kthread.h>
#include <elan/kcomm_stats.h>
#include <elan/devinfo.h>

typedef struct ep_callback
{
    struct ep_callback *Next;
    void              (*Routine)(void *, statemap_t *);
    void	       *Arg;
} EP_CALLBACK;

#define EP_CB_FLUSH_FILTERING		0
#define EP_CB_FLUSH_FLUSHING		1
#define EP_CB_PASSIVATED		2
#define EP_CB_FAILOVER			3
#define EP_CB_DISCONNECTING		4
#define EP_CB_DISCONNECTED		5
#define EP_CB_NODESET			6
#define EP_CB_COUNT			7

#endif /* !defined(__ELAN__) */

/* Small unreliable system message queues */
#define EP_SYSTEMQ_INTR			0			/* input queue for cluster membership generating an interrupt */
#define EP_SYSTEMQ_POLLED		1			/* input queue for cluster membership polled on clock tick */
#define EP_SYSTEMQ_MANAGER		2			/* input queue for manager messages */
#define EP_NUM_SYSTEMQ			64

#define EP_SYSTEMQ_ADDR(qnum)		(EP_SYSTEM_QUEUE_BASE + (qnum) * EP_QUEUE_DESC_SIZE)
#define EP_SYSTEMQ_DESC(base,qnum)	((base) + (qnum) * EP_QUEUE_DESC_SIZE)

#define EP_SYSTEMQ_MSG_ALIGN		64			/* message sizes aligned to 64 byte boundaries */
#define EP_SYSTEMQ_MSG_MAX		(4*64)			/* max message size */

/* Special flag for Version field to indicate message not
 * seen in main memory yet and time limit to poll for it */
#define EP_SYSTEMQ_UNRECEIVED			0xdeadbabe
#define EP_SYSTEMQ_UNRECEIVED_TLIMIT		16384			/* 1023 uS */

#ifndef __ELAN__

typedef void (EP_INPUTQ_HANDLER) (EP_RAIL *rail, void *arg, void *msg);
typedef void (EP_INPUTQ_CALLBACK) (EP_RAIL *rail, void *arg);

typedef struct ep_inputq
{
    unsigned long	q_hidden;				/* implementation hidden as ep3 or ep4 */
} EP_INPUTQ;

typedef struct ep_outputq
{
    unsigned long	q_hidden;				/* implementation hidden as ep3 or ep4 */
} EP_OUTPUTQ;

/* returned values for ep_outputq_state */
#define EP_OUTPUTQ_BUSY		0
#define EP_OUTPUTQ_FAILED	1
#define EP_OUTPUTQ_FINISHED	2

typedef struct ep_switch
{
    unsigned    present:1;
    unsigned	invalid:1;
    unsigned	link:3;
    unsigned	bcast:3;
    unsigned	lnr;
} EP_SWITCH;

/*
 * Network error fixup, flush, relocation messges
 */
typedef struct ep_map_nmd_body
{
    uint32_t		nFrags;
    EP_RAILMASK		Railmask;
    EP_NMD		Nmd[EP_MAXFRAG];
} EP_MAP_NMD_BODY;

typedef struct ep_failover_body
{
    EP_XID		Xid;
    EP_RAILMASK		Railmask;
} EP_FAILOVER_BODY;

typedef struct ep_failover_txd
{
    EP_XID		Xid;
    uint32_t		Rail;
    EP_ADDR		TxdRail;
} EP_FAILOVER_TXD;

typedef uint64_t EP_NETERR_COOKIE;

#define EP_PANIC_STRLEN		31

typedef struct ep_node_state
{
    unsigned char       State;
    unsigned char       NetworkErrorState;
    EP_RAILMASK         Railmask;
} EP_NODE_STATE;

#define EP_MANAGER_MSG_SIZE		(2 * EP_SYSTEMQ_MSG_ALIGN)

typedef struct ep_manager_msg_hdr
{
    EP_XID		Xid;					/* Message transaction id */

    uint16_t		NodeId;					/* Originating node number */
    uint16_t		DestId;					/* destination node id */

    uint16_t		Checksum;				/* Message checksum */
    uint8_t		Rail;					/* Rail message associated with */
    uint8_t		Type;					/* Message type */

    uint32_t		Pad;					/* pad to 32 bytes */

    uint32_t		Version;				/* Message Version */
} EP_MANAGER_MSG_HDR;

typedef union ep_manager_msg_body
{
    unsigned char       Space[EP_MANAGER_MSG_SIZE - sizeof (EP_MANAGER_MSG_HDR)];

    EP_NETERR_COOKIE	Cookies[2];				/* EP_MSG_TYPE_NETERR */
    EP_MAP_NMD_BODY	MapNmd;					/* EP_MSG_TYPE_MAP_NMD */
    EP_FAILOVER_BODY	Failover;				/* EP_MSG_TYPE_FAILOVER_REQUEST */
    EP_FAILOVER_TXD	FailoverTxd;				/* EP_MSG_TYPE_FAILOVER_RESPONSE */
    unsigned char       PanicReason[EP_PANIC_STRLEN+1];		/* EP_MSG_TYPE_REMOTE_PANIC */
    EP_NODE_STATE       NodeState;                              /* EP_MSG_TYPE_GET_NODE_STATE_RESPONSE */   
    EP_SERVICE          Service;                                /* EP_MSG_TYPE_GET_NODE_STATE */
} EP_MANAGER_MSG_BODY;

typedef struct ep_manager_msg
{
    EP_MANAGER_MSG_BODY Body;
    EP_MANAGER_MSG_HDR  Hdr;
} EP_MANAGER_MSG;

#define EP_MANAGER_MSG_VERSION				0xcad01000
#define EP_MANAGER_MSG_TYPE_REMOTE_PANIC		0x00
#define EP_MANAGER_MSG_TYPE_NETERR_REQUEST		0x01
#define EP_MANAGER_MSG_TYPE_NETERR_RESPONSE		0x02
#define EP_MANAGER_MSG_TYPE_FLUSH_REQUEST		0x03
#define EP_MANAGER_MSG_TYPE_FLUSH_RESPONSE		0x04
#define EP_MANAGER_MSG_TYPE_MAP_NMD_REQUEST		0x05
#define EP_MANAGER_MSG_TYPE_MAP_NMD_RESPONSE		0x06
#define EP_MANAGER_MSG_TYPE_FAILOVER_REQUEST		0x07
#define EP_MANAGER_MSG_TYPE_FAILOVER_RESPONSE		0x08
#define EP_MANAGER_MSG_TYPE_GET_NODE_STATE              0x09
#define EP_MANAGER_MSG_TYPE_GET_NODE_STATE_RESPONSE     0x0a

/* Message types which should only be sent when a rail is connected */
#define EP_MANAGER_MSG_TYPE_CONNECTED(type)		(((type) & 1) == 1)

#define EP_MANAGER_OUTPUTQ_SLOTS	128			/* # entries in outputq */
#define EP_MANAGER_INPUTQ_SLOTS		128			/* # entries in inputq */
#define EP_MANAGER_OUTPUTQ_RETRIES	31			/* # retries for manager messages */

/* XID's are allocated from a cache, which doesn't
 * require locking since it relies on the caller to
 * manage the locking for us.
 */
typedef struct ep_xid_cache
{
    struct list_head	Link;

    uint32_t		Handle;					/* my XID cache handle */
    uint64_t		Current;				/* range of XID.Unique we can allocate from */
    uint64_t		Last;

    void	      (*MessageHandler)(void *arg, EP_MANAGER_MSG *);
    void	       *Arg;
} EP_XID_CACHE;

#define EP_XID_CACHE_CHUNKS	(10000)

typedef struct ep_node_rail
{
    struct list_head    Link;					/* can be linked on work lists */

    unsigned char       State;					/* node connection state */
    unsigned char	NetworkErrorState;			/* reasons for keeping the context filter up */
    unsigned char	MessageState;				/* state of messages during passivate/relocate */

    EP_XID		MsgXid;					/* neterr/flush transaction id */
    long		NextRunTime;				/* time to drop context filter for destroyed dma packet, or to send next request */
    EP_NETERR_COOKIE	NetworkErrorCookies[2];			/* identify cookie for destroyed atomic packet */

    uint32_t		Cookie;					/* per-node network error cookie */
    spinlock_t		CookieLock;				/* and spinlock for it. */

    struct list_head    StalledDmas;				/* list of stalled DMAs */
} EP_NODE_RAIL;

#define EP_NODE_DISCONNECTED		0			/* node is disconnected */
#define EP_NODE_CONNECTING		1			/* awaiting connection */
#define EP_NODE_CONNECTED		2			/* node is connected */
#define EP_NODE_LEAVING_CONNECTED	3			/* node is starting to disconnect */
#define EP_NODE_LOCAL_PASSIVATE 	4			/* flushing context filter/run queues */
#define EP_NODE_REMOTE_PASSIVATE	5			/* stalling for neterr flush */
#define EP_NODE_PASSIVATED		6			/* relocating active/passive messages */
#define EP_NODE_DISCONNECTING		7			/* entering disconncted - abort remaining comms */
#define EP_NODE_NUM_STATES		8

#define EP_NODE_NETERR_ATOMIC_PACKET	(1 << 0)
#define EP_NODE_NETERR_DMA_PACKET	(1 << 1)

#define EP_NODE_PASSIVE_MESSAGES	(1 << 0)
#define EP_NODE_ACTIVE_MESSAGES		(1 << 1)

/*
 * Kernel thread code is loaded as a table.
 */
typedef struct ep_symbol
{
    char   *name;
    EP_ADDR value;
} EP_SYMBOL;

typedef struct ep_code
{
    u_char        *text;
    u_int	   text_size;
    u_char        *data;
    u_int	   data_size;
    u_char        *rodata;
    u_int	   rodata_size;
    EP_SYMBOL     *symbols;
    
    int		   ntext;
    sdramaddr_t    pptext;
    EP_ADDR	   etext;
    sdramaddr_t   _stext;
    sdramaddr_t	  _rodata;

    int		   ndata;
    sdramaddr_t    ppdata;
    EP_ADDR	   edata;
    sdramaddr_t   _sdata;
} EP_CODE;

typedef struct ep_switchstate
{
    unsigned char       linkid;
    unsigned char       LNR;
    unsigned char       bcast;
    unsigned char       uplink;
} EP_SWITCHSTATE;

typedef struct ep_rail_ops
{
    void	(*DestroyRail) (EP_RAIL *rail);

    int 	(*StartRail) (EP_RAIL *rail);
    void	(*StallRail) (EP_RAIL *rail);
    void	(*StopRail) (EP_RAIL *rail);

    sdramaddr_t (*SdramAlloc) (EP_RAIL *rail, EP_ADDR addr, unsigned size);
    void        (*SdramFree) (EP_RAIL *rail, sdramaddr_t addr, unsigned size);
    void        (*SdramWriteb) (EP_RAIL *rail, sdramaddr_t addr, unsigned char val);
    
    void	(*KaddrMap) (EP_RAIL *rail, EP_ADDR eaddr, virtaddr_t kaddr, unsigned len, unsigned int perm, int ep_attr);
    void	(*SdramMap) (EP_RAIL *rail, EP_ADDR eaddr, sdramaddr_t saddr, unsigned len, unsigned int perm, int ep_attr);
    void	(*Unmap) (EP_RAIL *rail, EP_ADDR eaddr, unsigned len);

    void       *(*DvmaReserve) (EP_RAIL *rail, EP_ADDR eaddr, unsigned npages);
    void	(*DvmaRelease) (EP_RAIL *rail, EP_ADDR eaddr, unsigned npages, void *private);
    void	(*DvmaSetPte) (EP_RAIL *rail, void *private, unsigned index, physaddr_t phys, unsigned int perm);
    physaddr_t	(*DvmaReadPte) (EP_RAIL *rail, void *private, unsigned index);
    void	(*DvmaUnload)(EP_RAIL *rail, void *private, unsigned index, unsigned npages);
    void	(*FlushTlb) (EP_RAIL *rail);

    int 	(*ProbeRoute) (EP_RAIL *r, int level, int sw, int nodeid, int *linkup, 
			       int *linkdown, int attempts, EP_SWITCH *lsw);
    void	(*PositionFound) (EP_RAIL *rail, ELAN_POSITION *pos);
    int		(*CheckPosition) (EP_RAIL *rail);
    void	(*NeterrFixup) (EP_RAIL *rail, unsigned int nodeId, EP_NETERR_COOKIE *cookies);

    void	(*LoadSystemRoute) (EP_RAIL *rail, unsigned int vp, unsigned int lowNode, unsigned int highNode);

    void	(*LoadNodeRoute) (EP_RAIL *rail, unsigned nodeId);
    void	(*UnloadNodeRoute) (EP_RAIL *rail, unsigned nodeId);
    void	(*LowerFilter) (EP_RAIL *rail, unsigned nodeId);
    void	(*RaiseFilter) (EP_RAIL *rail, unsigned nodeId);
    void	(*NodeDisconnected) (EP_RAIL *rail, unsigned nodeId);

    void	(*FlushFilters) (EP_RAIL *rail);
    void	(*FlushQueues) (EP_RAIL *rail);


    EP_INPUTQ  *(*AllocInputQ) (EP_RAIL *rail, unsigned qnum, unsigned slotSize, unsigned slotCount,
				void (*callback)(EP_RAIL *rail, void *arg), void *arg);
    void	(*FreeInputQ) (EP_RAIL *rail, EP_INPUTQ *q);
    void	(*EnableInputQ) (EP_RAIL *rail, EP_INPUTQ *q);
    void	(*DisableInputQ) (EP_RAIL *rail, EP_INPUTQ *q);
    int		(*PollInputQ) (EP_RAIL *rail, EP_INPUTQ *q, int maxCount, EP_INPUTQ_HANDLER *handler, void *arg);

    EP_OUTPUTQ *(*AllocOutputQ) (EP_RAIL *rail, unsigned slotSize, unsigned slotCount);
    void	(*FreeOutputQ) (EP_RAIL *rail, EP_OUTPUTQ *outputq);
    void       *(*OutputQMsg) (EP_RAIL *rail, EP_OUTPUTQ *outputq, unsigned slotNum);
    int         (*OutputQState) (EP_RAIL *rail, EP_OUTPUTQ *outputq, unsigned slotNum);
    int		(*OutputQSend) (EP_RAIL *rail, EP_OUTPUTQ *outputq, unsigned slotNum, unsigned size,
				unsigned vp, unsigned qnum, unsigned retries);

    void        (*FillOutStats) (EP_RAIL *rail, char *str);
    void	(*Debug) (EP_RAIL *rail);

} EP_RAIL_OPS;

#define ep_alloc_inputq(rail,qnum,slotSize,slotCount,callback,arg) \
	(rail)->Operations.AllocInputQ(rail,qnum,slotSize,slotCount,callback,arg)
#define ep_free_inputq(rail,inputq) \
	(rail)->Operations.FreeInputQ(rail,inputq)
#define ep_enable_inputq(rail,inputq) \
	(rail)->Operations.EnableInputQ(rail,inputq)
#define ep_disable_inputq(rail,inputq) \
	(rail)->Operations.DisableInputQ(rail,inputq)
#define ep_poll_inputq(rail,inputq,maxCount,handler,arg) \
	(rail)->Operations.PollInputQ(rail,inputq,maxCount,handler,arg)
#define ep_alloc_outputq(rail,slotSize,slotCount)\
	(rail)->Operations.AllocOutputQ(rail,slotSize,slotCount)
#define ep_free_outputq(rail,outputq)\
	(rail)->Operations.FreeOutputQ(rail,outputq)
#define ep_outputq_msg(rail,outputq,slotNum)\
	(rail)->Operations.OutputQMsg(rail,outputq,slotNum)
#define ep_outputq_state(rail,outputq,slotNum)\
	(rail)->Operations.OutputQState(rail,outputq,slotNum)
#define ep_outputq_send(rail,outputq,slotNum,size,vp,qnum,retries)\
	(rail)->Operations.OutputQSend(rail,outputq,slotNum,size,vp,qnum,retries)

struct ep_rail
{
    EP_SYS	       *System;					/* "system" we've attached to */

    unsigned char	Number;					/* Rail number */
    unsigned char       State;					/* Rail state */
    char		Name[32];				/* Rail name */

    struct list_head    ManagerLink;				/* linked on ManagedRails list */

    ELAN_DEVINFO	Devinfo;				/* Device information for this rail */
    ELAN_POSITION       Position;				/* Position on switch device is connected to */

    EP_RAIL_OPS	        Operations;				/* device specific operations */
    EP_RAIL_STATS	Stats;					/* statistics */

    EP_ALLOC            ElanAllocator;				/* per-rail elan memory allocator */
    EP_ALLOC            MainAllocator;				/* per-rail main memory allocator */

    unsigned		TlbFlushRequired;			/* lazy TLB flushing */

    int 	        SwitchBroadcastLevel;			/* current switch level ok for broadcast */
    unsigned long       SwitchBroadcastLevelTick;

    int			SwitchProbeLevel;			/* result of last switch probe */
    EP_SWITCHSTATE      SwitchState[ELAN_MAX_LEVELS];
    EP_SWITCHSTATE      SwitchLast[ELAN_MAX_LEVELS];
    unsigned long       SwitchProbeTick[ELAN_MAX_LEVELS];
    
    /* Node disconnecting/connecting state */
    EP_CALLBACK        *CallbackList[EP_CB_COUNT];		/* List of callbacks */
    kmutex_t	        CallbackLock;				/*   and lock for it. */
    unsigned	        CallbackStep;				/*  step through UpdateConnectionState. */

    /* back pointer for cluster membership */
    void	       *ClusterRail;

    /* Per node state for message passing */
    EP_NODE_RAIL       *Nodes;					/* array of per-node state */
    statemap_t         *NodeSet;				/* per-rail statemap of connected nodes */
    statemap_t	       *NodeChangeMap;				/* statemap of nodes to being connected/disconnected */
    statemap_t	       *NodeChangeTmp;				/*   and temporary copies */

    struct list_head    NetworkErrorList;			/* list of nodes resolving network errors */
    struct list_head    LocalPassivateList;			/* list of nodes in state LOCAL_PASSIVATE */
    struct list_head    RemotePassivateList;			/* list of nodes waiting for remote network error flush */
    struct list_head    PassivatedList;				/* list of nodes performing message relocation */
    struct list_head    DisconnectingList;			/* list of nodes transitioning to disconnected */

    EP_XID_CACHE	XidCache;				/* XID cache for node messages (single threaded access) */

    /* Manager messages */
    EP_INPUTQ	       *ManagerInputQ;
    EP_OUTPUTQ	       *ManagerOutputQ;
    unsigned		ManagerOutputQNextSlot;
    spinlock_t		ManagerOutputQLock;

    /* /proc entries */
    struct proc_dir_entry *ProcDir;
    struct proc_dir_entry *SvcIndicatorDir;
    int                    CallbackRegistered;
};

/* values for State */
#define EP_RAIL_STATE_UNINITIALISED	0			/* device uninitialised */
#define EP_RAIL_STATE_STARTED		1			/* device started but network position unknown */
#define EP_RAIL_STATE_RUNNING		2			/* device started and position known */
#define EP_RAIL_STATE_INCOMPATIBLE	3			/* device started, but position incompatible */

typedef struct ep_rail_entry
{
    struct list_head	Link;
    EP_RAIL	       *Rail;
} EP_RAIL_ENTRY;

typedef struct ep_subsys
{
    EP_SYS	       *Sys;

    struct list_head	Link;					/* Linked on sys->Subsystems */
    char	       *Name;					/* Name to lookup */
    
    void	       (*Destroy)    (struct ep_subsys *subsys, EP_SYS *sys);

    int		       (*AddRail)    (struct ep_subsys *subsys, EP_SYS *sys, EP_RAIL *rail);
    void	       (*RemoveRail) (struct ep_subsys *subsys, EP_SYS *sys, EP_RAIL *rail);
} EP_SUBSYS;

typedef struct ep_node
{
    EP_RAILMASK		ConnectedRails;
} EP_NODE;

struct ep_sys
{
    EP_RAIL         *Rails[EP_MAX_RAILS];			/* array of all available devices */

    kmutex_t	     StartStopLock;				/* lock for starting stopping rails */

    ELAN_POSITION    Position;					/* primary node position */

    EP_NMH_TABLE     MappingTable;				/* Network mapping handle table */

    EP_ALLOC	     Allocator;					/* shared main memory allocator */

    EP_DVMA_STATE    DvmaState;					/* dvma state */

    kmutex_t	     SubsysLock;                               /* lock on the Subsytems list */
    struct list_head Subsystems;				/* list of subsystems */

    /* device manager state */
    struct list_head ManagedRails;				/* list of managed devices */
    EP_KTHREAD       ManagerThread;				/* place for manager thread to sleep */

    /* global node state */
    spinlock_t	     NodeLock;					/* spinlock for node state (including per-device node state) */
    EP_NODE	    *Nodes;					/* system wide node state */
    statemap_t      *NodeSet;					/* system wide nodeset */
    struct list_head NodesetCallbackList;			/* list of "callbacks" */

    /* Transaction Id */
    struct list_head XidCacheList;				/* list of XID caches */
    uint32_t 	     XidGeneration;				/* XID generation number (distinguishes reboots) */
    uint32_t	     XidHandle;					/* XID handles (distinguishes XID caches) */
    uint64_t	     XidNext;					/* next XID to prime cache */
    spinlock_t	     XidLock;					/*   and it's spinlock  */

    /* Shutdown/Panic */
    unsigned int     Shutdown;					/* node has shutdown/panic'd */
};

#if defined(DEBUG_ASSERT)
extern int ep_assfail (EP_RAIL *rail, const char *string, const char *func, const char *file, const int line);
extern int sdram_assert;
extern int assfail_mode;

#define EP_ASSERT(rail, EX)	do { \
    if (!(EX) && ep_assfail ((EP_RAIL *) (rail), #EX, __FUNCTION__, __FILE__, __LINE__)) { \
	BUG(); \
    } \
} while (0)
#define EP_ASSFAIL(rail,EX)	do { \
   if (ep_assfail ((EP_RAIL *) (rail), EX, __FUNCTION__, __FILE__, __LINE__)) { \
	BUG(); \
    } \
} while (0)
#define SDRAM_ASSERT(EX)	(sdram_assert ? (EX) : 1)
#else
#define EP_ASSERT(rail, EX)	((void) 0)
#define EP_ASSFAIL(rail,str)	((void) 0)
#define SDRAM_ASSERT(EX)	(1)
#endif

/* conf_osdep.c */
extern EP_SYS    *ep_system(void);
extern void       ep_mod_dec_usecount (void);
extern void       ep_mod_inc_usecount (void);

/* procfs_osdep.c */
extern struct proc_dir_entry *ep_procfs_root;
extern struct proc_dir_entry *ep_config_root;

/* kcomm.c */
extern int        ep_sys_init (EP_SYS *sys);
extern void       ep_sys_fini (EP_SYS *sys);
extern void	  ep_shutdown (EP_SYS *sys);
extern int        ep_init_rail (EP_SYS *sys, EP_RAIL *rail);
extern void       ep_destroy_rail (EP_RAIL *rail);
extern int        ep_start_rail (EP_RAIL *rail);
extern void       ep_stop_rail (EP_RAIL *rail);

extern void       ep_connect_node (EP_RAIL *rail, int nodeId);
extern int        ep_disconnect_node (EP_RAIL *rail, int nodeId);

extern EP_XID     ep_xid_cache_alloc (EP_SYS *sys, EP_XID_CACHE *cache);
extern void       ep_xid_cache_init (EP_SYS *sys, EP_XID_CACHE *cache);
extern void       ep_xid_cache_destroy (EP_SYS *sys, EP_XID_CACHE *cache);

extern int        ep_send_message (EP_RAIL *rail, int nodeId, int type, EP_XID xid, EP_MANAGER_MSG_BODY *body);

extern void       ep_panic_node (EP_SYS *sys, int nodeId, unsigned char *reason);

extern void	  ep_subsys_add (EP_SYS *sys, EP_SUBSYS *subsys);
extern void	  ep_subsys_del (EP_SYS *sys, EP_SUBSYS *subsys);
extern EP_SUBSYS *ep_subsys_find (EP_SYS *sys, char *name);

extern void       DisplayNodes (EP_RAIL *rail);

extern void       ep_fillout_stats(EP_RAIL *rail, char *str);

/* neterr.c */
extern void       ep_queue_network_error (EP_RAIL *rail, int nodeId, int what, int channel, EP_NETERR_COOKIE cookie);

/* kcomm_elan3.c */
extern unsigned int ep3_create_rails (EP_SYS *sys, unsigned int disabled);

/* kcomm_elan4.c */
extern unsigned int ep4_create_rails (EP_SYS *sys, unsigned int disabled);

/* probenetwork.c */
extern int	  ProbeNetwork (EP_RAIL *rail, ELAN_POSITION *pos);
extern void	  CheckPosition (EP_RAIL *rail);

extern uint16_t   CheckSum (char *msg, int nob);

/* threadcode.c */
extern EP_ADDR    ep_symbol (EP_CODE *code, char *name);
extern int        ep_loadcode (EP_RAIL *rail, EP_CODE *code);
extern void       ep_unloadcode (EP_RAIL *rail, EP_CODE *code);

/* Public interface */
/* debug.c */
extern int              ep_sprintf_bitmap (char *str, unsigned nbytes, bitmap_t *bitmap, int base, int count, int off);
extern void             ep_display_bitmap (char *prefix, char *tag, bitmap_t *bitmap, unsigned base, unsigned nbits);

/* epcomms.c */
extern int              ep_waitfor_nodeid (EP_SYS *sys);
extern int              ep_nodeid (EP_SYS *sys);
extern int              ep_numnodes (EP_SYS *sys);

/* railhints.c */
extern int              ep_pickRail(EP_RAILMASK railmask);

/* support.c */
extern int              ep_register_nodeset_callback (EP_SYS *sys, void (*routine)(void *, statemap_t *), void *arg);
extern void             ep_remove_nodeset_callback (EP_SYS *sys, void (*routine)(void *, statemap_t *), void *arg);
extern void             ep_call_nodeset_callbacks (EP_SYS *sys, statemap_t *map);

extern int              ep_register_callback (EP_RAIL *rail, unsigned idx, void (*routine)(void *, statemap_t *), void *arg);
extern void             ep_remove_callback (EP_RAIL *rail, unsigned idx, void (*routine)(void *, statemap_t *), void *arg);
extern void             ep_call_callbacks (EP_RAIL *rail, unsigned idx, statemap_t *);
extern unsigned int     ep_backoff (EP_BACKOFF *backoff, int type);

#endif /* !__ELAN__ */

#endif /* __ELAN_KCOMM_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
