/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __EP_EPSTATS_H
#define __EP_EPSTATS_H

#ident "$Id: kcomm_stats.h,v 1.5.2.1 2005/05/31 11:42:43 mike Exp $"
/*      $Source: /cvs/master/quadrics/epmod/kcomm_stats.h,v $ */

#define EP_BUCKET_SLOTS			8

#define BucketStat(obj,stat,size)	((size) < 128   ? (obj)->Stats.stat[0]++ : \
					 (size) < 512   ? (obj)->Stats.stat[1]++ : \
					 (size) < 1024  ? (obj)->Stats.stat[2]++ : \
					 (size) < 8192  ? (obj)->Stats.stat[3]++ : \
					 (size) < 16384 ? (obj)->Stats.stat[4]++ : \
					 (size) < 32768 ? (obj)->Stats.stat[5]++ : \
					 (size) < 65536 ? (obj)->Stats.stat[6]++ : \
					 (obj)->Stats.stat[7]++)
#define IncrStat(obj,stat)		((obj)->Stats.stat++)


#define EP3_NUM_DMA_FAIL		11	/* NOTE - the same as EP_NUM_RETRIES */

#define ADD_STAT(STATS,STAT,VALUE) { unsigned long now = lbolt;\
   STATS.STAT.total += VALUE; \
   if  ( (long)( now - STATS.STAT.last_time) > HZ ){ \
 STATS.STAT.last_per_sec = ( STATS.STAT.total - STATS.STAT.last_count)/ ( ( (long)( now - STATS.STAT.last_time) + (HZ/2)) / HZ);\
 STATS.STAT.last_time = now; \
 STATS.STAT.last_count = STATS.STAT.total; \
   }} \

#define INC_STAT(STATS,STAT) ADD_STAT(STATS,STAT,1)

#define GET_STAT_PER_SEC(STATS, STAT) (  (( lbolt - STATS.STAT.last_time ) < (HZ * 5)) ? STATS.STAT.last_per_sec : 0 )
#define GET_STAT_TOTAL(STATS, STAT) ( STATS.STAT.total )

struct ep_stats_count 
{
    unsigned long total;
    unsigned long last_time;
    unsigned long last_count;
    unsigned long last_per_sec;
};

typedef struct ep_stats_count  	EP_STATS_COUNT;

typedef struct ep3_rail_stats
{
    unsigned long	IssueDmaFail[EP3_NUM_DMA_FAIL];

    unsigned long	DmaQueueLength[EP_BUCKET_SLOTS];
    unsigned long	CprocDmaQueueOverflow;
    unsigned long	DprocDmaQueueOverflow;
    unsigned long	IprocDmaQueueOverflow;
    unsigned long	CprocEventQueueOverflow;
    unsigned long	DprocEventQueueOverflow;
    unsigned long	IprocEventQueueOverflow;

    unsigned long	QueueingPacketTrap;
    unsigned long	DmaIdentifyTrap;
    unsigned long	ThreadIdentifyTrap;
    unsigned long	DmaPacketTrap;
} EP3_RAIL_STATS;

typedef struct ep4_rail_stats
{
    unsigned long       somestatsgohere;
} EP4_RAIL_STATS;

typedef struct ep_rail_stats
{
    unsigned long	SendMessageFailed;
    unsigned long	NeterrAtomicPacket;
    unsigned long       NeterrDmaPacket;

    EP_STATS_COUNT      rx;
    EP_STATS_COUNT      rx_len;

    EP_STATS_COUNT      tx;
    EP_STATS_COUNT      tx_len;

} EP_RAIL_STATS;

typedef struct ep_cm_rail_stats
{
    /* cluster membership statistics */
    unsigned long	HeartbeatsSent;
    unsigned long	HeartbeatsRcvd;
    
    unsigned long	RetryHeartbeat;
    unsigned long	RejoinRequest;
    unsigned long	RejoinTooSlow;
    unsigned long	LaunchMessageFail;
    unsigned long	MapChangesSent;

    /* Heartbeat scheduling stats */
    unsigned long	HeartbeatOverdue;
} EP_CM_RAIL_STATS;

typedef struct ep_comms_rail_stats
{
    /* kernel comms large message statistics */
    unsigned long	TxEnveEvent;
    unsigned long	TxDataEvent;
    unsigned long	TxDoneEvent;
    unsigned long	RxDoneEvent;
    unsigned long	MulticastTxDone;
    unsigned long	QueueReceive;

    unsigned long	TxEnveRetry;
    unsigned long	TxDataRetry;
    unsigned long	TxDoneRetry;
    unsigned long	RxThrdEvent;
    unsigned long	RxDataRetry;
    unsigned long	RxDoneRetry;
    unsigned long	StallThread;
    unsigned long	ThrdWaiting;
    unsigned long	CompleteEnvelope;

    unsigned long	NoFreeTxds;
    unsigned long	NoFreeRxds;

    unsigned long	LockRcvrTrapped;
} EP_COMMS_RAIL_STATS;

typedef struct ep_comms_stats
{
    unsigned long	DataXmit[8];
    unsigned long	McastXmit[8];
    unsigned long	RPCXmit[8];
    unsigned long	RPCPut[8];
    unsigned long	RPCGet[8];
    unsigned long	CompleteRPC[8];
    unsigned long	RxData[8];
    unsigned long	RxMcast[8];

    unsigned long	NoFreeTxds;
    unsigned long	NoFreeRxds;
} EP_COMMS_STATS;

#endif /* __EP_EPSTATS_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
