/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_NMH_H
#define __ELAN3_NMH_H

#ident "@(#)$Id: nmh.h,v 1.7 2004/01/06 10:29:55 david Exp $"
/*      $Source: /cvs/master/quadrics/epmod/nmh.h,v $*/


/* Forward declarations */
typedef struct ep_nmd		EP_NMD;
typedef struct ep_nmh_ops	EP_NMH_OPS;
typedef struct ep_nmh		EP_NMH;

/* Railmask held in 16 bit field (packs with nodeId into NMD */
typedef uint16_t		EP_RAILMASK;

#define EP_RAIL2RAILMASK(rnum)	(1 << (rnum))
#define EP_RAILMASK_ALL		0xffff

/* kernel comms elan network address */
typedef uint32_t		EP_ADDR;

/* network mapping descriptor - this is returned to the user from a map operation,
 * and is what is passed to all communication functions */
struct ep_nmd
{
    EP_ADDR	nmd_addr;					/* base address */
    uint32_t	nmd_len;					/* size in bytes */
    uint32_t	nmd_attr;					/* nodeid << 16 | railmask */
};

#define EP_NMD_ATTR(nodeid,railmask)	(((nodeid) << 16) | (railmask))
#define EP_NMD_NODEID(nmd)		((nmd)->nmd_attr >> 16)
#define EP_NMD_RAILMASK(nmd)		((nmd)->nmd_attr & EP_RAILMASK_ALL)

#if !defined(__ELAN__)

struct ep_nmh_ops
{
    int	   (*op_map_rails) (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, EP_RAILMASK mask);	/* add mappings to different rail(s) */

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
    uint16_t (*op_calc_check_sum) (EP_SYS *sys, EP_NMH *nmh, EP_NMD *nmd, uint16_t check_sum); /* calculates check sum              */
#endif
};

struct ep_nmh
{
    EP_NMD	     nmh_nmd;					/* public field */
    struct list_head nmh_link;					/* linked on hash table */
    EP_NMH_OPS	    *nmh_ops;					/* operations to perform on object */
};

#define EP_NMH_NUMHASH			(32 - 11 + 1)		/* one hash table for each power of 2 above pagesize */
#define EP_NMH_HASHSIZE			(64)			/* max size of each hash table */

typedef struct ep_nmh_table
{
    struct list_head *tbl_hash[EP_NMH_NUMHASH];
    unsigned	      tbl_size[EP_NMH_NUMHASH];
} EP_NMH_TABLE;

extern int         ep_nmh_init (EP_NMH_TABLE *tbl);
extern void        ep_nmh_fini (EP_NMH_TABLE *tbl);

extern void        ep_nmh_insert (EP_NMH_TABLE *tbl, EP_NMH *nmd);
extern void        ep_nmh_remove (EP_NMH_TABLE *tbl, EP_NMH *nmd);
extern EP_NMH     *ep_nmh_find (EP_NMH_TABLE *tbl, EP_NMD *nmh);

#if ! defined(CONFIG_EP_NO_CHECK_SUM)
extern uint32_t    ep_nmd_calc_data_check_sum(EP_SYS *sys, EP_NMD *nmd, int nFrags);
#endif

/* Public interface */
extern EP_RAILMASK ep_nmd2railmask (EP_NMD *frags, int nFrags);
extern void        ep_nmd_subset (EP_NMD *subset, EP_NMD *nmd, unsigned off, unsigned len);
extern int	   ep_nmd_merge (EP_NMD *merged, EP_NMD *a, EP_NMD *b);
extern int         ep_nmd_map_rails (EP_SYS *sys, EP_NMD *nmd, unsigned railmask);

#endif /* __ELAN__ */

#endif /* __ELAN3_NMH_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
