/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: stats.h,v 1.5 2003/09/24 13:55:37 david Exp $"
/*      $Source: /cvs/master/quadrics/elanmod/modsrc/stats.h,v $*/

#ifndef __ELAN_STATS_H
#define __ELAN_STATS_H


/* non-kernel headings */
#define ELAN_STATS_NAME_MAX_LEN ((uint)64)
typedef unsigned int ELAN_STATS_IDX;

typedef struct elan_stats_map
{
	char entry_name[ELAN_STATS_NAME_MAX_LEN];
	int  index;
} ELAN_STATS_MAP;

#if defined(__KERNEL__)

/* stats callbacks */
#define ELAN_STATS_OPS_VERSION ((u_int)1)
typedef struct elan_stats_ops
{
	u_int  ops_version;

	int (*elan_stats_get_name)    (void * arg, uint index,   caddr_t  name);
	int (*elan_stats_get_block)   (void * arg, uint entries, ulong   *values);
	int (*elan_stats_clear_block) (void * arg);

} ELAN_STATS_OPS;

typedef struct elan_stats_struct
{
	struct list_head   node;

	ELAN_STATS_IDX     statidx;
	char               block_name[ELAN_STATS_NAME_MAX_LEN];
	uint               num_entries;
	ELAN_STATS_OPS *ops;
	void              *arg;

} ELAN_STATS_STRUCT;

/* stats.c */
extern int                   elan_stats_register   (ELAN_STATS_IDX    *statidx, 
						       char              *block_name, 
						       uint               num_entries,
						       ELAN_STATS_OPS *ops,
						       void              *arg);

extern int                   elan_stats_deregister  (ELAN_STATS_IDX  statidx);
extern ELAN_STATS_STRUCT *elan_stats_find        (ELAN_STATS_IDX  statidx);
extern ELAN_STATS_STRUCT *elan_stats_find_by_name(caddr_t         block_name);
extern ELAN_STATS_STRUCT *elan_stats_find_next   (ELAN_STATS_IDX  statidx);


/* elan_stats.c */
extern int elan_stats_get_next_index (ELAN_STATS_IDX statidx, ELAN_STATS_IDX *next_statidx);

extern int elan_stats_find_index     (caddr_t  block_name, ELAN_STATS_IDX *statidx, uint *num_entries);

extern int elan_stats_get_block_info (ELAN_STATS_IDX statidx, caddr_t block_name, uint *num_entries);

extern int elan_stats_get_index_name (ELAN_STATS_IDX statidx, uint index, caddr_t name);

extern int elan_stats_get_block      (ELAN_STATS_IDX statidx, uint entries, ulong   *values);

extern int elan_stats_clear_block    (ELAN_STATS_IDX statidx);

#endif /* __KERNEL__ */

#endif /* __ELAN_STATS_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
