/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_ELAN3MMU_H
#define __ELAN3_ELAN3MMU_H

#ident "$Id: elan3mmu.h,v 1.41 2004/12/14 10:19:32 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elan3mmu.h,v $*/


#include <elan3/pte.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct elan3mmu_global_stats
{
    int		version;
    int		pteload;
    int		pteunload;
    int		ptereload;

    int		streamable_alloc;
    int		streamable_free;
    int		streamable_alloc_failed;

    int		num_ptbl_level[4]; /* number of level N  ptbls */

    int		create_ptbl_failed;			/* count of ptbl creation failure */

    int         lX_alloc_l3;	                        /* count of l3 ptbls used as lX */
    int         lX_freed_l3; 	                        /* count of lX ptbls freed as l3 */

    int		l2_alloc_l3;				/* count of l3 ptbls used as l2 */
    int		l2_freed_l3;				/* count of l2 ptbls freed as l3 */

    int		stolen_ptbls;				/* count of l3 ptbls stolen */
} ELAN3MMU_GLOBAL_STATS;

#define ELAN3MMU_STATS_VERSION		1

#define ELAN3MMU_STAT(what)		(elan3mmu_global_stats.what++)
#define ELAN3MMU_SET_STAT(what,count)	(elan3mmu_global_stats.what = count)

#ifdef __KERNEL__

#define ELAN3_PT_SHIFT	(ELAN3_L2_SHIFT + 2)

typedef struct elan3_ptbl
{
    struct elan3_ptbl	 *ptbl_parent;				/* Parent page table, or next on freelist */
    struct elan3mmu	 *ptbl_elan3mmu;			/* elan3mmu we're allocated for */
    E3_Addr		  ptbl_base;				/* Virtual address we're mapping */
    u_char		  ptbl_index;				/* Index in ptbl group */
    u_char		  ptbl_valid;				/* Number of valid entries */
    u_char		  ptbl_flags;				/* Flags, defined below. */
    u_char		  ptbl_spare;
} ELAN3_PTBL;

#define ptbl_next	ptbl_parent				/* Parent pointer is next pointer when on free list */

#define PTBL_LEVEL_X            0x00
#define PTBL_LEVEL_1		0x01
#define PTBL_LEVEL_2		0x02
#define PTBL_LEVEL_3		0x03
#define PTBL_LEVEL_MASK		0x03
#define PTBL_LOCKED		0x04				/* Page table is locked,  protects all fields */
#define PTBL_KEEP		0x08				/* This ptbl is not to be stolen */
#define PTBL_ALLOCED		0x10				/* This ptbl has been allocated, and is not free */
#define PTBL_GROUPED		0x20				/* This ptbl is a member of a group of ptbls */
#define PTBL_KERNEL		0x80				/* This ptbl is allocated for the kernel */

#define PTBL_LEVEL(flags)	((flags) & PTBL_LEVEL_MASK)
#define PTBL_IS_LOCKED(flags)	(((flags) & (PTBL_LOCKED|PTBL_ALLOCED)) == (PTBL_LOCKED|PTBL_ALLOCED))

#if ELAN3_PAGE_SHIFT == 13
#  define PTBL_GROUP_SIZE	8192				/* page table groups are 8k bytes */
#  define PTBLS_PER_GROUP_L1	8				/* Number of level 1 tables in a group */
#  define PTBLS_PER_GROUP_L2	32				/*   ... level 2 */
#  define PTBLS_PER_GROUP_L3	32				/*   ... level 3 */
#  define PTBLS_PER_GROUP_LX	32				/*   ... level X */
#  define PTBLS_PER_GROUP_MAX	32				/*  max of l1,l2,l3,lX */
#else
#  define PTBL_GROUP_SIZE	4096				/* page table groups are 4k bytes */
#  define PTBLS_PER_GROUP_L1	4				/* Number of level 1 tables in a group */
#  define PTBLS_PER_GROUP_L2	16				/*   ... level 2 */
#  define PTBLS_PER_GROUP_L3	8				/*   ... level 3 */
#  define PTBLS_PER_GROUP_LX	16				/*   ... level X */
#  define PTBLS_PER_GROUP_MAX	16				/*  max of l1,l2,l3,lX */
#endif

#define HMES_PER_GROUP		(PTBLS_PER_GROUP_L3*ELAN3_L3_ENTRIES)

#if ELAN3_PAGE_SHIFT == 13
#  define PTBLS_PER_PTBL_L1	4				/* 256 PTPs */
#  define PTBLS_PER_PTBL_L2	1				/* 64 PTPs */
#  define PTBLS_PER_PTBL_L3	1				/* 32 PTEs */
#else
#  define PTBLS_PER_PTBL_L1	4				/* 256 PTPs */
#  define PTBLS_PER_PTBL_L2	1				/* 64 PTPs */
#  define PTBLS_PER_PTBL_L3	2				/* 64 PTEs */
#endif

#define ELAN3_LX_ENTRIES     (32) 
#define PTBLS_PER_PTBL_LX   (1)	

#define L1_VA_PER_PTBL	(ELAN3_L1_SIZE*(ELAN3_L1_ENTRIES/PTBLS_PER_PTBL_L1))	/* 4 ptbl for L1 */
#define L2_VA_PER_PTBL	(ELAN3_L2_SIZE*(ELAN3_L2_ENTRIES/PTBLS_PER_PTBL_L2))	/* 1 ptbl for L2 */
#define L3_VA_PER_PTBL	(ELAN3_L3_SIZE*(ELAN3_L3_ENTRIES/PTBLS_PER_PTBL_L3))	/* 1 ptbl for L3 */

typedef struct elan3_ptbl_gr
{
    struct elan3_ptbl_gr *pg_next;				/* Next in list. */
    int			 pg_level;				/* Level PG allocated for */
    sdramaddr_t		 pg_addr;				/* sdram offset of ptes/ptps */    
    ELAN3_PTBL		 pg_ptbls[PTBLS_PER_GROUP_MAX];		/* The actual page tables */
} ELAN3_PTBL_GR;


/*
 * The elan3mmu structure is the mmu dependant hardware address translation
 * structure linked to the address space structure to show the translatioms
 * provided by the elan for an address sapce.
 *
 * We also have a doubly linked list of 'regions' which allow the 
 * elan3mmu code to determine the access permissions for the elan 
 * dependant on the virtual address that the translation is being
 * loaded at.
 */

typedef struct elan3mmu_rgn
{
    struct elan3mmu_rgn *rgn_mnext;				/* Doubly linked list of regions */
    struct elan3mmu_rgn *rgn_mprev;				/*   sorted on main address */ 
    caddr_t		 rgn_mbase;				/* main address of base of region */

    struct elan3mmu_rgn *rgn_enext;				/* Doubly linked list of regions */
    struct elan3mmu_rgn *rgn_eprev;				/*   sorted on elan address */
    E3_Addr		 rgn_ebase;				/* elan address of base of region */

    u_int		 rgn_len;				/* length of region */
    u_int		 rgn_perm;				/* elan access permission */
} ELAN3MMU_RGN;

typedef struct elan3mmu
{
    spinlock_t		    elan3mmu_lock;			/* spinlock lock for regions */
    ELAN3MMU_RGN           *elan3mmu_mrgns;			/* Doubly linked list of memory regions */
    ELAN3MMU_RGN	   *elan3mmu_mtail;			/* Last memory region on list */
    ELAN3MMU_RGN	   *elan3mmu_mrgnlast;			/* Last region 'hit' */

    ELAN3MMU_RGN           *elan3mmu_ergns;			/* Doubly linked list of memory regions */
    ELAN3MMU_RGN	   *elan3mmu_etail;			/* Last memory region on list */
    ELAN3MMU_RGN	   *elan3mmu_ergnlast;			/* Last region 'hit' */

    struct elan3_dev        *elan3mmu_dev;			/* Elan device we're using. */
    struct elan3_ctxt	   *elan3mmu_ctxt;			/* Elan ctxt we're associated with */

    sdramaddr_t             elan3mmu_ctp;			/* Context table entry for our context */
    ELAN3_PTBL	           *elan3mmu_l1ptbl;			/* Level 1 Page table (first of 4) */

    spinlock_t	            elan3mmu_lXptbl_lock;               /* spinlock for level X table list */
    ELAN3_PTBL              *elan3mmu_lXptbl;                    /* Level X Page table list         */

#ifdef LINUX
    struct mm_struct       *elan3mmu_coproc_mm;			/* Linux mm we're mapping */
#endif
} ELAN3MMU;

_NOTE(LOCK_ORDER(elan3mmu::elan3mmu_lock elan3_dev::IntrLock))

_NOTE(MUTEX_PROTECTS_DATA(elan3mmu::elan3mmu_lock,
			  elan3mmu::elan3mmu_mrgns elan3mmu::elan3mmu_mtail
			  elan3mmu::elan3mmu_ergns elan3mmu::elan3mmu_etail))
/* protected by dev->IntrLock for read by device driver */
_NOTE(DATA_READABLE_WITHOUT_LOCK(elan3mmu::elan3mmu_mrgns elan3mmu::elan3mmu_mtail
				 elan3mmu::elan3mmu_ergns elan3mmu::elan3mmu_etail))

_NOTE(SCHEME_PROTECTS_DATA("only set to valid region", 
			   elan3mmu::elan3mmu_ergnlast elan3mmu::elan3mmu_mrgnlast))

_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::IntrLock,
			  elan3mmu::elan3mmu_l1ptbl 
			  elan3mmu::elan3mmu_ctp 
			  elan3mmu::elan3mmu_dev))

_NOTE(DATA_READABLE_WITHOUT_LOCK(elan3mmu::elan3mmu_l1ptbl
				 elan3mmu::elan3mmu_ctp 
				 elan3mmu::elan3mmu_dev))

/*
 * Macros for accessing ptes/ptbls/ptbl_grs
 */

#define OFFSETOF(object,member)	/* calculate offset of structure member */ \
	((size_t) (&(((object *)0)->member)))
#define PTBL_TO_GR(ptbl)	/* convert ptbl to ptbl group */ \
	((ELAN3_PTBL_GR *) ((caddr_t) ((ptbl) - (ptbl)->ptbl_index) - OFFSETOF(ELAN3_PTBL_GR,pg_ptbls[0])))
#define PTBL_TO_PTADDR(ptbl)	/* convert ptbl to a ptp pointing at it */ \
        (PTBL_TO_GR(ptbl)->pg_addr + ((ptbl)->ptbl_index<<ELAN3_PT_SHIFT))
#define PTE_TO_HME(ptbl,pte)	/* convert pte to corresponding hme */ \
        (PTBL_TO_GR(ptbl)->pg_hmes + ((pte) - (ELAN3_PTE *) PTBL_TO_GR(ptbl)->pg_vaddr))
#define HME_TO_PTE(ptebl,hme)	/* convert hme to corresponding pte */ \
        ((ELAN3_PTE *) PTBL_TO_GR(ptbl)->pg_vaddr + ((hme) - (PTBL_TO_GR(ptbl)->pg_hmes)))


/* Flags for lock_ptbl */
#define LK_PTBL_NOWAIT		0x1
#define LK_PTBL_FAILOK		0x2

/* Return values for lock_ptbl */
#define LK_PTBL_OK		0x0
#define LK_PTBL_MISMATCH	0x1
#define LK_PTBL_FAILED		0x2

/* Flags for elan3mmu_ptesync */
#define	NO_MLIST_LOCK	0
#define	MLIST_LOCKED	1

/* Flags for elan3mmu_pteload */
#define PTE_LOAD		0x00
#define PTE_LOAD_LOCK		0x01				/* translation should be locked */
#define PTE_LOAD_NOSYNC		0x02				/* ref/mod bits should not be sync'ed to page */
#define PTE_NO_SLEEP            0x04                            /* true if we cant sleep */
#define PTE_NO_STEAL		0x08				/* true if we don't want to steal ptbls */

#define PTE_LOAD_ENDIAN_MASK	0x10				/* mask for endian-ness */
#define PTE_LOAD_LITTLE_ENDIAN	0x00				/* translation is to little-endian memory */
#define PTE_LOAD_BIG_ENDIAN	0x10				/* translation is to big-endian memory */


/* Flags for elan3mmu_unload */
#define PTE_UNLOAD		0x00
#define PTE_UNLOAD_UNLOCK	0x01
#define PTE_UNLOAD_NOFLUSH	0x02
#define PTE_UNLOAD_NOSYNC	0x04

extern int	    elan3mmu_debug;
#ifdef DEBUG_PRINTF
#  define HAT_PRINTF0(n,msg)		 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg)             : (void) 0)
#  define HAT_PRINTF1(n,msg,a)		 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a)           : (void) 0)
#  define HAT_PRINTF2(n,msg,a,b)	 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a,b)         : (void) 0)
#  define HAT_PRINTF3(n,msg,a,b,c)	 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a,b,c)       : (void) 0)
#  define HAT_PRINTF4(n,msg,a,b,c,d)	 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a,b,c,d)     : (void) 0)
#  define HAT_PRINTF5(n,msg,a,b,c,d,e)	 ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a,b,c,d,e)   : (void) 0)
#  define HAT_PRINTF6(n,msg,a,b,c,d,e,f) ((elan3mmu_debug & n) ? (void) elan3_debugf (NULL, DBG_HAT, msg,a,b,c,d,e,f) : (void) 0)
#  ifdef LINUX
#    define HAT_PRINTF(n,args...)        ((elan3mmu_debug & n) ? (void) elan3_debugf(NULL, DBG_HAT, ##args) : (void) 0)
#  endif
#else
#  define HAT_PRINTF0(n,msg)
#  define HAT_PRINTF1(n,msg,a)
#  define HAT_PRINTF2(n,msg,a,b)
#  define HAT_PRINTF3(n,msg,a,b,c)
#  define HAT_PRINTF4(n,msg,a,b,c,d)
#  define HAT_PRINTF5(n,msg,a,b,c,d,e)
#  define HAT_PRINTF6(n,msg,a,b,c,d,e,f)
#  ifdef LINUX
#    define HAT_PRINTF(n,args...)
#  endif
#endif

/* elan3mmu_generic.c */
extern ELAN3MMU_GLOBAL_STATS elan3mmu_global_stats;

extern void	     elan3mmu_init (void);
extern void	     elan3mmu_fini (void);

extern ELAN3MMU	   *elan3mmu_alloc (struct elan3_ctxt *ctxt);
extern void	    elan3mmu_free (ELAN3MMU *elan3mmu);

extern void          elan3mmu_set_context_filter (ELAN3_DEV *dev, int ctx, int disabled, E3_uint32 Pend, E3_uint32 *Maskp);
extern int	     elan3mmu_attach (ELAN3_DEV *dev, int ctx, ELAN3MMU *elan3mmu, sdramaddr_t routeTable, E3_uint32 routeMask);
extern void	     elan3mmu_detach (ELAN3_DEV *dev, int ctx);

extern ELAN3MMU_RGN *elan3mmu_findrgn_elan (ELAN3MMU *elan3mmu, E3_Addr addr, int tail);
extern int           elan3mmu_addrgn_elan (ELAN3MMU *elan3mmu, ELAN3MMU_RGN *nrgn);
extern ELAN3MMU_RGN *elan3mmu_removergn_elan (ELAN3MMU *elan3mmu, E3_Addr addr);
extern ELAN3MMU_RGN *elan3mmu_rgnat_elan (ELAN3MMU *elan3mmu, E3_Addr addr);
extern ELAN3MMU_RGN *elan3mmu_findrgn_main (ELAN3MMU *elan3mmu, caddr_t addr, int tail);
extern int           elan3mmu_addrgn_main (ELAN3MMU *elan3mmu, ELAN3MMU_RGN *nrgn);
extern ELAN3MMU_RGN *elan3mmu_removergn_main (ELAN3MMU *elan3mmu, caddr_t addr);
extern ELAN3MMU_RGN *elan3mmu_rgnat_main (ELAN3MMU *elan3mmu, caddr_t addr);

extern int	     elan3mmu_setperm (ELAN3MMU *elan3mmu, caddr_t maddr, E3_Addr eaddr, u_int len, u_int perm);
extern void	     elan3mmu_clrperm (ELAN3MMU *elan3mmu, E3_Addr addr, u_int len);
extern int	     elan3mmu_checkperm (ELAN3MMU *elan3mmu, E3_Addr addr, u_int len, u_int access);
extern caddr_t	     elan3mmu_mainaddr (ELAN3MMU *elan3mmu, E3_Addr addr);
extern E3_Addr	     elan3mmu_elanaddr (ELAN3MMU *elan3mmu, caddr_t addr);

extern void          elan3mmu_expand (ELAN3MMU *elan3mmu, E3_Addr addr, int len, int level, int attr);
extern void          elan3mmu_reserve (ELAN3MMU *elan3mmu, E3_Addr addr, u_int npages, sdramaddr_t *);
extern void          elan3mmu_release (ELAN3MMU *elan3mmu, E3_Addr addr, u_int npages, sdramaddr_t *);

extern void          elan3mmu_pteload (ELAN3MMU *elan3mmu, int level, E3_Addr addr, physaddr_t paddr, int perm, int attr);
extern void	     elan3mmu_unload (ELAN3MMU *elan3mmu, E3_Addr addr, u_int len, int flags);
extern void	     elan3mmu_sync (ELAN3MMU *elan3mmu, E3_Addr addr, u_int len, u_int clearflag);
extern void	     elan3mmu_pteunload (ELAN3_PTBL *ptbl, sdramaddr_t pte, int flags, int got_mlist_lock);
extern void	     elan3mmu_ptesync (ELAN3_PTBL *ptbl, sdramaddr_t pte, int flags, int got_mlist_lock);
extern sdramaddr_t   elan3mmu_ptp2pte (ELAN3MMU *elan3mmu, sdramaddr_t ptp, int level);
extern sdramaddr_t   elan3mmu_ptefind (ELAN3MMU *elan3mmu, E3_Addr, int *level, ELAN3_PTBL **pptbl, spinlock_t **plock, unsigned long *flags);
extern sdramaddr_t   elan3mmu_ptealloc (ELAN3MMU *elan3mmu, E3_Addr, int level, ELAN3_PTBL **pptbl, spinlock_t **plock, int attr, unsigned long *flags);
extern void	     elan3mmu_l1inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l1ptbl, int flags);
extern int           elan3mmu_l2inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l2ptbl, int flags, E3_Addr addr, spinlock_t **pl2lock, unsigned long *lock_flags);
extern int           elan3mmu_l3inval (ELAN3MMU *elan3mmu, ELAN3_PTBL *l3ptbl, int flags, E3_Addr addr, spinlock_t **pl3lock, unsigned long *lock_flags);

extern void          elan3mmu_free_l1ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags);
extern void          elan3mmu_free_l2ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags);
extern void          elan3mmu_free_l3ptbl (ELAN3_DEV *dev, ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags);

extern int	     elan3mmu_lock_this_ptbl (ELAN3_PTBL *ptbl, int flag, spinlock_t **plock, unsigned long *flags);
extern int           elan3mmu_lock_ptbl (ELAN3_PTBL *ptbl, u_int flag, ELAN3MMU *elan3mmu, E3_Addr va, int level, spinlock_t **plock, unsigned long *flags);
extern void	     elan3mmu_unlock_ptbl (ELAN3_PTBL *ptbl, spinlock_t *lock, unsigned long flags);

/* elan3mmu_osdep.c */
extern void	     elan3mmu_init_osdep (void);
extern void	     elan3mmu_fini_osdep (void);
extern void	     elan3mmu_alloc_osdep (ELAN3MMU *elan3mmu);
extern void	     elan3mmu_free_osdep (ELAN3MMU *elan3mmu);
extern ELAN3_PTE     elan3mmu_phys_to_pte (ELAN3_DEV *dev, physaddr_t paddr, int perm);
extern ELAN3_PTE     elan3mmu_kernel_invalid_pte (ELAN3MMU *elan3mmu);

#if defined (DIGITAL_UNIX)
#  include <elan3/elan3mmu_dunix.h>
#elif defined (LINUX)
#  include <elan3/elan3mmu_linux.h>
#endif

#endif /* __KERNEL__ */

#ifdef __cplusplus
}
#endif

#endif /* __ELAN3_ELAN3MMU_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
