/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_ELANCTXT_H
#define _ELAN3_ELANCTXT_H

#ident "$Id: elanctxt.h,v 1.81 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elanctxt.h,v $*/

#ifdef __cplusplus
extern "C" {
#endif

#include <elan3/elanregs.h>
#include <elan3/vmseg.h>

#define BumpUserStat(ctxt, stat)	((ctxt)->FlagPage->stat++)

#if defined(__LITTLE_ENDIAN__)

typedef union _CProcTrapBuf
{
    E3_uint64 Align64;
    struct
    {
	E3_uint32 Areg;
	E3_uint32 Breg;
    } r;
    struct
    {
	E3_uint32 Addr;
	E3_uint32 ContextType;
    } s;
} CProcTrapBuf_BE;

typedef E3_EventInt        E3_EventInt_BE;
typedef E3_IprocTrapHeader E3_IprocTrapHeader_BE;
typedef E3_IprocTrapData   E3_IprocTrapData_BE;
typedef E3_FaultSave 	   E3_FaultSave_BE;

typedef union
{
    E3_uint64	Align64;
    E3_DMA      s;
} E3_DMA_BE;

typedef E3_ThreadQueue     E3_ThreadQueue_BE;

#else

/* "Big-Endian" data structures copied by 64 bit loads, these are 32 bit word flipped */
/* from the corresponding data structure. */

typedef union _CProcTrapBuf
{
    E3_uint64 Align64;
    struct
    {
	E3_uint32 Breg;
	E3_uint32 Areg;
    } r;
    struct
    {
	E3_uint32 ContextType;
	E3_uint32 Addr;
    } s;
} CProcTrapBuf_BE;

typedef union _E3_EventInt_BE
{
    E3_uint64	  Align64;
    struct {
	E3_uint32 EventContext;	/* Bits 16 to 28 */
	E3_uint32 IntCookie;
    } s;
} E3_EventInt_BE;

typedef union _E3_IprocTrapHeader_BE
{
   E3_uint64		 Align64;

   struct
   {
      E3_uint32	 	 TrAddr;
      E3_TrTypeCntx 	 TrTypeCntx;
      union
      {
	 E3_IProcStatus_Reg u_IProcStatus;
	 E3_uint32	    u_TrData1;
      } ipsotd;
      E3_uint32	 	 TrData0;
   } s;
} E3_IprocTrapHeader_BE;

typedef E3_IprocTrapData E3_IprocTrapData_BE;

typedef union _E3_FaultSave_be
{
    E3_uint64			Align64;
    struct {
	volatile E3_uint32	FaultContext;
	E3_FaultStatusReg	FSR;
	volatile E3_uint32	EventAddress;
	volatile E3_uint32	FaultAddress;
    } s;
} E3_FaultSave_BE;

typedef union _e3_dma_be
{
    E3_uint64		Align64;
    struct {
	E3_uint32	dma_size;
	E3_DmaType	dma_u;
	E3_Addr		dma_dest;
	E3_Addr		dma_source;
	E3_CookieVProc	dma_destCookieProc;
	E3_Addr		dma_destEvent;
	E3_CookieVProc	dma_srcCookieProc;
	E3_Addr		dma_srcEvent;
    } s;
} E3_DMA_BE;

typedef union _E3_ThreadQueue_BE
{
   E3_uint64	Align64;
   struct
   {
       /* copied by 64 bit copy from elan to main */
       E3_uint32 :3;		/* Bits 29 to 31 */
       E3_uint32 Context:13;	/* Bits 16 to 28 */
       E3_uint32 :16;		/* Bits 0  to 15 */
       E3_Addr	 Thread;	/* Bits 32 to 63 */
   } s;
} E3_ThreadQueue_BE;

#endif /* defined(LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__) */

typedef struct neterr_msg
{
    E3_uint32		Rail;					/* Rail error received on */
    ELAN_CAPABILITY	SrcCapability;				/* Capability of source of packet */
    ELAN_CAPABILITY	DstCapability;				/* Capability of dest of packet */

    E3_uint32		DstProcess;				/* Virtual Process of dest of packet */
    E3_Addr		CookieAddr;				/* Cookie Address (or NULL for DMA) */
    E3_uint32		CookieVProc;				/* Cookie and VP (identifies DMA) */
    E3_uint32		NextCookie;				/* Next Cookie value (for thread) */
    E3_uint32		WaitForEop;				/* Wait for EOP transaction */
} NETERR_MSG;

#ifdef __KERNEL__

/*
 * Associated with each input channel can be a network error
 * resolver structure, which can be queued on the network 
 * error resolver threads to perform RPCs to the other kernels
 * when a network error occurs with an identify transaction
 * included
 */
typedef struct neterr_resolver
{
    struct neterr_resolver *Next;

    spinlock_t	            Lock;

    struct elan3_ctxt       *Ctxt;
    ELAN_LOCATION	    Location;

    int			    Completed;
    int			    Status;
    long		    Timestamp;

    NETERR_MSG	            Message;
} NETERR_RESOLVER;


typedef struct neterr_fixup
{
    struct neterr_fixup	   *Next;

    kcondvar_t		    Wait;
    int			    Completed;
    int			    Status;

    NETERR_MSG		    Message;
} NETERR_FIXUP;

#endif /* __KERNEL__ */

/* Each of the following structures must be padded to a whole */
/* number of 64 bit words since the kernel uses 64 bit load/stores */
/* to transfer the elan register state. */
typedef struct command_trap
{
    E3_Status_Reg	Status;					/* 4  bytes */
    E3_uint32		Pad;					/* 4  bytes */
    E3_FaultSave_BE	FaultSave;				/* 16 bytes */
    CProcTrapBuf_BE    	TrapBuf;				/* 8  bytes */
} COMMAND_TRAP;

typedef struct thread_trap
{
    E3_uint32		Registers[32];				/* 128 bytes */
#define REG_GLOBALS	0
#define REG_OUTS	8
#define REG_LOCALS	16
#define REG_INS		24

    E3_FaultSave_BE	FaultSave;				/* 16 bytes */
    E3_FaultSave_BE     DataFaultSave;				/* 16 bytes */
    E3_FaultSave_BE     InstFaultSave;				/* 16 bytes */
    E3_FaultSave_BE     OpenFaultSave;				/* 16 bytes */
    
    E3_Status_Reg	Status;					/* 4 bytes */

    E3_Addr		pc;					/* 4 bytes */
    E3_Addr		npc;					/* 4 bytes */
    E3_Addr		StartPC;				/* 4 bytes */
    E3_Addr		sp;					/* 4 bytes */
    E3_uint32		mi;					/* 4 bytes */
    E3_TrapBits		TrapBits;				/* 4 bytes */
    E3_DirtyBits	DirtyBits;				/* 4 bytes */
} THREAD_TRAP;

typedef struct dma_trap
{
    E3_DMA_BE		Desc;					/* 32 bytes */
    E3_FaultSave_BE	FaultSave;				/* 16 bytes */
    E3_FaultSave_BE	Data0;					/* 16 bytes */
    E3_FaultSave_BE	Data1;					/* 16 bytes */
    E3_FaultSave_BE	Data2;					/* 16 bytes */
    E3_FaultSave_BE	Data3;					/* 16 bytes */
    E3_Status_Reg	Status;					/* 4 bytes */
    E3_DmaInfo		PacketInfo;				/* 4 bytes */
} DMA_TRAP;

typedef struct input_trap
{
    E3_uint32		   State;				/* 4 bytes */
    E3_Status_Reg	   Status;				/* 4 bytes */
    E3_FaultSave_BE	   FaultSave;				/* 16 bytes */
    
    u_int		   NumTransactions;			/* 4 bytes */
    u_int		   Overflow;				/* 4 bytes */
    u_int		   AckSent;				/* 4 bytes */
    u_int		   BadTransaction;			/* 4 bytes */

    E3_IprocTrapHeader_BE *TrappedTransaction;			/* 4 bytes */
    E3_IprocTrapData_BE   *TrappedDataBuffer;			/* 4 bytes */
    E3_IprocTrapHeader_BE *WaitForEopTransaction;		/* 4 bytes */
    E3_IprocTrapData_BE   *WaitForEopDataBuffer;		/* 4 bytes */
    E3_IprocTrapHeader_BE *DmaIdentifyTransaction;		/* 4 bytes */
    E3_IprocTrapHeader_BE *ThreadIdentifyTransaction;		/* 4 bytes */
    E3_Addr		   LockQueuePointer;			/* 4 bytes */
    E3_Addr		   UnlockQueuePointer;			/* 4 bytes */

    E3_IprocTrapHeader_BE  Transactions[MAX_TRAPPED_TRANS];	/* n * 8 bytes */
    E3_IprocTrapData_BE	   DataBuffers[MAX_TRAPPED_TRANS];	/* n * 64 bytes */
} INPUT_TRAP;

typedef struct input_fault_save
{
    struct input_fault_save *Next;
    E3_Addr	             Addr;
    E3_uint32	             Count;
} INPUT_FAULT_SAVE;

#define NUM_INPUT_FAULT_SAVE	32
#define MIN_INPUT_FAULT_PAGES	8
#define MAX_INPUT_FAULT_PAGES	128

typedef E3_uint32 EVENT_COOKIE;

#ifdef __KERNEL__

typedef struct event_cookie_entry
{
    struct event_cookie_entry *ent_next;
    struct event_cookie_entry *ent_prev;

    spinlock_t		       ent_lock;
    unsigned		       ent_ref;

    EVENT_COOKIE	       ent_cookie;
    EVENT_COOKIE	       ent_fired;
    kcondvar_t		       ent_wait;
} EVENT_COOKIE_ENTRY;

typedef struct event_cookie_table
{
    struct event_cookie_table *tbl_next;
    struct event_cookie_table *tbl_prev;

    unsigned long              tbl_task;
    unsigned long              tbl_handle;

    spinlock_t	               tbl_lock;
    unsigned	               tbl_ref;
    EVENT_COOKIE_ENTRY        *tbl_entries;
} EVENT_COOKIE_TABLE;

#define NBYTES_PER_SMALL_ROUTE	8
#define NBYTES_PER_LARGE_ROUTE	16

#define ROUTE_BLOCK_SIZE	ELAN3_PAGE_SIZE
#define NROUTES_PER_BLOCK	(ROUTE_BLOCK_SIZE/NBYTES_PER_LARGE_ROUTE)

typedef struct elan3_routes
{
    struct elan3_routes		*Next;					/* Can be chained together */

    sdramaddr_t			 Routes;				/* sdram offset of route entries */
    bitmap_t			 Bitmap[BT_BITOUL(NROUTES_PER_BLOCK)];	/* Bitmap of which entries are used */
} ELAN3_ROUTES; 


typedef struct elan3_route_table
{
    spinlock_t		 Lock;				/* Route lock */
    sdramaddr_t		 Table;				/* Kernel address for route table */
    u_int		 Size;				/* # entries in route table */

    ELAN3_ROUTES	*LargeRoutes;			/* Large routes */
} ELAN3_ROUTE_TABLE;

typedef struct elan3_vpseg
{
    struct elan3_vpseg		*Next;
    int				 Process;			/* Virtual process */
    int				 Entries;			/*  and # processes */
    int				 Type;				/* Type of cookie */

    union
    {
	
	ELAN_CAPABILITY	Capability;			/* Capability of remote segment */
#  define SegCapability		SegUnion.Capability
	struct {
	    u_short	        LowProc;			/* Base process number */
	    u_short	        HighProc;			/*   and high process number */
#  define SegLowProc		SegUnion.BROADCAST.LowProc
#  define SegHighProc		SegUnion.BROADCAST.HighProc
	} BROADCAST;
    } SegUnion;
} ELAN3_VPSEG;

#define ELAN3_VPSEG_UNINT	0				/* Unitialised */
#define ELAN3_VPSEG_P2P		1				/* Point to Point */
#define ELAN3_VPSEG_BROADCAST	2				/* Broadcast */

#define NUM_LISTS	7					/* Number of "swap" lists */

typedef struct elan3_ctxt
{
    struct elan3_ctxt    *Next;					/* can be queued on a task */
    struct elan3_ctxt    *Prev;

    CtxtHandle		 Handle;				/* user handle */
    int			 RefCnt;				/* reference count */

    ELAN3MMU		*Elan3mmu;				/* elan3mmu allocated for Elan translations */

    struct elan3_ops     *Operations;				/* User supplied helper functions */
    void		*Private;				/* Users private pointer */

    int			 Status;				/* Status (guarded by dev_mutex) */
    int			 OthersState;				/* State of halt queueing for dma/thread */
    int			 LwpCount;				/* Number of lwp's running */

    ELAN3_DEV		*Device;				/* Elan device */

    ELAN_CAPABILITY	 Capability;				/* Capability I've attached as */
    ELAN_POSITION	 Position;				/* Position when I was created */
    
    ELAN3_VPSEG		*VpSegs;				/* List of virtual process segments */
    ELAN3_ROUTE_TABLE    *RouteTable;

    krwlock_t		 VpLock;				/* Reader/writer lock for vp list */
    kmutex_t		 SwapListsLock;				/* mutex to lock swap lists */
    kmutex_t		 CmdLock;				/* mutex to lock trapped dma command */
    kmutex_t		 CmdPortLock;				/* mutex to load/unload commandport xlation */

    kcondvar_t		 Wait;					/* Condition variable to sleep on */
    kcondvar_t		 CommandPortWait;			/* Condition variable to wait for commandport */
    kcondvar_t		 LwpWait;				/* Condition variable to wait for lwps to stop */
    kcondvar_t		 HaltWait;				/* Condition variable to wait for halt */
    int			 Halted;				/*  and flag for halt cv */

    caddr_t		 CommandPageMapping;			/* user virtual address for command page mapping */
    ioaddr_t             CommandPage;				/* Elan command port mapping page */
    DeviceMappingHandle  CommandPageHandle;			/* DDI Handle */
    ioaddr_t		 CommandPort;				/* Elan command port */
    void		*CommandPortItem;			/* Item we're re-issuing to commandport */

    ELAN3_FLAGSTATS      *FlagPage;				/* Page visible to user process */

    COMMAND_TRAP	*CommandTraps;				/* Command port traps */
    ELAN3_SPLIT_QUEUE     CommandTrapQ;
								   
    CProcTrapBuf_BE	*Commands;				/* Overflowed commands */
    ELAN3_QUEUE           CommandQ;

    THREAD_TRAP		*ThreadTraps;				/* Thread processor traps */
    ELAN3_QUEUE		 ThreadTrapQ;
    
    DMA_TRAP		*DmaTraps;				/* Dma processor tra[ed */
    ELAN3_QUEUE		 DmaTrapQ;

    INPUT_TRAP		 Input0Trap;				/* Inputter channel 0 trap */
    INPUT_TRAP		 Input1Trap;				/* Inputter channel 1 trap */
    NETERR_RESOLVER	*Input0Resolver;			/* Inputter channel 0 network error resolver */
    NETERR_RESOLVER	*Input1Resolver;			/* Inputter channel 1 network error resolver */

    INPUT_FAULT_SAVE	 InputFaults[NUM_INPUT_FAULT_SAVE];	/* stored writeblock addresses */
    INPUT_FAULT_SAVE    *InputFaultList;			/* organized in list for LRU */
    spinlock_t		 InputFaultLock;			/* and lock for list */

    kmutex_t		 NetworkErrorLock;
    NETERR_FIXUP	*NetworkErrorFixups;

    EVENT_COOKIE        *EventCookies;				/* Event cookies. */
    ELAN3_QUEUE		 EventCookieQ;

    E3_Addr		*SwapThreads;				/* Swapped Thread Queue */
    ELAN3_QUEUE		 SwapThreadQ;

    E3_DMA_BE		*SwapDmas;				/* Swapped Dmas Queue */
    ELAN3_QUEUE		 SwapDmaQ;

    int			 ItemCount[NUM_LISTS];			/* Count of items on each swap list */
    int			 inhibit;				/* if set lwp not to reload translations */

    int                  Disabled;
} ELAN3_CTXT;

_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::IntrLock, 
			  elan3_ctxt::Status elan3_ctxt::OthersState
			  elan3_ctxt::CommandTrapQ elan3_ctxt::CommandQ elan3_ctxt::ThreadTrapQ elan3_ctxt::DmaTrapQ 
			  elan3_ctxt::Input0Trap elan3_ctxt::Input1Trap elan3_ctxt::EventCookieQ elan3_ctxt::SwapThreadQ 
			  elan3_ctxt::SwapDmaQ elan3_ctxt::CommandPortItem elan3_ctxt::LwpCount))
_NOTE(MUTEX_PROTECTS_DATA(elan3_ctxt::SwapListsLock, 
			  elan3_ctxt::ItemCount))
_NOTE(RWLOCK_PROTECTS_DATA(elan3_ctxt::VpLock, 
			   elan3_ctxt::VpSegs elan3_vpseg::Next elan3_vpseg::Process 
			   elan3_vpseg::Entries elan3_vpseg::Type))

_NOTE(DATA_READABLE_WITHOUT_LOCK(elan3_ctxt::ItemCount elan3_ctxt::Status elan3_ctxt::CommandPortItem))

_NOTE(LOCK_ORDER(elan3_ctxt::SwapListsLock elan3_ctxt::CmdLock elan3_dev::IntrLock))
_NOTE(LOCK_ORDER(elan3_ctxt::SwapListsLock as::a_lock))			/* implicit by pagefault */

#define CTXT_DETACHED				(1 << 0)		/* Context is detached. */
#define CTXT_NO_LWPS				(1 << 1)		/* No lwp's to handle faults */
#define CTXT_EXITING				(1 << 2)		/* User process is exiting */

#define CTXT_SWAPPING_OUT			(1 << 3)		/* Context is swapping out */
#define CTXT_SWAPPED_OUT			(1 << 4)		/* Context is swapped out */

#define CTXT_SWAP_FREE				(1 << 5)		/* Swap buffer is free */
#define CTXT_SWAP_VALID				(1 << 6)		/* Swap buffer has queue entries in it */

#define CTXT_DMA_QUEUE_FULL			(1 << 7)		/* Dma trap queue is full */
#define CTXT_THREAD_QUEUE_FULL			(1 << 8)		/* Thread trap queue is full */
#define CTXT_EVENT_QUEUE_FULL			(1 << 9)		/* Event interrupt queue is full */
#define CTXT_COMMAND_OVERFLOW_ERROR		(1 << 10)		/* Trap queue overflow */

#define CTXT_SWAP_WANTED			(1 << 11)		/* Some one wanted to swap */
#define CTXT_WAITING_SWAPIN			(1 << 12)		/* Someone waiting on swapin */

#define CTXT_WAITING_COMMAND			(1 << 13)		/* swgelan waiting on command port */
#define CTXT_COMMAND_MAPPED_MAIN		(1 << 14)		/* segelan has mapped command port */

#define CTXT_QUEUES_EMPTY			(1 << 15)		/* dma/thread run queues are empty */
#define CTXT_QUEUES_EMPTYING			(1 << 16)		/* dma/thread run queues are being emptied */

#define CTXT_USER_FILTERING			(1 << 17)		/* user requested context filter */

#define CTXT_KERNEL				(1 << 18)		/* context is a kernel context */
#define CTXT_COMMAND_MAPPED_ELAN		(1 << 19)		/* command port is mapped for elan */
#define CTXT_FIXUP_NETERR			(1 << 20)		/* fixing up a network error */


#define CTXT_SWAPPED_REASONS		(CTXT_NO_LWPS	|		\
					 CTXT_DETACHED	|		\
					 CTXT_EXITING   |		\
					 CTXT_FIXUP_NETERR)

#define CTXT_OTHERS_REASONS		(CTXT_EVENT_QUEUE_FULL  |	\
					 CTXT_DMA_QUEUE_FULL    |	\
					 CTXT_THREAD_QUEUE_FULL |	\
					 CTXT_COMMAND_OVERFLOW_ERROR |	\
					 CTXT_SWAPPED_REASONS)

#define CTXT_INPUTTER_REASONS		(CTXT_USER_FILTERING |		\
					 CTXT_OTHERS_REASONS)

#define CTXT_COMMAND_MAPPED		(CTXT_COMMAND_MAPPED_MAIN |	\
					 CTXT_COMMAND_MAPPED_ELAN)

#define CTXT_IS_KERNEL(ctxt)		((ctxt)->Status & CTXT_KERNEL)

/*
 * State values for ctxt_inputterState/ctxt_commandportStats
 */
#define CTXT_STATE_OK			0
#define CTXT_STATE_TRAPPED		1		/* Inputter channel 0 trapped */
#define CTXT_STATE_RESOLVING		2		/* An LWP is resolving the trap */
#define CTXT_STATE_NEEDS_RESTART	3		/* Th trapped packet needs to be executed */
#define CTXT_STATE_NETWORK_ERROR	4		/* We're waiting on an RPC for the identify transaction */
#define CTXT_STATE_EXECUTING		5		/* An LWP is executing the trapped packet */

/*
 * State values for OthersState.
 */
#define CTXT_OTHERS_RUNNING		0
#define CTXT_OTHERS_HALTING		1
#define CTXT_OTHERS_SWAPPING		2
#define CTXT_OTHERS_HALTING_MORE	3
#define CTXT_OTHERS_SWAPPING_MORE	4
#define CTXT_OTHERS_SWAPPED		5

typedef struct elan3_ops
{
    u_int  Version;

    int	 (*Exception)	(ELAN3_CTXT *ctxt, int type, int proc, void *trap, va_list ap);

    /* swap item list functions */
    int  (*GetWordItem)		(ELAN3_CTXT *ctxt, int list, void **itemp, E3_uint32 *valuep);
    int  (*GetBlockItem)	(ELAN3_CTXT *ctxt, int list, void **itemp, E3_Addr *valuep);
    void (*PutWordItem)		(ELAN3_CTXT *ctxt, int list, E3_Addr value);
    void (*PutBlockItem)	(ELAN3_CTXT *ctxt, int list, E3_uint32 *ptr);
    void (*PutbackItem)		(ELAN3_CTXT *ctxt, int list, void *item);
    void (*FreeWordItem)	(ELAN3_CTXT *ctxt, void *item);
    void (*FreeBlockItem)	(ELAN3_CTXT *ctxt, void *item);
    int  (*CountItems)		(ELAN3_CTXT *ctxt, int list);

    /* event interrupt cookie */
    int  (*Event)		(ELAN3_CTXT *ctxt, E3_uint32 cookie, int flag);

    /* swapin/swapout functions. */
    void (*Swapin)		(ELAN3_CTXT *ctxt);
    void (*Swapout)		(ELAN3_CTXT *ctxt);

    /* Free of private data */
    void (*FreePrivate)		(ELAN3_CTXT *ctxt);

    /* Fixup a network error */
    int  (*FixupNetworkError)	(ELAN3_CTXT *ctxt, NETERR_FIXUP *nef);

    /* Interrupt handler trap interface */
    int  (*DProcTrap)		(ELAN3_CTXT *ctxt, DMA_TRAP *trap);
    int  (*TProcTrap)		(ELAN3_CTXT *ctxt, THREAD_TRAP *trap);
    int	 (*IProcTrap)		(ELAN3_CTXT *ctxt, INPUT_TRAP *trap, int chan);
    int	 (*CProcTrap)		(ELAN3_CTXT *ctxt, COMMAND_TRAP *trap);
    int  (*CProcReissue)        (ELAN3_CTXT *ctxt, CProcTrapBuf_BE *TrapBuf);

    /* User memory access functions */
    int	      (*StartFaultCheck)(ELAN3_CTXT *ctxt);
    void      (*EndFaultCheck)  (ELAN3_CTXT *ctxt);

    E3_uint8  (*Load8)		(ELAN3_CTXT *ctxt, E3_Addr addr);
    void      (*Store8)		(ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint8 val);
    E3_uint16 (*Load16)		(ELAN3_CTXT *ctxt, E3_Addr addr);
    void      (*Store16)	(ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint16 val);
    E3_uint32 (*Load32)		(ELAN3_CTXT *ctxt, E3_Addr addr);
    void      (*Store32)	(ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint32 val);
    E3_uint64 (*Load64)		(ELAN3_CTXT *ctxt, E3_Addr addr);
    void      (*Store64)	(ELAN3_CTXT *ctxt, E3_Addr addr, E3_uint64 val);

} ELAN3_OPS;

#define ELAN3_OPS_VERSION	0xdeef0001

/*
 * Flags for ops_event.
 */
#define OP_INTR			0		/* Called from interrupt handler */
#define OP_LWP			1		/* Called from "lwp" */

/*
 * Return codes for "ops" functions.
 */
#define OP_DEFER		0		/* Defer to next lower interrupt */
#define OP_IGNORE		1		/* No event hander, so ignore it */
#define OP_HANDLED		2		/* Handled event (resume thread) */
#define OP_FAILED		3		/* Failed */

#define ELAN3_CALL_OP(ctxt,fn)				((ctxt)->Operations && (ctxt)->Operations->fn) ? (ctxt)->Operations->fn 

#define ELAN3_OP_EXCEPTION(ctxt,type,proc,trap,ap)	(ELAN3_CALL_OP(ctxt,Exception)    (ctxt,type,proc,trap,ap)	: OP_IGNORE)
#define ELAN3_OP_GET_WORD_ITEM(ctxt,list,itemp,valuep)	(ELAN3_CALL_OP(ctxt,GetWordItem)  (ctxt,list,itemp,valuep)	: 0)
#define ELAN3_OP_GET_BLOCK_ITEM(ctxt,list,itemp,valuep)	(ELAN3_CALL_OP(ctxt,GetBlockItem) (ctxt,list,itemp,valuep)	: 0)
#define ELAN3_OP_PUT_WORD_ITEM(ctxt,list,value)		(ELAN3_CALL_OP(ctxt,PutWordItem)  (ctxt,list,value)		: (void)0)
#define ELAN3_OP_PUT_BLOCK_ITEM(ctxt,list,ptr)		(ELAN3_CALL_OP(ctxt,PutBlockItem) (ctxt,list,ptr)		: (void)0)
#define ELAN3_OP_PUTBACK_ITEM(ctxt,list,item)		(ELAN3_CALL_OP(ctxt,PutbackItem)  (ctxt,list,item)		: (void)0)
#define ELAN3_OP_FREE_WORD_ITEM(ctxt,item)		(ELAN3_CALL_OP(ctxt,FreeWordItem) (ctxt,item)			: (void)0)
#define ELAN3_OP_FREE_BLOCK_ITEM(ctxt,item)		(ELAN3_CALL_OP(ctxt,FreeBlockItem)(ctxt,item)			: (void)0)
#define ELAN3_OP_COUNT_ITEMS(ctxt,list)			(ELAN3_CALL_OP(ctxt,CountItems)(ctxt,list)			: 0)
#define ELAN3_OP_EVENT(ctxt,cookie,flag)		(ELAN3_CALL_OP(ctxt,Event)(ctxt,cookie,flag)			: OP_IGNORE)
#define ELAN3_OP_SWAPIN(ctxt)				(ELAN3_CALL_OP(ctxt,Swapin)(ctxt)				: (void)0)
#define ELAN3_OP_SWAPOUT(ctxt)				(ELAN3_CALL_OP(ctxt,Swapout)(ctxt)				: (void)0)
#define ELAN3_OP_FREE_PRIVATE(ctxt)			(ELAN3_CALL_OP(ctxt,FreePrivate)(ctxt)				: (void)0)
#define ELAN3_OP_FIXUP_NETWORK_ERROR(ctxt, nef)		(ELAN3_CALL_OP(ctxt,FixupNetworkError)(ctxt,nef)			: OP_FAILED)

#define ELAN3_OP_DPROC_TRAP(ctxt, trap)			(ELAN3_CALL_OP(ctxt,DProcTrap)(ctxt,trap)			: OP_DEFER)
#define ELAN3_OP_TPROC_TRAP(ctxt, trap)			(ELAN3_CALL_OP(ctxt,TProcTrap)(ctxt,trap)			: OP_DEFER)
#define ELAN3_OP_IPROC_TRAP(ctxt, trap, chan)		(ELAN3_CALL_OP(ctxt,IProcTrap)(ctxt,trap,chan)			: OP_DEFER)
#define ELAN3_OP_CPROC_TRAP(ctxt, trap)			(ELAN3_CALL_OP(ctxt,CProcTrap)(ctxt,trap)			: OP_DEFER)
#define ELAN3_OP_CPROC_REISSUE(ctxt,tbuf)		(ELAN3_CALL_OP(ctxt,CProcReissue)(ctxt, tbuf)			: OP_DEFER)

#define ELAN3_OP_START_FAULT_CHECK(ctxt)		(ELAN3_CALL_OP(ctxt,StartFaultCheck)(ctxt)			: 0)
#define ELAN3_OP_END_FAULT_CHECK(ctxt)			(ELAN3_CALL_OP(ctxt,EndFaultCheck)(ctxt)				: (void)0)
#define ELAN3_OP_LOAD8(ctxt,addr)			(ELAN3_CALL_OP(ctxt,Load8)(ctxt,addr)				: 0)
#define ELAN3_OP_STORE8(ctxt,addr,val)			(ELAN3_CALL_OP(ctxt,Store8)(ctxt,addr,val)			: (void)0)
#define ELAN3_OP_LOAD16(ctxt,addr)			(ELAN3_CALL_OP(ctxt,Load16)(ctxt,addr)				: 0)
#define ELAN3_OP_STORE16(ctxt,addr,val)			(ELAN3_CALL_OP(ctxt,Store16)(ctxt,addr,val)			: (void)0)
#define ELAN3_OP_LOAD32(ctxt,addr)			(ELAN3_CALL_OP(ctxt,Load32)(ctxt,addr)				: 0)
#define ELAN3_OP_STORE32(ctxt,addr,val)			(ELAN3_CALL_OP(ctxt,Store32)(ctxt,addr,val)			: (void)0)
#define ELAN3_OP_LOAD64(ctxt,addr)			(ELAN3_CALL_OP(ctxt,Load64)(ctxt,addr)				: 0)
#define ELAN3_OP_STORE64(ctxt,addr,val)			(ELAN3_CALL_OP(ctxt,Store64)(ctxt,addr,val)			: (void)0)

#endif /* __KERNEL__ */

/* "list" arguement to ops functions */
#define LIST_DMA_PTR		0
#define LIST_DMA_DESC		1
#define LIST_THREAD    		2
#define LIST_COMMAND		3
#define LIST_SETEVENT		4
#define LIST_FREE_WORD		5
#define LIST_FREE_BLOCK		6

#define MAX_LISTS		7

#if defined(__KERNEL__) && MAX_LISTS != NUM_LISTS
#  error Check NUM_LISTS == MAX_LISTS
#endif

/*
 * Values for the 'type' field to PostException().
 */
#define EXCEPTION_INVALID_ADDR		1		/* FaultArea, res */
#define EXCEPTION_UNIMP_INSTR		2		/* instr */
#define EXCEPTION_INVALID_PROCESS	3		/* proc, res */
#define EXCEPTION_SIMULATION_FAILED	4		/* */
#define EXCEPTION_UNIMPLEMENTED		5		/* */
#define EXCEPTION_SWAP_FAULT		6		/* */
#define EXCEPTION_SWAP_FAILED		7		/* */
#define EXCEPTION_BAD_PACKET		8		/* */
#define EXCEPTION_FAULTED		9		/* addr */
#define EXCEPTION_QUEUE_OVERFLOW	10		/* FaultArea, TrapType */
#define EXCEPTION_COMMAND_OVERFLOW	11		/* count */
#define EXCEPTION_DMA_RETRY_FAIL	12		/* */
#define EXCEPTION_CHAINED_EVENT		13		/* EventAddr */
#define EXCEPTION_THREAD_KILLED		14		/* */
#define EXCEPTION_CANNOT_SAVE_THREAD	15
#define EXCEPTION_BAD_SYSCALL		16		/* */
#define EXCEPTION_DEBUG			17
#define EXCEPTION_BAD_EVENT		18		/* */
#define EXCEPTION_NETWORK_ERROR		19		/* rvp */
#define EXCEPTION_BUS_ERROR		20
#define EXCEPTION_COOKIE_ERROR		21
#define EXCEPTION_PACKET_TIMEOUT	22
#define EXCEPTION_BAD_DMA		23		/* */
#define EXCEPTION_ENOMEM		24

/*
 * Values for the 'proc' field to ElanException().
 */
#define COMMAND_PROC			1
#define THREAD_PROC			2
#define DMA_PROC			3
#define INPUT_PROC			4
#define EVENT_PROC			5

/* Flags to IssueDmaCommand */
#define ISSUE_COMMAND_FOR_CPROC		1
#define ISSUE_COMMAND_CANT_WAIT		2

/* Return code from IssueDmaCommand.*/
#define ISSUE_COMMAND_OK		0
#define ISSUE_COMMAND_TRAPPED		1
#define ISSUE_COMMAND_RETRY		2
#define ISSUE_COMMAND_WAIT		3

#ifdef __KERNEL__

extern ELAN3_CTXT *elan3_alloc(ELAN3_DEV *dev, int kernel);
extern void	   elan3_free      (ELAN3_CTXT *ctxt);

extern int	   elan3_attach    (ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap);
extern int         elan3_doattach  (ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap);
extern void	   elan3_detach    (ELAN3_CTXT *ctxt);
extern void        elan3_dodetach  (ELAN3_CTXT *ctxt);

extern int	   elan3_addvp     (ELAN3_CTXT *ctxt, int process, ELAN_CAPABILITY *cap);
extern int	   elan3_removevp  (ELAN3_CTXT *ctxt, int process);
extern int	   elan3_addbcastvp(ELAN3_CTXT *ctxt, int process, int base, int count);

extern int         elan3_process   (ELAN3_CTXT *ctxt);

extern int	   elan3_load_route (ELAN3_CTXT *ctxt, int process, E3_uint16 *flits);
extern int	   elan3_check_route(ELAN3_CTXT *ctxt, int process, E3_uint16 *flits, E3_uint32 *routeError);

extern int	   elan3_lwp       (ELAN3_CTXT *ctxt);

extern void	   elan3_swapin (ELAN3_CTXT *ctxt, int reason);
extern void	   elan3_swapout (ELAN3_CTXT *ctxt, int reason);
extern int         elan3_pagefault (ELAN3_CTXT *ctxt, E3_FaultSave_BE *FaultSave, int npages);
extern void        elan3_block_inputter (ELAN3_CTXT *ctxt, int block);


extern E3_Addr     elan3_init_thread (ELAN3_DEV *dev, E3_Addr fn, E3_Addr addr, sdramaddr_t stack, int stackSize, int nargs, ...);

extern void	   SetInputterState (ELAN3_CTXT *ctxt, E3_uint32 Pend, E3_uint32 *Maskp);
extern void	   SetInputterStateForContext (ELAN3_CTXT *ctxt, E3_uint32 Pend, E3_uint32 *Maskp);
extern void        UnloadCommandPageMapping (ELAN3_CTXT *ctxt);
extern void	   StartSwapoutContext (ELAN3_CTXT *ctxt, E3_uint32 Pend, E3_uint32 *Maskp);

extern int	   HandleExceptions (ELAN3_CTXT *ctxt, unsigned long *flags);
extern int	   RestartContext (ELAN3_CTXT *ctxt, unsigned long *flags);
extern int         CheckCommandQueueFlushed (ELAN3_CTXT *ctxt, E3_uint32 cflags, int how, unsigned long *flags);
extern int	   IssueCommand (ELAN3_CTXT *ctxt, unsigned cmdoff, E3_Addr value, int flags);
extern int	   IssueDmaCommand (ELAN3_CTXT *ctxt, E3_Addr value, void *item, int flags);
extern int         WaitForDmaCommand (ELAN3_CTXT *ctxt, void *item, int flags);
extern void	   FixupEventTrap (ELAN3_CTXT *ctxt, int proc, void *trap, E3_uint32 TrapType, 
				   E3_FaultSave_BE *FaultSaveArea, int flags);
extern int	   SimulateBlockCopy (ELAN3_CTXT *ctxt, E3_Addr EventAddress);
extern void	   ReissueEvent (ELAN3_CTXT *ctxt, E3_Addr addr,int flags);
extern int         SetEventsNeedRestart (ELAN3_CTXT *ctxt);
extern void        RestartSetEvents (ELAN3_CTXT *ctxt);
extern int	   RunEventType (ELAN3_CTXT *ctxt, E3_FaultSave_BE *FaultSaveArea, E3_uint32 EventType);
extern void        WakeupLwp (ELAN3_DEV *dev, void *arg);
extern void	   QueueEventInterrupt (ELAN3_CTXT *ctxt, E3_uint32 cookie);
extern int         WaitForCommandPort (ELAN3_CTXT *ctxt);

extern int 	   ElanException (ELAN3_CTXT *ctxt, int type, int proc, void *trap, ...);

/* context_osdep.c */
extern int	   LoadElanTranslation (ELAN3_CTXT *ctxt, E3_Addr elanAddr, int len, int protFault, int writeable);
extern void	   LoadCommandPortTranslation (ELAN3_CTXT *ctxt);

#if defined(DIGITAL_UNIX)
/* seg_elan.c */
extern caddr_t	   elan3_segelan3_create (ELAN3_CTXT *ctxt);
extern void	   elan3_segelan3_destroy (ELAN3_CTXT *ctxt);
extern int         elan3_segelan3_map (ELAN3_CTXT *ctxt);
extern void        elan3_segelan3_unmap (ELAN3_CTXT *ctxt);

/* seg_elanmem.c */
extern int	   elan3_segelanmem_create (ELAN3_DEV *dev, unsigned object, unsigned off, vm_offset_t *addrp, int len);
#endif /* defined(DIGITAL_UNIX) */

/* route_table.c */
extern ELAN3_ROUTE_TABLE *AllocateRouteTable (ELAN3_DEV *dev, int size);
extern void              FreeRouteTable  (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl);
extern int               LoadRoute       (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int vp, int ctxnum, int nflits, E3_uint16 *flits);
extern int               GetRoute        (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int process, E3_uint16 *flits);
extern void	         InvalidateRoute (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int vp);
extern void	         ValidateRoute   (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int vp);
extern void	         ClearRoute      (ELAN3_DEV *dev, ELAN3_ROUTE_TABLE *tbl, int vp);

extern int               GenerateRoute (ELAN_POSITION *pos, E3_uint16 *flits, int lowid, int highid, int timeout, int highPri);
extern int               GenerateProbeRoute (E3_uint16 *flits, int nodeid, int level, int *linkup, int *linkdown, int adaptive);
extern int               GenerateCheckRoute (ELAN_POSITION *pos, E3_uint16 *flits, int level, int adaptive);

/* virtual_process.c */
extern ELAN_LOCATION  ProcessToLocation     (ELAN3_CTXT *ctxt, ELAN3_VPSEG *seg, int process, ELAN_CAPABILITY *cap);
extern int	      ResolveVirtualProcess (ELAN3_CTXT *ctxt, int process);
extern caddr_t        CapabilityString      (ELAN_CAPABILITY *cap);
extern void           UnloadVirtualProcess  (ELAN3_CTXT *ctxt, ELAN_CAPABILITY *cap);

extern int           elan3_get_route   (ELAN3_CTXT *ctxt, int process, E3_uint16 *flits);
extern int           elan3_reset_route (ELAN3_CTXT *ctxt, int process);

/* cproc.c */
extern int	  NextCProcTrap (ELAN3_CTXT *ctxt, COMMAND_TRAP *trap);
extern void	  ResolveCProcTrap (ELAN3_CTXT *ctxt);
extern int	  RestartCProcTrap (ELAN3_CTXT *ctxt);

/* iproc.c */
extern void       InspectIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap);
extern void	  ResolveIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, NETERR_RESOLVER **rvp);
extern int	  RestartIProcTrap (ELAN3_CTXT *ctxt, INPUT_TRAP *trap);
extern char      *IProcTrapString (E3_IprocTrapHeader_BE *hdrp, E3_IprocTrapData *datap);
extern void       SimulateUnlockQueue (ELAN3_CTXT *ctxt, E3_Addr QueuePointer, int SentAck);

/* tproc.c */
extern int	  NextTProcTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap);
extern void	  ResolveTProcTrap (ELAN3_CTXT *ctxt, THREAD_TRAP *trap);
extern int	  TProcNeedsRestart (ELAN3_CTXT *ctxt);
extern void	  RestartTProcItems (ELAN3_CTXT *ctxt);
extern E3_Addr    SaveThreadToStack (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int SkipInstruction);
extern void       ReissueStackPointer (ELAN3_CTXT *ctxt, E3_Addr StackPointer);

/* tprocinsts.c */
extern int        RollThreadToClose (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, E3_uint32 PAckVal);

/* tproc_osdep.c */
extern int        ThreadSyscall (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int *skip);
extern int	  ThreadElancall (ELAN3_CTXT *ctxt, THREAD_TRAP *trap, int *skip);

/* dproc.c */
extern int	  NextDProcTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap);
extern void	  ResolveDProcTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap);
extern int	  DProcNeedsRestart (ELAN3_CTXT *ctxt);
extern void	  RestartDProcItems (ELAN3_CTXT *ctxt);
extern void       RestartDmaDesc (ELAN3_CTXT *ctxt, E3_DMA_BE *desc);
extern void       RestartDmaTrap (ELAN3_CTXT *ctxt, DMA_TRAP *trap);
extern void	  RestartDmaPtr (ELAN3_CTXT *ctxt, E3_Addr ptr);

/* network_error.c */
extern void       InitialiseNetworkErrorResolver (void);
extern void       FinaliseNetworkErrorResolver (void);
extern int        QueueNetworkErrorResolver (ELAN3_CTXT *ctxt, INPUT_TRAP *trap, NETERR_RESOLVER **rvpp);
extern void	  FreeNetworkErrorResolver (NETERR_RESOLVER *rvp);
extern void       CancelNetworkErrorResolver (NETERR_RESOLVER *rvp);
extern int	  ExecuteNetworkErrorFixup (NETERR_MSG *msg);
extern void	  CompleteNetworkErrorFixup (ELAN3_CTXT *ctxt, NETERR_FIXUP *nef, int status);

extern int        AddNeterrServerSyscall (int elanId, void *configp, void *addrp, char *namep);

/* eventcookie.c */
extern void                cookie_init(void);
extern void                cookie_fini(void);
extern EVENT_COOKIE_TABLE *cookie_alloc_table (unsigned long task, unsigned long handle);
extern void                cookie_free_table (EVENT_COOKIE_TABLE *tbl);
extern int                 cookie_alloc_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie);
extern int                 cookie_free_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie);
extern int                 cookie_fire_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie);
extern int                 cookie_wait_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie);
extern int                 cookie_arm_cookie (EVENT_COOKIE_TABLE *tbl, EVENT_COOKIE cookie);

/* routecheck.c */
extern int elan3_route_check          (ELAN3_CTXT *ctxt, E3_uint16 *flits, int destNode);
extern int elan3_route_broadcast_check(ELAN3_CTXT *ctxt, E3_uint16 *flitsA, int lowNode, int highNode);


#endif /* __KERNEL__ */

#ifdef __cplusplus
}
#endif

#endif /* _ELAN3_ELANCTXT_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
