/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_ELANDEV_H
#define __ELAN3_ELANDEV_H

#ident "$Id: elandev.h,v 1.76.2.1 2005/07/20 11:35:20 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elandev.h,v $ */

#include <elan/bitmap.h>
#include <elan/devinfo.h>
#include <elan/stats.h>

#if defined(DIGITAL_UNIX)
#  include <elan3/elandev_dunix.h>
#elif defined(LINUX)
#  include <elan3/elandev_linux.h>
#elif defined(SOLARIS)
#  include <elan3/elandev_solaris.h>
#endif

#ifndef TRUE
#  define TRUE 1
#endif
#ifndef FALSE
#  define FALSE 0
#endif

/*
 * Elan base address registers defined as follows :
 */
#define ELAN3_BAR_SDRAM		0
#define ELAN3_BAR_COMMAND_PORT	1
#define ELAN3_BAR_REGISTERS	2
#define ELAN3_BAR_EBUS		3

/* Macro to generate 'offset' to mmap "mem" device */
#define OFF_TO_SPACE(off)	((off) >> 28)
#define OFF_TO_OFFSET(off)	((off) & 0x0FFFFFFF)
#define GEN_OFF(space,off)	(((space) << 28) | ((off) & 0x0FFFFFFF))

#ifdef __KERNEL__

/*
 * Elan EBUS is configured as follows :
 */
#define ELAN3_EBUS_ROM_OFFSET		0x000000		/* rom */
#define ELAN3_EBUS_INTPAL_OFFSET	0x180000		/* interrupt pal (write only) */

#define ELAN3_EBUS_ROM_SIZE		0x100000

/*
 * Elan SDRAM is arranged as follows :
 */
#define ELAN3_TANDQ_SIZE		0x0020000		/* Trap And Queue Size */
#define ELAN3_CONTEXT_SIZE		0x0010000		/* Context Table Size */
#define ELAN3_COMMAND_TRAP_SIZE		0x0010000		/* Command Port Trap Size */

#ifdef MPSAS
#define ELAN3_LN2_NUM_CONTEXTS	8				/* Support 256 contexts */
#else
#define ELAN3_LN2_NUM_CONTEXTS	12				/* Support 4096 contexts */
#endif
#define ELAN3_NUM_CONTEXTS	(1 << ELAN3_LN2_NUM_CONTEXTS)	/* Entries in context table */

#define ELAN3_SDRAM_NUM_BANKS	4				/* Elan supports 4 Banks of Sdram */
#define ELAN3_SDRAM_BANK_SHIFT	26				/* each of which can be 64 mbytes ? */
#define ELAN3_SDRAM_BANK_SIZE	(1 << ELAN3_SDRAM_BANK_SHIFT)

#define ELAN3_MAX_CACHE_SIZE	(64 * 1024)			/* Maximum cache size */
#define ELAN3_CACHE_SIZE	(64 * 4 * E3_CACHELINE_SIZE)	/* Elan3 has 8K cache */

#ifndef offsetof
#define offsetof(s, m)		(size_t)(&(((s *)0)->m))
#endif

/*
 * circular queue and macros to access members.
 */
typedef struct
{
    u_int	q_back;			/* Next free space */
    u_int	q_front;		/* First object to remove */
    u_int	q_size;			/* Size of queue */
    u_int	q_count;		/* Current number of entries */
    u_int	q_slop;			/* FULL <=> (count+slop) == size */
} ELAN3_QUEUE;

typedef struct 
{
    u_int	q_back;			/* Next free space */
    u_int	q_middle;		/* Middle pointer */
    u_int	q_front;		/* First object to remove */
    u_int	q_size;			/* Size of queue */
    u_int	q_count;		/* Current number of entries */
    u_int	q_slop;			/* FULL <=> (count+slop) == size */
} ELAN3_SPLIT_QUEUE;

#define ELAN3_QUEUE_INIT(q,num,slop)	((q).q_size = (num), (q).q_slop = (slop)+1, (q).q_front = (q).q_back = 0, (q).q_count = 0)
#define ELAN3_QUEUE_FULL(q)		((q).q_count == ((q).q_size - (q).q_slop))
#define ELAN3_QUEUE_REALLY_FULL(q)	((q).q_count == (q).q_size - 1)
#define ELAN3_QUEUE_EMPTY(q)		((q).q_count == 0)
#define ELAN3_QUEUE_FRONT_EMPTY(q)	((q).q_front == (q).q_middle)
#define ELAN3_QUEUE_BACK_EMPTY(q)	((q).q_middle == (q).q_back)
#define ELAN3_QUEUE_ADD(q)		((q).q_back = ((q).q_back+1) % (q).q_size, (q).q_count++)
#define ELAN3_QUEUE_REMOVE(q)		((q).q_front = ((q).q_front+1) % (q).q_size, (q).q_count--)
#define ELAN3_QUEUE_ADD_FRONT(q)		((q).q_front = ((q).q_front-1) % (q).q_size, (q).q_count++)
#define ELAN3_QUEUE_CONSUME(q)		((q).q_middle = ((q).q_middle+1) % (q).q_size)
#define ELAN3_QUEUE_FRONT(q,qArea)	(&(qArea)[(q).q_front])
#define ELAN3_QUEUE_MIDDLE(q,qArea)	(&(qArea)[(q).q_middle])
#define ELAN3_QUEUE_BACK(q,qArea)	(&(qArea)[(q).q_back])

#define SDRAM_MIN_BLOCK_SHIFT	10
#define SDRAM_NUM_FREE_LISTS	17				/* allows max 64Mb block */
#define SDRAM_MIN_BLOCK_SIZE	(1 << SDRAM_MIN_BLOCK_SHIFT)
#define SDRAM_MAX_BLOCK_SIZE	(SDRAM_MIN_BLOCK_SIZE << (SDRAM_NUM_FREE_LISTS-1))
#define SDRAM_FREELIST_TRIGGER	32

typedef struct elan3_sdram_bank
{
    u_int		Size;					/* Size of bank of memory */

    ioaddr_t		Mapping;				/* Where mapped in the kernel */
    DeviceMappingHandle Handle;					/* and mapping handle */

    struct elan3_ptbl_gr **PtblGroups;
    
    bitmap_t	       *Bitmaps[SDRAM_NUM_FREE_LISTS];
} ELAN3_SDRAM_BANK;

typedef struct elan3_haltop
{
    struct elan3_haltop	  *Next;				/* Chain to next in list. */
    E3_uint32		   Mask;				/* Interrupt mask to see before calling function */
    
    void		 (*Function)(void *, void *);		/* Function to call */
    void		  *Arguement;				/* Arguement to pass to function */
} ELAN3_HALTOP;

#define HALTOP_BATCH	32

#endif /* __KERNEL__ */

typedef struct elan3_stats
{
    u_long	Version;					/* version field */
    u_long	Interrupts;				        /* count of elan interrupts */
    u_long	TlbFlushes;					/* count of tlb flushes */
    u_long	InvalidContext;					/* count of traps with invalid context */
    u_long	ComQueueHalfFull;				/* count of interrupts due to com queue being half full */

    u_long	CProcTraps;					/* count of cproc traps */
    u_long	DProcTraps;					/* count of dproc traps */
    u_long	TProcTraps;					/* cound of tproc traps */
    u_long	IProcTraps;					/* count of iproc traps */
    u_long	EventInterrupts;				/* count of event interrupts */

    u_long	PageFaults;					/* count of elan page faults */

    /* inputter related */
    u_long	EopBadAcks;					/* count of EOP_BAD_ACKs */
    u_long	EopResets;					/* count of EOP_ERROR_RESET */
    u_long      InputterBadLength;				/* count of BadLength */
    u_long      InputterCRCDiscards;				/* count of CRC_STATUS_DISCARD */
    u_long      InputterCRCErrors;				/* count of CRC_STATUS_ERROR */
    u_long      InputterCRCBad;					/* count of CRC_STATUS_BAD */
    u_long	DmaNetworkErrors;				/* count of errors in dma data */
    u_long	DmaIdentifyNetworkErrors;			/* count of errors after dma identify */
    u_long	ThreadIdentifyNetworkErrors;			/* count of errors after thread identify */

    /* dma related */
    u_long	DmaRetries;					/* count of dma retries (due to retry fail count) */    
    u_long	DmaOutputTimeouts;				/* count of dma output timeouts */
    u_long	DmaPacketAckErrors;				/* count of dma packet ack errors */

    /* thread related */
    u_long	ForcedTProcTraps;				/* count of forced tproc traps */
    u_long	TrapForTooManyInsts;				/* count of too many instruction traps */
    u_long	ThreadOutputTimeouts;				/* count of thread output timeouts */
    u_long       ThreadPacketAckErrors;				/* count of thread packet ack errors */

    /* link related */
    u_long	LockError;					/* count of RegPtr->Exts.LinkErrorTypes:LS_LockError */
    u_long	DeskewError;					/* count of RegPtr->Exts.LinkErrorTypes:LS_DeskewError */
    u_long	PhaseError;					/* count of RegPtr->Exts.LinkErrorTypes:LS_PhaseError */
    u_long	DataError;					/* count of RegPtr->Exts.LinkErrorTypes:LS_DataError */
    u_long	FifoOvFlow0;					/* count of RegPtr->Exts.LinkErrorTypes:LS_FifoOvFlow0 */
    u_long	FifoOvFlow1;					/* count of RegPtr->Exts.LinkErrorTypes:LS_FifoOvFlow1 */
    u_long	LinkErrorValue;					/* link error value on data error */

    /* memory related */
    u_long	CorrectableErrors;				/* count of correctable ecc errors */
    u_long	UncorrectableErrors;				/* count of uncorrectable ecc errors */
    u_long       MultipleErrors;					/* count of multiple ecc errors */
    u_long	SdramBytesFree;					/* count of sdram bytes free */
    
    /* Interrupt related */
    u_long	LongestInterrupt;				/* length of longest interrupt in ticks */

    u_long	EventPunts;					/* count of punts of event interrupts to thread */
    u_long	EventRescheds;					/* count of reschedules of event interrupt thread */
} ELAN3_STATS;

#define ELAN3_STATS_VERSION	(ulong)2
#define ELAN3_NUM_STATS		(sizeof (ELAN3_STATS)/sizeof (u_long))

#define ELAN3_STATS_DEV_FMT   "elan3_stats_dev_%d"

#ifdef __KERNEL__

#define BumpStat(dev,stat)	((dev)->Stats.stat++)

typedef struct elan3_level_ptbl_block
{
    spinlock_t		    PtblLock;			        /* Page table freelist lock */
    int			    PtblTotal;			        /* Count of level N page tables allocated */
    int		            PtblFreeCount;		        /* Count of free level N page tables */
    struct elan3_ptbl	   *PtblFreeList;			/* Free level N page tables */
    struct elan3_ptbl_gr	   *PtblGroupList;	         	/* List of Groups of level N page tables */
} ELAN3_LEVEL_PTBL_BLOCK;
 
typedef struct elan3_dev
{
    ELAN3_DEV_OSDEP	    Osdep;				/* OS specific entries */
    int		       	    Instance;				/* Device number */
    ELAN_DEVINFO            Devinfo;	                        
    ELAN_POSITION	    Position;				/* position in switch network (for user code) */
    ELAN_DEV_IDX	    DeviceIdx;				/* device index registered with elanmod */

    int			    ThreadsShouldStop;			/* flag that kernel threads should stop */

    spinlock_t		    IntrLock;
    spinlock_t		    TlbLock;
    spinlock_t		    CProcLock;
    kcondvar_t		    IntrWait;				/* place event interrupt thread sleeps */
    unsigned		    EventInterruptThreadStarted:1;	/* event interrupt thread started */
    unsigned		    EventInterruptThreadStopped:1;	/* event interrupt thread stopped */
    
    DeviceMappingHandle	    RegHandle;				/* DDI Handle */
    ioaddr_t	            RegPtr;				/* Elan Registers */

    volatile E3_uint32	    InterruptMask;			/* copy of RegPtr->InterruptMask */
    volatile E3_uint32	    Event_Int_Queue_FPtr;		/* copy of RegPtr->Event_Int_Queue_FPtr */
    volatile E3_uint32      SchCntReg;				/* copy of RegPtr->SchCntReg */
    volatile E3_uint32      Cache_Control_Reg;			/* true value for RegPtr->Cache_Control_Reg */
    
    ELAN3_SDRAM_BANK	    SdramBanks[ELAN3_SDRAM_NUM_BANKS];	/* Elan sdram banks */
    spinlock_t		    SdramLock;				/* Sdram allocator */
    sdramaddr_t		    SdramFreeLists[SDRAM_NUM_FREE_LISTS];
    unsigned		    SdramFreeCounts[SDRAM_NUM_FREE_LISTS];
		
    sdramaddr_t		    TAndQBase;				/* Trap and Queue area */
    sdramaddr_t		    ContextTable;			/* Elan Context Table */
    u_int		    ContextTableSize;			/* # entries in context table */

    struct elan3_ctxt      **CtxtTable;                         /* array of ctxt pointers or nulls */

    sdramaddr_t		    CommandPortTraps[2];		/* Command port trap overflow */
    int			    CurrentCommandPortTrap;		/* Which overflow queue we're using */
    
    u_int		    HaltAllCount;			/* Count of reasons to halt context 0 queues */
    u_int		    HaltNonContext0Count;		/* Count of reasons to halt non-context 0 queues */
    u_int		    HaltDmaDequeueCount;		/* Count of reasons to halt dma from dequeuing */
    u_int		    HaltThreadCount;			/* Count of reasons to halt the thread processor */
    u_int		    FlushCommandCount;			/* Count of reasons to flush command queues */
    u_int		    DiscardAllCount;			/* Count of reasons to discard context 0 */
    u_int		    DiscardNonContext0Count;		/* Count of reasons to discard non context 0 */

    struct thread_trap	   *ThreadTrap;				/* Thread Processor trap space */
    struct dma_trap	   *DmaTrap;				/* DMA Processor trap space */

    spinlock_t		    FreeHaltLock;			/* Lock for haltop free list */
    ELAN3_HALTOP		   *FreeHaltOperations;			/* Free list of haltops */
    u_int		    NumHaltOperations;			/* Number of haltops allocated */
    u_int		    ReservedHaltOperations;		/* Number of haltops reserved */

    ELAN3_HALTOP		   *HaltOperations;			/* List of operations to call */
    ELAN3_HALTOP		  **HaltOperationsTailpp;		/* Pointer to last "next" pointer in list */
    E3_uint32		    HaltOperationsMask;			/* Or of all bits in list of operations */

    physaddr_t		    SdramPhysBase;			/* Physical address of SDRAM */
    physaddr_t		    SdramPhysMask;			/* and mask of significant bits */ 
    
    physaddr_t		    PciPhysBase;			/* physical address of local PCI segment */
    physaddr_t		    PciPhysMask;			/* and mask of significant bits */

    long		    ErrorTime;				/* lbolt at last error (link,ecc etc) */
    long		    ErrorsPerTick;			/* count of errors for this tick */
    timer_fn_t		    ErrorTimeoutId;			/* id of timeout when errors masked out */
    timer_fn_t		    DmaPollTimeoutId;			/* id of timeout to poll for "bad" dmas */
    int			    FilterHaltQueued;

    /*
     * HAT layer specific entries.
     */
    ELAN3_LEVEL_PTBL_BLOCK   Level[4];
    spinlock_t		    PtblGroupLock;			/* Lock for Page Table group lists */
    struct elan3_ptbl_gr    *Level3PtblGroupHand;		/* Hand for ptbl stealing */

    /*
     * Per-Context Information structures.
     */
    struct elan3_info	   *Infos;				/* List of "infos" for this device */

    char                    LinkShutdown;                       /* link forced into reset by panic/shutdown/dump */

    /*
     * Device statistics.
     */
    ELAN3_STATS		    Stats;
    ELAN_STATS_IDX          StatsIndex;

    struct {
	E3_Regs		   *RegPtr;
	char		   *Sdram[ELAN3_SDRAM_NUM_BANKS];
    } PanicState;
} ELAN3_DEV;

#define ELAN3_DEV_CTX_TABLE(dev,ctxtn) ( (dev)->CtxtTable[ (ctxtn) &  MAX_ROOT_CONTEXT_MASK] )

/* macros for accessing dev->RegPtr.Tags/Sets. */
#define write_cache_tag(dev,what,val)	writeq (val, (void *) (dev->RegPtr + offsetof (E3_Regs, Tags.what)))
#define read_cache_tag(dev,what)	readq ((void *) (dev->RegPtr + offsetof (E3_Regs, Tags.what)))
#define write_cache_set(dev,what,val)	writeq (val, (void *) (dev->RegPtr + offsetof (E3_Regs, Sets.what)))
#define read_cache_set(dev,what)	readq ((void *) (dev->RegPtr + offsetof (E3_Regs, Sets.what)))

/* macros for accessing dev->RegPtr.Regs. */
#define write_reg64(dev,what,val)	writeq (val, (void *) (dev->RegPtr + offsetof (E3_Regs, Regs.what)))
#define write_reg32(dev,what,val)	writel (val, (void *) (dev->RegPtr + offsetof (E3_Regs, Regs.what)))
#define read_reg64(dev,what)		readq ((void *) (dev->RegPtr + offsetof (E3_Regs, Regs.what)))
#define read_reg32(dev,what)		readl ((void *) (dev->RegPtr + offsetof (E3_Regs, Regs.what)))

/* macros for accessing dev->RegPtr.uRegs. */
#define write_ureg64(dev,what,val)	writeq (val, (void *) (dev->RegPtr + offsetof (E3_Regs, URegs.what)))
#define write_ureg32(dev,what,val)	writel (val, (void *) (dev->RegPtr + offsetof (E3_Regs, URegs.what)))
#define read_ureg64(dev,what)		readq ((void *) (dev->RegPtr + offsetof (E3_Regs, URegs.what)))
#define read_ureg32(dev,what)		readl ((void *) (dev->RegPtr + offsetof (E3_Regs, URegs.what)))

/* macros for accessing dma descriptor/thread regs */
#define copy_dma_regs(dev, desc) \
MACRO_BEGIN \
    register int i;  \
    for (i = 0; i < sizeof (E3_DMA)/sizeof(E3_uint64); i++) \
	((E3_uint64 *) desc)[i] = readq ((void *)(dev->RegPtr + offsetof (E3_Regs, Regs.Dma_Desc) + i*sizeof (E3_uint64))); \
MACRO_END

#define copy_thread_regs(dev, regs) \
MACRO_BEGIN \
    register int i;  \
    for (i = 0; i < (32*sizeof (E3_uint32))/sizeof(E3_uint64); i++) \
	((E3_uint64 *) regs)[i] = readq ((void *)(dev->RegPtr + offsetof (E3_Regs, Regs.Globals[0]) + i*sizeof (E3_uint64))); \
MACRO_END

_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::IntrLock, 
			  _E3_DataBusMap::Exts _E3_DataBusMap::Input_Context_Fil_Flush
			  elan3_dev::CurrentCommandPortTrap elan3_dev::HaltAllCount elan3_dev::HaltDmaDequeueCount
			  elan3_dev::FlushCommandCount elan3_dev::DiscardAllCount elan3_dev::DiscardNonContext0Count
			  elan3_dev::HaltOperations elan3_dev::HaltOperationsMask))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::TlbLock, 
			  _E3_DataBusMap::Cache_Control_Reg))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::InfoLock, 
			  elan3_dev::Infos elan3_dev::InfoTable))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::FreeHaltLock, 
			  elan3_dev::FreeHaltOperations elan3_dev::NumHaltOperations elan3_dev::ReservedHaltOperations))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::PageFreeListLock, 
			  elan3_dev::PageFreeList elan3_dev::PageFreeListSize))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::Level1PtblLock,
			  elan3_dev::Level1PtblTotal elan3_dev::Level1PtblFreeCount elan3_dev::Level1PtblFreeList))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::Level2PtblLock,
			  elan3_dev::Level2PtblTotal elan3_dev::Level2PtblFreeCount elan3_dev::Level2PtblFreeList))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::Level3PtblLock,
			  elan3_dev::Level3PtblTotal elan3_dev::Level3PtblFreeCount elan3_dev::Level3PtblFreeList))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::PtblGroupLock,
			  elan3_dev::Level1PtblGroupList elan3_dev::Level2PtblGroupList elan3_dev::Level3PtblGroupList))

_NOTE(DATA_READABLE_WITHOUT_LOCK(elan3_dev::InfoTable elan3_dev::Level1PtblFreeList
				 elan3_dev::Level2PtblFreeList elan3_dev::Level3PtblFreeList))

_NOTE(LOCK_ORDER(elan3_dev::InfoLock elan3_dev::IntrLock))
_NOTE(LOCK_ORDER(as::a_lock elan3_dev::InfoLock))
_NOTE(LOCK_ORDER(as::a_lock elan3_dev::IntrLock))

#define SET_INT_MASK(dev,Mask)		MACRO_BEGIN write_reg32 (dev, Exts.InterruptMask, ((dev)->InterruptMask = (Mask)));  mmiob(); MACRO_END
#define ENABLE_INT_MASK(dev, bits)	MACRO_BEGIN write_reg32 (dev, Exts.InterruptMask, ((dev->InterruptMask |= (bits)))); mmiob(); MACRO_END
#define DISABLE_INT_MASK(dev, bits)	MACRO_BEGIN write_reg32 (dev, Exts.InterruptMask, ((dev->InterruptMask &= ~(bits)))); mmiob(); MACRO_END

#define INIT_SCHED_STATUS(dev, val) \
MACRO_BEGIN \
	(dev)->SchCntReg = (val); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg); \
	mmiob(); \
MACRO_END

#define SET_SCHED_STATUS(dev, val) \
MACRO_BEGIN \
	ASSERT (((val) & HaltStopAndExtTestMask) == (val)); \
	(dev)->SchCntReg |= (val); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg); \
        mmiob (); \
MACRO_END

#define CLEAR_SCHED_STATUS(dev, val) \
MACRO_BEGIN \
	ASSERT (((val) & HaltStopAndExtTestMask) == (val)); \
	(dev)->SchCntReg &= ~(val); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg); \
        mmiob(); \
MACRO_END

#define MODIFY_SCHED_STATUS(dev, SetBits, ClearBits) \
MACRO_BEGIN \
	ASSERT ((((SetBits)|(ClearBits)) & HaltStopAndExtTestMask) == ((SetBits)|(ClearBits))); \
	(dev)->SchCntReg = (((dev)->SchCntReg | (SetBits)) & ~(ClearBits)); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg); \
	mmiob(); \
MACRO_END

#define PULSE_SCHED_STATUS(dev, RestartBits) \
MACRO_BEGIN \
	ASSERT (((RestartBits) & HaltStopAndExtTestMask) == 0); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg | (RestartBits)); \
	mmiob(); \
MACRO_END

#define SET_SCHED_LINK_VALUE(dev, enabled, val) \
MACRO_BEGIN \
	(dev)->SchCntReg = (((dev)->SchCntReg & HaltAndStopMask) | ((enabled) ? LinkBoundaryScan : 0) | LinkSetValue(val, 0)); \
	write_reg32 (dev, Exts.SchCntReg, (dev)->SchCntReg); \
	mmiob(); \
MACRO_END

#ifdef DEBUG_ASSERT
#  define ELAN3_ASSERT(dev, EX)	((void)((EX) || elan3_assfail(dev, #EX, __FILE__, __LINE__)))
#else
#  define ELAN3_ASSERT(dev, EX)
#endif

/* elandev_generic.c */
extern int        InitialiseElan (ELAN3_DEV *dev, ioaddr_t CmdPort);
extern void       FinaliseElan (ELAN3_DEV *dev);
extern int        InterruptHandler (ELAN3_DEV *dev);
extern void       PollForDmaHungup (void *arg);

extern int        SetLinkBoundaryScan (ELAN3_DEV *dev);
extern void       ClearLinkBoundaryScan (ELAN3_DEV *dev);
extern int        WriteBoundaryScanValue (ELAN3_DEV *dev, int value);
extern int        ReadBoundaryScanValue(ELAN3_DEV *dev, int link);

extern int        ReadVitalProductData (ELAN3_DEV *dev, int *CasLatency);

extern struct elan3_ptbl_gr *ElanGetPtblGr (ELAN3_DEV *dev, sdramaddr_t offset);
extern void       ElanSetPtblGr (ELAN3_DEV *dev, sdramaddr_t offset, struct elan3_ptbl_gr *ptg);

extern void       ElanFlushTlb (ELAN3_DEV *dev);

extern void       SetSchedStatusRegister (ELAN3_DEV *dev, E3_uint32 Pend, volatile E3_uint32 *Maskp);
extern void	  FreeHaltOperation (ELAN3_DEV *dev, ELAN3_HALTOP *op);
extern int	  ReserveHaltOperations (ELAN3_DEV *dev, int count, int cansleep);
extern void	  ReleaseHaltOperations (ELAN3_DEV *dev, int count);
extern void	  ProcessHaltOperations (ELAN3_DEV *dev, E3_uint32 Pend);
extern void	  QueueHaltOperation (ELAN3_DEV *dev, E3_uint32 Pend, volatile E3_uint32 *Maskp,
				      E3_uint32 ReqMask, void (*Function)(ELAN3_DEV *, void *), void *Arguement);

extern int        ComputePosition (ELAN_POSITION *pos, unsigned NodeId, unsigned NumNodes, unsigned numDownLinksVal);

extern caddr_t	  MiToName (int mi);
extern void	  ElanBusError (ELAN3_DEV *dev);

extern void	  TriggerLsa (ELAN3_DEV *dev);

extern ELAN3_DEV  *elan3_device (int instance);
extern int	  DeviceRegisterSize (ELAN3_DEV *dev, int rnumber, int *sizep);
extern int	  MapDeviceRegister (ELAN3_DEV *dev, int rnumber, ioaddr_t *addrp, int offset, 
				     int len, DeviceMappingHandle *handlep);
extern void       UnmapDeviceRegister (ELAN3_DEV *dev, DeviceMappingHandle *handlep);


/* sdram.c */
/* sdram accessing functions - define 4 different types for 8,16,32,64 bit accesses */
extern unsigned char      elan3_sdram_readb (ELAN3_DEV *dev, sdramaddr_t ptr);
extern unsigned short     elan3_sdram_readw (ELAN3_DEV *dev, sdramaddr_t ptr);
extern unsigned int       elan3_sdram_readl (ELAN3_DEV *dev, sdramaddr_t ptr);
extern unsigned long long elan3_sdram_readq (ELAN3_DEV *dev, sdramaddr_t ptr);
extern void               elan3_sdram_writeb (ELAN3_DEV *dev, sdramaddr_t ptr, unsigned char val);
extern void               elan3_sdram_writew (ELAN3_DEV *dev, sdramaddr_t ptr, unsigned short val);
extern void               elan3_sdram_writel (ELAN3_DEV *dev, sdramaddr_t ptr, unsigned int val);
extern void               elan3_sdram_writeq (ELAN3_DEV *dev, sdramaddr_t ptr, unsigned long long val);

extern void		  elan3_sdram_zerob_sdram (ELAN3_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan3_sdram_zerow_sdram (ELAN3_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan3_sdram_zerol_sdram (ELAN3_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan3_sdram_zeroq_sdram (ELAN3_DEV *dev, sdramaddr_t ptr, int nbytes);

extern void               elan3_sdram_copyb_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan3_sdram_copyw_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan3_sdram_copyl_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan3_sdram_copyq_from_sdram (ELAN3_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan3_sdram_copyb_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan3_sdram_copyw_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan3_sdram_copyl_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan3_sdram_copyq_to_sdram (ELAN3_DEV *dev, void *from, sdramaddr_t to, int nbytes);

extern void	          elan3_sdram_init (ELAN3_DEV *dev);
extern void               elan3_sdram_fini (ELAN3_DEV *dev);
extern void	          elan3_sdram_add (ELAN3_DEV *dev, sdramaddr_t base, sdramaddr_t top);
extern sdramaddr_t        elan3_sdram_alloc (ELAN3_DEV *dev, int nbytes);
extern void               elan3_sdram_free (ELAN3_DEV *dev, sdramaddr_t ptr, int nbytes);
extern physaddr_t         elan3_sdram_to_phys (ELAN3_DEV *dev, sdramaddr_t addr);

/* cproc.c */
extern void	  HandleCProcTrap (ELAN3_DEV *dev, E3_uint32 Pend, E3_uint32 *Mask);

/* iproc.c */
extern void	  HandleIProcTrap (ELAN3_DEV *dev, int Channel, E3_uint32 Pend, sdramaddr_t FaultSaveOff, 
				   sdramaddr_t TransactionsOff, sdramaddr_t DataOff);

/* tproc.c */
extern int 	  HandleTProcTrap (ELAN3_DEV *dev, E3_uint32 *RestartBits);
extern void	  DeliverTProcTrap (ELAN3_DEV *dev, struct thread_trap *threadTrap, E3_uint32 Pend);

/* dproc.c */
extern int 	  HandleDProcTrap (ELAN3_DEV *dev, E3_uint32 *RestartBits);
extern void	  DeliverDProcTrap (ELAN3_DEV *dev, struct dma_trap *dmaTrap, E3_uint32 Pend);

#if defined(LINUX)
/* procfs_linux.h */
extern struct proc_dir_entry *elan3_procfs_root;
extern struct proc_dir_entry *elan3_config_root;

extern void elan3_procfs_init(void);
extern void elan3_procfs_fini(void);
extern void elan3_procfs_device_init (ELAN3_DEV *dev);
extern void elan3_procfs_device_fini (ELAN3_DEV *dev);
#endif /* defined(LINUX) */

/* elan3_osdep.c */
extern int        BackToBackMaster;
extern int        BackToBackSlave;

#define ELAN_REG_REC_MAX (100)
#define ELAN_REG_REC(REG)  {                                         \
elan_reg_rec_file [elan_reg_rec_index] = __FILE__;                   \
elan_reg_rec_line [elan_reg_rec_index] = __LINE__;                   \
elan_reg_rec_reg  [elan_reg_rec_index] = REG;                        \
elan_reg_rec_cpu  [elan_reg_rec_index] = smp_processor_id();         \
elan_reg_rec_lbolt[elan_reg_rec_index] = lbolt;                      \
elan_reg_rec_index = ((elan_reg_rec_index+1) % ELAN_REG_REC_MAX);}

extern char *    elan_reg_rec_file [ELAN_REG_REC_MAX];
extern int       elan_reg_rec_line [ELAN_REG_REC_MAX];
extern long      elan_reg_rec_lbolt[ELAN_REG_REC_MAX];
extern int       elan_reg_rec_cpu  [ELAN_REG_REC_MAX];
extern E3_uint32 elan_reg_rec_reg  [ELAN_REG_REC_MAX];
extern int       elan_reg_rec_index;
 
#endif /* __KERNEL__ */


#define ELAN3_PROCFS_ROOT          "/proc/qsnet/elan3"
#define ELAN3_PROCFS_VERSION       "/proc/qsnet/elan3/version"
#define ELAN3_PROCFS_DEBUG         "/proc/qsnet/elan3/config/elandebug"
#define ELAN3_PROCFS_DEBUG_CONSOLE "/proc/qsnet/elan3/config/elandebug_console"
#define ELAN3_PROCFS_DEBUG_BUFFER  "/proc/qsnet/elan3/config/elandebug_buffer"
#define ELAN3_PROCFS_MMU_DEBUG     "/proc/qsnet/elan3/config/elan3mmu_debug"
#define ELAN3_PROCFS_PUNT_LOOPS    "/proc/qsnet/elan3/config/eventint_punt_loops"

#define ELAN3_PROCFS_DEVICE_STATS_FMT    "/proc/qsnet/elan3/device%d/stats"
#define ELAN3_PROCFS_DEVICE_POSITION_FMT "/proc/qsnet/elan3/device%d/position"
#define ELAN3_PROCFS_DEVICE_NODESET_FMT  "/proc/qsnet/elan3/device%d/nodeset"

#endif /* __ELAN3_ELANDEV_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
