/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

/*
 * Header file for internal slave mapping of the ELAN3 registers
 */

#ifndef _ELAN3_ELANREGS_H
#define _ELAN3_ELANREGS_H

#ident "$Id: elanregs.h,v 1.87 2004/04/22 12:27:21 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elanregs.h,v $*/

#include <elan3/e3types.h>
#include <elan3/dma.h>
#include <elan3/elanuregs.h>

#define MAX_ROOT_CONTEXT_MASK	0xfff
#define SYS_CONTEXT_BIT		0x1000
#define ALL_CONTEXT_BITS	(MAX_ROOT_CONTEXT_MASK | SYS_CONTEXT_BIT)
#define ROOT_TAB_OFFSET(Cntxt)	(((Cntxt) & MAX_ROOT_CONTEXT_MASK) << 4)
#define CLEAR_SYS_BIT(Cntxt)	((Cntxt) & ~SYS_CONTEXT_BIT)

#define E3_CACHELINE_SIZE	(32)
#define E3_CACHE_SIZE		(8192)

typedef volatile struct _E3_CacheSets
{
   E3_uint64	Set0[256];	/* 2k bytes per set */
   E3_uint64	Set1[256];	/* 2k bytes per set */
   E3_uint64	Set2[256];	/* 2k bytes per set */
   E3_uint64	Set3[256];	/* 2k bytes per set */
} E3_CacheSets;

typedef union e3_cache_tag
{
   E3_uint64	Value;
   struct {
#if defined(__LITTLE_ENDIAN__)
       E3_uint32 pad2:8;		/* Undefined value when read */
       E3_uint32 LineError:1;		/* A line error has occured */
       E3_uint32 Modified:1;		/* Cache data is modified */
       E3_uint32 FillPending:1;		/* Pipelined fill occuring*/
       E3_uint32 AddrTag27to11:17;	/* Tag address bits 27 to 11 */
       E3_uint32 pad1:4;		/* Undefined value when read */
       E3_uint32 pad0;			/* Undefined value when read */
#else
       E3_uint32 pad0;			/* Undefined value when read */
       E3_uint32 pad1:4;		/* Undefined value when read */
       E3_uint32 AddrTag27to11:17;	/* Tag address bits 27 to 11 */
       E3_uint32 FillPending:1;		/* Pipelined fill occuring*/
       E3_uint32 Modified:1;		/* Cache data is modified */
       E3_uint32 LineError:1;		/* A line error has occured */
       E3_uint32 pad2:8;		/* Undefined value when read */
#endif
   } s;
} E3_CacheTag;

#define E3_NumCacheLines	64
#define E3_NumCacheSets		4

typedef volatile struct _E3_CacheTags
{
   E3_CacheTag Tags[E3_NumCacheLines][E3_NumCacheSets];	/* 2k bytes per set */
} E3_CacheTags;

typedef union E3_IProcStatus_Reg
{
    E3_uint32 Status;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 TrapType:8;		/* iprocs trap ucode address */
	E3_uint32 SuspendAddr:8;	/* iprocs suspend address */
	E3_uint32 EopType:2;		/* Type of Eop Received */
	E3_uint32 QueueingPacket:1;	/* receiving a queueing packet */
	E3_uint32 AckSent:1;		/* a packet ack has been sent */
	E3_uint32 Reject:1;		/* a packet nack has been sent */
	E3_uint32 CrcStatus:2;		/* Crc Status value */
	E3_uint32 BadLength:1;		/* Eop was received in a bad place */
	E3_uint32 Chan1:1;		/* This packet received on v chan1 */
	E3_uint32 First:1;		/* This is the first transaction in the packet */
	E3_uint32 Last:1;		/* This is the last transaction in the packet */
	E3_uint32 Unused:2;
	E3_uint32 WakeupFunction:3;	/* iprocs wakeup function */
#else
	E3_uint32 WakeupFunction:3;	/* iprocs wakeup function */
	E3_uint32 Unused:2;
	E3_uint32 Last:1;		/* This is the last transaction in the packet */
	E3_uint32 First:1;		/* This is the first transaction in the packet */
	E3_uint32 Chan1:1;		/* This packet received on v chan1 */
	E3_uint32 BadLength:1;		/* Eop was received in a bad place */
	E3_uint32 CrcStatus:2;		/* Crc Status value */
	E3_uint32 Reject:1;		/* a packet nack has been sent */
	E3_uint32 AckSent:1;		/* a packet ack has been sent */
	E3_uint32 QueueingPacket:1;	/* receiving a queueing packet */
	E3_uint32 EopType:2;		/* Type of Eop Received */
	E3_uint32 SuspendAddr:8;	/* iprocs suspend address */
	E3_uint32 TrapType:8;		/* iprocs trap ucode address */
#endif
    } s;
} E3_IProcStatus_Reg;

#define CRC_STATUS_GOOD    (0 << 21)
#define CRC_STATUS_DISCARD (1 << 21)
#define CRC_STATUS_ERROR   (2 << 21)
#define CRC_STATUS_BAD     (3 << 21)

#define CRC_MASK	   (3 << 21)

#define EOP_GOOD	   (1 << 16)
#define EOP_BADACK	   (2 << 16)
#define EOP_ERROR_RESET	   (3 << 16)

#define E3_IPS_LastTrans	(1 << 26)
#define E3_IPS_FirstTrans	(1 << 25)
#define E3_IPS_VChan1		(1 << 24)
#define E3_IPS_BadLength	(1 << 23)
#define E3_IPS_CrcMask		(3 << 21)
#define E3_IPS_Rejected		(1 << 20)
#define E3_IPS_AckSent		(1 << 19)
#define E3_IPS_QueueingPacket	(1 << 18)
#define E3_IPS_EopType		(3 << 16)

typedef union E3_Status_Reg
{
    E3_uint32 Status;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 TrapType:8;		/* procs trap ucode address */
	E3_uint32 SuspendAddr:8;	/* procs suspend address */
	E3_uint32 Context:13;		/* procs current context */
	E3_uint32 WakeupFunction:3;	/* procs wakeup function */
#else
	E3_uint32 WakeupFunction:3;	/* procs wakeup function */
	E3_uint32 Context:13;		/* procs current context */
	E3_uint32 SuspendAddr:8;	/* procs suspend address */
	E3_uint32 TrapType:8;		/* procs trap ucode address */
#endif
    } s;
} E3_Status_Reg;

/* values for WakeupFunction */
#define SleepOneTick			0
#define WakeupToSendTransOrEop		1
#define SleepOneTickThenRunnable	2
#define WakeupNever			4
/* extra dma wakeup functions */
#define WakupeToSendTransOrEop		1
#define WakeupForPacketAck		3
#define WakeupToSendTrans		5
/* extra thread wakup function */
#define WakeupStopped			3
/* extra cproc wakup function */
#define WakeupSetEvent			3

#define GET_STATUS_CONTEXT(Ptr)      ((Ptr.Status >> 16) & 0x1fff)
#define GET_STATUS_SUSPEND_ADDR(Ptr) ((Ptr.Status >> 8) & 0xff)
#define GET_STATUS_TRAPTYPE(Ptr)     ((E3_uint32)(Ptr.Status & 0xff))

/*
 * Interrupt register bits
 */
#define INT_PciMemErr			(1<<15) /* Pci memory access error */
#define INT_SDRamInt			(1<<14) /* SDRam ECC interrupt */
#define INT_EventInterrupt		(1<<13) /* Event Interrupt */
#define INT_LinkError			(1<<12) /* Link Error */
#define INT_ComQueue			(1<<11) /* a comm queue half full */
#define INT_TProcHalted			(1<<10) /* Tproc Halted */
#define INT_DProcHalted			(1<<9) /* Dmas Halted */
#define INT_DiscardingNonSysCntx	(1<<8) /* Inputters Discarding Non-SysCntx */
#define INT_DiscardingSysCntx		(1<<7) /* Inputters Discarding SysCntx */
#define INT_TProc			(1<<6) /* tproc interrupt */
#define INT_CProc			(1<<5) /* cproc interrupt */
#define INT_DProc			(1<<4) /* dproc interrupt */
#define INT_IProcCh1NonSysCntx		(1<<3) /* iproc non-SysCntx interrupt */
#define INT_IProcCh1SysCntx		(1<<2) /* iproc SysCntx interrupt */
#define INT_IProcCh0NonSysCntx		(1<<1) /* iproc non-SysCntx interrupt */
#define INT_IProcCh0SysCntx		(1<<0) /* iproc SysCntx interrupt */

#define INT_Inputters		(INT_IProcCh0SysCntx | INT_IProcCh0NonSysCntx | INT_IProcCh1SysCntx | INT_IProcCh1NonSysCntx)
#define INT_Discarding  	(INT_DiscardingSysCntx | INT_DiscardingNonSysCntx)
#define INT_Halted		(INT_DProcHalted | INT_TProcHalted)
#define INT_ErrorInterrupts	(INT_PciMemErr | INT_SDRamInt | INT_LinkError)

/*
 * Link state bits.
 */
#define LS_LinkNotReady	(1 << 0) /* Link is in reset or recovering from an error */
#define LS_Locked	(1 << 1) /* Linkinput PLL is locked */
#define LS_LockError	(1 << 2) /* Linkinput PLL was unable to lock onto the input clock. */
#define LS_DeskewError	(1 << 3) /* Linkinput was unable to Deskew all the inputs. (Broken wire?) */
#define LS_PhaseError	(1 << 4) /* Linkinput Phase alignment error. */
#define LS_DataError	(1 << 5) /* Received value was neither good data or a token. */
#define LS_FifoOvFlow0	(1 << 6) /* Channel 0 input fifo overflowed. */
#define LS_FifoOvFlow1	(1 << 7) /* Channel 1 input fifo overflowed. */

/*
 * Link State Constant defines, used for writing to LinkSetValue
 */

#define LRS_DataDel0		0x0
#define LRS_DataDel1		0x1
#define LRS_DataDel2		0x2
#define LRS_DataDel3		0x3
#define LRS_DataDel4		0x4
#define LRS_DataDel5		0x5
#define LRS_DataDel6		0x6
#define LRS_DataDel7		0x7
#define LRS_DataDel8		0x8
#define LRS_PllDelValue		0x9
#define LRS_ClockEven		0xA
#define LRS_ClockOdd		0xB
#define LRS_ErrorLSW		0xC
#define LRS_ErrorMSW		0xD
#define LRS_FinCoarseDeskew	0xE
#define LRS_LinkInValue		0xF
#define LRS_NumLinkDels         0x10

#define LRS_Pllfast             0x40
 
union Sched_Status
{
    E3_uint32 Status;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 StopNonSysCntxs:1;
	E3_uint32 FlushCommandQueues:1;
	E3_uint32 HaltDmas:1;
	E3_uint32 HaltDmaDequeue:1;
	E3_uint32 HaltThread:1;
	E3_uint32 CProcStop:1;
	E3_uint32 DiscardSysCntxIn:1;
	E3_uint32 DiscardNonSysCntxIn:1;
	E3_uint32 RestartCh0SysCntx:1;
	E3_uint32 RestartCh0NonSysCntx:1;
	E3_uint32 RestartCh1SysCntx:1;
	E3_uint32 RestartCh1NonSysCntx:1;
	E3_uint32 RestartDProc:1;
	E3_uint32 RestartTProc:1;
	E3_uint32 RestartCProc:1;
	E3_uint32 ClearLinkErrorInt:1;
	E3_uint32 :3;
	E3_uint32 LinkSetValue:10; 
	E3_uint32 FixLinkDelays:1;
	E3_uint32 LinkBoundaryScan:1;
#else
	E3_uint32 LinkBoundaryScan:1;
	E3_uint32 FixLinkDelays:1;
	E3_uint32 LinkSetValue:10; 
	E3_uint32 :3;
	E3_uint32 ClearLinkErrorInt:1;
	E3_uint32 RestartCProc:1;
	E3_uint32 RestartTProc:1;
	E3_uint32 RestartDProc:1;
	E3_uint32 RestartCh1NonSysCntx:1;
	E3_uint32 RestartCh1SysCntx:1;
	E3_uint32 RestartCh0NonSysCntx:1;
	E3_uint32 RestartCh0SysCntx:1;
	E3_uint32 DiscardNonSysCntxIn:1;
	E3_uint32 DiscardSysCntxIn:1;
	E3_uint32 CProcStop:1;
	E3_uint32 HaltThread:1;
	E3_uint32 HaltDmaDequeue:1;
	E3_uint32 HaltDmas:1;
	E3_uint32 FlushCommandQueues:1;
	E3_uint32 StopNonSysCntxs:1;
#endif
    } s;
};

#define LinkBoundaryScan	((E3_uint32) 1<<31) /* Clears the link error interrupt */
#define FixLinkDelays		((E3_uint32) 1<<30) /* Clears the link error interrupt */
#define LinkSetValue(Val, OldVal) ((E3_uint32) (((Val) & 0x3ff) << 20) | ((OldVal) & ((~0x3ff) << 20)))

#define ClearLinkErrorInt	((E3_uint32) 1<<16) /* Clears the link error interrupt */
#define RestartCProc		((E3_uint32) 1<<15) /* Clears command proc interrupt */
#define RestartTProc		((E3_uint32) 1<<14) /* Clears thread interrupt */
#define RestartDProc		((E3_uint32) 1<<13) /* Clears dma0 interrupt */
#define RestartCh1NonSysCntx	((E3_uint32) 1<<12) /* Clears interrupt */
#define RestartCh1SysCntx	((E3_uint32) 1<<11) /* Clears interrupt */
#define RestartCh0NonSysCntx	((E3_uint32) 1<<10) /* Clears interrupt */
#define RestartCh0SysCntx	((E3_uint32) 1<<9) /* Clears interrupt */
#define CProcStopped		((E3_uint32) 1<<9) /* Read value only */

#define TraceSetEvents		((E3_uint32) 1<<8)
#define DiscardNonSysCntxIn	((E3_uint32) 1<<7)
#define DiscardSysCntxIn	((E3_uint32) 1<<6)
#define CProcStop		((E3_uint32) 1<<5) /* Will empty all the command port queues. */
#define HaltThread		((E3_uint32) 1<<4) /* Will stop the thread proc and clear the tproc command queue */
#define HaltDmaDequeue		((E3_uint32) 1<<3) /* Will stop the dmaers starting new dma's. */
#define HaltDmas		((E3_uint32) 1<<2) /* Will stop the dmaers and clear the dma command queues */
#define FlushCommandQueues	((E3_uint32) 1<<1) /* Causes the command ports to be flushed. */
#define StopNonSysCntxs		((E3_uint32) 1<<0) /* Prevents a non-SysCntx from starting. */

/* Initial value of schedule status register */
#define LinkResetToken		0x00F

#define Sched_Initial_Value	(LinkBoundaryScan | (LinkResetToken << 20) | \
				 DiscardSysCntxIn | DiscardNonSysCntxIn | HaltThread | HaltDmas)

#define StopDmaQueues	     (HaltDmaDequeue | HaltDmas | \
			      DiscardNonSysCntxIn | DiscardSysCntxIn)
#define CheckDmaQueueStopped (INT_DiscardingNonSysCntx | INT_DiscardingSysCntx | INT_DProcHalted)

#define HaltStopAndExtTestMask	0xfff001ff
#define HaltAndStopMask		0x000001ff


#define DmaComQueueNotEmpty	(1<<0)
#define ThreadComQueueNotEmpty	(1<<1)
#define EventComQueueNotEmpty	(1<<2)
#define DmaComQueueHalfFull	(1<<3)
#define ThreadComQueueHalfFull	(1<<4)
#define EventComQueueHalfFull	(1<<5)
#define DmaComQueueError	(1<<6)
#define ThreadComQueueError	(1<<7)
#define EventComQueueError	(1<<8)

#define ComQueueNotEmpty	(DmaComQueueNotEmpty | ThreadComQueueNotEmpty | EventComQueueNotEmpty)
#define ComQueueError		(DmaComQueueError | ThreadComQueueError | EventComQueueError)

typedef union _E3_DmaInfo
{
    E3_uint32  Value;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 DmaOutputOpen:1;	/* The packet is currently open */
	E3_uint32 :7;
	E3_uint32 TimeSliceCount:2;	/* Time left to timeslice */
	E3_uint32 UseRemotePriv:1;	/* Set for remote read dmas */
	E3_uint32 DmaLastPacket:1;	/* Set for the last packet of a dma */
	E3_uint32 PacketAckValue:2;	/* Packet ack type. Valid if AckBufferValid set. */
	E3_uint32 PacketTimeout:1;	/* Packet timeout. Sent an EopError. Valid if AckBufferValid set. */
	E3_uint32 AckBufferValid:1;	/* Packet ack is valid. */
	E3_uint32 :16;			/* read as Zero */
#else
	E3_uint32 :16;			/* read as Zero */
	E3_uint32 AckBufferValid:1;	/* Packet ack is valid. */
	E3_uint32 PacketTimeout:1;	/* Packet timeout. Sent an EopError. Valid if AckBufferValid set. */
	E3_uint32 PacketAckValue:2;	/* Packet ack type. Valid if AckBufferValid set. */
	E3_uint32 DmaLastPacket:1;	/* Set for the last packet of a dma */
	E3_uint32 UseRemotePriv:1;	/* Set for remote read dmas */
	E3_uint32 TimeSliceCount:2;	/* Time left to timeslice */
	E3_uint32 :7;
	E3_uint32 DmaOutputOpen:1;	/* The packet is currently open */
#endif
    } s;
} E3_DmaInfo;

typedef volatile struct _E3_DmaRds
{
   E3_uint32		DMA_Source4to0AndTwoReads;
   E3_uint32		pad13;
   E3_uint32		DMA_BytesToRead;
   E3_uint32		pad14;
   E3_uint32		DMA_MinusPacketSize;
   E3_uint32		pad15;
   E3_uint32		DMA_MaxMinusPacketSize;
   E3_uint32		pad16;
   E3_uint32		DMA_DmaOutputOpen;
   E3_uint32		pad16a;
   E3_DmaInfo		DMA_PacketInfo;
   E3_uint32		pad17[7];
   E3_uint32		IProcTrapBase;
   E3_uint32		pad18;
   E3_uint32		IProcBlockTrapBase;
   E3_uint32		pad19[11];
} E3_DmaRds;
   
typedef volatile struct _E3_DmaWrs
{
   E3_uint64		pad0;
   E3_uint64		LdAlignment;
   E3_uint64		ResetAckNLdBytesToWr;
   E3_uint64		SetAckNLdBytesToWr;
   E3_uint64		LdBytesToRd;
   E3_uint64		LdDmaType;
   E3_uint64		SendRoutes;
   E3_uint64		SendEop;
   E3_uint64		pad1[8];
} E3_DmaWrs;

typedef volatile struct _E3_Exts
{
   E3_uint32		CurrContext;				/* 0x12a00 */
   E3_uint32		pad0;
   E3_Status_Reg	DProcStatus;				/* 0x12a08 */
   E3_uint32		pad1;
   E3_Status_Reg	CProcStatus;				/* 0x12a10 */
   E3_uint32		pad2;
   E3_Status_Reg	TProcStatus;				/* 0x12a18 */
   E3_uint32		pad3;
   E3_IProcStatus_Reg	IProcStatus;				/* 0x12a20 */
   E3_uint32		pad4[3];

   E3_uint32		IProcTypeContext;			/* 0x12a30 */
   E3_uint32		pad5;
   E3_uint32		IProcTransAddr;				/* 0x12a38 */
   E3_uint32		pad6;
   E3_uint32		IProcCurrTransData0;			/* 0x12a40 */
   E3_uint32		pad7;
   E3_uint32		IProcCurrTransData1;			/* 0x12a48 */
   E3_uint32		pad8;

   E3_uint32		SchCntReg;				/* 0x12a50 */
   E3_uint32		pad9;
   E3_uint32		InterruptReg;				/* 0x12a58 */
   E3_uint32		pad10;
   E3_uint32		InterruptMask;				/* 0x12a60 */
   E3_uint32		pad11;
   E3_uint32		LinkErrorTypes;				/* 0x12a68 */
   E3_uint32		pad12[3];
   E3_uint32		LinkState;	/* a read here returens the DataDel value for the */
					/* link that has just been defined by a write to */
					/* Regs.Exts.SchCntReg.LinkSetValue */
   E3_uint32		pad13;

   union							/* 0x12a80 */
   {
      E3_DmaWrs		DmaWrs;
      E3_DmaRds		DmaRds;
   } Dmas;
} E3_Exts;

typedef union com_port_entry
{
    E3_uint64  type;
    struct
    {
	E3_uint32 Address;		/* Command VAddr */
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 Context0Issue:1;	/* Issue was for context 0 */
	E3_uint32 EventNotCommand:1;	/* Issue address bit 3 */
	E3_uint32 RemoteDesc:1;		/* Issue address bit 5 */
	E3_uint32 :13;	       	 	/* read as Zero */
	E3_uint32 Context:12;		/* Command Context */
	E3_uint32 :4;			/* read as Zero */
#else
	E3_uint32 :4;			/* read as Zero */
	E3_uint32 Context:12;		/* Command Context */
	E3_uint32 :13;	       	 	/* read as Zero */
	E3_uint32 RemoteDesc:1;		/* Issue address bit 5 */
	E3_uint32 EventNotCommand:1;	/* Issue address bit 3 */
	E3_uint32 Context0Issue:1;	/* Issue was for context 0 */
#endif
    } s;
} E3_ComPortEntry;

/* control reg bits */
#define CONT_MMU_ENABLE		(1 << 0) /* bit 0 enables mmu */
#define CONT_ENABLE_8K_PAGES	(1 << 1) /* When set smallest page is 8k instead of 4k. */
#define CONT_EN_ALL_SETS	(1 << 2) /* enable cache */
#define CONT_CACHE_LEVEL0	(1 << 3) /* cache context table */
#define CONT_CACHE_LEVEL1	(1 << 4) /* cache up level 1 PTD/PTE */
#define CONT_CACHE_LEVEL2	(1 << 5) /* cache up level 2 PTD/PTE */
#define CONT_CACHE_LEVEL3	(1 << 6) /* cache up level 3 PTD/PTE */
#define CONT_CACHE_TRAPS	(1 << 7) /* cache up traps */
#define CONT_CACHE_LEV0_ROUTES	(1 << 8) /* cache up small routes */
#define CONT_CACHE_LEV1_ROUTES	(1 << 9) /* cache up large routes */
#define CONT_CACHE_ALL		(CONT_CACHE_LEVEL0 | CONT_CACHE_LEVEL1 | CONT_CACHE_LEVEL2 | \
				 CONT_CACHE_LEVEL3 | CONT_CACHE_TRAPS | \
				 CONT_CACHE_LEV0_ROUTES | CONT_CACHE_LEV1_ROUTES)

#define CONT_SYNCHRONOUS	(1 << 10) /* PCI running sync */
#define CONT_SER		(1 << 11) /* Single bit output (Elan1 SER bit) */
#define CONT_SIR		(1 << 12) /* Writing 1 resets elan. */

#define CONT_PSYCHO_MODE	(1 << 13) /* Enables all the perversion required by psycho */
#define CONT_ENABLE_ECC		(1 << 14) /* Enables error detecting on the ECC */
#define CONT_SDRAM_TESTING	(1 << 15) /* Switches to test mode for checking EEC data bits */

/* defines SDRam CasLatency. Once set will not change again unless reset is reasserted. */
/* 1 = Cas Latency is 3, 0 = Cas Latency is 2 */
#define CAS_LATENCY_2		(0 << 16)
#define CAS_LATENCY_3		(1 << 16)
#define REFRESH_RATE_2US	(0 << 17) /* defines 2us SDRam Refresh rate. */
#define REFRESH_RATE_4US	(1 << 17) /* defines 4us SDRam Refresh rate. */
#define REFRESH_RATE_8US	(2 << 17) /* defines 8us SDRam Refresh rate. */
#define REFRESH_RATE_16US	(3 << 17) /* defines 16us SDRam Refresh rate. */

#define CONT_PCI_ERR		(1 << 19) /* Read 1 if PCI Error */
#define CONT_CLEAR_PCI_ERROR	(1 << 19) /* Clears an PCI error. */

/* Will cause the PCI error bit to become set. This is used to force the threads proc
   and the uProc to start to stall. */
#define CONT_SET_PCI_ERROR	(1 << 20)

/* Writes SDram control reg when set. Also starts SDram memory system refreshing. */
#define SETUP_SDRAM		(1 << 21)

/* Flushes the tlb */
#define MMU_FLUSH		(1 << 22)
/* and read back when it's finished */
#define MMU_FLUSHED		(1 << 0)

/* Clears any ECC error detected by SDRam interface */
#define CLEAR_SDRAM_ERROR	(1 << 23)

#define ECC_ADDR_MASK		0x0ffffff8
#define ECC_UE_MASK		0x1 
#define ECC_CE_MASK		0x2
#define ECC_ME_MASK		0x4 
#define ECC_SYN_MASK		0xff

/* define page table entry bit fields */
#define TLB_PageSizeBits	(3 << 0)
#define TLB_ACCBits		(7 << 2)
#define TLB_LocalBit		(1 << 5)
#define TLB_PCI64BitTargetBit	(1 << 6)
#define TLB_PCIBigEndianBit	(1 << 7)

#define TLB_ModifiedBit		(1 << 55)
#define TLB_ReferencedBit	(1 << 63)

/* Used to read values from the tlb. */
#define TLB_TlbReadCntBitsSh	56
#define TLB_UseSelAddrSh	(1ULL << 60)
#define TLB_WriteTlbLine	(1ULL << 61)

#define TLB_SEL_LINE(LineNo) (TLB_UseSelAddrSh | \
			      ((E3_uint64)((LineNo) & 0xf) << TLB_TlbReadCntBitsSh))

typedef union _E3_CacheContReg
{
    E3_uint32 ContReg;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 MMU_Enable:1;		/* wr 1 to enable the MMU */
	E3_uint32 Set8kPages:1;		/* wr 1 smallest page is 8k. */
	E3_uint32 EnableAllSets:1;	/* wr 1 All the cache sets are enabled */
	E3_uint32 Cache_Level0:1;	/* wr 1 lev0 page tabs will be cached */
	E3_uint32 Cache_Level1:1;	/* wr 1 lev1 page tabs will be cached */
	E3_uint32 Cache_Level2:1;	/* wr 1 lev2 page tabs will be cached */
	E3_uint32 Cache_Level3:1;	/* wr 1 lev3 page tabs will be cached */
	E3_uint32 Cache_Traps:1;	/* wr 1 trap info will be cached */
	E3_uint32 Cache_Lev0_Routes:1;	/* wr 1 small routes will be cached */
	E3_uint32 Cache_Lev1_Routes:1;	/* wr 1 big routes will be cached */
	E3_uint32 PCI_Synchronous:1;	/* Pci and sys clocks are running synchronously*/
	E3_uint32 SER:1;		/* 1 bit output port */
	E3_uint32 SIR:1;		/* write 1 will reset elan */
	E3_uint32 PsychoMode:1;		/* Enables psycho perversion mode. */
	E3_uint32 CasLatency:1;		/* 1=cas latency=3, 1=cas latency=2 */
	E3_uint32 RefreshRate:2;	/* 0=2us, 1=4us, 2=8us, 3=16us */
	E3_uint32 Pci_Err:1;		/* pci error. Write 1 clears err */
	E3_uint32 Set_Pci_Error:1;	/* Will simulate an Pci error */
	E3_uint32 StartSDRam:1;		/* Starts the sdram subsystem */
	E3_uint32 FlushTlb:1;		/* Flush the contence of the tlb */
	E3_uint32 :11;
#else
	E3_uint32 :11;
	E3_uint32 FlushTlb:1;		/* Flush the contence of the tlb */
	E3_uint32 StartSDRam:1;		/* Starts the sdram subsystem */
	E3_uint32 Set_Pci_Error:1;	/* Will simulate an Pci error */
	E3_uint32 Pci_Err:1;		/* pci error. Write 1 clears err */
	E3_uint32 RefreshRate:2;	/* 0=2us, 1=4us, 2=8us, 3=16us */
	E3_uint32 CasLatency:1;		/* 1=cas latency=3, 1=cas latency=2 */
	E3_uint32 PsychoMode:1;		/* Enables psycho perversion mode. */
	E3_uint32 SIR:1;		/* write 1 will reset elan */
	E3_uint32 SER:1;		/* 1 bit output port */
	E3_uint32 PCI_Synchronous:1;	/* Pci and sys clocks are running synchronously*/
	E3_uint32 Cache_Lev1_Routes:1;	/* wr 1 big routes will be cached */
	E3_uint32 Cache_Lev0_Routes:1;	/* wr 1 small routes will be cached */
	E3_uint32 Cache_Traps:1;	/* wr 1 trap info will be cached */
	E3_uint32 Cache_Level3:1;	/* wr 1 lev3 page tabs will be cached */
	E3_uint32 Cache_Level2:1;	/* wr 1 lev2 page tabs will be cached */
	E3_uint32 Cache_Level1:1;	/* wr 1 lev1 page tabs will be cached */
	E3_uint32 Cache_Level0:1;	/* wr 1 lev0 page tabs will be cached */
	E3_uint32 EnableAllSets:1;	/* wr 1 All the cache sets are enabled */
	E3_uint32 Set8kPages:1;		/* wr 1 smallest page is 8k. */
	E3_uint32 MMU_Enable:1;		/* wr 1 to enable the MMU */
#endif
    } s;
} E3_CacheContReg;

typedef union _E3_TrapBits
{
    volatile E3_uint32 Bits;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 ForcedTProcTrap:1;	 /* The theads proc has been halted */
	E3_uint32 InstAccessException:1; /* An instruction access exception */
	E3_uint32 Unimplemented:1;	 /* Unimplemented instruction executed */
	E3_uint32 DataAccessException:1; /* A data access exception */	

	E3_uint32 ThreadTimeout:1;	 /* The threads outputer has timed out */
	E3_uint32 OpenException:1;	 /* Invalid sequence of open, sendtr or close */
	E3_uint32 OpenRouteFetch:1;	 /* Fault while fetching routes for previous open*/
	E3_uint32 TrapForTooManyInsts:1; /* Thread has been executing for too long */
	
	E3_uint32 PacketAckValue:2;	 /* Packet ack type. Valid if AckBufferValid set. */
	E3_uint32 PacketTimeout:1;	 /* Packet timeout. Sent an EopError. Valid if AckBufferValid set. */

	E3_uint32 AckBufferValid:1;	 /* The PacketAckValue bits are valid */
	E3_uint32 OutputWasOpen:1;	 /* The output was open when tproc trapped */
	E3_uint32 TProcDeschedule:2;	 /* The reason the tproc stopped running. */
	E3_uint32 :17;
#else
	E3_uint32 :17;
	E3_uint32 TProcDeschedule:2;	 /* The reason the tproc stopped running. */
	E3_uint32 OutputWasOpen:1;	 /* The output was open when tproc trapped */
	E3_uint32 AckBufferValid:1;	 /* The PacketAckValue bits are valid */
	
	E3_uint32 PacketTimeout:1;	 /* Packet timeout. Sent an EopError. Valid if AckBufferValid set. */
	E3_uint32 PacketAckValue:2;	 /* Packet ack type. Valid if AckBufferValid set. */
	
	E3_uint32 TrapForTooManyInsts:1; /* Thread has been executing for too long */
	E3_uint32 OpenRouteFetch:1;	 /* Fault while fetching routes for previous open*/
	E3_uint32 OpenException:1;	 /* Invalid sequence of open, sendtr or close */
	E3_uint32 ThreadTimeout:1;	 /* The threads outputer has timed out */

	E3_uint32 DataAccessException:1; /* A data access exception */
	E3_uint32 Unimplemented:1;	 /* Unimplemented instruction executed */
	E3_uint32 InstAccessException:1; /* An instruction access exception */
	E3_uint32 ForcedTProcTrap:1;	 /* The theads proc has been halted */
#endif
    } s;
} E3_TrapBits;

typedef union _E3_DirtyBits
{
    volatile E3_uint32 Bits;
    struct
    {
#if defined(__LITTLE_ENDIAN__)
	E3_uint32 GlobalsDirty:8;
	E3_uint32 OutsDirty:8;		/* will always read as dirty. */
	E3_uint32 LocalsDirty:8;
	E3_uint32 InsDirty:8;
#else
	E3_uint32 InsDirty:8;
	E3_uint32 LocalsDirty:8;
	E3_uint32 OutsDirty:8;		/* will always read as dirty. */
	E3_uint32 GlobalsDirty:8;
#endif
    } s;
} E3_DirtyBits;

#define E3_TProcDescheduleMask    0x6000
#define E3_TProcDescheduleWait    0x2000
#define E3_TProcDescheduleSuspend 0x4000
#define E3_TProcDescheduleBreak   0x6000

#define E3_TrapBitsMask   	  0x7fff

#define ThreadRestartFromTrapBit	1
#define ThreadReloadAllRegs		2

#define E3_PAckOk	0
#define E3_PAckTestFail	1
#define E3_PAckDiscard	2
#define E3_PAckError	3

typedef volatile struct _E3_DataBusMap
{
   E3_uint64		 Dma_Alignment_Port[8];		/* 0x00002800 */
   E3_uint32		 pad0[0x30];			/* 0x00002840 */

   E3_uint32		 Input_Trans0_Data[0x10];	/* 0x00002900 */
   E3_uint32		 Input_Trans1_Data[0x10];
   E3_uint32		 Input_Trans2_Data[0x10];
   E3_uint32		 Input_Trans3_Data[0x10];

/* this is the start of the exts directly addressable from the ucode. */
   E3_Exts		 Exts;				/* 0x00002a00 */

/* this is the start of the registers directly addressable from the ucode. */
   E3_DMA		 Dma_Desc;			/* 0x00002b00 */

   E3_uint32		 Dma_Last_Packet_Size;		/* 0x00002b20 */
   E3_uint32		 Dma_This_Packet_Size;		/* 0x00002b24 */
   E3_uint32		 Dma_Tmp_Source;		/* 0x00002b28 */
   E3_uint32		 Dma_Tmp_Dest;			/* 0x00002b2c */

   E3_Addr		 Thread_SP_Save_Ptr;	/* points to the thread desched save word. */
   E3_uint32		 Dma_Desc_Size_InProg;		/* 0x00002b34 */

   E3_uint32		 Thread_Desc_SP;		/* 0x00002b38 */
   E3_uint32		 Thread_Desc_Context;		/* 0x00002b3c */

   E3_uint32		 uCode_TMP[0x10];		/* 0x00002b40 */

   E3_uint32		 TProc_NonSysCntx_FPtr;		/* 0x00002b80 */
   E3_uint32		 TProc_NonSysCntx_BPtr;		/* 0x00002b84 */
   E3_uint32		 TProc_SysCntx_FPtr;		/* 0x00002b88 */
   E3_uint32		 TProc_SysCntx_BPtr;		/* 0x00002b8c */
   E3_uint32		 DProc_NonSysCntx_FPtr;		/* 0x00002b90 */
   E3_uint32		 DProc_NonSysCntx_BPtr;		/* 0x00002b94 */
   E3_uint32		 DProc_SysCntx_FPtr;		/* 0x00002b98 */
   E3_uint32		 DProc_SysCntx_BPtr;		/* 0x00002b9c */

   E3_uint32		 Input_Trap_Base;		/* 0x00002ba0 */
   E3_uint32		 Input_Queue_Offset;		/* 0x00002ba4 */
   E3_uint32		 CProc_TrapSave_Addr;		/* 0x00002ba8 */
   E3_uint32		 Input_Queue_Addr;		/* 0x00002bac */
   E3_uint32		 uCode_TMP10;			/* 0x00002bb0 */
   E3_uint32		 uCode_TMP11;			/* 0x00002bb4 */
   E3_uint32		 Event_Trace_Ptr;		/* 0x00002bb8 */
   E3_uint32		 Event_Trace_Mask;		/* 0x00002bbc */

   E3_ComPortEntry	 DmaComQueue[3];		/* 0x00002bc0 */

   E3_uint32		 Event_Int_Queue_FPtr;		/* 0x00002bd8 */
   E3_uint32		 Event_Int_Queue_BPtr;		/* 0x00002bdc */

   E3_ComPortEntry	 ThreadComQueue[2];		/* 0x00002be0 */
   E3_ComPortEntry	 SetEventComQueue[2];		/* 0x00002bf0 */

   E3_uint32		 pad1[96];			/* 0x00002c00 */
   E3_uint32		 ComQueueStatus;		/* 0x00002d80 */
   E3_uint32		 pad2[31];			/* 0x00002d84 */

/* These are the internal registers of the threads proc. */
   E3_uint32		 Globals[8];			/* 0x00002e00 */
   E3_uint32		 Outs[8];
   E3_uint32		 Locals[8];
   E3_uint32		 Ins[8];

   E3_uint32		 pad3[16];

   E3_uint32		 IBufferReg[4];

   E3_uint32		 ExecuteNPC;
   E3_uint32		 ExecutePC;

   E3_uint32		 StartPC;
   E3_uint32		 pad4;

   E3_uint32		 StartnPC;
   E3_uint32		 pad5;

   E3_TrapBits		 TrapBits;
   E3_DirtyBits		 DirtyBits;
   E3_uint64		 LoadDataReg;
   E3_uint64		 StoreDataReg;

   E3_uint32		 ECC_STATUS0;
   E3_uint32		 ECC_STATUS1;
   E3_uint32		 pad6[0xe];

/* Pci slave port regs */
   E3_uint32		 PciSlaveReadCache[0x10];

   E3_uint32		 Fault_Base_Ptr;
   E3_uint32		 pad7;
   E3_uint32		 Context_Ptr;
   E3_uint32		 pad8;
   E3_uint32 		 Input_Context_Filter;	    /* write only, No data */
   E3_uint32		 Input_Context_Fil_Flush;   /* write only, No data */
   E3_CacheContReg	 Cache_Control_Reg;
   E3_uint32		 pad9;

   E3_uint64		 Tlb_Line_Value;
   
   E3_uint32		 Walk_Datareg1;
   E3_uint32		 Walk_VAddr_Tab_Base;
   E3_uint32		 Walk_Datareg;
   E3_uint32		 Walk_ContextReg;
   E3_uint32		 Walk_FaultAddr;
   E3_uint32		 Walk_EventAddr;

/* outputers output cont ext registers. */
   E3_uint64		 Dma_Route_012345_Context;
   E3_uint64		 pad10;
   E3_uint64		 Dma_Route_01234567;
   E3_uint64		 Dma_Route_89ABCDEF;

   E3_uint64		 Thread_Route_012345_Context;
   E3_uint64		 pad11;
   E3_uint64		 Thread_Route_01234567;
   E3_uint64		 Thread_Route_89ABCDEF;
} E3_DataBusMap;

typedef volatile struct _E3_Regs
{
   E3_CacheSets		  Sets;				/* 0x00000000 */
   E3_CacheTags		  Tags;				/* 0x00002000 */
   E3_DataBusMap	  Regs;				/* 0x00002800 */
   E3_uint32		  pad1[0x400];
   E3_User_Regs  	  URegs;
} E3_Regs;

#define MAX_TRAPPED_TRANS	16
#define TRANS_DATA_WORDS	16
#define TRANS_DATA_BYTES	64

/*
 * Event interrupt
 */
typedef volatile union _E3_EventInt
{
   E3_uint64	 ForceAlign;
   struct {
       E3_uint32 IntCookie;
       E3_uint32 EventContext;	/* Bits 16 to 28 */
    } s;
} E3_EventInt;

#define GET_EVENT_CONTEXT(Ptr) ((Ptr->s.EventContext >> 16) & MAX_ROOT_CONTEXT_MASK)

typedef volatile union _E3_ThreadQueue
{
   E3_uint64	ForceAlign;
   struct
   {
       E3_Addr	 Thread;
#if defined(__LITTLE_ENDIAN__)
       E3_uint32 :16;		/* Bits 0  to 15 */
       E3_uint32 Context:13;	/* Bits 16 to 28 */
       E3_uint32 :3;		/* Bits 29 to 31 */
#else
       E3_uint32 :3;		/* Bits 29 to 31 */
       E3_uint32 Context:13;	/* Bits 16 to 28 */
       E3_uint32 :16;		/* Bits 0  to 15 */
#endif
   } s;
} E3_ThreadQueue;

typedef volatile union _E3_FaultStatusReg
{
   E3_uint32 Status;
   struct
   {
#if defined(__LITTLE_ENDIAN__)
      E3_uint32 AccTypePerm:3;	/* Access permission. See below. Bits 0 to 2 */
      E3_uint32 AccSize:4;	/* Access size. See below for different types. Bits 3 to 6 */
      E3_uint32 WrAcc:1;	/* Access was a write. Bit 7 */
      E3_uint32 NonAllocAcc:1;	/* Access was a cache non allocate type. Bit 8 */
      E3_uint32 BlkDataType:2;	/* Data size used for endian flips. Bits 9 to 10 */
      E3_uint32 RdLine:1;	/* Access was a dma read line. Bit 11 */
      E3_uint32 RdMult:1;	/* Access was a dma read multiple. Bit 12 */
      E3_uint32 Walking:1;	/* The fault occued when walking. Bit 13 */
      E3_uint32 Level:2;	/* Page table level when the fault occued. Bits 14 to 15 */
      E3_uint32 ProtFault:1;	/* A protection fault occured. Bit 16 */
      E3_uint32 FaultPte:2;	/* Page table type when the fault occured. Bit 17 */
      E3_uint32 AlignmentErr:1;	/* Address alignment did not match the access size. Bit 19 */
      E3_uint32 VProcSizeErr:1;	/* VProc number is out of range. Bit 20 */
      E3_uint32 WalkBadData:1;	/* Memory CRC error during a walk. Bit 21 */
      E3_uint32 :10;		/* Bits 22 to 31 */
#else
      E3_uint32 :10;		/* Bits 22 to 31 */
      E3_uint32 WalkBadData:1;	/* Memory CRC error during a walk. Bit 21 */
      E3_uint32 VProcSizeErr:1;	/* VProc number is out of range. Bit 20 */
      E3_uint32 AlignmentErr:1;	/* Address alignment did not match the access size. Bit 19 */
      E3_uint32 FaultPte:2;	/* Page table type when the fault occured. Bit 17 */
      E3_uint32 ProtFault:1;	/* A protection fault occured. Bit 16 */
      E3_uint32 Level:2;	/* Page table level when the fault occued. Bits 14 to 15 */
      E3_uint32 Walking:1;	/* The fault occued when walking. Bit 13 */
      E3_uint32 RdMult:1;	/* Access was a dma read multiple. Bit 12 */
      E3_uint32 RdLine:1;	/* Access was a dma read line. Bit 11 */
      E3_uint32 BlkDataType:2;	/* Data size used for endian flips. Bits 9 to 10 */
      E3_uint32 NonAllocAcc:1;	/* Access was a cache non allocate type. Bit 8 */
      E3_uint32 WrAcc:1;	/* Access was a write. Bit 7 */
      E3_uint32 AccSize:4;	/* Access size. See below for different types. Bits 3 to 6 */
      E3_uint32 AccTypePerm:3;	/* Access permission. See below. Bits 0 to 2 */
#endif
   } s;
} E3_FaultStatusReg;

typedef union _E3_FaultSave
{
   E3_uint64		 ForceAlign;
   struct {
      E3_FaultStatusReg	 FSR;
      volatile E3_uint32 FaultContext;
      volatile E3_uint32 FaultAddress;
      volatile E3_uint32 EventAddress;
   } s;
} E3_FaultSave;

/* MMU fault status reg bit positions. */
#define FSR_WritePermBit	0	/* 1=Write access perm, 0=Read access perm */
#define FSR_RemotePermBit	1	/* 1=Remote access perm, 0=local access perm */
#define FSR_EventPermBit	2	/* 1=Event access perm, 0=data access perm */
#define FSR_Size0Bit		3
#define FSR_Size1Bit		4
#define FSR_Size2Bit		5
#define FSR_Size3Bit		6
#define FSR_WriteAccBit		7	/* 1=Write access, 0=Read access. */
#define FSR_NonAllocBit		8	/* 1=Do not fill cache with this data */
#define FSR_BlkDataTy0Bit	9
#define FSR_BlkDataTy1Bit	10
#define FSR_ReadLineBit		11
#define FSR_ReadMultipleBit	12

#define FSR_PermMask		(0xf << FSR_WritePermBit)
#define FSR_SizeMask		(0xf << FSR_Size0Bit)
#define FSR_AccTypeMask		(3 << FSR_WriteAccBit)
#define FSR_BlkDataTyMask	(3 << FSR_BlkDataTy0Bit)
#define FSR_PciAccTyMask	(3 << FSR_ReadLineBit)
#define FSR_Walking		(0x1 << 13)
#define FSR_Level_Mask		(0x3 << 14)
#define FSR_ProtFault		(0x1 << 16)
#define FSR_FaultPTEType	(0x2 << 17)
#define FSR_AddrSizeError	(0x1 << 19)
#define FSR_VProcSizeError	(0x1 << 20)
#define FSR_WalkBadData		(0x1 << 21)

#define FSR_PermRead		0
#define FSR_PermWrite		1
#define FSR_PermRemoteRead	2
#define FSR_PermRemoteWrite	3
#define FSR_PermEventRd		4
#define FSR_PermEventWr		5
#define FSR_PermRemoteEventRd	6
#define FSR_PermRemoteEventWr	7

/* AT size values for each access type */
#define FSR_Word		(0x0 << FSR_Size0Bit)
#define FSR_DWord		(0x1 << FSR_Size0Bit)
#define FSR_QWord		(0x2 << FSR_Size0Bit)
#define FSR_Block32		(0x3 << FSR_Size0Bit)
#define FSR_ReservedBlock	(0x6 << FSR_Size0Bit)
#define FSR_Block64		(0x7 << FSR_Size0Bit)
#define FSR_GetCntxFilter	(0x8 << FSR_Size0Bit)
#define FSR_QueueDWord		(0x9 << FSR_Size0Bit)
#define FSR_RouteFetch		(0xa << FSR_Size0Bit)
#define FSR_QueueBlock		(0xb << FSR_Size0Bit)
#define FSR_Block32PartWrite	(0xe << FSR_Size0Bit)
#define FSR_Block64PartWrite	(0xf << FSR_Size0Bit)

#define FSR_AllocRead		(0 << FSR_WriteAccBit)
#define FSR_AllocWrite		(1 << FSR_WriteAccBit)
#define FSR_NonAllocRd		(2 << FSR_WriteAccBit)
#define FSR_NonAllocWr		(3 << FSR_WriteAccBit)

#define FSR_TypeByte		(0 << FSR_BlkDataTy0Bit)
#define FSR_TypeHWord		(1 << FSR_BlkDataTy0Bit)
#define FSR_TypeWord		(2 << FSR_BlkDataTy0Bit)
#define FSR_TypeDWord		(3 << FSR_BlkDataTy0Bit)

typedef union E3_TrTypeCntx
{
   E3_uint32 TypeContext;
   struct
   {
#if defined(__LITTLE_ENDIAN__)
      E3_uint32 Type:16;		/* Transaction type field */
      E3_uint32 Context:13;		/* Transaction context */
      E3_uint32 TypeCntxInvalid:1;	/* Bit  29 */
      E3_uint32 StatusRegValid:1;	/* Bit  30 */
      E3_uint32 LastTrappedTrans:1;	/* Bit  31 */
#else
      E3_uint32 LastTrappedTrans:1;	/* Bit  31 */
      E3_uint32 StatusRegValid:1;	/* Bit  30 */
      E3_uint32 TypeCntxInvalid:1;	/* Bit  29 */
      E3_uint32 Context:13;		/* Transaction context */
      E3_uint32 Type:16;		/* Transaction type field */
#endif
   } s;
} E3_TrTypeCntx;

#define GET_TRAP_TYPE(Ptr)    (Ptr.TypeContext & 0xfff)
#define GET_TRAP_CONTEXT(Ptr) ((Ptr.TypeContext >> 16) & 0x1fff)

/* Words have been swapped for big endian access when fetched with dword access from elan.*/
typedef union _E3_IprocTrapHeader
{
   E3_uint64	forceAlign;

   struct
   {
      E3_TrTypeCntx 	 TrTypeCntx;
      E3_uint32	 	 TrAddr;
      E3_uint32	 	 TrData0;
      union
      {
	 E3_IProcStatus_Reg u_IProcStatus;
	 E3_uint32	    u_TrData1;
      } ipsotd;
   } s;
} E3_IprocTrapHeader;

#define IProcTrapStatus ipsotd.u_IProcStatus
#define TrData1	        ipsotd.u_TrData1

typedef struct E3_IprocTrapData
{
   E3_uint32 TrData[TRANS_DATA_WORDS];
} E3_IprocTrapData;

/*
 * 64 kbytes of elan local memory. Must be aligned on a 64k boundary
 */
#define E3_NonSysCntxQueueSize	0x400
#define E3_SysCntxQueueSize	0x100

typedef struct _E3_TrapAndQueue
{
   E3_DMA		NonSysCntxDmaQueue[E3_NonSysCntxQueueSize];			/* 0x000000 */
   E3_DMA		SysCntxDmaQueue[E3_SysCntxQueueSize];				/* 0x008000 */
   E3_EventInt		EventIntQueue[E3_NonSysCntxQueueSize];				/* 0x00A000 */
   E3_ThreadQueue	NonSysCntxThreadQueue[E3_NonSysCntxQueueSize];			/* 0x00C000 */	
   E3_ThreadQueue	SysCntxThreadQueue[E3_SysCntxQueueSize];			/* 0x00E000 */
   E3_FaultSave		IProcSysCntx;							/* 0x00E800 */
   E3_Addr		Thread_SP_Save;							/* 0x00E810 */
   E3_uint32		dummy0[3];							/* 0x00E814 */
   E3_FaultSave		ThreadProcData;							/* 0x00E820 */
   E3_FaultSave		ThreadProcInst;							/* 0x00E830 */
   E3_FaultSave		dummy1[2];							/* 0x00E840 */	
   E3_FaultSave		ThreadProcOpen;							/* 0x00E860 */
   E3_FaultSave		dummy2;								/* 0x00E870 */
   E3_FaultSave		IProcNonSysCntx;						/* 0x00E880 */
   E3_FaultSave		DProc;								/* 0x00E890 */
   E3_FaultSave		CProc;								/* 0x00E8A0 */
   E3_FaultSave		TProc;								/* 0x00E8B0 */
   E3_FaultSave		DProcData0;							/* 0x00E8C0 */
   E3_FaultSave		DProcData1;							/* 0x00E8D0 */
   E3_FaultSave		DProcData2;							/* 0x00E8E0 */
   E3_FaultSave		DProcData3;							/* 0x00E8F0 */
   E3_uint32		dummy3[0xc0];							/* 0x00E900 */
   E3_IprocTrapHeader	VCh0_C0_TrHead[MAX_TRAPPED_TRANS];
   E3_IprocTrapHeader	VCh0_NonC0_TrHead[MAX_TRAPPED_TRANS];
   E3_IprocTrapHeader	VCh1_C0_TrHead[MAX_TRAPPED_TRANS];
   E3_IprocTrapHeader	VCh1_NonC0_TrHead[MAX_TRAPPED_TRANS];
   E3_IprocTrapData	VCh0_C0_TrData[MAX_TRAPPED_TRANS];
   E3_IprocTrapData	VCh0_NonC0_TrData[MAX_TRAPPED_TRANS];
   E3_IprocTrapData	VCh1_C0_TrData[MAX_TRAPPED_TRANS];
   E3_IprocTrapData	VCh1_NonC0_TrData[MAX_TRAPPED_TRANS];
   E3_uint64		DmaOverflowQueueSpace[0x1000];
   E3_uint64		ThreadOverflowQueueSpace[0x800];
   E3_uint64		EventOverflowQueueSpace[0x800];
} E3_TrapAndQueue;


typedef struct _E3_ContextControlBlock 
{
   E3_uint32	rootPTP;
   E3_uint32	filter;
   E3_uint32	VPT_ptr;
   E3_uint32	VPT_mask;
} E3_ContextControlBlock;

#define E3_CCB_CNTX0		(0x20000000)
#define E3_CCB_DISCARD_ALL	(0x40000000)
#define E3_CCB_ACKOK_ALL	(0x80000000)
#define E3_CCB_MASK		(0xc0000000)

#define E3_NUM_CONTEXT_0	(0x20)

/* Macros to manipulate event queue pointers */
/*     generate index in EventIntQueue */
#define E3_EVENT_INTQ_INDEX(fptr)	(((fptr) & 0x1fff) >> 3)
/*     generate next fptr */
#define E3_EVENT_INTQ_NEXT(fptr)	((((fptr) + 8) & ~0x4000) | 0x2000)


#endif /* notdef _ELAN3_ELANREGS_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
