/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_ELANSYSCALL_H
#define __ELAN3_ELANSYSCALL_H

#ident "$Id: elansyscall.h,v 1.34 2004/06/07 13:50:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elansyscall.h,v $*/

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _ASM

typedef struct sys_word_item
{
    struct sys_word_item *Next;
    E3_uint32	    	  Value;
} SYS_WORD_ITEM;

typedef struct sys_block_item
{
    struct sys_block_item *Next;
    E3_uint32		  *Pointer;
} SYS_BLOCK_ITEM;

typedef struct sys_swap_space
{
    int		 Magic;
    void	*ItemListsHead[MAX_LISTS];
    void       **ItemListsTailp[MAX_LISTS];
} SYS_SWAP_SPACE;

typedef struct sys_exception
{
    int			Type;
    int			Proc;
    u_long		Res;
    u_long		Value;
    E3_FaultSave_BE	FaultArea;
    
    union
    {
	DMA_TRAP	Dma;
	THREAD_TRAP	Thread;
	COMMAND_TRAP	Command;
	INPUT_TRAP	Input;
    }			Union;
} SYS_EXCEPTION;

typedef struct sys_exception_space
{
    struct sys_exception_space *Next;
    int			        Magic;
    int			        Front;
    int			        Back;
    int			        Count;
    int			        Overflow;
    SYS_EXCEPTION	        Exceptions[1];
} SYS_EXCEPTION_SPACE;

#ifdef __KERNEL__

typedef struct sys_ctxt
{
    SYS_SWAP_SPACE      *Swap;
    SYS_EXCEPTION_SPACE *Exceptions;
    kmutex_t	         Lock;

    spinlock_t	         WaitLock;
    kcondvar_t	         NetworkErrorWait;

    int		         Armed;
    int		         Backoff;
    long	         Time;

    u_long		 Flags;
    int                  signal;

    EVENT_COOKIE_TABLE  *Table;
} SYS_CTXT;

extern SYS_CTXT *sys_init (ELAN3_CTXT *ctxt);
extern int       sys_waitevent (ELAN3_CTXT *ctxt, E3_Event *event);
extern void      sys_addException (SYS_CTXT *sctx, int type, int proc, caddr_t ptr, int size, 
				   E3_FaultSave_BE *, u_long res, u_long value);
extern int       sys_getException (SYS_CTXT *sctx, SYS_EXCEPTION *ex);

/* returns -ve error or ELAN_CAP_OK or ELAN_CAP_RMS */
/* use = ELAN_USER_ATTACH, ELAN_USER_P2P, ELAN_USER_BROADCAST */
extern int  elan3_validate_cap (ELAN3_DEV *dev, ELAN_CAPABILITY *cap ,int use);

#endif /* __KERNEL__ */

#endif /* _ASM */

/* values for "Flags" */
#define ELAN3_SYS_FLAG_DMA_BADVP		1
#define ELAN3_SYS_FLAG_THREAD_BADVP	2
#define ELAN3_SYS_FLAG_DMAFAIL		4
#define ELAN3_SYS_FLAG_NETERR		8

#define SYS_SWAP_MAGIC		0xB23C52DF
#define SYS_EXCEPTION_MAGIC	0xC34D63E0

#define EXCEPTION_GLOBAL_STRING	"elan3_exceptions"
#define EXCEPTION_ABORT_STRING  "elan3_abortstring"

#ifdef __cplusplus
}
#endif

#endif /* __ELAN3_ELANSYSCALL_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
