/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_ELANUREGS_H
#define __ELAN3_ELANUREGS_H

#ident "$Id: elanuregs.h,v 1.10 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elanuregs.h,v $*/

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Statistic control reg values
 * Each 4-bit nibble of the control word specifies what statistic
 * is to be recorded in each of the 8 statistic counters
 */

/* Count reg 0 */
#define STC_INPUT_TRANSACTIONS		0
#define STP_DMA_EOP_WAIT_ACK		1
#define STP_THREAD_RUNNING		2
#define STP_UCODE_WAIT_MEM		3
#define STC_CACHE_WRITE_BACKS		4
#define STC_PCI_SLAVE_READS		5
#define STC_REG0_UNUSED6		6
#define STP_REG0_UNUSED7		7

#define STATS_REG0_NAMES {		\
	  "STC_INPUT_TRANSACTIONS",	\
	  "STP_DMA_EOP_WAIT_ACK",	\
	  "STP_THREAD_RUNNING",		\
	  "STP_UCODE_WAIT_MEM",		\
	  "STC_CACHE_WRITE_BACKS",	\
	  "STC_PCI_SLAVE_READS",	\
	  "STC_REG0_UNUSED6",		\
	  "STP_REG0_UNUSED7"		\
}

/* Count reg 1 */
#define STC_INPUT_WRITE_BLOCKS		(0 << 4)
#define STP_DMA_DATA_TRANSMITTING	(1 << 4)
#define STP_THEAD_WAITING_INST		(2 << 4)
#define STC_REG1_UNUSED3		(3 << 4)
#define STP_FETCHING_ROUTES		(4 << 4)
#define STC_REG1_UNUSED5		(5 << 4)
#define STC_PCI_SLAVE_WRITES		(6 << 4)
#define STP_PCI_SLAVE_READ_WAITING	(7 << 4)

#define STATS_REG1_NAMES {		\
      "STC_INPUT_WRITE_BLOCKS",		\
	  "STP_DMA_DATA_TRANSMITTING",	\
	  "STP_THEAD_WAITING_INST",	\
	  "STC_REG1_UNUSED3",		\
	  "STP_FETCHING_ROUTES",	\
	  "STC_REG1_UNUSED5",		\
	  "STC_PCI_SLAVE_WRITES",	\
	  "STP_PCI_SLAVE_READ_WAITING"	\
}

/* Count reg 2 */
#define STC_INPUT_PKTS			(0 << 8)
#define STP_DMA_WAITING_MEM		(1 << 8)
#define STP_THREAD_WAIT_OPEN_PKT	(2 << 8)
#define STC_REG2_UNUSED3		(3 << 8)
#define STC_ROUTE_FETCHES		(4 << 8)
#define STC_CACHE_NON_ALLOC_MISSES	(5 << 8)
#define STC_REG2_UNUSED6		(6 << 8)
#define STP_PCI_SLAVE_WRITE_WAITING	(7 << 8)

#define STATS_REG2_NAMES {		\
      "STC_INPUT_PKTS",			\
	  "STP_DMA_WAITING_MEM",	\
	  "STP_THREAD_WAIT_OPEN_PKT",	\
	  "STC_REG2_UNUSED3",		\
	  "STC_ROUTE_FETCHES",		\
	  "STC_CACHE_NON_ALLOC_MISSES",	\
	  "STC_REG2_UNUSED6",		\
	  "STP_PCI_SLAVE_WRITE_WAITING"	\
}

/* Count reg 3 */
#define STC_INPUT_PKTS_REJECTED		(0 << 12)
#define STP_DMA_WAIT_NETWORK_BUSY	(1 << 12)
#define STP_THREAD_WAIT_PACK		(2 << 12)
#define STP_UCODE_BLOCKED_UCODE		(3 << 12)
#define STC_TLB_HITS			(4 << 12)
#define STC_REG3_UNUSED5		(5 << 12)
#define STC_PCI_MASTER_READS		(6 << 12)
#define STP_PCI_MASTER_WRITE_WAITING	(7 << 12)

#define STATS_REG3_NAMES {		\
      "STC_INPUT_PKTS_REJECTED",	\
	  "STP_DMA_WAIT_NETWORK_BUSY",	\
	  "STP_THREAD_WAIT_PACK",	\
	  "STP_UCODE_BLOCKED_UCODE",	\
	  "STC_TLB_HITS",		\
	  "STC_REG3_UNUSED5",		\
	  "STC_PCI_MASTER_READS",	\
	  "STP_PCI_MASTER_WRITE_WAITING"\
}

/* Count reg 4 */
#define STP_INPUT_DATA_TRANSMITTING	(0 << 16)
#define STC_DMA_NON_CTX0_PKTS		(1 << 16)
#define STP_THREAD_EOP_WAIT_ACK		(2 << 16)
#define STP_UCODE_DPROC_RUNNING		(3 << 16)
#define STC_TLB_MEM_WALKS		(4 << 16)
#define STC_REG4_UNUSED5		(5 << 16)
#define STC_PCI_MASTER_WRITES		(6 << 16)
#define STP_PCI_MASTER_READ_WAITING	(7 << 16)

#define STATS_REG4_NAMES {		\
      "STP_INPUT_DATA_TRANSMITTING",	\
	  "STC_DMA_NON_CTX0_PKTS",	\
	  "STP_THREAD_EOP_WAIT_ACK",	\
	  "STP_UCODE_DPROC_RUNNING",	\
	  "STC_TLB_MEM_WALKS",		\
	  "STC_REG4_UNUSED5",		\
	  "STC_PCI_MASTER_WRITES",	\
	  "STP_PCI_MASTER_READ_WAITING"	\
}

/* Count reg 5 */
#define STP_INPUT_WAITING_NETWORK_DATA	(0 << 20)
#define STC_DMA_NON_CTX0_PKTS_REJECTED	(1 << 20)
#define STP_THREAD_WAITING_DATA		(2 << 20)
#define STP_UCODE_CPROC_RUNNING		(3 << 20)
#define STP_THREAD_TRANSMITTING_DATA	(4 << 20)
#define STP_PCI_WAITING_MAIN		(5 << 20)
#define STC_REG5_UNUSED6		(6 << 20)
#define STC_REG5_UNUSED7		(7 << 20)

#define STATS_REG5_NAMES {			\
      "STP_INPUT_WAITING_NETWORK_DATA",		\
	  "STC_DMA_NON_CTX0_PKTS_REJECTED",	\
	  "STP_THREAD_WAITING_DATA",		\
	  "STP_UCODE_CPROC_RUNNING",		\
	  "STP_THREAD_TRANSMITTING_DATA",	\
	  "STP_PCI_WAITING_MAIN",		\
	  "STC_REG5_UNUSED6",			\
	  "STC_REG5_UNUSED7"			\
}

/* Count reg 6 */
#define STP_INPUT_WAITING_MEMORY	(0 << 24)
#define STC_DMA_CTX0_PKTS		(1 << 24)
#define STP_THREAD_WAITING_MEMORY	(2 << 24)
#define STP_UCODE_TPROC_RUNNING		(3 << 24)
#define STC_CACHE_HITS			(4 << 24)
#define STP_PCI_WAITING_ELAN		(5 << 24)
#define STC_REG6_UNUSED4		(6 << 24)
#define STC_REG6_UNUSED7		(7 << 24)

#define STATS_REG6_NAMES {		\
      "STP_INPUT_WAITING_MEMORY",	\
	  "STC_DMA_CTX0_PKTS",		\
	  "STP_THREAD_WAITING_MEMORY",	\
	  "STP_UCODE_TPROC_RUNNING",	\
	  "STC_CACHE_HITS",		\
	  "STP_PCI_WAITING_ELAN",	\
	  "STC_REG6_UNUSED4",		\
	  "STC_REG6_UNUSED7"		\
}

/* Count reg 7 */
#define STC_INPUT_CTX_FILTER_FILL	(0 << 28)	
#define STC_DMA_CTX0_PKTS_REJECTED	(1 << 28)
#define STP_THREAD_WAIT_NETWORK_BUSY	(2 << 28)
#define STP_UCODE_IPROC_RUNNING		(3 << 28)
#define STP_TLB_MEM_WALKING		(4 << 28)
#define STC_CACHE_ALLOC_MISSES		(5 << 28)
#define STP_PCI_DATA_TRANSFER		(6 << 28)
#define STC_REG7_UNUSED7		(7 << 28)

#define STATS_REG7_NAMES {		\
      "STC_INPUT_CTX_FILTER_FILL",	\
	  "STC_DMA_CTX0_PKTS_REJECTED",	\
	  "STP_THREAD_WAIT_NETWORK_BUSY",\
	  "STP_UCODE_IPROC_RUNNING",	\
	  "STP_TLB_MEM_WALKING",	\
	  "STC_CACHE_ALLOC_MISSES",	\
	  "STP_PCI_DATA_TRANSFER",	\
	  "STC_REG7_UNUSED7"		\
}

#define STATS_REG_NAMES { \
    STATS_REG0_NAMES, \
    STATS_REG1_NAMES, \
    STATS_REG2_NAMES, \
    STATS_REG3_NAMES, \
    STATS_REG4_NAMES, \
    STATS_REG5_NAMES, \
    STATS_REG6_NAMES, \
    STATS_REG7_NAMES, \
}

extern const char *elan3_stats_names[8][8];

#define ELAN3_STATS_NAME(COUNT, CONTROL) (elan3_stats_names[(COUNT)][(CONTROL) & 7])

typedef volatile union e3_StatsControl
{
   E3_uint32 StatsControl;
   struct
   {
#if defined(__LITTLE_ENDIAN__)
      E3_uint32 StatCont0:4;
      E3_uint32 StatCont1:4;
      E3_uint32 StatCont2:4;
      E3_uint32 StatCont3:4;
      E3_uint32 StatCont4:4;
      E3_uint32 StatCont5:4;
      E3_uint32 StatCont6:4;
      E3_uint32 StatCont7:4;
#else
      E3_uint32 StatCont7:4;
      E3_uint32 StatCont6:4;
      E3_uint32 StatCont5:4;
      E3_uint32 StatCont4:4;
      E3_uint32 StatCont3:4;
      E3_uint32 StatCont2:4;
      E3_uint32 StatCont1:4;
      E3_uint32 StatCont0:4;
#endif
   } s;
} E3_StatsControl;

typedef volatile union e3_StatsCount
{
   E3_uint64    ClockStat; 
   struct
   {
       E3_uint32 ClockLSW;	/* read only */
       E3_uint32 StatsCount;
   } s;
} E3_StatsCount;

typedef volatile union e3_clock
{
   E3_uint64 NanoSecClock;
   struct
   {
      E3_uint32 ClockLSW;
      E3_uint32 ClockMSW;
   } s;
} E3_Clock;
#define E3_TIME( X ) ((X).NanoSecClock)

typedef volatile struct _E3_User_Regs
{
   E3_StatsCount	StatCounts[8];
   E3_StatsCount	InstCount;
   E3_uint32		pad0;
   E3_StatsControl	StatCont;
   E3_Clock		Clock;
   E3_uint32		pad1[0x7ea];
} E3_User_Regs;

typedef volatile struct _E3_CommandPort 
{
   E3_Addr		PutDma;		/* 0x000 */
   E3_uint32		Pad1;
   E3_Addr		GetDma;		/* 0x008 */
   E3_uint32		Pad2;
   E3_Addr		RunThread;	/* 0x010 */
   E3_uint32		Pad3[3];
   E3_Addr		WaitEvent0;	/* 0x020 */
   E3_uint32		Pad4;
   E3_Addr 		WaitEvent1;	/* 0x028 */
   E3_uint32		Pad5;
   E3_Addr 		SetEvent;	/* 0x030 */
   E3_uint32		Pad6[3];
   E3_uint32		Pad7[0x7f0];	/* Fill out to an 8K page */
} E3_CommandPort;
/* Should have the new structures for the top four pages of the elan3 space */

#define E3_COMMANDPORT_SIZE	(sizeof (E3_CommandPort))

#ifdef __cplusplus
}
#endif

#endif /* __ELAN3_ELANUREGS_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
