/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_ELANVP_H
#define _ELAN3_ELANVP_H

#ident "$Id: elanvp.h,v 1.45 2004/06/18 09:28:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/elanvp.h,v $ */

#include <elan3/e3types.h>
#include <elan/bitmap.h>
#include <elan/capability.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Context number allocation.
 * [0-31]	system contexts
 * [32-63]	hardware test
 * [64-1023]	available
 * [1024-2047]	RMS allocatable
 * [2048-4095]	kernel comms data contexts
 */
#define ELAN3_KCOMM_CONTEXT_NUM		0x001			/* old kernel comms context (system) */
#define ELAN3_CM_CONTEXT_NUM		0x002			/* new cluster member ship comms context (system) */
#define ELAN3_MRF_CONTEXT_NUM		0x003			/* multi-rail kernel comms context */
#define ELAN3_DMARING_BASE_CONTEXT_NUM	0x010			/* 16 contexts for dma ring issue (system) */
#define ELAN3_DMARING_TOP_CONTEXT_NUM	0x01f

#define ELAN3_HWTEST_BASE_CONTEXT_NUM	0x020			/* reserved for hardware test */
#define ELAN3_HWTEST_TOP_CONTEXT_NUM	0x03f

#define ELAN3_KCOMM_BASE_CONTEXT_NUM	0x800			/* kernel comms data transfer contexts */
#define ELAN3_KCOMM_TOP_CONTEXT_NUM	0xfff

#define ELAN3_HWTEST_CONTEXT(ctx)     	((ctx) >= ELAN3_HWTEST_BASE_CONTEXT_NUM && \
					 (ctx) <= ELAN3_HWTEST_TOP_CONTEXT_NUM)    

#define ELAN3_SYSTEM_CONTEXT(ctx)	(((ctx) & SYS_CONTEXT_BIT) != 0 || \
					 (ctx) < E3_NUM_CONTEXT_0 || \
					 (ctx) >= ELAN3_KCOMM_BASE_CONTEXT_NUM)

/* Maximum number of virtual processes */
#define ELAN3_MAX_VPS		(16384)

#define ELAN3_INVALID_PROCESS	(0x7fffffff)		/* A GUARANTEED invalid process # */
#define ELAN3_INVALID_NODE	(0xFFFF)
#define ELAN3_INVALID_CONTEXT	(0xFFFF)



#if defined(__KERNEL__) && !defined(__ELAN3__)

/*
 * Contexts are accessible via Elan capabilities,
 * for each context that can be "attached" to there
 * is a ELAN3_CTXT_INFO structure created by its
 * "owner".  This also "remembers" all remote 
 * segments that have "blazed" a trail to it.
 *
 * If the "owner" goes away the soft info is 
 * destroyed when it is no longer "attached" or 
 * "referenced" by a remote segment.
 *
 * If the owner changes the capability, then 
 * the soft info must be not "referenced" or 
 * "attached" before a new process can "attach"
 * to it.
 */

_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::InfoLock,
			  elan3_info::Next elan3_info::Prev elan3_info::Device elan3_info::Owner
			  elan3_info::Capability elan3_info::AttachedCapability elan3_info::Context))
_NOTE(MUTEX_PROTECTS_DATA(elan3_dev::IntrLock,
			  elan3_info::Nacking elan3_info::Disabled))
_NOTE(DATA_READABLE_WITHOUT_LOCK(elan3_info::Context elan3_info::Device elan3_info::Capability))

#endif /* __KERNEL__ */

#define LOW_ROUTE_PRIORITY	0
#define HIGH_ROUTE_PRIORITY	1

#define DEFAULT_ROUTE_TIMEOUT	3
#define DEFAULT_ROUTE_PRIORITY	LOW_ROUTE_PRIORITY


/* a small route is 4 flits (8 bytes), a big route  */
/* is 8 flits (16 bytes) - each packed route is 4 bits */
/* so giving us a maximum of 28 as flit0 does not contain */
/* packed routes */
#define MAX_FLITS		8
#define MAX_PACKED		28

/* bit definitions for 64 bit route pointer */
#define ROUTE_VALID		(1ULL << 63)
#define ROUTE_PTR		(1ULL << 62)
#define ROUTE_CTXT_SHIFT	48
#define ROUTE_PTR_MASK		((1ull << ROUTE_CTXT_SHIFT)-1)
#define ROUTE_GET_CTXT          ((VAL >> ROUTE_CTXT_SHIFT) & 0x3fff )

#define SMALL_ROUTE(flits, context)	(((E3_uint64) (flits)[0] <<  0) | ((E3_uint64) (flits)[1] << 16) | \
					 ((E3_uint64) (flits)[2] << 32) | ((E3_uint64) (context) << ROUTE_CTXT_SHIFT) | \
					 ROUTE_VALID)

#define BIG_ROUTE_PTR(paddr, context)	((E3_uint64) (paddr) | ((E3_uint64) context << ROUTE_CTXT_SHIFT) | ROUTE_VALID | ROUTE_PTR)

#define BIG_ROUTE0(flits)		(((E3_uint64) (flits)[0] <<  0) | ((E3_uint64) (flits)[1] << 16) | \
					 ((E3_uint64) (flits)[2] << 32) | ((E3_uint64) (flits)[3] << 48))
#define BIG_ROUTE1(flits)		(((E3_uint64) (flits)[4] <<  0) | ((E3_uint64) (flits)[5] << 16) | \
					 ((E3_uint64) (flits)[6] << 32) | ((E3_uint64) (flits)[7] << 48))


/* defines for first flit of a route */
#define FIRST_HIGH_PRI		(1 << 15)
#define FIRST_AGE(Val)		((Val) << 11)
#define FIRST_TIMEOUT(Val)	((Val) << 9)
#define FIRST_PACKED(X)		((X) << 7)
#define FIRST_ROUTE(Val)	(Val)
#define FIRST_ADAPTIVE		(0x30)
#define FIRST_BCAST_TREE	(0x20)
#define FIRST_MYLINK		(0x10)
#define FIRST_BCAST(Top, Bot)	(0x40 | ((Top) << 3) | (Bot))

/* defines for 3 bit packed entries for subsequent flits */
#define PACKED_ROUTE(Val)	(8 | (Val))
#define PACKED_ADAPTIVE		(3)
#define PACKED_BCAST_TREE	(2)
#define PACKED_MYLINK		(1)
#define PACKED_BCAST0(Top,Bot)	(4 | (Bot & 3))
#define PACKED_BCAST1(Top,Bot)	((Top << 1) | (Bot >> 2))

/* ---------------------------------------------------------- 
 * elan3_route functions 
 * return ELAN3_ROUTE_xxx codes
 * ---------------------------------------------------------- */

#define ELAN3_ROUTE_SUCCESS        (0x00)
#define ELAN3_ROUTE_SYSCALL_FAILED (0x01)
#define ELAN3_ROUTE_INVALID        (0x02)
#define ELAN3_ROUTE_TOO_LONG       (0x04)
#define ELAN3_ROUTE_LOAD_FAILED    (0x08)
#define ELAN3_ROUTE_PROC_RANGE     (0x0f)
#define ELAN3_ROUTE_INVALID_LEVEL  (0x10)
#define ELAN3_ROUTE_OCILATES       (0x20)
#define ELAN3_ROUTE_WRONG_DEST     (0x40)
#define ELAN3_ROUTE_TURN_LEVEL     (0x80)
#define ELAN3_ROUTE_NODEID_UNKNOWN (0xf0)

#ifdef __cplusplus
}
#endif

#endif /* _ELAN3_ELANVP_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
