/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_EVENTS_H
#define _ELAN3_EVENTS_H

#ident "$Id: events.h,v 1.45 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/events.h,v $*/

/*
 * Alignments for events, event queues and blockcopy blocks.
 */
#define E3_EVENT_ALIGN 		(8)
#define E3_QUEUE_ALIGN		(32)
#define E3_BLK_ALIGN		(64)
#define E3_BLK_SIZE		(64)
#define E3_BLK_PATTERN 		(0xfeedface)

#define E3_EVENT_FREE		((0 << 4) | EV_WCOPY)
#define E3_EVENT_PENDING	((1 << 4) | EV_WCOPY)
#define E3_EVENT_ACTIVE		((2 << 4) | EV_WCOPY)
#define E3_EVENT_FIRED		((3 << 4) | EV_WCOPY)
#define E3_EVENT_FAILED		((4 << 4) | EV_WCOPY)
#define E3_EVENT_DONE		((5 << 4) | EV_WCOPY)
#define E3_EVENT_PRIVATE	((6 << 4) | EV_WCOPY)

/*
 * Event values and masks
 *
 * Block Copy event	xxxxxxxxxxxxxxxx1
 * Chained event 	30 bit ptr ....0x
 * Event interrupt	29 bit cookie 01x
 * Dma event		28 bit ptr   011x
 * thread event		28 bit ptr   111x
 */
#define EV_CLEAR		(0x00000000)
#define EV_TYPE_BCOPY		(0x00000001)
#define EV_TYPE_CHAIN		(0x00000000)
#define EV_TYPE_EVIRQ		(0x00000002)
#define EV_TYPE_DMA		(0x00000006)
#define EV_TYPE_THREAD		(0x0000000e)

#define EV_TYPE_BCOPY_BYTE	(0)
#define EV_TYPE_BCOPY_HWORD	(1)
#define EV_TYPE_BCOPY_WORD	(2)
#define EV_TYPE_BCOPY_DWORD	(3)

/*
 * Data type is in the lowest two bits of the Dest pointer.
 */
#define EV_BCOPY_DTYPE_MASK	(3)
#define EV_WCOPY		(1)	/* [DestWord] = Source */
#define EV_BCOPY		(0)	/* [DestBlock] = [SourceBlock] */

#define EV_TYPE_MASK		(0x0000000e)
#define EV_TYPE_MASK_BCOPY	(0x00000001)
#define EV_TYPE_MASK_CHAIN	(0x00000002)
#define EV_TYPE_MASK_EVIRQ	(0x00000006)
#define EV_TYPE_MASK_DMA	(0x0000000e)
#define EV_TYPE_MASK_THREAD	(0x0000000e)
#define EV_TYPE_MASK2		(0x0000000f)

/*
 * Min/Max size for Elan queue entries 
 */
#define E3_QUEUE_MIN	E3_BLK_SIZE
#define E3_QUEUE_MAX	(E3_BLK_SIZE * 5)

/*
 * Elan queue state bits
 */
#define E3_QUEUE_FULL	(1<<0)
#define E3_QUEUE_LOCKED	(1<<8)

#ifndef _ASM

typedef union _E3_Event
{
   E3_uint64	ev_Int64;
   struct {
      volatile E3_int32	u_Count;
      E3_uint32		u_Type;
   } ev_u;
} E3_Event;

typedef union _E3_BlockCopyEvent
{
   E3_uint64 ev_ForceAlign;
   struct E3_BlockCopyEvent_u {
      volatile E3_int32	u_Count;
      E3_uint32		u_Type;
      E3_Addr	        u_Source;
      E3_Addr	        u_Dest;	  /* lowest bits are the data type for endian conversion */
   } ev_u;
} E3_BlockCopyEvent;

#define ev_Type   ev_u.u_Type
#define ev_Count  ev_u.u_Count
#define ev_Source ev_u.u_Source
#define ev_Dest   ev_u.u_Dest

typedef union _E3_WaitEvent0
{
   E3_uint64            we_ForceAlign;
   struct {
	E3_Addr		u_EventLoc;
	E3_int32	u_WaitCount;
   } we_u;
} E3_WaitEvent0;
#define we_EventLoc we_u.u_EventLoc
#define we_WaitCount we_u.u_WaitCount

typedef union _E3_Event_Blk
{
    E3_uint8  eb_Bytes[E3_BLK_SIZE];
    E3_uint32 eb_Int32[E3_BLK_SIZE/sizeof (E3_uint32)];
    E3_uint64 eb_Int64[E3_BLK_SIZE/sizeof (E3_uint64)];
} E3_Event_Blk;

/* We make eb_done the last word of the blk
 * so that we can guarantee the rest of the blk is
 * correct when this value is set.
 * However, when the TPORT code copies the envelope
 * info into the blk, it uses a dword endian type.
 * Thus we must correct for this when initialising
 * the pattern in the Elan SDRAM blk (eeb_done)
 */
#define eb_done eb_Int32[15]
#define eeb_done eb_Int32[15^WordEndianFlip]

#define EVENT_WORD_READY(WORD) (*((volatile E3_uint32 *) WORD) != 0)
#define EVENT_BLK_READY(BLK) (((volatile E3_Event_Blk *) (BLK))->eb_done != 0)
#define EVENT_READY(EVENT)   (((volatile E3_Event *) (EVENT))->ev_Count <= 0)

#define ELAN3_WAIT_EVENT (0)
#define ELAN3_POLL_EVENT (-1)

#define SETUP_EVENT_TYPE(ptr,typeval) (((unsigned long)(ptr)) | (typeval))

#define E3_RESET_BCOPY_BLOCK(BLK) 				  			\
	do {										\
		(BLK)->eb_done = 0;							\
	} while (0)

typedef struct e3_queue
{
   volatile E3_uint32	q_state;	/* queue is full=bit0, queue is locked=bit8 */
   volatile E3_Addr	q_bptr;		/* block aligned ptr to current back item */
   E3_uint32		q_size;		/* size of queue item; 0x1 <= size <= (0x40 * 5) */
   E3_Addr		q_top;		/* block aligned ptr to last queue item */
   E3_Addr		q_base;		/* block aligned ptr to first queue item */
   volatile E3_Addr	q_fptr;		/* block aligned ptr to current front item */
   E3_Event		q_event;	/* queue event */
} E3_Queue;

typedef struct e3_blockcopy_queue
{
   volatile E3_uint32	q_state;	/* queue is full=bit0, queue is locked=bit8 */
   volatile E3_Addr	q_bptr;		/* block aligned ptr to current back item */
   E3_uint32		q_size;		/* size of queue item; 0x1 <= size <= (0x40 * 5) */
   E3_Addr		q_top;		/* block aligned ptr to last queue item */
   E3_Addr		q_base;		/* block aligned ptr to first queue item */
   volatile E3_Addr	q_fptr;		/* block aligned ptr to current front item */
   E3_BlockCopyEvent	q_event;	/* queue event */
   E3_uint32		q_pad[6];
} E3_BlockCopyQueue;

#define E3_QUEUE_EVENT_OFFSET	24
#define QUEUE_FULL(Q)		((Q)->q_state & E3_QUEUE_FULL)		

#endif /* ! _ASM */

#endif /* _ELAN3_EVENTS_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
