/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_INTRINSICS_H
#define _ELAN3_INTRINSICS_H

#ident "$Id: intrinsics.h,v 1.35 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/intrinsics.h,v $ */

#include <elan3/e3types.h>
#include <elan3/events.h>

/* 
 * This file contains definitions of the macros for accessing the QSW
 * specific instructions, as if they were functions.
 * The results from the function 
 */

#define C_ACK_OK	0			/* return from c_close() */
#define C_ACK_TESTFAIL	1			/* return from c_close() */
#define C_ACK_DISCARD	2			/* return from c_close() */
#define C_ACK_ERROR	3			/* return from c_close() */

/*
 * Elan asi's for tproc block accesses
 */
#define EASI_BYTE	0
#define EASI_HALF	1
#define EASI_WORD	2
#define EASI_DOUBLE	3

#if defined(__ELAN3__) && !defined (_ASM)

extern inline void c_abort(void) 
{
    asm volatile (".word 0x0000		! die you thread you " : : );
}

extern inline void c_suspend(void) 
{
    asm volatile (
	"set 1f, %%i7			! RevB bug fix. get address of the wakeup inst\n"
	"andcc %%i7,0x4,%%g0		! RevB bug fix. check alignment\n"
	"bne 1f				! RevB bug fix. jump to other alignment\n"
	"nop				! RevB bug fix. delay slot\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"suspend			! do the real suspend\n"
	"1: add %%i7,5*4,%%i7		! RevB bug fix. Point i7 to first ldblock\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"suspend			! do the real suspend\n" : : );
}

extern inline int c_close(void) 
{
    register int rc asm("o0");

    asm volatile ("close %0" : "=r" (rc) : );

    return (rc);
}

extern inline int c_close_cookie(volatile E3_uint32 *cookiep, E3_uint32 next)
{
    register int rc asm("o0");

    asm volatile ("close 	%0		! close the packet\n"
		  "bz,a		1f		! ack received\n"
		  "st		%1, [%2]	! update cookie on ack\n"
		  "1:				! label for not-ack\n"
		  : "=r" (rc) : "r" (next), "r" (cookiep));

    return (rc);
}

extern inline void c_break_busywait(void)
{
    asm volatile (
	"breaktest			! test to see if break necessary\n"
	"bpos 1f			! no other thread ready\n"
	"nop				! delay slot\n"
	"sub     %%sp,3*8*4,%%sp	! Space to save the registers\n"
	"stblock %%g0,[%%sp+0]		! save the globals\n"
	"stblock %%i0,[%%sp+8*4]	! save the ins\n"
	"stblock %%l0,[%%sp+16*4]	! save the locals\n"
	"set 2f, %%i7			! RevB bug fix. get address of the wakeup inst\n"
	"andcc %%i7,0x4,%%g0		! RevB bug fix. check alignment\n"
	"bne 3f				! RevB bug fix. jump to other alignment\n"
	"nop				! RevB bug fix. delay slot\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"break				! do the real break\n"
	"2: b 4f			! RevB bug fix. Branch over other alignment case\n"
	" ldblock [%%sp+16*4],%%l0	! RevB bug fix. restore locals in delay slot\n"
	"3: add %%i7,5*4,%%i7		! RevB bug fix. Point i7 to first ldblock\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"break				! do the real break\n"
	"ldblock [%%sp+16*4],%%l0	! restore locals\n"
	"4: ldblock [%%sp+8*4], %%i0	! restore ins\n"
	"ldblock [%%sp+0],%%g0		! restore globals\n"
	"add     %%sp,3*8*4,%%sp	! restore stack pointer\n"
	"1: " : : );
}

extern inline void c_break(void)
{
    asm volatile (
	"breaktest			! test to see if break necessary\n"
	"bne 1f				! haven't exceeded our inst count yet\n"
	"nop				! delay slot\n"
	"sub     %%sp,3*8*4,%%sp	! Space to save the registers\n"
	"stblock %%g0,[%%sp+0]		! save the globals\n"
	"stblock %%i0,[%%sp+8*4]	! save the ins\n"
	"stblock %%l0,[%%sp+16*4]	! save the locals\n"
	"set 2f, %%i7			! RevB bug fix. get address of the wakeup inst\n"
	"andcc %%i7,0x4,%%g0		! RevB bug fix. check alignment\n"
	"bne 3f				! RevB bug fix. jump to other alignment\n"
	"nop				! RevB bug fix. delay slot\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"break				! do the real break\n"
	"2: b 4f			! RevB bug fix. Branch over other alignment case\n"
	" ldblock [%%sp+16*4],%%l0	! RevB bug fix. restore locals in delay slot\n"
	"3: add %%i7,5*4,%%i7		! RevB bug fix. Point i7 to first ldblock\n"
	"ldd [%%i7],%%i6		! RevB bug fix. data fetch of instructions\n"
	"break				! do the real break\n"
	"ldblock [%%sp+16*4],%%l0	! restore locals\n"
	"4: ldblock [%%sp+8*4], %%i0	! restore ins\n"
	"ldblock [%%sp+0],%%g0		! restore globals\n"
	"add     %%sp,3*8*4,%%sp	! restore stack pointer\n"
	"1: " : : );
}

extern inline void c_open( const int arg ) 
{
    asm volatile ("open %0" : : "r" (arg) );
    asm volatile ("nop; nop; nop; nop");
    asm volatile ("nop; nop; nop; nop");
    asm volatile ("nop; nop; nop; nop");
    asm volatile ("nop; nop; nop; nop");
    asm volatile ("nop; nop; nop; nop");
    asm volatile ("nop; nop; nop; nop");
}

extern inline void c_waitevent( volatile E3_Event *const ptr,
				const int count) 
{
    register volatile E3_Event *a_unlikely asm("o0") = ptr;
    register int a_very_unlikely asm("o1") = count;

    asm volatile (
        "sub     %%sp,1*8*4,%%sp	! Space to save the registers\n"
        "stblock %%i0,[%%sp+0]		! save the ins\n"
	"set	2f, %%i7		! RevB bug fix. get address of the wakeup inst\n"
	"andcc %%i7,0x4,%%g0		! RevB bug fix. check alignment\n"
	"bne 3f				! RevB bug fix. jump to other alignment\n"
	"nop				! RevB bug fix. delay slot\n"
	"ldd [%%i7],%%i4		! RevB bug fix. data fetch of instructions\n"
        "waitevent			! do the business\n"
	"2: b 4f			! RevB bug fix. Branch over other alignment case\n"
        "  ldblock [%%sp+0],%%i0	! RevB bug fix. restore ins in delay slot\n"
	"3: add %%i7,5*4,%%i7		! RevB bug fix. Point i7 to first ldblock\n"
	"ldd [%%i7],%%i4		! RevB bug fix. data fetch of instructions\n"
        "waitevent			! do the business\n"
        "ldblock [%%sp+0],%%i0		! restore ins\n"
        "4: add     %%sp,1*8*4,%%sp	! restore stack pointer\n"
        : /* no outputs */
        : /* inputs */ "r" (a_unlikely), "r" (a_very_unlikely)
        : /* clobbered */ "g0", "g1", "g2", "g3", "g4", "g5", "g6", "g7",
      			  "l0", "l1", "l2", "l3", "l4", "l5", "l6", "l7" );

}

#define c_sendtrans0(type,dest)			\
	asm volatile ("sendtrans %0, %%g0, %1" : : "i" (type), "r" (dest))

#define c_sendtrans1(type,dest,arg)		\
	asm volatile ("sendtrans %0, %2, %1" : : "i" (type), "r" (dest), "r" (arg))

#define c_sendtrans2(type,dest,arg1,arg2)	\
	do {					\
	     register const unsigned long a_unlikely_1 asm("o4") = arg1;			\
	     register const unsigned long a_unlikely_2 asm("o5") = arg2;			\
	     asm volatile ("sendtrans %0, %2, %1"					\
		 : : "i" (type), "r" (dest), "r" (a_unlikely_1), "r" (a_unlikely_2));	\
	} while(0)

#define c_sendmem(type,dest,ptr)		\
	asm volatile ("sendtrans %0, [%2], %1" : : "i" (type), "r" (dest), "r" (ptr))

/* Copy a single 64-byte block (src blk is read using a BYTE endian type) */
extern inline void elan3_copy64b(void *src, void *dst)
{
    /* Copy 64 bytes using ldblock/stblock
     * We save and restore the locals/ins because if we don't gcc
     * really makes a bad job of optimisising the rest of the thread code!
     *
     * We force the parameters in g5, g6 so that they aren't
     * trashed by the loadblk32 into the locals/ins
     */
    register void *tmp1 asm("g5") = src;
    register void *tmp2 asm("g6") = dst;

    asm volatile (
	"and     %%sp,63,%%g7		! Calculate stack alignment\n"
	"sub     %%sp,2*8*4,%%sp	! Space to save the registers\n"
	"sub     %%sp,%%g7,%%sp		! align stack\n" 
	"stblock64 %%l0,[%%sp]		! save the locals and ins\n"
	"ldblock64a [%0]%2,%%l0		! load 64-byte block into locals/ins\n"
	"stblock64a %%l0,[%1]%2		! store 64-byte block from local/ins\n"
	"ldblock64 [%%sp],%%l0		! restore locals and ins\n"
	"add     %%sp,%%g7, %%sp	! undo alignment\n"
	"add     %%sp,2*8*4,%%sp	! restore stack pointer\n"
	: /* outputs */
	: /* inputs */ "r" (tmp1), "r" (tmp2), "n" (EASI_BYTE)
	: /* clobbered */ "g5", "g6", "g7" );
}

/* Copy a single 64-byte block (src blk is read using a WORD endian type) */
extern inline void elan3_copy64w(void *src, void *dst)
{
    /* Copy 64 bytes using ldblock/stblock
     * We save and restore the locals/ins because if we don't gcc
     * really makes a bad job of optimisising the rest of the thread code!
     *
     * We force the parameters in g5, g6 so that they aren't
     * trashed by the loadblk32 into the locals/ins
     */
    register void *tmp1 asm("g5") = src;
    register void *tmp2 asm("g6") = dst;

    asm volatile (
	"and     %%sp,63,%%g7		! Calculate stack alignment\n"
	"sub     %%sp,2*8*4,%%sp	! Space to save the registers\n"
	"sub     %%sp,%%g7,%%sp		! align stack\n" 
	"stblock64 %%l0,[%%sp]		! save the locals and ins\n"
	"ldblock64a [%0]%2,%%l0		! load 64-byte block into locals/ins\n"
	"stblock64a %%l0,[%1]%2		! store 64-byte block from local/ins\n"
	"ldblock64 [%%sp],%%l0		! restore locals and ins\n"
	"add     %%sp,%%g7, %%sp	! undo alignment\n"
	"add     %%sp,2*8*4,%%sp	! restore stack pointer\n"
	: /* outputs */
	: /* inputs */ "r" (tmp1), "r" (tmp2), "n" (EASI_WORD)
	: /* clobbered */ "g5", "g6", "g7" );
}

/* Read a 64-bit value with a WORD (32-bit) endian type */
extern inline E3_uint64 elan3_read64w( volatile E3_uint64 *const ptr )
{
    E3_uint64 result;

    asm volatile (
	"ldblock8a [%1]%2, %0\n"
	: /* outputs */ "=r" (result)
	: /* inputs */ "r" (ptr), "n" (EASI_WORD) );

    return( result );
}

/* Read a 64-bit value with a DOUBLEWORD (64-bit) endian type */
extern inline E3_uint64 elan3_read64dw( volatile E3_uint64 *const ptr )
{
    E3_uint64 result;

    asm volatile (
	"ldblock8a [%1]%2, %0\n"
	: /* outputs */ "=r" (result)
	: /* inputs */ "r" (ptr), "n" (EASI_DOUBLE) );

    return( result );
}

/* Write a 32-bit value with a WORD (32-bit) endian type */
extern inline void elan3_write64w( volatile E3_uint64 *const ptr, E3_uint64 value )
{
    asm volatile (
	"stblock8a %1, [%0]%2\n"
	: /* no outputs */
	: /* inputs */ "r" (ptr), "r" (value), "n" (EASI_WORD) );
}

/* Write a 64-bit value with a DOUBLEWORD (64-bit) endian type */
extern inline void elan3_write64dw( volatile E3_uint64 *const ptr, E3_uint64 value )
{
    asm volatile (
	"stblock8a %1, [%0]%2\n"
	: /* no outputs */
	: /* inputs */ "r" (ptr), "r" (value), "n" (EASI_DOUBLE) );
}

extern inline E3_uint32 c_swap(volatile E3_uint32 *source, E3_uint32 result)
{
   asm volatile("swap [%1],%0\n"
   		: "=r" (result)
		: "r" (source) ,"0" (result)
		: "memory");
   return result;
}

extern inline E3_uint32 c_swap_save(volatile E3_uint32 *source, const E3_uint32 result)
{
   register E3_uint32 a_unlikely;
   asm volatile("" : "=r" (a_unlikely) : );

   asm volatile("mov %2,%0; swap [%1],%0\n"
   		: "=r" (a_unlikely)
		: "r" (source) ,"r" (result), "0" (a_unlikely)
		: "memory");
   return a_unlikely;
}
#endif /* (__ELAN3__) && !(_ASM) */

#endif /* _ELAN3_INTRINSICS_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
