/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN3_PTE_H
#define __ELAN3_PTE_H

#ident "$Id: pte.h,v 1.26 2003/09/24 13:57:24 david Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/pte.h,v $*/

#ifdef	__cplusplus
extern          "C"
{
#endif

#include <elan3/e3types.h>
#include <elan3/perm.h>

typedef E3_uint64 ELAN3_PTE;
typedef E3_uint32 ELAN3_PTP;

#define ELAN3_PTE_SIZE		(8)
#define ELAN3_PTP_SIZE		(4)

#define ELAN3_PTE_REF		((E3_uint64) 1 << 63)		/* 63      - referenced bit */
#define ELAN3_PTE_MOD		((E3_uint64) 1 << 55)		/* 55      - modified bit */
#define ELAN3_RM_MASK		(ELAN3_PTE_REF | ELAN3_PTE_MOD)

#define ELAN3_PTE_PFN_MASK	0x0000fffffffff000ull		/* [12:48] - Physical address */

#define ELAN3_PTE_BIG_ENDIAN	0x80				/* 7       - big endian */
#define ELAN3_PTE_64_BIT		0x40				/* 6       - 64 bit pci address */
#define ELAN3_PTE_LOCAL		0x20				/* 5       - local sdram */

#define ELAN3_PTE_PERM_MASK	0x1c				/* [2:4]   - Permissions */
#define ELAN3_PTE_PERM_SHIFT	   2

#define ELAN3_ET_MASK		0x3
#define ELAN3_ET_INVALID		0x0					/* [0:1] */
#define ELAN3_ET_PTP		0x1
#define ELAN3_ET_PTE		0x2

#define ELAN3_INVALID_PTP	((ELAN3_PTP) 0)
#define ELAN3_INVALID_PTE	((ELAN3_PTE) 0)

#define ELAN3_PTP_TYPE(ptp)	((ptp) & ELAN3_ET_MASK)
#define ELAN3_PTE_TYPE(pte)	((pte) & ELAN3_ET_MASK)
#define ELAN3_PTE_PERM(pte)	((pte) & ELAN3_PTE_PERM_MASK)
#define ELAN3_PTE_VALID(pte)	(((pte) & ELAN3_ET_MASK) == ELAN3_ET_PTE)
#define ELAN3_PTE_ISREF(pte)	((pte) & ELAN3_PTE_REF)
#define ELAN3_PTE_ISMOD(pte)	((pte) & ELAN3_PTE_MOD)
#define ELAN3_PTE_WRITEABLE(pte)	(ELAN3_PERM_WRITEABLE(ELAN3_PTE_PERM(pte)))

#define ELAN3_PERM_WRITEABLE(perm)	((perm) == ELAN3_PERM_NOREMOTE || (perm) > ELAN3_PERM_REMOTEREAD)
#define ELAN3_PERM_REMOTE(perm)		((perm) > ELAN3_PERM_NOREMOTE)

#define ELAN3_PERM_READONLY(perm)	((perm) == ELAN3_PERM_NOREMOTE ? ELAN3_PERM_LOCAL_READ : \
					 (perm) > ELAN3_PERM_REMOTEREAD ? ELAN3_PERM_READ : (perm))
#if PAGE_SHIFT == 12
#  define ELAN3_PAGE_SHIFT	12
#else
#  define ELAN3_PAGE_SHIFT	13
#endif

#define ELAN3_PAGE_SIZE		(1 << ELAN3_PAGE_SHIFT)
#define ELAN3_PAGE_OFFSET	(ELAN3_PAGE_SIZE-1)
#define ELAN3_PAGE_MASK		(~ELAN3_PAGE_OFFSET)

#if ELAN3_PAGE_SHIFT == 13
#  define ELAN3_L3_SHIFT		5
#else
#  define ELAN3_L3_SHIFT		6
#endif
#define ELAN3_L2_SHIFT		6
#define ELAN3_L1_SHIFT		8

/* Number of entries in a given level ptbl */
#define ELAN3_L3_ENTRIES		(1 << ELAN3_L3_SHIFT)
#define ELAN3_L2_ENTRIES		(1 << ELAN3_L2_SHIFT)
#define ELAN3_L1_ENTRIES		(1 << ELAN3_L1_SHIFT)

/* Virtual address spanned by each entry */
#define ELAN3_L3_SIZE		(1 << (ELAN3_PAGE_SHIFT))
#define ELAN3_L2_SIZE		(1 << (ELAN3_L3_SHIFT+ELAN3_PAGE_SHIFT))
#define ELAN3_L1_SIZE		(1 << (ELAN3_L3_SHIFT+ELAN3_L2_SHIFT+ELAN3_PAGE_SHIFT))

/* Virtual address size of page table */
#define ELAN3_L1_PTSIZE          (ELAN3_L1_ENTRIES * ELAN3_L1_SIZE)
#define ELAN3_L3_PTSIZE		(ELAN3_L3_ENTRIES * ELAN3_L3_SIZE)
#define ELAN3_L2_PTSIZE		(ELAN3_L2_ENTRIES * ELAN3_L2_SIZE)

/* Mask for offset into page table */
#define ELAN3_L1_PTOFFSET        ((ELAN3_L1_SIZE*ELAN3_L1_ENTRIES)-1)
#define ELAN3_L3_PTOFFSET	((ELAN3_L3_SIZE*ELAN3_L3_ENTRIES)-1)
#define ELAN3_L2_PTOFFSET	((ELAN3_L2_SIZE*ELAN3_L2_ENTRIES)-1)

#define ELAN3_L1_INDEX(addr)	(((E3_Addr) (addr) & 0xFF000000) >> (ELAN3_L2_SHIFT+ELAN3_L3_SHIFT+ELAN3_PAGE_SHIFT))
#define ELAN3_L2_INDEX(addr)	(((E3_Addr) (addr) & 0x00FD0000) >> (ELAN3_L3_SHIFT+ELAN3_PAGE_SHIFT))
#define ELAN3_L3_INDEX(addr)	(((E3_Addr) (addr) & 0x0003F000) >> ELAN3_PAGE_SHIFT)

#define	ELAN3_L1_BASE(addr)	(((E3_Addr)(addr)) & 0x00000000)
#define	ELAN3_L2_BASE(addr)	(((E3_Addr)(addr)) & 0xFF000000)
#define	ELAN3_L3_BASE(addr)	(((E3_Addr)(addr)) & 0xFFFC0000)

/* Convert a page table pointer entry to the PT */
#define PTP_TO_PT_PADDR(ptp)	((E3_Addr)(ptp & 0xFFFFFFFC))

#ifdef __KERNEL__
/*
 * incompatible access for permission macro.
 */
extern  u_char  elan3mmu_permissionTable[8];
#define ELAN3_INCOMPAT_ACCESS(perm,access) (! (elan3mmu_permissionTable[(perm)>>ELAN3_PTE_PERM_SHIFT] & (1 << (access))))

#define elan3_readptp(dev, ptp)		(elan3_sdram_readl (dev, ptp))
#define elan3_writeptp(dev, ptp, value)	(elan3_sdram_writel (dev, ptp, value))
#define elan3_readpte(dev, pte)		(elan3_sdram_readq (dev, pte))
#define elan3_writepte(dev,pte, value)	(elan3_sdram_writeq (dev, pte, value))

#define elan3_invalidatepte(dev, pte)	(elan3_sdram_writel (dev, pte, 0))
#define elan3_modifypte(dev,pte,new)	(elan3_sdram_writel (dev, pte, (int) (new)))
#define elan3_clrref(dev,pte)		(elan3_sdram_writeb (dev, pte + 7)

#endif	/* __KERNEL__ */

#ifdef __cplusplus
}
#endif

#endif /* __ELAN3_PTE_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
