/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_THREAD_H
#define _ELAN3_THREAD_H

#ident "$Id: thread.h,v 1.17 2002/08/09 11:23:34 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/thread.h,v $*/

/* Alignment for a stack frame */
#define E3_STACK_ALIGN		(64)

typedef struct _E3_Frame {
    E3_uint32	       fr_local[8];		/* saved locals (not used) */
    E3_uint32	       fr_arg[6];		/* saved arguements o0 -> o5 */
    E3_Addr	       fr_savefp;		/* saved frame pointer o6 */
    E3_Addr	       fr_savepc;		/* saved program counter o7 */
    E3_Addr	       fr_stret;		/* stuct return addr */
    E3_uint32	       fr_argd[6];		/* arg dump area */
    E3_uint32	       fr_argx[1];		/* array of args past the sixth */
} E3_Frame;

typedef struct _E3_Stack {
    E3_uint32		Locals[8];
    E3_uint32		Ins[8];
    E3_uint32		Globals[8];
    E3_uint32		Outs[8];
} E3_Stack;

typedef struct _E3_OutsRegs {
   E3_uint32	o[8];				/* o6 == pc, o7 == fptr */
} E3_OutsRegs;

/*
 * "Magic" value for stack pointer to be ignored.
 */
#define VanishingStackPointer	0x42


/*
 * When the Elan traps the N & Z CC bits are held in the NPC
 * and the V & C bits are in the PC
 */
#define PSR_C_BIT	(1)
#define PSR_V_BIT	(2)
#define PSR_Z_BIT	(1)
#define PSR_N_BIT	(2)
#define CC_MASK		(3)
#define PC_MASK 	(~3)
#define SP_MASK		(~3)

/*
 * Threads processor Opcodes.
 */
#define OPCODE_MASK		(0xC1F80000)
#define OPCODE_IMM		(1 << 13)

#define OPCODE_CLASS(instr)	((instr) & 0xC0000000)
#define OPCODE_CLASS_0		0x00000000
#define OPCODE_CLASS_1		0x40000000
#define OPCODE_CLASS_2		0x80000000
#define OPCODE_CLASS_3		0xC0000000

#define OPCODE_CPOP		0x81B00000
#define OPCODE_Ticc		0x81D00000

#define OPCODE_FCODE_SHIFT	19
#define OPCODE_FCODE_MASK	0x1f
#define OPCODE_NOT_ALUOP	0x01000000

#define OPCODE_SLL		0x81280000
#define OPCODE_SRL		0x81300000
#define OPCODE_SRA		0x81380000

#define OPCODE_OPEN		0x81600000
#define OPCODE_CLOSE		0x81680000
#define OPCODE_BREAKTEST	0x81700000

#define OPCODE_BREAK		0x81a00000
#define OPCODE_SUSPEND		0x81a80000
#define OPCODE_WAIT		0x81b00000

#define OPCODE_JMPL		0x81c00000

#define OPCODE_LD		0xC0000000
#define OPCODE_LDD		0xC0180000

#define OPCODE_LDBLOCK16	0xC0900000
#define OPCODE_LDBLOCK32	0xC0800000
#define OPCODE_LDBLOCK64	0xC0980000

#define OPCODE_ST		0xC0200000
#define OPCODE_STD		0xC0380000

#define OPCODE_SWAP		0xC0780000

#define OPCODE_STBLOCK16	0xC0b00000
#define OPCODE_STBLOCK32	0xC0a00000
#define OPCODE_STBLOCK64	0xC0b80000

#define OPCODE_CLASS0_MASK	0xC1C00000
#define OPCODE_SETHI		0x01000000
#define OPCODE_BICC		0x00800000
#define OPCODE_SENDREG		0x01800000
#define OPCODE_SENDMEM		0x01c00000

#define OPCODE_BICC_BN		0x00000000
#define OPCODE_BICC_BE		0x02000000
#define OPCODE_BICC_BLE		0x04000000
#define OPCODE_BICC_BL		0x06000000
#define OPCODE_BICC_BLEU	0x08000000
#define OPCODE_BICC_BCS		0x0A000000
#define OPCODE_BICC_BNEG	0x0C000000
#define OPCODE_BICC_BVS		0x0E000000

#define OPCODE_BICC_MASK	0x0E000000
#define OPCODE_BICC_ANNUL	0x20000000

#define INSTR_RS2(instr)	(((instr) >>  0) & 0x1F)
#define INSTR_RS1(instr)	(((instr) >> 14) & 0x1F)
#define INSTR_RD(instr)		(((instr) >> 25) & 0x1F)
#define INSTR_IMM(instr)	(((instr) & 0x1000) ? ((instr) & 0xFFF) | 0xFFFFF000 : (instr) & 0xFFF)

#define Ticc_COND(instr)	INSTR_RD(instr)
#define Ticc_TA			8

#endif /* _ELAN3_THREAD_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
