/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN3_TRTYPE_H
#define _ELAN3_TRTYPE_H

#ident "$Id: trtype.h,v 1.13 2002/08/09 11:23:34 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan3mod/elan3/elan3/trtype.h,v $ */

/*<15>	ackNow  */
#define TR_SENDACK	(1 << 15)

#define TR_SIZE_SHIFT	12
#define TR_SIZE_MASK	7

/*<14:12> Size	0, 1, 2, 4, 8, 16, 32, 64  Double Words
          Bit 14 is forced to zero currently so that only size 0, 1, 2, 4 are
	  allowed    */

#define TR_SIZE0	(0 << TR_SIZE_SHIFT)
#define TR_SIZE1	(1 << TR_SIZE_SHIFT)
#define TR_SIZE2	(2 << TR_SIZE_SHIFT)
#define TR_SIZE4	(3 << TR_SIZE_SHIFT)
#define TR_SIZE8	(4 << TR_SIZE_SHIFT)

#define TR_64_BIT_ADDR	(1 << 11)
#define TR_LAST_TRANS	(1 << 10)

#define TR_WRITEBLOCK_BIT	(1 << 9)
#define TR_WRITEBLOCK		(TR_WRITEBLOCK_BIT | TR_SIZE8)


#define TR_WRITEBLOCK_SIZE	64

/*
 * write-block
 */
/*	WriteBlock 	<8:7>	Data type
			<6:0> 	Part write size */
#define TR_TYPE_SHIFT	7
#define TR_TYPE_MASK	((1 << 2) - 1)

#define TR_TYPE_BYTE	0
#define TR_TYPE_SHORT	1
#define TR_TYPE_WORD	2
#define TR_TYPE_DWORD	3

#define TR_PARTSIZE_MASK ((1 << 7) -1)

#define TR_WAIT_FOR_EOP	(1 << 8)

/*
 * trace-route format 
 */
#define TR_TRACEROUTE0_CHANID(val)		((val) & 1)			/* 0     Chan Id */
#define TR_TRACEROUTE0_LINKID(val)		(((val) >> 1) & 7)		/* 1:3   Link Id */
#define TR_TRACEROUTE0_REVID(val)		(((val) >> 4) & 7)		/* 4:6   Revision ID */
#define TR_TRACEROUTE0_BCAST_TOP_PIN(val)	(((val) >> 7) & 1)		/* 7     Broadcast Top Pin (REV B) */
#define TR_TRACEROUTE0_LNR(val)			((val) >> 8)			/* 8:15  Global Link Not Ready */

#define TR_TRACEROUTE1_PRIO(val)		((val & 0xF))			/* 0:3   Arrival Priority (REV A) */
#define TR_TRACEROUTE1_AGE(val)			(((val) >> 4) & 0xF)		/* 4:7   Priority Held(Age) (REV A) */
#define TR_TRACEROUTE1_ROUTE_SELECTED(val)	((val) & 0xFF)			/* 0:7   Arrival age (REV B) */
#define TR_TRACEROUTE1_BCAST_TOP(val)		(((val) >> 8) & 7)		/* 8:10  Broadcast Top */
#define TR_TRACEROUTE1_ADAPT(val)		(((val) >> 12) & 3)		/* 12:13 This Adaptive Value (REV A) */
#define TR_TRACEROUTE1_BCAST_BOT(val)		(((val) >> 12) & 7)		/* 12:14 Broadcast Bottom (REV B) */

#define TR_TRACEROUTE2_ARRIVAL_AGE(val)		((val) & 0xF)			/* 0:3	 Arrival Age (REV B) */
#define TR_TRACEROUTE2_CURR_AGE(val)		(((val) >> 4) & 0xF)		/* 4:7   Current Age (REV B) */
#define TR_TRACEROUTE2_BUSY(val)		(((val) >> 8) & 0xFF)		/* 8:15  Busy (REV B) */

#define TR_TRACEROUTE_SIZE	32
#define TR_TRACEROUTE_ENTRIES	(TR_TRACEROUTE_SIZE/2)

/*
 * non-write block
 */
#define TR_OPCODE_MASK		(((1 << 8) - 1) |			\
				 (TR_SIZE_MASK << TR_SIZE_SHIFT) |	\
				 TR_WRITEBLOCK_BIT)

#define TR_NOP_TRANS		(0x0 | TR_SIZE0)
#define TR_SETEVENT		(0x0 | TR_SIZE0 | TR_SENDACK | TR_LAST_TRANS)
#define TR_REMOTEDMA		(0x1 | TR_SIZE4 | TR_SENDACK | TR_LAST_TRANS)
#define TR_LOCKQUEUE		(0x2 | TR_SIZE0)
#define TR_UNLOCKQUEUE		(0x3 | TR_SIZE0 | TR_SENDACK | TR_LAST_TRANS)

#define TR_SENDDISCARD		(0x4 | TR_SIZE0)
#define TR_TRACEROUTE		(0x5 | TR_SIZE4)

#define TR_DMAIDENTIFY		(0x6 | TR_SIZE0)
#define TR_THREADIDENTIFY	(0x7 | TR_SIZE1)

#define TR_GTE			(0x8 | TR_SIZE1)
#define TR_LT			(0x9 | TR_SIZE1)
#define TR_EQ			(0xA | TR_SIZE1)
#define TR_NEQ			(0xB | TR_SIZE1)

#define TR_WRITEWORD 		(0xC | TR_SIZE1)
#define TR_WRITEDOUBLEWORD	(0xD | TR_SIZE1)
#define TR_TESTANDWRITE 	(0xE | TR_SIZE1)
#define TR_ATOMICADDWORD	(0xF | TR_SIZE1 | TR_SENDACK | TR_LAST_TRANS)
#define TR_OPCODE_TYPE_MASK	0xff


#endif /* notdef _ELAN3_TRTYPE_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
