/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_COMMANDS_H
#define __ELAN4_COMMANDS_H

#ident "$Id: commands.h,v 1.29 2004/06/16 15:45:02 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/commands.h,v $*/

/*
 * This header file describes the command format for the Elan 4
 *    See CommandFormat.doc
 */

/*
 * Number of channels in traced elanlib_trace.c
 */
#define TRACE_MAX_CHANNELS	2

/*
 * Define encoding for the commands issued into the command queues
 */
#define RUN_THREAD_CMD       0x00
#define OPEN_STEN_PKT_CMD    0x01
#define WRITE_DWORD_CMD      0x02
#define ADD_DWORD_CMD        0x03
#define COPY64_CMD           0x05
#define GUARD_CMD            0x06
#define SET_EVENT_CMD        0x07
#define SEND_TRANS_CMD       0x09
#define INTERRUPT_CMD        0x0d
#define RUN_DMA_CMD          0x0e
#define SET_EVENTN_CMD       0x0f
#define NOP_CMD		     0x17
#define MAKE_EXT_CLEAN_CMD   0x37
#define WAIT_EVENT_CMD       0x1f

/*
 * Define the portion of the data word the user is NOT
 * allowed to use. This varies with Commmand type
 */
#define RUN_THREAD_CMD_MASK       0x03
#define OPEN_STEN_PKT_CMD_MASK    0x0f
#define WRITE_DWORD_CMD_MASK      0x07
#define ADD_DWORD_CMD_MASK        0x07
#define COPY64_CMD_MASK           0x0f
#define GUARD_CMD_MASK            0x0f
#define SET_EVENT_CMD_MASK        0x1f
#define SEND_TRANS_CMD_MASK       0x1f
#define INTERRUPT_CMD_MASK        0x0f
#define RUN_DMA_CMD_MASK          0x0f
#define SET_EVENTN_CMD_MASK       0x1f
#define NOP_CMD_MASK		  0x3f
#define MAKE_EXT_CLEAN_MASK	  0x3f
#define WAIT_EVENT_CMD_MASK       0x1f

#define COPY64_DATA_TYPE_SHIFT    0x4
#define COPY64_DTYPE_BYTE	  (0 << COPY64_DATA_TYPE_SHIFT)
#define COPY64_DTYPE_SHORT	  (1 << COPY64_DATA_TYPE_SHIFT)
#define COPY64_DTYPE_WORD	  (2 << COPY64_DATA_TYPE_SHIFT)
#define COPY64_DTYPE_LONG	  (3 << COPY64_DATA_TYPE_SHIFT)

/*
 * SET_EVENTN - word 1 has following form
 * [63:5]	Event Address
 * [4:0]	Part Set Value.
 */
#define SET_EVENT_PART_SET_MASK      0x1f

/* OPEN_STEN_PKT_CMD 
 *   [63:32]	Vproc
 *   [31]	Use Test
 *   [30:28]	unused
 *   [27:21]	Test Acceptable PAck code
 *   [20:16]	Test Ack Channel Number
 *   [15:9]	Acceptable PAck code
 *   [8:4]	Ack Channel Number (1 bit on Elan4)
 *   [3:0]	Command type
 */
/* Acceptable PAck code */
#define PACK_OK			(1 << 0)
#define PACK_TESTFAIL		(1 << 1)
#define PACK_DISCARD		(1 << 2)
#define RESTART_COUNT_ZERO	(1 << 3)
#define PACK_ERROR		(1 << 7)
#define PACK_TIMEOUT		(1 << 8)

/*
 *#ifndef USE_DIRTY_COMMANDS
 *#define USE_DIRTY_COMMANDS
 *#endif
 */
#ifdef USE_DIRTY_COMMANDS
#define OPEN_PACKET_USED_MASK    0x00000000780f00e0ULL
#define SEND_TRANS_USED_MASK     0xffffffff0000fff0ULL
#define COPY64_WRITE_USED_MASK   0x000000000000000fULL
#define MAIN_INT_USED_MASK       0x0000000000003ff0ULL
#define GUARD_USED_MASK          0xfffffe007000fde0ULL
#define DMA_TYPESIZE_USED_MASK   0x000000000000fff0ULL
#define SETEVENTN_USED_MASK      0xffffffffffffffe0ULL
#define NOP_USED_MASK            0xffffffffffffffc0ULL
#define EXT_CLEAN_USED_MASK      0xffffffffffffffc0ULL
#define WAIT_CNT_TYPE_USED_MASK  0x00000000fffff800ULL
#else
#define OPEN_PACKET_USED_MASK    0x0ULL
#define SEND_TRANS_USED_MASK     0x0ULL
#define COPY64_WRITE_USED_MASK   0x0ULL
#define MAIN_INT_USED_MASK       0x0ULL
#define GUARD_USED_MASK          0x0ULL
#define DMA_TYPESIZE_USED_MASK   0x0ULL
#define SETEVENTN_USED_MASK      0x0ULL
#define NOP_USED_MASK            0x0ULL
#define EXT_CLEAN_USED_MASK      0x0ULL
#define WAIT_CNT_TYPE_USED_MASK  0x0ULL
#endif

#define OPEN_PACKET(chan, code, vproc) \
	((((chan) & 1) << 4) | (((code) & 0x7f) << 9) | ((E4_uint64)(vproc) << 32) | OPEN_STEN_PKT_CMD)

#define OPEN_PACKET_TEST(chan, code, vproc, tchan, tcode) \
	((((chan) & 1) << 4) | (((code) & 0x7f) << 9) | ((E4_uint64)(vproc) << 32) | \
	 (((tchan) & 1) << 16) | (((tcode) & 0x7f) << 21) | (((E4_uint64) 1) << 31) | OPEN_STEN_PKT_CMD)

/*
 * GUARD_CMD
 *   [63:41]	unused
 *   [40]	Reset Restart Fail Count	// only performed if the Guard executes the next command.
 *   [39:32]	New Restart Fail Count value
 *   [31]	Use Test
 *   [30:28]	unused
 *   [27:21]	Test Acceptable PAck code
 *   [20:16]	Test Ack Channel Number
 *   [15:9]	unused
 *   [8:4]	Ack Channel Number
 *   [3:0]	Command type
 */
/* GUARD_CHANNEL(chan)
 */
#define GUARD_ALL_CHANNELS	((1 << 9) | GUARD_CMD)
#define GUARD_CHANNEL(chan)	((((chan) & 1) << 4) | GUARD_CMD)
#define GUARD_TEST(chan,code)	((1ull << 31) | (((code) & 0x7f) << 21) | (((chan) & 1) << 16))
#define GUARD_RESET(count)	((1ull << 40) | ((((E4_uint64) count) & 0xff) << 32))

#define GUARD_CHANNEL_TEST(chan,tchan,tcode) \
	((((chan) & 1) << 4) | (((tchan) & 1) << 16) | (((tcode) & 0x7f) << 21) | \
	 (((E4_uint64) 1) << 31) | GUARD_CMD)

/*
 * SEND_TRANS_CMD
 * [63:32]	unused
 * [31:16]	transaction type
 * [15:4]	unused
 * [3:0]	Command type
 */
#define SEND_TRANS(TransType)	(((TransType) << 16) | SEND_TRANS_CMD)

/*
 * Command port trace debug levels
 */
#define TRACE_CMD_BUFFER	0x01
#define TRACE_CMD_TYPE		0x02
#define TRACE_CHANNEL_OPENS	0x04
#define TRACE_GUARDED_ATOMICS	0x08
#define TRACE_CMD_TIMEOUT	0x10

/*
 * Commands that should be preceeded by a GUARD_CMD.
 */
#define IS_ATOMIC_CMD(cmd) 								\
   ((cmd) == RUN_THREAD_CMD || (cmd) == ADD_DWORD_CMD || (cmd) == INTERRUPT_CMD || 	\
    (cmd) == RUN_DMA_CMD    || (cmd) == SET_EVENT_CMD || (cmd) == SET_EVENTN_CMD || 	\
    (cmd) == WAIT_EVENT_CMD)

#ifndef _ASM

/*
 * These structures are used to build event copy command streams. They are intended to be included
 * in a larger structure to form a self documenting command sequence that can be easily coped and manipulated.
 */

typedef struct e4_runthreadcmd
{
   E4_Addr	PC;
   E4_uint64	r[6];
} E4_RunThreadCmd;

typedef E4_uint64 E4_OpenCmd;

typedef struct e4_writecmd
{
   E4_Addr	WriteAddr;
   E4_uint64	WriteValue;
} E4_WriteCmd;

typedef struct e4_addcmd
{
   E4_Addr	AddAddr;
   E4_uint64	AddValue;
} E4_AddCmd;

typedef struct e4_copycmd
{
   E4_Addr	SrcAddr;
   E4_Addr	DstAddr;
} E4_CopyCmd;

typedef E4_uint64 E4_GaurdCmd;
typedef E4_uint64 E4_SetEventCmd;

/*
 * The data to this command must be declared as a vector after the use of this.
 */
typedef struct e4_sendtranscmd
{
   E4_Addr	Type;
   E4_Addr	Addr;
} E4_SendTransCmd;

typedef E4_uint64 E4_IntCmd;

/* The normal Dma struc can be used here. */

typedef struct e4_seteventncmd
{
   E4_Addr	Event;
   E4_Addr	SetCount;
} E4_SetEventNCmd;

typedef E4_uint64 E4_NopCmd;
typedef E4_uint64 E4_MakeExtCleanCmd;

typedef struct e4_waitcmd
{
   E4_Addr	ev_Event;
   E4_Addr	ev_CountType;
   E4_Addr	ev_Params[2];
} E4_WaitCmd;

#endif /* _ASM */

#endif /* __ELAN4_COMMANDS_H  */

