/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_ELANDEV_H
#define __ELAN4_ELANDEV_H

#ident "$Id: device.h,v 1.92.2.2 2005/08/09 09:57:04 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/device.h,v $ */

#include <elan/devinfo.h>
#include <elan/capability.h>

#include <elan4/pci.h>
#include <elan4/sdram.h>
#include <elan4/dma.h>
#include <elan4/events.h>
#include <elan4/registers.h>

#include <elan4/mmu.h>
#include <elan4/trap.h>
#include <elan4/stats.h>
#include <elan4/neterr.h>

#ifdef CONFIG_MPSAS
#include <elan4/mpsas.h>
#endif

#if defined(LINUX)
#include <elan4/device_Linux.h>
#elif defined(TRU64UNIX)
#include <elan4/device_OSF1.h>
#elif defined(SOLARIS)
#include <elan4/device_SunOS.h>
#endif

/*
 * Network context number allocation.
 * [0]          neterr fixup system context
 * [1]          kernel comms system context
 * [2048-4095]	kernel comms data contexts
 */
#define ELAN4_NETERR_CONTEXT_NUM	0x00			/* network error fixup context number */
#define ELAN4_KCOMM_CONTEXT_NUM		0x01			/* kernel comms context number */
#define ELAN4_KCOMM_BASE_CONTEXT_NUM	0x800			/* kernel comms data transfer contexts */
#define ELAN4_KCOMM_TOP_CONTEXT_NUM	0xfff

#define ELAN4_SYSTEM_CONTEXT(ctx)  ((ctx) >= ELAN4_KCOMM_BASE_CONTEXT_NUM)

typedef void (ELAN4_HALTFN)(struct elan4_dev *dev, void *arg);

typedef struct elan4_haltop
{
    struct list_head    op_link;				/* chain on a list */
    E4_uint32		op_mask;				/* Interrupt mask to see before calling function */
    
    ELAN4_HALTFN       *op_function;				/* function to call */
    void	       *op_arg;					/* arguement to pass to function */
} ELAN4_HALTOP;

typedef void (ELAN4_DMA_FLUSHFN)(struct elan4_dev *dev, void *arg, int qfull);

typedef struct elan4_dma_flushop
{
    struct list_head    op_link;				/* chain on a list */
    ELAN4_DMA_FLUSHFN  *op_function;				/* function to call */
    void	       *op_arg;					/* arguement to pass to function */
} ELAN4_DMA_FLUSHOP;

typedef void (ELAN4_INTFN)(struct elan4_dev *dev, void *arg);

typedef struct elan4_intop
{
    struct list_head    op_link;				/* chain on a list */
    ELAN4_INTFN        *op_function;				/* function to call */
    void	       *op_arg;					/* arguement to pass to function */
    E4_uint64		op_cookie;				/* and main interrupt cookie */
} ELAN4_INTOP;

typedef struct elan4_eccerrs
{
    E4_uint64        EccStatus;
    E4_uint64        ConfigReg;
    E4_uint32        ErrorCount;
} ELAN4_ECCERRS;

#define SDRAM_MIN_BLOCK_SHIFT	10
#define SDRAM_NUM_FREE_LISTS	19					/* allows max 256 Mb block */
#define SDRAM_MIN_BLOCK_SIZE	(1 << SDRAM_MIN_BLOCK_SHIFT)
#define SDRAM_MAX_BLOCK_SIZE	(SDRAM_MIN_BLOCK_SIZE << (SDRAM_NUM_FREE_LISTS-1))

#if PAGE_SHIFT < 13
#define SDRAM_PAGE_SIZE		8192
#define SDRAM_PGOFF_OFFSET	1
#define SDRAM_PGOFF_MASK	(~SDRAM_PGOFF_OFFSET)
#else
#define SDRAM_PAGE_SIZE		PAGE_SIZE
#define SDRAM_PGOFF_OFFSET	0
#define SDRAM_PGOFF_MASK	(~SDRAM_PGOFF_OFFSET)
#endif

typedef struct elan4_sdram
{
    sdramaddr_t 	b_base;					/* offset in sdram bar */
    unsigned		b_size;					/* size of bank */
    ioaddr_t		b_ioaddr;				/* ioaddr where mapped into the kernel */
    ELAN4_MAP_HANDLE	b_handle;				/*    and mapping handle */
    bitmap_t	       *b_bitmaps[SDRAM_NUM_FREE_LISTS];	/* buddy allocator bitmaps */
} ELAN4_SDRAM_BANK;

/* command queue */
typedef struct elan4_cq 
{
    struct elan4_cqa    *cq_cqa;					/* command queue allocator this belongs to */
    unsigned		 cq_idx;					/*  and which command queue this is */

    sdramaddr_t		 cq_space;					/* sdram backing up command queue */
    unsigned		 cq_size;					/* size value */
    unsigned		 cq_perm;					/* permissions */
    ioaddr_t		 cq_mapping;					/* mapping of command queue page */
    ELAN4_MAP_HANDLE	 cq_handle;					/*    and mapping handle */
} ELAN4_CQ;

/* cqtype flags to elan4_alloccq() */
#define CQ_Priority	(1 << 0)
#define CQ_Reorder	(1 << 1)

/* command queues are allocated in chunks,so that all the
 * command ports are in a single system page */
#define ELAN4_CQ_PER_CQA	MAX(1, (PAGESIZE/CQ_CommandMappingSize))

/* maximum number of command queues per context */
#define ELAN4_MAX_CQA		(256 / ELAN4_CQ_PER_CQA)

typedef struct elan4_cqa
{
    struct list_head	cqa_link;					/* linked together */
    bitmap_t		cqa_bitmap[BT_BITOUL(ELAN4_CQ_PER_CQA)];	/* bitmap of which are free */
    unsigned int        cqa_type;					/* allocation type */
    unsigned int	cqa_cqnum;					/* base cq number */
    unsigned int	cqa_ref;					/* "mappings" to a queue */
    unsigned int	cqa_idx;					/* index number */
    ELAN4_CQ		cqa_cq[ELAN4_CQ_PER_CQA];			/* command queue entries */
} ELAN4_CQA;

#define elan4_cq2num(cq)	((cq)->cq_cqa->cqa_cqnum + (cq)->cq_idx)
#define elan4_cq2idx(cq)	((cq)->cq_cqa->cqa_idx * ELAN4_CQ_PER_CQA + (cq)->cq_idx)

#ifdef ELAN4_LARGE_PAGE_SUPPORT
#define NUM_HASH_TABLES		2
#else
#define NUM_HASH_TABLES		1
#endif

typedef struct elan4_ctxt_trans_index
{
    int        tbl;
    int        index;
} ELAN4_CTXT_TRANS_INDEX;

#define ELAN4_CTXT_MAX_SHUFFLE (10)

typedef struct elan4_ctxt
{
    struct elan4_dev      *ctxt_dev;					/* device we're associated with */
    struct list_head       ctxt_link;					/* chained on device */
    
    struct elan4_trap_ops *ctxt_ops;					/* client specific operations */

    unsigned int	   ctxt_features;				/* device features this context is using */
    signed int	           ctxt_num;					/* local context number */

    struct list_head	   ctxt_cqalist;				/* link list of command queue allocators */
    bitmap_t	      	   ctxt_cqamap[BT_BITOUL(ELAN4_MAX_CQA)];	/*   bitmap for allocating cqa_idx */

    ELAN4_HASH_ENTRY     **ctxt_mmuhash[NUM_HASH_TABLES];		/* software hash tables */
    spinlock_t		   ctxt_mmulock;				/*   and spinlock. */

    struct proc_dir_entry *procdir;
    ELAN4_CTXT_TRANS_INDEX trans_index[NUM_HASH_TABLES];                /* place to stash info for /proc/qsnet/elan4/deviceN/ctxt/N/translations_N */

    int                    shuffle_needed[NUM_HASH_TABLES];                  /* true when there are entries in shuffle array */
    int                    shuffle[NUM_HASH_TABLES][ELAN4_CTXT_MAX_SHUFFLE]; /* hashidx's that need shuffling or -1 = none. if all set then shuffle ALL hashidx's */
} ELAN4_CTXT;

typedef struct elan4_trap_ops
{
    void       (*op_eproc_trap) (ELAN4_CTXT *ctxt, E4_uint64 status);
    void       (*op_cproc_trap) (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned cqnum);
    void       (*op_dproc_trap) (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit);
    void       (*op_tproc_trap) (ELAN4_CTXT *ctxt, E4_uint64 status);
    void       (*op_iproc_trap) (ELAN4_CTXT *ctxt, E4_uint64 status, unsigned unit);
    void       (*op_interrupt)  (ELAN4_CTXT *ctxt, E4_uint64 cookie);
    void       (*op_neterrmsg)  (ELAN4_CTXT *ctxt, ELAN4_NETERR_MSG *msg);
    void       (*op_need_shuffle) (ELAN4_CTXT *ctxt, int tbl, int index);
} ELAN4_TRAP_OPS;

typedef struct elan4_route_table
{
    spinlock_t  tbl_lock;
    unsigned	tbl_size;
    sdramaddr_t tbl_entries;
} ELAN4_ROUTE_TABLE;

#define DEV_STASH_ROUTE_COUNT 20

typedef struct elan4_route_ringbuf {
    int start;
    int end;
    E4_VirtualProcessEntry routes[DEV_STASH_ROUTE_COUNT]; 
} ELAN4_ROUTE_RINGBUF;

#define elan4_ringbuf_init(ringbuf) memset(&ringbuf, 0, sizeof(ELAN4_ROUTE_RINGBUF));

typedef struct elan4_trans_index
{
    int tbl;
} ELAN4_TRANS_INDEX;

#define ELAN4_TRANS_STATS_NUM_BUCKETS 7
typedef struct elan4_trans_stats
{
    int tbl;
    int buckets[ELAN4_TRANS_STATS_NUM_BUCKETS];
} ELAN4_TRANS_STATS;

typedef struct elan4_dev
{
    ELAN4_CTXT		 dev_ctxt;					/* context for device operations */

    ELAN4_DEV_OSDEP	 dev_osdep;					/* OS specific entries */

    int			 dev_instance;					/* device number */
    ELAN_DEVINFO	 dev_devinfo;					/* device information (revision etc */
    ELAN_POSITION	 dev_position;					/* position connected to switch */
    ELAN_DEV_IDX	 dev_idx;					/* device idx registered with elanmod */

    kmutex_t		 dev_lock;					/* lock for device state/references */
    unsigned		 dev_state;					/* device state */
    unsigned		 dev_references;				/*  # references */

    ioaddr_t       	 dev_regs;					/* Mapping of device registers */
    ELAN4_MAP_HANDLE	 dev_regs_handle;
    ioaddr_t		 dev_rom;					/* Mapping of rom */
    ELAN4_MAP_HANDLE	 dev_rom_handle;
    ioaddr_t		 dev_i2c;					/* Mapping of I2C registers */
    ELAN4_MAP_HANDLE	 dev_i2c_handle;
    
    E4_uint64		 dev_sdram_cfg;					/* SDRAM config value (from ROM) */
    E4_uint64		 dev_sdram_initial_ecc_val;			/* power on ECC register value */
    int			 dev_sdram_numbanks;				/* # banks of sdram */
    ELAN4_SDRAM_BANK	 dev_sdram_banks[SDRAM_MAX_BANKS];		/* Mapping of sdram banks */
    spinlock_t		 dev_sdram_lock;				/* spinlock for buddy allocator */
    sdramaddr_t		 dev_sdram_freelists[SDRAM_NUM_FREE_LISTS];
    unsigned		 dev_sdram_freecounts[SDRAM_NUM_FREE_LISTS];

    physaddr_t		 dev_regs_phys;					/* physaddr of registers */
    physaddr_t		 dev_sdram_phys;				/*  and of sdram */

    sdramaddr_t		 dev_cacheflush_space;				/* sdram reserved for cache flush operation */

    sdramaddr_t		 dev_faultarea;					/* fault areas for each unit */
    sdramaddr_t		 dev_inputtraparea;				/* trap area for trapped transactions */
    sdramaddr_t		 dev_ctxtable;					/* context table (E4_ContextControlBlock) */
    int		       	 dev_ctxtableshift;				/* and size (in bits) */

    E4_uint32		 dev_syscontrol;				/* copy of system control register */
    spinlock_t		 dev_syscontrol_lock;				/*   spinlock to sequentialise modifications */
    unsigned		 dev_direct_map_pci_writes;			/*   # counts for CONT_DIRECT_MAP_PCI_WRITES */

    volatile E4_uint32	 dev_intmask;					/* copy of interrupt mask register */
    spinlock_t		 dev_intmask_lock;				/*   spinlock to sequentialise modifications */

    /* i2c section */
    spinlock_t		 dev_i2c_lock;					/* spinlock for i2c operations */
    unsigned int         dev_i2c_led_disabled;				/* count of reasons led auto update disabled */

    /* mmu section */
    unsigned		 dev_pagesizeval[NUM_HASH_TABLES];		/* page size value */
    unsigned		 dev_pageshift[NUM_HASH_TABLES];			/* pageshift in bits. */
    unsigned		 dev_hashsize[NUM_HASH_TABLES];			/* # entries in mmu hash table */
    sdramaddr_t		 dev_hashtable[NUM_HASH_TABLES];		/* mmu hash table */
    ELAN4_HASH_ENTRY	*dev_mmuhash[NUM_HASH_TABLES];			/*   and software shadow */
    ELAN4_HASH_ENTRY    *dev_mmufree_count;		  	        /*   space holder - will use to indicate if there is a free slot in chain */
    ELAN4_HASH_ENTRY    *dev_mmufreelist;				/*   and free blocks */
    spinlock_t           dev_mmulock;
    E4_uint16		 dev_topaddr[4];				/* top address values */
    unsigned char	 dev_topaddrvalid;
    unsigned char	 dev_topaddrmode;
    unsigned char	 dev_pteval;					/* allow setting of relaxed order/dont snoop attributes */

    unsigned		 dev_rsvd_hashmask[NUM_HASH_TABLES];
    unsigned		 dev_rsvd_hashval[NUM_HASH_TABLES];

    ELAN4_TRANS_INDEX    trans_index[NUM_HASH_TABLES];                  /* place to stash info for /proc/qsnet/elan4/deviceN/translations_N */
    ELAN4_TRANS_STATS    trans_stats[NUM_HASH_TABLES];                  /* place to stash info for /proc/qsnet/elan4/deviceN/stats/translations_N */
    /* run queues */
    sdramaddr_t	      	 dev_comqlowpri;				/* CProc low & high pri run queues */
    sdramaddr_t		 dev_comqhighpri;

    sdramaddr_t		 dev_dmaqlowpri;				/* DProc,TProc,Interrupt queues */
    sdramaddr_t		 dev_dmaqhighpri;
    sdramaddr_t		 dev_threadqlowpri;
    sdramaddr_t		 dev_threadqhighpri;
    sdramaddr_t		 dev_interruptq;

    E4_uint32	         dev_interruptq_nfptr;				/* cache next main interrupt fptr */
    struct list_head     dev_interruptq_list;				/*   list of operations to call when space in interruptq*/

    /* command queue section */
    sdramaddr_t		 dev_cqaddr;					/* SDRAM address of command queues */
    unsigned		 dev_cqoffset;					/* offset for command queue alignment constraints */
    unsigned		 dev_cqcount;					/* number of command queue descriptors */
    bitmap_t		*dev_cqamap;					/* bitmap for allocation */
    spinlock_t		 dev_cqlock;					/* spinlock to protect bitmap */
    unsigned		 dev_cqreorder;					/* offset for first re-ordering queue with mtrr */

    /* halt operation section */
    struct list_head     dev_haltop_list;				/* list of operations to call when units halted */
    E4_uint32		 dev_haltop_mask;				/* mask of which ones to halt */
    E4_uint32		 dev_haltop_active;				/* mask of which haltops are executing */
    spinlock_t		 dev_haltop_lock;				/*    and their spinlock */
    struct timer_list    dev_haltop_timer;                              /* timer looking for haltop locked in list */

    struct {
	struct list_head list;						/* list of halt operations for DMAs */
	ELAN4_CQ	*cq;						/*   and command queue's */
	ELAN4_INTOP      intop;						/*   and main interrupt op */
	E4_uint64	 status;					/*   status register (when waiting for intop)*/
    }			 dev_dma_flushop[2];

    unsigned		 dev_halt_all_count;				/* count of reasons to halt all units */
    unsigned		 dev_halt_lowpri_count;				/* count of reasons to halt lowpri queues */
    unsigned		 dev_halt_cproc_count;				/* count of reasons to halt command processor */
    unsigned		 dev_halt_dproc_count;				/* count of reasons to halt dma processor */
    unsigned		 dev_halt_tproc_count;				/* count of reasons to halt thread processor */
    unsigned		 dev_discard_all_count;				/* count of reasons to discard all packets */
    unsigned		 dev_discard_lowpri_count;			/* count of reasons to discard non-system packets */
    unsigned		 dev_discard_highpri_count;			/* count of reasons to discard system packets */

    E4_uint32		 dev_schedstatus;				/* copy of schedule status register */

    /* local context allocation section */
    spinlock_t		 dev_ctxlock;					/* spinlock to protect bitmap */
    bitmap_t		*dev_ctxmap;					/* bitmap for local context allocation */

    spinlock_t		 dev_ctxt_lock;					/* spinlock to protect context list */
    struct list_head     dev_ctxt_list;					/* linked list of contexts */

    /* locks to sequentialise interrupt handling */
    spinlock_t		 dev_trap_lock;					/* spinlock while handling a trap */
    spinlock_t		 dev_requeue_lock;				/* spinlock sequentialising cproc requeue */

    /* error rate interrupt section */
    long		 dev_error_time;				/* lbolt at start of sampling period */
    unsigned		 dev_errors_per_period;				/* errors so far this sampling period */
    timer_fn_t		 dev_error_timeoutid;				/* timeout to re-enable error interrupts */
    timer_fn_t		 dev_linkerr_timeoutid;				/* timeout to clear link error led */

    /* kernel threads */
    unsigned		 dev_stop_threads:1;				/* kernel threads should exit */

    /* main interrupt thread */
    kcondvar_t		 dev_mainint_wait;				/* place for mainevent interrupt thread to sleep */
    spinlock_t		 dev_mainint_lock;				/*   and it's spinlock */
    unsigned		 dev_mainint_started:1;
    unsigned		 dev_mainint_stopped:1;

    /* device context - this is used to flush insert cache/instruction cache/dmas & threads */
    ELAN4_CPROC_TRAP     dev_cproc_trap;				/* space to extract cproc trap into */

    struct list_head     dev_intop_list;				/* list of main interrupt operations */
    spinlock_t		 dev_intop_lock;				/*   and spinlock */
    E4_uint64		 dev_intop_cookie;				/*   and next cookie to use */

    spinlock_t		 dev_flush_lock;				/* spinlock for flushing */
    kcondvar_t		 dev_flush_wait;				/*  and place to sleep */

    ELAN4_CQ		*dev_flush_cq[COMMAND_INSERTER_CACHE_ENTRIES];	/* command queues to flush the insert cache */
    ELAN4_INTOP          dev_flush_op[COMMAND_INSERTER_CACHE_ENTRIES];	/* and a main interrupt operation for each one */
    unsigned		 dev_flush_finished;				/* flush command finished */

    ELAN4_HALTOP	 dev_iflush_haltop;				/* halt operation for icache flush */
    unsigned		 dev_iflush_queued:1;				/* icache haltop queued */

    ELAN4_ROUTE_TABLE   *dev_routetable;				/* virtual process table (for dma queue flush)*/
    sdramaddr_t          dev_sdrampages[2];				/* pages of sdram to hold suspend code sequence */
    E4_Addr		 dev_tproc_suspend;				/*  st8suspend instruction */
    E4_Addr		 dev_tproc_space;				/*     and target memory */

    sdramaddr_t		 dev_neterr_inputq;				/* network error input queue descriptor & event */
    sdramaddr_t		 dev_neterr_slots;				/* network error message slots */
    ELAN4_CQ		*dev_neterr_msgcq;				/* command queue for sending messages */
    ELAN4_CQ		*dev_neterr_intcq;				/* command queue for message received interrupt */
    ELAN4_INTOP		 dev_neterr_intop;				/*   and it's main interrupt operation */
    E4_uint64		 dev_neterr_queued;				/* # message queued in msgcq */
    spinlock_t           dev_neterr_lock;				/*   and spinlock .... */

    ELAN4_DEV_STATS	 dev_stats;					/* device statistics */
    ELAN4_ECCERRS	 dev_sdramerrs[30];				/* last few sdram errors for procfs */

    unsigned int 	*dev_ack_errors;				/* Map of source of dproc ack errors */
    ELAN4_ROUTE_RINGBUF  dev_ack_error_routes;
    unsigned int        *dev_dproc_timeout;                             /* Ditto dproc timeout errors */
    ELAN4_ROUTE_RINGBUF  dev_dproc_timeout_routes;
    unsigned int        *dev_cproc_timeout;                             /* Ditto cproc timeout errors */
    ELAN4_ROUTE_RINGBUF  dev_cproc_timeout_routes;

    unsigned		 dev_linkerr_signalled;				/* linkerror signalled to switch controller */

    struct list_head     dev_hc_list;                                   /* list of the allocated hash_chunks */

    ELAN4_IPROC_TRAP     dev_iproc_trap;				/* space for iproc trap */
} ELAN4_DEV;

/* values for dev_state */
#define ELAN4_STATE_STOPPED		(1 << 0)			/* device initialised but not started */
#define ELAN4_STATE_STARTING		(1 << 1)			/* device in process of starting */
#define ELAN4_STATE_STARTED		(1 << 2)			/* device started */
#define ELAN4_STATE_STOPPING		(1 << 3)			/* device in process of stopping */

extern __inline__ unsigned long long
__elan4_readq (ELAN4_DEV *dev, ioaddr_t addr)
{
#if defined(__i386)
    if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_64BIT_READ)
    {
	uint64_t save[2];
	uint64_t rval;
	unsigned long flags, cr0;
	
	local_irq_save (flags);

	/* Save FPU state */
        asm volatile("mov %%cr0,%0 ; clts\n" : "=r" (cr0));
	
	/* GNAT 7726: Save 128-bit xmm0 register value */
	asm volatile ("movupd %%xmm0,%0\n" : "=m" (save[0]));
	
	/* Do a 64-bit PCI read */
	asm volatile ("sfence\n"
	              "movq (%1), %%xmm0\n"
		      "movq %%xmm0, %0\n"
		      "sfence\n"
		      : "=m" (rval) : "r" (addr) : "memory");
	
	/* GNAT 7726: Restore 128-bit xmm0 register value */
	asm volatile("movupd %0,%%xmm0\n" : : "m" (save[0]));
	
	/* Restore FPU state */
	asm volatile("mov %0,%%cr0\n" : : "r" (cr0));
	
	local_irq_restore(flags);
	
	return rval;
    }
#endif
    return readq ((void *)addr);
}

extern __inline__ unsigned int
__elan4_readl (ELAN4_DEV *dev, ioaddr_t addr)
{
    if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_64BIT_READ)
    {
	uint64_t val = __elan4_readq (dev, ((unsigned long) addr & ~7));
	return ((val >> (((unsigned long) addr & 7) << 3)) & 0xffffffff);
    }
    return readl ((void *)addr);
}

extern __inline__ unsigned int
__elan4_readw (ELAN4_DEV *dev, ioaddr_t addr)
{
    if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_64BIT_READ)
    {
	uint64_t val = __elan4_readq (dev, ((unsigned long) addr & ~7));
	return ((val >> (((unsigned long) addr & 7) << 3)) & 0xffff);
    }
    return readw ((void *)addr);
}

extern __inline__ unsigned int
__elan4_readb (ELAN4_DEV *dev, ioaddr_t addr)
{
    if (dev->dev_devinfo.dev_params.values[ELAN4_PARAM_DRIVER_FEATURES] & ELAN4_FEATURE_64BIT_READ)
    {
	uint64_t val = __elan4_readq (dev, ((unsigned long) addr & ~7));
	return ((val >> (((unsigned long) addr & 7) << 3)) & 0xff);
    }
    return readb ((void *)addr);
}

/* macros for accessing dev->dev_regs.Tags. */
#define write_tag(dev,what,val)		writeq (val, (void *) (dev->dev_regs + offsetof (E4_Registers, Tags.what)))
#define read_tag(dev,what)		__elan4_readq (dev, dev->dev_regs + offsetof (E4_Registers, Tags.what))

/* macros for accessing dev->dev_regs.Regs. */
#define write_reg64(dev,what,val)	writeq (val, (void *) (dev->dev_regs + offsetof (E4_Registers, Regs.what)))
#define write_reg32(dev,what,val)	writel (val, (void *) (dev->dev_regs + offsetof (E4_Registers, Regs.what)))
#define read_reg64(dev,what)		__elan4_readq (dev, dev->dev_regs + offsetof (E4_Registers, Regs.what))
#define read_reg32(dev,what)		__elan4_readl (dev, dev->dev_regs + offsetof (E4_Registers, Regs.what))

/* macros for accessing dev->dev_regs.uRegs. */
#define write_ureg64(dev,what,val)	writeq (val, (void *) (dev->dev_regs + offsetof (E4_Registers, uRegs.what)))
#define write_ureg32(dev,what,val)	writel (val, (void *) (dev->dev_regs + offsetof (E4_Registers, uRegs.what)))
#define read_ureg64(dev,what)		__elan4_readq (dev, dev->dev_regs + offsetof (E4_Registers, uRegs.what))
#define read_ureg32(dev,what)		__elan4_readl (dev, dev->dev_regs + offsetof (E4_Registers, uRegs.what))

/* macros for accessing dev->dev_i2c */
#define write_i2c(dev,what,val)		writeb (val, (void *) (dev->dev_i2c + offsetof (E4_I2C, what)))
#define read_i2c(dev,what)		__elan4_readb (dev, dev->dev_i2c + offsetof (E4_I2C, what))

/* macros for accessing dev->dev_rom */
#define read_ebus_rom(dev,off)		__elan4_readb (dev, dev->dev_rom + off)

/* PIO flush operations - ensure writes to registers/sdram are ordered */
#ifdef CONFIG_IA64_SGI_SN2
#define pioflush_reg(dev)		read_reg32(dev,InterruptReg)
#define pioflush_sdram(dev)		elan4_sdram_readl(dev, 0)
#else
#define pioflush_reg(dev)		mb()
#define pioflush_sdram(dev)		mb()
#endif

/* macros for manipulating the interrupt mask register */
#define SET_INT_MASK(dev,value)	\
do { \
    write_reg32(dev, InterruptMask, (dev)->dev_intmask = (value)); \
    pioflush_reg(dev);\
} while (0)

#define CHANGE_INT_MASK(dev, value) \
do { \
    if ((dev)->dev_intmask != (value)) \
    {\
	write_reg32 (dev, InterruptMask, (dev)->dev_intmask = (value));\
	pioflush_reg(dev);\
    }\
} while (0)

#define ENABLE_INT_MASK(dev,value) \
do { \
    unsigned long flags; \
 \
    spin_lock_irqsave (&(dev)->dev_intmask_lock, flags); \
    write_reg32(dev, InterruptMask, (dev)->dev_intmask |= (value)); \
    pioflush_reg(dev);\
    spin_unlock_irqrestore (&(dev)->dev_intmask_lock, flags); \
} while (0)

#define DISABLE_INT_MASK(dev,value) \
do { \
    unsigned long flags; \
 \
    spin_lock_irqsave (&(dev)->dev_intmask_lock, flags); \
    write_reg32(dev, InterruptMask, (dev)->dev_intmask &= ~(value)); \
    pioflush_reg(dev);\
    spin_unlock_irqrestore (&(dev)->dev_intmask_lock, flags); \
} while (0)

#define SET_SYSCONTROL(dev,what,value) \
do { \
    unsigned long flags; \
\
    spin_lock_irqsave (&(dev)->dev_syscontrol_lock, flags); \
    if ((dev)->what++ == 0) \
        write_reg64 (dev, SysControlReg, (dev)->dev_syscontrol |= (value)); \
    pioflush_reg(dev);\
    spin_unlock_irqrestore (&(dev)->dev_syscontrol_lock, flags); \
} while (0)

#define CLEAR_SYSCONTROL(dev,what,value) \
do { \
    unsigned long flags; \
\
    spin_lock_irqsave (&(dev)->dev_syscontrol_lock, flags); \
    if (--(dev)->what == 0)\
       write_reg64 (dev, SysControlReg, (dev)->dev_syscontrol &= ~(value)); \
    pioflush_reg (dev); \
    spin_unlock_irqrestore (&(dev)->dev_syscontrol_lock, flags); \
} while (0)

#define PULSE_SYSCONTROL(dev,value) \
do { \
    unsigned long flags; \
\
    spin_lock_irqsave (&(dev)->dev_syscontrol_lock, flags); \
    write_reg64 (dev, SysControlReg, (dev)->dev_syscontrol | (value)); \
    pioflush_reg (dev); \
    spin_unlock_irqrestore (&(dev)->dev_syscontrol_lock, flags); \
} while (0)

#define CHANGE_SYSCONTROL(dev,add,sub) \
do { \
    unsigned long flags; \
\
    spin_lock_irqsave (&(dev)->dev_syscontrol_lock, flags); \
    dev->dev_syscontrol |= (add);\
    dev->dev_syscontrol &= ~(sub);\
    write_reg64 (dev, SysControlReg, (dev)->dev_syscontrol);\
    pioflush_reg (dev); \
    spin_unlock_irqrestore (&(dev)->dev_syscontrol_lock, flags); \
} while (0)

#define SET_SCHED_STATUS(dev, value)\
do {\
    write_reg32 (dev, SchedStatus.Status, (dev)->dev_schedstatus = (value));\
    pioflush_reg (dev);\
} while (0)

#define CHANGE_SCHED_STATUS(dev, value)\
do {\
    if ((dev)->dev_schedstatus != (value))\
    {\
	write_reg32 (dev, SchedStatus.Status, (dev)->dev_schedstatus = (value));\
	pioflush_reg (dev);\
    }\
} while (0)

#define PULSE_SCHED_RESTART(dev,value)\
do {\
    write_reg32 (dev, SchedStatus.Restart, value);\
    pioflush_reg (dev);\
} while (0)

/* device context elan address space */
#define DEVICE_TPROC_SUSPEND_ADDR		(0x1000000000000000ull)
#define DEVICE_TPROC_SPACE_ADDR			(0x1000000000000000ull + SDRAM_PAGE_SIZE)
#if defined(__LITTLE_ENDIAN__)
#  define DEVICE_TPROC_SUSPEND_INSTR		0xd3f040c0 /* st64suspend %r16, [%r1] */
#else
#  define DEVICE_TPROC_SUSPEND_INSTR		0xc040f0d3 /* st64suspend %r16, [%r1] */
#endif

#define DEVICE_NETERR_INPUTQ_ADDR		(0x2000000000000000ull)
#define DEVICE_NETERR_INTCQ_ADDR		(0x2000000000000000ull + SDRAM_PAGE_SIZE)
#define DEVICE_NETERR_SLOTS_ADDR		(0x2000000000000000ull + SDRAM_PAGE_SIZE*2)

/*
 * Interrupt operation cookie space
 * [50:48]	type
 * [47:0]	value
 */
#define INTOP_PERSISTENT			(0x1000000000000ull)
#define INTOP_ONESHOT				(0x2000000000000ull)
#define INTOP_TYPE_MASK				(0x3000000000000ull)
#define INTOP_VALUE_MASK			(0x0ffffffffffffull)

/* functions for accessing sdram - sdram.c */
extern unsigned char      elan4_sdram_readb (ELAN4_DEV *dev, sdramaddr_t ptr);
extern unsigned short     elan4_sdram_readw (ELAN4_DEV *dev, sdramaddr_t ptr);
extern unsigned int       elan4_sdram_readl (ELAN4_DEV *dev, sdramaddr_t ptr);
extern unsigned long long elan4_sdram_readq (ELAN4_DEV *dev, sdramaddr_t ptr);
extern void               elan4_sdram_writeb (ELAN4_DEV *dev, sdramaddr_t ptr, unsigned char val);
extern void               elan4_sdram_writew (ELAN4_DEV *dev, sdramaddr_t ptr, unsigned short val);
extern void               elan4_sdram_writel (ELAN4_DEV *dev, sdramaddr_t ptr, unsigned int val);
extern void               elan4_sdram_writeq (ELAN4_DEV *dev, sdramaddr_t ptr, unsigned long long val);

extern void		  elan4_sdram_zerob_sdram (ELAN4_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan4_sdram_zerow_sdram (ELAN4_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan4_sdram_zerol_sdram (ELAN4_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void		  elan4_sdram_zeroq_sdram (ELAN4_DEV *dev, sdramaddr_t ptr, int nbytes);

extern void               elan4_sdram_copyb_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan4_sdram_copyw_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan4_sdram_copyl_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan4_sdram_copyq_from_sdram (ELAN4_DEV *dev, sdramaddr_t from, void *to, int nbytes);
extern void               elan4_sdram_copyb_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan4_sdram_copyw_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan4_sdram_copyl_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes);
extern void               elan4_sdram_copyq_to_sdram (ELAN4_DEV *dev, void *from, sdramaddr_t to, int nbytes);

/* device.c - configuration */
extern unsigned int elan4_hash_0_size_val;
extern unsigned int elan4_hash_1_size_val;
extern unsigned int elan4_ctxt_table_shift;
extern unsigned int elan4_ln2_max_cqs;
extern unsigned int elan4_dmaq_highpri_size;
extern unsigned int elan4_threadq_highpri_size;
extern unsigned int elan4_dmaq_lowpri_size;
extern unsigned int elan4_threadq_lowpri_size;
extern unsigned int elan4_interruptq_size;
extern unsigned int elan4_mainint_punt_loops;
extern unsigned int elan4_mainint_resched_ticks;
extern unsigned int elan4_linkport_lock;
extern unsigned int elan4_eccerr_recheck;

/* device.c */
extern void               elan4_set_schedstatus (ELAN4_DEV *dev, E4_uint32 intreg);
extern void               elan4_queue_haltop (ELAN4_DEV *dev, ELAN4_HALTOP *op);
extern void		  elan4_queue_intop (ELAN4_DEV *dev, ELAN4_CQ *cq, ELAN4_INTOP *op);
extern void		  elan4_register_intop (ELAN4_DEV *dev, ELAN4_INTOP *op);
extern void		  elan4_deregister_intop (ELAN4_DEV *dev, ELAN4_INTOP *op);
extern void		  elan4_queue_dma_flushop (ELAN4_DEV *dev, ELAN4_DMA_FLUSHOP *op, int hipri);
extern void		  elan4_queue_mainintop (ELAN4_DEV *dev, ELAN4_INTOP *op);

extern int                elan4_1msi0 (ELAN4_DEV *dev);

extern int                elan4_insertctxt (ELAN4_DEV *dev, ELAN4_CTXT *ctxt, ELAN4_TRAP_OPS *ops);
extern void               elan4_removectxt (ELAN4_DEV *dev, ELAN4_CTXT *ctxt);
extern ELAN4_CTXT        *elan4_localctxt (ELAN4_DEV *dev, unsigned num);
extern ELAN4_CTXT        *elan4_networkctxt (ELAN4_DEV *dev, unsigned num);

extern int                elan4_attach_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum);
extern void               elan4_detach_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum);
extern void		  elan4_set_filter (ELAN4_CTXT *ctxt, unsigned int ctxnum, E4_uint32 state);
extern void		  elan4_set_routetable (ELAN4_CTXT *ctxt, ELAN4_ROUTE_TABLE *tbl);

extern ELAN4_CQA *        elan4_getcqa (ELAN4_CTXT *ctxt, unsigned int idx);
extern void               elan4_putcqa (ELAN4_CTXT *ctxt, unsigned int idx);
extern ELAN4_CQ          *elan4_alloccq (ELAN4_CTXT *ctxt, unsigned cqsize, unsigned cqperm, unsigned cqtype);
extern void               elan4_freecq (ELAN4_CTXT *ctxt, ELAN4_CQ *cq);
extern void               elan4_restartcq (ELAN4_DEV *dev, ELAN4_CQ *cq);
extern void               elan4_flushcq (ELAN4_DEV *dev, ELAN4_CQ *cq);
extern void               elan4_updatecq (ELAN4_DEV *dev, ELAN4_CQ *cq, unsigned perm, unsigned restart);

extern void		  elan4_flush_icache (ELAN4_CTXT *ctxt);
extern void		  elan4_flush_icache_halted (ELAN4_CTXT *ctxt);

extern int                elan4_initialise_device (ELAN4_DEV *dev);
extern void               elan4_finalise_device (ELAN4_DEV *dev);
extern int                elan4_start_device (ELAN4_DEV *dev);
extern void               elan4_stop_device (ELAN4_DEV *dev);

extern int		  elan4_compute_position (ELAN_POSITION *pos, unsigned nodeid, unsigned numnodes, unsigned aritiyval);
extern int		  elan4_get_position (ELAN4_DEV *dev, ELAN_POSITION *pos);
extern int		  elan4_set_position (ELAN4_DEV *dev, ELAN_POSITION *pos);
extern void		  elan4_get_params   (ELAN4_DEV *dev, ELAN_PARAMS *params, unsigned short *mask);
extern void		  elan4_set_params   (ELAN4_DEV *dev, ELAN_PARAMS *params, unsigned short mask);


extern int                elan4_read_vpd(ELAN4_DEV *dev, unsigned char *tag, unsigned char *result) ;

extern void               proc_insertctxt(ELAN4_DEV *dev,ELAN4_CTXT *ctxt);
extern void               proc_removectxt(ELAN4_DEV *dev,ELAN4_CTXT *ctxt);

extern int                elan4_route2str (E4_VirtualProcessEntry *route, char *routeStr);
extern void               elan4_hardware_lock_check(ELAN4_DEV *dev, char *from);

/* device_osdep.c */
extern unsigned int	  elan4_pll_cfg;
extern int		  elan4_pll_div;
extern int		  elan4_mod45disable;
extern int                assfail_mode;

extern int                elan4_pciinit (ELAN4_DEV *dev);
extern void               elan4_pcifini (ELAN4_DEV *dev);
extern void		  elan4_updatepll (ELAN4_DEV *dev, unsigned int val);
extern void               elan4_pcierror (ELAN4_DEV *dev);

extern ELAN4_DEV	 *elan4_reference_device (int instance, int state);
extern void		  elan4_dereference_device (ELAN4_DEV *dev);

extern ioaddr_t           elan4_map_device (ELAN4_DEV *dev, unsigned bar, unsigned off, unsigned size, ELAN4_MAP_HANDLE *handlep);
extern void               elan4_unmap_device (ELAN4_DEV *dev, ioaddr_t ptr, unsigned size, ELAN4_MAP_HANDLE *handlep);
extern unsigned long      elan4_resource_len (ELAN4_DEV *dev, unsigned bar);

extern void               elan4_configure_writecombining (ELAN4_DEV *dev);
extern void		  elan4_unconfigure_writecombining (ELAN4_DEV *dev);

/* i2c.c */
extern int		  i2c_disable_auto_led_update (ELAN4_DEV *dev);
extern void		  i2c_enable_auto_led_update (ELAN4_DEV *dev);
extern int		  i2c_write (ELAN4_DEV *dev, unsigned int addr, unsigned int count, unsigned char *data);
extern int		  i2c_read (ELAN4_DEV *dev, unsigned int addr, unsigned int count, unsigned char *data);
extern int		  i2c_writereg (ELAN4_DEV *dev, unsigned int addr, unsigned int reg, unsigned int count, unsigned char *data);
extern int		  i2c_readreg (ELAN4_DEV *dev, unsigned int addr, unsigned int reg, unsigned int count, unsigned char *data);
extern int		  i2c_read_rom (ELAN4_DEV *dev, unsigned int addr, unsigned int count, unsigned char *data);

#if defined(__linux__)
/* procfs_Linux.c */
extern void 		  elan4_procfs_device_init (ELAN4_DEV *dev);
extern void 		  elan4_procfs_device_fini (ELAN4_DEV *dev);
extern void 		  elan4_procfs_init(void);
extern void 		  elan4_procfs_fini(void);

extern struct proc_dir_entry *elan4_procfs_root;
extern struct proc_dir_entry *elan4_config_root;
#endif

/* sdram.c */
extern void	          elan4_sdram_init (ELAN4_DEV *dev);
extern void               elan4_sdram_fini (ELAN4_DEV *dev);
extern void               elan4_sdram_setup_delay_lines (ELAN4_DEV *dev, int factor);
extern int                elan4_sdram_init_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank);
extern void               elan4_sdram_fini_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank);
extern void	          elan4_sdram_add_bank (ELAN4_DEV *dev, ELAN4_SDRAM_BANK *bank);
extern sdramaddr_t        elan4_sdram_alloc (ELAN4_DEV *dev, int nbytes);
extern void               elan4_sdram_free (ELAN4_DEV *dev, sdramaddr_t ptr, int nbytes);
extern void               elan4_sdram_flushcache (ELAN4_DEV *dev, sdramaddr_t base, int nbytes);
extern char              *elan4_sdramerr2str (ELAN4_DEV *dev, E4_uint64 status, E4_uint64 ConfigReg, char *str);

/* traps.c */
extern void               elan4_display_eproc_trap (void *type, int mode, char *str, ELAN4_EPROC_TRAP *trap);
extern void               elan4_display_cproc_trap (void *type, int mode, char *str, ELAN4_CPROC_TRAP *trap);
extern void               elan4_display_dproc_trap (void *type, int mode, char *str, ELAN4_DPROC_TRAP *trap);
extern void               elan4_display_tproc_trap (void *type, int mode, char *str, ELAN4_TPROC_TRAP *trap);
extern void               elan4_display_iproc_trap (void *type, int mode, char *str, ELAN4_IPROC_TRAP *trap);


extern void               elan4_extract_eproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_EPROC_TRAP *trap, int iswaitevent);
extern void               elan4_extract_cproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_CPROC_TRAP *trap, unsigned cqnum);
extern void               elan4_extract_dproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_DPROC_TRAP *trap, unsigned unit);
extern void               elan4_extract_tproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_TPROC_TRAP *trap);
extern void               elan4_extract_iproc_trap (ELAN4_DEV *dev, E4_uint64 status, ELAN4_IPROC_TRAP *trap, unsigned unit);
extern void elan4_ringbuf_store(ELAN4_ROUTE_RINGBUF *ringbuf, E4_VirtualProcessEntry *route, ELAN4_DEV *dev);
extern int                cproc_open_extract_vp (ELAN4_DEV *dev, ELAN4_CQ *cq, int chan);

extern void               elan4_inspect_iproc_trap (ELAN4_IPROC_TRAP *trap);
extern E4_uint64          elan4_trapped_open_command (ELAN4_DEV *dev, ELAN4_CQ *cq);

/* mmu.c */
extern void               elan4mmu_flush_tlb (ELAN4_DEV *dev);
extern ELAN4_HASH_ENTRY  *elan4mmu_ptealloc (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, unsigned int *tagidxp);
extern int                elan4mmu_pteload (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, int type, E4_uint64 pte);
extern int                elan4mmu_pteload_page (ELAN4_CTXT *ctxt, int tbl, E4_Addr vaddr, struct page *page, int perm);
extern void               elan4mmu_pteunload (ELAN4_CTXT *ctxt, ELAN4_HASH_ENTRY *he, unsigned int tagidx, unsigned int pteidx);
extern void               elan4mmu_unload_range (ELAN4_CTXT *ctxt, int tbl, E4_Addr start, unsigned long len);
extern void               elan4mmu_invalidate_ctxt (ELAN4_CTXT *ctxt);

extern ELAN4_HASH_CACHE  *elan4mmu_reserve (ELAN4_CTXT *ctxt, int tbl, E4_Addr start, unsigned int npages, int cansleep);
extern void               elan4mmu_release (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc);
extern void               elan4mmu_set_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx, E4_uint64 newpte);
extern E4_uint64          elan4mmu_get_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx);
extern void               elan4mmu_clear_pte (ELAN4_CTXT *ctxt, ELAN4_HASH_CACHE *hc, unsigned int idx);

extern int                elan4mmu_display_mmuhash(ELAN4_DEV *dev, int tlb, int *index, char *page, int count);
extern int                elan4mmu_display_ctxt_mmuhash(ELAN4_CTXT *ctxt, int tlb, int *index, char *page, int count);
extern int                elan4mmu_display_bucket_mmuhash(ELAN4_DEV *dev, int tlb, int *buckets, int nBuckets, char *page, int count);
extern void               elan4mmu_do_shuffle(ELAN4_CTXT *ctxt, int tbl);
extern void               elan4mmu_set_shuffle(ELAN4_CTXT *ctxt, int tbl, int hashidx);

/* mmu_osdep.c */
extern int                elan4mmu_sdram_aliascheck (ELAN4_CTXT *ctxt, E4_Addr addr, sdramaddr_t phys);
extern int                elan4mmu_alloc_topaddr (ELAN4_DEV *dev, physaddr_t paddr, unsigned type);
extern E4_uint64          elan4mmu_phys2pte (ELAN4_DEV *dev, physaddr_t phys, unsigned perm);
extern physaddr_t	  elan4mmu_pte2phys (ELAN4_DEV *dev, E4_uint64 pte);

/* neterr.c */
extern int                elan4_neterr_init (ELAN4_DEV *dev);
extern void               elan4_neterr_destroy (ELAN4_DEV *dev);
extern int                elan4_neterr_sendmsg (ELAN4_DEV *dev, unsigned int nodeid, unsigned int retries, ELAN4_NETERR_MSG *msg);
extern int                elan4_neterr_iproc_trap (ELAN4_DEV *dev, ELAN4_IPROC_TRAP *trap);

/* routetable.c */
extern ELAN4_ROUTE_TABLE *elan4_alloc_routetable (ELAN4_DEV *dev, unsigned size);
extern void               elan4_free_routetable (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl);
extern void               elan4_write_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp, E4_VirtualProcessEntry *entry);
extern void               elan4_read_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp, E4_VirtualProcessEntry *entry);
extern void               elan4_invalidate_route (ELAN4_DEV *dev, ELAN4_ROUTE_TABLE *tbl, unsigned vp);
extern int                elan4_generate_route (ELAN_POSITION *pos, E4_VirtualProcessEntry *route, unsigned ctxnum,
						unsigned lowid, unsigned highid, unsigned options);
extern int		  elan4_check_route (ELAN_POSITION *pos, ELAN_LOCATION location, E4_VirtualProcessEntry *route, unsigned flags);

/* user.c */
extern int        __categorise_command (E4_uint64 command, int *cmdSize);
extern int        __whole_command (sdramaddr_t *commandPtr, sdramaddr_t insertPtr, unsigned int cqSize, unsigned int cmdSize);

/* debug.c */
extern int        elan4_assfail (ELAN4_CTXT *ctxt, const char *ex, const char *func, const char *file, const int line);
extern int        elan4_debug_trigger (ELAN4_CTXT *ctxt, const char *func, const char *file, const int line, const char *fmt, ...);

#if defined(DEBUG_ASSERT)
#define ELAN4_ASSERT(ctxt,EXPR)			do { \
    if (!(EX) && elan4_assfail (ctxt, #EXPR, __FUNCTION__, __FILE__, __LINE__)) { \
	BUG(); \
    } \
} while (0)
#else
#define ELAN4_ASSERT(ctxt,EXPR)			((void) 0)
#endif

#define ELAN4_DEBUG_TRIGGER(ctxt,fmt, args...)	do {\
    if (elan4_debug_trigger (ctxt, __FUNCTION__, __FILE__, __LINE__, fmt, ##args)) \
	BUG();\
} while (0)

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_ELANDEV_H */
