/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_ELANDEV_LINUX_H
#define __ELAN4_ELANDEV_LINUX_H

#ident "$Id: device_Linux.h,v 1.26 2005/04/05 15:29:28 robin Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/device_Linux.h,v $*/

#include <qsnet/autoconf.h>

#if !defined(NO_COPROC)				/* The older coproc kernel patch is applied */
#include <linux/coproc.h>

#define ioproc_ops		coproc_ops_struct
#define ioproc_register_ops	register_coproc_ops
#define ioproc_unregister_ops	unregister_coproc_ops

#define IOPROC_MM_STRUCT_ARG	1
#define IOPROC_PATCH_APPLIED	1

#elif !defined(NO_IOPROC)			/* The new ioproc kernel patch is applied */
#include <linux/ioproc.h>

#define IOPROC_PATCH_APPLIED	1
#endif


#if defined(MPSAS)
#include <elan4/mpsas.h>
#endif

#if defined(CONFIG_DEVFS_FS)
#include <linux/devfs_fs_kernel.h>
#endif

#define ELAN4_MAJOR              61
#define ELAN4_NAME               "elan4"
#define ELAN4_MAX_CONTROLLER     16	       /* limited to 4 bits */
 
/* OS dependant component of ELAN4_DEV struct */
typedef struct elan4_dev_osdep
{
    struct pci_dev	  *pdev;			/* PCI config data */

    struct proc_dir_entry *procdir;
    struct proc_dir_entry *configdir;
    struct proc_dir_entry *statsdir;
    struct proc_dir_entry *ctxtdir;

#if defined(CONFIG_DEVFS_FS)
    devfs_handle_t devfs_control;
    devfs_handle_t devfs_sdram;
    devfs_handle_t devfs_user;
#endif

#if defined(CONFIG_MTRR)
    int			   sdram_mtrr;
    int			   regs_mtrr;
#endif
} ELAN4_DEV_OSDEP;

/* /dev/elan/rmsX */

/* /dev/elan4/controlX */
typedef struct control_private
{
    struct elan4_dev   *pr_dev;
    unsigned		pr_boundary_scan;
} CONTROL_PRIVATE;

/* /dev/elan4/sdramX */
typedef struct mem_page
{
    struct mem_page *pg_next;
    sdramaddr_t      pg_addr;
    unsigned long    pg_pgoff;
    unsigned	     pg_ref;
} MEM_PAGE;

#define MEM_HASH_SIZE	32
#define MEM_HASH(pgoff)	((pgoff) & (MEM_HASH_SIZE-1))

typedef struct mem_private
{
    struct elan4_dev *pr_dev;
    MEM_PAGE         *pr_pages[MEM_HASH_SIZE];
    spinlock_t        pr_lock;
} MEM_PRIVATE;

/* /dev/elan4/userX */
typedef struct user_private
{
    atomic_t	      pr_ref;
    struct user_ctxt *pr_uctx;
    struct mm_struct *pr_mm;

#if defined(IOPROC_PATCH_APPLIED)
    struct ioproc_ops pr_ioproc;
#endif
} USER_PRIVATE;

/* No mapping handles on linux */
typedef void *ELAN4_MAP_HANDLE;

#define ELAN4_TASK_HANDLE()	((unsigned long) current->mm)

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_ELANDEV_LINUX_H */
