/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_DMA_H
#define __ELAN4_DMA_H

#ident "$Id: dma.h,v 1.16 2003/09/04 12:39:17 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/dma.h,v $*/

#include <elan4/types.h>

/* Alignment for a DMA descriptor */
#define E4_DMA_ALIGN		(64)

/* Maximum size of a single DMA ((1 << 31)-1) */
#define E4_MAX_DMA_SIZE		(0x7fffffff)

/* 
 * dma_typeSize
 *
 * [63:32]	Size
 * [31]		unused
 * [30]		IsRemote
 * [29]		QueueWrite
 * [28]		ShmemWrite
 * [27:26]	DataType
 * [25]		Broadcast
 * [24]		AlignPackets
 * [23:16]	FailCount
 * [15:14]	unused
 * [13:0]	Context
 */

#define DMA_FailCount(val)	(((val) & 0xff) << 16)
#define DMA_AlignPackets	(1 << 24)
#define DMA_Broadcast		(1 << 25)
#define DMA_ShMemWrite		(1 << 28)
#define DMA_QueueWrite		(1 << 29)
#define DMA_IsRemote		(1 << 30)
#define DMA_Context(val)	((unsigned) (val) & 0x3ff)
#define DMA_ContextMask		0x3fffull
#define Dma_TypeSizeMask	0xfffffffffff00000ull

#define DMA_DataTypeByte	(E4_DATATYPE_BYTE  << 26)
#define DMA_DataTypeShort	(E4_DATATYPE_SHORT << 26)
#define DMA_DataTypeWord	(E4_DATATYPE_WORD  << 26)
#define DMA_DataTypeLong	(E4_DATATYPE_DWORD << 26)

#define E4_DMA_TYPE_SIZE(size, dataType, flags, failCount)	\
    ((((E4_uint64)(size)) << 32) |  ((dataType) & DMA_DataTypeLong) | \
     (flags) | DMA_FailCount(failCount))

typedef volatile struct e4_dma
{
    E4_uint64		dma_typeSize;
    E4_uint64		dma_cookie;
    E4_uint64		dma_vproc;
    E4_Addr		dma_srcAddr;
    E4_Addr		dma_dstAddr;
    E4_Addr		dma_srcEvent;
    E4_Addr		dma_dstEvent;
} E4_DMA;

/* Same as above but padded to 64-bytes */
typedef volatile struct e4_dma64
{
    E4_uint64		dma_typeSize;
    E4_uint64		dma_cookie;
    E4_uint64		dma_vproc;
    E4_Addr		dma_srcAddr;
    E4_Addr		dma_dstAddr;
    E4_Addr		dma_srcEvent;
    E4_Addr		dma_dstEvent;
    E4_Addr		dma_pad;
} E4_DMA64;

#endif /* __ELAN4_DMA_H */
