/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_EVENTS_H
#define __ELAN4_EVENTS_H

#ident "$Id: events.h,v 1.22 2004/06/23 11:07:18 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/events.h,v $*/

#define E4_EVENT_ALIGN		32
#define E4_EVENTBLOCK_SIZE	64

#ifndef _ASM
/*
 * Event locations must be aligned to a 32 byte boundary. It is very much more efficent to place
 * them in elan local memory but is not essential.
 */
typedef struct _E4_Event
{
    volatile E4_uint64	ev_CountAndType;
    E4_uint64	        ev_Params[2];
} E4_Event;

/* Same as above but padded to correct Event alignment */
typedef struct _E4_Event32
{
    volatile E4_uint64	ev_CountAndType;
    E4_uint64	        ev_Params[2];
    E4_uint64	        ev_pad;
} E4_Event32;

/*
 * An E4_EVENTBLOCK_SIZE aligned block of Main or Elan memory
 */
typedef union _E4_Event_Blk
{
    /* Padded to 64-bytes in case a cache-line write is more efficient */
    volatile E4_uint8  eb_unit8[E4_EVENTBLOCK_SIZE];
    volatile E4_uint32 eb_uint32[E4_EVENTBLOCK_SIZE/sizeof(E4_uint32)];
    volatile E4_uint64 eb_uint64[E4_EVENTBLOCK_SIZE/sizeof(E4_uint64)];
} E4_Event_Blk;
#define eb_done       eb_uint32[14]
#define eb_done_dword eb_uint64[7]

#endif /* ! _ASM */

/*
 * ev_CountAndType
 *  [63:31]   Count
 *  [10]      CopyType
 *  [9:8]     DataType
 *  [7:0]     CopySize
 */
#define E4_EVENT_TYPE_MASK	0x00000000ffffffffull
#define E4_EVENT_COUNT_MASK	0xffffffff00000000ull
#define E4_EVENT_COUNT_SHIFT	32
#define E4_EVENT_COPY_TYPE_MASK	(1 << 10)
#define E4_EVENT_DATA_TYPE_MASK	(3 << 8)
#define E4_EVENT_COPY_SIZE_MASK	(0xff)

/* CopyType */
#define E4_EVENT_COPY		(0 << 10)
#define E4_EVENT_WRITE		(1 << 10)

/* DataType */
#define E4_EVENT_DTYPE_BYTE	(0 << 8)
#define E4_EVENT_DTYPE_SHORT	(1 << 8)
#define E4_EVENT_DTYPE_WORD	(2 << 8)
#define E4_EVENT_DTYPE_LONG	(3 << 8)

#define EVENT_COUNT(EventPtr)	((E4_int32)(elan4_load64 (&(EventPtr)->ev_CountAndType) >> E4_EVENT_COUNT_SHIFT))
#define EVENT_TYPE(EventPtr)	((E4_uint32)(elan4_load64 (&(EventPtr)->ev_CountAndType) & E4_EVENT_TYPE_MASK))

#define E4_WAITEVENT_COUNT_TYPE_VALUE(Count, EventType, DataType, CopySize) \
	(((E4_uint64)(Count) << E4_EVENT_COUNT_SHIFT) | (EventType) | (DataType) | (CopySize))

#define E4_EVENT_TYPE_VALUE(EventType, DataType, CopySize)	\
	((EventType) | (DataType) | (CopySize))

#define E4_EVENT_INIT_VALUE(InitialCount, EventType, DataType, CopySize)	\
	(((E4_uint64)(InitialCount) << E4_EVENT_COUNT_SHIFT) | E4_EVENT_TYPE_VALUE(EventType, DataType, CopySize))

#define ev_CopySource	ev_Params[0]
#define ev_CopyDest	ev_Params[1]
#define ev_WritePtr	ev_Params[0]
#define ev_WriteValue	ev_Params[1]

#define EVENT_BLK_READY(BLK) ((BLK)->eb_done != 0)
#define EVENT_READY(EVENT)   ((E4_uint32)((((volatile E4_Event *) (EVENT))->ev_CountAndType) >> E4_EVENT_COUNT_SHIFT) >= 0)

#define ELAN_WAIT_EVENT (0)
#define ELAN_POLL_EVENT (-1)

#define E4_BLK_PATTERN	((E4_uint32)0xfeedface)

#define E4_INIT_COPY_EVENT(EVENT, BLK_ELAN, BLK, SIZE) 		  					\
	do {												\
	   elan4_store64 (E4_EVENT_INIT_VALUE(0, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, SIZE), &(EVENT)->ev_CountAndType); \
           elan4_store64 ((BLK_ELAN), &(EVENT)->ev_CopySource); \
    	   elan4_store64 ((BLK), &(EVENT)->ev_CopyDest); \
	} while (0)

#define E4_INIT_WRITE_EVENT(EVENT, DWORD) 				  				\
	do {												\
	    elan4_store64 (E4_EVENT_INIT_VALUE(0, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0), &(EVENT)->ev_CountAndType);	\
	    elan4_store64 ((DWORD), &(EVENT)->ev_WritePtr); \
    	    elan4_store64 ((E4_Addr) (E4_BLK_PATTERN), &(EVENT)->ev_WriteValue); \
	} while (0)

#define E4_RESET_BLK_EVENT(BLK) 					\
	do {								\
		(BLK)->eb_done = (0);					\
	} while (0)

#define E4_PRIME_BLK_EVENT(EVENT, COUNT)				\
    	do {								\
	   elan4_store64 (E4_EVENT_INIT_VALUE(COUNT, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, 8), &(EVENT)->ev_CountAndType);\
	} while (0)

#define E4_PRIME_COPY_EVENT(EVENT, SIZE, COUNT)				\
    	do {								\
	   elan4_store64 (E4_EVENT_INIT_VALUE(COUNT, E4_EVENT_COPY, E4_EVENT_DTYPE_LONG, (SIZE >> 3)), &(EVENT)->ev_CountAndType);\
	} while (0)

#define E4_PRIME_WRITE_EVENT(EVENT, COUNT)					\
    	do {									\
	   elan4_store64 (E4_EVENT_INIT_VALUE(COUNT, E4_EVENT_WRITE, E4_EVENT_DTYPE_LONG, 0), &(EVENT)->ev_CountAndType);\
	} while (0)

#ifndef _ASM

#define E4_INPUTQ_ALIGN			32	/* Descriptor must be 32-byte aligned */

typedef struct _E4_InputQueue
{
   volatile E4_Addr	q_bptr;		/* 64 bit aligned ptr to current back item */
   E4_Addr		q_fptr;		/* 64 bit aligned ptr to current front item */
   E4_uint64		q_control;	/* this defines the last item, item size, and offset back to the first item. */
   E4_Addr		q_event;	/* queue event */
} E4_InputQueue;

#define E4_INPUTQ_LASTITEM_MASK 	0x00000000ffffffffULL
#define E4_INPUTQ_ITEMSIZE_MASK		0x000000ff00000000ULL
#define E4_INPUTQ_LASTITEM_OFFSET_MASK	0xffffff0000000000ULL
#define E4_INPUTQ_LASTITEM_SHIFT	0
#define E4_INPUTQ_ITEMSIZE_SHIFT	32
#define E4_INPUTQ_LASTITEM_OFFSET_SHIFT	40

/*
 * Macro to initialise the InputQueue control word given the FirstItem, LastItem & ItemSize
 * FirstItem and LastItem are 64 bit double word aligned elan addresses.
 */
#define E4_InputQueueControl(FirstItem, LastItem, ItemSizeInBytes)\
   (((((E4_uint64)(LastItem)))                                                      & E4_INPUTQ_LASTITEM_MASK) |\
    ((((E4_uint64)(ItemSizeInBytes))        << (E4_INPUTQ_ITEMSIZE_SHIFT-3))        & E4_INPUTQ_ITEMSIZE_MASK)  |\
    ((((E4_uint64)((FirstItem)-(LastItem))) << (E4_INPUTQ_LASTITEM_OFFSET_SHIFT-3)) & E4_INPUTQ_LASTITEM_OFFSET_MASK))    

/* 
 * LastItemOffset is a sign extended -ve quantity with LastItemOffset[26:3] == q_control[63:40]
 * we sign extend this by setting LastItemOffset[63:27] to be #one.
 */
#define E4_InputQueueLastItemOffset(control)  ((((E4_int64) -1) << (64 - (E4_INPUTQ_LASTITEM_OFFSET_SHIFT-3))) | \
					       ((E4_int64) (((control) & E4_INPUTQ_LASTITEM_OFFSET_MASK) >> (E4_INPUTQ_LASTITEM_OFFSET_SHIFT-3))))
#define E4_InputQueueItemSize(control)	      (((control) & E4_INPUTQ_ITEMSIZE_MASK) >> (E4_INPUTQ_ITEMSIZE_SHIFT-3))

/*
 * Macro to increment the InputQ front pointer taking into account wrap 
 */
#define E4_InputQueueFptrIncrement(Q, FirstItem, LastItem, ItemSizeInBytes) \
	((Q)->q_fptr = ( ((Q)->q_fptr == (LastItem)) ? (FirstItem) : ((Q)->q_fptr + (ItemSizeInBytes))) )

#endif /* _ASM */

#endif /* __ELAN4_EVENTS_H */
