/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_IOCTL_H
#define __ELAN4_IOCTL_H

#ident "@(#)$Id: ioctl.h,v 1.29 2005/01/10 17:45:50 duncant Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/ioctl.h,v $*/

#include <elan/devinfo.h>
#include <elan/capability.h>

#include <elan4/dma.h>
#include <elan4/neterr.h>
#include <elan4/registers.h>
#include <elan4/intcookie.h>

#define ELAN4IO_CONTROL_PATHNAME	"/dev/elan4/control%d"
#define ELAN4IO_USER_PATHNAME		"/dev/elan4/user%d"
#define ELAN4IO_SDRAM_PATHNAME  	"/dev/elan4/sdram%d"
#define ELAN4IO_MAX_PATHNAMELEN	32

/*
 * NOTE - ioctl values 0->0x1f are defined for 
 *        generic/control usage.
 */

/* Macro to generate 'offset' to mmap "control" device */
#define OFF_TO_BAR(off)		(((off) >> 28) & 0xF)
#define OFF_TO_OFFSET(off)	((off) & 0x0FFFFFFF)
#define GEN_OFF(bar,off)	(((bar) << 28) | ((off) & 0x0FFFFFFF))

/* Definiations for generic ioctls */
#define ELAN4IO_GENERIC_BASE		0x00

typedef struct elan4io_stats_struct
{
    int		       which;
    unsigned long long ptr;					/* always pass pointer as 64 bit */
} ELAN4IO_STATS_STRUCT;

#define ELAN4IO_STATS			_IOR ('e', ELAN4IO_GENERIC_BASE + 0, ELAN4IO_STATS_STRUCT)
#define ELAN4IO_DEVINFO			_IOR ('e', ELAN4IO_GENERIC_BASE + 1, ELAN_DEVINFO)
#define ELAN4IO_POSITION		_IOR ('e', ELAN4IO_GENERIC_BASE + 2, ELAN_POSITION)


/* 
 * Definitions for /dev/elan4/controlX
 */
#define ELAN4IO_CONTROL_BASE		0x20

#define ELAN4IO_GET_POSITION		_IOR ('e', ELAN4IO_CONTROL_BASE + 0, ELAN_POSITION)
#define ELAN4IO_SET_POSITION		_IOW ('e', ELAN4IO_CONTROL_BASE + 1, ELAN_POSITION)
#define ELAN4IO_DEBUG_SNAPSHOT		_IOW ('e', ELAN4IO_CONTROL_BASE + 2, )

typedef struct elan4io_params_mask_struct
{
    unsigned short	p_mask;
    ELAN_PARAMS		p_params;
} ELAN4IO_PARAMS_STRUCT;
#define ELAN4IO_GET_PARAMS		_IOR ('e', ELAN4IO_CONTROL_BASE + 3, ELAN4IO_PARAMS_STRUCT)
#define ELAN4IO_SET_PARAMS		_IOW ('e', ELAN4IO_CONTROL_BASE + 4, ELAN4IO_PARAMS_STRUCT)

/* old versions - implicit p_mask == 3 */
#define ELAN4IO_OLD_GET_PARAMS		_IOR ('e', ELAN4IO_CONTROL_BASE + 3, ELAN_PARAMS)
#define ELAN4IO_OLD_SET_PARAMS		_IOW ('e', ELAN4IO_CONTROL_BASE + 4, ELAN_PARAMS)

/*
 * Definitions for /dev/elan4/userX
 */
#define ELAN4IO_USER_BASE		0x40

#define ELAN4IO_FREE			_IO   ('e', ELAN4IO_USER_BASE + 0)
#define ELAN4IO_ATTACH			_IOWR ('e', ELAN4IO_USER_BASE + 1, ELAN_CAPABILITY)
#define ELAN4IO_DETACH			_IOWR ('e', ELAN4IO_USER_BASE + 2, ELAN_CAPABILITY)
#define ELAN4IO_BLOCK_INPUTTER		_IO   ('e', ELAN4IO_USER_BASE + 3)

typedef struct elan4io_add_p2pvp_struct 
{
    unsigned        vp_process;
    ELAN_CAPABILITY vp_capability;
} ELAN4IO_ADD_P2PVP_STRUCT;

#define ELAN4IO_ADD_P2PVP		_IOW  ('e', ELAN4IO_USER_BASE + 4, ELAN4IO_ADD_P2PVP_STRUCT)

typedef struct elan4io_add_bcastvp_struct
{
    unsigned int	vp_process;
    unsigned int	vp_lowvp;
    unsigned int	vp_highvp;
} ELAN4IO_ADD_BCASTVP_STRUCT;

#define ELAN4IO_ADD_BCASTVP		_IOW  ('e', ELAN4IO_USER_BASE + 5, ELAN4IO_ADD_BCASTVP_STRUCT)

#define ELAN4IO_REMOVEVP		_IO   ('e', ELAN4IO_USER_BASE + 6)

typedef struct elan4io_route_struct
{
    unsigned int	   rt_process;
    unsigned int	   rt_error;
    E4_VirtualProcessEntry rt_route;
} ELAN4IO_ROUTE_STRUCT;

#define ELAN4IO_SET_ROUTE		_IOW  ('e', ELAN4IO_USER_BASE + 7, ELAN4IO_ROUTE_STRUCT)
#define ELAN4IO_RESET_ROUTE		_IOW  ('e', ELAN4IO_USER_BASE + 9, ELAN4IO_ROUTE_STRUCT)
#define ELAN4IO_GET_ROUTE		_IOWR ('e', ELAN4IO_USER_BASE + 8, ELAN4IO_ROUTE_STRUCT)
#define ELAN4IO_CHECK_ROUTE		_IOWR ('e', ELAN4IO_USER_BASE + 10, ELAN4IO_ROUTE_STRUCT)

typedef struct elan4io_alloc_cq_struct
{
    unsigned int cq_size;					/* input: size of queue */
    unsigned int cq_perm;					/* input: requested permissions */
    unsigned int cq_type;					/* input: queue type */
    unsigned int cq_indx;					/* output: queue number */
} ELAN4IO_ALLOCCQ_STRUCT;

#define ELAN4IO_ALLOCCQ			_IOWR ('e', ELAN4IO_USER_BASE + 11, ELAN4IO_ALLOCCQ_STRUCT)
#define ELAN4IO_FREECQ   		_IOWR ('e', ELAN4IO_USER_BASE + 12, unsigned)

#define ELAN4IO_CQ_TYPE_REORDER		1			/* revb reordering command queue */

typedef struct elan4io_perm_struct
{
    E4_Addr            ps_eaddr;
    E4_uint64          ps_len;
    unsigned long      ps_maddr;
    unsigned int       ps_perm;
} ELAN4IO_PERM_STRUCT;

typedef struct elan4io_perm_struct32
{
    E4_Addr            ps_eaddr;
    E4_uint64          ps_len;
    unsigned int       ps_maddr;
    unsigned int       ps_perm;
} ELAN4IO_PERM_STRUCT32;

#define ELAN4IO_SETPERM			_IOWR ('e', ELAN4IO_USER_BASE + 13, ELAN4IO_PERM_STRUCT)
#define ELAN4IO_SETPERM32		_IOWR ('e', ELAN4IO_USER_BASE + 13, ELAN4IO_PERM_STRUCT32)
#define ELAN4IO_CLRPERM			_IOWR ('e', ELAN4IO_USER_BASE + 14, ELAN4IO_PERM_STRUCT)
#define ELAN4IO_CLRPERM32		_IOWR ('e', ELAN4IO_USER_BASE + 14, ELAN4IO_PERM_STRUCT32)

typedef struct elan4io_trapsig_struct
{
    int		ts_signo;
} ELAN4IO_TRAPSIG_STRUCT;
#define ELAN4IO_TRAPSIG			_IOW  ('e', ELAN4IO_USER_BASE + 15, ELAN4IO_TRAPSIG_STRUCT)

typedef struct elan4io_traphandler_struct
{
    unsigned int       th_nticks;				/* number of ticks to sleep for next trap */
    unsigned int       th_proc;					/* elan processor involved */
    unsigned long      th_trapp;				/* space to store trap */
} ELAN4IO_TRAPHANDLER_STRUCT;

typedef struct elan4io_traphandler_struct32
{
    unsigned int       th_nticks;				/* number of ticks to sleep for next trap */
    unsigned int       th_proc;					/* elan processor involved */
    unsigned int       th_trapp;				/* space to store trap */
} ELAN4IO_TRAPHANDLER_STRUCT32;

#define ELAN4IO_TRAPHANDLER		_IOW  ('e', ELAN4IO_USER_BASE + 16, ELAN4IO_TRAPHANDLER_STRUCT)
#define ELAN4IO_TRAPHANDLER32		_IOW  ('e', ELAN4IO_USER_BASE + 16, ELAN4IO_TRAPHANDLER_STRUCT32)

typedef struct elan4io_required_mappings_struct
{
    E4_Addr    rm_upage_addr;					/* elan address of user page */
    E4_Addr    rm_trestart_addr;				/* elan address of tproc restart trampoline */
} ELAN4IO_REQUIRED_MAPPINGS_STRUCT;
#define ELAN4IO_REQUIRED_MAPPINGS	_IOW  ('e', ELAN4IO_USER_BASE + 17, ELAN4IO_REQUIRED_MAPPINGS_STRUCT)

typedef struct elan4io_resume_eproc_trap_struct
{
    E4_Addr             rs_addr;
} ELAN4IO_RESUME_EPROC_TRAP_STRUCT;
#define ELAN4IO_RESUME_EPROC_TRAP	_IOW  ('e', ELAN4IO_USER_BASE + 18, ELAN4IO_RESUME_EPROC_TRAP_STRUCT)

typedef struct elan4io_resume_cproc_trap_struct
{
    unsigned int	rs_indx;
} ELAN4IO_RESUME_CPROC_TRAP_STRUCT;
#define ELAN4IO_RESUME_CPROC_TRAP	_IOW  ('e', ELAN4IO_USER_BASE + 19, ELAN4IO_RESUME_CPROC_TRAP_STRUCT)

typedef struct elan4io_resume_dproc_trap_struct
{
    E4_DMA		rs_desc;
} ELAN4IO_RESUME_DPROC_TRAP_STRUCT;
#define ELAN4IO_RESUME_DPROC_TRAP	_IOW  ('e', ELAN4IO_USER_BASE + 20, ELAN4IO_RESUME_DPROC_TRAP_STRUCT)

typedef struct elan4io_resume_tproc_trap_struct
{
    E4_ThreadRegs	rs_regs;
} ELAN4IO_RESUME_TPROC_TRAP_STRUCT;
#define ELAN4IO_RESUME_TPROC_TRAP	_IOW  ('e', ELAN4IO_USER_BASE + 21, ELAN4IO_RESUME_TPROC_TRAP_STRUCT)

typedef struct elan4io_resume_iproc_trap_struct
{
    unsigned int	rs_channel;
    unsigned int	rs_trans;
    E4_IprocTrapHeader  rs_header;
    E4_IprocTrapData    rs_data;
} ELAN4IO_RESUME_IPROC_TRAP_STRUCT;
#define ELAN4IO_RESUME_IPROC_TRAP	_IOW  ('e', ELAN4IO_USER_BASE + 22, ELAN4IO_RESUME_IPROC_TRAP_STRUCT)

#define ELAN4IO_FLUSH_ICACHE		_IO   ('e', ELAN4IO_USER_BASE + 23)
#define ELAN4IO_STOP_CTXT		_IO   ('e', ELAN4IO_USER_BASE + 24)

#define ELAN4IO_ALLOC_INTCOOKIE		_IOW  ('e', ELAN4IO_USER_BASE + 25, ELAN4_INTCOOKIE)
#define ELAN4IO_FREE_INTCOOKIE		_IOW  ('e', ELAN4IO_USER_BASE + 26, ELAN4_INTCOOKIE)
#define ELAN4IO_ARM_INTCOOKIE		_IOW  ('e', ELAN4IO_USER_BASE + 27, ELAN4_INTCOOKIE)
#define ELAN4IO_WAIT_INTCOOKIE		_IOW  ('e', ELAN4IO_USER_BASE + 28, ELAN4_INTCOOKIE)

typedef struct elan4io_alloc_trap_queues_struct
{
    unsigned int	tq_ndproc_traps;
    unsigned int	tq_neproc_traps;
    unsigned int	tq_ntproc_traps;
    unsigned int	tq_nthreads;
    unsigned int	tq_ndmas;
} ELAN4IO_ALLOC_TRAP_QUEUES_STRUCT;
#define ELAN4IO_ALLOC_TRAP_QUEUES	_IOW  ('e', ELAN4IO_USER_BASE + 29, ELAN4IO_ALLOC_TRAP_QUEUES_STRUCT)

typedef struct elan4io_neterr_msg_struct
{
    unsigned int	nm_vp;
    unsigned int	nm_nctx;
    unsigned int	nm_retries;
    unsigned int        nm_pad;
    ELAN4_NETERR_MSG    nm_msg;
} ELAN4IO_NETERR_MSG_STRUCT;
#define ELAN4IO_NETERR_MSG		_IOW ('e', ELAN4IO_USER_BASE + 30, ELAN4IO_NETERR_MSG_STRUCT)

typedef struct elan4io_neterr_timer_struct 
{
    unsigned int	nt_usecs;
} ELAN4IO_NETERR_TIMER_STUCT;

#define ELAN4IO_NETERR_TIMER		_IO  ('e', ELAN4IO_USER_BASE + 31)

typedef struct elan4io_neterr_fixup_struct
{
    E4_uint64		nf_cookie;
    unsigned int	nf_waitforeop;
    unsigned int 	nf_sten;
    unsigned int	nf_vp;
    unsigned int	nf_pad;
} ELAN4IO_NETERR_FIXUP_STRUCT;

#define ELAN4IO_NETERR_FIXUP		_IOW ('e', ELAN4IO_USER_BASE + 32, ELAN4IO_NETERR_FIXUP_STRUCT)

typedef struct elan4io_firecap_struct 
{
    ELAN_CAPABILITY     fc_capability;
    ELAN4_INTCOOKIE     fc_cookie;
} ELAN4IO_FIRECAP_STRUCT;

#define ELAN4IO_FIRE_INTCOOKIE		_IOW  ('e', ELAN4IO_USER_BASE + 33, ELAN4IO_FIRECAP_STRUCT)

#define ELAN4IO_ALLOC_INTCOOKIE_TABLE	_IOW  ('e', ELAN4IO_USER_BASE + 34, ELAN_CAPABILITY)
#define ELAN4IO_FREE_INTCOOKIE_TABLE	_IO   ('e', ELAN4IO_USER_BASE + 35)

typedef struct elan4io_translation
{
    E4_Addr		tr_addr;
    unsigned long	tr_len;
    unsigned int	tr_access;
} ELAN4IO_TRANSLATION_STRUCT;

#define ELAN4IO_LOAD_TRANSLATION	_IOW  ('e', ELAN4IO_USER_BASE + 36, ELAN4IO_TRANSLATION_STRUCT)
#define ELAN4IO_UNLOAD_TRANSLATION	_IOW  ('e', ELAN4IO_USER_BASE + 37, ELAN4IO_TRANSLATION_STRUCT)

typedef struct elan4io_dumpcq_struct32
{
    E4_uint64 cq_space;      /* output: sdram addr of q, used to decode ptrs */
    E4_uint32 cq_size;       /* output: The real size of the command queue */
    E4_uint32 bufsize;       /* input: The size of the buffer to dump to */
    E4_uint32 cq_indx;       /* input: index of cq to dump */
    unsigned int buffer;     /* input: user address of rgs->buffer to dump to */
} ELAN4IO_DUMPCQ_STRUCT32;

typedef struct elan4io_dumpcq_struct
{
    E4_uint64 cq_space;      /* output: sdram addr of q, used to decode ptrs */
    E4_uint32 cq_size;       /* output: The real size of the command queue */
    E4_uint32 bufsize;       /* input: The size of the buffer to dump to */
    E4_uint32 cq_indx;       /* input: index of cq to dump */
    unsigned long buffer;    /* input: user address of rgs->buffer to dump to */
} ELAN4IO_DUMPCQ_STRUCT;

#define ELAN4IO_DUMPCQ			_IOWR ('e', ELAN4IO_USER_BASE + 38, ELAN4IO_DUMPCQ_STRUCT)
#define ELAN4IO_DUMPCQ32			_IOWR ('e', ELAN4IO_USER_BASE + 38, ELAN4IO_DUMPCQ_STRUCT32)

/* mmap offsets - - we define the file offset space as follows:
 *
 * page 0 - 4095 - command queues
 * page 4096	 - device user registers
 * page 4097	 - flag page/user stats
 * page 4098	 - device stats
 * page 4099     - tproc trampoline
 */

#define ELAN4_OFF_COMMAND_QUEUES	0
#define ELAN4_OFF_USER_REGS		4096
#define ELAN4_OFF_USER_PAGE		4097
#define ELAN4_OFF_DEVICE_STATS		4098
#define ELAN4_OFF_TPROC_TRAMPOLINE	4099


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_IOCTL_H */
