/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: mmu.h,v 1.14 2005/04/21 11:12:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/mmu.h,v $*/


#ifndef __ELAN4_MMU_H
#define __ELAN4_MMU_H

#include <linux/pci.h>

typedef union elan4_pte_page
{
    struct {
	struct page	*page;
	physaddr_t	 dma_addr;
    } _page;
#define pg_page		_page.page
#define pg_dma_addr	_page.dma_addr

} ELAN4_PTE_PAGE;

typedef struct elan4_hash_entry
{
    struct elan4_hash_entry	*he_next;
    struct elan4_hash_entry	*he_prev;

    sdramaddr_t			 he_entry;
    
    struct elan4_hash_entry	*he_chain[2];
    E4_uint64			 he_tag[2];
    E4_uint32			 he_pte[2];

    ELAN4_PTE_PAGE		 he_pg[2][4];
} ELAN4_HASH_ENTRY;

#define ELAN4_HENT_CHUNKS	16		/* SDRAM_MIN_BLOCK_SIZE/sizeof (E4_HashTableEntry) */

typedef struct elan4_hash_chunk
{
    struct list_head            hc_link;
    ELAN4_HASH_ENTRY		hc_hents[ELAN4_HENT_CHUNKS];
} ELAN4_HASH_CHUNK;

typedef struct elan4_hash_cache
{
    E4_Addr           hc_start;
    E4_Addr           hc_end;
    int		      hc_tbl;

    ELAN4_HASH_ENTRY *hc_hes[1];
} ELAN4_HASH_CACHE;

/* 
 * he_pte is really 4 bytes of pte "type" one for each pte
 * entry - however we declare it as an "int" so we can
 * easily determine that all 4 entries are invalid 
 */
#define HE_SET_PTE(he,tagidx,pteidx,val)	(((E4_uint8 *) &(he->he_pte[tagidx]))[pteidx] = (val))
#define HE_GET_PTE(he,tagidx,pteidx)		(((E4_uint8 *) &(he->he_pte[tagidx]))[pteidx])

#define HE_TYPE_INVALID				0
#define HE_TYPE_SDRAM				1
#define HE_TYPE_COMMAND				2
#define HE_TYPE_REGS				3
#define HE_TYPE_PAGE				4
#define HE_TYPE_OTHER				5
#define HE_TYPE_RESERVED			6

/*
 * he_tag has the following form :
 *     [63:27]	tag
 *     [20:17]  pte valid
 *     [16]     locked
 *     [15]     copy
 *     [14]     valid
 *     [13:0]	context
 */

#define HE_TAG_VALID		(1 << 14)
#define HE_TAG_COPY		(1 << 15)
#define HE_TAG_LOCKED		(1 << 16)

#define INVALID_CONTEXT		0

extern u_char elan4_permtable[];
#define ELAN4_INCOMPAT_ACCESS(perm,access) ((elan4_permtable[(perm)] & (1 << (access))) == 0)
extern u_char elan4_permreadonly[];
#define ELAN4_PERM_READONLY(perm)	  (elan4_permreadonly[(perm)])

extern int elan4_debug_mmu;

extern int elan4_mmuhash_chain_reduction;
extern int elan4_mmuhash_chain_end_reduce;
extern int elan4_mmuhash_chain_middle_reduce;
extern int elan4_mmuhash_chain_middle_fail;
extern int elan4_mmuhash_shuffle_attempts;
extern int elan4_mmuhash_shuffle_done;

#ifdef DEBUG_PRINTF
#  define MPRINTF(ctxt,lvl,args...)	(elan4_debug_mmu > (lvl) ? elan4_debugf(ctxt,DBG_MMU, ##args) : (void)0)
#else
#  define MPRINTF(ctxt,lvl,args...)	((void) 0)
#endif

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_MMU_H */
