/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_PCI_H 
#define __ELAN4_PCI_H

#ident "$Id: pci.h,v 1.32 2003/09/04 12:39:17 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/pci.h,v $*/

/* Elan has 2 64 bit bars */
#define ELAN4_BAR_SDRAM			0
#define ELAN4_BAR_REGISTERS		2

#define PCI_VENDOR_ID_QUADRICS		0x14fc
#define PCI_DEVICE_ID_ELAN3		0x0000
#define   PCI_REVISION_ID_ELAN3_REVA	0x0000
#define   PCI_REVISION_ID_ELAN3_REVB	0x0001
#define PCI_DEVICE_ID_ELAN4		0x0001
#define   PCI_REVISION_ID_ELAN4_REVA	0x0000
#define   PCI_REVISION_ID_ELAN4_REVB	0x0001

/* support standard pseudo bars */
#define ELAN4_PSEUDO_BAR_ROM		8

/* Elan PCI control
 configuration space register. ElanControlRegister */
#define PCI_ELAN_PARITY_ADDR_LO		0x40
#define PCI_ELAN_PARITY_ADDR_HI		0x44
#define PCI_ELAN_PARITY_TYPE		0x48
#define PCI_ELAN_CONTROL		0x4c
#define PCI_ELAN_PLL_CONTROL		0x50
#define PCI_ELAN_SPLIT_MESSAGE_ATTR	0x54
#define PCI_ELAN_SPLIT_MESSAGE_VALUE	0x54
#define PCI_ELAN_RAMBIST_FAILED		0x54
#define PCI_ELAN_TOPPHYSADDR(i)		(0x58 + ((i)<<1))

/*
 * [31]	   PciM66EN		This is set it the bus is running in PCI2.3 - 66MHz mode.
 * [30:28] InitPattern		This gives the PCI-X startup mode. See "Pci intialisation patterns" below.
 * [27]	   notBusIs64Bits	If set the bus is running 32 bits wide. If Clear it is a 64 bit bus.
 * [26:24] RamBistCntl		Used to control the Elan4 RAM BIST. Not acitive it zero.
 * [23]	   RamBistFinished	Only used when performing the RAM BIST test.
 * [22]	   SelectSplitMessAttr  See ECTRL_SELECT_SPLIT_MESS_ATTR below.
 * [21]	   ReceivedSplitCompError See ECTRL_REC_SPLIT_COMP_MESSAGE below
 * [20:16] WriteHighPriTime	Used with ReadHighPriTime to control the ratio of PCI master write to PCI master
 *				read bandwidth under heavy load. The high the value of WriteHighPriTime the longer
 *				the PCI write bursts will be allowed without interruption from a read transfer.
 * [15]    DisableCouplingTest	This is only used as part of the RAM BIST test. It effects the testing of the main
 *				cache tag RAMS.
 * [14:13] Not used		Will read as zero.
 * [12:8]  ReadHighPriTime	Used with WriteHighPriTime to control the ratio of PCI master write to PCI master
 *				read bandwidth under heavy load. The high the value of ReadHighPriTime the longer
 *				the PCI read bursts will be allowed without interruption from a write transfer.
 * [7] EnableLatencyCountReset  This bit effect the behaviour of disconnects due to the removal of GNT# after the latency
 *				counter has expired. If set it will allow the latency counter to be reset each time the
 *				GNT# is reasserted. If asserted it should provided improved bandwidth on the PCI bus
 *				without increasing the maximum latency another device would have for access to the bus.
 *				It will increase the average latency of other devices.
 * [6] ExtraMasterAddrBits	This bit used to control the physical PCI addresses generated by the MMU.
 * [5] ReducedPciDecode		If set the PCI local memory BAR will decode 256Mbytes of PCI address space. If clear it
 *				will decode 2Gbyte of PCI address space.
 * [4] ConfigInEBusRom		If set the constant values of the Elan4 PCI configuration space will be taken from the
 *				EEPROM. If clear the internal values will be used.
 * [3] EnableRd2_2Bursts	This bit only effects the behaviour of burst reads when the PCI bus is operating in
 *				PCI-2.2 mode. It allows adjacent reads to be merged into longer bursts for higher
 *				performance.
 * [2] SoftIntReset		If set this bit will cause the Elan4 to reset itself with the exception of the PCI
 *				configuation space. All internal state machines will be put into the reset state.
 * [1] EnableWrBursts		This bit allows much longer PCI-X write bursts. If set it will stop the Elan4 from
 *				being completely PCI-X compliant as the Elan4 may request a long PCI-X write burst that
 *				it does not complete. However it should significantly increase the maximum PCI-X write
 *				bandwidth and is unlikely to cause problems with many PCI-X bridge chips.
 * [0] InvertMSIPriority	This bit effect the way MSI interrupts are generated. It provides flexiblity to generate
 *				the MSI interrupts in a different way to allow for different implimentations of MSI
 *				logic and still give the correct priority of Elan4 interrupts.
 *
 *	{PciM66EN, InitPattern, notBusIs64Bits, RamBistCntl, RamBistFinished,
 *	 SelectSplitMessAttr, ReceivedSplitCompError, WriteHighPriTime,
 *	 DisableCouplingTest, 2'h0, ReadHighPriTime,
 *	 EnableLatencyCountReset, ExtraMasterAddrBits, ReducedPciDecode, ConfigInEBusRom,
 *	 EnableRd2_2Bursts, SoftIntReset, EnableWrBursts, InvertMSIPriority}
 */

#define ECTRL_INVERT_MSI_PRIO		(1 << 0)
#define ECTRL_ENABLE_WRITEBURSTS	(1 << 1)
#define ECTRL_SOFTWARE_INTERNAL_RESET	(1 << 2)
#define ECTRL_ENABLE_2_2READBURSTS	(1 << 3)
#define ECTRL_CONFIG_IN_EBUS_ROM	(1 << 4)
#define ECTRL_28_NOT_30_BIT_LOCAL_BAR	(1 << 5)
#define ECTRL_ExtraMasterAddrBits	(1 << 6)
#define ECTRL_ENABLE_LATENCY_RESET      (1 << 7)
#define ECTRL_DISABLE_COUPLING_TEST	(1 << 15)

/*
 * Ratio of the following two registers set the relative bandwidth given to intputer data
 * versus other PCI pci traffic when scheduling new PCI master accesses.
 */
#define ECTRL_OTHER_HIGH_PRI_TIME_SHIFT	(8)	/* Sets top 4 bits of 8 bit counter */
#define ECTRL_OTHER_HIGH_PRI_TIME_MASK	(0x1f)


#define ECTRL_IPROC_HIGH_PRI_TIME_SHIFT	(16)	/* Sets top 4 bits of 8 bit counter */
#define ECTRL_IPROC_HIGH_PRI_TIME_MASK	(0x1f)

/*
 * This is set if a split completion message is received.
 * This will cause a PCI error interrupt.
 * This error is cleared by writting a 1 to this bit.
 */
#define ECTRL_REC_SPLIT_COMP_MESSAGE	(1 << 21)
/*
 * This bit is used to select reading of either the Split message attribute value when
 * set or the split completion message data value from 0x54 in the config space
 * if the ECTRL_REC_SPLIT_COMP_MESSAGE bit is set. 0x54 returns the the BistFailed flags
 * if any of the BIST control bits are set (bits 26 to 24)
 */
#define ECTRL_SELECT_SPLIT_MESS_ATTR	(1 << 22)

// Internal RAM bist control bits.
// Three bits of state control the RAM BIST (Built in self test).
//
// These bits must not be set unless the ECTRL_SOFTWARE_INTERNAL_RESET bit has also been set!
//
// For a normal fast ram test assert ECTRL_BIST_FAST_TEST. 
// For a data retention test first write ECTRL_START_RETENTION_TEST then wait the retention period of
// at least 1ms and preferably much longer then write ECTRL_CONTINUE_RETENTION_TEST then wait
// again and finallly write ECTRL_FINISH_RETENTION_TEST.
// 
// The read only bit ECTRL_BIST_FINISHED_TEST can be polled to check that the test has compleated.
#define ECTRL_BIST_CTRL_SHIFT		(24)
#define ECTRL_BIST_CTRL_MASK		(7 << 24)

#define ECTRL_BIST_FAST_TEST		((7 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)	// old scheme
#define ECTRL_START_RETENTION_TEST	((1 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)
#define ECTRL_CONTINUE_RETENTION_TEST	((3 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)
#define ECTRL_FINISH_RETENTION_TEST	((7 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)

#define ECTRL_BIST_KICK_OFF		((1 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)	// new scheme
#define ECTRL_BIST_MOVE_ON_ODD		((3 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)
#define ECTRL_BIST_MOVE_ON_EVEN		((5 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)
#define ECTRL_BIST_SCREAM_THROUGH	((7 << 24) | ECTRL_SOFTWARE_INTERNAL_RESET)

#define ECTRL_CLEAR_BIST_TEST		(0 << 24)
#define ECTRL_BIST_FINISHED_TEST	(1 << 23)

// Read only current PCI bus type.
#define ECTRL_RUNNING_32BIT_MODE	(1 << 27)
#define ECTRL_INITIALISATION_MODE	(7 << 28)
#define ECTRL_RUNNING_M66EN_MODE	(1 << 31)

#define ECTRL_INIT_PATTERN_SHIFT	(28)
#define ECTRL_INIT_PATTERN_MASK		(0x7)

// Pci intialisation patterns
#define Pci2_2				(0 << 28)
#define PciX50To66MHz			(1 << 28)
#define PciX66to100MHz			(2 << 28)
#define PciX100to133MHz			(3 << 28)
#define PciXReserved1			(4 << 28)
#define PciXReserved2			(5 << 28)
#define PciXReserved3			(6 << 28)
#define PciXReserved4			(7 << 28)

/* Elan PCI pll and pad control configuration space register. ElanPllControlReg */
// This overrides the default PCI pll control settings.
#define PciPll_FeedForwardISel0		(1 << 0)	// Lsi name Z0
#define PciPll_FeedForwardISel1		(1 << 1)	// Lsi name Z1
#define PciPll_ChargePumpISel0		(1 << 2)	// Lsi name P0
#define PciPll_ChargePumpISel1		(1 << 3)	// Lsi name P1
#define PciPll_EnableAutoReset		(1 << 4)	// Lsi name ENARST
#define PciPll_RSEL200500		(1 << 5)	// Lsi name Range Select, 0: 100 - 250MHz, 1: 200 - 500MHz
#define PciPll_DivideFeedback		(1 << 6)	// Just used for test - This divides the shortcut feedback to the PCI PLL so that it can lock to the tester clock.
#define PciPll_CutFeedback		(1 << 7)	// Just used for test - This disables the shortcut feedback.

// This overrides the default PCI BZ controler settings.
#define PciBZ_UPDI			(0xf << 8)
#define PciBZ_WAIT_INT			(0xf << 12)

// This overrides the default Sys and SDRam pll control settings.
#define SysPll_FeedForwardISel0		(1 << 16)	// Lsi name P0     
#define SysPll_FeedForwardISel1		(1 << 17)	// Lsi name P1     
#define SysPll_ChargePumpISel0		(1 << 18)	// Lsi name Z0    
#define SysPll_ChargePumpISel1		(1 << 19)	// Lsi name Z1    
#define SysPll_EnableAutoReset		(1 << 20)	// Lsi name ENARST
#define SysPll_DivPhaseCompInBy2	(1 << 21)	// Lsi name NODIV (Should be DIV)
#define SysPll_PllTestClkSel		(1 << 22)	// If asserted the master clock source is not taken from the pll.

#define Pll_ForceEBusADTristate	        (1 << 23)	// Required to enable the testing of EnableAutoReset. Enables use of EBusAD[7] (rev A)
#define Pll_LinkErrDirectToSDA	        (1 << 23)	// Access to link error flag for triggering (rev B)


#define ECTRL_SYS_CLOCK_RATIO_SHIFT	(24)
// Config: with 800MHz		Speeds are 266 200 160 133.
//         0 = 133/133 (1:1)	6:6	1
//	   1 = 160/133 (6:5)	5:6	1.2
//         2 = 200/133 (3:2)	4:6	1.5
//	   3 = 266/133 (2:1)	3:6	2
//	   4 = 200/200 (1:1)	4:4	1
//	   5 = 266/200 (4:3)	3:4	1.33

// Config: with 600MHz		Speeds are 200 150 120 100
//         0 = 100/100 (1:1)	6:6	1
//	   1 = 120/100 (6:5)	5:6	1.2
//         2 = 150/100 (3:2)	4:6	1.5
//	   3 = 200/100 (2:1)	3:6	2
//	   4 = 150/150 (1:1)	4:4	1
//	   5 = 200/150 (4:3)	3:4	1.33

#define ECTRL_SYS_CLOCK_RATIO_SHIFT	(24)
#define ECTRL_SYS_CLOCK_RATIO_1_1Slow	(0 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_6_5	(1 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_3_2	(2 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_2_1	(3 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_1_1Fast	(4 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_4_3	(5 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_MAX_NORMAL	(6)				 	/* used to generate a valid random value */
#define GET_RANDOM_CLOCK_RATIO		(Random(ECTRL_SYS_CLOCK_MAX_NORMAL) << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_PLL_TEST	(6 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_TEST	(7 << ECTRL_SYS_CLOCK_RATIO_SHIFT)
#define ECTRL_SYS_CLOCK_RATIO_MASK	(7 << ECTRL_SYS_CLOCK_RATIO_SHIFT)

#endif /* __ELAN4_PCI_H */
