/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN4_REGISTERS_H
#define _ELAN4_REGISTERS_H

#ident "$Id: registers.h,v 1.120 2005/03/03 16:28:50 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/registers.h,v $*/

/*
 * Header file for internal slave mapping of the ELAN4 registers
 */

#define E4_CACHELINE_SIZE	(64)
#define E4_STACK_ALIGN		(64)

#ifndef _ASM

#include <elan4/types.h>
#include <elan4/dma.h>
#include <elan4/userregs.h>

typedef volatile struct _E4_CacheSets
{
   E4_uint64	Set0[1024];	/* 8k bytes per set */
   E4_uint64	Set1[1024];	/* 8k bytes per set */
   E4_uint64	Set2[1024];	/* 8k bytes per set */
   E4_uint64	Set3[1024];	/* 8k bytes per set */
} E4_CacheSets;

typedef union e4_cache_tag
{
   struct {
       E4_uint32 pad0;			/* Undefined value when read */
#if (BYTE_ORDER == LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__)
       E4_uint32 :10;						/* 0-9   - reserved */
       E4_uint32 LineError:1;					/* 10    - line error */
       E4_uint32 Modified:1;					/* 11    - modified */
       E4_uint32 FillPending:1;					/* 12    - fill pending */
       E4_uint32 AddrTag30to13:18;				/* 30-13 - tag */
       E4_uint32 :1;						/* 31    -  */
#else
       E4_uint32 :1;						/* 31    -  */
       E4_uint32 AddrTag30to13:18;				/* 30-13 - tag */
       E4_uint32 FillPending:1;					/* 12    - fill pending */
       E4_uint32 Modified:1;					/* 11    - modified */
       E4_uint32 LineError:1;					/* 10    - line error */
       E4_uint32 :10;						/* 0-9   - reserved */
#endif
   } s;
   E4_uint64	Value;
} E4_CacheTag;

typedef volatile struct _E4_CacheTags
{
   E4_CacheTag Tags[4][128];	/* 8k bytes per set, 64 byte cache line */
} E4_CacheTags;

#define E4_NumCacheSets		4
#define E4_NumCacheLines	128
#define E4_CacheLineSize	64
#define E4_CacheSize		(E4_NumCacheSets * E4_NumCacheLines * E4_CacheLineSize)
#define E4_CacheSetSize 	(E4_NumCacheLines * E4_CacheLineSize)

/*
 * Run Queue pointers 
 *
 * [62:35]	FrontPointer[30:3]
 * [33:32]	Size Value
 * [30:3]	BackPointer[30:3]
 */
#define E4_QueuePtrMask		(0x7ffffff8ULL)
#define E4_QueueSizeMask	3
#define E4_QueueEntrySize       sizeof (E4_uint64)

#define E4_Queue8KBytes		0
#define E4_Queue64KBytes	1
#define E4_Queue512KBytes	2
#define E4_Queue4MBytes		3

#define E4_QueueFrontValue(val,size)	((val) | (size))
#define E4_QueueValue(queue,size) 	(((E4_uint64) E4_QueueFrontValue(queue,size)) << 32 | ((E4_uint64) (queue)))

#define E4_QueueFrontPointer(val)	/* extract queue front pointer from register */\
	(((val) >> 32) & E4_QueuePtrMask)
#define E4_QueueBackPointer(val) 	/* extract queue back pointer from register */ \
	((val) & E4_QueuePtrMask)
#define E4_QueueSizeValue(val)		/* extract queue size value from register */ \
	(((val) >> 32) & E4_QueueSizeMask)
#define E4_QueueSize(value) 		/* queue size in bytes from size value */ \
	(1 << (((value)*3) + 13))
#define E4_QueueOffsetMask(fptr)\
        ((8192 << (((fptr) & E4_QueueSizeMask) << 3)) - 1)
#define E4_QueueOffset(fptr)\
        ((fptr) & E4_QueueOffsetMask(fptr))
#define E4_QueueFrontPointerInc(fptr)   \
        ( ((fptr) & ~E4_QueueOffsetMask(fptr)) | ((E4_QueueOffset(fptr) + 8) & E4_QueueOffsetMask(fptr)) )

typedef union _E4_QueuePtr
{
   E4_uint64	Value;
   struct {
       E4_uint32 Back;
       E4_uint32 Front;
   } s;
} E4_QueuePtr;

/*
 * DMA processor status register.
 *
 * [48]		FirstSendTrans 		Set for the first packet of a dma.
 * [47:46]	TimeSliceCount 		Time left to timeslice.
 * [45]		DmaLastPacket 		Set for the last packet of a dma.
 * [44]		CurrPrefetchDma 	Dma descriptor the prefetcher is valid for.
 * [43:39]	PrefetcherState 	Dma prefetcher's state machines value.
 * [38:33]	PacketAssemblyState 	Packet assembler's state machines value.
 * [32:31]	PrefetcherWakeupFnt 	Dma prefetcher's wakeup function.
 * [30:28]	PacketAssWakeupFnt 	Packet assembler's wakeup function.
 * [27]		AckBufferValid 		Packet ack is valid.
 * [26]		PrefetchedDataProblem 	Had either a data read fault or data error. Valid if AckBufferValid.
 * [25]		PrefetcherHalting 	Prefetch data about to stop for halt. Valid if AckBufferValid.
 * [24]		PacketTimeout 		Packet timeout. Sent an EopError. Valid if AckBufferValid set.
 * [23:22]	PacketAckValue 		Packet ack type. Valid if AckBufferValid set.
 * [21:20]	FaultUnitNo 		Set if the dma prefetcher has faulted.
 * [19:17]	TrapType 		Packet assembler's trap type.
 * [16]		PrefetcherFault 	Set if the dma prefetcher has faulted for this DMA unit.
 * [15]		Remote	 		The Dma had been issued remotly
 * [14]		Priority 		Running at high priority.
 * [13:0]	Context 		procs current context.
 */

#define DPROC_FirstSendTrans(s)		((unsigned)((s) >> 48) & 1)
#define DPROC_TimeSliceCount(s)		((unsigned)(((s) >> 46) & 3)
#define DPROC_DmaLastPacket(s)		((unsigned)((s) >> 45) & 1)
#define DPROC_CurrPrefetchDma(s)	((unsigned)((s) >> 44) & 1)
#define DPROC_PrefetcerState(s)		((unsigned)((s) >> 39) & 0x1f)
#define DPROC_PacketAssemblerState(s)	((unsigned)((s) >> 33) & 0x1f)
#define DPROC_PrefetcherWakeupFn(s)	((unsigned)((s) >> 31) & 3)
#define DPROC_PacketAssemblerWakeupFn(s)((unsigned)((s) >> 28) & 3)
#define DPROC_AckBufferValid(s)		((unsigned)((s) >> 27) & 1)
#define DPROC_PrefetcherDataProblem(s)	((unsigned)((s) >> 26) & 1)
#define DPROC_PrefetcherHalting(s)	((unsigned)((s) >> 25) & 1)
#define DPROC_PacketTimeout(s)		((unsigned)((s) >> 24) & 1)
#define DPROC_PacketAckValue(s)		((unsigned)((s) >> 22) & 3)
#define DPROC_FaultUnitNo(s)		((unsigned)((s) >> 20) & 3)
#define DPROC_TrapType(s)		((unsigned)((s) >> 17) & 7)
#define DPROC_PrefetcherFault(s)	((unsigned)((s) >> 16) & 1)
#define DPROC_Remote(s)			((unsigned)((s) >> 15) & 1)
#define DPROC_Priority(s)		((unsigned)((s) >> 14) & 1)
#define DPROC_Context(s)		((unsigned)(s) & 0x3fff)

/*
 * Command processor status register.
 *
 * [26:21]	CPState		procs current state.
 * [20]		WakeupFnt	procs wakeup function.
 * [19:16]	TrapValue	procs trap value.
 * [15]		Remote	 	Issued remotely.
 * [14]		Priority	Running at high priority.
 * [13:0]	Context		procs current context.
 */

#define CPROC_TrapType(s)		((unsigned)((s) >> 16) & 0xf)
#define CPROC_Remote(s)			((unsigned)((s) >> 15) & 0x1)
#define CPROC_Priority(s)		((unsigned)((s) >> 14) & 0x1)
#define CPROC_Context(s)		((unsigned)(s) & 0x3fff)

/*
 * Event processor status register.
 *
 * [34:30]	CPState		event procs current state.
 * [29:28]	WakeupFnt	event procs wakeup function.
 * [27:20]	EventCopySize	This is the number of DWords to still be copied on a copy dword event.
 * [19]		EProcPort1Fault	CUN_EventProc1 has taken a translation fault.
 * [18]		EProcPort0Fault	CUN_EventProc0 has taken a translation fault.
 * [17:16]	TrapValue	event proc's trap value.
 * [15]		Remote	 	Issued remotely.
 * [14]		Priority	Running at high priority.
 * [13:0]	Context		procs current context.
 */

#define EPROC_CPState(s)		((unsigned)((s) >> 30) & 0x1f)
#define EPROC_WakeupFunction(s)		((unsigned)((s) >> 28) & 3)
#define EPROC_CopySize(s)		((unsigned)((s) >> 20) & 0xFF)
#define EPROC_Port1Fault(s)		((unsigned)((s) >> 19) & 1)
#define EPROC_Port0Fault(s)		((unsigned)((s) >> 18) & 1)
#define EPROC_TrapType(s)		((unsigned)((s) >> 16) & 3)
#define EPROC_Remote(s)			((unsigned)((s) >> 15) & 1)
#define EPROC_Priority(s)		((unsigned)((s) >> 14) & 1)
#define EPROC_Context(s)		((unsigned)(s) & 0x3fff)

/*
 * Thread processor status register.
 *
 * [39:24]	MemPortBusy		16 bits of port busy flags for all FFU memory ports.
 * [23:21]	Reads as zero
 * [20:18]	TQState			State vector for thread queuing proc.
 * [17]		HighRunQueueFull	High priority run queue is full
 * [16]		LowRunQueueFull		Low priority run queue is full
 * [15]		ReadyHigh		More runable threads at high priority
 * [14]		ReadyLow		More runable threads at low priority
 * [13:0]	Context			procs current context.
 */
#define TPROC_HighRunQueueFull(s)	((unsigned)((s) >> 17) & 1)
#define TPROC_LowRunQueueFull(s)	((unsigned)((s) >> 16) & 1)
#define TPROC_ReadyHigh(s)		((unsigned)((s) >> 15) & 1)
#define TPROC_ReadyLow(s)		((unsigned)((s) >> 14) & 1)
#define TPROC_Context(s)        	((unsigned)((s) & 0x3fff))

/*
 * Input processor status register
 *
 * [55]		Last Trans (~EOP)
 * [54]		First Trans (~EOP)
 * [53]		Channel (~EOP) 
 * [52]		Bad Length (~EOP)
 * [51:50]	Trans CRC Status (~EOP)
 * [49:48]	EOP type
 * [47]		EOP trap
 * [46]		Trapping priority
 * [45]		Trapping Channel
 * [44:43]	Bad ack sent
 * [42:41]	Good ack sent
 * [40]		Queueing Packet (~EOP)
 * [39:36]	Channel trapped bits
 * [35:32]	IProc Trap Value
 * [31:16]	Network Context (~EOP)
 * [15:0]	Transaction Type (~EOP)
 */
#define IPROC_LastTrans(s)		((unsigned)((s) >> 55) & 0x1)
#define IPROC_FirstTrans(s)		((unsigned)((s) >> 54) & 0x1)
#define IPROC_Channel(s)		((unsigned)((s) >> 53) & 0x1)
#define IPROC_BadLength(s)		((unsigned)((s) >> 52) & 0x1)
#define IPROC_TransCRCStatus(s)		((unsigned)((s) >> 50) & 0x3)
#define IPROC_EOPType(s)		((unsigned)((s) >> 48) & 0x3)
#define IPROC_EOPTrap(s)		((unsigned)((s) >> 47) & 0x1)
#define IPROC_InputterPri(s)		((unsigned)((s) >> 46) & 0x1)
#define IPROC_InputterChan(s)		((unsigned)((s) >> 45) & 0x1)
#define IPROC_BadAckSent(s)		((unsigned)((s) >> 43) & 0x3)
#define IPROC_GoodAckSent(s)		((unsigned)((s) >> 41) & 0x3)
#define IPROC_QueueingPacket(s)		((unsigned)((s) >> 40) & 0x1)
#define IPROC_ChannelTrapped(s)		((unsigned)((s) >> 36) & 0xF)
#define IPROC_TrapValue(s)		((unsigned)((s) >> 32) & 0xF)
#define IPROC_NetworkContext(s)		((unsigned)((s) >> 16) & 0xFFFF)
#define IPROC_TransactionType(s)	((unsigned)(s) & 0xFFFF)

/* values for IPROC_TransCRCStatus */
#define CRC_STATUS_GOOD    (0)
#define CRC_STATUS_DISCARD (1)
#define CRC_STATUS_ERROR   (2)
#define CRC_STATUS_BAD     (3)

/* values for IPROC_EOPType */
#define EOP_GOOD	   (1)
#define EOP_BADACK	   (2)
#define EOP_ERROR_RESET	   (3)

/*
 * Interrupt register bits
 *
 * There are up to four sources of interrupt for the MSI port.
 * The Elan will request 4 ports but may only get either 2 or 1 port. The Interrupts are assigned
 * as shown below:
 * No Of MSI ints	Low Prioity							High Prioity
 * 	4		Event Ints	OtherInts		Inputer Ints		Hard Error ints.
 *		   i.e. 		Dproc, Tproc, Sten.	HighPri and LowPri	Link errs, ECC errs,
 *
 *	2		Event Ints	All other interrupts.
 *	1		All together.
 * 
 * It is not safe to change the number of sources of interrupt while there may be outstanding,
 * unserviced interrupts pending.
 * There two forms of encoding. This has been provided in case an MSI implimentation assumes either
 * a high value to have a high priority or a low value to have a high priority. This is controled
 * by a bit in the Elan Pci Control register.
 */
#define INT_LinkPortKeyFail		(1<<18)
#define INT_PciMemErr			(1<<17)
#define INT_SDRamInt			(1<<16)
#define INT_LinkError			(1<<15)
#define INT_IProcCh1HighPri		(1<<14)
#define INT_IProcCh0HighPri		(1<<13)
#define INT_IProcCh1LowPri		(1<<12)
#define INT_IProcCh0LowPri		(1<<11)
#define INT_DiscardingHighPri		(1<<10)
#define INT_DiscardingLowPri		(1<<9)
#define INT_CProcHalted			(1<<8)
#define INT_TProcHalted			(1<<7)
#define INT_DProcHalted			(1<<6)
#define INT_EProc			(1<<5)
#define INT_TProc			(1<<4)
#define INT_CProc			(1<<3)
#define INT_Dma1Proc			(1<<2)
#define INT_Dma0Proc			(1<<1)
#define INT_MainInterrupt		(1<<0)

#define INT_Units		(INT_EProc | INT_TProc | INT_CProc | INT_Dma1Proc | INT_Dma0Proc)
#define INT_Inputters		(INT_IProcCh1HighPri | INT_IProcCh0HighPri | INT_IProcCh1LowPri | INT_IProcCh0LowPri)
#define INT_Discarding  	(INT_DiscardingHighPri | INT_DiscardingLowPri)
#define INT_Halted		(INT_CProcHalted | INT_TProcHalted | INT_DProcHalted)
#define INT_ErrorInterrupts	(INT_PciMemErr | INT_SDRamInt | INT_LinkError)

#define INT_MSI0		INT_MainInterrupt
#define INT_MSI1		(INT_Units | INT_Discarding | INT_Halted)
#define INT_MSI2		(INT_Inputters)
#define INT_MSI3		(INT_ErrorInterrupts)

#define E4_INTERRUPT_REG_SHIFT	32
#define E4_INTERRUPT_MASK_MASK	(0xffffffffULL)

/*
 * Trap type values - see trapvalues.v
 */

#define CommandProcInserterError		0x1
#define CommandProcPermissionTrap		0x2
#define CommandProcSendTransInvalid		0x3
#define CommandProcSendTransExpected		0x4
#define CommandProcDmaQueueOverflow		0x5
#define CommandProcInterruptQueueOverflow	0x6
#define CommandProcMemoryFault			0x7
#define CommandProcRouteFetchFault		0x8
#define CommandProcFailCountZero		0x9
#define CommandProcAddressAlignment		0xa
#define CommandProcWaitTrap			0xb
#define CommandProcMultipleGuards		0xc
#define CommandProcOpenOnGuardedChan		0xd
#define CommandProcThreadQueueOverflow		0xe
#define CommandProcBadData                      0xf

#define DmaProcNoFault				0x0
#define DmaProcRouteFetchFault			0x1
#define DmaProcFailCountError			0x2
#define DmaProcPacketAckError			0x3
#define DmaProcRunQueueReadFault		0x4
#define DmaProcQueueOverflow			0x5

#define EventProcNoFault			0x0
#define EventProcAddressAlignment		0x1
#define EventProcMemoryFault			0x2
#define EventProcCountWrapError			0x3

#define InputNoFault				0x0
#define InputAddressAlignment			0x1
#define InputMemoryFault			0x2
#define InputInvalidTransType			0x3
#define InputDmaQueueOverflow			0x4
#define InputEventEngineTrapped			0x5
#define InputCrcErrorAfterPAckOk		0x6
#define InputEopErrorOnWaitForEop		0x7
#define InputEopErrorTrap			0x8
#define InputDiscardAfterAckOk			0x9
 
typedef struct _E4_Sched_Status
{
    E4_uint32 Status;
    E4_uint32 Restart;
} E4_Sched_Status;
 
typedef struct _E4_Input_Ptrs
{
    E4_uint32 ContextFilterTable;
    E4_uint32 TrapBasePtr;
} E4_Input_Ptrs;

#define SCH_StopLowPriQueues		(1 << 0)
#define SCH_DProcHalt			(1 << 1)
#define SCH_TProcHalt			(1 << 2)
#define SCH_CProcHalt			(1 << 3)

#define SCH_CProcTimeout600ns		(1 << 4)
#define SCH_CProcTimeout1p4us		(2 << 4)
#define SCH_CProcTimeout3p0us		(3 << 4)
#define SCH_CProcTimeout6p2us		(4 << 4)
#define SCH_CProcTimeout12p6us		(5 << 4)
#define SCH_CProcTimeout25p4us		(6 << 4)
#define SCH_CProcTimeout51p0us		(7 << 4)
#define SCH_DiscardLowPriInput		(1 << 7)
#define SCH_DiscardHighPriInput		(1 << 8)

#define SCH_DProcTimeslice64us		(0 << 9)
#define SCH_DProcTimeslice128us		(1 << 9)
#define SCH_DProcTimeslice256us		(2 << 9)
#define SCH_DProcTimeslice512us		(3 << 9)

#define SCH_Halt			(SCH_StopLowPriQueues | SCH_DProcHalt | SCH_TProcHalt | SCH_CProcHalt)
#define SCH_Discard			(SCH_DiscardLowPriInput | SCH_DiscardHighPriInput)

#define SCH_RestartCProc		(1 << 0)
#define SCH_RestartTProc		(1 << 1)
#define SCH_RestartEProc		(1 << 2)
#define SCH_RestartDma0Proc		(1 << 3)
#define SCH_RestartDma1Proc		(1 << 4)
#define SCH_RestartDmaPrefetchProc	(1 << 5)
#define SCH_RestartCh0LowPriInput	(1 << 6)
#define SCH_RestartCh1LowPriInput	(1 << 7)
#define SCH_RestartCh0HighPriInput	(1 << 8)
#define SCH_RestartCh1HighPriInput	(1 << 9)
#define SCH_ClearLinkErrorInt		(1 << 10)
#define SCH_ContextFilterFlush		(1 << 11)

/*
 * Link state bits.
 */
#define LS_LinkNotReady		(1 << 0) /* Link is in reset or recovering from an error */
#define LS_Locked		(1 << 1) /* Linkinput PLL is locked */
#define LS_LockError		(1 << 2) /* Linkinput PLL was unable to lock onto the input clock. */
#define LS_DeskewError		(1 << 3) /* Linkinput was unable to Deskew all the inputs. (Broken wire?) */
#define LS_PhaseError		(1 << 4) /* Linkinput Phase alignment error. */
#define LS_DataError		(1 << 5) /* Received value was neither good data or a token. */
#define LS_FifoOvFlow0		(1 << 6) /* Channel 0 input fifo overflowed. */
#define LS_FifoOvFlow1		(1 << 7) /* Channel 1 input fifo overflowed. */
#define LS_Mod45Changed		(1 << 8) /* Mod45 bit has changed. Error setr to force reset. */
#define LS_PAckNotSeenError	(1 << 9) /* PAck value not returned for this packet. */

/*
 * Link State Constant defines, used for writing to LinkSetValue
 */

#define LRS_DataDel0		0x0
#define LRS_DataDel1		0x1
#define LRS_DataDel2		0x2
#define LRS_DataDel3		0x3
#define LRS_DataDel4		0x4
#define LRS_DataDel5		0x5
#define LRS_DataDel6		0x6
#define LRS_DataDel7		0x7
#define LRS_DataDel8		0x8
#define LRS_LinkInValue		0x9
#define LRS_PllDelValue		0xA
#define LRS_ClockEven		0xB
#define LRS_ErrorVal8to0	0xC
#define LRS_ErrorVal17to9	0xD
#define LRS_ErrorVal26to18	0xE
#define LRS_ErrorVal35to27	0xF
#define LRS_NumLinkDels         0x10

#define LRS_Pllfast             0x40

typedef struct _E4_CommandControl
{
    volatile E4_uint32 CommandQueueDescsBase;
    volatile E4_uint32 CommandRequeuePtr;
} E4_CommandControl;

#define E4_CommandRequeueBusy		0x80000000	/* Test against read value of CommandRequeuePtr */
#define E4_CommandRequeueHighPri	0x1		/* Will requeue onto the high pri queue */
#define E4_QueueDescPtrMask		0x7fffffe0

typedef struct _E4_CommandQueueDesc
{
    E4_uint64 CQ_QueuePtrs;
    E4_uint64 CQ_HoldingValue;		/* 32 bit value for 32 bit accesses or OutOfOrderMask*/
    E4_uint64 CQ_AckBuffers;		/* Space for 32 4 bit ack buffer values. */
    E4_uint64 CQ_Control;
} E4_CommandQueueDesc;

/*
 * Rev A - CQ_QueuePtrs
 * [63]		Unused		Should be set to zero.
 * [62:51]	Unused		(reads as top of InsertPtr)
 * [50:35]	CompletedPtr	Completed pointer. This is alligned to a byte address.
 * [34]		Trapped		Will be set if the command has trapped.
 * [33:32]	Size		Size of queue.
 * [31]		Used		Will be set if the descriptor has been changed and written back by the elan.
 * [30:3]	InsertPtr	Insert pointer. This is alligned to a byte address.
 * [2]		TimedOut	Will be set if the queue timedout executing a command.
 * [1]		Priority	When set the queue runs at high priority.
 * [0]		Error		If this becomes set all new data written to the queue is * discarded.
 *
 * Rev B - CQ_QueuePtrs
 * [63]		TimedOut	Will be set if the queue timedout executing a command.
 * [62]		Priority	When set the queue runs at high priority.
 * [61]		QueueType	1=will accept unordered 64 bit PCI writes. 0=will accept ordered 32 or 64 bit PCI writes.
 * [60:51]	Unused		(reads as top of InsertPtr)
 * [50:35]	CompletedPtr	Completed pointer. This is alligned to a byte address.
 * [34]		Trapped		Will be set if the command has trapped.
 * [33:32]	Size		Size of queue.
 * [31]		Used		Will be set if the descriptor has been changed and written back by the elan.
 * [30:3]	InsertPtr	Insert pointer. This is alligned to a byte address.
 * [2]		OrderControl	Holds bit 8 of last PCI accesses. Used by a reordering queue.
 * [1:0]	ErrorType	This field has the current error status of the queue.
 */

/* Common between revA and RevB */
#define CQ_PtrMask		(0x7ffffff8)			/* 31 bit sdram address */
#define CQ_PtrOffsetMask	(0x7fff8)
#define CQ_PtrBaseMask		(0x7ff80000)

#define CQ_InsertPtrShift	(3 - 3)				/* InsertPtr is 64 bit aligned */
#define CQ_SizeShift		(32)
#  define CQ_Size1K		0
#  define CQ_Size8K		1
#  define CQ_Size64K		2
#  define CQ_Size512K		3
#  define CQ_SizeMask		3

#define CQ_CompletedPtrShift	(35 - 3)			/* CompletedPtr is 64 but aligned */

#define CQ_Used			(1ull << 31)
#define CQ_Trapped		(1ull << 34)

#define CQ_QueuePtrsValue(Size,Inserter,Completer) \
	(((E4_uint64) (Size) << CQ_SizeShift) | \
	 ((E4_uint64) (Inserter) << CQ_InsertPtrShift) | \
	 ((E4_uint64) (Completer) << CQ_CompletedPtrShift))

#define CQ_InsertPtr(QueuePtrs) \
	(((E4_uint64) QueuePtrs) & CQ_PtrMask)

#define CQ_CompletedPtr(QueuePtrs) \
	(((E4_uint32)((QueuePtrs) >> CQ_CompletedPtrShift) & CQ_PtrOffsetMask) | \
	 (CQ_InsertPtr(QueuePtrs) & CQ_PtrBaseMask))

#define CQ_Size(SizeVal)		(1024 * (1 << ((SizeVal)*3)))

/* Rev A specific */
#define CQ_RevA_Error			(1 << 0)
#define CQ_RevA_Priority		(1 << 1)
#define CQ_RevA_TimedOut		(1 << 2)

/* Rev B specific */
#define CQ_RevB_ErrorType(QueuePtr)	((QueuePtr) & (3 << 0))
#  define CQ_RevB_NoError		(0ull << 0)
#  define CQ_RevB_Overflowed		(1ull << 0)
#  define CQ_RevB_InvalidWriteSize	(2ull << 0)
#  define CQ_RevB_InvalidWriteOrder	(3ull << 0)
#define CQ_RevB_OrderControl		(1ull << 2)

#define CQ_RevB_QueueType(QueuePtr)	((QueuePtr) & (1ull << 61))
#  define CQ_RevB_ReorderingQueue	(1ull << 61)
#  define CQ_RevB_32bitWriteQueue	(0ull << 61)

#define CQ_RevB_Priority		(1ull << 62)
#define CQ_RevB_TimedOut		(1ull << 62)

/* 
 * CQ_AckBuffers - Packet Ack Values
 */
#define PackOk			(0x0)
#define PackTestFail		(0x1)
#define PackDiscard		(0x2)
#define PackError		(0x7)
#define PackTimeout		(0x8)
#define PackWaiting		(0xF)
#define PackValue(val,chan)	(((val) >> ((chan) * 4)) & 0xf)

/*
 * CQ_Control
 * [63:35]	ExtractPtr
 * [34]		Unused
 * [33:32]	ChannelNotCompleted
 * [31:24]	Permissions
 * [23:16] 	RestartCount		Decremented after each restart. Will trap when zero
 * [15:14]	Unused			Should be set to zero
 * [13:0]	Context
 */
#define CQ_Context(Control)		((E4_uint32) ((Control) >>  0) & 0x3fff)
#define CQ_RestartCount(Control)	((E4_uint32) ((Control) >> 16) & 0x7f)
#define CQ_ChannelNotCompleted(Control)	((E4_uint32) ((Control) >> 32) & 3)
#define CQ_ExtractPtr(Control)		((E4_uint32) ((Control) >> 32) & 0xFFFFFFF8)

#define CQ_RestartCountShift		16

#define CQ_SetEventEnableBit	(1 << 24)
#define CQ_WaitEventEnableBit	(1 << 25)
#define CQ_ModifyEnableBit	(1 << 26)
#define CQ_WriteEnableBit	(1 << 27)
#define CQ_ThreadStartEnableBit	(1 << 28)
#define CQ_DmaStartEnableBit	(1 << 29)
#define CQ_STENEnableBit	(1 << 30)
#define CQ_InterruptEnableBit	(1 << 31)
#define CQ_EnableAllBits        (0xFF000000)
#define CQ_PermissionMask	(0xFF000000)

#define CQ_ControlValue(Cntx, RestartCount, Permissions) \
	(((Cntx) & 0x3fff) | (((RestartCount) & 0xff) << 16) | ((Permissions) & CQ_PermissionMask))

/*
 * This file describes the slave address map of Elan4.
 *
 * Elan4 has two PCI 64 bit base address registers. One is setup for elan
 * local memory and the other is for the command port, elan registers and ebus.
 *
 * This file describes the command port, elan registers and ebus BAR. This is a
 * 26 bit base address register and is split up as follows:
 * 1 The ebus requires 21 bits of address. 26'h3e00000 to 26'h3ffffff
 * 2 The control regsiters requires 16 bits of address. 26'h3df0000 to 26'h3dfffff
 * 3 The command port has the rest. This give just under 8k command ports or about 123 per
 *   processor of a 64 node SMP.
 */

/* BAR1 contains the command queues followed by the registers and the Ebus - and is 26 bits */
/* each command queue has an 8K page associated with it */
#define CQ_CommandMappingSize		(1 << 13)
#define CQ_NumCommandDescs		((1 << (26 - 13)))
#define CQ_CommandDescsAlignment	((1 << (26 - 13)) * sizeof (E4_CommandQueueDesc))

/* control reg bits i.e. E4_DataBusMap.SysControlReg */
#define CONT_EN_ALL_SETS		(1ULL << 0) /* enable cache */
#define CONT_MMU_ENABLE			(1ULL << 1) /* bit 0 enables mmu */
#define CONT_CACHE_HASH_TABLE		(1ULL << 2) /* cache up hash table entries */
#define CONT_CACHE_CHAINS		(1ULL << 3) /* cache up chain entries */
#define CONT_CACHE_ROOT_CNTX		(1ULL << 4) /* cache root context table for routes and filters. */
#define CONT_CACHE_STEN_ROUTES		(1ULL << 5) /* cache up sten packet routes */
#define CONT_CACHE_DMA_ROUTES		(1ULL << 6) /* cache up dma packet routes */

#define CONT_CACHE_NONE		0ULL
#define CONT_CACHE_ALL		(CONT_CACHE_HASH_TABLE | CONT_CACHE_CHAINS | CONT_CACHE_ROOT_CNTX | \
				 CONT_CACHE_STEN_ROUTES | CONT_CACHE_DMA_ROUTES)

/* This controls the format size and position of the MMU hash tables. */
#define CONT_INHIBIT_MAX_CHAIN_ITEMS	(1ULL << 7)	/* Prevents the MaxChainItems value of 1024 from forcing a translation miss */
#define CONT_TABLE0_MASK_SIZE_SHIFT	8		/* Defines the size of hash table 0 */
#define CONT_TABLE0_PAGE_SIZE_SHIFT	13		/* Set the page size for hash table 0 */
#define CONT_TABLE1_MASK_SIZE_SHIFT	16		/* Defines the size of hash table 1 */
#define CONT_TABLE1_PAGE_SIZE_SHIFT	21		/* Set the page size for hash table 1 */
#define CONT_TWO_HASH_TABLES		(1ULL << 24)	/* Sets the MMU to use two hash tables. If not set only 0 used. */
#define CONT_2K_NOT_1K_DMA_PACKETS	(1ULL << 25)	/* Used to select the default DMA packet size. */
#define CONT_ALIGN_ALL_DMA_PACKETS	(1ULL << 26)	/* Will force all dma packets to be aligned to a page.*/
#define CONT_DIRECT_MAP_PCI_WRITES	(1ULL << 27)	/* Will force pci writes to write and flush the dcache.*/
#define CONT_TLB_FLUSH			(1ULL << 28)	/* Invalidates the TLB and indicates when flushed */
#define CONT_CLEAR_WALK_WROTE_TABLES	(1ULL << 29)	/* Used to guarantee that the elan is using new PTE values. */
#define CONT_ROUTE_FLUSH		(1ULL << 30)	/* Invalidates all route cache entries. */
#define CONT_CLEAR_LINKPORT_INT		(1ULL << 31)	/* Clears the Linkport key fail interrupt. Reads as 0. */
#define CONT_CLEAR_SDRAM_ERROR		(1ULL << 32)	/* Clears an EEC error interrupt. Reads as 0. */

/*
 * These are extra control bits used for testing the DLLs of the SDRAM interface. Most of the Sdram
 * control bits are defined in xsdram.h
 */
#define SDRAM_FIXED_DLL_DELAY_SHIFT	47
#define SDRAM_FIXED_DLL_DELAY_BITS	5
#define SDRAM_FIXED_DLL_DELAY_MASK	((1ULL << SDRAM_FIXED_DLL_DELAY_BITS) - 1ULL)
#define SDRAM_FIXED_DLL_DELAY(Value)	((SDRAM_FIXED_DLL_DELAY_MASK & (Value)) << SDRAM_FIXED_DLL_DELAY_SHIFT)
#define SDRAM_FIXED_DELAY_ENABLE	(1ULL << 52)
#define SDRAM_GET_DLL_DELAY(Value)	(((Value) >> SDRAM_FIXED_DLL_DELAY_SHIFT) & SDRAM_FIXED_DLL_DELAY_MASK)

#define SDRAM_166_DLL_CORRECTION_FACTOR	3	/* This is to allow for SSO and ringing on the DQ lines */
#define SDRAM_150_DLL_CORRECTION_FACTOR	2	/* This is to allow for SSO and ringing on the DQ lines */

#define PAGE_SIZE_4K	0x0
#define PAGE_SIZE_8K	0x1
#define PAGE_SIZE_64K	0x2
#define PAGE_SIZE_512K	0x3
#define PAGE_SIZE_2M	0x4
#define PAGE_SIZE_4M	0x5
#define PAGE_SIZE_64M	0x6
#define PAGE_SIZE_512M	0x7

#define PAGE_SIZE_MASK	0x7
#define PAGE_MASK_MASK	0x1f

/* control reg bits i.e. E4_DataBusMap.LinkControlReg */
#define LCONT_REVA_GREEN_LED	 	(1 << 0)
#define LCONT_REVA_YELLOW_LED	 	(1 << 1)
#define LCONT_REVA_RED_LED	 	(1 << 2)
#define LCONT_REVA_ENABLE_LED_DRIVE	(1 << 3) /* Enable manual setting of the Leds to the bits set above. */

#define LCONT_REVB_DISABLE_TLB_PREFETCH 	(1 << 0)
#define LCONT_REVB_DISABLE_CRC_ERROR_CHECKING	(1 << 1)


#define LCONT_EN_SYS_WRITES		(1 << 4) /* Enable linkport writes to sys registers. i.e. all of E4_DataBusMap. */
#define LCONT_EN_SYS_READS		(1 << 5) /* Enable linkport reads from sys registers. i.e. all of E4_DataBusMap. */
#define LCONT_EN_USER_WRITES		(1 << 6) /* Enable linkport writes to user registers. i.e. all of E4_User_Regs. */
#define LCONT_EN_USER_READS		(1 << 7) /* Enable linkport reads from user registers. i.e. all of E4_User_Regs. */

#define LCONT_TEST_VALUE_MASK		0x3ff	 /* Value used for test writes and link boundary scan. */
#define LCONT_TEST_VALUE_SHIFT		8
#define LCONT_TEST_VALUE(Value)		((LCONT_LINK_STATE_MASK & (Value)) << LCONT_TEST_VALUE_SHIFT)

/*
 * State read from LINK_STATE when TEST_VALUE is set to the following values.
 * TEST_VALUE	  LINK_STATE read	TEST_VALUE	  LINK_STATE read
 *    000     -   Data delay count 0	   008	     -  Data delay count 8
 *    001     -   Data delay count 1	   009	     -  Link in value
 *    002     -   Data delay count 2	   00a	     -  PLL delay
 *    003     -   Data delay count 3	   00b	     -  Clock Delay
 *    004     -   Data delay count 4	   00c	     ?  ErrorVal8to0
 *    005     -   Data delay count 5	   00d	     ?  ErrorVal17to9
 *    006     -   Data delay count 6	   00e	     ?  ErrorVal26to18
 *    007     -   Data delay count 7	   00f	     ?  ErrorVal35to27
 */

#define LCONT_TEST_CONTROL_MASK		0x3	/* Selects and controls the action of the LINK_STATE value. */
#define LCONT_TEST_CONTROL_SHIFT	18

#define LCONT_READ_ERRORS		0	/* {Mod45RequestChanged, FifoOverflowError, DataError, PhaseError,
						 * 	DeskewError, LockError, Locked, LinkNotReady} */
#define LCONT_READ_STATE		1	/* Read valus addressed by TEST_CONTROL value */
#define LCONT_FIX_LINK_DELAYS		2	/* Sets delays to TEST_CONTROL value */
#define LCONT_BOUNDARY_SCAN		3	/* Puts link into boundary scan. Outputs TEST_CONTROL value to link,
						 * reads LINK_STATE from link. */ 

#define LCONT_LINK_STATE_MASK		0x3ff	/* Read only */
#define LCONT_LINK_STATE_SHIFT		20	/* Read only */
#define LCONT_LINK_STATE(ControlRegValue)	(LCONT_LINK_STATE_MASK & ((ControlRegValue) >> LCONT_LINK_STATE_SHIFT))

/* control reg bits i.e. E4_DataBusMap.LinkContSettings */
#define LCONT_MOD45_DISABLE		(1 << 0) /* is set the link will try to run in TNB mode. */
#define LCONT_CONFIG_PHASE_MASK		0x7	/* This set the delay through the phase alignment buffer. */
#define LCONT_CONFIG_PHASE_SHIFT	1

#define LCONT_PLL_REF_VAL_BITS_MASK	0x7f	/* This is the divide value on the LinkIn clock to form the comms PLL */
#define LCONT_PLL_REF_VAL_BITS_SHIFT	4	/* reference clock. Div value is (n - 2). e.g. to Divide by 7 set to 5. */

#define LCONT_FORCE_COMMSCLK_LOCAL	(1 << 11) /* This must be set at one end of a back to back Elan configuration. */
#define LCONT_LVDS_VOLTAGE_BITS_MASK	0x3	/* This is used to set the voltage swing on the LVDS link output pads. */
#define LCONT_LVDS_VOLTAGE_BITS_SHIFT	12	/* reference clock. Div value is (n - 2). e.g. to Divide by 7 set to 5. */

#define LCONT_VOD_170			0	/* Approximate differential voltage swing in mV of link outputs into */
#define LCONT_VOD_360			1	/* a 100 ohm diferential load. */
#define LCONT_VOD_460			2
#define LCONT_VOD_550			3

#define LCONT_LVDS_TERMINATION_MASK	0x3	/* This set the resistor values of the internal single ended termation */
#define LCONT_LVDS_TERMINATION_SHIFT	14	/* resistors of the link input and comms input clcok. */

#define LCONT_TERM_55_OHM		0	/* Resistor values for internal termination of LVDS pads. */
#define LCONT_TERM_50_OHM		1
#define LCONT_TERM_AUTO_OHM		2	/* Should normally be set to auto. */
#define LCONT_TERM_45_OHM		3

#define LCONT_LVDS_EN_TERM_UPDATE	(1 << 47) /* This should be asserted and deasserted if LCONT_LVDS_TERMINATION is changed. */

/* Macros used to access and construct MMU hash table and chain entries. */
/*
 * Each hash entry is made up of a 64 byte block. Each entry hash two tags where each
 * tag has 4 PTE's. PTE's 0 to 2 use the bottom 48 bits of a 64 bit word and PTE 3
 * uses the top 16 bits of 3 64 bit words.
 *
 * These macros can be used to build a single PTE. PTE3 needs to be built into a 48 bit
 * object before they can be used.
 */
#define PTE_ENTRY_MASK		0x0000ffffffffffffULL
#define PTE_TYPE_MASK		0x000000000000000fULL	
#define PTE_PERM_MASK		0x00000000000000f0ULL
#define PTE_PERM_TYPE_MASK	0x00000000000000ffULL
#define PTE_REF_MASK		0x0000000000000100ULL
#define PTE_PPN_MASK		0x00007ffffffffe00ULL
#define PTE_MOD_MASK		0x0000800000000000ULL
#define PTE_TOPADDR_MASK	0x0000600000000000ULL

#define PTE_MOD_SHIFT		47
#define PTE_PPN_SHIFT		9
#define PTE_REF_SHIFT		8
#define PTE_PERM_SHIFT		4
#define PTE_TYPE_SHIFT		0

#define PTE_PADDR_SHIFT		(12 - 9)		/* Physical addresses are shifted down 3 this to go into the PTE */


/* Values required for tag 3 */
#define PTE_REF_3			0x0100000000000000ULL
#define PTE_MOD_3			0x8000000000000000ULL
#define PTE_ENTRY_MASK_3		0xffff000000000000ULL
#define PTE_PERM_TYPE_MASK_3		0x00ff000000000000ULL
#define PTE_ENTRY_3_FOR_0(NewPte)	((NewPte << (48)) & PTE_ENTRY_MASK_3)
#define PTE_ENTRY_3_FOR_1(NewPte)	((NewPte << (32)) & PTE_ENTRY_MASK_3)
#define PTE_ENTRY_3_FOR_2(NewPte)	((NewPte << (16)) & PTE_ENTRY_MASK_3)

/* Values required for the tags */
#define TAG_CONTEXT_MASK		0x0000000000003fffULL
#define TAG_ADDRESS_MASK		0xfffffffff8000000ULL
#define TAG_CHAINPTR_18TO6_MASK		0x0000000007ffc000ULL
#define TAG_CHAINPTR_LOW_SHIFT		(14 - 6)
#define TAG_CHAINPTR_30TO19_MASK	0x0000000003ffc000ULL
#define TAG_CHAINPTR_HIGH_SHIFT		(19 - 14)
#define TAG_COPY_BIT			0x0000000004000000ULL

/*
 * This takes number loaded into the control register and returns the page size as a power of two.
 */

#define E4_PAGE_SIZE_TABLE		E4_uint32 const PageSizeTable[] = {12, 13, 16, 19, 21, 22, 26, 29}
#define E4_PAGE_SIZE_TABLE_SIZE		(sizeof(PageSizeTable)/sizeof(PageSizeTable[0]))

/*
 * This macro generates a hash block index.
 *
 * Cntx		 This is the 14 bit context. It should not be larger than 14 bits.
 * VAddr	 This is the 64 bit virtual address. It does not require any masking and can be a byte address.
 * PageSize	 This is the value loaded into the control register for this hash table.
 * HashTableMask This should be set mask out upper bits past the end of the hash table.
 */
#define E4MMU_SHIFT_ADDR(VAddr, Shift) \
    ((((E4_uint32)(VAddr)) >> (Shift)) | (((E4_uint32)((VAddr) >> 32)) << (32 - (Shift))))

#define E4MMU_CONTEXT_SCRAMBLE(Cntx) \
	      ((((Cntx) << 8) | ((Cntx) >> 6)) ^ (((Cntx) << 15) | ((Cntx) << 1)))

#define E4MMU_HASH_INDEX(Cntx, VAddr, PageShift, HashTableMask)		\
	    ((E4MMU_SHIFT_ADDR(VAddr, (PageShift) + 2) ^ E4MMU_CONTEXT_SCRAMBLE(Cntx)) & (HashTableMask))

#define E4MMU_TAG(vaddr,ctx)	(((vaddr) & TAG_ADDRESS_MASK) | ((ctx) & TAG_CONTEXT_MASK))

#define E4MMU_TAG2VADDR(tag,hashidx,PageShift,HashTableMask)	\
		(((tag) & TAG_ADDRESS_MASK) | ((((hashidx) ^ E4MMU_CONTEXT_SCRAMBLE((tag) & TAG_CONTEXT_MASK)) & (HashTableMask)) << ((PageShift + 2))))

/*
 * Detailed bit descriptions for the tags and PTE's are better done with the macros
 * defined above.
 */
typedef struct _E4_HashTableEntry
{
   E4_uint64	Tag[2];
   E4_uint64	TagPTE[2][3];
} E4_HashTableEntry;

#define E4MMU_TAG_OFFSET(tag)		((tag) << 3)
#define E4MMU_PTE_LOW_OFFSET(tag,pte)	((((tag)*3 + (pte) + 2) << 3))
#define E4MMU_PTE_HIGH_OFFSET(tag,pte)	((((tag)*3 + (pte) + 2) << 3) + 4)
#define E4MMU_PTE3_WORD0_OFFSET(tag)	((((tag)*3 + 2) << 3) + 6)
#define E4MMU_PTE3_WORD1_OFFSET(tag)	((((tag)*3 + 3) << 3) + 6)
#define E4MMU_PTE3_WORD2_OFFSET(tag)	((((tag)*3 + 4) << 3) + 6)


/*
 * Hash0AddrBits is the size of the hash table in bytes as a power of 2.
 * e.g. 11 would give 32 hash entries where each entry is 64 bytes.
 */
#define SETUP_HASH_TABLES(Hash0PageSize, Hash0AddrBits, Hash1PageSize, Hash1AddrBits)	\
			  (((Hash0PageSize) << CONT_TABLE0_PAGE_SIZE_SHIFT) |	\
			   ((Hash0AddrBits) << CONT_TABLE0_MASK_SIZE_SHIFT) |	\
			   ((Hash1PageSize) << CONT_TABLE1_PAGE_SIZE_SHIFT) |	\
			   ((Hash1AddrBits) << CONT_TABLE1_MASK_SIZE_SHIFT))

/* ECC status register */
#define ECC_Addr(s)			((s) & 0x7ffffff8ULL)
#define ECC_Syndrome(s)			(((s) >> 32) & 0xffffULL)
#define ECC_RisingDQSSyndrome(s)	(((s) >> 32) & 0xffULL)
#define ECC_FallingDQSSyndrome(s)	(((s) >> 40) & 0xffULL)
#define ECC_UncorrectableErr(s) 	(((s) >> 48) & 1ULL)
#define ECC_MultUncorrectErrs(s)	(((s) >> 49) & 1ULL)
#define ECC_CorrectableErr(s)		(((s) >> 50) & 1ULL)
#define ECC_MultCorrectErrs(s)		(((s) >> 51) & 1ULL)

/* Permission type saved in a PTE. This is a four bit field */
#define PERM_Disabled		0x0
#define PERM_Unused		0x1
#define PERM_LocDataRead	0x2
#define PERM_LocDataWrite	0x3
#define PERM_LocRead		0x4
#define PERM_LocExecute		0x5
#define PERM_ReadOnly		0x6
#define PERM_LocWrite		0x7
#define PERM_LocEventOnly	0x8
#define PERM_LocEventWrite	0x9
#define PERM_RemoteEvent	0xa
#define PERM_RemoteAll		0xb
#define PERM_RemoteReadOnly	0xc
#define PERM_RemoteWriteLocRead	0xd
#define PERM_DataReadWrite	0xe
#define PERM_NoFault		0xf

#define PERM_Mask		0xf

/* Permission type hints to device driver */
#define PERM_Preload		0x10

#define PTE_SetPerm(Perm)	(((Perm) & PERM_Mask) << 4)

/* Control info saved in the lookup field of the TLB */
#define PTE_PciNotLocal		(1ULL << 0)		/* Directs the access to the PCI interface */
#define PTE_BigEndian		(1ULL << 1)		/* Valid for PCI entries only */
#define PTE_RelaxedOrder	(1ULL << 2)		/* Valid for PCI entries only */
#define PTE_DontSnoop		(1ULL << 3)		/* Valid for PCI entries only */

#define PTE_UseFixedSet		(1ULL << 1)		/* Value for non PCI entries only */
#define PTE_CommandQueue	(1ULL << 2)		/* Value for non PCI entries only */
#define PTE_SetFixedSetNo(Set)	((((Set) & 3) << 2) | PTE_UseFixedSet)

#define PTE_TypeBitsMask	(0xfULL)
#define PTE_PermissionTypeMask	(0xfULL << 4)
#define PTE_Referenced		(1ULL << 8)
#define PTE_PhysicalPageNoMask	(0x7ffffffffe00ULL)
#define PTE_Modified		(1ULL << 47)

#define PTE_PhysicalAddrShiftIntoPTE	(12 - 9)

/* define page table entry bit fields */
#define TLB_PageSizeBits	(3 << 0)
#define TLB_ACCBits		(7 << 2)
#define TLB_LocalBit		(1 << 5)
#define TLB_PCI64BitTargetBit	(1 << 6)
#define TLB_PCIBigEndianBit	(1 << 7)

#define TLB_ModifiedBit		(1 << 55)
#define TLB_ReferencedBit	(1 << 63)

/* Used to read values from the tlb. */
#define TLB_TlbReadCntBitsSh	56
#define TLB_UseSelAddrSh	(1ULL << 60)
#define TLB_WriteTlbLine	(1ULL << 61)

#define TLB_SEL_LINE(LineNo) (TLB_UseSelAddrSh | \
			      ((E4_uint64)((LineNo) & 0xf) << TLB_TlbReadCntBitsSh))

#define TLB_NUM_ENTRIES		16
/*
 * The following macros are used with the test access port (TlbLineValue) for the TLBs.
 */
#define TLV_DoPciAccess			(1ULL << 0)
#define TLV_CommandAccess		(1ULL << 1)
#define TLV_DoCacheAccess		(1ULL << 2)
#define TLV_notStartTLBWalk		(1ULL << 3)
#define TLV_UseFixedSet			(1ULL << 4)
#define TLV_BigEndian			(1ULL << 4)
#define TLV_RelaxedOrder		(1ULL << 5)
#define TLV_DontSnoop			(1ULL << 6)
#define TLV_FixedSetNo_MASK		(3ULL << 5)
#define TLV_PciTypeBits_MASK		(7ULL << 4)
#define TLV_LookupBits_MASK		(0x7fULL)
#define TLV_MissErr			(1ULL << 7)
#define TLV_TypeBits			(0xffULL)

#define TLV_PhysicalAddr_MASK		(0x3fffffffff000ULL)

#define TLV_TlbTesting			(1ULL << 51)
#define TLV_SelectUnitsTlbRead		(1ULL << 52)
#define TLV_SelectTProcTlbRead		(1ULL << 53)

#define TLV_TlbLineSelect_MASK		(0xf)
#define TLV_UnitsTlbLineSelect_SHIFT	(54)
#define TLV_TProcTlbLineSelect_SHIFT	(59)
#define TLV_EnableUnitsTlbRead		(1ULL << 58)
#define TLV_EnableTProcTlbRead		(1ULL << 63)

/*
 * Use this macro to enable direct testing of the Units TLB.
 * When Line is in the range 0 to 15 a TLB line is selected for reading or writing.
 * When Line is set to -1 the tlb will be activated to perform a match.
 */
#define TLV_UnitsTlbLineSel(Line) (((Line) == -1) ? 0ULL : \
    (TLV_EnableUnitsTlbRead | ((E4_uint64)((Line) & TLV_TlbLineSelect_MASK) << TLV_UnitsTlbLineSelect_SHIFT)))
#define TLV_TProcTlbLineSel(Line) (((Line) == -1) ? 0ULL : \
    (TLV_EnableTProcTlbRead | ((E4_uint64)((Line) & TLV_TlbLineSelect_MASK) << TLV_TProcTlbLineSelect_SHIFT)))
 
/* 
 * Thread_Trap_State
 *  see f_RegFileControl.v TProcStatus
 */
#define TS_HaltThread		      (1 << 0)
#define TS_TrapForTooManyInstructions (1 << 1)
#define TS_InstAccessException	      (1 << 2)
#define TS_Unimplemented	      (1 << 3)
#define TS_DataAccessException	      (1 << 4)
#define TS_DataAlignmentError	      (1 << 5)
#define TS_TrapForUsingBadData	      (1 << 6)
#define TS_TrapTypeMask		      (0x7f)
#define TS_DataPortNo(ts)	      (((ts) >> 7) & 7)
#define TS_TrappedFlag		      (1 << 10)
#define TS_MemLock		      (1 << 11)
#define TS_XCCshift		      12
#define TS_XCCmask		      0xff
#define TS_ICC(ts)		      (((ts) >> 12) & 15)
#define TS_XCC(ts)		      (((ts) >> 16) & 15)
#define TS_InstValid_F		      (1 << 20)
#define TS_InstValid_R		      (1 << 21)
#define TS_InstValid_E		      (1 << 22)
#define TS_InstValid_W		      (1 << 23)
#define TS_HighPriority		      (1 << 24)
#define TS_RemoteThread		      (1 << 25)
#define TS_TProcTranslationInProgress (1 << 26)
#define TS_MemLock_E		      (1 << 27)

/* Thread run queue entries */
typedef struct E4_ThreadRegs
{
    E4_uint64 Registers[7];
} E4_ThreadRegs;

typedef struct E4_TProcQueueEntry
{
    E4_ThreadRegs	Regs;			/* XXXX: jon check this */
    E4_uint64		Context;		/* XXXX: jon check this */
} E4_TProcQueueEntry;

typedef struct E4_DProcQueueEntry
{
    E4_DMA		Desc;
    E4_uint64		Pad;
} E4_DProcQueueEntry;

/*
 * Packet acknowledge values.
 */
#define E4_PAckOk	0
#define E4_PAckTestFail	1
#define E4_PAckDiscard	2
#define E4_PAckError	3

/*
 * return values from breaktest instruction.
 */
#define ICC_CARRY_BIT           (0x1ULL << 0)  /* Breaktest: Load pending         */
#define ICC_ZERO_BIT            (0x1ULL << 1)  /* Breaktest: Time to break        */
#define ICC_SIGNED_BIT          (0x1ULL << 2)  /* Breaktest: Another thread ready */
#define ICC_TPROC_RDY_LOW_PRI   (0x1ULL << 3)
#define ICC_TPROC_RDY_HIGH_PRI  (0x1ULL << 4)
#define ICC_RUNNING_HIGH_PRI    (0x1ULL << 5)
#define ICC_RUNNING_AS_REMOTE   (0x1ULL << 6)
#define ICC_TIME_TO_BREAK       (0x1ULL << 7)
#define ICC_RS1LOAD_PENDING     (0x1ULL << 8)
#define ICC_TPROC_HALT          (0x1ULL << 9)

/*
 * Main Interrupt cookies
 * [63:14]	user cookie
 * [13:0]	context
 */
#define E4_MAIN_INT_SHIFT		14
#define E4_MAIN_INT_COOKIE(cookie)	((cookie) >> E4_MAIN_INT_SHIFT)
#define E4_MAIN_INT_CTX(cookie)		((cookie) & 0x3FFF)

typedef E4_uint64 E4_MainIntEntry;

#define E4_MainIntEntrySize	sizeof (E4_MainIntEntry)

/*
 * The internal databus is 64 bits wide.
 * All writes to the internal registers MUST be made with 64 bit write operations.
 * These can be made up of pairs 32 bit writes on the PCI bus. The writes will be
 * treated as nops if they are performed with two separate 32 bit writes.
 */
typedef volatile struct _E4_DataBusMap
{
   E4_uint64		InputTrans[4][16];									/* 0x000 */

   E4_uint64		Dma0TransAddr;										/* 0x200 */
   E4_DMA		Dma0Desc;	/* Current Dma0 registers */						/* 0x208 */

   E4_uint64		Dma1TransAddr;										/* 0x240 */
   E4_DMA		Dma1Desc;	/* Current Dma1 registers */						/* 0x248 */
  
   E4_uint64		Dma0LastPacketSize;									/* 0x280 */
   E4_uint64		Dma0ThisPacketSize;									/* 0x288 */
   E4_uint64		Dma0DescSizeInProg;									/* 0x290 */
   E4_uint64		Dma0BytesToPrefetch;									/* 0x298 */
   E4_uint64		Dma0PrefetchAddr;									/* 0x2a0 */
   E4_uint64		EventCountAndType;									/* 0x2a8 */
   E4_uint64		EventParameters[2];									/* 0x2b0 */
  
   E4_uint64		Dma1LastPacketSize;									/* 0x2c0 */
   E4_uint64		Dma1ThisPacketSize;									/* 0x2c8 */
   E4_uint64		Dma1DescSizeInProg;									/* 0x2d0 */
   E4_uint64		Dma1BytesToPrefetch;									/* 0x2d8 */
   E4_uint64		Dma1PrefetchAddr;									/* 0x2e0 */
   E4_Input_Ptrs	InputTrapAndFilter;									/* 0x2e8 */
   E4_uint64		EventAddress;										/* 0x2f0 */
   E4_QueuePtr		MainIntQueuePtrs;									/* 0x2f8 */
   
   E4_uint64		Event_Copy[16];										/* 0x300 */

   E4_uint64		CommandCopy[7];										/* 0x380 */
   E4_uint64		CommandHold;										/* 0x3b8 */

   E4_uint64		InputQueueDesc[4];									/* 0x3c0 */

   /* Run queue Pointers */
   E4_uint64		DProcLowPriPtrs;									/* 0x3e0 */
   E4_uint64		DProcHighPriPtrs;									/* 0x3e8 */
   E4_uint64		TProcLowPriPtrs;									/* 0x3f0 */
   E4_uint64		TProcHighPriPtrs;									/* 0x3f8 */

   E4_uint64		CProcStatus;										/* 0x400 */
   E4_uint64		TProcStatus;										/* 0x408 */
   E4_uint64		IProcStatus;										/* 0x410 */
   E4_uint64		EProcStatus;										/* 0x418 */
   E4_uint64		DProc0Status;										/* 0x420 */
   E4_uint64		DProc1Status;										/* 0x428 */
   E4_Sched_Status	SchedStatus;										/* 0x430 */

   E4_uint64		LoadIProcCntxFilter;	/* Will load one of 4 cntx filter regs. Write only */		/* 0x438 */

   E4_CommandControl	CommandControl;										/* 0x440 */
   E4_uint64		CommandCacheTestPort;									/* 0x448 */
   E4_uint64		CommandLowPriRunPtrs;									/* 0x450 */
   E4_uint64		CommandHighPriRunPtrs;									/* 0x458 */
   E4_uint64		CommandSchedDataPort[4];								/* 0x460 */

   E4_uint64		DmaRouteBuffer[2][2];	/* Write only. Should not be written to. */			/* 0x480 */
   E4_uint64		StenRouteBuffer[2];	/* Write only. Should not be written to. */			/* 0x4a0 */
   E4_uint64		pad4[0x098 - 0x096];									/* 0x4b0 */

   E4_uint64		DmaAlignmentPort[8];	/* Write only. Should only be written to clear the prev reg. */	/* 0x4c0 */

   E4_uint64		MmuBlockEntry[8];	/* Used for hash table and chain fetches */			/* 0x500 */
   E4_uint64		WriteUnitsTlbLine[3];									/* 0x550 */
   E4_uint64		pad5;											/* 0x540 */
   E4_uint64		WriteTProcTlbLine[3];									/* 0x568 */
   E4_uint64		pad6;											/* 0x540 */

   E4_uint64		MmuTableBasePtrs;	/* Both tables packed into a single 64 bit value */		/* 0x580 */
   E4_uint64		MmuFaultAndRootCntxPtr;	/* Both packed into a single 64 bit value */			/* 0x588 */
   E4_uint64		UnitsVAddr;										/* 0x590 */
   E4_uint64		TProcVAddr;										/* 0x598 */
   E4_uint64		UnitsCntx;										/* 0x5a0 */
   E4_uint64		TProcCntx;		/* Read only. Writes access VProcCacheWritePort	*/		/* 0x5a8 */
   E4_uint64		FaultAddrReg;										/* 0x5b0 */
   E4_uint64		FaultTypeAndContextReg;									/* 0x5b8 */

   E4_uint32		SysControlReg;										/* 0x5c0 */
   E4_uint32		CacheTagValue;										/* 0x5c4 */
   E4_uint64		TlbLineValue;										/* 0x5c8 */
   E4_uint64		SDRamConfigReg;										/* 0x5d0 */
   E4_uint32		InterruptMask;										/* 0x5d8 */
   E4_uint32		InterruptReg;										/* 0x5dc */
   E4_uint64		SDRamECCStatus;										/* 0x5e0 */
   E4_uint32		LinkControlReg;										/* 0x5e8 */
   E4_uint32		LinkContSettings;									/* 0x5ec */
   E4_uint64		LinkPortKey;										/* 0x5f0 */
   E4_uint64		LinkPortLock;										/* 0x5f8 */

   E4_uint64		SDRamWriteBuffer[4][8];									/* 0x600 */
   E4_uint64		SDRamReadBuffer[4][8];									/* 0x700 */

   E4_uint64		TProcRegs[64];										/* 0x800 */
   E4_uint64		TProcStartUp[8];	/* Not to be used except by the elan itself */			/* 0xa00 */

   E4_uint64		LoadPending;										/* 0xa40 */
   E4_uint64		StortPending;										/* 0xa48 */
   E4_uint64		DirtyBits;										/* 0xa50 */
   E4_uint64		BadBits;										/* 0xa58 */

   E4_uint64		ICachePort_Cntl_Addr;									/* 0xa60 */
   E4_uint64		Thread_Trap_State;									/* 0xa68 */

/* Instruction buffer (4 * 32 bit words) */
   E4_uint64		nPC_W;											/* 0xa70 */
   E4_uint64		PC_W;											/* 0xa78 */

   E4_uint64		ICacheFillData[8];									/* 0xa80 */
   E4_uint64		ICachePort[8];										/* 0xac0 */

   E4_uint64		PciDataBufs[4][8];									/* 0xb00 */

   E4_uint64		CommandQueueBuffer[128];								/* 0xc00 */
} E4_DataBusMap;

/*
 * These macros are used to setup the thread pcoessors ICache.
 */
#define E4_ICacheTagAddrShift		6
#define E4_AccessICacheRams		1
#define E4_InvalidTagValue		0xffffffffffffffffULL
#define E4_ICacheSizeInBytes		(1024*16)
#define E4_ICacheLineSizeInBytes	(64)
#define E4_ICacheLines			(E4_ICacheSizeInBytes/E4_ICacheLineSizeInBytes)
#define E4_ICachePortSize		( (sizeof((E4_DataBusMap *) 0)->ICachePort) /	\
 					  (sizeof((E4_DataBusMap *) 0)->ICachePort[0]))

#define E4_ICacheFixupInsn		0xc0b02f95ull		/* st1 [%r0 +  0xf95] */
#define E4_ICacheFixupAddr		0xf95ull
#define E4_ICacheFixupOffset		0xfc0

/*
 * Event interrupt
 */
typedef volatile union _E4_EventInt
{
   E4_uint64	 ForceAlign;
   struct {
       E4_uint32 IntCookie;
       E4_uint32 EventContext;	/* Bits 16 to 28 */
    } s;
} E4_EventInt;

/*
 * The following are used to interpret a fault status register.
 */

/*
 * FSR[14:0] - AccessType
 *
 * T = Type bit
 * S = size bit. Size is in units of 64 bits or 8 bytes.
 * E = Byte end pointer. Used to define the last written byte of the last 64 bits written.
 * D = Data type bit. Used for endian conversion in the PCI interface.
 * C = Used by the cache to decide if this access should allocate a cache line.
 * d = Set if dma read or write data data. This is used to guarantee order at the PCI interface.
 * A = Access type used to check permissions by the MMU in a virtual access.
 * P = Part Write. If set some byte enables may be used. Effects the action of a cache miss.
 */

/* FSR[7:0] */
/* bit 7 => virtual write */
#define AT_VirtualWriteAccBit		(1 << 7)		/* AAADDdC1EEESSSS = Virtual Write */
#define AT_VirtualWriteSizeMask		0xf			/* size of write access (0 => 128 bytes) */
#define AT_VirtualWriteEndPtrShift	4			/* end byte pointer for part write block */
#define AT_VirtualWriteEndPtrMask	0x7

/* else bit 6 => virtual read */
#define AT_VirtualReadAccBit		(1 << 6)		/* AAADDdC01SSSSSS = Virtual Read */
#define AT_VirtualReadSizeMask		0x3f			/* size of read access (0 => 512 bytes) */

/* else => special access */
#define AT_SelBitsMask			0xf			/* Bits to select the type of acces from */
#define AT_SelBitsShift			0x4
#define AT_SpecialRd			(0x0 << 4)		/* AAADDdC0000TTTT = Special read Access */
#define AT_SpecialWr			(0x1 << 4)		/* AAADDdC0001TTTT = Special write Access */
#define AT_PhysicalRd			(0x2 << 4)		/* AAADDdC00100SSS = Physical Read */
#define AT_PhysicalWr			(0x3 << 4)		/* AAADDdC0011PSSS = Physical write */

#define AT_OtherSizeMask	 	0xf			/* Size bits used by all other accesses. 0=128 bytes */
#define AT_SpecialBitsMask	 	0xf			/* Bits used to define the special access types */
#define AT_CacheSizeBitsMask 		0x7			/* Size bits used for local accesses. 0=64 */
#define AT_CachePhysPartWriteBit	0x8			/* This bit is set if the access is a part write to the cache */

/* Special memory access operations */
#define AT_RegAccess			0x0
#define AT_GetCntxFilter		0xe			/* Only used by special reads */
#define AT_RouteFetch			0xf			/* Only used by special reads */

/* FSR[9:8] */
#define AT_NonAlloc			(1 << 8)		/* 1=Do not fill cache with this data */
#define AT_DmaData			(1 << 9)		/* This is a DMA read access. Required to guarantee dma read order. */

/* FSR[11:10] - Data Type - defines data type for endian conversion in PCI interface*/
#define AT_BlkDataTyMask		0x3
#define AT_BlkDataTyShift		10

#define AT_BlkDataType(FSR)		(((FSR) >> AT_BlkDataTyShift) & AT_BlkDataTyMask)
#define AT_TypeByte			0x0
#define AT_TypeHWord			0x1
#define AT_TypeWord			0x2
#define AT_TypeDWord			0x3

/* FSR[14:12] - Access Permissions */
#define AT_PermBitsMask			0x7
#define AT_PermBitsShift		12

#define AT_Perm(FSR)			(((FSR) >> AT_PermBitsShift) & AT_PermBitsMask)
#define AT_PermLocalDataRead		0x0
#define AT_PermLocalDataWrite		0x1
#define AT_PermRemoteRead		0x2
#define AT_PermRemoteWrite		0x3
#define AT_PermExecute			0x4
#define AT_PermLocalEvent		0x5
#define AT_PermRemoteEvent		0x7

/* FSR[22:15] - reason for fault */

#define FSR_WalkForThread		(1 << 15) /* The thread processor caused the fault */
#define FSR_Walking			(1 << 16) /* The fault was caused during a hash table access */
#define FSR_NoTranslationsFound		(1 << 17) /* The hash table did not contain a matching tag */
#define FSR_WalkingProtectionFault	(1 << 18) /* A protection fault was detected while walking */
#define FSR_HashTable1			(1 << 19) /* Was accessing hash table 1 not 0 */
#define FSR_RouteVProcErr		(1 << 20) /* This is an invalid vproc for a route fetch */
#define FSR_FaultForBadData		(1 << 21) /* Bad data (double bit ECC error) while performing a walk access */
#define FSR_FaultForMaxChainCount	(1 << 22) /* The Elan4 has walked a chain of 1024 items. */

typedef volatile struct _E4_FaultSave
{
    E4_uint64 FSRAndFaultContext;                 /* Bits 0-31 : FaultContext. Bits 32-63 : FaultStatus Register */
    E4_uint64 FaultAddress;
} E4_FaultSave;

#define FaultSaveContext(FSRAndFaultContext)	((E4_uint32) ((FSRAndFaultContext) & 0xFFFFFFFF))
#define FaultSaveFSR(FSRAndFaultContext)	((E4_uint32) ((FSRAndFaultContext) >> 32))

typedef union E4_TrTypeCntx
{
   E4_uint32 TypeContext;
   struct
   {
#if (BYTE_ORDER == LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__)
      E4_uint32 Type:16;		/* Transaction type field */
      E4_uint32 Context:13;		/* Transaction context */
      E4_uint32 TypeCntxInvalid:1;	/* Bit  29 */
      E4_uint32 StatusRegValid:1;	/* Bit  30 */
      E4_uint32 LastTrappedTrans:1;	/* Bit  31 */
#else
      E4_uint32 LastTrappedTrans:1;	/* Bit  31 */
      E4_uint32 StatusRegValid:1;	/* Bit  30 */
      E4_uint32 TypeCntxInvalid:1;	/* Bit  29 */
      E4_uint32 Context:13;		/* Transaction context */
      E4_uint32 Type:16;		/* Transaction type field */
#endif
   } s;
} E4_TrTypeCntx;

#define MAX_TRAPPED_TRANS	28
#define TRANS_DATA_DWORDS	16
#define TRANS_DATA_BYTES	128
#define NO_OF_INPUT_CHANNELS	4

#define CH0_LOW_PRI_CHAN	0
#define CH1_LOW_PRI_CHAN	1
#define CH0_HIGH_PRI_CHAN	2
#define CH1_HIGH_PRI_CHAN	3

/* Words have been swapped for big endian access when fetched with dword access from elan.*/
typedef struct _E4_IprocTrapHeader
{
   E4_uint64	TrAddr;
   E4_uint64	IProcStatusCntxAndTrType;
} E4_IprocTrapHeader;

typedef struct _E4_IprocTrapData
{
   E4_uint64 Data[TRANS_DATA_DWORDS];
} E4_IprocTrapData;

/*
 * This struct defines the trap state for the inputers. It requires a contiguous 16K byte block of local memory.
 * The channel bits have been grouped to the low end of the address to force all Identify cookies to use the
 * same cache line.
 */
typedef struct _E4_IprocTrapState
{
   E4_IprocTrapData   TrData[MAX_TRAPPED_TRANS][NO_OF_INPUT_CHANNELS];
   E4_IprocTrapHeader TrHeader[MAX_TRAPPED_TRANS][NO_OF_INPUT_CHANNELS];
   E4_uint64	      pad[8*NO_OF_INPUT_CHANNELS];
} E4_IprocTrapState;

/*
 * 64 kbytes of elan local memory. Must be aligned on a 64k boundary
 */
#define E4_LowPriQueueSize	0x400
#define E4_HighPriQueueSize	0x100

typedef struct _E4_FaultSaveArea
{
   E4_FaultSave		TProcData[8];
   E4_FaultSave		TProcInst;
   E4_FaultSave		Dummy[7];
   E4_FaultSave		SchedProc;
   E4_FaultSave		DProc;
   E4_FaultSave		EventProc;
   E4_FaultSave		IProc;
   E4_FaultSave		DProcData[4];
   E4_FaultSave		QReadData[8];
} E4_FaultSaveArea;

/* Macros to manipulate event queue pointers */
/*     generate index in EventIntQueue */
#define E4_EVENT_INTQ_INDEX(fptr)	(((fptr) & 0x1fff) >> 3)
/*     generate next fptr */
#define E4_EVENT_INTQ_NEXT(fptr)	((((fptr) + 8) & ~0x4000) | 0x2000)

typedef struct _E4_CommandPort
{
   volatile E4_uint64 Command[1024];	/* a whole 8k page */
} E4_CommandPort;

/*
 * This is the allocation of unit numbers within the ELAN. It is used to extract the fault address
 * and fault type after a unit has trapped on a memory fetch. Only units that can generate traps
 * have been included.
 */
#define CUN_TProcData0		0x00
#define CUN_TProcData1		0x01
#define CUN_TProcData2		0x02
#define CUN_TProcData3		0x03
#define CUN_TProcData4		0x04
#define CUN_TProcData5		0x05
#define CUN_TProcData6		0x06
#define CUN_TProcData7		0x07
#define CUN_TProcInst		0x08

/* memory current unit numbers
 * TProc data bus */
#define CUN_DProcPA0		0x10
#define CUN_DProcPA1		0x11
#define CUN_DProcPrefetch	0x12
#define CUN_CommandProc		0x13
#define CUN_DProcData0		0x14	/* Dma prefetch reads. */
#define CUN_DProcData1		0x15	/* Dma prefetch reads. */
#define CUN_DProcData2		0x16	/* Dma prefetch reads. */
#define CUN_DProcData3		0x17	/* Dma prefetch reads. */

#define CUN_IProcLowPri		0x18
#define CUN_IProcHighPri	0x19
#define CUN_Spare0		0x1A
#define CUN_Spare1		0x1B
#define CUN_Spare2		0x1C
#define CUN_ThreadQueue		0x1D
#define CUN_EventProc0		0x1e
#define CUN_EventProc1		0x1f

#define CUN_Entries		0x20

typedef struct E4_Registers
{
   E4_CacheTags		Tags;				/* 4k bytes  c000 -> cfff */
   E4_DataBusMap	Regs;				/* 4k bytes  d000 -> dfff */
   E4_User_Regs		uRegs;				/* 8k bytes  e000 -> ffff */
} E4_Registers;

#define I2cCntl_I2cPortWrite		(0 << 0)
#define I2cCntl_I2cPortRead		(1 << 0)
#define I2cCntl_I2cPortGenStopBit	(1 << 1)
#define I2cCntl_I2cPortGenRestartBit	(1 << 2)
#define I2cCntl_I2cPortAccFailed	(1 << 3)
#define I2cCntl_I2cStopped		(1 << 4)
#define I2cCntl_I2cWakeupFailed		(1 << 5)
#define I2cCntl_I2cFastMode		(1 << 6)
#define I2cCntl_I2cPortBusy		(1 << 7)

#define I2cCntl_LedI2cRegBase_Mask	0x7f
#define I2cCntl_I2cUpdatingLedReg	(1 << 7)

#define I2cCntl_InvertLedValues		(1 << 0)		/* read/write */
#define I2cCntl_LedRegWriteFailed	(1 << 1)		/* read only */
#define I2cCntl_EEPromLoadFailed	(1 << 2)		/* read only */
#define I2cCntl_InhibitI2CRom		(1 << 3)		/* read only */
#define I2cCntl_BadRomCrc		(1 << 4)		/* read only */
#define I2cCntl_MapInI2cConfigData	(1 << 5)		/* read/write */
#define I2cCntl_SampleNewLedValues	(1 << 6)		/* read/write */
#define I2cCntl_ClearLinkError		(1 << 7)		/* write only */

typedef struct E4_I2C
{
   volatile E4_uint8    I2cWrData;
   volatile E4_uint8    I2cRdData;
   volatile E4_uint8    I2cPortControl;
   volatile E4_uint8	I2cLedBase;
   volatile E4_uint8    I2cStatus;
   volatile E4_uint8    I2cLedsValue;
   volatile E4_uint16	I2cPad;
 
   E4_uint8		pad[256 - sizeof(E4_uint64)];

   E4_uint8		UnchangedElan4ConfigRegs[256];
   E4_uint8		I2cRomConfigShadowValues[256];
   E4_uint8		ChangedElan4ConfigRegs[256];
} E4_I2C;

typedef struct _E4_ContextControlBlock 
{
    E4_uint32 Filter;			/* Use a Network context to index for this value */
    E4_uint32 VirtualProcessTable;	/* Use a local context to index for this value */
} E4_ContextControlBlock;

/*
 * Filter
 *   [13:0]	Context
 *   [14]	DiscardAll
 *   [15]	AckAll
 *   [16]	HighPri
 *   [17]	CountStats
 *   [31:18]	Unused
 */
#define E4_FILTER_STATS		(1 << 17)
#define E4_FILTER_HIGH_PRI	(1 << 16)
#define E4_FILTER_ACKOK_ALL	(1 << 15)
#define E4_FILTER_DISCARD_ALL	(1 << 14)
#define E4_FILTER_CONTEXT_MASK	(0x3FFF)

/*
 * VirtualProcessTable
 *   [8:0]	Unused	
 *   [12:9]	Size       num vp entries = 512 << Size
 *   [30:13]	Pointer
 *   [31]	Valid
 */
#define E4_VPT_MIN_ENTRIES      512
#define E4_VPT_VALID		((unsigned)1 << 31)
#define E4_VPT_PTR_SHIFT	0
#define E4_VPT_SIZE_SHIFT	9
#define E4_VPT_SIZE_MASK        0xf
#define E4_VPT_NUM_VP(vpt_val)  (E4_VPT_MIN_ENTRIES << (((vpt_val) >> E4_VPT_SIZE_SHIFT) & E4_VPT_SIZE_MASK))
#define E4_VPT_VALUE(ptr,size)	(((ptr) << E4_VPT_PTR_SHIFT) | ((size) << E4_VPT_SIZE_SHIFT))


/* Virtual Process Table */
typedef struct _E4_VirtualProcessEntry
{
    E4_uint64	Values[2];
} E4_VirtualProcessEntry;

/*
 * Entries have the following format - rtX is a packed route 
 *
 * |rt11|rt10|rt9 |rt8 |rt7 |rt6 |rt5 |rt4 |rt3 |rt2 |rt2 |rt0 |PAAADD       RRRRRR|
 * |output context     |rt23|rt22|rt21|rt20|rt19|rt18|rt17|rt16|rt15|rt14|rt13|rt12|
 */

#define ROUTE_CTXT_SHIFT	48
#define ROUTE_CTXT_MASK		(~((1ull << ROUTE_CTXT_SHIFT)-1))
#define ROUTE_CTXT_VALUE(ctx)	(((E4_uint64) ctx) << ROUTE_CTXT_SHIFT)

#define ROUTE_PACKED_OFFSET	16
#define ROUTE_NUM_PACKED	24

/* defines for first flit of a route */
#define FIRST_TIMEOUT(Val)	((Val) << 14)			/* [15:14]  */
#define FIRST_SYSTEM_PACKET     (1 << 13)                       /* [13]     */
#define FIRST_FLOOD_PACKET      (1 << 12)                       /* [12]     */
#define FIRST_HIGH_PRI		(1 << 11)			/* [11]    */
#define FIRST_AGE(Val)		((Val) << 7)			/* [10:7] */
#define FIRST_OPTIONS_MASK	(0xFF80)

/* [6:0] unpacked 1st route value */
#define FIRST_INVALID		(0)
#define FIRST_ROUTE(Val)	(0x08 | (Val))
#define FIRST_ADAPTIVE		(0x30)
#define FIRST_BCAST_TREE	(0x20)
#define FIRST_MYLINK		(0x10)
#define FIRST_BCAST(Top, Bot)	(0x40 | ((Top) << 3) | (Bot))

/* defines for 3 bit packed entries for subsequent flits */
#define PACKED_INVALID		(0)
#define PACKED_ROUTE(Val)	(8 | (Val))
#define PACKED_ADAPTIVE		(3)
#define PACKED_BCAST_TREE	(2)
#define PACKED_MYLINK		(1)
#define PACKED_BCAST0(Top,Bot)	(4 | (Bot & 3))
#define PACKED_BCAST1(Top,Bot)	((Top << 1) | (Bot >> 2))

#endif /* _ASM */
/* The MMU root context pointer has a mask to bounds check 
 * it - this is computed as follows.
 */
#define E4_CONTEXT_MASK(num)	(((num) >= 0x2000) ? 0x00 :	\
 			         ((num) >= 0x1000) ? 0x80 :	\
 			         ((num) >= 0x0800) ? 0xc0 :	\
 			         ((num) >= 0x0400) ? 0xe0 :	\
 			         ((num) >= 0x0200) ? 0xf0 :	\
 			         ((num) >= 0x0100) ? 0xf8 :	\
 			         ((num) >= 0x0080) ? 0xfc :	\
 			         ((num) >= 0x0040) ? 0xfe : 0xff)
/*
 * This generates the size field for a virtual process table.
 * Size defined as 2^n no of 8K pages.
 * Single cycle route fetches are possible if the minimum vproc table size is 8k.
 */
#define E4_GEN_VPT_SIZE(Size) 	(((Size) & E4_VPT_SIZE_MASK) << E4_VPT_SIZE_SHIFT)

#define COMMAND_RUN_QUEUE_BITS		(13 + 2) /* 8K entries of 4 bytes. This is fixed in hardware. */
#define COMMAND_DESCS_SPACE_BITS	(13 + 5) /* 8K entries of 32 bytes. This is fixed in hardware. */
#define COMMAND_INSERTER_CACHE_ENTRIES	16

#define COM_TEST_PORT_ADDR_MASK		0xfULL
#define COM_TEST_PORT_ADDR_SH		0

/*
 * The flush register is accessed through the CommandControl register.
 * The address is naturally alligned. It also positions the command descriptors in memory.
 * When no command queues need flushing it should be or with COM_FLUSH_INVALID. This sets
 * it to the top command queue descriptor. This cannot be accessed from the PCI.
 */
#define COM_ENABLE_DEQUEUE		(1 << 4)
#define COM_FLUSH_DESCRIPTOR_MASK	0x7fffffe0ULL
#define COM_FLUSH_INVALID		0x0003ffe0ULL


/*
 * Elan4 BAR1 is split up as follows :
 *
 * RevA
 *     0x3f00000 EBUS other
 *     0x3e00000 EBUS ROM
 *     0x3dfc000 registers
 *     0x0000000 command ports
 *
 * RevB
 *     0x3ffc000 registers
 *     0x3ff8000 padding
 *     0x3ff6000 i2c registers
 *     0x0000000 command ports
 */
#define ELAN4_BAR1_SIZE			(1 << 26)	/* 64M */
#define ELAN4_REG_SIZE			(1 << 14)	/* 16K */

#define ELAN4_REVA_EBUS_SIZE		(1 << 21)	/* 2M */
#define ELAN4_REVA_EBUS_OFFSET		(ELAN4_BAR1_SIZE - ELAN4_REVA_EBUS_SIZE)
#define ELAN4_REVA_REG_OFFSET		(ELAN4_REVA_EBUS_OFFSET - ELAN4_REG_SIZE)
#define ELAN4_REVA_NUM_COMMAND_QUEUES	(ELAN4_REVA_REG_OFFSET >> 13)

#define ELAN4_REVA_EBUS_ROM_SIZE	(1 << 20)	/* 1M */
#define ELAN4_REVA_EBUS_ROM_OFFSET	0

#define ELAN4_REVB_I2C_PADDING		(1 << 14)	/* 16K */
#define ELAN4_REVB_I2C_SIZE		(1 << 13)	/* 8k */
#define ELAN4_REVB_REG_OFFSET		(ELAN4_BAR1_SIZE - ELAN4_REG_SIZE)
#define ELAN4_REVB_I2C_OFFSET		(ELAN4_REVB_REG_OFFSET - ELAN4_REVB_I2C_PADDING - ELAN4_REVB_I2C_SIZE)
#define ELAN4_REVB_NUM_COMMAND_QUEUES	(ELAN4_REVB_I2C_OFFSET >> 13)

#endif /* notdef _ELAN4_REGISTERS_H */
