/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: trap.h,v 1.10 2003/10/07 12:11:10 david Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/trap.h,v $*/

#ifndef __ELAN4_TRAP_H
#define __ELAN4_TRAP_H

/*
 * If the EProc Faults whilst performing an action (e.g. Read/Write on the data src or dest Addr)
 *  the Eproc increments the Addr(s) by a block size (64 bytes):
 *  1: Fault on Read: 
 *                     Src EventAddr = Read Addr + block
 *  2: Fault on Write:
 *                     Src EventAddr = Read Addr + block
 *                     Dst EventAddr = Read Addr + block
 *                     Size          = Size - block ndwords
 *  We must rewind the addr correctly to completely the transfer successfully
 */
#define EVENT_COPY_NDWORDS	0x8
#define EVENT_COPY_BLOCK_SIZE	0x40

typedef struct elan4_eproc_trap
{
    E4_uint64		tr_status;
    E4_FaultSave	tr_faultarea;
    E4_Event		tr_event;
    E4_Addr	        tr_eventaddr;
} ELAN4_EPROC_TRAP;

typedef struct elan4_cproc_trap
{
    E4_uint64		tr_status;					/* cproc status register */
    E4_uint64		tr_command;					/* cproc command */
    E4_CommandQueueDesc tr_qdesc;					/* copy of command queue descriptor */
    E4_FaultSave	tr_faultarea;					/* fault area for mmu traps */
    ELAN4_EPROC_TRAP	tr_eventtrap;					/* associated event trap (waitevent) */
} ELAN4_CPROC_TRAP;

typedef struct elan4_dproc_trap
{
    E4_DMA		tr_desc;
    E4_FaultSave	tr_packAssemFault;
    E4_FaultSave	tr_prefetchFault;
    E4_uint64		tr_status;
} ELAN4_DPROC_TRAP;

typedef struct elan4_tproc_trap
{
    E4_uint64		tr_regs[64];
    E4_FaultSave	tr_dataFault;
    E4_FaultSave	tr_instFault;
    E4_uint64		tr_status;
    E4_uint64		tr_state;
    E4_Addr		tr_pc;
    E4_Addr		tr_npc;
    E4_uint64		tr_dirty;
    E4_uint64		tr_bad;
} ELAN4_TPROC_TRAP;

typedef struct elan4_iproc_trap
{
    E4_uint32	          tr_numTransactions;
    E4_uint32		  tr_flags;
    E4_uint32		  tr_trappedTrans;
    E4_uint32		  tr_waitForEopTrans;
    E4_uint32		  tr_identifyTrans;
    E4_uint32		  tr_pad;

    E4_FaultSave          tr_faultarea;
    E4_IprocTrapHeader    tr_transactions[MAX_TRAPPED_TRANS];
    E4_IprocTrapData      tr_dataBuffers[MAX_TRAPPED_TRANS];
} ELAN4_IPROC_TRAP;

#define TR_FLAG_ACK_SENT	(1 << 0)
#define TR_FLAG_EOP_ERROR	(1 << 1)
#define TR_FLAG_BAD_TRANS	(1 << 2)
#define TR_FLAG_DMA_PACKET	(1 << 3)
#define TR_FLAG_EOP_BAD		(1 << 4)
#define TR_FLAG_TOOMANY_TRANS	(1 << 5)

#define TR_TRANS_INVALID	(0xffffffff)

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_TRAP_H */
