/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _ELAN4_TRTYPE_H
#define _ELAN4_TRTYPE_H

#ident "$Id: trtype.h,v 1.20 2004/02/06 10:38:21 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/trtype.h,v $*/

/*<15:11> Size field is used to give the number of additional 64 bit data values.
	  A value from 0 to 16 inclusive is valid. */

#include <elan4/types.h>

#define TR_SIZE_SHIFT		(11)
#define TR_SIZE_MASK		(0x1f << TR_SIZE_SHIFT)
#define SET_TR_SIZE(Size)	(((Size) << TR_SIZE_SHIFT) & TR_SIZE_MASK)

/* <10:9> Last Transaction and AckNow bits, marks the last transaction and
          enables a PACK_OK to be sent. */
#define TR_LAST_AND_SEND_ACK	(3 << 9)


/* <8>  Only valid on the last transaction. Delays execution until an EOP_GOOD is received.
 *      Any other EOP type will abort execution of this transaction. */
#define TR_WAIT_FOR_EOP		(1 << 8)

/*
 * Data type. This is used by transactions of variable data type. It controls any endian
 * converion required if the destiantion host processor has a big endian memory format.
 */
/*	WriteBlock 	<8:7>	Data type
			<6:0> 	Part write size */
#define TR_DATATYPE_SHIFT	(6)
#define TR_DATATYPE_MASK	((1 << 2) - 1)

#define TR_DATATYPE_BYTE	E4_DATATYPE_BYTE	
#define TR_DATATYPE_SHORT	E4_DATATYPE_SHORT
#define TR_DATATYPE_WORD	E4_DATATYPE_WORD	
#define TR_DATATYPE_DWORD	E4_DATATYPE_DWORD

/* <5:0> Transaction Type
 *       For Writeblock <5:3> 000 => Write, 0001 => Read
 *                      <2:0> End Byte Addr */
#define TR_OPCODE_MASK		0x3F
#define TR_BLOCK_OPCODE_MASK	0x38

#define TR_WRITEBLOCK		0x0
#define TR_ENDBYTE_MASK		0x7
#define TR_WRITE(Size, EndByte, DataType)						\
			(0x0 | SET_TR_SIZE(Size) | ((EndByte) & TR_ENDBYTE_MASK) |	\
			 (((DataType) & TR_DATATYPE_MASK) << TR_DATATYPE_SHIFT))

#define TR_NOP_TRANS		(0x10 | SET_TR_SIZE(0))
#define TR_SETEVENT		0x10
#define TR_SETEVENT_NOIDENT	(TR_SETEVENT | SET_TR_SIZE(0) | TR_LAST_AND_SEND_ACK)
#define TR_SETEVENT_IDENTIFY	(TR_SETEVENT | SET_TR_SIZE(1) | TR_LAST_AND_SEND_ACK)
#define TR_REMOTEDMA		(0x11 | SET_TR_SIZE(7) | TR_LAST_AND_SEND_ACK)
#define TR_SENDDISCARD		(0x12 | SET_TR_SIZE(0))

/*
 * Conditional transactions that might return PAckTestFail.
 * All will allow further exection of the packet if ([Address] operator DataValue) is true.
 * e.g. for TR_GTE further execution if ([Address] >= DataValue) is true.
 * These should be used where a definite TRUE/FALSE answer is required.
 */
#define TR_GTE			(0x14 | SET_TR_SIZE(1))
#define TR_LT			(0x15 | SET_TR_SIZE(1))
#define TR_EQ			(0x16 | SET_TR_SIZE(1))
#define TR_NEQ			(0x17 | SET_TR_SIZE(1))

/*
 * Conditional transactions that might return PAckDiscard.
 * All will allow further exection of the packet if ([Address] operator DataValue) is true.
 * e.g. for TR_GTE further execution if ([Address] >= DataValue) is true.
 * These should be used where eventually a TRUE answer is expected but the node might not be ready yet.
 * These can be mixed with the normal conditionals to allow a single packet to test for readyness and
 * a TRUE/FALSE answer.
 */
#define TR_GTE_DISCARD		(0x34 | SET_TR_SIZE(1))
#define TR_LT_DISCARD		(0x35 | SET_TR_SIZE(1))
#define TR_EQ_DISCARD		(0x36 | SET_TR_SIZE(1))
#define TR_NEQ_DISCARD		(0x37 | SET_TR_SIZE(1))

#define TR_TRACEROUTE_TRANS	0x18
#define TR_TRACEROUTE(Size)	(TR_TRACEROUTE_TRANS | (TR_DATATYPE_WORD << TR_DATATYPE_SHIFT) |SET_TR_SIZE(Size))
#define TR_IDENTIFY		(0x19 | SET_TR_SIZE(0))

#define TR_ADDWORD		(0x1c | SET_TR_SIZE(2) | TR_LAST_AND_SEND_ACK)
#define TR_INPUT_Q_COMMIT	(0x1d | SET_TR_SIZE(1) | TR_LAST_AND_SEND_ACK)
#define TR_TESTANDWRITE 	(0x1e | SET_TR_SIZE(3) | TR_LAST_AND_SEND_ACK)
#define TR_INPUT_Q_GETINDEX	(0x1f | SET_TR_SIZE(0))



/* TraceRoute formate */
#define TR_TRACEROUTE0_CHANID(val)		((val) & 1)			/* 0     Chan Id */
#define TR_TRACEROUTE0_LINKID(val)		(((val) >> 1) & 7)		/* 1:3   Link Id */
#define TR_TRACEROUTE0_REVID(val)		(((val) >> 4) & 7)		/* 4:6   Revision Id */
#define TR_TRACEROUTE0_BCAST_PIN(val)		(((val) >> 7) & 1)		/* 7     Bcast Top Pin */
#define TR_TRACEROUTE0_LNR(val)			(((val) >> 8) & 0xFF)		/* 8:15  Global Link Not Ready */

#define TR_TRACEROUTE1_ROUTES_SELECTED(val)	((val & 0xFF))			/* 0:7   Routes Selected */
#define TR_TRACEROUTE1_BCAST_TOP(val)		(((val) >> 8) & 7)		/* 8:10  Broadcast Top */
#define TR_TRACEROUTE1_BCAST_BOTTOM(val)	(((val) >> 12) & 7)		/* 12:14 Broadcast Bottom */

#endif /* _ELAN4_TRANSACTIONTYPE_H */
