/*
 *    Copyright (c) 2001-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: user.h,v 1.45 2005/04/21 11:12:06 mike Exp $"
/*      $Source: /cvs/master/quadrics/elan4mod/user.h,v $*/

#ifndef __ELAN4_USER_H
#define __ELAN4_USER_H

#include <elan/capability.h>
#include <elan4/usertrap.h>
#include <elan4/intcookie.h>

typedef struct trap_queue
{
    unsigned	q_back;			/* Next free space */
    unsigned	q_front;		/* First object to remove */
    unsigned	q_size;			/* Size of queue */
    unsigned	q_count;		/* Current number of entries */
    unsigned	q_slop;			/* FULL <=> (count+slop) == size */
} RING_QUEUE;

#define RING_QUEUE_INIT(q,num,slop)	((q).q_size = (num), (q).q_slop = (slop), (q).q_front = (q).q_back = 0, (q).q_count = 0)
#define RING_QUEUE_FULL(q)		((q).q_count >= ((q).q_size - (q).q_slop))
#define RING_QUEUE_REALLY_FULL(q)	((q).q_count == (q).q_size)
#define RING_QUEUE_EMPTY(q)		((q).q_count == 0)
#define RING_QUEUE_NEXT(q,indx)		((indx) = (((indx)+1) % (q).q_size))
#define RING_QUEUE_PREV(q,indx)		((indx) = (((indx)+(q).q_size-1) % (q).q_size))
#define RING_QUEUE_ADD(q)		(RING_QUEUE_NEXT(q ,(q).q_back),  (++(q).q_count) >= ((q).q_size - (q).q_slop))
#define RING_QUEUE_REMOVE(q)		(RING_QUEUE_NEXT(q, (q).q_front), (--(q).q_count) == 0)
#define RING_QUEUE_ADD_FRONT(q)		(RING_QUEUE_PREV(q, (q).q_front), (++(q).q_count) >= ((q).q_size - (q).q_slop))
#define RING_QUEUE_ENTRY(qArea,indx)	(&(qArea)[(indx)])
#define RING_QUEUE_FRONT(q,qArea)	RING_QUEUE_ENTRY(qArea, (q).q_front)
#define RING_QUEUE_BACK(q,qArea)	RING_QUEUE_ENTRY(qArea, (q).q_back)
#define RING_QUEUE_ITERATE(q,idx)	for (idx = (q).q_front; idx != (q).q_back; idx = (((idx) + 1) % (q).q_size))

typedef struct user_rgn
{
    struct user_rgn *rgn_mnext;					/* Doubly linked list of regions */
    struct user_rgn *rgn_mprev;					/*   sorted on main address */ 
    virtaddr_t       rgn_mbase;					/* main address of base of region */

    struct user_rgn *rgn_enext;					/* Doubly linked list of regions */
    struct user_rgn *rgn_eprev;					/*   sorted on elan address */
    E4_Addr	     rgn_ebase;					/* elan address of base of region */

    unsigned long    rgn_len;					/* length of region */
    unsigned	     rgn_perm;					/* elan access permission */
} USER_RGN;

typedef struct user_vpseg
{ 
    struct list_head	vps_link;

    unsigned short	vps_process;				/* virtual process number */
    unsigned short	vps_entries;				/*   and # virtual processes */

    unsigned 		vps_type;
    union
    {
	struct {
	    ELAN_CAPABILITY	   *cap;
	    E4_VirtualProcessEntry *routes;
	} p2p;
#define vps_p2p_cap	vps_u.p2p.cap
#define vps_p2p_routes  vps_u.p2p.routes

	struct {
	    unsigned short lowvp;
	    unsigned short highvp;
	} bcast;
#define vps_bcast_lowvp		vps_u.bcast.lowvp
#define vps_bcast_highvp	vps_u.bcast.highvp
    } vps_u;
} USER_VPSEG;

/* values for vps_type */
#define USER_VPSEG_P2P		0
#define USER_VPSEG_BCAST	1

typedef struct user_cq
{
    struct list_head ucq_link;

    ELAN4_CQ	    *ucq_cq;					/* the real command queue */

    unsigned char    ucq_state;					/* command queue state */
    unsigned char    ucq_errored;				/* command queue has errored */
    unsigned char    ucq_flags;					/* flags */
    ELAN4_CPROC_TRAP ucq_trap;					/* trap state */

    atomic_t	     ucq_ref;					/* # references to this cq (mmaps) */
} USER_CQ;

/* values for ucq_state */
#define UCQ_RUNNING	                 0			/* command queue is running */
#define UCQ_TRAPPED                  	 1			/* command queue has trapped */
#define UCQ_NEEDS_RESTART                2			/* command queue has trapped, and needs restarting */
#define UCQ_STOPPED			 3			/* command queue has trapped, and delivered to user */

/* values for ucq_flags */
#define UCQ_SYSTEM		(1 << 0)
#define UCQ_REORDER		(1 << 1)

extern int num_fault_save;
extern int min_fault_pages;
extern int max_fault_pages;

typedef struct fault_save
{
    struct fault_save           *next;
    E4_Addr                      addr;
    E4_uint32                    count;
} FAULT_SAVE;

typedef struct user_iproc_trap
{
    unsigned char     ut_state;
    ELAN4_IPROC_TRAP  ut_trap;
} USER_IPROC_TRAP;

/* values for ut_state */
#define UTS_IPROC_RUNNING			0
#define UTS_IPROC_TRAPPED			1
#define UTS_IPROC_RESOLVING			2
#define UTS_IPROC_EXECUTE_PACKET		3
#define UTS_IPROC_EXECUTING			4
#define UTS_IPROC_NETWORK_ERROR			5
#define UTS_IPROC_STOPPED			6

typedef struct user_ctxt_entry
{
    struct list_head	 cent_link;					/* entry chained on context */
    ELAN_CAPABILITY  	*cent_cap;					/* capability we attached with */
} USER_CTXT_ENTRY;

typedef struct user_ctxt
{
    ELAN4_CTXT         uctx_ctxt;				/* is also an elan context */

    spinlock_t	       uctx_spinlock;				/* spinlock for items used with interrupt handler */
    kcondvar_t	       uctx_wait;				/* place to sleep (traphandler/swapout/swapin/neterr fixup) */

    unsigned	       uctx_status;				/* status				(uctx_spinlock) */

    pid_t	       uctx_trap_pid;				/* pid to deliver signals to on trap */
    int		       uctx_trap_signo;				/* signal number to deliver */
    unsigned	       uctx_trap_state;				/* state of trap handling code */
    unsigned	       uctx_trap_count;				/* count of "thread" in user_trap_handler() */

    unsigned	       uctx_int_count;				/* # interrupts since last zeroed */
    unsigned long      uctx_int_start;				/* tick when int_count last zeroed */
    unsigned long      uctx_int_delay;				/* # ticks to delay next wakeup */
    struct timer_list  uctx_int_timer;				/* and timer to use to delay signal */
    struct timer_list  uctx_shuffle_timer;			/* and timer to use to delay shuffle signal */

    struct timer_list  uctx_neterr_timer;			/* network error timer */

    struct list_head   uctx_vpseg_list;				/* list of vp segments we've got */
    kmutex_t           uctx_vpseg_lock;				/*   and lock to protect it. */
    ELAN4_ROUTE_TABLE *uctx_routetable;				/* our virtual process table */
    ELAN_POSITION      uctx_position;				/* position in network */

    struct list_head   uctx_cent_list; 				/* list of attached network contexts */

    USER_CQ	      *uctx_ddcq;				/* command queue for re-issueing traps */
    E4_uint64	       uctx_ddcq_insertcnt;			/* # dwords inserted into command queue */
    E4_uint64          uctx_ddcq_completed;			/* last "completed" write was here */
    int		       uctx_ddcq_intr;				/* count of outstanding ddcq interrupts */

    ELAN4_HALTOP       uctx_haltop;				/* halt operation for flushing */
    ELAN4_DMA_FLUSHOP  uctx_dma_flushop;			/* flush operation for flushing dma runqueue */

    INTCOOKIE_TABLE   *uctx_intcookie_table;			/* table of interrupt cookies (shared with other uctxs for this task) */

    kmutex_t	       uctx_cqlock;				/* lock for create/destory cqs */
    struct list_head   uctx_cqlist;				/* list of command queues 		(uctx_cqlock,uctx_spinlock) */

    ELAN4_DPROC_TRAP  *uctx_dprocTraps;				/* queue of dproc traps to resolve/reissue */
    RING_QUEUE	       uctx_dprocTrapQ;

    ELAN4_TPROC_TRAP  *uctx_tprocTraps;				/* queue of tproc traps to resolve/reissue */
    RING_QUEUE         uctx_tprocTrapQ;

    ELAN4_EPROC_TRAP  *uctx_eprocTraps;				/* queue of eproc traps to resolve */
    RING_QUEUE	       uctx_eprocTrapQ;

    USER_IPROC_TRAP    uctx_iprocTrap[2];			/* input trap state, 1 per virtual channel */

    E4_DMA	      *uctx_dmas;				/* queue of dmas to restart */
    RING_QUEUE         uctx_dmaQ;
    
    E4_ThreadRegs     *uctx_threads;				/* queue of threads to restart */
    RING_QUEUE         uctx_threadQ;

    ELAN4_NETERR_MSG  *uctx_msgs;				/* queue of neterr messages */
    RING_QUEUE	       uctx_msgQ;
    kmutex_t	       uctx_rgnmutex;				/* lock for create/destroy regions */
    spinlock_t	       uctx_rgnlock;				/* spinlock to protect linked lists */
    USER_RGN	      *uctx_mrgns;				/* Doubly linked list of memory regions (uctx_rgnlock) */
    USER_RGN	      *uctx_mtail;				/* Last memory region on list 		(uctx_rgnlock) */
    USER_RGN	      *uctx_mrgnlast;				/* Last region 'hit' 			(uctx_rgnlock) */

    USER_RGN	      *uctx_ergns;				/* Doubly linked list of memory regions (uctx_rgnlock) */
    USER_RGN	      *uctx_etail;				/* Last memory region on list		(uctx_rgnlock) */
    USER_RGN	      *uctx_ergnlast;				/* Last region 'hit' 			(uctx_rgnlock) */

    ELAN4_USER_PAGE   *uctx_upage;				/* kernel page shared with user */
    sdramaddr_t	       uctx_trampoline;				/* sdram page for tproc trampoline */

    E4_Addr	       uctx_upage_addr;				/*   elan addr page mapped into */
    E4_Addr	       uctx_trestart_addr;			/* address of thread restart code */
    FAULT_SAVE         *uctx_faults;
    FAULT_SAVE         *uctx_fault_list;
    int                 uctx_num_fault_save;
    spinlock_t          uctx_fault_lock;
} USER_CTXT;

/* bit values for uctx_status */
#define UCTX_EXITING				(1 << 0)		/* context is exiting. */
#define UCTX_USER_FILTERING			(1 << 1)		/* user requested context filter */
#define UCTX_USER_STOPPED			(1 << 2)		/* user requested stop */

#define UCTX_SWAPPING				(1 << 3)		/* context is swapping out */
#define UCTX_SWAPPED				(1 << 4)		/* context is swapped out */

#define UCTX_STOPPING				(1 << 5)		/* stopping elan from running this context */
#define UCTX_STOPPED				(1 << 6)		/* elan no longer running this context */

#define UCTX_EPROC_QUEUE_FULL			(1 << 7)		/* reasons for stopping running */
#define UCTX_DPROC_QUEUE_FULL			(1 << 8)
#define UCTX_TPROC_QUEUE_FULL			(1 << 9)
#define UCTX_IPROC_CH0_TRAPPED			(1 << 10)
#define UCTX_IPROC_CH1_TRAPPED			(1 << 11)

#define UCTX_NETERR_TIMER			(1 << 12)
#define UCTX_NETERR_FIXUP			(1 << 13)

#define UCTX_EPROC_QUEUE_OVERFLOW		(1 << 14)
#define UCTX_DPROC_QUEUE_OVERFLOW		(1 << 15)
#define UCTX_TPROC_QUEUE_OVERFLOW		(1 << 16)

#define UCTX_EPROC_QUEUE_ERROR			(1 << 17)
#define UCTX_DPROC_QUEUE_ERROR			(1 << 18)
#define UCTX_TPROC_QUEUE_ERROR			(1 << 19)

#define UCTX_STOPPED_REASONS			(UCTX_EPROC_QUEUE_FULL | UCTX_DPROC_QUEUE_FULL | UCTX_TPROC_QUEUE_FULL)
#define UCTX_SWAPPED_REASONS			(UCTX_EXITING | UCTX_USER_STOPPED | UCTX_NETERR_FIXUP)
#define UCTX_NACKING_REASONS			(UCTX_USER_FILTERING | UCTX_IPROC_CH0_TRAPPED | UCTX_IPROC_CH1_TRAPPED)

#define UCTX_OVERFLOW_REASONS			(UCTX_EPROC_QUEUE_OVERFLOW | UCTX_DPROC_QUEUE_OVERFLOW | UCTX_TPROC_QUEUE_OVERFLOW)
#define UCTX_ERROR_REASONS			(UCTX_EPROC_QUEUE_ERROR | UCTX_DPROC_QUEUE_ERROR | UCTX_TPROC_QUEUE_ERROR)

#define UCTX_RUNNABLE(uctx)			(((uctx)->uctx_status & (UCTX_SWAPPED_REASONS | UCTX_STOPPED_REASONS)) == 0)
#define UCTX_NACKING(uctx)			(((uctx)->uctx_status & (UCTX_SWAPPED_REASONS | UCTX_STOPPED_REASONS | UCTX_NACKING_REASONS)) != 0)

/* values for uctx_trap_signalled */
#define UCTX_TRAP_IDLE				0
#define UCTX_TRAP_SLEEPING			1
#define UCTX_TRAP_SIGNALLED			2
#define UCTX_TRAP_ACTIVE			3

extern int        user_p2p_route_options;
extern int        user_bcast_route_options;
extern int	  user_dproc_retry_count;
extern int	  user_cproc_retry_count;
extern int	  user_ioproc_enabled;
extern int	  user_pagefault_enabled;

extern USER_CTXT *user_alloc (ELAN4_DEV *dev);
extern void       user_free (USER_CTXT *uctx);
extern void       user_swapout (USER_CTXT *uctx, unsigned reason);
extern void       user_swapin (USER_CTXT *uctx, unsigned reason);
extern int        user_attach (USER_CTXT *uctx, ELAN_CAPABILITY *cap);
extern void       user_detach (USER_CTXT *uctx, ELAN_CAPABILITY *cap);
extern void       user_block_inputter (USER_CTXT *uctx, unsigned blocked);
extern int        user_alloc_trap_queues (USER_CTXT *uctx, unsigned ndproc_traps, unsigned neproc_traps, 
					  unsigned ntproc_traps, unsigned nthreads, unsigned ndmas);

extern int        user_add_p2pvp (USER_CTXT *uctx, unsigned process, ELAN_CAPABILITY *cap);
extern int        user_add_bcastvp (USER_CTXT *uctx, unsigned process, unsigned lowvp, unsigned highvp);
extern int        user_removevp (USER_CTXT *uctx, unsigned process);

extern int        user_set_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route);
extern int        user_reset_route (USER_CTXT *uctx, unsigned process);
extern int        user_get_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route);
extern int        user_check_route (USER_CTXT *uctx, unsigned process, E4_VirtualProcessEntry *route, unsigned *error);
extern int	  user_send_neterr_msg (USER_CTXT *uctx, unsigned int vp, unsigned int nctx, unsigned int retries, ELAN4_NETERR_MSG *msg);
extern int        user_neterr_sten (USER_CTXT *uctx, unsigned int vp, E4_uint64 cookie, int waitforeop);
extern int        user_neterr_dma (USER_CTXT *uctx, unsigned int vp, E4_uint64 cookie, int waitforeop);

extern int        user_resume_eproc_trap (USER_CTXT *uctx, E4_Addr addr);
extern int        user_resume_cproc_trap (USER_CTXT *uctx, unsigned indx);
extern int        user_resume_dproc_trap (USER_CTXT *uctx, E4_DMA *dma);
extern int        user_resume_tproc_trap (USER_CTXT *uctx, E4_ThreadRegs *regs);
extern int        user_resume_iproc_trap (USER_CTXT *uctx, unsigned channel, unsigned trans,
					  E4_IprocTrapHeader *hdrp, E4_IprocTrapData *datap);

extern int        user_trap_handler (USER_CTXT *uctx, ELAN4_USER_TRAP *utrapp, int nticks);
extern USER_CQ   *user_findcq (USER_CTXT *uctx, unsigned num);
extern USER_CQ   *user_alloccq (USER_CTXT *uctx, unsigned size, unsigned perm, unsigned flags);
extern void       user_freecq (USER_CTXT *uctx, USER_CQ *cq);
extern void       user_dropcq (USER_CTXT *uctx, USER_CQ *cq);

/* user_osdep.c */
extern int        user_load_range (USER_CTXT *uctx, E4_Addr addr, unsigned long nbytes, E4_uint32 fsr);
extern void       user_update_main (USER_CTXT *uctx, struct mm_struct *mm, struct vm_area_struct *vma, unsigned long start, unsigned long len);
extern void       user_unload_main (USER_CTXT *uctx, unsigned long start, unsigned long len);


/* regions.c */
extern USER_RGN  *user_findrgn_elan (USER_CTXT *uctx, E4_Addr addr, int tail);
extern USER_RGN  *user_findrgn_main (USER_CTXT *uctx, virtaddr_t addr, int tail);
extern USER_RGN  *user_rgnat_elan (USER_CTXT *uctx, E4_Addr addr);
extern USER_RGN  *user_rgnat_main (USER_CTXT *uctx, virtaddr_t addr);
extern int        user_setperm (USER_CTXT *uctx, virtaddr_t maddr, E4_Addr eaddr, unsigned long len, unsigned perm);
extern void       user_clrperm (USER_CTXT *uctx, E4_Addr addr, unsigned long len);
extern int        user_checkperm (USER_CTXT *uctx, E4_Addr raddr, unsigned long rsize, unsigned access);
extern virtaddr_t user_elan2main (USER_CTXT *uctx, E4_Addr addr);
extern E4_Addr    user_main2elan (USER_CTXT *uctx, virtaddr_t addr);
extern void       user_preload_main (USER_CTXT *uctx, virtaddr_t addr, unsigned long len);
extern void       user_freergns (USER_CTXT *uctx);

/* user_ddcq.c */
extern int        user_ddcq_check (USER_CTXT *uctx, unsigned num);
extern int        user_ddcq_flush (USER_CTXT *uctx);
extern void       user_ddcq_intr (USER_CTXT *uctx);
extern void       user_ddcq_write_dword (USER_CTXT *uctx, E4_Addr addr, E4_uint64 value);
extern void       user_ddcq_interrupt (USER_CTXT *uctx, E4_uint64 cookie);
extern void       user_ddcq_run_dma (USER_CTXT *uctx, E4_DMA *dma);
extern void       user_ddcq_run_thread (USER_CTXT *uctx, E4_ThreadRegs *regs);
extern void       user_ddcq_setevent (USER_CTXT *uctx, E4_Addr addr);
extern void       user_ddcq_seteventn (USER_CTXT *uctx, E4_Addr addr, E4_uint32 count);
extern void       user_ddcq_waitevent (USER_CTXT *uctx, E4_Addr addr, E4_uint64 CountAndType, E4_uint64 Param0, E4_uint64 Param1);


/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
#endif /* __ELAN4_USER_H */
