/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2003 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __ELAN4_USERREGS_H
#define __ELAN4_USERREGS_H

#ident "$Id: userregs.h,v 1.15 2004/10/06 14:50:38 addy Exp $"
/*      $Source: /cvs/master/quadrics/elan4hdr/userregs.h,v $*/

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Statistic control reg values
 * Each 4-bit nibble of the control word specifies what statistic
 * is to be recorded in each of the 8 statistic counters
 */
#define COUNT_REG0_SHIFT   32ull
#define COUNT_REG1_SHIFT   36ull
#define COUNT_REG2_SHIFT   40ull
#define COUNT_REG3_SHIFT   44ull
#define COUNT_REG4_SHIFT   48ull
#define COUNT_REG5_SHIFT   52ull
#define COUNT_REG6_SHIFT   56ull
#define COUNT_REG7_SHIFT   60ull


/* Count reg 0 */
#define STC_INPUT_NON_WRITE_BLOCKS   	(0x0ull << COUNT_REG0_SHIFT)
#define STP_DMA_EOP_WAIT_ACK         	(0x1ull << COUNT_REG0_SHIFT)
#define STP_TPROC_RUNNING           	(0x2ull << COUNT_REG0_SHIFT)
#define STC_STEN_PKTS_OPEN              (0x3ull << COUNT_REG0_SHIFT)
#define STP_CPROC_HOLDS_FFU_DP       	(0x4ull << COUNT_REG0_SHIFT)
#define STC_TLB_TABLE_WALKS             (0x5ull << COUNT_REG0_SHIFT)
#define STC_CACHE_HITS                  (0x6ull << COUNT_REG0_SHIFT)
#define STC_PCI_SLAVE_READS             (0x7ull << COUNT_REG0_SHIFT)
#define STP_PCI_WAITING_FOR_GNT         (0x8ull << COUNT_REG0_SHIFT)
#define STP_SYS_CLOCK_RATE0	        (0xfull << COUNT_REG0_SHIFT)

#define STATS_REG0_NAMES {                     \
          "STC_INPUT_NON_WRITE_BLOCKS",        \
          "STP_DMA_EOP_WAIT_ACK",              \
          "STP_TPROC_RUNNING",                 \
          "STC_STEN_PKTS_OPEN",                \
          "STP_CPROC_HOLDS_FFU_DP",            \
          "STC_TLB_TABLE_WALKS",               \
          "STC_CACHE_HITS",                    \
          "STC_PCI_SLAVE_READS",               \
          "STP_PCI_WAITING_FOR_GNT",           \
          "STP_SYS_CLOCK_RATE0"                \
}

/* Count reg 1 */
#define STC_INPUT_WRITE_BLOCKS     	(0x0ull << COUNT_REG1_SHIFT)
#define STP_DMA_DATA_TRANSMITTING  	(0x1ull << COUNT_REG1_SHIFT)
#define STC_CPROC_VALUES_EXE       	(0x2ull << COUNT_REG1_SHIFT)
#define STC_STEN_TRANS_SENT        	(0x3ull << COUNT_REG1_SHIFT)
#define STP_TPROC_DQ_HOLDS_FFU_DP    	(0x4ull << COUNT_REG1_SHIFT)
#define STC_TPROC_TLB_HITS         	(0x5ull << COUNT_REG1_SHIFT)
#define STC_CACHE_ALLOC_MISSES     	(0x6ull << COUNT_REG1_SHIFT)
#define STP_PCI_MASTER_READ_WAITING	(0x7ull << COUNT_REG1_SHIFT)
#define STP_PCI_WAITING_FOR_DEVSEL      (0x8ull << COUNT_REG1_SHIFT)
#define STP_SYS_CLOCK_RATE1	        (0xfull << COUNT_REG1_SHIFT)

#define STATS_REG1_NAMES {		       \
          "STC_INPUT_WRITE_BLOCKS",            \
          "STP_DMA_DATA_TRANSMITTING",         \
          "STC_CPROC_VALUES_EXE",              \
          "STC_STEN_TRANS_SENT",               \
          "STP_TPROC_DQ_HOLDS_FFU_DP",         \
          "STC_TPROC_TLB_HITS",                \
          "STC_CACHE_ALLOC_MISSES",            \
          "STP_PCI_MASTER_READ_WAITING",       \
          "STP_PCI_WAITING_FOR_DEVSEL",        \
          "STP_SYS_CLOCK_RATE1"                \
}

/* Count reg 2 */
#define STC_INPUT_PKTS              	(0x0ull << COUNT_REG2_SHIFT)
#define STP_DMA_WAITING_MEM         	(0x1ull << COUNT_REG2_SHIFT)
#define STC_CPROC_TRANSFERS             (0x2ull << COUNT_REG2_SHIFT)
#define STP_STEN_WAIT_NETWORK_BUSY   	(0x3ull << COUNT_REG2_SHIFT)
#define STP_IPROC_HOLDS_FFU_DP      	(0x4ull << COUNT_REG2_SHIFT)
#define STC_UNITS_TLB_HITS          	(0x5ull << COUNT_REG2_SHIFT)
#define STC_CACHE_NON_ALLOC_MISSES      (0x6ull << COUNT_REG2_SHIFT)
#define STP_PCI_MASTER_WRITE_WAITING	(0x7ull << COUNT_REG2_SHIFT)
#define STC_PCI_OUT_OF_ORDER_SPLIT_COMP (0x8ull << COUNT_REG2_SHIFT)
#define STP_SYS_CLOCK_RATE2	        (0xfull << COUNT_REG2_SHIFT)

#define STATS_REG2_NAMES {		       \
          "STC_INPUT_PKTS",                    \
          "STP_DMA_WAITING_MEM",               \
          "STC_CPROC_TRANSFERS",               \
          "STP_STEN_WAIT_NETWORK_BUSY",        \
          "STP_IPROC_HOLDS_FFU_DP",            \
          "STC_UNITS_TLB_HITS",                \
          "STC_CACHE_NON_ALLOC_MISSES",        \
          "STP_PCI_MASTER_WRITE_WAITING",      \
          "STC_PCI_OUT_OF_ORDER_SPLIT_COMP",   \
          "STP_SYS_CLOCK_RATE2"                \
}

/* Count reg 3 */
#define STC_INPUT_PKTS_REJECTED         (0x0ull << COUNT_REG3_SHIFT)
#define STP_DMA_WAIT_NETWORK_BUSY       (0x1ull << COUNT_REG3_SHIFT)
#define STC_CPROC_PREFETCH_SDRAM        (0x2ull << COUNT_REG3_SHIFT)
#define STP_STEN_BLOCKED_ACKS_OR_VC     (0x3ull << COUNT_REG3_SHIFT)
#define STP_EPROC_HOLDS_FFU_DP          (0x4ull << COUNT_REG3_SHIFT)
#define STP_TPROC_BLOCKED_MEMSYS        (0x5ull << COUNT_REG3_SHIFT)
#define STC_CACHE_WRITE_BACKS           (0x6ull << COUNT_REG3_SHIFT)
#define STP_PCI_SLAVE_READ_WAITING      (0x7ull << COUNT_REG3_SHIFT)
#define STP_PCI_IDLE_CYCLES		(0x8ull << COUNT_REG3_SHIFT)
#define STP_SYS_CLOCK_RATE3	        (0xfull << COUNT_REG3_SHIFT)

#define STATS_REG3_NAMES {	               \
          "STC_INPUT_PKTS_REJECTED",           \
          "STP_DMA_WAIT_NETWORK_BUSY",         \
          "STC_CPROC_PREFETCH_SDRAM",          \
          "STP_STEN_BLOCKED_ACKS_OR_VC",       \
          "STP_EPROC_HOLDS_FFU_DP",            \
          "STP_TPROC_BLOCKED_MEMSYS",          \
          "STC_CACHE_WRITE_BACKS",             \
          "STP_PCI_SLAVE_READ_WAITING",        \
          "STP_PCI_IDLE_CYCLES",               \
          "STP_SYS_CLOCK_RATE3"                \
}

/* Count reg 4 */
#define STP_INPUT_DATA_TRANSMITTING	(0x0ull << COUNT_REG4_SHIFT)
#define STC_DMA_PKTS_ACCEPTED      	(0x1ull << COUNT_REG4_SHIFT)
#define STC_CPROC_FLUSH_REQ_SDRAM  	(0x2ull << COUNT_REG4_SHIFT)
#define STP_STEN_EOP_WAIT_ACK      	(0x3ull << COUNT_REG4_SHIFT)
#define STP_DMA_HOLDS_FFU_DP       	(0x4ull << COUNT_REG4_SHIFT)
#define STP_UNIT_BLOCKED_MEMSYS    	(0x5ull << COUNT_REG4_SHIFT)
#define STC_PCI_MASTER_READS       	(0x6ull << COUNT_REG4_SHIFT)
#define STP_PCI_SLAVE_WRITE_WAITING	(0x7ull << COUNT_REG4_SHIFT)
#define STC_INPUT_PACKETS_DISCARDED	(0x8ull << COUNT_REG4_SHIFT)
#define STP_SYS_CLOCK_RATE4	        (0xfull << COUNT_REG4_SHIFT)

#define STATS_REG4_NAMES {		       \
          "STP_INPUT_DATA_TRANSMITTING",       \
          "STC_DMA_PKTS_ACCEPTED",             \
          "STC_CPROC_FLUSH_REQ_SDRAM",         \
          "STP_STEN_EOP_WAIT_ACK",             \
          "STP_DMA_HOLDS_FFU_DP",              \
          "STP_UNIT_BLOCKED_MEMSYS",           \
          "STC_PCI_MASTER_READS",              \
          "STP_PCI_SLAVE_WRITE_WAITING",       \
          "STC_INPUT_PACKETS_DISCARDED",       \
          "STP_SYS_CLOCK_RATE4"                \
}

/* Count reg 5 */
#define STP_INPUT_WAITING_NETWORK_DATA  (0x0ull << COUNT_REG5_SHIFT)
#define STC_DMA_PKTS_REJECTED           (0x1ull << COUNT_REG5_SHIFT)
#define STC_CPROC_INSERT_CACHE_MISSES   (0x2ull << COUNT_REG5_SHIFT)
#define STP_STEN_TRANSMITTING_DATA      (0x3ull << COUNT_REG5_SHIFT)
#define FFU_BLOCKED_DIFF_FFU_PROC       (0x4ull << COUNT_REG5_SHIFT)
#define STP_TABLE_WALKS_BLOCKED_MEMSYS  (0x5ull << COUNT_REG5_SHIFT)
#define STC_PCI_MASTER_WRITES           (0x6ull << COUNT_REG5_SHIFT)
#define STP_PCI_MASTER_HOLDS_BUS        (0x7ull << COUNT_REG5_SHIFT)
#define STC_PCI_NO_SPLIT_COMPS		(0x8ull << COUNT_REG5_SHIFT)
#define STP_SYS_CLOCK_RATE5	        (0xfull << COUNT_REG5_SHIFT)

#define STATS_REG5_NAMES {		       \
          "STP_INPUT_WAITING_NETWORK_DATA",    \
          "STC_DMA_PKTS_REJECTED",             \
          "STC_CPROC_INSERT_CACHE_MISSES",     \
          "STP_STEN_TRANSMITTING_DATA",        \
          "FFU_BLOCKED_DIFF_FFU_PROC",         \
          "STP_TABLE_WALKS_BLOCKED_MEMSYS",    \
          "STC_PCI_MASTER_WRITES",             \
          "STP_PCI_MASTER_HOLDS_BUS",          \
          "STC_PCI_NO_SPLIT_COMPS",            \
          "STP_SYS_CLOCK_RATE5"                \
}

/* Count reg 6 */
#define STP_INPUT_BLOCKED_WAITING_TRANS (0x0ull << COUNT_REG6_SHIFT)
#define STP_TPROC_INST_STALL          	(0x1ull << COUNT_REG6_SHIFT)
#define STP_CPROC_WAITING_DESCHED     	(0x2ull << COUNT_REG6_SHIFT)
#define STP_STEN_PKT_OPEN_WAITING_DATA	(0x3ull << COUNT_REG6_SHIFT)
#define STP_TLB_HASH_TABLE_ACCESSES   	(0x4ull << COUNT_REG6_SHIFT)
#define STP_PCI_SLAVE_BLOCKED_MEMSYS  	(0x5ull << COUNT_REG6_SHIFT)
#define STP_PCI_TRANSFERRING_DATA       (0x6ull << COUNT_REG6_SHIFT)
#define STP_PCI_MASTER_WAITING_BUS      (0x7ull << COUNT_REG6_SHIFT)
#define STP_PCI_READ_LATENCY		(0x8ull << COUNT_REG6_SHIFT)
#define STP_SYS_CLOCK_RATE6	        (0xfull << COUNT_REG6_SHIFT)

#define STATS_REG6_NAMES {		       \
          "STP_INPUT_BLOCKED_WAITING_TRANS",   \
          "STP_TPROC_INST_STALL",              \
          "STP_CPROC_WAITING_DESCHED",         \
          "STP_STEN_PKT_OPEN_WAITING_DATA",    \
          "STP_TLB_HASH_TABLE_ACCESSES",       \
          "STP_PCI_SLAVE_BLOCKED_MEMSYS",      \
          "STP_PCI_TRANSFERRING_DATA",         \
          "STP_PCI_MASTER_WAITING_BUS",        \
          "STP_PCI_READ_LATENCY",              \
          "STP_SYS_CLOCK_RATE6"                \
}

/* Count reg 7 */
#define STC_INPUT_CTX_FILTER_FILL       (0x0ull << COUNT_REG7_SHIFT)	
#define STP_TPROC_LOAD_STORE_STALL      (0x1ull << COUNT_REG7_SHIFT)
#define STC_CPROC_TIMEOUTS              (0x2ull << COUNT_REG7_SHIFT)
#define STP_STEN_BLOCKED_NETWORK        (0x3ull << COUNT_REG7_SHIFT)
#define STP_TLB_CHAIN_ACCESSES          (0x4ull << COUNT_REG7_SHIFT)
#define STP_CPROC_SCHED_BLOCKED_MEMSYS  (0x5ull << COUNT_REG7_SHIFT)
#define STC_PCI_SLAVE_WRITES            (0x6ull << COUNT_REG7_SHIFT)
#define STC_PCI_DISCONNECTS_RETRIES     (0x7ull << COUNT_REG7_SHIFT)
#define STC_RING_OSCILLATOR		(0x8ull << COUNT_REG7_SHIFT)
#define STP_SYS_CLOCK_RATE7	        (0xfull << COUNT_REG7_SHIFT)

#define STATS_REG7_NAMES {		       \
          "STC_INPUT_CTX_FILTER_FILL",         \
          "STP_TPROC_LOAD_STORE_STALL",        \
          "STC_CPROC_TIMEOUTS",                \
          "STP_STEN_BLOCKED_NETWORK",          \
          "STP_TLB_CHAIN_ACCESSES",            \
          "STP_CPROC_SCHED_BLOCKED_MEMSYS",    \
          "STC_PCI_SLAVE_WRITES",              \
          "STC_PCI_DISCONNECTS_RETRIES",       \
          "STC_RING_OSCILLATOR",               \
          "STP_SYS_CLOCK_RATE7"                \
}

#define STATS_REG_NAMES { \
    STATS_REG0_NAMES, \
    STATS_REG1_NAMES, \
    STATS_REG2_NAMES, \
    STATS_REG3_NAMES, \
    STATS_REG4_NAMES, \
    STATS_REG5_NAMES, \
    STATS_REG6_NAMES, \
    STATS_REG7_NAMES, \
}


#define INPUT_PERF_STATS        (STC_INPUT_NON_WRITE_BLOCKS | STC_INPUT_WRITE_BLOCKS |              \
				 STC_INPUT_PKTS | STC_INPUT_PKTS_REJECTED |                         \
                                 STC_INPUT_CTX_FILTER_FILL | STP_INPUT_DATA_TRANSMITTING |           \
				 STP_INPUT_WAITING_NETWORK_DATA | STP_INPUT_BLOCKED_WAITING_TRANS | STC_INPUT_PACKETS_DISCARDED) 

#define DMA_PERF_STATS          (STC_DMA_PKTS_ACCEPTED | STC_DMA_PKTS_REJECTED |                    \
                                 STP_DMA_EOP_WAIT_ACK | STP_DMA_DATA_TRANSMITTING |                 \
				 STP_DMA_WAITING_MEM | STP_DMA_WAIT_NETWORK_BUSY)                 


#define TPROC_PERF_STATS        (STP_TPROC_RUNNING | STP_TPROC_INST_STALL |                         \
                                 STP_TPROC_LOAD_STORE_STALL)

#define CPROC_PERF_STATS        (STC_CPROC_VALUES_EXE | STC_CPROC_TRANSFERS |                       \
				 STC_CPROC_PREFETCH_SDRAM | STC_CPROC_FLUSH_REQ_SDRAM |             \
				 STC_CPROC_INSERT_CACHE_MISSES | STP_CPROC_WAITING_DESCHED |        \
				 STC_CPROC_TIMEOUTS)

#define STEN_PERF_STATS         (STC_STEN_PKTS_OPEN | STC_STEN_TRANS_SENT |                         \
				 STP_STEN_WAIT_NETWORK_BUSY | STP_STEN_BLOCKED_ACKS_OR_VC |         \
				 STP_STEN_EOP_WAIT_ACK | STP_STEN_TRANSMITTING_DATA |               \
				 STP_STEN_PKT_OPEN_WAITING_DATA | STP_STEN_BLOCKED_NETWORK)

#define FFU_PREF_STATS          (STP_CPROC_HOLDS_FFU_DP | STP_TPROC_DQ_HOLDS_FFU_DP |               \
				 STP_IPROC_HOLDS_FFU_DP | STP_EPROC_HOLDS_FFU_DP |                  \
				 STP_DMA_HOLDS_FFU_DP | FFU_BLOCKED_DIFF_FFU_PROC)

#define TABLE_WALK_PERF_STATS   (STC_TPROC_TLB_HITS | STC_UNITS_TLB_HITS |                          \
				 STP_TLB_HASH_TABLE_ACCESSES | STP_TLB_CHAIN_ACCESSES |             \
				 STC_TLB_TABLE_WALKS)

#define ADDRESS_ARB_PERF_STATS  (STP_UNIT_BLOCKED_MEMSYS | STP_TPROC_BLOCKED_MEMSYS |               \
				 STP_TABLE_WALKS_BLOCKED_MEMSYS | STP_CPROC_SCHED_BLOCKED_MEMSYS |  \
				 STP_PCI_SLAVE_BLOCKED_MEMSYS)

#define CACHE_PERF_STATS        (STC_CACHE_HITS | STC_CACHE_ALLOC_MISSES |                          \
				 STC_CACHE_NON_ALLOC_MISSES | STC_CACHE_WRITE_BACKS)


#define PCI_PERF_STATS          (STC_PCI_SLAVE_READS | STP_PCI_MASTER_READ_WAITING |                \
                                 STP_PCI_MASTER_WRITE_WAITING | STP_PCI_SLAVE_READ_WAITING |        \
                                 STP_PCI_SLAVE_WRITE_WAITING | STC_PCI_MASTER_WRITES |              \
                                 STP_PCI_TRANSFERRING_DATA | STC_PCI_SLAVE_WRITES)

#define PCIBUS_PERF_STATS       (STP_PCI_WAITING_FOR_GNT | STP_PCI_WAITING_FOR_DEVSEL |		    \
				 STC_PCI_OUT_OF_ORDER_SPLIT_COMP | STP_PCI_IDLE_CYCLES | 	    \
				 STC_PCI_MASTER_READS | STP_PCI_MASTER_HOLDS_BUS | 		    \
				 STP_PCI_MASTER_WAITING_BUS | STC_PCI_DISCONNECTS_RETRIES)

				 
    extern const char *elan_stats_names[8][10];

#define ELAN_STATS_NAME(COUNT, CONTROL) (elan_stats_names[(COUNT)][(CONTROL) & 7])

    typedef volatile union e4_StatsControl
    {
	E4_uint64 StatsControl;
	struct
	{
#if (BYTE_ORDER == LITTLE_ENDIAN) || defined(__LITTLE_ENDIAN__)
	    E4_uint32 StatCont0:4;
	    E4_uint32 StatCont1:4;
	    E4_uint32 StatCont2:4;
	    E4_uint32 StatCont3:4;
	    E4_uint32 StatCont4:4;
	    E4_uint32 StatCont5:4;
	    E4_uint32 StatCont6:4;
	    E4_uint32 StatCont7:4;
#else
	    E4_uint32 StatCont7:4;
	    E4_uint32 StatCont6:4;
	    E4_uint32 StatCont5:4;

	    E4_uint32 StatCont4:4;
	    E4_uint32 StatCont3:4;
	    E4_uint32 StatCont2:4;
	    E4_uint32 StatCont1:4;
	    E4_uint32 StatCont0:4;
#endif
	    E4_uint32 pad;
	} s;
    } E4_StatsControl;

typedef volatile union e4_StatsCount
{
   E4_uint64    ClockStat; 
   struct
   {
       E4_uint32 ClockLSW;	/* read only */
       E4_uint32 StatsCount;
   } s;
} E4_StatsCount;

typedef volatile union e4_clock
{
   E4_uint64 NanoSecClock;
   struct
   {
      E4_uint32 ClockLSW;
      E4_uint32 ClockMSW;
   } s;
} E4_Clock;
#define E4_TIME( X ) ((X).NanoSecClock)

#define ELAN4_COMMS_CLOCK_FREQUENCY	660		/* In Mhz. This is half the bit rate. */
#define ELAN4_CLOCK_ADD_VALUE		200		/* For 200ns increment rate */
#define ELAN4_CLOCK_COMMS_DIV_VALUE	(((ELAN4_COMMS_CLOCK_FREQUENCY * ELAN4_CLOCK_ADD_VALUE) / (1000 * 4)) - 1)
#define ELAN4_CLOCK_TICK_RATE		((ELAN4_CLOCK_ADD_VALUE << 8) + ELAN4_CLOCK_COMMS_DIV_VALUE)

typedef volatile union e4_clocktickrate
{
   E4_uint64 NanoSecClock;
   struct
   {
      E4_uint32 pad1;
      E4_uint32 TickRates;
   } s;
} E4_ClockTickRate;

/*
 * This is made into an 8k byte object.
 */
typedef volatile struct _E4_User_Regs
{
   E4_StatsCount	StatCounts[8];
   E4_StatsCount	InstCount;
   E4_Clock		Clock;
   E4_StatsControl	StatCont;
   E4_ClockTickRate	ClockTickRate;
   E4_uint8		pad1[EightK - ((sizeof(E4_StatsCount)*9)+sizeof(E4_StatsControl)+
   					sizeof(E4_Clock)+sizeof(E4_ClockTickRate))];
} E4_User_Regs;

#ifdef __cplusplus
}
#endif

#endif /* __ELAN4_USERREGS_H */
