/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "$Id: jtagio.h,v 1.8 2004/12/16 10:39:27 lee Exp $"
/*             $Source: /cvs/master/quadrics/jtagmod/jtagio.h,v $*/


#ifndef __SYS_JTAGMOD_H
#define __SYS_JTAGMOD_H

#ifdef __cplusplus
extern "C" {
#endif

#define JTAG_MAX_CHIPS		8
#define JTAG_MAX_INSTR_LEN	8
#define JTAG_MAX_BITS           (JTAG_MAX_CHIPS * JTAG_MAX_INSTR_LEN)
#define JTAG_MAX_DATA_LEN	1024

#define JTAG_BYPASS		0xFF

#define I2C_ADDR_LEN		7				/* 7 bits of address */
#define I2C_DATA_LEN		8				/* 8 bits of data */
#define I2C_MAX_DATA_LEN	9				/* and upto 9 bytes worth */

#define BITS_PER_BYTE		8
#define JTAG_NBYTES(nbits)	(((nbits)+BITS_PER_BYTE-1)/BITS_PER_BYTE)
#define JTAG_BIT(v, num)	(((v)[(num) / BITS_PER_BYTE] >> ((num) % BITS_PER_BYTE)) & 1)
#define JTAG_SET_BIT(v, num)	((v)[(num) / BITS_PER_BYTE] |= (1 << ((num) % BITS_PER_BYTE)))
#define JTAG_CLR_BIT(v, num)	((v)[(num) / BITS_PER_BYTE] &= ~(1 << ((num) % BITS_PER_BYTE)))

#define RING_CLOCK_CARD		(0x3D)
#define RING_CLOCK_SHIFT	(0x3E)
#define RING_JTAG_LOOPBACK	(0x3F)
#define RING_MAX		(0x40)

#define RING_QUAD_BIT		(0x40)
#define RING_I2C_BIT		(0x80)

#define VALID_JTAG_RING(ring)	((ring) < 0x20 || (ring) == RING_JTAG_LOOPBACK)
#define VALID_I2C_RING(ring)	((ring) < 0x20 || (ring) == RING_CLOCK_CARD)


typedef struct jtag_value
{
    u_char	bytes[JTAG_NBYTES(JTAG_MAX_DATA_LEN)];
} JTAG_VALUE;

/* arguements to JTAG_SHIFT_IR/JTAG_SHIFT_DR */
typedef struct jtag_reset_args
{
    u_int	ring;
} JTAG_RESET_ARGS;

typedef struct jtag_shift_args
{
    u_int	ring;
    u_int	nbits;
    u_char     *value;
} JTAG_SHIFT_ARGS;

typedef struct i2c_args
{
    u_int	ring;
    u_int	device;
    u_int	reg;
    u_int	count;
    u_int	ok;
    u_char	data[I2C_MAX_DATA_LEN];
} I2C_ARGS;

/* values for 'ok' - the return value from i2c_xx functions */
#define I2C_OP_SUCCESS		0
#define I2C_OP_ERROR		1
#define I2C_OP_NOT_IDLE		2
#define I2C_OP_NO_DEVICE	3
#define I2C_OP_WRITE_TO_BIG	4
#define I2C_OP_BAD_RESOURCE	5

typedef struct i2c_clock_shift_args
{
    u_int	t;
    u_int	n;
    u_int	m;
} I2C_CLOCK_SHIFT_ARGS;

#define JTAG_RESET		_IOWR('j', '0', JTAG_RESET_ARGS)
#define JTAG_SHIFT_IR		_IOWR('j', '1', JTAG_SHIFT_ARGS)
#define JTAG_SHIFT_DR		_IOWR('j', '2', JTAG_SHIFT_ARGS)

#define I2C_CLOCK_SHIFT		_IOWR('j', '4', I2C_CLOCK_SHIFT_ARGS)
#define I2C_WRITE		_IOWR('j', '5', I2C_ARGS)
#define I2C_READ		_IOWR('j', '6', I2C_ARGS)
#define I2C_WRITEREG		_IOWR('j', '7', I2C_ARGS)
#define I2C_READREG		_IOWR('j', '8', I2C_ARGS)


#ifdef __cplusplus
}
#endif

#endif /* __SYS_JTAGMOD_H */
