/*
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Derived from exit_actn.c by
 *    Copyright (C) 2003 Quadrics Ltd.
 *
 */
#ifndef __LINUX_PTRACK_H
#define __LINUX_PTRACK_H

/* 
 * Process tracking - this allows a module to keep track of processes
 * in order that it can manage all tasks derived from a single process.
 */

#define PTRACK_PHASE_CLONE	1
#define PTRACK_PHASE_CLONE_FAIL	2
#define PTRACK_PHASE_EXEC	3
#define PTRACK_PHASE_EXIT      	4

#define PTRACK_FINISHED		0
#define PTRACK_INNHERIT		1
#define PTRACK_DENIED		2

#ifdef CONFIG_PTRACK

typedef int (*ptrack_callback_t)(void *arg, int phase, struct task_struct *child);

struct ptrack_desc {
       struct list_head        link;
       ptrack_callback_t       callback;
       void                   *arg;
};

extern int     ptrack_register (ptrack_callback_t callback, void *arg);
extern void    ptrack_deregister (ptrack_callback_t callback, void *arg);
extern int     ptrack_registered (ptrack_callback_t callback, void *arg);

extern int     ptrack_call_callbacks (int phase, struct task_struct *child);

#define INIT_TASK_PTRACK(tsk) \
	.ptrack_list = LIST_HEAD_INIT(tsk.ptrack_list)

#else
#define ptrack_call_callbacks (phase, child) (0)

#define INIT_TASK_PTRACK(tsk)

#endif

#endif /* __LINUX_PTRACK_H */
