/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef _QSNET_CONFIG_H
#define _QSNET_CONFIG_H

#ident "$Id: config.h,v 1.24 2005/04/28 22:59:31 robin Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/config.h,v $*/


/*
 * QSNET standard defines :
 *
 *   Target operating system defines
 *		SOLARIS
 *		TRU64UNIX/DIGITAL_UNIX
 *		LINUX
 *
 *   Target processor defines
 *		SPARC
 *		ALPHA
 *		I386
 *		IA64
 *		X86_64
 *
 *   Byte order defines
 *   		__LITTLE_ENDIAN__
 *   		__BIG_ENDIAN__
 *
 *   Data size defines
 *   		_LP64			- LP64 - long/pointer is 64 bits
 *   		_ILP32			- LP32 - long/pointer is 32 bits
 *
 *   Elan defines for main processor
 *   		__MAIN_LITTLE_ENDIAN__	- main byte order (for thread code)
 *   		__MAIN_BIG_ENDIAN__
 *   		_MAIN_LP64		- main long size (for thread code)
 *   		_MAIN_ILP32
 *
 *   Compiling for kernel (defined in makefile)
 *   		_KERNEL
 *
 */

#if defined(__LP64__) && !defined(_LP64)
#  define _LP64
#endif

#if defined(__arch64__) && !defined(_LP64) && !defined(_ILP32)
#  define _LP64
#endif

#if defined(__alpha__) && !defined(_LP64) && !defined(_ILP32)
#  define _LP64
#endif

#if !defined(__arch64__) && !defined(_ILP32) && !defined(_LP64)
#  define _ILP32
#endif

#if defined(__ELAN__) || defined(__ELAN3__)

#define __LITTLE_ENDIAN__

#if defined(__host_solaris) && defined(__host_sparc)
#define SOLARIS
#define SPARC
#define SOLARIS_SPARC
#define _MAIN_ILP32
#define __MAIN_BIG_ENDIAN__

#elif defined(__host_osf)
#define TRU64UNIX
#define DIGITAL_UNIX
#define ALPHA
#define _MAIN_LP64
#define __MAIN_LITTLE_ENDIAN__

#elif defined(__host_linux) && defined(__host_alpha)
#define LINUX
#define ALPHA
#define LINUX_ALPHA
#define _MAIN_LP64
#define __MAIN_LITTLE_ENDIAN__

#elif defined(__host_linux) && defined(__host_sparc)
#define LINUX
#define SPARC
#define LINUX_SPARC
#define __MAIN_BIG_ENDIAN__
#ifdef __KERNEL__
#  define _MAIN_LP64
#else
#  define _MAIN_ILP32
#endif

#elif defined(__host_linux) && defined(__host_i386)
#define LINUX
#define I386
#define LINUX_I386
#define _MAIN_ILP32
#define __MAIN_LITTLE_ENDIAN__

#elif defined(__host_linux) && defined(__host_ia64)
#define LINUX
#define IA64
#define LINUX_IA64
#define _MAIN_LP64
#define __MAIN_LITTLE_ENDIAN__

#elif defined(__host_linux) && defined(__host_x86_64)
#define LINUX
#define X86_64
#define LINUX_X86_64
#define _MAIN_LP64
#define __MAIN_LITTLE_ENDIAN__

#else
#error Cannot determine operating system/processor architecture.
#endif

#else /* !defined(__ELAN3__) */

#if (defined(sun) || defined(__sun)) && defined(sparc) && !defined(__sparcv9)	/* Sun Solaris 5.6 */
#define SOLARIS
#define SPARC
#define SOLARIS_SPARC
#ifndef __BIG_ENDIAN__
#define __BIG_ENDIAN__
#endif 

#elif (defined(sun) || defined(__sun)) && defined(sparc) && defined(__sparcv9)	/* Sun Solaris 5.7 */
#define SOLARIS
#define SPARC
#define SOLARIS_SPARC
#define __BIG_ENDIAN__

#elif defined(__osf__) && defined(__alpha)					/* Digital Unix */
#define TRU64UNIX
#define DIGITAL_UNIX
#define ALPHA
#define __LITTLE_ENDIAN__

#elif (defined(linux) || defined(__linux__)) && defined(__alpha)		/* Linux Alpha */

#define LINUX
#define ALPHA
#define LINUX_ALPHA
#define __LITTLE_ENDIAN__

#elif (defined(linux) || defined(__linux__)) && defined(__sparc)		/* Linux Sparc */

#define LINUX
#define SPARC
#define LINUX_SPARC
#define __BIG_ENDIAN__

#elif (defined(linux) || defined(__linux__)) && defined(__i386)			/* Linux i386 */

#define LINUX
#define I386
#define LINUX_I386
#define __LITTLE_ENDIAN__

#elif (defined(linux) || defined(__linux__)) && defined(__ia64)			/* Linux ia64 */

#define LINUX
#define IA64
#define LINUX_IA64
#define __LITTLE_ENDIAN__

#elif (defined(linux) || defined(__linux__)) && (defined(__x86_64) || defined(__x86_64__))	/* Linux x86_64 */

#define LINUX
#define X86_64
#define LINUX_X86_64
#define __LITTLE_ENDIAN__

#elif defined(__QNXNTO__)
#define QNX
#define I386
#define __LITTLE_ENDIAN__
#else
#error Cannot determine operating system/processor architecture.
#endif

#endif

#include <qsnet/workarounds.h>

#endif /* _QSNET_CONFIG_H */
