/*
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef _QSNET_DEBUG_H
#define _QSNET_DEBUG_H

#if defined(DIGITAL_UNIX) 
#include <kern/assert.h>
#elif defined(LINUX)
extern int qsnet_assfail (char *ex, const char *func, char *file, int line);

#define ASSERT(EX)	do { \
	if (!(EX) && qsnet_assfail (#EX, __FUNCTION__, __BASE_FILE__, __LINE__)) { \
		BUG(); \
	} \
} while (0)
#endif /* DIGITAL_UNIX */

/* debug.c */
extern void qsnet_debug_init(void);
extern void qsnet_debug_fini(void);
extern void qsnet_debug_disable(int);
extern void qsnet_debug_alloc(void);

#define QSNET_DEBUG_BUFFER  ((unsigned int)(0x01))
#define QSNET_DEBUG_CONSOLE ((unsigned int)(0x02))
#define QSNET_DEBUG_BUF_CON ( QSNET_DEBUG_BUFFER | QSNET_DEBUG_CONSOLE )

#ifdef __GNUC__
extern void qsnet_debugf      (unsigned int mode, const char *fmt, ...)
	__attribute__ ((format (printf,2,3)));
extern void kqsnet_debugf      (char *fmt, ...)
	__attribute__ ((format (printf,1,2)));
#else
extern void qsnet_debugf      (unsigned int mode, const char *fmt, ...);
extern void kqsnet_debugf     (char *fmt, ...);
#endif
extern void qsnet_vdebugf     (unsigned int mode, const char *prefix, const char *fmt,  va_list ap);
extern int  qsnet_debug_buffer(caddr_t ubuffer, int len);
extern int  qsnet_debug_dump  (void);
extern int  qsnet_debug_kmem  (void *handle);

extern void qsnet_debug_buffer_on(void);
extern void qsnet_debug_buffer_clear(void);
extern void qsnet_debug_buffer_mark(char *str);

#endif /* _QSNET_DEBUG_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
