/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __QSNET_KERNEL_LINUX_H
#define __QSNET_KERNEL_LINUX_H

#ident "$Id: kernel_linux.h,v 1.69.2.3 2005/09/07 14:35:03 mike Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/kernel_linux.h,v $*/

#include <linux/version.h>
#if defined(MODVERSIONS)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
#include <config/modversions.h>
#else
#include <linux/modversions.h>
#endif
#endif

#include <linux/autoconf.h>
#include <qsnet/module.h>

/* ASSERT(spin_is_locked(l)) would always fail on UP kernels */
#if defined(CONFIG_SMP)
#define SPINLOCK_HELD(l)	spin_is_locked(l)
#else
#define SPINLOCK_HELD(l)	(1) 
#endif

#include <asm/io.h>
#include <asm/uaccess.h>

#include <linux/types.h>
#include <linux/time.h>

#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <qsnet/module.h>

#include <linux/highmem.h>

#include <qsnet/mutex.h>
#include <qsnet/condvar.h>
#include <qsnet/crwlock.h>

#if defined(LINUX_ALPHA)
#  include <asm/core_tsunami.h>	/* for TSUNAMI_MEM */
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
#	undef 	MOD_INC_USE_COUNT
#	undef	MOD_DEC_USE_COUNT
#	define	MOD_INC_USE_COUNT
#	define	MOD_DEC_USE_COUNT
#endif

#define MIN(a,b)	((a) > (b) ? (b) : (a))
#define MAX(a,b)	((a) > (b) ? (a) : (b))

/* stray types */
typedef u64		   u_longlong_t;
typedef unsigned long	   uintptr_t;
typedef int		   bool_t;

typedef unsigned long	   virtaddr_t;				/* virtual address */
typedef unsigned long      ioaddr_t;				/* io address */
typedef unsigned long      sdramaddr_t;				/* elan sdram offset */

/* 386 kernel can be compiled with PAE enabled to use a 44 bit physical address */
#if defined(CONFIG_X86_PAE)
typedef unsigned long long physaddr_t;
#else
typedef unsigned long	   physaddr_t;
#endif

/* ticks since reboot, and tick freq */
#define lbolt		jiffies	
#define hz		HZ

/* System page size and friends */
#define PAGESIZE	PAGE_SIZE
#define PAGESHIFT	PAGE_SHIFT
#define PAGEOFFSET	(PAGE_SIZE - 1)
#define PAGEMASK	PAGE_MASK

#define PAGE_ALIGNED(a)	(((a) & PAGE_MASK) == a)

/* convert between bytes and pages */
#define btop(b)         ((unsigned long)(b) >> PAGE_SHIFT)	/* rnd down */ 
#define btopr(b)        btop(PAGE_ALIGN((unsigned long) b))	/* rnd up */
#define ptob(p)		((unsigned long)(p) << PAGE_SHIFT)

/* round up sz to the nearest multiple of blk */
#define roundup(sz,blk) ((blk) * ((sz) / (blk) + ((sz) % (blk) ? 1 : 0)))	

/* send a signal to a process */
#define psignal(pr,sig)	send_sig(sig,pr,0)

/* microsecond delay */
#define DELAY(us)	udelay(us)

/* macro macros */
#define MACRO_BEGIN     do {
#define MACRO_END       } while (0)

/* D-Unix compatable errno values */
#define ESUCCESS        0
#define EFAIL           255

/* ASSERT(NO_LOCKS_HELD) will be a no-op */
#define NO_LOCKS_HELD	1

/* misc */
typedef int		label_t;
#define on_fault(ljp)	((ljp) == NULL)
#define _NOTE(X)
#define no_fault()	((void) 0)
#define panicstr	0

/* return from system call is -EXXX on linux */
#define set_errno(e)	(-(e))

/* 
 * BSD-style byte ops 
 */

#define bcmp(src1,src2,len)		memcmp(src1,src2,len)
#define bzero(dst,len)	 		memset(dst,0,len)
#define bcopy(src,dst,len)	 	memcpy(dst,src,len)

#define preemptable_start		do { long must_yield_at = lbolt + (hz/10);
#define preemptable_end			} while (0)
#define preemptable_check()		do {\
                                            if ((lbolt - must_yield_at) > 0)\
					    {\
						preemptable_yield() ; \
						must_yield_at = lbolt + (hz/10);\
					    }\
					} while (0)

#define preemptable_yield()		schedule()

#define CURPROC()                       current
#define CURTHREAD()                     current
#define SUSER()				suser()

/* 64 bit IO operations on 32 bit intel cpus using MMX */
#if defined(LINUX_I386)
extern u64	    qsnet_readq (volatile u64 *ptr);
extern void	    qsnet_writeq (u64 value, volatile u64 *ptr);

#define readq(ptr)		qsnet_readq((void *) ptr)
#define writeq(val,ptr)		qsnet_writeq(val, (void *)ptr)
#endif

/*
 * Memory barriers
 */
#ifndef mmiob
#  define mmiob()			mb()
#endif

/* 
 * Exit handlers
 */
#define HANDLER_REGISTER(func,arg,flags)   xa_handler_register(func,arg,flags)
#define HANDLER_UNREGISTER(func,arg,flags) xa_handler_unregister(func,arg,flags)

/* 
 * KMEM_GETPAGES and KMEM_ALLOC both call kmem_alloc, which 
 * translates the call to kmalloc if < PAGE_SIZE, or vmalloc 
 * if >= PAGE_SIZE.  vmalloc will always return a page-aligned 
 * region rounded up to the nearest page, while kmalloc will 
 * return bits and pieces of a page.
 */

#ifdef KMEM_DEBUG
extern void          *qsnet_kmem_alloc_debug(int len, int sleep, int zerofill, char *file, int line);
extern void           qsnet_kmem_free_debug(void *ptr, int len, char *file, int line);
#define KMEM_ALLOC(ptr,type,len,sleep) \
	{ KMEM_ASSERT(sleep); (ptr)=(type)qsnet_kmem_alloc_debug(len,sleep,0,__FILE__,__LINE__); }
#define KMEM_ZALLOC(ptr,type,len,sleep) \
	{ KMEM_ASSERT(sleep); (ptr)=(type)qsnet_kmem_alloc_debug(len,sleep,1,__FILE__,__LINE__); }

#define KMEM_FREE(ptr,len) 		  qsnet_kmem_free_debug((void *)ptr,len,__FILE__,__LINE__)

#else

extern void          *qsnet_kmem_alloc(int len, int sleep, int zerofill);
extern void           qsnet_kmem_free(void *ptr, int len);

#define KMEM_ALLOC(ptr,type,len,sleep) \
	{ KMEM_ASSERT(sleep); (ptr)=(type)qsnet_kmem_alloc(len,sleep,0); }
#define KMEM_ZALLOC(ptr,type,len,sleep) \
	{ KMEM_ASSERT(sleep); (ptr)=(type)qsnet_kmem_alloc(len,sleep,1); }

#define KMEM_FREE(ptr,len) 		  qsnet_kmem_free((void *)ptr,len)

#endif
extern void       qsnet_kmem_display(void *handle);
extern physaddr_t kmem_to_phys(void *ptr);

#define KMEM_ASSERT(sleep) 		 ASSERT(!(in_interrupt() && sleep))


#define KMEM_GETPAGES(ptr,type,pgs,sleep) KMEM_ZALLOC(ptr,type,ptob(pgs),sleep)
#define KMEM_FREEPAGES(ptr,pgs) 	  KMEM_FREE(ptr,ptob(pgs));

/*
 * Copying from user space -> kernel space (perms checked)
 */
#define copyin(up,kp,size)		copy_from_user(kp,up,size)
#define copyin_noerr(up,kp,size)	copy_from_user(kp,up,size)

/* get_user() gets xfer width right */
#define fulinux(ret, up) 		(get_user(ret, (up)) == 0 ? ret : -1)
#define fulinuxp(ret, up) 		(get_user(ret, (up)) == 0 ? ret : NULL)

extern __inline__ int fubyte    (u8  *up) { u8  ret;   return fulinux(ret, up);}
extern __inline__ int fusword   (u16 *up) { u16 ret;   return fulinux(ret, up);}
extern __inline__ int fuword    (u32 *up) { u32 ret;   return fulinux(ret, up);}
#if BITS_PER_LONG > 32
extern __inline__ u64 fulonglong(u64 *up) { u64 ret;   return fulinux(ret, up);}
#else
extern __inline__ u64 fulonglong(u64 *up) { return ((u64) fuword((u32 *)up) | (((u64) fuword(((u32 *)up)+1))<<32)); }
#endif
extern __inline__ void *fuptr (void **up) { void *ret; return fulinuxp(ret,up);}

#define fubyte_noerr(up)		fubyte(up)
#define fusword_noerr(up)		fusword(up)
#define fuword_noerr(up)		fuword(up)
#define fulonglong_noerr(up)		fulonglong(up)
#define fuptr_noerr(up)			fuptr(up)

extern __inline__ int copyinstr(char *up, char *kp, int max, int *size)
{ 
	for (*size = 1; *size <= max; (*size)++) {
		if (get_user(*kp, up++) != 0)
			return EFAULT;	/* bad user space addr */
		if (*kp++ == '\0')
			return 0; 	/* success */
	}
	*size = max;
	return ENAMETOOLONG; 		/* runaway string */
}
 
/*
 * Copying from kernel space -> user space (perms checked)
 */

#define copyout(kp,up,size)		copy_to_user(up,kp,size)
#define copyout_noerr(kp,up,size)	copy_to_user(up,kp,size)

/* put_user() gets xfer width right */
#define sulinux(val, up) 		(put_user(val, (up)) == 0 ? 0 : -1)

extern __inline__ int subyte    (u8  *up, u8  val) { return sulinux(val, up); }
extern __inline__ int susword   (u16 *up, u16 val) { return sulinux(val, up); }
extern __inline__ int suword    (u32 *up, u32 val) { return sulinux(val, up); }
#if BITS_PER_LONG > 32
extern __inline__ int sulonglong(u64 *up, u64 val) { return sulinux(val, up); }
#else
extern __inline__ int sulonglong(u64 *up, u64 val) { return (suword((u32 *) up, (u32) val) == 0 ? 
							     suword(((u32 *) up)+1, (u32) (val >> 32)) : -1); }
#endif
extern __inline__ int suptr   (void **up,void *val){ return sulinux(val, up); }

#define subyte_noerr(up,val)		subyte(up,val)	
#define susword_noerr(up,val)		susword(up,val)	
#define suword_noerr(up,val)		suword(up,val)	
#define sulonglong_noerr(up,val)	sulonglong(up,val)	
#define suptr_noerr(up,val)		suptr(up,val)	

/*
 * /proc/qsnet interface
 */
extern inline int
str_append(char *buf, char *add, int size)
{
#define TRUNC_MSG 	 "[Output truncated]\n"
	int full = 0;
	int max = size - strlen(TRUNC_MSG) - strlen(add) - 1;

	if (strlen(buf) > max) {
		strcat(buf, TRUNC_MSG);
		full = 1;
	} else
		strcat(buf, add);
	return full;
}

/* Spinlocks */
#define spin_lock_destroy(l)		((void) 0)

/* Complex - Reader/Writer locks - we added <linux/crwlock.h> */
typedef crwlock_t 			krwlock_t;
#define krwlock_init(l)			crwlock_init(l)
#define krwlock_destroy(l)		crwlock_destroy(l)
#define krwlock_write(l)		crwlock_write(l)
#define krwlock_read(l)			crwlock_read(l)
#define krwlock_done(l)			crwlock_done(l)
#define krwlock_is_locked(l)		crwlock_held(l)
#define krwlock_is_write_locked(l)	crwlock_write_held(l)
#define krwlock_is_read_locked(l)	crwlock_read_held(l)

/*
 * Timeouts - Solaris style.
 */
typedef struct timer_list timer_fn_t;

extern inline void
schedule_timer_fn(timer_fn_t *timer, void (*fun)(void *), void *arg, long hz_delay)
{
	init_timer(timer);

	timer->function = (void (*)(unsigned long)) fun;
	timer->data     = (unsigned long) arg;
	timer->expires  = jiffies + hz_delay;

	add_timer(timer);
}

/* returns 1 if timer_fn was cancelled */
extern inline int
cancel_timer_fn(timer_fn_t *timer)
{
    return (del_timer_sync(timer));
}

extern inline int
timer_fn_queued(timer_fn_t *timer)
{
    return (timer_pending (timer));
}
/*
 * Hold/release CPU's.
 */

extern void 	cpu_hold_all(void);
extern void 	cpu_release_all(void);
#define CAPTURE_CPUS()		cpu_hold_all()
#define RELEASE_CPUS()		cpu_release_all()

#define IASSERT ASSERT

/* code to support multipage procfs entries */

typedef struct display_info {
    void (*func)(long, char *, ...);
    long arg;
} DisplayInfo;

typedef struct qsnet_proc_private 
{
	struct nodeset_private *pr_next;
	void                   *pr_user_data;
	char		       *pr_data;
	int                     pr_data_len;
	unsigned	        pr_off;
	unsigned	        pr_len;
	DisplayInfo             pr_di;
} QSNET_PROC_PRIVATE;

#endif /* __QSNET_KERNEL_LINUX_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
