/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2004 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __QSNET_KPTE_H
#define __QSNET_KPTE_H

#ident "@(#)$Id: kpte.h,v 1.5 2005/03/18 13:56:40 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/qsnet/kpte.h,v $*/

#include <qsnet/autoconf.h>

#ifdef NO_RMAP
#	define pte_offset_kernel pte_offset
#	define pte_offset_map	 pte_offset
#       define pte_unmap(A)      do { ; } while (0)
#endif

/* 
 * Pte stuff
 */
static __inline__ struct mm_struct *
get_kern_mm(void)
{
        return &init_mm;
}

static __inline__ pte_t *
find_pte_map(struct mm_struct *mm, unsigned long vaddr)
{
        pgd_t *pgd;
        pmd_t *pmd;
	pte_t *ptep;

/* XXXX - need to handle huge tlb code */
	pgd = pgd_offset(mm, vaddr);
	if (pgd_none(*pgd) || pgd_bad(*pgd))
		goto out;
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
	{
	    pud_t *pud = pud_offset(pgd, vaddr);
	    if (pud_none(*pud) || pud_bad(*pud))
		goto out;

	    pmd = pmd_offset(pud, vaddr);
	}
#else
	pmd = pmd_offset(pgd, vaddr);
#endif
	if (pmd_none(*pmd) || pmd_bad (*pmd))
		goto out;

	ptep = pte_offset_map (pmd, vaddr);
	if (! ptep)
		goto out;
	
	if (pte_present (*ptep))
		return ptep;

	pte_unmap (ptep);
out:
	return NULL;
}

static __inline__ pte_t *
find_pte_kernel(unsigned long vaddr)
{
        pgd_t *pgd;
        pmd_t *pmd;
	pte_t *pte;

	pgd = pgd_offset_k(vaddr);
	if (pgd && !pgd_none(*pgd)) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
	    pud_t *pud = pud_offset(pgd, vaddr);
	    if (pud && !pud_none(*pud)) {
	    	pmd = pmd_offset(pud, vaddr);
#else
		pmd = pmd_offset(pgd, vaddr);
#endif
	    	if (pmd && pmd_present(*pmd)) {
			pte = pte_offset_kernel(pmd, vaddr);
			if (pte && pte_present(*pte))
			    return (pte);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
		}
#endif
	    }
	}
	return (NULL);
}

static __inline__ physaddr_t
pte_phys(pte_t pte)
{
#if defined(LINUX_ALPHA)
	/* RedHat 7.1 2.4.3-12 
	 * They have now enabled Monster windows on Tsunami
	 * and so can use the Main's phys pte value 
	 */
	return (pte_val(pte) >> (32-PAGE_SHIFT));
#elif defined(LINUX_I386) || defined(LINUX_X86_64)
#if defined(_PAGE_NX)
	return (pte_val(pte) & ~((1 << PAGE_SHIFT)-1) & ~_PAGE_NX);
#else
	return (pte_val(pte) & ~((1 << PAGE_SHIFT)-1));
#endif
#elif defined(LINUX_SPARC)
	return (pte_val(pte) & _PAGE_PADDR);
#elif defined(LINUX_IA64)
	return (pte_val(pte) & _PFN_MASK);
#else
#error Unknown architecture
#endif
}

#ifndef page_to_pfn
#define page_to_pfn(page)	(pte_phys(mk_pte(page, __pgprot(0))) >> PAGE_SHIFT)
#endif

#endif /* __QSNET_KPTE_H */

/*
 * Local variables:
 * c-file-style: "stroustrup"
 * End:
 */
