/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *    Copyright (c) 2002-2004 by Quadrics Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __QSNET_KTHREAD_H
#define __QSNET_KTHREAD_H

#ident "@(#)$Id: kthread.h,v 1.1 2004/10/28 11:50:29 david Exp $ $Name: QSNETMODULES-5-11-3_20050907 $"
/*      $Source: /cvs/master/quadrics/qsnet/kthread.h,v $*/

#include <qsnet/autoconf.h>

/* 
 * kernel threads 
 */
extern __inline__ void
kernel_thread_init(char *comm)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
#ifndef NO_NPTL
#	define sigmask_lock 			sighand->siglock
#endif
	lock_kernel();
	daemonize();
        reparent_to_init();

        /* avoid getting signals */
        spin_lock_irq(&current->sigmask_lock);
        flush_signals(current);
        sigfillset(&current->blocked);
	
#ifdef NO_NPTL
        recalc_sigpending(current);
#else
        recalc_sigpending();
#endif

        spin_unlock_irq(&current->sigmask_lock);

	/* set our name for identification purposes */
	strncpy(current->comm, comm, sizeof(current->comm));

	unlock_kernel();
#else
	daemonize(comm);
#endif
}

extern __inline__ void *
kernel_thread_wrap(caddr_t stk, int stksize, void (*proc)(void *), void *arg)
{
        ASSERT(stk == NULL && stksize == 0);
        kernel_thread((int (*)(void *))proc, arg, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
        return (void *)1; /* non-null value */
}

#define kernel_thread_create(proc,arg)  kernel_thread_wrap(NULL,0,(void (*)(void *))proc,arg)
#define kernel_thread_exit()		((void) 0)
#define kernel_thread_become_highpri()	((void) 0)

#endif /* __QSNET_KTHREAD_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
