/*
 *    Copyright (c) 2003 by Quadrics Limited.
 * 
 *    For licensing information please see the supplied COPYING file
 *
 */

#ident "@(#)$Id: list.h,v 1.5 2003/10/27 13:55:33 david Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/list.h,v $*/

#ifndef __QSNET_LIST_H
#define __QSNET_LIST_H

/* Implementation of doubly linked lists - compatible with linux */
struct list_head 
{
    struct list_head *next;
    struct list_head *prev;
};

#if !defined(LINUX)
#if ! defined( offsetof ) 
#define offsetof(T,F) ((int )&(((T *)0)->F))
#endif

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)
#endif

#define list_entry(ptr, type, off) \
	((type *) ((unsigned long)(ptr) - offsetof (type,off)))

#define INIT_LIST_HEAD(list) \
MACRO_BEGIN \
	(list)->next = (list)->prev = (list); \
MACRO_END

#define list_add(new, list) \
MACRO_BEGIN \
	(list)->next->prev = (new); \
	(new)->next = (list)->next; \
	(new)->prev = (list); \
	(list)->next = (new); \
MACRO_END

#define list_add_tail(new, list) \
MACRO_BEGIN \
	(list)->prev->next = new; \
	(new)->prev = (list)->prev; \
	(new)->next = (list); \
	(list)->prev = (new); \
MACRO_END

#define list_del(entry) \
MACRO_BEGIN \
	(entry)->prev->next = (entry)->next; \
	(entry)->next->prev = (entry)->prev; \
MACRO_END

#define list_del_init(entry) \
MACRO_BEGIN \
	(entry)->prev->next = (entry)->next; \
	(entry)->next->prev = (entry)->prev; \
	(entry)->next = (entry)->prev = (entry); \
MACRO_END

#define list_empty(list) \
	((list)->next == (list))

#define list_for_each(pos,list) \
	for (pos = (list)->next; pos != (list); \
	     pos = (pos)->next)

#define list_for_each_safe(pos,n,list) \
	for (pos = (list)->next, n = (pos)->next; pos != (list); \
	     pos = n, n = (pos)->next)

#endif /* __QSNET_LIST_H */
