/*
 *    Copyright (C) 2000  Regents of the University of California
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if	!defined(_LINUX_MUTEX_H)
#define	_LINUX_MUTEX_H
#if	defined(__KERNEL__)

#include <asm/smp.h>
#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <qsnet/debug.h>
#include <linux/interrupt.h>
#include <linux/version.h>

#define PID_NONE	0

typedef struct
{
    struct semaphore sem;
    pid_t	     holder;
} kmutex_t;

extern __inline__ void
kmutex_init (kmutex_t *l)
{
#if	LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
    l->sem = MUTEX;
#else
    init_MUTEX(&l->sem);
#endif
    l->holder = PID_NONE;
}

extern __inline__ void
kmutex_destroy (kmutex_t *l) 
{
    ASSERT (l->holder == PID_NONE); 
}

extern __inline__ void
kmutex_lock (kmutex_t *l) 
{ 
    ASSERT(l->holder != current->pid);
    down (&l->sem);
    l->holder = current->pid; 
}

extern __inline__ void
kmutex_unlock (kmutex_t *l) 
{
    ASSERT(l->holder == current->pid);

    l->holder = PID_NONE;
    up (&l->sem);
}

extern __inline__ int
kmutex_trylock (kmutex_t *l) 
{
    if (down_trylock (&l->sem) == 0) 
    {
	l->holder = current->pid;
	return (1);
    }
    return (0);
}

extern __inline__ int
kmutex_is_locked (kmutex_t *l) 
{
    return (l->holder == current->pid);
}

#endif /* __KERNEL__ */
#endif /* _LINUX_MUTEX_H */
