/*
 *    Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 *    For licensing information please see the supplied COPYING file
 *
 */

#ifndef __PROCFS_LINUX_H
#define __PROCFS_LINUX_H

#ident "$Id: procfs_linux.h,v 1.13.2.2 2005/07/20 11:35:37 mike Exp $"
/*      $Source: /cvs/master/quadrics/qsnet/procfs_linux.h,v $ */

#if defined(__KERNEL__)

#include <qsnet/autoconf.h>
#include <qsnet/kernel_linux.h>
#include <linux/proc_fs.h>

extern gid_t qsnet_procfs_gid;

/* borrowed from fs/proc/proc_misc - helper for proc_read_int */
static inline int 
qsnet_proc_calc_metrics(char *page, char **start, off_t off, int count, int *eof, int len)
{
	if (len <= off+count) *eof = 1;
	*start = page + off;
	len -= off;
	if (len>count) len = count;
	if (len<0) len = 0;
	return len;
}

static inline int
qsnet_proc_write_int(struct file *file, const char *buf, unsigned long count, void *data)
{
	char tmpbuf[16];
	int  res = count;
	
	if (count > sizeof(tmpbuf) - 1)
		return (-EINVAL);
	
	MOD_INC_USE_COUNT;
	if (copy_from_user(tmpbuf, buf, count))
		res = -EFAULT;
	else
	{
		tmpbuf[count] = '\0'; 
		*(int *)data = simple_strtoul(tmpbuf, NULL, 0);
	}
	MOD_DEC_USE_COUNT;
	
	return (res);
}

static inline int
qsnet_proc_read_int(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len, res;
	
	MOD_INC_USE_COUNT;
	
	len = sprintf(page, "%d\n", *(int *)data);
	res = qsnet_proc_calc_metrics(page, start, off, count, eof, len);
	
	MOD_DEC_USE_COUNT;
	return (res);
}

static inline struct proc_dir_entry *
qsnet_proc_register_int(struct proc_dir_entry *dir, char *path, int *var, int read_only)
{
	struct proc_dir_entry *p;
	
	p = create_proc_entry(path, read_only ? S_IRUGO : S_IRUGO|S_IWUSR|S_IWGRP, dir);
	if (p) {
		if (! read_only) 
			p->write_proc = qsnet_proc_write_int;
		p->read_proc  = qsnet_proc_read_int;
		p->data       = var;
		p->owner      = THIS_MODULE;
		p->gid        = qsnet_procfs_gid;
	}
	return p;
}

static inline int
qsnet_proc_write_hex(struct file *file, const char *buf, unsigned long count, void *data)
{
	char tmpbuf[16];
	int  res = count;
	
	if (count > sizeof(tmpbuf) - 1)
		return (-EINVAL);
	
	MOD_INC_USE_COUNT;
	if (copy_from_user(tmpbuf, buf, count))
		res = -EFAULT;
	else
	{
		tmpbuf[count] = '\0'; 
		*(int *)data = simple_strtoul(tmpbuf, NULL, 0);
	}
	MOD_DEC_USE_COUNT;
	
	return (res);
}

static inline int
qsnet_proc_read_hex(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len, res;
	
	MOD_INC_USE_COUNT;
	
	len = sprintf(page, "0x%x\n", *(int *)data);
	res = qsnet_proc_calc_metrics(page, start, off, count, eof, len);
	
	MOD_DEC_USE_COUNT;
	return (res);
}

static inline struct proc_dir_entry *
qsnet_proc_register_hex(struct proc_dir_entry *dir, char *path, int *var, int read_only)
{
	struct proc_dir_entry *p;
	
	p = create_proc_entry(path, read_only ? S_IRUGO : S_IRUGO|S_IWUSR|S_IWGRP, dir);
	if (p) {
		if (! read_only) 
			p->write_proc = qsnet_proc_write_hex;
		p->read_proc  = qsnet_proc_read_hex;
		p->data       = var;
		p->owner      = THIS_MODULE;
		p->gid        = qsnet_procfs_gid;
	}
	return p;
}

#define QSNET_PROC_STR_LEN_MAX ((int)256)

static inline int
qsnet_proc_write_str(struct file *file, const char *buf, unsigned long count, void *data)
{
	int  res = count;
	
	if (count > (QSNET_PROC_STR_LEN_MAX - 1))
		return (-EINVAL);
	
	MOD_INC_USE_COUNT;
	if (copy_from_user((char *)data, buf, count))
		res = -EFAULT;
	else
	{
		((char *)data)[count] = '\0'; 
		/* remove linefeed */
		if ( (count) && (((char *)data)[count -1] == '\n'))
			((char *)data)[count -1] = '\0';
	}
	MOD_DEC_USE_COUNT;
	
	return (res);
}

static inline int
qsnet_proc_read_str(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len, res;
	
	if ( strlen(data) > (count + 1))
		return (-EINVAL);	

	MOD_INC_USE_COUNT;
	
	/* cant output too much */
	if ( strlen(data) > (count + 1))
	{
		MOD_DEC_USE_COUNT;
		return (-EINVAL);	
	}


	len = sprintf(page, "%s\n", (char *)data);
	if (len > count)
	{
		MOD_DEC_USE_COUNT;
		return (-EINVAL);	
	}

	res = qsnet_proc_calc_metrics(page, start, off, count, eof, len);
	
	MOD_DEC_USE_COUNT;
	return (res);
}

static inline struct proc_dir_entry *
qsnet_proc_register_str(struct proc_dir_entry *dir, char *path, char *var, int read_only)
{
	struct proc_dir_entry *p;
	
	p = create_proc_entry(path, read_only ? S_IRUGO : S_IRUGO|S_IWUSR|S_IWGRP, dir);
	if (p) {
		if (! read_only) 
			p->write_proc = qsnet_proc_write_str;
		p->read_proc  = qsnet_proc_read_str;
		p->data       = var;
		p->owner      = THIS_MODULE;
		p->gid        = qsnet_procfs_gid;
	}
	return p;
}

extern struct proc_dir_entry *qsnet_procfs_root; 
extern struct proc_dir_entry *qsnet_procfs_config;

/* code for procfs handling multipage requests */

void qsnet_proc_character_fill (long mode, char *fmt, ...);
int  qsnet_proc_release (struct inode *inode, struct file *file);

static inline ssize_t
qsnet_proc_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    QSNET_PROC_PRIVATE *pr  = (QSNET_PROC_PRIVATE *) file->private_data;
    int                 error;

    if (pr->pr_off >= pr->pr_len)
	return (0);

    if ((error = verify_area (VERIFY_WRITE, buf, count)) != 0)
	return (error);

    if (count >= (pr->pr_len - pr->pr_off))
	count = pr->pr_len - pr->pr_off;

    copy_to_user (buf, pr->pr_data + pr->pr_off, count);

    pr->pr_off += count;
    *ppos      += count;

    return (count);
}


#ifdef NO_PDE
static inline struct proc_dir_entry *PDE(const struct inode *inode)
{
    return inode->u.generic_ip;
}
#endif
#endif /* __KERNEL__ */

#define QSNET_PROCFS_IOCTL      "/proc/qsnet/ioctl"
#define QSNET_PROCFS_KMEM_DEBUG "/proc/qsnet/kmem_debug"
#define QSNET_PROCFS_VERSION    "/proc/qsnet/version"

#endif /* __PROCFS_LINUX_H */

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
