/*
 * Copyright (c) 1996-2002 by Quadrics Supercomputers World Ltd.
 *
 * For licensing information please see the supplied COPYING file
 *
 * rmscall.h:  user interface to rms kernel module
 *
 * $Id: rmscall.h,v 1.27.2.1 2005/07/28 10:49:09 robin Exp $
 * $Source: /cvs/master/quadrics/rmsmod/rmscall.h,v $
 *
 */

#ifndef RMSCALL_H_INCLUDED
#define RMSCALL_H_INCLUDED 1

#ident	"$Id: rmscall.h,v 1.27.2.1 2005/07/28 10:49:09 robin Exp $"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * flags for rms_fork_register
 *
 * RMS_IOF is not in a public header file 
 */
#define RMS_IOF 1		/* inherit on fork */

#ifndef __KERNEL__
#include <sys/types.h>
#endif

#include <qsnet/types.h>
#include <elan/capability.h>

#define MAXCOREPATHLEN 32

#if defined(SOLARIS)
typedef long long rmstime_t;
#else  /* DIGITAL_UNIX */
typedef long rmstime_t;
#endif

typedef enum {
    
    PRG_RUNNING  = 0x01,	/* program is running                  */
    PRG_ZOMBIE   = 0x02,	/* last process on a node has exited   */
    PRG_NODE     = 0x04,	/* stats are complete for this node    */
    PRG_KILLED   = 0x08,	/* program was killed                  */
    PRG_SUSPEND  = 0x10,        /* program is suspended                */
    PRG_ERROR    = 0x80	        /* error collecting stats              */

} PRGSTATUS_FLAGS;

/*
 * program time statistics extended in version 5 of the kernel module
 */
typedef struct {
    rmstime_t etime;		/* elapsed cpu time (milli-secs)       */
    rmstime_t atime;		/* allocated cpu time (cpu milli-secs) */
    rmstime_t utime;		/* user cpu time (cpu milli-secs)      */
    rmstime_t stime;		/* system cpu time (cpu milli-secs)    */
    int ncpus;			/* number of cpus allocated            */
    int flags;			/* program status flags                */
    int mem;			/* max memory size in MBytes           */
    int pageflts;		/* number of page faults               */
    rmstime_t memint;		/* memory integral                     */
} prgstats_old_t;

typedef struct {
    uint64_t etime;		/* elapsed cpu time (milli-secs)       */
    uint64_t atime;		/* allocated cpu time (cpu milli-secs) */
    uint64_t utime;		/* user cpu time (cpu milli-secs)      */
    uint64_t stime;		/* system cpu time (cpu milli-secs)    */
    uint64_t pageflts;		/* number of page faults               */
    uint64_t memint;		/* memory integral                     */
    uint64_t ebytes;		/* data transferred by the Elan(s)     */
    uint64_t exfers;		/* number of Elan data transfers       */
    uint64_t spare64[4];	/* expansion space                     */
    int ncpus;			/* number of cpus allocated            */
    int flags;			/* program status flags                */
    int mem;			/* max memory size in MBytes           */
    int spare32[5];             /* expansion space                     */
} prgstats_t;

int  rmsmod_init(void);
void rmsmod_fini(void);

int rms_setcorepath(caddr_t path);
int rms_getcorepath(pid_t pid, caddr_t path, int maxlen);
int rms_prgcreate(int id, uid_t uid, int cpus);
int rms_prgdestroy(int id);
int rms_prgids(int maxids, int *prgids, int *nprgs);
int rms_prginfo(int id, int maxpids, pid_t *pids, int *nprocs);
int rms_prgaddcap(int id, int index, ELAN_CAPABILITY *cap);

int rms_prgsuspend(int id);
int rms_prgresume(int id);
int rms_prgsignal(int id, int signo);

int rms_getprgid(pid_t pid, int *id);
int rms_ncaps(int *ncaps);
int rms_getcap(int index, ELAN_CAPABILITY *cap);
int rms_mycap(int *index);
int rms_setcap(int index, int ctx);
int rms_prefcap(int nprocess, int *index);

int   rms_prggetstats(int id, prgstats_t *stats);
void  rms_accumulatestats(prgstats_t *total, prgstats_t *stats);
char *rms_statsreport(prgstats_t *stats, char *buf);

int rms_elaninitdone(int vp);
int rms_prgelanpids(int id, int maxpids, int *vps, pid_t *pids, int *npids);
int rms_setelanstats(int id, uint64_t ebytes, uint64_t exfers);

int rms_setpset(int psid);
int rms_getpset(int id, int *psid);
int rms_modversion(void);

int rms_addproc(int id, pid_t pid);
int rms_removeproc(int id, pid_t pid);
int rms_ptrack_enabled(void);

#ifdef __cplusplus
}
#endif


#if defined(__KERNEL__)

int rms_init(void);
int rms_fini(void);
int rms_reconfigure(void);

extern int rms_debug;

#if 1
#define DBG(x) do if (rms_debug) x ; while (0)
#else
#define DBG(x)
#endif

#endif

#endif /* RMSCALL_H_INCLUDED */




